/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data;

import org.geotools.referencing.crs.DefaultGeographicCRS;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import schmitzm.geotools.grid.GridUtil;
import appl.data.LoadingException;

/**
 * Diese Klasse bildet eine Basis-Implementierung von {@link ReadableGrid}.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class AbstractReadableGrid implements ReadableGrid {
  /** Enthaelt das CRS des Rasters */
  protected CoordinateReferenceSystem crs = null;

  /**
   * Erzeugt ein AbstractReadableGrid.
   * @param crs CoordinateReferenceSystem
   */
  public AbstractReadableGrid(CoordinateReferenceSystem crs) {
    this.crs = (crs == null) ? DefaultGeographicCRS.WGS84 : crs;
  }

  /**
   * Creates an abstract ReadableGrid with a default CRS (WGS84)
   */
  public AbstractReadableGrid() {
    this(null);
  }

	/**
	 * This class does not support late loading itself!
	 *
	 * @see appl.data.LateLoadable#unloadData()
	 * @return false;
	 * @see appl.data.LateLoadable#isLateLoadable()
	 */
	public boolean isLateLoadable() {
		return false;
	}

	/**
	 * Does nothing!
	 *
	 * This class  does not support late loading!
	 *
	 * @see appl.data.LateLoadable#unloadData()
	 * @see appl.data.LateLoadable#loadData()
	 */
	public void loadData() throws LoadingException {
	}

	/**
	 * Does nothing!
	 *
	 * This class does not support late loading!
	 *
	 * @see appl.data.LateLoadable#unloadData()
	 */
	public void unloadData(){
	}

  /**
   * Liefert das CRS des Rasters.
   */
  public CoordinateReferenceSystem getCoordinateReferenceSystem() {
    return this.crs;
  }


   /**
   * Liefert die reale Breite einer Rasterzelle.
   * @return <code>getRealWidth() / getWidth()</code>
   */
  public double getCellWidth() {
    return getRealWidth() / getWidth();
  }

  /**
   * Liefert die reale Breite einer Rasterzelle.
   * @return <code>getRealHeight() / getHeight()</code>
   */
  public double getCellHeight() {
    return getRealHeight() / getHeight();
  }

  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   */
  public short getRasterSampleAsShort(int... cell) {
    return ((Number)getRasterSample(cell)).shortValue();
  }

  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   */
  public byte getRasterSampleAsByte(int... cell) {
    return ((Number)getRasterSample(cell)).byteValue();
  }


  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   */
  public int getRasterSampleAsInt(int... cell) {
    return ((Number)getRasterSample(cell)).intValue();
  }

  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   */
  public long getRasterSampleAsLong(int... cell) {
    return ((Number)getRasterSample(cell)).longValue();
  }


  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   */
  public float getRasterSampleAsFloat(int... cell) {
    return ((Number)getRasterSample(cell)).floatValue();
  }

  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   */
  public double getRasterSampleAsDouble(int... cell) {
    return ((Number)getRasterSample(cell)).doubleValue();
  }

  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * Liegt der Koordinatenwert genau auf der Grenze zwischen zwei Rasterzellen,
   * wird die naechst groessere Zelle gewaehlt (ausser die Grenze entspricht
   * dem Raster-Rand!).
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
   *            Koordinaten angegeben werden
   */
  public Object getGridSample(double... coord) {
    if ( coord.length < RASTER_DIM )
      throw new UnsupportedOperationException(RASTER_DIM+" Coordinates expected");
    // Koordinaten umrechnen in Zellen-Index
    int cellX = convertRealToRaster(coord[0],0);
    int cellY = convertRealToRaster(coord[1],1);
    return getRasterSample(cellX,cellY);
  }

  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   */
  public short getGridSampleAsShort(double... coord) {
    return ((Number)getGridSample(coord)).shortValue();
  }

  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   */
  public byte getGridSampleAsByte(double... coord){
    return ((Number)getGridSample(coord)).byteValue();
  }

  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   */
  public int getGridSampleAsInt(double... coord){
    return ((Number)getGridSample(coord)).intValue();
  }


  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   */
  public long getGridSampleAsLong(double... coord){
    return ((Number)getGridSample(coord)).longValue();
  }

  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   */
  public float getGridSampleAsFloat(double... coord){
    return ((Number)getGridSample(coord)).floatValue();
  }


  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   */
  public double getGridSampleAsDouble(double... coord){
    return ((Number)getGridSample(coord)).doubleValue();
  }

  /**
   * Konvertiert eine reale Koordinate in eine Zellennummer. Liegt der
   * Koordinatenwert genau auf der Grenze zwischen zwei Rasterzellen, wird
   * die naechst "groessere" Zellegewaehlt.
   * Ausnahme bildet der Rand des Rasters. Hier wird die kleinere Zelle
   * (also die letzte) herangezogen
   * @param coord Georeferenz-Koordinate
   * @param dim   Dimension, in der die Umrechnung erfolgen soll
   * @exception UnsupportedOperationException falls eine ungueltige Dimension
   *            angegeben wird (nur Werte 0 <= coord < {@link #RASTER_DIM} sind
   *            zulaessig
   */
  public int convertRealToRaster(final double coord, final int dim) {
    return convertRealToRaster(this,coord,dim);
  }

  /**
   * Konvertiert eine Zellennummer in reale Koordinate. Dabei wird die
   * Koordinate der Zellenmitte zurueckgegeben.
   * @param cell  Rasterzellen-Koordinate
   * @param dim   Dimension, in der die Umrechnung erfolgen soll
   * @exception UnsupportedOperationException falls eine ungueltige Dimension
   *            angegeben wird (nur Werte 0 <= coord < {@link #RASTER_DIM} sind
   *            zulaessig
   */
  public double convertRasterToReal(final int cell, final int dim) {
    return convertRasterToReal(this,cell,dim);
  }

  /**
   * Konvertiert eine reale Koordinate in eine Zellennummer. Liegt der
   * Koordinatenwert genau auf der Grenze zwischen zwei Rasterzellen, wird
   * die naechst "groessere" Zellegewaehlt.
   * Ausnahme bildet der Rand des Rasters. Hier wird die kleinere Zelle
   * (also die letzte) herangezogen.
   * @param grid  Raster fuer das die Umrechnung vorgenommen wird
   * @param coord Georeferenz-Koordinate
   * @param dim   Dimension, in der die Umrechnung erfolgen soll
   * @exception UnsupportedOperationException falls eine ungueltige Dimension
   *            angegeben wird (nur Werte 0 <= coord < {@link #RASTER_DIM} sind
   *            zulaessig
   */
  public static int convertRealToRaster(ReadableGrid grid, double coord, int dim) {
    if ( dim < 0 || dim >= RASTER_DIM )
      throw new UnsupportedOperationException(RASTER_DIM+" is the maximum dimension!");

    int    rasterMin    = (dim==0) ? grid.getMinX()      : grid.getMinY();
    double realMin      = (dim==0) ? grid.getX()         : grid.getY();
    int    rasterLength = (dim==0) ? grid.getWidth()     : grid.getHeight();
    double realLength   = (dim==0) ? grid.getRealWidth() : grid.getRealHeight();
    double cellSize     = (dim==0) ? grid.getCellWidth() : grid.getCellHeight();

    int cell = convertRealToRaster(coord, realMin, realLength, rasterMin, rasterLength);
    // In Y-Richtung liegt der Raster-Ursprung eines Grids OBEN, waehrend die
    // Latitude-Referenz die UNTERE Grid-Kante beschreibt
    // -> Umrechnen
    if ( dim == 1 )
      cell = rasterMin+rasterLength - cell - 1;
    return cell;
  }

  /**
   * Konvertiert eine Zellennummer in eine reale Koordinate. Dabei wird die
   * Koordinate der Zellenmitte zurueckgegeben.
   * @param grid  Raster fuer das die Umrechnung vorgenommen wird
   * @param cell  Rasterzellen-Koordinate
   * @param dim   Dimension, in der die Umrechnung erfolgen soll
   * @exception UnsupportedOperationException falls eine ungueltige Dimension
   *            angegeben wird (nur Werte 0 <= coord < {@link #RASTER_DIM} sind
   *            zulaessig
   */
  public static double convertRasterToReal(ReadableGrid grid, int cell, int dim) {
    if ( dim < 0 || dim >= RASTER_DIM )
      throw new UnsupportedOperationException(RASTER_DIM+" is the maximum dimension!");

    int    rasterMin    = (dim==0) ? grid.getMinX()      : grid.getMinY();
    double realMin      = (dim==0) ? grid.getX()         : grid.getY();
    int    rasterLength = (dim==0) ? grid.getWidth()     : grid.getHeight();
    double realLength   = (dim==0) ? grid.getRealWidth() : grid.getRealHeight();
    double cellSize     = (dim==0) ? grid.getCellWidth() : grid.getCellHeight();

    // In Y-Richtung liegt der Raster-Ursprung eines Grids OBEN, waehrend die
    // Longitude-Referenz die UNTERE Grid-Kante beschreibt
    // -> Y-Koordinate umrechnen in "Sicht aus UNTERER Ecke"
    if ( dim == 1 )
      cell = rasterMin+rasterLength - cell - 1;

    return convertRasterToReal(cell,rasterMin,realMin,cellSize);
  }

  /**
   * Konvertiert eine reale Koordinate in eine Zellennummer.
   * Siehe {@link GridUtil#convertRealToRaster(double, double, double, int, int)}.
   * @param coord Georeferenz-Koordinate, die umgerechnet werden soll
   * @param realMin    Minimale Welt-Koordinate (in der gewuenschten Richtung/Dimension) des
   *                   Grids in dem sich die Zelle befindet
   * @param realLength Reale Laenge (in der gewuenschten Richtung/Dimension) des
   *                   Grids in dem sich die Zelle befindet
   * @param rasterMin    Minimale Zellen-Nummer (in der gewuenschten Richtung/Dimension) des
   *                     Grids in dem sich die Zelle befindet
   * @param rasterLength Laenge in Zellen (in der gewuenschten Richtung/Dimension) des
   *                     Grids in dem sich die Zelle befindet
   */
  public static int convertRealToRaster(final double coord, final double realMin, final double realLength, final int rasterMin, final int rasterLength) {
    return GridUtil.convertRealToRaster(coord, realMin, realLength, rasterMin, rasterLength);
  }

  /**
   * Konvertiert eine Zellennummer in eine reale Koordinate.
   * Siehe {@link GridUtil#convertRasterToReal(int, int, double, double)}.
   * @param cell       Rasterzellen-Koordinate, die umgerechnet werden soll
   * @param rasterMin  Minimale Raster-Koordinate (in der gewuenschten Richtung/Dimension) des
   *                   Grids in dem sich die Zelle befindet
   * @param realMin    Minimale Welt-Koordinate (in der gewuenschten Richtung/Dimension) des
   *                   Grids in dem sich die Zelle befindet
   * @param cellSize   Zellengroesse (in der gewuenschten Richtung/Dimension) des
   *                   Grids in dem sich die Zelle befindet
   */
  public static double convertRasterToReal(final int cell, final int rasterMin, final double realMin, final double cellSize) {
    return GridUtil.convertRasterToReal(cell, rasterMin, realMin, cellSize);
  }
  
  /**
   * Vergleicht zwei Raster auf gleiche Struktur (Hoehe, Breite, Zell-Hoehe,
   * Zell-Breite, Sample-Type).
   * @param grid1 Raster 1
   * @param desc1 Beschreibung von Raster 1 (fuer Fehlermeldung)
   * @param grid2 Raster 2
   * @param desc2 Beschreibung von Raster 2 (fuer Fehlermeldung)
   * @param checkType bestimmt, ob auch der Sample-Type uebereinstimmen muss
   * @exception UnsupportedOperationException falls die Raster nicht gleich sind
   */
  public static void compareStructure(ReadableGrid grid1, String desc1, ReadableGrid grid2, String desc2, boolean checkType) {
    if ( grid1.getWidth() != grid2.getWidth() )
      throw new UnsupportedOperationException(createCompareMessage(DataUtil.RESOURCE.getString("RasterDim.GridWidth"),desc1,desc2));
    if ( grid1.getHeight() != grid2.getHeight() )
      throw new UnsupportedOperationException(createCompareMessage(DataUtil.RESOURCE.getString("RasterDim.GridHeight"),desc1,desc2));
    if ( grid1.getCellWidth() != grid2.getCellWidth() )
      throw new UnsupportedOperationException(createCompareMessage(DataUtil.RESOURCE.getString("RasterDim.CellWidth"),desc1,desc2));
    if ( grid1.getCellHeight() != grid2.getCellHeight() )
      throw new UnsupportedOperationException(createCompareMessage(DataUtil.RESOURCE.getString("RasterDim.CellHeight"),desc1,desc2));
    if ( checkType && grid1.getSampleType() != grid2.getSampleType() )
      throw new UnsupportedOperationException(createCompareMessage(DataUtil.RESOURCE.getString("RasterDim.SampleType"),desc1,desc2));
  }

  /**
   * Erzeugt eine Fehlermeldung fuer eine bestimmte inkompatible Raster-Struktur.
   */
  private static String createCompareMessage(String compType, String objDesc1, String objDesc2) {
    return DataUtil.RESOURCE.getString("RasterDim.ErrorMess",compType,objDesc1,objDesc2);
  }
}
