/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data;

import java.util.Enumeration;

import schmitzm.data.property.ListProperty;
import schmitzm.data.property.Properties;
import schmitzm.data.property.Property;
import schmitzm.data.property.PropertySet;
import schmitzm.data.property.ScalarProperty;

/**
 * Diese Klasse stellt die Struktur (nicht den Inhalt) eines "komplexen" Objekts
 * (Datentyps) dar. Ein "komplexes" Objekt kann z.B. eine {@linkplain ListProperty Liste}
 * sein, oder eine {@linkplain Properties Menge von Eigenschaften}.
 * Auch eine {@linkplain ScalarProperty skalare Eigenschaft} kann
 * bereits als komplexes Objekt aufgefasst werden, da sie einen Namen und ein
 * Unterobjekt enthaelt (welches wiederum komplex sein kann!).
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public interface ObjectStructure {
  /**
   * Liefert die Anzahl an Attributen, die der Datentyp hat. Bei einem
   * {@link Properties}-Objekt entspricht dies z.B. der Anzahl an Propertys.
   * Bei einer {@linkplain ListProperty Liste} von einfachen Elementen ist
   * die Anzahl 1.
   */
  public int getAttrCount();

  /**
   * Liefert die Typen aller Attribute des Objekttyps. Dies kann wiederum
   * eine <code>ObjectStructure</code> sein, oder eine <code>Class</code>-Instanz,
   * falls das Attribut nicht weiter strukturiert werden kann.
   * @return Aufzaehlung, in der jedes Element vom Typ <code>Class</code> oder
   *         <code>ObjectStructure</code> ist
   */
  public Enumeration getAttrTypes();

  /**
   * Prueft, ob der Objekttyp einen Namen besitzt. z.B. ist eine {@link Property}
   * i.d.R. benannt, ein {@link PropertySet} jedoch nicht.
   */
  public boolean isStructureNamed();

  /**
   * Liefert den Namen der <code>ObjectStructure</code>, falls sie benannt ist.
   * @return <code>null</code> falls die Struktur unbenannt ist
   * @see #isStructureNamed()
   */
  public String getStructureName();

  /**
   * Prueft, ob die Objektstruktur mehrere Instanzen aufnehmen kann. Dies ist
   * z.B. fuer {@linkplain ListProperty Listen} der Fall, nicht jedoch fuer
   * ein {@link PropertySet}, da dieses zwar verschiedene Attributwerte aufnimmt,
   * fuer jedes Attribut jedoch immer nur eine Auspraegung!
   */
  public boolean containsMultipleValues();

  /**
   * Prueft, ob die Objekt-Struktur mit einer anderen identisch ist.
   * @param object andere <code>ObjectStructure</code>
   */
  public boolean equalsInStructure(Object object);
}
