/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data;

import java.util.Enumeration;

/**
 * Diese Klasse stellt statische Methoden fuer die Arbeit mit
 * {@linkplain ObjectStructure Objekt-Strukturen} zur Verfuegung.
 * Hierzu gehoeren insbesondere Vergleichsmethoden, mit denen auf strukturelle
 * Gleichkeit oder Implikation getestet werden kann.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ObjectStructureUtil {
  /** Rueckgabe-Konstante fuer "zwei Strukturen sind identisch". */
  public static final int EQUAL = 1;
  /** Rueckgabe-Konstante fuer "die linke Strukturen umfasst die rechte". */
  public static final int LEFT_CONTAINS_RIGHT = 2;
  /** Rueckgabe-Konstante fuer "die rechte Strukturen umfasst die linke". */
  public static final int RIGHT_CONTAINS_LEFT = 3;
  /** Rueckgabe-Konstante fuer "weder umfasst die linke Strukur die rechte,
   *  noch umgekehrt" (Ueberschneidung ist dennoch moeglich!). */
  public static final int UNEQUAL = 4;
  /** Rueckgabe-Konstante fuer "die beiden Strukituren haben eine leere Schnittmenge".<br>
   *  <b>Bemerke</b>: wird z.Z. noch nicht verwendet!! */
  public static final int DISJOINT = 5;

  /**
   * Vergleicht zwei Objekt-Strukturen.
   * @param o1 "linke" Objekt-Struktur
   * @param o2 "rechte" Objekt-Struktur
   * @return eine der Konstanten {@link #EQUAL}, {@link #LEFT_CONTAINS_RIGHT},
   *         {@link #RIGHT_CONTAINS_LEFT}, {@link #UNEQUAL}
   * @see #checkStructureContainsStructure(ObjectStructure,ObjectStructure)
   */
  public static int compareObjectStructures(ObjectStructure o1, ObjectStructure o2) {
    boolean o1_cont_o2 = checkStructureContainsStructure(o1,o2);
    boolean o2_cont_o1 = checkStructureContainsStructure(o2,o1);
    if ( o1_cont_o2 && o2_cont_o1 ) return EQUAL;
    if ( o1_cont_o2  ) return LEFT_CONTAINS_RIGHT;
    if ( o2_cont_o1  ) return RIGHT_CONTAINS_LEFT;
    // Teil-Ueberschneidung wird (noch) nicht geprueft
    return UNEQUAL;
  }

  /**
   * Prueft, ob eine Objekt-Struktur eine andere umfasst.<br>
   * <u>Vorgehensweise</u>:<br>
   * <ol>
   * <li>Kann die eine Struktur mehrere Werte aufnehmen und die andere nicht
   *     ({@link ObjectStructure#containsMultipleValues()}), kann die linke
   *     Struktur die rechte auf keinen Fall enthalten.</li>
   * <li>Enhaelt die rechte Struktur mehr Attribute, als die linke, so kann die
   *     linke Struktur die rechte auf keinen Fall enthalten.</li>
   * <li>Enthaelt die linke Struktur ein Attribut, welches (wiederum strukturell)
   *     nicht in der rechten Struktur vorhanden ist, so kann die linke Struktur
   *     die rechte nicht enthalten.<br>
   *     Beim Vergleich der Attribute wird wiederum auf strukturelle Implikation
   *     geprueft:
   *     <ul>
   *     <li>Handelt es sich um zwei strukturierte Attribute, muss das "linke"
   *         Attribut das "rechte" strukturell umfassen (Rekursion!).</li>
   *     <li>Handelt es sich um zwei statische Attribute (Typ <code>Class</code>),
   *         muss die "linke" Klasse eine Spezialisierung (Ableitung) der
   *         "rechten" Klasse sein.</li>
   *     <li>Treffen beide Bedingungen nicht zu, kann das linke Attribut das
   *         rechte nicht umfassen.<li>
   *     </ul></li>
   * </ol>
   * @param o1 "linke" Objekt-Struktur
   * @param o2 "rechte" Objekt-Struktur
   */
  public static boolean checkStructureContainsStructure(ObjectStructure o1, ObjectStructure o2) {
    if ( o1.containsMultipleValues() != o2.containsMultipleValues() ||
         o2.getAttrCount() > o1.getAttrCount() )
      return false;

    // Pruefen, ob fuer jedes Attribut in o2 ein Attribut in o1 vorhanden ist,
    //  welches o2 beinhaltet
    for (Enumeration attr2 = o2.getAttrTypes(); attr2.hasMoreElements();)
      if ( !findAttrTypeInStrcucture(attr2.nextElement(),o1) )
        return false;

    return true;
  }

  /**
   * Prueft, ob ein Attribut in einer Objekt-Struktur (strukturell) enthalten ist.
   * Erlaeuterungen ueber die Vorgehensweise sind in der Dokumenation zu
   * {@link #checkStructureContainsStructure(ObjectStructure,ObjectStructure)}
   * zu finden.
   * @param type1 Attribut (<code>ObjectStructure</code>- oder
   *              <code>Class</code>-Instanz)
   * @param o2    Objekt-Struktur
   */
  private static boolean findAttrTypeInStrcucture(Object type1, ObjectStructure o2) {
    for (Enumeration attr2 = o2.getAttrTypes(); attr2.hasMoreElements();) {
      Object type2 = attr2.hasMoreElements();

      // wenn es sich bei beiden um Strukturen handelt und Typ2 den gesuchten
      // Typ1 impliziert, wurde Typ1 in der Struktur o2 gefunden
      if (type1 instanceof ObjectStructure &&
          type2 instanceof ObjectStructure &&
          checkStructureContainsStructure((ObjectStructure)type2,(ObjectStructure)type1) )
        return true;

      // wenn es sich bei beiden um einfache Klassen handelt und Typ2 den gesuchten
      // Typ1 impliziert, wurde Typ1 in der Struktur o2 gefunden
      else if (type1 instanceof Class &&
               type2 instanceof Class &&
               ((Class)type2).isAssignableFrom((Class)type1) )
             return true;
    }

    return false;
  }

}
