/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data;


/**
 * Diese Klasse implementiert Rechen-Operationen auf Rastern.
 * @see RasterOperationTree
 * @see RasterOperationTreeParser
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class RasterCalculator {
  /** Instanz des Parsers. */
  public static final RasterOperationTreeParser RASTER_CALC_PARSER = new RasterOperationTreeParser();

  /////////////////////////////////////////////////////////////////////
  /////////////////////   FORMEL AUSFUEHREN   /////////////////////////
  /////////////////////////////////////////////////////////////////////
  /**
   * Berechnet eine Formel auf jeder Zelle der Eingabe-Raster und speichert das
   * Ergebnis in der jeweiligen Zelle des Ausgabe-Rasters. Hierfuer wird der
   * Formel-String durch den {@link RasterOperationTreeParser} in einen
   * {@link RasterOperationTree} umgewandelt, der auf dann fuer jede Rasterzelle
   * des Output-Rasters ausgewertet wird.
   * @param rule       Formel
   * @param inRaster   Eingabe-Raster
   * @param inFilter   Eingabe-Filter
   * @param outRaster  Ausgabe-Raster
   * @see RasterOperationTreeParser
   */
  public static void calculate(String rule, ReadableGrid[] inRaster, RasterFilter[] inFilter, WritableGrid outRaster ) {
    RasterOperationTree opTree = RASTER_CALC_PARSER.parse(rule);
    if ( opTree == null )
      return;

    int minX = outRaster.getMinX();
    int minY = outRaster.getMinY();
    int maxX = minX + outRaster.getWidth() - 1;
    int maxY = minY + outRaster.getHeight() - 1;

    for (int y=minY; y<=maxY; y++)
      for (int x=minX; x<=maxX; x++) {
        // Berechnung starten mit dem ersten Operand
        double result = opTree.evaluate( x, y, inRaster, inFilter );
        outRaster.setRasterSample( (float)result, x, y );
      }
  }

  /**
   * Berechnet eine Formel auf jeder Zelle der Eingabe-Raster und speichert das
   * Ergebnis in der jeweiligen Zelle des Ausgabe-Rasters. Hierfuer wird der
   * Formel-String durch den {@link RasterOperationTreeParser} in einen
   * {@link RasterOperationTree} umgewandelt, der auf dann fuer jede Rasterzelle
   * des Output-Rasters ausgewertet wird.
   * @param rule      Formel
   * @param inRaster  Eingabe-Raster
   * @param outRaster Ausgabe-Raster
   * @see RasterOperationTreeParser
   */
  public static void calculate(String rule, ReadableGrid[] inRaster, WritableGrid outRaster ) {
    calculate( rule, inRaster, null, outRaster);
  }
  /**
   * Prueft die Syntax einer Raster-Formel auf Korrektheit.
   * @param rule Raster-Formel.
   */
  public static boolean checkRule(String rule) {
    try {
      checkRuleAndError(rule);
    } catch( Exception err ) {
      return false;
    }
    return true;
  }

  /**
   * Prueft die Syntax einer Raster-Formel auf Korrektheit.
   * @param rule Raster-Formel.
   * @exception IllegalArgumentException bei einem Fehler
   */
  public static void checkRuleAndError(String rule) {
    RASTER_CALC_PARSER.parse(rule);
  }
}
