/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data;

/**
 * Diese Klasse stellt eine (2-dimensionale) Filter-Matrix dar. Diese muss
 * in beiden Dimensionen eine ungerade Laenge haben, damit die Filter-Matrix
 * ein eindeutiges Zentrum besitzt.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class RasterFilter {
  /** Speichert die Filter-Matrix. */
  protected double[][] filter = null;
  /** Speichert die horizontale Laenge der Filter-Matrix. */
  protected int sizeX;
  /** Speichert die vertikale Laenge der Filter-Matrix. */
  protected int sizeY;
  /** Speichert die Summe ueber alle Eintraege der Filter-Matrix. */
  protected double weight;

  /**
   * Erzeugt einen neuen Raster-Filter. Die Filter-Matrix muss ein eindeutiges
   * Zentrum haben, d.h. die Array-Dimensionen muessen ungerade Laenge haben!
   * @param filter Filter-Matrix
   */
  public RasterFilter(double[][] filter) {
    if ( filter == null )
      throw new UnsupportedOperationException("null-Filter not allowed!");
    if ( filter.length % 2 != 1 || filter[0].length % 2 != 1 )
      throw new UnsupportedOperationException("Filter size must be odd in each dimension!");
    for (double[] filterCol : filter)
      if ( filterCol.length != filter[0].length )
        throw new UnsupportedOperationException("Filter must have the same size in each column!");

    this.filter = filter;
    this.sizeX  = filter.length;
    this.sizeY  = filter[0].length;
    this.weight = calcFilterWeight();
  }

  /**
   * Erzeugt einen leeren Raster-Filter. {@code sizeX} und {@code sizeY}
   * muessen ungerade sein, damit die Filter-Matrix ein eindeutiges Zentrum hat.
   * @param sizeX horizontale Laenge der Filter-Matrix
   * @param sizeY vertikale Laenge der Filter-Matrix
   */
  public RasterFilter(int sizeX, int sizeY) {
    this( new double[sizeX][sizeY] );
  }

  /**
   * Erzeugt einen leeren quadratischen Raster-Filter. {@code size} muss
   * ungerade sein, damit die Filter-Matrix ein eindeutiges Zentrum hat.
   * @param size horizontale/vertikale Laenge der Filter-Matrix
   */
  public RasterFilter(int size) {
    this( size, size );
  }

  /**
   * Liefert die horizontale Laenge der Filter-Matrix.
   */
  public int getFilterSizeX() {
    return sizeX;
  }

  /**
   * Liefert die vertikale Laenge der Filter-Matrix.
   */
  public int getFilterSizeY() {
    return sizeY;
  }

  /**
   * Liefert die Summe ueber alle Eintraege der Filter-Matrix.
   */
  public double getFilterWeight() {
    return weight;
  }

  /**
   * Berechnet die Summe ueber alle Eintraege der Filter-Matrix.
   */
  public double calcFilterWeight() {
    double sum = 0;
    for (double[] filterCol : filter)
      for (double v : filterCol)
        sum += v;
    return sum;
  }

  /**
   * Liefert einen Wert der Filter-Matrix.
   * @param x X-Koorindinate
   * @param y Y-Koordinate
   */
  public double getFilterValue(int x, int y) {
    return filter[x][y];
  }

  /**
   * Setzt einen Wert der Filter-Matrix und berechnet das
   * {@linkplain #getFilterWeight() Filter-Gewicht} neu.
   * @param value neuer Wert
   * @param x X-Koorindinate
   * @param y Y-Koordinate
   */
  public void setFilterValue(double value, int x, int y) {
    weight -= filter[x][y]; // Gesamtsumme um alten Wert reduzieren
    filter[x][y] = value;
    weight += value; // Gesamtsumme um neuen Wert erhoehen
  }

  /**
   * Wertet den Filter auf einer Rasterzelle aus, in dem die Summe der mit
   * den Filterwerten gewichteten Rasterwerte (um die Rasterzelle) gebildet
   * wird und diese mit der Summe der Filterwerte normiert wird (geteilt wird).
   * Am Rand des Rasters wird so verfahren, dass die Zellen ausserhalb des
   * Rasters ignoriert werden, sowohl in der gewichteten Summe, als auch in
   * der Filter-Summe.
   * @param raster Raster
   * @param x      X-Koordinate im Raster (Spaltenindex)
   * @param y      Y-Koordinate im Raster (Zeilenindex)
   * @return {@code Double.NaN} falls die Rasterzelle diesen Wert besitzt
   * @exception IllegalArgumentException falls ein {@code null}-Raster uebergeben wird
   *            oder die Koordinate ausserhalb des Rasters liegt
   */
  public double evaluate(ReadableGrid raster, int x, int y) {
    if ( raster == null )
      throw new IllegalArgumentException("Filter can not be evaluated on null-Raster!");

    // Aus Performanzgruenden werden moeglichst alle Berechnungen nur
    // einmal durchgefuehrt!
    final int minX = raster.getMinX();
    final int maxX = raster.getMinX() + raster.getWidth() - 1;
    final int minY = raster.getMinY();
    final int maxY = raster.getMinY() + raster.getHeight() - 1;
    if ( x < minX || x > maxX || y < minY || y > maxY )
      throw new IllegalArgumentException("Cell ("+x+"/"+y+") out of raster!");
    if ( Double.isNaN( raster.getRasterSampleAsDouble(x,y) ) )
      return Double.NaN;

    // Erste relevate Raster-Koordinate
    final int startX = x - sizeX/2;
    final int startY = y - sizeY/2;

    // Filter ueber Raster legen und Summen berechnen
    double filterSum   = 0; // Summe der tatsaechlich verwendeten Filterwerte (RAND!!)
    double cellSum     = 0; // Mit Filter gewichtete Summe der Rasterzellen
    int    cellX       = 0; // betrachtete X-Koordinate im Raster
    int    cellY       = 0; // betrachtete Y-Koordinate im Raster
    double cellValue   = 0; // Wert der betrachteten Zelle
    double filterValue = 0; // Wert der betrachteten Filter-Zelle
    for (int fy = 0; fy < sizeY; fy++) {
      cellY = startY + fy; // Raster-Koordinate
      if ( cellY >= minY && cellY <= maxY )
        for (int fx = 0; fx < sizeX; fx++) {
          cellX       = startX + fx; // Raster-Koordinate
          cellValue   = raster.getRasterSampleAsDouble(cellX,cellY);
          filterValue = filter[fx][fy];
          if ( cellX >= minX && cellX <= maxX && filterValue != 0 && !Double.isNaN(cellValue) ) {
            filterSum += filterValue;
            cellSum   += filterValue * cellValue;
          }
        }
    }
    // "Gefilterter" Zellenwert
    return (filterSum == 0) ? 0.0 : cellSum / filterSum;
  }
}
