/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data;

import schmitzm.lang.tree.BinaryTreeNode;
import schmitzm.lang.tree.OperationTree;
import schmitzm.lang.tree.TreeNode;

/**
 * Diese Klasse stellt einen Operator-Baum dar, in dem neben den von der
 * Oberklasse definitierten Operationen, Referenzen auf Raster und Filter
 * enthalten sein koennen.<br>
 * Der Operator-Baum wird auf einer bestimmten Raster-Koordinate ausgewertet.<br>
 * <br>
 * <b>Raster-Referenz:</b> {@link RasterReferenceNode}<br>
 * Eine Raster-Referenz wird durch einen {@code int}-Wert dargestellt, der von
 * einem {@code #} eingeleitet wird (z.B. {@code #0}). Bei der Auswertung des Operatorbaums
 * (auf einer bestimmten Raster-Koordinate) wird diese Referenz durch den
 * entsprechenden Zell-Wert des referenzierten Rasters ersetzt.<br>
 * <br>
 * <b>Filter-Referenz:</b> {@link FilterReferenceNode}<br>
 * Eine Filter-Referenz wird durch einen {@code int}-Wert dargestellt, der von
 * einem {@code F} eingeleitet wird (z.B. {@code F1(.)}) Die Auswertung eines
 * Filters kann nur direkt auf einer Raster-Referenz erfolgen, da ansonsten
 * keine zellen-weise Abarbeitung des Rasters mehr moeglich ist, sondern
 * komplette Raster als Zwischenergebnisse ausgewertet werden muessten.<br>
 * <b>Koordinaten-Referenz:</b> {@code X} oder {@code Y}<br>
 * Stellt einen Alias fuer die X- bzw. Y-Koordinate der Rasterzelle dar, die
 * verarbeitet wird.<br>
 * <b>NoData:</b> Konstante {@code NoData} und Funktion {@code isNoData(.)}<br>
 * Stellen einen Alias fuer {@code NaN}, bzw. {@code isNaN(.)} dar.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class RasterOperationTree extends OperationTree {
  private int                x           = 0;
  private int                y           = 0;
  private ReadableGrid[]     inputRaster = null;
  private RasterFilter[]     inputFilter = null;

  /**
   * Erzeugt einen neuen Raster-Operatorbaum
   * @param root Wurzelknoten
   */
  public RasterOperationTree(TreeNode root) {
    super(root);
  }

  /**
   * Nicht unterstuetzt!
   * @deprecated
   * @exception UnsupportedOperationException bei jedem Aufruf
   */
  public Object evaluate() {
    throw new UnsupportedOperationException("RasterOperationTree only supports evaluate(ReadableGrid[],int,int)");
  }

  /**
   * Wertet den Operatorbaum auf einer bestimmten Raster-Koordinate aus.
   * @param x X-Koordinate
   * @param y Y-Koordinate
   * @param inputRaster Eingabe-Raster auf die die #-Referenzen gemappt werden
   * @param inputFilter Eingabe-Filter auf die die F-Referenzen gemappt werden
   */
  public Double evaluate(int x, int y, ReadableGrid[] inputRaster, RasterFilter[] inputFilter) {
    this.x = x;
    this.y = y;
    this.inputRaster = inputRaster;
    this.inputFilter = inputFilter;
    Object result = super.evaluate();
    if ( result instanceof Boolean )
      result = (Boolean)result ? 1 : 0;
    if ( !(result instanceof Number) )
      throw new UnsupportedOperationException( DataUtil.RESOURCE.getString("RasterOperationTree.err.NonNumericResult", result) );
    return ((Number)result).doubleValue();
  }

  /**
   * Wertet den Operatorbaum auf einer bestimmten Raster-Koordinate aus.
   * @param x X-Koordinate
   * @param y Y-Koordinate
   * @param inputRaster Eingabe-Raster auf die die Referenzen gemappt werden
   */
  public Double evaluate(int x, int y, ReadableGrid[] inputRaster) {
    return evaluate(x,y,inputRaster);
  }

  /**
   * Wertet einen Knoten des Operator-Baums aus.
   * @param opTreeNode BinaryTreeNode
   * @return double
   */
  protected Object evaluate(TreeNode opTreeNode) {
    // Referenznummer auf Raster
    if ( opTreeNode instanceof RasterReferenceNode )
      return getInputRaster( ((RasterReferenceNode)opTreeNode).getObject() ).getRasterSampleAsDouble(x, y);

    // Referenznummer auf Filter
    if ( opTreeNode instanceof FilterReferenceNode ) {
      FilterReferenceNode filterRefereceNode = (FilterReferenceNode)opTreeNode;
      RasterFilter   filter      = getInputFilter( filterRefereceNode.getFilterNumber() );
      BinaryTreeNode operandNode = filterRefereceNode.getLeftChild();
      if ( !(operandNode instanceof RasterOperationTree.RasterReferenceNode) )
        throw new UnsupportedOperationException( DataUtil.RESOURCE.getString("RasterOperationTree.err.FilterReferenceNotAllowed") );
      ReadableGrid   raster      = getInputRaster( ((RasterReferenceNode)operandNode).getObject() );
      return filter.evaluate(raster,x, y);
    }
    return super.evaluate(opTreeNode);
  }

  /**
   * Wertet einen 1-stelligen Operator aus.
   * Erweitert die Methode der Oberklasse, um die folgenden Operatoren:
   * <ul>
   *   <li>{@code isNoData(.)}: Alias fuer die Funktion "isNaN(.)"</li>
   * </ul>
   * @param operator 1-stelliger Operator
   * @param operand  Operand auf den der Operator angewendet wird
   * @return <i>operator</li>( {@code operand} )
   */
  protected Double performOperation(String operator, Object operand) {
    operator = operator.toUpperCase();
    if ( operator.equals("ISNODATA") )
      operator = "ISNAN";

    Object result = super.performOperation(operator,operand);
    return result instanceof Number ? ((Number)result).doubleValue() : 0.0;
  }

  /**
   * Wertet einen 0-stelligen Operator (Alias oder Variable) aus.
   * Erweitert die Methode der Oberklasse, um die folgenden Operatoren:
   * <ul>
   *   <li>{@code X}: Alias fuer die horizontale Raster-Koordinate</li>
   *   <li>{@code Y}: Alias fuer die vertikale Raster-Koordinate</li>
   *   <li>{@code NoData}: Alias fuer die Konstante "NaN"</li>
   * </ul>
   * @param operator 0-stelliger Operator
   */
  protected Double performOperation(String operator) {
    operator = operator.toUpperCase();

    if ( operator.equalsIgnoreCase("X") )
      return Double.valueOf(x);
    if ( operator.equalsIgnoreCase("Y") )
      return Double.valueOf(y);
    if ( operator.equals("NODATA") )
      operator = "NAN";

    Object result = super.performOperation(operator);
    return result instanceof Number ? ((Number)result).doubleValue() : 0.0;
  }

  /**
   * Liefert eines der zur Verfuegung stehenden Eingabe-Raster.
   * @param rasterNo Raster-Nummer
   * @exception IllegalArgumentException falls eine ungueltige Raster-Nummer
   *            uebergeben wird
   */
  private ReadableGrid getInputRaster(int rasterNo) {
    if ( rasterNo >= inputRaster.length ) {
      String allowedStr = ( inputRaster.length > 0 ) ? DataUtil.RESOURCE.getString( "RasterOperationTree.err.AllowedIdx", inputRaster.length-1 )
                                                     : DataUtil.RESOURCE.getString( "RasterOperationTree.err.NoneAllowed");
      throw new IllegalArgumentException( DataUtil.RESOURCE.getString("RasterOperationTree.err.InvalidRasterReference",rasterNo,allowedStr) ); 
    }
    return inputRaster[rasterNo];
  }

  /**
   * Liefert einen der zur Verfuegung stehenden Eingabe-Filter.
   * @param rasterNo Filter-Nummer
   * @exception IllegalArgumentException falls eine ungueltige Filter-Nummer
   *            uebergeben wird
   */
  private RasterFilter getInputFilter(int filterNo) {
    if ( filterNo >= inputFilter.length ) {
      String allowedStr = ( inputFilter.length > 0 ) ? DataUtil.RESOURCE.getString( "RasterOperationTree.err.AllowedIdx", inputFilter.length-1 )
                                                     : DataUtil.RESOURCE.getString( "RasterOperationTree.err.NoneAllowed");
      throw new IllegalArgumentException( DataUtil.RESOURCE.getString("RasterOperationTree.err.InvalidFilterReference",inputFilter,allowedStr) ); 
    }
    return inputFilter[filterNo];
  }


  /**
   * Diese Knoten repraesentiert eine Referenz auf ein Raster im Operatorbaum.
   * Da es sich bei einem Zellwert um eine Konstante handelt, hat der Knoten
   * keine Kind-Knoten. Die Referenz wird durch einen {@code int} dargestellt.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class RasterReferenceNode extends BinaryTreeNode<Integer> {
    /**
     * Erzeugt einen neuen Referenz-Knoten
     * @param rasterNo Referenz-Nummer
     * @param parent   Vater-Knoten
     */
    public RasterReferenceNode(int rasterNo, BinaryTreeNode parent) {
      super(rasterNo, parent);
    }

    /**
     * Erzeugt einen neuen Referenz-Knoten
     * @param rasterNo Referenz-Nummer
     */
    public RasterReferenceNode(int rasterNo) {
      this(rasterNo, null);
    }

    /**
     * Macht nichts, da {@code RasterReferenceNode} immer einen Blatt-Knoten
     * darstellt.
     * @param i Index (beginnend bei 0)
     * @param child neuer Kind-Knoten
     */
    public void setChild(int i, TreeNode<Integer> child) {
    }
  }

  /**
   * Diese Knoten repraesentiert eine Referenz auf einen Filter im Operatorbaum.
   * Als Operator-Kennzeichen wird "F" verwendet. Die Referenz auf den Filter
   * wird in einem zusaetzlichen {@code int}-Wert gespeichert.
   * @see #getFilterNumber()
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class FilterReferenceNode extends OperationTree.UnaryOperatorNode {
    /** Speichert die Referenznummer auf den Filter. */
    protected int filterNo = 0;

    /**
     * Erzeugt einen neuen Referenz-Knoten
     * @param filterNo Referenz-Nummer
     * @param parent   Vater-Knoten
     * @param operand  Knoten der den (linker) Operand repraesentiert
     */
    public FilterReferenceNode(int filterNo, BinaryTreeNode parent, BinaryTreeNode operand) {
      super("F", parent, operand);
      setFilterNumber( filterNo );
    }

    /**
     * Erzeugt einen neuen Referenz-Knoten
     * @param filterNo Referenz-Nummer
     * @param operand  Knoten der den (linker) Operand repraesentiert
     */
    public FilterReferenceNode(int filterNo, BinaryTreeNode operand) {
      this(filterNo, null, operand);
    }

    /**
     * Setzt die Referenznummer auf einen Filter.
     * @param filterNo Referenznummer auf einen Filter
     */
    public void setFilterNumber(int filterNo) {
      this.filterNo = filterNo;
    }

    /**
     * Liefert die Referenznummer auf einen Filter.
     */
    public int getFilterNumber() {
      return filterNo;
    }
  }
}
