/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data;

import schmitzm.data.RasterOperationTree.FilterReferenceNode;
import schmitzm.data.RasterOperationTree.RasterReferenceNode;
import schmitzm.lang.tree.BinaryTreeNode;
import schmitzm.lang.tree.OperationTree;
import schmitzm.lang.tree.OperationTreeParser;

/**
 * Diese Klasse stellt einen Parser fuer einen {@linkplain RasterOperationTree Raster-Operatorbaum}
 * dar. Dieser erstellt einen {@linkplain RasterOperationTree Raster-Operatorbaum} aus einem
 * Formel-String, der neben den in {@link OperationTreeParser} beschriebenen
 * Komponenten Raster- und Filter-Referenzen enthalten darf.<br>
 * <br>
 * <b>Raster-Referenz:</b> {@link RasterReferenceNode}<br>
 * Eine Raster-Referenz wird durch einen {@code int}-Wert dargestellt, der von
 * einem {@code #} eingeleitet wird (z.B. {@code #0}). Bei der Auswertung des Operatorbaums
 * (auf einer bestimmten Raster-Koordinate) wird diese Referenz durch den
 * entsprechenden Zell-Wert des referenzierten Rasters ersetzt.<br>
 * <br>
 * <b>Filter-Referenz:</b> {@link FilterReferenceNode}<br>
 * Eine Filter-Referenz wird durch einen {@code int}-Wert dargestellt, der von
 * einem {@code F} eingeleitet wird (z.B. {@code F1(.)}) Die Auswertung eines Filters kann nur
 * direkt auf einer Raster-Referenz erfolgen, da ansonsten keine zellen-weise
 * Abarbeitung des Rasters mehr moeglich ist, sondern komplette Raster als
 * Zwischenergebnisse ausgewertet werden muessten.<br>
 * <b>Koordinaten-Referenz:</b> {@code X} oder {@code Y}<br>
 * Stellt einen Alias fuer die X- bzw. Y-Koordinate der Rasterzelle dar, die
 * verarbeitet wird.<br>
 * <b>NoData:</b> Konstante {@code NoData} und Funktion {@code isNoData(.)}<br>
 * Stellen einen Alias fuer {@code NaN}, bzw. {@code isNaN(.)} dar.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class RasterOperationTreeParser extends OperationTreeParser {
  /**
   * Erstellt einen Raster-Operator-Baum aus einem Formel-String.
   * @param rule Formel
   */
  public RasterOperationTree parse(String rule) {
    return new RasterOperationTree( super.parse(rule).getRoot() );
  }

//  /**
//   * Liefert einen LeerString, da der Strings im RasterCalculator keinen Sinn
//   * machen.
//   */
//  @Override
//  public String getStringEncapsulationChars() {
//    return "";
//  }

  /**
   * Parst das naechste Literal aus dem Tokenizer {@link #tok}.
   * Erweitert die Funktionalitaet der Oberklasse, so dass neben Konstanten
   * auch Referenznummern auf ein Raster (eingeleitet durch {@code #}) und
   * Referenznummern auf einen Filter (eingeleitet durch {@code F})
   * angegeben werden koennen. Zudem koennen die Konstanten {@code X} und {@code Y}
   * verwendet werden, um die horizontale und vertikale Raster-Koordinate zu
   * referenzieren und {@code NoData}, bzw. {@code isNoData(.)} als Alias fuer
   * {@code NaN}, bzw. {@code isNaN} verwendet werden.
   */
  protected BinaryTreeNode parseLiteral() {
    String token = nextNonWSToken();

    // Filter
    if (token.startsWith("F")) {
      BinaryTreeNode operandNode = parseRulePart();
      if ( !(operandNode instanceof RasterOperationTree.RasterReferenceNode) )
        throw new UnsupportedOperationException( DataUtil.RESOURCE.getString("RasterOperationTree.err.FilterReferenceNotAllowed") );
      return new RasterOperationTree.FilterReferenceNode(getIntFromString(token.substring(1)),operandNode);
    }

    // Raster-Indikator
    if (token.startsWith("#"))
      return new RasterOperationTree.RasterReferenceNode(getIntFromString(token.substring(1)));

    // X- oder Y-Koordinate, NODATA
    if ( token.equalsIgnoreCase("X") || token.equalsIgnoreCase("Y") ||
         token.equalsIgnoreCase("NODATA") ) {
      return new OperationTree.ConstantAliasNode(token);
    }

    // Check auf NoData
    if ( token.equalsIgnoreCase("ISNODATA") )
      return new OperationTree.UnaryOperatorNode(token, parseRulePart());

    // sonst: auf Literale der Oberklasse zurueckgreifen
    pushbackWithWSToken();
    return super.parseLiteral();
  }
  
  ////////////////////////////////////////////////////////////////
  /////////////////   ParserOperatorsHints   /////////////////////
  ////////////////////////////////////////////////////////////////
  /**
   * Initialises the list of available operators.
   * @see #getOperators()
   * @see #avOperators
   */
  @Override
  protected void initOperatorList() {
    super.initOperatorList();
    // Zur Verfuegung stehende Operatoren erweitern
    avOperators.add("X");           
    avOperators.add("Y");           
    avOperators.add("NoData");   
    avOperators.add("isNoData");   
    
    // Keys fuer Lokalisation neu aufbauen
    initOperatorKeys();
  }
  
  /**
   * Liefert die Anzahl an geklammerten Parametern, die ein Operator hat.
   * Erweitert die Methode der Oberklasse um den Operator {@code isNoData(.)}
   * und die Filter-Referenz {@code F.}.
   * @param op Operator
   * @return 1 fuer {@code isNoData} und {@code F..}, sonst den Wert der {@code super}-Methode
   */
  @Override
  public int getOperatorParameterCount(String op) {
    // Funktionen mit einem geklammerten Parameter
    if (op.equalsIgnoreCase("isNoData") || op.charAt(0) == 'F')
      return 1;
    return super.getOperatorParameterCount(op);
  }

}

