/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data;

import org.opengis.referencing.crs.CoordinateReferenceSystem;

import appl.data.LateLoadable;

/**
 * Dieses Interface stellt die Basis fuer ein georeferenziertes
 * Raster dar.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public interface ReadableGrid extends LateLoadable {
  /**
   * Die Dimension des Rasters (2).
   */
  public static final int RASTER_DIM = 2;

  /**
   * Zerstoert das Raster und gibt alle Ressourcen wieder frei.
   */
  public void dispose();

  /**
   * Liefert die Breite des Rasters (in Zellen).
   */
  public int getWidth();

  /**
   * Liefert die Hoehe des Rasters (in Zellen).
   */
  public int getHeight();

  /**
   * Liefert den Index der ersten (Südwest) Zelle in X-Richtung.
   */
  public int getMinX();

  /**
   * Liefert den Index der ersten Zelle (Südwest) in Y-Richtung.
   */
  public int getMinY();

  /**
   * Liefert das {@link CoordinateReferenceSystem} in dem das Raster
   * dargestellt ist.
   */
  public CoordinateReferenceSystem getCoordinateReferenceSystem();

  /**
   * Liefert die reale Breite des Rasters.
   */
  public double getRealWidth();

  /**
   * Liefert die reale Breite des Rasters.
   */
  public double getRealHeight();

  /**
   * Liefert die reale Breite einer Zelle.<br>
   * Sollte <code>{@link #getRealWidth()} / {@link #getWidth()} entsprechen!
   */
  public double getCellWidth();

  /**
   * Liefert die reale Hoehe einer Zelle.<br>
   * Sollte <code>{@link #getRealHeight()} / {@link #getHeight()} entsprechen!
   */
  public double getCellHeight();

  /**
   * Liefert die X-Koordinate (Longitude) der Georeferenz der linken unteren Ecke des Rasters (Südwest).
   */
  public double getX();

  /**
   * Liefert die Y-Koordinate (Latitude) der Georeferenz der linken unteren Ecke des Rasters (Südwest).
   */
  public double getY();

  /**
   * Liefert die Art der Daten, die im Raster gespeichert werden koennen. Diese
   * wird durch eine der TYPE-Konstanten in {@link java.awt.image.DataBuffer}
   * repraesentiert.
   */
  public int getSampleType();


  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   */
  public Object getRasterSample(int... cell);

  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   */
  public short getRasterSampleAsShort(int... cell);

  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   */
  public byte getRasterSampleAsByte(int... cell);

  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   */
  public int getRasterSampleAsInt(int... cell);

  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   */
  public long getRasterSampleAsLong(int... cell);

  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   */
  public float getRasterSampleAsFloat(int... cell);

  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   */
  public double getRasterSampleAsDouble(int... cell);

  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   */
  public Object getGridSample(double... coord);

  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   */
  public short getGridSampleAsShort(double... coord);

  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   */
  public byte getGridSampleAsByte(double... coord);

  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   */
  public int getGridSampleAsInt(double... coord);

  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   */
  public long getGridSampleAsLong(double... coord);

  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   */
  public float getGridSampleAsFloat(double... coord);

  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   */
  public double getGridSampleAsDouble(double... coord);

  /**
   * Konvertiert eine reale Koordinate in eine Zellennummer. Liegt der
   * Koordinatenwert genau auf der Grenze zwischen zwei Rasterzellen, wird
   * die naechst "groessere" Zellegewaehlt.
   * Ausnahme bildet der Rand des Rasters. Hier wird die kleinere Zelle
   * (also die letzte) herangezogen
   * @param coord Georeferenz-Koordinate
   * @param dim   Dimension, in der die Umrechnung erfolgen soll
   * @exception UnsupportedOperationException falls eine ungueltige Dimension
   *            angegeben wird (nur Werte 0 <= coord < {@link #RASTER_DIM} sind
   *            zulaessig
   */
  public int convertRealToRaster(double coord, int dim);

  /**
   * Konvertiert eine Zellennummer in reale Koordinate. Dabei wird die
   * Koordinate der Zellenmitte zurueckgegeben.
   * @param cell  Rasterzellen-Koordinate
   * @param dim   Dimension, in der die Umrechnung erfolgen soll
   * @exception UnsupportedOperationException falls eine ungueltige Dimension
   *            angegeben wird (nur Werte 0 <= coord < {@link #RASTER_DIM} sind
   *            zulaessig
   */
  public double convertRasterToReal(int cell, int dim);
}
