/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data;

import java.awt.geom.Rectangle2D;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferDouble;
import java.awt.image.DataBufferFloat;
import java.awt.image.DataBufferInt;
import java.awt.image.Raster;
import java.awt.image.RenderedImage;
import java.io.IOException;
import java.io.Serializable;

import org.opengis.referencing.crs.CoordinateReferenceSystem;

import appl.util.RasterMetaData;

/**
 * Diese Klasse stellt eine Implementierung von {@link WritableGrid} dar und
 * basiert auf einem einfachen {@code float}-Array.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class WritableGridArray extends AbstractWritableGrid implements WritableGrid, Serializable {
  /** Die Dimension des Rasters (2). */
  public static final int RASTER_DIM = 2;
  /** Speichert die Geo-Referenz (Latitude, Longitude, Breite, Hoehe) des
   *  Rasters. */
  protected transient Rectangle2D envelope;
  /** Speichert die kleinste horizontale Raster-Koordinate. */
  protected int minX = 0;
  /** Speichert die kleinste vertikale Raster-Koordinate. */
  protected int minY = 0;
  /** Speichert die Breite des Rasters in Zellen. */
  protected int rasterWidth = 0;
  /** Speichert die Hoehe des Rasters in Zellen. */
  protected int rasterHeight = 0;

  /**
   * Erzeugt ein neues Raster.
   * @param minX Raster-Index der ersten Zelle in X-Richtung
   * @param minY Raster-Index der ersten Zelle in Y-Richtung
   * @param rasterWidth Breite des Rasters (in Zellen)
   * @param rasterHeight Hoehe des Rasters (in Zellen)
   * @param envelope Georeferenz des Rasters (Longitude = X-Koordinate, Latitude = Y-Koordinate, Breite und Hoehe des Rasters in Metern)
   */
  public WritableGridArray(int minX, int minY, int rasterWidth, int rasterHeight, Rectangle2D envelope, CoordinateReferenceSystem crs) {
    super(crs);
    this.minX = minX;
    this.minY = minY;
    this.rasterWidth  = rasterWidth;
    this.rasterHeight = rasterHeight;
    this.envelope = envelope;
  }

  /**
   * Erzeugt ein neues Raster in WGS84.
   * @param minX Raster-Index der ersten Zelle in X-Richtung
   * @param minY Raster-Index der ersten Zelle in Y-Richtung
   * @param rasterWidth Breite des Rasters (in Zellen)
   * @param rasterHeight Hoehe des Rasters (in Zellen)
   * @param envelope Georeferenz des Rasters (Longitude = X-Koordinate, Latitude = Y-Koordinate, Breite und Hoehe des Rasters in Metern)
   */
//  public WritableGridArray(int minX, int minY, int rasterWidth, int rasterHeight, Rectangle2D envelope) {
//    this(minX, minY,rasterWidth,rasterHeight,envelope,null);
//  }

  /**
   * Erzeugt ein neues Raster in WGS84.
   * @param minX Raster-Index der ersten Zelle in X-Richtung
   * @param minY Raster-Index der ersten Zelle in Y-Richtung
   * @param rasterWidth Breite des Rasters (in Zellen)
   * @param rasterHeight Hoehe des Rasters (in Zellen)
   * @param lon Georeferenz (Longitude = X-Koordinate)
   * @param lat Georeferenz (Latitude = Y-Koordinate)
   * @param realWidth Breite des Rasters (in Metern)
   * @param realHeight Hoehe des Rasters (in Metern)
   */
//  public WritableGridArray(int minX, int minY, int rasterWidth, int rasterHeight, double lon, double lat, double realWidth, double realHeight) {
//    this( minX, minY, rasterWidth, rasterHeight, new Rectangle2D.Double(lon,lat,realWidth,realHeight) );
//  }

  /**
   * Erzeugt ein neues Raster in WGS84.
   * @param rasterWidth Breite des Rasters (in Zellen)
   * @param rasterHeight Hoehe des Rasters (in Zellen)
   * @param lon Georeferenz (Longitude = X-Koordinate)
   * @param lat Georeferenz (Latitude = Y-Koordinate)
   * @param realWidth Breite des Rasters (in Metern)
   * @param realHeight Hoehe des Rasters (in Metern)
   */
//  public WritableGridArray(int rasterWidth, int rasterHeight, double lon, double lat, double realWidth, double realHeight) {
//    this(0,0,rasterWidth,rasterHeight,lon,lat,realWidth,realHeight);
//  }

  /**
   * Erzeugt ein neues Raster in WGS84.
   * @param rasterWidth Breite des Rasters (in Zellen)
   * @param rasterHeight Hoehe des Rasters (in Zellen)
   * @param envelope Georeferenz des Rasters (Longitude = X-Koordinate, Latitude = Y-Koordinate, Breite und Hoehe des Rasters in Metern)
   */
//  public WritableGridArray(int rasterWidth, int rasterHeight, Rectangle2D envelope) {
//    this(0,0,rasterWidth,rasterHeight,envelope);
//  }

  /**
   * Erzeugt ein neues Raster.
   * @param image Daten
   * @param envelope Georeferenz und Ausdehnung des Rasters
   * @param crs CoordinateReferenceSystem fuer das Raster
   * @return Ein {@link WritableGridArray.Integer}, {@link WritableGridArray.Float} oder
   *         {@link WritableGridArray.Double}, je nach Datentyp des {@code image}-DataBuffers
   */
  public static WritableGridArray create(RenderedImage image, Rectangle2D envelope, CoordinateReferenceSystem crs) {
    return create(image.getData(),envelope,crs);
//    int width  = image.getWidth();
//    int height = image.getHeight();
//    switch (image.getData().getDataBuffer().getDataType()) {
//      case DataBuffer.TYPE_BYTE:
//      case DataBuffer.TYPE_SHORT:
//      case DataBuffer.TYPE_USHORT:
//      case DataBuffer.TYPE_INT:    return new WritableGridArray.Integer(width,height,envelope,crs,image.getData());
//      case DataBuffer.TYPE_FLOAT:  return new WritableGridArray.Float(width,height,envelope,crs,image.getData());
//      case DataBuffer.TYPE_DOUBLE: return new WritableGridArray.Double(width,height,envelope,crs,image.getData());
//
//    }
//    throw new UnsupportedOperationException("Data type of image not yet supported: "+image.getData().getDataBuffer().getDataType());
  }

  /**
   * Erzeugt ein neues Raster in WGS84.
   * @param image Daten
   * @param envelope Georeferenz und Ausdehnung des Rasters
   * @return Ein {@link WritableGridArray.Integer}, {@link WritableGridArray.Float} oder
   *         {@link WritableGridArray.Double}, je nach Datentyp des {@code image}-DataBuffers
   */
  public static WritableGridArray create(RenderedImage image, Rectangle2D envelope) {
    return create(image,envelope,null);
  }

  /**
   * Erzeugt ein neues Raster.
   * @param image Daten
   * @param envelope Georeferenz und Ausdehnung des Rasters
   * @param crs CoordinateReferenceSystem fuer das Raster
   * @return Ein {@link WritableGridArray.Integer}, {@link WritableGridArray.Float} oder
   *         {@link WritableGridArray.Double}, je nach Datentyp des {@code image}-DataBuffers
   */
  public static WritableGridArray create(Raster raster, Rectangle2D envelope, CoordinateReferenceSystem crs) {
    int width  = raster.getWidth();
    int height = raster.getHeight();
    switch (raster.getDataBuffer().getDataType()) {
      case DataBuffer.TYPE_BYTE:
      case DataBuffer.TYPE_SHORT:
      case DataBuffer.TYPE_USHORT:
      case DataBuffer.TYPE_INT:    return new WritableGridArray.Integer(width,height,envelope,crs,raster);
      case DataBuffer.TYPE_FLOAT:  return new WritableGridArray.Float(width,height,envelope,crs,raster);
      case DataBuffer.TYPE_DOUBLE: return new WritableGridArray.Double(width,height,envelope,crs,raster);

    }
    throw new UnsupportedOperationException("Data type of image not yet supported: "+raster.getDataBuffer().getDataType());
  }

  /**
   * Erzeugt ein neues Raster in WGS84.
   * @param raster Daten
   * @param envelope Georeferenz und Ausdehnung des Rasters
   * @return Ein {@link WritableGridArray.Integer}, {@link WritableGridArray.Float} oder
   *         {@link WritableGridArray.Double}, je nach Datentyp des {@code raster}-DataBuffers
   */
  public static WritableGridArray create(Raster raster, Rectangle2D envelope) {
    return create(raster,envelope,null);
  }

  /**
   * Creates an empty GridArray out of a {@link RasterMetaData} object.
   * @param metaData the MetaData
   * @param crs CoordinateReferenceSystem fuer das Raster
   * @return A {@link WritableGridArray.Integer}, {@link WritableGridArray.Float} or
   *         {@link WritableGridArray.Double}, depending on the Datatype
   */
  public static WritableGridArray createEmpty(RasterMetaData metaData, CoordinateReferenceSystem crs) {
    switch (metaData.getDataType()) {
      case DataBuffer.TYPE_INT:    return new WritableGridArray.Integer(metaData.getMinX(),metaData.getMinY(),metaData.getWidth(),metaData.getHeight(),metaData.getX(),metaData.getY(),metaData.getRealWidth(),metaData.getRealHeight(),crs,null);
      case DataBuffer.TYPE_FLOAT:  return new WritableGridArray.Float(metaData.getMinX(),metaData.getMinY(),metaData.getWidth(),metaData.getHeight(),metaData.getX(),metaData.getY(),metaData.getRealWidth(),metaData.getRealHeight(),crs,null);
      case DataBuffer.TYPE_DOUBLE: return new WritableGridArray.Double(metaData.getMinX(),metaData.getMinY(),metaData.getWidth(),metaData.getHeight(),metaData.getX(),metaData.getY(),metaData.getRealWidth(),metaData.getRealHeight(),crs,null);

    }
    throw new UnsupportedOperationException("Data type of image not yet supported!");
  }

  /**
   * Creates an empty GridArray out of a {@link RasterMetaData} object. WGS84 is
   * used for CRS.
   * @param metaData the MetaData
   * @return A {@link WritableGridArray.Integer}, {@link WritableGridArray.Float} or
   *         {@link WritableGridArray.Double}, depending on the Datatype
   */
  public static WritableGridArray createEmpty(RasterMetaData metaData) {
    return createEmpty(metaData,metaData.getCoordinateReferenceSystem());
  }

  /**
   * Liefert eine Referenz auf den kompletten Inhalt des Rasters in Form eines
   * {@link DataBuffer}.
   */
  public abstract DataBuffer getDataBuffer();

  /**
   * Liefert eine Referenz auf den kompletten Inhalt des Rasters.
   * @return einen 1-dimensionalen Array eines Basis-Datentyps
   */
  public abstract Object getData();

  /**
   * Liefert eine Kopie des kompletten Inhalts des Rasters.
   * @param data Array in den die Daten geschrieben werden (kann {@code null} sein!)
   * @return einen 2-dimensionaler Array eines Basis-Datentyps (z.B. {@code float[height][width]})
   */
  public abstract Object getData(Object data);

  /**
   * Macht zur Zeit noch nichts.
   * @todo WritableGridArray.dispose() implementieren
   */
  public void dispose() {
  }

  /**
   * Liefert die X-Koordinate der Georeferenz (Longitude) der linken unteren
   * Ecke des Rasters (Südwest).
   */
  public double getX() {
    return envelope.getX();
  }

  /**
   * Liefert die Y-Koordinate der Georeferenz (Latitude) der linken unteren
   * Ecke des Rasters (Südwest).
   */
  public double getY() {
    return envelope.getY();
  }

  /**
   * Liefert die reale Breite des Rasters.
   */
  public double getRealWidth() {
    return envelope.getWidth();
  }

  /**
   * Liefert die reale Breite des Rasters.
   */
  public double getRealHeight() {
    return envelope.getHeight();
  }

  /**
   * Liefert die Georeferenz (Lat/Lon) und Ausdehnung des Rasters.
   */
  public Rectangle2D getEnvelope() {
    return (Rectangle2D)envelope.clone();
  }

  /**
   * Liefert die Breite des Rasters (in Zellen).
   */
  public int getWidth() {
    return this.rasterWidth;
  }

  /**
   * Liefert die Hoehe des Rasters (in Zellen).
   */
  public int getHeight() {
    return this.rasterHeight;
  }

  /**
   * Liefert den Index der ersten (Südwest) Zelle in X-Richtung.
   */
  public int getMinX() {
    return this.minX;
  }

  /**
   * Liefert den Index der ersten Zelle (Südwest) in Y-Richtung.
   */
  public int getMinY() {
    return this.minY;
  }

  /** Serializes the Grid
   * @author Dominik Appl
   * @param s
   * @throws IOException
   */
  private synchronized void writeObject( java.io.ObjectOutputStream s )
  throws IOException
  {
	  //writing all attributes to the outputstream
	  s.defaultWriteObject();
	  //writing the transient envelope (which is not serializable)
	  s.writeDouble(envelope.getX());
	  s.writeDouble(envelope.getY());
	  s.writeDouble(envelope.getHeight());
	  s.writeDouble(envelope.getWidth());

  }

  /** Deserializes the Grid
   * @author Dominik Appl
   * @param s
   * @throws IOException
   */
  private synchronized void readObject( java.io.ObjectInputStream s )
   throws IOException, ClassNotFoundException
   {
	  //reading standard fields
	  s.defaultReadObject();
	  //reading envelope
	  double lat = s.readDouble();
	  double lon = s.readDouble();
	  double realHeight = s.readDouble();
	  double realWidth = s.readDouble();
	  envelope = new Rectangle2D.Double(lat,lon,realWidth,realHeight);
   }

  /////////////////////////////////////////////////////////////////////////
  ///////////////////////////   Float - Raster ////////////////////////////
  /////////////////////////////////////////////////////////////////////////
  public static class Float extends WritableGridArray {
    /** Speichert zeilenweise die Daten des {@code float}-Rasters ({@code data[y*w+x])} */
    protected float[] data = null;

    /**
     * Erzeugt ein neues {@code float}-Raster.
     * @param minX Raster-Index der ersten Zelle in X-Richtung
     * @param minY Raster-Index der ersten Zelle in Y-Richtung
     * @param rasterWidth Breite des Rasters (in Zellen)
     * @param rasterHeight Hoehe des Rasters (in Zellen)
     * @param envelope Georeferenz des Rasters (Longitude = X-Koordinate, Latitude = Y-Koordinate, Breite und Hoehe des Rasters in Metern)
     * @param crs Referenzsystem fuer das Raster
     * @param data Inhalt des Rasters (wenn nicht {@code null}, muss der Array
     *             mindestens die Groesse {@code rasterWidth * rasterHeight} haben)
     */
    public Float(int minX, int minY, int rasterWidth, int rasterHeight, Rectangle2D envelope, CoordinateReferenceSystem crs, float[] data) {
      super(minX,minY,rasterWidth,rasterHeight,envelope,crs);
      if ( data == null )
        data = new float[rasterHeight*rasterWidth];
      if ( data.length < rasterHeight*rasterWidth )
        throw new IllegalArgumentException("Array for raster content to small!");
      this.data = data;
    }

    /**
     * Erzeugt ein neues {@code float}-Raster.
     * @param minX Raster-Index der ersten Zelle in X-Richtung
     * @param minY Raster-Index der ersten Zelle in Y-Richtung
     * @param rasterWidth Breite des Rasters (in Zellen)
     * @param rasterHeight Hoehe des Rasters (in Zellen)
     * @param lon Georeferenz (Longitude = X-Koordinate)
     * @param lat Georeferenz (Latitude = Y-Koordinate)
     * @param realWidth Breite des Rasters (in Metern)
     * @param realHeight Hoehe des Rasters (in Metern)
     * @param crs Referenzsystem fuer das Raster
     * @param data Inhalt des Rasters (wenn nicht {@code null}, muss der Array
     *             mindestens die Groesse {@code rasterWidth * rasterHeight} haben)
     */
    public Float(int minX, int minY, int rasterWidth, int rasterHeight, double lon, double lat, double realWidth, double realHeight, CoordinateReferenceSystem crs, float[] data) {
      this( minX, minY, rasterWidth, rasterHeight, new Rectangle2D.Double(lon,lat,realWidth,realHeight),crs, data );
    }

    /**
     * Erzeugt ein neues leeres {@code float}-Raster.
     * @param rasterWidth Breite des Rasters (in Zellen)
     * @param rasterHeight Hoehe des Rasters (in Zellen)
     * @param lon Georeferenz (Longitude = X-Koordinate)
     * @param lat Georeferenz (Latitude = Y-Koordinate)
     * @param realWidth Breite des Rasters (in Metern)
     * @param realHeight Hoehe des Rasters (in Metern)
     * @param crs Referenzsystem fuer das Raster
     * @param data Inhalt des Rasters (wenn nicht {@code null}, muss der Array
     *             mindestens die Groesse {@code rasterWidth * rasterHeight} haben)
     */
    public Float(int rasterWidth, int rasterHeight, double lon, double lat, double realWidth, double realHeight, CoordinateReferenceSystem crs, float[] data) {
      this(0,0,rasterWidth,rasterHeight,lon,lat,realWidth,realHeight,crs,data);
    }

    /**
     * Erzeugt ein neues leeres {@code float}-Raster.
     * @param rasterWidth Breite des Rasters (in Zellen)
     * @param rasterHeight Hoehe des Rasters (in Zellen)
     * @param envelope Georeferenz des Rasters (Longitude = X-Koordinate, Latitude = Y-Koordinate, Breite und Hoehe des Rasters in Metern)
     * @param crs Referenzsystem fuer das Raster
     * @param data Inhalt des Rasters (wenn nicht {@code null}, muss der Array
     *             mindestens die Groesse {@code rasterWidth * rasterHeight} haben)
     */
    public Float(int rasterWidth, int rasterHeight, Rectangle2D envelope, CoordinateReferenceSystem crs, float[] data) {
      this(0,0,rasterWidth,rasterHeight,envelope,crs,data);
    }

    /**
     * Erzeugt ein neues leeres {@code float}-Raster.
     * @param rasterWidth Breite des Rasters (in Zellen)
     * @param rasterHeight Hoehe des Rasters (in Zellen)
     * @param envelope Georeferenz des Rasters (Longitude = X-Koordinate, Latitude = Y-Koordinate, Breite und Hoehe des Rasters in Metern)
     * @param crs Referenzsystem fuer das Raster
     * @param data Inhalt des Rasters (wenn nicht {@code null}, muss der Array
     *             mindestens die Groesse {@code rasterWidth * rasterHeight} haben)
     */
    public Float(int rasterWidth, int rasterHeight, Rectangle2D envelope, CoordinateReferenceSystem crs, float[][] data) {
      this(0,0,rasterWidth,rasterHeight,envelope,crs,null);
      // Daten in Array uebernehmen
      for (int y=0; y<rasterHeight; y++)
        for (int x=0; x<rasterWidth; x++)
          this.data[y*rasterWidth+x] = data[x][y];
    }

    /**
     * Erzeugt ein neues leeres {@code float}-Raster.
     * @param rasterWidth Breite des Rasters (in Zellen)
     * @param rasterHeight Hoehe des Rasters (in Zellen)
     * @param envelope Georeferenz des Rasters (Longitude = X-Koordinate, Latitude = Y-Koordinate, Breite und Hoehe des Rasters in Metern)
     * @param crs Referenzsystem fuer das Raster
     * @param data Inhalt des Rasters (wenn nicht {@code null}, muss das Raster
     *             mindestens die Groesse {@code rasterWidth * rasterHeight} haben)
     */
    public Float(int rasterWidth, int rasterHeight, Rectangle2D envelope, CoordinateReferenceSystem crs, Raster data) {
      this(0,0,rasterWidth,rasterHeight,envelope,crs,null);
      // Daten in Array uebernehmen
      data.getPixels(0,0,rasterWidth,rasterHeight,this.data);
    }

    /**
     * Liefert eine Referenz auf den kompletten Inhalt des Raster.
     * @return {@code float[height*width]}-Array.
     */
    public float[] getData() {
      return this.data;
    }

    /**
     * Liefert eine Referenz auf den kompletten Inhalt des Rasters in Form eines
     * {@link DataBuffer}.
     */
    public DataBufferFloat getDataBuffer() {
      return new DataBufferFloat(data,data.length);
    }

    /**
     * Liefert eine Kopie des kompletten Inhalts des Rasters.
     * @param data Array in den die Daten geschrieben werden (kann {@code null} sein!)
     * @return einen {@code float[height][width]}-Array.
     */
    public float[][] getData(Object data) {
      return getData((float[][])data);
    }
    /**
     * Liefert eine Kopie des kompletten Inhalts des Rasters.
     * @param data Array in den die Daten geschrieben werden (kann {@code null} sein!)
     * @return {@code float[height][width]}-Array.
     */
    public float[][] getData(float[][] data) {
      if ( data == null )
        data = new float[rasterHeight][rasterWidth];
      for (int y=0; y<rasterHeight; y++)
        for (int x=0; x<rasterWidth; x++)
          data[y][x] = this.data[y*rasterWidth+x];
      return data;
    }

    /**
     * Liefert eine Kopie des kompletten Inhalts des Rasters. Die {@code float}-Werte
     * des Rasters werden einfach zu {@code int} gecastet.
     * @param data Array in den die Daten geschrieben werden (kann {@code null} sein!)
     * @return {@code int[height][width]}-Array.
     */
    public int[][] getData(int[][] data) {
      if ( data == null )
        data = new int[rasterHeight][rasterWidth];
      for (int y=0; y<rasterHeight; y++)
        for (int x=0; x<rasterWidth; x++)
          data[y][x] = (int)this.data[y*rasterWidth+x];
      return data;
    }

    /**
     * Liefert eine Kopie des kompletten Inhalts des Rasters. Die {@code float}-Werte
     * des Rasters werden einfach zu {@code double} gecastet.
     * @param data Array in den die Daten geschrieben werden (kann {@code null} sein!)
     * @return {@code double[height][width]}-Array.
     */
    public double[][] getData(double[][] data) {
      if ( data == null )
        data = new double[rasterHeight][rasterWidth];
      for (int y=0; y<rasterHeight; y++)
        for (int x=0; x<rasterWidth; x++)
          data[y][x] = this.data[y*rasterWidth+x];
      return data;
    }

    /**
     * Liefert die Art der Daten, die im Raster gespeichert werden koennen. Diese
     * wird durch eine der TYPE-Konstanten in {@link DataBuffer}
     * repraesentiert.
     * @return immer {@link DataBuffer#TYPE_FLOAT DataBuffer.TYPE_FLOAT}
     */
    public int getSampleType() {
      return DataBuffer.TYPE_FLOAT;
    }

    /**
     * Setzt einen Wert im Raster ueber Raster-Koordinaten.
     * @param value neuer Wert
     * @param cell  2D-Raster-Koordinate (Zellenindizes, beginnend bei
     *              {@link #getMinX()} und {@link #getMinY()})
     * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
     *            Koordinaten angegeben werden
     */
    public void setRasterSample(Object value, int ...cell) {
      if (cell.length < RASTER_DIM)
        throw new UnsupportedOperationException(RASTER_DIM +
                                                " Coordinates expected");
      if (value == null)
        throw new UnsupportedOperationException(getClass().getSimpleName() +
                                                " can not store null-Objects");
      if (! (value instanceof java.lang.Number))
        throw new UnsupportedOperationException(getClass().getSimpleName() +
                                                " only can not store "+value.getClass().getSimpleName()+"-Objects");

      int x = cell[0] - minX;
      int y = cell[1] - minY;
      data[y*rasterWidth+x] = ((Number)value).floatValue();
    }

    /**
     * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
     * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
     *             {@link #getMinX()} und {@link #getMinY()})
     * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
     *            Koordinaten angegeben werden
     */
    public java.lang.Float getRasterSample(int ...cell) {
      if (cell.length < RASTER_DIM)
        throw new UnsupportedOperationException(RASTER_DIM + " Coordinates expected");
      int x = cell[0] - minX;
      int y = cell[1] - minY;
      return data[y*rasterWidth+x];
    }

    /**
     * Liefert einen Wert des Rasters ueber Raster-Koordinaten. Greift direkt
     * auf den Daten-Array zu und nutzt nicht die generischen Methoden der
     * Oberklasse.
     * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
     *             {@link #getMinX()} und {@link #getMinY()})
     * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
     *            Koordinaten angegeben werden
     */
    public float getRasterSampleAsFloat(int... cell) {
      int x = cell[0] - minX;
      int y = cell[1] - minY;
      return data[y*rasterWidth+x];
    }

    /**
     * Setzt einen Wert des Rasters ueber Raster-Koordinaten. Greift direkt
     * auf den Daten-Array zu und nutzt nicht die generischen Methoden der
     * Oberklasse.
     * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
     *             {@link #getMinX()} und {@link #getMinY()})
     * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
     *            Koordinaten angegeben werden
     */
    public void setRasterSample(float value, int... cell) {
      if ( cell.length < RASTER_DIM )
        throw new UnsupportedOperationException(RASTER_DIM+" Coordinates expected");
      int x = cell[0] - minX;
      int y = cell[1] - minY;
      data[y*rasterWidth+x] = value;
    }

    /**
     * Liefert einen Wert des Rasters ueber Geo-Koordinaten. Greift direkt
     * auf den Daten-Array zu und nutzt nicht die generischen Methoden der
     * Oberklasse.
     * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
     * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
     *            Koordinaten angegeben werden
     */
    public float getGridSampleAsFloat(double... coord){
      if ( coord.length < RASTER_DIM )
        throw new UnsupportedOperationException(RASTER_DIM+" Coordinates expected");
      // Koordinaten umrechnen in Zellen-Index
      int cellX = convertRealToRaster(coord[0],0);
      int cellY = convertRealToRaster(coord[1],1);
      return data[cellY*rasterWidth + cellX];
    }

    /**
     * Setzt einen Wert des Rasters ueber Geo-Koordinaten. Liegt der
     * Koordinatenwert genau auf der Grenze zwischen zwei Rasterzellen, wird die
     * naechst groessere Zelle gewaehlt (ausser die Grenze entspricht dem Raster-Rand!).
     * Greift direkt auf den Daten-Array zu und nutzt nicht die generischen Methoden der
     * Oberklasse.
     * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
     * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
     *            Koordinaten angegeben werden
     */
    public void setGridSample(float value, int... coord) {
      if ( coord.length < RASTER_DIM )
        throw new UnsupportedOperationException(RASTER_DIM+" Coordinates expected");
      // Koordinaten umrechnen in Zellen-Index
      int cellX = convertRealToRaster(coord[0],0);
      int cellY = convertRealToRaster(coord[1],1);
      data[cellY*rasterWidth+cellX] = value;
    }
  }

  /////////////////////////////////////////////////////////////////////////
  ///////////////////////////   Double - Raster ////////////////////////////
  /////////////////////////////////////////////////////////////////////////
  public static class Double extends WritableGridArray {
    /** Speichert zeilenweise die Daten des {@code double}-Rasters ({@code data[y*w+x])} */
    protected double[] data = null;

    /**
     * Erzeugt ein neues {@code double}-Raster.
     * @param minX Raster-Index der ersten Zelle in X-Richtung
     * @param minY Raster-Index der ersten Zelle in Y-Richtung
     * @param rasterWidth Breite des Rasters (in Zellen)
     * @param rasterHeight Hoehe des Rasters (in Zellen)
     * @param envelope Georeferenz des Rasters (Longitude = X-Koordinate, Latitude = Y-Koordinate, Breite und Hoehe des Rasters in Metern)
     * @param crs Referenzsystem fuer das Raster
     * @param data Inhalt des Rasters (wenn nicht {@code null}, muss der Array
     *             mindestens die Groesse {@code rasterWidth * rasterHeight} haben)
     */
    public Double(int minX, int minY, int rasterWidth, int rasterHeight, Rectangle2D envelope, CoordinateReferenceSystem crs, double[] data) {
      super(minX,minY,rasterWidth,rasterHeight,envelope,crs);
      if ( data == null )
        data = new double[rasterHeight*rasterWidth];
      if ( data.length < rasterHeight*rasterWidth )
        throw new IllegalArgumentException("Array for raster content to small!");
      this.data = data;
    }

    /**
     * Erzeugt ein neues {@code double}-Raster.
     * @param minX Raster-Index der ersten Zelle in X-Richtung
     * @param minY Raster-Index der ersten Zelle in Y-Richtung
     * @param rasterWidth Breite des Rasters (in Zellen)
     * @param rasterHeight Hoehe des Rasters (in Zellen)
     * @param lon Georeferenz (Longitude = X-Koordinate)
     * @param lat Georeferenz (Latitude = Y-Koordinate)
     * @param realWidth Breite des Rasters (in Metern)
     * @param realHeight Hoehe des Rasters (in Metern)
     * @param crs Referenzsystem fuer das Raster
     * @param data Inhalt des Rasters (wenn nicht {@code null}, muss der Array
     *             mindestens die Groesse {@code rasterWidth * rasterHeight} haben)
     */
    public Double(int minX, int minY, int rasterWidth, int rasterHeight, double lon, double lat, double realWidth, double realHeight, CoordinateReferenceSystem crs, double[] data) {
      this( minX, minY, rasterWidth, rasterHeight, new Rectangle2D.Double(lon,lat,realWidth,realHeight), crs, data );
    }

    /**
     * Erzeugt ein neues leeres {@code double}-Raster.
     * @param rasterWidth Breite des Rasters (in Zellen)
     * @param rasterHeight Hoehe des Rasters (in Zellen)
     * @param lon Georeferenz (Longitude = X-Koordinate)
     * @param lat Georeferenz (Latitude = Y-Koordinate)
     * @param realWidth Breite des Rasters (in Metern)
     * @param realHeight Hoehe des Rasters (in Metern)
     * @param crs Referenzsystem fuer das Raster
     * @param data Inhalt des Rasters (wenn nicht {@code null}, muss der Array
     *             mindestens die Groesse {@code rasterWidth * rasterHeight} haben)
     */
    public Double(int rasterWidth, int rasterHeight, double lon, double lat, double realWidth, double realHeight, CoordinateReferenceSystem crs, double[] data) {
      this(0,0,rasterWidth,rasterHeight,lon,lat,realWidth,realHeight,crs,data);
    }

    /**
     * Erzeugt ein neues leeres {@code double}-Raster.
     * @param rasterWidth Breite des Rasters (in Zellen)
     * @param rasterHeight Hoehe des Rasters (in Zellen)
     * @param envelope Georeferenz des Rasters (Longitude = X-Koordinate, Latitude = Y-Koordinate, Breite und Hoehe des Rasters in Metern)
     * @param crs Referenzsystem fuer das Raster
     * @param data Inhalt des Rasters (wenn nicht {@code null}, muss der Array
     *             mindestens die Groesse {@code rasterWidth * rasterHeight} haben)
     */
    public Double(int rasterWidth, int rasterHeight, Rectangle2D envelope, CoordinateReferenceSystem crs, double[] data) {
      this(0,0,rasterWidth,rasterHeight,envelope,crs,data);
    }

    /**
     * Erzeugt ein neues leeres {@code double}-Raster.
     * @param rasterWidth Breite des Rasters (in Zellen)
     * @param rasterHeight Hoehe des Rasters (in Zellen)
     * @param envelope Georeferenz des Rasters (Longitude = X-Koordinate, Latitude = Y-Koordinate, Breite und Hoehe des Rasters in Metern)
     * @param crs Referenzsystem fuer das Raster
     * @param data Inhalt des Rasters (wenn nicht {@code null}, muss der Array
     *             mindestens die Groesse {@code rasterWidth * rasterHeight} haben)
     */
    public Double(int rasterWidth, int rasterHeight, Rectangle2D envelope, CoordinateReferenceSystem crs, double[][] data) {
      this(0,0,rasterWidth,rasterHeight,envelope,crs,null);
      // Daten in Array uebernehmen
      for (int y=0; y<rasterHeight; y++)
        for (int x=0; x<rasterWidth; x++)
          this.data[y*rasterWidth+x] = data[x][y];
    }

    /**
     * Erzeugt ein neues leeres {@code double}-Raster.
     * @param rasterWidth Breite des Rasters (in Zellen)
     * @param rasterHeight Hoehe des Rasters (in Zellen)
     * @param envelope Georeferenz des Rasters (Longitude = X-Koordinate, Latitude = Y-Koordinate, Breite und Hoehe des Rasters in Metern)
     * @param crs Referenzsystem fuer das Raster
     * @param data Inhalt des Rasters (wenn nicht {@code null}, muss das Raster
     *             mindestens die Groesse {@code rasterWidth * rasterHeight} haben)
     */
    public Double(int rasterWidth, int rasterHeight, Rectangle2D envelope, CoordinateReferenceSystem crs, Raster data) {
      this(0,0,rasterWidth,rasterHeight,envelope,crs,null);
      // Daten in Array uebernehmen
      data.getPixels(0,0,rasterWidth,rasterHeight,this.data);
    }

    /**
     * Liefert eine Referenz auf den kompletten Inhalt des Raster.
     * @return {@code double[height*width]}-Array.
     */
    public double[] getData() {
      return this.data;
    }

    /**
     * Liefert eine Referenz auf den kompletten Inhalt des Rasters in Form eines
     * {@link DataBuffer}.
     */
    public DataBufferDouble getDataBuffer() {
      return new DataBufferDouble(data,data.length);
    }

    /**
     * Liefert eine Kopie des kompletten Inhalts des Rasters.
     * @param data Array in den die Daten geschrieben werden (kann {@code null} sein!)
     * @return einen {@code double[height][width]}-Array.
     */
    public double[][] getData(Object data) {
      return getData((double[][])data);
    }
    /**
     * Liefert eine Kopie des kompletten Inhalts des Rasters.
     * @param data Array in den die Daten geschrieben werden (kann {@code null} sein!)
     * @return {@code double[height][width]}-Array.
     */
    public double[][] getData(double[][] data) {
      if ( data == null )
        data = new double[rasterHeight][rasterWidth];
      for (int y=0; y<rasterHeight; y++)
        for (int x=0; x<rasterWidth; x++)
          data[y][x] = this.data[y*rasterWidth+x];
      return data;
    }

    /**
     * Liefert eine Kopie des kompletten Inhalts des Rasters. Die {@code double}-Werte
     * des Rasters werden einfach zu {@code float} gecastet.
     * @param data Array in den die Daten geschrieben werden (kann {@code null} sein!)
     * @return {@code float[height][width]}-Array.
     */
    public float[][] getData(float[][] data) {
      if ( data == null )
        data = new float[rasterHeight][rasterWidth];
      for (int y=0; y<rasterHeight; y++)
        for (int x=0; x<rasterWidth; x++)
          data[y][x] = (float)this.data[y*rasterWidth+x];
      return data;
    }

    /**
     * Liefert eine Kopie des kompletten Inhalts des Rasters. Die {@code double}-Werte
     * des Rasters werden einfach zu {@code int} gecastet.
     * @param data Array in den die Daten geschrieben werden (kann {@code null} sein!)
     * @return {@code int[height][width]}-Array.
     */
    public int[][] getData(int[][] data) {
      if ( data == null )
        data = new int[rasterHeight][rasterWidth];
      for (int y=0; y<rasterHeight; y++)
        for (int x=0; x<rasterWidth; x++)
          data[y][x] = (int)this.data[y*rasterWidth+x];
      return data;
    }

    /**
     * Liefert die Art der Daten, die im Raster gespeichert werden koennen. Diese
     * wird durch eine der TYPE-Konstanten in {@link DataBuffer}
     * repraesentiert.
     * @return immer {@link DataBuffer#TYPE_DOUBLE DataBuffer.TYPE_DOUBLE}
     */
    public int getSampleType() {
      return DataBuffer.TYPE_DOUBLE;
    }

    /**
     * Setzt einen Wert im Raster ueber Raster-Koordinaten.
     * @param value neuer Wert
     * @param cell  2D-Raster-Koordinate (Zellenindizes, beginnend bei
     *              {@link #getMinX()} und {@link #getMinY()})
     * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
     *            Koordinaten angegeben werden
     */
    public void setRasterSample(Object value, int ...cell) {
      if (cell.length < RASTER_DIM)
        throw new UnsupportedOperationException(RASTER_DIM + " Coordinates expected");
      if (value == null)
        throw new UnsupportedOperationException(getClass().getSimpleName() +
                                                " can not store null-Objects");
      if (! (value instanceof java.lang.Number))
        throw new UnsupportedOperationException(getClass().getSimpleName() +
                                                " only can not store "+value.getClass().getSimpleName()+"-Objects");
      int x = cell[0] - minX;
      int y = cell[1] - minY;
      data[y*rasterWidth+x] = ((java.lang.Number)value).doubleValue();
    }

    /**
     * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
     * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
     *             {@link #getMinX()} und {@link #getMinY()})
     * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
     *            Koordinaten angegeben werden
     */
    public java.lang.Double getRasterSample(int ...cell) {
      if (cell.length < RASTER_DIM)
        throw new UnsupportedOperationException(RASTER_DIM + " Coordinates expected");
      int x = cell[0] - minX;
      int y = cell[1] - minY;
      return data[y*rasterWidth+x];
    }

    /**
     * Liefert einen Wert des Rasters ueber Raster-Koordinaten. Greift direkt
     * auf den Daten-Array zu und nutzt nicht die generischen Methoden der
     * Oberklasse.
     * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
     *             {@link #getMinX()} und {@link #getMinY()})
     * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
     *            Koordinaten angegeben werden
     */
    public double getRasterSampleAsDouble(int... cell) {
      int x = cell[0] - minX;
      int y = cell[1] - minY;
      return data[y*rasterWidth+x];
    }

    /**
     * Setzt einen Wert des Rasters ueber Raster-Koordinaten. Greift direkt
     * auf den Daten-Array zu und nutzt nicht die generischen Methoden der
     * Oberklasse.
     * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
     *             {@link #getMinX()} und {@link #getMinY()})
     * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
     *            Koordinaten angegeben werden
     */
    public void setRasterSample(double value, int... cell) {
      if ( cell.length < RASTER_DIM )
        throw new UnsupportedOperationException(RASTER_DIM+" Coordinates expected");
      int x = cell[0] - minX;
      int y = cell[1] - minY;
      data[y*rasterWidth+x] = value;
    }

    /**
     * Liefert einen Wert des Rasters ueber Geo-Koordinaten. Greift direkt
     * auf den Daten-Array zu und nutzt nicht die generischen Methoden der
     * Oberklasse.
     * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
     * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
     *            Koordinaten angegeben werden
     */
    public double getGridSampleAsDouble(double... coord){
      if ( coord.length < RASTER_DIM )
        throw new UnsupportedOperationException(RASTER_DIM+" Coordinates expected");
      // Koordinaten umrechnen in Zellen-Index
      int cellX = convertRealToRaster(coord[0],0);
      int cellY = convertRealToRaster(coord[1],1);
      return data[cellY*rasterWidth + cellX];
    }

    /**
     * Setzt einen Wert des Rasters ueber Geo-Koordinaten. Liegt der
     * Koordinatenwert genau auf der Grenze zwischen zwei Rasterzellen, wird die
     * naechst groessere Zelle gewaehlt (ausser die Grenze entspricht dem Raster-Rand!).
     * Greift direkt auf den Daten-Array zu und nutzt nicht die generischen Methoden der
     * Oberklasse.
     * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
     * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
     *            Koordinaten angegeben werden
     */
    public void setGridSample(double value, int... coord) {
      if ( coord.length < RASTER_DIM )
        throw new UnsupportedOperationException(RASTER_DIM+" Coordinates expected");
      // Koordinaten umrechnen in Zellen-Index
      int cellX = convertRealToRaster(coord[0],0);
      int cellY = convertRealToRaster(coord[1],1);
      data[cellY*rasterWidth+cellX] = value;
    }
  }

  /////////////////////////////////////////////////////////////////////////
  ///////////////////////////   Integer - Raster ////////////////////////////
  /////////////////////////////////////////////////////////////////////////
  public static class Integer extends WritableGridArray {
    /** Speichert zeilenweise die Daten des {@code int}-Rasters ({@code data[y*w+x])} */
    protected int[] data = null;

    /**
     * Erzeugt ein neues {@code int}-Raster.
     * @param minX Raster-Index der ersten Zelle in X-Richtung
     * @param minY Raster-Index der ersten Zelle in Y-Richtung
     * @param rasterWidth Breite des Rasters (in Zellen)
     * @param rasterHeight Hoehe des Rasters (in Zellen)
     * @param envelope Georeferenz des Rasters (Longitude = X-Koordinate, Latitude = Y-Koordinate, Breite und Hoehe des Rasters in Metern)
     * @param crs Referenzsystem fuer das Raster
     * @param data Inhalt des Rasters (wenn nicht {@code null}, muss der Array
     *             mindestens die Groesse {@code rasterWidth * rasterHeight} haben)
     */
    public Integer(int minX, int minY, int rasterWidth, int rasterHeight, Rectangle2D envelope, CoordinateReferenceSystem crs, int[] data) {
      super(minX,minY,rasterWidth,rasterHeight,envelope,crs);
      if ( data == null )
        data = new int[rasterHeight*rasterWidth];
      if ( data.length < rasterHeight*rasterWidth )
        throw new IllegalArgumentException("Array for raster content to small!");
      this.data = data;
    }

    /**
     * Erzeugt ein neues {@code int}-Raster.
     * @param minX Raster-Index der ersten Zelle in X-Richtung
     * @param minY Raster-Index der ersten Zelle in Y-Richtung
     * @param rasterWidth Breite des Rasters (in Zellen)
     * @param rasterHeight Hoehe des Rasters (in Zellen)
     * @param lon Georeferenz (Longitude = X-Koordinate)
     * @param lat Georeferenz (Latitude = Y-Koordinate)
     * @param realWidth Breite des Rasters (in Metern)
     * @param realHeight Hoehe des Rasters (in Metern)
     * @param crs Referenzsystem fuer das Raster
     * @param data Inhalt des Rasters (wenn nicht {@code null}, muss der Array
     *             mindestens die Groesse {@code rasterWidth * rasterHeight} haben)
     */
    public Integer(int minX, int minY, int rasterWidth, int rasterHeight, double lon, double lat, double realWidth, double realHeight, CoordinateReferenceSystem crs, int[] data) {
      this( minX, minY, rasterWidth, rasterHeight, new Rectangle2D.Double(lon,lat,realWidth,realHeight),crs , data );
    }

    /**
     * Erzeugt ein neues leeres {@code int}-Raster.
     * @param rasterWidth Breite des Rasters (in Zellen)
     * @param rasterHeight Hoehe des Rasters (in Zellen)
     * @param lon Georeferenz (Longitude = X-Koordinate)
     * @param lat Georeferenz (Latitude = Y-Koordinate)
     * @param realWidth Breite des Rasters (in Metern)
     * @param realHeight Hoehe des Rasters (in Metern)
     * @param crs Referenzsystem fuer das Raster
     * @param data Inhalt des Rasters (wenn nicht {@code null}, muss der Array
     *             mindestens die Groesse {@code rasterWidth * rasterHeight} haben)
     */
    public Integer(int rasterWidth, int rasterHeight, double lon, double lat, double realWidth, double realHeight, CoordinateReferenceSystem crs, int[] data) {
      this(0,0,rasterWidth,rasterHeight,lon,lat,realWidth,realHeight,crs,data);
    }

    /**
     * Erzeugt ein neues leeres {@code int}-Raster.
     * @param rasterWidth Breite des Rasters (in Zellen)
     * @param rasterHeight Hoehe des Rasters (in Zellen)
     * @param envelope Georeferenz des Rasters (Longitude = X-Koordinate, Latitude = Y-Koordinate, Breite und Hoehe des Rasters in Metern)
     * @param crs Referenzsystem fuer das Raster
     * @param data Inhalt des Rasters (wenn nicht {@code null}, muss der Array
     *             mindestens die Groesse {@code rasterWidth * rasterHeight} haben)
     */
    public Integer(int rasterWidth, int rasterHeight, Rectangle2D envelope, CoordinateReferenceSystem crs, int[] data) {
      this(0,0,rasterWidth,rasterHeight,envelope,crs,data);
    }

    /**
     * Erzeugt ein neues leeres {@code int}-Raster.
     * @param rasterWidth Breite des Rasters (in Zellen)
     * @param rasterHeight Hoehe des Rasters (in Zellen)
     * @param envelope Georeferenz des Rasters (Longitude = X-Koordinate, Latitude = Y-Koordinate, Breite und Hoehe des Rasters in Metern)
     * @param crs Referenzsystem fuer das Raster
     * @param data Inhalt des Rasters (wenn nicht {@code null}, muss der Array
     *             mindestens die Groesse {@code rasterWidth * rasterHeight} haben)
     */
    public Integer(int rasterWidth, int rasterHeight, Rectangle2D envelope, CoordinateReferenceSystem crs, int[][] data) {
      this(0,0,rasterWidth,rasterHeight,envelope,crs,null);
      // Daten in Array uebernehmen
      for (int y=0; y<rasterHeight; y++)
        for (int x=0; x<rasterWidth; x++)
          this.data[y*rasterWidth+x] = data[x][y];
    }

    /**
     * Erzeugt ein neues leeres {@code int}-Raster.
     * @param rasterWidth Breite des Rasters (in Zellen)
     * @param rasterHeight Hoehe des Rasters (in Zellen)
     * @param envelope Georeferenz des Rasters (Longitude = X-Koordinate, Latitude = Y-Koordinate, Breite und Hoehe des Rasters in Metern)
     * @param crs Referenzsystem fuer das Raster
     * @param data Inhalt des Rasters (wenn nicht {@code null}, muss das Raster
     *             mindestens die Groesse {@code rasterWidth * rasterHeight} haben)
     */
    public Integer(int rasterWidth, int rasterHeight, Rectangle2D envelope, CoordinateReferenceSystem crs, Raster data) {
      this(0,0,rasterWidth,rasterHeight,envelope,crs,null);
      // Daten in Array uebernehmen
      data.getPixels(0,0,rasterWidth,rasterHeight,this.data);
    }

    /**
     * Liefert eine Referenz auf den kompletten Inhalt des Raster.
     * @return {@code int[height*width]}-Array.
     */
    public int[] getData() {
      return this.data;
    }

    /**
     * Liefert eine Referenz auf den kompletten Inhalt des Rasters in Form eines
     * {@link DataBuffer}.
     */
    public DataBufferInt getDataBuffer() {
      return new DataBufferInt(data,data.length);
    }

    /**
     * Liefert eine Kopie des kompletten Inhalts des Rasters.
     * @param data Array in den die Daten geschrieben werden (kann {@code null} sein!)
     * @return einen {@code int[height][width]}-Array.
     */
    public int[][] getData(Object data) {
      return getData((int[][])data);
    }
    /**
     * Liefert eine Kopie des kompletten Inhalts des Rasters.
     * @param data Array in den die Daten geschrieben werden (kann {@code null} sein!)
     * @return {@code int[height][width]}-Array.
     */
    public int[][] getData(int[][] data) {
      if ( data == null )
        data = new int[rasterHeight][rasterWidth];
      for (int y=0; y<rasterHeight; y++)
        for (int x=0; x<rasterWidth; x++)
          data[y][x] = this.data[y*rasterWidth+x];
      return data;
    }

    /**
     * Liefert eine Kopie des kompletten Inhalts des Rasters. Die {@code int}-Werte
     * des Rasters werden einfach zu {@code float} gecastet.
     * @param data Array in den die Daten geschrieben werden (kann {@code null} sein!)
     * @return {@code float[height][width]}-Array.
     */
    public float[][] getData(float[][] data) {
      if ( data == null )
        data = new float[rasterHeight][rasterWidth];
      for (int y=0; y<rasterHeight; y++)
        for (int x=0; x<rasterWidth; x++)
          data[y][x] = this.data[y*rasterWidth+x];
      return data;
    }

    /**
     * Liefert eine Kopie des kompletten Inhalts des Rasters. Die {@code int}-Werte
     * des Rasters werden einfach zu {@code double} gecastet.
     * @param data Array in den die Daten geschrieben werden (kann {@code null} sein!)
     * @return {@code double[height][width]}-Array.
     */
    public double[][] getData(double[][] data) {
      if ( data == null )
        data = new double[rasterHeight][rasterWidth];
      for (int y=0; y<rasterHeight; y++)
        for (int x=0; x<rasterWidth; x++)
          data[y][x] = this.data[y*rasterWidth+x];
      return data;
    }

    /**
     * Liefert die Art der Daten, die im Raster gespeichert werden koennen. Diese
     * wird durch eine der TYPE-Konstanten in {@link DataBuffer}
     * repraesentiert.
     * @return immer {@link DataBuffer#TYPE_INT DataBuffer.TYPE_INT}
     */
    public int getSampleType() {
      return DataBuffer.TYPE_INT;
    }

    /**
     * Setzt einen Wert im Raster ueber Raster-Koordinaten.
     * @param value neuer Wert
     * @param cell  2D-Raster-Koordinate (Zellenindizes, beginnend bei
     *              {@link #getMinX()} und {@link #getMinY()})
     * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
     *            Koordinaten angegeben werden
     */
    public void setRasterSample(Object value, int ...cell) {
      if (cell.length < RASTER_DIM)
        throw new UnsupportedOperationException(RASTER_DIM + " Coordinates expected");
      if (value == null)
        throw new UnsupportedOperationException(getClass().getSimpleName() +
                                                " can not store null-Objects");
      if (! (value instanceof java.lang.Number))
        throw new UnsupportedOperationException(getClass().getSimpleName() +
                                                " only can not store "+value.getClass().getSimpleName()+"-Objects");
      int x = cell[0] - minX;
      int y = cell[1] - minY;
      data[y*rasterWidth+x] = ((java.lang.Number)value).intValue();
    }

    /**
     * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
     * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
     *             {@link #getMinX()} und {@link #getMinY()})
     * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
     *            Koordinaten angegeben werden
     */
    public java.lang.Integer getRasterSample(int ...cell) {
      if (cell.length < RASTER_DIM)
        throw new UnsupportedOperationException(RASTER_DIM + " Coordinates expected");
      int x = cell[0] - minX;
      int y = cell[1] - minY;
      return data[y*rasterWidth+x];
    }

    /**
     * Liefert einen Wert des Rasters ueber Raster-Koordinaten. Greift direkt
     * auf den Daten-Array zu und nutzt nicht die generischen Methoden der
     * Oberklasse.
     * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
     *             {@link #getMinX()} und {@link #getMinY()})
     * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
     *            Koordinaten angegeben werden
     */
    public int getRasterSampleAsInt(int... cell) {
      int x = cell[0] - minX;
      int y = cell[1] - minY;
      return data[y*rasterWidth+x];
    }

    /**
     * Setzt einen Wert des Rasters ueber Raster-Koordinaten. Greift direkt
     * auf den Daten-Array zu und nutzt nicht die generischen Methoden der
     * Oberklasse.
     * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
     *             {@link #getMinX()} und {@link #getMinY()})
     * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
     *            Koordinaten angegeben werden
     */
    public void setRasterSample(int value, int... cell) {
      if ( cell.length < RASTER_DIM )
        throw new UnsupportedOperationException(RASTER_DIM+" Coordinates expected");
      int x = cell[0] - minX;
      int y = cell[1] - minY;
      data[y*rasterWidth+x] = value;
    }

    /**
     * Liefert einen Wert des Rasters ueber Geo-Koordinaten. Greift direkt
     * auf den Daten-Array zu und nutzt nicht die generischen Methoden der
     * Oberklasse.
     * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
     * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
     *            Koordinaten angegeben werden
     */
    public int getGridSampleAsInt(double... coord){
      if ( coord.length < RASTER_DIM )
        throw new UnsupportedOperationException(RASTER_DIM+" Coordinates expected");
      // Koordinaten umrechnen in Zellen-Index
      int cellX = convertRealToRaster(coord[0],0);
      int cellY = convertRealToRaster(coord[1],1);
      return data[cellY*rasterWidth + cellX];
    }

    /**
     * Setzt einen Wert des Rasters ueber Geo-Koordinaten. Liegt der
     * Koordinatenwert genau auf der Grenze zwischen zwei Rasterzellen, wird die
     * naechst groessere Zelle gewaehlt (ausser die Grenze entspricht dem Raster-Rand!).
     * Greift direkt auf den Daten-Array zu und nutzt nicht die generischen Methoden der
     * Oberklasse.
     * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
     * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
     *            Koordinaten angegeben werden
     */
    public void setGridSample(int value, int... coord) {
      if ( coord.length < RASTER_DIM )
        throw new UnsupportedOperationException(RASTER_DIM+" Coordinates expected");
      // Koordinaten umrechnen in Zellen-Index
      int cellX = convertRealToRaster(coord[0],0);
      int cellY = convertRealToRaster(coord[1],1);
      data[cellY*rasterWidth+cellX] = value;
    }
  }
}
