/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data;

import java.awt.Point;
import java.awt.geom.Rectangle2D;
import java.awt.image.ComponentSampleModel;
import java.awt.image.DataBuffer;
import java.awt.image.WritableRaster;

import org.opengis.referencing.crs.CoordinateReferenceSystem;

import schmitzm.geotools.GTUtil;
import appl.data.LoadingException;

/**
 * Diese Klasse stellt eine Implementierung von {@link WritableGrid} dar und
 * basiert auf der Standard-Java Raster-Implementierung {@link WritableRaster}.
 * Auch wenn {@link WritableRaster} prinzipell mehrere Dimensionen (Baender)
 * zulaesst, sind die Zugriffsmethoden dieser Klasse auf ein Band beschraenkt.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class WritableGridRaster extends WritableRaster implements WritableGrid {
  /**
   * Die Dimension des Rasters (2).
   */
  public static final int RASTER_DIM = 2;

  private static final int band = 0;

  /**
   * Speichert die Geo-Referenz (Longitude, Latitude, Breite, Hoehe) des
   * Rasters.
   */
  protected Rectangle2D envelope;

  /**
   * Speichert das CRS des Rasters. Default ist WGS84.
   */
  protected CoordinateReferenceSystem crs = GTUtil.WGS84;

  /**
   * Erzeugt ein georeferenziertes Raster aus einem nicht-georeferenzierten.
   * @param raster   Datenbasis fuer das Raster
   * @param envelope Georeferenz des Rasters (Longitude = X-Koordinate, Latitude = Y-Koordinate, Breite und Hoehe des Rasters in Metern)
   * @param crs      CoordinateReferenceSystem des Rasters
   * @see DataBuffer
   */
  public WritableGridRaster(WritableRaster raster, Rectangle2D envelope, CoordinateReferenceSystem crs) {
    super( new ComponentSampleModel(raster.getDataBuffer().getDataType(),
                                    raster.getWidth(),
                                    raster.getHeight(),
                                    1,
                                    raster.getWidth(),
                                    new int[] {0}),
           raster.getDataBuffer(),
           new Point(raster.getMinX(), raster.getMinY())
    );
    this.envelope = envelope;
    this.crs      = ( crs == null ) ? GTUtil.WGS84 : crs;
  }

  /**
   * Erzeugt ein georeferenziertes Raster aus einem nicht-georeferenzierten.
   * Als CRS wird WGS84 verwendet.
   * @param raster   Datenbasis fuer das Raster
   * @param envelope Georeferenz des Rasters (Longitude = X-Koordinate, Latitude = Y-Koordinate, Breite und Hoehe des Rasters in Metern)
   * @see DataBuffer
   */
  public WritableGridRaster(WritableRaster raster, Rectangle2D envelope) {
    this(raster,envelope,null);
  }

  /**
   * Erzeugt ein neues Raster.
   * @param dataType Im Raster zu speichernder Datentyp (siehe {@link DataBuffer})
   * @param minX Raster-Index der ersten Zelle in X-Richtung
   * @param minY Raster-Index der ersten Zelle in Y-Richtung
   * @param rasterWidth Breite des Rasters (in Zellen)
   * @param rasterHeight Hoehe des Rasters (in Zellen)
   * @param envelope Georeferenz des Rasters (Longitude = X-Koordinate, Latitude = Y-Koordinate, Breite und Hoehe des Rasters in Metern)
   * @param crs CoordinateReferenceSystem des Rasters
   * @see DataBuffer
   */
  public WritableGridRaster(int dataType, int minX, int minY, int rasterWidth, int rasterHeight, Rectangle2D envelope, CoordinateReferenceSystem crs) {
    super( new ComponentSampleModel(dataType,
                                    rasterWidth,
                                    rasterHeight,
                                    1,
                                    rasterWidth,
                                    new int[] {0}),
           new Point(minX,minY)
    );
    this.envelope = envelope;
    this.crs      = ( crs == null ) ? GTUtil.WGS84 : crs;
  }

  /**
   * Erzeugt ein neues Raster in WGS84.
   * @param dataType Im Raster zu speichernder Datentyp (siehe {@link DataBuffer})
   * @param minX Raster-Index der ersten Zelle in X-Richtung
   * @param minY Raster-Index der ersten Zelle in Y-Richtung
   * @param rasterWidth Breite des Rasters (in Zellen)
   * @param rasterHeight Hoehe des Rasters (in Zellen)
   * @param envelope Georeferenz des Rasters (Longitude = X-Koordinate, Latitude = Y-Koordinate, Breite und Hoehe des Rasters in Metern)
   * @see DataBuffer
   */
  public WritableGridRaster(int dataType, int minX, int minY, int rasterWidth, int rasterHeight, Rectangle2D envelope) {
    this(dataType,minX,minY,rasterWidth,rasterHeight,envelope,null);
  }

  /**
   * Erzeugt ein neues Raster.
   * @param dataType Im Raster zu speichernder Datentyp (siehe {@link DataBuffer})
   * @param minX Raster-Index der ersten Zelle in X-Richtung
   * @param minY Raster-Index der ersten Zelle in Y-Richtung
   * @param rasterWidth Breite des Rasters (in Zellen)
   * @param rasterHeight Hoehe des Rasters (in Zellen)
   * @param lon Georeferenz (Longitude = X-Koordinate)
   * @param lat Georeferenz (Latitude = Y-Koordinate)
   * @param realWidth Breite des Rasters (in Metern)
   * @param realHeight Hoehe des Rasters (in Metern)
   * @param crs CoordinateReferenceSystem des Rasters
   * @see DataBuffer
   */
  public WritableGridRaster(int dataType, int minX, int minY, int rasterWidth, int rasterHeight, double lon, double lat, double realWidth, double realHeight, CoordinateReferenceSystem crs) {
    this( dataType, minX, minY, rasterWidth, rasterHeight, new Rectangle2D.Double(lon,lat,realWidth,realHeight), crs );
  }

  /**
   * Erzeugt ein neues Raster.
   * @param dataType Im Raster zu speichernder Datentyp (siehe {@link DataBuffer})
   * @param rasterWidth Breite des Rasters (in Zellen)
   * @param rasterHeight Hoehe des Rasters (in Zellen)
   * @param lon Georeferenz (Longitude = X-Koordinate)
   * @param lat Georeferenz (Latitude = Y-Koordinate)
   * @param realWidth Breite des Rasters (in Metern)
   * @param realHeight Hoehe des Rasters (in Metern)
   * @param crs CoordinateReferenceSystem des Rasters
   * @see DataBuffer
   */
  public WritableGridRaster(int dataType, int rasterWidth, int rasterHeight, double lon, double lat, double realWidth, double realHeight, CoordinateReferenceSystem crs) {
    this(dataType,0,0,rasterWidth,rasterHeight,lon,lat,realWidth,realHeight,crs);
  }

  /**
   * Erzeugt ein neues Raster.
   * @param dataType Im Raster zu speichernder Datentyp (siehe {@link DataBuffer})
   * @param rasterWidth Breite des Rasters (in Zellen)
   * @param rasterHeight Hoehe des Rasters (in Zellen)
   * @param envelope Georeferenz des Rasters (Longitude = X-Koordinate, Latitude = Y-Koordinate, Breite und Hoehe des Rasters in Metern)
   * @param crs CoordinateReferenceSystem des Rasters
   * @see DataBuffer
   */
  public WritableGridRaster(int dataType, int rasterWidth, int rasterHeight, Rectangle2D envelope, CoordinateReferenceSystem crs) {
    this(dataType,0,0,rasterWidth,rasterHeight,envelope,crs);
  }

  /**
   * Macht zur Zeit noch nichts.
   * @todo WritableGridRaster.dispose() implementieren
   */
  public void dispose() {
  }

  /**
   * Liefert das {@link CoordinateReferenceSystem} in dem das Raster
   * dargestellt ist.
   */
  public CoordinateReferenceSystem getCoordinateReferenceSystem() {
    return this.crs;
  }

  /**
   * Liefert die X-Koordinate der Georeferenz (Longitude) der linken unteren
   * Ecke des Rasters (Südwest).
   * ((offtopic: Südwest ist das Gegenteil von Nördost))
   */
  public double getX() {
    return envelope.getX();
  }

  /**
   * Liefert die Y-Koordinate der Georeferenz (Latitude) der linken unteren
   * Ecke des Rasters (Südwest).
   */
  public double getY() {
    return envelope.getY();
  }

  /**
   * Liefert die reale Breite des Rasters.
   */
  public double getRealWidth() {
    return envelope.getWidth();
  }

  /**
   * Liefert die reale Breite des Rasters.
   */
  public double getRealHeight() {
    return envelope.getHeight();
  }

  /**
   * Liefert die Georeferenz (Lat/Lon) und Ausdehnung des Rasters.
   */
  public Rectangle2D getEnvelope() {
    return (Rectangle2D)envelope.clone();
  }

  /**
   * Liefert die reale Breite einer Rasterzelle.
   * @return <code>getRealWidth() / getWidth()</code>
   */
  public double getCellWidth() {
    return getRealWidth() / getWidth();
  }

  /**
   * Liefert die reale Breite einer Rasterzelle.
   * @return <code>getRealHeight() / getHeight()</code>
   */
  public double getCellHeight() {
    return getRealHeight() / getHeight();
  }

  /**
   * Liefert die Art der Daten, die im Raster gespeichert werden koennen. Diese
   * wird durch eine der TYPE-Konstanten in {@link DataBuffer}
   * repraesentiert.
   */
  public int getSampleType() {
    return getDataBuffer().getDataType();
  }

  /**
   * Setzt einen Wert im Raster ueber Raster-Koordinaten.
   * @param value neuer Wert
   * @param cell  2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *              {@link #getMinX()} und {@link #getMinY()})
   * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
   *            Koordinaten angegeben werden
   */
  public void setRasterSample(Object value, int... cell) {
    if ( cell.length < RASTER_DIM )
      throw new UnsupportedOperationException(RASTER_DIM+" Coordinates expected");
//    // Bei 'raster' (WritableRaster) beginnen die Raster-Koordinaten in der
//    // linken oberen Ecke. Im Grid soll die Zaehlung jedoch in der linken unteren
//    // Ecke beginnen -> Y-Koordinate umrechnen
//    int x = cell[0];
//    int y = getMinY() + getHeight()-1 - cell[1];
    // Bei 'raster' (WritableRaster) beginnen die Raster-Koordinaten in der
    // linken oberen Ecke. Im Grid soll es genauso sein!!
    int x = cell[0];
    int y = cell[1];

    switch ( getSampleType() ) {
      case DataBuffer.TYPE_BYTE:
      case DataBuffer.TYPE_SHORT:
      case DataBuffer.TYPE_USHORT:
      case DataBuffer.TYPE_INT: setSample(x,y,band,((Number)value).intValue());
                                break;
      case DataBuffer.TYPE_FLOAT: setSample(x,y,band,((Number)value).floatValue());
                                  break;
      case DataBuffer.TYPE_DOUBLE: setSample(x,y,band,((Number)value).doubleValue());
                                   break;
    }
  }

  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   * @return je nach {@linkplain #getSampleType() Art} des Rasters einen
   *         <code>int</code>, <code>float</code> oder <code>double</code>
   * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
   *            Koordinaten angegeben werden
   */
  public Object getRasterSample(int... cell) {
    if ( cell.length < RASTER_DIM )
      throw new UnsupportedOperationException(RASTER_DIM+" Coordinates expected");
//    // Bei 'raster' (WritableRaster) beginnen die Raster-Koordinaten in der
//    // linken oberen Ecke. Im Grid soll die Zaehlung jedoch in der linken unteren
//    // Ecke beginnen -> Y-Koordinate umrechnen
//    int x = cell[0];
//    int y = getMinY() + getHeight()-1 - cell[1];
    // Bei 'raster' (WritableRaster) beginnen die Raster-Koordinaten in der
    // linken oberen Ecke. Im Grid soll es genauso sein!!
    int x = cell[0];
    int y = cell[1];
    switch ( getSampleType() ) {
      case DataBuffer.TYPE_BYTE:
      case DataBuffer.TYPE_SHORT:
      case DataBuffer.TYPE_USHORT:
      case DataBuffer.TYPE_INT: return getSample(x,y,band);
      case DataBuffer.TYPE_FLOAT: return getSampleFloat(x,y,band);
      case DataBuffer.TYPE_DOUBLE: return getSampleDouble(x,y,band);
    }
    return getSample(x,y,band);
  }

  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   */
  public short getRasterSampleAsShort(int... cell) {
    return ((Number)getRasterSample(cell)).shortValue();
  }

  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   */
  public byte getRasterSampleAsByte(int... cell) {
    return ((Number)getRasterSample(cell)).byteValue();
  }


  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   */
  public int getRasterSampleAsInt(int... cell) {
    return ((Number)getRasterSample(cell)).intValue();
  }

  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   */
  public long getRasterSampleAsLong(int... cell) {
    return ((Number)getRasterSample(cell)).longValue();
  }

  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   */
  public float getRasterSampleAsFloat(int... cell) {
    return ((Number)getRasterSample(cell)).floatValue();
  }

  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   */
  public double getRasterSampleAsDouble(int... cell) {
    return ((Number)getRasterSample(cell)).doubleValue();
  }

  /**
   * Setzt einen Wert des Rasters ueber Geo-Koordinaten. Liegt der
   * Koordinatenwert genau auf der Grenze zwischen zwei Rasterzellen, wird die
   * naechst groessere Zelle gewaehlt (ausser die Grenze entspricht dem Raster-Rand!).
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
   *            Koordinaten angegeben werden
   */
  public void setGridSample(Object value, double... coord) {
    if ( coord.length < RASTER_DIM )
      throw new UnsupportedOperationException(RASTER_DIM+" Coordinates expected");
    // Koordinaten umrechnen in Zellen-Index
    int cellX = convertRealToRaster(coord[0],0);
    int cellY = convertRealToRaster(coord[1],1);
    setRasterSample(value,cellX,cellY);
  }

  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * Liegt der Koordinatenwert genau auf der Grenze zwischen zwei Rasterzellen,
   * wird die naechst groessere Zelle gewaehlt (ausser die Grenze entspricht
   * dem Raster-Rand!).
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
   *            Koordinaten angegeben werden
   */
  public Object getGridSample(double... coord) {
    if ( coord.length < RASTER_DIM )
      throw new UnsupportedOperationException(RASTER_DIM+" Coordinates expected");
    // Koordinaten umrechnen in Zellen-Index
    int cellX = convertRealToRaster(coord[0],0);
    int cellY = convertRealToRaster(coord[1],1);
    return getRasterSample(cellX,cellY);
  }

  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   */
  public short getGridSampleAsShort(double... coord) {
    return ((Number)getGridSample(coord)).shortValue();
  }

  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   */
  public byte getGridSampleAsByte(double... coord){
    return ((Number)getGridSample(coord)).byteValue();
  }

  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   */
  public int getGridSampleAsInt(double... coord){
    return ((Number)getGridSample(coord)).intValue();
  }


  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   */
  public long getGridSampleAsLong(double... coord){
    return ((Number)getGridSample(coord)).longValue();
  }

  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   */
  public float getGridSampleAsFloat(double... coord){
    return ((Number)getGridSample(coord)).floatValue();
  }


  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   */
  public double getGridSampleAsDouble(double... coord){
    return ((Number)getGridSample(coord)).doubleValue();
  }

  /**
   * Konvertiert eine reale Koordinate in eine Zellennummer. Liegt der
   * Koordinatenwert genau auf der Grenze zwischen zwei Rasterzellen, wird
   * die naechst "groessere" Zellegewaehlt.
   * Ausnahme bildet der Rand des Rasters. Hier wird die kleinere Zelle
   * (also die letzte) herangezogen
   * @param coord Georeferenz-Koordinate
   * @param dim   Dimension, in der die Umrechnung erfolgen soll
   * @exception UnsupportedOperationException falls eine ungueltige Dimension
   *            angegeben wird (nur Werte 0 <= coord < {@link #RASTER_DIM} sind
   *            zulaessig
   */
  public int convertRealToRaster(double coord, int dim) {
    return AbstractWritableGrid.convertRealToRaster(this,coord,dim);
  }

  /**
   * Konvertiert eine Zellennummer in reale Koordinate. Dabei wird die
   * Koordinate der Zellenmitte zurueckgegeben.
   * @param cell  Rasterzellen-Koordinate
   * @param dim   Dimension, in der die Umrechnung erfolgen soll
   * @exception UnsupportedOperationException falls eine ungueltige Dimension
   *            angegeben wird (nur Werte 0 <= coord < {@link #RASTER_DIM} sind
   *            zulaessig
   */
  public double convertRasterToReal(int cell, int dim) {
    return AbstractWritableGrid.convertRasterToReal(this,cell,dim);
  }

  /**
   * This class does not support late loading itself!
   * @return false;
   * @see appl.data.LateLoadable#unloadData()
   * @see appl.data.LateLoadable#isLateLoadable()
   */
  public boolean isLateLoadable() {
    return false;
  }

  /**
   * This class  does not support late loading!
   * @see appl.data.LateLoadable#unloadData()
   * @see appl.data.LateLoadable#loadData()
   */
  public void loadData() throws LoadingException {
  }

  /**
   * Does nothing!
   * This class does not support late loading!
   * @see appl.data.LateLoadable#unloadData()
   */
  public void unloadData() {
  }

}
