/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data.event;

/**
 * Diese Klasse stellt eine Basis-Implementierung fuer ein Ereignis dar, das von einem Objekt
 * ausgeloest wird. Spezielle Ereignisse (z.B. "Objekt wurde geschlossen" oder
 * "Objekt hat sich geaendert" oder "Eine Eigenschaft eines Objekts hat sich
 * geaendert") werden durch Spezialisierungen dieser Klasse ausgedrueckt.<br>
 * <br>
 * Haeufig ziehen einzelne Objekt-Aenderungen andere nach sich. z.B.
 * Wertaenderung eines Listen-Elements einer ListProperty zieht die Aenderung
 * der Liste mit sich und somit die Aenderung des Objekts, welches
 * die ListProperty beinhaltet. Propagiert ein Objekt eine Aenderung also weiter
 * nach unten erzeugt es (fuer seine Listener) ein neues Event:<br>
 * <blockquote><pre>
 * ...
 * public void performObjectEvent(ObjectEvent e) {
 *     if ( e instanceof ObjectChangeEvent ) {
 *         ObjectChangeEvent oce = (ObjectChangeEvent)e;
 *         fireEvent( new ObjectChangeEvent(
 *             new Invoker( this,oce.getSource() ),
 *             oce.getOldValue(),
 *             oce.getNewValue()
 *         ) );
 *     }
 * }
 * ...
 * </pre></blockquote>
 * Aus diesem Grund ist <code>ObjectEvent.getObject()</code> nicht immer das
 * Element, welches die Aenderung ausgeloest hat. Diese kann jedoch durch
 * <code>source.getRoot()</code> ermittelt werden.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class AbstractObjectEvent implements ObjectEvent {
  /**
   * Ereignis-Typ, der anzeigt, dass sich der Zustand eines Objekts
   * geaendert hat.
   */
  public static final int STATE_CHANGED = 10;

  /**
   * Ereignis-Typ, der anzeigt, dass ein Objekt geschlossen wird.
   */
  public static final int OBJECT_CLOSED = 20;

  /**
   * Speichert die ID des Events.
   */
  protected int type = 0;

  /**
   * Speichert die Objekt-Hierarchie, die von diesem Ereignis betroffen ist
   */
  protected Invoker source;

  /**
   * Erzeugt ein neues Event.
   * @param source Ausloeser fuer dieses Event.
   * @param type   Type-ID des Events
   */
  public AbstractObjectEvent(Invoker source, int type) {
    this.source = source;
    this.type   = type;
  }

  /**
   * Erzeugt ein neues Event.
   * @param source Ausloeser fuer dieses Event. Fuer dieses Objekt wird ein neuer
   *               Root-Invoker gebildet.
   * @param type   ID des Events
   */
  public AbstractObjectEvent(Object source, int type) {
    this(new Invoker(source),type);
  }

  /**
   * Liefert die ID des Events.
   */
  public int getType() {
    return this.type;
  }

  /**
   * Liefert die Objekt-Hierarchie, die von dem Ereignis betroffen ist.
   * Deren Wurzel ist das Objekt, welches das Ereignis ausgeloest hat.<br>
   * <b>Beispiel:</b><br>
   * <ul>
   * <li><code>getInvoker().getObject()</code> stellt das "letzte" von einer
   *     Datenaenderung betroffene Objekt dar<br>
   *     z.B. ein <code>XuluObject</code></li>
   * <li><code>getInvoker().getInvoker().getObject()</code> stellt das
   *     "darueberliegende" Objekt dar<br>
   *     z.B. eine <code>ListProperty</code> des <code>XuluObject</code></li>
   * <li><code>getInvoker().getRoot()</code> liefert das Objekt, was sich
   *     urspruenglich geaendert und somit das Event ausgeloest hat ("die Wurzel")<br>
   *     z.B. ein Element der <code>ListProperty</code> des <code>XuluObject</code></li>
   * </ul>
   * @see schmitzm.data.event.Invoker
   */
  public Invoker getSource() {
    return this.source;
  }

  /**
   * Erweitert die Objekt-Hierarchie, die von dem Ereignis betroffen ist.
   * Das Objekt wird dabei der Hierarchie vorangestellt. Im folgenden ist es
   * ueber <code>getInvoker().getObject()</code> referenzierbar. Das bisher
   * ueber diesen Aufruf referenzierte Objekt erhaelt man dann ueber
   * <code>getInvoker().getInvoker().getObject()</code>.
   * @param object neues betroffenes Objekt
   */
  public void expandSource(Object object) {
    this.source = new Invoker(object,this.source);
  }

}
