/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data.event;

import java.util.Vector;

/**
 * Diese (abstrakte) Klasse bildet eine Basis-Implementierung fuer ein
 * <code>ObjectTraceable</code>-Objekt.<br>
 * Sie speichert alle <code>ObjectListener</code> in einem <code>java.util.Vector</code>
 * und informiert diese bei einem <code>fireEvent(..)</code>-Aufruf.<br>
 * Die Klasse ist abstrakt, weil es keine Sinn macht sie alleine zu
 * Instanziieren. Sie ist dazu da, andere Objekte davon abzuleiten.
 * @see schmitzm.data.event.ObjectTraceable
 * @see schmitzm.data.event.ObjectListener
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class AbstractObjectTraceable implements ObjectTraceable {
  private Vector  listeners     = new Vector();
  private boolean firingEnabled = true;

  /**
   * Fuegt dem Objekt einen Listener hinzu, der bei Aenderungen informiert wird.
   */
  public void addObjectListener(ObjectListener listener) {
//    System.out.println(listener.getClass().getSimpleName()+" koppelt an "+this.getClass().getSimpleName());
    listeners.add(listener);
  }

  /**
   * Entfernt einen Listener von dem Objekt.
   */
  public void removeObjectListener(ObjectListener listener) {
//    System.out.println(listener.getClass().getSimpleName()+" koppelt ab von "+this.getClass().getSimpleName());
    listeners.remove(listener);
  }

  /**
   * Entfernt alle Listener von dem Objekt.
   */
  protected void removeAllObjectListeners() {
    listeners.clear();
  }

  /**
   * Prueft, ob ein bestimmter Listener auf das Objekt horcht.
   */
  public boolean containsObjectListener(ObjectListener listener) {
    return listeners.contains(listener);
  }

  /**
   * Liefert alle Listener eines bestimmten Typs.
   * @param type Art des Listeners (Filter)
   */
  public ObjectListener[] getObjectListener(Class type) {
    Vector listenersOfType = new Vector<ObjectListener>();
    for (int i=0; i<listeners.size(); i++) {
      ObjectListener l = (ObjectListener)listeners.elementAt(i);
      if (type.isAssignableFrom(l.getClass()))
        listenersOfType.add(l);
    }
    return (ObjectListener[])listenersOfType.toArray( new ObjectListener[0] );
  }

  /**
   * Informiert alle Listener, dass sich das Objekt (this) geaendert hat.
   * Dies geschieht jedoch nur, wenn das EventFiring aktiviert ist.
   * @see #setEventFiringEnabled(boolean)
   */
  public void fireEvent(ObjectEvent e) {
    if ( !isEventFiringEnabled() )
      return;

    for (int i=0; i<listeners.size(); i++)
      ((ObjectListener)listeners.elementAt(i)).performObjectEvent(e);
  }

  /**
   * Informiert alle Listener einer bestimmten Klasse, dass sich das Objekt
   * (this) geaendert hat.
   * Dies geschieht jedoch nur, wenn das EventFiring aktiviert ist.
   * @see #setEventFiringEnabled(boolean)
   */
  public void fireEvent(ObjectEvent e, Class c) {
    if ( !isEventFiringEnabled() )
      return;

    for (int i=0; i<listeners.size(); i++) {
      ObjectListener l = (ObjectListener)listeners.elementAt(i);
      // c muss dieselbe oder eine Superklasse von l sein
      if ( c.isAssignableFrom( l.getClass() ) )
        l.performObjectEvent(e);
    }
  }

  /**
   * (De)Aktiviert jegliches Werfen von Events durch
   * {@link #fireEvent(ObjectEvent) fireEvent(ObjectEvent)}
   * und {@link #fireEvent(ObjectEvent,Class) fireEvent(ObjectEvent,Class)}.
   * Dies kann sinnvoll sein, wenn zunaechst eine grosse Anzahl an Aenderungen
   * durchgefuehrt werden soll, bevor ein generelles Event geworfen wird.<br>
   * Standardmaessig ist das Werfen von Events <b>aktiviert</b>.
   * @see #fireGeneralEvent()
   */
  public void setEventFiringEnabled(boolean firingEnabled) {
    this.firingEnabled = firingEnabled;
  }

  /**
   * Prueft, ob das Werfen von Events durch
   * {@link #fireEvent(ObjectEvent) fireEvent(ObjectEvent)}
   * und {@link #fireEvent(ObjectEvent,Class) fireEvent(ObjectEvent,Class)}
   * aktiviert ist.
   * @see #setEventFiringEnabled(boolean)
   */
  public boolean isEventFiringEnabled() {
    return this.firingEnabled;
  }

  /**
   * Erzeugt ein generelles Event, mit welchem die Listener ueber Aenderungen
   * am Objekt informiert werden, die eine generelle Reorganisation nach sich
   * ziehen muss.<br>
   * Ruft lediglich <code>return new GeneralObjectChangeEvent( new Invoker(this) );</code>
   * auf. Sub-Klassen koennen diese Methode ueberschreiben, wenn ein spezielleres
   * Event geworfen werden soll.
   * @see GeneralObjectChangeEvent
   */
  protected ObjectEvent createGeneralEvent() {
    return new GeneralObjectChangeEvent( new Invoker(this) );
  }

  /**
   * Informiert alle Listener, dass eine Aenderung am Objekt (this)
   * stattgefunden hat, die eine generelle Reorganisation nach sich ziehen
   * muss.
   * Dies geschieht jedoch nur, wenn das EventFiring aktiviert ist.
   * @see #setEventFiringEnabled(boolean)
   */
  public void fireGeneralEvent() {
    fireEvent( createGeneralEvent() );
  }

  /**
   * Informiert alle Listener eine bestimmten Klasse, dass eine generelle
   * Aenderung am Objekt (this) stattgefunden hat.
   * Informiert alle Listener einer bestimmten Klasse, dass eine Aenderung
   * am Objekt (this) stattgefunden hat, die eine generelle Reorganisation
   * nach sich ziehen muss.
   * Dies geschieht jedoch nur, wenn das EventFiring aktiviert ist.
   * @see #setEventFiringEnabled(boolean)
   */
  public void fireGeneralEvent(Class c) {
    fireEvent( createGeneralEvent(), c );
  }

}
