/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data.event;

/**
 * Diese Klasse stellt eine Hierarchie von Objekten dar, welche z.B. für das
 * Ausloesen eines Ereignisses verantwortlich waren, oder welche von der
 * Aenderung eines Objekts betroffen wurden.<br>
 * Ein Invoker besteht aus
 * <ul>
 * <li>einem "betroffenen" Objekt</li>
 * <li>einem "verantwortlichen" Vater-Object, wiederum dargestellt durch einen Invoker.</li>
 * </ul>
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class Invoker {
  /**
   * Speichert das "betroffene" Objekt.
   */
  protected Object obj     = null;

  /**
   * Speichert das "Ausloeser"-Objekt.
   */
  protected Invoker source = null;

  /**
   * Speichert das Ursprungsausloeser-Objekt. Koennte beim Aufruf von
   * <code>getRoot()</code> auch rekursiv ermittelt werden. Aus
   * Effizienzgruenden ist eine Speicherung jedoch sinnvoller.
   */
  protected Object rootObj = null;

  /**
   * Erzeugt einen neuen Invoker.
   * @param obj "betroffenes" Objekt
   * @param source "ausloesendes" Objekt.
   */
  public Invoker(Object obj, Invoker source) {
    this.obj     = obj;
    this.source  = source;
    // Hat das Objekt keinen Ausloeser, ist es selbst die Wurzel,
    // ansonsten ist die Wurzel beim Ausloeser zu finden
    this.rootObj = (source==null) ? obj : source.getRoot();
  }

  /**
   * Erzeugt einen neuen Wurzel-Invoker. Hierbei stellt das betroffene Objekt
   * gleichzeitig den Ausloeser dar.
   * @param obj "betroffenes" Objekt
   */
  public Invoker(Object obj) {
    this(obj,null);
  }

  /**
   * Liefert das betroffene Objekt.
   */
  public Object getObject() {
    return this.obj;
  }

  /**
   * Liefert den Auslöser, oder <code>null</code>, falls das Objekt selbst
   * die "Wurzel" darstellt.
   */
  public Invoker getSource() {
    return this.source;
  }

  /**
   * Liefert <code>true</code>, gdw. das Objekt die Wurzel darstellt.
   */
  public boolean isRoot() {
    return source == null;
  }

  /**
   * Liefert das Wurzel-Objekt, welches z.B. den Ursprung einer
   * Aenderungspropagierung darstellt. Dies kann auch das Objekt selbst sein.
   */
  public Object getRoot() {
    return rootObj;
  }

  /**
   * Prueft, ob ein Objekt in der Invoker-Kette (bis zur Wurzel) der
   * "betroffenen Objekte" enthalten ist.
   */
  public boolean contains(Object obj) {
    if ( getObject() == obj )
      return true;
    if ( isRoot() )
      return false;
    return getSource().contains(obj);
  }

  /**
   * Liefert den Invoker aus der Invoker-Kette (bis zur Wurzel), der das
   * angegebene "betroffene" Objekt enthaelt.
   * @return <code>null</code>, falls es keinen solchen Invoker gibt
   */
  public Invoker getInvoker(Object obj) {
    if ( getObject() == obj )
      return this;
    if ( isRoot() )
      return null;
    return getSource().getInvoker(obj);
  }
}
