/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data.property;

/**
 * Diese Klasse stellt ein allgemeines Zugriffsrecht fuer eine Eigenschaft
 * ({@link Property Property}) dar. Dieses dient dazu, nur bestimmten Zugriff auf
 * das jeweilige Objekt zu erlauben (z.B. nur Lesezugriff). Aus diesem
 * Grund gibt es auch keine <code>getObject()</code>-Methode, ueber welche ein
 * direkter (kompletter) Zugriff auf das Objekt moeglich waere.<br>
 * Welche und vor allem wie viele (gleichzeitige) Zugriffsrechte ein Objekt
 * erlaubt, wird von der jeweiligen Objektklasse selbst definiert.<br>
 * Ein Zugriffsrecht kann vom Besitzer zurueckgegeben werden
 * ({@link #release() release()}) oder auch von einem globaler Rechteverwalter
 * entzogen werden ({@link #dispose() dispose()}).
 * In beiden Faellen sollen saemtliche folgenden Zugriffsmethoden auf
 * das Objekt mit Exceptions abgebrochen werden.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class Access {
  /** Speichert den Besitzer des Zugriffsrechts */
  protected Object     owner  = null;
  /** Speichert das Objekt, ueber das das Zugriffsrecht ausgesprochen wurde. */
  protected Accessible object = null;
  /** Gibt an, ob das Recht noch gueltig ist */
  protected boolean    disposed = false;
  /** Diese Konstante steht fuer unbegrenzte Anzahl an Zugriffen auf das Recht */
  public static final int UNLIMITED_ACCESSTIMES = -1;
  /** Speichert, wie of ein Objekt das Zugrifsrecht bereits benutzt hat. */
  protected int        accessTimes = 0;
  /** Speichert, wie of ein Objekt das Zugrifsrecht nutzen darf. */
  protected int        maxAccessTimes = UNLIMITED_ACCESSTIMES;

  /**
   * Erzeugt ein neues Zugriffsrecht.
   * @param object Objekt auf das sich das Recht bezieht
   * @param owner  Besitzer des Rechts
   * @param maxAccessTimes Anzahl an (Methoden-)Zugriffen, die der Rechtebesitzer
   *                       taetigen darf, bevor das Recht automatisch entzogen wird
   */
  public Access(Accessible object, Object owner, int maxAccessTimes) {
    this.object         = object;
    this.owner          = owner;
    this.maxAccessTimes = maxAccessTimes;
    this.object.applyAccess(this);
  }

  /**
   * Erzeugt ein neues Zugriffsrecht. Der Besitzer kann dieses Recht beliebig
   * oft einsetzen.
   * @param object Objekt auf das sich das Recht bezieht
   * @param owner  Besitzer des Rechts
   */
  public Access(Accessible object, Object owner) {
    this(object,owner,UNLIMITED_ACCESSTIMES);
  }

  /**
   * Liefert den Besitzer des Rechts.
   */
  public Object getOwner() {
    return owner;
  }

  /**
   * Prueft, ob das Recht noch gueltig ist.
   */
  public boolean isDisposed() {
    return disposed;
  }

  /**
   * Wirft eine <code>AccessViolationException</code>, wenn das Zugriffsrecht
   * nicht mehr gueltig ist. Andernfalls passiert nichts.
   */
  public void checkDisposed() {
    if ( this.isDisposed() )
      throw new AccessViolationException(this,this.getClass().getName()+" is disposed!!");
  }

  /**
   * Macht das Recht ungueltig. Entspricht {@link #release()}.
   */
  public void dispose() {
    release();
  }

  /**
   * Gibt das Recht auf. Das betroffene Objekt wird informiert und
   * sollte entsprechend reagieren.<br>
   * Alle Referenzen auf das betroffene Objekt und auf den Besitzer des
   * Rechts werden entfernt!! {@link #getOwner()} liefert im folgenden also
   * <code>null</code>.
   * @see schmitzm.data.property.Accessible#releaseAccess(Access)
   */
  public void release() {
    this.disposed = true;
    if (object != null)
      object.releaseAccess(this);
    this.owner  = null;
    this.object = null;
  }

  /**
   * Vergleicht das Zugriffsrecht mit einem anderen. Zwei Rechte sind gleich,
   * wenn:
   * <ol>
   * <li>Die Klassen identisch sind.</li>
   * <li>Die Rechtebesitzer identisch sind.</li>
   * <li>Die Rechte sich auf dasselbe Objekt beziehen.</li>
   * </ol>
   * @param access Ein Zugriffsrecht
   */
  public boolean equals(Object access) {
    return access != null &&
           access.getClass().equals( this.getClass() ) &&
           ((Access)access).owner  == this.owner &&
           ((Access)access).object == this.object;
  }

  /**
   * Liefert die Anzahl, die der Besitzer berechtigt ist, das Zugriffsrecht
   * (ueber eine Methode) einzusetzen.
   * @see #UNLIMITED_ACCESSTIMES
   */
  public int getMaxAccessTimes() {
    return this.maxAccessTimes;
  }

  /**
   * Liefert die Anzahl an Zugriffen, die der Besitzer bereits (ueber eine
   * Methode) getaetigt hat.
   */
  public int getAccessTimes() {
    return this.accessTimes;
  }

  /**
   * Erhoeht die Anzahl an (Methoden-)Zugriffen auf das Recht und prueft, ob die
   * maximal erlaubte Anzahl erreicht/ueberschritten ist.
   * Ist dies der Fall wird das Zugriffsrecht ungueltig gemacht.
   * @see #dispose()
   */
  protected void incAndCheckMaxAccessTimesReached() {
    if ( !isDisposed() ) {
      accessTimes++;
      if (getMaxAccessTimes() != UNLIMITED_ACCESSTIMES &&
          getAccessTimes() >= getMaxAccessTimes())
        dispose();
    }
  }
}
