/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data.property;

import schmitzm.data.event.ObjectTraceable;

/**
 * Diese Interface sollten alle Objekte implementieren, die sich <b>dynamisch</b>
 * aus einzelnen Eigenschaften ({@link Property Properties}) zusammensetzen.
 * Die Methoden des Interface liefern sowohl den Zugriff auf die aktuellen Eigenschaften
 * des Objekts, ermoeglichen aber auch das dynamische Hinzufuegen oder Entfernen von
 * Eigenschaften.<br>
 * Zu beachten ist, dass es der jeweiligen Implementierung ueberlassen ist,
 * ob beim Entfernen einer Eigenschaft auch die komplette Eigenschaft (mittels
 * {@link Property#dispose()}) geschlossen wird. Je nach Implementierung muss
 * dies manuell geschehen. Deshalb liefern die remove-Methoden eine Referenz
 * auf die entfernte Eigenschaft zurueck.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public interface DynamicProperties extends Properties, ObjectTraceable {
  /**
   * Fuegt dem Objekt eine Eigenschaft hinzu. Das Objekt sollte darauf
   * achten, dass es noch keine Eigenschaft mit diesem Namen besitzt,
   * damit die Eindeutigkeit des Names gewaehrleistet ist.
   * @param prop Eigenschaft
   * @exception schmitzm.data.property.PropertyException falls das
   *            Objekt bereits eine Eigenschaft mit dem Namen der neuen
   *            Eigenschaft besitzt
   */
  public void addProperty(Property prop);

  /**
   * Entfernt eine Eigenschaft von dem Objekt.
   * @param name Name der Eigenschaft
   * @return Die entferne Eigenschaft oder <code>null</code> falls das
   *         Objekt keine Eigenschaft mit dem Namen besitzt
   */
  public Property removeProperty(String name);

  /**
   * Entfernt eine Eigenschaft von dem Objekt. Falls das Objekt diese
   * Eigenschaft nicht besitzt, passiert nichts.
   * @param prop zu entfernende Eigenschaft
   * @return Die entferne Eigenschaft oder <code>null</code> falls das Objekt
   *         die Eigenschaft nicht besitzt
   */
  public Property removeProperty(Property prop);

  /**
   * Entfernt Eigenschaften von dem Objekt. Falls das Objekt eine dieser
   * Eigenschaften nicht besitzt, passiert fuer die entsprechende Eigenschaft
   * nichts.
   * @param prop zu entfernende Eigenschaft
   * @return Die entfernten Eigenschaften (kann eine leere Eigenschaftsmenge
   *         sein, wenn das Objekt keine der uebergebenen Eigenschaften
   *         besitzt.
   */
  public Properties removeProperties(Properties prop);
}
