/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data.property;

import java.util.Vector;

import schmitzm.data.ObjectStructure;
import schmitzm.data.event.Invoker;
import schmitzm.data.event.ObjectChangeEvent;

/**
 * Diese Klasse stellt eine Listen-Eigenschaft dar, die Objekte eines bestimmten
 * Typs aufnehmen kann. Die Groesse der Liste steht a priori noch nicht fest.
 * Es koennen beliebig Objekte hingefügt oder entfernt werden.<br>
 * Intern wird die Liste als <code>java.util.Vector</code> organisiert.<br>
 * Der Zugriff auf die Liste kann durch Zugriffsrechte kontrolliert werden.
 * Ein Zugriff (lesend oder schreibend) ist nur ueber ein entsprechendes
 * Zugriffsrecht ({@link Access}) moeglich. Standardmaessig ist unbegrenzter
 * Zugriff eingestellt. Durch Angabe von {@link ValuePropertyAccessParameters}
 * kann jedoch benutzerdefiniert festgelegt werden, wie viele Objekte gleichzeitig
 * Lese- und/oder Schreibzugriff erlangen duerfen.
 * @see #getReadAccess(Object)
 * @see #getWriteAccess(Object)
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ListProperty extends ValueProperty {

  /**
   * Speichert die Listen-Elemente.
   */
  protected Vector elements = new Vector();

  /**
   * Erzeugt eine Listen-Eigenschaft.
   * @param name Bezeichnung der Eigenschaft
   * @param elementType Daten-Typ der Elemente, die die <code>ListProperty</code>
   *        aufnehmen kann
   * @param params Parameter fuer die Verwaltung des Zugriffsrechts
   * @see schmitzm.data.property.ValuePropertyAccessParameters
   */
  public ListProperty(String name, ValuePropertyType elementType, ValuePropertyAccessParameters params) {
    super( name, elementType, params );
  }

  /**
   * Erzeugt eine Listen-Eigenschaft. Der Zugriff auf die Eigenschaft ist
   * uneingeschraengt (beliebig viele Lese- und beliebig viele Schreibrechte).
   * @param name Bezeichnung der Eigenschaft
   * @param elementType Daten-Typ der Elemente, die die <code>ListProperty</code>
   *        aufnehmen kann
   * @see schmitzm.data.property.ValuePropertyAccessParameters#UNLIMITED_ACCESS
   */
  public ListProperty(String name, ValuePropertyType elementType) {
    this( name, elementType, ValuePropertyAccessParameters.UNLIMITED_ACCESS );
  }

  /**
   * Erzeugt eine Listen-Eigenschaft. Als Typ kann auch ein Build-in-Type angegeben
   * werden (z.B. int.class).
   * @param name Bezeichnung der Eigenschaft
   * @param elementType Daten-Typ der Elemente, die die <code>ListProperty</code>
   *        aufnehmen kann
   * @param params Parameter fuer die Verwaltung des Zugriffsrechts
   * @see schmitzm.data.property.ValuePropertyAccessParameters
   */
  public ListProperty(String name, Class elementType, ValuePropertyAccessParameters params) {
    super( name, elementType, params );
  }

  /**
   * Erzeugt eine Listen-Eigenschaft. Als Typ kann auch ein Build-in-Type angegeben
   * werden (z.B. int.class). Der Zugriff auf die Eigenschaft ist
   * uneingeschraengt (beliebig viele Lese- und beliebig viele Schreibrechte).
   * @param name Bezeichnung der Eigenschaft
   * @param elementType Daten-Typ der Elemente, die die <code>ListProperty</code>
   *        aufnehmen kann
   * @see schmitzm.data.property.ValuePropertyAccessParameters#UNLIMITED_ACCESS
   */
  public ListProperty(String name, Class elementType) {
    this( name, elementType, ValuePropertyAccessParameters.UNLIMITED_ACCESS );
  }

  /**
   * Erzeugt eine Listen-Eigenschaft.
   * @param name Bezeichnung der Eigenschaft
   * @param sample Bestimmt die Element-Struktur, die in der <code>ListProperty</code>
   *               gespeichert werden koennen
   * @param params Parameter fuer die Verwaltung des Zugriffsrechts
   * @see schmitzm.data.property.ValuePropertyAccessParameters
   */
  public ListProperty(String name, ObjectStructure sample, ValuePropertyAccessParameters params) {
    this( name, new ValuePropertyType(sample), params );
  }

  /**
   * Erzeugt eine Listen-Eigenschaft. Der Zugriff auf die Eigenschaft ist
   * uneingeschraengt (beliebig viele Lese- und beliebig viele Schreibrechte).
   * @param name Bezeichnung der Eigenschaft
   * @param sample Bestimmt die Element-Struktur, die in der <code>ListProperty</code>
   *               gespeichert werden koennen
   * @see schmitzm.data.property.ValuePropertyAccessParameters#UNLIMITED_ACCESS
   */
  public ListProperty(String name, ObjectStructure sample) {
    this( name, sample, ValuePropertyAccessParameters.UNLIMITED_ACCESS );
  }

  /**
   * Erzeugt eine Listen-Eigenschaft.
   * @param name Bezeichnung der Eigenschaft
   * @param sampleData Bestimmt die Element-Struktur, die in der <code>ListProperty</code>
   *                   gespeichert werden koennen (das Objekt wird <b>nicht</b>
   *                   in die Liste aufgenommen!!)
   * @param params Parameter fuer die Verwaltung des Zugriffsrechts
   * @see schmitzm.data.property.ValuePropertyAccessParameters
   */
  public ListProperty(String name, Object sampleData, ValuePropertyAccessParameters params) {
    this( name, ValuePropertyType.createValuePropertyType(sampleData), params );
  }

  /**
   * Erzeugt eine Listen-Eigenschaft. Der Zugriff auf die Eigenschaft ist
   * uneingeschraengt (beliebig viele Lese- und beliebig viele Schreibrechte).
   * @param name Bezeichnung der Eigenschaft
   * @param sampleData Bestimmt die Element-Struktur, die in der <code>ListProperty</code>
   *                   gespeichert werden koennen (das Objekt wird <b>nicht</b>
   *                   in die Liste aufgenommen!!)
   * @see schmitzm.data.property.ValuePropertyAccessParameters#UNLIMITED_ACCESS
   */
  public ListProperty(String name, Object sampleData) {
    this( name, sampleData, ValuePropertyAccessParameters.UNLIMITED_ACCESS );
  }

  /**
   * Erzeugt eine Listen-Eigenschaft. Als Elemente kann diese saemtliche Objekte
   * aufnehmen, die von <code>java.lang.Object</code> abgeleitet sind
   * (also <u>keine</u> Build-in-Types!!).
   * @param name Bezeichnung der Eigenschaft
   * @param params Parameter fuer die Verwaltung des Zugriffsrechts
   * @see schmitzm.data.property.ValuePropertyAccessParameters
   */
  public ListProperty(String name, ValuePropertyAccessParameters params) {
    this(name,Object.class,params);
  }

  /**
   * Erzeugt eine Listen-Eigenschaft. Als Elemente kann diese saemtliche Objekte
   * aufnehmen, die von <code>java.lang.Object</code> abgeleitet sind
   * (also <u>keine</u> Build-in-Types!!). Der Zugriff auf die Eigenschaft ist
   * uneingeschraengt (beliebig viele Lese- und beliebig viele Schreibrechte).
   * @param name Bezeichnung der Eigenschaft
   * @see schmitzm.data.property.ValuePropertyAccessParameters#UNLIMITED_ACCESS
   */
  public ListProperty(String name) {
    this(name,Object.class);
  }

  /**
   * Zerstoert die Property und die Property-Werte. Handelt es sich bei den
   * Werten der Liste wiederum Propertys, werden auch deren
   * <code>dispose()</code>-Methoden aufgerufen.
   */
  public void dispose() {
    for (int i=0; elements!=null && !elements.isEmpty(); i++) {
      Object o = elements.firstElement();
      this.removeValue(0);
      if ( o instanceof Property )
        ((Property)o).dispose();
    }
    super.dispose();
  }

  /**
   * Liefert eine neue, leerer <code>ListProperty</code> mit identischem
   * Namen, Typ und gleichen Zugriffsparemetern.
   */
  public ListProperty cloneStructure() {
    return new ListProperty( getName(), getPropertyType(), getAccessParameters() );
  }

  /**
   * Liefert ein Leserecht auf die Liste.
   * Wie viele Leserechte gleichzeitig erteilt werden koennen ist ueber die
   * Access-Parametern festgelegt.<br>
   * Neben den "normalen" Lesezugriffen (getter auf bestimmtes Objekt) muessen
   * fuer Listen weitere Zugriffsmethoden bereitgestellt werden. Deshalb liefert
   * diese Methode kein {@link PropertyReadAccess}, sondern ein
   * {@link ListPropertyReadAccess}.
   * @param owner Objekt, welches das Recht beantragt
   * @see #applyAccess(Access)
   * @see #getAccessParameters()
   */
  public ListPropertyReadAccess getReadAccess(Object owner) {
    return new ListPropertyReadAccess(this,owner);
  }

  /**
   * Liefert ein Schreibrecht auf die Property.
   * Wie viele Schreibrechte gleichzeitig erteilt werden koennen ist ueber die
   * Access-Parametern festgelegt.<br>
   * Neben den "normalen" Schreibzugriffen (setter auf bestimmtes Objekt) muessen
   * fuer Listen weitere Zugriffsmethoden bereitgestellt werden. Deshalb liefert
   * diese Methode kein {@link PropertyWriteAccess}, sondern ein
   * {@link ListPropertyWriteAccess}.
   * @param owner Objekt, welches das Recht beantragt
   * @see #applyAccess(Access)
   * @see #getAccessParameters()
   */
  public ListPropertyWriteAccess getWriteAccess(Object owner) {
    return new ListPropertyWriteAccess(this,owner);
  }

  /**
   * Liefert ein Leserecht auf die Liste, welches nach einmaligem
   * Zugriff automatisch ungueltig wird.<br>
   * Wie viele Leserechte gleichzeitig erteilt werden koennen ist ueber die
   * Access-Parametern festgelegt.<br>
   * Neben den "normalen" Lesezugriffen (getter auf bestimmtes Objekt) muessen
   * fuer Listen weitere Zugriffsmethoden bereitgestellt werden. Deshalb liefert
   * diese Methode kein {@link PropertyReadAccess}, sondern ein
   * {@link ListPropertyReadAccess}.
   * @see #applyAccess(Access)
   * @see #getAccessParameters()
   */
  public ListPropertyReadAccess getOneTimeReadAccess() {
    return new ListPropertyReadAccess(this,this,1);
  }

  /**
   * Liefert ein Schreibrecht auf die Property, welches nach einmaligem
   * Zugriff automatisch ungueltig wird.<br>
   * Wie viele Schreibrechte gleichzeitig erteilt werden koennen ist ueber die
   * Access-Parametern festgelegt.<br>
   * Neben den "normalen" Schreibzugriffen (setter auf bestimmtes Objekt) muessen
   * fuer Listen weitere Zugriffsmethoden bereitgestellt werden. Deshalb liefert
   * diese Methode kein {@link PropertyWriteAccess}, sondern ein
   * {@link ListPropertyWriteAccess}.
   * @see #applyAccess(Access)
   * @see #getAccessParameters()
   */
  public ListPropertyWriteAccess getOneTimeWriteAccess() {
    return new ListPropertyWriteAccess(this,this,1);
  }

  /**
   * Liefert die Anzahl an Zugriffsrechten einer bestimmten Art, die aktuell
   * fuer das Objekt verteilt sind.<br>
   * Muss {@link ValueProperty#getAccessCount(Class)} ueberschreiben, da dieser
   * Klasse Instanzen von {@link ListPropertyReadAccess} und {@link ListPropertyWriteAccess}
   * verteilt.
   */
  public int getAccessCount(Class c) {
    int accessCount = 0;
    // Wird als Typ ListPropertyReadAccess (ListPropertyReadAccess) angegeben,
    // wuerde die super-Methode 0 zurueckgeben, da diese kein Obertyp von
    // PropertyReadAccess (PropertyReadAccess) ist!!
    if ( c==null || c.isAssignableFrom(ListPropertyReadAccess.class) )
      accessCount += super.getAccessCount(PropertyReadAccess.class);
    if ( c==null || c.isAssignableFrom(ListPropertyWriteAccess.class) )
      accessCount += super.getAccessCount(PropertyWriteAccess.class);
    return accessCount;
  }

  ///////////////////////////////////////////////////////////////////
  // Implementierung von ValueProperty
  ///////////////////////////////////////////////////////////////////

  /**
   * Liefert ein Element der Liste. Kann (seit JDK1.5.0) auch genutzt werden,
   * wenn es sich bei der Property um einen Build-in-Type handelt.<br>
   * <b>Da eine Listen-Eigenschaft 1-dim. ist, muss als
   * <code>coords</code>-Parameter genau 1 Wert angegeben werden!!</b>
   * @param coords Listen-Index
   * @exception java.lang.UnsupportedOperationException falls nicht genau ein
   *            <code>coords</code>-Parameter angegeben wird
   */
  protected Object getValue(int... coords) {
    if (coords.length != 1)
      throw new UnsupportedOperationException("A ListProperty contains values in one dimension. Thus one and only one parameter are allowed for getValue(..).");
    return elements.elementAt(coords[0]);
  }

  /**
   * Belegt ein Element der Liste. Kann (seit JDK1.5.0) auch genutzt werden,
   * wenn es sich bei der Property um einen Build-in-Type handelt.<br>
   * Iniziiert ein <code>ObjectChangeEvent</code> fuer die Liste.
   * <b>Da eine Listen-Eigenschaft 1-dim. ist, muss als
   * <code>coords</code>-Parameter genau 1 Wert angegeben werden!!</b>
   * @param value  neuer Wert fuer das Listen-Element
   * @param coords Listen-Index
   * @exception java.lang.UnsupportedOperationException falls nicht genau ein
   *            <code>coords</code>-Parameter angegeben wird
   * @exception java.lang.ClassCastException falls das angegebene Objekt nicht
   *            zum Property-Typ passt
   */
  protected void setValue(Object value, int... coords) {
    if (coords.length != 1)
      throw new UnsupportedOperationException("A ListProperty contains values in one dimension. Thus one and only one parameter are allowed for getValue(..).");
    if ( value!=null && !this.isValid( value.getClass() ) )
      throw new ClassCastException(value.getClass().getName().concat(" can not be stored in a ListProperty[").concat(this.getType().getName()).concat("]"));

    // Die set-Methode darf auch fuer einen ueber die Liste
    // hinaus reichenden Index aufgerufen. In diesem Fall wird
    // einfach ein Objekt hinzugefuegt.
    if ( coords[0] == getCount() ) {
      insertValue(value,coords[0]);
    } else {
      // altes Objekt merken fuer Event
      Object oldValue = getValue(coords[0]);
      elements.setElementAt(value, coords[0]);

      this.fireEvent(new ObjectChangeEvent(
          new Invoker(this),
          oldValue,
          value
      ));
    }
  }

  ///////////////////////////////////////////////////////////////////
  // Neue Listen-Funktionen
  ///////////////////////////////////////////////////////////////////

  /**
   * Liefert den Listen-Index des ersten Vorkommens eines Elements.
   * @param value Listen-Element
   * @return -1 falls das Element nicht in der Liste vorhanden ist.
   */
  protected int indexOf(Object value) {
    return elements.indexOf(value);
  }

  /**
   * Liefert die aktuelle Anzahl an Elementen, die in der Liste gespeichert
   * sind.
   */
  protected int getCount() {
    return elements.size();
  }

  /**
   * Belegt ein Element der Liste. Kann (seit JDK1.5.0) auch genutzt werden,
   * wenn es sich bei der Property um einen Build-in-Type handelt.<br>
   * Iniziiert ein <code>ObjectChangeEvent</code> fuer die Liste.
   * @param value  neuer Wert fuer das Listen-Element
   * @param index  Listen-Index fuer das neue Element (alle bestehenden Elemente
   *               ab (einschliesslich) dieser Position werden nach hinten
   *               verschoben)
   * @exception java.lang.ClassCastException falls das angegebene Objekt nicht
   *            zum Property-Typ passt
   */
  protected void insertValue(Object value, int index) {
    if (!isValid(value.getClass()))
      throw new ClassCastException(value.getClass().getName().concat(
          " can not be stored in a ListProperty[").concat(getType().getName()).concat("]"));

    elements.insertElementAt(value, index);

    fireEvent(new ObjectChangeEvent(
        new Invoker(this),
        null,
        value
    ));
  }

  /**
   * Entfernt ein Element aus der Liste.<br>
   * Iniziiert ein <code>ObjectChangeEvent</code> fuer die Liste.
   * @param index Listen-Index des zu loeschenden Elements
   */
  protected void removeValue(int index) {
    fireEvent(new ObjectChangeEvent(
        new Invoker(this),
        getValue(index),
        null
    ));
    elements.remove(index);
  }

  /**
   * Entfernt alle Instanzen des angegebenen Objekts aus der Liste (sofern
   * es vorhanden ist).<br>
   * Iniziiert ein <code>ObjectChangeEvent</code> fuer die Liste.
   * @param value das zu entfernende Objekt
   * @return <code>false</code> falls dass Objekt nicht in der Liste vorhanden war
   */
  protected boolean removeValue(Object value) {
    int count = 0;
    for (int i = 0; (i = this.indexOf(value)) >= 0; count++)
      removeValue(i);
    return count > 0;
  }

  ///////////////////////////////////////////////////////////////////
  // Implementierung von ObjectStructure
  ///////////////////////////////////////////////////////////////////
  /**
   * Liefert <code>true</code>, da eine Listen-Eigenschaft immer
   * aus mehreren Werten besteht.
   */
  public boolean containsMultipleValues() {
    return true;
  }

}
