/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data.property;

/**
 * Diese Klasse stellt ein Recht auf Lesezugriff fuer eine
 * {@link ListProperty} dar. Da Listen dynamisch aufgebaut sind, erweitert sie
 * hierzu den {@link PropertyReadAccess} um <code>getCount</code>-
 * und <code>indexOf</code>-Methoden.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ListPropertyReadAccess extends PropertyReadAccess {
  /**
   * Erzeugt ein neues Lesezugriffsrecht fuer Listen. Die Anzahl an Zugriffen
   * fuer den Besitzer ist unbegrenzt.
   * @param object Instanz von <code>ListProperty</code> auf die sich das Recht bezieht
   * @param owner  Besitzer des Rechts
   * @see schmitzm.data.property.ListProperty
   * @exception schmitzm.data.property.AccessViolationException falls das
   *            angegebene Objekt <b>keine</b> <code>ListProperty</code> ist
   */
  public ListPropertyReadAccess(Accessible object, Object owner) {
    this(object, owner, UNLIMITED_ACCESSTIMES);
  }

  /**
   * Erzeugt ein neues Lesezugriffsrecht fuer Listen.
   * @param object Instanz von <code>ListProperty</code> auf die sich das Recht bezieht
   * @param owner  Besitzer des Rechts
   * @param maxAccessTimes Anzahl an (Methoden-)Zugriffen, die der Rechtebesitzer
   *                       taetigen darf, bevor das Recht automatisch entzogen wird
   * @see schmitzm.data.property.ListProperty
   * @exception schmitzm.data.property.AccessViolationException falls das
   *            angegebene Objekt <b>keine</b> <code>ListProperty</code> ist
   */
  public ListPropertyReadAccess(Accessible object, Object owner, int maxAccessTimes) {
    super(object, owner, maxAccessTimes);
    if (! (object instanceof ListProperty))
      throw new AccessViolationException(this,this.getClass().getName() +" can only be instantiated for ListProperties");
  }

  /**
   * Liefert die aktuelle Anzahl an Elementen, die in der Liste gespeichert
   * sind.
   * @exception schmitzm.data.property.AccessViolationException falls
   *            das Leserecht auf der Property nicht mehr gueltig ist
   */
  public int getCount() {
    checkDisposed();
    int ret = ((ListProperty)object).getCount();
    incAndCheckMaxAccessTimesReached();
    return ret;
  }

  /**
   * Checkt, ob die Liste leer ist.
   * @exception schmitzm.data.property.AccessViolationException falls
   *            das Leserecht auf der Property nicht mehr gueltig ist
   */
  public boolean isEmpty() {
    return getCount() == 0;
  }

  /**
   * Liefert den Listen-Index des ersten Vorkommens eines Elements.
   * @param value Listen-Element
   * @return -1 falls das Element nicht in der Liste vorhanden ist.
   * @exception schmitzm.data.property.AccessViolationException falls
   *            das Leserecht auf der Property nicht mehr gueltig ist
   */
  public int indexOf(Object value) {
    checkDisposed();
    int ret = ((ListProperty)object).indexOf(value);
    incAndCheckMaxAccessTimesReached();
    return ret;
  }

  /**
   * Checkt, ob ein Objekt in der Liste vorhanden ist.
   * @exception schmitzm.data.property.AccessViolationException falls
   *            das Leserecht auf der Property nicht mehr gueltig ist
   */
  public boolean contains(Object value) {
    return indexOf(value) >= 0;
  }
}
