/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data.property;


/**
 * Diese Klasse stellt ein Recht auf Schreibzugriff fuer eine
 * {@link ListProperty} dar. Da Listen dynamisch aufgebaut sind, erweitert sie
 * hierzu den {@link PropertyWriteAccess} um <code>insert</code>-
 * und <code>remove</code>-Methoden.<br>
 * <b>Bemerkung:</b><br>
 * Um auch den kompletten Lesezugriff zu ermoeglichen, werden zusaetzlich auch
 * noch einmal alle (neuen) Methoden von {@link ListPropertyReadAccess}
 * implementiert. Dies ist erforderlich, da diese Klasse von
 * {@link PropertyWriteAccess} abgeleitet werden muss und doppelte Ableitung
 * leider nicht moeglich ist.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ListPropertyWriteAccess extends PropertyWriteAccess {

  /**
   * Erzeugt ein neues Schreibzugriffsrecht fuer Listen. Die Anzahl an Zugriffen
   * fuer den Besitzer ist unbegrenzt.
   * @param object Instanz von <code>ListProperty</code> auf die sich das Recht bezieht
   * @param owner  Besitzer des Rechts
   * @see schmitzm.data.property.ListProperty
   * @exception schmitzm.data.property.AccessViolationException falls das
   *            angegebene Objekt <b>keine</b> <code>ListProperty</code> ist
   */
  public ListPropertyWriteAccess(Accessible object, Object owner) {
    this(object, owner,UNLIMITED_ACCESSTIMES);
  }

  /**
   * Erzeugt ein neues Schreibzugriffsrecht fuer Listen.
   * @param object Instanz von <code>ListProperty</code> auf die sich das Recht bezieht
   * @param owner  Besitzer des Rechts
   * @param maxAccessTimes Anzahl an (Methoden-)Zugriffen, die der Rechtebesitzer
   *                       taetigen darf, bevor das Recht automatisch entzogen wird
   * @see schmitzm.data.property.ListProperty
   * @exception schmitzm.data.property.AccessViolationException falls das
   *            angegebene Objekt <b>keine</b> <code>ListProperty</code> ist
   */
  public ListPropertyWriteAccess(Accessible object, Object owner, int maxAccessTimes) {
    super(object, owner, maxAccessTimes);
    if (! (object instanceof ListProperty))
      throw new AccessViolationException(this,
          this.getClass().getName() + " can only be instantiated for ListProperties");
  }

  /**
   * Fuegt der Liste ein Element (am Ende) hinzu. Kann (seit JDK1.5.0) auch
   * genutzt werden, wenn es sich bei der Property um einen Build-in-Type handelt.<br>
   * Iniziiert ein <code>ObjectChangeEvent</code> fuer die Liste.
   * @param value  neuer Wert fuer die Liste
   * @exception java.lang.UnsupportedOperationException falls nicht genau ein
   *            <code>coords</code>-Parameter angegeben wird
   * @exception java.lang.ClassCastException falls das angegebene Objekt nicht
   *            zum Property-Typ passt
   * @exception schmitzm.data.property.AccessViolationException falls
   *            das Schreibrecht auf der Property nicht mehr gueltig ist
   */
  public void addValue(Object value) {
// Kann so nicht gemacht werden, da bei OneTimeWriteAccess das Recht
// bei Aufruf der this.getCount()-Methode bereits "verbraucht" wird!!
// Das anschliessende this.insertValue(..) fuehrt dann zu einer
// AccessViolationException, da kein Recht mehr besteht!! :-)
// Loesung: >> direkt auf getCount-Methode des Objekts zugreifen!
//  insertValue(value, getCount());
    insertValue(value,((ListProperty)object).getCount());
  }

  /**
   * Belegt ein Element der Liste. Kann (seit JDK1.5.0) auch genutzt werden,
   * wenn es sich bei der Property um einen Build-in-Type handelt.<br>
   * Iniziiert ein <code>ObjectChangeEvent</code> fuer die Liste.
   * @param value  neuer Wert fuer das Listen-Element
   * @param index  Listen-Index fuer das neue Element (alle bestehenden Elemente
   *               ab (einschliesslich) dieser Position werden nach hinten
   *               verschoben)
   * @exception java.lang.ClassCastException falls das angegebene Objekt nicht
   *            zum Property-Typ passt
   * @exception schmitzm.data.property.AccessViolationException falls
   *            das Schreibrecht auf der Property nicht mehr gueltig ist
   */
  public void insertValue(Object value, int index) {
    checkDisposed();
    ((ListProperty)this.object).insertValue(value,index);
    incAndCheckMaxAccessTimesReached();
  }

  /**
   * Entfernt ein Element aus der Liste.<br>
   * Iniziiert ein <code>ObjectChangeEvent</code> fuer die Liste.
   * @param index Listen-Index des zu loeschenden Elements
   * @exception schmitzm.data.property.AccessViolationException falls
   *            das Schreibrecht auf der Property nicht mehr gueltig ist
   */
  public void removeValue(int index) {
    checkDisposed();
    ((ListProperty)this.object).removeValue(index);
    incAndCheckMaxAccessTimesReached();
  }

  /**
   * Entfernt alle Instanzen des angegebenen Objekts aus der Liste (sofern
   * es vorhanden ist).<br>
   * Iniziiert ein <code>ObjectChangeEvent</code> fuer die Liste.
   * @param value das zu entfernende Objekt
   * @return <code>false</code> falls dass Objekt nicht in der Liste vorhanden war
   * @exception schmitzm.data.property.AccessViolationException falls
   *            das Schreibrecht auf der Property nicht mehr gueltig ist
   */
  public boolean removeValue(Object value) {
    checkDisposed();
    boolean ret = ((ListProperty)this.object).removeValue(value);
    incAndCheckMaxAccessTimesReached();
    return ret;
  }

  /**
   * Entfernt alle Objekte aus der Liste.<br>
   * Iniziiert ein <code>ObjectChangeEvent</code> fuer jedes geloeschte
   * Element der Liste.
   * @exception schmitzm.data.property.AccessViolationException falls
   *            das Schreibrecht auf der Property nicht mehr gueltig ist
   */
  public void removeAll() {
    checkDisposed();
    while ( ((ListProperty)this.object).getCount() > 0 )
      ((ListProperty)this.object).removeValue(0);
    incAndCheckMaxAccessTimesReached();
  }

  ////////////////////////////////////////////////////////////////////////
  // Folgende Methoden sind identisch zur ListPropertyReadAcces.
  // Muessen leider doppelt implememtiert werden, da ListPropertyWriteAccess
  // bereits von PropertyWriteAccess abgeleitet werden muss und doppelte
  // Ableitung leider nicht moeglich ist.
  ////////////////////////////////////////////////////////////////////////
  /**
   * Liefert die aktuelle Anzahl an Elementen, die in der Liste gespeichert
   * sind.<br>
   * Identisch zu {@link ListPropertyReadAccess#getCount()}.
   * @exception schmitzm.data.property.AccessViolationException falls
   *            das Schreibrecht auf der Property nicht mehr gueltig ist
   */
  public int getCount() {
    checkDisposed();
    int ret = ((ListProperty)object).getCount();
    incAndCheckMaxAccessTimesReached();
    return ret;
  }

  /**
   * Checkt, ob die Liste leer ist.<br>
   * Identisch zu {@link ListPropertyReadAccess#isEmpty()}.
   * @exception schmitzm.data.property.AccessViolationException falls
   *            das Schreibrecht auf der Property nicht mehr gueltig ist
   */
  public boolean isEmpty() {
    return getCount() == 0;
  }

  /**
   * Liefert den Listen-Index des ersten Vorkommens eines Elements.<br>
   * Identisch zu {@link ListPropertyReadAccess#indexOf(Object)}.
   * @param value Listen-Element
   * @return -1 falls das Element nicht in der Liste vorhanden ist.
   * @exception schmitzm.data.property.AccessViolationException falls
   *            das Schreibrecht auf der Property nicht mehr gueltig ist
   */
  public int indexOf(Object value) {
    checkDisposed();
    int ret = ((ListProperty)object).indexOf(value);
    incAndCheckMaxAccessTimesReached();
    return ret;
  }

  /**
   * Checkt, ob ein Objekt in der Liste vorhanden ist.<br>
   * Identisch zu {@link ListPropertyReadAccess#contains(Object)}.
   * @exception schmitzm.data.property.AccessViolationException falls
   *            das Schreibrecht auf der Property nicht mehr gueltig ist
   */
  public boolean contains(Object value) {
    return indexOf(value) >= 0;
  }
}
