/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data.property;

import schmitzm.data.ObjectStructure;
import schmitzm.data.event.Invoker;
import schmitzm.data.event.ObjectChangeEvent;
import schmitzm.lang.MultiDimArray;

/**
 * Diese Klasse stellt eine Matrix-Eigenschaft dar, die Objekte eines bestimmten
 * Typs aufnehmen kann. Die Dimension und Groesse der Matrix wird bei der
 * Instanziierung festgelegt und kann nicht geaendert werden.<br>
 * Der Zugriff auf die Property kann durch Zugriffsrechte kontrolliert werden.
 * Ein Zugriff (lesend oder schreibend) ist nur ueber ein entsprechendes
 * Zugriffsrecht ({@link Access}) moeglich. Standardmaessig ist unbegrenzter
 * Zugriff eingestellt. Durch Angabe von {@link ValuePropertyAccessParameters}
 * kann jedoch benutzerdefiniert festgelegt werden, wie viele Objekte gleichzeitig
 * Lese- und/oder Schreibzugriff erlangen duerfen.
 * @see schmitzm.data.property.ValueProperty#getReadAccess(Object)
 * @see schmitzm.data.property.ValueProperty#getWriteAccess(Object)
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class MatrixProperty extends ValueProperty {
  // Groesse der Matrix merken (aus Effizienzgruenden!)
  private int[] size = null;

  /**
   * Speichert die Matrix als <code>java.lang.reflect.Array</code>.
   * @see java.lang.reflect.Array
   */
  protected Object matrix = null;

  /**
   * Speichert die Groessenangabe der Matrix als <code>String</code> der Form
   * <code>"(d1,d2,d3,...)</code>.
   */
  protected String sizeDesc = "";

  /**
   * Erzeugt eine Matrix-Eigenschaft. Die Dimension der Matrix ist implizit
   * durch die Größe des <code>size</code>-Arrays gegeben
   * @param name Name der Eigenschaft
   * @param size Groesse der Matrix in jeder Dimension
   * @param elementType Matrix kann nur Elemente dieses Daten-Typs aufnehmen
   * @param params Parameter fuer die Verwaltung des Zugriffsrechts
   * @see schmitzm.data.property.ValuePropertyAccessParameters
   */
  public MatrixProperty(String name, int[] size, ValuePropertyType elementType, ValuePropertyAccessParameters params) {
    super(name, elementType, params);
    this.size      = size.clone();
    this.matrix    = MultiDimArray.newInstance(elementType.getType(),size);
    // Groessenbeschreibung zusammensetzen
    sizeDesc = "(";
    for (int d=0; d<size.length; d++) {
      if ( d > 0 ) sizeDesc+=",";
      sizeDesc+=size[d];
    }
    sizeDesc+=")";
  }

  /**
   * Erzeugt eine Matrix-Eigenschaft. Die Dimension der Matrix ist implizit
   * durch die Größe des <code>size</code>-Arrays gegeben.
   * Der Zugriff auf die Eigenschaft ist uneingeschraengt (beliebig viele
   * Lese- und beliebig viele Schreibrechte).
   * @param name Name der Eigenschaft
   * @param size Groesse der Matrix in jeder Dimension
   * @param elementType Matrix kann nur Elemente dieses Daten-Typs aufnehmen
   * @see schmitzm.data.property.ValuePropertyAccessParameters#UNLIMITED_ACCESS
   */
  public MatrixProperty(String name, int[] size, ValuePropertyType elementType) {
    this(name, size, elementType, ValuePropertyAccessParameters.UNLIMITED_ACCESS);
  }

  /**
   * Erzeugt eine Matrix-Eigenschaft. Als Typ kann auch ein Build-in-Type
   * angegeben werden (z.B. int.class). Die Dimension der Matrix ist implizit
   * durch die Größe des <code>size</code>-Arrays gegeben
   * @param name Name der Eigenschaft
   * @param size Groesse der Matrix in jeder Dimension
   * @param elementType Matrix kann nur Elemente dieses Daten-Typs aufnehmen
   * @param params Parameter fuer die Verwaltung des Zugriffsrechts
   * @see schmitzm.data.property.ValuePropertyAccessParameters
   */
  public MatrixProperty(String name, int[] size, Class elementType, ValuePropertyAccessParameters params) {
    this(name,size,new ValuePropertyType(elementType),params);
  }

  /**
   * Erzeugt eine Matrix-Eigenschaft. Als Typ kann auch ein Build-in-Type
   * angegeben werden (z.B. int.class). Die Dimension der Matrix ist implizit
   * durch die Größe des <code>size</code>-Arrays gegeben.
   * Der Zugriff auf die Eigenschaft ist uneingeschraengt (beliebig viele
   * Lese- und beliebig viele Schreibrechte).
   * @param name Name der Eigenschaft
   * @param size Groesse der Matrix in jeder Dimension
   * @param elementType Matrix kann nur Elemente dieses Daten-Typs aufnehmen
   * @see schmitzm.data.property.ValuePropertyAccessParameters#UNLIMITED_ACCESS
   */
  public MatrixProperty(String name, int[] size, Class elementType) {
    this(name, size, elementType, ValuePropertyAccessParameters.UNLIMITED_ACCESS);
  }

  /**
   * Erzeugt eine Matrix-Eigenschaft. Als Typ kann auch ein Build-in-Type
   * angegeben werden (z.B. int.class). Die Dimension der Matrix ist implizit
   * durch die Größe des <code>size</code>-Arrays gegeben
   * @param name Name der Eigenschaft
   * @param size Groesse der Matrix in jeder Dimension
   * @param sample Matrix kann nur Elemente dieser Struktur aufnehmen
   * @param params Parameter fuer die Verwaltung des Zugriffsrechts
   * @see schmitzm.data.property.ValuePropertyAccessParameters
   */
  public MatrixProperty(String name, int[] size, ObjectStructure sample, ValuePropertyAccessParameters params) {
    this(name,size,new ValuePropertyType(sample),params);
  }

  /**
   * Erzeugt eine Matrix-Eigenschaft. Als Typ kann auch ein Build-in-Type
   * angegeben werden (z.B. int.class). Die Dimension der Matrix ist implizit
   * durch die Größe des <code>size</code>-Arrays gegeben.
   * Der Zugriff auf die Eigenschaft ist uneingeschraengt (beliebig viele
   * Lese- und beliebig viele Schreibrechte).
   * @param name Name der Eigenschaft
   * @param size Groesse der Matrix in jeder Dimension
   * @param sample Matrix kann nur Elemente dieser Struktur aufnehmen
   * @see schmitzm.data.property.ValuePropertyAccessParameters#UNLIMITED_ACCESS
   */
  public MatrixProperty(String name, int[] size, ObjectStructure sample) {
    this(name, size, sample, ValuePropertyAccessParameters.UNLIMITED_ACCESS);
  }

  /**
   * Erzeugt eine Matrix-Eigenschaft. Als Typ kann auch ein Build-in-Type
   * angegeben werden (z.B. int.class). Die Dimension der Matrix ist implizit
   * durch die Größe des <code>size</code>-Arrays gegeben
   * @param name Name der Eigenschaft
   * @param size Groesse der Matrix in jeder Dimension
   * @param sampleData Bestimmt die Element-Struktur, die in der <code>MatrixProperty</code>
   *                   gespeichert werden kann (das Objekt wird <b>nicht</b>
   *                   in die Matrix aufgenommen!!)
   * @param params Parameter fuer die Verwaltung des Zugriffsrechts
   * @see schmitzm.data.property.ValuePropertyAccessParameters
   */
  public MatrixProperty(String name, int[] size, Object sampleData, ValuePropertyAccessParameters params) {
    this(name,size,ValuePropertyType.createValuePropertyType(sampleData),params);
  }

  /**
   * Erzeugt eine Matrix-Eigenschaft. Als Typ kann auch ein Build-in-Type
   * angegeben werden (z.B. int.class). Die Dimension der Matrix ist implizit
   * durch die Größe des <code>size</code>-Arrays gegeben.
   * Der Zugriff auf die Eigenschaft ist uneingeschraengt (beliebig viele
   * Lese- und beliebig viele Schreibrechte).
   * @param name Name der Eigenschaft
   * @param size Groesse der Matrix in jeder Dimension
   * @param sampleData Bestimmt die Element-Struktur, die in der <code>MatrixProperty</code>
   *                   gespeichert werden kann (das Objekt wird <b>nicht</b>
   *                   in die Matrix aufgenommen!!)
   * @see schmitzm.data.property.ValuePropertyAccessParameters#UNLIMITED_ACCESS
   */
  public MatrixProperty(String name, int[] size, Object sampleData) {
    this(name, size, sampleData, ValuePropertyAccessParameters.UNLIMITED_ACCESS);
  }

  /**
   * Erzeugt eine Matrix-Eigenschaft. Als Elemente kann diese saemtliche Objekte
   * aufnehmen, die von <code>java.lang.Object</code> abgeleitet sind
   * (also <u>keine</u> Build-in-Types!!). Die Dimension der Matrix ist implizit
   * durch die Größe des <code>size</code>-Arrays gegeben
   * @param name Name der Eigenschaft
   * @param size Groesse der Matrix in jeder Dimension
   * @param params Parameter fuer die Verwaltung des Zugriffsrechts
   * @see schmitzm.data.property.ValuePropertyAccessParameters
   */

  public MatrixProperty(String name, int[] size, ValuePropertyAccessParameters params) {
    this(name, size, Object.class,params);
  }

  /**
   * Erzeugt eine Matrix-Eigenschaft. Als Elemente kann diese saemtliche Objekte
   * aufnehmen, die von <code>java.lang.Object</code> abgeleitet sind
   * (also <u>keine</u> Build-in-Types!!). Die Dimension der Matrix ist implizit
   * durch die Größe des <code>size</code>-Arrays gegeben.
   * Der Zugriff auf die Eigenschaft ist uneingeschraengt (beliebig viele
   * Lese- und beliebig viele Schreibrechte).
   * @param name Name der Eigenschaft
   * @param size Groesse der Matrix in jeder Dimension
   * @see schmitzm.data.property.ValuePropertyAccessParameters#UNLIMITED_ACCESS
   */

  public MatrixProperty(String name, int[] size) {
    this(name, size, Object.class, ValuePropertyAccessParameters.UNLIMITED_ACCESS);
  }

  /**
   * Zerstoert die Matrix, indem die auf <code>null</code> gesetzt wird.
   */
  public void dispose() {
    matrix = null;
    super.dispose();
  }

  /**
   * Liefert eine neue, leerer <code>MatrixProperty</code> mit identischem
   * Namen, Typ, geleicher Groesse und gleichen Zugriffsparemetern.
   */
  public MatrixProperty cloneStructure() {
    return new MatrixProperty( getName(), getSize(), getPropertyType(), getAccessParameters() );
  }

  ///////////////////////////////////////////////////////////////////
  // Implementierung von ValueProperty
  ///////////////////////////////////////////////////////////////////

  /**
   * Liefert ein Element der Matrix. Kann (seit JDK1.5.0) auch genutzt werden,
   * wenn es sich bei der Property um einen Build-in-Type handelt.
   * @param coords Position (Koordinaten) der Matrix
   * @exception java.lang.UnsupportedOperationException falls die Anzahl der
   *            Koordinaten nicht genau der Matrix-Dimension entspricht
   */
  protected Object getValue(int... coords) {
    if (coords.length != getDimension())
      throw new UnsupportedOperationException("The MatrixProperty is ".concat(""+getDimension()).concat("-dimensional. Thus exactly ").concat(""+getDimension()).concat(" parameters are needed for getValue(..).") );
    return MultiDimArray.get(matrix,coords);
  }

  /**
   * Setzt ein Element der Matrix. Kann (seit JDK1.5.0) auch genutzt werden,
   * wenn es sich bei der Property um einen Build-in-Type handelt.<br>
   * Iniziiert ein <code>ObjectChangeEvent</code> fuer die Liste.
   * @param value  Wert fuer das Matrix-Element
   * @param coords Position (Koordinaten) der Matrix
   * @exception java.lang.UnsupportedOperationException falls die Anzahl der
   *            Koordinaten nicht genau der Matrix-Dimension entspricht
   * @exception java.lang.ClassCastException falls das angegebene Objekt nicht
   *            zum Property-Typ passt
   */
  protected void setValue(Object value, int... coords) {
    if (coords.length != getDimension())
      throw new UnsupportedOperationException("The MatrixProperty is ".concat(""+getDimension()).concat("-dimensional. Thus exactly ").concat(""+getDimension()).concat(" parameters are needed for setValue(..).") );
    if ( value!=null && !this.isValid( value.getClass() ) )
      throw new ClassCastException(value.getClass().getName().concat(" can not be stored in a MatrixProperty[").concat(this.getType().getName()).concat("]"));

    // altes Objekt merken fuer Event
    Object oldValue = getValue(coords);
    MultiDimArray.set(matrix,coords,value);

    this.fireEvent( new ObjectChangeEvent(
      new Invoker(this),
      oldValue,
      value
    ) );
  }

  ///////////////////////////////////////////////////////////////////
  // Neue Matrix-Funktionen
  ///////////////////////////////////////////////////////////////////
  /**
   * Liefert die Dimension der Matrix.
   */
  public int getDimension() {
    return this.size.length;
  }

  /**
   * Liefert die Groesse der Matrix in einer Dimension.
   * @param dim Index der Dimension (beginnend bei 0)
   */
  public int getSize(int dim) {
    return size[dim];
  }

  /**
   * Liefert die Groesse der Matrix in allen Dimensionen.
   */
  public int[] getSize() {
    return size.clone();
  }

  /**
   * Liefert die Groesse der Matrix in allen Dimensionen als Beschreibung.
   */
  public String getSizeText() {
    return sizeDesc;
  }

  ///////////////////////////////////////////////////////////////////
  // Implementierung von ObjectStructure
  ///////////////////////////////////////////////////////////////////
  /**
   * Liefert <code>true</code>, da eine Matrix-Eigenschaft immer
   * aus mehreren Werten besteht.
   */
  public boolean containsMultipleValues() {
    return true;
  }

//  /**
//   * Liefert <code>true</code>, da neben dem Klassennamen auch die groesse
//   * der Matrix ausgewiesen werden soll.
//   */
//  public boolean isStructureNamed() {
//    return true;
//  }
//
//  /**
//   * Liefert <code>true</code>, da neben dem Klassennamen auch die groesse
//   * der Matrix ausgewiesen werden soll.
//   */
//  public String getStructureName() {
//    return this.getClass().getSimpleName().concat( getSizeText() );
//  }
}
