/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data.property;

import schmitzm.data.ObjectStructure;

/**
 * Diese Interface sollten alle Objekte implementieren, die sich (fest) aus
 * einzelnen Eigenschaften ({@link Property Properties}) zusammensetzen.
 * Die Methoden des Interface liefern den Zugriff auf die einzelnen Eigenschaften.
 * Das Hinzufuegen oder Entfernen von Eigenschaften ist durch dieses Interface
 * nicht moeglich.
 * @see schmitzm.data.property.DynamicProperties
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public interface Properties extends ObjectStructure {
  /**
   * Liefert die Anzahl an Eigenschaften des Objekts.
   */
  public int getPropertyCount();

  /**
   * Liefert die Eigenschaften des Objekts als Array. Wenn das Objekt (noch)
   * keine Eigenschaften besitzt, sollte ein leerer Array (und nicht <code>null</code>!)
   * zurueckgegeben werden!
   */
  public Property[] getProperties();

  /**
   * Liefert alle Eigenschaften des Objekts, die einem bestimmten
   * Typ angehoeren. Als Typ kann sowohl ein bestimmter <code>Property</code>-Untertyp
   * angegeben werden (z.B. {@link ListProperty}) oder ein Objekt-Typ (Typ
   * den eine Property aufnehmen kann; z.B. <code>int.class</code>).<br>
   * Da die Methode wiederum eine <code>Properties</code>-Instanz liefert,
   * lassen sich recht einfach verschiedene Filter hintereinander anwenden:<br>
   * <br>
   * <center><code>
   * Properties.getProperties(ListProperty.class).getProperties(int.class).getPropertyNames()
   * </code></center>
   * <br>
   * liefert z.B. die Namen aller Listen-Eigenschaften eines Objekts, die
   * <code>int</code>-Werte aufnehmen koennen.<br>
   * <b>Bemerke:</b><br>
   * Wenn das Objekt keine Eigenschaften des angegebenen Typs besitzt, sollte
   * eine leere Property-Menge (z.B. {@link PropertySet}) zurueckgegeben werden
   * und <b>nicht</b> <code>null</code>!
   */
  public Properties getProperties(Class type);

  /**
   * Liefert die Namen aller Eigenschaften des Objekts.
   */
  public String[] getPropertyNames();

  /**
   * Liefert die Typen aller Eigenschaften des Objekts.
   */
  public PropertyType[] getPropertyTypes();

  /**
   * Liefert eine Eigenschaft des Objekts.
   * @param name Name der Eigenschaft
   * @return <code>null</code> falls das Objekt keine Eigenschaft mit dem
   *         Namen besitzt
   */
  public Property getProperty(String name);

  /**
   * Prueft, ob das Objekt eine bestimmte Eigenschaft besitzt.
   */
  public boolean containsProperty(String name);
}
