/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data.property;

import schmitzm.data.event.AbstractObjectTraceable;
import schmitzm.data.event.Invoker;
import schmitzm.data.event.ObjectCloseEvent;

/**
 * Diese Klasse stellt den Obertyp fuer eine Eigenschaft dar. Eine Eigenschaft
 * besteht aus 3 Teilen/Feldern:
 * <ul>
 * <li>ein Name, der die Eigenschaft (in einer Menge von Eigenschaften identifiziert)</li>
 * <li>eine Klasse, die den Typ der Eigenschaft beschreibt</code></li>
 * <li>ein Wert, als aktuelle Auspr?gung der Eigenschaft</li>
 * </ul>
 * Properties werden an verschiedenen Stellen eingesetzt. z.B. bei der Beschreibung
 * allgemeiner Xulu-Datenobjekte oder allgemeiner Modell-Ereignisse (Events).<br>
 * An eine Property koennen beliebig <code>ObjectListener</code> gekoppelt werden.
 * Sobald die Eigenschaft (?ber <code>setValue(..)</code>) ihren Wert aendert,
 * werden alle angeschlossenen <code>ObjectListener</code> mit einem
 * <code>ObjectChangeEvent</code> informiert.<br>
 * <br>
 * Es gibt verschiedenste Arten, wie eine Eigenschaft aufgebaut ist und wie sie
 * ihre Auspraegung(en) verwaltet (z.B. ein skalarer Wert oder eine ganze
 * Liste/Matrix). Entsprechend variieren auch die Zugriffsmethoden.
 * Aus diesem Grund spezifiziert diese Klasse noch keine Methoden, um auf die
 * Auspraegung der Eigenschaft (lesend oder schreibend) zuzugreifen.
 * Dies ist durch die speziellen Unterklassen zu implementieren.<br><br>
 * Der Zugriff auf die Property kann durch Zugriffsrechte kontrolliert werden.
 * Ein Zugriff (z.B. lesend oder schreibend) ist nur ueber ein entsprechendes
 * Zugriffsrecht ({@link Access}) moeglich. Da an dieser Stelle (in dieser Oberklasse)
 * die Art der Eigenschaft noch nicht spezifiziert ist, sind die entsprechenden
 * Methoden fuer das {@link Accessible}-Interface von den Unterklassen
 * zu implementieren.
  * <br><br>
  * Aus Vereinfachungsgruenden implementiert diese Klasse auch das Interface
  * {@link Properties}. Auf den ersten Blick macht dies zwar keinen Sinn, da
  * eine Property zunaechst ja immer genau eine Eigenschaft repraesentiert.
  * An vielen Stellen ist es jedoch praktisch, eine einzelne Eigenschaft unmittelbar
  * als (einelementige) Menge von Propertys behandeln zu koennen, ohne zunaechst
  * ein {@link PropertySet} erzeugen und die Eigenschaft einfuegen zu muessen.
 * @see schmitzm.data.event.ObjectListener
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class Property extends AbstractObjectTraceable implements Properties, Accessible {

  /**
   * Speichert den Namen der Eigenschaft.
   */
  protected String name = "";

  /**
   * Speichert den Daten-Typ der Eigenschaft.
   */
  protected PropertyType type = null;

  /**
   * Erzeugt eine Eigenschaft. Als Typ kann auch ein Build-in-Type angegeben
   * werden (z.B. <code>int.class</code>).
   * @param name Name der Eigenschaft
   * @param type Daten-Typ der Eigenschaft
   */
  public Property(String name, PropertyType type) {
    this.name = name;
    this.type = type;
  }

  /**
   * Liefert den Namen der Eigenschaft.
   */
  public String getName() {
    return this.name;
  }

  /**
   * Liefert die Klasse des Daten-Typs, der in der Eigenschaft gespeichert werden kann.
   */
  public Class getType() {
    return getPropertyType().getType();
  }

  /**
   * Liefert den Daten-Typ, der in der Eigenschaft gespeichert werden kann.
   */
  public PropertyType getPropertyType() {
    return type;
  }

  /**
   * Checkt, ob die Property einen Wert vom Typ <code>type</code> aufnehmen
   * kann.
   * @see schmitzm.data.property.PropertyType#isValid(Class)
   */
  public boolean isValid(Class type) {
    return getPropertyType().isValid(type);
  }

  /**
   * Erzeugt einen neue Property, die von der Struktur her identisch ist
   * mit der Property. Der Inhalt der Property muss <b>nicht</b> kopiert werden.
   */
  public abstract Property cloneStructure();


  ///////////////////////////////////////////////////////////////////
  // Implementierung von Accessible
  ///////////////////////////////////////////////////////////////////
  /**
   * "Schliesst" die Eigenschaft. Diese Eigenschaft entzieht zunaecht saemtliche
   * noch aktiven Zugriffsrechte und informiert dann  alle angeschlossenen
   * <code>ObjectListener</code>. Anschliessend werden alle <code>ObjectListener</code>
   * von dem Objekt entfernt.<br>
   * Unterklassen sollten diese Methode ueberschreiben wenn weitere Aktionen
   * beim "Schliessen" einer Eigenschaft notwendig sind.
   */
  public void dispose() {
    disposeAllAccess();
    this.fireEvent( new ObjectCloseEvent( new Invoker(this) ));
    this.removeAllObjectListeners();
  }

  /**
   * Prueft, ob aktuell Zugriffsrechte auf das Objekt verteilt sind.
   */
  public boolean hasAccess() {
    return getAccessCount() > 0;
  }

  /**
   * Prueft, ob aktuell Zugriffsrechte einer bestimmten Art
   * auf das Objekt verteilt sind.
   */
  public boolean hasAccess(Class c) {
    return getAccessCount(c) > 0;
  }

  ///////////////////////////////////////////////////////////////////
  // Implementierung von Properties
  ///////////////////////////////////////////////////////////////////
  /**
   * Liefert immer 1.
   */
  public int getPropertyCount() {
    return 1;
  }

  /**
   * Liefert einen 1-elementigen Array, in dem nur die Eigenschaft selbst
   * enthalten ist.
   */
  public Property[] getProperties() {
    return new Property[] { this };
  }

  /**
   * Liefert <code>this</code>, falls, der angegebene Typ eine Oberklasse
   * dieser Property ist, oder eine Oberklasse des Typs dieser Property.
   * Andernfalls wird ein leeres {@link PropertySet} zurueckgegeben.
   */
  public Properties getProperties(Class type) {
    if ( type.isAssignableFrom( this.getType() ) ||
         type.isAssignableFrom( this.getClass() ) )
      return this;
    return new PropertySet();
  }

  /**
   * Liefert einen 1-elementigen Array, in dem nur der Name der Eigenschaft selbst
   * enthalten ist.
   */
  public String[] getPropertyNames() {
    return new String[] { getName() };
  }

  /**
   * Liefert einen 1-elementigen Array, in dem nur der Typ der Eigenschaft
   * selbst enthalten ist.
   */
  public PropertyType[] getPropertyTypes() {
    return new PropertyType[] { getPropertyType() };
  }

  /**
   * Liefert die Eigenschaft selbst (<code>this</code>), falls deren Name
   * angegeben wird, ansonsten <code>null</code>.
   */
  public Property getProperty(String name) {
    return containsProperty(name) ? this : null;
  }

  /**
   * Liefert <code>true</code>, falls der Name der Eigenschaft selbst
   * angegeben wird.
   */
  public boolean containsProperty(String name) {
    return name.equals(getName());
  }

}
