/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data.property;

/**
 * Diese Klasse stellt ein Recht auf Lesezugriff fuer eine
 * {@link ValueProperty} dar.
 * Hierzu implementiert sie die entsprechenden <code>getter</code>-Methoden.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class PropertyReadAccess extends Access {
  private ValueProperty property       = null;

  /**
   * Erzeugt ein neues Lesezugriffsrecht. Die Anzahl an Zugriffen fuer den
   * Besitzer ist unbegrenzt.
   * @param object Instanz von <code>ValueProperty</code> auf die sich das Recht bezieht
   * @param owner  Besitzer des Rechts
   * @see schmitzm.data.property.ValueProperty
   * @exception schmitzm.data.property.AccessViolationException falls das
   *            angegebene Objekt <b>keine</b> <code>ValueProperty</code> ist
   */
  public PropertyReadAccess(Accessible object, Object owner) {
    this(object,owner,UNLIMITED_ACCESSTIMES);
  }

  /**
   * Erzeugt ein neues Lesezugriffsrecht.
   * @param object Instanz von <code>ValueProperty</code> auf die sich das Recht bezieht
   * @param owner  Besitzer des Rechts
   * @param maxAccessTimes Anzahl an (Methoden-)Zugriffen, die der Rechtebesitzer
   *                       taetigen darf, bevor das Recht automatisch entzogen wird
   * @see schmitzm.data.property.ValueProperty
   * @exception schmitzm.data.property.AccessViolationException falls das
   *            angegebene Objekt <b>keine</b> <code>ValueProperty</code> ist
   */
  public PropertyReadAccess(Accessible object, Object owner, int maxAccessTimes) {
    super(object,owner,maxAccessTimes);
    if (!(object instanceof ValueProperty))
      throw new AccessViolationException(this,this.getClass().getName()+" can only be instantiated for ValueProperties");
    this.property = (ValueProperty)object;
  }

  /**
   * Liefert den/einen Wert der Eigenschaft.<br>
   * <code>coords</code> spezifizieren (optional) die "Koordinaten", an denen
   * der Wert in der Eigenschaft zu finden ist.
   * Fuer skalare Eigenschaften darf/braucht diese Angabe nicht gemacht zu
   * werden. Fuer Listen und 1-dim. Matrizen (Arrays) darf nur ein Wert
   * angegeben werden. Fuer mehr-dimensionale Matrizen muessen entsprechend
   * der Dimension mehr Koordinaten angegeben werden (als einzelne Parameter oder
   * als ein Array). z.B. fuer 2-dim. Matrix:<br>
   * <code>getValue(10,13)</code> oder <code>getValue( new int[] {10,13} )</code><br>
   * <br>
   * Kann (seit JDK1.5.0) auch genutzt werden, wenn es sich bei der Property um einen
   * Build-in-Type handelt.
   * @param coords optionale Koordinaten
   * @exception schmitzm.data.property.AccessViolationException falls
   *            das Leserecht auf der Property nicht mehr gueltig ist
   */
  public Object getValue(int... coords) {
    checkDisposed();
    Object ret = this.property.getValue(coords);
    incAndCheckMaxAccessTimesReached();
    return ret;
  }

  /**
   * Liefert den/einen Wert der Eigenschaft als <code>char</code>.
   * Kann nur genutzt werden, wenn der Objekt-Typ ein Build-In-Type ist.
   * @param coords (optionale) Koordinaten (siehe <code>getValue(..)</code>
   *               fuer Erlaeuterungen)
   * @exception java.lang.ClassCastException falls es sich nicht um eine
   *            <code>char</code>-Property handelt.
   * @exception schmitzm.data.property.AccessViolationException falls
   *            das Leserecht auf der Property nicht mehr gueltig ist
   * @see #getValue(int...)
   */
  public char getValueAsChar(int... coords) {
    return ((Character)getValue(coords)).charValue();
  }

  /**
   * Liefert den/einen aktuellen Wert der Eigenschaft als <code>short</code>.
   * Kann nur genutzt werden, wenn der Objekt-Typ ein Build-In-Type ist.
   * @param coords (optionale) Koordinaten (siehe <code>getValue(..)</code>
   *               fuer Erlaeuterungen)
   * @exception java.lang.ClassCastException falls es sich nicht um eine
   *            <code>short</code>-Property handelt.
   * @exception schmitzm.data.property.AccessViolationException falls
   *            das Leserecht auf der Property nicht mehr gueltig ist
   * @see #getValue(int...)
   */
  public short getValueAsShort(int... coords) {
    return ((Number)getValue(coords)).shortValue();
  }

  /**
   * Liefert den/einen aktuellen Wert der Eigenschaft als <code>byte</code>.
   * Kann nur genutzt werden, wenn der Objekt-Typ ein Build-In-Type ist.
   * @param coords (optionale) Koordinaten (siehe <code>getValue(..)</code>
   *               fuer Erlaeuterungen)
   * @exception java.lang.ClassCastException falls es sich nicht um eine
   *            <code>byte</code>-Property handelt.
   * @exception schmitzm.data.property.AccessViolationException falls
   *            das Leserecht auf der Property nicht mehr gueltig ist
   * @see #getValue(int...)
   */
  public byte getValueAsByte(int... coords) {
    return ((Number)getValue(coords)).byteValue();
  }

  /**
   * Liefert den/einen aktuellen Wert der Eigenschaft als <code>int</code>.
   * Kann nur genutzt werden, wenn der Objekt-Typ ein Build-In-Type ist.
   * @param coords (optionale) Koordinaten (siehe <code>getValue(..)</code>
   *               fuer Erlaeuterungen)
   * @exception java.lang.ClassCastException falls es sich nicht um eine
   *            <code>int</code>-Property handelt.
   * @exception schmitzm.data.property.AccessViolationException falls
   *            das Leserecht auf der Property nicht mehr gueltig ist
   * @see #getValue(int...)
   */
  public int getValueAsInt(int... coords) {
    return ((Number)getValue(coords)).intValue();
  }

  /**
   * Liefert den/einen aktuellen Wert der Eigenschaft als <code>long</code>.
   * Kann nur genutzt werden, wenn der Objekt-Typ ein Build-In-Type ist.
   * @param coords (optionale) Koordinaten (siehe <code>getValue(..)</code>
   *               fuer Erlaeuterungen)
   * @exception java.lang.ClassCastException falls es sich nicht um eine
   *            <code>long</code>-Property handelt.
   * @exception schmitzm.data.property.AccessViolationException falls
   *            das Leserecht auf der Property nicht mehr gueltig ist
   * @see #getValue(int...)
   */
  public long getValueAsLong(int... coords) {
    return ((Number)getValue(coords)).longValue();
  }

  /**
   * Liefert den/einen aktuellen Wert der Eigenschaft als <code>float</code>.
   * Kann nur genutzt werden, wenn der Objekt-Typ ein Build-In-Type ist.
   * @param coords (optionale) Koordinaten (siehe <code>getValue(..)</code>
   *               fuer Erlaeuterungen)
   * @exception java.lang.ClassCastException falls es sich nicht um eine
   *            <code>float</code>-Property handelt.
   * @exception schmitzm.data.property.AccessViolationException falls
   *            das Leserecht auf der Property nicht mehr gueltig ist
   * @see #getValue(int...)
   */
  public float getValueAsFloat(int... coords) {
    return ((Number)getValue(coords)).floatValue();
  }

  /**
   * Liefert den/einen aktuellen Wert der Eigenschaft als <code>double</code>.
   * Kann nur genutzt werden, wenn der Objekt-Typ ein Build-In-Type ist.
   * @param coords (optionale) Koordinaten (siehe <code>getValue(..)</code>
   *               fuer Erlaeuterungen)
   * @exception java.lang.ClassCastException falls es sich nicht um eine
   *            <code>double</code>-Property handelt.
   * @exception schmitzm.data.property.AccessViolationException falls
   *            das Leserecht auf der Property nicht mehr gueltig ist
   * @see #getValue(int...)
   */
  public double getValueAsDouble(int... coords) {
    return ((Number)getValue(coords)).doubleValue();
  }

  /**
   * Liefert den/einen aktuellen Wert der Eigenschaft als <code>boolean</code>.
   * Kann nur genutzt werden, wenn der Objekt-Typ ein Build-In-Type ist.
   * @param coords (optionale) Koordinaten (siehe <code>getValue(..)</code>
   *               fuer Erlaeuterungen)
   * @exception java.lang.ClassCastException falls es sich nicht um eine
   *            <code>boolean</code>-Property handelt.
   * @exception schmitzm.data.property.AccessViolationException falls
   *            das Leserecht auf der Property nicht mehr gueltig ist
   * @see #getValue(int...)
   */
  public boolean getValueAsBoolean(int... coords) {
    return ((Boolean)getValue(coords)).booleanValue();
  }

}
