/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data.property;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import schmitzm.data.ObjectStructure;
import schmitzm.data.ObjectStructureUtil;
import schmitzm.data.event.AbstractObjectTraceable;
import schmitzm.data.event.Invoker;
import schmitzm.data.event.ObjectChangeEvent;
import schmitzm.data.event.ObjectTraceable;

/**
 * Diese Klasse stellt eine Menge an Eigenschaften dar. Die einzelnen Properties
 * werden parallel in einer {@link java.util.Hashtable} und einem
 * {@link java.util.Vector} verwaltet. Die Hashtable dient dazu, die Properties
 * effizient ueber ihren Namen ansprechen zu koennen. Der Vector repraesentiert
 * eine Ordnung und wird deshalb verwendet, wenn alle Eigenschaften
 * angesprochen werden (z.B. {@link #getProperties()} oder {@link #getPropertyNames()}).<br>
 * Zu beachten ist, dass beim Entfernen einer Eigenschaft aus der Menge
 * <b>nicht</b> die komplette Eigenschaft (mittels {@link Property#dispose()})
 * geschlossen wird! Dies bietet den Vorteil dass
 * eine Eigenschaft einfach von einer Menge in eine andere verschoben werden kann.
 * Soll eine Eigenschaft (und vor allem ihre Ressourcen) komplett freigegeben
 * werden, muss dies manuell geschehen. Hierfuer liefern die remove-Methoden eine
 * Referenz auf die entfernte Eigenschaft zurueck.<br>
 * <code>PropertySet</code> implementiert das Interface {@link ObjectTraceable},
 * so dass Aenderungen an der Menge durch Listener verfolgt werden koennen.
 * Es werden jedoch <b>keine</b> Aenderungen an den in der Menge enthaltenen
 * Eigenschaften propagiert, sondern nur Aenderungen an der Eigenschaftsmenge
 * selbst, also Hinzufuegen oder Entfernen einer Property. In beiden Faellen
 * wird ein {@link ObjectChangeEvent} geworfen, wobei entweder
 * {@link ObjectChangeEvent#getNewValue()} (beim Entfernen) oder {@link ObjectChangeEvent#getOldValue()}
 * (beim Hinzufuegen) <code>null</code> liefert.
 * @see schmitzm.data.property.Property
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class PropertySet extends AbstractObjectTraceable implements DynamicProperties, Cloneable{
  private Hashtable contentHash = null;
  private Vector    contentVec  = null;

  /**
   * Erzeugt eine leere Menge von Eigenschaften aus einer Vorlage. Es wird jedoch
   * nur die Struktur der Vorlage kopiert, <b>nicht</b> die darin enthaltenen
   * Eigenschaften!!
   * @param propSet Vorlage fuer die Property-Menge
   */
  public PropertySet(PropertySet propSet) {
    this.contentHash =  (Hashtable)propSet.contentHash.clone();
    this.contentVec  =  (Vector)propSet.contentVec.clone();
  }

  /**
   * Erzeugt eine leere Menge von Eigenschaften.
   */
  public PropertySet() {
    this.contentHash = new Hashtable();
    this.contentVec  = new Vector();
  }

  /**
   * Erzeugt eine Kopie der Menge. Dabei wird jedoch nicht der Inhalt - also
   * die Properties selbst - kopiert, sondern nur das Hashtable-Objekt!!
   */
  public Object clone() {
    return new PropertySet(this);
  }

  /**
   * Liefert die Anzahl an Eigenschaften.
   */
  public int getPropertyCount() {
    return contentVec.size();
  }

  /**
   * Liefert die Eigenschaften der Menge als Array.
   */
  public Property[] getProperties() {
//    return (Property[])contentHash.values().toArray(new Property[] {});
    return (Property[])contentVec.toArray(new Property[] {});
  }

  /**
   * Liefert alle Eigenschaften, die einem bestimmten
   * Typ angehoeren. Als Typ kann sowohl ein bestimmter <code>Property</code>-Untertyp
   * angegeben werden (z.B. {@link ListProperty}), als auch ein Objekt-Typ (Typ
   * den eine Property aufnehmen kann; z.B. <code>int.class</code>).<br>
   * Da die Methode wiederum eine <code>Properties</code>-Instanz liefert,
   * lassen sich recht einfach verschiedene Filter hintereinander anwenden:<br>
   * <br>
   * <center><code>
   * Properties.getProperties(ListProperty.class).getProperties(int.class).getPropertyNames()
   * </code></center>
   * <br>
   * liefert z.B. die Namen aller Listen-Eigenschaften, die <code>int</code>-Werte
   * aufnehmen koennen.
   */
  public Properties getProperties(Class type) {
    PropertySet set = new PropertySet();
    for(Enumeration e=contentVec.elements();e.hasMoreElements();) {
      Property p = (Property)e.nextElement();
      // Ist der angegebene Typ eine Property (oder ein Untertyp),
      // wird direkt auf die Art der Property geprueft (1), ansonsten auf den
      // Typ, den die Property beinhalten kann (2)
      if ( Property.class.isAssignableFrom(type) ) {
        // (1) type Superklasse der Property-Art?
        if ( type.isAssignableFrom( p.getClass() ) )
          set.addProperty(p);
      } else {
        // (2) type Superklasse des Property-Typs?
        if ( type.isAssignableFrom( p.getType() ) )
          set.addProperty(p);
      }
    }

    return set;
  }

  /**
   * Liefert die Namen aller Eigenschaften.
   */
  public String[] getPropertyNames() {
//    return (String[])content.keySet().toArray(new String[0]);
    String[] name = new String[ getPropertyCount() ];
    Enumeration e = contentVec.elements();
    for(int i=0; e.hasMoreElements(); i++)
      name[i] = ((Property)e.nextElement()).getName();
    return name;
  }

  /**
   * Liefert die Typen aller Eigenschaften.
   */
  public PropertyType[] getPropertyTypes() {
    PropertyType[]  type = new PropertyType[ getPropertyCount() ];
    Enumeration e = contentVec.elements();
    for(int i=0; e.hasMoreElements(); i++)
      type[i] = ((Property)e.nextElement()).getPropertyType();
    return type;
  }

  /**
   * Liefert eine bestimmte Eigenschaft der Menge.
   * @param name Name der Eigenschaft
   * @return <code>null</code> falls die Menge keine Eigenschaft dieses
   *         Namens enthaelt
   */
  public Property getProperty(String name) {
    return (Property)contentHash.get(name);
  }

  /**
   * Prueft, ob die Menge eine bestimmte Eigenschaft enthaelt.
   * @param name Name der Eigenschaft
   */
  public boolean containsProperty(String name) {
    return getProperty(name)!=null;
  }

  /**
   * Fuegt der Menge eine Eigenschaft hinzu. Dabei wird auf Eindeutigkeit
   * des Namens geachtet!
   * @param prop Eigenschaft
   * @exception schmitzm.data.property.PropertyException falls bereits
   *            eine Eigenschaft dieses Namens in der Menge vorhanden ist
   */
  public void addProperty(Property prop) {
    if ( containsProperty(prop.getName()) )
      throw new PropertyException(prop,"PropertySet already contains a Property named '".concat(prop.getName()).concat("'"));
    contentHash.put(prop.getName(),prop);
    contentVec.add(prop);
    fireEvent(new ObjectChangeEvent(new Invoker(this),null,prop));
  }

  /**
   * Entfernt eine Eigenschaft aus der Menge.
   * @param name Name der Eigenschaft
   * @return Die entferne Eigenschaft oder <code>null</code>, falls das
   *         die Menge keine Eigenschaft mit dem Namen enthaelt
   */
  public Property removeProperty(String name) {
    Property prop = (Property)contentHash.remove(name);
    contentVec.remove(prop);
    fireEvent(new ObjectChangeEvent(new Invoker(this),prop,null));
    return prop;
  }

  /**
   * Entfernt eine Eigenschaft aus der Menge.
   * Falls die Menge eine solche Eigenschaft nicht besitzt, passiert nichts.
   * @param prop zu entfernende Eigenschaft
   * @return Die entferne Eigenschaft oder <code>null</code> falls das Objekt
   *         die Eigenschaft nicht besitzt
   */
  public Property removeProperty(Property prop) {
    return removeProperty(prop.getName());
  }

  /**
   * Entfernt Eigenschaften aus der Menge. Falls die Menge eine dieser
   * Eigenschaften nicht enthaelt, passiert fuer die entsprechende Eigenschaft
   * nichts.
   * @param prop zu entfernende Eigenschaft
   * @return Die entfernten Eigenschaften (kann eine leere Eigenschaftsmenge
   *         sein, wenn die Menge <u>keine</u> der uebergebenen Eigenschaften
   *         besitzt.
   */
  public Properties removeProperties(Properties prop) {
    Property    p   = null;
    PropertySet set = new PropertySet();
    for (int i=0; i<prop.getProperties().length; i++)
      if ( (p = removeProperty(prop.getProperties()[i]))!=null )
        set.addProperty(p);
    return set;
  }

  ///////////////////////////////////////////////////////////////////
  // Implementierung von ObjectStructure
  ///////////////////////////////////////////////////////////////////
  /**
   * Liefert <code>false</code>.
   */
  public boolean isStructureNamed() {
    return false;
  }

  /**
   * Liefert <code>null</code>.
   */
  public String getStructureName() {
    return null;
  }

  /**
   * Liefert die Anzah1 an Properties in der Menge.
   * @see #getPropertyCount()
   */
  public int getAttrCount() {
    return getPropertyCount();
  }

  /**
   * Liefert eine Liste, die genau so viele Elemente enthaelt, wie die
   * Eigenschaftsmenge Propertys hat. Alle Listen-Elemente sind vom Typ
   * {@link ObjectStructure}.
   */
  public Enumeration<ObjectStructure> getAttrTypes() {
    return contentVec.elements();
  }

  /**
   * Vergleicht die Struktur der Property mit einer anderen auf Gleichheit.
   * @param object eine anderes {@link ObjectStructure}-Objekt
   * @return <code>false</code> falls das uebergebene Objekt keine
   *         {@link ObjectStructure} ist
   * @see ValuePropertyType#equalsInStructure(Object)
   */
  public boolean equalsInStructure(Object object) {
    if ( !(object instanceof ObjectStructure) )
      return false;
    return ObjectStructureUtil.compareObjectStructures(this,(ObjectStructure)object) == ObjectStructureUtil.EQUAL;
  }

  /**
   * Liefert <code>false</code>, da eine Eigenschaftsmenge von jeder Eigenschaft
   * immer nur eine Instanz enthaelt.
   */
  public boolean containsMultipleValues() {
    return false;
  }

}
