/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data.property;

import schmitzm.temp.BaseTypeUtil;

/**
 * Diese Klasse stellt allgemein den (strukturellen) Typ einer Eigenschaft dar.
 * Sie selbst implementiert lediglich die Speicherung eines Klassennamens.
 * Entscheidender sind die abgeleiteten Untertypen:<br>
 * Fuer alle Wert-Eigenschaften ist im {@linkplain ValuePropertyType Typ} z.B.
 * (falls vorhanden) die Unterstruktur gespeichert.
 * @see ValuePropertyType
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class PropertyType {
  /**
   * Speichert den eigentlichen Typ der Property.
   */
  protected Class type = null;

  /**
   * Erzeugt einen neuen Eigenschaftstyp
   * @param type Objekt-Typ den die Eigenschaft aufnehmen kann
   */
  public PropertyType(Class type) {
    this.type = type;
  }

  /**
   * Liefert den Typ, den die Eigenschaft aufnehmen kann.
   */
  public Class getType() {
    return this.type;
  }

  /**
   * Checkt, ob die Property einen Wert vom Typ <code>type</code> aufnehmen
   * kann. Dies ist der Fall, wenn <code>type</code> dieselbe oder eine
   * Subklasse des Property-Typs ist.<br>
   * Ausserdem gelten fuer Build-In-Types auch die entsprechenden Objekte
   * (z.B. <code>Integer</code>-Instanz fuer in <code>int</code>-Property)
   * als gueltig.
   * @see schmitzm.data.property.Property#isValid(Class)
   * @see schmitzm.temp.BaseTypeUtil#isBuildInType(Class)
   * @see schmitzm.temp.BaseTypeUtil#isEquivalent(Class,Class)
   */
  public boolean isValid(Class type) {
    return getType().isAssignableFrom(type) ||
           BaseTypeUtil.isBuildInType(getType()) && BaseTypeUtil.isEquivalent(getType(),type);
//    return getType().equals(char.class)    && Character.class.isAssignableFrom(type) ||
//           getType().equals(short.class)   && Short.class.isAssignableFrom(type) ||
//           getType().equals(byte.class)    && Byte.class.isAssignableFrom(type) ||
//           getType().equals(int.class)     && Integer.class.isAssignableFrom(type) ||
//           getType().equals(long.class)    && Long.class.isAssignableFrom(type) ||
//           getType().equals(float.class)   && Float.class.isAssignableFrom(type) ||
//           getType().equals(double.class)  && Double.class.isAssignableFrom(type) ||
//           getType().equals(boolean.class) && Boolean.class.isAssignableFrom(type) ||
//           getType().isAssignableFrom(type);
  }

  /**
   * Liefert den Namen der Typ-Klasse als Beschreibung des Eigenschaftstyp.
   */
  public String toString() {
    return getType().getName();
  }
}
