/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data.property;

import java.util.Vector;

import schmitzm.lang.LangUtil;
import schmitzm.temp.BaseTypeUtil;

/**
 * Diese Klasse beinhaltet statische Hilfsmethoden fuer das Arbeiten mit
 * {@link Property} und {@link Properties}.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class PropertyUtil {

  /**
   * Liefert eine Eigenschaft eines {@link Properties}-Objekt.
   * Enthaelt das {@link Properties}-Objekt keine Eigenschaft mit dem
   * angegebenen Namen, wird dieser sukzessive verkuerzt (Punkt-Notation)
   * bis im Objekt eine entsprechende Eigenschaft gefunden wird.
   * In diesem Fall muss es sich beim Wert dieser Eigenschaft
   * wiederum um ein {@link Properties}-Objekt handeln, um den "abgeschnittenen"
   * Teil des Eigenschaft-Names rekursiv zu verarbeiten.
   * @param propObject {@link Properties}-Objekt
   * @param propSuffix Name der Eigenschaft (bzw. mehrere Eigenschaften
   *                   durch "." getrennt)
   * @return {@code null} wenn die Eigenschaft nicht gefunden werden kann
   */
  public static Property getProperty(Properties propObject, String propSuffix) {
    return getProperty(propObject, propSuffix, '.');
  }

  /**
   * Liefert den Wert eine Eigenschaft. Enthaelt das {@link Properties}-Objekt
   * keine Eigenschaft mit dem angegebenen Namen, wird dieser sukzessive
   * verkuerzt (Punkt-Notation) bis im Objekt eine entsprechende Eigenschaft
   * gefunden wird. In diesem Fall muss es sich beim Wert dieser Eigenschaft
   * wiederum um ein {@link Properties}-Objekt handeln, um den "abgeschnittenen"
   * Teil des Eigenschaft-Names rekursiv zu verarbeiten.
   * @param propObject {@link Properties}-Objekt
   * @param propSuffix Name der Eigenschaft (bzw. mehrere Eigenschaften
   *                   getrennt durch das {@code sep}-Zeichen)
   * @param sep        Trenn-Zeichen, mit dem in {@link propSuffix} mehere Eigenschaften
   *                   von einander getrennt sind
   * @return {@code null} wenn die Eigenschaft nicht gefunden werden kann
   */
  public static Property getProperty(Properties propObject, String propSuffix, char sep) {
    if ( propSuffix == null )
      propSuffix = "";

    String propName = propSuffix;
    int[]  arrayCoord = new int[0];
    propSuffix = "";
    while ( !propName.equals("") ) {
      Property prop = propObject.getProperty(propName);
      if ( prop == null ) {
        arrayCoord = getArrayCoordinates(propName);
        if ( arrayCoord.length > 0 ) {
          // cut the array coordinates
          propName = cutArrayCoordinates(propName);
          prop     = propObject.getProperty(propName);
        }
      }
      // Property found
      if ( prop != null ) {
        // no more property names exist >> the object is the requested property (Juhuuuu!)
        if ( propSuffix.equals("") )
          return prop;

        // more property names exist >> property must be a ScalarProperty containing
        //                              a Properties to perform the property suffix
        if ( prop instanceof ValueProperty ) {
          Object propValue = ((ValueProperty)prop).getOneTimeReadAccess().getValue(arrayCoord);
          if ( propValue instanceof Properties )
            return getProperty((Properties)propValue, propSuffix, sep);
        }
      }

      // cut at last "." and try the prefix as property name
      int tLastDotIdx = propName.lastIndexOf(sep);
      if ( !propSuffix.equals("") )
        propSuffix += sep;
      propSuffix += propName.substring(tLastDotIdx+1);
      propName   =  tLastDotIdx > 0 ? propName.substring(0,tLastDotIdx) : "";
    }
    // Property not found in object
    return null;
  }

  /**
   * Ermittelt die am Ende eines Strings in runden Klammern spezifierten
   * Koordinaten.
   * @param objectSpec ein String
   * @return int[0], wenn der String keine Klammerung am Ende enthaelt.
   */
  public static int[] getArrayCoordinates(String objectSpec) {
    objectSpec = objectSpec.trim();
    int[] arrayCoord = new int[0];
    if ( objectSpec.endsWith(")") && objectSpec.contains("(")) {
      int openIdx = objectSpec.lastIndexOf("(");
      String[] arrayCoordStr = objectSpec.substring(openIdx+1,objectSpec.length()-1).split(",");
      arrayCoord = new int[arrayCoordStr.length];
      for (int i=0; i<arrayCoord.length; i++)
        arrayCoord[i] = Integer.parseInt(arrayCoordStr[i].trim());
    }
    return arrayCoord;
  }

  /**
   * Entfernt eine etwaige Koordinaten-Definition (in runden Klammern) am
   * Ende eines Strings.
   * @param objectSpec ein String
   * @return den original String, wenn dieser keine Koordinaten-Definition
   *         enthaelt
   */
  public static String cutArrayCoordinates(String objectSpec) {
    if ( objectSpec.endsWith(")") && objectSpec.contains("("))
      return objectSpec.substring(0, objectSpec.lastIndexOf("("));
    return objectSpec;
  }


  /**
   * Wraps a {@link Property} to an appropriate java object.
   * <ul>
   *   <li>{@link ScalarProperty}: wrapped to its (single) value</li>
   *   <li>{@link ListProperty}: wrapped to an appropriate {@link Vector} which
   *                             contains the {@link ListProperty} values</li>
   *   <li>{@link MatrixProperty}: wrapped to an appropriate native {@code Object[]..[]}
   *                               which contains the {@link MatrixProperty} values</li>
   * </ul>
   * @param prop   a property
   * @param coord  <b>if specified, the appropiate element of the {@link ListProperty}
   *               or {@link MatrixProperty} is returned, instead of the complete
   *               list or matrix!</b>
   * @return the Property itself, if no wrap can be performed
   */
  public static Object getPropertyValue(Property prop, int... coord) {
    if ( coord == null )
      coord = new int[0];

    //*********** ScalarProperty **************
    if ( prop instanceof ScalarProperty ) {
      return ((ScalarProperty)prop).getOneTimeReadAccess().getValue();
    }

    //*********** ListProperty **************
    if ( prop instanceof ListProperty ) {
      // if coordinates are given, the list element is returned
      if (coord.length > 0)
        return ((ListProperty)prop).getOneTimeReadAccess().getValue(coord);
      // return all list elements as Vector
      ListPropertyReadAccess prop_RA = ((ListProperty)prop).getReadAccess(null);
      Vector vector = new Vector( prop_RA.getCount() );
      for (int i=0; i<prop_RA.getCount(); i++)
        vector.add( prop_RA.getValue(i) );
      prop_RA.release();
      return vector;
    }

    //*********** MatrixProperty **************
    if ( prop instanceof MatrixProperty ) {
      // if coordinates are given, the matrix element is returned
      if (coord.length > 0)
        return ((MatrixProperty)prop).getOneTimeReadAccess().getValue(coord);

      // challenge: MatrixProperty is generic and its dimension and
      //            data type is not fixed
      //            --> the creation of the target native array
      //                must be realised generic too
      MatrixProperty     matrixProp = (MatrixProperty)prop;
      PropertyReadAccess prop_RA    = (matrixProp).getReadAccess(null);
      int[]              dimSizes   = matrixProp.getSize(); // Size of the dimensions
      Object             array      = LangUtil.createArray( dimSizes );
      // Calculate the total number of elements in the matrix
      int elemCount = 1;
      for ( int dimSize : dimSizes )
        elemCount *= dimSize;

      // put the elements in the array
      coord = new int[ dimSizes.length ];
      for ( int i=0; i<elemCount; i++ ) {
        // set the element at coordinate tCoord
        LangUtil.setArrayValue(
            array,
            prop_RA.getValue( coord ),
            coord
        );

        // increment the coordinate for the element
        for (int d = 1; d < dimSizes.length; d++)
          if (++coord[d] < dimSizes[d])
            break; // Coordinates in dimensionen > d not influenced by the increment
          else
            coord[d] = 0; // Initialize coordinate in dimension d and continue
                          // to increment in dimension d+1 erhoehen
      }
      prop_RA.release();
      return array;
    }

    return prop;
  }

  /**
   * Sets the value of a {@link Property} with an appropriate java object.
   * <ul>
   *   <li>{@link ScalarProperty}: sets the property's (single) value to {@code aValue}</li>
   *   <li>{@link ListProperty}: clears the list and fills it with the
   *                             elements of {@code aValue} (must be an {@code Object[]}
   *                             or an {@link Iterable})</li>
   *   <li>{@link MatrixProperty}: fills the matrix with the elements of
   *                               {@code aValue} (must be an {@code Object[]})</li>
   * </ul>
   * @param prop   a property
   * @param value  a value the property is set to
   * @param coord  <b>if specified, only the appropiate element of the {@link ListProperty}
   *               or {@link MatrixProperty} is set with {@code aValue},
   *               instead of the complete list or matrix!</b>
   */
  public static void setPropertyValue(ValueProperty prop, Object value, int... coord) {
    if ( coord == null )
      coord = new int[0];

    //*********** ScalarProperty **************
    if (prop instanceof ScalarProperty)
      ((ScalarProperty)prop).getOneTimeWriteAccess().setValue(
          convertToPropertyType(value, prop.getType())
      );

    //*********** ListProperty **************
    if (prop instanceof ListProperty) {
      // if coordinates are given, the matrix element is set
      if (coord.length > 0) {
        ((ListProperty)prop).getOneTimeWriteAccess().setValue(
            convertToPropertyType(value, prop.getType()),
            coord
        );
        return;
      }
      ListPropertyWriteAccess prop_WA = ((ListProperty)prop).getWriteAccess(null);
      // Clear the ListProperty
      prop_WA.removeAll();
      // Put all elements of aValue into the ListProperty
      if ( value instanceof String ) {
        // String vom Format "{ ... , ... , ... }", dann String
        // splitten
        String sepString = ((String) value).trim();
        if ( sepString.startsWith("{") && sepString.endsWith("}") )
          value = sepString.substring(1, sepString.length()-1).split(",");
      }
      if ( value instanceof Object[] )
        for (Object arrayValue : (Object[])value)
          prop_WA.addValue( convertToPropertyType(arrayValue, prop.getType()) );
      else if ( value instanceof Iterable )
        for (Object element : (Iterable)value)
          prop_WA.addValue( convertToPropertyType(element, prop.getType()) );
      prop_WA.release();
    }
    //*********** MatrixProperty **************
    if (prop instanceof MatrixProperty) {
      // if coordinates are given, the matrix element is set
      if (coord.length > 0) {
        ((MatrixProperty)prop).getOneTimeWriteAccess().setValue(
            convertToPropertyType(value, prop.getType()),
            coord
        );
        return;
      }

      MatrixProperty      matrixProp = (MatrixProperty)prop;
      PropertyWriteAccess prop_WA    = (matrixProp).getWriteAccess(null);
      int[]               dimSizes   = LangUtil.getArrayLength(value); // Size of the dimensions
      if ( dimSizes.length != matrixProp.getDimension() )
        throw new UnsupportedOperationException("Dimension of MatrixProperty differs from specified array: "+matrixProp.getDimension()+" <> "+dimSizes.length);

      // Calculate the total number of elements in the matrix
      int elemCount = 1;
      for ( int dimSize : dimSizes )
        elemCount *= dimSize;

      // put the array elements in the MatrixProperty
      coord = new int[ dimSizes.length ];
      for ( int i=0; i<elemCount; i++ ) {
        // set the MatrixProperty at coordinate tCoord
        prop_WA.setValue(
          convertToPropertyType(LangUtil.getArrayValue( value, coord ), prop.getType()),
          coord
        );

        // increment the coordinate for the element
        for (int d = 1; d < dimSizes.length; d++)
          if (++coord[d] < dimSizes[d])
            break; // Coordinates in dimensionen > d not influenced by the increment
          else
            coord[d] = 0; // Initialize coordinate in dimension d and continue
                          // to increment in dimension d+1 erhoehen
      }
      prop_WA.release();
    }
  }

  private static Object convertToPropertyType(Object value, Class neededType) {
    // try to set the object directly to the property
    if ( !neededType.isInstance(value) ) {
      if ( value instanceof String )
        value = BaseTypeUtil.convertFromString((String)value, neededType);
    }
    return value;
  }

}
