/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data.property;

/**
 * Diese Klasse stellt ein Recht auf Schreibzugriff fuer eine
 * {@link ValueProperty} dar.
 * Hierzu implementiert sie die entsprechenden <code>setter</code>-Methoden.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class PropertyWriteAccess extends PropertyReadAccess {
  private ValueProperty property = null;

  /**
   * Erzeugt ein neues Schreibzugriffsrecht. Die Anzahl an Zugriffen fuer den
   * Besitzer ist unbegrenzt.
   * @param object Instanz von <code>ValueProperty</code> auf die sich das Recht bezieht
   * @param owner  Besitzer des Rechts
   * @see schmitzm.data.property.ValueProperty
   * @exception schmitzm.data.property.AccessViolationException falls das
   *            angegebene Objekt <b>keine</b> <code>ValueProperty</code> ist
   */
  public PropertyWriteAccess(Accessible object, Object owner) {
    this(object,owner,UNLIMITED_ACCESSTIMES);
  }

  /**
   * Erzeugt ein neues Schreibzugriffsrecht.
   * @param object Instanz von <code>ValueProperty</code> auf die sich das Recht bezieht
   * @param owner  Besitzer des Rechts
   * @param maxAccessTimes Anzahl an (Methoden-)Zugriffen, die der Rechtebesitzer
   *                       taetigen darf, bevor das Recht automatisch entzogen wird
   * @see schmitzm.data.property.ValueProperty
   * @exception schmitzm.data.property.AccessViolationException falls das
   *            angegebene Objekt <b>keine</b> <code>ValueProperty</code> ist
   */
  public PropertyWriteAccess(Accessible object, Object owner, int maxAccessTimes) {
    super(object,owner,maxAccessTimes);
    if (!(object instanceof ValueProperty))
      throw new AccessViolationException(this,this.getClass().getName()+" can only be instantiated for ValueProperties");
    property = (ValueProperty)object;
  }

  /**
   * Setzt den/einen Wert der Eigenschaft.<br>
   * <code>coords</code> spezifizieren (optional) die "Koordinaten", an denen
   * der Wert in der Eigenschaft zu finden ist.
   * Fuer skalare Eigenschaften darf/braucht diese Angabe nicht gemacht zu
   * werden. Fuer Listen und 1-dim. Matrizen (Arrays) darf nur ein Wert
   * angegeben werden. Fuer mehr-dimensionale Matrizen muessen entsprechend
   * der Dimension mehr Koordinaten angegeben werden (als einzelne Parameter oder
   * als ein Array). z.B. fuer 2-dim. Matrix:<br>
   * <code>getValue(10,13)</code> oder <code>getValue( new int[] {10,13} )</code><br>
   * <br>
   * Kann (seit JDK1.5.0) auch genutzt werden, wenn es sich bei der Property um einen
   * Build-in-Type handelt.
   * Iniziiert ein <code>ObjectChangeEvent</code> fuer die Property.
   * @param value  neuer Wert fuer die Eigenschaft
   * @param coords optionale Koordinaten
   * @exception java.lang.ClassCastException falls das angegebene Objekt nicht
   *            zum Property-Typ passt
   * @exception schmitzm.data.property.AccessViolationException falls
   *            das Schreibrecht auf der Property nicht mehr gueltig ist
   */
  public void setValue(Object value, int... coords) {
    checkDisposed();
    this.property.setValue(value,coords);
    incAndCheckMaxAccessTimesReached();
  }

}
