/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data.property;

import schmitzm.data.event.Invoker;
import schmitzm.data.event.ObjectChangeEvent;

/**
 * Diese Klasse stellt eine Eigenschaft dar, welche genau einen Wert
 * (in Form eines Objekts) darstellt.<br>
 * Der Zugriff auf die Property kann durch Zugriffsrechte kontrolliert werden.
 * Ein Zugriff (lesend oder schreibend) ist nur ueber ein entsprechendes
 * Zugriffsrecht ({@link Access}) moeglich. Standardmaessig ist unbegrenzter
 * Zugriff eingestellt. Durch Angabe von {@link ValuePropertyAccessParameters}
 * kann jedoch benutzerdefiniert festgelegt werden, wie viele Objekte gleichzeitig
 * Lese- und/oder Schreibzugriff erlangen duerfen.
 * @see schmitzm.data.property.ValueProperty#getReadAccess(Object)
 * @see schmitzm.data.property.ValueProperty#getWriteAccess(Object)
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ScalarProperty extends ValueProperty {
  /**
   * Speichert die Auspraegung der Eigenschaft.
   */
  protected Object value = null;

  /**
   * Erzeugt eine Eigenschaft.
   * @param name Bezeichnung der Eigenschaft
   * @param elementType Daten-Typ, den die Eigenschaft aufnehmen kann
   * @param params Parameter fuer die Verwaltung des Zugriffsrechts
   * @see schmitzm.data.property.ValuePropertyAccessParameters
   */
  public ScalarProperty(String name, ValuePropertyType elementType, ValuePropertyAccessParameters params) {
    super( name, elementType, params );
  }

  /**
   * Erzeugt eine Eigenschaft. Der Zugriff auf die Eigenschaft ist
   * uneingeschraengt (beliebig viele Lese- und beliebig viele Schreibrechte).
   * @param name Bezeichnung der Eigenschaft
   * @param elementType Daten-Typ, den die Eigenschaft aufnehmen kann
   * @see schmitzm.data.property.ValuePropertyAccessParameters#UNLIMITED_ACCESS
   */
  public ScalarProperty(String name, ValuePropertyType elementType) {
    this( name, elementType, ValuePropertyAccessParameters.UNLIMITED_ACCESS );
  }

  /**
   * Erzeugt eine Eigenschaft. Als Typ kann auch ein Build-in-Type angegeben
   * werden (z.B. int.class).
   * @param name Bezeichnung der Eigenschaft
   * @param elementType Daten-Typ, den die Eigenschaft aufnehmen kann
   * @param params Parameter fuer die Verwaltung des Zugriffsrechts
   * @see schmitzm.data.property.ValuePropertyAccessParameters
   */
  public ScalarProperty(String name, Class elementType, ValuePropertyAccessParameters params) {
    super( name, elementType, params );
  }

  /**
   * Erzeugt eine Eigenschaft. Als Typ kann auch ein Build-in-Type angegeben
   * werden (z.B. <code>int.class</code>). Der Zugriff auf die Eigenschaft ist
   * uneingeschraengt (beliebig viele Lese- und beliebig viele Schreibrechte).
   * @param name Bezeichnung der Eigenschaft
   * @param type Daten-Typ der Eigenschaft
   * @see schmitzm.data.property.ValuePropertyAccessParameters#UNLIMITED_ACCESS
   */
  public ScalarProperty(String name, Class type) {
    this(name,type,ValuePropertyAccessParameters.UNLIMITED_ACCESS);
  }

  /**
   * Erzeugt eine Eigenschaft. Der Datentyp ist implizit durch das Wert-Objekt
   * gegeben. Kann auch fuer Build-in-Werte (<code>int</code>, <code>char</code>, ...)
   * verwendet werden.
   * @param name  Bezeichnung der Eigenschaft
   * @param value Initialer Wert der Eigenschaft
   * @param params Parameter fuer die Verwaltung des Zugriffsrechts
   * @see schmitzm.data.property.ValuePropertyAccessParameters
   */
  public ScalarProperty(String name, Object value, ValuePropertyAccessParameters params) {
    this(name,ValuePropertyType.createValuePropertyType(value),params);
    setValue(value);
  }

  /**
   * Erzeugt eine Eigenschaft. Der Datentyp ist implizit durch das Wert-Objekt
   * gegeben. Kann auch fuer Build-in-Werte (<code>int</code>, <code>char</code>, ...)
   * verwendet werden. Der Zugriff auf die Eigenschaft ist
   * uneingeschraengt (beliebig viele Lese- und beliebig viele Schreibrechte).
   * @param name  Bezeichnung der Eigenschaft
   * @param value Initialer Wert der Eigenschaft
   * @see schmitzm.data.property.ValuePropertyAccessParameters#UNLIMITED_ACCESS
   */
  public ScalarProperty(String name, Object value) {
    this(name,value,ValuePropertyAccessParameters.UNLIMITED_ACCESS);
  }

  /**
   * Liefert eine neue, leerer <code>ScalarProperty</code> mit identischem
   * Namen, Typ und gleichen Zugriffsparemetern.
   */
  public ScalarProperty cloneStructure() {
    return new ScalarProperty( getName(), getPropertyType(), getAccessParameters() );
  }

  /**
   * Zerstoert die Property und den Property-Wert. Handelt es sich bei dem
   * Wert der Property wiederum um eine Property, wird auch deren
   * <code>dispose()</code>-Methode aufgerufen.
   */
  public void dispose() {
    if ( value instanceof Property )
      ((Property)value).dispose();
    value = null;
    super.dispose();
  }

  ///////////////////////////////////////////////////////////////////
  // Implementierung von ValueProperty
  ///////////////////////////////////////////////////////////////////
  /**
   * Liefert den aktuellen Wert der Eigenschaft. Kann (seit JDK1.5.0) auch
   * genutzt werden, wenn es sich bei der Property um einen Build-in-Type
   * handelt.<br>
   * <b>Da eine skalare Eigenschaft genau einen Wert enthaelt, duerfen keine
   * <code>coords</code>-Parameter angegeben werden!!</b>
   * @exception java.lang.UnsupportedOperationException falls ein oder mehrere
   *            <code>coords</code>-Parameter angegeben werden
   */
  protected Object getValue(int... coords) {
    if (coords.length > 0)
      throw new UnsupportedOperationException("A ScalarProperty only contains one value. Thus no parameters are allowed for getValue(..).");
    return value;
  }

  /**
   * Setzt den Wert der Eigenschaft. Kann (seit JDK1.5.0) auch genutzt werden,
   * wenn es sich bei der Property um einen Build-in-Type handelt.<br>
   * Iniziiert ein <code>ObjectChangeEvent</code> fuer die Property.
   * <b>Da eine skalare Eigenschaft genau einen Wert enthaelt, duerfen keine
   * <code>coords</code>-Parameter angegeben werden!!</b>
   * @param value neuer Wert fuer die Eigenschaft
   * @exception java.lang.UnsupportedOperationException falls ein oder mehrere
   *            <code>coords</code>-Parameter angegeben werden
   * @exception java.lang.ClassCastException falls das angegebene Objekt nicht
   *            zum Property-Typ passt
   */
  protected void setValue(Object value, int... coords) {
    if (coords.length > 0)
      throw new UnsupportedOperationException("A ScalarProperty only contains one value. Thus no parameters are allowed for getValue(..).");
    if ( value!=null && !this.isValid( value.getClass() ) )
      throw new ClassCastException(value.getClass().getName().concat(" can not be stored in a ScalarProperty[").concat(this.getType().getName()).concat("]"));

    this.fireEvent( new ObjectChangeEvent(
      new Invoker(this),
      getValue(),
      this.value = value
    ) );
  }

  ///////////////////////////////////////////////////////////////////
  // Implementierung von ObjectStructure
  ///////////////////////////////////////////////////////////////////
  /**
   * Liefert <code>false</code>, das eine skalare Eigenschaft immer
   * nur aus einem Wert besteht.
   */
  public boolean containsMultipleValues() {
    return false;
  }
}
