/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data.property;

import java.util.Enumeration;
import java.util.NoSuchElementException;
import java.util.Vector;

import schmitzm.data.ObjectStructure;

/**
 * Diese Klasse stellt eine Oberklasse fuer alle Eigenschaften dar, die
 * einen oder mehrere Wert speichern koennen.
 * <b>Die Implementierung basiert auf Sun JDK 1.5.0 und nutzt optionale
 * Methoden-Parameter.</b>
 * Hierdurch kann (wahlweise) fuer skalare Properties auf Parameter verzichtet
 * werden (z.B. <code>getValue()</code> oder <code>setValue(<i>someObject</i>)</code>).
 * Fuer 2-dim. Matrizen koennen derselben Methode 2 zusaetzliche Parameter
 * (oder ein 2-dim. Array) uebergeben werden, um die Position innerhalb der
 * Matrix zu spezifizieren (z.B. <code>getValue(10,13)</code> oder
 * <code>setValue(<i>someObject</i>,10,13)</code>).<br><br>
 * Die jeweilige Sub-Klasse muss nur noch die Art der Speicherung (z.B. skalar,
 * Liste oder Matrix), sowie je eine setter- und getter-Methode implementieren.<br>
 * Der Zugriff auf die Property kann durch Zugriffsrechte kontrolliert werden.
 * Ein Zugriff (lesend oder schreibend) ist nur ueber ein entsprechendes
 * Zugriffsrecht ({@link Access}) moeglich. Standardmaessig ist unbegrenzter
 * Zugriff eingestellt. Durch Angabe von {@link ValuePropertyAccessParameters}
 * kann jedoch benutzerdefiniert festgelegt werden, wie viele Objekte gleichzeitig
 * Lese- und/oder Schreibzugriff erlangen duerfen.
 * @see #getReadAccess(Object)
 * @see #getWriteAccess(Object)
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class ValueProperty extends Property implements ObjectStructure {
  private Vector readAccess = new Vector<PropertyReadAccess>();
  private Vector writeAccess = new Vector<PropertyWriteAccess>();
  private ValuePropertyAccessParameters accessParams = null;

  /**
   * Erzeugt eine Eigenschaft.
   * @param name   Bezeichnung der Eigenschaft
   * @param type   Typ der Eigenschaft
   * @param params Parameter fuer die Verwaltung des Zugriffsrechts
   * @see schmitzm.data.property.ValuePropertyAccessParameters
   */
  public ValueProperty(String name, PropertyType type, ValuePropertyAccessParameters params) {
    super(name,type);
    this.accessParams = params;
  }

  /**
   * Erzeugt eine Eigenschaft. Der Zugriff auf die Eigenschaft ist
   * uneingeschraengt (beliebig viele Lese- und beliebig viele Schreibrechte).
   * @param name   Bezeichnung der Eigenschaft
   * @param type   Typ der Eigenschaft
   * @see schmitzm.data.property.ValuePropertyAccessParameters#UNLIMITED_ACCESS
   */
  public ValueProperty(String name, PropertyType type) {
    this(name,type,ValuePropertyAccessParameters.UNLIMITED_ACCESS);
  }

  /**
   * Erzeugt eine Eigenschaft.
   * @param name   Bezeichnung der Eigenschaft
   * @param sample Bestimmt die Element-Struktur, die in der Eigenschaft
   *               gespeichert werden kann
   * @param params Parameter fuer die Verwaltung des Zugriffsrechts
   * @see schmitzm.data.property.ValuePropertyAccessParameters
   */
  public ValueProperty(String name, ObjectStructure sample, ValuePropertyAccessParameters params) {
    this(name,new ValuePropertyType(sample),params);
  }

  /**
   * Erzeugt eine Eigenschaft. Der Zugriff auf die Eigenschaft ist
   * uneingeschraengt (beliebig viele Lese- und beliebig viele Schreibrechte).
   * @param name   Bezeichnung der Eigenschaft
   * @param sample Bestimmt die Element-Struktur, die in der Eigenschaft
   *               gespeichert werden kann
   * @see schmitzm.data.property.ValuePropertyAccessParameters#UNLIMITED_ACCESS
   */
  public ValueProperty(String name, ObjectStructure sample) {
    this(name,new ValuePropertyType(sample),ValuePropertyAccessParameters.UNLIMITED_ACCESS);
  }

  /**
   * Erzeugt eine Eigenschaft. Als Typ kann auch ein Build-in-Type angegeben
   * werden (z.B. <code>int.class</code>).
   * @param name   Bezeichnung der Eigenschaft
   * @param type   Daten-Typ der Eigenschaft
   * @param params Parameter fuer die Verwaltung des Zugriffsrechts
   * @see schmitzm.data.property.ValuePropertyAccessParameters
   */
  public ValueProperty(String name, Class type, ValuePropertyAccessParameters params) {
    this(name,ValuePropertyType.createValuePropertyType(type),params);
  }

  /**
   * Erzeugt eine Eigenschaft. Als Typ kann auch ein Build-in-Type angegeben
   * werden (z.B. <code>int.class</code>). Der Zugriff auf die Eigenschaft ist
   * uneingeschraengt (beliebig viele Lese- und beliebig viele Schreibrechte).
   * @param name Bezeichnung der Eigenschaft
   * @param type Daten-Typ der Eigenschaft
   * @see schmitzm.data.property.ValuePropertyAccessParameters#UNLIMITED_ACCESS
   */
  public ValueProperty(String name, Class type) {
    this(name,type,ValuePropertyAccessParameters.UNLIMITED_ACCESS);
  }

  /**
   * Liefert den Daten-Typ, der in der Eigenschaft gespeichert werden kann.
   */
  public ValuePropertyType getPropertyType() {
    return (ValuePropertyType)super.getPropertyType();
  }

  ///////////////////////////////////////////////////////////////////
  // Implementierung fuer Accessible
  ///////////////////////////////////////////////////////////////////

  /**
   * Liefert die Parameter (max. Lese/Schreibzugriff) der Property.
   * @return ValuePropertyAccessParameters
   */
  public ValuePropertyAccessParameters getAccessParameters() {
    return accessParams;
  }

  /**
   * Wird aufgerufen, wenn ein Zugriffsrecht (Lesend oder Schreibend)
   * angefordert wird. Diese Methode prueft, ob das Recht erteilt werden kann.
   * Wie viele Lese- und Schreibrechte erteilt werden koennen ist ueber die
   * Access-Parametern festgelegt.
   * @param access Zugriffsrecht welches erteilt werden soll
   * @exception schmitzm.data.property.AccessViolationException falls
   *            das angegebene Recht nicht zur Property passt
   * @see #getAccessParameters()
   */
  public void applyAccess(Access access) {
    if ( access.object != this )
      throw new AccessViolationException(access,"Access does not belong to an instance of "+this.getClass());

    if ( access instanceof PropertyReadAccess )
      // Jedes Leserecht wird nur einmal vermerkt
      if ( !readAccess.contains(access) ) {
        // Leserechte schon alle vergeben?
        if ( !getAccessParameters().isReadUnlimited() && readAccess.size() == getAccessParameters().getSimultanReadAccess() )
          throw new AccessViolationException(access,this.getClass().getName().concat(" can only share ReadAccess to ").concat(String.valueOf(getAccessParameters().getSimultanWriteAccess())).concat(" object"));
        readAccess.add(access);
      }

    if ( access instanceof PropertyWriteAccess )
      // Jedes Schreibrecht wird nur einmal vermerkt
      if ( !writeAccess.contains(access) ) {
        // Schreibrechte schon alle vergeben?
        if ( !getAccessParameters().isWriteUnlimited() && writeAccess.size() == getAccessParameters().getSimultanWriteAccess() )
          throw new AccessViolationException(access,this.getClass().getName().concat(" can only share WriteAccess to ").concat(String.valueOf(getAccessParameters().getSimultanWriteAccess())).concat(" object"));
        writeAccess.add(access);
      }
  }

  /**
   * Wird aufgerufen, wenn ein Zugriffsrecht (Lesend oder Schreibend)
   * wieder zurueckgegeben wird.
   * @param access Zugriffsrecht welches zurueckgegeben werden soll
   * @exception schmitzm.data.property.AccessViolationException falls
   *            das angegebene Recht nicht zur Property passt
   */
  public void releaseAccess(Access access) {
    if ( access.object != this )
      throw new AccessViolationException(access,"Access does not belong to an instance of "+this.getClass());

    if ( access instanceof PropertyReadAccess )
      readAccess.remove(access);
    if ( access instanceof PropertyWriteAccess )
      writeAccess.remove(access);
  }

  /**
   * Entzieht saemtlichen Zugriffsrechten, die aktuell fuer das Objekt
   * verteilt sind, die Gueltigkeit.
   */
  public void disposeAllAccess() {
    // Bemerke: das Entfernen aus dem Vecor geschieht durch
    //          releaseAccess(.) welche von Access.dispose()
    //          aufgerufen wird!!
    for (;!readAccess.isEmpty();)
      ((Access)readAccess.elementAt(0)).dispose();
    for (;!writeAccess.isEmpty();)
      ((Access)writeAccess.elementAt(0)).dispose();
  }

  /**
   * Liefert die Anzahl an Zugriffsrechten einer bestimmten Art, die aktuell
   * fuer das Objekt verteilt sind.
   */
  public int getAccessCount(Class c) {
    int accessCount = 0;
    if ( c==null || c.isAssignableFrom(PropertyReadAccess.class) )
      accessCount += readAccess.size();
    if ( c==null || c.isAssignableFrom(PropertyWriteAccess.class) )
      accessCount += writeAccess.size();
    return accessCount;
  }

  /**
   * Liefert die Gesamz-Anzahl an Zugriffsrechten, die aktuell fuer das Objekt
   * verteilt sind.
   */
  public int getAccessCount() {
    return getAccessCount(null);
  }


  /**
   * Liefert ein Leserecht auf die Property.
   * Wie viele Leserechte gleichzeitig erteilt werden koennen ist ueber die
   * Access-Parametern festgelegt.
   * @param owner Objekt, welches das Recht beantragt
   * @see #applyAccess(Access)
   * @see #getAccessParameters()
   */
  public PropertyReadAccess getReadAccess(Object owner) {
    return new PropertyReadAccess(this,owner);
  }

  /**
   * Liefert ein Schreibrecht auf die Property.
   * Wie viele Schreibrechte gleichzeitig erteilt werden koennen ist ueber die
   * Access-Parametern festgelegt.
   * @param owner Objekt, welches das Recht beantragt
   * @see #applyAccess(Access)
   * @see #getAccessParameters()
   */
  public PropertyWriteAccess getWriteAccess(Object owner) {
    return new PropertyWriteAccess(this,owner);
  }

  /**
   * Liefert ein Leserecht auf die Property, welches nach einmaligem
   * Zugriff automatisch ungueltig wird.<br>
   * Wie viele Leserechte gleichzeitig erteilt werden koennen ist ueber die
   * Access-Parametern festgelegt.
   * @see #applyAccess(Access)
   * @see #getAccessParameters()
   */
  public PropertyReadAccess getOneTimeReadAccess() {
    return new PropertyReadAccess(this,this,1);
  }

  /**
   * Liefert ein Schreibrecht auf die Property, welches nach einmaligem
   * Zugriff automatisch ungueltig wird.<br>
   * Wie viele Schreibrechte gleichzeitig erteilt werden koennen ist ueber die
   * Access-Parametern festgelegt.
   * @see #applyAccess(Access)
   * @see #getAccessParameters()
   */
  public PropertyWriteAccess getOneTimeWriteAccess() {
    return new PropertyWriteAccess(this,this,1);
  }

  ///////////////////////////////////////////////////////////////////
  // Deklarationen fuer ValueProperty
  ///////////////////////////////////////////////////////////////////

  /**
   * Liefert den/einen Wert der Eigenschaft.<br>
   * <code>coords</code> spezifizieren (optional) die "Koordinaten", an denen
   * der Wert in der Eigenschaft zu finden ist.
   * Fuer skalare Eigenschaften darf/braucht diese Angabe nicht gemacht zu
   * werden. Fuer Listen und 1-dim. Matrizen (Arrays) darf nur ein Wert
   * angegeben werden. Fuer mehr-dimensionale Matrizen muessen entsprechend
   * der Dimension mehr Koordinaten angegeben werden (als einzelne Parameter oder
   * als ein Array). z.B. fuer 2-dim. Matrix:<br>
   * <code>getValue(10,13)</code> oder <code>getValue( new int[] {10,13} )</code><br>
   * <br>
   * Kann (seit JDK1.5.0) auch genutzt werden, wenn es sich bei der Property um einen
   * Build-in-Type handelt.
   * @param coords optionale Koordinaten
   */
  protected abstract Object getValue(int... coords);

  /**
   * Setzt den/einen Wert der Eigenschaft.<br>
   * <code>coords</code> spezifizieren (optional) die "Koordinaten", an denen
   * der Wert in der Eigenschaft zu finden ist.
   * Fuer skalare Eigenschaften darf/braucht diese Angabe nicht gemacht zu
   * werden. Fuer Listen und 1-dim. Matrizen (Arrays) darf nur ein Wert
   * angegeben werden. Fuer mehr-dimensionale Matrizen muessen entsprechend
   * der Dimension mehr Koordinaten angegeben werden (als einzelne Parameter oder
   * als ein Array). z.B. fuer 2-dim. Matrix:<br>
   * <code>getValue(10,13)</code> oder <code>getValue( new int[] {10,13} )</code><br>
   * <br>
   * Kann (seit JDK1.5.0) auch genutzt werden, wenn es sich bei der Property um einen
   * Build-in-Type handelt.
   * Iniziiert ein <code>ObjectChangeEvent</code> fuer die Property.
   * @param value  neuer Wert fuer die Eigenschaft
   * @param coords optionale Koordinaten
   * @exception java.lang.ClassCastException falls das angegebene Objekt nicht
   *            zum Property-Typ passt
   */
  protected abstract void setValue(Object value, int... coords);


  ///////////////////////////////////////////////////////////////////
  // Implementierung von ObjectStructure
  ///////////////////////////////////////////////////////////////////
  /**
   * Liefert <code>true</code>, da jede Property einen Namen besitzt.
   */
  public boolean isStructureNamed() {
    return true;
  }

  /**
   * Liefert den Namen der Eigenschaft.
   * @see #getName()
   */
  public String getStructureName() {
    return getName();
  }

  /**
   * Liefert 1, da eine Property immer genau eine Eigenschaft darstellt.
   */
  public int getAttrCount() {
    return 1;
  }

  /**
   * Liefert eine Liste, in der genau 1 Element enthalten ist. Dabei handelt es
   * sich um eine {@link ObjectStructure}, falls die Property einen strukturierten
   * Inhalt hat, oder andernfalls eine {@link Class}, die den "allgemeinen"
   * Eigenschaftstyp darstellt.
   * @see ValuePropertyType#getTypeStructure()
   * @see ValuePropertyType#getType()
   */
  public Enumeration getAttrTypes() {
    return new Enumeration() {
      private int nextCalls = 0;

      public boolean hasMoreElements() {
        return nextCalls == 0;
      }

      public Object nextElement() {
        if (!hasMoreElements())
          throw new NoSuchElementException(getClass().getSimpleName().concat(" only contains ").concat( String.valueOf(getAttrCount()) ).concat(" Attribute!"));
        nextCalls++;
        return getPropertyType().isStructured() ? getPropertyType().getTypeStructure() : getType();
      }
    };
  }

  /**
   * Vergleicht die Struktur der Property mit einer anderen auf Gleichheit.
   * @param object kann eine andere {@link ValueProperty} sein oder ein
   *               {@link ValuePropertyType}
   * @see ValuePropertyType#equalsInStructure(Object)
   */
  public boolean equalsInStructure(Object object) {
    return getPropertyType().equalsInStructure( object );
  }
}
