/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data.property;

/**
 * Diese Klasse stellt Rechte-Parameter fuer Wert-Eigenschaften (<code>ValueProperty</code>)
 * dar. Fuer eine Wert-Eigenschaft koennen Lese- und Schreibrechte vergeben
 * werden. Ueber die Parameter dieser Klasse kann festgelegt werden, wie
 * viele Objekte gleichzeitig das jeweilige Recht besitzenm duerfen.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ValuePropertyAccessParameters {
  /** Konstante, die fuer unbegrenzten (Lese- oder Schreib-)Zugriff steht. */
  public static final int UNLIMITED = -1;
  /** Vordefinierte Parameter fuer unbegrenzten Lese- und unbegrenzten Schreibzugriff. */
  public static final ValuePropertyAccessParameters UNLIMITED_ACCESS = new ValuePropertyAccessParameters(UNLIMITED,UNLIMITED);
  /** Vordefinierte Parameter fuer unbegrenzten Lese- und isolierten Schreibzugriff
   * (max. ein Schreibzugriffsrecht zu jedem Zeitpunkt). */
  public static final ValuePropertyAccessParameters DEFAULT_ACCESS   = new ValuePropertyAccessParameters(UNLIMITED,1);

  private int simRead = 0;
  private int simWrite = 0;

  /**
   * Erzeugt neue Parameter fuer Wert-Eigenschaften.
   * @param simultanRead Anzahl an Objekten, die glz. lesen duerfen
   * @param simultanWrite Anzahl an Objekten, die glz. schreiben duerfen
   */
  public ValuePropertyAccessParameters(int simultanRead, int simultanWrite) {
    this.simRead = simultanRead;
    this.simWrite = simultanWrite;
  }

  /**
   * Liefert die Anzahl an Objekten die gleichzeitig lesen duerfen.
   * @return <code>UNLIMITED</code> falls beliebig viele Objekte glz. lesen duerfen
   */
  public int getSimultanReadAccess() {
    return this.simRead;
  }

  /**
   * Prueft, ob beliebig viele Objekte gleichzeitig lesen duerfen.
   */
  public boolean isReadUnlimited() {
    return getSimultanReadAccess()==UNLIMITED;
  }

  /**
   * Liefert die Anzahl an Objekten die gleichzeitig schreiben duerfen.
   * @return <code>UNLIMITED</code> falls beliebig viele Objekte glz. schreiben duerfen
   */
  public int getSimultanWriteAccess() {
    return this.simWrite;
  }

  /**
   * Prueft, ob beliebig viele Objekte gleichzeitig schreiben duerfen.
   */
  public boolean isWriteUnlimited() {
    return getSimultanWriteAccess()==UNLIMITED;
  }

}
