/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.data.property;

import schmitzm.data.ObjectStructure;
import schmitzm.data.ObjectStructureUtil;
import schmitzm.temp.BaseTypeUtil;

/**
 * Diese Klasse repraesentiert den Datentyp einer {@linkplain ValueProperty Wert-Eigenschaft}.
 * Der Datentyp kann aus einer einfachen Klasse bestehen, oder aus oder aus
 * einer komplexen  {@linkplain ObjectStructure Objekt-Struktur}.
 * <b>Bemerke:</b><br>
 * Fuer Basis-Datentypen (<code>int</code>, <code>double</code>,
 * <code>String</code>, ...) sollten keine neuen
 * <code>ValuePropertyType</code>-Instanzen erzeugt werden, sondern die
 * bereits implementierten Konstanten verwendet werden!
 * @see #TYPE_BYTE
 * @see #TYPE_SHORT
 * @see #TYPE_INT
 * @see #TYPE_LONG
 * @see #TYPE_FLOAT
 * @see #TYPE_DOUBLE
 * @see #TYPE_BOOLEAN
 * @see #TYPE_CHAR
 * @see #TYPE_STRING
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ValuePropertyType extends PropertyType {
  /**
   * Speichert die Struktur der Objekte, die fuer den Property-Wert zulaessig
   * ist.
   */
  protected ObjectStructure structure = null;

  /** Repraesentiert eine vorgefertigte {@link Byte}-Eigenschaft. **/
  public static final ValuePropertyType TYPE_BYTE    = new ValuePropertyType(Byte.class);
  /** Repraesentiert eine vorgefertigte {@link Short}-Eigenschaft. **/
  public static final ValuePropertyType TYPE_SHORT   = new ValuePropertyType(Short.class);
  /** Repraesentiert eine vorgefertigte {@link Integer}-Eigenschaft. **/
  public static final ValuePropertyType TYPE_INT     = new ValuePropertyType(Integer.class);
  /** Repraesentiert eine vorgefertigte {@link Long}-Eigenschaft. **/
  public static final ValuePropertyType TYPE_LONG    = new ValuePropertyType(Long.class);
  /** Repraesentiert eine vorgefertigte {@link Float}-Eigenschaft. **/
  public static final ValuePropertyType TYPE_FLOAT   = new ValuePropertyType(Float.class);
  /** Repraesentiert eine vorgefertigte {@link Double}-Eigenschaft. **/
  public static final ValuePropertyType TYPE_DOUBLE  = new ValuePropertyType(Double.class);
  /** Repraesentiert eine vorgefertigte {@link Boolean}-Eigenschaft. **/
  public static final ValuePropertyType TYPE_BOOLEAN = new ValuePropertyType(Boolean.class);
  /** Repraesentiert eine vorgefertigte {@link Character}-Eigenschaft. **/
  public static final ValuePropertyType TYPE_CHAR    = new ValuePropertyType(Character.class);
  /** Repraesentiert eine vorgefertigte {@link String}-Eigenschaft. **/
  public static final ValuePropertyType TYPE_STRING  = new ValuePropertyType(String.class);

  /**
   * Liefert einen passenden Property-Typ zu einem Datentyp oder einem
   * Beispiel-Objekt. Handelt es sich um
   * einen Basis-Datentyp (<code>Integer</code>, <code>Double</code>,
   * <code>String</code>, ...) wird eine der konstanten
   * <code>ValuePropertyType</code>-Instanzen zurueckgegeben. Andernfalls wird
   * versucht den Beispielwert ueber den Standard-Konstruktor der angegebenen
   * Klasse zu erzeugen. Schlaegt dies fehl, wird der Beispiel-Wert auf
   * <code>null</code> gesetzt.
   * @param sample Typ den die Eigenschaft aufnehmen soll (<code>Class</code>,
   *               <code>ObjectStructure</code> oder Beispiel-<code>Object</code>)
   * @see #TYPE_BYTE
   * @see #TYPE_SHORT
   * @see #TYPE_INT
   * @see #TYPE_LONG
   * @see #TYPE_FLOAT
   * @see #TYPE_DOUBLE
   * @see #TYPE_BOOLEAN
   * @see #TYPE_CHAR
   * @see #TYPE_STRING
   */
  public static ValuePropertyType createValuePropertyType(Object sample) {
    if ( sample == null )
      throw new UnsupportedOperationException("Can not create a PropertyType from null");

    // Wenn es sich um eine Structure handelt, diese verwenden
    if ( sample instanceof ObjectStructure )
      return new ValuePropertyType( (ObjectStructure)sample );

    // Fuer Basis-Datentypen die Konstanten verwenden
    // Bemerke: Leider werden fuer Class-Objekte nicht die BaseTypeUtil-Methoden
    //          fuer Class aufgerufen, sondern fuer Object!! (warum?)
    //          Deshalb muss hier leider fallunterschieden und explizit
    //          gecastet werden
    if ( sample instanceof Class ) {
      if (BaseTypeUtil.isByte((Class)sample))
        return TYPE_BYTE;
      if (BaseTypeUtil.isShort((Class)sample))
        return TYPE_SHORT;
      if (BaseTypeUtil.isInteger((Class)sample))
        return TYPE_INT;
      if (BaseTypeUtil.isLong((Class)sample))
        return TYPE_LONG;
      if (BaseTypeUtil.isFloat((Class)sample))
        return TYPE_FLOAT;
      if (BaseTypeUtil.isDouble((Class)sample))
        return TYPE_DOUBLE;
      if (BaseTypeUtil.isBoolean((Class)sample))
        return TYPE_BOOLEAN;
      if (BaseTypeUtil.isCharacter((Class)sample))
        return TYPE_CHAR;
      if (BaseTypeUtil.isString((Class)sample))
        return TYPE_STRING;
      // sonst Property-Typ neu erzeugen
      return new ValuePropertyType((Class)sample);
    } else {
      if (BaseTypeUtil.isByte(sample))
        return TYPE_BYTE;
      if (BaseTypeUtil.isShort(sample))
        return TYPE_SHORT;
      if (BaseTypeUtil.isInteger(sample))
        return TYPE_INT;
      if (BaseTypeUtil.isLong(sample))
        return TYPE_LONG;
      if (BaseTypeUtil.isFloat(sample))
        return TYPE_FLOAT;
      if (BaseTypeUtil.isDouble(sample))
        return TYPE_DOUBLE;
      if (BaseTypeUtil.isBoolean(sample))
        return TYPE_BOOLEAN;
      if (BaseTypeUtil.isCharacter(sample))
        return TYPE_CHAR;
      if (BaseTypeUtil.isString(sample))
        return TYPE_STRING;
      // sonst Property-Typ neu erzeugen
      return new ValuePropertyType(sample.getClass());
    }
  }

  /**
   * Erzeugt einen neuen Eigenschaftstyp.
   * <b>Bemerke:</b><br>
   * Fuer Basis-Datentypen (<code>Integer</code>, <code>Double</code>,
   * <code>String</code>, ...) sollte nicht dieser Konstruktor verwendet werden,
   * sondern eine der konstanten <code>ValuePropertyType</code>-Instanzen.
   * @param sample Beispielwert
   * @exception java.lang.UnsupportedOperationException falls der Beispiel-Wert
   *            keine Instanz des Eigenschaftstyp ist.
   * @see #TYPE_BYTE
   * @see #TYPE_SHORT
   * @see #TYPE_INT
   * @see #TYPE_LONG
   * @see #TYPE_FLOAT
   * @see #TYPE_DOUBLE
   * @see #TYPE_BOOLEAN
   * @see #TYPE_CHAR
   * @see #TYPE_STRING
   */
  public ValuePropertyType(ObjectStructure sample) {
    super(sample.getClass());
    this.structure = sample;
  }

  /**
   * Erzeugt einen neuen Eigenschaftstyp. Der Beispielwert wird ueber den
   * Standard-Konstruktor der Typ-Klasse zu instanziieren versucht. Schlaegt
   * dies fehl, wird der Beispiel-Wert auf <code>null</code> gesetzt.<br>
   * <b>Bemerke:</b><br>
   * Fuer Basis-Datentypen (<code>Integer</code>, <code>Double</code>,
   * <code>String</code>, ...) sollte nicht dieser Konstruktor verwendet werden,
   * sondern eine der konstanten <code>ValuePropertyType</code>-Instanzen.
   * @param type Typ den die Eigenschaft aufnehmen kann
   * @see #TYPE_BYTE
   * @see #TYPE_SHORT
   * @see #TYPE_INT
   * @see #TYPE_LONG
   * @see #TYPE_FLOAT
   * @see #TYPE_DOUBLE
   * @see #TYPE_BOOLEAN
   * @see #TYPE_CHAR
   * @see #TYPE_STRING
   */
  public ValuePropertyType(Class type) {
    super(type);
    this.structure = null;
  }

  /**
   * Prueft, ob der Eigenschaftstyp aus einer festen Struktur besteht.
   * @return <code>false</code> falls der Eigenschaftstyp lediglich durch eine Klasse
   *         festgelegt ist
   */
  public boolean isStructured() {
    return getTypeStructure() != null;
  }

  /**
   * Liefert die (feste) Struktur des Eigenschaftstyp.
   * @return <code>null</code> falls der Eigenschaftstyp lediglich durch eine Klasse
   *         festgelegt ist
   */
  public ObjectStructure getTypeStructure() {
    return structure;
  }

  /**
   * Prueft, ob der Datentyp mit einem anderen <b>strukturell</b> identisch ist.
   * Der Inhalt der Eigenschaft wird nicht beachtet
   * @param typeObj kann eine <code>ValuePropertyType</code>-Instanz sein, oder
   *                auch eine {@link ValueProperty}-Instanz
   */
  public boolean equalsInStructure(Object typeObj) {
    // ValuePropertyType-Instanz aus uebergebenen Objekt ermitteln
    ValuePropertyType type = null;
    if ( typeObj instanceof ValuePropertyType )
      type = (ValuePropertyType)typeObj;
    if ( typeObj instanceof ValueProperty )
      type = ((ValueProperty)typeObj).getPropertyType();
    // ein vergleichbarere Typ
    if ( type == null )
      return false;

    // wenn beide strukturiert sind, die Struktur vergleichen
    if ( isStructured() ) {
      if ( !type.isStructured() )
          return false;
      return ObjectStructureUtil.compareObjectStructures(getTypeStructure(),type.getTypeStructure()) == ObjectStructureUtil.EQUAL;
    } else {
      // wenn beide nicht strukturiert sind, die Klassen vergleichen
      if ( type.isStructured() )
          return false;
      return getType().equals( type.getType() );
    }
  }

 }
