/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.feature;

import org.geotools.filter.AbstractFilterImpl;
import org.geotools.filter.Filter;
import org.geotools.filter.FilterFactoryImpl;
import org.opengis.feature.simple.SimpleFeature;

/**
 * Diese Klasse stellt einen {@link Filter} dar, der einen Attributwert mit einem
 * konstanten Vergleichswert vergleicht.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class AttributeFilter extends AbstractFilterImpl implements Cloneable {
  /** Attribut-Filter, der auf "Attributwert = Konstante" prueft. */
  public static final AttributeFilter EQUALS = new AttributeFilter(CompareType.EQUALS);
  /** Attribut-Filter, der auf "Attributwert > Konstante" prueft. */
  public static final AttributeFilter GT = new AttributeFilter(CompareType.GT);
  /** Attribut-Filter, der auf "Attributwert >= Konstante" prueft. */
  public static final AttributeFilter GE = new AttributeFilter(CompareType.GE);
  /** Attribut-Filter, der auf "Attributwert < Konstante" prueft. */
  public static final AttributeFilter LT = new AttributeFilter(CompareType.LT);
  /** Attribut-Filter, der auf "Attributwert <= Konstante" prueft. */
  public static final AttributeFilter LE = new AttributeFilter(CompareType.LE);

  /** Typ der Vergleichsfunktion fuer den Filter. */
  public static enum CompareType {
    /** Vergleich auf Gleichheit (Attributwert = Konstante). */
    EQUALS,
    /** Groesser-Vergleich (Attributwert > Konstante). */
    GT,
    /** Groesser/Gleich-Vergleich (Attributwert >= Konstante). */
    GE,
    /** Kleiner-Vergleich (Attributwert < Konstante). */
    LT,
    /** Kleiner/Gleich-Vergleich (Attributwert <= Konstante). */
    LE,
  }

  /** Art des Vergleichs in {@link #evaluate(SimpleFeature)}. */
  protected CompareType compType = null;
  /** Name des SimpleFeature-Attributs, das mit der {@linkplain #compValue Konstante}
   *  verglichen wird.*/
  protected String attrName = "";
  /** Konstante, mit der das {@linkplain #attrName SimpleFeature-Attibut}
   *  vergleichen wird. */
  protected Object compValue = null;

  /**
   * Erzeugt einen neuen Filter. Das Attribut und die Konstante muessen
   * nachtraeglich durch {@link #setAttributeName(String)} und
   * {@link #setCompareValue(Object)} gesetzt werden.
   * @param compType bestimmt die Vergleichs-Funktion
   * @see #CompareType
   */
  public AttributeFilter(CompareType compType) {
    this(compType, null, null);
  }

  /**
   * Erzeugt einen neuen Filter.
   * @param compType bestimmt die Vergleichs-Funktion
   * @param attrName Name des Attributs, welches verglichen wird
   * @param compValue Konstante mit der das Attribut verglichen wird
   */
  public AttributeFilter(CompareType compType, String attrName, Object compValue) {
    super( new FilterFactoryImpl() );
    this.compType = compType;
    setAttributeName(attrName);
    setCompareValue(compValue);
  }

  /**
   * Liefert einen Filter, der die inverse Funktion des Filters darstellt.
   */
  public AttributeFilter inverse() {
    switch ( compType ) {
      case EQUALS: return EQUALS;
      case GT:     return LE;
      case GE:     return LT;
      case LT:     return GE;
      case LE:     return GT;
    }
    throw new UnsupportedOperationException("Compare type can not be inverted: "+compType);
  }

  /**
   * Setzt die Konstante, mit der der Attribut-Wert verglichen wird.
   * @param compValue Vergleichswert
   */
  public void setCompareValue(Object compValue) {
    this.compValue = compValue;
  }

  /**
   * Liefert die Konstante, mit der der Attribut-Wert verglichen wird.
   */
  public Object getCompareValue() {
    return this.compValue;
  }

  /**
   * Setzt das Attribut, das mit der Konstanten verglichen wird.
   * @param attrName Attribut-Name
   */
  public void setAttributeName(String attrName) {
    this.attrName = attrName;
  }

  /**
   * Liefert das Attribut, das mit der Konstanten verglichen wird.
   */
  public Object getAttributeName() {
    return this.attrName;
  }

  /**
   * Liefert den Vergleichstype mit dem das Attribut und die Konstanten
   * verglichen werden.
   */
  public CompareType getCompareType() {
    return this.compType;
  }

  /**
   * Fuehrt den Vergleich durch. Liefert {@code false}, wenn der Attribut-Name
   * oder die Konstante nicht gesetzt ist oder der Attribut-Wert {@code null} ist.
   * @param feature SimpleFeature dessen Attribut mit der Konstanten verglichen wird
   */
  public boolean evaluate(SimpleFeature feature) {
    if ( attrName == null || compValue == null )
      return false;
    Object attrValue = feature.getAttribute(attrName);
    if ( attrValue == null )
      return false;

    Object attributeValue = attrValue;
    Object compareValue   = compValue;
    // Wenn Attribut- und Vergleichswert numerisch sind, dann werden die
    // Zahlen vergleichen
    if ( attrValue instanceof Number && compValue instanceof Number ) {
      attributeValue = ((Number)attrValue).doubleValue();
      compareValue   = ((Number)compValue).doubleValue();
    } else if ( attrValue instanceof Number && compValue instanceof String ) {
      // Versuchen, "compValue" in Zahl umzuwandeln
      try {
        compareValue   = Double.parseDouble( (String)compValue );
        attributeValue = ((Number)attrValue).doubleValue();
      } catch (NumberFormatException err) {
      }
    } else if ( attrValue instanceof String && compValue instanceof Number ) {
      // Versuchen, "attrValue" in Zahl umzuwandeln
      try {
        attributeValue = Double.parseDouble( (String)attrValue );
        compareValue   = ((Number)compValue).doubleValue();
      } catch (NumberFormatException err) {
      }
    }

    // Vergleich durchfuehren
    switch( compType ) {
      case EQUALS: return attributeValue.equals( compareValue );
      case GT:
      case GE:
      case LT:
      case LE: if ( !(attrValue instanceof Number) || !(compValue instanceof Number) )
                 return false;
               double aVal = (Double)attributeValue;
               double cVal = (Double)compareValue;
               switch (compType) {
                 case GT: return aVal >  cVal;
                 case GE: return aVal >= cVal;
                 case LT: return aVal <  cVal;
                 case LE: return aVal <= cVal;
               }
    }
    // sollte nie erreicht werden
    throw new UnsupportedOperationException("Unknown compare Type: "+compType);
  }

  /**
   * Fuehrt den Vergleich durch. Liefert {@code false}, wenn der Attribut-Name
   * oder die Konstante nicht gesetzt ist oder der Attribut-Wert {@code null} ist.
   * @param feature SimpleFeature dessen Attribut mit der Konstanten verglichen wird
   * @exception IllegalArgumentException wenn das uebergebene Objekt kein
   *            {@link SimpleFeature} ist
   */
  public boolean evaluate(Object feature) {
    if ( feature != null && !(feature instanceof SimpleFeature) )
      throw new IllegalArgumentException("evaluate(Object) can only be appied ob SimpleFeature objects: "+feature.getClass().getSimpleName());
    return evaluate((SimpleFeature)feature);
  }

  /**
   * Erzeugt einen {@link AttributeFilter} mit dem gleichen Vergleichstyp,
   * dem gleichen Attributnamen und der gleichen Vergleichskonstante wie
   * die Instanz.
   */
  public AttributeFilter clone() {
    return new AttributeFilter( compType, attrName, compValue );
  }

}

