/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/

package schmitzm.geotools.feature;

import java.util.HashMap;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.geotools.filter.text.cql2.CQL;
import org.geotools.filter.text.cql2.CQLException;
import org.opengis.filter.Filter;

import schmitzm.swing.SwingUtil;

/**
 * This parser creates a CQL-{@link Filter} from a rule string. It also provides
 * information about available operators and their tooltips.
 * 
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 * @author Stefan Krüger
 * 
 */
public class CQLFilterParser implements FilterParser {
	
	private static Logger LOGGER = Logger.getLogger(CQLFilterParser.class);
	
	private final Vector<String> operators = new Vector<String>();

	/**
	 * Map operator Strings to Keys that can be found in the .properties
	 * resource files
	 **/
	private final HashMap<String, String> opKeys = new HashMap<String, String>();

	/**
	 * Creates a new parser.
	 * 
	 * 
	 * CQLFitlerParser.OpDesc.not=NOT CQLFitlerParser.OpDesc.or=OR
	 * CQLFitlerParser.OpDesc.and=AND CQLFitlerParser.OpDesc.eq=\=
	 * CQLFitlerParser.OpDesc.ne=<> CQLFitlerParser.OpDesc.lt=<
	 * CQLFitlerParser.OpDesc.le=<= CQLFitlerParser.OpDesc.gt=>
	 * CQLFitlerParser.OpDesc.like=LIKE CQLFitlerParser.OpDesc.ge=>=
	 */
	public CQLFilterParser() {
		addOp("<>", "ne");
		addOp("<", "lt");
		addOp("<=", "le");
		addOp("=", "eq");
		addOp("LIKE", "like");
		addOp(">=", "ge");
		addOp(">", "gt");
		addOp("NOT", "not");
		addOp("OR", "or");
		addOp("AND", "and");
	}

	private void addOp(String operator, String opKey) {
		operators.add(operator);
		opKeys.put(operator, opKey);
	}

	/**
	 * Creates a CQL-{@link Filter} for the rule string.
	 * 
	 * @param rule
	 *            a rule string
	 * @return Filter.INCLUDE if the rule is empty
	 */
	@Override
	public Filter parseFilter(String rule) {
		if (rule == null || rule.trim().equals("")
				|| rule.equalsIgnoreCase("Filter.INCLUDE"))
			return Filter.INCLUDE;
		if (rule.equalsIgnoreCase("Filter.EXCLUDE"))
			return Filter.EXCLUDE;
		try {
			return CQL.toFilter(rule);
		} catch (CQLException err) {
			LOGGER.error("Exception while parsing : " + rule, err);
			return Filter.INCLUDE;
		}
	}

	/**
	 * Returns a description that can be used for a tooltip
	 */
	@Override
	public String getOperatorDescription(String operator) {
		return SwingUtil.RESOURCE.getString("CQLFitlerParser.OpTooltip."
				+ opKeys.get(operator));
	}

	/**
	 * Returns always 0
	 */
	@Override
	public int getOperatorParameterCount(String operator) {
		return 0;
	}

	/**
	 * The title as it might be used in the GUI
	 */
	@Override
	public String getOperatorTitle(String operator) {
		return SwingUtil.RESOURCE.getString("CQLFitlerParser.OpDesc."
				+ opKeys.get(operator));
	}

	/**
	 * a list of String operators as they may be inserted into the CQL filter.
	 */
	@Override
	public Vector<String> getOperators() {
		return operators;
	}

}
