/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.feature;

import org.opengis.feature.simple.SimpleFeature;

import schmitzm.lang.tree.BinaryTreeNode;
import schmitzm.lang.tree.OperationTree;
import schmitzm.lang.tree.TreeNode;

/**
 * Diese Klasse stellt einen Operator-Baum dar, in dem neben den von der
 * Oberklasse definitierten Operationen, Referenzen auf {@link SimpleFeature}-Attribute
 * enthalten sein koennen. Der Operator-Baum wird auf einem einzelnen {@link SimpleFeature}
 * ausgewertet. <b>Zur Zeit koennen nur numerische Attribute referenziert werden
 * und keine Strings!</b><br>
 * <br>
 * <b>Referenz auf Attribut-Name:</b> {@link AttributeNameReferenceNode}<br>
 * Die Referenz auf einen Attribut-Namen wird durch einen {@code String}-Wert
 * dargestellt, der von einem {@code $} eingeleitet wird (z.B. {@code $citizens}).<br>
 * <br>
 * <b>Referenz auf Attribut-Index:</b> {@link AttributeIndexReferenceNode}<br>
 * Die Referenz auf einen Attribut-Index (beginnend bei 0) wird durch einen
 * {@code int}-Wert dargestellt, der von einem {@code #} eingeleitet wird (z.B. {@code #7}).<br>
 * <br>
 * Bei der Auswertung des Operatorbaums wird die Referenz durch
 * den entsprechenden Attribut-Wert des Features ersetzt.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class FeatureOperationTree extends OperationTree {
  private SimpleFeature feature = null;

  /**
   * Erzeugt einen neuen Raster-Operatorbaum
   * @param root Wurzelknoten
   */
  public FeatureOperationTree(TreeNode root) {
    super(root);
  }

  /**
   * Nicht unterstuetzt!
   * @deprecated
   * @exception UnsupportedOperationException bei jedem Aufruf
   */
  public Object evaluate() {
    throw new UnsupportedOperationException("FeatureOperationTree only supports evaluate(SimpleFeature)");
  }

  /**
   * Wertet den Operatorbaum auf einem {@link SimpleFeature} aus.
   * @param feature ein SimpleFeature
   */
  public Object evaluate(SimpleFeature feature) {
    this.feature = feature;
    return super.evaluate();
  }

  /**
   * Wertet einen Knoten des Operator-Baums aus.
   * @param opTreeNode BinaryTreeNode
   * @return double
   */
  protected Object evaluate(TreeNode opTreeNode) {
    // Attribut-Index
    if ( opTreeNode instanceof AttributeIndexReferenceNode ) {
      int attrIdx = ((AttributeIndexReferenceNode)opTreeNode).getObject();
      Object object = feature.getAttribute( attrIdx );
      if ( object == null )
        throw new UnsupportedOperationException( FeatureUtil.RESOURCE.getString("FeatureOperationTree.err.NullAttr",attrIdx) );
//      if ( !(object instanceof Number) )
//        throw new UnsupportedOperationException("FeatureOperationTree only can evaluate numeric Attributes: Attribute "+attrIdx+" is of type "+object.getClass().getSimpleName());
//      return ((Number)object).doubleValue();
      return object;
    }

    // Attribut-Name
    if ( opTreeNode instanceof AttributeNameReferenceNode ) {
      String attrName = ((AttributeNameReferenceNode)opTreeNode).getObject();
      Object object = feature.getAttribute( attrName );
      if ( feature.getFeatureType().getDescriptor( attrName ) == null )
        throw new UnsupportedOperationException( FeatureUtil.RESOURCE.getString("FeatureOperationTree.err.UnknownAttr",attrName) );
      if ( object == null )
        throw new UnsupportedOperationException( FeatureUtil.RESOURCE.getString("FeatureOperationTree.err.NullAttr",attrName) );
//      if ( !(object instanceof Number) )
//        throw new UnsupportedOperationException("FeatureOperationTree can only evaluate numeric Attributes: '"+attrName+"' is of type "+object.getClass().getSimpleName());
//      return ((Number)object).doubleValue();
      return object;
    }

    return super.evaluate(opTreeNode);
  }

  /**
   * Diese Knoten repraesentiert eine Namens-Referenz auf ein {@link SimpleFeature}-Attribut
   * im Operatorbaum. Da es sich dabei um eine Konstante handelt, hat der Knoten
   * keine Kind-Knoten. Die Referenz wird durch den Attribut-Namen ({@code String})
   * dargestellt.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class AttributeNameReferenceNode extends BinaryTreeNode<String> {
    /**
     * Erzeugt einen neuen Referenz-Knoten
     * @param attrName Attribut-Name
     * @param parent   Vater-Knoten
     */
    public AttributeNameReferenceNode(String attrName, BinaryTreeNode parent) {
      super(attrName, parent);
    }

    /**
     * Erzeugt einen neuen Referenz-Knoten
     * @param attrName Attribut-Name
     */
    public AttributeNameReferenceNode(String attrName) {
      this(attrName, null);
    }

    /**
     * Macht nichts, da {@code AttributeReferenceNode} immer einen Blatt-Knoten
     * darstellt.
     * @param i Index (beginnend bei 0)
     * @param child neuer Kind-Knoten
     */
    public void setChild(int i, TreeNode<String> child) {
    }
  }

  /**
   * Diese Knoten repraesentiert eine Index-Referenz auf ein {@link SimpleFeature}-Attribut
   * im Operatorbaum. Da es sich dabei um eine Konstante handelt, hat der Knoten
   * keine Kind-Knoten. Die Referenz wird durch den Attribut-Namen ({@code String})
   * dargestellt.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class AttributeIndexReferenceNode extends BinaryTreeNode<Integer> {
    /**
     * Erzeugt einen neuen Referenz-Knoten
     * @param attrIdx  Attribut-Index (beginnend bei 0)
     * @param parent   Vater-Knoten
     */
    public AttributeIndexReferenceNode(int attrIdx, BinaryTreeNode parent) {
      super(attrIdx, parent);
    }

    /**
     * Erzeugt einen neuen Referenz-Knoten
     * @param attrIdx  Attribut-Index (beginnend bei 0)
     */
    public AttributeIndexReferenceNode(int attrIdx) {
      this(attrIdx, null);
    }

    /**
     * Macht nichts, da {@code AttributeIndexReferenceNode} immer einen Blatt-Knoten
     * darstellt.
     * @param i Index (beginnend bei 0)
     * @param child neuer Kind-Knoten
     */
    public void setChild(int i, TreeNode<Integer> child) {
    }
  }

}
