/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.feature;

import org.geotools.filter.FilterAbstract;
import org.geotools.filter.FilterFactoryImpl;
import org.opengis.feature.simple.SimpleFeature;

/**
 * Dieser Filter nutzt einen {@link FeatureOperationTree}, um die
 * {@link #evaluate(SimpleFeature)}-Methode auszuwerten. Ein SimpleFeature erfuellt die
 * Filter-Bedingung, wenn das Ergebnis des auf dem SimpleFeature ausgewerteten
 * Operator-Baums ungleich 0 ist.
 * @see FeatureOperationTree
 * @see FeatureOperationTreeParser
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class FeatureOperationTreeFilter extends FilterAbstract {//AbstractFilterImpl {
  /** Formel, die auf dem SimpleFeature ausgewertet wird. */
  protected String rule = null;
  /** Operator-Baum zur Formel {@link #rule}. */
  protected FeatureOperationTree opTree = null;

  /**
   * Erstellt einen neuen Filter.
   * @param rule Formel, die auf dem SimpleFeature ausgewertet wird.
   * @see FeatureOperationTreeParser
   */
  public FeatureOperationTreeFilter(String rule) {
    this( rule, new FeatureOperationTreeParser() );
  }

  /**
   * Erstellt einen neuen Filter.
   * @param rule Formel, die auf dem SimpleFeature ausgewertet wird.
   * @param parser Parser, der aus der Formel einen Operator-Baum erstellt
   */
  public FeatureOperationTreeFilter(String rule, FeatureOperationTreeParser parser) {
    super( new FilterFactoryImpl() );
    this.rule   = rule;
    this.opTree = parser.parse(rule);
  }

    /**
	 * Prueft, ob ein {@link SimpleFeature} dem Filter entspricht. Dies ist der Fall,
	 * wenn das Formel-Ergebnis fuer das SimpleFeature ungleich 0 ist.
	 * 
	 * @param feature
	 *            zu ueberpruefendes SimpleFeature
	 * @return {@code true}, wenn das SimpleFeature dem Filter entspricht
	 */
    @Override
	public boolean evaluate(SimpleFeature feature) {
// Exception sollte hier nicht ausgegeben werden, da bei allgemeinen Fehlern
// die Exception fuer JEDES SimpleFeature ausgegeben wuerde!	  
//	  try{
	    Object result = opTree.evaluate(feature);
	    return result instanceof Boolean && (Boolean) result
	        || result instanceof Number  && ((Number) result).doubleValue() != 0;
//  	  } catch (Exception e){
//  	    e.printStackTrace(); 
//  	    return false;
//  	  }
	}

  /**
   * Prueft, ob ein {@link SimpleFeature} dem Filter entspricht. Dies ist der Fall,
   * wenn das Formel-Ergebnis fuer das SimpleFeature ungleich 0 ist.
   * @param feature zu ueberpruefendes SimpleFeature
   * @return {@code true}, wenn die Auswertung
   * @exception IllegalArgumentException wenn das uebergebene Objekt kein
   *            {@link SimpleFeature} ist
   */
  @Override
  public boolean evaluate(Object feature) {
    if ( feature != null && !(feature instanceof SimpleFeature) )
      throw new IllegalArgumentException("evaluate(Object) can only be applied to SimpleFeature objects: "+feature.getClass().getSimpleName());
    return evaluate((SimpleFeature)feature);
  }

  /**
   * Liefert die Formel, die den Filter definitiert.
   */
  public String getRule() {
    return rule;
  }

}
