/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.feature;

import org.opengis.filter.Filter;

import schmitzm.lang.tree.BinaryTreeNode;
import schmitzm.lang.tree.OperationTreeParser;

/**
 * Diese Klasse stellt einen Parser fuer einen {@linkplain FeatureOperationTree SimpleFeature-Operatorbaum}
 * dar. Dieser erstellt einen {@linkplain FeatureOperationTree SimpleFeature-Operatorbaum} aus einem
 * Formel-String, der neben den in {@link OperationTreeParser} beschriebenen
 * Komponenten Attribut-Referenzen enthalten darf.<br>
 * <br>
 * <b>Referenz auf Attribut-Name:</b> {@link FeatureOperationTree.AttributeNameReferenceNode}<br>
 * Die Referenz auf einen Attribut-Namen wird durch einen {@code String}-Wert
 * dargestellt, der von einem {@code $} eingeleitet wird (z.B. {@code $citizens}).<br>
 * <br>
 * <b>Referenz auf Attribut-Index:</b> {@link FeatureOperationTree.AttributeIndexReferenceNode}<br>
 * Die Referenz auf einen Attribut-Index (beginnend bei 0) wird durch einen
 * {@code int}-Wert dargestellt, der von einem {@code #} eingeleitet wird (z.B. {@code #7}).<br>
 * <br>
 * Bei der Auswertung des Operatorbaums wird die Referenz durch
 * den entsprechenden Attribut-Wert des Features ersetzt.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class FeatureOperationTreeParser extends OperationTreeParser implements FilterParser {
  /**
   * Erstellt einen SimpleFeature-Operator-Baum aus einem Formel-String.
   * @param rule Formel
   */
  public FeatureOperationTree parse(String rule) {
    // Leere Formel wird als "immer wahr" behandelt
    if ( rule == null || rule.trim().equals("") )
      rule = "1";
    return new FeatureOperationTree( super.parse(rule).getRoot() );
  }

  /**
   * Parst das naechste Literal aus dem Tokenizer {@link #tok}.
   * Erweitert die Funktionalitaet der Oberklasse, so dass neben Konstanten
   * auch Referenznummern und -namen auf ein SimpleFeature-Attribut (eingeleitet
   * durch {@code #} und {@code $}) angegeben werden koennen.
   */
  protected BinaryTreeNode parseLiteral() {
    String token = nextNonWSToken();

    // Attribut-Name
    if (token.startsWith("$"))
      return new FeatureOperationTree.AttributeNameReferenceNode(token.substring(1));

    // Attribut-Index
    if (token.startsWith("#"))
      return new FeatureOperationTree.AttributeIndexReferenceNode(getIntFromString(token.substring(1)));

    // sonst: auf Literale der Oberklasse zurueckgreifen
    pushbackWithWSToken();
    return super.parseLiteral();
  }
  
  /**
   * Creates a {@link FeatureOperationTreeFilter} for the rule string.
   * @param rule a rule string
   * @return TRUE filter, if rule is {@code null} or empty
   */
  @Override
  public Filter parseFilter(String rule) {
    if (rule == null || rule.trim().equals(""))
      rule = "1";
    return new FeatureOperationTreeFilter(rule);
  }
}

