/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.feature;

import javax.swing.table.TableModel;

import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.simple.SimpleFeatureType;
import org.opengis.feature.type.AttributeDescriptor;

import schmitzm.geotools.gui.GeotoolsGUIUtil;
import schmitzm.swing.table.AbstractTableModel;
import schmitzm.temp.BaseTypeUtil;

/**
 * Diese Klasse stellt ein {@link TableModel} auf einem einzelnen {@link SimpleFeature}
 * dar. Dieses definiert drei Spalten:
 * <ol>
 *   <li>Attribut-Name</li>
 *   <li>Attribut-Typ</li>
 *   <li>Wert des Attributs als Text-Eingabe</li>
 * </ol>
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class FeatureTableModel extends AbstractTableModel {
  /** {@link SimpleFeature} das in der Tabelle dargestellt wird. */
  protected SimpleFeature feature = null;
  /** {@link SimpleFeatureType} der in der Tabelle dargestellt wird. */
  protected SimpleFeatureType featureType = null;

  /**
   * Erstellt ein leeres Tabellen-Modell.
   */
  public FeatureTableModel() {
    this((SimpleFeature)null);
  }

  /**
   * Erstellt ein neues Tabellen-Modell.
   * @param feature dargestelltes SimpleFeature
   */
  public FeatureTableModel(SimpleFeature feature) {
    this( feature != null ? feature.getFeatureType() : null);
    setFeature( feature );
  }

  /**
   * Erstellt ein neues Tabellen-Modell.
   * @param type ein SimpleFeatureType
   */
  public FeatureTableModel(SimpleFeatureType type) {
    super();
    setFeature( type );
  }

  /**
   * Setzt das {@link SimpleFeature}, das in der Tabelle dargestellt wird.
   * @param feature {@link SimpleFeature}
   */
  public void setFeature(SimpleFeature feature) {
    this.feature     = feature;
    this.featureType = feature != null ? feature.getFeatureType() : null;
    this.fireTableDataChanged();
  }

  /**
   * Setzt das {@link SimpleFeature}, das in der Tabelle dargestellt wird.
   * @param fType {@link SimpleFeatureType} fuer den eine Standard-SimpleFeature
   *              angezeigt wird
   */
  public void setFeature(SimpleFeatureType fType) {
    try {
      if ( fType != null )
        setFeature( FeatureUtil.createSampleFeature(fType) );
      else
        setFeature( (SimpleFeature)null );

    } catch (Exception err) {
      throw new RuntimeException(err);
    }
  }

  /**
   * Liefert den {@link SimpleFeatureType}, der in der Tabelle dargestellt wird.
   */
  public SimpleFeatureType getFeatureType() {
    return this.featureType;
  }

  /**
   * Liefert das {@link SimpleFeature}, das in der Tabelle dargestellt wird.
   */
  public SimpleFeature getFeature() {
    return this.feature;
  }

  /**
   * Liefert die Spaltennamen der Tabelle.
   */
  public String[] createColumnNames() {
    return new String[] {
        GeotoolsGUIUtil.RESOURCE.getString("schmitzm.geotools.feature.FeatureTableModel.AttrName"),
        GeotoolsGUIUtil.RESOURCE.getString("schmitzm.geotools.feature.FeatureTableModel.AttrType"),
        GeotoolsGUIUtil.RESOURCE.getString("schmitzm.geotools.feature.FeatureTableModel.AttrValue")
    };
  }


  /**
   * Liefert die Anzahl an Zeilen. Diese entspricht der Anzahl an Attributen
   * des {@link SimpleFeature}.
   */
  public int getRowCount() {
    return featureType != null ? featureType.getAttributeCount() : 0;
  }

  /**
   * Liefert einen Wert der Tabelle.
   * @param rowIndex  Zeilen-Index (beginnend bei 0)
   * @param columnIndex Spalten-Index (beginnend bei 0)
   */
  public Object getValueAt(int rowIndex, int columnIndex) {
    AttributeDescriptor aType = featureType.getAttributeDescriptors().get(rowIndex);
    switch ( columnIndex ) {
      case 0: return aType.getLocalName();
      case 1: return aType.getType().getBinding().getSimpleName();
      case 2: return feature.getAttribute(rowIndex);
    }
    return null;
  }

  /**
   * Setzt einen Wert der Tabelle. Nur fuer Spalte "Wert" (2) relevant.
   * @param rowIndex  Zeilen-Index (beginnend bei 0)
   * @param columnIndex Spalten-Index (beginnend bei 0)
   */
  @Override
  public void setValueAt(Object value, int rowIndex, int columnIndex) {
    AttributeDescriptor aType = featureType.getAttributeDescriptors().get(rowIndex);
    if ( "".equals(value) )
      value = null;
    final Class<?> bindingClass = aType.getType().getBinding();
	if ( value != null &&  Number.class.isAssignableFrom(bindingClass) )
      value = BaseTypeUtil.convertFromString(value.toString(), bindingClass);
    if ( value == null && !aType.isNillable() )
      value = FeatureUtil.getDefaultAttributeValue(bindingClass);

    try {
      switch ( columnIndex ) {
        case 2: feature.setAttribute(rowIndex, value);
      }
    } catch (Exception err) {
      throw new RuntimeException(err);
    }
  }

  /**
   * Liefert {@code true} nur fuer die Spalte "Wert" (2) und nur
   * dann, wenn es sich um eine numerisches oder String-Attribut
   * handelt.
   */
  @Override
  public boolean isCellEditable(int rowIndex, int columnIndex) {
    Class aClass = featureType.getAttributeDescriptors().get(rowIndex).getType().getBinding();
    return columnIndex == 2 && BaseTypeUtil.isBaseType( aClass );
  }

}
