/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.feature;

import java.util.Vector;

import javax.swing.table.TableModel;

import org.opengis.feature.simple.SimpleFeatureType;
import org.opengis.feature.type.AttributeDescriptor;

import schmitzm.geotools.gui.GeotoolsGUIUtil;
import schmitzm.swing.table.AbstractTableModel;


/**
 * Diese Klasse stellt ein {@link TableModel} auf einem {@link SimpleFeatureType}
 * dar. Dieses definiert zwei Spalten, in denen der Name (0) und der
 * Typ (1) der Attribute dargestellt werden.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class FeatureTypeTableModel extends AbstractTableModel {
  /** {@link SimpleFeatureType} das in der Tabelle dargestellt wird. */
  protected SimpleFeatureType featureType = null;
  /** Bestimmt die angezeigten Features */
  protected AttributeTypeFilter attrFilter = AttributeTypeFilter.ALL;

  /** Speichert die (die gefilterten!) Attibut-Typen des Features.*/
  protected Vector<AttributeDescriptor> attrTypes = new Vector<AttributeDescriptor>();

  /**
   * Erstellt ein leeres Tabellen-Modell.
   */
  public FeatureTypeTableModel() {
    this(null);
  }

  /**
   * Erstellt ein neues Tabellen-Modell.
   * @param type ein SimpleFeatureType
   */
  public FeatureTypeTableModel(SimpleFeatureType type) {
    super();
    setFeatureType( type );
  }

  /**
   * Setzt den {@link SimpleFeatureType}, der in der Tabelle dargestellt wird.
   * @param ftype {@link SimpleFeatureType}
   */
  public void setFeatureType(SimpleFeatureType ftype) {
    this.featureType = ftype;
    // Attribut-Typen in Array speichern
    attrTypes.clear();
    if ( featureType != null )
      for (int i=0; i<featureType.getAttributeCount(); i++) {
        AttributeDescriptor type = featureType.getAttributeDescriptors().get(i);
        if ( attrFilter == null || attrFilter.accept( type ) )
          attrTypes.add( type );
      }
    this.fireTableDataChanged();
  }

  /**
   * Liefert den Filter, der die dargestellten Attribute bestimmt.
   */
  public AttributeTypeFilter getAttributeFilter() {
    return attrFilter;
  }

  /**
   * Setzt den Filter, der die dargestellten Attribute bestimmt.
   * @param attrFilter Filter
   */
  public void setAttributeFilter(AttributeTypeFilter attrFilter) {
    this.attrFilter = attrFilter;
    // Filter neu anwenden
    setFeatureType( featureType );
  }

  /**
   * Liefert den {@link SimpleFeatureType}, der in der Tabelle dargestellt wird.
   */
  public SimpleFeatureType getFeatureType() {
    return this.featureType;
  }



  /**
   * Liefert die Spaltennamen der Tabelle.
   */
  public String[] createColumnNames() {
    return new String[] {
        GeotoolsGUIUtil.RESOURCE.getString("schmitzm.geotools.feature.FeatureTableModel.AttrName"),
        GeotoolsGUIUtil.RESOURCE.getString("schmitzm.geotools.feature.FeatureTableModel.AttrType"),
    };
  }


  /**
   * Liefert die Anzahl an Zeilen. Diese entspricht der Anzahl an Attributen
   * des {@link SimpleFeatureType}.
   */
  public int getRowCount() {
    return /*featureType*/attrTypes != null ? attrTypes.size() : 0;
  }

  /**
   * Liefert einen Wert der Tabelle.
   * @param rowIndex  Zeilen-Index (beginnend bei 0)
   * @param columnIndex Spalten-Index (beginnend bei 0)
   */
  public Object getValueAt(int rowIndex, int columnIndex) {
    AttributeDescriptor aType = attrTypes.elementAt(rowIndex);
    switch ( columnIndex ) {
      case 0: return aType.getLocalName();
      case 1: return aType.getType().getBinding().getSimpleName();
    }
    return null;
  }
}
