/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.feature;

import hep.aida.bin.DynamicBin1D;

import java.awt.Color;
import java.io.IOException;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Random;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.Vector;

import javax.measure.quantity.Length;
import javax.measure.unit.SI;
import javax.measure.unit.Unit;

import org.apache.log4j.Logger;
import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.coverage.grid.io.AbstractGridCoverage2DReader;
import org.geotools.data.DataUtilities;
import org.geotools.data.DefaultQuery;
import org.geotools.data.FeatureSource;
import org.geotools.data.Query;
import org.geotools.data.memory.MemoryDataStore;
import org.geotools.factory.CommonFactoryFinder;
import org.geotools.factory.GeoTools;
import org.geotools.feature.AttributeTypeFactory;
import org.geotools.feature.DefaultFeatureCollection;
import org.geotools.feature.DefaultFeatureCollections;
import org.geotools.feature.FeatureCollection;
import org.geotools.feature.FeatureIterator;
import org.geotools.feature.GeometryAttributeType;
import org.geotools.feature.NameImpl;
import org.geotools.feature.SchemaException;
import org.geotools.feature.simple.SimpleFeatureBuilder;
import org.geotools.feature.simple.SimpleFeatureTypeBuilder;
import org.geotools.filter.FilterAbstract;
import org.geotools.map.MapLayer;
import org.geotools.resources.CRSUtilities;
import org.geotools.resources.coverage.FeatureUtilities;
import org.geotools.styling.Graphic;
import org.geotools.styling.Mark;
import org.geotools.styling.Style;
import org.geotools.styling.StyleBuilder;
import org.geotools.styling.Symbolizer;
import org.geotools.util.SimpleInternationalString;
import org.opengis.coverage.grid.GridCoverageReader;
import org.opengis.feature.IllegalAttributeException;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.simple.SimpleFeatureType;
import org.opengis.feature.type.AttributeDescriptor;
import org.opengis.feature.type.GeometryDescriptor;
import org.opengis.feature.type.Name;
import org.opengis.filter.Filter;
import org.opengis.filter.FilterFactory;
import org.opengis.filter.FilterFactory2;
import org.opengis.parameter.GeneralParameterValue;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import schmitzm.geotools.FilterUtil;
import schmitzm.lang.LangUtil;
import schmitzm.lang.ResourceProvider;
import skrueger.geotools.StyledFeaturesInterface;

import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.MultiLineString;
import com.vividsolutions.jts.geom.MultiPoint;
import com.vividsolutions.jts.geom.MultiPolygon;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.geom.Polygon;

/**
 * Diese Klasse beinhaltet statische Methoden zum Arbeiten mit Features.
 * 
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *         (University of Bonn/Germany)
 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons Kr&uuml;ger</a>
 * @version 1.1
 */
public class FeatureUtil {
	/**
	 * {@link ResourceProvider}, der die Lokalisation fuer Komponenten des
	 * Package {@code schmitzm.geotools.feature} zur Verfuegung stellt. Diese
	 * sind in properties-Dateien unter {@code
	 * schmitzm.geotools.feature.resource.locales} hinterlegt.
	 */
	public static ResourceProvider RESOURCE = new ResourceProvider(LangUtil
			.extendPackagePath(FeatureUtil.class,
					"resource.locales.FeatureResourceBundle"), Locale.ENGLISH);

	private static final NameImpl GC_NAME = new NameImpl(
			"http://www.opengis.net/gml", "GridCoverage");

	/**
	 * Convenience method to access the {@link ResourceProvider}.
	 */
	public static String R(String key, Object... values) {
		return RESOURCE.getString(key, values);
	}

	private static final Logger LOGGER = LangUtil
			.createLogger(FeatureUtil.class);

	private static final String DEFAULT_VECTOR_STYLE_NAME = "default vector style";

	private static StyleBuilder STYLE_BUILDER = new StyleBuilder();

	/** Instance of JTS-GeometryFactory. */
	public static final GeometryFactory GEOMETRY_FACTORY = new GeometryFactory();

	/** Instance of AttributeTypeFactory. */
	public static final AttributeTypeFactory ATTRTYPE_FACTORY = AttributeTypeFactory
			.defaultInstance();

	/** Instance of {@link FilterFactory}. */
	public static final org.opengis.filter.FilterFactory FILTER_FACTORY = CommonFactoryFinder
			.getFilterFactory(GeoTools.getDefaultHints());

	/** Instance of {@link FilterFactory2}. */
	public static final FilterFactory2 FILTER_FACTORY2 = CommonFactoryFinder
			.getFilterFactory2(GeoTools.getDefaultHints());

	/**
	 * Caches the already created {@link SimpleFeatureBuilder
	 * SimpleFeatureBuilders} for reusing in
	 * {@link #createFeature(SimpleFeatureType, String, Object[])}.
	 */
	private static Map<SimpleFeatureType, SimpleFeatureBuilder> featureBuilderCache = new HashMap<SimpleFeatureType, SimpleFeatureBuilder>();

	/** Join-Types. */
	public static enum JoinType {
		/**
		 * The feature of the left collection is taken into the result, even
		 * there is no join feature in the right collection.
		 */
		LEFT_OUTER,
		/**
		 * The feature of the right collection is taken into the result, even
		 * there is no join feature in the left collection.
		 */
		RIGHT_OUTER,
		/**
		 * Only features with a join feature in the other collection are taken
		 * into the result (Default).
		 */
		FULL
	}

	/** The geometry type of a {@link FeatureCollection} */
	public static enum GeometryForm {
		/** Point */
		POINT,
		/** Line or LineString */
		LINE,
		/** Polygon */
		POLYGON
	}

	/**
	 * Enthaelt die {@link AttributeDescriptor AttributeTypes}, fuer die
	 * automatisch Werte generiert werden koennen.
	 */
	private static HashMap<AttributeDescriptor, AutoValueGenerator<?>> autoAttrValueGenerators = new HashMap<AttributeDescriptor, AutoValueGenerator<?>>();

	/**
	 * Determines the kind of geometry of a {@link SimpleFeatureType}.
	 * 
	 * @param fType
	 *            a feature type
	 */
	public static GeometryForm getGeometryForm(SimpleFeatureType fType) {
		if (fType.getGeometryDescriptor() == null)
			return null;

		GeometryDescriptor geometryType = fType.getGeometryDescriptor();
		return getGeometryForm(geometryType);
	}

	/**
	 * Determines the kind of geometry of a {@link GeometryDescriptor}.
	 */
	public static GeometryForm getGeometryForm(GeometryDescriptor gDesc) {
		return getGeometryForm(gDesc.getType().getBinding());
	}

	/**
	 * Determines the kind of geometry of a {@link GeometryAttributeType}.
	 * 
	 * @param geometryType
	 *            a {@link GeometryAttributeType}
	 */
	public static GeometryForm getGeometryForm(
			GeometryAttributeType geometryType) {

		return getGeometryForm(geometryType.getBinding());
	}

	/**
	 * Determines the kind of geometry of a {@link Class}.
	 */
	public static GeometryForm getGeometryForm(Class<?> geometryType) {

		if (Point.class.isAssignableFrom(geometryType)
				|| MultiPoint.class.isAssignableFrom(geometryType))
			return GeometryForm.POINT;
		else if (LineString.class.isAssignableFrom(geometryType)
				|| MultiLineString.class.isAssignableFrom(geometryType))
			return GeometryForm.LINE;
		else if (Polygon.class.isAssignableFrom(geometryType)
				|| MultiPolygon.class.isAssignableFrom(geometryType))
			return GeometryForm.POLYGON;

		throw new UnsupportedOperationException("Unknown geometry type: "
				+ geometryType.getName());
	}

	/**
	 * Determines the kind of geometry of a {@link FeatureCollection}.
	 * 
	 * @param fc
	 *            a feature collection
	 */
	public static GeometryForm getGeometryForm(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc) {
		return getGeometryForm(fc.getSchema());
	}

	/**
	 * Determines the kind of geometry of a {@link FeatureSource}.
	 * 
	 * @param fs
	 *            a feature source
	 */
	public static GeometryForm getGeometryForm(
			FeatureSource<SimpleFeatureType, SimpleFeature> fs) {
		return getGeometryForm(fs.getSchema());
	}

	/**
	 * Determines the kind of geometry of a {@link MapLayer}.
	 * 
	 * @param layer
	 *            a map layer
	 */
	public static GeometryForm getGeometryForm(MapLayer layer) {
		return getGeometryForm((SimpleFeatureType) layer.getFeatureSource().getSchema());
	}

	/**
	 * Extrahiert alle Geometrien aus einem SimpleFeature.
	 * 
	 * @param f
	 *            SimpleFeature
	 * @return Array aller Geometrien im SimpleFeature.
	 */
	public static Geometry[] extractGeometriesToGeometry(SimpleFeature f) {
		Vector<Object> geomVec = new Vector<Object>();

		for (int j = 0; j < f.getFeatureType().getAttributeCount(); j++)
			// if ( f.getFeatureType().getAttributeType(j).isGeometry() )
			if (f.getFeatureType() instanceof GeometryAttributeType)
				geomVec.add(f.getAttribute(j));

		Geometry[] geomArr = new Geometry[geomVec.size()];
		for (int i = 0; i < geomVec.size(); i++)
			geomArr[i] = (Geometry) geomVec.elementAt(i);
		return geomArr;
	}

	/**
	 * Extrahiert alle Default-Geometrien aus einer FeatureCollection.
	 * 
	 * @param fc
	 *            FeatureCollection
	 * @return Array aller Geometrien in der FeatureCollection.
	 */
	public static Vector<Geometry> extractGeometries(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc,
			Vector<Geometry> result) {
		if (result == null)
			result = new Vector<Geometry>();

		FeatureIterator<SimpleFeature> fi = fc.features();
		try {
			for (; fi.hasNext();) {
				SimpleFeature f = fi.next();
				result.add((Geometry) f.getDefaultGeometry());
			}
		} finally {
			fc.close(fi);
		}
		return result;
	}

	/**
	 * Erzeugt einen Standard-Style fuer eine {@link FeatureCollection} Und
	 * setzt eine default namen
	 * 
	 * @param fc
	 *            FeatureCollection
	 */
	public static Style createDefaultStyle(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc) {
		return createDefaultStyle(fc.getSchema().getGeometryDescriptor());
	}

	/**
	 * Erzeugt einen Standard-Style fuer einen {@link GeometryAttributeType} Und
	 * setzt eine default Namen.
	 * 
	 * @param geometryAttrib
	 *            GeometryAttributeType
	 */
	public static Style createDefaultStyle(GeometryDescriptor geometryAttrib) {
		Style style = null;

		if (geometryAttrib != null
				&& (com.vividsolutions.jts.geom.Polygon.class
						.isAssignableFrom(geometryAttrib.getType().getBinding()) || com.vividsolutions.jts.geom.MultiPolygon.class
						.isAssignableFrom(geometryAttrib.getType().getBinding())))
			style = createPolygonStyle(Color.ORANGE, Color.BLACK, 1);
		else if (geometryAttrib != null
				&& (com.vividsolutions.jts.geom.Point.class
						.isAssignableFrom(geometryAttrib.getType().getBinding()) || com.vividsolutions.jts.geom.MultiPoint.class
						.isAssignableFrom(geometryAttrib.getType().getBinding())))
			style = createPointStyle(Color.RED);
		else
			style = createLineStyle(Color.BLUE, 1);

		style.setName(DEFAULT_VECTOR_STYLE_NAME);

		// 26 style.setTitle(DEFAULT_VECTOR_STYLE_NAME);
		style.getDescription().setTitle(
				new SimpleInternationalString(DEFAULT_VECTOR_STYLE_NAME));
		return style;
	}

	/**
	 * Erzeugt einen Polygon-Style.
	 * 
	 * @param fillColor
	 *            Fuell-Farbe
	 * @param borderColor
	 *            Rand-Farbe
	 * @param borderWidth
	 *            Breite des Rands
	 */
	public static Style createPolygonStyle(Color fillColor, Color borderColor,
			double borderWidth) {
		final Symbolizer symb = STYLE_BUILDER.createPolygonSymbolizer(
				fillColor, borderColor, borderWidth);
		return STYLE_BUILDER.createStyle(symb);
	}

	/**
	 * Erzeugt einen (randlosen) Punkt-Style in Form eines Kreises.
	 * 
	 * @param color
	 *            Farbe des Punkts
	 */
	public static Style createPointStyle(Color fillColor) {
		return createPointStyle(fillColor, fillColor, 5);
	}

	/**
	 * Erzeugt einen (randlosen) Punkt-Style in Form eines Kreises.
	 * 
	 * @param fillColor
	 *            Fuell-Farbe des Punkts
	 * @param borderColor
	 *            Rand-Farbe des Punkts
	 * @param width
	 *            Groesse des Punkts
	 */
	public static Style createPointStyle(Color fillColor, Color borderColor,
			double width) {
		return createPointStyle(StyleBuilder.MARK_CIRCLE, fillColor,
				borderColor, 1, 1, width, 0);
	}

	/**
	 * Erzeugt einen Punkt-Style.
	 * 
	 * @param color
	 *            Farbe des Punkts
	 * @param markStyle
	 *            Darstellungsweise des Punkts
	 * @param opacity
	 * @see StyleBuilder#MARK_ARROW;
	 * @see StyleBuilder#MARK_CIRCLE;
	 * @see StyleBuilder#MARK_CROSS;
	 * @see StyleBuilder#MARK_SQUARE;
	 * @see StyleBuilder#MARK_STAR;
	 * @see StyleBuilder#MARK_TRIANGLE;
	 * @see StyleBuilder#MARK_X;
	 */
	public static Style createPointStyle(String markStyle, Color fillColor,
			Color borderColor, double borderWidth, double opacity, double size,
			double rotation) {
		final Mark mark = STYLE_BUILDER.createMark(markStyle, fillColor,
				borderColor, borderWidth);
		final Graphic g = STYLE_BUILDER.createGraphic(null, mark, null,
				opacity, size, rotation);
		final Symbolizer symb = STYLE_BUILDER.createPointSymbolizer(g);
		return STYLE_BUILDER.createStyle(symb);
	}

	/**
	 * Erzeugt einen Linien-Style.
	 * 
	 * @param lineColor
	 *            Farbe der Linie
	 */
	public static Style createLineStyle(Color lineColor) {
		return createLineStyle(lineColor, 1);
	}

	/**
	 * Erzeugt einen Linien-Style.
	 * 
	 * @param lineColor
	 *            Farbe der Linie
	 * @param lineWidth
	 *            Breite der Linie
	 */
	public static Style createLineStyle(Color lineColor, double lineWidth) {
		final Symbolizer symb = STYLE_BUILDER.createLineSymbolizer(lineColor,
				lineWidth);
		return STYLE_BUILDER.createStyle(symb);
	}

	/**
	 * Erzeugt einen Array von {@link SimpleFeature Features} auf einer
	 * {@link FeatureCollection}.<br>
	 * Aufgrund von Problemen bei der Verwendung von Filtern ({@code .size()}
	 * liefert nicht die Anzahl der gefilterten Features; {@code .toArray}
	 * liefert manchmal {@code null}), wird der Array "naiv" ueber Durchlaufen
	 * eines {@link FeatureIterator} kopiert.
	 * 
	 * @param fc
	 *            eine SimpleFeature Collection
	 * @return einen leeren Array, wenn die Collection leer oder {@code null}
	 *         ist
	 */
	public static SimpleFeature[] featuresToArray(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc) {
		return featuresToArray(fc, true);
	}

	/**
	 * Erzeugt einen Array von {@link SimpleFeature Features} auf einer
	 * {@link FeatureCollection}.
	 * 
	 * @param fc
	 *            eine SimpleFeature Collection
	 * @param includeNullFeatures
	 *            wenn {@code false} werden NULL-Features in der Collection
	 *            nicht in den Array kopiert
	 * @return einen leeren Array, wenn die Collection leer oder {@code null}
	 *         ist
	 */
	public static SimpleFeature[] featuresToArray(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc,
			boolean includeNullFeatures) {
		if (fc == null || fc.isEmpty())
			return new SimpleFeature[0];

		// long startTime = System.currentTimeMillis();

		// Naiv alle Features durchgehen.
		final Vector<SimpleFeature> fv = new Vector<SimpleFeature>();
		final FeatureIterator<SimpleFeature> fi = fc.features();
		try {

			while (fi.hasNext()) {
				try {
					final SimpleFeature f = fi.next();
					if (includeNullFeatures || f != null)
						fv.add(f);
				} catch (Exception e) {
					LOGGER.info("error reading a feature, but continuing",e);
					// If reading one features fails (as it happens a lot with
					// OSM data, we still want the others to appear
					// Caused by: java.lang.IllegalArgumentException: Points of
					// LinearRing do not form a closed linestring
					// at
					// com.vividsolutions.jts.geom.LinearRing.validateConstruction(LinearRing.java:105)
					// at
					// com.vividsolutions.jts.geom.LinearRing.<init>(LinearRing.java:100)
					// at
					// com.vividsolutions.jts.geom.GeometryFactory.createLinearRing(GeometryFactory.java:339)
					// at
					// org.geotools.data.shapefile.shp.PolygonHandler.read(PolygonHandler.java:188)
				}
			}
		} finally {
			fc.close(fi);
		}

		// LOGGER.info("took "+ (System.currentTimeMillis()-startTime)
		// +"to put features into vector");

		return fv.toArray(new SimpleFeature[0]);
	}
	

	/**
	 * Clones an {@link SimpleFeatureType}.
	 * 
	 * @param fType
	 *            type to clone
	 * @param fTypeName
	 *            the name for the clone (if {@code null} the name is taken from
	 *            {@code fType})
	 * @param nillable
	 *            nillable property for <b>all</b> attributes (if {@code null}
	 *            the nillable property is taken individually from the source
	 *            attributes)
	 */
	public static SimpleFeatureType cloneFeatureType(SimpleFeatureType fType,
			String fTypeName, Boolean nillable) throws SchemaException {
		if (fTypeName == null)
			fTypeName = fType.getTypeName();

		SimpleFeatureTypeBuilder builder = new SimpleFeatureTypeBuilder();
		builder.setName(fType.getTypeName());
		for (AttributeDescriptor aType : fType.getAttributeDescriptors()) {
			AttributeDescriptor aTypeClone = cloneAttributeType(aType, nillable);
			builder.add(aTypeClone);
			if (aTypeClone instanceof GeometryAttributeType
					&& builder.getDefaultGeometry() == null)
				builder.setDefaultGeometry(aTypeClone.getLocalName());
		}
		return builder.buildFeatureType();
	}

	/**
	 * Clones an {@link AttributeDescriptor}.
	 * 
	 * @param aType
	 *            type to clone
	 * @param nillable
	 *            the clone gets this value for its nillable property
	 *            independently of {@code aType.isNillable()} (if {@code null}
	 *            the nillable property is taken from {@code aType})
	 * 
	 */
	public static AttributeDescriptor cloneAttributeType(
			AttributeDescriptor aType, Boolean nillable) {
		// If no special nillable property is specified, take it from the
		// given AttributeDescriptor
		if (nillable == null)
			nillable = aType.isNillable();

		// Create a default value for the attribute
		Object defaultValue = nillable ? null : getDefaultAttributeValue(aType);

		Object metaData = null;
		// if it is a GeometryAttributeType, the CRS must be stored
		// in the meta data
		if (aType instanceof GeometryAttributeType)
			metaData = ((GeometryAttributeType) aType).getCoordinateSystem();

		// combine the restrictions of the attribute type for the
		// AttributeTypeFactory
		Filter restrictions = FilterUtil.FILTER_FAC.and(aType.getType()
				.getRestrictions());

		// Create the clone
		return AttributeTypeFactory.newAttributeType(aType.getLocalName(),
				aType.getType().getBinding(), nillable, restrictions,
				defaultValue, metaData);
	}
	
    /**
     * Applies the following restrictions to an attribute name.
     * <ul>
     *   <li>Truncates white spaces at start and end.</li>
     *   <li>Truncates name to 10 characters</li>
     *   <li>Transforms name to upper case letters</li>
     * </ul>
     * @param attrName an attribute name
     * @see #checkAttributeNameRestrictionsAndError(String)
     */
    public static String applyAttributeNameRestrictions(String attrName) {
      // NULL not allowed
      if ( attrName != null ) {
        attrName = attrName.trim();
        attrName = attrName.toUpperCase();
        if ( attrName.length() > 10 )
          attrName = attrName.substring(0,10);
      }
      return attrName;
    }
    
    /**
     * Throws an exception if an attribute name does not satisfy all of
     * the following restrictions.
     * <ul>
     *   <li>{@code null} and empty string not allowed</li>
     *   <li>Maximum of 10 characters</li>
     *   <li>Starting with a letter (a-z,A-Z)</li>
     *   <li>Only combination of letters (a-z,A-Z), digits (0-9) and '_'</li>
     * </ul>
     * @param attrName an attribute name
     * @see #checkAttributeNameRestrictionsAndError(String)
     */
	public static void checkAttributeNameRestrictionsAndError(String attrName) throws Exception {
	  // NULL not allowed
	  if ( attrName == null || attrName.trim().equals("") )
	    throw new Exception( RESOURCE.getString("org.geotools.err.attr.name.null") );
	  // Maximum 10 characters
	  if ( attrName.length() > 10 )
	    throw new Exception( RESOURCE.getString("org.geotools.err.attr.name.length") );
	  // Digit not allowed as first character
	  if ( !attrName.matches("[a-zA-Z].*") )
	    throw new Exception( RESOURCE.getString("org.geotools.err.attr.name.start") );
      // Only "normal" characters, digits and '_' allowed
      if ( !attrName.matches("\\w+") )
        throw new Exception( RESOURCE.getString("org.geotools.err.attr.name.chars") );
	}

	/**
	 * Checks whether an attribute name satisfies all restrictions.
	 * @param attrName an attribute name
	 * @see #checkAttributeNameRestrictionsAndError(String)
	 */
	public static boolean checkAttributeNameRestrictions(String attrName) {
	  try {
	    checkAttributeNameRestrictionsAndError(attrName);
	  } catch (Exception err) {
	    return false;
	  }
	  return true;
	}

	
	/**
	 * Erweitert das Schema einer {@link FeatureCollection} um eine Reihe von
	 * Attributen.
	 * 
	 * @param fType
	 *            zu erweiterndes SimpleFeature-Schema
	 * @param aTypes
	 *            Typen der neuen Attribute
	 * @return eine <b>neue</b> Instanz von {@link SimpleFeatureType}
	 * @throws SchemaException
	 *             falls das Erweitern des SimpleFeature-Schemas scheitert
	 * @throws NullPointerException
	 *             falls {@code fType = null}
	 */
	public static SimpleFeatureType extendFeatureType(SimpleFeatureType fType,
			List<AttributeDescriptor> aTypes) throws SchemaException {
		return extendFeatureType(fType, aTypes
				.toArray(new AttributeDescriptor[0]));
	}

	/**
	 * Erweitert das Schema einer {@link FeatureCollection} um eine Reihe von
	 * Attributen.
	 * 
	 * @param fType
	 *            zu erweiterndes SimpleFeature-Schema
	 * @param aTypes
	 *            Typen der neuen Attribute
	 * @return eine <b>neue</b> Instanz von {@link SimpleFeatureType}
	 * @throws SchemaException
	 *             falls das Erweitern des SimpleFeature-Schemas scheitert
	 * @throws NullPointerException
	 *             falls {@code fType = null}
	 */
	public static SimpleFeatureType extendFeatureType(SimpleFeatureType fType,
			AttributeDescriptor... aTypes) throws SchemaException {
		// BEMERKUNG: Folgende Methode funktioniert, fuegt die abgeleiteten
		// (alten)
		// Attribute aber hinten an die neuen Attribute an. Dies ist
		// nicht gewuenscht und erschwert auch den Aufbau des fValues-Array
		// SimpleFeatureType resultType = FeatureTypeBuilder.newFeatureType(
		// aTypes, // new Attributes
		// fType.getTypeName(), // new type name
		// null, // namespace
		// false, // abstact?
		// new SimpleFeatureType[] { fType },
		// null
		// );

		SimpleFeatureTypeBuilder builder = new SimpleFeatureTypeBuilder();
		builder.setName(fType.getName());
		builder.addAll(fType.getAttributeDescriptors()); // old attributes
		for (AttributeDescriptor aType : aTypes) {
			// combine the restrictions of the attribute type for the
			// AttributeTypeFactory
			Filter restrictions = FilterUtil.FILTER_FAC.and(aType.getType()
					.getRestrictions());

			try {
				builder.add(aType);
				if (aType instanceof GeometryAttributeType
						&& builder.getDefaultGeometry() == null)
					builder.setDefaultGeometry(aType.getLocalName());
			} catch (IllegalArgumentException err) {
				builder.add(AttributeTypeFactory.newAttributeType(aType
						.getLocalName()
						+ "_2", aType.getType().getBinding(), aType
						.isNillable(), restrictions, aType.getDefaultValue(),
						null));
			}
		}
		SimpleFeatureType resultType = builder.buildFeatureType();

		// As long as DataUtilities.defaultValue(AttributeDescriptor) does
		// not handle the default value correctly for not-nillable
		// attributes, create a new SimpleFeatureType with all attributes
		// nillable
		// not necessary as long a correct default value can be created
		// in createFeatureType(ResultSetMetaData, ...)
		// resultType = cloneFeatureType(resultType, null, true);

		return resultType;
	}

	/**
	 * Orderes {@link SimpleFeature Features} according to an attribute
	 * (ascending).
	 * 
	 * @param fi
	 *            the features to order
	 * @param attrName
	 *            the attribute to order by
	 * @return a list of ordered features
	 */
	public static Vector<SimpleFeature> sortFeatures(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc,
			String attrName) {
		final FeatureCollection<SimpleFeatureType, SimpleFeature> source = fc;
		FeatureIterator<SimpleFeature> iterator = source.features();
		try {
			return sortFeatures(iterator, attrName);
		} finally {
			source.close(iterator);
		}
	}

	/**
	 * Orderes {@link SimpleFeature Features} according to an attribute
	 * (ascending).
	 * 
	 * @warn When using this method, it's your responsibility t close the
	 *       iterator afterwards!
	 * @see #sortFeatures(FeatureCollection, String)
	 * 
	 * @param fi
	 *            an iterator for the features
	 * @param attrName
	 *            the attribute to order by
	 * @return a list of ordered features
	 */
	protected static Vector<SimpleFeature> sortFeatures(
			FeatureIterator<SimpleFeature> fi, String attrName) {
		// First create a SortedMap with CollisionList
		SortedMap<Comparable<?>, Vector<SimpleFeature>> sortedFeatureLists = new TreeMap<Comparable<?>, Vector<SimpleFeature>>();
		for (; fi.hasNext();) {
			SimpleFeature f = fi.next();
			// Check whether attribute value is Comparable
			Object attrValue = f.getAttribute(attrName);
			if (attrValue != null && !(attrValue instanceof Comparable))
				throw new UnsupportedOperationException(
						"SimpleFeature sort only supported for Comparable attributes: "
								+ LangUtil.getSimpleClassName(attrValue));
			// Determine X value as sort attribute (NULL not permitted for
			// XYDateset!)
			Comparable<?> xValue = (Comparable<?>) attrValue;
			if (xValue == null)
				continue;

			// Check whether collision list exists
			Vector<SimpleFeature> collList = sortedFeatureLists.get(xValue);
			if (collList == null) {
				collList = new Vector<SimpleFeature>();
				sortedFeatureLists.put(xValue, collList);
			}
			// Add feature to collision list
			collList.add(f);
		}

		// Put the (now ordered) features from the collision lists
		// into an temporary list to create a list of Features
		Vector<SimpleFeature> sortedFeatures = new Vector<SimpleFeature>();
		for (Vector<SimpleFeature> collList : sortedFeatureLists.values())
			for (SimpleFeature f : collList)
				sortedFeatures.add(f);

		return sortedFeatures;
	}

	/**
	 * Erweitert das Schema einer {@link FeatureCollection} um eine Reihe von
	 * Attributen.
	 * 
	 * @param fc
	 *            zu erweiternde FeatureCollection
	 * @param aTypes
	 *            Typen der neuen Attribute
	 * @return eine <b>neue</b> Instanz von {@link FeatureCollection}
	 * @throws SchemaException
	 *             falls das Erweitern des SimpleFeature-Schemas scheitert
	 * @throws IllegalAttributeException
	 *             falls das Einfuegen der Features in die neue
	 *             {@link FeatureCollection} scheitert
	 */
	public static FeatureCollection<SimpleFeatureType, SimpleFeature> extendFeatureCollection(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc,
			AttributeDescriptor... aTypes) throws SchemaException,
			IllegalAttributeException {
		// Schema erstellen
		SimpleFeatureType fType = fc.getSchema();
		SimpleFeatureType resultType = extendFeatureType(fType, aTypes);

		// Neue Collection erstellen
		FeatureCollection<SimpleFeatureType, SimpleFeature> resultFc = DefaultFeatureCollections
				.newCollection();

		FeatureIterator<SimpleFeature> fi = fc.features();
		try {

			// Array fuer die Attribut-Werte eines Features
			List<Object> fValues = new ArrayList<Object>(resultType
					.getAttributeCount());
			for (; fi.hasNext();) {
				// Werte der alten Attribute in Array schreiben
				fValues = fi.next().getAttributes();

				// Default-Werte der neuen Attribute in Array schreiben
				for (int i = fType.getAttributeCount(); i < fValues.size(); i++)
					fValues.set(i, resultType.getAttributeDescriptors().get(i)
							.getDefaultValue());
				// Erweitertes SimpleFeature erzeugen und FeatureCollection
				// fuellen
				resultFc.add(createFeature(resultType, fValues));
			}
		} finally {
			fc.close(fi);
		}

		return resultFc;
	}

	/**
	 * Fuehrt eine Join-Operation zwischen zwei {@link FeatureCollection} durch.
	 * Hierfuer wird eine einfacher Nested-Loop-Join durchgefuehrt. In den
	 * optionalen Parametern koennen Attribut-Namen angegeben werden, die in die
	 * Ergebnis-Colletion uebernommen werden (Projektion).<br>
	 * <b>Achtung:</b> Wird diese Option verwendet, sollte sichergestellt sein,
	 * dass {@code fc1} und {@code fc2} KEINE gleichen Attributnamen hat!
	 * Ansonsten wird (aus technischen Gruenden) das Attribut von {@code fc2}
	 * uebernommen.
	 * 
	 * @param fc1
	 *            erste FeatureCollection
	 * @param joinAttr1
	 *            Attribut-Name in der ersten FeatureCollection
	 * @param compareOp
	 *            Operation, mit der der JOIN-Vergleich durchgefuehrt wird
	 * @param fc2
	 *            zweiten FeatureCollection
	 * @param joinAttr2
	 *            Attribut-Name in der zweiten FeatureCollection
	 * @param projAttr
	 *            Attribute (von {@code fc1} und {@code fc2}), die in das
	 *            Ergebnis übernommen werden (Projektion)
	 * @return eine <b>neue</b> Instanz von {@link FeatureCollection}
	 * @throws SchemaException
	 *             falls das Erweitern des SimpleFeature-Schemas scheitert
	 * @throws IllegalAttributeException
	 *             falls das Einfuegen der Features in die neue
	 *             {@link FeatureCollection} scheitert
	 */
	public static FeatureCollection<SimpleFeatureType, SimpleFeature> joinFeatureCollection(
			JoinType joinType,
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc1,
			String joinAttr1, AttributeFilter compareOp,
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc2,
			String joinAttr2, String... projAttr) throws SchemaException,
			IllegalAttributeException {
		// RIGHT-OUTER-JOIN reaslisiert als "inverser" LEFT-OUTER-JOIN
		// Bemerkung: nur Uebergangsloesung, da nicht ganz sauber!!
		// Attribut-Reihenfolge wird vertauscht...
		if (joinType == JoinType.RIGHT_OUTER)
			// throw new
			// IllegalArgumentException("Join type RIGHT_OUTER not yet supported.");
			return joinFeatureCollection(JoinType.LEFT_OUTER, fc2, joinAttr2,
					compareOp.inverse(), fc1, joinAttr1, projAttr);

		// Flag, ob Projektion erfolgen soll
		boolean projection = projAttr.length > 0;

		// Schema erstellen
		SimpleFeatureType fType1 = fc1.getSchema();
		SimpleFeatureType fType2 = fc2.getSchema();
		SimpleFeatureType resultType = null;
		if (fType1 == null)
			throw new UnsupportedOperationException(
					"JOIN can not be performed. FeatureCollection 1 seems to be empty.");
		if (fType2 == null)
			throw new UnsupportedOperationException(
					"JOIN can not be performed. FeatureCollection 2 seems to be empty.");
		if (fType1.getDescriptor(joinAttr1) == null)
			throw new UnsupportedOperationException(
					"JOIN can not be performed. FeatureCollection 1 does not contain the JOIN-Attribute: "
							+ joinAttr1);
		if (fType2.getDescriptor(joinAttr2) == null)
			throw new UnsupportedOperationException(
					"JOIN can not be performed. FeatureCollection 2 does not contain the JOIN-Attribute: "
							+ joinAttr2);

		// Wenn keine Projektion, dann alle Attribute von fc1 mit denen
		// von fc2 vereinen; sonst neuen SimpleFeatureType aufbauen
		if (!projection)
			resultType = extendFeatureType(fType1, fType2
					.getAttributeDescriptors());
		else {
			// Leeren SimpleFeature-Type erzeugen
			SimpleFeatureTypeBuilder builder = new SimpleFeatureTypeBuilder();
			builder.setName(fType1.getName());
			// Projektions-Attribute hinzufuegen
			for (String attrName : projAttr) {
				// erste Wahl: Attribut von fc2 uebernehmen (da beim Belegen
				// von "fValues" der Attributwert aus fc1 in der inneren
				// FOR-Schleife ggf. durch den Attributwert von fc2
				// ueberschrieben wird!)
				AttributeDescriptor aType = fType2.getDescriptor(attrName);
				// zweite Wahl: Attribut von fc1 uebernehmen
				if (aType == null)
					aType = fType1.getDescriptor(attrName);
				// Attribut in den SimpleFeatureType einfuegen
				if (aType != null)
					builder.add(aType);
			}
			// SimpleFeatureType erzeugen
			resultType = builder.buildFeatureType();
		}

		// Neue Collection erstellen
		DefaultFeatureCollection resultFc = (DefaultFeatureCollection) DefaultFeatureCollections
				.newCollection();
		FeatureIterator<SimpleFeature> fi = fc1.features();
		try {
			// Array fuer die Attribut-Werte eines Features
			Object[] fValues = new Object[resultType.getAttributeCount()];

			// NestedLoopJoin: Alle Features der ersten Collection durchgehen
			// und die passenden in der zweiten Collection suchen
			AttributeFilter filter = compareOp.clone();
			for (SimpleFeature f1 = null; fi.hasNext();) {
				f1 = fi.next();
				// Werte der 1. Features in Array schreiben
				if (!projection)
					// Alle Attribut-Werte aus fc1
					getAttributeValues(f1, fValues, 0);
				else
					// Nur Werte der Projektions-Attribute aus fc1
					getAttributeValues(f1, fValues, 0, projAttr);

				// Passende Features in 2. Colletion suchen
				filter.setAttributeName(joinAttr2);
				filter.setCompareValue(f1.getAttribute(joinAttr1));
				FeatureCollection<SimpleFeatureType, SimpleFeature> filteredCollection = fc2
						.subCollection(filter);
				FeatureIterator<SimpleFeature> joinedFeaturesIterator = filteredCollection
						.features();
				try {

					// Wenn LEFT-OUTER-JOIN und JOIN nicht erfolgreich, ein
					// Dummy
					// SimpleFeature mit Default-Werten in die Collection
					// einfuegen
					// Sonst: JOIN-Paare "normal" bilden
					if (!joinedFeaturesIterator.hasNext()
							&& joinType == JoinType.LEFT_OUTER) {
						if (!projection)
							getDefaultAttributeValues(fType2, fValues, fType1
									.getAttributeCount());
						else
							getDefaultAttributeValues(fType2, fValues, 0,
									projAttr);
						// Erweitertes SimpleFeature erzeugen und
						// FeatureCollection
						// fuellen
						resultFc.add(createFeature(resultType, fValues));
					} else {
						for (SimpleFeature f2 = null; joinedFeaturesIterator
								.hasNext();) {
							f2 = joinedFeaturesIterator.next();
							// Werte der 2. Features in Array schreiben
							if (!projection)
								getAttributeValues(f2, fValues, fType1
										.getAttributeCount());
							else
								getAttributeValues(f2, fValues, 0, projAttr);
							// Erweitertes SimpleFeature erzeugen und
							// FeatureCollection
							// fuellen
							SimpleFeature feature = createFeature(resultType,
									fValues);
							resultFc.add(feature);
						}
					}
				} finally {
					filteredCollection.close(joinedFeaturesIterator);
				}
			}
		} finally {
			fc1.close(fi);
		}

		return resultFc;
	}

	/**
	 * Fuehrt eine (Full-Outer-) Join-Operation zwischen zwei
	 * {@link FeatureCollection} durch. Hierfuer wird eine einfacher
	 * Nested-Loop-Join durchgefuehrt. In den optionalen Parametern koennen
	 * Attribut-Namen angegeben werden, die in die Ergebnis-Colletion
	 * uebernommen werden (Projektion).<br>
	 * <b>Achtung:</b> Wird diese Option verwendet, sollte sichergestellt sein,
	 * dass {@code fc1} und {@code fc2} KEINE gleichen Attributnamen hat!
	 * Ansonsten wird (aus technischen Gruenden) das Attribut von {@code fc2}
	 * uebernommen.
	 * 
	 * @param fc1
	 *            erste FeatureCollection
	 * @param joinAttr1
	 *            Attribut-Name in der ersten FeatureCollection
	 * @param compareOp
	 *            Operation, mit der der JOIN-Vergleich durchgefuehrt wird
	 * @param fc2
	 *            zweiten FeatureCollection
	 * @param joinAttr2
	 *            Attribut-Name in der zweiten FeatureCollection
	 * @param projAttr
	 *            Attribute (von {@code fc1} und {@code fc2}), die in das
	 *            Ergebnis übernommen werden (Projektion)
	 * @return eine <b>neue</b> Instanz von {@link FeatureCollection}
	 * @throws SchemaException
	 *             falls das Erweitern des SimpleFeature-Schemas scheitert
	 * @throws IllegalAttributeException
	 *             falls das Einfuegen der Features in die neue
	 *             {@link FeatureCollection} scheitert
	 */
	public static FeatureCollection<SimpleFeatureType, SimpleFeature> joinFeatureCollection(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc1,
			String joinAttr1, AttributeFilter compareOp,
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc2,
			String joinAttr2, String... projAttr) throws SchemaException,
			IllegalAttributeException {
		return joinFeatureCollection(JoinType.FULL, fc1, joinAttr1, compareOp,
				fc2, joinAttr2, projAttr);
	}

	/**
	 * Fuehrt eine (Full-Outer-) Equi-Join-Operation zwischen zwei
	 * {@link FeatureCollection} durch.
	 * 
	 * @param fc1
	 *            erste FeatureCollection
	 * @param joinAttr1
	 *            Attribut-Name in der ersten FeatureCollection
	 * @param fc2
	 *            zweiten FeatureCollection
	 * @param joinAttr2
	 *            Attribut-Name in der zweiten FeatureCollection
	 * @param projAttr
	 *            Attribute (von {@code fc1} und {@code fc2}), die in das
	 *            Ergebnis übernommen werden (Projektion)
	 * @return eine <b>neue</b> Instanz von {@link FeatureCollection}
	 * @throws SchemaException
	 *             falls das Erweitern des SimpleFeature-Schemas scheitert
	 * @throws IllegalAttributeException
	 *             falls das Einfuegen der Features in die neue
	 *             {@link FeatureCollection} scheitert
	 * @see #joinFeatureCollection(FeatureCollection, String, AttributeFilter,
	 *      FeatureCollection, String)
	 * @see #AttributeFilter
	 */
	public static FeatureCollection<SimpleFeatureType, SimpleFeature> joinFeatureCollection(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc1,
			String joinAttr1,
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc2,
			String joinAttr2, String... projAttr) throws SchemaException,
			IllegalAttributeException {
		return joinFeatureCollection(fc1, joinAttr1, AttributeFilter.EQUALS,
				fc2, joinAttr2, projAttr);
	}

	/**
	 * Liefert die Default-Werte fuer alle Attribute eines
	 * {@link SimpleFeatureType}.
	 * 
	 * @param ft
	 *            ein SimpleFeature-Type
	 */
	public static Object[] getDefaultAttributeValues(SimpleFeatureType ft) {
		return getDefaultAttributeValues(ft, null, 0);
	}

	/**
	 * Liefert die Default-Werte fuer eine Reihen von Attributwerten eines
	 * FeatureTypes.
	 * 
	 * @param ft
	 *            ein SimpleFeature-Type
	 * @param destArray
	 *            Array, in dem die Objekte abgelegt werden (wenn {@code null},
	 *            wird ein neuer Array erzeugt)
	 * @param destStartIdx
	 *            Start-Index in {@code destArray} ab dem die Attributwerte
	 *            abgelegt werden
	 * @param attrName
	 *            Attribute aus {@link f}
	 * @return {@code destArray} oder einen neuen Array {@code
	 *         Object[destStartIdx + attrName.length]}
	 */
	public static Object[] getDefaultAttributeValues(SimpleFeatureType ft,
			Object[] destArray, int destStartIdx, String... attrName) {
		// Wenn keine Attribute angegeben sind, alle verwenden
		int attrCount = attrName.length > 0 ? attrName.length : ft
				.getAttributeCount();

		if (destArray == null)
			destArray = new Object[destStartIdx + attrCount];
		if (destArray.length < destStartIdx + attrCount)
			throw new IllegalArgumentException("Destination array "
					+ destArray.length + " with start index " + destStartIdx
					+ " too small for " + attrName.length
					+ " attribute values!");

		// Wenn keine Attribute angegeben sind, alle verwenden
		for (int i = 0; i < attrCount; i++) {
			AttributeDescriptor attrType = attrName.length > 0 ? ft
					.getDescriptor(attrName[i]) : ft.getDescriptor(i);
			// wenn attrType = null, dann gibt es das Attribut im
			// SimpleFeatureType nicht
			// -> destArray nicht ueberschreiben
			if (attrType != null)
				destArray[destStartIdx + i] = getDefaultAttributeValue(attrType);
		}
		return destArray;
	}

	/**
	 * Liefert den Default-Wert fuer ein Attribut.
	 * 
	 * @param attrType
	 *            Attribut-Typ
	 * @return {@code null}, wenn das Attribut NILLABLE ist
	 */
	public static Object getDefaultAttributeValue(AttributeDescriptor attrType) {
		return (getDefaultAttributeValue(attrType, true));
	}

	/**
	 * Liefert den Default-Wert fuer ein Attribut.
	 * 
	 * @param attrType
	 *            Attribut-Typ
	 * @param allowNull
	 *            wenn {@code false}, wird in jedem Fall ein Wert ungleich
	 *            {@code null} geliefert, auch wenn das Attribut NILLABLE ist
	 */
	public static Object getDefaultAttributeValue(AttributeDescriptor attrType,
			boolean allowNull) {
		// AutoValueGenerator avg = getAutoValueGenerator(attrType);
		// Object attrValue = (avg == null) ? attrType.createDefaultValue() :
		// avg.getNextValue();
		Object attrValue = attrType.getDefaultValue();
		if (attrValue == null && (!attrType.isNillable() || !allowNull)) {
			attrValue = getDefaultAttributeValue(attrType.getType()
					.getBinding());
			if (attrValue == null)
				LOGGER
						.warn("Could not create default value for not-null attribute '"
								+ attrType.getLocalName()
								+ "': "
								+ attrType.getType().getBinding()
										.getSimpleName());
		}
		return attrValue;
	}

	/**
	 * Liefert den Default-Wert fuer ein Attribut.
	 * 
	 * @param attrType
	 *            Attribut-Typ
	 * @return {@code null}, falls {@link DataUtilities#defaultValue(Class)}
	 *         keinen Defaul-Wert ermitteln kann (anstelle einer Exception!)
	 */
	public static Object getDefaultAttributeValue(Class<?> attrType) {
		Object attrValue = null;
		try {
			// DataUtilities.defaultValue(.) kann BigDecimal und BigInteger
			// nicht verarbeiten (gibt null zurueck!)
			if (BigDecimal.class.isAssignableFrom(attrType))
				return new BigDecimal(0.0);
			if (BigInteger.class.isAssignableFrom(attrType))
				return new BigInteger("0");
			attrValue = DataUtilities.defaultValue(attrType);
		} catch (IllegalArgumentException err) {
			// no default value could be generated for this type
			// --> ignore that
		}
		return attrValue;
	}

	/**
	 * Liefert eine Reihen von Attributwerten eines Features.
	 * 
	 * @param f
	 *            ein SimpleFeature
	 * @param destArray
	 *            Array, in dem die Objekte abgelegt werden (wenn {@code null},
	 *            wird ein neuer Array erzeugt)
	 * @param destStartIdx
	 *            Start-Index in {@code destArray} ab dem die Attributwerte
	 *            abgelegt werden
	 * @param attrName
	 *            Attribute des Features, deren Werte geliefert werden (wenn
	 *            {@code null} werden alle Attribute geliefert)
	 * @return {@code destArray} oder einen neuen Array {@code
	 *         Object[destStartIdx + attrName.length]}
	 */
	public static Object[] getAttributeValues(SimpleFeature f,
			Object[] destArray, int destStartIdx, String... attrName) {
		if (destArray == null)
			destArray = new Object[destStartIdx + attrName.length];
		if (destArray.length < destStartIdx + attrName.length)
			throw new IllegalArgumentException("Destination array "
					+ destArray.length + " with start index " + destStartIdx
					+ " too small for " + attrName.length
					+ " attribute values!");

		// Wenn keine Attribute angegeben sind, alle verwenden
		int attrCount = attrName.length > 0 ? attrName.length : f
				.getAttributeCount();
		for (int i = 0; i < attrCount; i++) {
			Object attr = attrName.length > 0 ? f.getAttribute(attrName[i]) : f
					.getAttribute(i);
			// wenn attr = null, dann gibt es das Attribut im SimpleFeature
			// nicht
			// -> destArray nicht ueberschreiben
			if (attr != null)
				destArray[destStartIdx + i] = attr;
		}
		return destArray;
	}

	/**
	 * Liefert eine Reihen von Attributwerten eines Features.
	 * 
	 * @param f
	 *            ein SimpleFeature
	 * @param attrName
	 *            Attribute des Features, deren Werte geliefert werden (wenn
	 *            {@code null} werden alle Attribute geliefert)
	 * @return neuer Array {@code Object[attrName.length]}
	 */
	public static Object[] getAttributeValues(SimpleFeature f,
			String... attrName) {
		return getAttributeValues(f, null, 0, attrName);
	}

	/**
	 * Liefert eine Reihen von Attributwerten eines Features.
	 * 
	 * @param f
	 *            ein SimpleFeature
	 * @param destMap
	 *            Map, in dem die Objekte abgelegt werden (wenn {@code null},
	 *            wird eine neue {@link HashMap} erzeugt)
	 * @param attrName
	 *            Attribute des Features, deren Werte geliefert werden (wenn
	 *            {@code null} werden alle Attribute geliefert)
	 * @return {@code destArray} oder einen neuen Array {@code
	 *         Object[destStartIdx + attrName.length]}
	 */
	public static Map<String, Object> getAttributeValues(SimpleFeature f,
			Map<String, Object> destMap, String... attrName) {
		if (destMap == null)
			destMap = new HashMap<String, Object>();

		if (attrName.length > 0)
			for (int i = 0; i < attrName.length; i++)
				destMap.put(attrName[i], f.getAttribute(attrName[i]));
		else
			for (int i = 0; i < f.getAttributeCount(); i++)
				destMap.put(f.getFeatureType().getName().getLocalPart(), f
						.getAttribute(i));

		return destMap;
	}

	/**
	 * Fuehrt eine Join-Operation zwischen einer {@link FeatureCollection} und
	 * einem {@link ResultSet} durch, in dem Zusatz-Informationen hinterlegt.
	 * Hierzu wird das ResultSet zunaecht in eine {@link FeatureCollection}
	 * umgewandelt und anschliessend ein Left-Outer-EquiJoin zwischen den beiden
	 * Collections ausgewertet.
	 * 
	 * @param fc
	 *            eine FeatureCollection
	 * @param fcJoinAttr
	 *            JOIN-Attribut in der FeatureCollection
	 * @param rs
	 *            ResultSet mit Zusatz-Informationen
	 * @param rsJoinAttr
	 *            JOIN-Attribut im ResultSet
	 * @return eine <b>neue</b> Instanz von {@link FeatureCollection}
	 * @throws SchemaException
	 *             falls das Erweitern des SimpleFeature-Schemas scheitert
	 * @throws IllegalAttributeException
	 *             falls das Einfuegen der Features in die neue
	 *             {@link FeatureCollection} scheitert
	 */
	public static FeatureCollection<SimpleFeatureType, SimpleFeature> joinFeatureCollection(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc,
			String fcJoinAttr, ResultSet rs, String rsJoinAttr,
			String... projAttr) throws SchemaException,
			IllegalAttributeException, SQLException {
		return joinFeatureCollection(fc, fcJoinAttr, AttributeFilter.EQUALS,
				rs, rsJoinAttr);
	}

	/**
	 * Fuehrt eine Join-Operation zwischen einer {@link FeatureCollection} und
	 * einem {@link ResultSet} durch, in dem Zusatz-Informationen hinterlegt.
	 * Hierzu wird das ResultSet zunaecht in eine {@link FeatureCollection}
	 * umgewandelt und anschliessend ein Left-Outer-Join zwischen den beiden
	 * Collections ausgewertet.
	 * 
	 * @param fc
	 *            eine FeatureCollection
	 * @param fcJoinAttr
	 *            JOIN-Attribut in der FeatureCollection
	 * @param compareOp
	 *            Operation, mit der der JOIN-Vergleich durchgefuehrt wird
	 * @param rs
	 *            ResultSet mit Zusatz-Informationen
	 * @param rsJoinAttr
	 *            JOIN-Attribut im ResultSet
	 * @return eine <b>neue</b> Instanz von {@link FeatureCollection}
	 * @throws SchemaException
	 *             falls das Erweitern des SimpleFeature-Schemas scheitert
	 * @throws IllegalAttributeException
	 *             falls das Einfuegen der Features in die neue
	 *             {@link FeatureCollection} scheitert
	 */
	public static FeatureCollection<SimpleFeatureType, SimpleFeature> joinFeatureCollection(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc,
			String fcJoinAttr, AttributeFilter compareOp, ResultSet rs,
			String rsJoinAttr, String... projAttr) throws SchemaException,
			IllegalAttributeException, SQLException {
		return joinFeatureCollection(JoinType.LEFT_OUTER, fc, fcJoinAttr,
				compareOp, rs, rsJoinAttr, projAttr);

	}

	/**
	 * Fuehrt eine Join-Operation zwischen einer {@link FeatureCollection} und
	 * einem {@link ResultSet} durch, in dem Zusatz-Informationen hinterlegt.
	 * Hierzu wird das ResultSet zunaecht in eine {@link FeatureCollection}
	 * umgewandelt und anschliessend ein Join zwischen den beiden Collections
	 * ausgewertet.
	 * 
	 * @param joinType
	 *            Join-Typ
	 * @param fc
	 *            eine FeatureCollection
	 * @param fcJoinAttr
	 *            JOIN-Attribut in der FeatureCollection
	 * @param compareOp
	 *            Operation, mit der der JOIN-Vergleich durchgefuehrt wird
	 * @param rs
	 *            ResultSet mit Zusatz-Informationen
	 * @param rsJoinAttr
	 *            JOIN-Attribut im ResultSet
	 * @return eine <b>neue</b> Instanz von {@link FeatureCollection}
	 * @throws SchemaException
	 *             falls das Erweitern des SimpleFeature-Schemas scheitert
	 * @throws IllegalAttributeException
	 *             falls das Einfuegen der Features in die neue
	 *             {@link FeatureCollection} scheitert
	 */
	public static FeatureCollection<SimpleFeatureType, SimpleFeature> joinFeatureCollection(
			JoinType joinType,
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc,
			String fcJoinAttr, AttributeFilter compareOp, ResultSet rs,
			String rsJoinAttr, String... projAttr) throws SchemaException,
			IllegalAttributeException, SQLException {
		if (fcJoinAttr == null || fcJoinAttr.trim().equals("")) {
			// Erste Nicht-Geometrie-Spalte suchen
			for (AttributeDescriptor aType : fc.getSchema()
					.getAttributeDescriptors()) {
				if (aType != fc.getSchema().getGeometryDescriptor()) {
					fcJoinAttr = aType.getName().getLocalPart();
					break;
				}
			}
			LOGGER
					.warn("Join-Attribut for FeatureCollection not set. First attribute used: "
							+ fcJoinAttr);
		}
		if (rsJoinAttr == null || rsJoinAttr.trim().equals("")) {
			rsJoinAttr = rs.getMetaData().getColumnName(1);
			LOGGER
					.warn("Join-Attribut for ResultSet not set. First column used: "
							+ rsJoinAttr);
		}

		// FeatureCollection aus ResultSet erzeugen
		FeatureCollection<SimpleFeatureType, SimpleFeature> rsFeatureCollection = createFeatureCollection(
				rs,
				// fuer den JOIN muss das Join-Attribut vorhanden sein!
				projAttr.length > 0 ? LangUtil
						.extendArray(projAttr, rsJoinAttr) : projAttr);

		// JOIN durchfuehren
		return joinFeatureCollection(joinType, // Join-Typ
				fc, // FeatureCollection
				fcJoinAttr, // Join-Attr. in FeatureCollection
				compareOp, // Join-Operator
				rsFeatureCollection, // ResultSet als FeatureCollection
				rsJoinAttr, // Join-Attr. in ResultSet
				projAttr // Projektions-Attribute
		);
	}

	/**
	 * Erzeugt eine {@link FeatureCollection} aus einem {@linkplain ResultSet
	 * SQL-Result}.
	 * 
	 * @param rs
	 *            SQL-Result
	 * @param projAttr
	 *            nur diese Attribute des ResultSet werden in die
	 *            FeatureCollection uebernommen (wenn nicht angegeben, werden
	 *            ALLE Attribute uebernommen)
	 * @throws SQLException
	 *             falls der Zugriff auf die SQL-Metadaten scheitert
	 * @throws SchemaException
	 *             falls das Erzeugen des SimpleFeatureType scheitert
	 * @throws IllegalAttributeException
	 *             falls das Erzeugen eines Features scheitert
	 */
	public static FeatureCollection<SimpleFeatureType, SimpleFeature> createFeatureCollection(
			ResultSet rs, String... projAttr) throws SchemaException,
			IllegalAttributeException, SQLException {
		// Schema der FeatureCollection erzeugen
		SimpleFeatureType rsFeatureType = createFeatureType(rs.getMetaData(),
				projAttr);

		// Array fuer die Attribut-Werte eines Features
		Object[] rsFeatureValues = new Object[rsFeatureType.getAttributeCount()];

		// ResultSet in FeatureCollection uebertragen
		FeatureCollection<SimpleFeatureType, SimpleFeature> rsFeatureCollection = DefaultFeatureCollections
				.newCollection();
		for (; rs.next();) {
			for (int i = 0; i < rsFeatureType.getAttributeCount(); i++)
				rsFeatureValues[i] = rs.getObject(rsFeatureType
						.getDescriptor(i).getName().getLocalPart());
			rsFeatureCollection.add(createFeature(rsFeatureType,
					rsFeatureValues));
		}

		return rsFeatureCollection;
	}

	/**
	 * Erzeugt einen {@link SimpleFeatureType} aus einem
	 * {@linkplain ResultSetMetaData SQL-Result}.
	 * 
	 * @param metaData
	 *            SQL-Metadaten des {@link ResultSet}
	 * @param projAttr
	 *            nur diese Attribute des ResultSet werden in den
	 *            SimpleFeatureType uebernommen (wenn nicht angegeben, werden
	 *            ALLE Attribute uebernommen)
	 * @throws SQLException
	 *             falls der Zugriff auf die SQL-Metadaten scheitert
	 * @throws SchemaException
	 *             falls das Erzeugen des SimpleFeatureType scheitert
	 */
	public static SimpleFeatureType createFeatureType(
			ResultSetMetaData metaData, String... projAttr)
			throws SQLException, SchemaException {
		try {
			return createFeatureType(metaData, null, Object.class, projAttr);
		} catch (ClassNotFoundException err) {
			// KANN NICHT VORKOMMEN, da Object.class als Standard-Klasse
			// angegeben
			// ist. Exception wird nur abgefangen, damit sie nicht in
			// throw-clause
			// deklariert (und in uebergeordneter Methode) abgefangen werden
			// muss
			return null;
		}
	}

	/**
	 * Erzeugt einen {@link SimpleFeatureType} aus einem {@linkplain ResultSet
	 * SQL-Result}.
	 * 
	 * @param metaData
	 *            SQL-Metadaten des {@link ResultSet}
	 * @param featureTypeName
	 *            Name fuer den SimpleFeatureType (kann {@code null} sein)
	 * @param defaultAttrClass
	 *            Klasse, die verwendet als Attribut-Typ wird, wenn die
	 *            entsprechende Klasse des ResultSet nicht zur Verfuegung steht
	 *            (wenn {@code null}, wird eine Exception geworfen)
	 * @param projAttr
	 *            nur diese Attribute des ResultSet werden in den
	 *            SimpleFeatureType uebernommen (wenn nicht angegeben, werden
	 *            ALLE Attribute uebernommen)
	 * @throws SQLException
	 *             falls der Zugriff auf die SQL-Metadaten scheitert
	 * @throws SchemaException
	 *             falls das Erzeugen des SimpleFeatureType scheitert
	 * @throws ClassNotFoundException
	 *             falls eine Attribut-Klasse aus dem ResultSet nicht erzeugt
	 *             werden kann und keine Default-Klasse angegeben ist
	 */
	public static SimpleFeatureType createFeatureType(
			ResultSetMetaData metaData, String featureTypeName,
			Class<?> defaultAttrClass, String... projAttr) throws SQLException,
			SchemaException, ClassNotFoundException {
		if (featureTypeName == null)
			featureTypeName = "ResultSetFeatureType";

		// Projektions-Attribute in ArrayList uebertragen
		Set<String> projAttrList = new HashSet<String>();
		for (String attrName : projAttr)
			projAttrList.add(attrName);

		// Fuer alle Attribute aus Metadaten einen AttributeDescriptor erzeugen
		// und diesen in SimpleFeatureType uebernehmen
		SimpleFeatureTypeBuilder featureTypeFac = new SimpleFeatureTypeBuilder();
		featureTypeFac.setName(featureTypeName);
		for (int i = 1; i <= metaData.getColumnCount(); i++) {
			String columnName = metaData.getColumnName(i);
			// Attribut nur uebernehmen, wenn als Projektions-Attribut angegeben
			if (projAttr.length > 0 && !projAttrList.contains(columnName))
				continue;

			Class<?> columnClass = defaultAttrClass;
			try {
				columnClass = Class.forName(metaData.getColumnClassName(i));
			} catch (ClassNotFoundException err) {
				if (defaultAttrClass == null)
					throw err;
				// Fehler Ignorieren und Object als Attribut-Klasse verwenden
				LOGGER.warn("Column '" + columnName + "' (" + i + "): Class '"
						+ metaData.getColumnClassName(i) + "' not found. '"
						+ columnClass.getName() + "' used.");
			}
			boolean nillable = metaData.isNullable(i) != ResultSetMetaData.columnNoNulls;
			Object defaultValue = getDefaultAttributeValue(columnClass);
			if (defaultValue == null && !nillable)
				LOGGER
						.warn("Could not create default value for not-null attribute '"
								+ columnName
								+ "': "
								+ columnClass.getSimpleName());

			// Add an attribute for the column to the SimpleFeatureType
			featureTypeFac.add(AttributeTypeFactory.newAttributeType(
					columnName, columnClass, nillable, null, // no restrictions
					defaultValue, null // no meta data
					));
		}
		return featureTypeFac.buildFeatureType();
	}

	/**
	 * Erzeugt einen {@link SimpleFeatureType} aus Beispiel-Attribut-Werten.
	 * 
	 * @param featureTypeName
	 *            Name fuer den SimpleFeatureType (kann {@code null} sein)
	 * @param attrValues
	 *            Attribut-Werte
	 * @param defaultAttrClass
	 *            Standard-Typ, der fuer ein Attribut verwendet wird, wenn der
	 *            angegebene Attribut-Wert {@code null} ist.
	 * @throws SchemaException
	 *             falls das Erzeugen des SimpleFeatureType scheitert
	 */
	public static SimpleFeatureType createFeatureType(String featureTypeName,
			Map<String, Object> attrValues, Class<?> defaultAttrClass)
			throws SchemaException {
		if (featureTypeName == null)
			featureTypeName = "SimpleFeatureType";

		SimpleFeatureTypeBuilder featureTypeFac = new SimpleFeatureTypeBuilder();
		featureTypeFac.setName(featureTypeName);
		for (String attrName : attrValues.keySet()) {
			Object attrValue = attrValues.get(attrName);
			Class<?> attrClass = attrValue == null ? defaultAttrClass
					: attrValue.getClass();
			featureTypeFac.add(AttributeTypeFactory.newAttributeType(attrName,
					attrClass));
		}
		return featureTypeFac.buildFeatureType();
	}

	/**
	 * Creates a default {@link SimpleFeatureType} for a given Geometry class
	 * 
	 * @param geomType
	 *            LineString.class, Point.class or Polygon.class from com.jts...
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static SimpleFeatureType createFeatureType(final Class<?> geomType) {
		// Create the only attribute
		final AttributeDescriptor geometryAttribut = AttributeTypeFactory
				.newAttributeType("the_geom", geomType);
		// Create the FeatureType with the only attribute
		final SimpleFeatureTypeBuilder builder = new SimpleFeatureTypeBuilder();
		builder.setName(geomType.getSimpleName());
		builder.add(geometryAttribut);
		return builder.buildFeatureType();
	}

	/**
	 * Erzeugt einen Geometrie-Attribut-Typ.
	 * 
	 * @param geomType
	 *            Art der Geometrie
	 * @param defGeom
	 *            Standard-Wert
	 * @param crs
	 *            Koordinatensystem
	 */
	public static GeometryAttributeType createGeometryAttributeType(
			Class<? extends Geometry> geomType, Geometry defGeom,
			CoordinateReferenceSystem crs) {
		return createGeometryAttributeType(null, geomType, defGeom, crs);
	}

	/**
	 * Erzeugt einen Geometrie-Attribut-Typ.
	 * 
	 * @param attrName
	 *            Name des attribute (wenn {@code null}, wird "the_geom"
	 *            verwendet)
	 * @param geomType
	 *            Art der Geometrie
	 * @param defGeom
	 *            Standard-Wert
	 * @param crs
	 *            Koordinatensystem
	 */
	public static GeometryAttributeType createGeometryAttributeType(
			String attrName, Class<? extends Geometry> geomType,
			Geometry defGeom, CoordinateReferenceSystem crs) {
		// return new GeometricAttributeType(
		// "the_geom",
		// geomType,
		// false,
		// defGeom,
		// crs,
		// null
		// );
		if (attrName == null)
			attrName = "the_geom";

		return (GeometryAttributeType) AttributeTypeFactory.newAttributeType(
				attrName, geomType, false, // nillable
				null, // restrictions
				defGeom, crs // metadata used for CRS
				);
	}

	/**
	 * Returns a {@link SimpleFeatureBuilder} for a given schema from cache. If
	 * schema is not already used, a new builder is created and cached.
	 * 
	 * @param schema
	 *            the feature schema
	 * @param forceNewBuilder
	 *            forces to create and cache a new builder instance
	 */
	private static SimpleFeatureBuilder getFeatureBuilderFromCache(
			SimpleFeatureType schema, boolean forceNewBuilder) {
		SimpleFeatureBuilder builder = featureBuilderCache.get(schema);
		if (builder == null || forceNewBuilder) {
			builder = new SimpleFeatureBuilder(schema);
			featureBuilderCache.put(schema, builder);
		}
		return builder;
	}

	/**
	 * Registriert einen {@link AutoValueGenerator} fuer einen
	 * {@link AttributeDescriptor}, so dass
	 * {@link #getNextAutoValue(AttributeDescriptor)} fuer diesen
	 * {@link AttributeDescriptor} verwendet werden kann.
	 * 
	 * @param aType
	 *            ein {@link AttributeDescriptor}
	 * @param generator
	 *            generiert automatische Attribut-Werte
	 */
	public static void registerAutoValueGenerator(AttributeDescriptor aType,
			AutoValueGenerator<?> generator) {
		autoAttrValueGenerators.put(aType, generator);
	}

	/**
	 * Entfernt einen {@link AutoValueGenerator} fuer einen
	 * {@link AttributeDescriptor}.
	 * 
	 * @param aType
	 *            ein {@link AttributeDescriptor}
	 */
	public static void unregisterAutoValueGenerator(AttributeDescriptor aType) {
		if (autoAttrValueGenerators.containsKey(aType))
			autoAttrValueGenerators.remove(aType);
	}

	/**
	 * Liefert den {@link AutoValueGenerator} fuer einen
	 * {@link AttributeDescriptor}.
	 * 
	 * @param aType
	 *            ein {@link AttributeDescriptor}
	 * @return {@code null}, wenn fuer den {@link AttributeDescriptor} noch kein
	 *         {@link AutoValueGenerator} registriert worden ist
	 */
	public static AutoValueGenerator<?> getAutoValueGenerator(
			AttributeDescriptor aType) {
		return autoAttrValueGenerators.get(aType);
	}

	/**
	 * Liefert den naechsten Auto-Wert fuer einen {@link AttributeDescriptor}.
	 * 
	 * @param aType
	 *            ein {@link AttributeDescriptor}
	 * @exception UnsupportedOperationException
	 *                falls fuer das Attribut noch kein
	 *                {@link AutoValueGenerator} registriert ist
	 */
	public static Object getNextAutoValue(AttributeDescriptor aType) {
		AutoValueGenerator<?> avg = getAutoValueGenerator(aType);
		if (avg == null)
			throw new UnsupportedOperationException(
					"No AutoValueGenerator registered for attribute type: "
							+ aType.getLocalName());
		return avg.getNextValue();
	}

	/**
	 * Creates a new feature for a given schema.
	 * 
	 * @param schema
	 *            schema for the feature
	 * @param forceNewBuilder
	 *            forces the creation of a new {@link SimpleFeatureBuilder}
	 * @param fID
	 *            feature id for the new feature
	 * @param attrValues
	 *            values of the attributes
	 */
	public static SimpleFeature createFeature(SimpleFeatureType schema,
			boolean forceNewBuilder, String fID, Object... attrValues) {
		SimpleFeatureBuilder builder = getFeatureBuilderFromCache(schema,
				forceNewBuilder);
		return builder.buildFeature(fID, attrValues);
	}

	/**
	 * Creates a new feature for a given schema.
	 * 
	 * @param schema
	 *            schema for the feature
	 * @param fID
	 *            feature id for the new feature
	 * @param attrValues
	 *            values of the attributes
	 */
	public static SimpleFeature createFeature(SimpleFeatureType schema,
			String fID, Object... attrValues) {
		return createFeature(schema, false, fID, attrValues);
	}

	/**
	 * Creates a new feature for a given schema with a random generated id.
	 * 
	 * @param schema
	 *            schema for the feature
	 * @param attrValues
	 *            values of the attributes
	 */
	public static SimpleFeature createFeature(SimpleFeatureType schema,
			Object... attrValues) {
		return createFeature(schema, "FID_" + new Random().nextInt(),
				attrValues);
	}

	/**
	 * Creates a sample SimpleFeature instance of the given
	 * {@link SimpleFeatureType}
	 * 
	 * @param schema
	 *            the schema for which to create a sample SimpleFeature instance
	 * @author SK
	 * @throws org.opengis.feature.IllegalAttributeException
	 */
	public static SimpleFeature createSampleFeature(SimpleFeatureType schema,
			Object... values) {
		SimpleFeature sampleFeature;

		//MS: Sind keine Attribut-Werte angegeben, Standard-Werte (insb.
		//    fuer die etwaige Geometrie generieren)
		if ( values.length == 0 )
		  values = getDefaultAttributeValues(schema);

		try {
			SimpleFeatureBuilder builder = new SimpleFeatureBuilder(schema);
			sampleFeature = builder.buildFeature("createSampleFeature_"
					+ new Random().nextInt(), values);

		} catch (org.opengis.feature.IllegalAttributeException e) {
			// Thrown if the SimpleFeatureType does not validate the attributes
			// Not very likely if we just create a sample feature, so we make it
			// a runtime exception.
			throw new RuntimeException(e);
		}

		return sampleFeature;
	}

	/**
	 * Erstellt eine {@link FeatureSource} basierend auf einem
	 * {@link MemoryDataStore} für eine übergebene {@link FeatureCollection}
	 * 
	 * @param features
	 */
	public static FeatureSource<SimpleFeatureType, SimpleFeature> createMemoryFeatureSource(
			FeatureCollection<SimpleFeatureType, SimpleFeature> features) {
		try {
			return new MemoryDataStore(features).getFeatureSource(features
					.getSchema().getTypeName());
		} catch (IOException e) {
			LOGGER
					.error(
							"Creating a memory datastore for the allFeaturesCollection",
							e);
			throw new RuntimeException(
					"Creating a memory datastore for the allFeaturesCollection",
					e);
		}
	}

	/**
	 * Liefert das Objekt ({@link GridCoverage2D} oder {@link FeatureCollection}
	 * oder GridCoverager oder {@link GridCoverageReader} auf dem ein Layer
	 * basiert. Ein Raster-Layer zeichnen sich dadurch aus, dass die zugrunde
	 * liegende {@link FeatureCollection} nur ein SimpleFeature enthaelt, das
	 * genau ein Attribut mit Namen "GridCoverage" und Typ {@code
	 * GridCoverage2D} oder {@link AbstractGridCoverageReader} hat. Sind diese
	 * Bedingungen erfuellt, wird das 2. Attribut zurueckgegeben, ansonsten die
	 * {@link FeatureCollection}.
	 * 
	 * @see {@link FeatureUtilities#wrapGridCoverage(GridCoverage2D)} and
	 *      {@link FeatureUtilities#wrapGridCoverageReader(AbstractGridCoverage2DReader, GeneralParameterValue[])}
	 * 
	 * @param styledFeatures
	 *            ein Layer
	 * @return {@code null}, falls das Objekt nicht ermittelt werden kann (da
	 *         ein Fehler aufgetreten ist).
	 * @param featureSource
	 * @return
	 */
	public static Object getWrappedGeoObject(
			FeatureSource<SimpleFeatureType, SimpleFeature> featureSource) {
		try {
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc = featureSource
					.getFeatures();
			return getWrappedGeoObject(fc);

		} catch (Exception err) {
			LOGGER.warn(err.getMessage(), err);
			return null;
		}
	}

	/**
	 * Liefert das Objekt ({@link GridCoverage2D} oder {@link FeatureCollection}
	 * oder GridCoverager oder {@link GridCoverageReader} auf dem ein Layer
	 * basiert. Ein Raster-Layer zeichnen sich dadurch aus, dass die zugrunde
	 * liegende {@link FeatureCollection} nur ein SimpleFeature enthaelt, das
	 * genau ein Attribut mit Namen "GridCoverage" und Typ {@code
	 * GridCoverage2D} oder {@link AbstractGridCoverageReader} hat. Sind diese
	 * Bedingungen erfuellt, wird das 2. Attribut zurueckgegeben, ansonsten die
	 * {@link FeatureCollection}.
	 * 
	 * @see {@link FeatureUtilities#wrapGridCoverage(GridCoverage2D)} and
	 *      {@link FeatureUtilities#wrapGridCoverageReader(AbstractGridCoverage2DReader, GeneralParameterValue[])}
	 * 
	 * @param styledFeatures
	 *            ein Layer
	 * @return {@code null}, falls das Objekt nicht ermittelt werden kann (da
	 *         ein Fehler aufgetreten ist).
	 * @param featureSource
	 * @return
	 */
	public static Object getWrappedGeoObject(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc) {

		try {

			// RasterLayer muss genau ein SimpleFeature beinhalten
			// Ist dies nicht der Fall wird die FeatureCollection zurueckgegeben
			if (fc == null || fc.size() != 1) {
				return fc;
			}

			// Ist dies nicht der Fall wird die FeatureCollection zurueckgegeben
			SimpleFeatureType ftype = fc.getSchema();

			if (ftype == null || !GC_NAME.equals(ftype.getName()))
				return fc;

			FeatureIterator<SimpleFeature> fIt = fc.features();
			try {
				SimpleFeature onlyFeature = fIt.next();

				// GridCoverage2D or ImagePyramidReader
				return onlyFeature.getAttribute(1);

			} finally {
				fIt.close();
			}

		} catch (Exception err) {
			LOGGER.warn(err.getMessage(), err);
			return null;
		}
	}

	/**
	 * This method determines the average distance to next centroid.
	 * 
	 * @param styledFeatures
	 */
	public static Double calcAvgNN(StyledFeaturesInterface<?> styledFeatures) {
		try {
			final Integer SAMPLESIZE = 60;

			final DynamicBin1D dists = new DynamicBin1D();

			final Integer countAll = styledFeatures
					.getFeatureCollectionFiltered().size();
			if (countAll <= 1)
				return Double.MAX_VALUE;

			final Random rand = new Random();
			final double ratio = SAMPLESIZE.doubleValue()
					/ countAll.doubleValue();
			// final double ratio = countAll > SAMPLESIZE ?
			// countAll.doubleValue() / SAMPLESIZE.doubleValue() : 1.;
			LOGGER.debug("ratio " + ratio);

			Filter filter = FeatureUtil.FILTER_FACTORY2.and(styledFeatures
					.getFilter(), new FilterAbstract(
					FeatureUtil.FILTER_FACTORY2) {

				@Override
				public boolean evaluate(Object object) {
					return rand.nextDouble() <= ratio;
				}

			});
			final Query random100query = new DefaultQuery(styledFeatures
					.getSchema().getTypeName(), filter, SAMPLESIZE, null,
					"random query for " + ratio);

			final GeometryForm geometryForm = FeatureUtil
					.getGeometryForm(styledFeatures.getSchema());

			Iterator<SimpleFeature> it;
			it = styledFeatures.getFeatureSource().getFeatures(random100query)
					.iterator();

			while (it.hasNext()) {

				SimpleFeature f1 = it.next();

				switch (geometryForm) {
				case POINT:
					final FeatureCollection<SimpleFeatureType, SimpleFeature> allFeatures = styledFeatures
							.getFeatureCollectionFiltered();

					dists.add(extractedPoint(allFeatures, f1));
					break;
				case LINE:
					dists.add(((Geometry) f1.getDefaultGeometry()).getLength());
					break;
				case POLYGON:
					dists.add(((Geometry) f1.getDefaultGeometry()).getLength());
					break;
				}
			}

			final Double avgNN = dists.mean();
			CoordinateReferenceSystem crs = styledFeatures.getCrs();
			Unit<Length> unit = (Unit<Length>) CRSUtilities.getUnit(crs
					.getCoordinateSystem());
			LOGGER.info("avgNN for " + styledFeatures.getTitle() + " is "
					+ avgNN + " in " + unit);

			// Measure<Double,Length> lengthNativeUnit = Measure.valueOf(avgNN,
			// unit);

			double inMeter = unit.getConverterTo(SI.METER).convert(avgNN);

			double inPixels = inMeter * 100 * 10 / 0.28;

			return inPixels;

		} catch (IOException e) {
			throw new RuntimeException("Can calculate AvgNN!", e);
		}
	}

	private static double extractedPoint(
			final FeatureCollection<SimpleFeatureType, SimpleFeature> allFeatures,
			SimpleFeature f1) {

		Double nearestDist = Double.MAX_VALUE;

		Point centroid1 = getCentroid(f1.getDefaultGeometry());

		// iterate over all Features again and determine the
		// distance:
		Iterator<SimpleFeature> it2 = allFeatures.iterator();
		// Iterator<SimpleFeature> it2 =
		// styledFeatures.getFeatureSource()
		// .getFeatures(random100query).iterator();
		while (it2.hasNext()) {
			SimpleFeature f2 = it2.next();

			if (f2.getID().equals(f1.getID()))
				continue;

			Point centroid2 = getCentroid(f2.getDefaultGeometry());

			final double distance = centroid1.distance(centroid2);

			nearestDist = Math.min(nearestDist, distance);
		}

		return nearestDist;
	}

	/**
	 * Returns the centroid for any JTS {@link Geometry}
	 */
	private static Point getCentroid(Object someGeometry) {
		Geometry g = (Geometry) someGeometry;
		return g.getCentroid();
	}

	public static Object getLayerSourceObject(MapLayer mapLayer) {
		return getWrappedGeoObject((FeatureSource<SimpleFeatureType, SimpleFeature>) mapLayer
				.getFeatureSource());
	}

	/**
	 * Checks whether the given attribute local name exists in the schema. If no
	 * exact match is found, an ingnoreCase match is performed. If the schema
	 * has no attributes, an exception is thrown.
	 * 
	 * @param schema
	 *            SimpleFeatureType to check against
	 * @param name
	 *            localname of an attribte to look for
	 * 
	 * @return If no match is found, the first attribute is returned.
	 */
	public static Name findBestMatchingAttributeFallBackFirst(
			SimpleFeatureType schema, String localName) {

		Name bestmatch = findBestMatchingAttribute(schema, localName);

		if (bestmatch == null)
			return schema.getAttributeDescriptors().get(0).getName();
		return bestmatch;
	}

	/**
	 * Checks whether the given attribute local name exists in the schema. If no
	 * exact match is found, an ingnoreCase match is performed. If the schema
	 * has no attributes, an exception is thrown.
	 * 
	 * @param schema
	 *            SimpleFeatureType to check against
	 * @param name
	 *            localname of an attribte to look for
	 * 
	 * @return If no match is found, <code>null</code> is returned.
	 */
	public static Name findBestMatchingAttribute(SimpleFeatureType schema,
			String localName) {
		List<AttributeDescriptor> attributeDescriptors = schema
				.getAttributeDescriptors();
		if (attributeDescriptors.size() == 0) {
			LOGGER.error("The schmema has no attributes!");
		}

		// Checking for exact match
		for (AttributeDescriptor d : attributeDescriptors) {
			if (d.getName().getLocalPart().equals(localName))
				return d.getName();
		}
		// Checking for irgnoreCase match
		for (AttributeDescriptor d : attributeDescriptors) {
			if (d.getName().getLocalPart().equalsIgnoreCase(localName)) {
				LOGGER.error("Corrected attributeName '" + localName + "' to "
						+ d.getName().getLocalPart());
				return d.getName();
			}
		}
		return null;
	}

	// /**
	// * Extrahiert alle Geometrien aus einer FeatureCollection. Fuer jedes
	// * Geometry-Attribut der FeatureCollection wird eine GeometrieCollection
	// * (<code>JTSGeometries</code>) angelegt, in der die Geometrien des
	// jeweiligen
	// * Attributs zusammengefasst werden.
	// * @param fc FeatureCollection
	// * @return Array aller Geometrien im SimpleFeature.
	// * @see org.geotools.feature.FeatureCollection
	// * @see org.geotools.renderer.geom.JTSGeometries
	// */
	// public static GeometryCollection[]
	// extractGeometriesToGeometryCollection(FeatureCollection fc) throws
	// org.opengis.referencing.operation.TransformException{
	// FeatureIterator it = fc.features(); // Zeiger durch die einzelnen
	// Features
	// Vector gcVec = new Vector(); // Speichert die Collections
	// int nextGCIdx = 0; // verweist innerhaln eines Features
	// // auf die naechste zu verwendende
	// // Collection
	//
	// // alle Features nach Geometrien durchsuchen
	// for (int i=0; it.hasNext(); i++) {
	// SimpleFeature f = it.next();
	//
	// // Das erste Geometry-Attribut des Features wird in GeometryCollection
	// // eins gespeichert, das zweite in der zweiten, usw.
	// // Bei jedem SimpleFeature wird also wieder von vorne begonnen.
	// nextGCIdx = 0;
	//
	// // Alle Attribute auf Geometrien checken
	// for (int j=0; j<f.getFeatureType().getAttributeCount(); j++) {
	// // Handelt es sich um eine Geometrie, wird sie in einer
	// // GeometryCollection gespeichert (fuer jedes Geometry-Attr.
	// // gibt es eine eigene Collection!!)
	// // if ( f.getFeatureType().getAttributeType(j).isGeometry() ){
	// if ( f.getFeatureType() instanceof GeometryAttributeType ) {
	// // gibt es noch keine Collection fuer das Attr., wird eine
	// // neue erstellt
	// if ( nextGCIdx >= gcVec.size() )
	// gcVec.add( new JTSGeometries() );
	// // Geometrie in Collection einfuegen
	// ((JTSGeometries)gcVec.elementAt(nextGCIdx++)).add((Geometry)f.getAttribute(j));
	// }
	// }
	// }
	//
	// GeometryCollection[] gcArr = new GeometryCollection[gcVec.size()];
	// for (int i=0; i<gcVec.size(); i++)
	// gcArr[i] = (GeometryCollection)gcVec.elementAt(i);
	// return gcArr;
	// }

}
