/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.grid;

import java.awt.Color;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.geom.Point2D;
import java.awt.image.ComponentSampleModel;
import java.awt.image.DataBuffer;
import java.awt.image.Raster;
import java.awt.image.RenderedImage;
import java.awt.image.WritableRaster;
import java.text.DecimalFormat;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Category;
import org.apache.log4j.Logger;
import org.geotools.coverage.CoverageFactoryFinder;
import org.geotools.coverage.GridSampleDimension;
import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.coverage.grid.GridCoverageFactory;
import org.geotools.coverage.grid.GridGeometry2D;
import org.geotools.geometry.Envelope2D;
import org.geotools.styling.ColorMap;
import org.geotools.styling.RasterSymbolizer;
import org.geotools.styling.Style;
import org.geotools.styling.StyleBuilder;
import org.opengis.coverage.PointOutsideCoverageException;
import org.opengis.coverage.SampleDimension;
import org.opengis.coverage.grid.GridEnvelope;
import org.opengis.geometry.Envelope;

import schmitzm.data.WritableGrid;
import schmitzm.data.WritableGridArray;
import schmitzm.geotools.GTUtil;
import schmitzm.geotools.styling.StylingUtil;
import schmitzm.lang.LangUtil;
import schmitzm.swing.SwingUtil;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.LineString;

/**
 * Diese Klasse stellt statische Funktionen fuer die Arbeit mit
 * {@linkplain GridCoverage2D Rasterdaten} zur Verfuegung.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.01
 */
public class GridUtil {

        /** log4j initialize */
        private static final Category cCat = Logger.getLogger(GridUtil.class.getName());

      public static final GridCoverageFactory GRID_FAC = new GridCoverageFactory();

	/**
	 * The default name for style created by createDefaultStyle();
	 */
	  public static final String DEFAULT_RASTER_STYLE_NAME = "default raster style";
	  public static final String DEFAULT_RASTER_STYLE_TITLE = "default raster style";
	  public static final String UNNAMED_RASTER_STYLE_NAME = "unnamed raster style";
	  public static final String UNTITLED_RASTER_STYLE_TITLE = "untitled raster style";

	  /**
	   * 11 Default colors used by {@link #createDiscreteStyle(GridCoverage2D, double, Color...)} if
	   * no colors are given.
	   * <ol>
	   *   <li>White</li>
	   *   <li>Red</li>
	   *   <li>Blue</li>
	   *   <li>Yellow</li>
	   *   <li>Green</li>
	   *   <li>Gray</li>
	   *   <li>Orange</li>
	   *   <li>Pink</li>
	   *   <li>Magenta</li>
       *   <li>Light gray</li>
	   *   <li>Cyan</li>
	   *   <li>Black</li>
	   * </ol>
	   */
      public static final Color[] DEFAULT_COLORS = new Color[] {
          Color.WHITE,
          Color.RED,
          Color.BLUE,
          Color.YELLOW,
          Color.GREEN,
          Color.GRAY,
          Color.ORANGE,
          Color.PINK,
          Color.MAGENTA,
          Color.LIGHT_GRAY,
          Color.CYAN,
          Color.BLACK
      };


      /**
	   * Liefert statistische Daten ueber ein Grid und erstellt ein Histogramm
	   * @param gc ein Raster
	   * @param band zu untersuchendes Band
	   * @param histogramSteps Schritte des Histogramms, -1 wenn alle Ausprägungen, 0 für ohne Histogramm
	   * @param histogrammDigits Anzahl an Nachkommastellen auf die die Histogramm-Klassen gerundet werden
	   *                         (negative Werte = Vorkommastellen; null = kein Runden). Nur relevant, wenn
	   *                         {@code hishistogramSteps} = -1
	   * @author Martin Schmitz
	   * @author Andreas Enders
	   */
	  public static GridStatistic determineStatistic(GridCoverage2D gc, int band, int histogramSteps, Integer histogrammDigits) {
	    GridStatistic gs = new GridStatistic();
	    GridEnvelope  gr = gc.getGridGeometry().getGridRange();
	    Envelope2D    ev = gc.getEnvelope2D();
	    gs.minX          = gr.getLow(0);
	    gs.minY          = gr.getLow(1);
	    gs.maxX          = gr.getHigh(0);
	    gs.maxY          = gr.getHigh(1);
	    gs.widthC        = gs.maxX - gs.minX;
	    gs.heightC       = gs.maxY - gs.minY;
	    gs.widthR        = ev.getLength(0);
	    gs.heightR       = ev.getLength(1);
	    gs.latSW         = ev.getX();
	    gs.lonSW         = ev.getY();
	    gs.cellWidth     = (gs.widthC != 0) ? gs.widthR / gs.widthC : 0;
	    gs.cellHeight    = (gs.heightC != 0) ? gs.heightR / gs.heightC : 0;

	    gs.nodataCnt     = 0;
	    gs.cellCnt       = 0;
	    gs.noDataVal     = gc.getSampleDimensions()[band].getNoDataValues();
	    if ( gs.noDataVal == null )
	      gs.noDataVal = new double[0];
	    gs.sumValue      = 0;


	    // Check auf NoData-Zellen erfolgt ueber Hash-Tabelle
	    Hashtable isNoData = new Hashtable<Double,Boolean>(gs.noDataVal.length);
	    for (int i=0; i<gs.noDataVal.length; i++)
	      isNoData.put(gs.noDataVal[i],true);

	    // Objekte fuer Koordinate
	    Point2D  p        = new Point2D.Float();
	    double[] val      = new double[ gc.getSampleDimensions().length ];

	    double lon = gs.lonSW + gs.cellHeight/2;
	    for (int y=gs.minY; y<gs.maxY; y++, lon+=gs.cellHeight) {
	      double lat = gs.latSW + gs.cellWidth/2;
	      for (int x=gs.minX; x<gs.maxX; x++, lat+=gs.cellWidth) {
	        p.setLocation(lat,lon);
	        gc.evaluate(p,val);
	        // NoData-Zellen gehen nicht in die Statistik ein
	        if ( Double.isNaN(val[band]) || isNoData.containsKey(val[band]) )
	          gs.nodataCnt++;
	        else {
	          gs.minValue = Math.min(gs.minValue, val[band]);
	          gs.maxValue = Math.max(gs.maxValue, val[band]);
	          gs.sumValue += val[band];
	          gs.cellCnt++;
	        }
	      }
	    }
	    //histogramm berechnen, wenn nicht ausgeschaltet
	    if (histogramSteps != 0)
	    {
	      lon = gs.lonSW + gs.cellHeight/2;
	      for (int y=gs.minY; y<gs.maxY; y++, lon+=gs.cellHeight)
	      {
	        double lat = gs.latSW + gs.cellWidth/2;
	        for (int x=gs.minX; x<gs.maxX; x++, lat+=gs.cellWidth)
	        {
	          p.setLocation(lat,lon);
	          gc.evaluate(p,val);
	          // NoData-Zellen gehen nicht in die Statistik ein
	          if ( !Double.isNaN(val[band]) && !isNoData.containsKey(val[band]) )
	          {
	            double tVal = val[band];
	            if (histogramSteps < 0)
	            {
	              //String tValue = ""+(new Double(tVal).longValue());
	              String value = String.valueOf( histogrammDigits != null ? LangUtil.round(tVal,histogrammDigits) : tVal );
	              Integer pixelCount = gs.histogramm.get(value);
	              if (pixelCount == null)
	                pixelCount = 1;
	              else
	                pixelCount++;
	              gs.histogramm.put(value, pixelCount);
	            }
	            else
	            {
	              double stepRange = (gs.maxValue - gs.minValue)/histogramSteps;
	              int step = new Double((tVal - gs.minValue)/stepRange).intValue();
	              String stepName = ""+gs.minValue + step * stepRange;
	              Integer number = gs.histogramm.get(stepName);
	              if (number == null) number = 1;
	              else number++;
	              gs.histogramm.put(stepName, number);
	            }
	          }
	        }
	      }
              cCat.debug("Histogram created with map: "+gs.histogramm);
	    }
	    // Durchschnittswert berechnen (nicht ueber NoData-Zellen)
	    gs.avgValue = (gs.cellCnt != 0) ? gs.sumValue/gs.cellCnt : 0;

	    return gs;
	  }


      /**
       * Liefert statistische Daten ueber ein Grid und erstellt ein Histogramm.
       * Die Histogramm-Klassen werden auf ganze Zahlen gerundet.
       * @param gc ein Raster
       * @param band zu untersuchendes Band
       * @param histogramSteps Schritte des Histogramms, -1 wenn alle Ausprägungen, 0 für ohne Histogramm
       * @author Martin Schmitz
       * @author Andreas Enders
       */
      public static GridStatistic determineStatistic(GridCoverage2D gc, int band, int histogramSteps) {
        return determineStatistic(gc, band, histogramSteps, 0);
      }

  /**
   * Liefert statistische Daten ueber ein Grid, ohne ein Histogram zu berechnen.
   * @param gc ein Raster
   * @param band zu untersuchendes Band
   */
  public static GridStatistic determineStatistic(GridCoverage2D gc, int band) {
	  return determineStatistic(gc, band, 0);
  }

  /**
   * Liefert statistische Daten ueber ein Grid, das in Zonen unterteilt ist.
   * Die Zonen werden durch einen {@link Float}-Wert identifiziert, damit
   * auch NoData-Werte in einem Raster einer Zone (naemlich der
   * Zone {@code Float.NaN}) zugewiesen werden koennen.
   * @param valueGC Raster ueber das die Statistik erzeugt wird
   * @param zoneGC Raster, das die Zonen bestimmt
   */
  public static GridZoneStatistic<Float> determineZoneStatistic(GridCoverage2D valueGC, GridCoverage2D zoneGC) {
    return determineZoneStatistic(valueGC, 0, zoneGC, 0);
  }

  /**
   * Liefert statistische Daten ueber ein Grid, das in Zonen unterteilt ist.
   * Die Zonen werden durch einen {@link Float}-Wert identifiziert, damit
   * auch NoData-Werte in einem Raster einer Zone (naemlich der
   * Zone {@code Float.NaN}) zugewiesen werden koennen.
   * @param valueGC Raster ueber das die Statistik erzeugt wird
   * @param valueBand Band ueber das die Statistik erzeugt wird
   * @param zoneGC Raster, das die Zonen bestimmt
   * @param zoneBand Band, das die Zonen bestimmt
   */
  public static GridZoneStatistic<Float> determineZoneStatistic(GridCoverage2D valueGC, int valueBand, GridCoverage2D zoneGC, int zoneBand) {
    final ReadableGridCoverage zoneGrid  = ReadableGridCoverage.create(zoneGC, 0);

    // Performanz: Raster-Ausmasse in lokalen Variablen speichern
    final float         cellHeight = (float)zoneGrid.getCellHeight();
    final float         cellWidth  = (float)zoneGrid.getCellHeight();
    final float         minX       = (float)zoneGrid.getX() + cellWidth/2;
    final float         maxX       = (float)(zoneGrid.getX() + zoneGrid.getRealWidth());
    final float         minY       = (float)zoneGrid.getY() + cellHeight/2;
    final float         maxY       = (float)(zoneGrid.getY() + zoneGrid.getRealHeight());

//    // Solution without WritableGrid
//    final Envelope2D    env        = zoneGC.getEnvelope2D();
//    final GridRange     gr         = zoneGC.getGridGeometry().getGridRange();
//    final float         height     = (float)env.getHeight();
//    final float         width      = (float)env.getWidth();
//    final int           heightC    = Math.abs( gr.getLower(1) - gr.getUpper(1) );
//    final int           widthC     = Math.abs( gr.getLower(0) - gr.getUpper(0) );
//    final float         cellHeight = heightC != 0 ? height / heightC : 0;
//    final float         cellWidth  = widthC  != 0 ? width  / widthC  : 0;
//    final float         minX       = (float)env.getX() + cellWidth/2;
//    final float         maxX       = (float)(env.getX() + width);
//    final float         minY       = (float)env.getY() + cellHeight/2;
//    final float         maxY       = (float)(env.getY() + height);

    final Point2D.Float loc        = new Point2D.Float();
    // Arrays zum Speichern der Rasterwerte pro Band
    final float[]  zoneGridVal  = new float[zoneGC.getNumSampleDimensions()];
    final float[]  valueGridVal = new float[zoneGC.getNumSampleDimensions()];
    // Variablen zum Speichern des Band-Werts
    float zoneNo  = 0;
    float cellVal = 0;

    // Zonen-Statistik berechnen
    GridZoneStatistic    zoneStat  = new GridZoneStatistic<Float>();
    for (loc.y=minY; loc.y<maxY; loc.y+=cellHeight)
      for (loc.x=minX; loc.x<=maxX; loc.x+=cellWidth) {
        zoneGC.evaluate(loc, zoneGridVal);
        zoneNo = zoneGridVal[zoneBand];
        try {
          valueGC.evaluate(loc, valueGridVal);
          cellVal = valueGridVal[valueBand];
          zoneStat.addValueToZoneStatistic(zoneNo, cellVal);
        } catch (PointOutsideCoverageException err) {
          // Zonen-Zelle liegt ausserhalb des Wert-Rasters
          // --> Zelle ignorieren
        }
      }

    return zoneStat;
  }

  /**
   * Erzeugt einen Standard-Style fuer ein {@link GridCoverage2D}
   * und setzt den default-Namen: GridUtil.DEFAULT_RASTER_STYLE_NAME
   * Und eine title : GridUtil.DEFAULT_RASTER_STYLE_TITLE
   *
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @author Stefan Alfons Krüger (DEFAULT_RASTER_STYLE_NAME, DEFAULT_RASTER_STYLE_TITLE)
   */
  public static Style createDefaultStyle() {
    Style defaultStyle = createStyle(null,1.0);
    defaultStyle.setName(DEFAULT_RASTER_STYLE_NAME);
    defaultStyle.setTitle(DEFAULT_RASTER_STYLE_TITLE);
	return( defaultStyle );
  }

  /**
   * Erzeugt einen Standard-Style fuer ein {@link GridCoverage2D}
   * @param colorMap Farb-Palette
   * @param opacity  Transparenz (1.0 = vollfarbig)
   * @return ein Standard-Style ( {@code StyleBuilder.createRasterSymbolizer()}) falls
   *         keine ColorMap angegeben wird
   */
  public static Style createStyle(ColorMap colorMap, double opacity) {
    StyleBuilder builder = StylingUtil.STYLE_BUILDER;
    RasterSymbolizer symb = colorMap != null ? builder.createRasterSymbolizer(colorMap,opacity) : builder.createRasterSymbolizer();
    return( builder.createStyle( symb ) );
  }

  /**
   * Creates an interpolative style for a {@link GridCoverage2D}. The given colors
   * are interpolated from the minimum to the maximum raster value.
   * @param grid     the raster
   * @param opacity  opacity for the style
   * @param colors   colors to interpolate with (at least 2)
   */
  public static Style createInterpolativeStyle(GridCoverage2D grid, double opacity, Color... colors) {
    if ( colors.length < 2 )
      throw new IllegalArgumentException("At least 2 colors must be specified for an interpolative style.");
    GridStatistic gridStat  = GridUtil.determineStatistic(grid, 0);
    double   intervalLength = gridStat.maxValue - gridStat.minValue;
    String[] labels         = new String[colors.length];
    double[] values         = new double[colors.length];
    for (int i=0; i<colors.length; i++) {
      labels[i] = ""; // all categories must have the same label
      values[i] = gridStat.minValue + intervalLength * i/(colors.length-1.0);
    }
    ColorMap colorMap = StylingUtil.STYLE_BUILDER.createColorMap(
        labels,
        values,
        colors,
        ColorMap.TYPE_RAMP
    );
    RasterSymbolizer rasterSymb = StylingUtil.STYLE_BUILDER.createRasterSymbolizer(colorMap, opacity);
    return StylingUtil.STYLE_BUILDER.createStyle( rasterSymb );
  }

  /**
   * Creates a discrete style for a {@link GridCoverage2D}. For every value
   * in the raster another color is choosen. If the raster contains more values
   * than colors are given, all surplus values are mapped to the "last" color.
   * @param grid          the raster
   * @param opacity       opacity for the style
   * @param digits        number of digits the grid value classes (and legend) are
   *                      rounded to (null means no round; >= 0 means digits after comma;
   *                      < 0 means digits before comma)
   * @param colorPalette  colors for the values (the smallest value is mapped to index 0, the second
   *                      smallest to index 1, ...); if no colors are specified {@link #DEFAULT_COLORS}
   *                      are used
   */
  public static Style createDiscreteStyle(GridCoverage2D grid, double opacity, Integer digits, Color... colorPalette) {
    if ( colorPalette.length == 0 )
      colorPalette = DEFAULT_COLORS;
    DecimalFormat decFormat = digits != null ? new DecimalFormat( SwingUtil.getNumberFormatPattern(digits) ) : null;
    GridStatistic gridStat  = GridUtil.determineStatistic(grid, 0, -1, digits);

    int      valueCount = gridStat.histogramm.size();
    String[] labels     = new String[valueCount];
    double[] values     = new double[valueCount];
    Color[]  colors     = new Color[valueCount];
    Color    lastColor  = colorPalette[colorPalette.length-1];

    int i = -1;
    for (String key : gridStat.histogramm.keySet()) {
      i++;
      values[i] = Double.parseDouble(key);
      colors[i] = i < colorPalette.length ? colorPalette[i] : lastColor;
      if ( decFormat != null )
        labels[i] = decFormat.format(values[i]);
      else
        labels[i] = String.valueOf(values[i]);
      //labels[i] += " ("+gridStat.histogramm.get(key)+" pixel)";
    }

    ColorMap tColorMap = StylingUtil.STYLE_BUILDER.createColorMap(
        labels,
        values,
        colors,
        ColorMap.TYPE_VALUES
    );

    RasterSymbolizer rasterSymb = StylingUtil.STYLE_BUILDER.createRasterSymbolizer(tColorMap, opacity);
    return StylingUtil.STYLE_BUILDER.createStyle( rasterSymb );
  }

  /**
   * Kopiert ein {@link GridCoverage2D} komplett.
   * @param source Quell-Raster
   */
  public static GridCoverage2D copyGridCoverage(GridCoverage2D source) {
    //return createGridCoverage(source,source.getEnvelope());
    return GRID_FAC.create(
        source.getName(),
        source.getRenderedImage().copyData(null),
        source.getEnvelope()
    );
  }

  /**
   * Kopiert ein {@link GridCoverage2D} komplett, dessen Raster-Koordinaten
   * (obere-linke Ecke) transformiert wird.
   * @param source Quell-Raster
   * @param minX   neue Raster-Koordinate fuer die obere linke Ecke
   * @param minY   neue Raster-Koordinate fuer die obere linke Ecke
   */
  public static GridCoverage2D copyGridCoverage(GridCoverage2D source, int minX, int minY) {
    return GRID_FAC.create(
                      source.getName(),
                      source.getRenderedImage().copyData(null).createWritableTranslatedChild(minX, minY),
                      source.getEnvelope()
           );
  }

  /**
   * Erzeugt ein neues {@link GridCoverage2D} mit neuer {@link SampleDimension}.
   * @param gc        Raster (mit nur einem Band)
   * @param sampleDim neue {@link SampleDimension}
   * @return eine neue {@link GridCoverage2D}-Instanz
   */
  public static GridCoverage2D resampleGridCoverage(GridCoverage2D gc, GridSampleDimension sampleDim) {
    return resampleGridCoverage(gc, new GridSampleDimension[] {sampleDim});
  }

  /**
   * Erzeugt ein neues {@link GridCoverage2D} mit neuer {@link SampleDimension}.
   * @param gc        Raster
   * @param sampleDim neue {@link SampleDimension}
   * @return eine neue {@link GridCoverage2D}-Instanz
   */
  public static GridCoverage2D resampleGridCoverage(GridCoverage2D gc, GridSampleDimension[] sampleDim) {
    final Map properties = new HashMap();
    //  properties.put("GC_NODATA", -9999.0);

    return CoverageFactoryFinder.getGridCoverageFactory(null).create(
        gc.getName(),
        gc.getRenderedImage(),
        (GridGeometry2D)gc.getGridGeometry(),
        sampleDim,
        null, //new GridCoverage[] { raster },
        properties
    );
  }

  /**
   * Erzeugt ein {@link GridCoverage2D} aus einem anderen. Das neue Raster
   * hat maximal die Ausmasse des Quell-Rasters.
   * @param source Quell-Raster
   * @param bboxEnv Bounding-Box (beschreibt die den Ausschnitt des Quell-Rasters aus dem
   *                das neue Raster erzeugt wird)
   * @return {@code null} falls sich das Quell-Raster und die Bounding-Box nicht
   *         ueberschneiden.
   */
  public static GridCoverage2D createGridCoverage(GridCoverage2D source, Envelope bboxEnv) {
    Envelope2D sourceEnv = source.getEnvelope2D();
    // Subset nur bzgl. des Bereichs in dem auch das Raster liegt
    bboxEnv = GTUtil.intersectEnvelope(sourceEnv,bboxEnv,source.getCoordinateReferenceSystem());
    if ( bboxEnv.getLength(0) == 0 && bboxEnv.getLength(1) == 0 )
      return null;

    // Daten des Source-Raster
    double realMinX     = sourceEnv.getX();
    double realMinY     = sourceEnv.getY();
    double realWidth    = sourceEnv.getWidth();
    double realHeight   = sourceEnv.getHeight();
    int    rasterMinX   = source.getRenderedImage().getMinX();
    int    rasterMinY   = source.getRenderedImage().getMinY();
    int    rasterWidth  = source.getRenderedImage().getWidth();
    int    rasterHeight = source.getRenderedImage().getHeight();
    double cellWidth    = realWidth / rasterWidth;
    double cellHeight   = realHeight / rasterHeight;

    // Envelope in Raster-Koordinaten (bzgl. unterer linker Ecke!)
    int subsetLX = convertRealToRaster(bboxEnv.getLowerCorner().getOrdinate(0),realMinX,realWidth,rasterMinX,rasterWidth);
    int subsetLY = convertRealToRaster(bboxEnv.getLowerCorner().getOrdinate(1),realMinY,realHeight,rasterMinY,rasterHeight);
    int subsetUX = convertRealToRaster(bboxEnv.getUpperCorner().getOrdinate(0),realMinX,realWidth,rasterMinX,rasterWidth);
    int subsetUY = convertRealToRaster(bboxEnv.getUpperCorner().getOrdinate(1),realMinY,realHeight,rasterMinY,rasterHeight);
    int subsetW  = Math.abs(subsetLX-subsetUX)+1;
    int subsetH  = Math.abs(subsetLY-subsetUY)+1;
    // obere linke Ecke des Subsets in Raster-Koordinaten
    int ulX = Math.min(subsetLX,subsetUX);
    int ulY = rasterMinY+rasterHeight - Math.max(subsetLY,subsetUY) -1;

    // WritableRaster fuer Subset erstellen
    Rectangle       rect = new Rectangle(ulX,ulY,subsetW,subsetH);
    Raster          r    = source.getRenderedImage().getData( rect );
    WritableRaster  wr   = createWritableRaster(r);

    // Envelope fuer Subset erzeugen (bzgl. untere liner Ecke!)
    Envelope subsetEnv = new Envelope2D(
        source.getCoordinateReferenceSystem(),
        convertRasterToReal(subsetLX,rasterMinX,realMinX,cellWidth)-cellWidth/2,
        convertRasterToReal(subsetLY,rasterMinY,realMinY,cellHeight)-cellHeight/2,
        subsetW * cellWidth,
        subsetH * cellHeight
    );

    return new GridCoverageFactory().create("",wr,subsetEnv);
  }


  /**
   * Erzeugt ein {@link WritableRaster} aus einem {@link Raster}.
   * @param r Raster
   * @param minX Start-Index in X-Richtung
   * @param minY Start-Index in Y-Richtung
   */
  public static WritableRaster createWritableRaster(Raster r, int minX, int minY) {
    return Raster.createWritableRaster(
        new ComponentSampleModel(r.getDataBuffer().getDataType(),
                                 r.getWidth(),
                                 r.getHeight(),
                                 1,
                                 r.getWidth(),
                                 new int[] {0}),
        r.getDataBuffer(),
        new Point(minX,minY)
    );
  }

  /**
   * Erzeugt ein {@link WritableRaster} aus einem {@link Raster}.
   * @param r Raster
   */
  public static WritableRaster createWritableRaster(Raster r) {
    return createWritableRaster(r,0,0);
  }
 
  /**
   * Konvertiert ein Raster-Objekt in eine {@link GridCoverage2D}.
   * Es werden folgende Objekt-Typen unterstuetzt:
   * <ul>
   *   <li>Instanzen von {@link GridCoverage2D}</li>
   *   <li>Instanzen von {@link WritableRaster} <u>und</u> {@link WritableGrid}</li>
   *   <li>Instanzen von {@link WritableGridArray}</li>
   * </ul>
   * @param obj Object
   * @exception IllegalArgumentException falls das Objekt nicht konvertiert werden kann
   */
  public static GridCoverage2D convertToGridCoverage2D(Object obj) {
    if ( obj instanceof GridCoverage2D )
      return (GridCoverage2D)obj;

    if ( obj instanceof WritableRaster  && obj instanceof WritableGrid ) {
      // GT-GridCoverage aus WritableGridRaster erzeugen
      WritableRaster wr = (WritableRaster)obj;
      WritableGrid   wg = (WritableGrid)obj;
      Envelope2D     ev = new Envelope2D(wg.getCoordinateReferenceSystem(),wg.getX(),wg.getY(),wg.getRealWidth(),wg.getRealHeight());
      return new GridCoverageFactory().create("",wr,ev);
    }

    if ( obj instanceof WritableGridArray ) {
      // GT-GridCoverage aus WritableGridRaster erzeugen
      WritableGridArray wga = (WritableGridArray)obj;
      Envelope2D        ev = GTUtil.createEnvelope2D(wga.getEnvelope(),wga.getCoordinateReferenceSystem());
      WritableRaster    wr = Raster.createWritableRaster(
          new ComponentSampleModel(wga.getSampleType(),
                                   wga.getWidth(),
                                   wga.getHeight(),
                                   1,
                                   wga.getWidth(),
                                   new int[] {0}
          ),
          wga.getDataBuffer(),
          new Point(0,0)
      );
      return new GridCoverageFactory().create("",wr,ev);
    }

    throw new IllegalArgumentException("Object can not be converted to GridCoverage2D: "+(obj==null ? "null" : obj.getClass().getSimpleName()));

  }


  /**
   * Converts a real world coordinate to the row/column number of a raster.
   * If the coordinate value is exactly located on the border between two
   * cells, the next greater raster cell is returned (except the raster border:
   * the lower (=last) cell is returned).<br>
   * This method can be applied on all dimensions (X or Y or ...), if the
   * accordingly input parameter for the dimension are given.<br>
   * <b>Note:</b>
   * Because this method is independed from the dimension, it assumes that the
   * "beginning" of the raster coordinates is located in the same corner like
   * the lat/lon-refernce!! 
   * @param coord        geographic reference (world coordinate), which is converted
   * @param realMin      minimal world coordinate of the raster
   * @param realLength   length of the raster in "world coordinates"
   * @param rasterMin    index of the first raster cell (normally 0)
   * @param rasterLength size of the raster (in cells)
   * @exception UnsupportedOperationException falls als {@code rasterLength} 0
   *            uebergeben wird
   */
  public static int convertRealToRaster(final double coord, final double realMin, final double realLength, final int rasterMin, final int rasterLength) {
    if ( rasterLength <= 0 )
      throw new UnsupportedOperationException("Length of raster must be > 0!");
    final double cellSize = realLength / rasterLength;

    // Wenn Koordinate genau den Rasterrand trifft, wird die letzte
    // Zelle zurueckgegeben...
    if ( coord == realMin + realLength )
      return rasterMin + rasterLength - 1;
    // ... sonst die naechste Zelle
    // Bemerkung: Einfaches Abschneiden (int) reicht nicht, es muss die
    //            naechstkleinere Zahl ermittelt werden, denn sonst gibt
    //            es die Zelle 0 zweimal, z.B.
    //              (coord-realMin)/cellSize =  0.8 --> Zelle 0 (korrekt)
    //              (coord-realMin)/cellSize = -0.8 --> Zelle 0 (falsch)
    //            Im zweiten Fall muss auf -1 "gerundet" werden!
    return ((int)Math.floor( (coord-realMin) / cellSize ) ) + rasterMin;
  }

  /**
   * Converts a raster cell index to "real" world coordinates. The coordinate
   * of the cell center is returned.<br> 
   * This method can be applied on all dimensions (X or Y or ...), if the
   * accordingly input parameter for the dimension are given.<br>
   * <b>Note:</b>
   * Because this method is independed from the dimension, it assumes that the
   * "beginning" of the raster coordinates is located in the same corner like
   * the lat/lon-refernce!! 
   * @param cell       raster index, which is converted
   * @param rasterMin  minimal index of the raster
   * @param realMin    minimal world coordinate of the raster
   * @param cellSize   size of one cell in "world coordinates"
   */
  public static double convertRasterToReal(final int cell, final int rasterMin, final double realMin, final double cellSize) {
    return realMin + (cell-rasterMin)*cellSize + 0.5*cellSize;
  }

  /**
   * Converts a geo reference to raster coordinates.
   * @param gc   source raster
   * @param geoX Longitude to be converted
   * @param geoY Latitude to be converted
   * @return 2-dimensional array with the X-index in element 0 and the Y-index in
   *         element 1
   */
  public static int[] convertRealToRaster(GridCoverage2D gc, double geoX, double geoY) {
    return convertRealToRaster(gc, new double[] { geoX, geoY } );
  }

  /**
   * Converts a geo reference to raster coordinates.
   * @param env      geo reference of the raster
   * @param image    image of the raster
   * @param geoX     Longitude to be converted
   * @param geoY     Latitude to be converted
   * @return 2-dimensional array with the X-index in element 0 and the Y-index in
   *         element 1
   */
  public static int[] convertRealToRaster(Envelope2D env, RenderedImage image, double geoX, double geoY) {
    return convertRealToRaster(env, image, new double[] { geoX, geoY } );
  }

  /**
   * Converts a geo reference to raster coordinates.
   * @param gc   source raster
   * @param geoCoord geo reference (Longitude,Latitude) to be converted
   * @return 2-dimensional array with the X-index in element 0 and the Y-index in
   *         element 1
   */
  public static int[] convertRealToRaster(GridCoverage2D gc, double[] geoCoord) {
    return convertRealToRaster(gc.getEnvelope2D(), gc.getRenderedImage(), geoCoord);
  }

  /**
   * Converts a geo reference to raster coordinates.
   * @param env      geo reference of the raster
   * @param image    image of the raster
   * @param geoCoord geo reference (Longitude,Latitude) to be converted
   * @return 2-dimensional array with the X-index in element 0 and the Y-index in
   *         element 1
   */
  public static int[] convertRealToRaster(Envelope2D env, RenderedImage image, double[] geoCoord) {
    final double realMinX = env.getX();
    final double realMinY = env.getY();
    final double realWidth = env.getWidth();
    final double realHeight = env.getHeight();
    final int    rasterMinX = image.getMinX();
    final int    rasterMinY = image.getMinY();
    final int    rasterWidth = image.getWidth();
    final int    rasterHeight = image.getHeight();

    int cellX = convertRealToRaster(geoCoord[0],realMinX,realWidth,rasterMinX,rasterWidth);
    int cellY = convertRealToRaster(geoCoord[1],realMinY,realHeight,rasterMinY,rasterHeight);
    // Ursprung der Raster-Koord. ist OBEN LINKS
    cellY = rasterMinY+rasterHeight - cellY - 1;

    return new int[] {cellX, cellY};
  }


  /**
   * Determines the raster cells, which are intersected by a {@link LineString}.
   * @param gc the raster
   * @param ls {@link LineString} to check
   * @param result Set the intersected cells are inserted in (can be {@code null})
   * @return an empty set if the LineString is located outside the raster
   */
  public static Set<Point> getOverlappingCells(GridCoverage2D gc, LineString ls, Set<Point> result) {
    if ( result == null )
      result = new HashSet<Point>();

    Envelope2D gridEnv      = gc.getEnvelope2D();
    double     realMinX     = gridEnv.getX();
    double     realMinY     = gridEnv.getY();
    double     realWidth    = gridEnv.getWidth();
    double     realHeight   = gridEnv.getHeight();
    double     realMaxX     = realMinX + realWidth;
    double     realMaxY     = realMinY + realHeight;
    int        rasterMinX   = gc.getRenderedImage().getMinX();
    int        rasterMinY   = gc.getRenderedImage().getMinY();
    int        rasterWidth  = gc.getRenderedImage().getWidth();
    int        rasterHeight = gc.getRenderedImage().getHeight();
    double     cellWidth    = realWidth / rasterWidth;
    double     cellHeight   = realHeight / rasterHeight;

    Coordinate[]    cellBounds = new Coordinate[] { new Coordinate(), new Coordinate(), new Coordinate(), new Coordinate() };
    GeometryFactory geomFac    = new GeometryFactory();

    // determine the cells of the LineStrings start/end point
    com.vividsolutions.jts.geom.Envelope lsBB = ls.getEnvelopeInternal();
    double bbMinX = Math.max(realMinX, lsBB.getMinX());
    double bbMinY = Math.max(realMinY, lsBB.getMinY());
    double bbMaxX = Math.min(realMaxX, lsBB.getMaxX());
    double bbMaxY = Math.min(realMaxY, lsBB.getMaxY());
    int[] lsMinCell = convertRealToRaster(gc,bbMinX,bbMaxY);
    int[] lsMaxCell = convertRealToRaster(gc,bbMaxX,bbMinY);

    bbMinX = realMinX + lsMinCell[0] * cellWidth;
    bbMaxX = realMinX + (lsMaxCell[0]+1) * cellWidth;
    bbMinY = realMinY + (rasterHeight-(lsMaxCell[1]+1)) * cellHeight;
    bbMaxY = realMinY + (rasterHeight-lsMinCell[1]) * cellHeight;

    // Check all raster cells inside the BB of the LineString, whether they
    // intersect the LineString
    for (double y = bbMinY; y<=bbMaxY; y+=cellHeight)
      for (double x = bbMinX; x<=bbMaxX; x+=cellWidth) {
        cellBounds[0].x = x;
        cellBounds[0].y = y;
        cellBounds[1].x = x+cellWidth;
        cellBounds[1].y = y;
        cellBounds[2].x = x+cellWidth;
        cellBounds[2].y = y+cellHeight;
        cellBounds[3]   = cellBounds[0];
        if ( ls.intersects(geomFac.createLinearRing(cellBounds)) ) {
          int[] cell = convertRealToRaster(gc, x, y);
          result.add( new Point(cell[0], cell[1]) );
        }

      }

    return result;
  }

  /**
   * Marks all raster cells, which are intersected by a {@link Geometry}.
   * @param gc the raster
   * @param g  Geometry, which is checked for intersection
   * @param markValue value the intersected cells are marked with
   * @param result Raster, the intersected cells are marked in (can be {@code null})
   */
  public static WritableRaster getOverlappingCells(GridCoverage2D gc, Geometry g, Number markValue, WritableRaster result) {
    if ( result == null ) {
      result = Raster.createWritableRaster(
                     gc.getRenderedImage().getSampleModel(),
                     new Point(
                         gc.getRenderedImage().getMinX(),
                         gc.getRenderedImage().getMinX()
                     )
      );
    }

    Envelope2D gridEnv      = gc.getEnvelope2D();
    int        sampleType   = gc.getRenderedImage().getSampleModel().getDataType();
    int        band         = 0;
    double     realMinX     = gridEnv.getX();
    double     realMinY     = gridEnv.getY();
    double     realWidth    = gridEnv.getWidth();
    double     realHeight   = gridEnv.getHeight();
    double     realMaxX     = realMinX + realWidth;
    double     realMaxY     = realMinY + realHeight;
    int        rasterMinX   = gc.getRenderedImage().getMinX();
    int        rasterMinY   = gc.getRenderedImage().getMinY();
    int        rasterWidth  = gc.getRenderedImage().getWidth();
    int        rasterHeight = gc.getRenderedImage().getHeight();
    double     cellWidth    = realWidth / rasterWidth;
    double     cellHeight   = realHeight / rasterHeight;


    // Determine the raster cells inside the Geometry-BB
    // -> only these are possibly intersected
    com.vividsolutions.jts.geom.Envelope lsBB = g.getEnvelopeInternal();
    double bbMinX = Math.max(realMinX, lsBB.getMinX());
    double bbMinY = Math.max(realMinY, lsBB.getMinY());
    double bbMaxX = Math.min(realMaxX, lsBB.getMaxX());
    double bbMaxY = Math.min(realMaxY, lsBB.getMaxY());
    int[] lsMinCell = convertRealToRaster(gc,bbMinX,bbMaxY);
    int[] lsMaxCell = convertRealToRaster(gc,bbMaxX,bbMinY);

    bbMinX = realMinX + lsMinCell[0] * cellWidth;
    bbMaxX = realMinX + (lsMaxCell[0]+1) * cellWidth;
    bbMinY = realMinY + (rasterHeight-(lsMaxCell[1]+1)) * cellHeight;
    bbMaxY = realMinY + (rasterHeight-lsMinCell[1]) * cellHeight;

    // Check all raster cells inside the BB of the Geometry, whether they
    // really intersect the Geometry
    Coordinate[]    cellBounds = new Coordinate[] { new Coordinate(), new Coordinate(), new Coordinate(), new Coordinate(), new Coordinate() };
    GeometryFactory geomFac    = new GeometryFactory();
    for (double y = bbMinY; y<=bbMaxY; y+=cellHeight)
      for (double x = bbMinX; x<=bbMaxX; x+=cellWidth) {
        cellBounds[0].x = x;
        cellBounds[0].y = y;
        cellBounds[1].x = x+cellWidth;
        cellBounds[1].y = y;
        cellBounds[2].x = x+cellWidth;
        cellBounds[2].y = y+cellHeight;
        cellBounds[3].x = x;
        cellBounds[3].y = y+cellHeight;
        cellBounds[4]   = cellBounds[0];
        if ( g.intersects(geomFac.createLinearRing(cellBounds)) ) {
          int[] cell = convertRealToRaster(gc, x, y);
          switch ( sampleType ) {
            case DataBuffer.TYPE_BYTE:
            case DataBuffer.TYPE_SHORT:
            case DataBuffer.TYPE_USHORT:
            case DataBuffer.TYPE_INT:
                        result.setSample(cell[0],cell[1],band,markValue.intValue());
                        break;
            case DataBuffer.TYPE_FLOAT:
                        result.setSample(cell[0],cell[1],band,markValue.floatValue());
                        break;
            case DataBuffer.TYPE_DOUBLE:
                        result.setSample(cell[0],cell[1],band,markValue.doubleValue());
                        break;
          }
        }
      }

    return result;
  }


  /**
   * Marks all raster cells, which are intersected by a {@link Geometry}.
   * @param gc the raster
   * @param g  Geometry, checked for intersection
   * @param markValue value the intersected cells are marked with
   * @return <b>a NEW instance of {@link GridCoverage2D}!!</b>
   */
  public static GridCoverage2D getOverlappingCells(GridCoverage2D gc, Geometry g, Number markValue) {
    WritableRaster result = getOverlappingCells(gc,g,markValue,null);
    return new GridCoverageFactory().create("",result,gc.getEnvelope());
  }
}
