/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.grid;

import java.util.SortedMap;
import java.util.SortedSet;
import java.util.TreeMap;
import java.util.TreeSet;

/**
 * Diese Klasse stellt Informationen ueber ein Raster dar, das in Zonen
 * unterteilt ist.
 * <ul>
 *   <li>Anzahl an Zellen pro Zone</li>
 *   <li>Summe der Zellwerte pro Zone</li>
 *   <li>Durchschnitt der Zellwerte pro Zone</li>
 *   <li>Anzahl an Zellen eines bestimmten Werts pro Zone</li>
 * </ul>
 * Der Typ, durch den die Zonen identifiziert werden, wird zum Instanzieerungszeitpunkt
 * durch {@code <Z>} festgelegt.
 * @see GridUtil#determineZoneStatistic(GridCoverage2D,GridCoverage2D)}.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class GridZoneStatistic<Z> {
  /** Speichert Nummern der Zonen. */
  protected SortedSet<Z>  zones = null;
  /** Speichert die Anzahl der Zellen pro Zone und Wert. */
  protected SortedMap<Z, SortedMap<Double,Integer>> detailCellCounts = null;
  /** Speichert die Anzahl der Zellen pro Zone. */
  protected SortedMap<Z, Integer> totalCellCounts = null;
  /** Speichert die Summe der Zellwerte pro Zone. */
  protected SortedMap<Z, Double>  cellSums = null;
  /** Speichert den Durchschnitt der Zellwerte pro Zone. */
  protected SortedMap<Z, Double>  cellAvgs = null;

  /**
   * Erzeugt eine leere {@code GridZoneStatistic}. Die Werte der Zonen muessen
   * mittles {@link setZoneStatistic(int,int,double} gesetzt werden.
   */
  public GridZoneStatistic() {
    this( new TreeMap<Z, Integer>(),  new TreeMap<Z, SortedMap<Double,Integer>>(), new TreeMap<Z, Double>(), new TreeMap<Z, Double>());
  }

  /**
   * Erzeugt eine {@code GridZoneStatistic}.
   * @param cellCounts Anzahl an Zellen pro Zone
   * @param cellSums   Summe der Zellwerte pro Zone
   */
  public GridZoneStatistic(SortedMap<Z, Integer> totalCellCounts, SortedMap<Z, SortedMap<Double,Integer>> detailCellCounts, SortedMap<Z, Double> cellSums) {
    this(totalCellCounts, detailCellCounts, cellSums, new TreeMap<Z, Double>());
    for (Z zone : getZones()) {
      int    cellCount = getTotalCellCount(zone);
      double cellSum   = getCellSum(zone);
      cellAvgs.put(zone, cellCount != 0 ? cellSum / cellCount : 0);
    }
  }

  /**
   * Erzeugt eine {@code GridZoneStatistic}.
   * @param cellCounts Anzahl an Zellen pro Zone
   * @param cellSums   Summe der Zellwerte pro Zone
   * @param cellAvgs   Durchschnittswert der Zellen pro Zone
   */
  public GridZoneStatistic(SortedMap<Z, Integer> totalCellCounts, SortedMap<Z, SortedMap<Double,Integer>> detailCellCounts, SortedMap<Z, Double> cellSums, SortedMap<Z, Double> cellAvgs) {
    this.cellSums   = cellSums;
    this.cellAvgs   = cellAvgs;
    this.detailCellCounts = detailCellCounts;
    this.totalCellCounts = totalCellCounts;
    this.zones      = new TreeSet( totalCellCounts.keySet() );
  }

  /**
   * Liefert die Anzahl an Zonen.
   */
  public int getZoneCount() {
    return zones.size();
  }

  /**
   * Liefert die Nummern der Zonen.
   */
  public SortedSet<Z> getZones() {
    return zones;
  }

  /**
   * Prueft, ob es eine bestimmte Zone gibt.
   * @param zone Nummer der Zone
   */
  public boolean containsZone(Z zone) {
    return zones.contains(zone);
  }

  /**
   * Leert die Statistik fuer eine Zone.
   * @param zone Nummer der Zone
   */
  public void clearZoneStatistic(Z zone) {
    zones.remove(zone);
    cellAvgs.remove(zone);
    cellSums.remove(zone);
    totalCellCounts.remove(zone);
    detailCellCounts.remove(zone);
  }

  /**
   * Leert die Statistik aller Zonen.
   */
  public void clearZoneStatistic() {
    while (!zones.isEmpty())
      clearZoneStatistic( zones.first() );
  }

  /**
   * Fuegt der Statistik einer Zone einen Wert hinzu.
   * @param zone Nummer der Zone
   * @param cellValue Zell-Wert
   */
  public void addValueToZoneStatistic(Z zone, double cellValue)
  {
    // Anzahl der Zellen des bestimmten Werts in Zone erhoehen
    if ( !detailCellCounts.containsKey(zone) )
      detailCellCounts.put(zone, new TreeMap<Double,Integer>());
    SortedMap<Double,Integer> detailCellCount = this.detailCellCounts.get(zone);
    detailCellCount.put(cellValue, getDetailCellCount(zone,cellValue)+1);
    // Gesamtanzahl der Zonenwerte erhoehen
    int cellCount = getTotalCellCount(zone)+1;
    this.totalCellCounts.put(zone, cellCount);
    // Gesamtsumme der Zonenwerte erhoehen
    double cellSum = getCellSum(zone)+cellValue;
    this.cellSums.put(zone, cellSum);
    // Durchschnittswert der Zone erneuern
    this.cellAvgs.put(zone, cellCount != 0 ? cellSum/cellCount : 0 );
    // ggf. neue Zone in Zonen-Liste aufnehmen
    this.zones.add(zone);
  }

  /**
   * Liefert die Anzahl an Zellen in einer Zone
   * @param zone Nummer der Zone
   * @return 0 falls es die Zone nicht gibt
   */
  public int getTotalCellCount(Z zone) {
    return (int)getValueFromZoneMap(zone, totalCellCounts);
  }

  /**
   * Liefert die Anzahl an Zellen eines bestimmten Werts in einer Zone.
   * @param zone Nummer der Zone
   * @param cellValue Zell-Wert
   * @return 0 falls es die Zone nicht gibt
   */
  public int getDetailCellCount(Z zone, double cellValue) {
    SortedMap<Double,Integer> cellCounts = detailCellCounts.get(zone);
    if ( cellCounts == null )
      return 0;
    Integer count = cellCounts.get(cellValue);
    return count == null ? 0 : count.intValue();
  }

  /**
   * Liefert die Zellenanzahl differenziert nach Zellwert fuer alle Zonen.
   */
  public SortedMap<Z, SortedMap<Double,Integer>> getDetailCellCountMap() {
    return this.detailCellCounts;
  }

  /**
   * Liefert die Zellenanzahl differenziert nach Zellwert fuer eine Zone.
   * @param zone Nummer der Zone
   */
  public SortedMap<Double,Integer> getDetailCellCountMap(Z zone) {
    return this.detailCellCounts.get(zone);
  }

  /**
   * Liefert Anzahl an verschiedenen Zellen fuer eine Zone.
   * @param zone Nummer der Zone
   */
  public int getValueCount(Z zone) {
    if ( !containsZone(zone) )
      return 0;
    return this.detailCellCounts.get(zone).size();
  }

  /**
   * Liefert die verschiedenen Werte, die in einer Zone vorkommen.
   * @param zone Nummer der Zone
   */
  public SortedSet<Double> getValues(Z zone) {
    if ( !containsZone(zone) )
      return new TreeSet<Double>();
    return new TreeSet<Double>( this.detailCellCounts.get(zone).keySet() );
  }

  /**
   * Liefert die Gesamt-Zellenanzahl pro Zone als Map.
   */
  public SortedMap<Z, Integer> getTotalCellCountMap() {
    return this.totalCellCounts;
  }

  /**
   * Liefert die Summe der Zell-Werte in einer Zone
   * @param zone Nummer der Zone
   * @return 0 falls es die Zone nicht gibt
   */
  public double getCellSum(Z zone) {
    return getValueFromZoneMap(zone, cellSums);
  }

  /**
   * Liefert die Summen der Zellwerte pro Zone als Map.
   */
  public SortedMap<Z, Double> getCellSumMap() {
    return this.cellSums;
  }

  /**
   * Liefert den Durchschnitt der Zell-Werte in einer Zone
   * @param zone Nummer der Zone
   * @return 0 falls es die Zone nicht gibt
   */
  public double getCellAvg(Z zone) {
    return getValueFromZoneMap(zone, cellAvgs);
  }

  /**
   * Liefert die Durchschnittswerte der Zellen pro Zone als Map.
   */
  public SortedMap<Z, Double> getCellAvgMap() {
    return this.cellAvgs;
  }

  /**
   * Liefert einen Wert aus einer Zonen-Map.
   * @param zone Nummer der Zone
   * @param map eine Map
   * @return 0, falls es in der Map keinen Eintrag fuer die Zone gibt
   */
  private <T extends Number> double getValueFromZoneMap(Z zone, SortedMap<Z, T> map) {
    Number value = map.get(zone);
    return value != null ? value.doubleValue() : 0;
  }
}
