/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.grid;

import java.awt.image.DataBuffer;
import java.awt.image.Raster;
import java.util.Map;

import javax.media.jai.PlanarImage;

import org.geotools.coverage.GridSampleDimension;
import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.coverage.grid.GridGeometry2D;
import org.geotools.factory.Hints;
import org.geotools.geometry.Envelope2D;
import org.opengis.coverage.grid.GridCoverage;

import schmitzm.data.AbstractReadableGrid;
import schmitzm.data.ReadableGrid;
import appl.data.LoadingException;

/**
 * Diese Klasse stellt ein {@linkplain GridCoverage2D GeoTools-GridCoverage} (2D) dar,
 * welches auf einem {@link Raster} basiert und darauf direkten
 * Lese- und Schreibzugriff liefert.<br>
 * Auch wenn {@link Raster} prinzipell mehrere Dimensionen (Baender)
 * zulaesst, sind die Zugriffsmethoden dieser Klasse auf ein Band beschraenkt.<br>
 * <br>
 * <br>Um Instanzen dieser Klasse zu erzeugen, sollten die Factory-Methoden
 * {@code create(.)} verwendet werden!</b>
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ReadableGridCoverage extends GridCoverage2D implements ReadableGrid {
  /**
   * Speichert die Datenbasis des Grids.
   */
  protected Raster raster = null;

  /**
   * Speichert das Band des {@link #raster WritableRasters}, welches durch dieses
   * Grid angesprochen wird.
   */
  protected int band   = 0;

  /**
   * Speichert die Informationen ueber die GeoReferenz des Rasters.
   */
  protected Envelope2D envelope = null;

  /**
   * Ruft den Standard-Konstruktor der von {@link GridCoverage2D} auf (ohne {@link Hint Hints}).
   * @param name          Name des Grids
   * @param image         Image-Daten
   * @param gridGeometry  Georeferenz und CRS
   * @param bands         Sample-Dimensions fuer jedes Band (kann {@code null} sein)
   * @param sources       Quell-Grids (kann {@code null} sein
   * @param properties    Properties fuer das Grid (kann {@code null} sein)
   * @param band          Band der Datenbasis, auf das die getter/setter referenziert sind
   */
  protected ReadableGridCoverage(CharSequence name, PlanarImage image, GridGeometry2D gridGeometry, GridSampleDimension[] bands, GridCoverage[] sources, Map<?,?> properties, int band) {
    this( name, image, gridGeometry, bands, sources, properties, null, band );
  }

  /**
   * Ruft den Standard-Konstruktor der von {@link GridCoverage2D} auf.
   * @param name          Name des Grids
   * @param image         Image-Daten
   * @param gridGeometry  Georeferenz und CRS
   * @param bands         Sample-Dimensions fuer jedes Band (kann {@code null} sein)
   * @param sources       Quell-Grids (kann {@code null} sein
   * @param properties    Properties fuer das Grid (kann {@code null} sein)
   * @param hints         (Rendering?) hints (kann {@code null} sein)
   * @param band          Band der Datenbasis, auf das die getter/setter referenziert sind
   */
  protected ReadableGridCoverage(CharSequence name, PlanarImage image, GridGeometry2D gridGeometry, GridSampleDimension[] bands, GridCoverage[] sources, Map<?,?> properties, Hints hints, int band) {
    super( name, image, gridGeometry, bands, sources, properties, hints );
    this.raster   = getRenderedImage().getData();
    this.band     = band;
    this.envelope = getEnvelope2D();
  }

  /**
   * Erzeugt ein neues Grid.
   * @param gc   Datenbasis fuer das Grid
   * @param band Band der Datenbasis, auf das die getter/setter referenziert sind
   */
  public static ReadableGridCoverage create(GridCoverage2D gc, int band) {
    return new ReadableGridCoverage(
      gc.getName().toString(),
      PlanarImage.wrapRenderedImage(gc.getRenderedImage()),
      (GridGeometry2D)gc.getGridGeometry(),
      gc.getSampleDimensions(),
      null,
      null,
      band
    );
  }

  /**
   * Liefert eine direkte Referenz auf die Datenbasis.
   */
  public Raster getRaster() {
    return raster;
  }

  /**
   * Liefert die Breite des Rasters (in Zellen).
   */
  public int getWidth() {
    return raster.getWidth();
  }

  /**
   * Liefert die Hoehe des Rasters (in Zellen).
   */
  public int getHeight() {
    return raster.getHeight();
  }

  /**
   * Liefert den Index der ersten Zelle (Südwest) in X-Richtung.
   */
  public int getMinX() {
    return raster.getMinX();
  }

  /**
   * Liefert den Index der ersten Zelle (Südwest) in Y-Richtung.
   */
  public int getMinY() {
    return raster.getMinY();
  }

  /**
   * Liefert die reale Breite des Rasters.
   */
  public double getRealWidth() {
    return envelope.getLength(0);
  }

  /**
   * Liefert die reale Breite des Rasters.
   */
  public double getRealHeight() {
    return envelope.getLength(1);
  }

  /**
   * Liefert die reale Breite einer Rasterzelle.
   * @return <code>getRealWidth() / getWidth()</code>
   */
  public double getCellWidth() {
    return getRealWidth() / getWidth();
  }

  /**
   * Liefert die reale Breite einer Rasterzelle.
   * @return <code>getRealHeight() / getHeight()</code>
   */
  public double getCellHeight() {
    return getRealHeight() / getHeight();
  }

  /**
   * Liefert die X-Koordinate der Georeferenz (Longitude) der linken unteren
   * Ecke des Rasters (Südwest).
   */
  public double getX() {
    return envelope.getX();
  }

  /**
   * Liefert die Y-Koordinate der Georeferenz (Latitude) der linken unteren
   * Ecke des Rasters (Südwest).
   */
  public double getY() {
    return envelope.getY();
  }

  /**
   * Liefert die Art der Daten, die im Raster gespeichert werden koennen. Diese
   * wird durch eine der TYPE-Konstanten in {@link DataBuffer}
   * repraesentiert.
   */
  public int getSampleType() {
    return raster.getDataBuffer().getDataType();
  }

  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   * @return je nach {@linkplain #getSampleType() Art} des Rasters einen
   *         <code>int</code>, <code>float</code> oder <code>double</code>
   * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
   *            Koordinaten angegeben werden
   */
  public Object getRasterSample(int... cell) {
    if ( cell.length < RASTER_DIM )
      throw new UnsupportedOperationException(RASTER_DIM+" Coordinates expected");
//    // Bei 'raster' (WritableRaster) beginnen die Raster-Koordinaten in der
//    // linken oberen Ecke. Im Grid soll die Zaehlung jedoch in der linken unteren
//    // Ecke beginnen -> Y-Koordinate umrechnen
//    int x = cell[0];
//    int y = getMinY() + getHeight()-1 - cell[1];
    // Bei 'raster' (WritableRaster) beginnen die Raster-Koordinaten in der
    // linken oberen Ecke. Im Grid soll es genauso sein!
    int x = cell[0];
    int y = cell[1];
    Object value = null;
    switch ( getSampleType() ) {
      case DataBuffer.TYPE_BYTE:
      case DataBuffer.TYPE_SHORT:
      case DataBuffer.TYPE_USHORT:
      case DataBuffer.TYPE_INT:    value = raster.getSample(x,y,band); break;
      case DataBuffer.TYPE_FLOAT:  value = raster.getSampleFloat(x,y,band); break;
      case DataBuffer.TYPE_DOUBLE: value = raster.getSampleDouble(x,y,band); break;
      default:                     value = raster.getSample(x,y,band); break;
    }
    return value;
  }

  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   */
  public short getRasterSampleAsShort(int... cell) {
    return ((Number)getRasterSample(cell)).shortValue();
  }

  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   */
  public byte getRasterSampleAsByte(int... cell) {
    return ((Number)getRasterSample(cell)).byteValue();
  }


  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   */
  public int getRasterSampleAsInt(int... cell) {
    return ((Number)getRasterSample(cell)).intValue();
  }

  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   */
  public long getRasterSampleAsLong(int... cell) {
    return ((Number)getRasterSample(cell)).longValue();
  }

  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   */
  public float getRasterSampleAsFloat(int... cell) {
    return ((Number)getRasterSample(cell)).floatValue();
  }

  /**
   * Liefert einen Wert des Rasters ueber Raster-Koordinaten.
   * @param cell 2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *             {@link #getMinX()} und {@link #getMinY()})
   */
  public double getRasterSampleAsDouble(int... cell) {
    return ((Number)getRasterSample(cell)).doubleValue();
  }

  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * Liegt der Koordinatenwert genau auf der Grenze zwischen zwei Rasterzellen,
   * wird die naechst groessere Zelle gewaehlt (ausser die Grenze entspricht
   * dem Raster-Rand!).
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
   *            Koordinaten angegeben werden
   */
  public Object getGridSample(double... coord) {
    if ( coord.length < RASTER_DIM )
      throw new UnsupportedOperationException(RASTER_DIM+" Coordinates expected");
    // Koordinaten umrechnen in Zellen-Index
    int cellX = convertRealToRaster(coord[0],0);
    int cellY = convertRealToRaster(coord[1],1);
    return getRasterSample(cellX,cellY);
  }

  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   */
  public short getGridSampleAsShort(double... coord) {
    return ((Number)getGridSample(coord)).shortValue();
  }

  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   */
  public byte getGridSampleAsByte(double... coord){
    return ((Number)getGridSample(coord)).byteValue();
  }

  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   */
  public int getGridSampleAsInt(double... coord){
    return ((Number)getGridSample(coord)).intValue();
  }


  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   */
  public long getGridSampleAsLong(double... coord){
    return ((Number)getGridSample(coord)).longValue();
  }

  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   */
  public float getGridSampleAsFloat(double... coord){
    return ((Number)getGridSample(coord)).floatValue();
  }


  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten.
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   */
  public double getGridSampleAsDouble(double... coord){
    return ((Number)getGridSample(coord)).doubleValue();
  }

  /**
   * Konvertiert eine reale Koordinate in eine Zellennummer. Liegt der
   * Koordinatenwert genau auf der Grenze zwischen zwei Rasterzellen, wird
   * die naechst "groessere" Zellegewaehlt.
   * Ausnahme bildet der Rand des Rasters. Hier wird die kleinere Zelle
   * (also die letzte) herangezogen
   * @param coord Georeferenz-Koordinate
   * @param dim   Dimension, in der die Umrechnung erfolgen soll
   * @exception UnsupportedOperationException falls eine ungueltige Dimension
   *            angegeben wird (nur Werte 0 <= coord < {@link #RASTER_DIM} sind
   *            zulaessig
   */
  public int convertRealToRaster(double coord, int dim) {
    return AbstractReadableGrid.convertRealToRaster(this,coord,dim);
  }

  /**
   * Konvertiert eine Zellennummer in reale Koordinate. Dabei wird die
   * Koordinate der Zellenmitte zurueckgegeben.
   * @param cell  Rasterzellen-Koordinate
   * @param dim   Dimension, in der die Umrechnung erfolgen soll
   * @exception UnsupportedOperationException falls eine ungueltige Dimension
   *            angegeben wird (nur Werte 0 <= coord < {@link #RASTER_DIM} sind
   *            zulaessig
   */
  public double convertRasterToReal(int cell, int dim) {
    return AbstractReadableGrid.convertRasterToReal(this,cell,dim);
  }

  /**
   * This class does not support late loading!
   * @see appl.data.LateLoadable#unloadData()
   * @return false;
   * @see appl.data.LateLoadable#isLateLoadable()
   */
  public boolean isLateLoadable() {
    return false;
  }

  /**
   * Does nothing!
   * This class does not support late loading!
   * @see appl.data.LateLoadable#unloadData()
   * @see appl.data.LateLoadable#loadData()
   */
  public void loadData() throws LoadingException {
  }

  /**
   * Does nothing!
   * This class does not support late loading!
   * @see appl.data.LateLoadable#unloadData()
   */
  public void unloadData() {
  }

  /**
   * Calls {@link #dispose(boolean) dispose(false}.
   */
  @Override
  public void dispose() {
    dispose(false);
  }
}
