/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.grid;

import java.awt.Point;
import java.awt.image.DataBuffer;
import java.awt.image.WritableRaster;
import java.util.Map;

import javax.media.jai.PlanarImage;
import javax.media.jai.RasterFactory;

import org.geotools.coverage.GridSampleDimension;
import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.coverage.grid.GridCoverageFactory;
import org.geotools.coverage.grid.GridGeometry2D;
import org.geotools.geometry.Envelope2D;
import org.opengis.coverage.grid.GridCoverage;

import schmitzm.data.WritableGrid;

/**
 * Diese Klasse stellt ein {@linkplain GridCoverage2D GeoTools-GridCoverage} (2D) dar,
 * welches auf einem {@link WritableRaster} basiert und darauf direkten
 * Lese- und Schreibzugriff liefert.<br>
 * Auch wenn {@link WritableRaster} prinzipell mehrere Dimensionen (Baender)
 * zulaesst, sind die Zugriffsmethoden dieser Klasse auf ein Band beschraenkt.<br>
 * <br>
 * <br>Um Instanzen dieser Klasse zu erzeugen, sollten die Factory-Methoden
 * {@code create(.)} verwendet werden!</b>
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class WritableGridCoverage extends ReadableGridCoverage implements WritableGrid {
  /**
   * Speichert die Datenbasis des Grids als {@link WritableRaster}.
   * Entspricht {@link ReadableGridCoverage#raster}.
   */
  protected WritableRaster writableRaster = null;

  /**
   * Ruft den Standard-Konstruktor der von {@link GridCoverage2D} auf.
   * @param name          Name des Grids
   * @param image         Image-Daten (muss ein {@link WritableRaster} beinhalten)
   * @param gridGeometry  Georeferenz und CRS
   * @param bands         Sample-Dimensions fuer jedes Band (kann {@code null} sein)
   * @param sources       Quell-Grids (kann {@code null} sein
   * @param properties    Properties fuer das Grid (kann {@code null} sein)
   * @param band          Band der Datenbasis, auf das die getter/setter referenziert sind
   */
  protected WritableGridCoverage(CharSequence name, PlanarImage image, GridGeometry2D gridGeometry, GridSampleDimension[] bands, GridCoverage[] sources, Map properties, int band) {
    super( name, image, gridGeometry, bands, sources, properties, band );
    if ( !(raster instanceof WritableRaster) )
      throw new UnsupportedOperationException("Image for WritableGridCoverage mus provide a WritableRaster!");
    this.writableRaster = (WritableRaster)raster;
  }

  /**
   * Erzeugt ein neues Grid.
   * @param gc   Datenbasis fuer das Grid
   * @param band Band der Datenbasis, auf das die getter/setter referenziert sind
   */
  public static WritableGridCoverage create(GridCoverage2D gc, int band) {
    return new WritableGridCoverage(
      gc.getName().toString(),
      PlanarImage.wrapRenderedImage(gc.getRenderedImage()),
      (GridGeometry2D)gc.getGridGeometry(),
      gc.getSampleDimensions(),
      null,
      null,
      band
    );
  }

  /**
   * Erzeugt ein neues Grid.
   * @param name     Name fuer das Grid
   * @param raster   Datenbasis fuer das Grid
   * @param band     Band der Datenbasis, auf das referenziert wird
   * @param envelope GeoReferenz fuer das Raster
   */
  public static WritableGridCoverage create(String name, WritableRaster raster, int band, Envelope2D envelope) {
    return create( new GridCoverageFactory().create(name,raster,envelope), band  );
  }

  /**
   * Erzeugt ein neues Grid. Die Spalten sind von 0 bis <code>w-1</code>, die
   * Zeilen von 0 bis <code>h-1</code> durchnummeriert.
   * @param name     Name fuer das Grid
   * @param type     Datentyp, der im Raster gespeichert ist (z.B. {@link DataBuffer#TYPE_INT})
   * @param w        Breite des Rasters in Zellen
   * @param h        Hoehe des Rasters in Zellen
   * @param envelope GeoReferenz fuer das Raster
   */
  public static WritableGridCoverage create(String name, int type, int w, int h, Envelope2D envelope) {
    return create(name,type,w,h,0,0,envelope);
  }

  /**
   * Erzeugt ein neues Grid.
   * @param name     Name fuer das Grid
   * @param type     Datentyp, der im Raster gespeichert ist (z.B. {@link DataBuffer#TYPE_INT})
   * @param w        Breite des Rasters in Zellen
   * @param h        Hoehe des Rasters in Zellen
   * @param x0       Index, mit dem die linke Spalte des Rasters angesprochen wird
   * @param y0       Index, mit dem die oberste Zeile des Rasters angesprochen wird
   * @param envelope GeoReferenz fuer das Raster
   */
  public static WritableGridCoverage create(String name, int type, int w, int h, int x0, int y0, Envelope2D envelope) {
    return create(name,RasterFactory.createBandedRaster(type,w,h,1,new Point(x0,y0)), envelope);
  }

  /**
   * Erzeugt ein neues Grid. Es wird automatisch auf Band 0 des WritableRasters
   * Bezug genommen.
   * @param name     Name fuer das Grid
   * @param raster   Datenbasis fuer das Grid
   * @param envelope GeoReferenz fuer das Raster
   */
  public static WritableGridCoverage create(String name, WritableRaster raster, Envelope2D envelope) {
    return create(name,raster,0,envelope);
  }

//  /**
//   * Erzeugt eine neue Instanz, basierend auf dem aktuellen Raster.
//   * @deprecated Methode dient nur als Test/Notloesung und wird (hoffentlich)
//   *             bald wieder entfernt werden koennen!
//   */
//  public WritableGridCoverage createNew() {
////    return new WritableGridCoverage(this.getName().toString(),raster,envelope);
//    return create(this.getName().toString(),raster,envelope);
//  }

  /**
   * Liefert eine direkte Referenz auf die Datenbasis.
   */
  public WritableRaster getRaster() {
    return writableRaster;
  }

  /**
   * Setzt einen Wert im Raster ueber Raster-Koordinaten.
   * @param value neuer Wert
   * @param cell  2D-Raster-Koordinate (Zellenindizes, beginnend bei
   *              {@link #getMinX()} und {@link #getMinY()})
   * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
   *            Koordinaten angegeben werden
   */
  public void setRasterSample(Object value, int... cell) {
    if ( cell.length < RASTER_DIM )
      throw new UnsupportedOperationException(RASTER_DIM+" Coordinates expected");
//    // Bei 'raster' (WritableRaster) beginnen die Raster-Koordinaten in der
//    // linken oberen Ecke. Im Grid soll die Zaehlung jedoch in der linken unteren
//    // Ecke beginnen -> Y-Koordinate umrechnen
//    int x = cell[0];
//    int y = getMinY() + getHeight()-1 - cell[1];
    // Bei 'raster' (WritableRaster) beginnen die Raster-Koordinaten in der
    // linken oberen Ecke. Im Grid soll es genauso sein!
    int x = cell[0];
    int y = cell[1];

    switch ( getSampleType() ) {
      case DataBuffer.TYPE_BYTE:
      case DataBuffer.TYPE_SHORT:
      case DataBuffer.TYPE_USHORT:
      case DataBuffer.TYPE_INT: writableRaster.setSample(x,y,band,((Number)value).intValue());
                                break;
      case DataBuffer.TYPE_FLOAT: writableRaster.setSample(x,y,band,((Number)value).floatValue());
                                  break;
      case DataBuffer.TYPE_DOUBLE: writableRaster.setSample(x,y,band,((Number)value).doubleValue());
                                   break;
    }
  }

  /**
   * Liefert einen Wert des Rasters ueber Geo-Koordinaten. Liegt der
   * Koordinatenwert genau auf der Grenze zwischen zwei Rasterzellen, wird die
   * naechst groessere Zelle gewaehlt (ausser die Grenze entspricht dem Raster-Rand!).
   * @param coord  2D-Raster-Koordinate (Lat/Lon-GeoReferenz)
   * @exception UnsupportedOperationException falls zu {@linkplain #RASTER_DIM wenig}
   *            Koordinaten angegeben werden
   */
  public void setGridSample(Object value, double... coord) {
    if ( coord.length < RASTER_DIM )
      throw new UnsupportedOperationException(RASTER_DIM+" Coordinates expected");
    // Koordinaten umrechnen in Zellen-Index
    int cellX = convertRealToRaster(coord[0],0);
    int cellY = convertRealToRaster(coord[1],1);
    setRasterSample(value,cellX,cellY);
  }
}
