/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Frame;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Map;
import java.util.Set;
import java.util.SortedMap;
import java.util.TreeMap;

import javax.swing.AbstractAction;
import javax.swing.AbstractButton;
import javax.swing.Action;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;

import org.apache.log4j.Logger;
import org.geotools.referencing.CRS;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import schmitzm.geotools.GTUtil;
import schmitzm.lang.LangUtil;
import schmitzm.lang.ResourceProvider;
import schmitzm.swing.ButtonGroup;
import schmitzm.swing.ExceptionDialog;
import schmitzm.swing.InputOption;
import schmitzm.swing.JPanel;
import schmitzm.swing.SelectionInputOption;
import schmitzm.swing.StatusDialog;
import schmitzm.swing.SwingUtil;
import schmitzm.swing.SwingWorker;
import schmitzm.swing.event.InputOptionAdapter;

/**
 * Dieser Dialog stellt verschiedene {@link CoordinateReferenceSystem} (CRS)
 * zur Auswahl. Neben einer Auswahl von vordefinierten CRS, wird WGS-84, sowie
 * (optional) ein beliebiges Standard-CRS dargestellt. Zudem kann in einem
 * Text-Feld die WKT-Spezifikation manuell eingegeben oder editiert werden.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class CRSSelectionDialog extends JDialog {
  private   Logger           LOGGER = LangUtil.createLogger(this);
  private   ResourceProvider RES = GeotoolsGUIUtil.RESOURCE;

  /** Beinhaltet die vordefinierten CRS. Die Map beinhaltet pro Authority eine
   *  Map mit den CRS. */
  protected static SortedMap<String,SortedMap<String,CoordinateReferenceSystem>> predefinedCRS = null;

  /** Label in dem die Meldung angezeigt wird. */
  protected JTextArea  messageLabel = null;
  /** Radio-Button fuer das WGS84-CRS. */
  protected JRadioButton wgs84Button = null;
  /** Radio-Button fuer das Standard-CRS. */
  protected JRadioButton defaultButton = null;
  /** Radio-Button fuer ein vordefiniertes CRS. */
  protected JRadioButton predefinedButton = null;
  /** Auswahl-Feld fuer die Authority. */
  protected SelectionInputOption.Combo<String> authoritySelect = null;
  /** Auswahl-Feld fuer das Authority-CRS. */
  protected SelectionInputOption.Combo<CoordinateReferenceSystem> crsSelect = null;
  /** Radio-Button fuer ein benutzerdefiniertes CRS. */
  protected JRadioButton userButton = null;
  /** Gruppe der Auswahl-Buttons. */
  protected ButtonGroup buttonGroup = null;
  /** Bereich, in dem das CRS als WKT angezeigt wird. */
  protected JTextArea wktTextArea = null;
  /** ScrollPane, in dem sich die TextArea fuer die WKT-Definition befinden. */
  protected JScrollPane wktScrollPane = null;

  /** Die Checkbox "Immer diese Auswahl treffen" des Dialogs. */
  protected JCheckBox rememberCheckbox = null;
  /** Der OK-Button des Dialogs. */
  protected JButton okButton = null;
  /** Der Abbrechen-Button des Dialogs. */
  protected JButton cancelButton = null;

  /** Das als "Standard" angebotene CRS.  */
  protected CoordinateReferenceSystem defaultCRS = null;
  /** Das ausgewaehlte CRS.  */
  protected CoordinateReferenceSystem selectedCRS = null;

  /** Panel (mit {@link GridBagLayout}), in dem die Meldung und Radio-Buttons
   *  angezeigt werden. */
  protected JPanel contentPane = null;
  /** Dem Dialog uebergeordnete Komponente. */
  protected Component parent = null;

  /**
   * Erzeugt einen neuen Fehler-Dialog. Der Dialog wird relativ zum Parent-Fenster
   * zentriert.
   * @param parent         uebergeordnetes Fenster (kann <code>null</code> sein!)
   * @param message        einleitende Meldung (kann <code>null</code> sein!)
   * @param cancelAllowed  wenn {@code false}, kann der Dialog nicht abgebrochen werden
   * @param rememberOption wenn {@code true} wird eine Checkbox "Immer diese Auswahl treffen"
   *                       angezeigt
   * @param defaultCRS     das als Standard angebotene CRS (kann {@code null} sein)
   */
  public CRSSelectionDialog(Component parent, String message, boolean cancelAllowed, boolean rememberOption, CoordinateReferenceSystem defaultCRS) {
    super((Frame)null,true);
    initPredefinedCRS();

    this.defaultCRS     = defaultCRS;
    this.parent         = parent;

    // Vorlagen-Dialog erzeugen
    if ( message != null ) {
      this.messageLabel = new JTextArea();
      this.messageLabel.setWrapStyleWord(true);
      this.messageLabel.setFont( new JLabel().getFont() );
      this.messageLabel.setBackground( new JLabel("").getBackground() );
      this.messageLabel.setText(message);
      this.messageLabel.setEditable(false);
    }
    this.wktTextArea      = new JTextArea(10,60);
    this.wktTextArea.setWrapStyleWord(true);
    this.wktScrollPane    = new JScrollPane(wktTextArea);
    this.buttonGroup      = new ButtonGroup();
    this.wgs84Button      = createRadioButton( RES.getString("schmitzm.geotools.gui.CRSSelectionDialog.button.wgs84") );
    this.userButton       = createRadioButton( RES.getString("schmitzm.geotools.gui.CRSSelectionDialog.button.userDefined") );
    this.predefinedButton = createRadioButton( RES.getString("schmitzm.geotools.gui.CRSSelectionDialog.button.predefined") );
    if ( defaultCRS != null ) {
      this.defaultButton = createRadioButton( RES.getString("schmitzm.geotools.gui.CRSSelectionDialog.button.default", defaultCRS.getName().toString()) );
      this.defaultButton.setSelected(true);
    } else
      this.wgs84Button.setSelected(true);
    this.crsSelect        = new SelectionInputOption.Combo<CoordinateReferenceSystem>(null,true);
    this.authoritySelect  = new SelectionInputOption.Combo<String>(null,true);
    this.authoritySelect.setSelectionObjects(predefinedCRS.keySet().toArray(), null);
    this.authoritySelect.setSelectedItem(0);
    SwingUtil.setPreferredWidth( this.authoritySelect, 100 );
    SwingUtil.setMinimumWidth( this.authoritySelect, 100 );
    // Wenn Authority geaendet wird, die CRS-Auswahl aktualisieren
    this.authoritySelect.addInputOptionListener( new InputOptionAdapter() {
      public void optionChanged(InputOption option, Object oldValue, Object newValue) {
        Map<String, CoordinateReferenceSystem> crsMap = predefinedCRS.get((String)newValue);
        crsSelect.setSelectionObjects(
            crsMap.values().toArray(new CoordinateReferenceSystem[0]),
            crsMap.keySet().toArray(new String[0])
        );
        // Das erste CRS auswaehlen
        if ( crsSelect.getSelectedItemCount() > 0 )
          crsSelect.setSelectedIndex(0);
      }
    });
    // Wenn CRS geaendert wird, das WKT-Feld aktualisieren
    this.crsSelect.addInputOptionListener( new InputOptionAdapter() {
      public void optionChanged(InputOption option, Object oldValue, Object newValue) {
        resetWKT( (CoordinateReferenceSystem)newValue );
      }
    });
    // Wenn Authority vorhanden, die erste vorbelegen
    if ( authoritySelect.getSelectedItemCount() > 0 )
      authoritySelect.setSelectedIndex(0);

    // Checkbox erstellen
    if ( rememberOption )
      this.rememberCheckbox = new JCheckBox( SwingUtil.RESOURCE.getString("RememberChoice") );

    // Dialog-Button erstellen
    this.okButton     = new JButton( SwingUtil.RESOURCE.getString("Ok") );
    this.cancelButton = new JButton( SwingUtil.RESOURCE.getString("Cancel") );
    this.cancelButton.addActionListener( new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        performButton(cancelButton);
      }
    });
    this.okButton.addActionListener( new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        performButton(okButton);
      }
    } );

    // GUI aufbauen
    this.contentPane = new JPanel();
    this.contentPane.setLayout( new GridBagLayout() );
    if ( messageLabel != null )
      this.contentPane.add(messageLabel, new GridBagConstraints(0,0,3,1,0.0,0.0,GridBagConstraints.WEST,GridBagConstraints.BOTH,new Insets(0,0,0,0),0,0));
    this.contentPane.add(wgs84Button, new GridBagConstraints(0,1,3,1,0.0,0.0,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(0,0,0,0),0,0));
    if ( defaultButton != null )
      this.contentPane.add(defaultButton, new GridBagConstraints(0,2,3,1,0.0,0.0,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(0,0,0,0),0,0));
    this.contentPane.add(predefinedButton, new GridBagConstraints(0,3,1,1,0.0,0.0,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(0,0,0,0),0,0));
    this.contentPane.add(authoritySelect, new GridBagConstraints(1,3,1,1,0.0,0.0,GridBagConstraints.WEST,GridBagConstraints.BOTH,new Insets(0,0,0,0),0,0));
    this.contentPane.add(crsSelect, new GridBagConstraints(2,3,1,1,1.0,0.0,GridBagConstraints.WEST,GridBagConstraints.BOTH,new Insets(0,0,0,0),0,0));
    this.contentPane.add(userButton, new GridBagConstraints(0,4,3,1,0.0,0.0,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(0,0,0,0),0,0));
    this.contentPane.add(wktScrollPane, new GridBagConstraints(0,5,3,1,1.0,1.0,GridBagConstraints.WEST,GridBagConstraints.BOTH,new Insets(0,0,0,0),0,0));
    if ( rememberCheckbox != null )
      this.contentPane.add(rememberCheckbox, new GridBagConstraints(0,6,3,1,0.0,0.0,GridBagConstraints.WEST,GridBagConstraints.HORIZONTAL,new Insets(0,0,0,0),0,0));

    // Dialog erzeugen
    JOptionPane  pane = new JOptionPane(
      contentPane,
      JOptionPane.QUESTION_MESSAGE,
      JOptionPane.DEFAULT_OPTION,
      null,
      cancelAllowed ? new Object[] {okButton, cancelButton} : new Object[] { okButton }
    );
    JDialog dialog = pane.createDialog(parent,RES.getString("schmitzm.geotools.gui.CRSSelectionDialog.title"));

    // Dialog nach Vorlage initialisieren
    this.setDefaultCloseOperation( DO_NOTHING_ON_CLOSE );
    this.setTitle( dialog.getTitle() );
    this.getContentPane().setLayout(new BorderLayout());
    this.getContentPane().add(dialog.getContentPane(), BorderLayout.CENTER);

    // Sofern vorhanden, den Default-Button als Standard auswaehlen
    performButton( defaultButton != null ? defaultButton : wgs84Button );

    // Dialog in der Mitte des Parent-Fenster anordnen
    pack();
    SwingUtil.setRelativeFramePosition(this,SwingUtil.getParentWindow(parent),0.5,0.5);
  }

  /**
   * Initialisiert die zur Verfuegung stehenden CRS.
   */
  protected void initPredefinedCRS() {
    if ( predefinedCRS != null )
      return;

    // Initalisierung der CRS-Auswahl in einem SwingWorker
    SwingWorker.Work work = new SwingWorker.Work() {
      public TreeMap<String, SortedMap<String,CoordinateReferenceSystem>> execute() {
        TreeMap<String, SortedMap<String,CoordinateReferenceSystem>> authorityCRS = new TreeMap<String,SortedMap<String,CoordinateReferenceSystem>>();
        for ( String authority : (Set<String>)CRS.getSupportedAuthorities(true) ) {
          SortedMap<String, CoordinateReferenceSystem> crs = GTUtil.getAvailableCRSByName(authority, false, true);
          if ( !crs.isEmpty() )
            authorityCRS.put(authority, crs);
        }
        return authorityCRS;
      }
      public void performError(Throwable err) {
        if ( err instanceof ThreadDeath )
          return;
        super.performError(err);
      }
    };
    SwingWorker worker = new SwingWorker(
        work,
        new StatusDialog(
            parent,
            RES.getString("schmitzm.geotools.gui.CRSSelectionDialog.init.crs.title"),
            RES.getString("schmitzm.geotools.gui.CRSSelectionDialog.init.crs.mess")
        )
    );
    try {
      worker.start();
    } catch (Exception err) {
      // ignore this
    }
    predefinedCRS = (TreeMap<String, SortedMap<String,CoordinateReferenceSystem>>)worker.getWorkResult();
    if ( predefinedCRS == null )
      predefinedCRS = new TreeMap<String,SortedMap<String,CoordinateReferenceSystem>>();
  }

  /**
   * Fuehrt die Aktion eines Buttons aus (Radio- oder Dialog-Button).
   * @param button der aktivierte Button
   */
  protected void performButton(AbstractButton button) {
    if ( button instanceof JRadioButton ) {
      // Auswahl der Authority und des CRS nur aktiv, wenn Radio-Button
      // fuer Authority ausgewaehlt ist
      this.authoritySelect.setEnabled( button == predefinedButton );
      this.crsSelect.setEnabled( button == predefinedButton );
      // Benutzerdefinerte Eingabe nur aktiv, wenn der entsprechende
      // Radio-Button ausgewaehlt ist
      this.wktTextArea.setEnabled( button == userButton );
    }

    // Wenn Auswahl WGS84, Default oder Authority, dann das jeweilige WKT
    // einblenden
    if ( button == wgs84Button )
      resetWKT( GTUtil.WGS84 );
    if ( button == defaultButton )
      resetWKT( defaultCRS );
    if ( button == predefinedButton )
      resetWKT( (CoordinateReferenceSystem)crsSelect.getValue() );
    // Wenn Auswahl Benutzerdefiniert, dann direkt auf die WKT springen
    if ( button == userButton )
      wktTextArea.grabFocus();

    // Abbrechen --> Dialog ohne CRS-Auswahl beenden
    if ( button == cancelButton ) {
      selectedCRS = null;
      setVisible(false);
    }
    // OK --> Rueckgabe-CRS belegen
    if ( button == okButton ) {
      try {
        selectedCRS = null;
        if ( wgs84Button.isSelected() )
          selectedCRS = GTUtil.WGS84;
        if ( defaultButton != null && defaultButton.isSelected() )
          selectedCRS = CRSSelectionDialog.this.defaultCRS;
        if ( predefinedButton.isSelected() )
          selectedCRS = (CoordinateReferenceSystem) crsSelect.getValue();
        if ( userButton.isSelected() && !"".equals(wktTextArea.getText()) )
          selectedCRS = CRS.parseWKT(wktTextArea.getText());
        if ( selectedCRS == null )
          JOptionPane.showMessageDialog(
              this,
              RES.getString("schmitzm.geotools.gui.CRSSelectionDialog.mandatory"),
              RES.getString("schmitzm.geotools.gui.CRSSelectionDialog.title"),
              JOptionPane.ERROR_MESSAGE
          );
        else
          setVisible(false);
      } catch (Exception err) {
        selectedCRS = null;
        ExceptionDialog.show(this, err);
        // Nach Fehler zu der entsprechenden Komponente springen
        if ( userButton.isSelected() )
          wktTextArea.grabFocus();
        else if ( predefinedButton.isSelected() )
          predefinedButton.grabFocus();
        else if ( buttonGroup.getSelectedButton() != null )
          buttonGroup.getSelectedButton().grabFocus();
      }
    }
  }

  /**
   * Erzeugt einen neuen Radio-Button und fuegt diesen in die {@link #buttonGroup} ein.
   * @param label Label fuer den Radio-Button
   */
  protected JRadioButton createRadioButton(String label) {
    Action action = new AbstractAction() {
      public void actionPerformed(ActionEvent e) {
        performButton( (AbstractButton)e.getSource() );
      }
    };
    action.putValue(Action.SHORT_DESCRIPTION, label);
    action.putValue(Action.LONG_DESCRIPTION, label);
    action.putValue(Action.NAME, label);
    JRadioButton button = new JRadioButton(action);
    this.buttonGroup.add(button);
    return button;
  }

  /**
   * Belegt das WKT-Feld neu. Jedoch nur, wenn sich der Text aendert,
   * damit nach Moeglichkeit die Caret-Position erhalten bleibt.
   * @param crs darzustellendes CRS
   */
  protected void resetWKT(CoordinateReferenceSystem crs) {
    String wkt = "";
    try {
      wkt = ( crs != null && crs.toWKT() != null ) ? crs.toWKT().toString() : "";
    } catch (Exception err) {
      LOGGER.warn(err.getMessage());
    }

    // Nur neu setzen, wenn auch Aenderung stattfindet (damit ggf.
    // die Caret-Position erhalten bleibt)
    if ( wktTextArea.getText() == null || !wktTextArea.getText().equals(wkt) ) {
      wktTextArea.setText( wkt );
      wktTextArea.setCaretPosition(0);
    }
  }

  /**
   * Liefert die angezeigte Meldung.
   */
  public String getMessage() {
    return  messageLabel.getText();
  }

  /**
   * Liefert den ausgewaehlen Radio-Button.
   */
  public AbstractButton getSelectedButton() {
    return buttonGroup.getSelectedButton();
  }

  /**
   * Prueft, ob der Dialog abgebrochen wurde.
   */
  public boolean isCanceled() {
    return !isVisible() && getSelectedButton() == cancelButton;
  }

  /**
   * Prueft, ob der Dialog via OK beendet wurde.
   */
  public boolean isConfirmed() {
    return !isVisible() && getSelectedButton() == okButton;
  }

  /**
   * Prueft, ob die "Immer diese Auswahl treffen" Option im Dialog
   * angewaehlt ist.
   * @return {@code false} solange der OK-Button noch nicht betaetigt wurde
   */
  public boolean isRememberOptionSet() {
    return isConfirmed() && rememberCheckbox != null && rememberCheckbox.isSelected();
  }

  /**
   * Liefert das ausgewaehlte CRS.
   * @return {@code null} solange der OK-Button noch nicht betaetigt wurde
   */
  public CoordinateReferenceSystem getCRS() {
    return this.selectedCRS;
  }

  /**
   * Zeigt einen CRS-Auswahl-Dialog an.
   * @param parent         uebergeordnetes Fenster (kann <code>null</code> sein!)
   * @param message        einleitende Meldung (kann <code>null</code> sein!)
   * @param cancelAllowed  wenn {@code false}, kann der Dialog nicht abgebrochen werden
   * @param rememberOption wenn {@code true} wird eine Checkbox "Immer diese Auswahl treffen"
   *                       angezeigt
   * @param defaultCRS     das als Standard angebotene CRS (kann {@code null} sein)
   */
  public static CoordinateReferenceSystem show(Component parent, String message, boolean cancelAllowed, boolean alwaysOption, CoordinateReferenceSystem defaultCRS) {
    CRSSelectionDialog dialog = new CRSSelectionDialog(parent,message,cancelAllowed,alwaysOption, defaultCRS);
    dialog.setVisible(true);
    return dialog.getCRS();
  }

  /**
   * Zeigt einen CRS-Auswahl-Dialog an, der abgebrochen werden kann. Die Eigenschaft
   * "immer diese Auswahl treffen" kann nicht eingestellt werden.
   * @param parent         uebergeordnetes Fenster (kann <code>null</code> sein!)
   * @param message        einleitende Meldung (kann <code>null</code> sein!)
   * @param defaultCRS     das als Standard angebotene CRS (kann {@code null} sein)
   * @see CRS#decode(String, boolean)
   */
  public static CoordinateReferenceSystem show(Component parent, String message, CoordinateReferenceSystem defaultCRS) {
    return show(parent,message,true,false,defaultCRS);
  }
}
