/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/

package schmitzm.geotools.gui;

import java.awt.BorderLayout;

import javax.swing.JScrollPane;

import org.geotools.styling.ColorMap;

import schmitzm.swing.InputOption;
import schmitzm.swing.JPanel;
import schmitzm.swing.SelectionInputOption;
import schmitzm.swing.event.InputOptionAdapter;

/**
 * Besides a {@linkplain ColorMapTable table} for the color map entries
 * this panel contains components to controll other {@link ColorMap}
 * properties. 
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *
 */
public class ColorMapPanel extends JPanel {
  /** Holds the color map table. */
  protected ColorMapTable colMapTable = null;
  /** Holds the combo box to chose the color map type (Ramp, Intervals,
   *  Values).  */
  protected SelectionInputOption<Integer> colMapType = null;
  
  /**
   * Creates a new panel with no color map.
   */
  public ColorMapPanel() {
    this(null);
  }

  /**
   * Creates a new panel.
   * @param colorMap color map to display
   */
  public ColorMapPanel(ColorMap colorMap) {
    this(colorMap,true);
  }

  /**
   * Creates a new panel.
   * @param colorMap color map to display
   * @param initGUI  indicates whether {@link #initGUI()} is calles at the
   *                 end of the constructor
   */
  protected ColorMapPanel(ColorMap colorMap, boolean initGUI) {
    super();
    if ( initGUI )
      initGUI();
    setColorMap(colorMap);
  }

  /**
   * Initializes the GUI.
   */
  protected void initGUI() {
    setLayout( new BorderLayout() );
    this.colMapTable = new ColorMapTable();
    this.colMapType  = new SelectionInputOption.Combo<Integer>(
        null,
        true,
        new Integer[] { ColorMap.TYPE_RAMP, ColorMap.TYPE_INTERVALS, ColorMap.TYPE_VALUES },
        new String[] { GeotoolsGUIUtil.R("org.geotools.styling.ColorMap.TYPE_RAMP"),
                       GeotoolsGUIUtil.R("org.geotools.styling.ColorMap.TYPE_INTERVALS"),
                       GeotoolsGUIUtil.R("org.geotools.styling.ColorMap.TYPE_VALUES") }
    );
    this.colMapType.addInputOptionListener( new InputOptionAdapter() {
      @Override
      public void optionChanged(InputOption inputOption, Object oldValue, Object newValue) {
        colMapTable.getColorMap().setType((Integer)newValue);
      }
    });
    
    this.add( colMapType, BorderLayout.NORTH );
    this.add( new JScrollPane(colMapTable), BorderLayout.CENTER );

  }

  /**
   * Returns the displayed {@link ColorMap}.
   */
  public ColorMap getColorMap() {
    return colMapTable.getColorMap();
  }

  /**
   * Sets the {@link ColorMap} to display.
   * @param colorMap a color map
   */
  public void setColorMap(ColorMap colorMap) {
    colMapTable.setColorMap(colorMap);
    if ( colorMap != null )
      colMapType.setSelectedItem(colorMap.getType());
  }


}
