/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import java.awt.Color;
import java.util.Map;

import javax.swing.JColorChooser;

import org.geotools.styling.ColorMap;
import org.geotools.styling.ColorMapEntry;

import schmitzm.geotools.styling.StylingUtil;
import schmitzm.swing.CaptionsChangeable;
import schmitzm.swing.table.AbstractMutableTableModel;
import schmitzm.swing.table.ColorRenderer;
import schmitzm.swing.table.MutableTable;

/**
 * Diese Klasse stellt eine Tabelle dar, in der eine {@link ColorMap} dargestellt
 * und veraendert werden kann.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of  Bonn/Germany)
 * @version 1.0
 */
public class ColorMapTable extends MutableTable implements CaptionsChangeable {
    /** Key, um den 1. Tabellenkopf-Eintrag "Quantity" in der {@link CaptionsChangeable}-Map anzusprechen.
     *  @see #resetCaptions(Map)*/
  public static final String TABLEHEADER_QUANTITY = ColorMapTable.class.getName()+".Header.QUANTITY";
  /** Key, um den 2. Tabellenkopf-Eintrag "Color" in der {@link CaptionsChangeable}-Map anzusprechen.
   *  @see #resetCaptions(Map)*/
  public static final String TABLEHEADER_COLOR = ColorMapTable.class.getName()+".Header.COLOR";
  /** Key, um den 3. Tabellenkopf-Eintrag "Label" in der {@link CaptionsChangeable}-Map anzusprechen.
   *  @see #resetCaptions(Map)*/
  public static final String TABLEHEADER_LABEL = ColorMapTable.class.getName()+".Header.LABEL";

  /**
   * Erzeugt eine neue Tabelle
   * @param colorMap darzustellende Farb-Palette
   */
  public ColorMapTable(ColorMap colorMap) {
    super( new ColorMapTableModel(colorMap), MutableTable.ITEM_ADD | MutableTable.ITEM_REMOVE );
    ((ColorMapTableModel)getModel()).table = this;
    this.setDefaultRenderer(Color.class, new ColorRenderer(true));
    this.setCellSelectionEnabled(true);
  }

  /**
   * Erzeugt eine neue Tabelle. Die dargestellte Farb-Palette muss anschliessend
   * noch mit {@link #setColorMap(ColorMap)} gesetzt werden.
   */
  public ColorMapTable() {
    this(null);
  }

  /**
   * Ruft {@link #performChange()} auf, auch wenn die CHANGE-Option nicht
   * aktiviert ist.
   */
  public void performDoubleClick() {
    performChange();
  }

  /**
   * Liefert die in der Tabelle dargestelle Farb-Palette.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   */
  public ColorMap getColorMap() {
	  //return StylingUtil.clearColorMapLabels(  ((ColorMapTableModel)getModel()).colMap  );
    return ((ColorMapTableModel)getModel()).colMap;
  }

  /**
   * Setzt die in der Tabelle dargestelle Farb-Palette.
   * @param colorMap darzustellende Farb-Palette
   */
  public void setColorMap(ColorMap colorMap) {
    ((ColorMapTableModel)getModel()).colMap = colorMap;
    ((ColorMapTableModel)getModel()).fireTableDataChanged();
  }

  /**
   * Setzt die Bezeichnungen des Tabellenheaders neu.
   * @param captionMap Map mit neuen Bezeichnungen fuer Headers
   */
  public void resetCaptions(Map<String,Object> captionMap) {
    String[] header =  ((ColorMapTableModel)getModel()).createColumnNames();

    Object caption = captionMap.get( TABLEHEADER_QUANTITY );
    if ( caption != null )
      header[0] = caption.toString();
    caption = captionMap.get( TABLEHEADER_COLOR );
    if ( caption != null )
      header[1] = caption.toString();
    caption = captionMap.get( TABLEHEADER_LABEL );
    if ( caption != null )
      header[2] = caption.toString();

    ((ColorMapTableModel)getModel()).fireTableStructureChanged();
  }

  /**
   * Diese Klasse stellt ein Daten-Modell fuer die Farbpaletten-Tabelle
   * dar.
   */
  private static class ColorMapTableModel extends AbstractMutableTableModel {
    protected MutableTable table   = null;
    protected ColorMap     colMap   = null;
    protected Class[]      colClass = new Class[]  {String.class, Color.class, String.class};

    /**
     * Erzeugt ein neues TableModel.
     * @param colorMap Farb-Palette
     */
    public ColorMapTableModel(ColorMap colorMap) {
      super();
      this.colMap = colorMap;
    }

    /**
     * Liefert die Spaltennamen der Tabelle.
     */
    @Override
    public String[] createColumnNames() {
      return new String[] {
          GeotoolsGUIUtil.RESOURCE.getString(TABLEHEADER_QUANTITY),
          GeotoolsGUIUtil.RESOURCE.getString(TABLEHEADER_COLOR),
          GeotoolsGUIUtil.RESOURCE.getString(TABLEHEADER_LABEL)
      };
    }

    /**
     * Fuer Spalte 1 wird ein Farb-Dialog aufgerufen. Spalte 0 (Wert) und 2
     * werden direkt in der Tabelle geaendert.
     * @param row Zeilennummer (beginnend bei 0)
     * @param col Spaltennummer (beginnend bei 0)
     */
    public void performChangeData(int row, int col) {
      if ( col != 1 )
        return;
      ColorMapEntry entry = colMap.getColorMapEntry(row);
      Color color = JColorChooser.showDialog(table,"Choose color",StylingUtil.getColorFromColorMapEntry(entry));
      if ( color != null )
        entry.setColor( StylingUtil.STYLE_BUILDER.colorExpression(color) );
      //changeColorMapEntry( colMap.getColorMapEntry(row) );
    }

    /**
     * Loescht einen Eintrag aus der Farb-Palette.
     * @param row Zeilennummer (beginnend bei 0)
     */
    public void performRemoveRow(int row) {
      colMap = StylingUtil.removeColorMapEntry(colMap, row);
    }

    /**
     * Ruft einen Dialog auf, in dem ein neuer Wert und eine neue Farbe
     * angegeben werden kann.
     */
    public void performAddRow() {
      colMap = StylingUtil.addColorMapEntryAndSort(
          colMap,
          StylingUtil.createColorMapEntry("",0.0,Color.WHITE,1.0)
      );
    }

/*
    private void changeColorMapEntry(ColorMapEntry oldEntry) {
      // Default-Wert fuer den Dialog ist zunaechst der alte Eintrag
      ColorMapEntry defaultEntry = oldEntry;
      while (true) {
        // Dialog anzeigen um Eintrag zu aendern/erzeugen
        ColorMapEntry newEntry = null;
        try {
          newEntry = showColorMapEntryDialog(defaultEntry);
        } catch (Exception err) {
          ExceptionDialog.show(table,err,"Error","Color map entry incorrect!");
          // nochmal versuchen (WHILE-Schleife fortfuehren)
          continue;
        }
        // wenn Dialog abgebrochen wurde oder der Eintrag nicht
        // geaendert wurde, wird nichts gemacht
        if (newEntry == null || newEntry.equals(oldEntry))
          break;

        // Eintrag einfuegen
        if ( oldEntry == null )
          // neuen Eintrag am Ende einfuegen
          colMap.addColorMapEntry( newEntry );
        else {
          // alte gegen neue Category ersetzen
          oldEntry.setColor( newEntry.getColor() );
          oldEntry.setQuantity( newEntry.getQuantity() );
        }
        this.fireTableDataChanged();
        break;
      }
    }

    private ColorMapEntry showColorMapEntryDialog(ColorMapEntry entry) throws Exception {
      Double defValue = (entry != null) ? StylingUtil.getQuantityFromColorMapEntry(entry) : 0.0;
      Color  defColor = (entry != null) ? StylingUtil.getColorFromColorMapEntry(entry) : null;

      Object[] obj = MultipleOptionPane.showMultipleInputDialog(
          table,
          "New quantity and color",
          new InputOption[] {
            new ManualInputOption.Double("Quantity",true,defValue),
            new ColorInputOption("Color",true,defColor)
          }
      );
      // Dialog abgebrochen
      if ( obj == null )
        return null;
      // Neuen Eintrag erzeugen
      return StylingUtil.createColorMapEntry("",(Double)obj[0],(Color)obj[1],1.0);
    }
*/
    /**
     * Liefert die Anzahl an Tabellenzeilen.
     */
    public int getRowCount() {
      if ( colMap == null )
        return 0;
      return colMap.getColorMapEntries().length;
    }

    /**
     * Liefert den Typ einer Spalte.
     * @param col Spaltennummer (beginnend mit 0)
     */
    public Class getColumnClass(int col) {
      return colClass[col];
    }

    /**
     * Spezifiert, ob eine Zelle editierbar ist.
     * @param row Zeilennummer (beginnend bei 0)
     * @param col Spaltennummer (beginnend bei 0)
     * @return <code>true</code> fuer Spalte 0 und 2, sonst <code>false</code>
     */
    public boolean isCellEditable(int row, int col) {
      return col == 0 || col == 2;
    }

    /**
     * Liefert einen Wert der Tabelle
     * @param row Zeilennummer (beginnend bei 0)
     * @param col Spaltennummer (beginnend bei 0)
     */
    public Object getValueAt(int row, int col) {
      switch( col ) {
        case 0: return StylingUtil.getQuantityFromColorMapEntry( colMap.getColorMapEntry(row) );
        case 1: return StylingUtil.getColorFromColorMapEntry( colMap.getColorMapEntry(row) );
        case 2: return colMap.getColorMapEntry(row).getLabel();
      }
      return null;
    }

    /**
     * Setzt einen Wert der Tabelle. Diese Methode macht nichts fuer Werte in
     * Spalte 1.
     * @param obj neuer Wert fuer die Zelle
     * @param row Zeilennummer (beginnend bei 0)
     * @param col Spaltennummer (beginnend bei 0)
     */
    public void setValueAt(Object obj, int row, int col) {
      try {
        switch( col ) {
          case 0: colMap.getColorMapEntry(row).setQuantity( StylingUtil.STYLE_BUILDER.literalExpression(Double.parseDouble((String)obj)) );
                  // if quantity has changed, sort the color map
                  colMap = StylingUtil.sortColorMap(colMap);
                  fireTableDataChanged();
                  break;
          case 2: colMap.getColorMapEntry(row).setLabel( (String)obj );
                  break;
        }
      } catch (Exception err) {
      }
    }
  }
}
