/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.Map;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.event.AncestorEvent;
import javax.swing.event.AncestorListener;

import org.apache.log4j.Logger;
import org.geotools.data.store.EmptyFeatureCollection;
import org.geotools.feature.FeatureCollection;
import org.geotools.map.DefaultMapLayer;
import org.geotools.styling.Style;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.simple.SimpleFeatureType;
import org.opengis.filter.Filter;

import schmitzm.geotools.feature.AttributeTypeFilter;
import schmitzm.geotools.feature.FeatureOperationTree;
import schmitzm.geotools.feature.FeatureOperationTreeFilter;
import schmitzm.geotools.feature.FeatureOperationTreeParser;
import schmitzm.geotools.feature.FilterParser;
import schmitzm.geotools.styling.StylingUtil;
import schmitzm.swing.ExceptionDialog;
import schmitzm.swing.SwingUtil;

/**
 * Diese Klasse stellt ein Panel zur Verfuegung, mit der ein {@link Filter} in
 * Form einer arithmetischen (und boolschen) Formel erstellt werden kann. Wie
 * der String geparst wird (z.B. CQL), wird durch einen {@link FilterParser}
 * bestimmt.<br>
 * Neben den Komponenten zur Definition/Eingabe des Filters enthaelt das Panel
 * einen Vorschau-Bereich, in dem eine {@link FeatureCollection} angezeigt wird,
 * auf der der Filter angewandt wird.
 * 
 * @see FeatureOperationTree
 * @see FeatureOperationTreeParser
 * @see FeatureOperationTreeFilter
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *         (University of Bonn/Germany)
 * @version 1.0
 */
public class FeatureCollectionFilterPanel extends FeatureFilterPanel {
	static Logger LOGGER = Logger.getLogger(FeatureCollectionFilterPanel.class);
	/**
	 * Konstante fuer die Layout-Constraints des Vorschau-Bereich.
	 * 
	 * @see #layoutConstraints
	 * @see #previewPanel
	 */
	public static final String PREVIEW_PANEL = FeatureCollectionFilterPanel.class
			.getName()
			+ "filterPreviewPanel";
	/**
	 * Konstante fuer die Layout-Constraints des Testen-Button.
	 * 
	 * @see #layoutConstraints
	 * @see #resetCaptions(Map)
	 * @see #testButton
	 */
	public static final String TEST_BUTTON = FeatureCollectionFilterPanel.class
			.getName()
			+ ".TestButton";
	/**
	 * Konstante fuer die Layout-Constraints des Testen-Button.
	 * 
	 * @see #layoutConstraints
	 * @see #testButton
	 */
	public static final String TESTRESULT_LABEL = FeatureCollectionFilterPanel.class
			.getName()
			+ ".TestResultLabel";
	/**
	 * Konstante fuer die Layout-Constraints des Details-Button.
	 * 
	 * @see #layoutConstraints
	 * @see #testResultDetails
	 */
	public static final String TESTRESULTDETAILS_BUTTON = FeatureCollectionFilterPanel.class
			.getName()
			+ ".TestResultDetailsButton";
	/**
	 * Konstante fuer den Tooltip des Details-Button.
	 * 
	 * @see #testResultDetails
	 */
	public static final String TESTRESULTDETAILS_TOOLTIP = FeatureCollectionFilterPanel.class
			.getName()
			+ ".TestResultDetailsToolTip";

	/**
	 * Constant for the BorderTitle of the results panel
	 */
	public static final String RESULTS_BORDER_TITLE = FeatureCollectionFilterPanel.class
			.getName()
			+ ".ResultsBorderTitle";

	/**
	 * Constant for the BorderTitle of the filter definition part
	 */
	public static final String FILER_DEFINITION_BORDER_TITLE = FeatureCollectionFilterPanel.class
			.getName()
			+ ".FilterDefinitionBorderTitle";
	/**
	 * Constant for the ToolTip of the preview mappane
	 */
	public static final String PREVIEW_MAPPANE_TOOLTIP = FeatureCollectionFilterPanel.class
			.getName()
			+ ".PreviewMapPaneToolTip";

	/** Panel fuer Filter-Vorschau. */
	private FeatureTablePane previewPanel = null;
	/** Button zum Testen der Formel */
	protected JButton testButton = null;
	/** Label mit Ergebnis des Formel-Tests */
	protected JLabel testResult = null;
	/** Button fuer Details des Formel-Tests-Ergebnis */
	protected JButton testResultDetails = null;

	/** Speichert, ob im Vorschau-Panel eine grafische Anzeige erscheinen soll */
	private boolean geomPrev = false;
	/** Speichert, die FeatureCollection */
	private FeatureCollection<SimpleFeatureType, SimpleFeature> fullFeatureCollection = null;

	/** Speichert den letzten Fehler in der Filter-Regel. */
	protected Throwable testError = null;

	/**
	 * A Panel with a titled border that contains all buttons used to define the
	 * filter
	 */
	protected JPanel defineFilterPanel;
	private AttributeTypeFilter attributeFilter = AttributeTypeFilter.NO_GEOMETRY;

	/**
	 * Erzeugt ein neues Panel.
	 * 
	 * @param parser
	 *            Parser um aus dem im Panel eingegebenen String einen
	 *            {@link Filter} zu erstellen.
	 * @param fc
	 *            definiert die zur Verfuegung gestellten
	 *            SimpleFeature-Attribute
	 * @param geomPrev
	 *            bestimmt, ob ein Preview-Bereich angezeigt wird ({@code true})
	 *            oder nicht ({@code false})
	 * @see FeatureCollection#getSchema()
	 */
	public FeatureCollectionFilterPanel(FilterParser parser,
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc,
			boolean geomPrev, AttributeTypeFilter attributeFilter) {
		this(parser, fc, geomPrev, true, attributeFilter);
	}

	/**
	 * Erzeugt ein neues Panel
	 * 
	 * @param parser
	 *            Parser um aus dem im Panel eingegebenen String einen
	 *            {@link Filter} zu erstellen.
	 * @param fc
	 *            definiert die zur Verfuegung gestellten
	 *            SimpleFeature-Attribute
	 * @param initGUI
	 *            Flag, ob {@link #initGUI()} am Ende des Konstruktor aufgerufen
	 *            werden soll (wenn {@code false} muss die explizit durch die
	 *            Unterklasse erfolgen!)
	 * @param attributeFilter
	 *            If null <code>null</code>,
	 *            {@value AttributeTypeFilter#NO_GEOMETRY} will be used.
	 */
	protected FeatureCollectionFilterPanel(FilterParser parser,
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc,
			boolean geomPrev, boolean initGUI,
			AttributeTypeFilter attributeFilter) {
		super(parser, fc.getSchema(), false);
		this.geomPrev = geomPrev;

		// Layout-Anordnung fuer GUI
		layoutConstraints.get(ATTRIBUTE_TABLE).gridwidth = 3;
		layoutConstraints.get(ATTRIBUTE_TABLE).weighty = 0.3;
		layoutConstraints.get(RULE_TEXTFIELD).gridwidth = 2;
		layoutConstraints.get(OPERATOR_COMBOBOX).gridx = 2;
		layoutConstraints.put(TEST_BUTTON, new GridBagConstraints(0, 4, 1, 1,
				0, 0.0, GridBagConstraints.WEST, GridBagConstraints.NONE,
				new Insets(5, 10, 10, 10), 0, 0));
		layoutConstraints
				.put(TESTRESULT_LABEL, new GridBagConstraints(1, 4, 2, 1, 1.0,
						0.0, GridBagConstraints.WEST,
						GridBagConstraints.HORIZONTAL,
						new Insets(5, 10, 10, 10), 0, 0));
		layoutConstraints.put(TESTRESULTDETAILS_BUTTON, new GridBagConstraints(
				2, 4, 1, 1, 0, 0.0, GridBagConstraints.EAST,
				GridBagConstraints.NONE, new Insets(5, 10, 10, 10), 0, 0));
		layoutConstraints.put(PREVIEW_PANEL, new GridBagConstraints(0, 5, 3, 1,
				0, 0.7, GridBagConstraints.CENTER, GridBagConstraints.BOTH,
				new Insets(5, 10, 10, 10), 0, 0));
		
		if (attributeFilter != null)
			this.attributeFilter = attributeFilter;

		if (initGUI)
			initGUI();

		setFeatureCollection(fc);
	}

	/**
	 * Initalisiert die GUI des Fensters. Der "Start-Button" der durch die
	 * Oberklasse definiert wird, wird dabei aus dem Panel entfernt!
	 */
	protected void initGUI() {
		super.initGUI();

		// Button zum Testen des Filters
		testButton = new JButton(GeotoolsGUIUtil.R(TEST_BUTTON));
		testButton.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				performRuleTest();
			}
		});

		// Button fuer Details des Test-Ergebnis
		testResultDetails = new JButton(GeotoolsGUIUtil
				.R(TESTRESULTDETAILS_BUTTON));
		testResultDetails.setVisible(false);
		testResultDetails.setToolTipText(GeotoolsGUIUtil
				.R(TESTRESULTDETAILS_TOOLTIP));
		testResultDetails.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				ExceptionDialog.show(FeatureCollectionFilterPanel.this,
						testError, true);
			}
		});

		// Label fuer Fehler-Meldungen
		testResult = new JLabel("");
		// Schrift: rot und fett
		testResult.setForeground(Color.RED);
		testResult.setFont(testResult.getFont().deriveFont(Font.BOLD));

		if (getPreviewPanel() == null) {

			setPreviewPanel(new FeatureTablePane(null, geomPrev) {

				// In der Tabelle sollen einzelne Werte selektiert werden
				// koennen
				// und diese bei einem Doppelklick in die Formel uebernommen
				// werden
				@Override
				protected void initGUI(boolean geomPreview) {
					super.initGUI(geomPreview);

					if (geomPreview){
						mapPane
						.setToolTipText(GeotoolsGUIUtil
								.R(FeatureCollectionFilterPanel.PREVIEW_MAPPANE_TOOLTIP));
					}

					// nur einzelne Zellen duerfen selektiert werden
					featuresTable.setColumnSelectionAllowed(true);
					featuresTable
							.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
					// beim Klick auf eine Tabellenzelle, soll der Wert in die
					// Formel
					// uebernommen werden
					featuresTable.addMouseListener(new MouseAdapter() {
						public void mouseClicked(MouseEvent e) {
							if (e.getButton() == MouseEvent.BUTTON1
									&& e.getClickCount() == 2)
								insertValueInRule(featuresTable.getValueAt(
										featuresTable.getSelectedRow(),
										featuresTable.getSelectedColumn()));
						}
					});

					// /*
					// * Der TableModelListener aus super.initGUI() will bei
					// einem
					// * tableChanged event nur die selektieren anzeigen. In der
					// * vorschau eines filters läuft das anders.
					// */
					// final int length =
					// featuresTableModel.getTableModelListeners().length;
					// if (length > 0)
					// featuresTableModel
					// .removeTableModelListener(featuresTableModel
					// .getTableModelListeners()[length - 1]);
					//
					// // Wenn neue FeatureCollection gesetzt wird, alle
					// Features
					// // in Vorschau anzeigen.
					// featuresTableModel
					// .addTableModelListener(new TableModelListener() {
					// public void tableChanged(TableModelEvent e) {
					// showFeaturesInMap(getFeatureCollection());
					// }
					// });
				}

				// Im Vorschau-Panel sollen nicht die in der Tabelle selektieren
				// Features angezeigt werden, sondern immer ALLE Features der
				// Tabelle (also das Resultat des Filters)
				@Override
				public void performListSelection() {
				}

				@Override
				/*
				 * This method differs from the super-method! This method does
				 * not care about selections in the table. It shows the complete
				 * table (= paramter fc) as a selection on top of the complete
				 * feature collection (which is rendered in gray).
				 */
				protected void showFeaturesInMap() {
					
					if (geomPrev == false) return;

					if (getPreviewPanel() == null)
						return;

					FeatureCollection<SimpleFeatureType, SimpleFeature> specialFeatureCollection = getFeatureCollection();

					final FeatureCollection<SimpleFeatureType, SimpleFeature> allFeatures = FeatureCollectionFilterPanel.this
							.getFeatureCollection();
					//
					// if (specialFeatureCollection != null && allFeatures !=
					// null)
					// LOGGER.debug("The result to show contains "
					// + specialFeatureCollection.size()
					// + " of total " + allFeatures.size());

					if (getPreviewPanel().mapPane.getMapContext()
							.getLayerCount() == 0) {

						if (allFeatures != null) {
							/**
							 * Als unterstes layer sollen alle Features in grau
							 * dargestellt werden. Dieses Layer soll nur einmal
							 * eingefügt werden
							 */
							getPreviewPanel().mapPane.getMapContext().addLayer(
									new DefaultMapLayer(allFeatures,
											StylingUtil.createStyleSimple(
													allFeatures,
													Color.lightGray,
													Color.darkGray)));

							// This thows an exception... We have to move to GT
							// filters i guess.
							getPreviewPanel().mapPane.getMapContext()
									.setAreaOfInterest(allFeatures.getBounds());
						}
					}

					if (specialFeatureCollection != null && allFeatures != null) {
						/**
						 * Der Inhalt der Tabelle soll in rot darüber gezeichnet
						 * werden. Dieses layer wird jedes mal neu erstellt.
						 */
						if (getPreviewPanel().mapPane.getMapContext()
								.getLayerCount() > 1)
							getPreviewPanel().mapPane.getMapContext()
									.removeLayer(1);

						// nur neues Layer erzeugen, wenn FeatureCollection
						// NICHT leer (sonst Exception im StreamingRenderer!)
						if (specialFeatureCollection.size() > 0)
							getPreviewPanel().mapPane.getMapContext().addLayer(
									new DefaultMapLayer(
											specialFeatureCollection,
											StylingUtil.createStyleSimple(
													specialFeatureCollection,
													Color.red, Color.white)));
					}
				}
			});

			// The PreviewPanel gets a Border saying what it is
			getPreviewPanel().setBorder(
					BorderFactory.createTitledBorder(GeotoolsGUIUtil
							.R(RESULTS_BORDER_TITLE)));

			// SwingUtil.setPreferredWidth(getPreviewPanel(), 600);
		}
		
		// Der Filter kann im Construktor übergeben werden und ist sonst per default ein NO_GEOMETRY-Filter
		getPreviewPanel().setAttributeFilter(attributeFilter);

		SwingUtil.setPreferredHeight(getPreviewPanel(), 150);

		/**
		 * Move all the Components that have been added so far into a new panel
		 * that has a nice border.
		 */
		defineFilterPanel = new JPanel(getLayout());
		// SwingUtil.setPreferredWidth(defineFilterPanel, 600);
		defineFilterPanel.setBorder(BorderFactory
				.createTitledBorder(GeotoolsGUIUtil
						.R(FILER_DEFINITION_BORDER_TITLE)));
		GridBagLayout gpl = (GridBagLayout) getLayout();
		for (Component c : getComponents()) {
			defineFilterPanel.add(c, gpl.getConstraints(c));
		}
		removeAll(); // From the contentpane
		setLayout(new BorderLayout());

		// zusaetzliche Komponenten einfuegen
		defineFilterPanel.add(testButton, layoutConstraints.get(TEST_BUTTON));
		defineFilterPanel.add(testResultDetails, layoutConstraints
				.get(TESTRESULTDETAILS_BUTTON));
		defineFilterPanel.add(testResult, layoutConstraints
				.get(TESTRESULT_LABEL));

		add(defineFilterPanel, BorderLayout.NORTH);

		add(getPreviewPanel(), BorderLayout.CENTER);

		// listen to the moment, the panel is inserted into a frame
		// and then set the TEST-Button the default button
		addAncestorListener(new AncestorListener() {
			@Override
			public void ancestorAdded(AncestorEvent event) {
				if (getRootPane() != null
						&& getRootPane().getDefaultButton() == null)
					getRootPane().setDefaultButton(testButton);

			}

			@Override
			public void ancestorMoved(AncestorEvent event) {
			}

			@Override
			public void ancestorRemoved(AncestorEvent event) {
			}
		});
	}

	/**
	 * Setzt den Filter, der die dargestellten Attribute bestimmt.
	 * 
	 * @param attrFilter
	 *            Filter
	 */
	public void setAttributeFilter(AttributeTypeFilter attrFilter) {
		super.setAttributeFilter(attrFilter);
		// Filter auch auf die in der Vorschau dargestellten
		// Attribut-Spalten anwenden
		getPreviewPanel().setAttributeFilter(attrFilter);
	}

	/**
	 * Setzt die Labels des Panels neu.
	 * 
	 * @param captionMap
	 *            Map
	 * @see FeatureFilterPanel#resetCaptions(Map)
	 */
	public void resetCaptions(Map<String, Object> captionMap) {
		super.resetCaptions(captionMap);
		SwingUtil.resetCaption(testButton, captionMap.get(TEST_BUTTON));

		/**
		 * TODO ?! resteCaption doesn't handle BorderTitle yet
		 * SwingUtil.resetCaption(defineFilterPanel.getBorder(),
		 * captionMap.get(FILER_DEFINITION_BORDER_TITLE)); //
		 * defineFilterPanel.setBorder(
		 * BorderFactory.createTitledBorder(GeotoolsGUIUtil.RESOURCE //
		 * .getString(FILER_DEFINITION_BORDER_TITLE)));
		 */
	}

	/**
	 * Fuegt an der Curserposition einen Wert in die Formel ein.
	 * 
	 * @param value
	 *            Object
	 */
	protected void insertValueInRule(Object value) {
		if (value == null)
			value = "";

		String ruleValue = value.toString();
		// Nur numerische Werte direkt in Formel einfuegen.
		// Andere Werte als String in Anfuehrungsstrichen gekapselt.
		if (!(value instanceof Number))
			// ruleValue = "\"" + ruleValue + "\"";
			ruleValue = "'" + ruleValue + "'"; // more common (e.g. for CQL)
		// Wert in Formel einfuegen
		performOperatorInsert(ruleValue, (JTextField) this.rule
				.getInputComponent());
		// Fokus zurueck auf Formel-Feld
		rule.grabFocus();
	}

	/**
	 * Setzt die aktuell im Dialog eingegebene Formel und aktualisiert die
	 * Vorschau.
	 * 
	 * @param rule
	 *            Formel als String
	 */
	public void setRule(String rule) {
		super.setRule(rule);
		performRuleTest();
	}

	/**
	 * Wird ausgefuehrt, wenn der Testen-Button gedrueckt wird. Erzeugt einen
	 * Filter aus der angegebenen Formel, wertet diese auf der
	 * {@link FeatureCollection} aus und zeigt das Resultat im
	 * {@linkplain #previewPanel Vorschau-Bereich} an. <br>
	 * Tritt ein Fehler auf, wird die entsprechende Meldung im Label
	 * {@link #testResult} (neben dem Button) angezeigt.
	 * 
	 */
	protected void performRuleTest() {
		try {
			FeatureCollection<SimpleFeatureType, SimpleFeature> subCollection = filterFeatureCollection();
			getPreviewPanel().setFeatureCollection(subCollection);
			resetComponentsAfterTest(null);
		} catch (Exception err) {
			getPreviewPanel().setFeatureCollection(null);
			resetComponentsAfterTest(err);
		}
		rule.grabFocus();
	}

	/**
	 * (De)aktiviert Label und Button fuer die Fehlermeldung entsprechend einer
	 * etwaigen Fehlermeldung. Wird ausgefuehrt nachdem der "Test"-Button
	 * geklickt wurde.
	 * 
	 * @param err
	 *            Fehler (kann {@code null} sein)
	 */
	protected void resetComponentsAfterTest(Throwable err) {
		testError = err;
		testResult.setText(err != null ? err.getMessage() : "");
		testResultDetails.setVisible(err != null);
	}

	/**
	 * Setzt die {@link FeatureCollection}, die (gefiltert) in der Vorschau
	 * angezeigt wird.
	 * 
	 * @param fc
	 *            eine {@link FeatureCollection}, die zum SimpleFeature-Type
	 *            (Schema) des Formel-Panels passt
	 * @exception IllegalArgumentException
	 *                falls die FeatureCollection nicht zum Schema des Panels
	 *                passt
	 * @see FeatureCollection#getSchema()
	 */
	public void setFeatureCollection(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc) {
		if (getFeatureType() == null)
			setFeatureType(fc.getSchema());
		if (!fc.getSchema().equals(getFeatureType()))
			throw new IllegalArgumentException(
					"FeatureCollection does not fit to schema of the panel!");
		this.fullFeatureCollection = fc;

		if (getPreviewPanel() != null) {
			getPreviewPanel().setFeatureCollection(fc);
		}
	}

	/**
	 * Liefert die ungefilterte {@link FeatureCollection}, auf der gefiltert
	 * wird.
	 */
	public FeatureCollection<SimpleFeatureType, SimpleFeature> getFeatureCollection() {
		return fullFeatureCollection;
	}

	/**
	 * Setzt den Style, in dem die Features in der Vorschau dargestellt werden.
	 * 
	 * @param style
	 *            Style fuer die Features
	 */
	public void setFeatureStyle(Style style) {
		getPreviewPanel().setFeatureStyle(style);
	}

	/**
	 * Liefert den Style, in dem die Features in der Vorschau dargestellt
	 * werden.
	 */
	public Style getFeatureStyle() {
		return getPreviewPanel().getFeatureStyle();
	}

	/**
	 * Wendet die im Panel eingetragene Filter-Formel auf die
	 * {@link FeatureCollection} des Vorschau-Fensters an und liefert die
	 * entsprechende Sub-Collection.
	 * 
	 * @see FeatureCollection#subCollection(Filter)
	 */
	public FeatureCollection<SimpleFeatureType, SimpleFeature> filterFeatureCollection() {
		final Filter filterParsed = createFilter();

		if (filterParsed == Filter.EXCLUDE) {
			// If no Filter String has been entered, just show all Features
			return new EmptyFeatureCollection(getFeatureType());
		}
		return getFeatureCollection().subCollection(filterParsed);
	}

	/**
	 * Wendet die im Panel eingetragene Filter-Formel auf eine
	 * {@link FeatureCollection} and und liefert die entsprechende
	 * Sub-Collection.
	 * 
	 * @see FeatureCollection#subCollection(Filter)
	 */
	public FeatureCollection<SimpleFeatureType, SimpleFeature> filterFeatureCollection(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc) {
		return fc.subCollection(createFilter());
	}

	/**
	 * Öffentliche Methode um die Belegung der Vorschaukomponente zu ändern.
	 * Sollte vor dem Aufruf von {@link #initGUI()} gesetzt werden, da diese
	 * Methode {@link #getPreviewPanel()} benutzt.
	 * 
	 * @param previewPanel
	 *            Das zu setzende {@link FeatureTablePane} oder eine eigene
	 *            Erweiterung
	 * @author SK
	 */
	public void setPreviewPanel(FeatureTablePane previewPanel) {
		this.previewPanel = previewPanel;
	}

	/**
	 * Liefert die aktuell gesetzt Vorschaukomponente. Kann auch
	 * <code>null</code> zurückgeben. Wir in {@link #initGUI()} verwendet.
	 */
	public FeatureTablePane getPreviewPanel() {
		return previewPanel;
	}
}
