/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import java.awt.GridBagConstraints;
import java.awt.Insets;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.Map;

import javax.swing.JLabel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;

import org.geotools.feature.FeatureCollection;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.simple.SimpleFeatureType;
import org.opengis.filter.Filter;

import schmitzm.geotools.feature.AttributeTypeFilter;
import schmitzm.geotools.feature.FeatureTypeTableModel;
import schmitzm.geotools.feature.FilterParser;
import schmitzm.lang.tree.OperationTreeParser;
import schmitzm.swing.OperationTreePanel;
import schmitzm.swing.SwingUtil;

/**
 * Diese Klasse stellt ein Panel zur Verfuegung, mit der ein
 * {@link Filter} in Form einer arithmetischen (und boolschen) Formel
 * erstellt werden kann. Wie der String geparst wird (z.B. CQL), wird durch
 * einen {@link FilterParser} bestimmt.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *         (University of Bonn/Germany)
 * @version 1.0
 */
public class FeatureFilterPanel extends OperationTreePanel {
	/**
	 * Konstante fuer das Label zur Tabelle der SimpleFeature-Attribute.
	 * 
	 * @see #layoutConstraints
	 * @see #resetCaptions(Map)
	 */
	public static final String ATTRIBUTE_LABEL = FeatureFilterPanel.class
			.getName()
			+ ".FeatureAttributeLabel";
	/**
	 * Konstante fuer die Tabelle der SimpleFeature-Attribute.
	 * 
	 * @see #layoutConstraints
	 */
	public static final String ATTRIBUTE_TABLE = FeatureFilterPanel.class
			.getName()
			+ ".FeatureAttributeTable";

	/** Label ueber der Tabelle mit den SimpleFeature-Attributen. */
	protected JLabel attributeLabel = null;
	/** Tabelle in der die SimpleFeature-Attribute angegeben werden. */
	protected JTable attributeTable = null;
	/** Tabellen-Modell, das den Inhalt der SimpleFeature-Attribut-Tabelle bestimmt */
	protected FeatureTypeTableModel attributeTableModel = null;
	private JScrollPane attributeScrollPane = null;
	/** Parser um aus dem im Panel eingegebenen String einen {@link Filter}
	 *  zu erstellen. */
	protected FilterParser filterParser = null;
	
	/**
	 * Erzeugt ein neues Panel.
	 * @param parser
	 *            Parser um aus dem im Panel eingegebenen String einen {@link Filter}
     *            zu erstellen.
	 * @param ftype
	 *            definiert die zur Verfuegung gestellten Feature-Attribute
	 * @see FeatureCollection#getSchema()
	 */
	public FeatureFilterPanel(FilterParser parser, SimpleFeatureType ftype) {
		this(parser,ftype, true);
	}

	/**
	 * Erzeugt ein neues Panel.
     * @param parser
     *            Parser um aus dem im Panel eingegebenen String einen {@link Filter}
     *            zu erstellen.
	 * @param fc
	 *            definiert die zur Verfuegung gestellten Feature-Attribute
	 * @see FeatureCollection#getSchema()
	 */
	public FeatureFilterPanel(FilterParser parser, FeatureCollection<SimpleFeatureType,SimpleFeature> fc) {
		this(parser,fc.getSchema());
	}

	/**
	 * Erzeugt ein neues Panel
     * @param parser
     *            Parser um aus dem im Panel eingegebenen String einen {@link Filter}
     *            zu erstellen.
	 * @param ftype
	 *            definiert die zur Verfuegung gestellten Feature-Attribute
	 * @param initGUI
	 *            Flag, ob {@link #initGUI()} am Ende des Konstruktor aufgerufen
	 *            werden soll (wenn {@code false} muss die explizit durch die
	 *            Unterklasse erfolgen!)
	 */
	protected FeatureFilterPanel(FilterParser parser, SimpleFeatureType ftype, boolean initGUI) {
		super(parser,false);
		setFilterParser(parser);

		// Zusaetzliche Operatoren und Konstanten
		// avOperators.add(...); avOperatorsDesc.put(..., ...);

		// Layout-Anordnung fuer GUI
		layoutConstraints.put(ATTRIBUTE_LABEL, new GridBagConstraints(0, 0, 2,
				1, 0, 0, GridBagConstraints.NORTHWEST, GridBagConstraints.NONE,
				new Insets(5, 10, 0, 10), 0, 0));
		layoutConstraints.put(ATTRIBUTE_TABLE, new GridBagConstraints(0, 1, 2,
				1, 1.0, 1.0, GridBagConstraints.CENTER,
				GridBagConstraints.BOTH, new Insets(0, 10, 5, 10), 0, 0));
		layoutConstraints.put(RULE_TEXTFIELD, new GridBagConstraints(0, 2, 1,
				1, 1.0, 0, GridBagConstraints.WEST,
				GridBagConstraints.HORIZONTAL, new Insets(0, 10, 5, 10), 0, 0));
		layoutConstraints.put(OPERATOR_COMBOBOX, new GridBagConstraints(1, 2,
				1, 1, 0.0, 0, GridBagConstraints.SOUTHEAST,
				GridBagConstraints.HORIZONTAL, new Insets(0, 10, 5, 10), 2, 2));
		layoutConstraints.put(START_BUTTON, new GridBagConstraints(0, 3, 1, 1,
				0.0, 0, GridBagConstraints.WEST, GridBagConstraints.NONE,
				new Insets(0, 10, 5, 10), 0, 0));

		if (initGUI) {
			initGUI();
			setFeatureType(ftype);
		}
	}

	/**
	 * Initalisiert die GUI des Fensters. Der "Start-Button" der durch die
	 * Oberklasse definiert wird, wird dabei aus dem Panel entfernt!
	 */
	protected void initGUI() {
		super.initGUI();

		attributeTableModel = new FeatureTypeTableModel();
		attributeTableModel.setAttributeFilter(AttributeTypeFilter.NO_GEOMETRY);
		attributeTable = new JTable(attributeTableModel);
		// SK.Begin 09-03-05
		// No Change, just the comment, that you can re-add the Header with
		// attributeTable.setTableHeader( new JTableHeader(
		// attributeTable.getColumnModel() ) );
		// SK.Ende 09-03-05
		attributeTable.setTableHeader(null);
		
		attributeTable.setShowGrid(false);
		attributeTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
		attributeScrollPane = new JScrollPane(attributeTable);
		SwingUtil.setPreferredHeight(attributeScrollPane, 50);

		// Bei Doppelklick auf Tabelle wird der Attribut-Name in die Formel
		// eingefuegt
		attributeTable.addMouseListener(new MouseAdapter() {
			public void mouseClicked(MouseEvent e) {
				// Nur auf Doppelklick reagieren
				if (e.getButton() != MouseEvent.BUTTON1
						|| e.getClickCount() != 2)
					return;
				// Selektiertes Attribut ermitteln
				String attrName = (String) attributeTableModel.getValueAt(
						attributeTable.getSelectedRow(), 0);
				
				
				/* Nur beim OperationTreeParser wird ein $ vor Variablen benötigt */
				String prefix = "";
				if (filterParser instanceof OperationTreeParser) {
					prefix = "$";
				}
				
				// Attribut-Namen in Formel einfuegen.
				performOperatorInsert(prefix + attrName, (JTextField) rule
						.getInputComponent());
				
				// Fokus zurueck auf Formel-Feld
				rule.grabFocus();
			}
		});

		// kein "startButton"
		remove(this.startButton);
		// zusaetzliche Komponenten einfuegen
		this.attributeLabel = new JLabel(GeotoolsGUIUtil.RESOURCE
				.getString("Attributes")
				+ ":");
		add(attributeLabel, layoutConstraints.get(ATTRIBUTE_LABEL));
		add(attributeScrollPane, layoutConstraints.get(ATTRIBUTE_TABLE));
	}

	/**
	 * Setzt die Labels des Panels neu.
	 * 
	 * @param captionMap
	 *            Map
	 * @see OperationTreePanel#resetCaptions(Map)
	 */
	public void resetCaptions(Map<String, Object> captionMap) {
		super.resetCaptions(captionMap);
		SwingUtil.resetCaption(attributeLabel, captionMap.get(ATTRIBUTE_LABEL));
	}

	/**
	 * Setzt den {@link SimpleFeatureType}, dessen Attribute fuer die Filter-Formel
	 * angeboten werden.
	 * 
	 * @param ftype
	 *            ein {@link SimpleFeatureType}
	 */
	public void setFeatureType(SimpleFeatureType ftype) {
		attributeTableModel.setFeatureType(ftype);
	}

	/**
	 * Liefert den {@link SimpleFeatureType}, dessen Attribute fuer die Filter-Formel
	 * angeboten werden.
	 */
	public SimpleFeatureType getFeatureType() {
		return attributeTableModel.getFeatureType();
	}

	/**
	 * Liefert den Filter, der die dargestellten Attribute bestimmt.
	 */
	public AttributeTypeFilter getAttributeFilter() {
		return attributeTableModel.getAttributeFilter();
	}

	/**
	 * Setzt den Filter, der die dargestellten Attribute bestimmt.
	 * 
	 * @param attrFilter
	 *            Filter
	 * @see FeatureTypeTableModel#setAttributeFilter(AttributeTypeFilter)
	 */
	public void setAttributeFilter(AttributeTypeFilter attrFilter) {
		attributeTableModel.setAttributeFilter(attrFilter);
	}

	/**
	 * Liefert die im Panel eingetragene Formel.
	 */
	public String getRule() {
		return rule.getValue();
	}

	/**
	 * Setzt die im Panel eingegebene Formel.
	 * 
	 * @param ruleString
	 *            Formel als String
	 */
	public void setRule(String ruleString) {
		rule.setValue(ruleString);
	}

	/**
	 * Setzt den Parser, mit dem aus der im Panel eingegebenen
	 * Formel ein {@link Filter} erstellt wird.
	 * @param parser der Parser
	 */
	public void setFilterParser(FilterParser parser) {
	  if ( parser == null )
	    throw new NullPointerException("FilterParser is not allowed to be NULL!");
	  this.filterParser = parser;
	}
	
    /**
     * Liefert den Parser, mit dem aus der im Panel eingegebenen
     * Formel ein {@link Filter} erstellt wird.
     */
    public FilterParser getFilterParser() {
      return this.filterParser;
    }

    /**
     * Liefert einen {@link Filter} zu der im Panel eingetragene Formel.
     * 
     * @return TRUE-Regel, falls im Panel keine Formel eingetragen ist.
     */
    public Filter createFilter() {
      return filterParser.parseFilter( getRule() );
    }
}
