/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import javax.swing.JScrollPane;
import javax.swing.JTable;

import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.simple.SimpleFeatureType;

import schmitzm.geotools.feature.FeatureTableModel;
import schmitzm.swing.InputOption;
import schmitzm.swing.MultipleOptionPane;
import schmitzm.swing.SwingUtil;

/**
 * {@linkplain InputOption Eingabe-Option} zur Definition eines {@link SimpleFeature} im
 * {@link MultipleOptionPane}.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 */
public class FeatureInputOption extends InputOption<SimpleFeature> {
  /** Beinhaltet die Tabelle, in der die Attribute eingegeben
   *  werden. */
  protected JTable inpTable;
  /** Beinhaltet das Tabellen-Modell, in dem die Attribute verwaltet
   *  werden. */
  protected FeatureTableModel inpTableModel;

  /**
   * Erzeugt einen neue Eingabe-Option.
   * @param label Ueberschrift
   * @param inputNeeded bestimmt, ob eine Eingabe erforderlich ist
   * @param feature initalial dargestelltes {@link SimpleFeature}
   */
  public FeatureInputOption(String label, boolean inputNeeded, SimpleFeature feature) {
    this(label, inputNeeded, feature != null ? feature.getFeatureType() : null);
    setValue(feature);
  }

  /**
   * Erzeugt einen neue Eingabe-Option.
   * @param label Ueberschrift
   * @param inputNeeded bestimmt, ob eine Eingabe erforderlich ist
   * @param fType bestimmt die dargestellten Attribute (kann {@code null} sein)
   */
  public FeatureInputOption(String label, boolean inputNeeded, SimpleFeatureType fType) {
    super(label, inputNeeded, true);
    inpTableModel.setFeature(fType);
  }

  /**
   * Erzeugt eine neue Instanz der Eingabe-Komponente. z.B. ein Text-Eingabefeld
   * oder eine Combo-Box.
   */
  protected JScrollPane createInputComponent() {
    this.inpTableModel = new FeatureTableModel();
    this.inpTable      = new JTable( inpTableModel );
    this.inpTable.setColumnSelectionAllowed(false);
    this.inpTable.setRowSelectionAllowed(false);
    SwingUtil.setPreferredHeight(this,200);
    SwingUtil.setPreferredWidth(this,200);
    return new JScrollPane(inpTable);
  }

  /**
   * Liefert das aktuell in der Tabelle definierte {@link SimpleFeature}.
   */
  protected SimpleFeature performGetValue() {
    // Workaround: Editieren der zuletzt editierten Zelle
    //             beenden, damit Wert in SimpleFeature uebernommen
    //             wird
    inpTable.editCellAt(-1, -1);
    return inpTableModel.getFeature();
  }
  /**
   * Initalisiert die Tabelle mit einem neuen {@link SimpleFeature}.
   * @param newValue neues {@link SimpleFeature}
   * @return immer {@code true}
   */
  protected boolean performSetValue(SimpleFeature newValue) {
    inpTableModel.setFeature(newValue);
    return true;
  }

  /**
   * Prueft, ob ein SimpleFeature definiert ist.
   */
  protected boolean performIsInputEmpty() {
    return inpTableModel.getFeature() == null;
  }

  /**
   * Prueft, ob die aktuelle Eingabe leer ist. Liefert immer {@code true}.
   */
  protected boolean performIsInputValid() {
    return true;
  }

}
