/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.text.NumberFormat;

import javax.swing.BorderFactory;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableModelListener;

import org.apache.log4j.Logger;
import org.geotools.feature.FeatureCollection;
import org.geotools.map.DefaultMapLayer;
import org.geotools.styling.Style;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.simple.SimpleFeatureType;

import schmitzm.geotools.feature.AttributeTypeFilter;
import schmitzm.geotools.styling.StylingUtil;
import schmitzm.swing.JPanel;
import schmitzm.swing.SortableJTable;
import schmitzm.swing.SwingUtil;
import schmitzm.swing.table.DoubleRenderer;

/**
 * Diese Komponente stellt eine Tabelle dar, in der die Attribute einer
 * {@link FeatureCollection} dargestellt werden. Optional werden links neben der
 * Tabelle die in der Tabelle ausgewaehlten Features grafisch dargestellt.<br>
 * <br>
 * <b>Bemerkung:</b><br>
 * Als {@code TableModel} fuer die SimpleFeature-Tabelle verwendet diese Klasse ein
 * eigenes internes {@link FeatureCollectionTableModel}, welches effizienter arbeitet, als
 * {@link org.geotools.swing.table.FeatureTableModel
 * org.geotools.swing.table.FeatureTableModel}.
 * 
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *         (University of Bonn/Germany)
 * @version 1.0
 */
public class FeatureTablePane extends JPanel {
	final static Logger LOGGER = Logger.getLogger(FeatureTablePane.class);

	/** Tabelle in der die Features angezeigt werden. */
	protected SortableJTable featuresTable = null;

	/** Tabellen-Modell der SimpleFeature-Tabelle. */
	protected FeatureCollectionTableModel featuresTableModel = null;
	/** Preview-Bereich fuer die in der Tabelle selektierten Features. */
	protected SelectableXMapPane mapPane = null;
	/** Style, in dem die Features in der Karte dargestellt werden */
	protected Style featureStyle = null;

	/**
	 * Constant for the ToolTip of the preview mappane
	 */
	public static final String PREVIEW_MAPPANE_TOOLTIP = FeatureTablePane.class
			.getName()
			+ ".PreviewMapPaneToolTip";

	/**
	 * Splitpane, das die Karten-Vorschau von der Tabelle trennt ({@code null}
	 * wenn Karten-Vorschau deaktiviert ist).
	 */
	protected JSplitPane splitPane = null;
	private JScrollPane featuresTableScrollPane;

	public JScrollPane getFeaturesTableScrollPane() {
		return featuresTableScrollPane;
	}

	/**
	 * Erzeugt einen neue Komponente mit Preview-Bereich.
	 */
	public FeatureTablePane() {
		this(false);
	}

	/**
	 * Erzeugt einen neue Komponente mit Preview-Bereich.
	 * 
	 * @param fc
	 *            angezeigte Features
	 */
	public FeatureTablePane(FeatureCollection<SimpleFeatureType,SimpleFeature> fc) {
		this(fc, true);
	}

	/**
	 * Erzeugt einen neue Komponente.
	 * 
	 * @param geomPreview
	 *            bestimmt, ob ein Preview-Bereich angezeigt wird ({@code true})
	 *            oder nicht ({@code false})
	 */
	public FeatureTablePane(boolean geomPreview) {
		this(null, geomPreview);
	}

	/**
	 * Erzeugt einen neue Komponente.
	 * 
	 * @param fc
	 *            angezeigte Features
	 * @param geomPreview
	 *            bestimmt, ob ein Preview-Bereich angezeigt wird ({@code true})
	 *            oder nicht ({@code false})
	 */
	public FeatureTablePane(FeatureCollection<SimpleFeatureType,SimpleFeature> fc, boolean geomPreview) {
		this(fc, null, geomPreview);
	}

	/**
	 * Erzeugt einen neue Komponente.
	 * 
	 * @param fc
	 *            angezeigte Features
	 * @param style
	 *            Style, in dem die Features in der Karte dargestellt werden
	 * @param geomPreview
	 *            bestimmt, ob ein Preview-Bereich angezeigt wird ({@code true})
	 *            oder nicht ({@code false})
	 */
	public FeatureTablePane(FeatureCollection<SimpleFeatureType,SimpleFeature> fc, Style style,
			boolean geomPreview) {
		this(new FeatureCollectionTableModel(fc), style, geomPreview);
	}

	/**
	 * Erzeugt einen neue Komponente.
	 * 
	 * @param model
	 *            verwendetes TableModel (kann {@code null} sein)
	 * @param fc
	 *            angezeigte Features
	 * @param style
	 *            Style, in dem die Features in der Karte dargestellt werden
	 * @param geomPreview
	 *            bestimmt, ob ein Preview-Bereich angezeigt wird ({@code true})
	 *            oder nicht ({@code false})
	 */
	public FeatureTablePane(FeatureCollectionTableModel model, Style style,
			boolean geomPreview) {
		super();
		this.featureStyle = style;
		this.featuresTableModel = (model != null) ? model
				: new FeatureCollectionTableModel();

		initGUI(geomPreview);

		featuresTableScrollPane = new JScrollPane(featuresTable,
				JScrollPane.VERTICAL_SCROLLBAR_ALWAYS,
				JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);

		// Komponenten dem Pane hinzufuegen
		if (mapPane != null) {
			splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT);
			// splitPane.setInnerBorder(null);
			splitPane.setLeftComponent(mapPane);
			splitPane.setRightComponent(featuresTableScrollPane);
			splitPane.setOneTouchExpandable(true);

			// splitPane.setDividerLocation(0.3); // SK: Does not have an
			// effect because the JSplitPane is not visible yet! Before the
			// JSplitPane is visible, only the divider location can only be
			// set in absolute int pixels.

			splitPane.setDividerLocation(mapPane.getPreferredSize().width);

			add(splitPane);

			showFeaturesInMap();
		} else {
			add(featuresTableScrollPane);
		}
	}

	/**
	 * Initalisiert die GUI.
	 * 
	 * @param geomPreview
	 *            bestimmt, ob ein Preview-Bereich angezeigt wird ({@code true})
	 *            oder nicht ({@code false})
	 */
	protected void initGUI(boolean geomPreview) {
		setLayout(new BorderLayout());
		// MapPane fuer Preview der in der Tabelle selektieren Features
		
		if (geomPreview) {
			this.mapPane = new SelectableXMapPane();
			
			// {
			// // Bei Links-Klick auf das gesamte Layer zoomen
			// public void mouseClicked(MouseEvent e) {
			// if (getState() == JMapPane.RESET && e.getButton() == e.BUTTON1) {
			// MapLayer layer = getTopLayer();
			// if (layer != null)
			// try {
			// setMapArea(layer.getFeatureSource().getBounds());
			// refresh();
			// } catch (Exception err) {
			// }
			// } else {
			// super.mouseClicked(e);
			// }
			// }
			// };
			mapPane.setToolTipText(GeotoolsGUIUtil.R(PREVIEW_MAPPANE_TOOLTIP));
//			mapPane.setState(SelectableXMapPane.NONE);
			mapPane.setMinimumSize(new Dimension(100, 100));
			SwingUtil.setPreferredWidth(mapPane, 200);
			mapPane.setBorder(BorderFactory.createLoweredBevelBorder());

		}

		// Tabelle fuer FeatureCollection
		featuresTable = new SortableJTable();

		featuresTable.setModel(featuresTableModel);
		featuresTable.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
		featuresTable.setColumnSelectionAllowed(false);
		featuresTable.setSelectionMode(ListSelectionModel.MULTIPLE_INTERVAL_SELECTION);
		// Wenn Auswahl in Tabelle sich aendert, Vorschau anpassen
		featuresTable.getSelectionModel().addListSelectionListener(
				new ListSelectionListener() {
					public void valueChanged(ListSelectionEvent e) {
					  performListSelection();
					}
				});

		// Wenn neue FeatureCollection gesetzt wird, werden trotzdem nur die
		// selektierten rot angezeigt. Der rest grau.
		featuresTableModel.addTableModelListener(new TableModelListener() {
			public void tableChanged(TableModelEvent e) {
				
				LOGGER.info("tableChanged");
				
				// showFeaturesInMap(getSelectedFeatures());
				if (mapPane != null) {
					// falls diese TableContextChange durch einen neu
					// angewendeten filter passiert ist, soll sich auch das
					// hintergund layer ändern.
					if (mapPane.getMapContext() != null)
						mapPane.getMapContext().clearLayerList();
				}
				showFeaturesInMap();
			}
		});
		// Format fuer Double auf 3 Nachkommastellen setzen
		NumberFormat formatter = NumberFormat.getNumberInstance();
		formatter.setMinimumFractionDigits(3);
		formatter.setMaximumFractionDigits(3);
		DoubleRenderer doubleRenderer = new DoubleRenderer(formatter,null);
		featuresTable.setDefaultRenderer(Double.class, doubleRenderer);
	}

	/**
	 * Liefert den Filter, der die dargestellten Attribut-Spalten bestimmt.
	 */
	public AttributeTypeFilter getAttributeFilter() {
		return featuresTableModel.getAttributeFilter();
	}

	/**
	 * Setzt den Filter, der die dargestellten Attribut-Spalten bestimmt.
	 * 
	 * @param attrFilter
	 *            Filter
	 */
	public void setAttributeFilter(AttributeTypeFilter attrFilter) {
		featuresTableModel.setAttributeFilter(attrFilter);
	}

	/**
	 * Prueft, ob der Geometry-Preview angezeigt wird.
	 */
	public boolean isGeometryPreviewVisible() {
		return mapPane != null && mapPane.isVisible();
	}

	/**
	 * Setzt die angezeigte {@link FeatureCollection}
	 * 
	 * @param fc
	 *            anzuzeigende Features
	 */
	public void setFeatureCollection(FeatureCollection<SimpleFeatureType,SimpleFeature> fc) {
		featuresTableModel.setFeatureCollection(fc);
	}

	/**
	 * Liefert die Attribut-Tabelle um deren Eigenschaften (z.B. Spaltenbreite)
	 * zu aendern.
	 */
	public SortableJTable getTable() {
		return featuresTable;
	}

	/**
	 * Liefert die angezeigten Features.
	 */
	public FeatureCollection<SimpleFeatureType,SimpleFeature> getFeatureCollection() {
		return featuresTableModel.getFeatureCollection();
	}

	/**
	 * Setzt den Style, in dem die Features in der Karte dargestellt werden.
	 * 
	 * @param style
	 *            Style fuer die Features
	 */
	public void setFeatureStyle(Style style) {
		this.featureStyle = style;
	}

	/**
	 * Liefert den Style, in dem die Features in der Karte dargestellt werden.
	 */
	public Style getFeatureStyle() {
		return this.featureStyle;
	}

	/**
	 * Liefert die in der Tabelle selektierten Features.
	 */
	public FeatureCollection<SimpleFeatureType,SimpleFeature> getSelectedFeatures() {
		return featuresTableModel.getFeaturesAsCollection(featuresTable
				.getSelectedModelRows());
	}

	/**
	 * Zeigt die aktuell in der Tabelle selektierten Features in rot im
	 * Preview-Bereich an.
	 */
	protected void performListSelection() {
		// FeatureCollection subCollection = getSelectedFeatures();
		// // Wenn keine Auswahl getaetigt ist, alle Features anzeigen
		// if (subCollection == null || subCollection.isEmpty())
		// subCollection = getFeatureCollection();
		showFeaturesInMap();
	}

	/**
	 * Zeigt einen die übergebene {@link FeatureCollection} rot in der Karte an.
	 * Darunter liegen in grau alle Features. Diese Funktion hat keinen Einfluss
	 * auf die in der Tabelle angezeigten Features!!
	 * 
	 * @param fc
	 *            eine {@link FeatureCollection}
	 */
	protected void showFeaturesInMap() {
		if (mapPane == null)
			return;

		FeatureCollection<SimpleFeatureType,SimpleFeature> selectedFeatures = getSelectedFeatures();

		/**
		 * Das layer mit allen features soll nicht immer neu eingefügt werden.
		 * Deshalb wird es nur einmal eingefügt.
		 */
		if (mapPane.getMapContext().getLayerCount() == 0) {

			final FeatureCollection<SimpleFeatureType,SimpleFeature> allFeatures = getFeatureCollection();

			/**
			 * Add ALL features to the JMapPane as the base layer and style them
			 * in grey
			 */
			if (allFeatures != null && mapPane != null) {
				
// Let's try not to use a MemoryFeatureCollection				
//				mapPane.getMapContext().addLayer(
//						new DefaultMapLayer(FeatureUtil.createMemoryFeatureSource(allFeatures), StylingUtil
//								.createStyleSimple(allFeatures,
//										Color.lightGray, Color.darkGray)));

				mapPane.getMapContext().addLayer(
				new DefaultMapLayer(allFeatures, StylingUtil
						.createStyleSimple(allFeatures,
								Color.lightGray, Color.darkGray)));
				
			}
		}

		// Das zweite Layer entfernen (wenn es schon vorhanden ist)
		if (mapPane.getMapContext().getLayerCount() > 1)
			mapPane.getMapContext().removeLayer(1);

		// Und ggf. neu hinzufügen
		if (selectedFeatures != null && selectedFeatures.size() > 0  &&
		    selectedFeatures.getSchema().getGeometryDescriptor() != null) {

			if (getFeatureStyle() == null) {
				setFeatureStyle(StylingUtil.createStyleSimple(selectedFeatures,
						Color.red, Color.white));
			}
			mapPane.getMapContext().addLayer(selectedFeatures, getFeatureStyle());

			/**
			 * TODO Das schmeisst eine blöde exception
			 * java.lang.UnsupportedOperationException: Unbekanntes
			 * SimpleFeature-Attribut: POP2002 at
			 * schmitzm.geotools.feature.FeatureOperationTree
			 * .evaluate(FeatureOperationTree.java:92) at
			 * schmitzm.lang.tree.OperationTree.evaluate(OperationTree.java:181)
			 */
//			mapPane.zoomTo(getFeatureCollection()); //
		}

		// Vorschaukarte aktualisieren
		mapPane.repaint();
	}
}
