/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. KrÃ¼ger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import javax.swing.JScrollPane;
import javax.swing.ListSelectionModel;

import org.opengis.feature.simple.SimpleFeatureType;

import schmitzm.geotools.feature.FeatureTypeBuilderTableModel;
import schmitzm.swing.InputOption;
import schmitzm.swing.MultipleOptionPane;
import schmitzm.swing.SwingUtil;
import schmitzm.swing.table.MutableTable;

/**
 * {@linkplain InputOption Eingabe-Option} zur Definition eines {@link SimpleFeatureType} im
 * {@link MultipleOptionPane}.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 */
public class FeatureTypeInputOption extends InputOption<SimpleFeatureType> {
  /** Beinhaltet die Tabelle, in der die Attribute eingegeben
   *  werden. */
  protected MutableTable inpTable;
  /** Beinhaltet das Tabellen-Modell, in dem die Attribute verwaltet
   *  werden. */
  protected FeatureTypeBuilderTableModel inpTableModel;
  /** Beinhaltet den Namen der beim Erzeugen des {@link SimpleFeatureType}
   *  gesetzt wird. */
  protected String fTypeName = null;

  /**
   * Erzeugt einen neue Eingabe-Option.
   * @param label Ueberschrift
   * @param inputNeeded bestimmt, ob eine Eingabe erforderlich ist
   */
  public FeatureTypeInputOption(String label, boolean inputNeeded) {
    this(label, inputNeeded, null);
  }


  /**
   * Erzeugt einen neue Eingabe-Option.
   * @param label Ueberschrift
   * @param inputNeeded bestimmt, ob eine Eingabe erforderlich ist
   * @param fType initalial dargestellter {@link SimpleFeatureType} (kann {@code null} sein)
   */
  public FeatureTypeInputOption(String label, boolean inputNeeded, SimpleFeatureType fType) {
    super(label, inputNeeded, true);
    if (fType != null) {
      setValue(fType);
    }
  }

  /**
   * Erzeugt eine neue Instanz der Eingabe-Komponente. z.B. ein Text-Eingabefeld
   * oder eine Combo-Box.
   */
  protected JScrollPane createInputComponent() {
    this.inpTableModel = new FeatureTypeBuilderTableModel();
    this.inpTable      = new MutableTable( inpTableModel, MutableTable.ITEM_ADD | MutableTable.ITEM_REMOVE );
    this.inpTable.setColumnSelectionAllowed(false);
    this.inpTable.setSelectionMode( ListSelectionModel.MULTIPLE_INTERVAL_SELECTION );
    SwingUtil.setPreferredHeight(this,200);
    SwingUtil.setPreferredWidth(this,400);
    return new JScrollPane(inpTable);
  }

  /**
   * Liefert den aktuellen in der Tabelle definierten {@link SimpleFeatureType}.
   */
  protected SimpleFeatureType performGetValue() {
    return inpTableModel.createFeatureType();
  }
  /**
   * Initalisiert die Tabelle mit einem neuen {@link SimpleFeatureType}.
   * @param newValue neuer {@link SimpleFeatureType}
   * @return immer {@code true}
   */
  protected boolean performSetValue(SimpleFeatureType newValue) {
    inpTableModel.setFeatureType(newValue, newValue.getTypeName());
    return true;
  }

  /**
   * Prueft, ob ein Attribut definiert ist.
   */
  protected boolean performIsInputEmpty() {
    return inpTable.getRowCount() == 0;
  }

  /**
   * Prueft, ob die aktuelle Eingabe leer ist. Wirft eine Exception
   * wenn {@link FeatureTypeBuilderTableModel#createFeatureType()}
   * eine Exception wirft.
   * @return immer {@code true}
   */
  protected boolean performIsInputValid() {
    inpTableModel.createFeatureType();
    return true;
  }

}
