/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.util.HashMap;

import org.geotools.map.MapContext;
import org.geotools.renderer.lite.StreamingRenderer;
import org.geotools.renderer.shape.ShapefileRenderer;

import schmitzm.geotools.GTUtil;
import schmitzm.geotools.map.event.JMapPaneListener;
import schmitzm.geotools.map.event.MapAreaChangedEvent;
import schmitzm.geotools.map.event.MapPaneEvent;
import schmitzm.geotools.map.event.ScaleChangedEvent;
import schmitzm.swing.JPanel;
import schmitzm.swing.SwingUtil;
import skrueger.geotools.XMapPane;

/**
 * Das {@code GeoMapPane} erweitert das {@link SelectableXMapPane} um einen
 * Massstab-Balken, sowie ein horizontales und vertikales Koordinaten-Raster
 * (Grid), in dem die Geo-Referenz des angezeigten Karten-Bereichs angezeigt wird. 
 * 
 * @see ScalePane
 * @see GridPanel
 * @author Martin Schmitz, SK
 * @version 1.1
 */
public class GeoMapPane extends JPanel {

	/**
	 * Konstante fuer das vertikale Koordinaten-Grid.
	 * 
	 * @see #layoutConstraints
	 */
	public static final String DLC_VGRID = "vertGrid";
	/**
	 * Konstante fuer das horizontale Koordinaten-Grid.
	 * 
	 * @see #layoutConstraints
	 */
	public static final String DLC_HGRID = "horGrid";
	/**
	 * Konstante fuer die Karte.
	 * 
	 * @see #layoutConstraints
	 */
	public static final String DLC_MAP = "mapPane";
	/**
	 * Konstante fuer den Massstab-Balken.
	 * 
	 * @see #layoutConstraints
	 */
	public static final String DLC_SCALE = "scalePane";
	/**
	 * Werte fuer die grafische Anordnung der Layout-Komponenten. Ueber die
	 * Konstanten {@link #DLC_MAP}, {@link #DLC_VGRID}, {@link #DLC_HGRID} und
	 * {@link #DLC_SCALE} koennen die Constraints (am Anfang von
	 * {@link #guiInit()}) veraendert oder erweitert werden.
	 */
	protected HashMap<String, GridBagConstraints> layoutConstraints = new HashMap<String, GridBagConstraints>();

	/** Karten-Bereich des {@code GeoMapPane}. */
	protected SelectableXMapPane mapPane = null;
	
	/** Panel, das den Kartenbereich ({@link #mapPane}) enthaelt.
	 *  Hierueber lassen sich bestimmte Eigenschaften (z.B. Border) steuern,
	 *  die aufgrund des optimierten Renderings fuer das {@link XMapPane}
	 *  nicht direkt gesetzt werden koennen. */
	protected JPanel mapPanePanel = null;
	
	/** Massstab-Balken */
	private ScalePane scalePane = null;
	/**
	 * Vertikale Koordinaten-Leiste (Grid). Zeigt die Latitude-Koordinate in
	 * Grad und Minutenm an.
	 */
	private GridPanel vertGrid = null;
	/**
	 * Horizontale Koordinaten-Leiste (Grid). Zeigt die Longitude-Koordinate in
	 * Grad und Minutenm an.
	 */
	private GridPanel horGrid = null;

	/**
	 * A flag indicating if dispose() was already called. If true, then further
	 * use of this {@link GeoMapPane} is undefined.
	 */
	private boolean disposed = false;
	
	/**
	 * Erzeugt ein neues {@code GeoMapPane}.
	 */
	public GeoMapPane() {
		this(null, null, null, null);
	}

	/**
	 * Erzeugt ein neues {@code GeoMapPane}. Dieser Konstruktor bietet die
	 * Moeglichkeit, alternative Implementierungen der einzelnen Komponenten
	 * darzustellen. <b>Die uebergebenen Parameter koennen {@code null} sein! In
	 * diesem Fall wird die entsprechende Standard-Implementierung
	 * verwendet.</b>
	 * 
	 * @param mapPane
	 *            Karten-Bereich
	 * @param vGrid
	 *            vertikales Koordinaten-Raster
	 * @param hGrid
	 *            horizontales Koordinaten-Raster
	 * @param scalePane
	 *            Massstab-Balken
	 * @param renderer
	 *            Allows you to choose between the {@link StreamingRenderer}
	 *            which is cool for creating PDF, SVG.. And the much faster
	 *            {@link ShapefileRenderer}. Defaults to {@link GTUtil#createGTRenderer(MapContext)}
	 */
	public GeoMapPane(SelectableXMapPane mapPane, GridPanel vGrid, GridPanel hGrid,
			ScalePane scalePane) {

	    // Massstab-Balken
		this.scalePane = scalePane;
		
		setOpaque(true);

		if (vGrid != null && !vGrid.isVertical())
			throw new IllegalArgumentException(
					"GridPanel for vertical grid must be of type GridPanel.VERTICAL!!");
		if (hGrid != null && !hGrid.isHorizontal())
			throw new IllegalArgumentException(
					"GridPanel for horizontal grid must be of type GridPanel.HORIZONTAL!!");

		// Karte
		this.mapPane = (mapPane != null) ? mapPane : new SelectableXMapPane();
		// Koordinaten-Leisten
		this.setVertGrid((vGrid != null) ? vGrid : new GridPanel(
				GridPanel.VERTICAL, this.mapPane));
		this.setHorGrid((hGrid != null) ? hGrid : new GridPanel(
				GridPanel.HORIZONTAL, this.mapPane));
		

		// Standard-Werte fuer GUI-Anordnung setzen
		layoutConstraints.put(DLC_MAP, new GridBagConstraints(1, 0, 1, 1, 1.0,
				1.0, GridBagConstraints.CENTER, GridBagConstraints.BOTH,
				new Insets(0, 0, 0, 0), 0, 0));
		layoutConstraints.put(DLC_VGRID, new GridBagConstraints(0, 0, 1, 1,
				0.0, 1.0, GridBagConstraints.CENTER, GridBagConstraints.BOTH,
				new Insets(0, 0, 0, 0), 0, 0));
		layoutConstraints.put(DLC_HGRID, new GridBagConstraints(1, 1, 1, 1,
				1.0, 0.0, GridBagConstraints.CENTER, GridBagConstraints.BOTH,
				new Insets(0, 0, 0, 0), 0, 0));
		layoutConstraints.put(DLC_SCALE, new GridBagConstraints(1, 2, 1, 1,
				0.0, 0.0, GridBagConstraints.WEST, GridBagConstraints.BOTH,
				new Insets(2, 5, 2, 5), 0, 0));
		// GUI initialisieren
		guiInit();
	}

	/**
	 * Wird vom Konstruktor aufgerufen und initialisiert die grafische
	 * Darstellung/Anordnung der einzelnen GUI-Komponenten in einem
	 * {@link GridBagLayout}. Die Constraints fuer die Anordnung der Komponenten
	 * sind in {@link #layoutConstraints} hinterlegt und koennen zu beginn
	 * dieser Methode ueberschrieben oder erweitert werden.
	 */
	protected void guiInit() {
		this.setLayout(new GridBagLayout());

		// Karten-Darestellung initialisieren
//        this.mapPane.setBorder(BorderFactory.createLoweredBevelBorder());
		SwingUtil.setPreferredWidth(this.mapPane, 200);

		// MapListener that listens to Scale and MapArea changes
		this.mapPane.addMapPaneListener(new JMapPaneListener() {
			public void performMapPaneEvent(MapPaneEvent e) {
				if (e instanceof ScaleChangedEvent) {
					ScaleChangedEvent sce = (ScaleChangedEvent) e;
					getScalePane().setScale(sce.getNewScale());
				}
				if (e instanceof MapAreaChangedEvent) {
					getVertGrid().repaint();
					getHorGrid().repaint();
				}
			}
		});
		
		// MapPane nochmal in ein Panel einfuegen, ueber das z.B.
		// die Border gesteuert werden kann (aufgrund des optimierten
		// Renderings des XMapPane koennen einige Eigenschaften nicht
		// direkt auf dem XMapPane gesetzt werden!)
        mapPanePanel = new JPanel( new BorderLayout() );
// Hier nur die "Moeglichkeit" schaffen, Border aber noch nicht setzen!
// Default: keine Border, aber Grid-Balken; wird z.B. in LayeredMapPane geaendert
//        mapPanePanel.setBorder( BorderFactory.createLoweredBevelBorder() );
        setGridBarVisible(false);
        mapPanePanel.add(mapPane,BorderLayout.CENTER);

		// Komponenten in Layout anordnen
		this.add(getVertGrid(), layoutConstraints.get(DLC_VGRID));
		this.add(getHorGrid(), layoutConstraints.get(DLC_HGRID));
		this.add(getScalePane(), layoutConstraints.get(DLC_SCALE));
		// WICHTIG: mapPane als LETZTES einfuegen, damit es beim Re-Paint ALS
		// ERSTES
		// aktualisiert wird und die Aenderungen (Area, Scale) im
		// Update des Massstab-Balken und der der Koordinaten-Leiste
		// bereits zur Verfuegung stehen!
        this.add(mapPanePanel, layoutConstraints.get(DLC_MAP));
	}

	/**
	 * Aktualisiert die Karten-Darstellung.
	 * 
	 * @see SelectableXMapPane#refresh()
	 */
	public void refreshMap() {
		mapPane.refresh();
	}

  	/**
  	 * Prueft, ob neben den Koordinaten-Abschnitten die Leiste dargestellt wird.
  	 */
	public boolean isGridBarVisible() {
	  return horGrid != null && horGrid.isGridBarVisible();
	}
	  
  	/**
  	 * Bestimmt, ob neben den Koordinaten-Abschnitten die Leiste dargestellt wird.
  	 */
	public void setGridBarVisible(boolean visible) {
	  if (horGrid != null)
	    horGrid.setGridBarVisible(visible);
      if (vertGrid != null)
        vertGrid.setGridBarVisible(visible);
  	}

	/**
	 * Nach dem {@code super}-Aufruf, wird der Massstab neu gesetzt (und somit
	 * neu angezeigt), falls sich der Massstab der Karte geaendert hat. Dies ist
	 * ein Workaround, damit der Massstab auch beim allerersten anzeigen korrekt
	 * angezeigt wird (ohne {@link ScaleChangedEvent}).
	 */
	public void paint(Graphics g) {
		super.paint(g);
		if (getScalePane().getScaleInMeters() != mapPane.getScale()
				&& mapPane.getScale() > 0)
			getScalePane().setScale(mapPane.getScale());
	}

	/**
	 * Liefert das {@link SelectableXMapPane}, in dem die Karten dargestellt werden.
	 */
	public final SelectableXMapPane getMapPane() {
		return mapPane;
	}

    /**
     * Liefert das {@link JPanel}, in dem der Kartenbereich gekapselt ist.
     * Hierbei handelt es sich nicht direkt um das {@link XMapPane}, sondern
     * um ein neues Panel, was wiederum das {@link XMapPane} enthaelt!
     */
    public final JPanel getMapPanePanel() {
        return mapPanePanel;
    }

    /**
	 * Liefert den {@link MapContext} der die einzelnen Karten-Layer verwaltet.
	 */
	public final MapContext getMapContext() {
		return mapPane.getMapContext();
	}

	/**
	 * Setzt die Hintergrundfarbe des kompletten {@link GeoMapPane}.
	 * 
	 * @param color
	 *            Hintergrundfarbe
	 */
	public void setBackground(Color color) {
		setBackground(color, true);
	}

	/**
	 * @return If dispose() has been called. If
	 *         <code>true<(code>, further use of this {@link GeoMapPane} is undefined
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public boolean isDisposed() {
		return disposed;
	}

	/**
	 * Should be called when the {@link GeoMapPane} is not needed no more to
	 * help the GarbageCollector
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public void dispose() {
		
		if (isDisposed())
			return;
		scalePane=null;
		horGrid=null;
		vertGrid=null;
		// Mach MapPane.dispose schon= getMapContext().clearLayerList();
		mapPane.dispose();
		removeAll();
		disposed = true;
	}
	
	/**
	 * Nuetzlich wenn die Componente gedruckt (z.B. wenn ein Screenshot gemacht
	 * wird) wird. Dann werden wird der Hintergrund auf WEISS gesetzt.
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	@Override
	public void print(Graphics g) {
		final Color orig = getBackground();
		setBackground(Color.WHITE);
		// wrap in try/finally so that we always restore the state
		try {
			super.print(g);
		} finally {
			setBackground(orig);
		}
	}

	public void setVertGrid(GridPanel vertGrid) {
		this.vertGrid = vertGrid;
	}

	public GridPanel getVertGrid() {
		return vertGrid;
	}

	public void setHorGrid(GridPanel horGrid) {
		this.horGrid = horGrid;
	}

	public GridPanel getHorGrid() {
		return horGrid;
	}

	public void setScalePane(ScalePane scalePane) {
		this.scalePane = scalePane;
	}

	public ScalePane getScalePane() {
		if (scalePane == null) {
			scalePane = new ScalePane();
		}
		return scalePane;
	}


}
