/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.geom.Point2D;
import java.text.DecimalFormat;
import java.util.Locale;

import javax.swing.JLabel;
import javax.swing.event.MouseInputListener;

import org.geotools.geometry.DirectPosition2D;
import org.geotools.resources.CRSUtilities;
import org.geotools.swing.event.MapMouseEvent;

import schmitzm.lang.LangUtil;
import schmitzm.lang.ResourceProvider;

/**
 * Diese Klasse stellt ein {@link JLabel} dar, in dem (2dimensionale) Geo-Koordinaten
 * angezeigt werden. Dabei werden die Koordinaten auf eine bestimmte Anzahl
 * an Nachkommastellen gerundet.<br>
 * Die Klasse fungiert als {@link MouseListener} und {@link MouseMotionListener} und
 * kann so direkt an ein {@link SelectableXMapPane} gekoppelt werden. Die Koordinaten-Darstellung
 * im Label aktualisiert sich somit automatisch, sobald sich die Maus ueber die
 * Karte bewegt. Wird ein Kartenbereich selektiert (gedrueckte linke Maustaste),
 * werden neben der aktuellen Position auch die Koordinaten des Selektionsstart
 * angezeigt.<br>
 * <b>Bemerke:</b><br>
 * Eine Instanz des <code>GeoPositionLabel</code> muss dem {@link SelectableXMapPane} sowohl
 * als {@linkplain SelectableXMapPane#addMouseListener(MouseListener) MouseListener}, als
 * auch als {@linkplain SelectableXMapPane#addMouseMotionListener(MouseMotionListener) MouseMotionListener}
 * hinzugefuegt werden. Ansonsten bekommt es nicht alle notwendigen Informationen
 * mit!
 * @see #setFractionDigits(int)
 * @see #getFractionDigits()
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class GeoPositionLabel extends JLabel implements MouseInputListener {

  // Faktor mit dem die Koordinaten vor dem Runden multipliziert und nach
  // dem Runden dividiert werden
  // --> bestimmt die dargestellten Nachkommastellen
  private double fracFactor = 0.0;

  /** Speichert das Format fuer die dargstellten Koordinaten */
  protected final DecimalFormat decForm = new DecimalFormat();

  /** Speichert die Koordinaten, die in dem Moment aktuell sind, wenn die linke
   *  Maustaste gedruckt wird. Und zwar solange, bis die Taste wieder losgelassen
   *  wird. Danach wieder <code>null</code>. */
  protected Point2D selStartCoord = null;

  /**
   * Ein Text der vor die Koorinatendarstellung geschrieben wird, z.b.: Koordinate:
   */
  private String preText = GeotoolsGUIUtil.RESOURCE
	.getString(GeoPositionLabel.class
			.getName()
			+ ".LABEL_PREFIX");

  /**
   * Erzeugt ein neues Label. Die Koordinaten werden ohne Nachkommastellen
   * dargestellt.
   */
  public GeoPositionLabel() {
    this(0);
  }

	protected static ResourceProvider RESOURCE = new ResourceProvider(LangUtil
			.extendPackagePath(GeotoolsGUIUtil.class,
					"resource.locales.GTResourceBundle"), Locale.ENGLISH);

  /**
   * Erzeugt ein neues Label.
   * @param fracDigits Anzahl an Nachkommastellen, auf die die Koordinaten gerundet
   *                   werden
   */
  public GeoPositionLabel(final int fracDigits) {
    super(" ");
    // eine Stelle vor dem Komma soll immer dargestellt werden
    this.decForm.setMinimumIntegerDigits(1);
    // Nachkommastellen setzen
    this.setFractionDigits(fracDigits);
  }

  /**
   * Stellt die Koordinaten im Label dar, wenn es sich bei dem Event um
   * ein {@link GeoMouseEvent} handelt  oder das Event von einem {@link SelectableXMapPane}
   * ausgeloest wurde.<br>
   * Wird von {@link #mousePressed(MouseEvent)}, {@link #mouseMoved(MouseEvent)}
   * und {@link #mouseDragged(MouseEvent)} aufgerufen.
   * @see #createGeoPositionString(Point2D,Point2D)
   */
  protected void displayCoordinates(final MouseEvent e) {
    if ( e==null || !(e instanceof MapMouseEvent) && !(e.getSource() instanceof SelectableXMapPane) )
      return;

    // aktuelle Geo-Position ermitteln und runden
    DirectPosition2D actCoord = SelectableXMapPane.getMapCoordinatesFromEvent(e);
    if ( actCoord == null )
      return;
    final double actLat = Math.round(actCoord.getX() * fracFactor) / fracFactor;
    final double actLon = Math.round(actCoord.getY() * fracFactor) / fracFactor;
    // Position des Selektion-Startpunkts runden
    double startLat = 0.0;
    double startLon = 0.0;
    if (selStartCoord != null) {
      startLat = Math.round(selStartCoord.getX() * fracFactor) / fracFactor;
      startLon = Math.round(selStartCoord.getY() * fracFactor) / fracFactor;
    }
    
    // Wir erfahren erst jetzt das CRS und setzen den ToolTip
    if (getToolTipText() == null) {
    	setToolTipText( preText+RESOURCE.getString(
				"schmitzm.geotools.gui.GridPanelFormatter.ToolTipTemplate",
				GeotoolsGUIUtil.getTitleForCRS(actCoord.getCoordinateReferenceSystem()), CRSUtilities.getUnit(actCoord.getCoordinateReferenceSystem()
						.getCoordinateSystem())));
    }
    

    
    // Label setzen
    setText(preText+createGeoPositionString(
        new Point2D.Double(actLat, actLon),
        selStartCoord == null ? null : new Point2D.Double(startLat, startLon)
    ));
    
    
  }

  /**
   * Erzeugt den String, in dem die Koordinaten dargestellt werden. Unterklassen
   * koennen diese Methode ueberschreiben, um die Darstellung zu aendern.
   * Die uebergebenen Koordinaten sind bereits entsprechend der gesetzten
   * Nachkommastellen gerundet!
   * @param actCoord aktuelle Geoposition des Mauszeigers
   * @param selStartCoord Geoposition des Selektion-Startpunkt (<code>null</code>
   *        falls die aktuell kein Bereich selektiert wird)
   */
  protected String createGeoPositionString(final Point2D actCoord, final Point2D selStartCoord) {
    String coordText = "";
    // wenn die linke Maustaste gedrueckt ist (also ein Bereich
    // selektiert wird), sind in selStartCoord die Startkoordinaten
    // gespeichert
    if ( selStartCoord != null )
      coordText = decForm.format(selStartCoord.getX())+" / "+decForm.format(selStartCoord.getY())+" - ";
    coordText = coordText + decForm.format(actCoord.getX())+" / "+decForm.format(actCoord.getY());

    return coordText;
  }

  /**
   * Setzt die Anzahl an Nachkommastellen, die fuer die Koordinaten
   * dargestellt werden
   * @param fracDigits Anzahl an Nachkommastellen
   */
  public void setFractionDigits(final int fracDigits) {
    this.fracFactor = Math.pow(10.0,fracDigits);
    this.decForm.setMinimumFractionDigits(fracDigits);
    this.decForm.setMaximumFractionDigits(fracDigits);
  }

  /**
   * Liefert die Anzahl an Nachkommastellen, die fuer die Koordinaten
   * dargestellt werden
   */
  public int getFractionDigits() {
    return (int)Math.log10(this.fracFactor);
  }

  /**
   * Wird aufgerufen, wenn der Maus-Button gedrueckt wird.
   * Handelt es sich bei dem Event um ein {@link GeoMouseEvent} oder das Event
   * von einem {@link SelectableXMapPane} ausgeloest wurde, werden die
   * aktuellen Koordinaten in {@link #selStartCoord} gespeichert.
   */
  public void mousePressed(final MouseEvent e) {
    final Point2D p = SelectableXMapPane.getMapCoordinatesFromEvent(e);
    if ( p != null && e.getButton() == MouseEvent.BUTTON1) {
      selStartCoord = p;
      //MS: on simple click the Envelope-Coordinates should not
      //    be shown!
      //displayCoordinates(e);
    }
  }

  /**
   * Wird aufgerufen, wenn der Maus-Button wieder losgelassen wird.
   * Setzt {@link #selStartCoord} auf <code>null</code>.
   */
  public void mouseReleased(final MouseEvent e) {
    selStartCoord = null;
  }

  /**
   * Macht nichts.
   */
  public void mouseClicked(final MouseEvent e) {
  }

  /**
   * Wird aufgerufen, sobald die Maus bewegt wird.
   * Stellt die Koordinaten im Label dar, wenn es sich bei dem Event um ein
   * {@link GeoMouseEvent} handelt oder das Event von einem {@link SelectableXMapPane}
   * ausgeloest wurde.
   */
  public void mouseMoved(final MouseEvent e) {
    displayCoordinates(e);
  }

  /**
   * Wird aufgerufen, sobald die Maus bei gedrueckter Taste bewegt wird.
   * Stellt die Koordinaten im Label dar, wenn es sich bei dem Event um ein
   * {@link GeoMouseEvent} handelt oder das Event von einem {@link SelectableXMapPane}
   * ausgeloest wurde.
   */
  public void mouseDragged(final MouseEvent e) {
    displayCoordinates(e);
  }

  /**
   * Macht nichts.
   */
  public void mouseEntered(final MouseEvent e) {
  }

  /**
   * Macht nichts.
   */
  public void mouseExited(final MouseEvent e) {
  }
}
