/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import java.awt.Component;
import java.io.File;
import java.net.URL;
import java.util.Locale;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.filechooser.FileFilter;
import javax.swing.filechooser.FileNameExtensionFilter;

import org.apache.log4j.Logger;
import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.feature.FeatureCollection;
import org.opengis.referencing.ReferenceIdentifier;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import schmitzm.geotools.io.GeoExportUtil;
import schmitzm.io.IOUtil;
import schmitzm.lang.LangUtil;
import schmitzm.lang.ResourceProvider;
import schmitzm.swing.ExceptionDialog;
import schmitzm.swing.SwingUtil;

/**
 * Diese Klasse enthaelt statische Hilfsmethoden im Bereich Geotools GUI.
 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class GeotoolsGUIUtil {
  private static Logger LOGGER = Logger.getLogger( GeotoolsGUIUtil.class.getName() );

  /** {@link ResourceProvider}, der die Lokalisation fuer GUI-Komponenten
   *  des Package {@code schmitzm.geotools.gui} zur Verfuegung stellt. Diese sind
   *  in properties-Datein unter {@code schmitzm.geotools.gui.resource.locales}
   *  hinterlegt. */
  public static ResourceProvider RESOURCE = new ResourceProvider( LangUtil.extendPackagePath(GeotoolsGUIUtil.class,"resource.locales.GTResourceBundle"), Locale.ENGLISH );

	/**
	 * Convenience method to access the translations in the {@link ResourceProvider}
	 * resources.
	 * 
	 * @param key
	 *            the key for the properties file
	 * @param values
	 *            optional values
	 */
	public static String R(String key, Object... values) {
		return RESOURCE.getString(key, values);
	}

  
  /** {@link FileNameExtensionFilter} fuer ArcInfoAscii-Raster. */
  public static final FileNameExtensionFilter FILTER_RASTER_ASCII = new FileNameExtensionFilter("ASCII-Raster (*.asc)","asc");
  /** {@link FileNameExtensionFilter} fuer GeoTiff-Raster. */
  public static final FileNameExtensionFilter FILTER_RASTER_GEOTIFF = new FileNameExtensionFilter("GeoTiff (*.tif)","tif");
  /** {@link FileNameExtensionFilter} fuer Shape-Files. */
  public static final FileNameExtensionFilter FILTER_FEATURE_SHAPE = new FileNameExtensionFilter("Shape-File (*.shp)","shp");
  
  /**
   * Zeigt einen Dialog zum Auswaehlen einer Datei an.
   * @param parent dem Dialog uebergeordnete GUI-Komponente
   * @param dialogType {@code JFileChooser.SAVE_DIALOG} oder {@code JFileChooser.OPEN_DIALOG}
   * @param title Titel, der im Dialog angezeigt wird (kann {@code null} sein)
   * @param initialDir Verzeichnis, das angezeigt wird (kann {@code null} sein)
   * @param filter (optionale) Filter, die im Dialog angewaehlt werden koennen
   * @return den FileChooser, in dem die Datei ausgewaehlt ist oder {@code null}
   *         falls der Dialog abgebrochen wurde
   */
  private static JFileChooser chooseFile(Component parent, int dialogType, String title, File initialDir, FileFilter... filter) {
    final JFileChooser fileChooser = new JFileChooser(initialDir);
    fileChooser.setFileSelectionMode(JFileChooser.FILES_ONLY);
    fileChooser.setMultiSelectionEnabled( false );
    for (FileFilter f : filter)
      fileChooser.addChoosableFileFilter(f);
    if ( filter.length > 0 ) {
      fileChooser.setFileFilter(filter[0]);
      fileChooser.setAcceptAllFileFilterUsed(false);
    }
    if ( title != null )
      fileChooser.setDialogTitle(title);
    int retValue = 0;
    if ( dialogType == JFileChooser.SAVE_DIALOG )
      retValue = fileChooser.showSaveDialog(parent);
    else if ( dialogType == JFileChooser.OPEN_DIALOG )
      retValue = fileChooser.showOpenDialog(parent);
    else
      throw new IllegalArgumentException("Unsupported dialog type: "+dialogType);

    if ( retValue != JFileChooser.APPROVE_OPTION )
      return null;

    return fileChooser;
  }

  /**
   * Zeigt einen Dialog zum Auswaehlen einer Export-Datei an. Die
   * Datei-Erweiterung wird entsprechend des ausgewaehlten Filters
   * automatisch angehaengt. Existiert die Datei bereits, wird zum Bestaetigen#
   * des Ueberschreibens aufgefordert.
   * @param parent dem Dialog uebergeordnete GUI-Komponente
   * @param dialogType {@code JFileChooser.SAVE_DIALOG} oder {@code JFileChooser.OPEN_DIALOG}
   * @param title Titel, der im Dialog angezeigt wird (kann {@code null} sein)
   * @param initialDir Verzeichnis, das angezeigt wird (kann {@code null} sein)
   * @param filter (optionale) Filter, die im Dialog angewaehlt werden koennen
   * @return den FileChooser, in dem die Datei ausgewaehlt ist oder {@code null}
   *         falls der Dialog abgebrochen wurde
   */
  private static JFileChooser chooseExportFile(Component parent, String title, File initialDir, FileNameExtensionFilter... filter) {
    JFileChooser expFileChooser = null;
    do {
      // Export-Datei auswaehlen
      expFileChooser = chooseFile(
        parent,
        JFileChooser.SAVE_DIALOG,
        title,
        initialDir,
        filter
      );
      if ( expFileChooser == null )
        return null;
      // Verzeichnis merken
      initialDir = expFileChooser.getCurrentDirectory();

      // Ueberschreiben bestaetigen
      File   expFile    = expFileChooser.getSelectedFile();
      String expFileExt = ((FileNameExtensionFilter)expFileChooser.getFileFilter()).getExtensions()[0];
      expFile = IOUtil.appendFileExt(expFile, expFileExt);
      expFileChooser.setSelectedFile(expFile);

      if ( !expFile.exists() || approveFileOverwrite(parent,expFile.getName()) )
        break;
    } while ( true );

    return expFileChooser;

  }
  /**
   * Zeigt einen Dialog an, in dem das Ueberschreiben einer Datei bestaetigt
   * werden muss.
   * @param parent Uebergeordnete GUI-Komponente fuer die Meldung
   * @param filename Bezeichnung der Datei
   * @return {@code true}, falls das Ueberschreiben mit OK bestaetigt wurde
   */
  public static boolean approveFileOverwrite(Component parent, String filename) {
    int retValue = JOptionPane.showConfirmDialog(
      parent,
      filename,
      SwingUtil.RESOURCE.getString("Overwrite"),
      JOptionPane.OK_CANCEL_OPTION,
      JOptionPane.QUESTION_MESSAGE
    );
    return retValue == JOptionPane.OK_OPTION;
  }

  /**
   * Exportiert eine {@link FeatureCollection} in ein Shape-File. Dabei
   * wird ein Anwender-Dialog geoeffnet, um die Ziel-Datei auszuwaehlen.
   * Tritt ein Fehler beim Export auf, wird dieser in einem Dialog angezeigt.
   * @param parent dem Dialog uebergeordnete GUI-Komponente (kann {@code null} sein)
   * @param fc zu exportierende {@link FeatureCollection}
   * @param desc Beschreibung der FeatureCollection, die im Dialog-Titel angezeigt
   *             wird (kann {@code null} sein)
   * @return die Export-Datei, oder {@code null} falls der Export abgebrochen
   *         oder aufgrund eines Fehlers nicht durchgefuehrt wurde
   */
  public static URL exportFeatureCollection(Component parent, FeatureCollection fc, String desc) {
    String title = RESOURCE.getString("schmitzm.geotools.gui.GeotoolsGUIUtil.SaveFeature");
    if ( desc != null && !desc.trim().equals("") )
      title += " [" + desc.trim() + "]";

    // Export-Datei auswaehlen
    JFileChooser destFileChooser = chooseExportFile(
      parent,
      title,
      null,
      FILTER_FEATURE_SHAPE
    );
    if ( destFileChooser == null )
      return null;
    File destFile = destFileChooser.getSelectedFile();

    // FeatureCollection exportieren, je nach angewaehltem Format
    try {
      FileFilter fileFilter = destFileChooser.getFileFilter();
      if ( fileFilter == FILTER_FEATURE_SHAPE )
        GeoExportUtil.writeFeaturesToShapeFile(fc, destFile);
      else
        throw new UnsupportedOperationException("Chosen file format not supported: "+fileFilter.getDescription());
      return destFile.toURI().toURL();
    } catch (Exception err) {
      ExceptionDialog.show(parent,err);
      return null;
    }
  }

  /**
   * Exportiert ein {@link GridCoverage2D} in eine Raster-Datei. Dabei
   * wird ein Anwender-Dialog geoeffnet, um die Ziel-Datei und den Datei-Typ
   * auszuwaehlen.
   * Tritt ein Fehler beim Export auf, wird dieser in einem Dialog angezeigt.
   * @param parent dem Dialog uebergeordnete GUI-Komponente (kann {@code null} sein)
   * @param gc zu exportierendes Raster
   * @param desc Beschreibung des Rasters, die im Dialog-Titel angezeigt
   *             wird (kann {@code null} sein)
   * @return die Export-Datei, oder {@code null} falls der Export abgebrochen
   *         oder aufgrund eines Fehlers nicht durchgefuehrt wurde
   */
  public static URL exportGridCoverage2D(Component parent, GridCoverage2D gc, String desc) {
    String title = RESOURCE.getString("schmitzm.geotools.gui.GeotoolsGUIUtil.SaveRaster");
    if ( desc != null && !desc.trim().equals("") )
      title += " [" + desc.trim() + "]";
    
    // Export-Datei auswaehlen
    JFileChooser destFileChooser = chooseExportFile(
      parent,
      title,
      null,
      FILTER_RASTER_ASCII,
      FILTER_RASTER_GEOTIFF
    );
    if ( destFileChooser == null )
      return null;
    File destFile = destFileChooser.getSelectedFile();

    // Raster exportieren, je nach angewaehltem Format
    try {
      FileFilter fileFilter = destFileChooser.getFileFilter();
      if ( fileFilter == FILTER_RASTER_ASCII )
        GeoExportUtil.writeGridToArcInfoASCII(gc, destFile);
      else if ( fileFilter == FILTER_RASTER_GEOTIFF )
        GeoExportUtil.writeGridToGeoTiff(gc, destFile);
      else
        throw new UnsupportedOperationException("Chosen file format not supported: "+fileFilter.getDescription());
      return destFile.toURI().toURL();
    } catch (Exception err) {
      ExceptionDialog.show(parent,err);
      return null;
    }
  }

  /**
	 * @return a good guess for a human readable string description of this CRS. An empty string in the worst case.
	 */
	public static String getTitleForCRS(CoordinateReferenceSystem crs) {
		// try {
		// final ReferenceIdentifier name = mapCrs.getName();
		// if (name != null)
		// return name.getCode();
		// else
		// return mapCrs.toString();
		// } catch (final Exception e) {
		// return mapCrs.toString();
		// }
		
		String layerCrsName = "";
		
		if (crs == null) return layerCrsName;
		
		try {
			final ReferenceIdentifier name = crs.getName();
			if (name != null)
				layerCrsName = name.getCode();
			else
				layerCrsName = crs.toString();
		} catch (final Exception e) {
			layerCrsName = crs.toString();
		}

		return layerCrsName;

	}

}
