/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.geom.AffineTransform;

import org.apache.log4j.Logger;
import org.geotools.referencing.CRS;
import org.geotools.referencing.operation.projection.PointOutsideEnvelopeException;
import org.opengis.referencing.FactoryException;
import org.opengis.referencing.crs.CoordinateReferenceSystem;
import org.opengis.referencing.operation.MathTransform;

import schmitzm.lang.LangUtil;
import schmitzm.swing.JPanel;
import schmitzm.swing.SwingUtil;

/**
 * Diese Klasse stellt eine horizontale oder vertikale Koordinaten-Leiste (Grid)
 * dar, die an ein {@link SelectableXMapPane} gekoppelt ist.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class GridPanel extends JPanel {
  private final Logger LOGGER = LangUtil.createLogger(this);

  /** Laenge der Grid-Abschnitts-Linien.  */
  private static final int GRID_LINE_SIZE = 10;
//  /** Container-Breite bei vertikalem Grid. */
//  public static final int VERT_WIDTH     = 50;
//  /** Container-Hoehe bei horizontalem Grid. */
//  public static final int HOR_HEIGHT     = 20;

  /** Flag fuer ein horizontales Grid. */
  public static final int HORIZONTAL = 1;
  /** Flag fuer ein vertikales Grid. */
  public static final int VERTICAL = 2;

  /** Speichert die Orientierung des Grid (Horizontal oder Vertikal). */
  private int orientation = 0;

  /** Karte, an der die Koordinaten-Leiste ausgerichtet wird. */
  protected SelectableXMapPane mapPane = null;

  /** Transformation von Karten-CRS zu CRS des Koordinaten-Rasters */
  protected MathTransform mapToGrid = null;
  /** Transformation von CRS des Koordinaten-Rasters zu Karten-CRS*/
  protected MathTransform gridToMap = null;
  /** Speichert das CRS, fuer das die Transformationen erstellt wurden. */
  private CoordinateReferenceSystem transformedMapCRS = null;
  /** Enthaelt Informationen, in welchem CRS die Koordinaten-Leiste dargestellt wird,
   *  und entsprechend, wie die Koordinaten in diesem CRS dargestellt werden. */
  protected GridPanelFormatter formatter = null;
  /** Flag, um das Verdrehen der Koorinaten nach der Projektion zu erzwingen.
   *  Kann in manchen Situationen sinnvoll sein. */
  protected boolean forceXYIndexSwitch = false;
  
  /** Speichert, ob neben den Koordinaten-Abschnitten auch eine Leiste
   *  dargestellt wird (Defaul: {@code true}). */
  protected boolean gridBarVisible = true;
  
  /**
   * Erzeugt eine WGS84-Koordinaten-Leiste fuer ein {@link SelectableXMapPane}.
   * @param orientation Orientierung ({@link #HORIZONTAL} oder {@link #VERTICAL})
   * @param mapPane     Karte an der die Koordinaten-Leiste ausgerichtet wird
   * @see GridPanelFormatter_LatLon1
   */
  public GridPanel(int orientation, SelectableXMapPane mapPane) {
    this(orientation,mapPane,null);
  }

  /**
   * Erzeugt eine Koordinaten-Leiste fuer ein {@link SelectableXMapPane}.
   * @param orientation Orientierung ({@link #HORIZONTAL} oder {@link #VERTICAL})
   * @param mapPane     Karte an der die Koordinaten-Leiste ausgerichtet wird
   * @param formatter   definiert das CRS und Koordinaten-Format fuer die Koordinaten
   *                    Leiste (wenn {@code null} wird ein {@link GridPanelFormatter_LatLon1}
   *                    verwendet)
   */
  public GridPanel(int orientation, SelectableXMapPane mapPane, GridPanelFormatter formatter) {
    this.orientation = orientation;
    this.mapPane     = mapPane;
    setGridFormatter(formatter);
    setXYIndexSwitchForced(false);
    
    if ( isVertical() )
      SwingUtil.setPreferredWidth(this,getGridFormatter().getWidthForVertical(this));
    else
      SwingUtil.setPreferredHeight(this,getGridFormatter().getHeightForHorizontal(this));
  }

  /**
   * Liefert die Orientierung der Koordinaten-Leiste.
   * @see #HORIZONTAL
   * @see #VERTICAL
   */
  public int getOrientation() {
    return orientation;
  }

  /**
   * Prueft, ob es sich um eine horizontale Koordinaten-Leiste handelt.
   */
  public boolean isHorizontal() {
    return this.orientation == HORIZONTAL;
  }

  /**
   * Prueft, ob es sich um eine vertikale Koordinaten-Leiste handelt.
   */
  public boolean isVertical() {
    return this.orientation == VERTICAL;
  }
  
  /**
   * Prueft, ob neben den Koordinaten-Abschnitten die Leiste dargestellt wird.
   */
  public boolean isGridBarVisible() {
    return gridBarVisible;
  }
  
  /**
   * Bestimmt, ob neben den Koordinaten-Abschnitten die Leiste dargestellt wird.
   */
  public void setGridBarVisible(boolean visible) {
    gridBarVisible = visible;
    repaint();
  }

  /**
   * Liefert den {@link GridPanelFormatter}, mit dem die Darstellung der
   * Koordinaten bestimmt wird.
   */
  public GridPanelFormatter getGridFormatter() {
    return this.formatter;
  }

  /**
   * Aendert den {@link GridPanelFormatter}, mit dem die Darstellung der
   * Koordinaten bestimmt wird.
   * @param formatter neuer {@link GridPanelFormatter} (wenn {@code null}
   *                  wird ein {@link GridPanelFormatter_LatLon1} verwendet)
   */
  public void setGridFormatter(GridPanelFormatter formatter) {
//    GridPanelFormatter oldFormatter = getGridFormatter();
    if ( formatter == null )
      formatter = new GridPanelFormatter_LatLon1();
    this.formatter = formatter;
    // wenn sich der Formatter geaendert hat, die Transformation initialisieren,
    // diese beim naechsten paint(.) erneuert werden
    
// SK:  Es kann auch der selbe GridFormatter mit einem anderen CRS sein.. deshalb erstmal trivial immer updaten.
//    if ( getGridFormatter() != oldFormatter || ) {
      mapToGrid = null;
      gridToMap = null;
//    }
      
      if ( isVertical() )
        SwingUtil.setPreferredWidth(this,getGridFormatter().getWidthForVertical(this));
      else
        SwingUtil.setPreferredHeight(this,getGridFormatter().getHeightForHorizontal(this));

      /**
       * Die Component darüber muss die neue preferred width/height beachten
       */
      if (getParent() != null) {
    	  getParent().invalidate();
    	  getParent().repaint();
      }
      
   	  setToolTipText(formatter.getTooltip());
  }
  
  /**
   * Liefert {@code true}, wenn das Verdrehen der X- und Y-Koordinate nach
   * der Transformation erzwungen wird (Default: {@code false}).
   */
  public boolean isXYIndexSwitchForced() {
    return forceXYIndexSwitch;
  }

  /**
   * Ermoeglicht, das Verdrehen der X- und Y-Koordinate nach
   * der Transformation zu erzwingen (Default: {@code false}).
   */
  public void setXYIndexSwitchForced(boolean forceXYSwitch) {
    this.forceXYIndexSwitch = forceXYSwitch;
  }

  /**
   * Zeichnet die Koordinaten-Leiste.
   * @param g2 Graphics
   */
  public void paint(Graphics g) {
    super.paint(g);
    
    Graphics2D g2 = (Graphics2D)g;
    g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
    
    
    if ( getParent() == null || !mapPane.isWellDefined())
      return;

    CoordinateReferenceSystem mapCRS = mapPane.getMapContext().getCoordinateReferenceSystem();   // (aktuelles) CRS des MapPane
    CoordinateReferenceSystem gridCRS = formatter.getCRS();  // CRS fuer die Grad/Min/Sek-Anzeige
//    try {
//      final CoordinateReferenceSystem UTM   = CRS.decode("EPSG:32631"); // utm zone 31n
//      final CoordinateReferenceSystem WGS84 = CRS.decode("EPSG:4326"); // lat/lon (WGS84)
////      final CoordinateReferenceSystem WGS84 = org.geotools.referencing.crs.DefaultGeographicCRS.WGS84;// lat/lon
////      final CoordinateReferenceSystem NAD83 = CRS.decode("EPSG:100002"); // lat/lon
////      mapCRS  = mapPane.getContext().getCoordinateReferenceSystem();
//      mapCRS = UTM;
//      gridCRS = WGS84;
//    } catch (Exception err) {
//      return;
//    }

//    gridCRS = // lat/lon
//    LOGGER.debug("mapCRS  = "+LangUtil.getSimpleClassName(mapCRS));
//    LOGGER.debug("gridCRS = "+LangUtil.getSimpleClassName(gridCRS));
//    LOGGER.debug("mapCRS: "+mapCRS);
//    LOGGER.debug("gridCRS: "+gridCRS);

    double     minX   = mapPane.getMapArea().getMinX();
    double     minY   = mapPane.getMapArea().getMinY();
    double     maxX   = mapPane.getMapArea().getMaxX();
    double     maxY   = mapPane.getMapArea().getMaxY();

// commented out, because it was unused
//    double     width  = mapPane.getMapArea().getWidth();
//    double     height = mapPane.getMapArea().getHeight();

    // Minimum/Maximum des Grids in MapPane-CRS (MCRS)
    double[] mapMin_MCRS = new double[] {minX,minY};
    double[] mapMax_MCRS = isVertical() ? new double[] {minX,maxY} : new double[] {maxX,minY};
    // Minimum/Maximum des Grids in Grid-CRS (GCRS)
    double[] mapMin_GCRS = new double[2];
    double[] mapMax_GCRS = new double[2];

    try {
      try {
        // Der Effizienz halber, die Transformation nur neu ermitteln
        // wenn sich das Karten-CRS geaendert hat
        if ( mapToGrid == null || gridToMap == null || transformedMapCRS == null || !transformedMapCRS.equals(mapCRS) ) {
          mapToGrid = CRS.findMathTransform(mapCRS, gridCRS, true);
          gridToMap = CRS.findMathTransform(gridCRS, mapCRS, true);
          // CRS merken, fuer das Transformation erstellt wurde
          transformedMapCRS = mapCRS;
        }
      } catch (FactoryException  err) {
        LOGGER.warn("CRS-Transformation to grid CRS ("+gridCRS.getName()+") failed: "+err.getMessage());
        mapToGrid = CRS.findMathTransform(mapCRS, gridCRS, true);
        gridToMap = CRS.findMathTransform(gridCRS, mapCRS, true);
        LOGGER.warn("CRS-Transformation without datum shift");
        // CRS merken, fuer das Transformation erstellt wurde
        transformedMapCRS = mapCRS;
      }
      mapToGrid.transform( mapMin_MCRS, 0, mapMin_GCRS, 0, 1  );
      mapToGrid.transform( mapMax_MCRS, 0, mapMax_GCRS, 0, 1  );
    } catch (PointOutsideEnvelopeException err) {
      return;
    } catch (Exception err) {
      LOGGER.error("CRS-Transformation to grid CRS ("+gridCRS.getName()+") failed: "+err.getMessage());
      LOGGER.debug("CRS-Transformation to grid CRS ("+gridCRS.getName()+") failed: "+err.getMessage(),err);
    }

//   // Ergebnis von Transform: {Latitude,Longitude}
//   int idx = isVertical() ? 0 : 1;
   // SEHR SEHR MERKWUERDIG: Mal liefert Transform {Latitude,Longitude}, und
   //                        manchmal {Longitude,Latitude}!!
   // WORKAROUND (klappt natuerlich nicht mit allen CRS!):
   //       Wenn  X_MCRS > Y_MCRS, dann muss auch X_GCRS > Y_GCRS
//  int idx = determineCompatibleArrayIndex(mapMin_MCRS,mapMin_GCRS,mapMax_MCRS,mapMax_GCRS);

  // Neuer Versuch des WORKAROUND:
  // Default: {Longitude,Latitude}, also {X,Y}
  int idx = isHorizontal() ? 0 : 1;
  
  
  // Wenn die CRS verdreht sind, ist auch das Resultat der Transformation verdreht  
//  boolean switchXY = isXYIndexSwitchForced() || !(mapCRS instanceof GeographicCRS ^ gridCRS instanceof GeographicCRS);
//  if ( switchXY ) {
//    LOGGER.debug("Coordinates switched!");
//    idx = 1-idx;
//  }
  
//   LOGGER.debug( isVertical() ? "V" : "H" );
//   LOGGER.debug( "Min MCRS: "+mapMin_MCRS[0]+", "+mapMin_MCRS[1]+"         GCRS:"+mapMin_GCRS[0]+", "+mapMin_GCRS[1]);
//   LOGGER.debug( "Max MCRS: "+mapMax_MCRS[0]+", "+mapMax_MCRS[1]+"         GCRS:"+mapMax_GCRS[0]+", "+mapMax_GCRS[1]);
//   LOGGER.debug( "  --> "+idx+"\n");

   double mapMin = mapMin_GCRS[idx];
   double mapMax = mapMax_GCRS[idx];
   // Maximale Groesse des Grid in Pixel
   int maxSize = isHorizontal() ? getParent().getWidth() : getParent().getHeight();

   // Passende Grid-Unterteilung in Grad
   double gridDist = formatter.determineGridDistance(this, mapMin,mapMax,maxSize);
   
   // Leiste zeichnen
   if ( isGridBarVisible() ) {
     if ( isHorizontal() )
       g2.drawLine(0,0,maxSize,0);
     else
       g2.drawLine(formatter.getWidthForVertical(this)-1,0,formatter.getWidthForVertical(this)-1,maxSize);
   }   

   // Abschnitte zeichnen
   double   firstGrid = Math.ceil(mapMin/gridDist)*gridDist; // 1. Abschnitt in Lat bzw. Lon
   double[] grid_GCRS = mapMin_GCRS.clone(); // Ausgangspunkt
   double[] grid_MCRS = new double[2];
   AffineTransform mapToWin = mapPane.getScreenToWorld();
   for (double gridPos=firstGrid; gridPos<=mapMax; gridPos+=gridDist) {
     grid_GCRS[idx] = gridPos;
     // In Map-CRS und dann in Fenster-Koordinaten transformieren
     double[] grid_WINDOW = new double[2];
     try {
       gridToMap.transform(grid_GCRS, 0, grid_MCRS, 0, 1);
       mapToWin.inverseTransform(grid_MCRS,0,grid_WINDOW,0,1);
     } catch (Throwable err) {
       LOGGER.warn("Error ignored during paint of coordinate grid: "+err);
       continue;
     }
    
     // Grid-Abschnitt zeichnen
     int x1 = isVertical() ? formatter.getWidthForVertical(this)-GRID_LINE_SIZE : (int)grid_WINDOW[0];
     int y1 = isVertical() ? (int)grid_WINDOW[1] : 0;
     int x2 = isVertical() ? x1+GRID_LINE_SIZE : x1;
     int y2 = isVertical() ? y1 : y1+GRID_LINE_SIZE;
     g2.drawLine(x1,y1,x2,y2);
     
     // Koordinate anzeigen, evt. Font setzen
     if (formatter.getFont()  != null) {
    	 g2.setFont( formatter.getFont() );
     }
     
     
     // TODO it would be great if we could print this right aligned     
     // see: g2.getFontMetrics().stringWidth(coordStr);
     
     String coordStr = formatter.formatCoordinate(this, grid_GCRS[idx]);
     
     
     x1 = isVertical() ?    5 : x1 + 5;
     y1 = isVertical() ? y1-2 : y1 + 12;
     
//     .string(coordStr, g2); 
     
     g2.drawString( coordStr, x1, y1 );
   }
  }


//  /**
//   * Diese Methode stellt einen Workaround fuer einen sehr sehr merkwuerdigen
//   * {@link MathTransform}-Effekt dar. Mal liefert Transform die Koordinaten
//   * als {Latitude,Longitude}-Array, und manchmal als {Longitude,Latitude}!
//   * Diese Methode versucht zu ermitteln, in welchem Array-Index die
//   * benoetigte Koordinate (je nachdem ob es sich um ein horizontales oder
//   * vertikales Grid handelt) tatsaechlich befindet.
//   * @param coord_SCRS Koordinate (in Quell-CRS)
//   * @param coord_DCRS transformierte Koordinate (in Ziel-CRS)
//   */
//  private int determineCompatibleArrayIndex(double[] coord_SCRS, double[] coord_DCRS) {
//    // SEHR SEHR MERKWUERDIG: Mal liefert Transform {Latitude,Longitude}, und
//    //                        manchmal {Longitude,Latitude}!!
//    // WORKAROUND (klappt natuerlich nicht mit allen CRS!):
//    //       Wenn  X_sourceCRS > Y_sourceCRS, dann muss auch X_destCRS > Y_destCRS
//    int idx = 0;
//    if (coord_SCRS[0] > coord_SCRS[1] && coord_DCRS[0] > coord_DCRS[1] ||
//        coord_SCRS[0] < coord_SCRS[1] && coord_DCRS[0] < coord_DCRS[1] ||
//        coord_SCRS[0] == coord_SCRS[1] && coord_DCRS[0] == coord_DCRS[1])
//      idx = isHorizontal() ? 0 : 1;
//    else
//      idx = isHorizontal() ? 1 : 0;
//    return idx;
//  }

//  /**
//   * Diese Methode stellt einen Workaround fuer einen sehr sehr merkwuerdigen
//   * {@link MathTransform}-Effekt dar. Mal liefert Transform die Koordinaten
//   * als {Latitude,Longitude}-Array, und manchmal als {Longitude,Latitude}!
//   * Diese Methode versucht zu ermitteln, in welchem Array-Index die
//   * benoetigte Koordinate (je nachdem ob es sich um ein horizontales oder
//   * vertikales Grid handelt) tatsaechlich befindet.
//   */
//  private int determineCompatibleArrayIndex(double[] mapMin_MCRS, double[] mapMin_GCRS,double[] mapMax_MCRS, double[] mapMax_GCRS) {
//    // SEHR SEHR MERKWUERDIG: Mal liefert Transform {Latitude,Longitude}, und
//    //                        manchmal {Longitude,Latitude}!!
//    // WORKAROUND (klappt natuerlich nicht mit allen CRS!):
//    //       Wenn  X_sourceCRS > Y_sourceCRS, dann muss auch X_destCRS > Y_destCRS
//
//
//    double diff0_GCRS = mapMax_GCRS[0] - mapMin_GCRS[0];
//    double diff1_GCRS = mapMax_GCRS[1] - mapMin_GCRS[1];
//
//    return diff0_GCRS > diff1_GCRS ? 0 : 1;
//  }
  
  /**
   * Nuetzlich wenn die Componente gedruckt (z.B. wenn ein Screenshot gemacht wird) wird. Dann werden wird der
   * Hintergrund auf WEISS gesetzt.
   *
   * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
   *         Kr&uuml;ger</a>
   */
  @Override
  public void print(Graphics g) {
      Color orig = getBackground();
      setBackground(Color.WHITE);

      // wrap in try/finally so that we always restore the state
      try {
          super.print(g);
      } finally {
          setBackground(orig);
      }
  }

}
