/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 *
 * This file is part of the SCHMITZM library - a collection of utility
 * classes based on Java 1.6, focusing (not only) on Java Swing
 * and the Geotools library.
 *
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/

package schmitzm.geotools.gui;

import java.awt.Font;
import java.util.Locale;

import org.apache.log4j.Logger;
import org.geotools.resources.CRSUtilities;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import schmitzm.lang.LangUtil;
import schmitzm.lang.ResourceProvider;
import sun.swing.SwingUtilities2;

/**
 * This class defines how the {@link GridPanel} shows the coordinate grid and
 * the coordinates.
 * 
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 */
public abstract class GridPanelFormatter {
	private static final Font defaultFont = new Font(Font.MONOSPACED,
			Font.PLAIN, 11);

	/**
	 * @return List of instances of all available formatters.
	 */
	public final static Class<? extends GridPanelFormatter>[] FORMATTERS = new Class[] {
			GridPanelFormatter_LatLon1.class, GridPanelFormatter_DHDN.class };

	private static final Logger LOGGER = LangUtil
			.createLogger(GridPanelFormatter.class);

	protected static ResourceProvider RESOURCE = new ResourceProvider(LangUtil
			.extendPackagePath(GeotoolsGUIUtil.class,
					"resource.locales.GTResourceBundle"), Locale.ENGLISH);

	public static GridPanelFormatter getFormatterByID(String ID) {
		for (Class<? extends GridPanelFormatter> x : FORMATTERS) {

			GridPanelFormatter newInstance;
			try {
				newInstance = x.newInstance();
			} catch (Exception e) {
				throw new RuntimeException(e);
			}

			if (newInstance.getId().equals(ID))
				return newInstance;
		}

		return null;
	}

	/** CRS the formatter is created for. */
	protected CoordinateReferenceSystem crs = null;

	/**
	 * Creates a new formatter.
	 * 
	 * @param crs
	 *            the CRS the formatter can format
	 */
	public GridPanelFormatter(CoordinateReferenceSystem crs) {
		this.crs = crs;
	}

	/**
	 * Determines a "fitting" fragmentation for the grid.
	 * 
	 * @param gridPanel
	 *            indicates (amongst others) the grid orientation
	 * @param mapMinCoord
	 *            minimal coordinate to show in the grid (in the grid CRS)
	 * @param mapMaxCoord
	 *            maximal coordinate to show in the grid (in the grid CRS)
	 * @param panelSize
	 *            size of the grid panel in pixels
	 */
	public abstract double determineGridDistance(GridPanel panel,
			double mapMinCoord, double mapMaxCoord, int panelSize);

	/**
	 * Formats a coordinate for the grid CRS.
	 * 
	 * @param coord
	 *            the coordinate
	 * @param gridPanel
	 *            indicates (amongst others) the grid orientation
	 * @return
	 */
	public abstract String formatCoordinate(GridPanel panel, double coord);

	/**
	 * Returns the CRS the formatter can format.
	 */
	public CoordinateReferenceSystem getCRS() {
		return crs;
	}

	public Font getFont() {
		return defaultFont;
	}

	/**
	 * Returns the default height for a horizontal Grid
	 */
	public int getHeightForHorizontal(GridPanel panel) {
		return SwingUtilities2.getFontMetrics(panel, getFont()).getHeight() + 2;
	}

	/**
	 * A String ID that can be used to identify a Formatter-class. Take care,
	 * that the string is compatible with XML and doesn't containg freaky chars.
	 */
	public abstract String getId();

	/** A human readable title **/
	public abstract String getTitle();

	/**
	 * Returns the default width for a vertical Grid
	 */
	public abstract int getWidthForVertical(GridPanel panel);

	/**
	 * Returns the CRS the formatter can format.
	 */
	public void setCRS(CoordinateReferenceSystem crs) {
		this.crs = crs;
	}

	/**
	 * Returns a tooltip String to applied to the {@link GridPanel}s. May be
	 * <code>null</code>
	 */
	public String getTooltip() {
		if (crs == null)
			return null;
		return RESOURCE.getString(
				"schmitzm.geotools.gui.GridPanelFormatter.ToolTipTemplate",
				GeotoolsGUIUtil.getTitleForCRS(crs), CRSUtilities.getUnit(crs
						.getCoordinateSystem()));
	}

}
