/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 *
 * This file is part of the SCHMITZM library - a collection of utility
 * classes based on Java 1.6, focusing (not only) on Java Swing
 * and the Geotools library.
 *
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 *
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/

package schmitzm.geotools.gui;

import java.awt.Font;
import java.text.DecimalFormat;

import org.geotools.referencing.crs.DefaultGeographicCRS;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import schmitzm.lang.ResourceProvider;
import sun.swing.SwingUtilities2;

/**
 * This class defines a {@link GridPanelFormatter} for the WGS84
 * {@link CoordinateReferenceSystem}. The coordinates are formated in degrees
 * and minutes along with the wind direction abbreviation (E/W, N/S). grid and
 * the coordinates. * @author <a href="mailto:martin.schmitz@koeln.de">Martin
 * Schmitz</a>
 */
public class GridPanelFormatter_LatLon1 extends GridPanelFormatter {
	private static final DecimalFormat MINUTE_FORMAT = new DecimalFormat("00");
	private static final Font font = new Font(Font.MONOSPACED, Font.PLAIN, 11);
	private static final ResourceProvider RES = GeotoolsGUIUtil.RESOURCE;

	/**
	 * Creates a new formatter.
	 * 
	 * @param crs
	 *            the CRS the formatter can format
	 */
	public GridPanelFormatter_LatLon1() {
		super(DefaultGeographicCRS.WGS84);
	}

	/**
	 * Determines a "fitting" fragmentation for the grid.
	 * 
	 * @param mapMin_LL
	 *            minimal Lat/Lon coordinate to show in the grid
	 * @param mapMax_LL
	 *            maximal Lat/Lon coordinate to show in the grid
	 * @param panelSize
	 *            size of the grid panel in pixels
	 */
	@Override
	public double determineGridDistance(GridPanel panel, double mapMin_LL,
			double mapMax_LL, int panelSize) {
		// Differenz zwischen Min/Max-Koordinaten = darzustellender Bereich in
		// Grad
		double diff = Math.abs(mapMin_LL - mapMax_LL);
		// Distanz zwischen zwei Unterteilungen in Grad
		double dist = 0.0;
		if (diff < 10 / 60.0) // < 10 Min. --> 1 Min.-Aufteilung
			dist = 1 / 60.0;
		else if (diff < 30 / 60.0) // < 30 Min. --> 5 Min.-Aufteilung
			dist = 5 / 60.0;
		else if (diff < 1) // < 1? --> 10 Min.-Aufteilung
			dist = 10 / 60.0;
		else if (diff < 3) // < 3? --> 15 Min.-Aufteilung
			dist = 15 / 60.0;
		else if (diff < 5) // < 5? --> 30 Min.-Aufteilung
			dist = 30 / 60.0;
		else if (diff < 11) // < 11? --> 1?-Aufteilung
			dist = 1;
		else
			// sonst --> 2?-Aufteilung
			dist = 2.;

		// Solange der Abstand zwischen 2 Unterteilungen weniger als 40 Pixel
		// betraegt --> Unterteilung verdoppeln
		while (diff != 0 && dist != 0 && panelSize / (diff / dist) < 40)
			dist *= 2.;

		// Solange der Abstand zwischen 2 Unterteilungen mehr als 120 Pixel
		// betraegt --> Unterteilung * 0.5
		while (diff != 0 && dist != 0 && panelSize / (diff / dist) > 120)
			dist /= 2.;

		return dist;
	}

	/**
	 * Formats a coordinate to degrees and minutes.
	 * 
	 * @param coord
	 *            the lat/lon coordinate
	 * @param gridPanel
	 *            indicates (amongst others) the grid orientation
	 * @return
	 */
	@Override
	public String formatCoordinate(GridPanel panel, double coord) {
		int degrees = (int) coord;
		int minutes = (int) Math.round((coord - degrees) * 60);
		// Aufgrund des Minuten-Systems kann schlecht im Vorhinein gerundet
		// werden. Deshalb wird im Nachhinein ein etwaiger Rundungsfehler
		// ausgeglichen
		if (Math.abs(minutes) >= 60) {
			degrees += Math.signum(minutes);
			minutes -= Math.signum(minutes) * 60;
		}
		int deegree = Math.abs(degrees);
		int minute = Math.abs(minutes);
		String coordStr = deegree + "\u00b0" + MINUTE_FORMAT.format(minute) + "'";
		if (degrees != 0 || minutes != 0) {
			if (panel.isVertical())
				coordStr += RES
						.getString((coord < 0) ? "schmitzm.geotools.gui.GeotoolsGUIUtil.SOUTH.Abb"
								: "schmitzm.geotools.gui.GeotoolsGUIUtil.NORTH.Abb");
			else
				coordStr += RES
						.getString((coord < 0) ? "schmitzm.geotools.gui.GeotoolsGUIUtil.WEST.Abb"
								: "schmitzm.geotools.gui.GeotoolsGUIUtil.EAST.Abb");
		}
		return coordStr;
	}

	@Override
	public Font getFont() {
		return font;
	}

	@Override
	public String getId() {
		return "1_lat_lon";
	}

	@Override
	public String getTitle() {
		return RESOURCE
				.getString("schmitzm.geotools.gui.GridPanelFormatter_LatLon1.title");
	}

	@Override
	public int getWidthForVertical(GridPanel panel) {
		return SwingUtilities2.getFontMetrics(panel, font).stringWidth(
				"90\u00b030' N") + 2;
	}

}
