/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.util.Vector;

import javax.swing.BorderFactory;
import javax.swing.JFrame;
import javax.swing.JPanel;

import org.geotools.feature.FeatureCollection;
import org.geotools.map.MapContext;
import org.geotools.map.MapLayer;
import org.geotools.map.event.MapLayerListEvent;
import org.geotools.map.event.MapLayerListListener;

import schmitzm.geotools.map.event.FeatureSelectedEvent;
import schmitzm.geotools.map.event.GridCoverageSelectedEvent;
import schmitzm.geotools.map.event.JMapPaneListener;
import schmitzm.geotools.map.event.MapPaneEvent;
import schmitzm.swing.SelectionInputOption;
import schmitzm.swing.SwingUtil;

/**
 * Diese Klasse stellt ein Fenster dar, in dem layer-basiert Objekte grafisch
 * dargestellt werden koennen. Hinzugefuegt werden die Objekte direkt in das
 * {@link LayeredMapPane#addLayer(Object,String)}<br>
 * Welche Objekte dargestellt werden koennen, ist der Beschreibung der
 * {@link LayeredMapPane#isVisualisable(Object)}-Methode zu entnehmen.<br>
 * Das Fenster besteht aus 3 Komponenten:
 * <ol>
 * <li>Eine Map ({@link SelectableXMapPane}) zu grafischen Darstellung der Layer
 * </li>
 * <li>Eine Liste mit Steuerungskomponenten, ueber die die einzelnen Layer
 * angesprochen werden koennen (ein/ausblenden, zoomen, ...).</li>
 * <li>Eine Status-Zeile, in der die Koordinaten der aktuellen Mausposition
 * angezeigt werden, sowie der Raster-Wert des obersten Rasters unterhalb des
 * Mauszeigers.</li>
 * </ol>
 * Die grafischen Layer (in der Map) koennen wahlweise (de)aktiviert werden.
 * Dies geschieht durch setzen/entfernen eines Haekchens in der entsprechenden
 * Steuerungskomponente. Diese enthaelt zudem ein Kontextmenue, ueber welches
 * <ul>
 * <li>das jeweilige Layer aus der Map entfernt werden kann</li>
 * <li>das jeweilige Layer in der Map eine Ebene nach oben/unten verschoben
 * werden kann</li>
 * <li>die Map sogezoomt werden kann, dass das jeweilige Layer komplett
 * angezeigt wird</li>
 * </ul>
 * Um Layer einzufuegen koennen die {@code addLayer(.)}-Methoden des
 * {@link #getLayeredMapPane() LayeredMapPane} (dabei wird ein Default-Style
 * verwendet) oder die entsprechenden Methoden des {@link MapContext} ({@code
 * getLayeredMapPane().getMapPane().getContext()}).
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *         (University of Bonn/Germany)
 * @version 1.0
 */
public class LayeredMapFrame extends JFrame {
  private JPanel contentPane = null;
  /** Karten- und Layer-Kontroll-Bereich. */
  protected LayeredMapPane layeredMapPane = null;
  private SelectableXMapPane mapPane = null;
  private MapContext mapContext = null;
  /** Toolbar zur Kontroller der Karten-Aktionen (z.B. Zoom oder Select). */
  protected MapActionControlPane mapControl = null;
  /** Status-Balken. */
  protected MapPaneStatusBar statusBar = null;
  /** Fenster fuer SimpleFeature-Details */
  protected FeatureCollectionFrame featuresFrame = null;
  /**
   * Auswahlfeld fuer das Raster, fuer welches die Koordinaten angezeigt werden.
   */
  protected SelectionInputOption<MapLayer> rasterComboBox = null;

  /**
   * Erzeugt ein neues (leeres) Map-Fenster.
   */
  public LayeredMapFrame() {
    this(null, "");
  }

  /**
   * Erzeugt ein neues (leeres) Map-Fenster.
   * @param lmp {@link LayeredMapPane} welches zur Anzeige der Karten verwendet
   *          wird (wenn {@code null} wird eine neue {@link LayeredMapPane}
   *          -Instanz erzeugt)
   */
  public LayeredMapFrame(LayeredMapPane lmp) {
    this(lmp, "");
  }

  /**
   * Erzeugt ein neues (leeres) Map-Fenster.
   * @param title Titel des Fensters
   */
  public LayeredMapFrame(String title) {
    this(null, title);
  }

  /**
   * Erzeugt ein neues (leeres) Map-Fenster.
   * @param title Titel des Fensters
   * @param lmp {@link LayeredMapPane} welches zur Anzeige der Karten verwendet
   *          wird (wenn {@code null} wird eine neue {@link LayeredMapPane}
   *          -Instanz erzeugt)
   */
  public LayeredMapFrame(LayeredMapPane lmp, String title) {
    super();
    this.setSize(new Dimension(500, 400));
    this.setTitle(title);
    contentPane = (JPanel) this.getContentPane();
    contentPane.setLayout(new GridBagLayout());

    // Karte und Layer-Kontrolle
    layeredMapPane = lmp != null ? lmp : new LayeredMapPane();
    mapPane = layeredMapPane.getMapPane();
    mapContext = mapPane.getMapContext();
    mapControl = new MapActionControlPane(mapPane, MapActionControlPane.VERTICAL);
    mapControl.setFloatable(false);

    // unter Layer-Liste eine ComboBox mit Raster-Auswahl einfuegen
    this.rasterComboBox = new SelectionInputOption.Combo<MapLayer>("", false);
    this.mapContext.addMapLayerListListener(new MapLayerListListener() {
      public void layerChanged(MapLayerListEvent e) {
        updateRasterComboBox();
      }

      public void layerMoved(MapLayerListEvent e) {
      }

      public void layerAdded(MapLayerListEvent e) {
        updateRasterComboBox();
      }

      public void layerRemoved(MapLayerListEvent e) {
        updateRasterComboBox();
      }
    });
    this.layeredMapPane.horSplitPane.getContainer(0).add(rasterComboBox,
                                                         BorderLayout.SOUTH);

    // Spezielles RasterPositionLabel in dem die Koordinaten des in der
    // ComboBox ausgewaehlten Rasters angezeigt werden
    RasterPositionLabel rpLabel = new RasterPositionLabel(1) {
      protected MapLayer determineRasterLayer(final SelectableXMapPane mapPane) {
        return rasterComboBox.getValue();
      }
    };

    // Status-Zeile mit Raster-Auswahlfeld
    statusBar = new MapPaneStatusBar(mapPane, rpLabel, null);
    statusBar.setBorder(BorderFactory.createCompoundBorder(
                                                           BorderFactory.createLoweredBevelBorder(),
                                                           BorderFactory.createEmptyBorder(
                                                                                           2,
                                                                                           5,
                                                                                           2,
                                                                                           5)));

    // Status-Zeile mit Raster-Auswahlfeld kombinieren
    JPanel statusBarContainer = new JPanel();
    statusBarContainer.setLayout(new BorderLayout());
    statusBarContainer.add(rasterComboBox, BorderLayout.WEST);
    statusBarContainer.add(statusBar, BorderLayout.CENTER);

    // Root-SplitPane in Fenster einfuegen
    contentPane.add(layeredMapPane, new GridBagConstraints(0, 0, 1, 1, 1.0,
        1.0, GridBagConstraints.CENTER, GridBagConstraints.BOTH, new Insets(0,
            0, 0, 0), 0, 0));
    contentPane.add(mapControl, new GridBagConstraints(1, 0, 1, 1, 0.0, 0.0,
        GridBagConstraints.CENTER, GridBagConstraints.BOTH, new Insets(2, 0, 0,
            0), 0, 0));
    contentPane.add(statusBarContainer, new GridBagConstraints(0, 1, 2, 1, 1.0,
        0.0, GridBagConstraints.CENTER, GridBagConstraints.BOTH, new Insets(2,
            0, 0, 0), 0, 0));

    // Fenster fuer SimpleFeature-Details
    featuresFrame = new FeatureCollectionFrame(null, true);
    featuresFrame.setSize(new Dimension(400, 200));
    // Ausgewaehlte Features werden im Detail-Frame angezeigt
    mapPane.addMapPaneListener(new JMapPaneListener() {
      public void performMapPaneEvent(MapPaneEvent e) {
        // Wenn Features ueber die Maus aus der Karte ausgewaehlt werden,
        // oeffnet sich ein Detail-Fenster
        if (e instanceof FeatureSelectedEvent && e.getSourceObject() == mapPane) {
          FeatureSelectedEvent fse = (FeatureSelectedEvent) e;
          FeatureCollection fc = fse.getSelectionResult();
          featuresFrame.setFeatureCollection(fc);
          featuresFrame.setTitle(fse.getSourceLayer().getTitle() + " [" +
                                 fse.getSelectionRange() + "]");
          if (!featuresFrame.isVisible()) {
            SwingUtil.setRelativeFramePosition(featuresFrame, 1, 0);
          }
          featuresFrame.setVisible(true);
        }

        if (e instanceof GridCoverageSelectedEvent) {
          // ...
        }
      }
    });
  }

  protected void updateRasterComboBox() {
    // Letzte Auswahl merken
    MapLayer lastSelection = (MapLayer) rasterComboBox.getValue();
    // Alle Raster-Layer und Titel ermitteln
    Vector<MapLayer> rasterLayer = new Vector<MapLayer>();
    Vector<String> rasterLayerDesc = new Vector<String>();
    for (MapLayer layer : mapPane.getMapContext().getLayers())
      if (mapPane.isGridCoverageLayer(layer)) {
        rasterLayer.add(layer);
        rasterLayerDesc.add(layer.getTitle());
      }
    // Auswahl neu setzen
    rasterComboBox.setSelectionObjects(rasterLayer.toArray(new MapLayer[0]),
                                       rasterLayerDesc.toArray());
    // Wenn nur ein Raster zur Verfuegung steht, dieses autom. auswaehlen
    if (rasterLayer.size() == 1)
      lastSelection = rasterLayer.firstElement();
    // Letzte Auswahl setzen
    rasterComboBox.setSelectedItem(lastSelection);
  }

  /**
   * Liefert den Karten- und Kontroll-Bereich des Fensters.
   */
  public LayeredMapPane getLayeredMapPane() {
    return this.layeredMapPane;
  }

  /**
   * Liefert den Status-Bereich des Fensters.
   */
  public MapPaneStatusBar getStatusBar() {
    return this.statusBar;
  }

//  /**
//   * Liefert den Toolbar fuer die Karten-Aktionen.
//   */
//  public MapActionControlPane getToolBar() {
//    return this.mapControl;
//  }
}
