/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import java.awt.BorderLayout;
import java.util.Hashtable;

import javax.swing.BorderFactory;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;

import org.geotools.coverage.Category;
import org.geotools.coverage.GridSampleDimension;
import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.coverage.grid.GridCoverageFactory;
import org.geotools.feature.FeatureCollection;
import org.geotools.map.MapContext;
import org.geotools.map.MapLayer;
import org.geotools.map.event.MapLayerListEvent;
import org.geotools.map.event.MapLayerListListener;
import org.geotools.styling.ColorMap;
import org.geotools.styling.Style;
import org.opengis.geometry.Envelope;

import schmitzm.geotools.feature.FeatureUtil;
import schmitzm.geotools.grid.GridUtil;
import schmitzm.geotools.styling.ColorMapManager;
import schmitzm.swing.JPanel;
import schmitzm.swing.MultiSplitPane;

/**
 * Diese Klasse stellt ein Panel dar, in dem layer-basiert Objekte grafisch
 * dargestellt werden koennen.<br>
 * Welche Objekte dargestellt werden koennen, ist der Beschreibung der
 * {@link #isVisualisable(Object)}-Methode zu entnehmen.<br>
 * Das Fenster besteht aus 2 durch einen Divider voneinander getrennten
 * Komponenten:
 * <ol>
 * <li>Eine Map ({@link SelectableXMapPane}) zu grafischen Darstellung der Layer
 * </li>
 * <li>Eine Liste mit Steuerungskomponenten, ueber die die einzelnen Layer
 * angesprochen werden koennen (ein/ausblenden, zoomen, ...).</li>
 * </ol>
 * Die grafischen Layer (in der Map) koennen wahlweise (de)aktiviert werden.
 * Dies geschieht durch setzen/entfernen eines Haekchens in der entsprechenden
 * Steuerungskomponente. Diese enthaelt zudem ein Kontextmenue, ueber welches
 * <ul>
 * <li>das jeweilige Layer aus der Map entfernt werden kann</li>
 * <li>das jeweilige Layer in der Map eine Ebene nach oben/unten verschoben
 * werden kann</li>
 * <li>die Map sogezoomt werden kann, dass das jeweilige Layer komplett
 * angezeigt wird</li>
 * </ul>
 * Darueberhinaus verwaltet das {@code LayerMapPane} eine Liste von
 * {@link ColorMapManager Farbpaletten} fuer Rasterdaten, die durch die
 * Layer-Kontrolle den Raster-Layern zugewiesen werden koennen.<br>
 * Um Layer einzufuegen koennen die {@code addLayer(.)}-Methoden dieser Klasse
 * verwendet werden (dabei wird ein Default-Style verwendet) oder die
 * entsprechenden Methoden des {@link MapContext} ({@code
 * getMapPane().getContext()}).
 * @author Martin Schmitz
 * @version 1.0
 */
public class LayeredMapPane extends JPanel {
  protected MultiSplitPane horSplitPane = new MultiSplitPane(2,
      JSplitPane.HORIZONTAL_SPLIT);
  /**
   * Komponente, in der die Karten, der Massstab und das Koordinaten-Raster
   * dargestellt werden
   */
  protected GeoMapPane geoMapPane = null;
  private MapContext mapContext = null;
  private SelectableXMapPane mapPane = null;
  private Hashtable<MapLayer, Object> layerObjects = new Hashtable<MapLayer, Object>();
  /** Komponente, in der die Layer-Kontrolle dargestellt ist. */
  protected MapContextControlPane layerControlList = null;
  /** Auswahl an Farb-Paletten fuer Raster-Daten */
  protected ColorMapManager colorMaps = null;

  /**
   * Erzeugt eine neue Komponente.
   */
  public LayeredMapPane() {
    this(null);
  }

  /**
   * Erzeugt eine neue Komponente.
   * @param geoMapPane Komponente, die zum Anzeigen der Karten verwendet wird
   *          (wenn {@code null} wird eine neue {@link GeoMapPane}-Instanz
   *          erzeugt)
   */
  public LayeredMapPane(GeoMapPane geoMapPane) {
    super();
    if (geoMapPane == null)
      geoMapPane = new GeoMapPane();
    this.setLayout(new BorderLayout());

    // Kartenbereich ueber Border "versenken"
    geoMapPane.getMapPanePanel()
              .setBorder(BorderFactory.createLoweredBevelBorder());
    // Aufgrund der Border werden die Koordinaten-Leisten
    // nicht mehr benoetigt
    geoMapPane.setGridBarVisible(false);

    // rechter Bereich: Map, Grids und Scale
    this.geoMapPane = geoMapPane;
    this.mapPane = geoMapPane.getMapPane();
    this.mapContext = mapPane.getMapContext();
    this.mapContext.addMapLayerListListener(new MapLayerListListener() {
      public void layerAdded(MapLayerListEvent e) {
      }

      public void layerChanged(MapLayerListEvent e) {
      }

      public void layerMoved(MapLayerListEvent e) {
      }

      public void layerRemoved(MapLayerListEvent e) {
        layerObjects.remove(e.getLayer());
      }
    });

    // linker Bereich: Layer-Liste mit Auswahl der Layer
    this.colorMaps = new ColorMapManager();
    this.layerControlList = new MapContextControlPane(mapPane, colorMaps);

    // horizontales SplitPane initialisieren
    for (int i = 0; i < horSplitPane.getContainerCount(); i++)
      horSplitPane.getContainer(i).setLayout(new BorderLayout());
    horSplitPane.setResizeWeigth(new double[] { 0.2, 0.8 });
    horSplitPane.setDividerSize(5);
    horSplitPane.setInnerBorder(null);
    horSplitPane.getContainer(0).add(new JScrollPane(layerControlList),
                                     BorderLayout.CENTER);
    horSplitPane.getContainer(1).add(geoMapPane);

    this.add(horSplitPane);
  }

  /**
   * Liefert den Karten-Bereich der Komponente.
   */
  public SelectableXMapPane getMapPane() {
    return mapPane;
  }

  /**
   * Liefert das in einem Layer dargestellte Objekt.
   * @param layer ein Layer
   */
  public Object getMapObject(MapLayer layer) {
    return layerObjects.get(layer);
  }

  /**
   * Liefert den Manager fuer die zur Verfuegung stehenden Farbpaletten.
   */
  public ColorMapManager getColorMapManager() {
    return this.colorMaps;
  }

  /**
   * Prueft, ob Instanzen einer bestimmten Klasse von {@link LayeredMapFrame}
   * visualisiert werden koennen. Dies gilt fuer
   * <ul>
   * <li>{@link GridCoverage2D org.geotools.coverage.grid.GridCoverage2D}</li>
   * <li>{@link FeatureCollection org.geotools.feature.FeatureCollection}</li>
   * </ul>
   */
  public boolean isVisualisable(Class c) {
    return GridCoverage2D.class.isAssignableFrom(c) ||
           FeatureCollection.class.isAssignableFrom(c);
  }

  /**
   * Prueft, ob eine Objekt-Instanz von {@link LayeredMapFrame} visualisiert
   * werden kann. Dies gilt fuer Instanzen von
   * <ul>
   * <li>{@link GridCoverage2D org.geotools.coverage.grid.GridCoverage2D}</li>
   * <li>{@link FeatureCollection org.geotools.feature.FeatureCollection}</li>
   * </ul>
   */
  public boolean isVisualisable(Object o) {
    return o instanceof GridCoverage2D || o instanceof FeatureCollection;
  }

  /**
   * Fuegt ein Raster-Layer (als oberstes Layer) ein.
   * @param gc ein Grid-Coverage
   * @param desc Beschreibung fuer das Raster
   * @param style Darstellungs-Style fuer das Layer
   */
  public MapLayer addLayer(GridCoverage2D gc, String desc, Style style) {
    // GridSampleDimension gsd = gc.getSampleDimensions()[0];
    // for (int i = 0; i < gsd.getCategories().size(); i++) {
    // Category cat = (Category) gsd.getCategories().get(i);
    // System.out.print("'" + cat.getName() + "'\t" + cat.getRange().toString()
    // +"\t"+cat.isQuantitative()+"\t");
    // for (int j = 0; j < cat.getColors().length; j++)
    // System.out.print(cat.getColors()[j].toString() + ",");
    // System.out.println("\n           > " + cat.toString() + "  (" +
    // cat.getClass().getName() + ")");
    // }

    // // Farbiges Grid erzeugen
    // Category[] c = new Category[0];
    // try {
    // c = new Category[] {
    // // (Category) gsd.getCategories().get(0),
    // // (Category) gsd.getCategories().get(1)
    // CoverageUtil.createGeophysicsCategory( new Category("", new
    // Color[]{Color.WHITE, Color.BLACK}, new NumberRange(0.0,255.0),1,0), true
    // ),
    // CoverageUtil.createGeophysicsCategory( new Category("No data",
    // Color.YELLOW, Float.NaN), false ),
    // // CoverageUtil.createGeophysicsCategory( new Category("No data", new
    // Color(0, 0, 0), Float.NaN), true ),
    // // CoverageUtil.createGeophysicsCategory( new Category("0", Color.BLACK,
    // 0.0f), true ),
    // // CoverageUtil.createGeophysicsCategory( new Category("1", Color.BLACK,
    // 1.0f), true ),
    // // CoverageUtil.createGeophysicsCategory( new Category("2", Color.BLACK,
    // 2.0f), true ),
    // // CoverageUtil.createGeophysicsCategory( new Category("3", Color.BLACK,
    // 3.0f), true ),
    // // CoverageUtil.createGeophysicsCategory( new Category("4", Color.BLACK,
    // 4.0f), true ),
    // // CoverageUtil.createGeophysicsCategory( new Category("5", Color.BLACK,
    // 5.0f), true )
    // };
    // } catch ( Exception err ) {
    // err.printStackTrace();
    // }
    // gc = recolorGrid(gc,0,c);
    //
    // gsd = gc.getSampleDimensions()[0];
    // for (int i = 0; i < gsd.getCategories().size(); i++) {
    // Category cat = (Category) gsd.getCategories().get(i);
    // System.out.print("'" + cat.getName() + "'\t" + cat.getRange().toString()
    // +"\t"+cat.isQuantitative()+"\t");
    // for (int j = 0; j < cat.getColors().length; j++)
    // System.out.print(cat.getColors()[j].toString() + ",");
    // System.out.println("\n           > " + cat.toString() + "  (" +
    // cat.getClass().getName() + ")");
    // }

    // Falls kein Style angegeben ist, aus der Standard-Farbpalete erstellen
    if (style == null) {
      ColorMap colorMap = colorMaps.getDefaultColorMap();
      style = GridUtil.createStyle(colorMap, 1.0);
    }
    // Layer erzeugen
    mapContext.addLayer(gc, style);
    MapLayer newLayer = mapContext.getLayer(mapContext.getLayerCount() - 1);
    newLayer.setTitle(desc);

    // neuer Anzeigebereich: Das komplette Raster
    if (mapPane.getMapArea() == null) {
      Envelope e = gc.getEnvelope();
      com.vividsolutions.jts.geom.Envelope newArea = new com.vividsolutions.jts.geom.Envelope(
          e.getUpperCorner().getOrdinate(0), // X1
          e.getLowerCorner().getOrdinate(0), // X2
          e.getUpperCorner().getOrdinate(1), // Y1
          e.getLowerCorner().getOrdinate(1) // Y2
      );
      mapPane.setMapArea(newArea);
    }

    // Anzeige aktualisieren
    mapPane.repaint();

    // Wenn Rendering geklappt hat, zum Layer gehoerendes Objekt merken
    layerObjects.put(newLayer, gc);

    return newLayer;
  }

  /**
   * Fuegt ein FeatureCollection-Layer (als oberstes Layer) ein.
   * @param fc eine SimpleFeature-Collection
   * @param desc Beschreibung fuer die SimpleFeature-Collection
   * @param style Darstellungs-Style fuer das Layer
   */
  public MapLayer addLayer(FeatureCollection fc, String desc, Style style) {
    if (style == null)
      style = FeatureUtil.createDefaultStyle(fc);
    // Layer erzeugen
    mapContext.addLayer(fc, style);
    MapLayer newLayer = mapContext.getLayer(mapContext.getLayerCount() - 1);
    newLayer.setTitle(desc);

    // neuer Anzeigebereich: Die komplette FeatureCollection
    if (mapPane.getMapArea() == null)
      mapPane.setMapArea(fc.getBounds());

    // Anzeige aktualisieren
    mapPane.repaint();

    // Wenn Rendering geklappt hat, zum Layer gehoerendes Objekt merken
    layerObjects.put(newLayer, fc);

    return newLayer;
  }

  /**
   * Fuegt ein Layer (als oberstes Layer) ein.
   * @param obj ein (darstellbares) Objekt
   * @param desc Beschreibung fuer das Objekt
   * @param style Darstellungs-Style fuer das Layer
   * @exception UnsupportedOperationException falls ein nicht-darstellbares
   *              Objekt uebergeben wird
   * @see #isVisualisable(Object)
   * @see #isVisualisable(Class)
   */
  public MapLayer addLayer(Object obj, String desc, Style style) {
    if (obj instanceof GridCoverage2D)
      return addLayer((GridCoverage2D) obj, desc, style);
    if (obj instanceof FeatureCollection)
      return addLayer((FeatureCollection) obj, desc, style);
    throw new UnsupportedOperationException(
        "LayeredMapFrame can not visualise objects of class " +
            obj.getClass().getName());
  }

  /**
   * Fuegt ein Layer (als oberstes Layer) ein. Es wird ein Standard-Style zur
   * Darstellung verwendet.
   * @param obj ein (darstellbares) Objekt
   * @param desc Beschreibung fuer das Objekt
   * @exception UnsupportedOperationException falls ein nicht-darstellbares
   *              Objekt uebergeben wird
   * @see #isVisualisable(Object)
   * @see #isVisualisable(Class)
   */
  public MapLayer addLayer(Object obj, String desc) {
    return addLayer(obj, desc, null);
  }

  /**
   * Erzeugt ein neues {@link GridCoverage2D} mit einer neuen Farb-Zuordnung.
   * @param grid Basis-Grid
   * @param band Band, dessen Farbe veraendert werden soll
   * @param cat Rasterwert/Farb-Zuordnungen
   */
  private static GridCoverage2D recolorGrid(GridCoverage2D grid, int band,
      Category[] cat) {
    GridSampleDimension[] gsd = grid.getSampleDimensions();
    gsd[band] = new GridSampleDimension(gsd[band].getDescription(), cat,
        gsd[band].getUnits());

    return new GridCoverageFactory().create(grid.getName(),
                                            grid.getRenderedImage(),
                                            grid.getEnvelope(), gsd,
                                            new GridCoverage2D[] { grid }, null);
  }

}
