/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import java.awt.Component;
import java.awt.event.ActionEvent;
import java.util.Map;

import javax.swing.AbstractAction;
import javax.swing.ButtonGroup;
import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JToggleButton;
import javax.swing.JToolBar;

import schmitzm.geotools.map.event.FeatureSelectedEvent;
import schmitzm.swing.CaptionsChangeable;
import schmitzm.swing.SwingUtil;
import skrueger.geotools.XMapPaneTool;

/**
 * Diese Klasse stellt einen {@link JToolBar} dar, mit dem zwischen den
 * verschiedenen "Klick"- und "Drag"-Aktionen des {@link SelectableXMapPane} gewechselt werden
 * kann:
 * <ul>
 *   <li><b>Info:</b>
 *       <ul>
 *         <li>Links-Klick:  nichts</li>
 *         <li>Links-Drag: nichts</li>
 *         <li>Rechts-Klick: nichts</li>
 *         <li>Rechts-Drag: Karten-Ausschnitt verschieben</li>
 *       </ul></li>
 *   <li><b>Zoom:</b>
 *       <ul>
 *         <li>Links-Klick:  Zoom in</li>
 *         <li>Links-Drag: Zoom in auf selektierten Bereich</li>
 *         <li>Rechts-Klick: Zoom out</li>
 *         <li>Rechts-Drag: Karten-Ausschnitt verschieben</li>
 *       </ul></li>
 *   <li><b>SelectTop:</b><br>
 *       Die Auswahl-Aktionen beziehen sich auf das oberste sichtbare SimpleFeature
 *       und fuehren zu einem {@link FeatureSelectedEvent}:
 *       <ul>
 *         <li>Links-Klick: einzelnes SimpleFeature selektieren</li>
 *         <li>Links-Drag: alle Features im selektierten Bereich auswaehlen</li>
 *         <li>Rechts-Klick: nichts</li>
 *         <li>Rechts-Drag: Karten-Ausschnitt verschieben</li>
 *       </ul></li>
 *   <li><b>SelectAll:</b><br>
 *       Wie <i>SelectTop</i>. Die Auswahl-Aktionen beziehen sich jedoch auf
 *       auf alle prinzipiell sichtbaren Layer (evt. verdeckte Features unterer
 *       Layer werden auch ausgewaehlt!)
 *       </li>
 * </ul>
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class MapActionControlPane extends JToolBar implements CaptionsChangeable {
  /** Konstante um die Aktion "Info" in der Mask der zur Verfuegung stehenden
   *  Buttons anzusprechen. */
  public static final int ACTION_INFO = 1;
  /** Konstante um die Aktion "Zoom" in der Mask der zur Verfuegung stehenden
   *  Buttons anzusprechen. */
  public static final int ACTION_ZOOM_IN = 2;
  /** Konstante um die Aktion "SelectAll" in der Mask der zur Verfuegung stehenden
   *  Buttons anzusprechen. */
  public static final int ACTION_SELECT_ALL = 4;
  /** Konstante um die Aktion "SelectTop" in der Mask der zur Verfuegung stehenden
   *  Buttons anzusprechen. */
  public static final int ACTION_SELECT_TOP = 8;
  /** Konstante um die ALLE Aktionen in der Mask der zur Verfuegung stehenden
   *  Buttons anzusprechen. */
  public static final int ACTION_ALL = 0xFFFFFFFF;

  /** Key-Konstante um die Aktion "Info" anzusprechen.
   *  @see #resetCaptions(Map) */
  public static final String INFO = MapActionControlPane.class.getName()+".INFO";
  /** Key-Konstante um die Aktion "Zoom" anzusprechen.
   *  @see #resetCaptions(Map) */
  public static final String ZOOM_IN = MapActionControlPane.class.getName()+".ZOOM_IN";
  /** Key-Konstante um die Aktion "SelectAll" anzusprechen.
   *  @see #resetCaptions(Map) */
  public static final String SELECT_ALL = MapActionControlPane.class.getName()+".SELECT_ALL";
  /** Key-Konstante um die Aktion "SelectTop" anzusprechen.
   *  @see #resetCaptions(Map) */
  public static final String SELECT_TOP = MapActionControlPane.class.getName()+".SELECT_TOP";


  /** {@link SelectableXMapPane} das gesteuert wird. */
  protected SelectableXMapPane mapPane = null;
  /** Button fuer Info-Aktion. */
  protected JToggleButton infoState = null;
  /** Button fuer Zoom-Aktion. */
  protected JToggleButton zoomState = null;
  /** Button fuer SelectTop-Aktion. */
  protected JToggleButton selectTopState = null;
  /** Button fuer SelectAll-Aktion. */
  protected JToggleButton selectAllState = null;
  /** Maske, die die zur Verfuegung stehenden Aktionen codiert */
  protected int actionMask = 0;

  /**
   * Erzeugt eine horizontale Steuer-Komponente, die (noch) keinem {@link SelectableXMapPane}
   * zugeordnet ist.
   */
  public MapActionControlPane() {
    this( null, HORIZONTAL );
  }

  /**
   * Erzeugt eine horizontale Steuer-Komponente.
   * @param mapPane {@link SelectableXMapPane} das gesteuert wird
   */
  public MapActionControlPane(SelectableXMapPane mapPane) {
    this( mapPane, HORIZONTAL );
  }

  /**
   * Erzeugt eine Steuer-Komponente.
   * @param mapPane {@link SelectableXMapPane} das gesteuert wird
   * @param orientation Orientierung der Komponente ({@link #HORIZONTAL}/{@link #VERTICAL})
   * @param actionMask definiert, welche Aktionen (in Form von Buttons) angezeigt werden
   *                   (OR-Verknuepfung der {@code ACTION}-Konstanten
   */
  public MapActionControlPane(SelectableXMapPane mapPane, int orientation, int actionMask) {
    super(orientation);
    this.actionMask = actionMask;
    // Button erzeugen und in Gruppe einfuegen, damit immer nur
    // eine aktiviert ist
    infoState      = new JToggleButton( new Action_InfoState() );
    zoomState      = new JToggleButton( new Action_ZoomState() );
    selectTopState = new JToggleButton( new Action_SelectOnTopLayerState() );
    selectAllState = new JToggleButton( new Action_SelectOnAllLayerState() );
    ButtonGroup bGroup = new ButtonGroup();
    bGroup.add(infoState);
    bGroup.add(zoomState);
    bGroup.add(selectTopState);
    bGroup.add(selectAllState);
    // Button dem ToolBar hinzufuegen
    if ( isActionVisible(ACTION_INFO) )
      this.add(infoState);
    if ( isActionVisible(ACTION_ZOOM_IN) )
      this.add(zoomState);
    if ( isActionVisible(ACTION_SELECT_TOP) )
      this.add(selectTopState);
    if ( isActionVisible(ACTION_SELECT_ALL) )
      this.add(selectAllState);

    setMapPane(mapPane);
  }

  /**
   * Erzeugt eine Steuer-Komponente. Alle Aktionen sind sichtbar.
   * @param mapPane {@link SelectableXMapPane} das gesteuert wird
   * @param orientation Orientierung der Komponente ({@link #HORIZONTAL}/{@link #VERTICAL})
   */
  public MapActionControlPane(SelectableXMapPane mapPane, int orientation) {
    this(mapPane,orientation,ACTION_ALL);
  }

	/**
	 * Setzt die Aktivierung der Aktionen entsprechend den Einstellungen des
	 * {@code JMapPane}.
	 * 
	 * @see SelectableXMapPane#getWindowSelectionState()
	 */
	public void resetActions() {
		for (int i = 0; i < getComponentCount(); i++) {
			Component comp = getComponent(i);
			if (comp instanceof JButton)
				((JButton) comp).setSelected(false);
		}

		// Buttons entsprechend dem MapPane-Status einstellen
		if (mapPane != null) {
			if (mapPane.getTool() == XMapPaneTool.ZOOM_IN
					|| mapPane.getTool() == XMapPaneTool.ZOOM_OUT) {
				if (isActionVisible(ACTION_ZOOM_IN))
					zoomState.doClick();
			} else if (mapPane.getTool() == XMapPaneTool.SELECTION_TOP_LAYER) {
				if (isActionVisible(ACTION_SELECT_TOP))
					selectTopState.doClick();
			} else if (mapPane.getTool() == XMapPaneTool.SELECTION_ALL_LAYERS) {

				if (isActionVisible(ACTION_SELECT_ALL))
					selectAllState.doClick();
			} else if (mapPane.getTool() == XMapPaneTool.NO_ACTION)
				if (isActionVisible(ACTION_INFO))
					infoState.doClick();
		}
	}

  /**
   * Prueft, ob eine Aktion (Button) zur Verfuegung steht.
   * @param action Aktion codiert durch eine {@code ACTION}-Konstante
   */
  public boolean isActionVisible(int action) {
    return (actionMask & action) > 0;
  }

  /**
   * Setzt das {@link SelectableXMapPane}, das durch diese Komponente gesteuert wird.
   */
  public void setMapPane(SelectableXMapPane mapPane) {
    this.mapPane = mapPane;
    resetActions();
  }

  /**
   * Liefert das {@link SelectableXMapPane}, das durch diese Komponente gesteuert wird.
   */
  public SelectableXMapPane getMapPane() {
    return this.mapPane;
  }

  /**
   * Belegt die Beschriftungen der Aktionen neu.
   * @param newCaptions enhaelt die neuen Beschriftungen
   * @see #INFO
   * @see #ZOOM_IN
   * @see #SELECT_TOP
   * @see #SELECT_ALL
   */
  public void resetCaptions(Map<String,Object> newCaptions) {
    if ( newCaptions == null )
      return;
    Object caption = newCaptions.get(INFO);
    if ( caption != null )
      infoState.getAction().putValue(AbstractAction.SHORT_DESCRIPTION,caption);
    caption = newCaptions.get(ZOOM_IN);
    if ( caption != null )
      zoomState.getAction().putValue(AbstractAction.SHORT_DESCRIPTION,caption);
    caption = newCaptions.get(SELECT_TOP);
    if ( caption != null )
      selectTopState.getAction().putValue(AbstractAction.SHORT_DESCRIPTION,caption);
    caption = newCaptions.get(SELECT_ALL);
    if ( caption != null )
      selectAllState.getAction().putValue(AbstractAction.SHORT_DESCRIPTION,caption);
  }

  ///////////////////////////////////////////////////////////////////
  /////////////////////   Button Aktionen   /////////////////////////
  ///////////////////////////////////////////////////////////////////
  /**
   * Aktion "Info" des {@link MapActionControlPane}.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  protected class Action_InfoState extends AbstractAction {
    /**
     * Erzeugt einen neue Info-Aktion
     */
    public Action_InfoState() {
      super();
      Icon icon = SwingUtil.createImageIconFromResourcePath("resource/icons/small/info.gif","Info");
      this.putValue( SHORT_DESCRIPTION, GeotoolsGUIUtil.RESOURCE.getString( INFO ) );
      this.putValue( SMALL_ICON, icon );
      if ( icon == null )
        this.putValue( NAME, "Info" );
    }

    /**
     * Fuehrt die Info-Aktion aus.
     * <ol>
     *   <li>{@code JMapPane.setWindowSelectionState( JMapPane.NONE )}</li>
     *   <li>{@code JMapPane.setState( JMapPane.Select )}</li>
     *   <li>{@code JMapPane.setHighlight( true )}</li>
     * </ol>
     * @see SelectableXMapPane
     * @param e das ActionEvent
     */
    public void actionPerformed(ActionEvent e) {
      getMapPane().setTool( XMapPaneTool.NO_ACTION );
    }
  }

  /**
   * Aktion "Zoom" des {@link MapActionControlPane}.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  protected class Action_ZoomState extends AbstractAction {

    /**
     * Erzeugt einen neue Zoom-Aktion
     */
    public Action_ZoomState() {
      super();
      Icon icon = SwingUtil.createImageIconFromResourcePath("resource/icons/small/zoom.gif","ZoomIn");
      this.putValue( SHORT_DESCRIPTION, GeotoolsGUIUtil.RESOURCE.getString( ZOOM_IN ) );
      this.putValue( SMALL_ICON, icon );
      if ( icon == null )
        this.putValue( NAME, "ZoomIn" );
    }

    /**
     * Fuehrt die Zoom-Aktion aus.
     * <ol>
     *   <li>{@code JMapPane.setWindowSelectionState( JMapPane.ZOOM )}</li>
     *   <li>{@code JMapPane.setState( JMapPane.ZoomIn )}</li>
     *   <li>{@code JMapPane.setHighlight( false )}</li>
     * </ol>
     * @see SelectableXMapPane
     * @param e das ActionEvent
     */
    public void actionPerformed(ActionEvent e) {
      getMapPane().setTool( XMapPaneTool.ZOOM_IN );
    }
  }


  /**
   * Aktion "SelectTop" des {@link MapActionControlPane}.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  protected class Action_SelectOnTopLayerState extends AbstractAction {
    /**
     * Erzeugt einen neue SelectTop-Aktion
     */
    public Action_SelectOnTopLayerState() {
      super();
      Icon icon = SwingUtil.createImageIconFromResourcePath("resource/icons/small/select_normal.gif","SelectTop");
      this.putValue( SHORT_DESCRIPTION, GeotoolsGUIUtil.RESOURCE.getString( SELECT_TOP ) );
      this.putValue( SMALL_ICON, icon );
      if ( icon == null )
        this.putValue( NAME, "SelectTop" );
    }

    /**
     * Fuehrt die SelectTop-Aktion aus.
     * <ol>
     *   <li>{@code JMapPane.setWindowSelectionState( JMapPane.SELECT_TOP )}</li>
     *   <li>{@code JMapPane.setState( JMapPane.Select )}</li>
     *   <li>{@code JMapPane.setHighlight( true )}</li>
     * </ol>
     * @see SelectableXMapPane
     * @param e das ActionEvent
     */
    public void actionPerformed(ActionEvent e) {
      getMapPane().setTool( XMapPaneTool.SELECTION_TOP_LAYER);
    }
  }

  /**
   * Aktion "SelectAll" des {@link MapActionControlPane}.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  protected class Action_SelectOnAllLayerState extends AbstractAction {
    /**
     * Erzeugt einen neue SelectAll-Aktion
     */
    public Action_SelectOnAllLayerState() {
      super();
      Icon icon = SwingUtil.createImageIconFromResourcePath("resource/icons/small/select_multi.gif","SelectAll");
      this.putValue( SHORT_DESCRIPTION, GeotoolsGUIUtil.RESOURCE.getString( SELECT_ALL ) );
      this.putValue( SMALL_ICON, icon );
      if ( icon == null )
        this.putValue( NAME, "SelectAll" );
    }

    /**
     * Fuehrt die SelectAll-Aktion aus.
     * <ol>
     *   <li>{@code JMapPane.setWindowSelectionState( JMapPane.SELECT_ALL )}</li>
     *   <li>{@code JMapPane.setState( JMapPane.Select )}</li>
     *   <li>{@code JMapPane.setHighlight( true )}</li>
     * </ol>
     * @see SelectableXMapPane
     * @param e das ActionEvent
     */
    public void actionPerformed(ActionEvent e) {
      getMapPane().setTool( XMapPaneTool.SELECTION_ALL_LAYERS);
    }
  }

}
