/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import java.awt.Graphics;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Iterator;
import java.util.Map;

import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.JDialog;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.event.MenuEvent;
import javax.swing.event.MenuListener;

import org.apache.log4j.Logger;
import org.geotools.map.MapContext;
import org.geotools.map.MapLayer;
import org.geotools.map.event.MapLayerEvent;
import org.geotools.map.event.MapLayerListEvent;
import org.geotools.map.event.MapLayerListListener;
import org.geotools.map.event.MapLayerListener;
import org.geotools.styling.ColorMap;

import schmitzm.geotools.grid.GridUtil;
import schmitzm.geotools.map.event.MapLayerAdapter;
import schmitzm.geotools.styling.ColorMapManager;
import schmitzm.geotools.styling.StylingUtil;
import schmitzm.lang.LangUtil;
import schmitzm.swing.ButtonGroup;
import schmitzm.swing.CaptionsChangeable;
import schmitzm.swing.SwingUtil;
import schmitzm.swing.event.WindowEventConnector;
import schmitzm.swing.menu.ConnectedPopupMenu;

/**
 * Diese Komponente ist an ein {@link SelectableXMapPane} gekoppelt und stellt die
 * dargestellten Layer in Form eine Liste dar.
 * 
 * @author Martin Schmitz
 * @version 1.0
 */
public class MapContextControlPane extends JPanel {
	protected final Logger LOGGER = LangUtil.createLogger(this);

	/**
	 * Key, um den Menue-Eintrag "Move layer up" in der
	 * {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String MENU_MOVE_LAYER_UP = MapContextControlPane.class
			.getName()
			+ ".Menu.MOVE_UP";
	/**
	 * Key, um den Menue-Eintrag "Move layer down" in der
	 * {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String MENU_MOVE_LAYER_DOWN = MapContextControlPane.class
			.getName()
			+ ".Menu.MOVE_DOWN";
	/**
	 * Key, um den Menue-Eintrag "Zoom to layer" in der
	 * {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String MENU_ZOOM_TO_LAYER = MapContextControlPane.class
			.getName()
			+ ".Menu.ZOOM_TO";
	/**
	 * Key, um den Menue-Eintrag "SimpleFeature-Filter..." in der
	 * {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String MENU_FILTER_LAYER = MapContextControlPane.class
			.getName()
			+ ".Menu.FILTER";
	/**
	 * Key, um den Menue-Eintrag "Recolor..." in der {@link CaptionsChangeable}
	 * -Map anzusprechen.
	 */
	public static final String MENU_RECOLOR_LAYER = MapContextControlPane.class
			.getName()
			+ ".Menu.RECOLOR";
	/**
	 * Key, um den Menue-Eintrag "Remove layer" in der
	 * {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String MENU_REMOVE_LAYER = MapContextControlPane.class
			.getName()
			+ ".Menu.REMOVE";
	/**
	 * Key, um den Menue-Eintrag "Show all layers" in der
	 * {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String MENU_SHOW_ALL_LAYERS = MapContextControlPane.class
			.getName()
			+ ".Menu.SHOWALL";
	/**
	 * Key, um den Menue-Eintrag "Hide all layers" in der
	 * {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String MENU_HIDE_ALL_LAYERS = MapContextControlPane.class
			.getName()
			+ ".Menu.HIDEALL";
	/**
	 * Key, um den Menue-Eintrag "Invert all layers" in der
	 * {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String MENU_INVERT_ALL_LAYERS = MapContextControlPane.class
			.getName()
			+ ".Menu.INVERTALL";
	/**
	 * Key, um den Unter-Menue-Eintrag "Customize color map" in der
	 * {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String MENU_CUSTOMIZE_COLOR = MapContextControlPane.class
			.getName()
			+ ".Menu.CUSTOMIZE";
	/**
	 * Key, um den Titel des Farbpaletten-Dialogs in der
	 * {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String CM_DIALOG_TITLE = MapContextControlPane.class
			.getName()
			+ ".ColorMapDialog.TITLE";
	/**
	 * Key, um den 1. Tabellenkopf-Eintrag "Quantity" des Farbpaletten-Dialogs
	 * in der {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String CM_DIALOG_TABLEHEADER_QUANTITY = ColorMapTable.TABLEHEADER_QUANTITY;
	/**
	 * Key, um den 2. Tabellenkopf-Eintrag "Color" des Farbpaletten-Dialogs in
	 * der {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String CM_DIALOG_TABLEHEADER_COLOR = ColorMapTable.TABLEHEADER_COLOR;
	/**
	 * Key, um den 3. Tabellenkopf-Eintrag "Label" des Farbpaletten-Dialogs in
	 * der {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String CM_DIALOG_TABLEHEADER_LABEL = ColorMapTable.TABLEHEADER_LABEL;
	/**
	 * Key, um den OK-Button des Farbpaletten-Dialogs in der
	 * {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String CM_DIALOG_OK = MapContextControlPane.class
			.getName()
			+ ".ColorMapDialog.OK";
	/**
	 * Key, um den ABBRECHEN-Button des Farbpaletten-Dialogs in der
	 * {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String CM_DIALOG_CANCEL = MapContextControlPane.class
			.getName()
			+ ".ColorMapDialog.CANCEL";
	/**
	 * Key, um den SPEICHERN-Button des Farbpaletten-Dialogs in der
	 * {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String CM_DIALOG_SAVE = MapContextControlPane.class
			.getName()
			+ ".ColorMapDialog.SAVE";
	/**
	 * Key, um den ÜBERNEHMEN-Button des Farbpaletten-Dialogs in der
	 * {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String CM_DIALOG_APPLY = MapContextControlPane.class
			.getName()
			+ ".ColorMapDialog.APPLY";
	/**
	 * Key, um den Titel des Farbpaletten-Speichern-Dialogs in der
	 * {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String CM_SAVE_DIALOG_TITLE = MapContextControlPane.class
			.getName()
			+ ".SaveColorMapDialog.TITLE";
	/**
	 * Key, um den Text des Farbpaletten-Speichern-Dialogs in der
	 * {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String CM_SAVE_DIALOG_QUESTION = MapContextControlPane.class
			.getName()
			+ ".SaveColorMapDialog.QUESTION";
	/**
	 * Key, um den Titel des FeatureCollection-Filter-Dialogs in der
	 * {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String FF_DIALOG_TITLE = FeatureLayerFilterDialog.DIALOG_TITLE;
	/**
	 * Key, um das Attribute-Label des FeatureCollection-Filter-Dialogs in der
	 * {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String FF_DIALOG_ATTRIBUTE = FeatureCollectionFilterPanel.ATTRIBUTE_LABEL;
	/**
	 * Key, um das Formel-Label des FeatureCollection-Filter-Dialogs in der
	 * {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String FF_DIALOG_RULE = FeatureCollectionFilterPanel.RULE_LABEL;
	/**
	 * Key, um den Formel-Feld-Tooltip des FeatureCollection-Filter-Dialogs in
	 * der {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String FF_DIALOG_RULE_TOOLTIP = FeatureCollectionFilterPanel.RULE_TOOLTIP;
	/**
	 * Key, um das Operatoren-Label des FeatureCollection-Filter-Dialogs in der
	 * {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String FF_DIALOG_OPERATORS = FeatureCollectionFilterPanel.OPERATOR_LABEL;
	/**
	 * Key, um den Testen-Button des FeatureCollection-Filter-Dialogs in der
	 * {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String FF_DIALOG_TEST = FeatureCollectionFilterPanel.TEST_BUTTON;
	/**
	 * Key, um den OK-Button des FeatureCollection-Filter-Dialogs in der
	 * {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String FF_DIALOG_OK = FeatureLayerFilterDialog.OK_BUTTON;
	/**
	 * Key, um den ABBRECHEN-Button des FeatureCollection-Filter-Dialogs in der
	 * {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String FF_DIALOG_CANCEL = FeatureLayerFilterDialog.CANCEL_BUTTON;
	/**
	 * Key, um den ÜBERNEHMEN-Button des FeatureCollection-Filter-Dialogs in der
	 * {@link CaptionsChangeable}-Map anzusprechen.
	 */
	public static final String FF_DIALOG_APPLY = FeatureLayerFilterDialog.APPLY_BUTTON;

	/** Karte, deren Layer kontrolliert werden. */
	protected SelectableXMapPane mapPane = null;
	/** Farbpaletten, die durch die Kontrolle zugewiesen werden koennen. */
	private ColorMapManager colorMaps = null;

	/**
	 * Erzeugt eine neue Layer-Kontrolle.
	 * 
	 * @param mapPane
	 *            Karten-Komponente, deren Layer kontrolliert wird
	 * @param colorMaps
	 *            Raster-Farbpaletten, die durch die Kontrolle zugewiesen werden
	 *            koennen
	 */
	public MapContextControlPane(SelectableXMapPane mapPane, ColorMapManager colorMaps) {
		super();
		// this.parentFrame = SwingUtil.getParentWindow(this);
		this.mapPane = mapPane;
		this.colorMaps = colorMaps;

		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));
		// Auf den MapContext lauschen und Steuerungskomponenten hinzufuegen,
		// loeschen, verschieben
		mapPane.getMapContext().addMapLayerListListener(
				new MapLayerListListener() {
					public void layerAdded(MapLayerListEvent e) {
						addLayerControl(getComponentCount() - e.getToIndex(), e
								.getLayer());
					}

					public void layerChanged(MapLayerListEvent e) {
						// Bezeichnung der Checkbox aendern
						((JCheckBox) getComponent(convertIndex(e.getFromIndex())))
								.setText(e.getLayer().getTitle());
					}

					public void layerMoved(MapLayerListEvent e) {
						int from = convertIndex(e.getFromIndex());
						int to = convertIndex(e.getToIndex());
						moveLayerControl(from, to);
					}

					public void layerRemoved(MapLayerListEvent e) {
						int from = convertIndex(e.getFromIndex());
						removeLayerControl(from);
					}
				});
	}

	/**
	 * Liefert die Farbpaletten, die durch die Kontroll-Komponente zugewiesen
	 * werden koennen.
	 */
	public ColorMapManager getColorMapManager() {
		return this.colorMaps;
	}

	/**
	 * Konvertiert einen Listen-Index des MapContext zu einem Index der
	 * MapContextControlList (oder layerObjects-Liste). Dies ist notwendig, da
	 * im MapContext der Index 0 fuer das unterste Layer steht, in der
	 * MapContextControlList (und layerObjects-Liste) jedoch fuer das oberste
	 * Layer!
	 */
	private int convertIndex(int mapContextIdx) {
		return getComponentCount() - 1 - mapContextIdx;
	}

	/**
	 * Fuegt eine neue Checkbox in die Liste der Kontroll-Komponenten ein
	 * 
	 * @param idx
	 *            Stelle an der die Komponente eingefuegt wird (0 ist das
	 *            oberste Layer)
	 * @param layer
	 *            das neue Layer
	 */
	private void addLayerControl(int idx, final MapLayer layer) {
		MapLayerControl control = new MapLayerControl(layer);
		add(control, idx);
		validate();
	}

	/**
	 * Bewegt eine Kontroll-Komponente
	 * 
	 * @param from
	 *            Index der zu verschiebenden Komponente (0 ist die oberste)
	 * @param to
	 *            Index der Ziel-Position
	 */
	private void moveLayerControl(int from, int to) {
		JCheckBox cb = (JCheckBox) getComponent(from);
		remove(cb);
		add(cb, to);
		validate();
	}

	/**
	 * Entfernt eine Kontroll-Komponente
	 * 
	 * @param idx
	 *            Index der zu loeschenden Komponente (0 ist die oberste)
	 */
	private void removeLayerControl(int idx) {
		remove(idx);
		validate();
		repaint();
	}

	// ////////////////////////////////////////////////////////////////////////
	// Kontext-Menue fuer Layer-Aktionen
	// ////////////////////////////////////////////////////////////////////////
	private class MapLayerControl extends JCheckBox implements
			CaptionsChangeable, ActionListener {
		private MapLayer layer = null;
		// Flag, ob bereits ein WindowListener im uebergeordneten Fenster
		// eingetragen wurde (geschieht beim ersten Anzeigen)
		private boolean windowListenerAdded = false;
		// Menu-Items
		private JMenuItem moveLayerUp = null;
		private JMenuItem moveLayerDown = null;
		private JMenuItem zoomToLayer = null;
		private JMenuItem filterLayer = null;
		private ColorMapMenu recolorLayer = null;
		private JMenuItem removeLayer = null;
		private JMenuItem showAllLayers = null;
		private JMenuItem hideAllLayers = null;
		private JMenuItem invertAllLayers = null;
		private FeatureLayerFilterDialog filterDialog = null;

		public MapLayerControl(final MapLayer layer) {
			super(layer.getTitle(), layer.isVisible());
			this.layer = layer;
			this.addActionListener(this); // (un)check Checkbox

			// Popup-Menue erzeugen, die die Funktionen des Layers steuert
			final ConnectedPopupMenu menu = new ConnectedPopupMenu();
			moveLayerUp = new JMenuItem(GeotoolsGUIUtil.RESOURCE
					.getString(MENU_MOVE_LAYER_UP));
			moveLayerDown = new JMenuItem(GeotoolsGUIUtil.RESOURCE
					.getString(MENU_MOVE_LAYER_DOWN));
			zoomToLayer = new JMenuItem(GeotoolsGUIUtil.RESOURCE
					.getString(MENU_ZOOM_TO_LAYER));
			filterLayer = new JMenuItem(GeotoolsGUIUtil.RESOURCE
					.getString(MENU_FILTER_LAYER));
			recolorLayer = new ColorMapMenu(GeotoolsGUIUtil.RESOURCE
					.getString(MENU_RECOLOR_LAYER), layer);
			removeLayer = new JMenuItem(GeotoolsGUIUtil.RESOURCE
					.getString(MENU_REMOVE_LAYER));
			showAllLayers = new JMenuItem(GeotoolsGUIUtil.RESOURCE
					.getString(MENU_SHOW_ALL_LAYERS));
			hideAllLayers = new JMenuItem(GeotoolsGUIUtil.RESOURCE
					.getString(MENU_HIDE_ALL_LAYERS));
			invertAllLayers = new JMenuItem(GeotoolsGUIUtil.RESOURCE
					.getString(MENU_INVERT_ALL_LAYERS));
			moveLayerDown.addActionListener(this);
			moveLayerUp.addActionListener(this);
			zoomToLayer.addActionListener(this);
			removeLayer.addActionListener(this);
			showAllLayers.addActionListener(this);
			hideAllLayers.addActionListener(this);
			invertAllLayers.addActionListener(this);
			filterLayer.addActionListener(this);
			filterLayer.setEnabled(!SelectableXMapPane.isGridCoverageLayer(layer));
			if (filterLayer.isEnabled())
				try {
					filterDialog = new FeatureLayerFilterDialog(null, mapPane,
							layer, true);
					filterDialog.setModal(false);
				} catch (Exception err) {
					LangUtil
							.logDebugError(
									LOGGER,
									"Layer "
											+ layer.getTitle()
											+ ": FeatureFilterFrame could not be created!",
									err);
					filterLayer.setEnabled(false);
				}
			menu.add(moveLayerUp);
			menu.add(moveLayerDown);
			menu.add(zoomToLayer);
			menu.add(filterLayer);
			menu.add(recolorLayer);
			menu.addSeparator();
			menu.add(removeLayer);
			menu.addSeparator();
			menu.add(showAllLayers);
			menu.add(hideAllLayers);
			menu.add(invertAllLayers);

			menu.connectTo(this);

			layer.addMapLayerListener(new MapLayerListener() {
				@Override
				public void layerChanged(MapLayerEvent e) {
				}

				@Override
				public void layerHidden(MapLayerEvent e) {
					setSelected(false);
				}

				@Override
				public void layerShown(MapLayerEvent e) {
					setSelected(true);
				}

				@Override
				public void layerDeselected(MapLayerEvent event) {
					// TODO Auto-generated method stub
				}

				@Override
				public void layerSelected(MapLayerEvent event) {
					// TODO Auto-generated method stub
				}
			});
		}

		/**
		 * Ruft die {@code super}-Methode auf. Zudem wird (beim ersten Anzeigen)
		 * ein {@link WindowEventConnector} in das uebergeordnete Fenster
		 * eingetragen, der dessen Aktionen mit denen des
		 * SimpleFeature-Filter-Fensters verbindet.
		 * 
		 * @param g
		 *            Graphics
		 */
		public void paint(Graphics g) {
			super.paint(g);
			// Wenn das Fenster geschlossen/minimiert wird, soll auch das
			// SimpleFeature-Filter-Fenster geschlossen/minimiert werden
			if (!windowListenerAdded && filterDialog != null) {
				Window parentFrame = SwingUtil.getParentWindow(this);
				if (parentFrame != null) {
					parentFrame.addWindowListener(new WindowEventConnector(
							filterDialog));
					windowListenerAdded = true;
				}
			}
		}

		/**
		 * Initiiert die Checkbox- und Menu-Aktionen, je nachdem, welche Quelle
		 * das ActionEvent hat.
		 * 
		 * @param e
		 *            ActionEvent
		 */
		public void actionPerformed(ActionEvent e) {
			if (e == null)
				return;
			Object source = e.getSource();
			MapContext context = mapPane.getMapContext();
			int currIdx = context.indexOf(layer);

			// Checkbox (un)checked -> (in)visible layer
			if (source == this)
				layer.setVisible(!layer.isVisible());
			// Menu "Move layer up"
			if (source == moveLayerUp && currIdx + 1 < context.getLayerCount())
				context.moveLayer(currIdx, currIdx + 1);
			// Menu "Move layer down"
			if (source == moveLayerDown && currIdx > 0)
				context.moveLayer(currIdx, currIdx - 1);
			// Menu "Zoom to layer"
			if (source == zoomToLayer)
				mapPane.zoomToLayer(layer);
			// Menu "Remove layer"
			if (source == removeLayer)
				context.removeLayer(layer);
			// Menu "Show all layer"
			if (source == showAllLayers)
				for (MapLayer layer : context.getLayers())
					layer.setVisible(true);
			// Menu "Hide all layer"
			if (source == hideAllLayers)
				for (MapLayer layer : context.getLayers())
					layer.setVisible(false);
			// Menu "Invert all layers"
			if (source == invertAllLayers)
				for (MapLayer layer : context.getLayers())
					layer.setVisible(!layer.isVisible());
			// Menu "Filter layer"
			if (source == filterLayer) {
				filterDialog.toFront();
				filterDialog.setVisible(true);
			}
			mapPane.refresh();
		}

		/**
		 * Belegt die Bezeichnungen der Menue-Eintraege neu.
		 * 
		 * @param captionMap
		 *            Map mit neuen Beschriftungen
		 */
		public void resetCaptions(Map<String, Object> captionMap) {
			SwingUtil.resetCaption(moveLayerUp, captionMap
					.get(MENU_MOVE_LAYER_UP));
			SwingUtil.resetCaption(moveLayerDown, captionMap
					.get(MENU_MOVE_LAYER_DOWN));
			SwingUtil.resetCaption(zoomToLayer, captionMap
					.get(MENU_ZOOM_TO_LAYER));
			SwingUtil.resetCaption(recolorLayer, captionMap
					.get(MENU_RECOLOR_LAYER));
			SwingUtil.resetCaption(removeLayer, captionMap
					.get(MENU_REMOVE_LAYER));
			SwingUtil.resetCaption(showAllLayers, captionMap
					.get(MENU_SHOW_ALL_LAYERS));
			SwingUtil.resetCaption(hideAllLayers, captionMap
					.get(MENU_HIDE_ALL_LAYERS));
			SwingUtil.resetCaption(invertAllLayers, captionMap
					.get(MENU_INVERT_ALL_LAYERS));
			SwingUtil.resetCaption(filterLayer, captionMap
					.get(MENU_FILTER_LAYER));
			recolorLayer.resetCaptions(captionMap);
			if (filterDialog != null)
				filterDialog.resetCaptions(captionMap);
		}
	}

	// ////////////////////////////////////////////////////////////////////////
	// Menue fuer Farb-Paletten
	// ////////////////////////////////////////////////////////////////////////
	private class ColorMapMenu extends JMenu implements CaptionsChangeable {
		private MapLayer layer = null;
		private ButtonGroup menuItemGroup = null;
		// Flag, ob bereits ein WindowListener im uebergeordneten Fenster
		// eingetragen wurde (geschieht beim ersten Anzeigen)
		private boolean windowListenerAdded = false;
		// statischer Menue-Eintrag zum Anpassen der Farbtabelle
		private JMenuItem customizeMenuItem = null;
		// Button fuer Farbtabellen-Dialog (Instanzvariablen, damit
		// Beschriftung geaendert werden kann!!)
		private JButton okButton = null;
		private JButton cancelButton = null;
		private JButton applyButton = null;
		private JButton saveButton = null;
		// Farbtabellen-Dialog
//		private ColorMapTable customiseDialogTable = new ColorMapTable();
		private ColorMapPanel customiseDialogPanel = new ColorMapPanel();
		private JDialog customiseDialog = null;
		private String cmDialogTitle = GeotoolsGUIUtil.RESOURCE
				.getString(CM_DIALOG_TITLE);
		private String saveCMDialogTitle = GeotoolsGUIUtil.RESOURCE
				.getString(CM_SAVE_DIALOG_TITLE);
		private String saveCMDialogQuestion = GeotoolsGUIUtil.RESOURCE
				.getString(CM_SAVE_DIALOG_QUESTION);

		/**
		 * Aktion, wenn eine der registrierten Farb-Paletten ausgewaehlt wird.
		 */
		private ActionListener colorMapAction = new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				String colMapName = ((JCheckBoxMenuItem) e.getSource())
						.getText();
				ColorMap colMap = (colMapName != null) ? StylingUtil
						.cloneColorMap(colorMaps.get(colMapName)) : null;
				if (colMap != null) {
					layer.setStyle(GridUtil.createStyle(colMap, 1.0));
					customiseDialogPanel.setColorMap(colMap);
					mapPane.repaint();
					((JCheckBoxMenuItem) e.getSource()).setSelected(true);
				}
			}
		};

		/**
		 * Aktion, um die aktuelle Farb-Paletten anzupassen.
		 */
		private ActionListener customiseMapAction = new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				if (customiseDialog.isVisible()) {
					customiseDialog.toFront();
					return;
				}
				customiseDialogPanel.setColorMap(StylingUtil
						.cloneColorMap(StylingUtil.getColorMapFromStyle(layer
								.getStyle())));
				customiseDialog.setVisible(true);
			}
		};

		/**
		 * Erzeugt ein neues Menue.
		 * 
		 * @param desc
		 *            Beschreibung fuer das Menue
		 * @param layer
		 *            Layer, das durch das Menue gesteuert wird
		 */
		public ColorMapMenu(String desc, MapLayer layer) {
			super(desc);
			this.layer = layer;
			this.customiseDialog = createColorMapDialog();
			customizeMenuItem = new JMenuItem(GeotoolsGUIUtil.RESOURCE
					.getString(MENU_CUSTOMIZE_COLOR));
			customizeMenuItem.addActionListener(customiseMapAction);
			// Initialer Aufbau
			reorganise();
			// Sobald das Menue neu angezeigt wird, muss es neu aufgebaut
			// werden, wenn sich die vorgefertigten ColorMaps geaendert haben
			this.addMenuListener(new MenuListener() {
				private ColorMapManager oldColorMaps = (ColorMapManager) colorMaps
						.clone();

				public void menuCanceled(MenuEvent e) {
				}

				public void menuDeselected(MenuEvent e) {
				}

				public void menuSelected(MenuEvent e) {
					if (!colorMaps.equals(oldColorMaps)) {
						reorganise();
						oldColorMaps = (ColorMapManager) colorMaps.clone();
					}
				}
			});
		}

		/**
		 * Baut das Menue neu auf.
		 */
		public void reorganise() {
			// Alle bisherigen Eintraege loeschen
			this.removeAll();

			// nur fuer GridCoverage2D-Objekte kann die Farbe veraendert werden
			// FeatureTypeStyle[] fts = layer.getStyle().getFeatureTypeStyles();
			// if ( !(fts[0].getRules()[0].getSymbolizers()[0] instanceof
			// RasterSymbolizer) ) {
			if (!SelectableXMapPane.isGridCoverageLayer(layer)) {
				this.setEnabled(false);
				return;
			}

			this.setEnabled(true);
			this.menuItemGroup = new ButtonGroup();

			// Einen Menue-Eintrag fuer jede registrierte Farb-Palette
			Iterator<String> keys = colorMaps.keySet().iterator();
			for (int i = 0; keys.hasNext(); i++) {
				String colMapName = keys.next();
				JCheckBoxMenuItem item = new JCheckBoxMenuItem(colMapName);
				item.addActionListener(colorMapAction);
				menuItemGroup.add(item);
				// Eintrag selektieren, wenn er vorher selektiert war
				item.setSelected(StylingUtil.colorMapsEqual(colorMaps
						.get(colMapName), StylingUtil
						.getColorMapFromStyle(layer.getStyle())));
				this.add(item);

			}

			// Einen Menue-Eintrag, um die Farb-Palette anzupassen
			if (this.getItemCount() > 0)
				this.addSeparator();
			this.add(customizeMenuItem);
		}

		public void resetCaptions(Map<String, Object> captionMap) {
			SwingUtil.resetCaption(customizeMenuItem, captionMap
					.get(MENU_CUSTOMIZE_COLOR));
			SwingUtil.resetCaption(okButton, captionMap.get(CM_DIALOG_OK));
			SwingUtil.resetCaption(cancelButton, captionMap
					.get(CM_DIALOG_CANCEL));
			SwingUtil.resetCaption(saveButton, captionMap.get(CM_DIALOG_SAVE));
			SwingUtil
					.resetCaption(applyButton, captionMap.get(CM_DIALOG_APPLY));
			SwingUtil.resetCaption(customiseDialog, captionMap
					.get(CM_DIALOG_TITLE));

			Object caption = captionMap.get(CM_DIALOG_TITLE);
			if (caption != null) {
				cmDialogTitle = caption.toString();
				// Damit Layer-Bezeichnung in Titel uebernommen wird, ein
				// MapLayerEvent
				// ausloesen
				layer.setTitle(layer.getTitle());
			}
			caption = captionMap.get(CM_SAVE_DIALOG_TITLE);
			if (caption != null)
				saveCMDialogTitle = caption.toString();
			caption = captionMap.get(CM_SAVE_DIALOG_QUESTION);
			if (caption != null)
				saveCMDialogQuestion = caption.toString();
		}

		/**
		 * Ruft die {@code super}-Methode auf. Zudem wird (beim ersten Anzeigen)
		 * ein {@link WindowEventConnector} in das uebergeordnete Fenster
		 * eingetragen, der dessen Aktionen mit denen des Farbpaletten-Fensters
		 * verbindet.
		 * 
		 * @param g
		 *            Graphics
		 */
		public void paint(Graphics g) {
			super.paint(g);
			// Wenn das Fenster geschlossen/minimiert wird, soll auch das
			// Farbpaletten-Fenster geschlossen/minimiert werden
			if (!windowListenerAdded) {
				Window parentFrame = SwingUtil.getParentWindow(this);
				if (parentFrame != null) {
					parentFrame.addWindowListener(new WindowEventConnector(
							customiseDialog));
					windowListenerAdded = true;
				}
			}
		}

		/**
		 * Erzeugt einen Dialog, in dem die aktuelle Farb-Palette des (Raster-)
		 * Layers angezeigt wird und veraendert werden kann.
		 */
		private JDialog createColorMapDialog() {
			// Optionen des Dialogs (ActionListener folgen spaeter)
			okButton = new JButton(SwingUtil.RESOURCE.getString("Ok"));
			cancelButton = new JButton(SwingUtil.RESOURCE.getString("Cancel"));
			applyButton = new JButton(SwingUtil.RESOURCE.getString("Apply"));
			saveButton = new JButton(SwingUtil.RESOURCE.getString("Save"));
			// Dialog aufbauen
			JOptionPane pane = new JOptionPane(
					customiseDialogPanel, JOptionPane.QUESTION_MESSAGE,
					JOptionPane.DEFAULT_OPTION, null, new JButton[] { okButton,
							cancelButton, applyButton, saveButton }, okButton);
			final JDialog dialog = pane.createDialog(SwingUtil
					.getParentWindow(this), cmDialogTitle + " ["
					+ layer.getTitle() + "]");
			dialog.setResizable(true);
			dialog.setModal(false);
			dialog.setSize(380, 200);

			// Aktionen der Farbpaletten-Fenster-Button
			ActionListener action = new ActionListener() {
				public void actionPerformed(ActionEvent e) {
					// Ok- und Apply-Button > ColorMap uebernehmen
					if (e.getSource() == okButton
							|| e.getSource() == applyButton) {
						// Farbe nur uebernehmen, wenn sich die Farb-Palette
						// wirklich
						// geaendert hat
						// if ( !StylingUtil.colorMapsEqual(
						// customiseDialogTable.getColorMap(),
						// StylingUtil.getColorMapFromStyle(layer.getStyle()) )
						// ) {
						layer.setStyle(GridUtil.createStyle(StylingUtil
								.cloneColorMap(customiseDialogPanel
										.getColorMap()), 1.0));
						mapPane.repaint();
						menuItemGroup.setUnselected();
						// }
					}
					// Ok- und Cancel-Button > Dialog schliessen
					if (e.getSource() == okButton
							|| e.getSource() == cancelButton) {
						dialog.setVisible(false);
					}
					// Save-Button > Neue vordefinierte Palette erzeugen
					if (e.getSource() == saveButton) {
						String name = JOptionPane.showInputDialog(
								MapContextControlPane.this,
								saveCMDialogQuestion, saveCMDialogTitle,
								JOptionPane.QUESTION_MESSAGE);
						if (name != null && !name.trim().equals("")) {
							colorMaps.put(name, StylingUtil
									.cloneColorMap(customiseDialogPanel
											.getColorMap()));
						}
					}
				}
			};
			okButton.addActionListener(action);
			cancelButton.addActionListener(action);
			applyButton.addActionListener(action);
			saveButton.addActionListener(action);

			// Wenn sich die die Bezeichnung des Layers aendert, soll auch der
			// Titel des Farbpaletten-Fensters geaendert werden
			layer.addMapLayerListener(new MapLayerAdapter() {
				public void layerChanged(MapLayerEvent e) {
					// Bezeichnung der Checkbox aendern
					dialog.setTitle(cmDialogTitle + " [" + layer.getTitle()
							+ "]");
				}
			});

			return dialog;
		}
	}
}
