/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import java.awt.event.MouseEvent;

/**
 * Extends the geotools {@link org.geotools.swing.MouseSelectionTracker} with an
 * enable/disable functionality. Furthermore the tracker (optionally)
 * also reacts on single clicks.
 * @see #setSelectionMode(SelectionMode)
 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a>
 */
public abstract class MouseSelectionTracker extends skrueger.geotools.MouseSelectionTracker {
  /** Selection events the tracker reacts on. */
  public enum SelectionMode {
    /** Tracker reacts on single clicks as well as on
     *  window selections. */
    ALL,
    /** Tracker reacts only on single clicks. Window selections
     *  are ignored. */
    ONLY_SINGLE_CLICK,
    /** Tracker reacts only on window selection. Single clicks
     *  are ignored. */
    ONLY_WINDOW_SELECTION
  }

  /** Indicates whether the tracker reacts on mouse events. */
  protected boolean enabled = true;
  /** Specifies the selection events the tracker reacts on. */
  protected SelectionMode selectionMode = null;

  /**
   * Creates a new tracker, which reacts only on window
   * selection events (default geotools behavior for
   * {@link org.geotools.swing.MouseSelectionTracker}).
   */
  public MouseSelectionTracker() {
    this(SelectionMode.ONLY_WINDOW_SELECTION);
  }

  /**
   * Creates a new tracker.
   * @param mode  specifies the selection events the tracker reacts on
   */ 
  public MouseSelectionTracker(SelectionMode mode) {
    super();
    setSelectionMode(mode);
  }
  
  /**
   * Returns whether the tracker reacts on mouse events or not.
   */
  public boolean isEnabled() {
    return enabled;
  }
  
  /**
   * Sets whether the tracker reacts on mouse events or not.
   */
  public void setEnabled(boolean enabled) {
    this.enabled = enabled;
  }

  /**
   * Sets the selection events the tracker reacts on.
   * @param mode the mode to set
   */
  public void setSelectionMode(SelectionMode mode) {
    this.selectionMode = mode;
  }

  /**
   * Returns the selection events the tracker reacts on.
   */
  public SelectionMode getSelectionMode() {
    return selectionMode;
  }

  /**
   * Ignores the event if tracker is disabled or selection mode
   * is {@link SelectionMode#ONLY_WINDOW_SELECTION}. Otherwise
   * {@link #selectionPerformed(int, int, int, int)} is called for
   * the selected point.
   */
  @Override
  public void mouseClicked(MouseEvent e) {
    if ( SelectionMode.ONLY_WINDOW_SELECTION.equals(getSelectionMode()) ||
        !isEnabled() || e.getButton() != MouseEvent.BUTTON1 )
      return;
    super.mouseClicked(e);
  
    int clickX = e.getX();
    int clickY = e.getY();
    selectionPerformed(clickX, clickY, clickX, clickY);}

  /**
   * Ignores the event if tracker is disabled. Otherwise the
   * super method is called.
   */
  @Override
  public void mouseEntered(MouseEvent e) {
    if ( !isEnabled() )
      return;
    super.mouseEntered(e);
  }

  /**
   * Ignores the event if tracker is disabled. Otherwise the
   * super method is called.
   */
  @Override
  public void mouseExited(MouseEvent e) {
    if ( !isEnabled() )
      return;
    super.mouseExited(e);
  }
  
  /**
   * Ignores the event if tracker is disabled. Otherwise the
   * super method is called.
   */
  @Override
  public void mousePressed(MouseEvent e) {
    if ( !isEnabled() )
      return;
    super.mousePressed(e);
  }

  /**
   * Ignores the event if tracker is disabled. Otherwise the
   * super method is called.
   */
  @Override
  public void mouseReleased(MouseEvent e) {
    if ( !isEnabled() )
      return;
    super.mouseReleased(e);
  }

  /**
   * Ignores the event if tracker is disabled or selection mode
   * is {@link SelectionMode#ONLY_SINGLE_CLICK}, so that the
   * dragging rectangle is not drawn.
   */
  @Override
  public void mouseDragged(MouseEvent e) {
    if ( SelectionMode.ONLY_SINGLE_CLICK.equals( getSelectionMode() ) ||
        !isEnabled() )
      return;
    super.mouseDragged(e);
  }

  /**
   * Ignores the event if tracker is disabled. Otherwise the
   * super method is called.
   */
  @Override
  public void mouseMoved(MouseEvent e) {
    if ( !isEnabled() )
      return;
    super.mouseMoved(e);
  }

}
