/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionListener;
import java.awt.geom.Point2D;
import java.text.DecimalFormat;
import java.util.Hashtable;
import java.util.Map;

import javax.swing.JLabel;

import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.coverage.grid.io.AbstractGridCoverage2DReader;
import org.geotools.geometry.GeneralDirectPosition;
import org.geotools.map.MapLayer;
import org.geotools.referencing.CRS;
import org.opengis.referencing.crs.CoordinateReferenceSystem;
import org.opengis.referencing.operation.MathTransform;

import schmitzm.data.WritableGrid;
import schmitzm.geotools.JTSUtil;
import schmitzm.geotools.feature.FeatureUtil;
import schmitzm.geotools.grid.GridUtil;
import schmitzm.swing.CaptionsChangeable;

import com.vividsolutions.jts.geom.Coordinate;

/**
 * Diese Klasse stellt ein {@link JLabel} dar, in dem (2dimensionale)
 * Raster-Koordinaten und der Rasterwert an der entsprechenden Stelle angezeigt
 * werden.<br>
 * Die Klasse fungiert als {@link MouseMotionListener} und kann so direkt an ein
 * {@link SelectableXMapPane} gekoppelt werden. Die Koordinaten-Darstellung im
 * Label aktualisiert sich somit automatisch, sobald sich die Maus ueber die
 * Karte bewegt.
 * 
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *         (University of Bonn/Germany)
 * @version 1.1
 * 
 *          SK 01.09.07: finals fuer Variablen eingefuegt... die Funktion
 *          displayCoordinates() wird oft benutzt und es gab viele Variablen,
 *          die nie geaendert wurden.
 */
public class RasterPositionLabel extends JLabel implements MouseMotionListener,
		CaptionsChangeable {

	// Faktor mit dem die Koordinaten vor dem Runden multipliziert und nach
	// dem Runden dividiert werden
	// --> bestimmt die dargestellten Nachkommastellen
	private double fracFactor = 0.0;

	/** Speichert das Format fuer die dargstellten Koordinaten */
	protected DecimalFormat decForm = new DecimalFormat();
	/**
	 * Key, um das Label-Praefix in der {@link CaptionsChangeable}-Map
	 * anzusprechen.
	 * 
	 * @see #resetCaptions(Map)
	 * 
	 * SK 12.12.09: static entfernt, da sonst das Sprachenumschalten die Cpation nicht gewechselt hat.
	 */
	public final String LABEL_PREFIX = RasterPositionLabel.class
			.getName()
			+ ".LABEL_PREFIX";
	
	
	/**
	 * Praefix, das der Raster-Position vorangestellt wird.
	 * 
	 * @see #setLabelPrefix(String)
	 * 
	 * SK 12.12.09: static entfernt, da sonst das Sprachenumschalten die Cpation nicht gewechselt hat.
	 */
	public String labelPrefix = GeotoolsGUIUtil.RESOURCE
			.getString(LABEL_PREFIX);

	/**
	 * Erzeugt ein neues Label. Der Raster-Wert wird ohne Nachkommastellen
	 * dargestellt.
	 */
	public RasterPositionLabel() {
		this(0);
	}

	/**
	 * Erzeugt ein neues Label.
	 * 
	 * @param fracDigits
	 *            Anzahl an Nachkommastellen, auf der Rasterwert gerundet wird
	 */
	public RasterPositionLabel(final int fracDigits) {
		super(" ");
		// eine Stelle vor dem Komma soll immer dargestellt werden
		this.decForm.setMinimumIntegerDigits(1);
		// Nachkommastellen setzen
		this.setFractionDigits(fracDigits);
	}

	/**
	 * Stellt die Koordinaten und den Wert des obersten (sichtbaren) Rasters im
	 * Label dar, wenn das Event von einem {@link SelectableXMapPane} ausgeloest
	 * wurde.<br>
	 * Wird {@link #mouseMoved(MouseEvent)} und
	 * {@link #mouseDragged(MouseEvent)} aufgerufen.
	 */
	protected void displayCoordinates(final MouseEvent e) {
		if (e == null || !(e.getSource() instanceof SelectableXMapPane))
			return;

		final SelectableXMapPane mapPane = (SelectableXMapPane) e.getSource();

		// Welt-Koordinaten der Mausposition ermitteln (in CRS der Map)
		final Point2D actPos_MapCRS = SelectableXMapPane
				.getMapCoordinatesFromEvent(e);
		// Welt-Koordinaten der Mausposition ermitteln (in CRS der Map)
		if (actPos_MapCRS == null)
			return;

		// oberstes dargestelltes Raster suchen
		final MapLayer layer = determineRasterLayer(mapPane, actPos_MapCRS);
		if (layer == null) {
			setText("");
			return;
		}

		CoordinateReferenceSystem mapCrs = mapPane.getMapContext()
				.getCoordinateReferenceSystem();
		CoordinateReferenceSystem layerCrs = layer.getFeatureSource()
				.getSchema().getGeometryDescriptor()
				.getCoordinateReferenceSystem();

		// Objekt aus Layer herausholen
		final Object layerObj = FeatureUtil.getLayerSourceObject(layer);

		// We checked that it's a GridLayer.. so have no fear...
		// if ((layerObj == null)
		// || (!(layerObj instanceof GridCoverage2D) && !(layerObj instanceof
		// org.geotools.coverage.grid.io.AbstractGridCoverage2DReader)))
		// return;

		final StringBuffer valueStr = new StringBuffer();
		if (layerObj instanceof GridCoverage2D) {
			final GridCoverage2D gc = (GridCoverage2D) layerObj;
			double[] gcValue = new double[0];
			GeneralDirectPosition actPos_GridCRS;
			try {

				Coordinate posInLayerCRS = JTSUtil.transformCoordinate(
						new Coordinate(actPos_MapCRS.getX(), actPos_MapCRS
								.getY()), mapCrs, layerCrs);
				//				
				// // Koordinaten in CRS des Rasters umrechnen
				// final MathTransform mapToGrid = CRS.findMathTransform(mapCrs,
				// layerCrs);
				// mapToGrid.transform(new GeneralDirectPosition(actPos_MapCRS),
				// actPos_GridCRS);

				actPos_GridCRS = new GeneralDirectPosition(posInLayerCRS.x,
						posInLayerCRS.y);
				//				
				// Wert im Raster ermitteln
				gcValue = gc.evaluate(actPos_GridCRS, (double[]) null);
			} catch (final Exception err) {
				// Position out of Raster data
				return;
			}

			valueStr.append(labelPrefix == null ? "" : labelPrefix);
			// Raster-Koordinaten anzeigen
			int cellX = 0;
			int cellY = 0;
			if (gc instanceof WritableGrid) {
				// Bei einem WritableGrid koennen diese automatisch ermittelt
				// werden
				final WritableGrid wg = (WritableGrid) gc;
				cellX = wg
						.convertRealToRaster(actPos_GridCRS.getOrdinate(0), 0);
				cellY = wg
						.convertRealToRaster(actPos_GridCRS.getOrdinate(1), 1);
			} else {
				// sonst: Raster-Koordinaten ausrechnen
				final int cell[] = GridUtil.convertRealToRaster(gc,
						actPos_GridCRS.getCoordinate());
				cellX = cell[0];
				cellY = cell[1];
			}
			valueStr.append(" (").append(cellX).append("/").append(cellY)
					.append(")");

			valueStr.append(": ");
			for (int j = 0; j < gcValue.length; j++) {
				if (j > 0)
					valueStr.append(" / ");
				valueStr.append(Double.isNaN(gcValue[j]) ? gcValue[j] : decForm
						.format(gcValue[j]));
				// Debug: ungerundeten Raster-Wert anzeigen
				// valueStr.append("   Orig = " + gcValue[j]);
			}
		}

		// CHANGE SK 9.8.2008 START

		else if (layerObj instanceof org.geotools.coverage.grid.io.AbstractGridCoverage2DReader) {
			final GeneralDirectPosition actPos_GridCRS = new GeneralDirectPosition(
					2);
			try {

				AbstractGridCoverage2DReader gcr = (AbstractGridCoverage2DReader) layerObj;
				// Koordinaten in CRS des Rasters umrechnen
				final MathTransform mapToGrid = CRS.findMathTransform(mapPane
						.getMapContext().getCoordinateReferenceSystem(), gcr
						.getCrs());
				mapToGrid.transform(new GeneralDirectPosition(actPos_MapCRS),
						actPos_GridCRS);
				// Wert im Raster ermitteln
				Hashtable<MapLayer, double[]> foundGridCoverageValues = mapPane
						.findGridCoverageValues(actPos_GridCRS.toPoint2D(),
								SelectableXMapPane.SELECT_ALL);
				if (foundGridCoverageValues != null
						&& foundGridCoverageValues.size() > 0) {
					double[] gcValue = foundGridCoverageValues.values()
							.iterator().next();
					for (int j = 0; j < gcValue.length; j++) {
						if (j > 0)
							valueStr.append(" / ");
						valueStr.append(Double.isNaN(gcValue[j]) ? gcValue[j]
								: decForm.format(gcValue[j]));
					}
				}
			} catch (final Exception err) {
				// Position out of Raster data
			}

			// CHANGE SK 9.8.2008 END
		}

		// Text im Label anzeigen
		setText(valueStr.toString());
	}

	/**
	 * Ermittelt das Raster-Layer, dessen Koordinaten angezeigt werden.
	 * Standardmaessig wird das oberste sichtbare Raster-Layer zurueckgegeben.
	 * Sub-Klassen koennen diese Methode ueberschreiben, um ein anderes Layer zu
	 * verwenden.
	 * 
	 * @param mapPane
	 *            MapPane der angezeigten Layer.
	 * @param actPos_MapCRS
	 *            Position fuer die das erste schneidende raster layer gefunden
	 *            werden soll.
	 */
	protected MapLayer determineRasterLayer(final SelectableXMapPane mapPane,
			Point2D actPos_MapCRS) {

		// find the first grid layer intersecting with the click position
		for (int idx = mapPane.getMapContext().getLayerCount(); idx >= 1; idx--) {
			MapLayer layer = mapPane.getMapContext().getLayer(idx - 1);
			if (!layer.isVisible())
				continue;
			if (!SelectableXMapPane.isGridCoverageLayer(layer))
				continue;
			if (!mapPane.layerIntersectsCoord(layer, new Coordinate(
					actPos_MapCRS.getX(), actPos_MapCRS.getY())))
				continue;
			return layer;
		}
		return null;
	}

	/**
	 * Setzt den String, welcher der Rasterposition und dem Rasterwert
	 * vorangestellt wird.
	 * 
	 * @param label
	 *            neues Praefix
	 */
	public void setLabelPrefix(final String label) {
		labelPrefix = (label == null) ? "" : label;
	}

	/**
	 * Liefert den String, welcher der Rasterposition und dem Rasterwert
	 * vorangestellt wird.
	 */
	public String getLabelPrefix() {
		return labelPrefix;
	}

	/**
	 * Setzt das Label-Praefix neu, sofern in der Map ein Wert fuer
	 * {@link #LABEL_PREFIX} hinterlegt ist.
	 * 
	 * @param captionMap
	 *            neue Labels
	 */
	public void resetCaptions(final Map<String, Object> captionMap) {
		if (!captionMap.containsKey(LABEL_PREFIX))
			return;
		final Object newPrefix = captionMap.get(LABEL_PREFIX);
		setLabelPrefix(newPrefix == null ? null : newPrefix.toString());
	}

	/**
	 * Setzt die Anzahl an Nachkommastellen, die fuer den Rasterwert dargestellt
	 * werden
	 * 
	 * @param fracDigits
	 *            Anzahl an Nachkommastellen
	 */
	public void setFractionDigits(final int fracDigits) {
		this.fracFactor = Math.pow(10.0, fracDigits);
		this.decForm.setMinimumFractionDigits(fracDigits);
		this.decForm.setMaximumFractionDigits(fracDigits);
	}

	/**
	 * Liefert die Anzahl an Nachkommastellen, die fuer den Rasterwert
	 * dargestellt werden
	 */
	public int getFractionDigits() {
		return (int) Math.log10(this.fracFactor);
	}

	/**
	 * Wird aufgerufen, sobald die Maus bewegt wird. Stellt die Koordinaten und
	 * den Rasterwert im Label dar, wenn das Event von einem
	 * {@link SelectableXMapPane} ausgeloest wurde.
	 * 
	 * @see #displayCoordinates(MouseEvent)
	 */
	public void mouseMoved(final MouseEvent e) {
		displayCoordinates(e);
	}

	/**
	 * Wird aufgerufen, sobald die Maus bei gedrueckter Taste bewegt wird.
	 * Stellt die Koordinaten und den Rasterwert im Label dar, wenn das Event
	 * von einem {@link SelectableXMapPane} ausgeloest wurde.
	 * 
	 * @see #displayCoordinates(MouseEvent)
	 */
	public void mouseDragged(final MouseEvent e) {
		displayCoordinates(e);
	}
}
