/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.text.DecimalFormat;

import javax.swing.JLabel;

import schmitzm.swing.JPanel;
import schmitzm.swing.SwingUtil;

/**
 * Stellt einen horizontalen Massstab-Balken dar. Die dargestellte Breite
 * richtet sich nach der Groesse der Parent-Componente. Der Balken stellt immer
 * <b>volle</b> 10, 100, 1000, ... Meter dar.<br>
 * Diese Groesse wird rechts neben dem schwarz/weissen Balken angezeigt.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ScalePanel extends JPanel {
  private static final DecimalFormat numFormat = new DecimalFormat("###,###,##0");
  public final static Color[] barColor  = new Color[] { Color.BLACK, Color.WHITE };

  /** Beinhaltet die aktuell dargestellte Aufloesung. Der Wert entspricht den
   *  Metern, die durch einen Bildschirm-Pixel dargestellt werden.*/
  protected double scale = 1;

  /**
   * Erzeugt eine Mass-Stab.
   */
  public ScalePanel() {
    super();
    SwingUtil.setPreferredHeight(this,10);
  }

  /**
   * Liefert die dargestellte Aufloesung in Pixel pro Meter.
   */
  public double getScaleInPixels() {
    return this.scale!=0 ? 1/this.scale : 0;
  }

  /**
   * Liefert die dargestellte Aufloesung in Metern pro Pixel.
   */
  public double getScaleInMeters() {
    return this.scale;
  }

  /**
   * Setzt die Aufloesung des Mass-Stabs und aktualisiert die Anzeige der
   * Komponente.
   * @param meters Meter, die durch einen Bildschirm-Pixel repraesentiert werden
   */
  public void setScale(final double meters) {
    if ( meters <= 0 )
      throw new IllegalArgumentException("Meters must be greater zero!");
    this.scale = meters;
    repaint();
  }

  /**
   * Setzt die Aufloesung des Mass-Stabs und aktualisiert die Anzeige der
   * Komponente.
   * @param pixels Anzahl an Pixeln, die einen Meter repraesentieren werden
   * @exception Ill
   */
  public void setScale(final long pixels) {
    if ( pixels <= 0 )
      throw new IllegalArgumentException("Count of pixels must be greater zero!");
    setScale(1.0/pixels);
  }

  /**
   * Zeichnet den Massstab-Balken.
   * @param g Graphics
   */
  public void paint(final Graphics g) {
    if (getParent() == null )
      return;
    
    final Graphics2D g2 = (Graphics2D)g;
    g2.setRenderingHint(RenderingHints.KEY_TEXT_ANTIALIASING, RenderingHints.VALUE_TEXT_ANTIALIAS_ON);
    g2.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_QUALITY);
    
    super.paint(g2);

    // maximale Breite fuer Skala (abzgl. 25 Pixel fuer Meter-Angabe)
    final int    maxW_pixel    = getParent().getWidth() - 30;
    final double maxW_meter    = maxW_pixel * getScaleInMeters();
    // Ausmasse der Skala (stellt immer volle 10/100/1000/... Meter dar)
    final long   scaleW_meter = (long)Math.pow(10, (long)Math.log10(maxW_meter));
    final int    scaleW_pixel = (int)Math.round( scaleW_meter * getScaleInPixels() );
    final int    scaleH_pixel = getHeight()-1;

    // Aufteilung in der Skala Teil-Balken
    final int     tileScaleCount  = 4;
    final int     tileScaleWidth  = scaleW_pixel/tileScaleCount;
    final int     tileScaleHeight = scaleH_pixel;
    
    int     nextColorIdx = 0;

    // Teilbalken malen
    for (int i = 0; i < tileScaleCount; i++) {
      g2.setColor(barColor[nextColorIdx]);
      g2.fillRect(i * tileScaleWidth, 0, tileScaleWidth, tileScaleHeight);
      nextColorIdx = 1 - nextColorIdx;
    }
    // Rahmen um alle Balken
    g2.setColor(Color.BLACK);
    g2.drawRect(0, 0, tileScaleWidth*tileScaleCount, tileScaleHeight);
    // Laengen-Angabe
    g2.setFont( new JLabel().getFont() );
    
    //TODO This should be switchable from km to miles
    final String scaleString = scaleW_meter >= 1000 ? numFormat.format(scaleW_meter/1000)+"km" : numFormat.format(scaleW_meter)+"m";
    
    g2.drawString(scaleString,scaleW_pixel+3,tileScaleHeight);
  }
}
