/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.util.ArrayList;
import java.util.List;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollBar;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JToolBar;
import javax.swing.SortOrder;
import javax.swing.RowSorter.SortKey;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import org.geotools.feature.FeatureCollection;

import schmitzm.geotools.feature.FeatureUtil;
import schmitzm.swing.table.SelectionTableModel;

/**
 * Extends the {@link FeatureTablePane} with buttons and functionality to select
 * lines. Can be configured to set the regions on an external a {@link SelectableXMapPane}
 * . Can be configured to show a simple preview {@link SelectableXMapPane} to the left of
 * the table.
 * 
 * @author Stefan A. Krüger
 * 
 */
public class SelectableFeatureTablePane extends FeatureTablePane {

	public static final ImageIcon ICON_SELECTION_INVERT = new ImageIcon(
			SelectableFeatureTablePane.class
					.getResource("resource/icons/mActionInvertSelection.png"));
	
	public static final ImageIcon ICON_SELECTED_TO_TOP = new ImageIcon(
			SelectableFeatureTablePane.class
					.getResource("resource/icons/mActionSelectedToTop.png"));
	
	public static final ImageIcon ICON_UNSELECT = new ImageIcon(
			SelectableFeatureTablePane.class
					.getResource("resource/icons/mActionUnselectAttributes.png"));
	
	public static final ImageIcon ICON_ZOOM_TO_SELECTED = new ImageIcon(
			SelectableFeatureTablePane.class
					.getResource("resource/icons/mActionZoomToSelected.png"));
	
	
	private final SelectableXMapPane externalMapPane;
	
	private JLabel statusLabel;

    /**
     * @param fc
     *            {@link FeatureCollection} that holds the data.
     * @param geomPreview
     *            If <code>true</code>, a preview {@link SelectableXMapPane} is attached
     *            to the left of the table.
     * @param externalMapPane
     *            <code>null</code> if this component is NOT linked to an
     *            external JMapPane. If mapPane == <code>null</code>, the
     *            "ZoomToSelection" Button is automatically disabled. If a
     *            {@link SelectableXMapPane} is passed, the "ZoomToSelectedFeature"
     *            function will be enabled.
     */
    public SelectableFeatureTablePane(FeatureCollection fc, boolean geomPreview, SelectableXMapPane externalMapPane) {
      this(new FeatureCollectionTableModel(fc), geomPreview, externalMapPane);
    }

    /**
	 * @param model
	 *            {@link FeatureCollectionTableModel} that holds the data.
	 * @param geomPreview
	 *            If <code>true</code>, a preview {@link SelectableXMapPane} is attached
	 *            to the left of the table.
	 * @param externalMapPane
	 *            <code>null</code> if this component is NOT linked to an
	 *            external JMapPane. If mapPane == <code>null</code>, the
	 *            "ZoomToSelection" Button is automatically disabled. If a
	 *            {@link SelectableXMapPane} is passed, the "ZoomToSelectedFeature"
	 *            function will be enabled.
	 */
	public SelectableFeatureTablePane(FeatureCollectionTableModel model,
			boolean geomPreview, SelectableXMapPane externalMapPane) {
		super(model, null, geomPreview);

		this.externalMapPane = externalMapPane;

		JPanel newPanelArroundTable = new JPanel(new BorderLayout());

		if (splitPane != null) {
			newPanelArroundTable.add(splitPane.getRightComponent(),
					BorderLayout.CENTER);
			splitPane.setRightComponent(newPanelArroundTable);
			splitPane.setDividerLocation(0.33); // SK: Does have no effect because the JSplitPane is not yet rendered.
		} else {
			newPanelArroundTable.add(getComponent(0), BorderLayout.CENTER);
			removeAll();
			add(newPanelArroundTable);
		}

		// Change the sizes
		mapPane.setPreferredSize(new Dimension(200, 300));
		
		featuresTable.getParent().setPreferredSize(new Dimension(400, 300));
	    featuresTable.setAutoResizeMode( JTable.AUTO_RESIZE_OFF );

	    // In the NORTH, the ToolBar and a JLabel presenting the total and selected number of features is displayed
		JPanel north = new JPanel(new BorderLayout());
		north.add(getToolBar(), BorderLayout.WEST);
		north.add(getStatusLabel(), BorderLayout.EAST);
		newPanelArroundTable.add(north, BorderLayout.NORTH);
		newPanelArroundTable.add(getSearchBar(), BorderLayout.SOUTH);
		
		featuresTable.getSelectionModel().addListSelectionListener( new ListSelectionListener(){

			@Override
			public void valueChanged(ListSelectionEvent e) {
				if (!e.getValueIsAdjusting()){
					getStatusLabel();
				}
			}
			
		});
	}

	/**
	 * After calling the super-method the table model is encapsulated in
	 * a {@link SelectionTableModel} to show a selection column.
	 */
    @Override
    protected void initGUI(boolean geomPreview) {
      super.initGUI(geomPreview);
      // Reset the table model to show a selection column
      featuresTable.setModel( new SelectionTableModel(featuresTableModel, featuresTable) );
      // Because SortableJTable is used, the first col is always a checkbox.
      // TODO: How wide is a CheckBox in differnet L'n'F s?
      int selCol = featuresTable.convertColumnIndexToModel(0);
      featuresTable.getColumnModel().getColumn(selCol).setMaxWidth(17);
    }
    
    /**
	 * This Label present something like "40 of 56 Polygons selected". It is lazily constructed. Later calls will only update the existing label.  
	 */
	private JLabel getStatusLabel() {
		if (statusLabel == null){
			statusLabel = new JLabel();
		}
		
		int selectedAnz = getSelectedFeatures().size();
		
		// TODO Cache totalAnz and localizedGeometryTypeName. Do not recalc all the time!  
		int totalAnz = getFeatureCollection().size();
		
		String localizedGeometryTypeName = "rows";
		
		switch ( FeatureUtil.getGeometryForm(getFeatureCollection()) ){
		case LINE:
			localizedGeometryTypeName = GeotoolsGUIUtil.R("schmitzm.geotools.gui.SelectableFeatureTablePane.selectionStatus.lines");
			break;
		case POLYGON:
			localizedGeometryTypeName = GeotoolsGUIUtil.R("schmitzm.geotools.gui.SelectableFeatureTablePane.selectionStatus.polygons");
			break;
		case POINT:
			localizedGeometryTypeName = GeotoolsGUIUtil.R("schmitzm.geotools.gui.SelectableFeatureTablePane.selectionStatus.points");
			break;
		}
		
		statusLabel.setText( GeotoolsGUIUtil.R("schmitzm.geotools.gui.SelectableFeatureTablePane.selectionStatus", selectedAnz, totalAnz,localizedGeometryTypeName));
		
		return statusLabel;
	}

	private Component getSearchBar() {
		return new JLabel(""); // TODO Add ability to search here!
	}

	/**
	 * Creates a {@link JToolBar} with actions to perform on the map.
	 */
	protected JToolBar getToolBar() {
		JToolBar toolBar = new JToolBar();
		
		toolBar.setFloatable(false); // Sorry Martin ;-) 

		/**
		 * Add an Action to clear the selection
		 */
		AbstractAction clearSelectionAction = new AbstractAction("",
				ICON_UNSELECT) {

			@Override
			public void actionPerformed(ActionEvent e) {
				featuresTable.getSelectionModel().clearSelection();
			}

		};
		clearSelectionAction
				.putValue(
						Action.SHORT_DESCRIPTION,
						GeotoolsGUIUtil
								.R("schmitzm.geotools.gui.SelectableFeatureTablePane.button.clearSelection.tt"));
		toolBar.add(clearSelectionAction);

		/**
		 * Add an Action to move the selected items to the top
		 */
		AbstractAction selectionToTopAction = new AbstractAction("",
				ICON_SELECTED_TO_TOP) {

			// A bug? Sorting only works from the second time on..
			boolean isFirstClick = true;

			@Override
			public void actionPerformed(ActionEvent e) {

				// moves the select rows to the top by sorting according to the
				// first column

				List<SortKey> keys = new ArrayList<SortKey>(featuresTable
						.getRowSorter().getSortKeys());

				SortKey sortKey = new SortKey(0, SortOrder.DESCENDING);
				keys.add(0, sortKey);

				featuresTable.getRowSorter().setSortKeys(keys);
				
				// Move the scrollbars to the top
				JScrollPane scrollPane = getFeaturesTableScrollPane();
				JScrollBar verticalScrollBar = scrollPane
						.getVerticalScrollBar();
				JScrollBar horizontalScrollBar = scrollPane
						.getHorizontalScrollBar();
				verticalScrollBar.setValue(verticalScrollBar.getMinimum());
				horizontalScrollBar.setValue(horizontalScrollBar.getMinimum());


				// A bug? Sorting only works from the second time on..
				if (isFirstClick) {
					isFirstClick = false;
					actionPerformed(e);
				}
			}

		};
		selectionToTopAction
				.putValue(
						Action.SHORT_DESCRIPTION,
						GeotoolsGUIUtil
								.R("schmitzm.geotools.gui.SelectableFeatureTablePane.button.selectionToTop.tt"));
		toolBar.add(selectionToTopAction);

		/**
		 * Add an Action to move the selected items
		 */
		AbstractAction selectionInverseAction = new AbstractAction("",
				ICON_SELECTION_INVERT) {

			@Override
			public void actionPerformed(ActionEvent e) {

				setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
				
				featuresTable.getSelectionModel().setValueIsAdjusting(true);
				
				for (int i = 0; i < featuresTable.getModel().getRowCount(); i++) {
					if (featuresTable.getSelectionModel().isSelectedIndex(i)) {
					  featuresTable.getSelectionModel().removeSelectionInterval(
								i, i);
					} else {
					  featuresTable.getSelectionModel().addSelectionInterval(i,
								i);
					}
				}
				featuresTable.getSelectionModel().setValueIsAdjusting(false);
				setCursor(null);

			}

		};
		selectionInverseAction
				.putValue(
						Action.SHORT_DESCRIPTION,
						GeotoolsGUIUtil
								.R("schmitzm.geotools.gui.SelectableFeatureTablePane.button.invertSelection.tt"));
		toolBar.add(selectionInverseAction);

		/**
		 * Add an Action to move the selected items
		 */
		if (externalMapPane != null) {

			AbstractAction zoomToSelection = new AbstractAction("",
					ICON_ZOOM_TO_SELECTED) {

				@Override
				public void actionPerformed(ActionEvent e) {
					FeatureCollection selectedFeatures = getSelectedFeatures();
					if (selectedFeatures.size() == 0)
						selectedFeatures = getFeatureCollection();
					externalMapPane.zoomTo(selectedFeatures);
					externalMapPane.refresh();

				}

			};
			zoomToSelection
					.putValue(
							Action.SHORT_DESCRIPTION,
							GeotoolsGUIUtil
									.R("schmitzm.geotools.gui.SelectableFeatureTablePane.button.zoomToSelection.tt"));
			toolBar.add(zoomToSelection);
		}

		return toolBar;
	}

}
