/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import java.awt.BorderLayout;
import java.awt.LayoutManager;
import java.awt.Point;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.io.IOException;
import java.util.Collection;
import java.util.Enumeration;
import java.util.Hashtable;

import org.apache.log4j.Logger;
import org.geotools.coverage.grid.GeneralGridEnvelope;
import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.coverage.grid.GridGeometry2D;
import org.geotools.coverage.grid.io.AbstractGridCoverage2DReader;
import org.geotools.coverage.grid.io.AbstractGridFormat;
import org.geotools.data.FeatureSource;
import org.geotools.data.memory.MemoryFeatureCollection;
import org.geotools.feature.FeatureCollection;
import org.geotools.filter.GeometryFilterImpl;
import org.geotools.geometry.GeneralEnvelope;
import org.geotools.geometry.jts.JTS;
import org.geotools.geometry.jts.ReferencedEnvelope;
import org.geotools.map.DefaultMapContext;
import org.geotools.map.MapContext;
import org.geotools.map.MapLayer;
import org.geotools.parameter.Parameter;
import org.geotools.referencing.CRS;
import org.geotools.renderer.GTRenderer;
import org.geotools.renderer.lite.StreamingRenderer;
import org.opengis.coverage.CannotEvaluateException;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.simple.SimpleFeatureType;
import org.opengis.feature.type.PropertyDescriptor;
import org.opengis.filter.Filter;
import org.opengis.parameter.GeneralParameterValue;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import schmitzm.geotools.JTSUtil;
import schmitzm.geotools.feature.FeatureUtil;
import schmitzm.geotools.grid.GridUtil;
import schmitzm.geotools.map.event.FeatureSelectedEvent;
import schmitzm.geotools.map.event.GeneralSelectionEvent;
import schmitzm.geotools.map.event.GridCoverageSelectedEvent;
import schmitzm.geotools.map.event.GridCoverageValueSelectedEvent;
import schmitzm.geotools.map.event.JMapPaneListener;
import schmitzm.geotools.map.event.MapAreaChangedEvent;
import schmitzm.geotools.map.event.MapPaneEvent;
import schmitzm.geotools.map.event.ScaleChangedEvent;
import schmitzm.geotools.styling.StylingUtil;
import skrueger.geotools.GeomFilterGenerator;
import skrueger.geotools.XMapPane;
import skrueger.geotools.GeomFilterGenerator.BoundingBoxFilterGenerator;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Envelope;

/**
 * Diese Klasse erweitert die Geotools-Klasse
 * {@link org.geotools.swing.JMapPane} um folgende Features:
 * <ul>
 * <li>zusaetzliche Maus-Steuerungen:
 * <ul>
 * <li><b>Linksklick:</b> ueber {@link #setState(int)} eingestellte Aktion</li>
 * <li><b>Rechtsklick:</b> Zoom-Out um Faktor 2 (nur wenn Linksklick auf Zoom-In
 * eingestellt ist)</li>
 * <li><b>Drag mit linker Maustaste:</b> neuen Karten-Bereich selektieren oder
 * Features selektieren (siehe {@link #setState(int)})</li>
 * <li><b>Drag mit rechter Maustaste:</b> Karten-Bereich verschieben</li>
 * <li><b>Mausrad:</b> Zoom-In/Out ueber aktueller Position (Faktor 1.2)</li>
 * </ul>
 * </li>
 * <li>Ankoppeln von {@link JMapPaneListener} und Ausloesung diverser
 * Ereignisse:
 * <ul>
 * <li><b>{@link ScaleChangedEvent}:</b> Wird ausgeloest, wenn sich die
 * Aufloesung der angezeigten Karte aendert</li>
 * <li><b>{@link MapAreaChangedEvent}:</b> Wird ausgeloest, wenn sich die
 * Aufloesung angezeigte Karte-Ausschnitt aendert</li>
 * <li><b>{@link GeneralSelectionEvent}:</b> Wird ausgeloest, wenn der Anwender
 * einen Bereich aus der Karte ausgewaehlt hat (egal ob dabei gezoomt wurde,
 * Features/Raster selektiert wurden, oder nicht selektiert wurde)</li>
 * <li><b>{@link FeatureSelectedEvent}:</b> Wird ausgeloest, wenn der Anwender
 * Features aus der Karte ausgewaehlt hat</li>
 * <li><b>{@link GridCoverageSelectedEvent}:</b> Wird ausgeloest, wenn der
 * Anwender Raster-Bereiche aus der Karte ausgewaehlt hat</li>
 * </ul>
 * </li>
 * </ul>
 * Sofern eingeschaltet, erfolgt {@linkplain #setHighlight(boolean)
 * Highlighting} immer auf dem obersten sichtbaren Nicht-Raster-Layer.<br>
 * Darueberhinaus besteht ueber {@link #getScreenToWorld()} Zugriff auf eine
 * {@linkplain AffineTransform affine Transformation} mit der die aktuellen
 * Fenster-Koordinaten (z.B. eines <code>MouseEvent</code>) in
 * Karten-Koordinaten (Latitude/Longitude) umgerechnet werden koennen.
 * 
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *         (University of Bonn/Germany)
 * @version 1.0
 */
public class SelectableXMapPane extends XMapPane {
	

	/**
	 * "SimpleFeature-Auswahl auf allen (sichtbaren) Layern".
	 */
	public static final int SELECT_ALL = 103;
	/**
	 * "Auswahl nur eines Features, das erste sichtbare von Oben".
	 */
	public static final int SELECT_ONE_FROM_TOP = 104;
	/**
	 * "SimpleFeature-Auswahl auf dem obersten (sichtbaren) Layer".
	 */
	public static final int SELECT_TOP = 4;


	/** Logger for debug messages. */
	protected static final Logger LOGGER = Logger
			.getLogger(SelectableXMapPane.class);

	/**
	 * Flag fuer Modus "Nichts machen".
	 * 
	 * @see #setState(int)
	 * @see #setState(int)
	 */
	public static final int NONE = 100;

	/**
	 * Erzeugt ein neues MapPane.
	 * 
	 * @param layout
	 *            Layout-Manager fuer die GUI-Komponente (z.B.
	 *            {@link BorderLayout})
	 * @param isDoubleBuffered
	 *            siehe Konstruktor der
	 *            {@linkplain org.SelectableXMapPane.org.geotools.swing.JMapPane#JMapPane(LayoutManager,boolean,GTRenderer,MapContext)
	 *            Oberklasse}
	 * @param renderer
	 *            Renderer fuer die graphische Darestellung (z.B.
	 *            {@link StreamingRenderer})
	 * @param localContext
	 *            Verwaltung der einzelnen Layer (z.B. {@link DefaultMapContext}
	 *            ).
	 */
	public SelectableXMapPane(MapContext context) {
		super(context, null);
	}

	/**
	 * Erzeugt ein neues MapPane.
	 * 
	 * @param layout
	 *            Layout-Manager fuer die GUI-Komponente (z.B.
	 *            {@link BorderLayout})
	 * @param isDoubleBuffered
	 *            siehe Konstruktor der
	 *            {@linkplain org.SelectableXMapPane.org.geotools.swing.JMapPane#JMapPane(LayoutManager,boolean,GTRenderer,MapContext)
	 *            Oberklasse}
	 * @param renderer
	 *            Renderer fuer die graphische Darestellung (z.B.
	 *            {@link StreamingRenderer})
	 * @param localContext
	 *            Verwaltung der einzelnen Layer (z.B. {@link DefaultMapContext}
	 *            ).
	 */
	public SelectableXMapPane() {
		super(null, null);
	}

	/**
	 * Liefert die Anzahl der Einheiten, die ein Bildschirm-Pixel darstellt. Die
	 * Einheit ist die Grundeinheit des CRS
	 */
	public double getScale() {

		if (!isWellDefined())
			return 0.0;

		return getMapArea().getWidth() / getWidth();
	}

	/**
	 * Liefert oberste Layer (sichtbar oder unsichtbar).
	 */
	public MapLayer getTopLayer() {
		int count = getMapContext().getLayerCount();
		return count > 0 ? getMapContext().getLayer(count - 1) : null;
	}

	/**
	 * Liefert oberste sichtbare Layer.
	 */
	public MapLayer getTopVisibleLayer() {
		for (int i = getMapContext().getLayerCount() - 1; i >= 0; i--) {
			MapLayer layer = getMapContext().getLayer(i);
			if (layer.isVisible())
				return layer;
		}
		return null;
	}

	/**
	 * Liefert oberste sichtbare Raster-Layer.
	 */
	public MapLayer getTopVisibleGridCoverageLayer() {
		for (int i = getMapContext().getLayerCount() - 1; i >= 0; i--) {
			MapLayer layer = getMapContext().getLayer(i);
			if (layer.isVisible() && isGridCoverageLayer(layer))
				return layer;
		}
		return null;
	}

	/**
	 * Liefert oberste sichtbare Nicht-Raster-Layer.
	 */
	public MapLayer getTopVisibleNonGridCoverageLayer() {
		for (int i = getMapContext().getLayerCount() - 1; i >= 0; i--) {
			MapLayer layer = getMapContext().getLayer(i);
			if (layer.isVisible() && !isGridCoverageLayer(layer))
				return layer;
		}
		return null;
	}

	/**
	 * Liefert unterste Layer (sichtbar oder unsichtbar).
	 */
	public MapLayer getBottomLayer() {
		return getMapContext().getLayerCount() > 0 ? getMapContext()
				.getLayer(0) : null;
	}


	/**
	 * Ermittelt die Raster-Werte, die an einem Punkt liegen und erzeugt
	 * entsprechende {@link GridCoverageValueSelectedEvent
	 * GridCoverageValueSelectedEvents}. Beim Modus {@link #SELECT_TOP} wird nur
	 * das oberste sichtbare Layer durchsucht. Beim Modus {@link #SELECT_ALL}
	 * werden alle sichtbaren Layer durchsucht.
	 * <p>
	 * SK: 28.09.2007 Da ein Rasterlayer auch mehrere Baender haben kann, ist es
	 * sinnvoll, nicht <code>Hashtable MapLayer,Double </code> sondern
	 * <code>Hashtable MapLayer,Double[] </code> zurueckzugeben.
	 * 
	 * @param point
	 *            Geo-Referenz
	 * @param mode
	 *            Suchmodus
	 * @return eine leere {@link Hashtable} falls der Punkt {@code null} ist
	 * 
	 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
	 *         (University of Bonn/Germany)
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public boolean findGridCoverageValuesAndFireEvents(Point2D point, int mode) {
		Hashtable<MapLayer, double[]> result = findGridCoverageValues(point,
				mode);

		// Events ausloesen fuer jedes Layer
		for (Enumeration<MapLayer> e = result.keys(); e.hasMoreElements();) {
			MapLayer layer = e.nextElement();
			double[] values = result.get(layer);
			fireMapPaneEvent(new GridCoverageValueSelectedEvent(this, layer,
					point, values));
		}
		return !result.isEmpty();
	}

	/**
	 * Setzt die sichtbare Karte. Danach wird die {@linkplain AffineTransform
	 * Transformation} zwischen Fenster-Koordinaten und Karten-Koordinaten neu
	 * berechnet.<br>
	 * Loest ein {@link ScaleChangedEvent aus}
	 * 
	 * {@link #setMapArea(Envelope)} return <code>false</code>, falls durch die
	 * neue MapArea ein nicht gueltiger Massstab entstehen wuerde UND die
	 * bisherige maparea != null ist
	 * 
	 * @param newMapArea
	 *            neuer Kartenausschnitt im CRS des {@link MapContext}
	 * @see #resetTransforms()
	 * @see #getScreenToWorld()
	 * 
	 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
	 *         (University of Bonn/Germany)
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	@Override
	public boolean setMapArea(ReferencedEnvelope newMapArea) {
		double oldScale = getScale();

		boolean b = super.setMapArea(newMapArea);

		if (b) {
			if (getScale() > 0) // If the JPane has not been set yet, there is
				// no scale
				fireMapPaneEvent(new ScaleChangedEvent(this, oldScale,
						getScale()));
			fireMapPaneEvent(new MapAreaChangedEvent(this, oldMapArea,
					getMapArea()));
		}
		return b;
	}

	/**
	 * Sets the mapArea to smartly present the given features. Note: The method
	 * does not call {@link #repaint()} on the {@link SelectableXMapPane}.
	 */
	public void zoomTo(SimpleFeature feature) {
		final MemoryFeatureCollection mfc = new MemoryFeatureCollection(feature
				.getFeatureType());
		mfc.add(feature);
		zoomTo(mfc);
	}

	/**
	 * Testet (anhand der Features), ob das Objekt eines Layers eine
	 * Bounding-Box schneidet.
	 * 
	 * @param layer
	 *            ein Layer
	 * @param env
	 *            Bounding-Box in CRS des MapPane
	 */
	public boolean featureLayerIntersectsEnvelope(MapLayer layer, Envelope env) {
		try {
			FeatureSource<SimpleFeatureType, SimpleFeature> featureSource = (FeatureSource<SimpleFeatureType, SimpleFeature>) layer
					.getFeatureSource();
			GeometryFilterImpl filter = new BoundingBoxFilterGenerator(env,
					getMapContext().getCoordinateReferenceSystem())
					.adaptFilter(featureSource);
			return !layer.getFeatureSource().getFeatures(filter).isEmpty();
		} catch (Exception err) {
			return false;
		}
	}

	/**
	 * Ermittelt alle Features, die einen Filter erfuellen. Beim Modus
	 * {@link #SELECT_TOP} wird nur das oberste sichtbare Layer durchsucht. Beim
	 * Modus {@link #SELECT_ALL} werden alle sichtbaren Layer durchsucht.
	 * 
	 * 17.4.08, Stefan
	 * 
	 * @param iterator
	 *            Filter
	 * @param mode
	 *            Suchmodus
	 * @return eine leere {@link Hashtable} falls der Filter {@code null} ist
	 */
	protected Hashtable<MapLayer, FeatureCollection<SimpleFeatureType, SimpleFeature>> findFeatures(
			GeomFilterGenerator filterGenerator, int mode, Envelope env) {
		Hashtable<MapLayer, FeatureCollection<SimpleFeatureType, SimpleFeature>> result = new Hashtable<MapLayer, FeatureCollection<SimpleFeatureType, SimpleFeature>>();
		if (filterGenerator == null)
			return result;

		// Je nach Modus: Alle oder nur das oberste Layer
		MapContext context = getMapContext();
		MapLayer[] layerList = context.getLayers();
		for (int i = layerList.length - 1; i >= 0; i--) {
			MapLayer layer = layerList[i];
			if (!layer.isVisible())
				continue;

			// Bei einem Raster-Layer, das die BB schneidet, abbrechen, wenn nur
			// im
			// obersten (sichtbaren) Layer gesucht
			// wird.AbstractGridCoverage2DReader
			// Ansonsten Raster-Layer einfach uebergehen.
			if (isGridCoverageLayer(layer)) {
				if ((mode == SELECT_TOP || mode == SELECT_ONE_FROM_TOP)
						&& layerIntersectsEnvelope(layer, env))
					break;
				continue;
			}

			// Filter an Layer koppeln
			// WICHTIG: Dies darf erst geschehen, NACHDEM das
			// Schleifen-Abbruch-Kriterium
			// fuer die Raster-Layer geprueft wurde!!
			// Werden folgende Zeilen naemlich auf das FeatureSource des
			// Raster-Layers angewandt, dann "bricht" der Filter "irgendwie"
			// zusammen und auch die ZUVOR gefilterten FeatureCollections,
			// sind ploetzlich EMPTY!!!!!!!!!!!!!!!!!!!!!!!!!!!
			final FeatureSource<SimpleFeatureType, SimpleFeature> featureSource = (FeatureSource<SimpleFeatureType, SimpleFeature>) layer
					.getFeatureSource();
			final GeometryFilterImpl filter = filterGenerator
					.adaptFilter(featureSource);

			try {
				// Filter auf Layer des Features anwenden
				// FeatureCollection fc = layer.getFeatureSource().getFeatures(
				// FilterUtil.cloneFilter(filter) ); // KLAPPT (NOCH) NICHT
				FeatureCollection<SimpleFeatureType, SimpleFeature> fc = featureSource
						.getFeatures(filter);

				// Liefert eine FeatureCollection zurück, in welcher nur
				// Features enthalten sind, welche bei der aktuellen
				// Anzeigeskala aufgrund des Styles gerendert werden.
				fc = StylingUtil.filterSLDVisibleOnly(fc, layer.getStyle(),
						this);

				if (!fc.isEmpty()) {
					result.put(layer, fc);
					// Beim Modus "oberstes Layer selektieren" die Schleife
					// jetzt beenden, da wir sichtbare Features gefunden haben.
					if (mode == SELECT_TOP || mode == SELECT_ONE_FROM_TOP)
						break;
				}
			} catch (Exception err) {
				LOGGER.error("Looking for features:", err);
			}

		}

		return result;
	}

	/**
	 * Ermittelt alle Raster-Werte, die an einer bestimmten Geo-Position liegen.
	 * Beim Modus {@link #SELECT_TOP} oder {@link #SELECT_ONE_FROM_TOP} wird nur
	 * das oberste sichtbare Layer durchsucht. Beim Modus {@link #SELECT_ALL}
	 * werden alle sichtbaren Layer durchsucht.
	 * <p>
	 * SK: 28.09.2007 Da ein Rasterlayer auch mehrere Baender haben kann, ist es
	 * sinnvoll, nicht <code>Hashtable MapLayer,Double </code> sondern
	 * <code>Hashtable MapLayer,Double[] </code> zurueckzugeben.
	 * 
	 * @param point
	 *            Geo-Referenz
	 * @param mode
	 *            Suchmodus
	 * @return eine leere {@link Hashtable} falls keine Werte ermittelt werden
	 *         konnten
	 * 
	 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
	 *         (University of Bonn/Germany)
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	protected Hashtable<MapLayer, double[]> findGridCoverageValues(
			Point2D point, int mode) {
		Hashtable<MapLayer, double[]> result = new Hashtable<MapLayer, double[]>();

		if (point == null)
			return result;

		MapContext context = getMapContext();
		// Je nach Modus: Alle oder nur das oberste Layer
		MapLayer[] layerList = context.getLayers();
		for (int i = layerList.length - 1; i >= 0; i--) {
			MapLayer layer = layerList[i];
			if (!layer.isVisible())
				continue;
			Object layerObj = FeatureUtil
					.getWrappedGeoObject((FeatureSource<SimpleFeatureType, SimpleFeature>) layer
							.getFeatureSource());

			// Bei einem Nicht-Raster-Layer, das den Punkt beinhaltet,
			// abbrechen, wenn nur im
			// obersten (sichtbaren) Layer gesucht wird.
			// Ansonsten Nicht-Raster-Layer einfach uebergehen.
			// SK 29.9.07: Ein AbstractGridCoverage2DReader ist auch ein Raster
			if (!(layerObj instanceof GridCoverage2D || layerObj instanceof AbstractGridCoverage2DReader)) {
				final Envelope pointAsEnvelope = new Envelope(point.getX(),
						point.getX(), point.getY(), point.getY());
				if ((mode == SELECT_TOP || mode == SELECT_ONE_FROM_TOP)
						&& featureLayerIntersectsEnvelope(layer,
								pointAsEnvelope)) {
				}
				continue;
			}

			GridCoverage2D sourceGrid;

			if (layerObj instanceof AbstractGridCoverage2DReader) {
				// LOGGER.info("layerObj instanceof AbstractGridCoverage2DReader"
				// );
				AbstractGridCoverage2DReader reader = (AbstractGridCoverage2DReader) layerObj;

				Parameter readGG = new Parameter(
						AbstractGridFormat.READ_GRIDGEOMETRY2D);

				ReferencedEnvelope mapExtend = new org.geotools.geometry.jts.ReferencedEnvelope(
						mapArea, context.getCoordinateReferenceSystem());

				readGG.setValue(new GridGeometry2D(new GeneralGridEnvelope(
						getBounds()), mapExtend));

				try {
					sourceGrid = (GridCoverage2D) reader
							.read(new GeneralParameterValue[] { readGG });
				} catch (Exception e) {
					LOGGER.error(
							"read(new GeneralParameterValue[] { readGG })", e);
					continue;
				}
			} else {
				// Ein instanceof ist nicht noetig, da sonst schon break oder
				// continue aufgerufen worden waere
				sourceGrid = (GridCoverage2D) layerObj;
			}

			// getNumSampleDimensions gibt die Anzahl der Baender des Rasters
			// zurueck.
			double[] values = new double[sourceGrid.getNumSampleDimensions()];

			try {
				// Grid an Geo-Position auswerten
				if (!CRS.equalsIgnoreMetadata(getMapContext()
						.getCoordinateReferenceSystem(), sourceGrid
						.getCoordinateReferenceSystem())) {
					Coordinate pointInLayerCRS = JTSUtil.transformCoordinate(
							new Coordinate(point.getX(), point.getY()),
							getMapContext().getCoordinateReferenceSystem(),
							sourceGrid.getCoordinateReferenceSystem());
					sourceGrid.evaluate(new Point.Double(pointInLayerCRS.x,
							pointInLayerCRS.y), values);
				} else
					sourceGrid.evaluate(point, values);
			} catch (CannotEvaluateException err) {
				// z.B. Punkt ausserhalb des Rasters --> Layer uebergehen
				continue;
			} catch (Exception e) {
				LOGGER.error("sourceGrid.evaluate(point, values);", e);
				continue;
			}

			result.put(layer, values);
			// Beim Modus "oberstes Layer selektieren" die Schleife beenden
			if (mode == SELECT_TOP || mode == SELECT_ONE_FROM_TOP)
				break;
		}
		return result;
	}

	/**
	 * Propagiert ein Ereignis an alle angeschlossenen Listener.
	 * 
	 * @param e
	 *            Ereignis
	 */
	public void fireMapPaneEvent(MapPaneEvent e) {
		for (JMapPaneListener l : mapPaneListeners)
			l.performMapPaneEvent(e);
	}

	/**
	 * Verarbeitet die Selektion eines Karten-Ausschnitts. Erzeugt immer ein
	 * {@link GeneralSelectionEvent} fuer den ausgewaehlten Bereich. Wurden
	 * Features oder Raster selektiert, werden zudem
	 * {@link FeatureSelectedEvent FeatureSelectedEvents} (bzw.
	 * GridCoverageSelectedEvents GridCoverageSelectedEvents) ausgeloest.
	 * 
	 * @param ox
	 *            X-Koordinate der VON-Position
	 * @param oy
	 *            Y-Koordinate der VON-Position
	 * @param px
	 *            X-Koordinate der BIS-Position
	 * @param py
	 *            Y-Koordinate der BIS-Position
	 * 
	 * @param selectionMode
	 *            on of SELECT_TOP, SELECT_ONE_FROM_TOP, SELECT_ALL
	 */
	public void performSelectionEvent(int ox, int oy, int px, int py,
			int selectionMode) {

		if (getMapContext().getLayerCount() == 0)
			return;

		// Fenster-Koordinaten in Map-Koordinaten umwandeln
		Envelope env = tranformWindowToGeo(ox, oy, px, py);

		// Generelles Event ausloesen
		fireMapPaneEvent(new GeneralSelectionEvent(this, env));

		switch (selectionMode) {
		case SELECT_TOP:
		case SELECT_ONE_FROM_TOP:
		case SELECT_ALL: // Features selektieren

			boolean featuresFound = findFeaturesAndFireEvents(
					new BoundingBoxFilterGenerator(env, getMapContext()
							.getCoordinateReferenceSystem()), selectionMode,
					env);
			if (selectionMode == SELECT_ALL || !featuresFound)
				findGridCoverageSubsetsAndFireEvents(env, selectionMode);
			break;
		}
	}

	/**
	 * Ermittelt alle Features, die in einem Bereich liegen und erzeugt
	 * entsprechende {@link FeatureSelectedEvent FeatureSelectedEvents}. Beim
	 * Modus {@link #SELECT_TOP} wird nur das oberste sichtbare Layer
	 * durchsucht. Beim Modus {@link #SELECT_ALL} werden alle sichtbaren Layer
	 * durchsucht.
	 * 
	 * @param filterGenerator
	 *            adapts a filter to a concrete {@link FeatureSource}
	 * @param selectionMode
	 *            Suchmodus
	 * @param env
	 *            Bereich der durchsucht wird (fuer das Filtern irrelevant; wird
	 *            nur fuer Events benoetigt!)
	 */
	public boolean findFeaturesAndFireEvents(
			GeomFilterGenerator filterGenerator, int selectionMode, Envelope env) {
		Hashtable<MapLayer, FeatureCollection<SimpleFeatureType, SimpleFeature>> result = findVisibleFeatures(
				filterGenerator, selectionMode, env);

		// Events ausloesen fuer jedes Layer
		for (final Enumeration<MapLayer> e = result.keys(); e.hasMoreElements();) {

			final MapLayer layer = e.nextElement();
			
			System.out.println("cheing in "+layer.getTitle());

			final FeatureCollection<SimpleFeatureType, SimpleFeature> fc = result
					.get(layer);
			if (fc != null && !fc.isEmpty())
				fireMapPaneEvent(new FeatureSelectedEvent(this, layer, env, fc));
		}
		return !result.isEmpty();
	}

	/**
	 * Ermittelt alle Teil-Raster, die in einem Bereich liegen.
	 * BefindFeaturesAndFireEventsim Modus {@link #SELECT_TOP} wird nur das
	 * oberste sichtbare Layer durchsucht. Beim Modus {@link #SELECT_ALL} werden
	 * alle sichtbaren Layer durchsucht.
	 * 
	 * @param env
	 *            Bounding-Box
	 * @param selectionMode
	 *            Suchmodus
	 * @return eine leere {@link Hashtable} falls die Bounding-Box {@code null}
	 *         ist
	 */
	protected Hashtable<MapLayer, GridCoverage2D> findGridCoverageSubsets(
			Envelope env, int selectionMode) {
		Hashtable<MapLayer, GridCoverage2D> result = new Hashtable<MapLayer, GridCoverage2D>();
		if (env == null)
			return result;

		MapContext context = getMapContext();
		// Je nach Modus: Alle oder nur das oberste Layer
		MapLayer[] layerList = context.getLayers();
		for (int i = layerList.length - 1; i >= 0; i--) {
			MapLayer layer = layerList[i];
			Object layerObj = FeatureUtil
					.getWrappedGeoObject((FeatureSource<SimpleFeatureType, SimpleFeature>) layer
							.getFeatureSource());
			if (!layer.isVisible())
				continue;

			// Bei einem Nicht-Raster-Layer, das die BB schneidet, abbrechen,
			// wenn nur im obersten (sichtbaren) Layer gesucht wird.
			// Ansonsten Nicht-Raster-Layer einfach uebergehen.
			if (!(layerObj instanceof GridCoverage2D)) {
				if ((selectionMode == SELECT_TOP || selectionMode == SELECT_ONE_FROM_TOP)
						&& featureLayerIntersectsEnvelope(layer, env))
					break;
				continue;
			}

			GridCoverage2D sourceGrid = (GridCoverage2D) layerObj;
			com.vividsolutions.jts.geom.Envelope jtsEnv = env;
			org.geotools.geometry.Envelope2D gtEnv2D = JTS.getEnvelope2D(
					jtsEnv, sourceGrid.getCoordinateReferenceSystem());
			org.opengis.geometry.Envelope gtGenEnv = new GeneralEnvelope(
					(org.opengis.geometry.Envelope) gtEnv2D); // gt2-2.4.2

			// GridCoverage2D subsetGrid =
			// (GridCoverage2D)Operations.DEFAULT.crop(sourceGrid,gtGenEnv);
			GridCoverage2D subsetGrid = GridUtil.createGridCoverage(sourceGrid,
					gtGenEnv);
			if (subsetGrid != null)
				result.put(layer, subsetGrid);
			// Beim Modus "oberstes Layer selektieren" die Schleife beenden
			if (selectionMode == SELECT_TOP || selectionMode == SELECT_ONE_FROM_TOP)
				break;
		}
		return result;
	}

	/**
	 * Ermittelt alle Teil-Raster, die in einem Bereich liegen und erzeugt
	 * entsprechende {@link GridCoverageSelectedEvent
	 * GridCoverageSelectedEvents}. Beim Modus {@link #SELECT_TOP} wird nur das
	 * oberste sichtbare Layer durchsucht. Beim Modus {@link #SELECT_ALL} werden
	 * alle sichtbaren Layer durchsucht.
	 * 
	 * @param env
	 *            Bounding-Box
	 * @param mode
	 *            Suchmodus
	 * @return eine leere {@link Hashtable} falls die Bounding-Box {@code null}
	 *         ist
	 */
	public boolean findGridCoverageSubsetsAndFireEvents(final Envelope env,
			final int mode) {
		final Hashtable<MapLayer, GridCoverage2D> result = findGridCoverageSubsets(
				env, mode);
		// Events ausloesen fuer jedes Layer
		for (final Enumeration<MapLayer> e = result.keys(); e.hasMoreElements();) {
			final MapLayer layer = e.nextElement();
			final GridCoverage2D subset = result.get(layer);
			if (subset != null)
				fireMapPaneEvent(new GridCoverageSelectedEvent(this, layer,
						env, subset));
		}
		return !result.isEmpty();
	}

	/**
	 * Ermittelt alle sichtbaren Features, die einen Filter erfuellen. Beim
	 * Modus {@link #SELECT_TOP} wird nur das oberste sichtbare Layer
	 * durchsucht. Beim Modus {@link #SELECT_ALL} werden alle sichtbaren Layer
	 * durchsucht.<br/>
	 * In mode SELECT_ONE_FROM_TOP, the method will iterator over all non-raster
	 * {@link MapLayer}s and the first layer returning
	 * 
	 * @see #findFeatures(GeometryFilterImpl, int, Envelope)
	 * 
	 * @param filterGenerator
	 *            adapts the filter to a concrete FeatureSource
	 * @param selectionMode
	 *            Suchmodus
	 * @return eine leere {@link Hashtable} falls der Filter {@code null} ist
	 */
	public Hashtable<MapLayer, FeatureCollection<SimpleFeatureType, SimpleFeature>> findVisibleFeatures(
			GeomFilterGenerator filterGenerator, int selectionMode, Envelope env) {
		Hashtable<MapLayer, FeatureCollection<SimpleFeatureType, SimpleFeature>> result = new Hashtable<MapLayer, FeatureCollection<SimpleFeatureType, SimpleFeature>>();
		if (filterGenerator == null)
			throw new IllegalArgumentException(
					"filterGenerator may not be null");

		// Je nach Modus: Alle oder nur das oberste Layer
		MapContext context = getMapContext();
		MapLayer[] layerList = context.getLayers();
		for (int i = layerList.length - 1; i >= 0; i--) {
			MapLayer layer = layerList[i];
			System.out.println("searching in "+layer.getTitle());
			if (!layer.isVisible()) {
				System.out.println("   not visible!");
				continue;
			}

			
			// This should never happen, because the check should be performed
			// earlier already
			if (!isMapLayerSelectable(layer)) {
				LOGGER.debug("Ignoring layer " + layer.getTitle()
						+ " because it is not declared as selectable!");
				continue;
			}

			// LOGGER.debug("mode = " + mode);

			// Bei einem Raster-Layer, das die BB schneidet, abbrechen, wenn nur
			// im obersten (sichtbaren) Layer gesucht wird.
			// Ansonsten Raster-Layer einfach uebergehen.
			if (isGridCoverageLayer(layer)) {
				if ((selectionMode == SELECT_TOP || selectionMode == SELECT_ONE_FROM_TOP)
						&& layerIntersectsEnvelope(layer, env))
					break;
				continue;
			}

			// Filter an Layer koppeln
			// WICHTIG: Dies darf erst geschehen, NACHDEM das
			// Schleifen-Abbruch-Kriterium
			// fuer die Raster-Layer geprueft wurde!!
			// Werden folgende Zeilen naemlich auf das FeatureSource des
			// Raster-Layers angewandt, dann "bricht" der Filter "irgendwie"
			// zusammen und auch die ZUVOR gefilterten FeatureCollections,
			// sind ploetzlich EMPTY!!!!!!!!!!!!!!!!!!!!!!!!!!!
			final FeatureSource<SimpleFeatureType, SimpleFeature> featureSource = (FeatureSource<SimpleFeatureType, SimpleFeature>) layer
					.getFeatureSource();

			final GeometryFilterImpl generatedFilter = filterGenerator
					.adaptFilter(featureSource);

			/**
			 * If the layer has already set a filter in its query, we have to
			 * combine them.
			 */
			Filter filter = generatedFilter;
			if (layer.getQuery() != null) {
				final Filter layerFilter = layer.getQuery().getFilter();
				if (layerFilter != null && !layerFilter.equals(Filter.INCLUDE)) {
					filter = generatedFilter.and(layerFilter);
				}
			}

			try {
				FeatureCollection<SimpleFeatureType, SimpleFeature> fc = featureSource
						.getFeatures(filter);

				// If not empty, filter it for visbile features only
				if (!fc.isEmpty()) {
					fc = StylingUtil.filterSLDVisibleOnly(fc, layer.getStyle(),
							this);

					if (!fc.isEmpty()) {
						result.put(layer, fc);
						// Beim Modus "oberstes Layer selektieren" die Schleife
						// beenden
						if (selectionMode == SELECT_TOP || selectionMode == SELECT_ONE_FROM_TOP)
							break;
					}

				}
			} catch (IOException err) {
				LOGGER.error("applying the distanceWithin filter ", err);
			}
		}
		return result;
	}

	/**
	 * Testet (anhand der Bounding-Box), ob das Objekt eines Layers eine andere
	 * Bounding-Box schneidet. Die Bounding-Box des Layer-Objekts wird zunaechst
	 * in das CRS des MapPanes umgerechnets.<br>
	 * If the layer's bbox is <code>null</code>, returns <code>false</code>. *
	 * 
	 * @param layer
	 *            ein Layer
	 * @param env
	 *            Bounding-Box in CRS des MapPane
	 * 
	 *            TODO movve to some utility class
	 */
	public boolean layerIntersectsEnvelope(MapLayer layer, Envelope env) {
		try {
			// BB des Layers umrechnen von Layer-CRS in Map-CRS
			Envelope bounds_MapCRS = JTSUtil.transformEnvelope(layer
					.getFeatureSource().getBounds(), layer.getFeatureSource()
					.getSchema().getGeometryDescriptor()
					.getCoordinateReferenceSystem(), getMapContext()
					.getCoordinateReferenceSystem());

			// TODO warum kann bounds_MapCRS == null sein ?? ?SK fragt martin???
			if (bounds_MapCRS == null)
				return false;

			return bounds_MapCRS.intersects(env);
		} catch (Exception err) {
			return false;
		}
	}

	/**
	 * Testet (anhand der Bounding-Box), ob das Objekt eines Layers eine
	 * Coordinate schneidet. Die Bounding-Box des Layer-Objekts wird zunaechst
	 * in das CRS des MapPanes umgerechnets.<br/>
	 * If the layer's bbox is <code>null</code>, returns <code>false</code>.
	 * 
	 * @param layer
	 *            ein Layer
	 * @param coord
	 *            Coordinate in CRS des MapPane
	 * 
	 *            TODO movve to some utility class
	 */
	public boolean layerIntersectsCoord(MapLayer layer, Coordinate coord) {
		try {
			ReferencedEnvelope bounds = layer.getFeatureSource().getBounds();

			// Transform the layer bounds to map CRS if neeed
			CoordinateReferenceSystem layerCRS = layer.getFeatureSource()
					.getSchema().getGeometryDescriptor()
					.getCoordinateReferenceSystem();
			CoordinateReferenceSystem mapCRS = getMapContext()
					.getCoordinateReferenceSystem();
			if (!CRS.equalsIgnoreMetadata(layerCRS, mapCRS)) {

				bounds = JTSUtil.transformEnvelope(bounds, mapCRS);
			}

			if (bounds == null)
				return false;

			return bounds.intersects(coord);
		} catch (Exception err) {
			return false;
		}
	}

	/**
	 * Prueft, ob es sich bei einem Layer um ein Raster-Layer handelt. VOn SK an
	 * eine GT Methode abgegeben.
	 * 
	 * @param layer
	 *            zu ueberpruefendes Layer
	 */
	public static boolean isGridCoverageLayer(MapLayer layer) {
		String GRID_PACKAGE = "org.geotools.coverage.grid";
		String GRID_PACKAGE2 = "org.opengis.coverage.grid";

		// TODO when my patch is accepted, replace the whole method with
		// MapLayerUtil.isGridLayer again

		Collection<PropertyDescriptor> descriptors = layer.getFeatureSource()
				.getSchema().getDescriptors();
		for (PropertyDescriptor desc : descriptors) {
			String className = desc.getType().getBinding().getName();

			if (className.contains(GRID_PACKAGE)
					|| className.contains(GRID_PACKAGE2)) {
				return true;
			}
		}

		return false;
	}

}
