/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.gui;

import java.awt.Color;
import java.awt.Dimension;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import javax.swing.JLabel;
import javax.swing.JToolBar;

import org.apache.log4j.Logger;
import org.geotools.styling.Style;
import org.geotools.styling.StyleBuilder;

import schmitzm.geotools.feature.FeatureUtil;
import schmitzm.swing.ColorInputOption;
import schmitzm.swing.InputOption;
import schmitzm.swing.SelectionInputOption;
import schmitzm.swing.SwingUtil;
import schmitzm.swing.event.InputOptionListener;

/**
 * A toolbar to define simple parameters of a layer style.
 * <ul>
 *   <li>Fill color</li>
 *   <li>Border color</li>
 *   <li>Border width</li>
 *   <li>Mark style (dot, cross, ...)</li>
 *   <li>Mark size</li>
 * </ul>.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 */
public class StyleToolBar extends JToolBar implements InputOptionListener {
	protected static final Logger LOGGER = Logger.getLogger(JMapEditorToolBar.class.getName());
	
	private static final String RES_PREFIX = StyleToolBar.class.getName();
     
	/** Identifies the "fill color" (value type {@link Color}) for the
	 *  {@link PropertyChangeListener}.
	 *  @see #addPropertyChangeListener(PropertyChangeListener) */
    public static final String FILL_COLOR = "Fill";
    /** Identifies the "border color" (value type {@link Color}) for the
     *  {@link PropertyChangeListener}.
     *  @see #addPropertyChangeListener(PropertyChangeListener) */
    public static final String BORDER_COLOR = "BorderColor";
    /** Identifies the "border width" (value type {@code int}) for the
     *  {@link PropertyChangeListener}.
     *  @see #addPropertyChangeListener(PropertyChangeListener) */
    public static final String BORDER_WIDTH = "BorderWidth";
    /** Identifies the "mark style" (value type {@link String}) for the
     *  {@link PropertyChangeListener}.
     *  @see #addPropertyChangeListener(PropertyChangeListener) */
    public static final String MARK_STYLE = "MarkStyle";
    /** Identifies the "mark size" (value type {@link int}) for the
     *  {@link PropertyChangeListener}.
     *  @see #addPropertyChangeListener(PropertyChangeListener) */
    public static final String MARK_SIZE = "MarkSize";
	
    /** Label next to the fill color {@link ColorInputOption InputOption}. */
    protected JLabel fillColorLabel = null;
    /** {@link InputOption} to define the fill color. */
	protected ColorInputOption fillColor = null;
    /** Label next to the border color {@link ColorInputOption InputOption}. */
    protected JLabel borderColorLabel = null;
    /** {@link InputOption} to define the border color. */
	protected ColorInputOption borderColor = null;
	/** Label next to the border width {@link SelectionInputOption InputOption}. */
	protected JLabel borderWidthLabel = null;
    /** {@link InputOption} to define the border width. */
    protected SelectionInputOption.Combo<Integer> borderWidth = null;
    /** Label next to the mark style and size {@link SelectionInputOption InputOptions}. */
	protected JLabel markLabel = null;
    /** {@link InputOption} to define the mark style. */
	protected SelectionInputOption.Combo<String> markStyle = null;
    /** {@link InputOption} to define the mark size. */
    protected SelectionInputOption.Combo<Integer> markSize = null;

    /**
	 * Creates a new tool bar. Calls {@link #initGUI()}.
	 */
	public StyleToolBar() {
	  super("Defines a style", JToolBar.HORIZONTAL);
	  setFloatable(false);
	  setRollover(true);
	  initGUI();
	}
	
	/**
	 * Creates the GUI.
	 */
	protected void initGUI() {
	  removeAll();
	  
	  //#####  Fill color  #####
	  fillColorLabel = new JLabel( GeotoolsGUIUtil.RESOURCE.getString(RES_PREFIX+".FillColor") );
      fillColor = new ColorInputOption(null,true,Color.RED);
      fillColor.addInputOptionListener(this);
      SwingUtil.setPreferredWidth(fillColor, 40);
      SwingUtil.fixComponentSize(fillColor);
      //##### Border color #####
      borderColorLabel = new JLabel( GeotoolsGUIUtil.RESOURCE.getString(RES_PREFIX+".BorderColor") );
      borderColor = new ColorInputOption(null,true,Color.BLACK);
      borderColor.addInputOptionListener(this);
      SwingUtil.setPreferredWidth(borderColor, 40);
      SwingUtil.fixComponentSize(borderColor);
      //##### Border width #####
      borderWidthLabel = new JLabel( GeotoolsGUIUtil.RESOURCE.getString(RES_PREFIX+".BorderWidth") );
      borderWidth = new SelectionInputOption.Combo<Integer>(
          null,
          true,
          new Integer[] {1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20},
          (Integer)2,
          null
      );
      borderWidth.addInputOptionListener(this);
      SwingUtil.setPreferredWidth(borderWidth, 50);
      SwingUtil.fixComponentSize(borderWidth);
      //##### Mark style #####
      markLabel = new JLabel( GeotoolsGUIUtil.RESOURCE.getString(RES_PREFIX+".PointMark") );
      markStyle = new SelectionInputOption.Combo<String>(
          null,
          true,
          new String[] {
              StyleBuilder.MARK_ARROW, 
              StyleBuilder.MARK_CIRCLE, 
              StyleBuilder.MARK_CROSS,
              StyleBuilder.MARK_SQUARE,
              StyleBuilder.MARK_STAR,
              StyleBuilder.MARK_TRIANGLE,
              StyleBuilder.MARK_X
          },
          StyleBuilder.MARK_CIRCLE,
          new String[] {
              GeotoolsGUIUtil.RESOURCE.getString("org.geotools.styling.StyleBuilder.MARK_ARROW"),
              GeotoolsGUIUtil.RESOURCE.getString("org.geotools.styling.StyleBuilder.MARK_CIRCLE"),
              GeotoolsGUIUtil.RESOURCE.getString("org.geotools.styling.StyleBuilder.MARK_CROSS"),
              GeotoolsGUIUtil.RESOURCE.getString("org.geotools.styling.StyleBuilder.MARK_SQUARE"),
              GeotoolsGUIUtil.RESOURCE.getString("org.geotools.styling.StyleBuilder.MARK_STAR"),
              GeotoolsGUIUtil.RESOURCE.getString("org.geotools.styling.StyleBuilder.MARK_TRIANGLE"),
              GeotoolsGUIUtil.RESOURCE.getString("org.geotools.styling.StyleBuilder.MARK_X")
          }
      );
      markStyle.addInputOptionListener(this);
      SwingUtil.fixComponentSize(markStyle);
      markSize = new SelectionInputOption.Combo<Integer>(
          null,
          true,
          new Integer[] {1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20},
          (Integer)5,
          null
      );
      markSize.addInputOptionListener(this);
      SwingUtil.setPreferredWidth(markSize, 50);
      SwingUtil.fixComponentSize(markSize);

      //##### put the components to the tool bar #####
      add( fillColorLabel );
      addSeparator( new Dimension(2,0) );
      add( fillColor );
      addSeparator( new Dimension(10,0) );
      add( borderColorLabel );
      addSeparator( new Dimension(2,0) );
      add( borderColor );
      addSeparator( new Dimension(10,0) );
      add( borderWidthLabel );
      addSeparator( new Dimension(2,0) );
      add( borderWidth );
      addSeparator( new Dimension(10,0) );
      add( markLabel );
      addSeparator( new Dimension(2,0) );
      add( markStyle );
      addSeparator( new Dimension(2,0) );
      add( markSize );
	}
	
	/**
	 * Creates a point style with the currently set point style, size, fill and
	 * border color.
	 */
	public Style createPointStyle() {
	  return FeatureUtil.createPointStyle(
	      (String)markStyle.getValue(),
	      fillColor.getValue(),
	      borderColor.getValue(),
	      1.0,
	      1.0,
	      ((Number)markSize.getValue()).doubleValue(),
	      0.0
	  );
	}

    /**
     * Creates a line style with the currently set border color and
     * width.
     */
    public Style createLineStyle() {
      return FeatureUtil.createLineStyle(
          borderColor.getValue(),
          ((Number)borderWidth.getValue()).doubleValue()
      );
    }

    /**
     * Creates a polygon style with the currently set fill, border color and
     * width.
     */
    public Style createPolygonStyle() {
      return FeatureUtil.createPolygonStyle(
          fillColor.getValue(),
          borderColor.getValue(),
          ((Number)borderWidth.getValue()).doubleValue()
      );
    }

    /**
     * Called whenever one of the input options of this tool bar
     * has changed its value. Invokes a {@link PropertyChangeEvent}
     * according to the changed input option.
     * @see #FILL_COLOR
     * @see #BORDER_COLOR
     * @see #BORDER_WIDTH
     * @see #MARK_STYLE
     * @see #MARK_SIZE
     */
    public void optionChanged(InputOption option, Object oldValue, Object newValue) {
      String changedProperty = null;
      if ( option == fillColor )
        changedProperty = FILL_COLOR;
      if ( option == borderColor )
        changedProperty = BORDER_COLOR;
      if ( option == borderWidth )
        changedProperty = BORDER_WIDTH;
      if ( option == markStyle )
        changedProperty = MARK_STYLE;
      if ( option == markSize )
        changedProperty = MARK_SIZE;
      
      if ( changedProperty != null )
        firePropertyChange(changedProperty, oldValue, newValue);
    }

    /**
     * Called whenever one of the input options of this tool bar
     * lost focus. Does nothing.
     */
    public void optionLostFocus(InputOption option) {
    }

    /**
     * Called whenever one of the input options of this tool bar
     * gained focus. Does nothing.
     */
    public void optionGainedFocus(InputOption option) {}

}
