/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.io;

import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;

import javax.imageio.ImageIO;

import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.data.FeatureSource;
import org.geotools.data.FeatureStore;
import org.geotools.data.Transaction;
import org.geotools.data.shapefile.ShapefileDataStore;
import org.geotools.feature.FeatureCollection;
import org.geotools.gce.arcgrid.ArcGridWriter;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.simple.SimpleFeatureType;
import org.opengis.metadata.Identifier;
import org.opengis.parameter.GeneralParameterValue;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import schmitzm.data.WritableGridRaster;
import schmitzm.geotools.feature.FeatureCollectionReader;
import schmitzm.geotools.grid.GridUtil;
import schmitzm.io.IOUtil;

/**
 * In dieser Klasse sind Funktionen zum Datenexport von Geo-Daten
 * zusammengefasst.
 * 
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *         (University of Bonn/Germany)
 * @version 1.0
 */
public class GeoExportUtil {
	/**
	 * Diese Methode exportiert eine
	 * {@link org.geotools.feature.FeatureCollection} in das ShapeFile-Format (
	 * <code><i>name</i>.shp <i>name</i>.shx <i>name</i>.dbf</code>).<br>
	 * Baut auf folgenden Geotools-Klassen auf: <code>
	 * <ul>
	 * <li>{@link ShapefileDataStore       org.geotools.data.shapefile.ShapefileDataStore}</li>
   * <li>{@link FeatureSource            org.geotools.data.FeatureSource}</li>
   * <li>{@link FeatureStore             org.geotools.data.FeatureStore}</li>
   * <li>{@link Transaction              org.geotools.data.Transaction}</li>
   * <li>{@link FeatureCollection        org.geotools.feature.FeatureCollection}</li>
   * <li>{@link FeatureCollectionReader  schmitzm.geotools.feature.FeatureCollectionReader}</li>
   * </ul>
   * </code>
	 * 
	 * @param fc
	 *            zu exportierende FeatureCollection
	 * @param outFile
	 *            Dateiname (Basisname)
	 * @throws java.lang.Exception
	 *             bei irgendeinem Fehler
	 */
	public static void writeFeaturesToShapeFile(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc, File outFile)
			throws Exception {
		// DataStore fuer Ausgabe-Datei oeffnen
	    ShapefileDataStore shpStore = null;
	    try {
	      shpStore = new ShapefileDataStore(outFile.toURI().toURL());
	    } catch (IllegalArgumentException err) {
	      // ShapefileDataStore schmeisst eine IllegalArgumentException,
	      // wenn Dateiname nicht ".shp", ".dbf", ... ist
	      // -> Erweiterung anhaengen
	      outFile = IOUtil.appendFileExt(outFile, ".shp");
	      shpStore = new ShapefileDataStore(outFile.toURI().toURL());
	    }
		// shpStore.createSchema(featureReader.getFeatureType());
		shpStore.createSchema(fc.getSchema());
		// FeatureStore aus dem ShapeFile-DataStore ermitteln
		FeatureSource<SimpleFeatureType, SimpleFeature> source = shpStore
				.getFeatureSource();

		// TODO 26 @Martin?! Du hattest hier gecastet... ob das so noch geht?!
		// FeatureStore featureStore = (FeatureStore)source;
		FeatureStore<SimpleFeatureType, SimpleFeature> featureStore = (FeatureStore<SimpleFeatureType, SimpleFeature>) source;

		// Features schreiben
		Transaction transaction = featureStore.getTransaction();
		featureStore.addFeatures(fc);
		transaction.commit();
		transaction.close();
	}

	/**
	 * Diese Methode exportiert ein Raster in eine Datei im
	 * ArcInfoASCII-Grid-Format.<br>
	 * Baut auf folgenden Geotools-Klassen auf: <code>
	 * <ul>
	 *   <li>{@link org.geotools.coverage.grid.GridCoverage2D}</li>
   *   <li>{@link org.geotools.gce.arcgrid.ArcGridWriter}</li>
   * </ul>
   * </code>
	 * 
	 * @param grid
	 *            zu exportierendes Grid
	 * @param outFile
	 *            Ausgabe-Datei
	 * @throws java.lang.Exception
	 *             bei irgendeinem Fehler
	 */
	public static void writeGridToArcInfoASCII(GridCoverage2D grid, File outFile)
			throws Exception {

		ArcGridWriter arcGridWriter = new ArcGridWriter(outFile);
		GeneralParameterValue[] parameters = new GeneralParameterValue[] {};
		arcGridWriter.write(grid, parameters);
		arcGridWriter.dispose();

		// Raster raster = grid.getRenderedImage().getData();
		// Rectangle2D rect = grid.getEnvelope2D();
		// // Geo-Koordinaten
		// double x = rect.getX();
		// double y = rect.getY();
		// // Zellenbreite
		// double cellsize = rect.getWidth() / raster.getWidth();
		// // Exportieren
		// // throw new
		// UnsupportedOperationException("GeoExportUtil.writeGridToArcInfoASCII(.) has still to be breaked from ArcGridRaster dependency!");
		// ArcGridRaster writer = new ArcGridRaster(new PrintWriter( new
		// FileOutputStream(outFile) ));
		// writer.writeRaster(raster, x, y, cellsize, false);

		// Die peojektion wird vom ArcGridWriter automaitsch mit-geschrieben!
		// // Projektion schreiben
		// writeProjectionFile( grid.getCoordinateReferenceSystem(),
		// IOUtil.changeFileExt(outFile,"prj"));
	}

	/**
	 * Diese Methode exportiert ein Raster in eine Datei im GeoTiff-Format.
	 * Daneben wird ein entsprechendes World-File (.tfw) erstellt, in dem die
	 * Georeferenz und Aufloesung des TIF hinterlegt wird.
	 * 
	 * @param grid
	 *            zu exportierendes Grid
	 * @param output
	 *            Ausgabe-Datei
	 * @throws java.lang.Exception
	 *             bei irgendeinem Fehler
	 */
	public static void writeGridToGeoTiff(GridCoverage2D grid, File output)
			throws Exception {
		// for (int i=0; i<ImageIO.getWriterFormatNames().length;i++)
		// System.out.println(ImageIO.getWriterFormatNames()[i]);

		// GeoTiff-File schreiben
		ImageIO.write(grid.getRenderedImage(), "tif", output);
		// World-File schreiben
		writeWorldFile(grid.getRenderedImage().getWidth(), grid
				.getRenderedImage().getHeight(), grid.getEnvelope2D(), IOUtil
				.changeFileExt(output, "tfw"));
		// Projektion schreiben
		writeProjectionFile(grid.getCoordinateReferenceSystem(), IOUtil
				.changeFileExt(output, "prj"));
	}

	/**
	 * Diese Methode exportiert ein Raster in eine Datei im
	 * ArcInfoASCII-Grid-Format.
	 * 
	 * @param grid
	 *            zu exportierendes Grid
	 * @param outFile
	 *            Ausgabe-Datei
	 * @throws java.lang.Exception
	 *             bei irgendeinem Fehler
	 */
	public static void writeGridRasterToArcInfoASCII(WritableGridRaster grid,
			File outFile) throws Exception {
		writeGridToArcInfoASCII(GridUtil.convertToGridCoverage2D(grid), outFile);
	}

	/**
	 * Diese Methode exportiert ein Raster in eine Datei im GeoTiff-Format.
	 * Daneben wird ein entsprechendes World-File (.tfw) erstellt, in dem die
	 * Georeferenz und Aufloesung des TIF hinterlegt wird.
	 * 
	 * @param grid
	 *            zu exportierendes Grid
	 * @param output
	 *            Ausgabe-Datei
	 * @throws java.lang.Exception
	 *             bei irgendeinem Fehler
	 */
	public static void writeGridRasterToGeoTiff(WritableGridRaster grid,
			File output) throws Exception {
		// for (int i=0; i<ImageIO.getWriterFormatNames().length;i++)
		// System.out.println(ImageIO.getWriterFormatNames()[i]);

		// GeoTiff-File schreiben
		BufferedImage im = new BufferedImage(grid.getWidth(), grid.getHeight(),
				BufferedImage.TYPE_INT_RGB);
		im.setData(grid);
		ImageIO.write(im, "tif", output);
		// World-File schreiben
		writeWorldFile(grid.getWidth(), grid.getHeight(), grid.getEnvelope(),
				IOUtil.changeFileExt(output, "tfw"));
		// Projektion schreiben
		writeProjectionFile(grid.getCoordinateReferenceSystem(), IOUtil
				.changeFileExt(output, "prj"));
	}

	/**
	 * Schreibt ein World-File (.tfw) fuer ein Grid. Dabei handelt es sich um
	 * eine zeilenweise ASCII-Datei:<br>
	 * <ol>
	 * <li>Zellengroesse in X-Richtung (in Meter)</li>
	 * <li>Koeffizient fuer Rotation in Y-Richtung (<b>wird mit 0.0
	 * befuellt</b>!)</li>
	 * <li>Koeffizient fuer Rotation in X-Richtung (<b>wird mit 0.0
	 * befuellt</b>!)</li>
	 * <li>negative Zellengroesse in Y-Richtung (in Meter)</li>
	 * <li>Horizontale Geo-Referenz (Longitude) der nord-westlichen Ecke</li>
	 * <li>Vertikale Geo-Referenz (Latitude) der nord-westlichen Ecke</li>
	 * </ol>
	 * 
	 * @param rasterWidth
	 *            Breite des Rasters (in Zellen)
	 * @param rasterHeight
	 *            Hoehe des Rasters (in Zellen)
	 * @param envelope
	 *            Geo-Referenz des Rasters (Position und Ausdehnung)
	 * @param output
	 *            Datei in die das World-File geschrieben wird
	 * @exception IOException
	 *                falls die Datei nicht geschrieben werden kann
	 */
	public static void writeWorldFile(int rasterWidth, int rasterHeight,
			Rectangle2D envelope, File output) throws IOException {
		PrintWriter out = new PrintWriter(output);

		// Zeile 1: Zellengroesse in Meter
		out.println(envelope.getWidth() / rasterWidth);
		// Zeile 2: Rotation in Y-Richtung
		out.println("0.0");
		// Zeile 3: Rotation in X-Richtung
		out.println("0.0");
		// Zeile 4: ?
		out.println(-envelope.getHeight() / rasterHeight);
		// Zeile 5: X-Georeferenz (Longitude) WESTEN
		out.println(envelope.getX());
		// Zeile 6: Y-Georeferenz (Latitude) NORDEN (Im Envelope steht SUEDEN!!)
		out.println(envelope.getY() + envelope.getHeight());

		out.flush();
		out.close();
	}

	/**
	 * Schreibt ein Projektions-File (.prj) fuer ein
	 * {@link CoordinateReferenceSystem}.
	 * 
	 * @param crs
	 *            Koordinaten-System
	 * @param output
	 *            Datei in die die Projektion geschrieben wird
	 * @exception IOException
	 *                falls die Datei nicht geschrieben werden kann
	 */
	public static void writeProjectionFile(CoordinateReferenceSystem crs,
			File output) throws IOException {
		PrintWriter out = new PrintWriter(output);
		out.println(crs.toWKT());
		out.flush();
		out.close();
	}

	/**
	 * Schreibt ein Projektions-File (.prj) fuer ein
	 * {@link CoordinateReferenceSystem}. Wenn möglich wird die Schreibweise
	 * "EPSG:12345" anstelle von WKT verwendet. Wenn die Zieldatei existiert und
	 * bereits den gleichen Inahlt hat, wird Sie nicht verändert (SVN
	 * freundlich).
	 * 
	 * @param crs
	 *            Koordinaten-System
	 * @param outputFile
	 *            Datei in die die Projektion geschrieben wird
	 *            
	 * @exception IOException
	 *                falls die Datei nicht geschrieben werden kann
	 */
	public static void writeProjectionFilePrefereEPSG(
			CoordinateReferenceSystem crs, File outputFile) throws IOException {

		// Directories will be created if needed.
		outputFile.getParentFile().mkdirs();

		String whatToWrite = null;

		/**
		 * LetIf we can determine the EPSG code for this, let's save it as
		 * "EPSG:12345" to the file.
		 */
		if (!crs.getIdentifiers().isEmpty()) {
			Object next = crs.getIdentifiers().iterator().next();
			if (next instanceof Identifier) {
				Identifier identifier = (Identifier) next;
				if (identifier.getAuthority().getTitle().toString().equals(
						"European Petroleum Survey Group")
						|| identifier.toString().startsWith("EPSG:")) {
					whatToWrite = "EPSG:" + identifier.getCode();
				}
			}
		}

		if (whatToWrite == null) {
			// If we don't know the EPSG code for the CRS, save it as WKT
			whatToWrite = crs.toWKT();
		}

		if (outputFile.exists()) {
			String origCont = IOUtil.readFileAsString(outputFile);
			if (origCont.equals(whatToWrite)) {
				// Not writing anything because the file already contains the
				// same stuff.
				return;
			}
		}

		PrintWriter out = new PrintWriter(outputFile);
		try {
			out.println(whatToWrite);
			out.flush();
		} finally {
			out.close();
		}
	}

}
