/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.map.event;

import org.geotools.geometry.jts.ReferencedEnvelope;
import org.geotools.map.MapContext;
import org.geotools.map.event.MapBoundsEvent;
import org.geotools.map.event.MapBoundsListener;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

/**
 * This {@link MapBoundsListener} listens to {@link MapBoundsEvent MapBoundsEvents}
 * and transfers the change of {@link CoordinateReferenceSystem} and "AreaOfInterest"
 * to the connected {@link MapContext}, so this {@link MapContext} has always the
 * same CRS and AreaOfInterest like the {@link MapContex} this Listener is added to.
 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a>
 */
public class MapContextSynchronizer implements MapBoundsListener {
  /** {@link MapContext} the changes by a {@link MapBoundsEvent} are
   *  transfered to. */
  protected MapContext mapContext = null;

  /**
   * Creates a new synchronizer.
   * @param mapContext the {@link MapContext} the changes by a
   *                   {@link MapBoundsEvent} are transfered to
   */
  public MapContextSynchronizer(MapContext mapContext) {
    this.mapContext = mapContext;
  }

  /**
   * Invoked when the CRS or AreaOfInterest of a connected {@link MapContext}
   * changes.
   * @param e an event
   */
  public void mapBoundsChanged(MapBoundsEvent e) {
    if ( mapContext == null )
      return;

    MapContext                eventContext = (MapContext)e.getSource();
    ReferencedEnvelope        eventAOI     = eventContext.getAreaOfInterest();
    CoordinateReferenceSystem eventCRS     = eventContext.getCoordinateReferenceSystem();
    ReferencedEnvelope        currAOI      = mapContext.getAreaOfInterest();
    CoordinateReferenceSystem currCRS      = mapContext.getCoordinateReferenceSystem();

    try {
      // check the CRS for changes
      if ( currCRS == null || !currCRS.equals(eventCRS) )
        mapContext.setCoordinateReferenceSystem( eventCRS );
      // check the AreaOfInterst for changes
      if ( currAOI == null || !currAOI.equals(eventAOI) )
        mapContext.setAreaOfInterest( eventAOI );
    } catch (Exception err) {
      throw new RuntimeException(err);
    }
  }

}
