/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.geotools.styling;

import java.awt.Color;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Array;
import java.net.URL;
import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.Vector;

import javax.measure.unit.Unit;
import javax.xml.transform.TransformerException;

import org.apache.commons.io.IOUtils;
import org.apache.log4j.Logger;
import org.geotools.brewer.color.BrewerPalette;
import org.geotools.brewer.color.PaletteSuitability;
import org.geotools.brewer.color.SampleScheme;
import org.geotools.coverage.Category;
import org.geotools.coverage.GridSampleDimension;
import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.coverage.grid.io.AbstractGridCoverage2DReader;
import org.geotools.data.FeatureSource;
import org.geotools.data.memory.MemoryFeatureCollection;
import org.geotools.data.store.EmptyFeatureCollection;
import org.geotools.factory.CommonFactoryFinder;
import org.geotools.factory.GeoTools;
import org.geotools.feature.FeatureCollection;
import org.geotools.feature.GeometryAttributeType;
import org.geotools.filter.BinaryComparisonAbstract;
import org.geotools.filter.ConstantExpression;
import org.geotools.filter.FilterAttributeExtractor;
import org.geotools.filter.IsBetweenImpl;
import org.geotools.geometry.jts.ReferencedEnvelope;
import org.geotools.renderer.lite.RendererUtilities;
import org.geotools.resources.i18n.Vocabulary;
import org.geotools.resources.i18n.VocabularyKeys;
import org.geotools.styling.ColorMap;
import org.geotools.styling.ColorMapEntry;
import org.geotools.styling.ColorMapEntryImpl;
import org.geotools.styling.ColorMapImpl;
import org.geotools.styling.ExternalGraphic;
import org.geotools.styling.FeatureTypeStyle;
import org.geotools.styling.FeatureTypeStyleImpl;
import org.geotools.styling.Fill;
import org.geotools.styling.Graphic;
import org.geotools.styling.LineSymbolizer;
import org.geotools.styling.Mark;
import org.geotools.styling.PointSymbolizer;
import org.geotools.styling.PolygonSymbolizer;
import org.geotools.styling.RasterSymbolizer;
import org.geotools.styling.RasterSymbolizerImpl;
import org.geotools.styling.Rule;
import org.geotools.styling.RuleImpl;
import org.geotools.styling.SLDParser;
import org.geotools.styling.SLDTransformer;
import org.geotools.styling.Stroke;
import org.geotools.styling.Style;
import org.geotools.styling.StyleBuilder;
import org.geotools.styling.StyleFactory;
import org.geotools.styling.Symbolizer;
import org.geotools.styling.TextSymbolizer;
import org.geotools.styling.visitor.DuplicatingStyleVisitor;
import org.geotools.util.NumberRange;
import org.jdom.Element;
import org.jdom.output.XMLOutputter;
import org.opengis.coverage.grid.Grid;
import org.opengis.coverage.grid.GridCoverage;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.simple.SimpleFeatureType;
import org.opengis.feature.type.Name;
import org.opengis.filter.And;
import org.opengis.filter.Filter;
import org.opengis.filter.FilterFactory;
import org.opengis.filter.FilterFactory2;
import org.opengis.filter.PropertyIsEqualTo;
import org.opengis.filter.expression.Expression;
import org.opengis.filter.expression.Function;
import org.opengis.filter.expression.Literal;
import org.opengis.filter.expression.PropertyName;

import schmitzm.geotools.FilterUtil;
import schmitzm.geotools.feature.FeatureUtil;
import schmitzm.geotools.feature.FeatureUtil.GeometryForm;
import schmitzm.geotools.grid.GridUtil;
import schmitzm.io.IOUtil;
import schmitzm.lang.LangUtil;
import skrueger.geotools.StyledFeaturesInterface;
import skrueger.geotools.StyledRasterInterface;
import skrueger.geotools.XMapPane;
import skrueger.geotools.selection.FeatureMapLayerSelectionSynchronizer;

import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.MultiLineString;
import com.vividsolutions.jts.geom.MultiPoint;
import com.vividsolutions.jts.geom.MultiPolygon;
import com.vividsolutions.jts.geom.Polygon;

/**
 * Diese Klasse enthaelt Hilfsfunktionen zum GeoTools-Styling
 * 
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *         (University of Bonn/Germany)
 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons Kr&uuml;ger</a>
 * 
 * @version 1.1
 */
public class StylingUtil {
	private static final FilterFactory FILTER_FACTORY = FeatureUtil.FILTER_FACTORY;
	private static final Logger LOGGER = Logger.getLogger(StylingUtil.class);
	/** transparent color */
	public static final Color TRANSPARENT_COLOR = new Color(0, 0, 0, 0);

	/** Standard-Instanz eines {@link StyleBuilder}. */
	public static final StyleBuilder STYLE_BUILDER = new StyleBuilder();

	/** Standard-Instanz eines {@link SLDTransformer}. */
	public final static SLDTransformer SLDTRANSFORMER = new SLDTransformer();
	static {
		SLDTRANSFORMER.setEncoding(Charset.defaultCharset());
	}

	/** Standard-Instanz einer {@link StyleFactory} */
	public static final StyleFactory STYLE_FACTORY = CommonFactoryFinder
			.getStyleFactory(GeoTools.getDefaultHints());

	/**
	 * Standard-Instanz einer {@link StyleFactory}
	 * 
	 * @deprecated wurde ersetzt durch {@link #STYLE_FACTORY}
	 */
	public static final StyleFactory styleFactory = STYLE_FACTORY;

	/**
	 * TODO SK The following constants are used to speed up
	 * {@link #createSelectionStyle(Object)}. I will "redo" them soon and also
	 * document them. This is still under construction. TODO SK
	 */

	static final Literal size0 = FeatureUtil.FILTER_FACTORY2.literal(26);

	static final Literal size1 = FeatureUtil.FILTER_FACTORY2.literal(9);
	static final Literal size2 = FeatureUtil.FILTER_FACTORY2.literal(7);
	static final Literal size3 = FeatureUtil.FILTER_FACTORY2.literal(3);
	static final Literal zeroLit = FeatureUtil.FILTER_FACTORY2.literal(0);
	static final Literal halfLit = FeatureUtil.FILTER_FACTORY2.literal(.5);

	static final Graphic SELECTION_GRAPHIC1 = STYLE_FACTORY.createGraphic(
			new ExternalGraphic[0], new Mark[] { STYLE_FACTORY.createMark(
					FeatureUtil.FILTER_FACTORY2.literal("square"), null,
					STYLE_FACTORY.createFill(STYLE_BUILDER
							.colorExpression(Color.WHITE)), size1, zeroLit) },
			new org.geotools.styling.Symbol[0], FeatureUtil.FILTER_FACTORY2
					.literal(1), size1, zeroLit);

	static final Graphic BLINK_GRAPHIC1 = STYLE_FACTORY.createGraphic(
			new ExternalGraphic[0], new Mark[] { STYLE_FACTORY.createMark(
					FeatureUtil.FILTER_FACTORY2.literal("circle"), null,
					STYLE_FACTORY.createFill(STYLE_BUILDER
							.colorExpression(Color.WHITE)), size0, halfLit) },
			new org.geotools.styling.Symbol[0], FeatureUtil.FILTER_FACTORY2
					.literal(1), size0, zeroLit);

	static final Graphic SELECTION_GRAPHIC2 = STYLE_FACTORY.createGraphic(
			new ExternalGraphic[0], new Mark[] { STYLE_FACTORY.createMark(
					FeatureUtil.FILTER_FACTORY2.literal("circle"), null,
					STYLE_FACTORY.createFill(STYLE_BUILDER
							.colorExpression(Color.BLACK)), size2, zeroLit) },
			new org.geotools.styling.Symbol[0], FeatureUtil.FILTER_FACTORY2
					.literal(1), size2, zeroLit);

	static final Graphic SELECTION_GRAPHIC3 = STYLE_FACTORY.createGraphic(
			new ExternalGraphic[0], new Mark[] { STYLE_FACTORY.createMark(
					FeatureUtil.FILTER_FACTORY2.literal("circle"), null,
					STYLE_FACTORY.createFill(STYLE_BUILDER
							.colorExpression(Color.YELLOW.brighter())), size3,
					zeroLit) }, new org.geotools.styling.Symbol[0],
			FeatureUtil.FILTER_FACTORY2.literal(1), size3, zeroLit);

	/**
	 * Erstellt einen Default-Style fuer die Klassen/Interfaces:
	 * StyledFeaturesInterface, GridCoverage2D, FeatureCollection, FeatureSource
	 * und GeometryAttributeType.
	 * 
	 * @param object
	 *            {@link GridCoverage2D},
	 *            {@link org.geotools.coverage.grid.io.AbstractGridCoverage2DReader}
	 *            oder {@link FeatureCollection}
	 * @return {@code null} falls kein Style generiert werden kann
	 * 
	 */
	public static Style createDefaultStyle(Object object) {
		Style style = STYLE_BUILDER.createStyle(); // SK.. nicer default than
		// null !

		if (object instanceof StyledFeaturesInterface)
			style = FeatureUtil
					.createDefaultStyle(((StyledFeaturesInterface<?>) object)
							.getSchema().getGeometryDescriptor());

		if (object instanceof GridCoverage2D
				|| object instanceof AbstractGridCoverage2DReader
				|| object instanceof StyledRasterInterface)
			style = GridUtil.createDefaultStyle();

		if (object instanceof FeatureCollection)
			style = FeatureUtil
					.createDefaultStyle((FeatureCollection<SimpleFeatureType, SimpleFeature>) object);

		if (object instanceof FeatureSource)
			style = FeatureUtil.createDefaultStyle(((FeatureSource) object)
					.getSchema().getGeometryDescriptor());

		if (object instanceof GeometryAttributeType)
			style = FeatureUtil
					.createDefaultStyle((GeometryAttributeType) object);

		return style;
	}

	/**
	 * Erzeugt eine {@link Category} fuer die NoData-Werte transparent
	 * dargestellt werden.
	 * 
	 * @param geoValue
	 *            Geo-Wert, der NoData repraesentiert
	 */
	public static Category createNoDataCategory(double geoValue) {
		return createNoDataCategory(0, geoValue);
	}

	/**
	 * Erzeugt eine {@link Category} fuer die NoData-Werte transparent
	 * dargestellt werden.
	 * 
	 * @param value
	 *            Sample-Wert der Category
	 * @param geoValue
	 *            Geo-Wert, der NoData repraesentiert
	 */
	public static Category createNoDataCategory(int value, double geoValue) {
		return new Category(Vocabulary
				.formatInternational(VocabularyKeys.NODATA),
				new Color[] { new Color(0, 0, 0, 0) }, new NumberRange(value,
						value), new NumberRange(geoValue, geoValue));
	}

	/**
	 * Erzeugt eine {@link GridSampleDimension} aus einer {@link ColorMap}.
	 * 
	 * @param name
	 *            Name fuer die {@link GridSampleDimension}
	 * @param colorMap
	 *            fuer jeden Eintrag der {@link ColorMap} wird eine
	 *            {@link Category} erzeugt
	 * @param noDataValues
	 *            Werte fuer die zusaetzliche {@link Category Categorys}
	 *            erstellt werden, welche die transparent dargestellt werden
	 *            (wenn {@code null} wird keine zusaetzliche {@link Category}
	 *            erstellt)
	 * @param unit
	 *            Einheit der Werte in der {@link GridSampleDimension} (kann
	 *            {@code null} sein)
	 */
	public static GridSampleDimension createDiscreteGridSampleDimension(
			String name, ColorMap colorMap, double[] noDataValues, Unit<?> unit) {
		if (colorMap.getColorMapEntries().length == 0)
			return null;

		SortedMap<Integer, Category> categories = new TreeMap<Integer, Category>();
		if (noDataValues == null)
			noDataValues = new double[0];

		// Create a (discrete) Category for each ColorMapEntry
		int colorMapMin = Integer.MAX_VALUE;
		int colorMapMax = Integer.MIN_VALUE;
		for (ColorMapEntry cme : colorMap.getColorMapEntries()) {
			String label = cme.getLabel();
			Color color = getColorFromColorMapEntry(cme);
			double value = getQuantityFromColorMapEntry(cme);
			int intValue = (int) Math.round(value);
			colorMapMin = Math.min(colorMapMin, intValue);
			colorMapMax = Math.max(colorMapMax, intValue);
			// if ( label == null || label.trim().equals("") )
			// label = String.valueOf(intValue);
			if (label == null)
				label = "";
			categories.put(intValue, new Category(label, new Color[] { color },
					new NumberRange(intValue, intValue), new NumberRange(
							intValue, intValue)));
		}

		// If NoData-Value is set, create an additional Category
		for (double noDataValue : noDataValues) {
			int value = categories.lastKey() + 10; // value not already used
			categories.put(value, createNoDataCategory(value, noDataValue));
			colorMapMin = Math.min(colorMapMin, (int) noDataValue);
			colorMapMax = Math.max(colorMapMax, (int) noDataValue);
		}

		// Declare "all" values smaller/greater the color map values
		// automatically as NoData
		// int colorMapRange = colorMapMax - colorMapMin;
		// double lowerStart = colorMapMin - 10000*colorMapRange;
		// double higherEnd = colorMapMax + 10000*colorMapRange;
		final int lowerStart = Integer.MIN_VALUE;
		final int higherEnd = Integer.MAX_VALUE;
		categories.put(colorMapMax + 10, new Category("_autoNoData1_",
				new Color[] { new Color(0, 0, 0, 0) }, new NumberRange(
						colorMapMax + 1, colorMapMax + 2), new NumberRange(
						colorMapMax + 1, higherEnd)));
		if (colorMapMin < colorMapMax)
			categories.put(colorMapMin - 10, new Category("_autoNoData2_",
					new Color[] { new Color(0, 0, 0, 0) },
					// new NumberRange(colorMapMin-2, colorMapMin-1), //
					// logischer, aber klappt nicht!?
					NumberRange.create(colorMapMax + 3, colorMapMax + 4),
					// new NumberRange(colorMapMax + 3, colorMapMax + 4),
					NumberRange.create(lowerStart, colorMapMin - 1)
			// new NumberRange(lowerStart, colorMapMin - 1)
					));

		// Create the GridSampleDimension
		GridSampleDimension gsd = new GridSampleDimension(name, categories
				.values().toArray(new Category[0]), unit).geophysics(true);

		for (Category c : (List<Category>) gsd.getCategories())
			LOGGER.debug(c.toString());

		return gsd;
	}

	/**
	 * Liefert die Farbpalette zu einem Style, wenn der Style aus einem
	 * {@link RasterSymbolizer} erzeugt wurde.
	 * 
	 * @param style
	 *            Style Wenn <code>null</code>, dann wird style auf
	 *            {@link GridUtil}.createDefaultStyle() gesetzt
	 * @return <code>null</code> wenn der Style nicht auf einem
	 *         {@link RasterSymbolizer} basiert.
	 */
	public static ColorMap getColorMapFromStyle(final Style style) {
		final List<ColorMap> colorMapsFromStyle = getColorMapsFromStyle(style);
		if (colorMapsFromStyle.size() > 0)
			return colorMapsFromStyle.get(0);
		else
			return null;

	}

	/**
	 * Liefert alle Farbpaletten aus einem Style, wenn irgendo
	 * {@link RasterSymbolizer} enthalten sind.
	 * 
	 * @param style
	 *            Style Wenn <code>null</code>, dann wird style auf
	 *            {@link GridUtil}.createDefaultStyle() gesetzt
	 * @return Leere Liste wenn keine RasterSymbolizer enthalten ist.
	 */
	public static List<ColorMap> getColorMapsFromStyle(Style style) {

		if (style == null)
			style = GridUtil.createDefaultStyle();

		final List<RasterSymbolizer> rasterSymbolizers = getRasterSymbolizers(style);

		final List<ColorMap> colorMaps = new ArrayList<ColorMap>();

		for (final RasterSymbolizer rs : rasterSymbolizers) {
			if (rs.getColorMap() != null) {
				colorMaps.add(rs.getColorMap());
			}
		}

		return colorMaps;
	}

	/**
	 * Ermittelt fuer einen Wert den {@link ColorMapEntry} aus einer
	 * {@link ColorMap}.
	 * 
	 * @param value
	 *            Wert
	 * @return {@code null}, wenn in der {@link ColorMap} kein expliziter
	 *         Eintrag fuer den Wert hinterlegt ist
	 */
	public static ColorMapEntry findColorMapEntry(ColorMap colorMap,
			double value) {
		for (ColorMapEntry entry : colorMap.getColorMapEntries())
			if (value == getQuantityFromColorMapEntry(entry))
				return entry;
		return null;
	}

	/**
	 * Applies general properties (color map type and extended colors) to a
	 * {@link ColorMap}.
	 * 
	 * @param sourceColorMap
	 *            color map the properties are taken from
	 * @param destColormap
	 *            color map the properties are set
	 */
	public static void applyColorMapProperties(ColorMap sourceColorMap,
			ColorMap destColormap) {
		if (sourceColorMap == null || destColormap == null)
			return;
		destColormap.setExtendedColors(sourceColorMap.getExtendedColors());
		destColormap.setType(sourceColorMap.getType());
	}

	/**
	 * Unless {@link ColorMap} has no methods to remove an entry this method
	 * creates a new {@link ColorMap} with the identical entries except the one
	 * to remove.
	 * 
	 * @param colorMap
	 *            the color map to remove entries from
	 * @param idx
	 *            the entry index to remove
	 */
	public static ColorMap removeColorMapEntry(ColorMap colorMap, int idx) {
		ColorMap newColorMap = new ColorMapImpl();
		applyColorMapProperties(colorMap, newColorMap);

		ColorMapEntry[] entry = colorMap.getColorMapEntries();
		for (int i = 0; i < entry.length; i++)
			if (i != idx)
				newColorMap.addColorMapEntry(entry[i]);

		return newColorMap;
	}

	/**
	 * Unless {@link ColorMap} has no methods to remove an entry this method
	 * creates a new {@link ColorMap} with the identical entries except the one
	 * to remove.
	 * 
	 * @param colorMap
	 *            the color map to remove entries from
	 * @param colorMapEntry
	 *            the entry to remove
	 */
	public static ColorMap removeColorMapEntry(ColorMap colorMap,
			ColorMapEntry colorMapEntry) {
		ColorMap newColorMap = new ColorMapImpl();
		applyColorMapProperties(colorMap, newColorMap);

		ColorMapEntry[] entry = colorMap.getColorMapEntries();
		for (int i = 0; i < entry.length; i++)
			if (entry[i] != colorMapEntry)
				newColorMap.addColorMapEntry(entry[i]);

		return newColorMap;
	}

	/**
	 * Unless {@link ColorMap} is not sorted automatically, this method calls
	 * {@link #sortColorMap(ColorMap)} after inserting the color map entry.
	 * 
	 * @param colorMap
	 *            the color map to extend
	 * @param colorMapEntry
	 *            the new entry
	 */
	public static ColorMap addColorMapEntryAndSort(ColorMap colorMap,
			ColorMapEntry colorMapEntry) {
		colorMap.addColorMapEntry(colorMapEntry);
		return sortColorMap(colorMap);
	}

	/**
	 * Sorts a {@link ColorMap} according to the value of its entries.
	 * 
	 * @param colorMap
	 *            the color map to sort
	 */
	public static ColorMap sortColorMap(ColorMap colorMap) {
		ColorMap newColorMap = new ColorMapImpl();
		applyColorMapProperties(colorMap, newColorMap);

		// Put all color map entries in a sorted map (with
		// collision lists!)
		SortedMap<Double, Vector<ColorMapEntry>> sortedEntries = new TreeMap<Double, Vector<ColorMapEntry>>();
		for (ColorMapEntry entry : colorMap.getColorMapEntries()) {
			double quantity = getQuantityFromColorMapEntry(entry);
			Vector<ColorMapEntry> collisionList = sortedEntries.get(quantity);
			if (collisionList == null) {
				collisionList = new Vector<ColorMapEntry>();
				sortedEntries.put(quantity, collisionList);
			}
			collisionList.add(entry);
		}

		// According to the order of the sorted map, put all
		// color map entries to the new ColorMap
		for (Vector<ColorMapEntry> collisionList : sortedEntries.values())
			for (ColorMapEntry entry : collisionList)
				newColorMap.addColorMapEntry(entry);

		return newColorMap;
	}

	/**
	 * Returns the color map type specified by a given string.
	 * 
	 * @param typeStr
	 *            a string
	 * @see ColorMap#getType()
	 */
	public static int getColorMapType(String typeStr) {
		if (typeStr == null)
			throw new IllegalArgumentException("Unknown color map type: "
					+ typeStr);

		typeStr = typeStr.trim().toLowerCase();
		if (typeStr.equals("ramp") || typeStr.equals("type_ramp"))
			return ColorMap.TYPE_RAMP;
		if (typeStr.equals("intervals") || typeStr.equals("type_intervals"))
			return ColorMap.TYPE_INTERVALS;
		if (typeStr.equals("values") || typeStr.equals("type_values"))
			return ColorMap.TYPE_VALUES;

		throw new IllegalArgumentException("Unknown color map type: " + typeStr);
	}

	/**
	 * Kopiert eine {@link ColorMap}.
	 * 
	 * @return <code>null</code> wenn die uebergebene ColorMap <code>null</code>
	 *         ist
	 */
	public static ColorMap cloneColorMap(ColorMap colorMap) {
		// MS: Leider kommt der DuplicatingStyleVisitor nicht damit klar, dass
		// der SLDParser Value und Opacity als String setzt. Und selbst wenn
		// man dies fixed sind Value und Opacity nach dem Clonen NULL.
		// --> zurueck zu meinem alten Code!! :-)
		// DuplicatingStyleVisitor duplicatingStyleVisitor = new
		// DuplicatingStyleVisitor();
		// colorMap.accept(duplicatingStyleVisitor);
		// ColorMap clone = (ColorMap)duplicatingStyleVisitor.getCopy();
		// return clone;

		if (colorMap == null)
			return null;
		ColorMapEntry[] entry = colorMap.getColorMapEntries();
		ColorMap newColorMap = new ColorMapImpl();
		newColorMap.setType(colorMap.getType());
		for (int i = 0; i < entry.length; i++)
			newColorMap.addColorMapEntry(cloneColorMapEntry(entry[i]));
		return newColorMap;
	}

	/**
	 * Kopiert einen {@link ColorMapEntry}.
	 * 
	 * @return <code>null</code> wenn der uebergebene ColorMapEntry
	 *         <code>null</code> ist
	 */
	public static ColorMapEntry cloneColorMapEntry(ColorMapEntry colorMapEntry) {
		if (colorMapEntry == null)
			return null;

		return createColorMapEntry(colorMapEntry.getLabel(),
				getQuantityFromColorMapEntry(colorMapEntry),
				getColorFromColorMapEntry(colorMapEntry),
				getOpacityFromColorMapEntry(colorMapEntry));
	}

	/**
	 * Prueft, ob zwei {@link ColorMap}-Instanzen gleich sind.
	 * 
	 * @param cm1
	 *            eine Farbpalette
	 * @param cm2
	 *            eine andere Farbpalette
	 */
	public static boolean colorMapsEqual(ColorMap cm1, ColorMap cm2) {
		if (cm1 == null ^ cm2 == null)
			return false;
		if (cm1 == cm2)
			return true;

		ColorMapEntry[] cme1 = cm1.getColorMapEntries();
		ColorMapEntry[] cme2 = cm2.getColorMapEntries();
		if (cme1.length != cme2.length)
			return false;

		// Farbpaletten-Eintraege muessen gleich sein und in gleicher
		// Reihenfolge vorliegen!
		for (int i = 0; i < cme1.length; i++)
			if (!colorMapEntriesEqual(cme1[i], cme2[i]))
				return false;
		return true;
	}

	/**
	 * Prueft, ob zwei {@link ColorMapEntry}-Instanzen gleich sind. Dies ist der
	 * Fall, wenn ihnen der gleiche Wert, die gleiche Farbe, die gleiche
	 * Transparenz und das gleiche Label zugeordnet ist.
	 * 
	 * @param cme1
	 *            ein Farbpaletten-Eintrag
	 * @param cme2
	 *            ein anderer Farbpaletten-Eintrag
	 */
	public static boolean colorMapEntriesEqual(ColorMapEntry cme1,
			ColorMapEntry cme2) {
		if (cme1 == null ^ cme2 == null)
			return false;
		if (cme1 == cme2)
			return true;

		// Label auf Gleichheit testen
		if (cme1.getLabel() == null ^ cme2.getLabel() == null || cme1 != null
				&& !cme1.getLabel().equals(cme2.getLabel()))
			return false;
		// Farbe auf Gleichheit testen
		Color col1 = getColorFromColorMapEntry(cme1);
		Color col2 = getColorFromColorMapEntry(cme2);
		if (col1 == null ^ col2 == null || col1 != null && !col1.equals(col2))
			return false;
		// Wert auf Gleichheit testen
		Double quan1 = getQuantityFromColorMapEntry(cme1);
		Double quan2 = getQuantityFromColorMapEntry(cme2);
		if (quan1 == null ^ quan2 == null || quan1 != null
				&& !quan1.equals(quan2))
			return false;
		// Transparenz auf Gleichheit testen
		Double op1 = getOpacityFromColorMapEntry(cme1);
		Double op2 = getOpacityFromColorMapEntry(cme2);
		if (op1 == null ^ op2 == null || op1 != null && !op1.equals(op2))
			return false;

		return true;
	}

	/**
	 * Erzeugt einen {@link ColorMapEntry}.
	 * 
	 * @param label
	 *            Label fuer den Farbpaletten-Eintrag
	 * @param quantity
	 *            Wert fuer den Farbpaletten-Eintrag
	 * @param color
	 *            Farbe fuer den Farbpaletten-Eintrag
	 * @param opacity
	 *            Transparenz fuer die Farbe des Farbpaletten-Eintrag
	 */
	public static ColorMapEntry createColorMapEntry(String label,
			Double quantity, Color color, double opacity) {
		ColorMapEntry newEntry = new ColorMapEntryImpl();
		try {
			newEntry.setLabel(label);
			newEntry.setColor(STYLE_BUILDER.colorExpression(color));
			newEntry.setOpacity(STYLE_BUILDER.literalExpression(opacity));
			newEntry.setQuantity(STYLE_BUILDER.literalExpression(quantity));
		} catch (Exception err) {
		}
		return newEntry;
	}

	/**
	 * Kopiert einen {@link ColorMapEntry}.
	 * 
	 * @return <code>null</code> wenn der uebergebene ColorMapEntry
	 *         <code>null</code> ist
	 */
	public static ColorMapEntry ColorMapEntry(ColorMapEntry colorMapEntry) {
		DuplicatingStyleVisitor duplicatingStyleVisitor = new DuplicatingStyleVisitor();
		colorMapEntry.accept(duplicatingStyleVisitor);
		return (ColorMapEntry) duplicatingStyleVisitor.getCopy();

		// if (colorMapEntry == null)
		// return null;
		//
		// return createColorMapEntry(colorMapEntry.getLabel(),
		// getQuantityFromColorMapEntry(colorMapEntry),
		// getColorFromColorMapEntry(colorMapEntry),
		// getOpacityFromColorMapEntry(colorMapEntry));
	}

	/**
	 * Liefert den Wert eines Farbpaletten-Eintrag.
	 * 
	 * @param expression
	 *            Expression, die einen String oder einen Double liefert
	 * @return {@code null}, wenn <code>null</code> uebergeben wird
	 */
	public static Double getQuantityFromExpression(Expression expression) {
		// gt2-2.3.4:
		// if ( expression == null || expression.getValue(null) == null )
		// return null
		// Object exprValue = evaluate.getValue(null);
		// gt2-2.3.4:
		if (expression == null || expression.evaluate(null) == null)
			return null;
		Object exprValue = expression.evaluate(null);

		return (exprValue instanceof String) ? Double
				.valueOf((String) exprValue) : ((Number) exprValue)
				.doubleValue();
	}

	/**
	 * Liefert den Wert eines Farbpaletten-Eintrag.
	 * 
	 * @param entry
	 *            Farbpaletten-Eintrag
	 * @return {@code null}, wenn <code>null</code> uebergeben wird
	 */
	public static Double getQuantityFromColorMapEntry(ColorMapEntry entry) {
		if (entry == null)
			return null;
		return getQuantityFromExpression(entry.getQuantity());
	}

	/**
	 * Setzt den Wert eines Farbpaletten-Eintrag.
	 * 
	 * @param entry
	 *            Farbpaletten-Eintrag
	 * @param quantity
	 *            neuer Wert
	 */
	public static void setQuantityForColorMapEntry(ColorMapEntry entry,
			double quantity) {
		if (entry != null)
			entry.setQuantity(STYLE_BUILDER.literalExpression(quantity));
	}

	/**
	 * Liefert die Transparenz eines Farbpaletten-Eintrag.
	 * 
	 * @param expression
	 *            Expression, die einen String oder einen Double liefert
	 * @return 1.0, wenn <code>null</code> uebergeben wird
	 */
	public static Double getOpacityFromExpression(Expression expression) {
		// gt2-2.3.4:
		// if ( expression == null || expression.getValue(null) == null )
		// return 1.0
		// Object exprValue = evaluate.getValue(null);
		// gt2-2.3.4:
		if (expression == null || expression.evaluate(null) == null)
			return 1.0;
		Object exprValue = expression.evaluate(null);

		return (exprValue instanceof String) ? Double
				.valueOf((String) exprValue) : ((Number) exprValue)
				.doubleValue();
	}

	/**
	 * Liefert die Transparenz eines Farbpaletten-Eintrag.
	 * 
	 * @param entry
	 *            Farbpaletten-Eintrag
	 * @return 1.0, wenn <code>null</code> uebergeben wird
	 */
	public static Double getOpacityFromColorMapEntry(ColorMapEntry entry) {
		if (entry == null)
			return 1.0;
		return getOpacityFromExpression(entry.getOpacity());
	}

	/**
	 * Setzt die Transparenz eines Farbpaletten-Eintrag.
	 * 
	 * @param entry
	 *            Farbpaletten-Eintrag
	 * @param opacity
	 *            Transparenzwert
	 */
	public static void setOpacityForColorMapEntry(ColorMapEntry entry,
			double opacity) {
		if (entry != null)
			entry.setOpacity(STYLE_BUILDER.literalExpression(opacity));
	}

	/**
	 * Liefert die Farbe eines Farbpaletten-Eintrag.
	 * 
	 * @param expression
	 *            Expression, die einen String liefert
	 */
	public static Color getColorFromExpression(Expression expression) {
		if (expression == null)
			return null;

		if (expression instanceof ConstantExpression) {
			ConstantExpression a = (ConstantExpression) expression;
			Object obj = a.getValue();
			if (obj instanceof Color) {
				return (Color) obj;
			}
		}

		return Color.decode(expression.toString());

		// Old way Martin did it:
		// return Color.decode((String) expression.evaluate(null));

	}

	/**
	 * Liefert die Farbe eines Farbpaletten-Eintrag.
	 * 
	 * @param entry
	 *            Farbpaletten-Eintrag
	 */
	public static Color getColorFromColorMapEntry(ColorMapEntry entry) {
		// if ( entry == null || entry.getColor() == null ||
		// entry.getColor().getValue(null) == null ) // gt2-2.3.4
		if (entry == null || entry.getColor() == null
				|| entry.getColor().evaluate(null) == null) // gt2-2-4-2
			return null;

		// // Transparenz
		// Double opacity = getOpacityFromColorMapEntry(entry);

		return getColorFromExpression(entry.getColor());
	}

	/**
	 * Setzt die Farbe eines Farbpaletten-Eintrag.
	 * 
	 * @param entry
	 *            Farbpaletten-Eintrag
	 * @param color
	 *            eine Farbe
	 */
	public static void setColorForColorMapEntry(ColorMapEntry entry, Color color) {
		if (entry != null)
			entry.setColor(STYLE_BUILDER.colorExpression(color));
	}

	/**
	 * Erzeugt einen {@link Style} aus einem {@linkplain Element JDOM-Element},
	 * des eine SLD-Definition enthaelt
	 * 
	 * @param element
	 *            Element mit SLD-Definition
	 */
	public static Style createStyleFromSLD(Element element) {
		String xmlDefinition = new XMLOutputter().outputString(element);
		ByteArrayInputStream inputStream = null;
		try {
			inputStream = new ByteArrayInputStream(xmlDefinition.getBytes());
			Style[] style = loadSLD(inputStream);
			return style == null || style.length == 0 ? null : style[0];
		} finally {
			IOUtils.closeQuietly(inputStream);
		}
	}

	/**
	 * Creates a {@link Style} for a {@link Grid} layer from a {@link ColorMap}
	 * 
	 * @param colorMap
	 *            If null, then defaultStyle for Grid will be used
	 * @param name
	 *            The name to give to the Style. null will result in name=
	 *            {@link GridUtil#UNNAMED_RASTER_STYLE_NAME} or
	 *            GridUtil#DEFAULT_RASTER_STYLE_NAME} (if no colorMap is given).
	 * 
	 * @param title
	 *            The Title to give to the Style. null will result in title=
	 *            {@link GridUtil#UNTITLED_RASTER_STYLE_TITLE} or
	 *            {@link GridUtil#DEFAULT_RASTER_STYLE_TITLE} (if no colorMap is
	 *            given)
	 * 
	 * @return Always a {@link Style} that you can be applyed to a
	 *         {@link GridCoverage}.
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Style createStyleFromColorMap(ColorMap colorMap, String name,
			String title) {
		if (colorMap == null)
			return GridUtil.createDefaultStyle();

		if (name == null)
			// A colorMap was provided, but no name.
			name = GridUtil.UNNAMED_RASTER_STYLE_NAME;

		if (title == null)
			// A colorMap was provided, but no title.
			title = GridUtil.UNTITLED_RASTER_STYLE_TITLE;

		RasterSymbolizerImpl rasterSymbolizerImpl = new RasterSymbolizerImpl();

		// Entferne alle label informationen
		clearColorMapLabels(colorMap);

		rasterSymbolizerImpl.setColorMap(colorMap);
		Symbolizer[] symbolizers = { rasterSymbolizerImpl };
		RuleImpl ruleImpl = new RuleImpl(symbolizers) {
		};
		Rule[] rules = { ruleImpl };
		FeatureTypeStyleImpl featureTypeStyleImpl = new FeatureTypeStyleImpl(
				rules) {
		};

		Style style = STYLE_FACTORY.createStyle();
		style.setName(name);
		style.getDescription().setTitle(title);
		style.featureTypeStyles().add(featureTypeStyleImpl);
		return style;
	}

	/**
	 * Removes all label information from the {@link ColorMapEntry}s of the
	 * given {@link ColorMap}
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static ColorMap clearColorMapLabels(ColorMap colorMap) {
		for (ColorMapEntry cme : colorMap.getColorMapEntries()) {
			cme.setLabel(null);
		}
		return colorMap;
	}

	/**
	 * Creates a {@link Style} for a {@link Grid} layer from a {@link ColorMap}
	 * Title will be set to {@link GridUtil#UNTITLED_RASTER_STYLE_TITLE}
	 * 
	 * @param colorMap
	 *            If null, then defaultStyle for Grid will be used
	 * @param name
	 *            The name to give to the Style. null will result in name
	 *            {@link GridUtil#UNTITLED_RASTER_STYLE_TITLE}
	 * 
	 * @return Always a Style that you can apply to a {@link GridCoverage}.
	 *         Style has name {@link GridUtil#DEFAULT_RASTER_STYLE_NAME} if no
	 *         colormap was provided.
	 * 
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Style createStyleFromColorMap(ColorMap colorMap, String name) {
		return createStyleFromColorMap(colorMap, name, null);
	}

	/**
	 * Loads {@link Style}s from a SLD {@link InputStream}
	 * 
	 * @param url
	 *            {@link URL} to read the SLD from
	 * 
	 * @return An {@link Array} of {@link Style}s, can be length==0 if no
	 *         UserStyles in SLD file. null if file not exists
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Style[] loadSLD(URL url) {
		InputStream openStream = null;
		try {
			openStream = url.openStream();
			return loadSLD(openStream);
		} catch (IOException e) {
			return null;
		} finally {
			IOUtils.closeQuietly(openStream);
		}
	}

	/**
	 * Loads {@link Style}s from a SLD {@link InputStream}
	 * 
	 * @param inputStream
	 *            {@link InputStream} to read the SLD from
	 * 
	 * @return An {@link Array} of {@link Style}s, can be length==0. null if
	 *         file not exists
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Style[] loadSLD(InputStream inputStream) {

		Style[] styles = null;
		try {
			SLDParser stylereader = new SLDParser(StylingUtil.STYLE_FACTORY,
					inputStream);
			styles = stylereader.readXML();
			return styles;
		} catch (Exception e) {
			LOGGER
					.warn(
							" ... no styles recognized. Return 'new Style[] { null }' ",
							e);
			return new Style[] { null };
		}

	}

	/**
	 * Loads {@link Style}s from a SLD {@link File}
	 * 
	 * @param sldFile
	 *            {@link File} to read the SLD from
	 * 
	 * @return An {@link Array} of {@link Style}s, can be length==0
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Style[] loadSLD(File sldFile) throws FileNotFoundException {

		FileInputStream inputStream = null;
		try {
			inputStream = new FileInputStream(sldFile);
			final Style[] loadedSLD = loadSLD(inputStream);
			return loadedSLD;
		} finally {
			IOUtils.closeQuietly(inputStream);
		}
	}

	/**
	 * Saves the {@link Style} to OGC SLD. Overwrites any existing file. If a
	 * FeatureTypeStyle for selection is used, it is automatically removed. This
	 * method also checks, whether the style is actually differing from any
	 * existing style in the File.
	 * 
	 * @param origStyle
	 *            {@link Style} to save. Any selectino related FeatureTypeStyle
	 *            will be removed.
	 * @param charset
	 *            The charset to use for the XML, e.g. <?xml version="1.0"
	 *            encoding="UTF-8"?>. If <code>null</code>, non is set and the
	 *            system default is used.
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static final boolean saveStyleToSLD(Style origStyle,
			File exportFile, Charset charset) throws TransformerException,
			IOException {
	    // Wenn Datei nicht mit .sld endet, die Dateierweiterung
	    // anhängen
	    exportFile = IOUtil.appendFileExt(exportFile, ".sld");
	  
		SLDTRANSFORMER.setIndentation(2);
		// LOGGER.debug("default charset = "+ Charset.defaultCharset());
		// SLDTRANSFORMER.setEncoding(Charset.forName("UTF-8"));

		Style exportStyle = removeSelectionFeatureTypeStyle(origStyle);

		if (!isStyleDifferent(exportStyle, exportFile)) {
			// LOGGER.debug("Style " + exportFile +
			// " didn't change, not saving.");
			return false;
		}

		SLDTRANSFORMER.transform(exportStyle, new FileWriter(exportFile));

		return true;
	}

	/**
	 * Creates a copy of the given {@link Style}, removing any
	 * {@link FeatureTypeStyle}s that are only SELECTION related.
	 * 
	 * @see {@link FeatureMapLayerSelectionSynchronizer#SELECTION_STYLING_FTS_NAME}
	 */
	public static Style removeSelectionFeatureTypeStyle(final Style style) {

		// Create a copy of the style
		DuplicatingStyleVisitor duplVisitor = new DuplicatingStyleVisitor();
		duplVisitor.visit(style);
		Style cleanStyle = (Style) duplVisitor.getCopy();

		// Remove any selection-FeatureTypeStyle from the new Style
		for (int ii = 0; ii < style.featureTypeStyles().size(); ii++) {

			FeatureTypeStyle fts = style.featureTypeStyles().get(ii);

			if (fts.getName() != null
					&& fts
							.getName()
							.equals(
									FeatureMapLayerSelectionSynchronizer.SELECTION_STYLING_FTS_NAME)) {
				cleanStyle.featureTypeStyles().remove(ii);
				break;
			}
		}

		return cleanStyle;
	}

	/**
	 * Returns <code>null</code> or any {@link FeatureTypeStyle} that is
	 * SELECTION related.
	 * 
	 * @see {@link FeatureMapLayerSelectionSynchronizer#SELECTION_STYLING_FTS_NAME}
	 */
	public static FeatureTypeStyle getSelectionFeatureTypeStyle(
			final Style style) {

		if (style == null)
			return null;

		// Remove any selection-FeatureTypeStyle from the new Style
		for (int ii = 0; ii < style.featureTypeStyles().size(); ii++) {

			FeatureTypeStyle fts = style.featureTypeStyles().get(ii);

			if (fts.getName() != null
					&& fts
							.getName()
							.equals(
									FeatureMapLayerSelectionSynchronizer.SELECTION_STYLING_FTS_NAME)) {
				return style.featureTypeStyles().get(ii);
			}
		}

		return null;
	}

	/**
	 * Compares a given Style and a {@link File} containg a {@link Style}.
	 * 
	 * @param style1
	 *            The first {@link Style}
	 * @param style2file
	 *            A {@link File} pointing to the second {@link Style}
	 * @return <code>true</code> is they are different, ignoring any XML
	 *         fomatting.
	 */
	public static boolean isStyleDifferent(Style style1, File style2file) {
		SLDTRANSFORMER.setIndentation(2);

		try {

			if (!style2file.exists())
				return true;

			Style style2 = loadSLD(style2file)[0];

			return isStyleDifferent(style1, style2);

		} catch (Exception e) {
			LOGGER.debug("Comparing styles " + style1 + " and " + style2file
					+ " failed. So we assume they are different.", e);
			return true;
		}
	}

	/**
	 * Compares a given Style and a {@link File} containg a {@link Style}.
	 * 
	 * @param style1
	 *            The first {@link Style}
	 * @param style2
	 *            The second {@link Style} to compare to
	 * @return <code>true</code> is they are different, ignoring any XML
	 *         fomatting.
	 */
	public static boolean isStyleDifferent(Style style1, Style style2) {
		try {
			// SLDTRANSFORMER.setEncoding(Charset.forName("UTF-8"));
			// Transforming style2 to an XML String
			String style1string = SLDTRANSFORMER.transform(style1);
			String style2string = SLDTRANSFORMER.transform(style2);

			return !style1string.equals(style2string);

		} catch (Exception e) {
			LOGGER.debug("Compating styles " + style1 + " and " + style2
					+ " failed. So we assume they are different.", e);
			return true;
		}

	}

	/**
	 * Saves the {@link Style} to OGC SLD using UTF-8 as charset. Overwrites any
	 * existing file.
	 * 
	 * @param style
	 *            {@link Style} to save
	 * 
	 * @throws TransformerException
	 * @throws IOException
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static final boolean saveStyleToSLD(Style style, File exportFile)
			throws TransformerException, IOException {
		return saveStyleToSLD(style, exportFile, null);
	}

	/**
	 * SLD Rules können die Paramter MinScaleDenominator und MaxScaleDenominator
	 * enthalten. Dadurch können Elemente für manche Zoom-Stufen deaktiviert
	 * werden.
	 * 
	 * Kommentar: "Sichtbarkeit" bezieht es nicht darauf, ob die Elemente auf
	 * dem Kartenausschnitt sichtbar sind, sondern um die SLD Regeln, welche das
	 * SimpleFeature evt. nur Scalenabhängig zeichnen.<br/>
	 * SK 19.6.2009:Bei einem {@link PolygonSymbolizer} zählt nicht der Rand,
	 * sondern nur ob eine Füllung vorhanden ist! Man kann dann zwar leider
	 * nicht auf ein Rand eines Polygons ohne Fill klicken, aber dafür kann man
	 * durch die ungefüllte Fläche klicken.
	 * 
	 * @param fc
	 *            Die zu filternde FeatureCollection. Diese wird nicht
	 *            verändert.
	 * @param style
	 *            Der Style, mit dem die Features gerendert werden (z.b.
	 *            layer.getStyle() )
	 * 
	 * @param scaleDenominator
	 *            Der aktuelle ScaleDenomitor für den die Sichtbarkeit ermittelt
	 *            wird.
	 * 
	 * @see RendererUtilities.calculateOGCScale
	 * 
	 * @return Eine FeatureCollection in welcher nur die Features enthalten
	 *         sind, welche bei aktuellen Scale mit dem übergebenen Style
	 *         gerendert werden.
	 * 
	 *         TODO Was ist mit raster?!
	 * 
	 *         TODO Das sollte man besser machen: Zuerst die FIlter der
	 *         sichtbaren regeln raussuchen, und dann direkt am anfang damitmit
	 *         mitfiltern...
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static MemoryFeatureCollection filterSLDVisibleOnly(
			final FeatureCollection<SimpleFeatureType, SimpleFeature> fc,
			final Style style, final Double scaleDenominator) {

		// Eine im Speicher gehaltene FeatureCollection der sichtbaren
		final MemoryFeatureCollection fcVisible = new MemoryFeatureCollection(
				fc.getSchema());

		// Prüfen aller Features in der Collection
		final Iterator<SimpleFeature> fcIt = fc.iterator();

		try {

			while (fcIt.hasNext()) {
				final SimpleFeature feature = fcIt.next();

				for (final FeatureTypeStyle fts : style.featureTypeStyles()) {

					// Leave out FTSs that are selection related
					if (fts.getName() != null
							&& fts
									.getName()
									.equals(
											FeatureMapLayerSelectionSynchronizer.SELECTION_STYLING_FTS_NAME))
						continue;

					final List<Rule> rules = fts.rules();

					for (final Rule rule : rules) {
						final double maxScaleDenominator = rule
								.getMaxScaleDenominator();
						final double minScaleDenominator = rule
								.getMinScaleDenominator();

						// 1. Check: Trifft die Rule auf den aktuellen Scale der
						// JMapPane?
						if ((minScaleDenominator > scaleDenominator)
								|| (scaleDenominator > maxScaleDenominator)) {
							continue;
						}

						// 2. Check: Eine Rule ist gülrig, aber trifft der
						// Filter für dieses SimpleFeature?
						final Filter rFilter = rule.getFilter();
						if ((rFilter != null) && (!rFilter.evaluate(feature)))
							continue;

						// 3. Check: Passt einer der Symbolizer dieser Rule zu
						// dem
						// SimpleFeature? Dieser Check wird in
						// "real world" meistens klappen, es sei denn, ein
						// Hansel
						// gibt einen LineSymbolizer für ein PointLayer an.
						boolean passt = false;
						for (final Symbolizer symb : rule.getSymbolizers()) {

							final Geometry geom = (Geometry) feature
									.getDefaultGeometry();

							if ((geom instanceof MultiPoint)
									|| (geom instanceof com.vividsolutions.jts.geom.Point)) {
								if (symb instanceof PointSymbolizer) {
									if (((PointSymbolizer) symb).getGraphic() != null)
										passt = true;
									break;
								}
							} else if ((geom instanceof MultiLineString)
									|| (geom instanceof LineString)) {
								if (symb instanceof LineSymbolizer) {
									if (((LineSymbolizer) symb).getStroke() != null)
										passt = true;
									break;
								}
							} else if ((geom instanceof MultiPolygon)
									|| (geom instanceof Polygon)) {

								if (symb instanceof PolygonSymbolizer) {
									if (((PolygonSymbolizer) symb).getFill() != null)
										passt = true;
									break;
								}

								if (symb instanceof PointSymbolizer) {
									// This is a special case and not handles
									// 100% correct yet. Polygons visualized
									// with Pointsymbolizers should also have
									// different Filters when they are selected
									if (((PointSymbolizer) symb).getGraphic() != null)
										passt = true;
									break;
								}
							} else {
								LOGGER
										.warn("Unbekannter geom Typ! Gehe von sichbar aus!");
								passt = true;
							}
						}
						if (!passt)
							continue;

						// Nun sollte das feature wiklich sichtbar sein
						fcVisible.add(feature);
					}
				}
			}
		} catch (final java.lang.IllegalStateException e) {
			LOGGER.error("Iterating over the features", e);
			/**
			 * SK: 14.Apri 2009. It happened a few time to that fcIt.hasNext
			 * suddenly threw an exception for the "africa countries.shp":
			 * Exception in thread "AWT-EventQueue-0"
			 * java.lang.IllegalStateException: ShapeType changed illegally from
			 * Polygon to Undefined at org.geotools.
			 * data.shapefile.shp.ShapefileReader.nextRecord(ShapefileReader
			 * .java:452) at org.geotools.data.shapefile.indexed.
			 * IndexedShapefileDataStore$Reader
			 * .next(IndexedShapefileDataStore.java:1272) at
			 * org.geotools.data.FIDFeatureReader.next(FIDFeatureReader.java:92)
			 * at org .geotools.data.FilteringFeatureReader.hasNext(
			 * FilteringFeatureReader .java:125) at
			 * org.geotools.data.store.FeatureReaderIterator.hasNext(
			 * FeatureReaderIterator.java:44) at
			 * schmitzm.geotools.feature.FeatureUtil
			 * .filterSLDVisibleOnly(FeatureUtil.java:216) *
			 * 
			 */
		}

		// LOGGER.info("filterSLDVisibleOnly removed "+ (fc.size() -
		// fcVisible.size()) + " features.");

		return fcVisible;
	}

	/**
	 * @param style
	 *            A {@link Style} to search for the first {@link TextSymbolizer}
	 *            that will be used for the given {@link SimpleFeature}.
	 * 
	 * @author Stefan A. Krüger
	 * 
	 * @return <code>null</code> or the first {@link TextSymbolizer} found in
	 *         the style that applies to the {@link SimpleFeature}.
	 */
	public static TextSymbolizer getTextSymbolizer(final Style style,
			final SimpleFeature f) {
		if (f == null)
			return null;
		try {
			if (style != null) {

				for (final FeatureTypeStyle fts : style.featureTypeStyles()) {
					for (final Rule r : fts.rules()) {

						final Filter filter = r.getFilter();

						if (filter != null && (!filter.evaluate(f))) {
							continue;
						}

						for (final Symbolizer symb : r.getSymbolizers()) {
							if (symb instanceof TextSymbolizer) {
								return (TextSymbolizer) symb;
							}
						}

					}
				}

			}

		} catch (final Exception e) {
			LOGGER.error("error - filter API stuff?", e);
			return null;
		}
		return null;
	}

	/**
	 * @param style
	 *            A {@link Style} to search for all {@link TextSymbolizer}s . No
	 *            guarantee, that any one of them will ever be used for any
	 *            feature (think about filters).
	 * 
	 * @author Stefan A. Krüger
	 * 
	 * @return {@link List} or all {@link TextSymbolizer}s found in the
	 *         {@link Style}.
	 */
	public static List<TextSymbolizer> getVisibleTextSymbolizers(
			final Style style) {
		List<TextSymbolizer> results = new ArrayList<TextSymbolizer>();
		try {
			if (style != null) {

				for (final FeatureTypeStyle fts : style.featureTypeStyles()) {
					for (final Rule r : fts.rules()) {

						if (r.getFilter() != null) {

							/*
							 * We are comparing it to AsUtil.alwaysfalsefilter
							 * 
							 * public static final PropertyIsEqualTo
							 * allwaysFalseFilter = ff2.equals(ff2
							 * .literal("1"), ff2.literal("2"));
							 */
							try {

								Filter f = r.getFilter();
								if (f instanceof And) {
									And and = (And) f;
									if (and.getChildren() != null) {
										Filter candidateF = and.getChildren()
												.get(0);
										if (candidateF instanceof PropertyIsEqualTo) {
											final PropertyIsEqualTo compare = (PropertyIsEqualTo) candidateF;
											if (compare.getExpression1() instanceof Literal
													&& compare.getExpression2() instanceof Literal) {
												Literal test1 = (Literal) compare
														.getExpression1();
												Literal test2 = (Literal) compare
														.getExpression2();
												if (test1.toString()
														.equals("1")
														&& test2.toString()
																.equals("2")) {
													// This TextSymbolizer is
													// disabled using
													// ASUtil.allwaysFalseFilter
													// LOGGER.debug("Ignoring Rule "+
													// r
													// +" because the filter is "+f);
													continue;
												}
											}
										}
									}
								}
							} catch (Exception e) {
								LOGGER
										.debug(
												"Checking for textSymbolizer allwaysFalseFilter",
												e);
							}
						}

						results.addAll(getTextSymbolizers(r.getSymbolizers()));
					}
				}

			}

		} catch (final Exception e) {
			LOGGER.error("error - filter API stuff?", e);
			return results;
		}
		return results;
	}

	/**
	 * @param style
	 *            A {@link Style} to search for all {@link TextSymbolizer}s . No
	 *            guarantee, that any one of them will ever be used for any
	 *            feature (think about filters).
	 * 
	 * @author Stefan A. Krüger
	 * 
	 * @return {@link List} or all {@link TextSymbolizer}s found in the
	 *         {@link Style}.
	 */
	public static List<TextSymbolizer> getTextSymbolizers(final Style style) {
		List<TextSymbolizer> results = new ArrayList<TextSymbolizer>();
		try {
			if (style != null) {

				for (final FeatureTypeStyle fts : style.featureTypeStyles()) {
					for (final Rule r : fts.rules()) {
						results.addAll(getTextSymbolizers(r.getSymbolizers()));
					}
				}

			}

		} catch (final Exception e) {
			LOGGER.error("error - filter API stuff?", e);
			return results;
		}
		return results;
	}

	/**
	 * @param symbolizers
	 *            List of Symbolizers to search for all {@link TextSymbolizer}s
	 *            . No guarantee, that any one of them will ever be used for any
	 *            feature (think about filters).
	 * 
	 * @author Stefan A. Krüger
	 * 
	 * @return {@link List} or all {@link TextSymbolizer}s found in the given
	 *         symbolizers.
	 */
	public static List<TextSymbolizer> getTextSymbolizers(
			Symbolizer[] symbolizers) {
		List<TextSymbolizer> results = new ArrayList<TextSymbolizer>();
		try {
			if (symbolizers != null) {

				for (final Symbolizer symb : symbolizers) {
					if (symb instanceof TextSymbolizer) {
						results.add((TextSymbolizer) symb);
					}
				}

			}

		} catch (final Exception e) {
			LOGGER.error("error - filter API stuff?", e);
			return results;
		}
		return results;
	}

	// /**
	// * Sorts a {@link ColorMap} by its quantity values. This is mentioned in
	// the
	// * OGC SLD Definition, page 52:
	// * <p>
	// * Quote: For example, a DEM raster giving elevations in meters above sea
	// * level can be translated to a colored image with a ColorMap. The
	// quantity
	// * attributes of a color-map are used for translating between numeric
	// * matrixes and color rasters and the ColorMap entries should be in order
	// of
	// * increasing numeric quantity so that intermediate numeric values can be
	// * matched to a color (or be interpolated between two colors). Labels may
	// be
	// * used for legends or may be used in the future to match character
	// values.
	// * Not all systems can support opacity in colormaps. The default opacity
	// is
	// * 1.0 (fully opaque). Defaults for quantity and label are
	// system-dependent.
	// * </p>
	// *
	// * @param colorMap
	// * {@link ColorMap} to sort by the quantity-field of the
	// * {@link ColorMapEntry}s
	// * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	// Kr&uuml;ger</a>
	// */
	// public static void sort(RemovableColorMapImpl colorMap) {
	// ColorMapEntry[] colorMapEntries = colorMap.getColorMapEntries();
	// SortedMap<Double, ColorMapEntry> tmap = new TreeMap<Double,
	// ColorMapEntry>();
	// // Add to a SortedMap
	// for (ColorMapEntry cme : colorMapEntries) {
	// tmap.put(getQuantityFromColorMapEntry(cme), cme);
	// colorMap.removeColorMapEntry(cme);
	// }
	// // Recreate ColorMap from sorted quantity values
	// for (ColorMapEntry cme : tmap.values()) {
	// colorMap.addColorMapEntry(cme);
	// }
	// }

	// /**
	// * Lets assume that every quantity value is maximally listed once in a
	// * {@link ColorMap}, then it makes some sense to use it as a key.
	// *
	// * @param quantityString
	// * quantitiy as String
	// * @return The first colorMap that has the same Quantity ( compared as
	// * Strings)
	// *
	// * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	// Kr&uuml;ger</a>
	// */
	// public static ColorMapEntry getColorMapEntryByQuantityString(
	// ColorMap colorMap, String quantityString) {
	// for (ColorMapEntry cme : colorMap.getColorMapEntries()) {
	// if (quantityString.equals(cme.getQuantity().toString()))
	// return cme;
	// }
	// return null;
	// }

	/**
	 * Clones a Style by converting it to XML and reading it back in.
	 * 
	 * @param style
	 *            the {@link Style} to be copied.
	 */
	public static Style clone(Style style) {
		DuplicatingStyleVisitor duplicatingStyleVisitor = new DuplicatingStyleVisitor();
		style.accept(duplicatingStyleVisitor);
		return (Style) duplicatingStyleVisitor.getCopy();
	}

	/**
	 * Clones a {@link Symbolizer} using the {@link DuplicatingStyleVisitor}.
	 */
	public static Symbolizer clone(Symbolizer sym) {
		DuplicatingStyleVisitor duplicatingStyleVisitor = new DuplicatingStyleVisitor();
		sym.accept(duplicatingStyleVisitor);
		return (Symbolizer) duplicatingStyleVisitor.getCopy();
	}

	/**
	 * Clons a {@link Graphic} element
	 * 
	 * @param graphicFill
	 * @return
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Graphic clone(Graphic graphic) {
		DuplicatingStyleVisitor duplicatingStyleVisitor = new DuplicatingStyleVisitor();
		graphic.accept(duplicatingStyleVisitor);
		return (Graphic) duplicatingStyleVisitor.getCopy();
	}

	/**
	 * Creates a new {@link BrewerPalette} with only one scheme (which includes
	 * all colors). The suitablity of the palette is set to "unknown" for all
	 * viewer types.
	 * 
	 * @param name
	 *            name for the palette (also the description is set to this
	 *            value)
	 * @param colors
	 *            colors for the palette
	 */
	public static BrewerPalette createBrewerPalette(String name, Color[] colors) {
		BrewerPalette palette = new BrewerPalette();
		palette.setColors(colors);
		palette.setName(name);
		palette.setDescription(name);

		// set suitability to UNKNOWN for all viewers
		PaletteSuitability suitability = new PaletteSuitability();
		try {
			suitability.setSuitability(colors.length, new String[] { "?", "?",
					"?", "?", "?", "?" });
			palette.setPaletteSuitability(suitability);
		} catch (IOException e) {
			LOGGER.error(
					"Unabel to PaletteSuitability.setSuitablility for colors.length="
							+ colors, e);
		}

		// Create the trivial scheme for the palette (requested colors equals
		// number of colors)
		int[] schema = new int[colors.length];
		for (int j = 0; j < colors.length; j++)
			schema[j] = j;
		SampleScheme sampleScheme = new SampleScheme();

		// set the trivial scheme for all schemas
		for (int j = 2; j < schema.length; j++) {
			sampleScheme.setSampleScheme(j, schema);
			// sampleScheme.setSampleScheme(schema.length, schema);
		}

		palette.setColorScheme(sampleScheme);

		return palette;
	}

	/**
	 * @param geoObject
	 *            {@link GridCoverage2D}, {@link AbstractGridCoverage2DReader},
	 *            {@link FeatureCollection}, {@link GeometryAttributeType} or
	 *            {@link FeatureSource}
	 * 
	 * @return a {@link Style} that is suitable to identify features of the
	 *         given type as selected items.
	 */
	public static FeatureTypeStyle createSelectionStyle(Object geoObject) {

		if (geoObject instanceof GridCoverage2D
				|| geoObject instanceof org.geotools.coverage.grid.io.AbstractGridCoverage2DReader) {
			// Wenn irgendwann mal selection für raster möglich ist, dann hier
			// einen schöneren style erstellen.
			return GridUtil.createDefaultStyle().featureTypeStyles().get(0);
		}

		// We have vector-data. Now let's determine the type...
		GeometryForm geometryForm = null;

		if (geoObject instanceof FeatureCollection) {
			geometryForm = FeatureUtil
					.getGeometryForm((FeatureCollection<SimpleFeatureType, SimpleFeature>) geoObject);
		} else if (geoObject instanceof GeometryAttributeType) {
			geometryForm = FeatureUtil
					.getGeometryForm((GeometryAttributeType) geoObject);
		} else if (geoObject instanceof FeatureSource) {
			geometryForm = FeatureUtil
					.getGeometryForm((FeatureSource<SimpleFeatureType, SimpleFeature>) geoObject);
		}

		// /**
		// * Let's see if we have a cached version of the requested style.
		// */
		// if (selectionStyles.containsKey(geometryForm))
		// return selectionStyles.get(geometryForm);

		Symbolizer[] symbolizers = new Symbolizer[0];

		switch (geometryForm) {

		case POINT:
			PointSymbolizer ps = STYLE_FACTORY.createPointSymbolizer();
			ps.setGraphic(SELECTION_GRAPHIC1);
			symbolizers = LangUtil.extendArray(symbolizers, ps);

			PointSymbolizer ps2 = STYLE_FACTORY.createPointSymbolizer();
			ps2.setGraphic(SELECTION_GRAPHIC2);
			symbolizers = LangUtil.extendArray(symbolizers, ps2);

			PointSymbolizer ps3 = STYLE_FACTORY.createPointSymbolizer();
			ps3.setGraphic(SELECTION_GRAPHIC3);
			symbolizers = LangUtil.extendArray(symbolizers, ps3);

			break;
		case POLYGON: { // Black
			Stroke outline = STYLE_BUILDER.createStroke(Color.red, 3);
			Mark dots = STYLE_BUILDER.createMark("shape://backslash");
			dots.setStroke(STYLE_BUILDER.createStroke(Color.red, 3.));
			Graphic fillGraphic = STYLE_BUILDER.createGraphic(null, dots, null);
			fillGraphic.setSize(FILTER_FACTORY.literal(9.));
			Fill fill = STYLE_BUILDER.createFill();
			fill.setGraphicFill(fillGraphic);
			PolygonSymbolizer polS0 = STYLE_BUILDER.createPolygonSymbolizer(
					outline, fill);
			symbolizers = LangUtil.extendArray(symbolizers, polS0);
		}

			{ // White
				Stroke outline = STYLE_BUILDER.createStroke(Color.yellow
						.brighter(), 1);
				Mark dots = STYLE_BUILDER.createMark("shape://backslash");
				dots.setStroke(STYLE_BUILDER.createStroke(Color.yellow
						.brighter(), 1.));
				Graphic fillGraphic = STYLE_BUILDER.createGraphic(null, dots,
						null);
				fillGraphic.setSize(FILTER_FACTORY.literal(9.));
				Fill fill = STYLE_BUILDER.createFill();
				fill.setGraphicFill(fillGraphic);
				PolygonSymbolizer polS0 = STYLE_BUILDER
						.createPolygonSymbolizer(outline, fill);
				symbolizers = LangUtil.extendArray(symbolizers, polS0);
			}
			// PolygonSymbolizer polS0 =
			// STYLE_BUILDER.createPolygonSymbolizer(Color.RED);
			// symbolizers = LangUtil.extendArray(symbolizers, polS0);

			break;

		case LINE:
			// LineSymbolizer ls = STYLE_FACTORY.createLineSymbolizer();
			// ls.setStroke(SELECTION_STROKE1);
			// symbolizers = LangUtil.extendArray(symbolizers, ls);
			//
			// LineSymbolizer ls2 = STYLE_FACTORY.createLineSymbolizer();
			// ls2.setStroke(SELECTION_STROKE2);
			// symbolizers = LangUtil.extendArray(symbolizers, ls2);
			//
			// LineSymbolizer ls3 = STYLE_FACTORY.createLineSymbolizer();
			// ls3.setStroke(SELECTION_STROKE3);
			// symbolizers = LangUtil.extendArray(symbolizers, ls3);

			LineSymbolizer ls = STYLE_BUILDER.createLineSymbolizer(Color.red,
					3.);
			symbolizers = LangUtil.extendArray(symbolizers, ls);

			LineSymbolizer ls2 = STYLE_BUILDER.createLineSymbolizer(
					Color.white, 1.);
			symbolizers = LangUtil.extendArray(symbolizers, ls2);
			break;

		default:
			throw new IllegalArgumentException("Provide a suitable object.");
		}

		return STYLE_BUILDER.createFeatureTypeStyle(symbolizers, Double.NaN,
				Double.NaN);

	}

	public static Style createStyleSimple(Object geoObject, Color color,
			Color color2) {
		if (geoObject instanceof GridCoverage2D
				|| geoObject instanceof org.geotools.coverage.grid.io.AbstractGridCoverage2DReader) {
			// Wenn irgendwann mal selection für raster möglich ist, dann hier
			// einen schöneren style erstellen.
			return GridUtil.createDefaultStyle();
		}

		// We have vector-data. Now let's determine the type...
		GeometryForm geometryForm = null;

		if (geoObject instanceof FeatureCollection) {
			geometryForm = FeatureUtil
					.getGeometryForm((FeatureCollection<SimpleFeatureType, SimpleFeature>) geoObject);
		} else if (geoObject instanceof GeometryAttributeType) {
			geometryForm = FeatureUtil
					.getGeometryForm((GeometryAttributeType) geoObject);
		} else if (geoObject instanceof FeatureSource) {
			geometryForm = FeatureUtil
					.getGeometryForm((FeatureSource<SimpleFeatureType, SimpleFeature>) geoObject);
		}

		// /**
		// * Let's see if we have a cached version of the requested style.
		// */
		// if (selectionStyles.containsKey(geometryForm))
		// return selectionStyles.get(geometryForm);

		Style style = STYLE_FACTORY.createStyle();
		style.addFeatureTypeStyle(STYLE_FACTORY.createFeatureTypeStyle());
		Symbolizer[] symbolizers = new Symbolizer[0];

		switch (geometryForm) {

		case POINT:
			PointSymbolizer ps = STYLE_BUILDER
					.createPointSymbolizer(STYLE_BUILDER.createGraphic(null,
							STYLE_BUILDER.createMark("circle"), null));
			ps.getGraphic().getMarks()[0].setFill(STYLE_BUILDER.createFill(
					color, 0.5));
			ps.getGraphic().getMarks()[0].setSize(STYLE_BUILDER
					.literalExpression(8.));
			ps.getGraphic().setSize(STYLE_BUILDER.literalExpression(8.));
			ps.getGraphic().getMarks()[0].setStroke(STYLE_BUILDER
					.createStroke(color2));
			symbolizers = LangUtil.extendArray(symbolizers, ps);

			break;
		case POLYGON:
			PolygonSymbolizer polS = STYLE_BUILDER.createPolygonSymbolizer(
					STYLE_BUILDER.createStroke(color2), STYLE_BUILDER
							.createFill(color));
			symbolizers = LangUtil.extendArray(symbolizers, polS);

		case LINE:
			LineSymbolizer ls = STYLE_BUILDER
					.createLineSymbolizer(STYLE_BUILDER.createStroke(color));
			symbolizers = LangUtil.extendArray(symbolizers, ls);
			break;

		default:
			throw new IllegalArgumentException("Provide a suitable object.");
		}

		style.getFeatureTypeStyles()[0].setRules(new Rule[] { STYLE_FACTORY
				.createRule() });
		style.getFeatureTypeStyles()[0].getRules()[0]
				.setSymbolizers(symbolizers);

		return style;
	}

	//
	// /**
	// * Doubles the number of palettes by adding the reverse palette to every
	// * palette. The name of the reverse palette is the name of the original
	// * palette plus the letter R.
	// *
	// * @author SK
	// * @param palettes The original list of {@link BrewerPalette
	// BrewerPalettes}
	// */
	// public static BrewerPalette[] addReversePalettes(BrewerPalette[]
	// palettes) {
	// BrewerPalette[] newPalettes = new BrewerPalette[palettes.length * 2];
	//
	// int pos = 0;
	// for (BrewerPalette bp : palettes) {
	// newPalettes[pos] = bp;
	// pos++;
	// Color[] newColors = new Color[bp.getMaxColors()];
	//
	// List<Color> colorList =
	// java.util.Arrays.asList(bp.getColors(bp.getMaxColors()));
	// Collections.reverse(colorList);
	// colorList.toArray(newColors);
	//
	// try {
	// newPalettes[pos] = StylingUtil.createBrewerPalette(bp.getName()
	// + "R", newColors);
	// newPalettes[pos].setPaletteSuitability(bp
	// .getPaletteSuitability());
	// newPalettes[pos].setDescription(bp.getDescription()+" reversed");
	// newPalettes[pos].setType(bp.getType());
	// newPalettes[pos].setColorScheme(bp.getColorScheme());
	// } catch (IOException e) {
	// LOGGER.error("",e);
	// newPalettes[pos] = bp;
	// }
	//
	// pos++;
	// }
	//
	// return newPalettes;
	// }

	/**
	 * Replaces the "main" color in a given {@link Symbolizer} element
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static void replaceSymbolizerColor(Symbolizer symb, Color oldColor,
			Color newColor) {
		if (symb == null)
			return;

		if (symb instanceof PointSymbolizer) {
			PointSymbolizer ps = (PointSymbolizer) symb;
			replacePointSymbolizerColor(ps, oldColor, newColor);
		}

		if (symb instanceof PolygonSymbolizer) {
			PolygonSymbolizer ps = (PolygonSymbolizer) symb;
			replacePolygonSymbolizerColor(ps, oldColor, newColor);
		}

		if (symb instanceof LineSymbolizer) {
			LineSymbolizer ps = (LineSymbolizer) symb;
			replaceLineSymbolizerColor(ps, oldColor, newColor);
		}

	}

	/**
	 * Replaces the "main" color in a given {@link LineSymbolizer} element
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static void replaceLineSymbolizerColor(LineSymbolizer ps,
			Color oldColor, Color newColor) {
		if (ps == null)
			return;

		replaceStrokeColor(ps.getStroke(), oldColor, newColor);
	}

	/**
	 * Replaces the "main" color in a given {@link PointSymbolizer} element
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static void replacePointSymbolizerColor(PointSymbolizer ps,
			Color oldColor, Color newColor) {
		if (ps == null)
			return;

		replaceGraphicColor(ps.getGraphic(), oldColor, newColor);
	}

	/**
	 * Replaces the "main" color in a given {@link PolygonSymbolizer} element
	 * 
	 * @param oldColor
	 * @param newColor
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 * @param ps
	 */
	public static void replacePolygonSymbolizerColor(PolygonSymbolizer ps,
			Color oldColor, Color newColor) {
		replaceFillColor(ps.getFill(), oldColor, newColor);
		replaceStrokeColor(ps.getStroke(), oldColor, newColor);
	}

	/**
	 * Replaces the "main" color in a given {@link Stroke} element
	 * 
	 * @param stroke
	 * @param oldColor
	 * @param newColor
	 */
	public static void replaceStrokeColor(Stroke stroke, Color oldColor,
			Color newColor) {
		if (stroke == null)
			return;

		if ((stroke.getColor() != null)
				&& (StylingUtil.getColorFromExpression(stroke.getColor())
						.equals(oldColor)))
			stroke
					.setColor(StylingUtil.STYLE_BUILDER
							.colorExpression(newColor));

		replaceGraphicColor(stroke.getGraphicFill(), oldColor, newColor);

		replaceGraphicColor(stroke.getGraphicStroke(), oldColor, newColor);

	}

	/**
	 * Replaces the "main" color in a given {@link Fill} element
	 * 
	 * @param fill
	 * @param oldColor
	 * @param newColor
	 */
	public static void replaceFillColor(Fill fill, Color oldColor,
			Color newColor) {
		if (fill == null)
			return;

		if ((fill.getColor() != null)
				&& (getColorFromExpression(fill.getColor()).equals(oldColor)))
			fill.setColor(STYLE_BUILDER.colorExpression(newColor));

		replaceGraphicColor(fill.getGraphicFill(), oldColor, newColor);
		//
		// if ((fill.getBackgroundColor() != null)
		// && (getColorFromExpression(fill.getBackgroundColor())
		// .equals(oldColor)))
		// fill.setBackgroundColor(STYLE_BUILDER.colorExpression(newColor));
	}

	/**
	 * Replaces the "main" color in a given {@link Graphic} element
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static void replaceGraphicColor(Graphic graphic, Color oldColor,
			Color newColor) {
		if (graphic == null)
			return;

		if ((graphic.getMarks() != null) && (graphic.getMarks().length > 0)) {

			// Checking for Colors in Marks...
			for (Mark m : graphic.getMarks()) {

				replaceFillColor(m.getFill(), oldColor, newColor);

				replaceStrokeColor(m.getStroke(), oldColor, newColor);

			}
		}
	}

	/**
	 * @return the {@link Color} used in the {@link Graphic} or null, if an
	 *         {@link ExternalGraphic} is used.
	 * 
	 * @param graphic
	 *            If <code>null</code> returns <code>null</code>
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Color getGraphicColor(Graphic graphic) {
		if (graphic == null)
			return null;

		Color foundColor = null;

		if ((graphic.getMarks() != null) && (graphic.getMarks().length > 0)) {

			// Checking for Colors in Marks...
			for (Mark m : graphic.getMarks()) {

				if (foundColor == null)
					foundColor = getFillColor(m.getFill());

				if (foundColor == null)
					foundColor = getStrokeColor(m.getStroke());
			}

		}

		return foundColor;
	}

	/**
	 * @return the first {@link Color} used in a {@link Stroke}.
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Color getStrokeColor(Stroke stroke) {
		if (stroke == null)
			return null;

		if (stroke.getColor() != null)
			return StylingUtil.getColorFromExpression(stroke.getColor());

		if (getGraphicColor(stroke.getGraphicStroke()) != null)
			return getGraphicColor(stroke.getGraphicStroke());

		if (getGraphicColor(stroke.getGraphicFill()) != null)
			return getGraphicColor(stroke.getGraphicFill());

		return null;
	}

	/**
	 * @return the first {@link Color} used in a {@link Fill}.
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Color getFillColor(Fill fill) {
		if (fill == null)
			return null;

		Color foundColor = null;

		if (foundColor == null)
			foundColor = getColorFromExpression(fill.getColor());

		if (foundColor == null)
			foundColor = getGraphicColor(fill.getGraphicFill());

		return foundColor;
	}

	/**
	 * @return the first {@link Color} used in a {@link Symbolizer}.
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Color getSymbolizerColor(Symbolizer symb) {
		if (symb == null)
			return null;

		if (symb instanceof PointSymbolizer) {
			PointSymbolizer ps = (PointSymbolizer) symb;
			return getPointSymbolizerColor(ps);
		}

		if (symb instanceof PolygonSymbolizer) {
			PolygonSymbolizer ps = (PolygonSymbolizer) symb;
			return getPolygonSymbolizerColor(ps);
		}

		if (symb instanceof LineSymbolizer) {
			LineSymbolizer ps = (LineSymbolizer) symb;
			return getLineSymbolizerColor(ps);
		}

		return null;
	}

	/**
	 * @return the first {@link Color} used in a {@link LineSymbolizer}.
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Color getLineSymbolizerColor(LineSymbolizer ps) {
		if (ps == null)
			return null;

		Color foundColor = null;

		if (foundColor == null)
			foundColor = getStrokeColor(ps.getStroke());

		return foundColor;
	}

	/**
	 * @return the first {@link Color} used in a {@link PolygonSymbolizer}.
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Color getPolygonSymbolizerColor(PolygonSymbolizer ps) {
		if (ps == null)
			return null;

		Color foundColor = null;

		if (foundColor == null)
			foundColor = getFillColor(ps.getFill());

		if (foundColor == null)
			foundColor = getStrokeColor(ps.getStroke());

		return foundColor;
	}

	/**
	 * @return the first {@link Color} used in a {@link PointSymbolizer}.
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static Color getPointSymbolizerColor(PointSymbolizer ps) {
		if (ps == null)
			return null;

		return getGraphicColor(ps.getGraphic());
	}

	/***************************************************************************
	 * Copies all Values from one {@link TextSymbolizer} to another
	 * {@link TextSymbolizer}
	 * 
	 * @param from
	 *            {@link TextSymbolizer} source
	 * @param to
	 *            {@link TextSymbolizer} target. May not be <code>null</code>.
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static void copyAllValues(final TextSymbolizer from,
			final TextSymbolizer to) {
		to.setLabel(from.getLabel());
		to.setPriority(from.getPriority());

		// TODO Does not copy all that has to be copied!

		final FilterFactory2 ff2 = FeatureUtil.FILTER_FACTORY2;

		to.getFonts()[0].setFontFamily(ff2.literal(from.getFonts()[0]
				.getFontFamily().toString()));
		to.getFonts()[0].setFontSize(ff2.literal(from.getFonts()[0]
				.getFontSize().toString()));
		to.getFonts()[0].setFontWeight(ff2.literal(from.getFonts()[0]
				.getFontWeight().toString()));
		to.getFonts()[0].setFontStyle(ff2.literal(from.getFonts()[0]
				.getFontStyle().toString()));
	}

	/**
	 * Returns all {@link RasterSymbolizer} that are contained in a
	 * {@link Style}. {@link RasterSymbolizer} from {@link FeatureTypeStyle}s
	 * that are selection related are ignored.
	 */
	public static List<RasterSymbolizer> getRasterSymbolizers(Style style) {
		final List<RasterSymbolizer> rsList = new ArrayList<RasterSymbolizer>();
		for (FeatureTypeStyle fts : style.featureTypeStyles()) {

			// Leave out FTSs that are selection related
			if (fts.getName() != null
					&& fts
							.getName()
							.equals(
									FeatureMapLayerSelectionSynchronizer.SELECTION_STYLING_FTS_NAME))
				continue;

			for (Rule r : fts.rules()) {
				for (Symbolizer symb : r.getSymbolizers()) {
					if (symb instanceof RasterSymbolizer) {
						rsList.add((RasterSymbolizer) symb);
					}
				}
			}
		}

		return rsList;
	}

	/**
	 * Super smart metod ;-) Needs to be extended to look at the graphic width
	 * and Geometry type.
	 * 
	 * @param geometryDescriptor
	 * @param avgNn
	 */
	public static double getMaxScaleDenominator(Double avgNn) {

		return getMaxScaleDenominator(avgNn, null);
	}

	/**
	 * Super smart metod ;-) Needs to be extended to look at the graphic width
	 * and Geometry type.
	 * 
	 * @param geometryDescriptor
	 * @param avgNn
	 */
	public static double getMaxScaleDenominator(Double avgNn, GeometryForm form) {

		double FAKTOR = .2;
		// if (geometry.isAssignableFrom(GeometryCollection.class)) {
		// FAKTOR = 20;
		// }
		if (form == GeometryForm.LINE) {
			FAKTOR = 1;
		} else if (form == GeometryForm.POLYGON) {
			FAKTOR = .01;
		}

		if (Double.isNaN(avgNn) || Double.isInfinite(avgNn))
			return Double.MAX_VALUE;

		double maxScale = avgNn * FAKTOR;

		LOGGER.debug("maxScale for " + avgNn + " is " + maxScale);

		maxScale = Math.round(maxScale / 1000) * 1000.;

		LOGGER.debug("maxScale for " + avgNn + " is gerundet " + maxScale);
		//
		// switch (FeatureUtil.getGeometryForm(geometryDescriptor)) {
		// case POINT:
		// maxScale = avgNn * 10000.;
		// break;
		// case LINE:
		// maxScale = avgNn * 10000.;
		// break;
		// case POLYGON:
		// maxScale = avgNn * 10000.;
		// break;
		// }

		return maxScale;
	}

	/**
	 * SLD Rules können die Paramter MinScaleDenominator und MaxScaleDenominator
	 * enthalten. Dadurch können Elemente für manche Zoom-Stufen deaktiviert
	 * werden.
	 * 
	 * @param fc
	 *            Die zu filternde FeatureCollection. Diese wird nicht
	 *            verändert.
	 * @param style
	 *            Der Style, mit dem die Features gerendert werden (z.b.
	 *            layer.getStyle() )
	 * 
	 * @return Eine FeatureCollection in welcher nur die Features enthalten
	 *         sind, welche bei aktuellen Scale mit dem übergebenen Style
	 *         gerendert werden.
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static FeatureCollection<SimpleFeatureType, SimpleFeature> filterSLDVisibleOnly(
			final FeatureCollection<SimpleFeatureType, SimpleFeature> fc,
			final Style style, XMapPane xMapPane) {

		if (xMapPane == null || style == null)
			return new EmptyFeatureCollection(fc.getSchema());

		// Der "scaleDenominator" der aktuellen JMapPane
		Double scaleDenominator = RendererUtilities.calculateOGCScale(
				new ReferencedEnvelope(xMapPane.getMapArea(), xMapPane
						.getMapContext().getCoordinateReferenceSystem()),
				xMapPane.getBounds().width, null);

		return filterSLDVisibleOnly(fc, style, scaleDenominator);
	}

	/**
	 * Since GT2.6, the AttributeNames are case sensitive. Also the raster
	 * Styles need GeometryProperty set to "geom" to work. This method checks
	 * all referenced AttributeNames and checks them against the schema.
	 * 
	 * @retun <code>true</code> if something has been change, otherwise
	 *        <code>false</code>.
	 */
	public static Style correctPropertyNames(Style style,
			final SimpleFeatureType schema) {

		DuplicatingStyleVisitor dsv = new DuplicatingStyleVisitor() {
			public void visit(RasterSymbolizer sym) {
				if (!sym.getGeometryPropertyName().equals("geom")) {
					LOGGER
							.debug("Corrected one RasterSymbolizer with GeometryPropertyName '"
									+ sym.getGeometryPropertyName()
									+ "' to 'geom'.");
					sym.setGeometryPropertyName("geom");
				}
				super.visit(sym);
			};

			public void visit(LineSymbolizer sym) {
				sym.setGeometryPropertyName(FeatureUtil
						.findBestMatchingAttributeFallBackFirst(schema,
								sym.getGeometryPropertyName()).getLocalPart());
				super.visit(sym);
			}

			public void visit(PolygonSymbolizer sym) {
				sym.setGeometryPropertyName(FeatureUtil
						.findBestMatchingAttributeFallBackFirst(schema,
								sym.getGeometryPropertyName()).getLocalPart());
				super.visit(sym);
			}

			public void visit(PointSymbolizer sym) {
				sym.setGeometryPropertyName(FeatureUtil
						.findBestMatchingAttributeFallBackFirst(schema,
								sym.getGeometryPropertyName()).getLocalPart());
				super.visit(sym);
			}

			@Override
			public void visit(TextSymbolizer sym) {
				sym.setGeometryPropertyName(FeatureUtil
						.findBestMatchingAttributeFallBackFirst(schema,
								sym.getGeometryPropertyName()).getLocalPart());

				if (sym.getLabel() instanceof PropertyName) {
					PropertyName propertyname = (PropertyName) sym.getLabel();
					sym.setLabel(ff.property(FeatureUtil
							.findBestMatchingAttributeFallBackFirst(schema,
									propertyname.getPropertyName())));
				}

				if (sym.getPriority() instanceof PropertyName) {
					PropertyName propertyname = (PropertyName) sym
							.getPriority();
					sym.setPriority(ff.property(FeatureUtil
							.findBestMatchingAttributeFallBackFirst(schema,
									propertyname.getPropertyName())));
				}

				super.visit(sym);
			}

			@Override
			public Expression copy(Expression expression) {
				if (expression instanceof PropertyName) {
					PropertyName pName = (PropertyName) expression;

					Name correctedName = FeatureUtil
							.findBestMatchingAttributeFallBackFirst(schema,
									pName.getPropertyName());

					return ff.property(correctedName);
				}
				// if (expression instanceof AttributeExpression) {
				// AttributeExpression aExpression = (AttributeExpression)
				// expression;
				// Name correctedName = FeatureUtil
				// .findBestMatchingAttributeFallBackFirst(schema,
				// aExpression.getPropertyName());
				//
				// return new AttributeExpressionImpl(correctedName
				// .getLocalPart());
				// }
				return super.copy(expression);
			}

			@Override
			protected Filter copy(Filter filter) {

				if (filter instanceof BinaryComparisonAbstract) {
					BinaryComparisonAbstract binFilter = (BinaryComparisonAbstract) filter;
					binFilter.setExpression1(copy(binFilter.getExpression1()));
					binFilter.setExpression2(copy(binFilter.getExpression2()));

					// TODO noch viel mehr faelle!!
					if (filter instanceof IsBetweenImpl) {
						IsBetweenImpl isbetween = (IsBetweenImpl) filter;
						isbetween
								.setExpression(copy(isbetween.getExpression()));
					}

				}
				return (Filter) super.copy(filter);
			}
		};

		dsv.visit(style);

		Style copiedCleanStyle = (Style) dsv.getCopy();
		return copiedCleanStyle;
	}

	/**
	 * Geopublisher allows to use one or two attributes for labelling, like
	 * 
	 * <code>
	 
	    <sld:Label>
          <ogc:PropertyName>ORTSRATSBE</ogc:PropertyName>: <ogc:PropertyName>ORB_SCHLUE</ogc:PropertyName>
        </sld:Label>

        </code>
	 * 
	 * Returns the second attribute found in the label expression of the given
	 * {@link TextSymbolizer} or <code>null</code>.
	 */
	public static PropertyName getSecondPropertyName(
			final SimpleFeatureType ft, final TextSymbolizer textSymbolizer) {
		try {

			final Expression labelExpression = textSymbolizer.getLabel();
			if (labelExpression instanceof PropertyName) {
				// There is no second property
				return null;
			} else if (labelExpression instanceof Function) {
				final Function filterExpression = (Function) labelExpression;
				final FilterAttributeExtractor attExVid = new FilterAttributeExtractor(
						ft);

				attExVid.visit(filterExpression, null);

				final String[] attributeNames = attExVid.getAttributeNames();
				if (attributeNames.length <= 1)
					return null;

				String prop2 = null;
				if (filterExpression.toString().indexOf(attributeNames[0]) < filterExpression
						.toString().indexOf(attributeNames[1])) {
					prop2 = attributeNames[1];
				} else {
					prop2 = attributeNames[0];
				}

				LOGGER.debug("Thie second for \n" + filterExpression + "\n is "
						+ prop2);

				return FilterUtil.FILTER_FAC2.property(prop2);
			}
		}

		catch (final Exception e) {
			LOGGER.error("Reading the seconf property for labelling failed", e);
		}

		return null;
	}

	/**
	 * Geopublisher allows to use one or two attributes for labelling, like
	 * 
	 * <code>
	 
	    <sld:Label>
          <ogc:PropertyName>ORTSRATSBE</ogc:PropertyName>: <ogc:PropertyName>ORB_SCHLUE</ogc:PropertyName>
        </sld:Label>

        </code>
	 * 
	 * Returns the first attribute found in the label expression of the given
	 * {@link TextSymbolizer}. If non is found for some reason, the method tries
	 * to return the first attribute from the schema. We expect the layer to
	 * have some attributes, because otherwise AS shouldn't even create the
	 * LabellingTab.
	 */
	public static PropertyName getFirstPropertyName(final SimpleFeatureType ft,
			final TextSymbolizer textSymbolizer) {
		final Expression labelExpression = textSymbolizer.getLabel();
		try {
			if (labelExpression instanceof PropertyName) {
				// There is no second property
				return (PropertyName) labelExpression;
			} else if (labelExpression instanceof Function) {
				final Function filterExpression = (Function) labelExpression;
				final FilterAttributeExtractor attExVid = new FilterAttributeExtractor(
						ft);

				attExVid.visit(filterExpression, null);

				final String[] attributeNames = attExVid.getAttributeNames();
				String prop1 = null;
				if (attributeNames.length == 1) {
					// Easy
					prop1 = attributeNames[0];
				} else {

					if (filterExpression.toString().indexOf(attributeNames[0]) < filterExpression
							.toString().indexOf(attributeNames[1])) {
						prop1 = attributeNames[0];
						// prop2 = attributeNames[1];
					} else {
						prop1 = attributeNames[1];
						// prop2 = attributeNames[0];
					}

				}
				return FilterUtil.FILTER_FAC2.property(prop1);
			}
		} catch (final Exception e) {
			LOGGER.error("Reading the second property for labelling failed", e);
			return null;
		}

		// TODO this is not rockhard.. could NPE
		return FilterUtil.FILTER_FAC2.property(ft.getAttributeDescriptors()
				.get(1).getLocalName());
	}

	/**
	 * AtlasStyler has a fixed system that allows to use one or two label
	 * attributes. If two are used, the are seperated by ": ". This mehtod stes
	 * one or two ProperybaName fields.
	 * 
	 * @param symbolizer
	 *            the {@link TextSymbolizer} to call setLabel(Expression) on.
	 * @param literalLabelField1
	 *            may not be null!
	 * @param literalLabelField2
	 *            may be <code>null</code>
	 */
	public static void setDoublePropertyName(final TextSymbolizer symbolizer,
			final PropertyName literalLabelField1,
			final PropertyName literalLabelField2) {
		if (literalLabelField2 != null
				&& !literalLabelField2.toString().equalsIgnoreCase("-")) {
			final Literal trenner = FilterUtil.FILTER_FAC2.literal(": ");
			final Function f3 = FilterUtil.FILTER_FAC2.function("strConcat",
					trenner, literalLabelField2);
			final Function bothExpression = FilterUtil.FILTER_FAC2.function(
					"strConcat", literalLabelField1, f3);
			LOGGER.debug("Extended label expression is now: " + bothExpression);
			symbolizer.setLabel(bothExpression);
		} else {
			symbolizer.setLabel(literalLabelField1);
		}
	}

	/**
	 * @return A {@link FeatureTypeStyle} that can be used style the given
	 *         geoObject during a blink or highlight process.
	 */
	public static FeatureTypeStyle createBlinkFeatureTypeStyle(Object geoObject) {
		if (geoObject instanceof GridCoverage2D
				|| geoObject instanceof org.geotools.coverage.grid.io.AbstractGridCoverage2DReader) {
			// Wenn irgendwann mal selection für raster möglich ist, dann hier
			// einen schöneren style erstellen.
			return GridUtil.createDefaultStyle().featureTypeStyles().get(0);
		}

		// We have vector-data. Now let's determine the type...
		GeometryForm geometryForm = null;
		if (geoObject instanceof FeatureCollection) {
			geometryForm = FeatureUtil
					.getGeometryForm((FeatureCollection<SimpleFeatureType, SimpleFeature>) geoObject);
		} else if (geoObject instanceof GeometryAttributeType) {
			geometryForm = FeatureUtil
					.getGeometryForm((GeometryAttributeType) geoObject);
		} else if (geoObject instanceof FeatureSource) {
			geometryForm = FeatureUtil
					.getGeometryForm((FeatureSource<SimpleFeatureType, SimpleFeature>) geoObject);
		}

		Symbolizer[] symbolizers = new Symbolizer[0];

		Graphic bg2;
		switch (geometryForm) {

		case POINT:

			Literal size0 = FeatureUtil.FILTER_FACTORY2.literal(26);
			Literal size1 = FeatureUtil.FILTER_FACTORY2.literal(28);

			Graphic bg1 = STYLE_FACTORY.createGraphic(new ExternalGraphic[0],
					new Mark[] { STYLE_FACTORY.createMark(
							FeatureUtil.FILTER_FACTORY2.literal("circle"),
							STYLE_BUILDER.createStroke(Color.red, 2.), null,
							size1, halfLit) },
					new org.geotools.styling.Symbol[0],
					FeatureUtil.FILTER_FACTORY2.literal(1), size1, zeroLit);

			PointSymbolizer ps1 = STYLE_FACTORY.createPointSymbolizer();
			ps1.setGraphic(bg1);
			symbolizers = LangUtil.extendArray(symbolizers, ps1);

			bg2 = STYLE_FACTORY.createGraphic(new ExternalGraphic[0],
					new Mark[] { STYLE_FACTORY.createMark(
							FeatureUtil.FILTER_FACTORY2.literal("circle"),
							STYLE_BUILDER.createStroke(Color.black, 2.),
							STYLE_BUILDER.createFill(Color.WHITE, 0.3), size0,
							halfLit) }, new org.geotools.styling.Symbol[0],
					FeatureUtil.FILTER_FACTORY2.literal(1), size0, zeroLit);

			PointSymbolizer ps2 = STYLE_FACTORY.createPointSymbolizer();
			ps2.setGraphic(bg2);

			symbolizers = LangUtil.extendArray(symbolizers, ps2);
			break;
		case POLYGON:

			PolygonSymbolizer pol1 = STYLE_BUILDER.createPolygonSymbolizer(
					STYLE_BUILDER.createStroke(Color.red, 4), null);
			symbolizers = LangUtil.extendArray(symbolizers, pol1);

			PolygonSymbolizer pol2 = STYLE_BUILDER.createPolygonSymbolizer(
					STYLE_BUILDER.createStroke(Color.black, 2), STYLE_BUILDER
							.createFill(Color.WHITE, .5));
			symbolizers = LangUtil.extendArray(symbolizers, pol2);

			break;

		case LINE:

			LineSymbolizer ls = STYLE_BUILDER.createLineSymbolizer(Color.red,
					6.);
			symbolizers = LangUtil.extendArray(symbolizers, ls);

			LineSymbolizer ls2 = STYLE_BUILDER.createLineSymbolizer(
					Color.black, 4.);
			symbolizers = LangUtil.extendArray(symbolizers, ls2);

			LineSymbolizer ls3 = STYLE_BUILDER.createLineSymbolizer(
					Color.white, 2.);
			symbolizers = LangUtil.extendArray(symbolizers, ls3);
			break;

		default:
			throw new IllegalArgumentException("Provide a suitable object.");
		}

		return STYLE_BUILDER.createFeatureTypeStyle(symbolizers, Double.NaN,
				Double.NaN);
	}


}
