/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.io;

/**
 * <hr>
 * <small><i>Diese Klasse wurde urspruenglich 2002/2003 von <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *            fuer die ADAGIOS-Java-Library (Universitaet Bonn) erstellt, 2009 in das
 *            SCHMITZM-Projekt uebernommen und teilweise erweitert.</i></small>
 * <hr>
 * Der <code>BinaryInputBuffer</code> erweitert den <code>InputBuffer</code>
 * durch die Moeglichkeit, neben einzelnen Bytes, auch die Binaerdarstellung
 * ganze Zahlen (mehrere Bytes auf einen Schub) in den Buffer schreiben
 * zu koennen.<br>
 * Um in gleicherweise ganze Zahlen aus dem Buffer zu lesen, kann die
 * Klasse <code>adagios.types.BinaryInputStream</code> verwendet werden:<br>
 * <code>
 * BinaryInputBuffer buf = new BinaryInputBuffer(1000);<br>
 * BinaryInputStream inp = new BinaryInputStream( buf );<br>
 * buf.writeInt( 123456 );<br>
 * buf.writeShort( 1000 );<br>
 * int i = inp.readInt();<br>
 * int s = inp.readShort();<br>
 * </code><br>
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class BinaryInputBuffer extends InputBuffer {

  /**
   * Erzeugt einen neuen <code>BinaryInputBuffer</code> der Groesse 1024.
   */
  public BinaryInputBuffer() {
    this(1024);
  }

  /**
   * Erzeugt einen neuen <code>BinaryInputBuffer</code>.
   * @param byteSize Groesse des Buffers in Byte
   */
  public BinaryInputBuffer(int byteSize) {
    super(byteSize);
  }

  /**
   * Schreibt einen <code>char</code>-Wert (2 Byte) in den Buffer.
   * Hierbei wird das 2er-Komplement verwendet.
   * @param byteOrder BigEndian/LittleEndian (vgl. {@link BinaryUtil})
   * @param c         zu schreibender Char
   * @see schmitzm.io.BinaryUtil#XDR
   * @see schmitzm.io.BinaryUtil#NDR
   * @see schmitzm.io.BinaryUtil#convertLongToBytes(byte,long,int)
   */
  public void writeChar(byte byteOrder, char c) {
    write( BinaryUtil.convertLongToBytes(byteOrder,c,2) );
  }

  /**
   * Schreibt einen <code>char</code>-Wert (2 Byte) in den Buffer.
   * Hierbei wird das 2er-Komplement in <i>BigEndian</i> verwendet.
   * @param c zu schreibender Char
   * @see schmitzm.io.BinaryUtil#XDR
   * @see #writeChar(byte,char)
   */
  public void writeChar(char c) {
    writeChar( BinaryUtil.XDR, c );
  }

  /**
   * Schreibt einen <code>short</code>-Wert (2 Byte) in den Buffer.
   * Hierbei wird das 2er-Komplement verwendet.
   * @param byteOrder BigEndian/LittleEndian (vgl. {@link BinaryUtil})
   * @param s         zu schreibender Short
   * @see schmitzm.io.BinaryUtil#XDR
   * @see schmitzm.io.BinaryUtil#NDR
   * @see schmitzm.io.BinaryUtil#convertLongToBytes(byte,long,int)
   */
  public void writeShort(byte byteOrder, short s) {
    write( BinaryUtil.convertLongToBytes(byteOrder,s,2) );
  }

  /**
   * Schreibt einen <code>short</code>-Wert (2 Byte) in den Buffer.
   * Hierbei wird das 2er-Komplement in <i>BigEndian</i> verwendet.
   * @param s zu schreibender Short
   * @see schmitzm.io.BinaryUtil#XDR
   * @see #writeShort(byte,short)
   */
  public void writeShort(short s) {
    writeShort( BinaryUtil.XDR, s );
  }

  /**
   * Schreibt einen <code>int</code>-Wert (4 Byte) in den Buffer.
   * Hierbei wird das 2er-Komplement verwendet.
   * @param byteOrder BigEndian/LittleEndian (vgl. {@link BinaryUtil})
   * @param i         zu schreibender Int
   * @see schmitzm.io.BinaryUtil#XDR
   * @see schmitzm.io.BinaryUtil#NDR
   * @see schmitzm.io.BinaryUtil#convertLongToBytes(byte,long,int)
   */
  public void writeInt(byte byteOrder, int i) {
    write( BinaryUtil.convertLongToBytes(byteOrder,i,4) );
  }

  /**
   * Schreibt einen <code>int</code>-Wert (4 Byte) in den Buffer.
   * Hierbei wird das 2er-Komplement in <i>BigEndian</i> verwendet.
   * @param i zu schreibender Int
   * @see schmitzm.io.BinaryUtil#XDR
   * @see #writeInt(byte,int)
   */
  public void writeInt(int i) {
    writeInt( BinaryUtil.XDR, i );
  }

  /**
   * Schreibt einen <code>long</code>-Wert (8 Byte) in den Buffer.
   * Hierbei wird das 2er-Komplement verwendet.
   * @param byteOrder BigEndian/LittleEndian (vgl. {@link BinaryUtil})
   * @param l         zu schreibender Long
   * @see schmitzm.io.BinaryUtil#XDR
   * @see schmitzm.io.BinaryUtil#NDR
   * @see schmitzm.io.BinaryUtil#convertLongToBytes(byte,long,int)
   */
  public void writeLong(byte byteOrder, long l) {
    write( BinaryUtil.convertLongToBytes(byteOrder,l,8) );
  }

  /**
   * Schreibt einen <code>long</code>-Wert (8 Byte) in den Buffer.
   * Hierbei wird das 2er-Komplement in <i>BigEndian</i> verwendet.
   * @param l zu schreibender Long
   * @see schmitzm.io.BinaryUtil#XDR
   * @see #writeLong(byte,long)
   */
  public void writeLong(long l) {
    writeLong( BinaryUtil.XDR, l );
  }

  /**
   * Schreibt einen <code>float</code>-Wert (4 Byte) in den Buffer.
   * @param byteOrder BigEndian/LittleEndian (vgl. {@link BinaryUtil})
   * @param f         zu schreibender Float
   * @see schmitzm.io.BinaryUtil#XDR
   * @see schmitzm.io.BinaryUtil#NDR
   * @see schmitzm.io.BinaryUtil#convertFloatToBytes(byte,float)
   */
  public void writeFloat(byte byteOrder, float f) {
    write( BinaryUtil.convertFloatToBytes(byteOrder,f) );
  }

  /**
   * Schreibt einen <code>double</code>-Wert (8 Byte) in den Buffer.
   * @param byteOrder BigEndian/LittleEndian (vgl. {@link BinaryUtil})
   * @param d         zu schreibender Double
   * @see schmitzm.io.BinaryUtil#XDR
   * @see schmitzm.io.BinaryUtil#NDR
   * @see schmitzm.io.BinaryUtil#convertDoubleToBytes(byte,double)
   */
  public void writeDouble(byte byteOrder, double d) {
    write( BinaryUtil.convertDoubleToBytes(byteOrder,d) );
  }

  /**
   * Schreibt einen <code>float</code>-Wert (8 Byte) in den Buffer.
   * Hierbei wird <i>BigEndian</i> verwendet.
   * @param f zu schreibender Float
   * @see schmitzm.io.BinaryUtil#XDR
   * @see #writeFloat(byte,float)
   */
  public void writeFloat(float f) {
    writeFloat( BinaryUtil.XDR, f );
  }

}
