/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.io;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStream;

/**
 * <hr>
 * <small><i>Diese Klasse wurde urspruenglich 2002/2003 von <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *            fuer die ADAGIOS-Java-Library (Universitaet Bonn) erstellt, 2009 in das
 *            SCHMITZM-Projekt uebernommen und teilweise erweitert.</i></small>
 * <hr>
 * <code>BinaryInputStream</code> erweitert die Klasse
 * <code>java.io.DataInputStream</code> um Methoden, mit denen man Werte
 * wahlweise in <i>BigEndian</i> oder <i>LittleEndian</i> aus dem Stream
 * auslesen kann.<br>
 * Die geerbten Methoden von <code>DataInputStream</code> interpretieren
 * die Byte-Folgen immer als <i>BigEndian</i> (das erste Byte im Stream
 * ist das hoechstwertige Byte für den zu lesenenden Wert).
 * @see java.io.DataInputStream
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class BinaryInputStream extends DataInputStream {
  /**
   * Erzeugt einen neuen Binary-Eingabestream.
   */
  public BinaryInputStream(InputStream in) {
    super(in);
  }

  /**
   * Erzeugt einen neuen Binary-Eingabestream, in dem aus dem
   * <code>byte</code>-Array ein {@link ByteArrayInputStream}
   * erzeugt wird.
   * @see java.io.ByteArrayInputStream
   */
  public BinaryInputStream(byte[] bytes) {
    this( new ByteArrayInputStream(bytes) );
  }

  /**
   * Erzeugt einen neuen Binary-Eingabestream, in dem aus dem String
   * ein {@link ByteArrayInputStream} erzeugt wird.
   * @see java.io.ByteArrayInputStream
   */
  public BinaryInputStream(String str) {
    this( new ByteArrayInputStream(str.getBytes()) );
  }

  /**
   * Liefert einen {@link InputStream} als {@link BinaryInputStream}.
   * Ist der uebergebene <code>InputStream</code> bereits eine Instanz von
   * <code>BinaryInputStream</code> wird diese (gecastet) zurueckgegeben.
   * Ansonsten wird eine neue <code>BinaryInputStream</code>-Instanz aus dem
   * <code>InputStream</code> erzeugt.
   */
  public static BinaryInputStream fromInputStream(InputStream in) {
    if (in instanceof BinaryInputStream) return (BinaryInputStream)in;
    return new BinaryInputStream(in);
  }

  /**
   * Liest einen <code>char</code>-Wert (2 Bytes) aus dem Stream.
   * Dabei wird das 2er-Komplement verwendet.
   * @param byteOrder BigEndian/LittleEndian (vgl. {@link BinaryUtil})
   * @see BinaryUtil#XDR
   * @see BinaryUtil#NDR
   * @see BinaryUtil#convertBytesToLong(byte,InputStream,int)
   */
  public char readChar(byte byteOrder) throws IOException {
    return (char)BinaryUtil.convertBytesToLong(byteOrder,this,2);
  }

  /**
   * Liest einen <code>short</code>-Wert (2 Bytes) aus dem Stream.
   * Dabei wird das 2er-Komplement verwendet.
   * @param byteOrder BigEndian/LittleEndian (vgl. {@link BinaryUtil})
   * @see BinaryUtil#XDR
   * @see BinaryUtil#NDR
   * @see BinaryUtil#convertBytesToLong(byte,InputStream,int)
   */
  public short readShort(byte byteOrder) throws IOException {
    return (short)BinaryUtil.convertBytesToLong(byteOrder,this,2);
  }

  /**
   * Liest einen <code>short</code>-Wert (2 Bytes) aus dem Stream, der ohne
   * Vorzeichen interpretiert wird. Dabei wird das 2er-Komplement verwendet.
   * @param byteOrder BigEndian/LittleEndian (vgl. {@link BinaryUtil})
   * @see BinaryUtil#XDR
   * @see BinaryUtil#NDR
   * @see BinaryUtil#convertBytesToLong(byte,InputStream,int)
   */
  public int readUnsignedShort(byte byteOrder) throws IOException {
    return (int)BinaryUtil.convertBytesToLong(byteOrder,this,2);
  }

  /**
   * Liest einen <code>int</code>-Wert (4 Bytes) aus dem Stream.
   * Dabei wird das 2er-Komplement verwendet.
   * @param byteOrder BigEndian/LittleEndian (vgl. {@link BinaryUtil})
   * @see BinaryUtil#XDR
   * @see BinaryUtil#NDR
   * @see BinaryUtil#convertBytesToLong(byte,InputStream,int)
   */
  public int readInt(byte byteOrder) throws IOException {
    return (int)BinaryUtil.convertBytesToLong(byteOrder,this,4);
  }

  /**
   * Liest einen <code>long</code>-Wert (8 Bytes) aus dem Stream.
   * Dabei wird das 2er-Komplement verwendet.
   * @param byteOrder BigEndian/LittleEndian (vgl. {@link BinaryUtil})
   * @see BinaryUtil#XDR
   * @see BinaryUtil#NDR
   * @see BinaryUtil#convertBytesToLong(byte,InputStream,int)
   */
  public long readLong(byte byteOrder) throws IOException {
    return (long)BinaryUtil.convertBytesToLong(byteOrder,this,8);
  }

  /**
   * Liest einen <code>double</code>-Wert (8 Bytes) aus dem Stream.
   * @param byteOrder BigEndian/LittleEndian (vgl. {@link BinaryUtil})
   * @see BinaryUtil#XDR
   * @see BinaryUtil#NDR
   * @see BinaryUtil#convertBytesToDouble(byte,InputStream)
   */
  public double readDouble(byte byteOrder) throws IOException {
    return (double)BinaryUtil.convertBytesToDouble(byteOrder,this);
  }

  /**
   * Liest einen <code>float</code>-Wert (4 Bytes) aus dem Stream.
   * @param byteOrder BigEndian/LittleEndian (vgl. {@link BinaryUtil})
   * @see BinaryUtil#XDR
   * @see BinaryUtil#NDR
   * @see BinaryUtil#convertBytesToFloat(byte,InputStream)
   */
  public float readFloat(byte byteOrder) throws IOException {
    return (long)BinaryUtil.convertBytesToFloat(byteOrder,this);
  }

}
