/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.io;

import java.io.EOFException;
import java.io.IOException;
import java.io.InputStream;

/**
 * <hr>
 * <small><i>Diese Klasse wurde urspruenglich 2002/2003 von <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *            fuer die ADAGIOS-Java-Library (Universitaet Bonn) erstellt, 2009 in das
 *            SCHMITZM-Projekt uebernommen und teilweise erweitert.</i></small>
 * <hr>
 * Diese Klasse enthaelt statische Methoden, die das Arbeiten mit Binaerstrings
 * erleichtern.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class BinaryUtil {

  /**
   * Numerische Konstante fuer die ByteOrder 'Big Endian'.<br>
   * Dabei steht das hoechstwertige Byte im ERSTEN Byte einer Folge.
   */
  public static final byte XDR = 0;

  /**
   * Numerische Konstante fuer die ByteOrder 'Little Endian'.<br>
   * Dabei steht das niederstwertige Byte im ERSTEN Byte einer Folge.
   */
  public static final byte NDR = 1;


  /**
   * Konvertiert Bytes eines Streams in einen <code>long</code>-Wert.
   * @param byteOrder 0 = BigEndian = hoechstwertiges Byte zuerst,
   *                  1 = LittleEndian = niederwertigstes Byte zuerst
   * @param input     Eingabe-Stream
   * @param byteCount Anzahl der Bytes, die aus dem Stream gelesen und in einen
   *                  Long umgewandelt werden sollen
   * @exception java.io.IOException falls das Lesen aus dem Stream scheitert
   * @exception java.io.EOFException falls nicht genug Bytes auf dem Stream liegen
   */
  public static long convertBytesToLong(byte byteOrder, InputStream input, int byteCount) throws IOException {
    byte[] bytes = new byte[byteCount];
    if ( input.read(bytes) < byteCount )
      throw new EOFException("Not enough Bytes available on InputStream");

    return convertBytesToLong(byteOrder,bytes);
  }

  /**
   * Konvertiert eine Reihe von Bytes in einen <code>long</code>-Wert.<br>
   * Fuer negative Werte wird eine Darstellung im 2er-Komplement erwartet.
   * @param byteOrder 0 = BigEndian = hoechstwertiges Byte zuerst,
   *                  1 = LittleEndian = niederwertigstes Byte zuerst
   * @param bytes     Eingabe-Bytes
   */
  public static long convertBytesToLong(byte byteOrder, byte[] bytes) {
    if (bytes == null || bytes.length == 0)
      throw new ArrayIndexOutOfBoundsException("At least 1 Byte needed for convert to long!!");

    // Bemerkung:
    // ==========
    // Wegen des sukzessiven Bit-Shift, muss die Summierung mit dem
    // hoechstwertigen Byte begonnen werden.
    // Dies steht bei 'BigEndian' im ersten, bei 'LittleEndian' im
    // letzten Byte.
    // Der zusaetzliche Summand 256 (wenn bytes[i]<0), ist notwendig, da
    // der Wertebereich von 'byte' von -128 bis +127 reicht (nicht 0..255!!).
    // Ein einfacher Cast von 'byte' zu 'char' ist leider nicht ausreichend,
    // da 'char' durch 16Bit repraesentiert wird. Beim Cast wird das
    // Vorzeichen-Bit also einfach uebernommen und der Wert bleibt unveraendert.
    // Was gaebe ich in diesem Moment fuer das kleine Woertchen 'unsigned'!! :-)

    // Das erste Bit des hoechstwerigen Bytes ist das Vorzeichenbit. Um das
    // Vorzeichen im long-Ergebnis zu erhalten, muessen die fuehrenden Bits
    // (die durch den byte-Array nicht abgedeckt werden) mit diesem VZ-Bit
    // belegt werden.
    // Dies geht am Einfachsten in dem man das Ergebnis initial komplett mit
    // dem VZ-Bit belegt:
    // a) hoechstwertiges Byte in long konvertierten
    //    --> nun hat 'result' dasselbe VZ wie die in bytes[] kodierte Zahl
    // b) die Bits des hoechstwertigen Bytes eliminieren, die nicht das VZ-Bit
    //    sind (durch den Rechts-Shift wird linkt mit dem VZ-Bit aufgefuellt!)
    long result = (byteOrder == XDR) ? bytes[0] : bytes[bytes.length-1];
    result >>= 7;

    if ( byteOrder == XDR )
      // Big Endian (hoechstwertiges Byte zuerst im Array)
      for(int i=0; i<bytes.length; i++)
//      result = (result << 8) + bytes[i] + (bytes[i]<0 ? 256 : 0);
        result = (result << 8) | getUnsignedByte(bytes[i]);
    else
      // Little Endian (niederwertiges Byte zuerst im Array)
      for(int i=bytes.length-1; i >= 0; i--) {
//      result = (result << 8) + bytes[i] + (bytes[i]<0 ? 256 : 0);
        result = (result << 8) | getUnsignedByte(bytes[i]);
      }
    return result;
  }

  /**
   * Konvertiert 4 Bytes eines Streams in einen <code>float</code>-Wert.
   * @param byteOrder 0 = BigEndian = hoechstwertiges Byte zuerst,
   *                  1 = LittleEndian = niederwertigstes Byte zuerst
   * @param input     Eingabe-Stream
   * @exception java.io.IOException falls das Lesen aus dem Stream scheitert
   * @exception java.io.EOFException falls nicht genug Bytes auf dem Stream liegen
   */
  public static float convertBytesToFloat(byte byteOrder, InputStream input) throws IOException {
    byte[] bytes = new byte[4];
    if ( input.read(bytes) < 4 )
      throw new EOFException("Not enough Bytes available on InputStream");

    return convertBytesToFloat(byteOrder,bytes);
  }

  /**
   * Konvertiert 4 Bytes in einen <code>float</code>-Wert.<br>
   * @param byteOrder 0 = BigEndian = hoechstwertiges Byte zuerst,
   *                  1 = LittleEndian = niederwertigstes Byte zuerst
   * @param bytes     8 Eingabe-Bytes
   */
  public static float convertBytesToFloat(byte byteOrder, byte[] bytes) {
    if (bytes == null || bytes.length != 4)
      throw new ArrayIndexOutOfBoundsException("Exactly 4 Byte needed for convert to float!!");

    // Bytes wie einen Integer einlesen
    int intValue = (int)convertBytesToLong(byteOrder, bytes);
    // Bits in Float konvertieren
    return Float.intBitsToFloat( intValue );
  }

  /**
   * Konvertiert 8 Bytes eines Streams in einen <code>double</code>-Wert.
   * @param byteOrder 0 = BigEndian = hoechstwertiges Byte zuerst,
   *                  1 = LittleEndian = niederwertigstes Byte zuerst
   * @param input     Eingabe-Stream
   * @exception java.io.IOException falls das Lesen aus dem Stream scheitert
   * @exception java.io.EOFException falls nicht genug Bytes auf dem Stream liegen
   */
  public static double convertBytesToDouble(byte byteOrder, InputStream input) throws IOException {
    byte[] bytes = new byte[8];
    if ( input.read(bytes) < 8 )
      throw new EOFException("Not enough Bytes available on InputStream");

    return convertBytesToDouble(byteOrder,bytes);
  }

  /**
   * Konvertiert 8 Bytes in einen <code>double</code>-Wert.<br>
   * @param byteOrder 0 = BigEndian = hoechstwertiges Byte zuerst,
   *                  1 = LittleEndian = niederwertigstes Byte zuerst
   * @param bytes     8 Eingabe-Bytes
   */
  public static double convertBytesToDouble(byte byteOrder, byte[] bytes) {
    if (bytes == null || bytes.length != 8)
      throw new ArrayIndexOutOfBoundsException("Exactly 8 Byte needed for convert to double!!");

    // Bytes wie einen Long einlesen
    long longValue = convertBytesToLong(byteOrder, bytes);
    // Bits in Double konvertieren
    return Double.longBitsToDouble( longValue );
  }

  /**
   * Konvertiert einen <code>long</code>-Wert in eine Reihe von Bytes.
   * Fuer negative Werte wird das 2er-Komplement verwendet.
   * @param byteOrder 0 = BigEndian = hoechstwertiges Byte zuerst,
   *                  1 = LittleEndian = niederwertigstes Byte zuerst
   * @param value     Eingabewert
   * @param size      Groesse des Rueckgabe-Arrays; unbenoetigte Felder werden
   *                  aufgefuellt (fuer negative Werte mit 1, sonst mit 0);<br>
   *                  Bei <code>size=0</code> ist der Rueckgabe-Array genau
   *                  so gross wie noetig.
   * @exception java.lang.ArrayIndexOutOfBoundException falls der long-Wert
   *            nicht in den Rueckgabe-Array passt.
   */
  public static byte[] convertLongToBytes(byte byteOrder, long value, int size) {
    // 2er-Komplement fuer negative Zahlen:
    // a) Verwende den positiven Wert
    // b) Subtrahiere 1
    // c) Bestimme die Bytes nach B+V (Betrag+VZ)
    // d) invertiere alle Bits
    if ( value < 0) {
      long posValue = Math.abs(value) - 1;                        // a) & b)
      byte[] bytes = convertLongToBytes(byteOrder,posValue,size); // c)
      for (int i=0; i<bytes.length; i++) bytes[i] ^= -1;          // d)
      return bytes;
    }

    // Anzahl an Bytes, die fuer die Zahl benoetigt werden (im 2er-Komplement)
    // --> multiplikation mit 2 (Bitshift!), da ein Bit mehr fuer das
    //     Vorzeichen benoetigt wird
    int bytesNeeded = (int)(Math.log(value<<1)/Math.log(256) + 1);
    if ( bytesNeeded > size )
      throw new ArrayIndexOutOfBoundsException(bytesNeeded+" Bytes needed to store "+value+" in 2's complement!!");

    // Wert in B+V ausrechenen
    byte[] bytes = new byte[ size==0 ? bytesNeeded : size ];
    int i=0; // Array-Zaehler
    // erstmal in LittleEndian ausrechnen
    for(int potenz=0; value>0; i++, potenz++) {
//    bytes[i] = (byte)(value%256);
//    value /= 256;
      bytes[i] = (byte)(value & 255);
      value >>= 8; // naechstes Byte
    }

    // Big Endian (hoechstwertiges Byte zuerst)
    // --> Array-Reihenfolge vertauschen
    if ( byteOrder == XDR ) {
      for (i=0; i<bytes.length/2; i++){
        // Dreieckstausch
        byte help = bytes[i];
        bytes[i] = bytes[bytes.length-i-1];
        bytes[bytes.length-i-1] = help;
      }
    }

    return bytes;
  }

  /**
   * Konvertiert einen <code>long</code>-Wert in eine Reihe von Bytes.
   * Fuer negative Werte wird das 2er-Komplement verwendet.
   * Die Bytes werden allerdings nicht als <code>byte</code> zurueckgegeben,
   * sondern als <code>int</code>, der einen <code>unsigned byte</code>
   * simuliert.<br>
   * Es werden also immer positive Werte zurueckgegeben!!
   * @param byteOrder 0 = BigEndian = hoechstwertiges Byte zuerst,
   *                  1 = LittleEndian = niederwertigstes Byte zuerst
   * @param value     Eingabewert
   * @param size      Groesse des Rueckgabe-Arrays; unbenoetigte Felder werden
   *                  aufgefuellt (fuer negative Werte mit 1, sonst mit 0);<br>
   *                  Bei <code>size=0</code> ist der Rueckgabe-Array genau
   *                  so gross wie noetig.
   * @exception java.lang.ArrayIndexOutOfBoundException falls der long-Wert
   *            nicht in den Rueckgabe-Array passt.
   */
  public static int[] convertLongToUnsignedBytes(byte byteOrder, long value, int size) {
    return getUnsignedBytes( convertLongToBytes(byteOrder,value,size) );
  }

  /**
   * Konvertiert einen <code>byte</code>-Wert (-128..127) in ein
   * <i>unsigned</i> <code>byte</code> (0..255).
   * Der Rueckgabetyp muss hierfuer ein <code>int</code> sein.
   * @param b Wert zwischen -128 und 127
   * @return <code>(int)b & 255</code>
   */
  public static int getUnsignedByte(byte b) {
    return (int)b & 255;
  }

  /**
   * Konvertiert einen Array von <code>byte</code>-Werten (-128..127) in einen
   * Array von <i>unsigned</i> <code>bytes</code> (0..255).
   * Der Rueckgabetyp muss hierfuer ein <code>int</code> sein.
   * @param signedByte {@code byte}-Array
   */
  public static int[] getUnsignedBytes(byte signedByte[]) {
    int[] unsignedByte = new int[ signedByte.length ];
    for (int i=0; i<signedByte.length; i++)
      unsignedByte[i] = getUnsignedByte( signedByte[i] );
    return unsignedByte;
  }

  /**
   * Konvertiert einen <code>float</code>-Wert in 4 Bytes.
   * @param byteOrder 0 = BigEndian = hoechstwertiges Byte zuerst,
   *                  1 = LittleEndian = niederwertigstes Byte zuerst
   * @param value     Eingabewert
   * @return byte[4]
   */
  public static byte[] convertFloatToBytes(byte byteOrder, float value) {
    int intValue = Float.floatToIntBits(value);
    return convertLongToBytes(byteOrder, intValue, 4);
  }

  /**
   * Konvertiert einen <code>float</code>-Wert in 4 Bytes.
   * Die Bytes werden allerdings nicht als <code>byte</code> zurueckgegeben,
   * sondern als <code>int</code>, der einen <code>unsigned byte</code>
   * simuliert.<br>
   * Es werden also immer positive Werte zurueckgegeben!!
   * @param byteOrder 0 = BigEndian = hoechstwertiges Byte zuerst,
   *                  1 = LittleEndian = niederwertigstes Byte zuerst
   * @param value     Eingabewert
   * @return byte[4]
   */
  public static int[] convertFloatToUnsignedBytes(byte byteOrder, float value) {
    return getUnsignedBytes( convertFloatToBytes(byteOrder,value) );
  }

  /**
   * Konvertiert einen <code>double</code>-Wert in 8 Bytes.
   * @param byteOrder 0 = BigEndian = hoechstwertiges Byte zuerst,
   *                  1 = LittleEndian = niederwertigstes Byte zuerst
   * @param value     Eingabewert
   * @return byte[8]
   */
  public static byte[] convertDoubleToBytes(byte byteOrder, double value) {
    long longValue = Double.doubleToLongBits(value);
    return convertLongToBytes(byteOrder, longValue, 8);
  }

  /**
   * Konvertiert einen <code>double</code>-Wert in 8 Bytes.
   * Die Bytes werden allerdings nicht als <code>byte</code> zurueckgegeben,
   * sondern als <code>int</code>, der einen <code>unsigned byte</code>
   * simuliert.<br>
   * Es werden also immer positive Werte zurueckgegeben!!
   * @param byteOrder 0 = BigEndian = hoechstwertiges Byte zuerst,
   *                  1 = LittleEndian = niederwertigstes Byte zuerst
   * @param value     Eingabewert
   * @return byte[8]
   */
  public static int[] convertDoubleToUnsignedBytes(byte byteOrder, double value) {
    return getUnsignedBytes( convertDoubleToBytes(byteOrder,value) );
  }

}
