/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.io;

import java.io.IOException;
import java.io.InputStream;

/**
 * <hr>
 * <small><i>Diese Klasse wurde urspruenglich 2002/2003 von <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *            fuer die ADAGIOS-Java-Library (Universitaet Bonn) erstellt, 2009 in das
 *            SCHMITZM-Projekt uebernommen und teilweise erweitert.</i></small>
 * <hr>
 * Diese Klasse stellt einen Eingabe-Stream dar, der Daten auf Basis
 * mehrerer anderen <code>InputStream</code>-Instanzen liefert. Zunaechst
 * werden Bytes aus dem ersten Stream geliefert, wenn dieser Stream keine
 * Daten mehr liefern kann aus dem zweiten, usw.<br>
 * Sobald ein Stream einmal komplett geleert wurde, liefert die
 * <code>read()</code>-Methode nur noch Bytes aus einem der naechsten
 * Streams. Auch wenn der "leere" Stream spaeter wieder Daten liefern koennte,
 * werden diese nicht mehr beruecksichtigt, sondern einfach ignoriert.<br>
 * Diese Klasse kann z.B. verwendet werden, um bereits aus einem InputStream
 * gelesene Daten (z.B. einen Header) diesem wieder voranzustellen, um
 * den InputStream "als Original" weiterzugeben.<br>
 * <b><u>Beispiel</u>:</b><br>
 * Ein Klasse liest Text-Repraesentationenen <u>unterschiedlicher</u> Typen
 * aus einer Datei. Erst nachdem der Header eines jeden Objekts eingelesen
 * wurde, kann entschieden werden, welche Typ-Klasse instanziiert werden muss.
 * Um deren <code>initFromText(InputStream)</code>-Methode aufrufen zu koennen,
 * wird aber wieder der komplette InputStream inklusive Header benoetigt.<br>
 * Hierzu kann ein <code>CombinedInputStream</code>, zusammengesetzt aus einem
 * neuen <code>ByteArrayInputStream</code> fuer den bereits eingelesenen
 * Header und dem Eingabe-Stream der Datei, erzeugt werden, welcher der
 * <code>initFromText(InputStream)</code>-Methode uebergeben wird.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class CombinedInputStream extends InputStream {
  private InputStream[] input     = null;
  private int           actStream = 0;
  private int           avBytes   = 0;


  /**
   * Erzeugt einen neuen <code>CombinedInputStream</code>.
   * @param input Eingabe-Streams aus denen sukzessive gelesen wird
   * @exception java.io.IOException falls beim Ermitteln der insgesamt
   *            zur Verfuegung stehenden Bytes ein Fehler auftritt
   */
  public CombinedInputStream(InputStream[] input) throws IOException{
    this.input = input;
    regenerateAvailableBytes();
  }

  /**
   * Erzeugt einen neuen <code>CombinedInputStream</code>.
   * @param size Anzahl an Eingabe-Streams auf die der
   *        <code>CombinedInputStream</code> dimensioniert wird
   */
  public CombinedInputStream(int size) {
    this.input = new InputStream[size];
  }

  /**
   * Liefert die Anzahl an Streams aus der der <code>CombinedInputStream</code>
   * zusammengesetzt wurde.
   */
  public int getStreamCount() {
    return input.length;
  }

  /**
   * Setzt einen Teil-Stream. Die Anzahl der insgesamt zur Verfuegung stehenden
   * Bytes wird danach neu berechnet.
   * @see #regenerateAvailableBytes()
   */
  public void setInputStream(int i, InputStream stream) throws IOException {
    this.input[i] = stream;
    regenerateAvailableBytes();
  }

  /**
   * Errechnet die Anzahl der insgesamt zur Verfuegung stehenden Bytes neu.
   * Dabei werden nur der aktuelle Stream und die darauf folgenden beruecksichtigt.
   * Die bereits ausgelesenen Streams werden ignoriert, auch wenn in ihnen
   * nun evt. wieder Bytes zur Verfuegung stehen.
   */
  public void regenerateAvailableBytes() throws IOException {
    avBytes = 0;
    for (int i=actStream; i<input.length; i++)
      if ( input[i] != null ) avBytes += input[i].available();
  }

  /**
   * Liefert die gesamt Anzahl von Bytes, die in den Eingabe-Streams
   * noch zur Verfuegung stehen.<br>
   * <b>Bemerkung:</b><br>
   * Aus Effizienzgruenden wird die Anzahl der Verfuegung stehenden Bytes
   * nicht bei jedem Aufruf neu errechnet, sondern nur beim Setzen eines
   * neuen Streams (mittels <code>setInputStream(..)</code>). Danach wird
   * nach bei jedem <code>read()</code>-Aufruf ein Zaehler herabgesetzt.<br>
   * Aendert sich die Byte-Zahl eines Teil-Streams waehrend der Benutzung des
   * <code>CombinedInputStream</code>, sollte die Methode
   * <code>regenerateAvailableBytes()</code> aufgerufen werden, um die Anzahl
   * der zur Verfuegung stehenden Bytes neu zu berechnen!!
   * @exception java.io.IOException falls der Zugriff auf einen der beiden
   *            Eingabe-Streams scheitert
   * @see #regenerateAvailableBytes()
   */
  public int available() {
    return avBytes;
  }

  /**
   * Liefert ein (unsigned) Byte aus dem aktuellen Stream.
   * Solange der erste Stream noch Daten liefern kann, werden seine Bytes
   * zurueckgegeben. Danach die Bytes des zweiten Streams, usw.
   * @return -1, falls keiner der Streams noch Daten liefern kann
   * @exception java.io.IOException falls der Zugriff auf einen der
   *            Eingabe-Streams scheitert
   * @see #available()
   */
  public int read() throws IOException {
    while( actStream < input.length && input[actStream].available() < 1 ) actStream++;
    if ( actStream >= input.length ) return -1;
    avBytes--;
    return input[actStream].read();
  }

  /**
   * Schliesst alle Eingabe-Streams.
   * @exception java.io.IOException falls der Zugriff auf einen der
   *            Eingabe-Streams scheitert
   */
  public void close() throws IOException {
    for (int i=0; i<input.length; i++) this.input[i].close();
  }

}
