/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.io;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URL;

import org.apache.log4j.Logger;

/**
 * Diese Klasse stellt statische Methoden fuer die Ein/Ausgabe zur Verfuegung.
 * 
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *         (University of Bonn/Germany)
 * 
 * @version 1.0
 */
public class IOUtil {

	private static Logger LOGGER = Logger.getLogger(IOUtil.class.getName());

	/**
	 * {@link FileFilter}, der alle Dateien akzeptiert, aber keine Verzeichnisse.
	 */
	public static final FileFilter ALL_FILES_FILTER = createSimpleFileFilter(
			"*", FilterMode.FILES_ONLY);
	/**
	 * {@link FileFilter}, der alle Verzeichnisse akzeptiert, aber keine
	 * "normalen" Dateien.
	 */
	public static final FileFilter ALL_DIRS_FILTER = createSimpleFileFilter(
			"*", FilterMode.DIR_ONLY);
	/** {@link FileFilter}, der alle Dateien und Verzeichnisse akzeptiert. */
	public static final FileFilter ALL_FILTER = createSimpleFileFilter("*",
			FilterMode.ALL);

	/**
	 * Modi fuer Filter.
	 * 
	 * @see IOUtil#createSimpleFileFilter(String, FilterMode)
	 * @see IOUtil#createRegExFileFilter(String, FilterMode)
	 * @author mojays
	 * 
	 */
	public static enum FilterMode {
		/** Der Filter liefert nur Dateien, aber keine Verzeichnisse. */
		FILES_ONLY,
		/** Der Filter liefert nur Verzeichnisse, aber keine Dateien. */
		DIR_ONLY,
		/** Der Filter liefert Dateien und Verzeichnisse. */
		ALL
	}

	/**
	 * Prueft, ob ein {@link File} einen angegebenen {@link FilterMode}
	 * erfuellt.
	 * 
	 * @param file
	 *            eine Datei oder ein Verzeichnis
	 * @param mode
	 *            Filter-Modus (wenn {@code null} wird {@link FilterMode#ALL}
	 *            angenommen
	 * @return {@code false} wenn die angegebenen Datei {@code null} ist
	 */
	public static boolean fileFitsFilterMode(File file, FilterMode mode) {
		if (mode == null)
			mode = FilterMode.ALL;
		final boolean acceptFiles = FilterMode.ALL.equals(mode)
				|| FilterMode.FILES_ONLY.equals(mode);
		final boolean acceptDirs = FilterMode.ALL.equals(mode)
				|| FilterMode.DIR_ONLY.equals(mode);

		return file != null
				&& (file.isDirectory() && acceptDirs || acceptFiles);
	}

	/**
	 * Liefert den Index des Dateinames, an der die Dateinamen-Erweiterung
	 * (inkl. Punkt) beginnt.
	 * 
	 * @param file
	 *            Datei
	 */
	public static int getFileExtIdx(File file) {
		return getFileExtIdx(file.getAbsolutePath());
	}

	/**
	 * Liefert den Index des Dateinames, an der die Dateinamen-Erweiterung
	 * (inkl. Punkt) beginnt.
	 * 
	 * @param file
	 *            Datei
	 */
	public static int getFileExtIdx(String pathOrName) {
		int lastPointIdx = pathOrName.lastIndexOf('.');
		int lastSepIdx = pathOrName.lastIndexOf(File.separatorChar);
		if (lastPointIdx < lastSepIdx)
			return -1;
		return lastPointIdx;
	}

	/**
	 * Liefert die Dateinamen-Erweiterung (z.B. ".exe") einer Datei.
	 * 
	 * @param file
	 *            Datei
	 * @return Leerstring, falls die Datei keine Erweiterung hat
	 */
	public static String getFileExt(File file) {
		return getFileExt(file, true);
	}

	/**
	 * Liefert die Dateinamen-Erweiterung (z.B. ".exe") einer Datei.
	 * 
	 * @param file
	 *            Datei
	 * @param withDot
	 *            wenn {@code false} wird die Dateinamen-Erweiterung ohne den
	 *            fuehrenden Punkt zurueckgegeben (z.B. "exe" statt ".exe")
	 * @return Leerstring, falls die Datei keine Erweiterung hat
	 */
	public static String getFileExt(File file, boolean withDot) {
		int extIdx = getFileExtIdx(file);
		if (extIdx < 0)
			return "";
		if (!withDot)
			extIdx++;
		return file.getAbsolutePath().substring(extIdx);
	}

	/**
	 * Haengt an einen Dateinamen-Erweiterung eine Erweiterung an, sofern diese
	 * noch nicht angehaengt ist.
	 * 
	 * @param file
	 *            Datei
	 * @param newExt
	 *            neue Dateinamen-Erweiterung (mit oder ohne ".")
	 * @return neues {@link File}-Objekt
	 */
	public static File appendFileExt(File file, String newExt) {
		if (!newExt.startsWith("."))
			newExt = "." + newExt;
		String oldExt = getFileExt(file, true);
		if (!oldExt.equalsIgnoreCase(newExt))
			file = new File(file.getAbsolutePath() + newExt);
		return file;
	}

	/**
	 * Aendert die Dateinamen-Erweiterung (z.B. ".exe") einer Datei.
	 * 
	 * @param file
	 *            Datei
	 * @param newExt
	 *            neue Dateinamen-Erweiterung (ohne ".")
	 * @return neues {@link File}-Objekt
	 */
	public static File changeFileExt(File file, String newExt) {
		String baseName = getBaseFilePath(file);
		return new File(baseName + "." + newExt);
	}

	/**
	 * Changes the ending (e.g. ".sld") of a {@link URL}
	 * 
	 * @param url
	 *            {@link URL} like <code>file:/sds/a.bmp</code>
	 * @param postfix
	 *            New file extension for the {@link URL} without <code>.</code>
	 * 
	 * @return A new {@link URL} with new extension.
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 * 
	 */
	public static URL changeUrlExt(final URL url, final String postfix)
			throws IllegalArgumentException {
		String a = url.toExternalForm();
		final int lastDotPos = a.lastIndexOf('.');
		if (lastDotPos >= 0)
			a = a.substring(0, lastDotPos);
		a = a + "." + postfix;
		try {
			return new URL(a);
		} catch (final MalformedURLException e) {
			throw new IllegalArgumentException("can't create a new URL for "
					+ url + " with new extension " + postfix, e);
		}
	}

	/**
	 * Die Funktion soll der Funktion File.getParent() fuer URLs entsprechen.
	 * Die URL wird in einen Sting konvertiert und dann (kuerzer) neu
	 * zusammengesetzt. Falls eine Verkuerzung nicht moeglich ist, wird eine
	 * {@link MalformedURLException} geworfen
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 * 
	 * 
	 */
	public static URL getParentUrl(final URL url) throws MalformedURLException {
		String a = url.toExternalForm();
		final int lastDotPos = a.lastIndexOf('/');
		if (lastDotPos >= 0)
			a = a.substring(0, lastDotPos);

		/**
		 * The parent of jar:file:some!/bar.file is jar:file:some!/, not
		 * jar:file:some!
		 */
		if (a.endsWith("!"))
			a += "/";

		return new URL(a);
	}

	/**
	 * @deprecated use getParentUrl()
	 */
	public static URL getParentURL(URL url) throws MalformedURLException {
		return getParentUrl(url);
	}

	/**
	 * Erweitert eine {@link URL}. Bei der uebergebenen URL muss es sich um ein
	 * Verzeichnis handeln, sonst wird eine {@link MalformedURLException}
	 * geschmissen
	 * 
	 * @param base
	 * @param string
	 * 
	 * @throws MalformedURLException
	 */
	public static URL extendURL(URL base, String string)
			throws MalformedURLException {
		String a = base.toExternalForm();
		if (!a.endsWith("/"))
			a += "/";
		a += string;
		return new URL(a);
	}

	/**
	 * Liefert den Pfad ohne Erweiterung (z.B. ohne ".exe") einer Datei.
	 * 
	 * @param file
	 *            Datei
	 * 
	 * @see #getBaseFileName(File)
	 */
	public static String getBaseFilePath(File file) {
		int extIdx = getFileExtIdx(file);
		if (extIdx < 0)
			return file.getAbsolutePath();
		return file.getAbsolutePath().substring(0, extIdx);
	}

	/**
	 * Liefert den Dateinamen ohne Erweiterung (z.B. ohne ".exe") einer Datei.
	 * 
	 * @param file
	 *            Datei
	 * 
	 * @see #getBaseFilePath(File)
	 */
	public static String getBaseFileName(File file) {
		int extIdx = getFileExtIdx(file.getName());
		if (extIdx < 0)
			return file.getName();
		return file.getName().substring(0, extIdx);
	}

	/**
	 * Liefert das oberste Verzeichnis zu einer Datei (Haupt-Verzeichnis).
	 * 
	 * @param file
	 *            Datei
	 */
	public static File getRootDir(File file) {
		File temp = new File(file.getAbsolutePath());
		while (temp.getParentFile() != null)
			temp = temp.getParentFile();
		return temp;
	}

	/**
	 * Erzeugt ein neues File-Objekt relativ zu einer Pfad-Angabe.<br>
	 * Im Normalfall setzt sich der neue Datei-Pfad aus <code>baseDir</code> und
	 * <code>path</code> zusammen, also <i>newFile = <code>baseDir</code> +
	 * <code>path<code></i>.<br>
	 * <b>Ausnahme 1:</b><br>
	 * Beginnt <code>path</code> mit einem Slash (bzw. Backslash bei Windows),
	 * ist <i>newFile = ROOT(<code>baseDir</code>) + <code>path<code></i>,
	 * wobei <i>ROOT(<code>baseDir</code>)</i></center> das Hauptverzeichnis
	 * (Laufwerk) des Basis-Verzeichnisses bezeichnet.<br>
	 * <b>Ausnahme 2:</b><br>
	 * Stellt <code>path</code> bereits eine absolute Pfadangabe dar (inkl.
	 * Laufwerksbezeichnung), ist <i>newFile = <code>path<code></i>.
	 * 
	 * @param baseDir
	 *            Basis-Verzeichnis
	 * @param path
	 *            Pfad/Datei-Angabe relativ zu <code>baseDir</code>
	 * @see #getRootDir(File)
	 */
	public static File createRelativeFile(File baseDir, String path) {
		File file = new File(path);

		// In relPath steht absolute Pfadangabe -> BaseDir ignorieren
		if (path.equalsIgnoreCase(file.getAbsolutePath()))
			return file;

		// Source-Tag beginnt mit '\' -> Source-Angabe relativ zum
		// Haupt-Verzeichnis
		if (path.startsWith(File.separator))
			return new File(IOUtil.getRootDir(baseDir), path);

		// Sonst: BaseDir als Basis fuer Source verwenden
		return new File(baseDir, path);
	}

	/**
	 * Versucht einen Eingabe-Stream zu schliessen.<br>
	 * Dabei werden bei Misserfolg keine Exceptions geworfen! SK: bei
	 * <code>null</code> wird true zurueckgegeben
	 * 
	 * @param in
	 *            zu schliessender Stream
	 * @return <code>false</code> falls das Schliessen nicht erfolgreich war
	 */
	public static boolean closeInputStream(InputStream in) {
		if (in == null)
			return true;
		try {
			in.close();
		} catch (Exception err) {
			return false;
		}
		return true;
	}

	/**
	 * Versucht alle gegebenen Eingabe-Streams zu schliessen.<br>
	 * Dabei werden bei Misserfolg keine Exceptions geworfen!
	 * 
	 * @param in
	 *            zu schliessende Streams
	 * @return <code>true</code> falls das Schliessen aller Streams erfolgreich
	 *         war
	 */
	public static boolean closeInputStream(InputStream[] in) {
		boolean result = true;
		for (int i = 0; i < in.length; i++)
			result &= closeInputStream(in[i]);
		return result;
	}

	/**
	 * Versucht einen Ausgabe-Stream zu schliessen. Zuvor wird ein
	 * {@link OutputStream#flush() flush()} auf den Stream getaetigt.<br>
	 * Bei Misserfolg werden keine Exceptions geworfen! SK: bei
	 * <code>null</code> wird true zurueckgegeben *
	 * 
	 * @param out
	 *            zu schliessender Stream
	 * @return <code>false</code> falls das Schliessen nicht erfolgreich war
	 */
	public static boolean closeOutputStream(OutputStream out) {
		if (out == null)
			return true;
		try {
			out.flush();
			out.close();
		} catch (Exception err) {
			return false;
		}
		return true;
	}

	/**
	 * Prueft, ob eine (Datei-)Zeile mit einem Kommentarkennzeichen beginnt und
	 * deshalb ignoriert werden muss. Kommentar-Kennzeichen sind <code>//</code>
	 * , <code>#</code> und <code>|</code>.
	 * 
	 * @param line
	 *            Eingabe-Zeile
	 */
	public static boolean isCommentLine(String line) {
		return line.startsWith("//") || line.startsWith("#")
				|| line.startsWith("|");
	}

	/**
	 * Schreibt einen Eingabe-Stream in eine Datei.
	 * 
	 * @param file
	 *            Dateipfad fuer die Zieldatei
	 * @exception java.io.IOException
	 *                falls das Lesen aus dem Stream oder das Schreiben in die
	 *                Ausgabedatei scheitert
	 */
	public static void writeStreamToFile(InputStream input, String file)
			throws IOException {
		OutputStream output = new FileOutputStream(file);
		while (input.available() > 0)
			output.write(input.read());
		output.flush();
		output.close();
	}

	/**
	 * Erstellt einen Filter fuer einen regulaeren Ausdruck.
	 * 
	 * @param filterStr
	 *            definiert den Filter
	 */
	public static FileFilter createRegExFileFilter(final String regEx,
			final FilterMode mode) {
		return new FileFilter() {
			public boolean accept(File file) {
				return fileFitsFilterMode(file, mode)
						&& file.getName().matches(regEx);
			}
		};
	}

	/**
	 * Erstellt einen Filter fuer einen Filter-String. Dieser kann als Wildcard
	 * das Zeichen "*" enthalten. Alle anderen Zeichen werden direkt als Filter
	 * verwendet.
	 * 
	 * @param filterStr
	 *            definiert den Filter
	 */
	public static FileFilter createSimpleFileFilter(String filterStr,
			final FilterMode mode) {
		// Gross/Klein-Schreibung ignorieren
		filterStr = filterStr.toLowerCase();

		// ##### Filter definiert den "inneren" Teil eines Dateinames #####
		if (filterStr.startsWith("*") && filterStr.endsWith("*")) {
			// Filter besteht nur aus "*" --> alle Dateien akzeptieren
			if (filterStr.length() <= 2)
				return new FileFilter() {
					public boolean accept(File file) {
						return fileFitsFilterMode(file, mode);
					}
				};
			// "inneren" Teil des Filters extrahieren
			final String condStr = filterStr.substring(1,
					filterStr.length() - 1);
			return new FileFilter() {
				public boolean accept(File file) {
					return fileFitsFilterMode(file, mode)
							&& file.getName().toLowerCase().contains(condStr);
				}
			};
		}

		// ##### Filter definiert den Anfang eines Dateinames #####
		if (filterStr.endsWith("*")) {
			// Anfang des Filters extrahieren
			final String condStr = filterStr.substring(0,
					filterStr.length() - 1);
			return new FileFilter() {
				public boolean accept(File file) {
					return fileFitsFilterMode(file, mode)
							&& file.getName().toLowerCase().startsWith(condStr);
				}
			};
		}
		// ##### Filter definiert das Ende eines Dateinames #####
		if (filterStr.startsWith("*")) {
			// Ende des Filters extrahieren
			final String condStr = filterStr.substring(1);
			return new FileFilter() {
				public boolean accept(File file) {
					return fileFitsFilterMode(file, mode)
							&& file.getName().toLowerCase().endsWith(condStr);
				}
			};
		}
		// ##### Filter definiert den genauen Dateinamen #####
		final String condStr = filterStr;
		return new FileFilter() {
			public boolean accept(File file) {
				return fileFitsFilterMode(file, mode)
						&& file.getName().equalsIgnoreCase(condStr);
			}
		};
	}

	/**
	 * Loescht Dateien oder Verzeichnisse in einem Verzeichnis. Das Verzeichnis
	 * selbst wird dabei NICHT geloescht, auch wenn es am Ende leer ist!
	 * 
	 * @param dir
	 *            Verzeichnis in dem Dateien/Verzeichnisse geloescht werden
	 * @param filter
	 *            bestimmt, welche Dateien/Verzeichnisse geloescht werden
	 * @param recursive
	 *            wenn {@code true} werden auch alle Dateien/Verzeichnisse in
	 *            Unterverzeichnissen geloescht, die dem Filter entsprechen
	 * @param showFiles
	 *            wenn {@code true} werden die geloeschten Dateien auf der
	 *            Console ausgegeben
	 * @return Anzahl geloeschter Dateien und Verzeichnisse
	 */
	public static int deleteFiles(File dir, FileFilter filter,
			boolean recursive, boolean showFiles) {
		if (!dir.isDirectory())
			return 0;

		int delCount = 0;
		// Dateien im aktuellen Verzeichnis loeschen
		File[] files = filter != null ? dir.listFiles(filter) : dir.listFiles();
		for (File file : files) {
			// Verzeichnis zunaechst leeren
			if (file.isDirectory())
				delCount += deleteFiles(file, null, true, showFiles);
			boolean success = file.delete();
			if (success)
				delCount++;
			if (showFiles) {
				StringBuffer message = new StringBuffer(success ? "Deleted: "
						: "Error: ");
				try {
					message.append(file.getCanonicalPath());
				} catch (Exception err) {
					message.append(file.getAbsolutePath());
				}
			}
		}
		// Wenn rekusiv geloescht werden soll, alle Verzeichnisse im Verzeichnis
		// durchsuchen
		if (recursive) {
			File[] directories = dir.listFiles(ALL_DIRS_FILTER);
			for (File directory : directories)
				delCount += deleteFiles(directory, filter, recursive, showFiles);
		}
		return delCount;
	}

	/**
	 * Die Umwandlung einer URL in ein File-Objekt ist leider nicht trivial.
	 * 
	 * @see http 
	 *      ://weblogs.java.net/blog/kohsuke/archive/2007/04/how_to_convert.html
	 * 
	 * @param url
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static File urlToFile(URL url) {
		// LOGGER.debug("\nconverting " + url.toString() + " to ");
		File f;
		try {
			f = new File(url.toURI());
		} catch (Exception e) {
			// TODO Here we probably need more cases...
			f = new File(url.getPath().replace("%20", " "));
		}
		// LOGGER.debug("        to " + f.toString());
		return f;
	}

	/**
	 * Fuehrt verschiedene Funktionen aus.
	 * <ul>
	 * <li>{@code DELDIR name}<br>
	 * Loescht ein Verzeichnis aus allen Unterverzeichnissen.</li>
	 * </ul>
	 * 
	 * @param arg
	 */
	public static void main(final String[] arg) {
		if (arg.length == 0) {
			System.err.println("Missing arguments.");
			return;
		}

		String func = arg[0].toUpperCase();
		if (func.equals("DELDIR")) {
			if (arg.length < 2) {
				System.err.println("Missing arguments for DELDIR.");
			}
			File currDir = new File(".");
			FileFilter filter = createSimpleFileFilter(arg[1],
					FilterMode.DIR_ONLY);
			int delCount = deleteFiles(currDir, filter, true, true);
		} else {
			LOGGER.error("Unknown function: " + func);
		}
	}

	/**
	 * The method File getTempDir() caches the system's temp direcotry here.
	 * 
	 * @see #getTempDir()
	 */
	private static File tempDir;

	/**
	 * The system property <code>java.io.tmpdir</code> is not consistent on
	 * Linux and Windows. On Windows&Solaris the path ends with a slash, on
	 * Linux&MacOS it doesn't. This method deals with this potential pit-fall.<br/>
	 * The folder is cached in a static {@link File} field.
	 * 
	 * @return A {@link File} object pointing to the system's temp directory.
	 *         This method does some extra checks and returns a valid
	 *         {@link File}
	 * 
	 * @see
	 *      <code>http://rationalpi.wordpress.com/2007/01/26/javaiotmpdir-inconsitency/</code>
	 */
	public static File getTempDir() {

		if (tempDir == null) {
			String tempPath = System.getProperty("java.io.tmpdir", System
					.getenv("TEMP"));

			if (tempPath == null || tempPath.isEmpty()
					|| !new File(tempPath).exists()) {
				LOGGER.warn("Temporary directory can't be found. Using /tmp");
				tempPath = "/tmp";
			}

			while (tempPath.endsWith(File.separator))
				tempPath = tempPath.substring(0, tempPath.length() - 1);

			tempDir = new File(tempPath);

			if (!tempDir.exists())
				throw new IllegalStateException(
						"Temp direcotory can't be determined.");
		}

		return tempDir;
	}

	/**
	 * Test whether it is possible to access the given URL.
	 */
	public static boolean urlExists(final URL url) {
		try {
			url.openStream().close();
		} catch (final IOException e) {
			return false;
		}
		return true;
	}

	/**
	 * Reads the contents of a File into one String. Watch the size!
	 * 
	 * @param file
	 *            a File to read
	 * @return as String the content of the file.
	 * 
	 * @throws java.io.IOException
	 */
	public static String readFileAsString(File file) throws java.io.IOException {
		byte[] buffer = new byte[(int) file.length()];
		FileInputStream f = new FileInputStream(file);
		try {
			f.read(buffer);
			return new String(buffer);
		} finally {
			f.close();
		}
	}

	/**
	 * @return a {@link String} with the content of the {@link URL}. Do not use
	 *         this on long files! Returns <code>null</code> if an erro occured.
	 */
	public static String readURLasString(URL url) {
		try {

			InputStream openStream = url.openStream();
			try {

				InputStreamReader inStream = new InputStreamReader(openStream);
				try {

					BufferedReader inReader = new BufferedReader(inStream);
					try {
						String content = inReader.readLine();
						return content;
					} finally {
						inReader.close();
					}

				} finally {
					inStream.close();
				}
			} finally {
				openStream.close();
			}
		} catch (IOException e) {
			LOGGER.error("getURLasString " + url, e);
		}
		return null;
	}

}
