/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.io;

import java.io.IOException;
import java.io.InputStream;

/**
 * <hr>
 * <small><i>Diese Klasse wurde urspruenglich 2002/2003 von <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *            fuer die ADAGIOS-Java-Library (Universitaet Bonn) erstellt, 2009 in das
 *            SCHMITZM-Projekt uebernommen und teilweise erweitert.</i></small>
 * <hr>
 * Diese Klasse stellt einen InputStream dar, der sukzessive Bytes
 * aus einem Buffer liefert. Dieser Buffer kann permanent wieder
 * (auf)gefuellt werden. Dazu muss er nicht erst komplett geleert
 * worden sein!<br>
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class InputBuffer extends InputStream {
  protected byte[]  buffer      = null;
  protected int     readPos     = 0;
  protected int     writePos    = 0;
  protected boolean bufferFull  = false;

  /**
   * Erzeugt einen neuen <code>InputBuffer</code>.
   * @param byteSize Groesse des Buffers in Byte
   */
  public InputBuffer(int byteSize) {
    buffer = new byte[byteSize];
  }

  public int getBufferSize() {
    return buffer.length;
  }

  /**
   * Erhoeht die Zeiger-Position im Buffer. Ist das Ende des Buffers
   * erreicht, springt der Zeiger wieder auf die Position 0.
   */
  protected int incPosition(int pos) {
    return incPosition(pos,1);
  }

  /**
   * Erhoeht die Zeiger-Position im Buffer. Ist das Ende des Buffers
   * erreicht, springt der Zeiger wieder nach vorne.
   * @param pos aktuelle Zeiger-Position
   * @param hop Anzahl an Positionen, die vor gerueckt werden sollen
   */
  protected int incPosition(int pos, int hop) {
    pos += hop;
    return pos % getBufferSize();
  }

  /**
   * Schreibt ein Byte in den Buffer.
   * @param b ein Byte
   * @exception java.nio.BufferOverflowException falls der Buffer bereits voll ist
   */
  public void write(byte b) {
    if ( availableSpace() == 0 )
      throw new java.nio.BufferOverflowException();
    buffer[writePos] = b;
    // Schreib-Zeiger erhoehen
    writePos = incPosition(writePos);
    // wenn Schreib-Zeiger nun auf der aktuellen Lese-Position
    // steht, ist der Buffer voll
    bufferFull = ( writePos == readPos );
  }

  /**
   * Schreibt eine Reihe von Bytes in den Buffer.
   * @param bytes Menge von zu schreibenen Bytes
   * @exception java.nio.BufferOverflowException falls der Buffer bereits voll ist
   */
  public void write(byte[] bytes) {
    for (int i=0; i<bytes.length; i++)
      write(bytes[i]);
  }

  /**
   * Schreibt eine Reihe von Bytes in den Buffer. Diese Methode blockiert,
   * bis der InputStream die geforderte Anzahl an Bytes liefern kann.
   * @param input  Stream aus dem die Bytes gelesen werden
   * @param len    Anzahl zu lesender Bytes
   * @exception java.nio.BufferOverflowException falls nicht genuegend freier
   *            Platz im Buffer ist
   * @exception java.io.IOException falls das Lesen aus dem Eingabe-Stream
   *            scheitert
   */
  public void write(InputStream input, int len) throws IOException {
    if ( availableSpace() < len )
      throw new java.nio.BufferOverflowException();

    // Ist zwischen dem Schreib-Cursor und dem Buffer-Ende noch genug
    // Platz, koennen alle Bytes AN EINEM STUECK gespeichert werden
    if ( writePos + len-1 < getBufferSize() ) {
      input.read(buffer,writePos,len);
      incPosition(writePos,len);
      bufferFull = ( writePos == readPos );
    }
    else {
      // Ansonsten wird erst bis zum Buffer-Ende geschrieben
      int len1 = getBufferSize()-writePos;
      write(input, len1);
      // Nachdem dann der Cursor wieder am Anfang des Buffers
      // steht, wird der Rest an einem Stueck geschrieben
      write(input, len - len1);
    }
  }

  /**
   * Liefert die Anzahl an Bytes, die (noch) in den Buffer
   * geschrieben werden koennen.
   */
  public int availableSpace() {
    return Math.abs( getBufferSize() - available() );
  }

  /**
   * Liesst das naechste Byte aus dem Buffer. Der zurueckgelieferte
   * Wert liegt zwischen 0 und 255.
   * @return -1 falls der Buffer leer ist
   */
  public int read() throws IOException {
    if ( available() == 0 ) return -1;
    int b = BinaryUtil.getUnsignedByte( buffer[readPos] );
    readPos = incPosition(readPos);
    // Buffer hat auf jeden Fall jetzt wieder einen freien Platz
    bufferFull = false;
    return b;
  }

  /**
   * Liefert die Anzahl an Bytes, die (noch) aus dem Buffer
   * gelesen werden koennen.
   */
  public int available() {
    // wenn der Buffer voll ist, ist writePos = readPos
    // --> available() wuerde jedoch 0 liefern
    if ( bufferFull ) return getBufferSize();
    return ( writePos - readPos + getBufferSize() ) % getBufferSize();
  }

  /**
   * Entfernt eine Anzahl von Bytes aus dem Buffer. Liegen nicht mehr
   * genug Bytes im Buffer, wird dies ignoriert und der Buffer einfach
   * komplett geleert.
   * @param n Anzahl zu entfernender Bytes (muss ein <code>int</code> sein)
   * @return Anzahl an Bytes, die aus dem Buffer entfernt wurden
   */
  public long skip(long n) {
    int skipped = Math.min((int)n,available());
    readPos = incPosition(readPos,skipped);
    return skipped;
  }

  /**
   * Leert den Buffer.
   * @return Anzahl an Bytes, die noch im Buffer waren
   */
  public long clear() {
    return skip( this.available() );
  }

  /**
   * Liefert den aktuellen Inhalt des Buffers als <code>bytes</code>-Array.
   */
  public byte[] getContent() {
    byte[] buf = new byte[available()];
    for (int i=0; i<buf.length; i++)
      buf[i] = buffer[ (readPos+i)%buffer.length ];
    return buf;
  }
}
