/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.io;

import java.io.IOException;
import java.io.InputStream;

/**
 * <hr>
 * <small><i>Diese Klasse wurde urspruenglich 2002/2003 von <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *            fuer die ADAGIOS-Java-Library (Universitaet Bonn) erstellt, 2009 in das
 *            SCHMITZM-Projekt uebernommen und teilweise erweitert.</i></small>
 * <hr>
 * Diese Klasse repraesentiert einen Eingabe-Stream, der einen anderen
 * Eingabe-Stream kapselt und auf eine bestimmte Groesse begrenzt.
 * Somit ist es moeglich, einer Methode "vorzugaukeln" der ihr uebergebene
 * InputStream habe eine bestimmte Laenge, obwohl er in Wirklichkeit viel
 * groesser ist.<br>
 * Der Leser fragt sich nach dem Sinn? :-)<br>
 * <b>Beispiel:</b><br>
 * In einem InputStream A stehen nacheinander die Binaer-Repraesentationen
 * geometrischer Objekte. Die ersten N Bytes gehoeren zum ersten Objekt.<br>
 * Uebergibt man nun den gesamten Stream an die <code>initFromBinary</code>-Methode,
 * wuerde diese u.U. nicht nur die ersten N Bytes lesen, sondern versuchen,
 * den kompletten Stream als Objekt zu interpretieren!!<br>
 * Uebergibt man der <code>initFromBinary</code>-Methode jedoch einen
 * <code>LimitedInputStream</code> der auf dem Stream A basiert und auf N
 * Bytes begrenzt ist, sind die restlichen Bytes von A (die hinter dem N-ten
 * Byte stehen) fuer die <code>initFromBinary</code>-Methode transparent.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class LimitedInputStream extends InputStream {
  private InputStream input;
  private int         limit;
  private int         readCount;

  /**
   * Erzeugt einen neuen <code>LimitedInputStream</code>.
   * @param input Basis-Daten
   * @param limit Maximale Anzahl an Bytes, die aus dem Basis-Stream gelesen
   *              werden, bevor der <code>LimitedInputStream</code> sein
   *              Ende erreicht
   */
  public LimitedInputStream(InputStream input, int limit) {
    this.input     = input;
    this.readCount = 0;
    setLimit(limit);
  }

  /**
   * Liefert das Limit fuer den Stream.
   */
  public int getLimit() {
    return limit;
  }

  /**
   * Setzt das Limit fuer den Stream.
   * @param limit neues Limit (in Bytes)
   * @exception java.lang.IllegalArgumentException falls <code>limit < 0</code>
   */
  public void setLimit(int limit) {
    if ( limit < 0 )
      throw new IllegalArgumentException("Limit must be greater or equal 0");
    this.limit = limit;
  }

  /**
   * Liefert die Anzahl an Bytes, die noch aus dem <code>LimitedInputStream</code>
   * gelesen werden koennen.
   * @exception java.io.IOException falls der <code>available()</code>-Aufruf
   *            fuer den Basis-Stream scheitert
   */
  public int available() throws IOException {
    return Math.min(input.available(),Math.max(limit-readCount,0));
  }

  /**
   * Liefert ein Byte des Basis-Streams, solange das Limit (oder das Ende
   * des Basis-Stream) noch nicht erreicht ist.
   * @exception java.io.IOException falls der <code>read()</code>-Aufruf
   *            fuer den Basis-Stream scheitert
   */
  public int read() throws IOException {
    if ( available() <= 0 ) return -1;
    readCount++;
    return input.read();
  }
}
