/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.io;

import java.io.IOException;
import java.io.InputStream;
import java.io.PushbackInputStream;
import java.io.StreamTokenizer;

/**
 * <hr>
 * <small><i>Diese Klasse wurde urspruenglich 2002/2003 von <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *            fuer die ADAGIOS-Java-Library (Universitaet Bonn) erstellt, 2009 in das
 *            SCHMITZM-Projekt uebernommen und teilweise erweitert.</i></small>
 * <hr>
 * Diese Klasse erweitert einen <code>DataInputStream</code> um die
 * Eigenschaft des <code>StreamTokenizer</code>
 * @see java.io.StreamTokenizer
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */

public class TokenInputStream extends PushbackInputStream {
  /**
   * Instanz von <code>DefaultTokenDefinition</code>.
   * @see adagios.types.TokenInputStream.DefaultTokenDefinition
   */
  public static final TokenDefinition DEFAULT_TOKENDEF = new DefaultTokenDefinition();


  /**
   * Signalisiert, dass kein Token mehr gelesen werden konnte, da
   * das Ende des Basisstreams erreicht wurde.
   * @see java.io.StreamTokenizer#TT_EOF
   */
  public static final int TT_EOF = StreamTokenizer.TT_EOF;

  /**
   * Signalisiert, dass das gelesene Token numerisch ist und
   * in <code>nval</code> gespeichert ist.
   * @see java.io.StreamTokenizer#TT_NUMBER
   */
  public static final int TT_NUMBER = StreamTokenizer.TT_NUMBER;

  /**
   * Signalisiert, dass das gelesene Token alphanumerisch ist und
   * in <code>sval</code> gespeichert ist.
   * @see java.io.StreamTokenizer#TT_WORD
   */
  public static final int TT_WORD = StreamTokenizer.TT_WORD;

  /**
   * Enthaelt den Typ des letzten gelesenene Tokens.
   * @see #TT_EOF
   * @see #TT_NUMBER
   * @see #TT_WORD
   * @see java.io.StreamTokenizer#ttype
   */
  public int    ttype = 0;

  /**
   * Handelt es sich bei dem letzten Token um eine Zeichenkette, wird diese
   * in <code>sval</code> gespeichtert. Ansonsten ist <code>sval = null</code>.
   * @see java.io.StreamTokenizer#sval
   */
  public String sval  = null;

  /**
   * Handelt es sich bei dem letzten Token um einen numerischen Wert, wird dieser
   * in <code>nval</code> gespeichtert. Ansonsten ist <code>nval = 0.0</code>.
   * @see java.io.StreamTokenizer#nval
   */
  public double nval  = 0.0;

  private TokenDefinition def = new DefaultTokenDefinition();

  /**
   * Erzeugt einen neuen TokenInputStream.
   * Der Pushback-Buffer wird mit 64Byte initialisiert. Als Tokendefinition wird
   * <code>DefaultTokenDefinition</code> verwendet.
   * @param is BasisInputStream
   * @see #DEFAULT_TOKENDEF
   */
  public TokenInputStream(InputStream is) {
    this(is,64);
  }

  /**
   * Erzeugt einen neuen TokenInputStream. Als Tokendefinition wird
   * <code>DefaultTokenDefinition</code> verwendet.
   * @param is   BasisInputStream
   * @param size Groesse des Pushback-Buffers
   * @exception java.lang.IllegalArgumentException falls <code>size <= 0</code>
   * @see #DEFAULT_TOKENDEF
   */
  public TokenInputStream(InputStream is, int size) {
    this(is,size,DEFAULT_TOKENDEF);
  }

  /**
   * Erzeugt einen neuen TokenInputStream.
   * @param is   BasisInputStream
   * @param size Groesse des Pushback-Buffers
   * @param def  bestimmt wie die Token aufgebaut sind
   * @exception java.lang.IllegalArgumentException falls <code>size <= 0</code>
   * @see adagios.types.TokenInputStream.TokenDefinition
   * @see #DEFAULT_TOKENDEF
   */
  public TokenInputStream(InputStream is, int size, TokenDefinition def) {
    super(is,size);
    this.def = def;
  }

  /**
   * Liefert einen neuen <code>TokenInputStream</code>.
   * Handelt es sich bei dem uebergebenen Basis-Stream bereits um einen
   * <code>TokenInputStream</code> wird dieser (gecastet) zurueckgeliefert,
   * ansonsten wird eine neue Instanz erzeugt.
   * @param is Basis-Stream
   */
  public static TokenInputStream fromInputStream(InputStream is) {
    if ( is instanceof TokenInputStream )
      return (TokenInputStream)is;
    return new TokenInputStream(is);
  }

  /**
   * Liefert die zugrunde liegende Token-Definition.
   */
  public TokenDefinition getTokenDefinition() {
    return def;
  }

  /**
   * Liesst ein Token vom Stream. Dabei werden zunaecht alle Delimiter-Zeichen
   * aus dem Stream gelesen und ignoriert.
   * Das erste Nicht-Delimiter-Zeichen bestimmt, ob das Token numerisch ist,
   * oder nicht. Bei numerischen Tokens wird das "Token-Lesen" beim ersten
   * nicht-numerischen Zeichen abgebrochen. Bei nicht-numerische Tokens erst
   * beim naechsten Delimiter.<br>
   * Der Wert des Tokens wird der Variablen <code>sval</code> oder
   * <code>nval</code> hinterlegt.<br>
   * Ein (nicht-numerisches) Ein-Zeichen-Token wird direkt als <code>int</code>
   * im Attribut <code>ttype</code> gespeichert und nicht in <code>sval</code>
   * (vgl. <code>StreamTokenizer</code>).<br>
   * <b>Beispiel (Delimiter = " \n\t"):</b><br>
   * <code>"   abcd456bla   999.99k k88 9,88   "</code> liefert die Token<br>
   * <ul>
   * <li><code>sval  = "abcd456bla"</code></li>
   * <li><code>nval  = 999.99</code></li>
   * <li><code>sval  = "k"</code></li>
   * <li><code>sval  = "k88"</code></li>
   * <li><code>nval  = 9.0</code></li>
   * <li><code>ttype = ','</code></li>
   * <li><code>nval  = 88.0</code></li>
   * </ul>
   * @return Typ des Tokens (Inhalt von <code>ttype</code>)
   * @see #ttype
   * @see #sval
   * @see #nval
   * @see StreamTokenizer#ttype
   * @see StreamTokenizer#nextToken()
   */
  public int readToken() throws IOException {
    char         c            = 999;
    StringBuffer token        = new StringBuffer();

    // Stream bereits zu Ende
    if ( available() <= 0 )
      return determineGlobalVariables("");

    // alle Delimiter-Chars ignorieren
    while ( available()>0 && def.isDelimiter( c = (char)read() ) );
    // wenn das letzte gelesene Zeichen ein Delimiter war ist der Stream zu Ende,
    // ohne dass ein Token gelesen werden kann
    if ( def.isDelimiter(c) )
      return determineGlobalVariables("");

    // letztes gelesenes Zeichen gehoert bereits zum Token
    token.append(c);

    boolean numericToken = false;
    // handelt es sich um ein Single-Character, wird nicht weitergelesen
    if ( !def.isSingleChar(c) ) {
      // ob das Token numerisch ist, bestimmt das erste Zeichen
      numericToken = def.isNumericStarter(c);
      // alle Zeichen dem Token hinzufuegen, bis ein Delimiter, ein SingleChar
      // oder bei numerischen Tokens ein nicht-numerisches Zeichen erreicht wird
      while ( available()>0 && !def.isDelimiter( c = (char)read() )
                            && !def.isSingleChar( c )
                            && (!numericToken || def.isNumeric(c)) ) {
        token.append(c);
      }
      // letztes gelesenes Zeichen ist ein Delimiter oder ein Zeichen, dass
      // nicht zum Token "passt"
      // --> dieses muss zurueck auf den Stream
      unread(c);
    }
//    System.err.println("-> "+token.toString());

    return determineGlobalVariables(token.toString());
  }

  /**
   * Legt das letzte Token zurueck auf den Stream. Wiederholte Aufrufe legen
   * das letzte Token mehrmals zurueck auf den Stream!!
   * Die Methode wird ignoriert, falls den letze Aufruf von
   * <code>nextToken(..)</code> das Streamende ueberschritten hat.
   * @return <code>false</code>, falls das letzte Token nicht auf den Stream
   *         zurueckgelegt werden konnte
   * @exception java.io.IOException falls das Zuruecklegen der Bytes auf den
   *            BasisStream scheitert
   */
  public boolean pushBack() throws IOException {
    byte[] b = null;
    switch (ttype) {
      case StreamTokenizer.TT_NUMBER: // numerisches Token
                                      b = String.valueOf(nval).getBytes();
                                      break;
      case StreamTokenizer.TT_WORD: // alphanumerisches Token
                                    b = sval.getBytes();
                                    break;
      case StreamTokenizer.TT_EOL: // Token kann nicht rekonstruiert werden
      case StreamTokenizer.TT_EOF: return false;
      default: // ein Zeichen
               b = new byte[] { (byte)ttype };
    }

    unread(b);
    return true;
  }

  /**
   * Bestimmt den Typ des Tokens (Char, numerisch, alphanumerisch) und
   * belegt entsprechend die globalen Variablen.
   * @return neuer Wert von <code>ttype</code>
   */
  private int determineGlobalVariables(String token) {
    // keine Nicht-Delimiter mehr auf dem Stream
    if (token.length() == 0)
      return setGlobalVariables(StreamTokenizer.TT_EOF, null, 0.0);

    // ein Ein-Zeichen-Token wird als int im Typ-Feld gespeichert
    if ( token.length() == 1 && def.isSingleChar(token.charAt(0)) )
      return setGlobalVariables(token.charAt(0),null,0.0);

    try {
      // versuchen, dass Token in eine Zahl umzuwandeln
      return setGlobalVariables(StreamTokenizer.TT_NUMBER,null,Double.parseDouble(token));
    }
    catch (Exception err) {
      // scheitert der Versuch, dass Token in eine Zahl umzuwandeln, handelt
      // es sich bei dem Token um eine Zeichenkette
      return setGlobalVariables(StreamTokenizer.TT_WORD,token,0.0);
    }
  }

  /**
   * Belegt die globalen Variablen mit den angegebenen Werten.
   * @return neuer Wert von <code>ttype</code>
   */
  private int setGlobalVariables(int ttype, String sval, double nval) {
    this.ttype = ttype;
    this.sval  = sval;
    this.nval  = nval;
    return ttype;
  }

  /**
   * Diese Klasse bestimmt, wie die Zeichen in einem <code>TokenInputStream</code>
   * zu Token zusammengefasst werde.
   */
  public static interface TokenDefinition {
    /**
     * Bestimmt, welche Zeichen "Delimiter" sind. Delimiter-Zeichen trennen
     * Token voneinander und werden ignoriert.
     */
    public boolean isDelimiter(char c);

    /**
     * Bestimmt, welche Zeichen numerisch sind.<br>
     * z.B. 0 bis 9 und '.'
     */
    public boolean isNumeric(char c);

    /**
     * Bestimmt, welche Zeichen ein numerisches Token einleiten. Im
     * Gegensatz zu <code>isNumeric(..)</code> sollten z.B. keine
     * Kommas oder Punkte als "NumericStarter" deklariert werden.<br>
     * z.B. 0 bis 9
     */
    public boolean isNumericStarter(char c);

    /**
     * Bestimmt, ob ein Zeichen ein "Einzelnes Zeichen" ist. SingleChars werden
     * werden nicht zu Token zusammengefasst, sondern immer als eigenes
     * Token behandelt. SingleChars trennen also auch Token voneinander.
     */
    public boolean isSingleChar(char c);
  }

  /**
   * Diese Klasse stellt eine Standard-Definition fuer einen
   * <code>TokenInputStream</code> dar.
   * @see adagios.types.TokenInputStream
   * @see adagios.types.TokenInputStream.TokenDefinition
   */
  public static class DefaultTokenDefinition implements TokenDefinition {
    /**
     * Delimiter sind SPACE ' ', TAB '\t' und CR '\n'.
     */
    public boolean isDelimiter(char c) {
      return c == ' '   // SPACE
          || c == '\t'  // TAB
          || c == '\n'; // CR
    }

    /**
     * Numerische Token muessen mit '0'..'9' oder '-' beginnen.
     */
    public boolean isNumericStarter(char c) {
      return c >= '0' && c <= '9' || c == '-';
    }

    /**
     * Numerische Zeichen sind NumericStarter oder Punkt.
     */
    public boolean isNumeric(char c) {
      return isNumericStarter(c) || c == '.';
    }

    /**
     * SingleChars sind Komma, Semikolon, Plus, Minus, Gleich, Slash,
     * BackSlash, Doppelpunkt, Pipe, runde Klammern, eckige Klammern und
     * geschweifte Klammern.
     */
    public boolean isSingleChar(char c) {
      return c == ',' || c == ';'
          || c == '+'  || c == '=' || c == '/'
          || c == ':' || c == '\\' || c == '|'
          || c == '(' || c == ')'  || c == '[' || c == ']'
          || c == '{' || c == '}';
    }
  }
}
