/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.io.dyntxt;

import java.io.IOException;
import java.io.OutputStream;

/**
 * Diese Klasse repraesentiert ein allgemeines dynamisches Text-Element.
 * Jedes Element kann durch eine alphanumerische ID identifiziert werden (muss
 * nicht eindeutig sein!!).<br>
 * Bis auf das Wurzelelement sollten allen Text-Elementen ein Vater-Element
 * zugeordnet sein, damit ein einfaches abarbeiten des Wurzel-Elements automatisch
 * die Abarbeitung aller untergeordneten Elemente nach sich zieht.
 * @author Martin Schmitz
 * @version 1.0
 */
public abstract class DynamicElement {
  private DynamicElement father = null;
  private String         id     = "";

  /**
   * Erzeugt ein neues dynamisches Root-Element.
   * @param id identifieziert den Block gegenueber dem <code>DynamicInputProvider</code>
   */
  public DynamicElement(String id) {
    this(id,null);
  }

  /**
   * Erzeugt ein neues dynamisches Text-Element.
   * @param id identifieziert den Block gegenueber dem <code>DynamicInputProvider</code>
   * @param father anderes dynamisches Element, welches das Element beinhaltet
   */
  public DynamicElement(String id, DynamicElement father) {
    this.father = father;
    this.id     = id;
  }

  /**
   * Checkt, ob das Element ein Vater-Element hat (<code>false</code>)
   * oder nicht.
   */
  public boolean isRoot() {
    return getFather()==null;
  }

  /**
   * Liefert das uebergeordnete dynamische Element.
   */
  public DynamicElement getFather() {
    return father;
  }

  /**
   * Liefert die naechste uebergeordnete Element.
   * @param cls Klasse der das Element entsprechen soll (muss eine
   * @return <code>null</code>, falls dem Element keine Element der angegebenen
   *         Klasse uebergeordnet ist
   */
  public DynamicElement getFather(Class cls) {
    if ( getFather() == null ) return null;
    if ( cls.isInstance(getFather()) ) return getFather();
    return getFather().getFather(cls);
  }


  /**
   * Liefert die ID des dynamischen Elements.
   */
  public String getID() {
    return id;
  }

  /**
   * Fuehrt das dynamische Element (und alle etwaigen darin enthaltenen Elemente)
   * aus.
   * @param inputProvider verarbeitet die Loops und Fields
   * @param output        hier wird die Ausgabe reingeschrieben
   * @exception java.io.IOException falls das Schreiben in den Stream fehlschlaegt
   */
  public abstract void performElement(DynamicInputProvider inputProvider, OutputStream output) throws IOException;
}
