/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.io.dyntxt;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * Diese Klasse stellt einen dynamischen Text dar. Dieser wird durch ein
 * Text-Geruest (Template) definiert, in dem durch spezielle Tags Schleifen
 * und Felder eingebaut sind:
 * <ul>
 * <li><b><code>LOOP(<i>LoopID</i>)</code></b><br>
 *     Definiert den Beginn einer Schleife.</li>
 * <li><b><code>ENDLOOP()</code></b><br>
 *     Definiert das Ende einer Schleife.</li>
 * <li><b><code>FIELD(<i>FieldID</i>)</code></b><br>
 *     Definiert ein Eingabe-Feld.</li>
 * </ul>
 * <br>
 * Um die Tags von normalem Text zu unterscheiden, werden sie zusaetzlich von
 * einer speziellen Zeichenfolge (TAG_INTRO) eingeleitet. Standardmaessig
 * wird hierfuer das <b>$-Zeichen</b> verwendet.<br>
 * <b>Beachte:</b><br>
 * Wird ein benutzerdefinierter Delimiter verwendet, so muss dieser evt. mit
 * zusaetzlichen Backslashs versehen werden, wenn er Spezialzeichen der
 * regulaeren Ausdruecke (Klasse <code>Pattern</code>) beinhaltet!!<br>
 * <br>
 * Das Template wird zunaechst komplett eingelesen und dabei die Schleifen und
 * Felder identifiziert.<br>
 * Beim Aufruf von <code>generateDynamicText(..)</code> werden dann die Schleifen
 * und Felder ausgewertet und der Ausgabetext erzeugt.
 * Wann eine Schleife beendet wird, und welche Werte jeweils fuer die Felder
 * eingesetzt werden, entscheidet ein <code>DynamicInputProvider</code>.<br>
 * Die Ausgabe wird komplett in einem Buffer zwischengespeichert. Ueber einen
 * <code>InputStream</code> kann dann beliebig darauf zugegriffen werden.
 * @see #generateDynamicText(OutputStream,DynamicInputProvider)
 * @see schmitzm.io.dyntxt.DynamicInputProvider
 * @see java.util.regex.Pattern
 * @author Martin Schmitz
 * @version 1.0
 */

public class DynamicTextGenerator {
  /**
   * Leitet die Tags ein.
   */
  public static final String TAG_INTRO          = "$";
  /**
   * <code>Pattern</code>-konformes <code>TAG_INTRO</code>.
   * @see #TAG_INTRO
   * @see java.util.regex.Pattern
   */
  public static final String TAG_INTRO_REGEXPR  = "\\$";
  /**
   * Leitet im Template eine Schleife ein.
   */
  public static final String TAG_LOOPSTART  = "LOOP";
  /**
   * Beendet im Template eine Schleife.
   */
  public static final String TAG_LOOPEND    = "ENDLOOP";
  /**
   * Steht im Template fuer ein dynamisches Feld.
   */
  public static final String TAG_FIELD      = "FIELD";

  private DynamicBlock initialBlock     = null;
  private String       tagIntro        = "";
  private String       tagIntroRegExpr = "";

  /**
   * Erzeugt einen neuen dynamischen Text.
   * @param template Vorlage-Text mit Tags (Felder, Schleife)
   * @exception java.io.IOException falls das Lesen aus dem Eingabe-Stream
   *            fehlschlaegt
   */
  public DynamicTextGenerator(InputStream template) throws IOException {
    this(template,TAG_INTRO,TAG_INTRO_REGEXPR);
  }

  /**
   * Erzeugt einen neuen dynamischen Text.
   * @param template Vorlage-Text mit Tags (Felder, Schleife)
   * @param tagIntro String, der die Tags einleitet
   * @param tagIntroRegExpr String, der die Tags einleitet (so erweitert, dass er
   *        mit den Regulaeren Ausdruecken kompatibel ist)
   * @exception java.io.IOException falls das Lesen aus dem Eingabe-Stream
   *            fehlschlaegt
   * @see java.util.regex.Pattern
   */
  public DynamicTextGenerator(InputStream template, String tagIntro, String tagIntroRegExpr) throws IOException {
    this.tagIntro        = tagIntro;
    this.tagIntroRegExpr = tagIntroRegExpr;
    this.initialBlock    = createBlockFromTemplate(template);
  }

  /**
   * Generiert einen Text, in dem die dynamischen Felder der Vorlage durch
   * den <code>DynamicInputProvider</code> aufgeloest werden.
   * @param output Ausgabe-Stream fuer den dynamischen Text
   * @param inputProvider loest die dynamischen Elemente auf
   * @exception java.io.IOException falls das Schreiben in den Ausgabe-Stream
   *            fehlschlaegt
   */
  public void generateDynamicText(OutputStream output, DynamicInputProvider inputProvider) throws IOException {
    initialBlock.performElement(inputProvider,output);
  }

  /**
   * Liefert die Zeichenfolge, die die Tags im Template einleitet und sie beim
   * Parsen eindeutig vom Text unterscheidet.
   */
  public String getTagIntro() {
    return this.tagIntro;
  }

  /**
   * Liefert das um Sonderzeichen erweitere TagIntro. Diese (insbesondere
   * Backslashes) werden notwendig, wenn das TagIntro Zeichen enthaelt, die
   * in regulaeren Ausdruecken durch die Klasse <code>Pattern</code> speziell
   * interpretiert werden.
   * @see #getTagIntro()
   * @see java.util.regex.Pattern
   */
  public String getTagIntroForRegExpr() {
    return this.tagIntroRegExpr;
  }

  /**
   * Parst die Vorlage und erzeugt daraus einen Wurzel-Block.
   */
  private DynamicBlock createBlockFromTemplate(InputStream template) throws IOException {
    // Initialen Block erzeugen
    DynamicBlock rootBlock = new DynamicBlock("ROOT");
    // regulaeren Ausdruck fuer Tags erzeugen
    String loopStartPattern = getTagIntroForRegExpr()+TAG_LOOPSTART+"\\(\\w*\\)";
    String loopEndPattern   = getTagIntroForRegExpr()+TAG_LOOPEND+"\\(\\)";
    String fieldPattern     = getTagIntroForRegExpr()+TAG_FIELD+"\\(\\w*\\)";
    Pattern pattern = Pattern.compile("("+loopStartPattern+")|("+loopEndPattern+")|("+fieldPattern+")");

    DynamicBlock actualBlock = rootBlock;

    // Eingabe zeilenweise auslesen
    BufferedReader input = new BufferedReader( new InputStreamReader(template) );
    String line   = "";
    int    lineNo = 0;
    for(lineNo=1; input.ready() && (line=input.readLine())!=null; lineNo++ ) {
      int lastEnd = 0; // Ende des letzten Pattern-Match der Zeile (=Anfang)
      Matcher m = pattern.matcher(line);
      while( m.find() ) {
        // alles zwischen letzem Tag und neuem Tag wird zu Text-Feld
        String text = line.substring(lastEnd,m.start());
        if ( text.length() > 0 )
          actualBlock.add( new StaticText(actualBlock,text) );

        // neuer Tag
        String tag = m.group();
        // Ende des neuen Tags bestimmen
        lastEnd = m.end();
        // ID steht zwischen den beiden Klammern
        String id  = tag.substring(tag.indexOf('(')+1,tag.indexOf(')'));
//        System.err.println(m.start()+"\t"+m.end()+"\t"+tag+"\t"+id);
        // welcher Tag ist es?
        if ( tag.startsWith(getTagIntro()+TAG_LOOPSTART) ) {
          // Start Loop
          DynamicLoop loop = new DynamicLoop(id,actualBlock);
          actualBlock.add(loop);
          actualBlock = loop;
        } else if ( tag.startsWith(getTagIntro()+TAG_LOOPEND) ) {
          // End Loop
          actualBlock = (DynamicBlock)actualBlock.getFather();
          // ein ENDLOOP zuviel
          if (actualBlock==null)
            throw new IOException("Line "+lineNo+": ENDLOOP not expected");
        } else if ( tag.startsWith(getTagIntro()+TAG_FIELD) ) {
          // Field
          actualBlock.add( new DynamicField(id,actualBlock) );
        }
      }
      // alles zwischen letzem Tag und Zeilenende wird zu Text-Feld
      String text = line.substring(lastEnd)+"\n";
      if ( !actualBlock.isEmpty() || !text.equals("\n") )
        actualBlock.add( new StaticText(actualBlock,text) );
    }

    // ENDLOOP fehlt
    if (actualBlock!=rootBlock)
      throw new IOException("Line "+lineNo+++": ENDLOOP missing");

    return rootBlock;
  }
}
