/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.jfree.chart;

import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.util.HashSet;
import java.util.Set;

import org.jfree.chart.ChartPanel;
import org.jfree.chart.entity.CategoryItemEntity;
import org.jfree.chart.entity.ChartEntity;
import org.jfree.chart.entity.EntityCollection;
import org.jfree.chart.entity.XYItemEntity;

import schmitzm.geotools.gui.MouseSelectionTracker;

/**
 * This tracker listens to mouse window selections and single mouse
 * clicks on a {@link ChartPanel}. It determines all selected 
 * {@link XYItemEntity XYItemEntities} and
 * {@link CategoryItemEntity CategoryItemEntities}.
 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a>
 */
public abstract class ChartMouseSelectionTracker extends MouseSelectionTracker {
  /** Holds the panel connected to this listener. */
  protected ChartPanel chartPanel = null;
  
  /**
   * Creates a new listener.
   * @param panel the chart panel connected to the listener (used to
   *              determine the {@link EntityCollection})
   * @param mode  specifies the selection events the tracker reacts on
   */
  public ChartMouseSelectionTracker(ChartPanel panel, SelectionMode mode) {
    super(mode);
    this.chartPanel    = panel;
  }

  /**
   * Creates a new listener which reacts on single clicks as well
   * as on window selection events.
   * @param panel the chart panel connected to the listener (used to
   *              determine the {@link EntityCollection})
   */
  public ChartMouseSelectionTracker(ChartPanel panel) {
    this(panel, SelectionMode.ALL);
  }
  
  /**
   * Reacts on the window selection by determining all {@link XYItemEntity XYItemEntities}
   * intersected by the selection window.
   */
  @Override
  protected void selectionPerformed(int ox, int oy, int px, int py) {
    // Create rectangle for selection window
    Rectangle2D selectedArea = new Rectangle2D.Double(
        Math.min(ox,px),
        Math.min(oy,py),
        Math.abs(ox-px),
        Math.abs(oy-py) );
    // If not a window is selected, but a point, create also
    // an point object for the click location  
    Point2D selectedPoint = null;
    if ( ox == px && oy == py )
      selectedPoint = new Point2D.Double(ox,oy);
    
    // ignore window selection if ONLY_SINGLE_CLICK is set
    if ( selectedPoint == null &&
         SelectionMode.ONLY_SINGLE_CLICK.equals(getSelectionMode()) )
       return;

    // Check every chart data points if the selection area
    // contains it (or it contains the selected point)
    HashSet<ChartEntity> selectedEntities = new HashSet<ChartEntity>();
    for( Object entityObj : chartPanel.getChartRenderingInfo().getEntityCollection().getEntities() ) {
      ChartEntity entity = (ChartEntity)entityObj;
      // ignore all non-data entities
      if ( !(entity instanceof XYItemEntity) &&
           !(entity instanceof CategoryItemEntity) )
        continue;
      
      // on point click, check whether point is inside the
      // entity area, otherwise check intersection
      if ( selectedPoint != null && entity.getArea().contains(selectedPoint) || 
           selectedPoint == null && entity.getArea().intersects(selectedArea) )
        selectedEntities.add( entity );
    }
    selectionPerformed(selectedArea, selectedEntities);
  }
  
  /**
   * Called after {@link #selectionPerformed(int, int, int, int)} to inform
   * the listeners about the selected items.
   * @param rect             the selection range
   * @param selectedEntities the selected {@link XYItemEntity XYItemEntities}
   */
  protected abstract void selectionPerformed(Rectangle2D rect, Set<ChartEntity> selectedEntities);
}
