/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.jfree.chart;

import java.awt.event.MouseEvent;
import java.awt.geom.Rectangle2D;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import javax.swing.ImageIcon;

import org.apache.log4j.Logger;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.entity.CategoryItemEntity;
import org.jfree.chart.entity.ChartEntity;
import org.jfree.chart.entity.XYItemEntity;
import org.jfree.chart.event.ChartChangeEvent;
import org.jfree.chart.event.ChartChangeEventType;
import org.jfree.chart.plot.CategoryPlot;
import org.jfree.chart.plot.DatasetRenderingOrder;
import org.jfree.chart.plot.Plot;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.AbstractRenderer;
import org.jfree.chart.renderer.category.CategoryItemRenderer;
import org.jfree.chart.renderer.xy.XYItemRenderer;
import org.jfree.data.category.CategoryDataset;
import org.jfree.data.general.Dataset;
import org.jfree.data.xy.XYDataset;

import schmitzm.jfree.JFreeChartUtil;
import schmitzm.jfree.chart.renderer.PointSelectionCategoryRenderer;
import schmitzm.jfree.chart.renderer.PointSelectionXYRenderer;
import schmitzm.jfree.chart.renderer.SelectionRenderer;
import schmitzm.jfree.chart.selection.CategoryDatasetSelectionModel;
import schmitzm.jfree.chart.selection.DatasetSelectionModel;
import schmitzm.jfree.chart.selection.SeriesDatasetSelectionModel;
import schmitzm.jfree.feature.FeatureSeriesDatasetSelectionModel;
import schmitzm.lang.LangUtil;
import schmitzm.swing.SwingUtil;

/**
 * This class extends the {@link org.jfree.chart.ChartPanel} with some new
 * functionalities.
 * <ul>
 * <li>select chart points and highlight them</li>
 * <li>setting window selection mode to ZOOM or SELECT</li>
 * </ul>
 * 
 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a>
 * @version 1.0
 */
public class SelectableChartPanel extends org.jfree.chart.ChartPanel {
	/** Used for log and debug messages. */
	private static Logger LOGGER = LangUtil.createLogger(SelectableChartPanel.class);
	
	public static final ImageIcon ICON_ZOOM_TO_SELECTED = SwingUtil.createImageIconFromResourcePath(
	    JFreeChartUtil.class, "chart/resource/icons/chartActionZoomToSelected.png", null ); 

	/**
	 * Types of window selection on chart panel.
	 */
	public static enum WindowSelectionMode {
		/**
		 * Zoom on window selection (normal behavior of
		 * {@link org.jfree.chart.ChartPanel})
		 */
		ZOOM_IN_CHART,
		/**
		 * Swap the "selection-state" of data objects in the chart on window
		 * selection.
		 */
		SELECT_SWAP,
		/**
		 * Set the set of selected objects to the inclusions of the window
		 * selection.
		 */
		SELECT_SET,
		/** Add objects to the selection on window selection. */
		SELECT_ADD,
		/** Remove objects from the selection on window selection. */
		SELECT_REMOVE
	}

	/**
	 * Holds the behavior on window selection (Default:
	 * {@link WindowSelectionMode#ZOOM_IN_CHART}).
	 */
	protected WindowSelectionMode windowSelectionMode = WindowSelectionMode.ZOOM_IN_CHART;

	/** Listens to mouse window selection on the chart. */
	protected ChartMouseSelectionTracker selectionTracker = null;

	/**
	 * Creates a new chart panel.
	 */
	public SelectableChartPanel(JFreeChart chart) {
		super(chart);
		init();
	}

	/**
	 * Creates a new chart panel.
	 */
	public SelectableChartPanel(JFreeChart chart, boolean useBuffer) {
		super(chart, useBuffer);
		init();
	}

	/**
	 * Creates a new chart panel.
	 */
	public SelectableChartPanel(JFreeChart chart, boolean properties,
			boolean save, boolean print, boolean zoom, boolean tooltips) {
		super(chart, properties, save, print, zoom, tooltips);
		init();
	}

	/**
	 * Creates a new chart panel.
	 */
	public SelectableChartPanel(JFreeChart chart, int width, int height,
			int minimumDrawWidth, int minimumDrawHeight, int maximumDrawWidth,
			int maximumDrawHeight, boolean useBuffer, boolean properties,
			boolean save, boolean print, boolean zoom, boolean tooltips) {
		super(chart, width, height, minimumDrawWidth, minimumDrawHeight,
				maximumDrawWidth, maximumDrawHeight, useBuffer, properties,
				save, print, zoom, tooltips);
		init();
	}

	/**
	 * Called by every constructor to initialize the extensions of this panel.
	 */
	protected void init() {

		selectionTracker = new ChartMouseSelectionTracker(this) {

		    private <E,F> void performSelection(DatasetSelectionModel<?,E,F> selModel, E seriesKey, F item) {
              switch (windowSelectionMode) {
                case SELECT_SET: // Is cleared once earlier
                case SELECT_ADD:
                  if (!selModel.isItemSelected(seriesKey, item))
                      selModel.setItemSelected(seriesKey, item, true);
                  break;
                case SELECT_SWAP:
                  selModel.changeItemSelection(seriesKey, item);
                  break;
              case SELECT_REMOVE:
                  if (selModel.isItemSelected(seriesKey, item))
                    selModel.setItemSelected(seriesKey, item, false);
                  break;
              }

		      
		    }
			/**
			 * This method is only called, if the #WindowSelectionMode is not in
			 * Zoom-Mode. Otherwise this listener is disabled.
			 */
			@SuppressWarnings("fallthrough")
			public void selectionPerformed(Rectangle2D selectedArea,
					Set<ChartEntity> selectedEntities) {
				LOGGER.debug(selectedEntities.size() + " objects selected");

                Plot plot = getChart().getPlot();

				// remember all changed selection models to reset
				// "valueAdjusting" property at the end
				Set<DatasetSelectionModel<?,?,?>> changedModels = new HashSet<DatasetSelectionModel<?,?,?>>();
				// if selection mode is SELECT_SET all models change
				// (also these which are not in the selectedEntities
				// anymore!), because they must be cleared 
				if ( windowSelectionMode == WindowSelectionMode.SELECT_SET ) {
				  for ( DatasetSelectionModel<?,?,?> selModel : JFreeChartUtil.getSelectionModels(plot).values() ) {
				    changedModels.add(selModel);
                    selModel.setValueIsAdjusting(true);
                    selModel.clearSelection();
				  }
				}
				
				for (Iterator<ChartEntity> i = selectedEntities.iterator(); i.hasNext();) {
					ChartEntity e = i.next();
					Dataset dataset = null;
                    if ( e instanceof CategoryItemEntity )
                      dataset = ((CategoryItemEntity)e).getDataset();
                    if ( e instanceof XYItemEntity )
                      dataset = ((XYItemEntity)e).getDataset();
					DatasetSelectionModel<?,?,?> selModel = JFreeChartUtil.getSelectionModelForDataset(
					                                          plot, dataset);
					// set "valueAdjusting" property for selection model and
					// remember the
					// model to reset the property at the end
					if (selModel != null) {
						selModel.setValueIsAdjusting(true);
						changedModels.add(selModel);
					}
					// perform the selection change
					if (selModel instanceof SeriesDatasetSelectionModel) {
					    XYItemEntity  xyEntity  = (XYItemEntity)e;
						int           series    = xyEntity.getSeriesIndex();
						Comparable<?> seriesKey = xyEntity.getDataset().getSeriesKey(series);
						int           item      = xyEntity.getItem();
						performSelection(
						    (SeriesDatasetSelectionModel)selModel,
						    seriesKey,
						    item
						);
					} else if ( selModel instanceof CategoryDatasetSelectionModel ) {
                        CategoryItemEntity  catEntity = (CategoryItemEntity)e;
                        Comparable<?>       seriesKey = catEntity.getRowKey();
                        Comparable<?>       itemKey   = catEntity.getColumnKey();
                        performSelection(
                            (CategoryDatasetSelectionModel)selModel,
                            seriesKey,
                            itemKey
                        );
					} else {
						if (selModel == null)
							LOGGER.warn("Selection not supported for dataset: " + LangUtil.getSimpleClassName(dataset));
						else
							LOGGER.error("Selection not yet implemented for selection model: "+LangUtil.getSimpleClassName(selModel));
					}
				}
				// refresh the visualization
				refresh();
				// reset the "valueAdjusting" property to inform listeners
				// about the selection change
				for (DatasetSelectionModel<?,?,?> selModel : changedModels)
					selModel.setValueIsAdjusting(false);
			}

		};
		addMouseListener(selectionTracker);
		setWindowSelectionMode(WindowSelectionMode.ZOOM_IN_CHART);
	}

	/**
	 * Repaints the chart.
	 */
	public void refresh() {
		setRefreshBuffer(true);
		repaint();
	}

    /**
     * Sets the behavior of the chart panel on window selections via mouse.
     * @param mode new selection mode
     */
    public void setWindowSelectionMode(WindowSelectionMode mode) {
        this.windowSelectionMode = mode;
        updateStatesForWindowSelectionMode();
    }

    /**
     * Returns the behavior of the chart panel on window selections via mouse.
     * @param selectionMode new selection mode
     */
    public WindowSelectionMode getWindowSelectionMode() {
        return this.windowSelectionMode;
    }

    /**
	 * After setting a new chart the zoom relevant properties are reset
	 * according to the new chart. After calling the super-method, this methods
	 * resets the properties according to the current
	 * {@link #windowSelectionMode} to keep its functionality remaining.
	 * 
	 * @see #updateStatesForWindowSelectionMode()
	 */
	public void setChart(JFreeChart chart) {
		super.setChart(chart);
		// zoom relevant properties were reset according to
		// chart properties!
		// But we do not want them changed, so reset the states.
		updateStatesForWindowSelectionMode();

		// create renderers to highlight the selected items
		updateSelectionRenderers();
	}

	/**
	 * Configures the zoomable properties and {@link #selectionTracker}
	 * according to the current {@link #windowSelectionMode}.
	 */
	protected void updateStatesForWindowSelectionMode() {
		// enable/disable the zoom functionality
		setMouseZoomable(windowSelectionMode == WindowSelectionMode.ZOOM_IN_CHART,
				getFillZoomRectangle());
		// enable/disable the selection functionality
		if (selectionTracker != null)
		  selectionTracker.setEnabled(windowSelectionMode != WindowSelectionMode.ZOOM_IN_CHART);
	}
	
	/**
	 * This methods creates a {@link SelectionRenderer} for each {@link Dataset}
	 * in the plot to realize the highlight of the selected items.
	 */
	protected void updateSelectionRenderers() {
	  if ( getChart() == null || getChart().getPlot() == null )
	    return;
	  Plot plot = getChart().getPlot();
	  
	  if ( plot instanceof XYPlot ) {
        XYPlot xyPlot = (XYPlot)plot;
        // remove all existing SelectionRenderer
        for (int i=0; i<xyPlot.getRendererCount(); i++)
          if ( xyPlot.getRenderer(i) instanceof SelectionRenderer ) {
            xyPlot.setDataset(i, null);
            xyPlot.setRenderer(i, null, false);
          }
        // add an SelectionRenderer for every "normal" renderer
        int baseDatasetCount  = xyPlot.getDatasetCount();
        for (int i=0; i<baseDatasetCount; i++) {
          XYDataset      dataset         = xyPlot.getDataset(i);
          XYItemRenderer datasetRenderer = (XYItemRenderer)JFreeChartUtil.getRendererForDataset(xyPlot, i);
          // Create SelectionRenderer
          XYItemRenderer selRenderer     = (XYItemRenderer)createSelectionRenderer(dataset);
          int            selRendererIdx  = baseDatasetCount+i;
          // Add the new SelectionRenderer with the original dataset
          xyPlot.setDataset(selRendererIdx, dataset);
          xyPlot.setRenderer(selRendererIdx, selRenderer, false);
          applyRendererPropertiesToSelectionRenderer(dataset,(AbstractRenderer)datasetRenderer,(AbstractRenderer)selRenderer);
        }
        // revert the rendering order (0 to X) so that the SelectionRenderers
        // are rendered last and the selected items are shown in foreground
        xyPlot.setDatasetRenderingOrder(DatasetRenderingOrder.FORWARD);
	  }
	  
      if ( plot instanceof CategoryPlot ) {
        CategoryPlot catPlot = (CategoryPlot)plot;
        // remove all existing SelectionRenderer
        for (int i=0; i<catPlot.getRendererCount(); i++)
          if ( catPlot.getRenderer(i) instanceof SelectionRenderer ) {
            catPlot.setDataset(i, null);
            catPlot.setRenderer(i, null, false);
          }
        // add an SelectionRenderer for every "normal" renderer
        int baseDatasetCount  = catPlot.getDatasetCount();
        for (int i=0; i<baseDatasetCount; i++) {
          CategoryDataset      dataset         = catPlot.getDataset(i);
          CategoryItemRenderer datasetRenderer = (CategoryItemRenderer)JFreeChartUtil.getRendererForDataset(catPlot, i);
          // Create SelectionRenderer
          CategoryItemRenderer selRenderer     = (CategoryItemRenderer)createSelectionRenderer(dataset);
          int                  selRendererIdx  = baseDatasetCount+i;
          // Add the new SelectionRenderer with the original dataset
          catPlot.setDataset(selRendererIdx, dataset);
          catPlot.setRenderer(selRendererIdx, selRenderer, false);
          applyRendererPropertiesToSelectionRenderer(dataset,(AbstractRenderer)datasetRenderer,(AbstractRenderer)selRenderer);
        }
        // revert the rendering order (0 to X) so that the SelectionRenderers
        // are rendered last and the selected items are shown in foreground
        catPlot.setDatasetRenderingOrder(DatasetRenderingOrder.FORWARD);
      }
      
      // fire event about the renderer changes
      JFreeChartUtil.fireChangeEvent(plot);
	}

    /**
     * Creates an adequate selection renderer for a dataset. This implementation
     * returns a renderer which draws only the data points. Sub-classes can
     * override this method to create different renderers or renderers with
     * different selection models (e.g. {@link FeatureSeriesDatasetSelectionModel})
     * for special purposes.
     * @param dataset a dataset
     * @return besides {@link SelectionRenderer} the returned object must be
     *         an {@link AbstractRenderer}  
     */
    protected SelectionRenderer<?> createSelectionRenderer(Dataset dataset) {
      if ( dataset instanceof XYDataset )
        return new PointSelectionXYRenderer();
      if ( dataset instanceof CategoryDataset )
        return new PointSelectionCategoryRenderer();

      LOGGER.warn("No selection renderer provided for dataset: "+LangUtil.getSimpleClassName(dataset));
      return null;
    }
    
    /**
     * Applies some rendering properties to the (new) selection renderer.
     * This methods simply sets the series shape of the selection renderer
     * to the shape of the origin renderer, so that the highlighted data
     * items are drawn the same way as the origin data items.
     * @param dataset        the dataset for the selection renderer
     * @param originRenderer the origin renderer for the dataset (should not
     *                       be changed!)
     * @param selRenderer    the selection renderer to apply the properties to
     */
    protected <E> void  applyRendererPropertiesToSelectionRenderer(Dataset dataset, AbstractRenderer originRenderer, AbstractRenderer selRenderer) {
      if ( selRenderer == null || originRenderer == null )
        return;
      
      selRenderer.setBaseShape( originRenderer.lookupSeriesShape(0) );
      selRenderer.setBaseStroke( originRenderer.lookupSeriesStroke(0) );

      if ( dataset instanceof XYDataset ) {
        ((XYItemRenderer)selRenderer).setBaseToolTipGenerator( ((XYItemRenderer)originRenderer).getBaseToolTipGenerator() );
        for (int i=0; i<((XYDataset)dataset).getSeriesCount(); i++) {
          selRenderer.setSeriesShape(i, originRenderer.lookupSeriesShape(i));
          selRenderer.setSeriesStroke(i, originRenderer.lookupSeriesStroke(i));
        }
      }
      if ( dataset instanceof CategoryDataset ) {
        ((CategoryItemRenderer)selRenderer).setBaseToolTipGenerator( ((CategoryItemRenderer)originRenderer).getBaseToolTipGenerator() );
//        for (int i=0; i<((CategoryDataset)dataset).getSeriesCount(); i++)
//          if ( ((CategoryItemRenderer)originRenderer).getSeriesShape(i) != null )
//            ((CategoryItemRenderer)selRenderer).setSeriesShape(i, ((CategoryItemRenderer)originRenderer).getSeriesShape(i));
      }
    }
    /**
     * If a new dataset is added this method updates the selection
     * renderers.
     */
    @Override
    public void chartChanged(ChartChangeEvent e) {
      super.chartChanged(e);
      if ( ChartChangeEventType.NEW_DATASET.equals( e.getType() ) )
        updateSelectionRenderers();
    }

	/**
	 * Ignores the event if the window selection mode is not
	 * {@link WindowSelectionMode#ZOOM_IN_CHART}, to avoid conflicts between the two
	 * listeners.
	 */
	@Override
	public void mouseDragged(MouseEvent e) {
		if (getWindowSelectionMode() != WindowSelectionMode.ZOOM_IN_CHART)
			return;
		super.mouseDragged(e);
	}

	/**
	 * Ignores the event if the window selection mode is not
	 * {@link WindowSelectionMode#ZOOM_IN_CHART}, to avoid conflicts between the two
	 * listeners.
	 */
	@Override
	public void mouseMoved(MouseEvent e) {
		if (getWindowSelectionMode() != WindowSelectionMode.ZOOM_IN_CHART)
			return;
		super.mouseMoved(e);
	}
	
}
