/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.jfree.chart.renderer;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Paint;
import java.awt.geom.Rectangle2D;

import org.apache.log4j.Logger;
import org.jfree.chart.axis.CategoryAxis;
import org.jfree.chart.axis.ValueAxis;
import org.jfree.chart.plot.CategoryPlot;
import org.jfree.chart.plot.Plot;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.category.CategoryItemRendererState;
import org.jfree.chart.renderer.category.LineAndShapeRenderer;
import org.jfree.data.category.CategoryDataset;
import org.jfree.data.general.Dataset;
import org.jfree.data.xy.XYDataset;

import schmitzm.jfree.JFreeChartUtil;
import schmitzm.jfree.chart.selection.CategoryDatasetSelectionModel;
import schmitzm.jfree.chart.selection.SeriesDatasetSelectionModel;
import schmitzm.lang.LangUtil;

/**
 * Extends the {@link LineAndShapeRenderer} with the
 * {@link SelectionRenderer} functionality. Additionally redefines
 * the {@link #drawItem(Graphics2D, CategoryItemRendererState, Rectangle2D, CategoryPlot, CategoryAxis, ValueAxis, CategoryDataset, int, int, int)}
 * method to realize the highlight rendering of the selected
 * items. <b>Only the selected data points are rendered. No lines
 * ore areas!</b><br>
 * <b>Note:</b><br>
 * An adequate {@link SeriesDatasetSelectionModel} is automatically
 * created when a new {@link Plot} is set. Manually calling {@link #setSelectionModel(SeriesDatasetSelectionModel)}
 * is usually not necessary. Overwrite {@link #createAndSetSelectionModel()} to
 * avoid this behavior.<br>
 * If no selection model is set, the behavior of this renderer is exactly
 * like the super class.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 */
public class PointSelectionCategoryRenderer extends LineAndShapeRenderer implements SelectionRenderer<CategoryDatasetSelectionModel> {
  /** Used for log and debug messages. */
  protected Logger LOGGER = LangUtil.createLogger(this);

  /** Holds the selection model, the renderer uses to determine whether a
   *  data item is selected or not. */
  protected CategoryDatasetSelectionModel selectionModel = null;

  /** Holds the color for selected items (used for all series). */
  protected Paint selectionPaint = null;

  /** Holds the outline paint for selected items (used for all series). */
  protected Paint selectionOutlinePaint;
  
//  /** Holds the selection {@link Stroke} (used for all series). */
//  protected Stroke selectionStroke = null;

  /**
   * Creates a new renderer.
   */
  public PointSelectionCategoryRenderer() {
    super(false,true);
    init();
  }

  /**
   * Called by every constructor to initialize the extensions 
   * of this renderer.
   */
  protected void init() {
    // set a default highlight color for the primary series
    setSelectionPaint(Color.YELLOW);
    setSelectionOutlinePaint(Color.RED);
    setUseOutlinePaint(true);
  }
  
  /**
   * Returns always {@code false}, so that no series of the selection
   * renderer is visible in the legend.
   */
  @Override
  public boolean isSeriesVisibleInLegend(int series) {
    return false;
  }
 
  /**
   * Sets the selection model the renderer uses to determine whether
   * a data item is selected or not.
   * @param model the selection model
   */
  @Override
  public void setSelectionModel(CategoryDatasetSelectionModel model) {
    this.selectionModel = model;
  }
  
  /**
   * Returns the selection model the renderer uses to determine whether
   * a data item is selected or not.
   */
  @Override
  public CategoryDatasetSelectionModel getSelectionModel() {
    return selectionModel;
  }
  
  /**
   * Creates and sets an adequate {@link SeriesDatasetSelectionModel}
   * according to the {@link Dataset} of the current {@link Plot}. Automatically
   * called by {@link #setPlot(XYPlot)}.<br>
   * Sets the selection model to {@code null} if the renderer is not
   * connected to a plot or the plot contains no dataset.
   * @see #createSelectionModel(XYDataset) 
   */
  public void createAndSetSelectionModel() {
    Dataset newDataset = JFreeChartUtil.getDatasetForRenderer(getPlot(),this);
    if ( getSelectionModel() == null || getSelectionModel().getDataset() != newDataset )
      setSelectionModel( newDataset != null ? createSelectionModel(newDataset) : null );
  }

  /**
   * Creates an adequate {@link SeriesDatasetSelectionModel}
   * according to the given {@link Dataset}. Subclasses can overwrite this
   * method to create an alternative {@link SeriesDatasetSelectionModel}.
   * @param dataset must be a {@link XYDataset}
   * @throws IllegalArgumentException if {@code dataset} is no {@link XYDataset}
   */
  @Override
  public CategoryDatasetSelectionModel createSelectionModel(Dataset dataset) {
    if ( !(dataset instanceof CategoryDataset) )
      throw new IllegalArgumentException(LangUtil.getSimpleClassName(this)+".createSelectionModel(.) can only be applied for CategoryDataset: "+LangUtil.getSimpleClassName(dataset));
    return new CategoryDatasetSelectionModel((CategoryDataset)dataset);
  }

  /**
   * Returns the color, the selected data is rendered with. This color
   * is used for all series.
   * @return {@code null} if no selection color is defined
   *         for the series
   */
  public Paint getSelectionPaint() {
    return selectionPaint;
  }

  /**
   * Sets the color, the selected data is rendered with. This color
   * is used for all series.
   * @param paint color for the selected data
   */
  public void setSelectionPaint(Paint paint) {
    selectionPaint = paint;
  }
  
  /**
   * Returns the outline color for selected points. This paint is
   * used for all series.
   */
  @Override
  public Paint getSelectionOutlinePaint() {
      return selectionOutlinePaint;
  }
  
  /**
   * Sets the outline paint for selected points. This paint is
   * used for all series.
   * @param paint a color
   */
  @Override
  public void setSelectionOutlinePaint(Paint paint) {
      selectionOutlinePaint = paint;
  }

//  /**
//   * Returns the {@link Stroke}, the selected data is rendered with.
//   * @return {@code null} if no special stroke is set (the base stroke
//   *         should be used)
//   */
//  @Override
//  public Stroke getSelectionStroke() {
//    return selectionStroke;
//  }
//
//  /**
//   * Sets the {@link Stroke}, the selected data is rendered with.
//   * @param stroke {@link Stroke} for the selected data (set to {@code null}
//   *               if the base stroke should be used)
//   */
//  @Override
//  public void setSelectionStroke(Stroke stroke) {
//    this.selectionStroke = stroke;
//  }

  /**
   * Creates and sets a new {@link SeriesDatasetSelectionModel}, if no selection model
   * is set yet, or the selection model's dataset differs from the dataset of the
   * new plot.
   * @param plot the plot rendered by this renderer
   */
  @Override
  public void setPlot(CategoryPlot plot) {
    super.setPlot(plot);
    createAndSetSelectionModel();
  }
  
  /**
   * Sets the rendering color (temporary for the rendering process) to the
   * {@linkplain #getSelectionPaint(int) highlight color}, if the item is
   * defined as selected (by the {@linkplain #getSelectionModel() selection model}).
   */
  @Override
  public void drawItem(Graphics2D g,
      CategoryItemRendererState state,
      Rectangle2D dataArea,
      CategoryPlot plot,
      CategoryAxis domainAxis,
      ValueAxis rangeAxis,
      CategoryDataset dataset,
      int row, // this is the series
      int column, // this is the item index in the series
      int pass) {

    // if selection model and color is set and the item is selected
    // the highlighting selection color is set for the rendering
    // process; otherwise the item is ignored for rendering!!
    if ( getSelectionModel() != null ) {
      Paint selPaint  = getSelectionPaint();
      if ( selPaint != null ) {
        Comparable<?> seriesKey = dataset.getRowKey(row);
        Comparable<?> itemKey   = dataset.getColumnKey(column);
        if ( selectionModel.isItemSelected(seriesKey, itemKey) ) {
          setSeriesPaint( row, selPaint, false );
          // Apply the selection outline color 
          if (getSelectionOutlinePaint() != null) 
        	  setSeriesOutlinePaint( row, getSelectionOutlinePaint(), false );
          
          super.drawItem(g, state, dataArea, plot, domainAxis, rangeAxis, dataset, row, column, pass);
        }
      } else
        LOGGER.warn("No selection paint set. Series is not rendererd highlighted.");
    } else
      LOGGER.debug("No selection model set for renderer. All series rendererd normally.");
  }

}
