/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.jfree.chart.selection;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.apache.log4j.Logger;
import org.geotools.util.WeakHashSet;
import org.jfree.data.general.Dataset;
import org.jfree.data.general.DatasetChangeListener;
import org.jfree.data.general.SeriesDataset;

import schmitzm.lang.LangUtil;

/**
 * This class is a general implementation of {@link DatasetSelectionModel} and maintains whether
 * items of a {@link Dataset} are selected in a chart or not. The model connects to
 * the {@link Dataset} as {@link DatasetChangeListener} to recognize the moment a
 * series is removed. Subclasses must implement
 * {@link #datasetChanged(org.jfree.data.general.DatasetChangeEvent)} to react with an
 * automatic unselect.<br>
 * <br>
 * {@code D} = the dataset type the model handles<br>
 * {@code E} = the data type the series are identified with in the dataset<br>
 * {@code F} = the data type the data items are identified with in each series
 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 */
public abstract class AbstractDatasetSelectionModel<D extends Dataset, E, F> implements DatasetSelectionModel<D, E, F> {
  /** {@link Logger} to log waring, debug or error messages. */
  protected Logger LOGGER = LangUtil.createLogger(this);

  /** Holds the {@link Dataset} whose data can be selected. */
  protected D dataset = null;

  /** Contains a {@link Set} for each series. In each of these sets the
   *  selected items are stored. */
  protected Map<E,Set<F>> selectedSeriesIdx = new HashMap<E,Set<F>>();
  
  /** Holds the listeners to the {@link DatasetSelectionModel}. */
  protected WeakHashSet<DatasetSelectionListener> listeners = new WeakHashSet<DatasetSelectionListener>(DatasetSelectionListener.class);

  /** Indicates whether a selection change is a part of multiple
   *  changes. If {@code true} {@link #refreshSelection()} has NO effect
   *  until {@link #setValueIsAdjusting(boolean) setValueIsAdjusting(false)}
   *  is called. */
  private boolean valueIsAdjusting = false;

  /**
   * Creates a new selection model.
   * @param dataset {@link Dataset} whose data can be selected
   */
  public AbstractDatasetSelectionModel(D dataset) {
    this.dataset = dataset;
    dataset.addChangeListener(this);
  }
  
  /**
   * Creates a {@link AbstractDatasetSelectionModel} for a {@link Dataset}.
   * @param dataset a dataset
   * @return {@code null} if selection is not supported for dataset
   */
  public static AbstractDatasetSelectionModel<?,?,?> createInstanceFor(Dataset dataset) {
    if ( dataset instanceof SeriesDataset )
      return new SeriesDatasetSelectionModel((SeriesDataset)dataset);
    return null;
  }

  /**
   * Returns the {@link Dataset} whose data can be selected. 
   */
  @Override
  public D getDataset() {
    return dataset;
  }
  
  /**
   * Checks if an series item is selected.
   * @param seriesKey key of the series
   * @param item      the item in the series
   */
  @Override
  public boolean isItemSelected(E seriesKey, F item) {
    // determine the set for the series
    Set<F> selectedIdx = selectedSeriesIdx.get(seriesKey);
    // check whether item is selected
    return selectedIdx != null && selectedIdx.contains(item);  
  }

  /**
   * Sets whether a series item is selected or not.
   * @param seriesKey key of the series
   * @param item      the item in the series
   * @param selected  flag to (de)select the item
   */
  @Override
  public void setItemSelected(E seriesKey, F item, boolean selected) {
    // determine the set for the series
    Set<F> selectedIdx = selectedSeriesIdx.get(seriesKey);
    // create a set for the series if not yet existing
    if ( selectedIdx == null ) {
      selectedIdx = new HashSet<F>();
      selectedSeriesIdx.put(seriesKey, selectedIdx);
    }
    // add/remove item index
    if ( selected )
      selectedIdx.add(item);
    else
      selectedIdx.remove(item);
    
    // inform all listeners about the selection change
    refreshSelection();
  }

  /**
   * Changes the selection of a series item.
   * @param seriesKey key of the series
   * @param item      the item in the series
   * @return {@code true} if the item changed to selected 
   */
  @Override
  public boolean changeItemSelection(E seriesKey, F item) {
    boolean nowSelected = !isItemSelected(seriesKey, item);
    setItemSelected(seriesKey, item, nowSelected);
    return nowSelected;
  }

  /**
   * Clears the selection.
   */
  @Override
  public void clearSelection() {
    // clear all series sets
    for (Set<F> selectedIdx: selectedSeriesIdx.values())
      selectedIdx.clear();
    // clear the series map
    selectedSeriesIdx.clear();
    // inform all listeners about the selection change
    refreshSelection();
  }
  
  /**
   * Adds a listener which will be informed about changed on the
   * selection model.
   * @param listener the listener to be added
   */
  @Override
  public void addSelectionListener(DatasetSelectionListener listener) {
    listeners.add(listener);
  }
  
  /**
   * Removes a listener from the selection model.
   * @param listener the listener to be remove
   */
  @Override
  public void removeSelectionListener(DatasetSelectionListener listener) {
    listeners.remove(listener);
  }

  /**
   * Returns whether the current selection change is a part of multiple
   * changes. If {@code true} {@link #refreshSelection()} has no effect
   * and no {@link DatasetSelectionChangeEvent} is fired until
   * {@link #setValueIsAdjusting(boolean) setValueIsAdjusting(false)}
   * is called.
   */
  @Override
  public boolean getValueIsAdjusting() {
    return this.valueIsAdjusting;
  }

  /**
   * Sets whether the following selection changes are part of multiple
   * changes. If {@code true} {@link #refreshSelection()} has no effect
   * and no {@link DatasetSelectionChangeEvent} event is fired.
   * If the value is set from {@code true} to {@code false} an automatic
   * refresh is initiated.
   * @see #refreshSelection()
   */
  @Override
  public void setValueIsAdjusting(boolean valueIsAdjusting) {
    if ( valueIsAdjusting != this.valueIsAdjusting ) {
      this.valueIsAdjusting = valueIsAdjusting;
      if ( !this.valueIsAdjusting )
        refreshSelection();
    }
  }
  
  /**
   * Informs all listeners about a general selection change. Does nothing
   * if {@link #getValueIsAdjusting()} is {@code true}.
   */
  @Override
  public void refreshSelection() {
    if ( !getValueIsAdjusting() )
      for (DatasetSelectionListener l : listeners)
        l.selectionChanged( createSelectionChangeEvent() );
  }
  
  /**
   * Creates a {@link DatasetSelectionChangeEvent}.
   */
  protected DatasetSelectionChangeEvent createSelectionChangeEvent() {
    return new DatasetSelectionChangeEvent(this);
  }
}
