/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.jfree.chart.selection;

import org.jfree.data.general.Dataset;
import org.jfree.data.general.DatasetChangeEvent;
import org.jfree.data.general.DatasetChangeListener;

/**
 * This selection model maintains whether items of a {@link Dataset} are selected
 * in a chart or not. The model connects to the {@link Dataset} as {@link DatasetChangeListener}
 * to recognize the moment a series is removed. Subclasses must implement
 * {@link #datasetChanged(org.jfree.data.general.DatasetChangeEvent)} to react with an
 * automatic unselect.<br>
 * <br>
 * {@code D} = the dataset type the model handles<br>
 * {@code E} = the data type the series are identified with in the dataset<br>
 * {@code F} = the data type the data items are identified with in each series
 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 */
public interface DatasetSelectionModel<D extends Dataset, E, F> extends DatasetChangeListener {
  /**
   * Returns the {@link Dataset} whose data can be selected. 
   */
  public D getDataset();
  
  /**
   * Checks if an series item is selected.
   * @param seriesKey key of the series
   * @param item      the item in the series
   */
  public boolean isItemSelected(E seriesKey, F item);

  /**
   * Sets whether a series item is selected or not.
   * @param seriesKey key of the series
   * @param item      the item in the series
   * @param selected  flag to (de)select the item
   */
  public void setItemSelected(E seriesKey, F item, boolean selected);
  
  /**
   * Changes the selection of a series item.
   * @param seriesKey key of the series
   * @param item      the item in the series
   * @return {@code true} if the item changed to selected 
   */
  public boolean changeItemSelection(E seriesKey, F item);

  /**
   * Clears the selection.
   */
  public void clearSelection();
  
  /**
   * Adds a listener which will be informed about changed on the
   * selection model.
   * @param listener the listener to be added
   */
  public void addSelectionListener(DatasetSelectionListener listener);
  
  /**
   * Removes a listener from the selection model.
   * @param listener the listener to be remove
   */
  public void removeSelectionListener(DatasetSelectionListener listener);
  
  /**
   * Returns whether the current selection change is a part of multiple
   * changes. If {@code true} {@link #refreshSelection()} has no effect
   * and no {@link DatasetSelectionChangeEvent} is fired until
   * {@link #setValueIsAdjusting(boolean) setValueIsAdjusting(false)}
   * is called.
   */
  public boolean getValueIsAdjusting();

  /**
   * Sets whether the following selection changes are part of multiple
   * changes. If {@code true} {@link #refreshSelection()} has no effect
   * and no {@link DatasetSelectionChangeEvent} event is fired.
   * If the value is set from {@code true} to {@code false} an automatic
   * refresh is initiated.
   * @see #refreshSelection()
   */
  public void setValueIsAdjusting(boolean valueIsAdjusting);

  /**
   * Informs all listeners about a general selection change.
   */
  public void refreshSelection();
  
  
  /**
   * Listens for dataset changes. If a some data items are no longer present
   * in the dataset, the respective items are removed from selection list.
   */
  @Override
  public void datasetChanged(DatasetChangeEvent event);

}
