/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
/*
 *  SDSS Framework - This file is part of the Spatial Decision Support System Framework Platform
 *
 *  AbstractChartStyle.java
 *
 *  Responsible developper: Andreas Enders, Geographical Institut, Hydrology Research Group, Meckenheimer Allee 166, 53115 Bonn, Germany
 *                                          IMPETUS Project, www.impetus.uni-koeln.de
 *  Contact Information:                    info@andreas-enders.de
 */

package schmitzm.jfree.chart.style;

import java.awt.Color;
import java.util.HashMap;
import java.util.Map;
import java.util.Vector;

import org.apache.log4j.Category;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.Axis;
import org.jfree.chart.labels.CategoryToolTipGenerator;
import org.jfree.chart.labels.StandardCategoryToolTipGenerator;
import org.jfree.chart.labels.StandardXYToolTipGenerator;
import org.jfree.chart.labels.XYToolTipGenerator;
import org.jfree.chart.plot.CategoryPlot;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.AbstractRenderer;
import org.jfree.chart.renderer.category.CategoryItemRenderer;
import org.jfree.chart.renderer.xy.XYItemRenderer;
import org.jfree.chart.title.TextTitle;
import org.jfree.chart.urls.CategoryURLGenerator;
import org.jfree.chart.urls.StandardCategoryURLGenerator;
import org.jfree.chart.urls.StandardXYURLGenerator;
import org.jfree.chart.urls.XYURLGenerator;

import schmitzm.jfree.JFreeChartUtil;
import schmitzm.lang.LangUtil;
import skrueger.geotools.CopyableUtil;

/**
 * This class provides an abstract implementation for chart styles.
 * It implements the variables and methods for maintaining the type, the title(s)
 * and flags of the {@link ChartStyle} interface.
 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a>
 * @version 1.0
 */
public abstract class AbstractChartStyle implements ChartStyle {
  /** Tooltip generator used for XY-Charts. */
  protected XYToolTipGenerator TIPGEN_XY = new StandardXYToolTipGenerator();
  /** Tooltip generator used for Category-Charts. */
  protected CategoryToolTipGenerator TIPGEN_CAT = new StandardCategoryToolTipGenerator();
  /** URL generator used for XY-Charts. */
  protected XYURLGenerator URLGEN_XY = new StandardXYURLGenerator();
  /** URL generator used for Category-Charts. */
  protected CategoryURLGenerator URLGEN_CAT = new StandardCategoryURLGenerator();
  
  
  /** Logger for this class */
  protected final Category LOGGER = LangUtil.createLogger(this);
  /** Holds a (unique) ID for the style. */
  protected String id = "";
  /** Holds the chart type.
   *  @see  ChartStyle#ChartType */
  protected ChartType type = null;
  /** Holds the chart title. */
  protected ChartLabelStyle titleStyle = new ChartLabelStyle();
  /** Holds a long description for the chart. */
  protected ChartLabelStyle descStyle = null;
  /** Holds the background color for the chart. */
  protected Color bgPaint = null;
  /** Stores whether a border is shown around the whole chart. */
  protected boolean borderVisible = false;
  /** Holds the styles of the chart axis. */
  protected Map<Integer,ChartAxisStyle> axisStyle = new HashMap<Integer,ChartAxisStyle>();
  /** Holds the plot orientation (horizontal/vertical). */
  protected PlotOrientation orientation = null;
  /** Stores whether a legend is generated for the chart (Default: true). */
  protected boolean legend = true;
  /** Stores whether tooltips are generated for the chart data (Default: true). */
  protected boolean tooltips = true;
  /** Stores whether URLs are generated for the chart (Default: true). */
  protected boolean urls = true;
  /** Holds the styles for the chart renderers. */
  protected Map<Integer,ChartRendererStyle> rendererStyle = new HashMap<Integer,ChartRendererStyle>();
  /** Holds the maximum dataset index for which a renderer is specified. */
  protected int maxRendererDatasetIdx = 0;
  /** Holds the style for the chart's plot. */
  protected ChartPlotStyle plotStyle = null;

  /**
   * Creates a line style with default values.
   * @param id a (unique) ID for the style
   */
  public AbstractChartStyle(String id) {
    this( id, ChartType.LINE );
  }

  /**
   * Creates a style with default values.
   * @param id   a (unique) ID for the style
   * @param type style type
   */
  public AbstractChartStyle(String id, ChartType type) {
    this(id,type,null,null,null,true,true,true);
  }

  /**
   * Creates a style.
   * @param id       a (unique) ID for the style
   * @param type     style type
   * @param title    title for the chart
   * @param xTitle   title for the chart's X-axis
   * @param yTitle   title for the chart's Y-axis
   * @param legend   flag whether a legend is generated
   * @param tooltips flag whether toolstips are generated
   * @param urls     flag whether URLs are generated
   */
  public AbstractChartStyle(String    id,     ChartType type,
                            String    title,  String  xTitle,   String  yTitle,
                            boolean   legend, boolean tooltips, boolean urls) {
    this.id = id;
    setType(type);
    setLegend(legend);
    setTooltips(tooltips);
    setURLs(urls);
    getTitleStyle().setLabel(title);
    if ( xTitle != null )
      setAxisStyle(DOMAIN_AXIS, new ChartAxisStyle(xTitle,Color.BLACK,0.0,0.0) );
    if ( yTitle != null )
      setAxisStyle(RANGE_AXIS, new ChartAxisStyle(yTitle,Color.BLACK,0.0,0.0) );
  }
  
  /**
   * Creates a (deep) clone of this chart style.
   */
  @Override
  public abstract AbstractChartStyle copy();
  
  /**
   * Copies all properties of this chart style to another
   * chart style.
   * @param dest destination object (if {@code null} the copy
   *             is created by {@link #copy()})
   * @return {@code dest} or the new instance
   * @exception IllegalArgumentException if {@code dest} is no
   *            {@link AbstractChartStyle}
   */
  @Override
  public ChartStyle copyTo(ChartStyle dest) {
    if ( dest == null )
      return copy();
    if ( !(dest instanceof AbstractChartStyle) )
      throw new IllegalArgumentException(LangUtil.getSimpleClassName(this)+".copy(..) can not be applied on "+LangUtil.getSimpleClassName(dest));

    AbstractChartStyle destStyle = (AbstractChartStyle)dest;
    destStyle.id = getID();
    destStyle.setType( getType() );
    destStyle.setTitleStyle( CopyableUtil.copyOrCreate(getTitleStyle(), destStyle.getTitleStyle()) );
    destStyle.setDescStyle( CopyableUtil.copyOrCreate(getDescStyle(), destStyle.getDescStyle()) );
    destStyle.setBackground( getBackground() );
    destStyle.setBorderVisible( isBorderVisible() );
    destStyle.setOrientation( getOrientation() );
    destStyle.setLegend( hasLegend() );
    destStyle.setTooltips( hasTooltips() );
    destStyle.setURLs( hasURLs() );
    destStyle.setPlotStyle( CopyableUtil.copyOrCreate(getPlotStyle(), destStyle.getPlotStyle()) );
    
    destStyle.axisStyle.clear();
    for ( Integer axis : axisStyle.keySet() )
      destStyle.setAxisStyle(axis, getAxisStyle(axis).copy() );
    destStyle.rendererStyle.clear();
    for ( Integer renderer : rendererStyle.keySet() )
      destStyle.setRendererStyle(renderer, getRendererStyle(renderer).copy() );

    return destStyle;
  }

  /**
   * Returns the (unique) ID for the style.
   */
  public String getID() {
    return id;
  }

  /**
   * Returns the type of the chart style.
   */
  public ChartType getType() {
    return type;
  }

  /**
   * Sets the type of the chart style.
   * @param type the chart type
   * @see ChartStyle#ChartType
   */
  public void setType(ChartType type) {
    this.type = type;
  }

  /**
   * Returns the orientation of the chart.
   */
  public PlotOrientation getOrientation() {
    return orientation;
  }

  /**
   * Sets the orientation of the chart.
   * @param orientation the orientation of the chart
   */
  public void setOrientation(PlotOrientation orientation) {
    this.orientation = orientation;
  }

  /**
   * Returns the style of the charts title.
   */
  public ChartLabelStyle getTitleStyle() {
    return titleStyle;
  }
  
  /**
   * Sets the style of the charts title.
   * @param style the new style (if {@code null} the style is
   *        reset to an empty default style, so that the
   *        getter method always can be accessed!)
   */
  public void setTitleStyle(ChartLabelStyle style) {
    if ( style == null )
      style = new ChartLabelStyle();
    titleStyle = style;
  }

  /**
   * Returns the style of the long chart description.
   */
  public ChartLabelStyle getDescStyle() {
    return descStyle;
  }

  /**
   * Sets the long description for the chart.
   * @param descStyle the new chart description style
   */
  public void setDescStyle(ChartLabelStyle descStyle) {
    this.descStyle = descStyle;
  }

  /**
   * Returns the background color of the chart.
   */
  public Color getBackground() {
    return bgPaint;
  }
  
  /**
   * Sets the background color of the chart.
   * @param paint the background color for the chart
   */
  public void setBackground(Color paint) {
    this.bgPaint = paint;
  }

  /**
   * Returns whether a border is shown around the whole chart.
   */
  public boolean isBorderVisible() {
    return this.borderVisible;
  }
  
  /**
   * Sets whether a border is shown around the whole chart.
   */
  public void setBorderVisible(boolean visible) {
    this.borderVisible = visible;
  }

  /**
   * Returns whether the chart is configured to generate a legend.
   */
  public boolean hasLegend() {
    return legend;
  }

  /**
   * Sets whether the chart is configured to generate a legend.
   */
  public void setLegend(boolean legend) {
    this.legend = legend;
  }

  /**
   * Returns whether the chart is configured to generate tooltips for the data.
   */
  public boolean hasTooltips() {
    return tooltips;
  }

  /**
   * Sets whether the chart is configured to generate tooltips for the data.
   */
  public void setTooltips(boolean tooltips) {
    this.tooltips = tooltips;
  }

  /**
   * Returns whether the chart is configured to generate URLs.
   */
  public boolean hasURLs() {
    return urls;
  }

  /**
   * Sets whether the chart is configured to generate URLs.
   */
  public void setURLs(boolean urls) {
    this.urls = urls;
  }

  /**
   * Returns the axis count.
   */
  public int getAxisCount() {
    return axisStyle.size();
  }

  /**
   * Returns the style of a chart axis.
   * @param axis axis number (0=domain, 1=range, ...)
   * @return {@code null} if the chart has no such axis
   */
  public ChartAxisStyle getAxisStyle(int axis) {
    return axisStyle.get(axis);
  }
  
  /**
   * Sets the style of a chart axis.
   * @param axis axis number (0=domain, 1=range, ...)
   * @param style style for the axis
   */
  public void setAxisStyle(int axis, ChartAxisStyle style) {
    axisStyle.put(axis, style);
  }
  
  /**
   * Returns count of renderer for the chart.
   */
  public int getRendererCount() {
    return maxRendererDatasetIdx + 1;
  }

  /**
   * Returns the style of a dataset renderer.
   * @param dataset dataset number (starting with 0)
   * @return {@code null} if the chart has no such renderer
   */
  public ChartRendererStyle getRendererStyle(int dataset) {
    return rendererStyle.get(dataset);
  }
  
  /**
   * Sets the style of a dataset renderer.
   * @param dataset dataset number (starting with 0)
   * @param style style for the dataset renderer
   */
  public void setRendererStyle(int dataset, ChartRendererStyle style) {
    rendererStyle.put(dataset, style);
    maxRendererDatasetIdx = LangUtil.max( rendererStyle.keySet() );
  }

  /**
   * Returns the style of the chart's plot.
   * @return {@code null} if no such style is defined
   */
  public ChartPlotStyle getPlotStyle() {
    return plotStyle;
  }
  
  /**
   * Sets the style of the chart's plot.
   * @param style style for the plot
   */
  public void setPlotStyle(ChartPlotStyle style) {
    plotStyle = style;
  }

  /**
   * Applies the properties of this style to an existing chart.
   * @param chart chart to be styled
   */
  public void applyToChart(JFreeChart chart) {
    // Style the common attributes
    if ( getBackground() != null )
      chart.setBackgroundPaint( getBackground() );
    if ( chart.getPlot() instanceof XYPlot ) {
      XYPlot plot = chart.getXYPlot();
      plot.getRenderer().setBaseToolTipGenerator( hasTooltips() ? TIPGEN_XY : null );
      plot.getRenderer().setURLGenerator( hasURLs() ? URLGEN_XY : null );
      if ( getOrientation() != null )
        plot.setOrientation( getOrientation() );
      
      // style renderer
      for (int i=0; i<plot.getRendererCount(); i++) {
        XYItemRenderer     renderer      = plot.getRenderer(i);
        ChartRendererStyle rendererStyle = getRendererStyle(i);
        if ( renderer != null && rendererStyle != null )
          rendererStyle.applyToRenderer((AbstractRenderer)renderer);
      }
    } else if ( chart.getPlot() instanceof CategoryPlot ) {
      CategoryPlot plot = chart.getCategoryPlot();
      plot.getRenderer().setBaseToolTipGenerator( hasTooltips() ? TIPGEN_CAT : null );
      plot.getRenderer().setBaseItemURLGenerator( hasURLs() ? URLGEN_CAT : null );
      if ( getOrientation() != null )
        plot.setOrientation( getOrientation() );

      // style renderer
      for (int i=0; i<plot.getRendererCount(); i++) {
        CategoryItemRenderer renderer      = plot.getRenderer(i);
        ChartRendererStyle   rendererStyle = getRendererStyle(i);
        if ( renderer != null && rendererStyle != null )
          rendererStyle.applyToRenderer((AbstractRenderer)renderer);
      }
    }
    
    // style plot
    if ( getPlotStyle() != null )
      getPlotStyle().applyToPlot(chart.getPlot());

    // Style the legend
    JFreeChartUtil.setLegendVisible(chart, hasLegend());
    
    // Style the chart title
    TextTitle title = chart.getTitle();
    if ( title == null && getTitleStyle() != null ) {
      title = new TextTitle();
      chart.setTitle( title );
    }
    
    if ( getTitleStyle().getLabel() != null )
      title.setText( getTitleStyle().getLabel() );
    if ( getTitleStyle().getPaint() != null )
      title.setPaint( getTitleStyle().getPaint() );

    // Style the chart sub title with chart description
    if ( getDescStyle() != null ) {
      if ( chart.getSubtitleCount() == 0 || !(chart.getSubtitle(0) instanceof TextTitle) )
        chart.addSubtitle(0, new TextTitle());
      TextTitle subTitle = (TextTitle)chart.getSubtitle(0);
      if ( getDescStyle().getLabel() != null )
        subTitle.setText( getDescStyle().getLabel() );
      if ( getDescStyle().getPaint() != null )
        subTitle.setPaint( getDescStyle().getPaint() );
    }
    
    // Style the chart axis
    Vector<Axis> chartAxis = new Vector<Axis>();
    if ( chart.getPlot() instanceof XYPlot ) {
      XYPlot plot = chart.getXYPlot();
      if ( plot.getDomainAxisCount() > 0 )
        chartAxis.add( plot.getDomainAxis(0) );
      for (int i=0; i<plot.getRangeAxisCount(); i++)
        chartAxis.add( plot.getRangeAxis(i) );
    }
    if ( chart.getPlot() instanceof CategoryPlot ) {
      CategoryPlot plot = chart.getCategoryPlot();
      if ( plot.getDomainAxisCount() > 0 )
        chartAxis.add( plot.getDomainAxis(0) );
      for (int i=0; i<plot.getRangeAxisCount(); i++)
        chartAxis.add( plot.getRangeAxis(i) );
    }
    for (int i=0; i<chartAxis.size(); i++)
      if ( i<getAxisCount() )
        getAxisStyle(i).applyToAxis( chartAxis.elementAt(i) );
      else
        LOGGER.warn("Style contains no style definition for axis "+i);
  }
}
