/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
/*
 *  SDSS Framework - This file is part of the Spatial Decision Support System Framework Platform
 *
 *  BasicChartStyle.java
 *
 *  Responsible developper: Andreas Enders, Geographical Institut, Hydrology Research Group, Meckenheimer Allee 166, 53115 Bonn, Germany
 *                                          IMPETUS Project, www.impetus.uni-koeln.de
 *  Contact Information:                    info@andreas-enders.de
 */

package schmitzm.jfree.chart.style;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.plot.CategoryPlot;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.plot.XYPlot;
import org.jfree.data.category.CategoryDataset;
import org.jfree.data.general.Dataset;
import org.jfree.data.xy.IntervalXYDataset;
import org.jfree.data.xy.TableXYDataset;
import org.jfree.data.xy.XYBarDataset;
import org.jfree.data.xy.XYDataset;

import schmitzm.jfree.JFreeChartUtil;
import schmitzm.lang.LangUtil;

/**
 * This class provides a basis chart style for line, area an bar charts.
 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a>
 * @version 1.0
 */
public class BasicChartStyle extends AbstractChartStyle {
  /** Stores whether the chart is stacked. Line charts can never be stacked. */
  protected boolean stacked = false;
  /** Stores whether the chart is stepped. Bar charts are always stepped. */
  protected boolean stepped = false;

  /**
   * Creates a normal line style with default values.
   * @param id   a (unique) ID for the style
   */
  public BasicChartStyle(String id) {
    this(id, ChartType.LINE);
  }

  /**
   * Creates a normal, vertical line, bar or area style.
   * @param id   a (unique) ID for the style
   * @param type type of the chart layout
   */
  public BasicChartStyle(String id, ChartType type) {
    this(id, type, false, false, PlotOrientation.VERTICAL);
  }

  /**
   * Creates a line, area or bar chart style.
   * @param id          a (unique) ID for the style
   * @param type        style type
   * @param stacked     flag whether the chart is stacked (ignored for line charts)
   * @param stepped     flag whether the chart is stepped (ignored for bar charts)
   * @param orientation orientation of the chart ({@link PlotOrientation.VERTICAL} or {@link PlotOrientation.HORIZONTAL})
   */
  public BasicChartStyle(String id, ChartType type, boolean stacked, boolean stepped, PlotOrientation orientation) {
    this(id, type, stacked, stepped, orientation, null, null, null, true, true, true);
  }

  /**
   * Creates a chart style.
   * @param id       a (unique) ID for the style
   * @param type     style type
   * @param stacked  flag whether the chart is stacked (ignored for line charts)
   * @param stepped  flag whether the chart is stepped (ignored for bar charts)
   * @param orientation orientation of the chart ({@link PlotOrientation.VERTICAL} or {@link PlotOrientation.HORIZONTAL})
   * @param title    title for the chart
   * @param xTitle   title for the chart's X-axis
   * @param yTitle   title for the chart's Y-axis
   * @param legend   flag whether a legend is generated
   * @param tooltips flag whether toolstips are generated
   * @param urls     flag whether URLs are generated
   */
  public BasicChartStyle(String id, ChartType type, boolean stacked, boolean stepped,
                         PlotOrientation orientation,
                         String title,   String xTitle,    String yTitle,
                         boolean legend, boolean tooltips, boolean urls) {
    super(id, type, title, xTitle, yTitle, legend, tooltips, urls);
    setStacked( stacked );
    setStepped( stepped );
    setOrientation( orientation );
  }

  /**
   * Throws an {@link UnsupportedOperationException}.
   */
  protected void throwUnsupportedTypeException() {
    throw new IllegalArgumentException("Only ChartType.LINE, ChartType.AREA or ChartType.BAR allowed for "+LangUtil.getSimpleClassName(this)+"!");
  }

  /**
   * Creates a (deep) clone of this chart style.
   */
  @Override
  public BasicChartStyle copy() {
    return (BasicChartStyle) copyTo( new BasicChartStyle("") );
  }
  
  /**
   * Copies all properties of this chart style to another
   * chart style.
   * @param dest destination object (if {@code null} the copy
   *             is created by {@link #copy()})
   * @return {@code dest} or the new instance
   * @exception IllegalArgumentException if {@code dest} is no
   *            {@link AbstractChartStyle}
   */
  @Override
  public ChartStyle copyTo(ChartStyle dest) {
    // copy the super class properties
    dest = super.copyTo(dest);
    if ( dest instanceof BasicChartStyle ) {
      BasicChartStyle destBCS = (BasicChartStyle) dest;
      // copy this classes properties
      destBCS.setStacked( isStacked() );
      destBCS.setStepped( isStepped() );
    }
    return dest;
  }

  /**
   * Sets the type of the chart style.
   * @param type the chart type
   * @exception IllegalArgumentException if {@link aType} is not of
   *            {@link ChartType#LINE}, {@link ChartType#AREA} or
   *            {@link ChartType#BAR}
   */
  public void setType(ChartType type) {
    super.setType(type);
    switch( type ) {
      // supported types
      case LINE: stacked = false;
                 break;
      case BAR:  stepped = true;
                 break;
      case AREA: break;
      // unsupported types
      default: throwUnsupportedTypeException();
    }
    super.setType(type);
  }
  
  /**
   * Returns whether the chart is stacked.
   * @return {@code false} for line charts
   */
  public boolean isStacked() {
    return stacked;
  }

  /**
   * Sets whether the chart is stacked. Sets the {@link #stacked} property
   * to {@code false} for line charts, even {@code aStacked} is {@code true}.
   */
  public void setStacked(boolean stacked) {
    if ( type == ChartType.LINE && stacked ) {
      LOGGER.warn("Stacked property can not be enabled for line charts! Command ignored...");
      stacked = false;
    }
    this.stacked = stacked;

    if ( type == ChartType.AREA && stepped && stacked ) {
      LOGGER.warn("Area chart style can not be stepped and stacked at the same time. Stepped property automatically disabled...");
      stepped = false;
    }
  }

  /**
   * Returns whether the chart is stepped.
   * @return {@code true} for bar charts
   */
  public boolean isStepped() {
    return stepped;
  }

  /**
   * Sets whether the chart is stepped. Sets the {@link #stepped} property
   * to {@code true} for bar charts, even {@code stepped} is {@code false}.
   */
  public void setStepped(boolean stepped) {
    if ( type == ChartType.BAR && !stepped) {
      LOGGER.warn("Stepped property can not be disabled for bar charts. Command ignored...");
      stepped = true;
    }
    this.stepped = stepped;

    if ( type == ChartType.AREA && stacked && stepped ) {
      LOGGER.warn("Area chart style can not be stepped and stacked at the same time. Stacked property automatically disabled...");
      stacked = false;
    }
  }

  /**
   * Creates a chart for the given {@link Dataset} and applies the style.
   * @param dataset the data for the chart
   * @exception UnsupportedOperationException if the style can not be applied
   *            to the given dataset
   */
  public JFreeChart applyToDataset(Dataset dataset) {
    JFreeChart chart = null;
    switch ( type ) {
      case LINE: chart = createLineChart(dataset);
                 break;
      case AREA: chart = createAreaChart(dataset);
                 break;
      case BAR:  chart = createBarChart(dataset);
                 break;
      // unsupported types
      default: throwUnsupportedTypeException();
    }
    applyToChart(chart);
    return chart;
  }
  
  /**
   * Applies the style to an existing chart.
   * @param chart chart the style is applied to
   */
  public void applyToChart(JFreeChart chart) {
    // apply the stacked and stepped property
    if ( chart.getPlot() instanceof XYPlot )
      chart.getXYPlot().setRenderer( JFreeChartUtil.createXYRenderer(this) );
    else if ( chart.getPlot() instanceof CategoryPlot )
      chart.getCategoryPlot().setRenderer( JFreeChartUtil.createCategoryRenderer(this) );
    else
      LOGGER.warn("Properties of BasicChartStyle can not be applied to plot: "+chart.getPlot());
    // Note: It is important that the super-apply comes after
    //       the upper stuff, so that changes on the renderer
    //       effect the new (special stacked/stepped) renderer!!
    super.applyToChart(chart);
  }
  

  /**
   * Creates a line chart from an {@link XYDataset} or {@link CategoryDataset}
   * @param dataset Dataset
   */
  private JFreeChart createLineChart(Dataset dataset) {
    JFreeChart chart = null;
    // XYDataset -> XYLineChart
    if (dataset instanceof XYDataset) {
      chart = ChartFactory.createXYLineChart(
          null,null,null, // Title and axis will be configured in applyToChart(..)
          (XYDataset)dataset,
          orientation,
          legend,
          tooltips,
          urls
      );
    }
    // CategoryDataset -> LineChart
    if (dataset instanceof CategoryDataset) {
      chart = ChartFactory.createLineChart(
          null,null,null, // Title and axis will be configured in applyToChart(..)
          (CategoryDataset)dataset,
          orientation,
          legend,
          tooltips,
          urls
      );
    }
    return chart;
  }

  /**
   * Creates an area chart from an {@link XYDataset} or {@link CategoryDataset}
   * @param dataset Dataset
   */
  private JFreeChart createAreaChart(Dataset dataset) {
    JFreeChart chart = null;
    // XYDataset -> XYAreaChart
    if (dataset instanceof XYDataset) {
      chart = ChartFactory.createXYAreaChart(
          null,null,null, // Title and axis will be configured in applyToChart(..)
          (XYDataset) dataset,
          orientation,
          legend,
          tooltips,
          urls
          );
    }
    // CategoryDataset -> AreaChart
    if (dataset instanceof CategoryDataset) {
      chart = ChartFactory.createAreaChart(
          null,null,null, // Title and axis will be configured in applyToChart(..)
          (CategoryDataset)dataset,
          orientation,
          legend,
          tooltips,
          urls
      );
    }

    return chart;
  }

  /**
   * Creates a bar chart from an {@link XYDataset} or {@link CategoryDataset}
   * @param dataset Dataset
   */
  private JFreeChart createBarChart(Dataset dataset) {
    JFreeChart chart = null;
    // XYDataset -> XYBarChart
    if (dataset instanceof XYDataset) {
      IntervalXYDataset barDataset = null;
      if ( dataset instanceof TableXYDataset )
        barDataset = (IntervalXYDataset)dataset;
      else {
        barDataset = new XYBarDataset((XYDataset)dataset, 0.5d);
        // copy properties to new dataset
        JFreeChartUtil.applyDatasetProperties(barDataset,dataset);
      }
        
      chart = ChartFactory.createXYBarChart(
          null,null,false,null, // Title and axis will be configured in applyToChart(..)
          barDataset,
          PlotOrientation.VERTICAL,
          true,
          true,
          true
      );
    }
    // CategoryDataset -> BarChart
    if (dataset instanceof CategoryDataset) {
      chart = ChartFactory.createBarChart(
          null,null,null, // Title and axis will be configured in applyToChart(..)
          (CategoryDataset)dataset,
          orientation,
          legend,
          tooltips,
          urls
      );
    }

    return chart;
  }
}
