/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.jfree.chart.style;

import java.awt.Color;
import java.util.Locale;

import org.apache.log4j.Category;

import schmitzm.lang.LangUtil;
import skrueger.geotools.Copyable;
import skrueger.i8n.Translation;

/**
 * This class is a general super class for a design style of text displayed
 * in a chart. This can be an localized text according to {@link Translation} or
 * a simple (unlocalized) String. The {@link #setLabel(String)} method will
 * try to decode the string for multiple languages. If this is not possible
 * the core string is taken as label.
 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a>
 * @version 1.0
 */
public class ChartLabelStyle implements Copyable<ChartLabelStyle>{
  /** Logger for this class */
  protected final Category LOGGER = LangUtil.createLogger(this);

  /** Holds the label text. */
  protected Translation label = null;
  /** Holds the text color of the label. */
  protected Color paint = null;
  
  @Override
	public String toString() {
		if (label != null && paint != null)
			return "'" + label + "' in " + paint;
		return super.toString();
	}
  
  /**
   * Creates a new style with default values (empty label, color undefined).
   */
  public ChartLabelStyle() {
    this("");
  }

  /**
   * Creates a new style (color undefined).
   * @param label label text (localized or static label is provided)
   */
  public ChartLabelStyle(String label) {
    this(new Translation(label), null);
  }
    /**
   * Creates a new style.
   * @param label label text (localized or static label is provided)
   * @param color text color
   */
  public ChartLabelStyle(String label, Color color) {
    this(new Translation(label), color);
  }
  
  /**
   * Creates a new style.
   * @param label label text (static
   * @param color text color
   */
  public ChartLabelStyle(Translation label, Color color) {
    setLabelTranslation(label);
    setPaint(color);
  }
  
  /**
   * Creates a (deep) clone of this style.
   */
  @Override
  public ChartLabelStyle copy() {
    return copyTo( new ChartLabelStyle() );
  }
  
  /**
   * Copies all properties of this style to another one.
   * @param dest destination object (if {@code null} the copy
   *             is created by {@link #copy()})
   * @return {@code dest} or the new instance
   */
  @Override
  public ChartLabelStyle copyTo(ChartLabelStyle dest) {
    if ( dest == null )
      return copy();
    dest.setLabelTranslation( getLabelTranslation().copy() );
    dest.setPaint( getPaint() );
    return dest;
  }

  /**
   * Returns the label text in the default localization.
   * @see Locale#getDefault()
   */
  public String getLabel() {
    return label.toString();
  }

  /**
   * Sets the label text.
   * @param label the text for the label (localized or static label is provided)
   */
  public void setLabel(String label) {
    this.label = new Translation(label);
  }
  
  /**
   * Returns the label text for all localizations.
   */
  public Translation getLabelTranslation() {
    return label;
  }

  /**
   * Sets the label text.
   * @param label the text for the label
   */
  public void setLabelTranslation(Translation label) {
    this.label = label;
  }
  
  /**
   * Returns the color for the label text.
   */
  public Color getPaint() {
    return paint;
  }

  /**
   * Sets the color for the label text.
   * @param color Color for the label text
   */
  public void setPaint(Color color) {
    this.paint = color;
  }
  
}
