/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.jfree.chart.style;

import java.awt.BasicStroke;
import java.awt.Color;

import org.apache.log4j.Category;
import org.jfree.chart.plot.CategoryMarker;
import org.jfree.chart.plot.CategoryPlot;
import org.jfree.chart.plot.Plot;
import org.jfree.chart.plot.ValueMarker;
import org.jfree.chart.plot.XYPlot;
import org.jfree.ui.Layer;
import org.jfree.ui.RectangleInsets;

import schmitzm.jfree.JFreeChartUtil;
import schmitzm.lang.LangUtil;
import skrueger.geotools.Copyable;

/**
 * This class defines the rendering style the charts plot(s). The following
 * properties are available:
 * <ul>
 * <li>foreground transparency</li>
 * <li>background color and transparency</li>
 * <li>insets margins</li>
 * <li>domain grid line visibility and color (only applicable for {@link XYPlot}
 * and {@link CategoryPlot})</li>
 * <li>range grid line visibility and color (only applicable for {@link XYPlot}
 * and {@link CategoryPlot})</li>
 * <li>axis crosshair visibility and color (only applicable for
 * {@link XYPlot} and {@link CategoryPlot})</li>
 * <li>axis crosshair centering (only applicable for
 * {@link XYPlot} and {@link CategoryPlot})</li>
 * </ul>
 * 
 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a>
 * @version 1.0
 */
public class ChartPlotStyle implements Copyable<ChartPlotStyle> {
	private final ValueMarker    VALUE_MARKER = new ValueMarker(0., Color.black, new BasicStroke(1f));
    private final CategoryMarker CAT_MARKER   = new CategoryMarker(0., Color.black, new BasicStroke(1f));

	/** Logger for this class */
	protected final Category LOGGER = LangUtil.createLogger(this);

	/** Holds the inner margins of the plot. */
	protected RectangleInsets insets = null;;
	/** Holds the foreground transparency. */
	protected Float foregroundAlpha = 1.0f;
	/** Holds the background color of the plot. */
	protected Color backgroundPaint = null;
	/** Holds the background transparency. */
	protected Float backgroundAlpha = 1.0f;
	/** Holds whether the domain grid line is visible. */
	protected Boolean domainGridlineVisible = null;
	/** Holds the color of the domain grid line. */
	protected Color domainGridlinePaint = null;
	/** Holds whether the range grid line is visible. */
	protected Boolean rangeGridlineVisible = null;
	/** Holds the color of the range grid line. */
	protected Color rangeGridlinePaint = null;
	/** Holds whether crosshair lines are visible (on 0/0). */
	protected Boolean crosshairVisible = null;
	/** Holds the color of the range grid line. */
	protected Color crosshairPaint = null;
	/** Boolean determining whether the plot will be centered around the axis
	 *  origin and will have extends: Xmax = -Xmin and Ymax = -Ymin **/
	protected boolean centerOriginSymetrically = false;


	/**
	 * Creates a new style with default values.
	 */
	public ChartPlotStyle() {
	}

	/**
	 * Creates a (deep) clone of this style.
	 */
	@Override
	public ChartPlotStyle copy() {
		return copyTo(new ChartPlotStyle());
	}

	/**
	 * Copies all properties of this style to another one.
	 * 
	 * @param dest
	 *            destination object (if {@code null} the copy is created by
	 *            {@link #copy()})
	 * @return {@code dest} or the new instance
	 */
	@Override
	public ChartPlotStyle copyTo(ChartPlotStyle dest) {
		if (dest == null)
			return copy();

		dest.setInsets(getInsets());
		dest.setForegroundAlpha(getForegroundAlpha());
		dest.setBackgroundPaint(getBackgroundPaint());
		dest.setBackgroundAlpha(getBackgroundAlpha());
		dest.setDomainGridlineVisible(isDomainGridlineVisible());
		dest.setDomainGridlinePaint(getDomainGridlinePaint());
		dest.setRangeGridlineVisible(isRangeGridlineVisible());
		dest.setRangeGridlinePaint(getRangeGridlinePaint());
		dest.setCrosshairVisible(isCrosshairVisible());
		dest.setCrosshairPaint(getCrosshairPaint());
		dest.setCenterOriginSymetrically( isCenterOriginSymetrically() );

		return dest;
	}

	/**
	 * Returns the inner margin of the plot.
	 */
	public RectangleInsets getInsets() {
		return insets;
	}

	/**
	 * Sets the inner margin of the plot.
	 */
	public void setInsets(RectangleInsets insets) {
		this.insets = insets;
	}

	/**
	 * Returns the transparency of the foreground.
	 * 
	 * @return {@code null} if no specific transparency is set for the plot
	 */
	public Float getForegroundAlpha() {
		return foregroundAlpha;
	}

	/**
	 * Sets the transparency of the foreground.
	 * 
	 * @param alpha
	 *            transparency value (0..1); can be {@code null} to reset to a
	 *            non-specific value
	 */
	public void setForegroundAlpha(Float alpha) {
		this.foregroundAlpha = alpha;
	}

	/**
	 * Returns the transparency of the background.
	 * 
	 * @return {@code null} if no specific transparency is set for the plot
	 */
	public Float getBackgroundAlpha() {
		return backgroundAlpha;
	}

	/**
	 * Sets the transparency of the background.
	 * 
	 * @param alpha
	 *            transparency value (0..1); can be {@code null} to reset to a
	 *            non-specific value
	 */
	public void setBackgroundAlpha(Float alpha) {
		this.backgroundAlpha = alpha;
	}

	/**
	 * Returns the background color of the plot.
	 * 
	 * @return {@code null} if no specific color is set
	 */
	public Color getBackgroundPaint() {
		return backgroundPaint;
	}

	/**
	 * Sets the background color of the plot.
	 * 
	 * @param color
	 *            background color for the plot (can be {@code null} to reset to
	 *            a non-specific color)
	 */
	public void setBackgroundPaint(Color color) {
		this.backgroundPaint = color;
	}

	/**
	 * Returns whether the plot's domain grid line is visible.
	 * 
	 * @return {@code null} if property is not set for the plot
	 */
	public Boolean isDomainGridlineVisible() {
		return domainGridlineVisible;
	}

	/**
	 * Sets whether the plot's domain grid line is visible.
	 * 
	 * @param visible
	 *            indicates the visibility of the grid line (can be {@code null}
	 *            to reset to a non-specific property)
	 */
	public void setDomainGridlineVisible(Boolean visible) {
		domainGridlineVisible = visible;
	}

	/**
	 * Returns the color of the plot's domain grid line.
	 * 
	 * @return {@code null} if no specific color is set
	 */
	public Color getDomainGridlinePaint() {
		return domainGridlinePaint;
	}

	/**
	 * Sets the color of the plot's domain grid line.
	 * 
	 * @param color
	 *            color for the domain grid line (can be {@code null} to reset
	 *            to a non-specific color)
	 */
	public void setDomainGridlinePaint(Color color) {
		this.domainGridlinePaint = color;
	}

	/**
	 * Returns whether the plot's range grid line is visible.
	 * 
	 * @return {@code null} if property is not set for the plot
	 */
	public Boolean isRangeGridlineVisible() {
		return rangeGridlineVisible;
	}

	/**
	 * Sets whether the plot's range grid line is visible.
	 * 
	 * @param visible
	 *            indicates the visibility of the grid line (can be {@code null}
	 *            to reset to a non-specific property)
	 */
	public void setRangeGridlineVisible(Boolean visible) {
		rangeGridlineVisible = visible;
	}

	/**
	 * Returns the color of the plot's range grid line.
	 * 
	 * @return {@code null} if no specific color is set
	 */
	public Color getRangeGridlinePaint() {
		return rangeGridlinePaint;
	}

	/**
	 * Sets the color of the plot's range grid line.
	 * 
	 * @param color
	 *            color for the range grid line (can be {@code null} to reset to
	 *            a non-specific color)
	 */
	public void setRangeGridlinePaint(Color color) {
		this.rangeGridlinePaint = color;
	}

	/**
	 * Returns whether the plot's crosshair lines are visible.
	 * 
	 * @return {@code null} if property is not set for the plot
	 */
	public Boolean isCrosshairVisible() {
		return crosshairVisible;
	}

	/**
	 * Sets whether the plot's crosshair lines are visible.
	 * 
	 * @param visible
	 *            indicates the visibility of the crosshair lines (can be
	 *            {@code null} to reset to a non-specific property)
	 */
	public void setCrosshairVisible(Boolean visible) {
		crosshairVisible = visible;
	}

	/**
	 * Returns the color of the plot's crosshair lines.
	 * 
	 * @return {@code null} if no specific color is set
	 */
	public Color getCrosshairPaint() {
		return crosshairPaint;
	}

	/**
	 * Sets the color of the plot's crosshair lines.
	 * 
	 * @param color
	 *            color for the crosshair lines (can be {@code null} to reset to
	 *            a non-specific color)
	 */
	public void setCrosshairPaint(Color color) {
		this.crosshairPaint = color;
	}

	/**
	 * Applies the style to a specific plot.<br>
	 * Note: Some style properties can only be applied to specific plot types.
	 * 
	 * @param plot
	 *            a plot
	 */
	public void applyToPlot(Plot plot) {
		if (plot == null) {
			LOGGER.debug("No plot to be customised.");
			return;
		}

		plot.setNotify(false);
		// Apply common attributes
		if (getForegroundAlpha() != null)
			plot.setForegroundAlpha(getForegroundAlpha());
		if (getBackgroundAlpha() != null)
			plot.setBackgroundAlpha(getBackgroundAlpha());
		if (getBackgroundPaint() != null)
			plot.setBackgroundPaint(getBackgroundPaint());
		if (getInsets() != null)
			plot.setInsets(insets);

		// Apply margin according to the specific renderer type
		if (plot instanceof CategoryPlot)
			applyToCategoryPlot((CategoryPlot) plot);
		if (plot instanceof XYPlot)
			applyToXYPlot((XYPlot) plot);

		plot.setNotify(true);
		// notify the listeners about style change
		// JFreeChartUtil.fireChangeEvent(plot);
	}

	/**
	 * Applies style properties which can only be applied to
	 * {@link CategoryPlot}.<br>
	 * Called by {@link #applyToPlot(Plot)}.
	 * 
	 * @param plot
	 *            a {@link CategoryPlot}
	 */
	protected void applyToCategoryPlot(CategoryPlot plot) {
		if (isDomainGridlineVisible() != null)
			plot.setDomainGridlinesVisible(isDomainGridlineVisible());
		if (getDomainGridlinePaint() != null)
			plot.setDomainGridlinePaint(getDomainGridlinePaint());
		if (isRangeGridlineVisible() != null)
			plot.setRangeGridlinesVisible(isRangeGridlineVisible());
		if (getRangeGridlinePaint() != null)
			plot.setRangeGridlinePaint(getRangeGridlinePaint());

		if (isDomainGridlineVisible() != null)
			plot.setDomainGridlinesVisible(isDomainGridlineVisible());
		if (getDomainGridlinePaint() != null)
			plot.setDomainGridlinePaint(getDomainGridlinePaint());
		if (isRangeGridlineVisible() != null)
			plot.setRangeGridlinesVisible(isRangeGridlineVisible());
		if (getRangeGridlinePaint() != null)
			plot.setRangeGridlinePaint(getRangeGridlinePaint());
        if (isCrosshairVisible() != null && isCrosshairVisible() == true) {
          if (getCrosshairPaint() != null)
            CAT_MARKER.setPaint(getCrosshairPaint());
          if (plot.getDomainMarkers(Layer.FOREGROUND) == null ||
              !plot.getDomainMarkers(Layer.FOREGROUND).contains(CAT_MARKER))
            plot.addDomainMarker(CAT_MARKER);
          if (plot.getRangeMarkers(Layer.FOREGROUND) == null ||
              !plot.getRangeMarkers(Layer.FOREGROUND).contains(CAT_MARKER))
            plot.addRangeMarker(CAT_MARKER);
        } else {
          plot.removeDomainMarker(CAT_MARKER);
          plot.removeRangeMarker(CAT_MARKER);
        }
        // Axis origin centration
        plot.getRangeAxis().setAutoRange(!isCenterOriginSymetrically());
        if (isCenterOriginSymetrically()) {
            JFreeChartUtil.zoomAxisToFullExtend( plot.getRangeAxis(), 0.1, isCenterOriginSymetrically());
        }

	}

	/**
	 * Applies style properties which can only be applied to {@link XYPlot}.<br>
	 * Called by {@link #applyToPlot(Plot)}.
	 * 
	 * @param plot
	 *            a {@link XYPlot}
	 */
	protected void applyToXYPlot(XYPlot plot) {

	    // Grid lines
		if (isDomainGridlineVisible() != null)
			plot.setDomainGridlinesVisible(isDomainGridlineVisible());
		if (getDomainGridlinePaint() != null)
			plot.setDomainGridlinePaint(getDomainGridlinePaint());
		if (isRangeGridlineVisible() != null)
			plot.setRangeGridlinesVisible(isRangeGridlineVisible());
		if (getRangeGridlinePaint() != null)
			plot.setRangeGridlinePaint(getRangeGridlinePaint());
		// Axis origin crosshair lines
		if (isCrosshairVisible() != null && isCrosshairVisible() == true) {
		  if (getCrosshairPaint() != null)
		    VALUE_MARKER.setPaint(getCrosshairPaint());
		  if (plot.getDomainMarkers(Layer.FOREGROUND) == null ||
		      !plot.getDomainMarkers(Layer.FOREGROUND).contains(VALUE_MARKER))
		    plot.addDomainMarker(VALUE_MARKER);
		  if (plot.getRangeMarkers(Layer.FOREGROUND) == null ||
		      !plot.getRangeMarkers(Layer.FOREGROUND).contains(VALUE_MARKER))
		    plot.addRangeMarker(VALUE_MARKER);
		} else {
		  plot.removeDomainMarker(VALUE_MARKER);
		  plot.removeRangeMarker(VALUE_MARKER);
		}
        // Axis origin centration
		plot.getRangeAxis().setAutoRange(!isCenterOriginSymetrically());
		plot.getDomainAxis().setAutoRange(!isCenterOriginSymetrically());
		if (isCenterOriginSymetrically()) {
			JFreeChartUtil.zoomAxisToFullExtend( plot.getDomainAxis(), 0.1, isCenterOriginSymetrically());
			JFreeChartUtil.zoomAxisToFullExtend( plot.getRangeAxis(), 0.1, isCenterOriginSymetrically());
		}
// TEST
        plot.setDomainPannable(true);
        plot.setRangePannable(true);

	}
	

    /**
     * Sets whether the axis origin is centered in the plot.
     * @param originSym indicated whether the axis origin should
     *                  be centered in the plot 
     */
	public void setCenterOriginSymetrically(boolean originSym) {
		centerOriginSymetrically = originSym;
	}

	/**
	 * Returns whether the axis origin is centered in the plot. 
	 */
	public boolean isCenterOriginSymetrically() {
		return centerOriginSymetrically;
	}

}
