/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.jfree.chart.style;

import java.awt.Color;

import org.apache.log4j.Logger;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.data.general.Dataset;

import skrueger.geotools.Copyable;


/**
 * This interface is a general super class for a design style of a {@link JFreeChart}.
 * Sub classes should read the style from an xml element or something like that.
 * Maybe there could be a general implementation representing a style for
 * basic charts (bar, line and area charts) and special implementations for
 * special chart types (e.g. financial, statistical oder spider charts).
 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a>
 * @version 1.0
 */
public interface ChartStyle extends Copyable<ChartStyle> {
  final static Logger LOGGER = Logger.getLogger(ChartStyle.class);
  
  /** Constant for the domain axis (X). */
  public static final int DOMAIN_AXIS = 0;
  /** Constant for the domain axis (X). Synonym for {@link #DOMAIN_AXIS}.*/
  public static final int X_AXIS = DOMAIN_AXIS;
  /** Constant for the range axis (Y). */
  public static final int RANGE_AXIS  = 1;
  /** Constant for the range axis (Y). Synonym for {@link #RANGE_AXIS}. */
  public static final int Y_AXIS  = RANGE_AXIS;
  /** Constant for the range axis (Y). Synonym for {@link #RANGE_AXIS}. */
  public static final int Y_AXIS_LEFT  = RANGE_AXIS;
  /** Constant for the second range axis (Y). */
  public static final int RANGE_AXIS2 = 2;
  /** Constant for the second range axis (Y). Synonym for {@link #RANGE_AXIS2}. */
  public static final int Y_AXIS2 = RANGE_AXIS2;
  /** Constant for the second range axis (Y). Synonym for {@link #RANGE_AXIS2}. */
  public static final int Y_AXIS_RIGHT = RANGE_AXIS2;

  /**
   * Returns an (unique) ID for the style.
   */
  public String getID();

  /**
   * Returns the type of the chart style.
   */
  public ChartType getType();

  /**
   * Returns the orientation of the chart.
   * @see PlotOrientation#HORIZONTAL
   * @see PlotOrientation#VERTICAL
   */
  public PlotOrientation getOrientation();
  
  /**
   * Sets the orientation of the chart.
   * @param orientation the orientation of the chart
   */
  public void setOrientation(PlotOrientation orientation);

  /**
   * Returns the style of the chart title. <b>Should never return
   * {@code null}!!</b>
   */
  public ChartLabelStyle getTitleStyle();

  /**
   * Sets the title for the chart.
   * @param titleStyle the new chart title style
   */
  public void setTitleStyle(ChartLabelStyle titleStyle);

  /**
   * Returns the style of the long chart description.
   */
  public ChartLabelStyle getDescStyle();

  /**
   * Sets the long description for the chart.
   * @param descStyle the new chart description style
   */
  public void setDescStyle(ChartLabelStyle descStyle);

  /**
   * Returns the background color of the chart.
   */
  public Color getBackground();
  
  /**
   * Sets the background color of the chart.
   * @param paint the background color for the chart
   */
  public void setBackground(Color paint);

  /**
   * Returns whether a border is shown around the whole chart.
   */
  public boolean isBorderVisible();
  
  /**
   * Sets whether a border is shown around the whole chart.
   */
  public void setBorderVisible(boolean visible);
  
  /**
   * Returns whether the chart is configured to generate a legend.
   */
  public boolean hasLegend();

  /**
   * Sets whether the chart is configured to generate a legend.
   */
  public void setLegend(boolean legend);

  /**
   * Returns whether the chart is configured to generate tooltips for the data.
   */
  public boolean hasTooltips();

  /**
   * Sets whether the chart is configured to generate tooltips for the data.
   */
  public void setTooltips(boolean tooltips);

  /**
   * Returns whether the chart is configured to generate URLs.
   */
  public boolean hasURLs();

  /**
   * Sets whether the chart is configured to generate URLs.
   */
  public void setURLs(boolean urls);

  /**
   * Returns count of axis of the chart.
   */
  public int getAxisCount();

  /**
   * Returns the style of a chart axis.<br>
   * 0 = domain axis (X)<br>
   * 1 = first (left) range axis (Y)<br>
   * 2 = second (right) range axis (Y) if available
   * @return {@code null} if axis style is not available
   */
  public ChartAxisStyle getAxisStyle(int axis);

  /**
   * Sets the style of a chart axis.
   * @param axis axis number (0=domain, 1=range, ...)
   * @param style style for the axis
   */
  public void setAxisStyle(int axis, ChartAxisStyle style);

  
  /**
   * Returns count of renderer for the chart.
   */
  public int getRendererCount();

  /**
   * Returns the style of a dataset renderer.
   * @param dataset dataset number (starting with 0)
   * @return {@code null} if the chart has no such renderer
   */
  public ChartRendererStyle getRendererStyle(int dataset);
  
  /**
   * Sets the style of a dataset renderer.
   * @param dataset dataset number (starting with 0)
   * @param style style for the axis
   */
  public void setRendererStyle(int dataset, ChartRendererStyle style);
  
  /**
   * Returns the style of the chart's plot.
   * @return {@code null} if no such style is defined
   */
  public ChartPlotStyle getPlotStyle();
  
  /**
   * Sets the style of the chart's plot.
   * @param style style for the plot
   */
  public void setPlotStyle(ChartPlotStyle style);

  /**
   * Applies the style to an existing chart.
   * @param chart chart to be styled
   * @exception UnsupportedOperationException if the style can not be applied
   *            to the given chart
   */
  public void applyToChart(JFreeChart chart);

  /**
   * Generates a chart using the given data and the style represented by
   * this class.
   * @param dataset the data for the chart
   * @exception UnsupportedOperationException if the style can not be applied
   *            to the given dataset
   */
  public JFreeChart applyToDataset(Dataset dataset);
  
}
