/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.jfree.chart.style;

import java.awt.Color;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.text.DateFormat;
import java.text.DecimalFormat;
import java.text.Format;
import java.text.NumberFormat;
import java.text.SimpleDateFormat;
import java.util.List;
import java.util.Random;

import org.jdom.Document;
import org.jdom.Element;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.ui.RectangleInsets;

import schmitzm.xml.XMLUtil;

/**
 * This class defines a factory to create a chart style from XML.
 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a>
 * @version 1.0
 */
public class ChartStyleXMLFactory<E extends ChartStyle> {
  /**
   * Reads a chart definition from XML element. The chart style ID
   * is taken from "id" attribute.
   * @param element the XML element
   * @param factory factory to create the style with
   */
  public E createStyleFromXML(Element element) {
    return( createStyleFromXML(element,null) );
  }

  /**
   * Reads a chart definition from XML element.
   * @param element the XML element
   * @param id      the ID for the style (if {@code null} the
   *                ID is taken from "id" attribute)
   */
  public E createStyleFromXML(Element element, String id) {
    if ( id == null )
      id = XMLUtil.getAttribute(element, "id", String.valueOf(new Random().nextInt()));
    
    // Determine the chart type from XML
    String               typeStr   = XMLUtil.getAttribute(element, "type");
    ChartType chartType = ChartStyleUtil.getChartType(typeStr);

    // create default style for given type
    E chartStyle = createDefaultChartStyle(id, chartType);
    // apply the "rest" of the XML definition to style
    applyStyleFromXML(chartStyle, element);
    
    return chartStyle;
  }
  
  /**
   * Creates a default style for a chart type.
   * @param id   a (unique) ID for the style
   * @param type a chart type
   */
  public E createDefaultChartStyle(String id, ChartType type) {
    switch ( type ) {
      case AREA:  
      case LINE:  
      case BAR:  
      case POINT: return (E) new BasicChartStyle(id, type);
      case SCATTER: return (E) new ScatterChartStyle(id);
    }
    throw new UnsupportedOperationException("Style for this chart type not yet supported: "+type);
  }
  
  /**
   * Applies the chart style definition from XML (<b>except the type and id!!</b>)
   * to an existing {@link ChartStyle} object.
   * @param chartStyle an existing chart style ({@code null} not permitted!)
   * @param element    element to read the properties from
   * @exception UnsupportedOperationException if {@code chartStyle} or {@code element}
   *            is {@code null}
   */
  public void applyStyleFromXML(E chartStyle, Element element) {
    if ( chartStyle == null || element == null )
      throw new UnsupportedOperationException("Chart style and element must be given to apply the chart style from XML!");
    
    // Legend visible
    chartStyle.setLegend( XMLUtil.getBooleanAttribute(element,"legend",true) );
    chartStyle.setTooltips( XMLUtil.getBooleanAttribute(element,"tooltips",true) );
    chartStyle.setURLs( XMLUtil.getBooleanAttribute(element,"urls",true) );
    
    // Background color
    Color color = XMLUtil.getColorAttribute(element.getChild("background"),"paint");
    if ( color != null )
      chartStyle.setBackground( color );
    
    // Border around chart visible
    Boolean borderVisible = XMLUtil.getBooleanAttribute(element.getChild("border"),"visible");
    if ( borderVisible != null )
      chartStyle.setBorderVisible( borderVisible );

    // Plot orientation
    String orientationStr = XMLUtil.getAttribute(element,"orientation","horizontal");
    if ("horizontal".equalsIgnoreCase(orientationStr))
      chartStyle.setOrientation( PlotOrientation.HORIZONTAL );
    else if ("vertical".equalsIgnoreCase(orientationStr))
      chartStyle.setOrientation( PlotOrientation.VERTICAL );
    else
      throw new UnsupportedOperationException("Unknown orientation definition: "+orientationStr);

    // Title style
    if ( element.getChild("title") != null )
      chartStyle.setTitleStyle( createLabelStyleFromXML(element.getChild("title") ) );
    // Description style
    if ( element.getChild("desc") != null )
      chartStyle.setDescStyle( createLabelStyleFromXML(element.getChild("desc") ) );
    // Domain axis
    if ( element.getChild("domainAxis") != null )
      chartStyle.setAxisStyle( ChartStyle.DOMAIN_AXIS,createAxisStyleFromXML(element.getChild("domainAxis") ) );
    // Range axis
    if ( element.getChild("rangeAxis") != null )
      chartStyle.setAxisStyle( ChartStyle.RANGE_AXIS,createAxisStyleFromXML(element.getChild("rangeAxis") ) );
    // Second range axis
    if ( element.getChild("rangeAxis2") != null )
      chartStyle.setAxisStyle( ChartStyle.RANGE_AXIS2,createAxisStyleFromXML(element.getChild("rangeAxis2") ) );

    // Rendering of series
    List<Element> rendererElements = element.getChildren("renderer");
    for (int i=0; i<rendererElements.size(); i++)
      chartStyle.setRendererStyle(i,createRendererStyleFromXML(rendererElements.get(i)));
      
    // Plot
    if ( element.getChild("plot") != null )
      chartStyle.setPlotStyle( createPlotStyleFromXML(element.getChild("plot") ) );

    // apply type specific properties
    if ( chartStyle instanceof BasicChartStyle )
      applyStyleFromXML((BasicChartStyle)chartStyle, element);
      
  }
  
  /**
   * Applies the {@link BasicChartStyle} properties from XML (<b>not the {@link ChartStyle}
   * properties!)</b>) to an existing {@link BasicChartStyle} object.
   * @param chartStyle an existing chart style ({@code null} not permitted!)
   * @param element    element to read the properties from
   */
  private void applyStyleFromXML(BasicChartStyle chartStyle, Element element) {
    // stepped property
    Boolean stepped = XMLUtil.getBooleanAttribute(element, "stepped");
    if ( stepped != null )
      chartStyle.setStepped(stepped);
    // stacked property
    Boolean stacked = XMLUtil.getBooleanAttribute(element, "stacked");
    if ( stacked != null )
      chartStyle.setStacked(stacked);
  }
  
  /**
   * Reads the {@link ChartLabelStyle} properties from an XML element.
   * <ul>
   *   <li>Attribute "paint": color of the label</li>
   *   <li>Element content: the label title</li>
   * </ul>
   * @param element
   * @param style
   */
  private void applyLabelStyleFromXML(Element element, ChartLabelStyle style) {
    // color in attribute
    Color color = XMLUtil.getColorAttribute(element, "paint");
    if ( color != null )
      style.setPaint( color );
    
    // label in element data
    String label = element.getTextTrim();
    style.setLabel( label != null ? label : "" ); 
  }
  
  /**
   * Creates a {@link ChartLabelStyle} from XML element.
   * @param element an element
   */
  public ChartLabelStyle createLabelStyleFromXML(Element element) {
    ChartLabelStyle style = new ChartLabelStyle();
    
    // read the common ChartLabelStyle properties
    applyLabelStyleFromXML(element,style);
    
    return style;
  }
  
  /**
   * Creates a {@link ChartLabelStyle} from XML element.
   * @param element an element
   */
  public ChartAxisStyle createAxisStyleFromXML(Element element) {
    ChartAxisStyle style = new ChartAxisStyle();
    
    // read "visible" property from attribute
    Boolean visible = XMLUtil.getBooleanAttribute(element, "visible");
    if ( visible != null )
      style.setVisible( visible );
    // read "unit" property from attribute
    String unit = XMLUtil.getAttribute(element, "unit");
    if ( unit != null )
      style.setUnitString(unit);

    // read the axis title from child <title>
    Element axisTitleElem = element.getChild("title");
    applyLabelStyleFromXML(axisTitleElem,style);
    // read the axis title angle from child <title>
    Double labelAngle = XMLUtil.getDoubleAttribute(axisTitleElem, "angle");
    if ( labelAngle != null )
      style.setLabelAngle( labelAngle );

    // values format and angle from child element <valueLabels>
    Element valuesElem = element.getChild("valueLabels");
    if ( valuesElem != null ) {
      Double valuesAngle = XMLUtil.getDoubleAttribute(valuesElem, "valuesAngle");
      if ( valuesAngle != null )
        style.setValuesAngle( valuesAngle );
      String formatStr = XMLUtil.getAttribute(valuesElem, "numberFormat");
      if ( formatStr != null )
        style.setValuesFormat(new DecimalFormat(formatStr) );
      else {
        formatStr = XMLUtil.getAttribute(valuesElem, "dateFormat");
        if ( formatStr != null )
          style.setValuesFormat(new SimpleDateFormat(formatStr) );
      }
    }
    
    return style;
  }
  
  /**
   * Creates a {@link ChartRendererStyle} from XML element.
   * @param element an element
   */
  public ChartRendererStyle createRendererStyleFromXML(Element element) {
    ChartRendererStyle style = new ChartRendererStyle();
    
    // read "margin" property from attribute
    Double margin = XMLUtil.getDoubleAttribute(element, "margin");
    if ( margin != null )
      style.setMargin( margin );

    // read series properties
    List<Element> seriesElements = element.getChildren("series");
    for (int i=0; i<seriesElements.size(); i++) {
      Element seriesElement = seriesElements.get(i);
      // series paint
      Color seriesPaint = XMLUtil.getColorAttribute(seriesElement, "paint");
      if ( seriesPaint != null )
        style.setSeriesPaint(i, seriesPaint);

      // item label visibility
      Boolean labelsVisible = XMLUtil.getBooleanAttribute(seriesElement, "itemLabelsVisible");
      if ( labelsVisible != null )
        style.setSeriesItemLabelsVisible(i, labelsVisible);

      // item shapes visibility
      Boolean shapesVisible = XMLUtil.getBooleanAttribute(seriesElement, "shapesVisible");
      if ( shapesVisible != null )
        style.setSeriesShapesVisible(i, shapesVisible);
      
      // legend visibility
      Boolean legendVisible = XMLUtil.getBooleanAttribute(seriesElement, "legendVisible");
      if ( legendVisible != null )
        style.setSeriesLegendVisible(i, legendVisible);
      
      // legend label
      String legendLabel = XMLUtil.getAttribute(seriesElement, "legendLabel");
      if ( legendLabel != null )
        style.setSeriesLegendLabel(i, new ChartLabelStyle(legendLabel));
      
      // legend tooltip
      String legendTooltip = XMLUtil.getAttribute(seriesElement, "legendTooltip");
      if ( legendTooltip != null )
        style.setSeriesLegendTooltip(i, new ChartLabelStyle(legendTooltip));
    }
    return style;
  }

  /**
   * Creates a {@link ChartPlotStyle} from XML element.
   * @param element an element
   */
  public ChartPlotStyle createPlotStyleFromXML(Element element) {
    ChartPlotStyle style = new ChartPlotStyle();

    // Axis origin centration
    style.setCenterOriginSymetrically( XMLUtil.getBooleanAttribute(element, "centerAxisOrigin", false) );
    
    // foreground transparency
    Element foregroundElem = element.getChild("foreground");
    if ( foregroundElem != null ) {
      Float alpha = XMLUtil.getFloatAttribute(foregroundElem, "alpha");
      if ( alpha != null )
        style.setForegroundAlpha(alpha);
    }
    
    // background color and transparency
    Element backegroundElem = element.getChild("background");
    if ( backegroundElem != null ) {
      Float alpha = XMLUtil.getFloatAttribute(backegroundElem, "alpha");
      if ( alpha != null )
        style.setBackgroundAlpha(alpha);
      Color color = XMLUtil.getColorAttribute(backegroundElem, "paint");
      if ( color != null )
        style.setBackgroundPaint(color);
    }
    
    // Insets
    Element insetsElem = element.getChild("insets");
    if ( insetsElem != null ) {
      double top    = XMLUtil.getDoubleAttribute(insetsElem, "top",    0.0);
      double bottom = XMLUtil.getDoubleAttribute(insetsElem, "bottom", 0.0);
      double left   = XMLUtil.getDoubleAttribute(insetsElem, "left",   0.0);
      double right  = XMLUtil.getDoubleAttribute(insetsElem, "right",  0.0);
      style.setInsets( new RectangleInsets(top,left,bottom,right) );
    }
    
    // domain grid line color and visibility
    Element domainGridElem = element.getChild("domainGridline");
    if ( domainGridElem != null ) {
      Boolean visible = XMLUtil.getBooleanAttribute(domainGridElem, "visible");
      if ( visible != null )
        style.setDomainGridlineVisible(visible);
      Color color = XMLUtil.getColorAttribute(domainGridElem, "paint");
      if ( color != null )
        style.setDomainGridlinePaint(color);
    }

    // range grid line color and visibility
    Element rangeGridElem = element.getChild("rangeGridline");
    if ( rangeGridElem != null ) {
      Boolean visible = XMLUtil.getBooleanAttribute(rangeGridElem, "visible");
      if ( visible != null )
        style.setRangeGridlineVisible(visible);
      Color color = XMLUtil.getColorAttribute(rangeGridElem, "paint");
      if ( color != null )
        style.setRangeGridlinePaint(color);
    }
    
    // crosshair lines color and visiblilty
    Element crosshairElem = element.getChild("crosshairLines");
    if ( crosshairElem != null ) {
      Boolean visible = XMLUtil.getBooleanAttribute(crosshairElem, "visible");
      if ( visible != null )
        style.setCrosshairVisible(visible);
      Color color = XMLUtil.getColorAttribute(crosshairElem, "paint");
      if ( color != null )
        style.setCrosshairPaint(color);
    }

    return style;
  }
  
  /**
   * Stores a {@link ChartStyle} to an XML file.
   * @param style           style to store
   * @param rootElementName name for the root element (if {@code null} "ChartType" is 
   *                        used)
   * @param filePath        file to store the style to
   */
  public void writeStyleToFile(E style, String rootElementName, File filePath) throws IOException {
    OutputStreamWriter out = new OutputStreamWriter(
        new FileOutputStream(filePath), "utf-8");
    // Create XML-Document
    Element  rootElem = createElementFromStyle(style, rootElementName);
    Document document = new Document(rootElem); 
    // Write document
    XMLUtil.XML_OUTPUTTER.output(document,out);
    out.flush();
    out.close();

  }

  /**
   * Creates a XML {@link Element} representing the {@link ChartStyle}.
   * @param style           style to create an element for
   * @param rootElementName name for the root element (if {@code null} "ChartStyle" is 
   *                        used)
   */
  public Element createElementFromStyle(E style, String rootElementName) {
    if ( rootElementName == null )
      rootElementName = "ChartStyle";
    Element root = new Element(rootElementName);
    // write common attributes
    root.setAttribute("id", style.getID());
    root.setAttribute("type", style.getType().toString());
    root.setAttribute("legend", String.valueOf( style.hasLegend() ));
    root.setAttribute("tooltips", String.valueOf( style.hasTooltips() ));
    root.setAttribute("urls", String.valueOf( style.hasURLs() ));
    
    // Attribute for plot orientation
    if ( PlotOrientation.HORIZONTAL.equals( style.getOrientation() ) )
      root.setAttribute("orientation", "horizontal" );
    else if ( PlotOrientation.VERTICAL.equals( style.getOrientation() ) )
      root.setAttribute("orientation", "vertical" );
    
    // Element for background
    addChildToElement(root, "background", false, "paint", style.getBackground() );
    
    // Element for border
    addChildToElement(root, "border", false, "visible", style.isBorderVisible() );
    
    // Chart title
    addLabelChildToElement(root, "title", style.getTitleStyle() );

    // Chart description
    addLabelChildToElement(root, "desc", style.getDescStyle() );
    
    // Axis
    if ( style.getAxisCount() > 0 )
      addAxisChildToElement(root, "domainAxis", style.getAxisStyle(0) );
    if ( style.getAxisCount() > 1 )
      addAxisChildToElement(root, "rangeAxis", style.getAxisStyle(1) );
    if ( style.getAxisCount() > 2 )
      addAxisChildToElement(root, "rangeAxis2", style.getAxisStyle(2) );
    
    // Renderers
    for (int i=0; i<style.getRendererCount(); i++)
      addRendererChildToElement(root, style.getRendererStyle(i));
    
    // Plot
    if ( style.getPlotStyle() != null ) {
      Element        plotElem  = addChildToElement(root, "plot");
      ChartPlotStyle plotStyle = style.getPlotStyle();
      
      XMLUtil.setAttribute(plotElem, "centerAxisOrigin", plotStyle.isCenterOriginSymetrically(), false);
      
      addChildToElement(plotElem, "foreground", false,
          "alpha", plotStyle.getForegroundAlpha()
      );
      addChildToElement(plotElem, "background", false,
          "alpha", plotStyle.getBackgroundAlpha(),
          "paint", plotStyle.getBackgroundPaint()
      );
      if ( plotStyle.getInsets() != null )
        addChildToElement(plotElem, "insets", false,
            "top",    plotStyle.getInsets().getTop(),
            "bottom", plotStyle.getInsets().getBottom(),
            "left",   plotStyle.getInsets().getLeft(),
            "right",  plotStyle.getInsets().getRight()
        );
      addChildToElement(plotElem, "domainGridline", false,
          "visible", plotStyle.isDomainGridlineVisible(),
          "paint",   plotStyle.getDomainGridlinePaint()
      );
      addChildToElement(plotElem, "rangeGridline", false,
          "visible", plotStyle.isRangeGridlineVisible(),
          "paint",   plotStyle.getRangeGridlinePaint()
      );
      addChildToElement(plotElem, "crosshairLines", false,
          "visible", plotStyle.isCrosshairVisible(),
          "paint",   plotStyle.getCrosshairPaint()
      );
    }    
    
    // Type specific style properties
    if ( style instanceof BasicChartStyle ) {
      root.setAttribute("stepped", String.valueOf( ((BasicChartStyle)style).isStepped() ));
      root.setAttribute("stacked", String.valueOf( ((BasicChartStyle)style).isStacked() ));
    }
    if ( style instanceof ScatterChartStyle )
      root.setAttribute("regrLineVisible", String.valueOf( ((ScatterChartStyle)style).isRegressionLineVisible() ));
      
    
    return root;
  }
  
  /**
   * Creates an child element with no attributes. If an element is given,
   * the new child is automatically added.
   * @param element    if not {@code null} the new element is added as child to
   *                   this element
   * @param childName  name of the new child
   * @return the new child element
   */
  public static Element addChildToElement(Element element, String childName) {
    return addChildToElement(element,childName,true);
  }
  
  /**
   * Creates an child element and sets some attributes. If an element is given,
   * the new child is automatically added.
   * @param element    if not {@code null} the new element is added as child to
   *                   this element
   * @param childName  name of the new child
   * @param inclNullAttr if {@code false} attributes with value {@code null}
   *                     are not set; if ALL attributes are {@code null} also
   *                     the child is not created!
   * @param attributes attribute/value pairs for the new child element
   * @return the new child element
   */
  public static Element addChildToElement(Element element, String childName, boolean inclNullAttr, Object... attributes) {
    // Create child
    Element child = new Element(childName);
    // Set child attributes
    boolean anyAttrSet = false;
    for (int i=0; i<attributes.length; i+=2) {
      String attrName  = attributes[i].toString();
      Object attrValue = attributes[i+1];
      anyAttrSet |= XMLUtil.setAttribute(child, attrName, attrValue, inclNullAttr);
    }
    // if null attributes should be ignored and NO attribute is 
    // actually set, also the child is not created!
    if ( !anyAttrSet && !inclNullAttr )
      child = null;
    
    // If element is given, the child is added
    if ( element != null && child != null )
      element.addContent(child);

    return child;
  }

  /**
   * Creates an child element for a {@link ChartLabelStyle}. Besides the attributes
   * the label title is set as element content. If an element is given, the new child
   * is automatically added. 
   * @param element    if not {@code null} the new element is added as child to
   *                   this element
   * @param titleChildName  name of the new child
   * @param labelStyle the label style to create the element for
   * @return the new child element; {@code null} if {@code labelStyle} is {@code null}
   */
  private static Element addLabelChildToElement(Element element, String titleChildName, ChartLabelStyle labelStyle) {
    if ( labelStyle == null )
      return null;
    // Create and add child with paint attribute
    Element labelElem = addChildToElement(element, titleChildName);
    XMLUtil.setNotNullAttribute(labelElem, "paint", labelStyle.getPaint());
    // Set title to element content
    labelElem.setText( labelStyle.getLabelTranslation().toOneLine() );

    return labelElem;
  }

  /**
   * Creates an child element for a {@link ChartAxisStyle}.
   * If an element is given, the new child is automatically added. 
   * @param element    if not {@code null} the new element is added as child to
   *                   this element
   * @param axisChildName  name of the new child
   * @param axisStyle the axis style to create the element for
   * @return the new child element; {@code null} if {@code axisStyle} is {@code null}
   */
  private static Element addAxisChildToElement(Element element, String axisChildName, ChartAxisStyle axisStyle) {
    if ( axisStyle == null )
      return null;
    
    // Create and add child with visible and unit attribute
    Element axisElem = addChildToElement(element, axisChildName);
    XMLUtil.setNotNullAttribute(axisElem, "visible", axisStyle.isVisible());
    XMLUtil.setNotNullAttribute(axisElem, "unit", axisStyle.getUnitString());
    
    // Create and add child <title>
    Element titleElem = addLabelChildToElement(axisElem, "title", axisStyle);
    // Add angle attribute to <title> element
    XMLUtil.setNotNullAttribute(titleElem, "angle", axisStyle.getLabelAngle());
    
    // Create and add child <valueLabels>
    Element valueLabelsElem = addChildToElement(axisElem, "valueLabels");
    XMLUtil.setNotNullAttribute(valueLabelsElem, "valueAngle", axisStyle.getValuesAngle());

    Format format = axisStyle.getValuesFormat();
    if ( format instanceof DateFormat )
      valueLabelsElem.setAttribute("dateFormat", format.toString());
    if ( format instanceof NumberFormat )
      valueLabelsElem.setAttribute("numberFormat", format.toString());
    
    return axisElem;
  }

  /**
   * Creates an child element for a {@link ChartRendererStyle}.
   * If an element is given, the new child is automatically added. 
   * @param element    if not {@code null} the new element is added as child to
   *                   this element
   * @param rendererStyle the renderer style to create the element for
   * @return the new child element; if {@code rendererStyle} is {@code null} an empty
   *         element is returned and added (so that the order of the renderers according
   *         to the dataset is preserved)
   */
  private static Element addRendererChildToElement(Element element, ChartRendererStyle rendererStyle) {
    // Create and add child with visible attribute
    Element rendererElem = addChildToElement(element, "renderer"); 
    
    if ( rendererStyle != null ) {
      // Margin attribute
      XMLUtil.setNotNullAttribute(element, "margin", rendererStyle.getMargin());
      // Create <series> child for all series
      for ( int i=0; i<rendererStyle.getSeriesCount(); i++) {
        Element seriesElem = addChildToElement(rendererElem, "series");
        XMLUtil.setNotNullAttribute(seriesElem, "paint", rendererStyle.getSeriesPaint(i));
        XMLUtil.setNotNullAttribute(seriesElem, "itemLabelsVisible", rendererStyle.isSeriesItemLabelsVisible(i));
        XMLUtil.setNotNullAttribute(seriesElem, "shapesVisible", rendererStyle.isSeriesShapesVisible(i));
        XMLUtil.setNotNullAttribute(seriesElem, "legendVisible", rendererStyle.isSeriesLegendVisible(i));
        if ( rendererStyle.getSeriesLegendLabel(i) != null )
          XMLUtil.setNotNullAttribute(seriesElem, "legendLabel", rendererStyle.getSeriesLegendLabel(i).getLabelTranslation());
        if ( rendererStyle.getSeriesLegendTooltip(i) != null )
          XMLUtil.setNotNullAttribute(seriesElem, "legendTooltip", rendererStyle.getSeriesLegendTooltip(i).getLabelTranslation());
      }
    }
    
    return rendererElem;
  }
}
