/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.jfree.chart.style;

import java.awt.image.BufferedImage;
import java.net.URL;

import javax.swing.ImageIcon;

import org.apache.log4j.Logger;

import schmitzm.jfree.JFreeChartUtil;
import schmitzm.lang.ResourceProvider;

/**
 * {@link Enum} representing the style type of a chart (bar, line, area, ...).
 * It also meta-information on these chart types, like translated titles,
 * preview images and information about the data requirements.
 */
public enum ChartType {

	/** Represents all bar chart styles */
	BAR,
	/** Represents all line chart styles */
	LINE,
	/** Represents all area chart styles */
	AREA,
	/** Represents all point chart styles */
	POINT,
	/** Represents all pie chart styles */
	PIE,
	/** Represents all gantt chart styles */
	GANTT,
	/** Represents all timeseries chart styles */
	TIMESERIES,
	/** Represents all spider chart styles */
	SPIDER,
	/** Represents all scatter chart styles */
	SCATTER;

	final static Logger LOGGER = Logger.getLogger(ChartType.class);

	/**
	 * Returns an image of 300x200 pixels that shows a hypothetical chart of the
	 * given type. Never returns <code>null</code>, but rather a default image.
	 * The images are not cached, so if ut's called often, cache it yourself.
	 */
	public ImageIcon getPreviewIcon() {
		final String fileName = getClass().getSimpleName() + "_" + toString();

		final String imageResPath = "resource/images/" + fileName
				+ "_preview.png";

		// LOGGER.debug("Looking for image " + imageResPath);
		final URL imageUrl = JFreeChartUtil.class.getResource(imageResPath);
		
		if (imageUrl == null) {
			LOGGER.warn(imageResPath + "not found!");
			
			// Create a default icon
			final BufferedImage bufferedImage = new BufferedImage(300, 200,
					BufferedImage.TYPE_INT_RGB);
			bufferedImage.getGraphics().drawString("no preview available", 20,
					110); // i8n
			return new ImageIcon(bufferedImage);

		}
		return new ImageIcon(imageUrl);
	}

	/**
	 * Returns an image of 16x16 pixels that shows the basic appearance of this
	 * chart type.
	 */
	public ImageIcon getIcon() {
		final String fileName = getClass().getSimpleName() + "_" + toString();

		final String imageResPath = "resource/images/" + fileName
				+ ".png";

		// LOGGER.debug("Looking for image " + imageResPath);
		final URL imageUrl = JFreeChartUtil.class.getResource(imageResPath);

		if (imageUrl == null) {
			LOGGER.warn(imageResPath + "not found!");

			// Create a default icon
			final BufferedImage bufferedImage = new BufferedImage(16, 16,
					BufferedImage.TYPE_INT_RGB);
			return new ImageIcon(bufferedImage);

		}
		return new ImageIcon(imageUrl);
	}

	/**
	 * Returns a description of this kind of chart. Can be used for tool-tips.
	 * May return <code>null</code> if no localized String found.
	 */
	public String getDescription() {
		final String resource = JFreeChartUtil.R(getClass().getSimpleName()
				+ "_" + toString() + ".Desc");
		if (resource == null
				|| resource.equals(ResourceProvider.MISSING_RESOURCE_STRING))
			return null;
		return resource;
	}

	/**
	 * Returns a localized title of this kind of chart. If no localized string
	 * found, return the Enum.toString()
	 */
	public String getTitle() {
		final String resource = JFreeChartUtil.R(getClass().getSimpleName()
				+ "_" + toString() + ".Title");
		if (resource == null
				|| resource.equals(ResourceProvider.MISSING_RESOURCE_STRING))
			return toString();
		return resource;
	}

	/**
	 * @return Minimal required number of series.
	 */
	public int getMinDimensions() {
		switch (this) {
		case BAR:
		case LINE:
		case AREA:
		case POINT:
		case SCATTER:
			return 2;
		case PIE:
		default:
			return 1;
		}
	}

	/**
	 * @return Maximum allowed number of series. -1 = no limits
	 */
	public int getMaxDimensions() {
		switch (this) {
		case SCATTER:
			return 2;
		case PIE:
			return 1;
		case BAR:
		case LINE:
		case AREA:
		case POINT:
		default:
			return -1;
		}
	}

	/**
	 * Does the {@link ChartType} allow category data for the domain axis?
	 */
	public boolean isCategoryAllowedForDomainAxis() {
		switch (this) {
		case SCATTER:
			return false;
		case BAR:
		case LINE:
		case AREA:
		case POINT:
		default:
			return true;
		}
	}

	/**
	 * Returns a localized title for the objects drawn in this chart. For a
	 * bar-chart it would be "bar", for a line-chart it would be "line"
	 */
	public String getObjectName() {
		final String resource = JFreeChartUtil.R(getClass().getSimpleName()
				+ "_" + toString() + ".Object.Title");
		if (resource == null
				|| resource.equals(ResourceProvider.MISSING_RESOURCE_STRING))
			return getTitle();
		return resource;
	}

};
