/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.jfree.chart.style;

import java.awt.Color;

import org.jfree.chart.JFreeChart;
import org.jfree.chart.renderer.xy.XYItemRenderer;
import org.jfree.data.general.Dataset;
import org.jfree.data.xy.XYDataset;
import org.jfree.data.xy.XYSeriesCollection;

import schmitzm.jfree.JFreeChartUtil;
import schmitzm.lang.LangUtil;

/**
 * This class provides a chart style for scatter chart, which shows the data as
 * points. Besides the points it is possible to show a regression line ("line of best
 * fit") for the data points.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 */
public class ScatterChartStyle extends AbstractChartStyle {
  /** Stores whether a regression line ("line of best fit") for the
   *  data points is shown. */
  protected boolean regrLineVisible = true;

  /**
   * Creates a scatter chart style a regression line shown.
   * @param id   a (unique) ID for the style
   */
  public ScatterChartStyle(String id) {
    super(id, ChartType.SCATTER);
  }

  /**
   * Creates a (deep) clone of this chart style.
   */
  @Override
  public ScatterChartStyle copy() {
    return (ScatterChartStyle) copyTo( new ScatterChartStyle("") );
  }
  
  /**
   * Copies all properties of this chart style to another
   * chart style.
   * @param dest destination object (if {@code null} the copy
   *             is created by {@link #copy()})
   * @return {@code dest} or the new instance
   * @exception IllegalArgumentException if {@code dest} is no
   *            {@link AbstractChartStyle}
   */
  @Override
  public ChartStyle copyTo(ChartStyle dest) {
    // copy the super class properties
    dest = super.copyTo(dest);
    // copy this classes properties
    if ( dest instanceof ScatterChartStyle ) {
      ScatterChartStyle destSCS = (ScatterChartStyle) dest;
      destSCS.setRegressionLineVisible( isRegressionLineVisible() );
    }
    return dest;
  }

  /**
   * Returns whether a regression line is shown.
   */
  public boolean isRegressionLineVisible() {
    return regrLineVisible;
  }
  
  /**
   * Sets whether a regression line is shown.
   */
  public void setRegressionLineVisible(boolean visible) {
    this.regrLineVisible = visible;
  }
  

  /**
   * Creates a scatter chart for the given {@link Dataset} and applies the style.
   * @param dataset the data for the chart
   * @exception UnsupportedOperationException if the style can not be applied
   *            to the given dataset
   */
  @Override
  public JFreeChart applyToDataset(Dataset dataset) {
	  
    if ( !(dataset instanceof XYDataset) )
      throw new UnsupportedOperationException("ScatterChartStyle can only be applied to XYDataset: "+LangUtil.getSimpleClassName(dataset));

    // create default scatter chart
    JFreeChart chart = JFreeChartUtil.createRegressionChart(
        (XYDataset)dataset, "", "", "", isRegressionLineVisible()
    );
    
    // apply style
    applyToChart(chart);
    return chart;
  }

  /**
   * Applies the style to an existing chart.
   * @param chart chart the style is applied to
   * @exception UnsupportedOperationException if the style can not be applied
   *            to the given dataset
   */
  public void applyToChart(JFreeChart chart) {
    if ( chart.getXYPlot() == null )
      throw new UnsupportedOperationException("ScatterChartStyle can only be applied to XYPlot: "+LangUtil.getSimpleClassName(chart.getPlot()));

    // Create regression line if not already existing
    if ( isRegressionLineVisible() && chart.getXYPlot().getRendererCount() < 2 ) {
      // Create sample data for regression line for primary dataset (primary series)
      // -> plot function directly is not yet available in JFreeChart
      XYSeriesCollection dataset         = (XYSeriesCollection)chart.getXYPlot().getDataset();
      String             regressionTitle = getTitleStyle().getLabel() + " Regression line"; //i8n
      XYDataset          regressionData  = JFreeChartUtil.createRegressionLineDataset(
                                                 dataset, 0, regressionTitle, 2);
      // Add regression line to plot (with default color)
      if ( regressionData != null )
        JFreeChartUtil.addRegressionLineToPlot(chart.getXYPlot(), regressionData, Color.blue);
    }

    // apply "normal" style properties
    super.applyToChart(chart);

    // apply visiblity of regression line
    XYItemRenderer regrLineRenderer = chart.getXYPlot().getRenderer(1);
    if ( regrLineRenderer != null )
      regrLineRenderer.setSeriesVisible(0, isRegressionLineVisible() );
  }

}
