/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.jfree.feature;

import java.util.HashMap;
import java.util.Map;

import org.geotools.feature.FeatureCollection;
import org.jfree.data.general.Dataset;
import org.opengis.feature.simple.SimpleFeature;

/**
 * This class implements the mapping between {@link Dataset} items (points in chart)
 * and the corresponding {@link SimpleFeature Features}.<br>
 * <br>
 * {@code D} = the dataset type the model handles<br>
 * {@code E} = the data type the series are identified with in the dataset<br>
 * {@code F} = the data type the data items are identified with in each series
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 */
public abstract class Feature2DatasetMapping<D extends Dataset, E, F> {
  /** Holds the {@link FeatureCollection} for the mapping */
  protected FeatureCollection featureCollection = null;

  /** Holds the {@link Dataset} for the mapping */
  protected D dataset = null;

  /** Holds the data items ({@code Map<SeriesKey,ItemID>})
   *  for each feature ID (String). */
  protected Map<String,Map<E,F>> fid2did = new HashMap<String, Map<E,F>>();
  /** Holds for each series the feature ID (String) for each data item. */
  protected Map<E,Map<F,String>> did2fid = new HashMap<E,Map<F,String>>();

  
  /**
   * Creates a mapping. 
   * @param fc a {@link FeatureCollection} (can be {@code null}) 
   */
  public Feature2DatasetMapping(FeatureCollection fc, D dataset) {
    this.featureCollection = fc;
    this.dataset           = dataset;
  }

  /**
   * Returns the {@link FeatureCollection} the mapping is provided for.
   */
  public FeatureCollection getFeatureCollection() {
    return featureCollection;
  }

  /**
   * Returns the {@link Dataset} the mapping is provided for.
   */
  public D getDataset() {
    return dataset;
  }

  /**
   * Returns the key of the primary series.
   */
  public abstract E getPrimarySeriesKey();

  /**
   * Sets the mapping between a feature ID and a dataset item IN THE
   * PRIMARY SERIES.
   * @param featureID    feature ID
   * @param itemID       defines the dataset item
   */
  public void setMapping(String featureID, F itemID) {
    setMapping(featureID, null, itemID);
  }

  /**
   * Sets the mapping between a feature ID and a dataset item.
   * @param featureID    feature ID
   * @param seriesKey    defines the series in the dataset (if
   *                     {@code null} the primary series is used)
   * @param itemID       defines the dataset item in the series
   */
  public void setMapping(String featureID, E seriesKey, F itemID) {
    if ( seriesKey == null )
      seriesKey = getPrimarySeriesKey();
    // Mapping FeatureID --> (Series/Index)
    Map<E, F> itemIDs = fid2did.get(featureID);
    if ( itemIDs == null ) {
      itemIDs = new HashMap<E, F>();
      fid2did.put(featureID, itemIDs);
    }
    itemIDs.put(seriesKey, itemID);
    
    // Mapping (Series/Index) --> FeatureID
    Map<F,String> featureIDs = did2fid.get(seriesKey);
    if ( featureIDs == null ) {
      featureIDs = new HashMap<F, String>();
      did2fid.put(seriesKey, featureIDs);
    }
    featureIDs.put(itemID, featureID);
  }
  
  /**
   * Removes the mapping of a feature ID.
   * @param fID feature ID
   */
  public void removeMappingForFeatureID(String fID) {
    // remove mapping FeatureID --> (Series/Index)
    Map<E, F> itemIDs = fid2did.remove(fID);

    // remove mapping (Series/Index) --> FeatureID
    // for all item IDs referencing to the feature ID
    for (E seriesKey : itemIDs.keySet()) {
      Map<F, String> featureIDs = did2fid.get(seriesKey);
      for (F itemID : itemIDs.values())
        featureIDs.remove(itemID);
    }
  }
  
  /**
   * Removes the mapping for a dataset item IN THE
   * PRIMARY SERIES.
   * @param itemID    dataset item
   */
  public void removeMappingForDataItem(F itemID) {
    removeMappingForDataItem(null,itemID);
  }
  
  /**
   * Removes the mapping for a dataset item.
   * @param seriesKey defines the series in the dataset (if
   *                  {@code null} the primary series is used)
   * @param itemID    defines the dataset item in the series
   */
  public void removeMappingForDataItem(E seriesKey, F itemID) {
    if ( seriesKey == null )
      seriesKey = getPrimarySeriesKey();
    // remove mapping (Series/Index) --> FeatureID 
    Map<F, String> featureIDs = did2fid.get(seriesKey);
    String fID = null;
    if ( featureIDs != null )
      fID = featureIDs.remove(itemID);
    
    // remove mapping FeatureID --> (Series/Index)
    if ( fID != null ) {
      Map<E, F> itemIDs = fid2did.remove(fID);
      itemIDs.remove(seriesKey);
    }  

  }

  /**
   * Returns the data items for a given feature IN THE
   * PRIMARY SERIES.
   * @param fID a feature ID
   * @return {@code null} if no data item is registered for the feature ID
   */
  public F getDataID(String fID) {
    return getDataID(null,fID);
  }

  /**
   * Returns the data items for a given feature ID.
   * @param seriesKey  defines the series in the dataset
   * @param fID a feature ID
   * @return {@code null} if no data item is registered for the feature ID
   */
  public F getDataID(E seriesKey, String fID) {
    if ( seriesKey == null )
      seriesKey = getPrimarySeriesKey();
    Map<E, F> seriesIDs = getDataIDs(fID);
    if ( seriesIDs != null )
      return seriesIDs.get(seriesKey);
    return null;
    
  }

  /**
   * Returns the feature ID for a given data item IN THE
   * PRIMARY SERIES.
   * @param itemID a data item
   * @return {@code null} if no feature ID is registered for the data item
   */
  public String getFeatureID(F itemID) {
    return getFeatureID(null,itemID);
  }

  /**
   * Returns the data items for a given feature ID.
   * @param fID a feature ID
   * @return {@code null} if no data item is registered for the feature ID
   */
  public Map<E,F> getDataIDs(String fID) {
    return fid2did.get(fID);
  }
  
  /**
   * Returns the feature ID for a given data item.
   * @param seriesKey defines the series in the dataset
   * @param itemID    defines the dataset item in the series (if
   *                  {@code null} the primary series is used)
   * @return {@code null} if no feature ID is registered for the data item
   */
  public String getFeatureID(E seriesKey, F itemID) {
    if ( seriesKey == null )
      seriesKey = getPrimarySeriesKey();
    Map<F,String> featureIDs = did2fid.get(seriesKey);
    if ( featureIDs != null )
      return featureIDs.get(itemID); 
    return null;
  }
  
}
