/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.jfree.feature;

import java.util.HashSet;
import java.util.Set;

import org.jfree.data.category.CategoryDataset;
import org.jfree.data.general.DatasetGroup;

import schmitzm.jfree.chart.selection.CategoryDatasetSelectionModel;
import schmitzm.lang.LangUtil;

/**
 * This class extends the {@link CategoryDatasetSelectionModel} with the
 * {@link FeatureDatasetSelectionModel}, so that selection operations
 * can directly applied on feature IDs.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 */
public class FeatureCategoryDatasetSelectionModel extends CategoryDatasetSelectionModel implements FeatureDatasetSelectionModel<CategoryDataset,Comparable<?>,Comparable<?>> {
  /** Holds the mapping between features and the corresponding
   *  column/row keys. */
  protected Feature2CategoryDatasetMapping mapping = null;
  
  /**
   * Creates a new selection model.
   * @param dataset a dataset providing {@link FeatureDatasetMetaData}
   *                as dataset group containing a {@link Feature2CategoryDatasetMapping}
   */
  public FeatureCategoryDatasetSelectionModel(CategoryDataset dataset) {
    super(dataset);
    // DatasetGroup must provide meta data (containing the SimpleFeature<->DataItem mapping)
    DatasetGroup group = dataset.getGroup();
    if ( group == null || !(group instanceof FeatureDatasetMetaData) )
      throw new UnsupportedOperationException("Dataset must provide FeatureDatasetMetaData as DatasetGroup: "+LangUtil.getSimpleClassName(group));
    FeatureDatasetMetaData metadata = (FeatureDatasetMetaData)group;
    // Meta data must provide Feature2SeriesDatasetMapping
    Feature2DatasetMapping<?,?,?> mapping = metadata.getMapping();
    if ( mapping == null || !(mapping instanceof Feature2CategoryDatasetMapping) )
      throw new UnsupportedOperationException("Mapping (provided by FeatureDatasetMetaData) must be a Feature2CategoryDatasetMapping: "+LangUtil.getSimpleClassName(mapping));
    this.mapping = (Feature2CategoryDatasetMapping)mapping;
  }

  /**
   * Checks if the data item of a feature is selected in any series.
   * @param fID  feature ID
   */
  @Override
  public boolean isFeatureSelected(String fID) {
    for (Object seriesKey : dataset.getRowKeys() )
      if ( isFeatureSelected(fID,(Comparable<?>)seriesKey) )
        return true;
    return false;
    
  }
  
  /**
   * Checks if the data item of a feature is selected.
   * @param fID  feature ID
   * @param seriesKey series (row) to check for the feature ID
   */
  public boolean isFeatureSelected(String fID, Comparable<?> seriesKey) {
    Comparable<?> dataItem = mapping.getDataID(seriesKey,fID);
    return dataItem != null && isItemSelected(seriesKey, dataItem);
  }

  /**
   * Sets the selection state for a feature in all dataset series.
   * @param fID  feature ID
   * @param selected  flag to (de)select the item
   */
  @Override
  public void setItemSelected(String fID, boolean selected) {
    for (Object seriesKey : dataset.getRowKeys())
      setItemSelected(fID, (Comparable<?>)seriesKey, selected);
  }
  
  /**
   * Sets whether the data item of a feature is selected or not.
   * @param fID  feature ID
   * @param seriesKey series (row) to set the selection in
   * @param selected  flag to (de)select the item
   */
  public void setItemSelected(String fID, Comparable<?> seriesKey, boolean selected) {
    Comparable<?> dataItem = mapping.getDataID(seriesKey,fID);
    if ( dataItem == null )
      LOGGER.warn("SimpleFeature unknown in Feature2DatasetItemDatasetGroup: "+fID);
    else
      setItemSelected(seriesKey, dataItem, selected);
  }

  /**
   * Changes the selection of the data item of a feature in all
   * series.
   * @param fID  feature ID
   * @return {@code true} if at least one item changed to selected 
   */
  @Override
  public boolean changeItemSelection(String fID) {
    boolean anyChangedToTrue = false;
    for (Object seriesKey : dataset.getRowKeys())
      anyChangedToTrue |= changeItemSelection(fID, (Comparable<?>)seriesKey);
    return anyChangedToTrue;
  }

  /**
   * Changes the selection of the data item of a feature.
   * @param fID  feature ID
   * @param seriesKey series to change the selection in
   * @return {@code true} if the item changed to selected 
   */
  public boolean changeItemSelection(String fID, Comparable<?> seriesKey) {
    Comparable<?> dataItem = mapping.getDataID(seriesKey,fID);
    return changeItemSelection(seriesKey, dataItem);
  }

  /**
   * Returns the IDs of the selected features (in any series/row).
   */
  public Set<String> getSelectedFeatures() {
    Set<String> selectedFIDs = new HashSet<String>();
    for (Comparable<?> seriesKey : selectedSeriesIdx.keySet()) 
      for (Comparable<?> selIdx : selectedSeriesIdx.get(seriesKey)) {
        String selFID = mapping.getFeatureID(seriesKey,selIdx);
        selectedFIDs.add(selFID);
      }
    
    return selectedFIDs;
  }

}
