/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/

package schmitzm.jfree.feature;

import org.apache.log4j.Logger;
import org.jfree.chart.JFreeChart;
import org.jfree.data.category.CategoryDataset;
import org.jfree.data.general.Dataset;
import org.jfree.data.xy.XYDataset;

import schmitzm.jfree.chart.SelectableChartPanel;
import schmitzm.jfree.chart.renderer.PointSelectionCategoryRenderer;
import schmitzm.jfree.chart.renderer.PointSelectionXYRenderer;
import schmitzm.jfree.chart.renderer.SelectionRenderer;
import schmitzm.jfree.chart.selection.DatasetSelectionModel;
import schmitzm.lang.LangUtil;

/**
 * This panel is a special {@link SelectableChartPanel} which uses
 * special {@link DatasetSelectionModel DatasetSelectionModels} to
 * connect the chart data with features.
 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a>
 * @version 1.0
 */
public class FeatureChartPanel extends SelectableChartPanel {
  /** Used for log and debug messages. */
  private static Logger LOGGER = LangUtil.createLogger(FeatureChartPanel.class);

  /**
   * Creates a new chart panel.
   */
  public FeatureChartPanel(JFreeChart chart) {
      super(chart);
  }

  /**
   * Creates a new chart panel.
   */
  public FeatureChartPanel(JFreeChart chart, boolean useBuffer) {
      super(chart, useBuffer);
  }

  /**
   * Creates a new chart panel.
   */
  public FeatureChartPanel(JFreeChart chart, boolean properties,
          boolean save, boolean print, boolean zoom, boolean tooltips) {
      super(chart, properties, save, print, zoom, tooltips);
  }

  /**
   * Creates a new chart panel.
   */
  public FeatureChartPanel(JFreeChart chart, int width, int height,
          int minimumDrawWidth, int minimumDrawHeight, int maximumDrawWidth,
          int maximumDrawHeight, boolean useBuffer, boolean properties,
          boolean save, boolean print, boolean zoom, boolean tooltips) {
      super(chart, width, height, minimumDrawWidth, minimumDrawHeight,
              maximumDrawWidth, maximumDrawHeight, useBuffer, properties,
              save, print, zoom, tooltips);
  }

  /**
   * Creates an adequate selection renderer for a dataset. If provided this
   * method returns a renderer which uses a {@link FeatureDatasetSelectionModel}
   * for the dataset. Otherwise the super method is used.
   * @param dataset a dataset
   */
  protected SelectionRenderer<?> createSelectionRenderer(Dataset dataset) {
    // check whether dataset provides a Feature2DatasetMapping
    // if not: create a "normal" SelectionRenderer by super-class
    //         method (without "SimpleFeature-Synchronizing" functionality)
    if ( dataset.getGroup() instanceof FeatureDatasetMetaData &&
         ((FeatureDatasetMetaData)dataset.getGroup()).getMapping() != null ) {
    
      if ( dataset instanceof XYDataset )
        return new PointSelectionXYRenderer() {
          @Override
          public FeatureSeriesDatasetSelectionModel createSelectionModel(Dataset dataset) {
            if ( !(dataset instanceof XYDataset) )
              throw new IllegalArgumentException(LangUtil.getSimpleClassName(this)+".createSelectionModel(.) can only be applied for XYDataset: "+LangUtil.getSimpleClassName(dataset));
            return new FeatureSeriesDatasetSelectionModel((XYDataset)dataset);
          }
        };

        if ( dataset instanceof CategoryDataset )
          return new PointSelectionCategoryRenderer() {
            @Override
            public FeatureCategoryDatasetSelectionModel createSelectionModel(Dataset dataset) {
              if ( !(dataset instanceof CategoryDataset) )
                throw new IllegalArgumentException(LangUtil.getSimpleClassName(this)+".createSelectionModel(.) can only be applied for CategoryDataset: "+LangUtil.getSimpleClassName(dataset));
              return new FeatureCategoryDatasetSelectionModel((CategoryDataset)dataset);
            }
          };
    }
    LOGGER.warn("No Feature2DatasetMapping provided for dataset: "+LangUtil.getSimpleClassName(dataset));

    return super.createSelectionRenderer(dataset);
  }
}
