/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.jfree.feature.style;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.geotools.feature.FeatureCollection;
import org.jfree.chart.JFreeChart;
import org.jfree.data.category.CategoryDataset;
import org.jfree.data.general.Dataset;
import org.jfree.data.xy.XYDataset;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.simple.SimpleFeatureType;

import schmitzm.jfree.chart.style.AbstractChartStyle;
import schmitzm.jfree.chart.style.ChartStyle;
import schmitzm.lang.LangUtil;

/**
 * This interface extends the chart style with several functionalities
 * used to define a chart on a {@link FeatureCollection}.
 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a>
 * @version 1.0
 */
public interface FeatureChartStyle extends ChartStyle {
  /**
   * Returns the maximum number of feature attributes that can be
   * specified by this style.
   * @return -1 if there is no limit for range attributes
   */
  public int getMaxAttributeCount();

  /**
   * Returns the number of feature attributes defined in this style.
   */
  public int getAttributeCount();

  /**
   * Removes all style informations about an attribute and
   * reorganizes the attribute indexes so there is an continuous 
   * order. 
   * @param idx an attribute index
   */
  public void removeAttribute(int idx);

  /**
   * Returns the name of a feature attribute needed to create a
   * chart for this style.
   * @param idx attribute index (0=domain; 1=1st range series;
   *            2=2nd range series; ...)
   */
  public String getAttributeName(int idx);
  
  /**
   * Sets the name of a feature attribute needed to create a
   * chart for this style.
   * @param idx attribute index (0=domain; 1=1st range series;
   *            2=2nd range series; ...)
   * @param attrName feature attribute name 
   */
  public void setAttributeName(int idx, String attrName);
  
  /**
   * Creates a chart according to the given 
   * @param fc  a feature collection
   * @exception UnsupportedOperationException if the style can not be applied
   *            to the {@link FeatureCollection} (e.g. the {@link FeatureCollection}
   *            does not provide the required attributes)
   */
  public JFreeChart applyToFeatureCollection(FeatureCollection<SimpleFeatureType, SimpleFeature> fc);
  
  /**
   * Sets whether the features are sorted according to the
   * domain attribute (before creating a {@link Dataset}).
   */
  public void setSortDomainAxis(boolean sort);

  /**
   * Returns whether the features are sorted according to the
   * domain attribute (before creating a {@link Dataset}).
   * @return {@code false} as default
   */
  public boolean isSortDomainAxis();

  /** 
   * Sets whether a {@link CategoryDataset} is forced for
   * a numeric domain attribute. The default is to create {@link XYDataset}
   * for a numeric and {@link CategoryDataset} for a non-numeric domain
   * attribute.
   */
  public void setForceCategories(boolean forceCategories);

  /** 
   * Returns whether a {@link CategoryDataset} is forced for
   * a numeric domain attribute. The default is to create {@link XYDataset}
   * for a numeric and {@link CategoryDataset} for a non-numeric domain
   * attribute.
   * @return {@code false} as default
   */
  public boolean isForceCategories();
  
  /** 
   * Sets whether the attribute data is normalized for an
   * attribute (before creating a {@link Dataset}).
   * @param idx attribute index (0=domain; 1=1st range series;
   *            2=2nd range series; ...)
   * @param normalize indicates the normalize property
   */  
  public void setAttributeNormalized(int idx, Boolean normalize);
	
  /** 
   * Returns whether the attribute data is normalized for an
   * attribute (before creating a {@link Dataset}).
   * @param idx attribute index (0=domain; 1=1st range series;
   *            2=2nd range series; ...)
   * @param normalize indicates the normalize property
   * @return {@code false} as default
   */  
  public boolean isAttributeNormalized(int idx);
  
  /**
   * Sets the values, which are interpreted as "No Data". 
   * @param idx attribute index the "No Data" values are set for
   * @param noDataValues the "No Data" values
   */
  public void setNoDataValues(int idx, Set<Object> noDataValues);
  
  /**
   * Returns the values, which are interpreted as "No Data". 
   * @param idx attribute index the "No Data" values are returned for
   */
  public Set<Object> getNoDataValues(int idx);
  
  /**
   * Sets a value, which is interpreted as "No Data". 
   * @param idx attribute index the "No Data" value is set for
   * @param noDataValue the "No Data" value
   */
  public void addNoDataValue(int idx, Object noDataValue);

  /**
   * Removes a "No Data" value for an attribute. 
   * @param idx attribute index the "No Data" value is removed for
   * @param noDataValue the "No Data" value to remove
   * @return {@code false} if the value was not an "No Data" value 
   */
  public boolean removeNoDataValue(int idx, Object noDataValue);
  
  /**
   * Checks whether the given value is one of the "No data" values for
   * the attribute.
   * @param idx attribute index the "No Data" value is checked for
   * @param value an attribute value
   */
  public boolean isNoDataValue(int idx, Object value);

  /**
   * Checks whether the given value is one of the "No data" values for
   * the attribute. In this case this method returns {@code null}, otherwise
   * the value itself.
   * @param idx attribute index the "No Data" value is checked for
   * @param value an attribute value
   * @return {@code null} if the given value is one of the "No data" values 
   */
  public <T> T filterNoDataValue(int idx, T value);

  /**
   * This class defines a dummy implementation of {@link FeatureChartStyle} just
   * to maintain the properties of the interface {@link FeatureChartStyle}, so
   * sub classes of {@link FeatureChartStyle} which usually are derived from
   * a normal {@link ChartStyle} implementation must not implement the
   * {@link FeatureChartStyle} maintenance each. Instead they can create an
   * instance of this dummy an pipe their method implementations to the
   * dummy!<br>
   * <br>
   * The {@link #applyToFeatureCollection(FeatureCollection)} and 
   * {@link #applyToDataset(Dataset)} methods are not implemented by the dummy,
   * but throw an exception instead!! 
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
   */
  public static class Dummy extends AbstractChartStyle implements FeatureChartStyle {
    /** Indicates whether the features are sorted according to the
     *  domain attribute (before creating a {@link Dataset}). */
    protected boolean sortDomainAxis;
    /** Indicates whether a {@link CategoryDataset} is forced for
     *  a numeric domain attribute. The default is to create {@link XYDataset}
     *  for a numeric and {@link CategoryDataset} for a non-numeric domain
     *  attribute. */
    protected boolean forceCategories;

    /** Holds the attributes needed to specify the chart data
     *  from feature collection (0 = attribute for domain axis; others assigned
     *  to the range axis as series). */
    protected Map<Integer,String> attrNames = new HashMap<Integer,String>();
    
    /** Indicates for each attribute whether the attribute data is normalized
     *  (before creating a {@link Dataset}). */
    protected Map<Integer,Boolean> normalizeAttr = new HashMap<Integer,Boolean>();
    
    /** Holds the maximum number of attributes the style can be defined
     *  defined for (-1 = no limit). */
    protected int maxAttrCount = -1;

    /** Holds the number of attributes the style is defined for. */
    protected int maxAttrIdx = 0;
    
    /** Holds the "No Data" values for each attribute. */
    protected Map<Integer,Set<Object>> noDataValues = new HashMap<Integer, Set<Object>>();

    /**
     * Creates a new dummy. This constructor is protected so that only derived
     * classes can instantiate a dummy! There is no limit for the number
     * of attributes which can be defined by the style.
     * @param id a (unique) ID for the style
     */
    protected Dummy(String id) {
      this(id,-1);
    }

    /**
     * Creates a new dummy. This constructor is protected so that only derived
     * classes can instantiate a dummy!
     * @param id  a (unique) ID for the style
     * @param maxAttrCount attribute count needed to specify the chart data
     *                     from feature collection
     */
    protected Dummy(String id, int maxAttrCount) {
      super(id);
      this.maxAttrCount = maxAttrCount;
    }

    /**
     * Updates the local variable for the number of attributes
     * the maximum index (key) of the several hash maps. 
     */
    private void updateAttributeCount() {
      // attribute names define the number of attributes, also
      // if there are "more" normalize constraints specified
      maxAttrIdx = LangUtil.max( attrNames.keySet(),
                                 normalizeAttr.keySet(),
                                 noDataValues.keySet() );
      
      if ( maxAttrCount > 0 && maxAttrIdx >= maxAttrCount )
        throw new IllegalArgumentException("Only "+maxAttrCount+" attributes can be specified for "+LangUtil.getSimpleClassName(this));
    }
    
    /**
     * Removes all style informations about an attribute and
     * reorganizes the attribute indexes so there is an continuous 
     * order. 
     * @param idx an attribute
     */
    @Override
    public void removeAttribute(int idx) {
      // clear sets and maps of the removed attribute
      Set<Object> noDataValues = getNoDataValues(idx);
      if ( noDataValues != null )
        noDataValues.clear();
      
      // remove the attribute by move the "greater" attributes
      // one position "forward"
      int maxIdx = getAttributeCount()-1;
      for (int i=idx; i<maxIdx; i++) {
        this.attrNames.put(i, this.attrNames.get(i+1));
        this.normalizeAttr.put(i, this.normalizeAttr.get(i+1));
        this.noDataValues.put(i, this.noDataValues.get(i+1));
      }
      
      // delete the last attribute, because now is is stored
      // one position forward
      this.attrNames.remove(maxIdx);
      this.normalizeAttr.remove(maxIdx);
      this.noDataValues.remove(maxIdx);
      
      updateAttributeCount();
    }
    
    /**
     * Creates a (deep) clone of this style. The properties
     * of the super class ({@link AbstractChartStyle}) are <b>ignored</b>
     * because they are unused for the dummy.
     */
    @Override
    public AbstractChartStyle copy() {
      return (AbstractChartStyle)copyTo( (FeatureChartStyle)new Dummy("") );
    }
    
    /**
     * Copies all properties of this style to another one. The properties
     * of the super class ({@link AbstractChartStyle}) are <b>ignored</b>
     * because they are unused for the dummy.
     * @param dest destination object (if {@code null} the copy
     *             is created by {@link #copy()})
     * @return {@code dest} or the new instance
     */
    @Override
    public ChartStyle copyTo(ChartStyle dest) {
      // !! do NOT copy the super class properties  !!
      // !! copy only this classes properties       !!
      // !! Reason: the dummy does not maintain the !!
      // !!         super class properties and we   !!
      //            do not want to overwrite them   !!
      if ( dest instanceof FeatureChartStyle ) {
        FeatureChartStyle destFCS = (FeatureChartStyle) dest;
        destFCS.setSortDomainAxis( isSortDomainAxis() );
        destFCS.setForceCategories( isForceCategories() );

        // Clear the attribute names and attribute normalization
        int max = destFCS.getMaxAttributeCount();
        for (int i=0; i<max; i++) {
          destFCS.setAttributeName(i, null);
          destFCS.setAttributeNormalized(i, null);
          
          // Added by SK. In this special case we don't 
          destFCS.setNoDataValues(i, getNoDataValues(i));
        }
        // Copy attribute names and normalization
        for (Integer idx : attrNames.keySet() )
          if ( getAttributeName(idx) != null )
            destFCS.setAttributeName(idx, getAttributeName(idx));
        for (Integer idx : normalizeAttr.keySet() )
          if ( isAttributeNormalized(idx) )
            destFCS.setAttributeNormalized(idx, isAttributeNormalized(idx));
        
      }
      
      return dest;
    }

    /**
     * Returns the maximum number of feature attributes that can be
     * specified by this style.
     * @return -1 if there is no limit for range attributes
     */
    @Override
    public int getMaxAttributeCount() {
      return maxAttrCount;
    }

    /**
     * Returns the number of feature attributes defined in this style.
     */
    @Override
    public int getAttributeCount() {
      return maxAttrIdx+1;
    }

    /**
     * Returns the name of a feature attribute needed to create a
     * chart for this style.
     * @param idx attribute index (0=domain; 1=1st range series;
     *            2=2nd range series; ...)
     */
    @Override
   public String getAttributeName(int idx) {
      return attrNames.get(idx);
    }

    /**
     * Sets the name of a feature attribute needed to create a
     * chart for this style.
     * @param idx attribute index (0=domain; 1=1st range series;
     *            2=2nd range series; ...)
     * @param attrName feature attribute name 
     */
    @Override
    public void setAttributeName(int idx, String attrName) {
      attrNames.put(idx,attrName);
      updateAttributeCount();
    }


    /**
     * Sets whether the features are sorted according to the
     * domain attribute (before creating a {@link Dataset}).
     */
    @Override
  	public void setSortDomainAxis(boolean sortDomainAxis){
		this.sortDomainAxis = sortDomainAxis;
  	}
  	
    /**
     * Returns whether the features are sorted according to the
     * domain attribute (before creating a {@link Dataset}).
     * @return {@code false} as default
     */
    @Override
    public boolean isSortDomainAxis() {
		return sortDomainAxis;
	}
    
    /** 
     * Sets whether a {@link CategoryDataset} is forced for
     * a numeric domain attribute. The default is to create {@link XYDataset}
     * for a numeric and {@link CategoryDataset} for a non-numeric domain
     * attribute.
     */
    @Override
  	public void setForceCategories(boolean forceCategories){
		this.forceCategories = forceCategories;
  	}
  	
    /** 
     * Returns whether a {@link CategoryDataset} is forced for
     * a numeric domain attribute. The default is to create {@link XYDataset}
     * for a numeric and {@link CategoryDataset} for a non-numeric domain
     * attribute.
     * @return {@code false} as default
     */
    @Override
    public boolean isForceCategories(){
    	return forceCategories;
    }

    /** 
     * Sets whether the attribute data is normalized for an
     * attribute (before creating a {@link Dataset}).
  	 * @param idx attribute index (0=domain; 1=1st range series;
  	 *            2=2nd range series; ...)
  	 * @param normalize indicates the normalize property
  	 */  
    @Override
  	public void setAttributeNormalized(int idx, Boolean normalize){
      normalizeAttr.put(idx, normalize);
      updateAttributeCount();
  	}

    /** 
     * Returns whether the attribute data is normalized for an
     * attribute (before creating a {@link Dataset}).
     * @param idx attribute index (0=domain; 1=1st range series;
     *            2=2nd range series; ...)
     * @return {@code false} as default
     */  
  	@Override
	public boolean isAttributeNormalized(int idx) {
	  Boolean normalize = normalizeAttr.get(idx);
	  return normalize != null && normalize;
	}

    /**
     * Sets the values, which are interpreted as "No Data". 
     * @param idx attribute index the "No Data" values are set for
     * @param noDataValues the "No Data" values
     */
    public void setNoDataValues(int idx, Set<Object> noDataValues) {
      this.noDataValues.put(idx, noDataValues);
      updateAttributeCount();
    }
    
    /**
     * Returns the values, which are interpreted as "No Data". 
     * @param idx attribute index the "No Data" values are returned for
     */
    public Set<Object> getNoDataValues(int idx) {
      return this.noDataValues.get(idx);
    }
    
    /**
     * Sets a value, which is interpreted as "No Data". 
     * @param idx attribute index the "No Data" value is set for
     * @param noDataValue the "No Data" value
     */
    public void addNoDataValue(int idx, Object noDataValue) {
      Set<Object> noDataValues = getNoDataValues(idx);
      if ( noDataValues == null ) {
        noDataValues = new HashSet<Object>();
        setNoDataValues(idx, noDataValues);
      }
      noDataValues.add(noDataValue);
    }

    /**
     * Removes a "No Data" value for an attribute. 
     * @param idx attribute index the "No Data" value is removed for
     * @param noDataValue the "No Data" value to remove
     * @return {@code false} if the value was not an "No Data" value 
     */
    public boolean removeNoDataValue(int idx, Object noDataValue) {
      Set<Object> noDataValues = getNoDataValues(idx);
      if ( noDataValues == null )
        return false;
      return noDataValues.remove(noDataValue);
    }

    /**
     * Checks whether the given value is one of the "No data" values for
     * the attribute.
     * @param idx attribute index the "No Data" value is checked for
     * @param value an attribute value
     */
    public boolean isNoDataValue(int idx, Object value) {
      Set<Object> noDataValues = getNoDataValues(idx);
      if ( noDataValues == null )
        return false;
      return noDataValues.contains(value);
    }

    /**
     * Checks whether the given value is one of the "No data" values for
     * the attribute. In this case this method returns {@code null}, otherwise
     * the value itself.
     * @param idx attribute index the "No Data" value is checked for
     * @param value an attribute value
     * @return {@code null} if the given value is one of the "No data" values 
     */
    public <T> T filterNoDataValue(int idx, T value) {
      if ( isNoDataValue(idx, value) )
        return null;
      return value;
    }

    /**
     * Does nothing, but always throws a {@link UnsupportedOperationException},
     * because the dummy can not provide this functionality.
     */
    @Override
    public JFreeChart applyToDataset(Dataset dataset) {
      throw new UnsupportedOperationException("FeatureChartStyle.Dummy does not implement applyToDataset(..)");
    }

    /**
     * Does nothing, but always throws a {@link UnsupportedOperationException},
     * because the dummy can not provide this functionality.
     */
    @Override
    public JFreeChart applyToFeatureCollection(FeatureCollection<SimpleFeatureType, SimpleFeature> fc) {
      throw new UnsupportedOperationException("FeatureChartStyle.Dummy does not implement applyToFeatureCollection(..)");
    }

  }
  
}

