/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.jfree.feature.style;

import java.util.List;

import org.geotools.feature.FeatureCollection;
import org.jdom.Element;

import schmitzm.jfree.chart.style.ChartStyle;
import schmitzm.jfree.chart.style.ChartStyleXMLFactory;
import schmitzm.jfree.chart.style.ChartType;
import schmitzm.xml.XMLUtil;

/**
 * This class defines a factory to create a chart style from XML
 * specialized for {@link FeatureCollection}.
 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a>
 * @version 1.0
 */
public class FeatureChartStyleXMLFactory extends ChartStyleXMLFactory<FeatureChartStyle> {

  /**
   * Applies the chart style definition from XML (<b>except the type and id!!</b>)
   * to an existing {@link ChartStyle} object.
   * @param chartStyle an existing chart style ({@code null} not permitted!)
   * @param element    element to read the properties from
   */
  @Override
  public void applyStyleFromXML(FeatureChartStyle chartStyle, Element element) {
    // Apply the normal style properties
    super.applyStyleFromXML(chartStyle, element);
    
    // Attribute definitions in <attributes> element
    Element featureElement = element.getChild("featureAttributes");
    if ( featureElement == null )
      throw new UnsupportedOperationException("<featureAttributes> element necessary for FeatureChartStyle!");

    // Read domain attribute
    Element domainAttrElem = featureElement.getChild("domainAttr");
    if ( domainAttrElem == null )
      throw new UnsupportedOperationException("Attribute <domainAttr> required for FeatureChartType");
    String domainAttrName = XMLUtil.getAttribute(domainAttrElem, "name");
    if ( domainAttrName == null )
      throw new UnsupportedOperationException("Attribute 'name' necessary for <domainAttr> element of FeatureChartStyle!");
    chartStyle.setAttributeName(0, domainAttrName);
    chartStyle.setAttributeNormalized(0, XMLUtil.getBooleanAttribute(domainAttrElem, "normalize", false));
    chartStyle.setSortDomainAxis(XMLUtil.getBooleanAttribute(domainAttrElem, "sort", false));
    chartStyle.setForceCategories(XMLUtil.getBooleanAttribute(domainAttrElem, "forceCategories", false));
    
    // Read all range attribute definitions
    int rangeAttrNo = 0;
    for (Element featureAttrElem : (List<Element>)featureElement.getChildren("rangeAttr") ) {
      String featureAttrName = XMLUtil.getAttribute(featureAttrElem, "name");
      if ( featureAttrName == null )
        throw new UnsupportedOperationException("Attribute 'name' necessary for <rangeAttr> element of FeatureChartStyle!");
      // Apply attributes to style
      rangeAttrNo++;
      chartStyle.setAttributeName(rangeAttrNo, featureAttrName);
      chartStyle.setAttributeNormalized(rangeAttrNo, XMLUtil.getBooleanAttribute(domainAttrElem, "normalize", false));
    }
  }

  /**
   * Creates a default style for a chart type.
   * @param id   a (unique) ID for the style
   * @param type a chart type
   */
  @Override
  public FeatureChartStyle createDefaultChartStyle(String id, ChartType type) {
    // Create special FeatureChartStyles
    switch ( type ) {
      case AREA:  
      case LINE:  
      case BAR:  
      case POINT: return new FeatureBasicChartStyle(id, type);
      case SCATTER: return new FeatureScatterChartStyle(id);
    }
    throw new UnsupportedOperationException("Style for this chart type not yet supported for features: "+type);
  }


  /**
   * Creates a XML {@link Element} representing the {@link FeatureChartStylee}.
   * @param style           style to create an element for
   * @param rootElementName name for the root element (if {@code null} "ChartType" is 
   *                        used)
   */
  @Override
  public Element createElementFromStyle(FeatureChartStyle style, String rootElementName) {
    Element root = super.createElementFromStyle(style, rootElementName);
    
    // Create and add element <featureAttributes>
    Element attrElem = addChildToElement(root, "featureAttributes");
    
    // Domain attribute definition
    if ( style.getAttributeCount() > 0 )
      addChildToElement(attrElem, "domainAttr", false,
          "name", style.getAttributeName(0),
          "normalize", style.isAttributeNormalized(0),
          "sort", style.isSortDomainAxis(),
          "forceCategories", style.isForceCategories()
      );
    
    // Range attributes definition
    for (int i=1; i<style.getAttributeCount(); i++)
      addChildToElement(attrElem, "rangeAttr", false,
          "name", style.getAttributeName(i),
          "normalize", style.isAttributeNormalized(i)
      );

    return root;
  }


}
