/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.jfree.feature.style;

import hep.aida.bin.DynamicBin1D;
import hep.aida.bin.StaticBin1D;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.geotools.feature.FeatureCollection;
import org.geotools.feature.FeatureIterator;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.plot.CategoryPlot;
import org.jfree.chart.plot.Plot;
import org.jfree.chart.plot.XYPlot;
import org.jfree.chart.renderer.category.CategoryItemRenderer;
import org.jfree.chart.renderer.xy.XYItemRenderer;
import org.jfree.data.category.CategoryDataset;
import org.jfree.data.category.DefaultCategoryDataset;
import org.jfree.data.general.Dataset;
import org.jfree.data.xy.XYDataset;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.simple.SimpleFeatureType;
import org.opengis.feature.type.AttributeDescriptor;
import org.opengis.feature.type.Name;

import schmitzm.geotools.feature.FeatureUtil;
import schmitzm.geotools.feature.PipedFeatureIterator;
import schmitzm.jfree.chart.selection.DatasetSelectionModel;
import schmitzm.jfree.chart.selection.DatasetSelectionModelProvider;
import schmitzm.jfree.chart.style.ChartStyle;
import schmitzm.jfree.chart.style.ChartStyleXMLFactory;
import schmitzm.jfree.feature.Feature2CategoryDatasetMapping;
import schmitzm.jfree.feature.Feature2SeriesDatasetMapping;
import schmitzm.jfree.feature.FeatureDatasetMetaData;
import schmitzm.jfree.feature.FeatureDatasetSelectionModel;

/**
 * This class contains static utility methods related to chart styles based on
 * {@link FeatureCollection}.
 * 
 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a>
 * @version 1.0
 */
public class FeatureChartUtil {
	final static Logger LOGGER = Logger.getLogger(FeatureChartUtil.class);

	/** Instance of {@link ChartStyleXMLFactory}. */
	public static final FeatureChartStyleXMLFactory FEATURE_CHART_STYLE_FACTORY = new FeatureChartStyleXMLFactory();

	/**
	 * Returns all {@link DatasetSelectionModel DatasetSelectionModels} that can
	 * be reached via the renderers of a chart.
	 * 
	 * @param chart
	 *            a chart
	 */
	public static List<FeatureDatasetSelectionModel<?, ?, ?>> getFeatureDatasetSelectionModelFor(
			JFreeChart chart) {
		ArrayList<FeatureDatasetSelectionModel<?, ?, ?>> renderers = new ArrayList<FeatureDatasetSelectionModel<?, ?, ?>>();

		Plot plot = chart.getPlot();

		/** Collect renderers from XYPlot */
		if (plot instanceof XYPlot) {
			XYPlot xyplot = (XYPlot) plot;

			for (int i = 0; i < xyplot.getRendererCount(); i++) {
				XYItemRenderer renderer = xyplot.getRenderer(i);
				if (renderer instanceof DatasetSelectionModelProvider
						&& ((DatasetSelectionModelProvider) renderer)
								.getSelectionModel() instanceof FeatureDatasetSelectionModel) {
					renderers
							.add((FeatureDatasetSelectionModel) ((DatasetSelectionModelProvider) renderer)
									.getSelectionModel());
				}
			}
		}
		/** Collect renderers from CategoryPlot */
		if (plot instanceof CategoryPlot) {
			CategoryPlot catPlot = (CategoryPlot) plot;

			for (int i = 0; i < catPlot.getRendererCount(); i++) {
				CategoryItemRenderer renderer = catPlot.getRenderer(i);
				if (renderer instanceof DatasetSelectionModelProvider
						&& ((DatasetSelectionModelProvider) renderer)
								.getSelectionModel() instanceof FeatureDatasetSelectionModel) {
					renderers
							.add((FeatureDatasetSelectionModel) ((DatasetSelectionModelProvider) renderer)
									.getSelectionModel());
				}
			}
		}

		return renderers;
	}

	/**
	 * Creates a {@link Dataset} for 1 or more attributes of a
	 * {@link FeatureCollection}. According to the feature attribute type the
	 * method decides whether a {@link XYDataset} or a {@link CategoryDataset}
	 * is created:<br>
	 * In case of an non-numeric X attribute a {@link CategoryDataset} is
	 * created always. Otherwise the default is to create a {@link XYDataset}.
	 * The flag {@code forceCat} can be used to create {@link CategoryDataset}
	 * for numeric X attributes.
	 * 
	 * @param fc
	 *            a {@link FeatureCollection}
	 * @param forceCat
	 *            forces a {@link CategoryDataset} also for numeric X attributes
	 * @param sort
	 *            sorts the features according to {@code xAttr} before creating
	 *            the dataset
	 * @param xAttr
	 *            feature attribute used for the X-value
	 * @param yAttr
	 *            feature attribute(s) used for the Y-value (at least one; for
	 *            each a series is created in the dataset)
	 */
	public static Dataset createDataset(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc,
			FeatureChartStyle style) {
		String xAttr = style.getAttributeName(0);
		// Check the X attribute to differ between XY- and CategoryDataset
		AttributeDescriptor xDesc = fc.getSchema().getDescriptor(xAttr);
		if (xDesc == null)
			throw new UnsupportedOperationException("Unknown attribute: "
					+ xAttr);

		// If domain attribute is numeric (and category not forced) create
		// a XYDataset
		if (!style.isForceCategories()
				&& Number.class.isAssignableFrom(xDesc.getType().getBinding()))
			return createXYDataset(fc, style);
		else
			return createCategoryDataset(fc, style);
	}

	/**
	 * Checks whether an attribute exists and whether it is of a specific type.
	 * Otherwise an {@link IllegalArgumentException} is thrown.
	 * 
	 * @param fType
	 *            a feature type
	 * @param aName
	 *            the attribute name checked for numeric type
	 * @param typeToCheck
	 *            the type the attribute must support (if {@code null} only the
	 *            attribute existing is checked!)
	 */
	private static void checkAttributeType(SimpleFeatureType fType,
			String aName, Class<?> typeToCheck, String... errDesc) {
		AttributeDescriptor aDesc = fType.getDescriptor(aName);
		if (aDesc == null)
			throw new UnsupportedOperationException("Unknown attribute: "
					+ aName);

		String attrDesc = errDesc.length > 0 ? errDesc[0] : "Attribute";
		String datasetDesc = errDesc.length > 1 ? errDesc[1] : "Dataset";

		if (typeToCheck != null
				&& !typeToCheck.isAssignableFrom(aDesc.getType().getBinding()))
			throw new UnsupportedOperationException(attrDesc + " must be "
					+ typeToCheck.getSimpleName() + " for " + datasetDesc
					+ ": " + aName);
	}

	/**
	 * Creates a {@link XYDataset} for 2 (or more) attributes of a
	 * {@link FeatureCollection}. XYDateset can only be created for
	 * <b>numeric</b> attributes.
	 * 
	 * @param fc
	 *            a {@link FeatureCollection}
	 * @param style
	 *            defines the attributes used to create the dataset from, as
	 *            well as the sorting and normalization properties
	 * @throws IllegalArgumentException
	 *             if less then 2 attributes are specified
	 * @throws UnsupportedOperationException
	 *             if attributes are not numeric
	 */
	public static XYSeriesCollection createXYDataset(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc,
			FeatureChartStyle chartStyle) {
		int attrCount = chartStyle.getAttributeCount();
		if (attrCount < 2)
			throw new IllegalArgumentException(
					"FeatureChartStyle must define at least 2 attributes to create XYDataset: "
							+ attrCount);
		String xAttrName = chartStyle.getAttributeName(0);
		// check data types of X attribute
		checkAttributeType(fc.getSchema(), xAttrName, Number.class,
				"Domain attribute", "XYDataset");
		// check data types of Y attribute
		for (int i = 1; i < attrCount; i++)
			checkAttributeType(fc.getSchema(), chartStyle.getAttributeName(i),
					Number.class, "Range attribute", "XYDataset");

		// Create a series for each range attribute (note: the 0
		// array element useless, but it is more comfortable to
		// use an array of the same size as "attributes"
		XYSeries[] xySeries = new XYSeries[attrCount];
		for (int i = 1; i < xySeries.length; i++)
			xySeries[i] = new XYSeries(chartStyle.getAttributeName(i), false,
					true); // IMPORTANT: Do not use autosort!!

		// Create a new dataset and insert the series
		XYSeriesCollection dataset = new XYSeriesCollection();
		Feature2SeriesDatasetMapping mapping = new Feature2SeriesDatasetMapping(
				fc, dataset);
		dataset.setGroup(new FeatureDatasetMetaData(mapping));
		for (int i = 1; i < xySeries.length; i++)
			dataset.addSeries(xySeries[i]);

		// Calculate any statistics needed for normalization.
		HashMap<String, StaticBin1D> statisticsForNormalization = calcStatisticsForNormalization(
				fc, chartStyle);

		// If dataset should be sorted, the features must be sorted first
		// according to the domain attribute. The "autoSort" functionality
		// of XYSeries can NOT BE USED because of the following reason:
		// To realize the synchronized feature/dataitem a mapping
		// between the feature IDs and the data item index (in the series)
		// must be created. Using the "autoSort" functionality, the
		// indices inside the series change with every XYSeries.add(..)
		// call, so that the mapping is destroyed!!
		//
		// Only solution: Pre-sort the features so that the XYSeries internal
		// order does not change!
		Iterator<SimpleFeature> fi = null;
		if (chartStyle.isSortDomainAxis()) {
			Vector<SimpleFeature> sortedFeatures = FeatureUtil.sortFeatures(fc,
					xAttrName);
			// Create an Iterator for the sorted Features
			fi = sortedFeatures.iterator();
		} else
			fi = new PipedFeatureIterator(fc.features());

		// Iterate the FeatureCollection and fill the series
		int datasetIdx = 0;
		for (; fi.hasNext();) {
			SimpleFeature f = fi.next();
			// Determine X value (NULL not permitted for XYDateset!)
			Number xValue = (Number) f.getAttribute(xAttrName);
			
			// Filter out NODATA values
			xValue = chartStyle.filterNoDataValue(0, xValue);
			if (xValue == null)
				continue;

			
			// Determine the Y values and fill the series
			for (int attrIdx = 1; attrIdx < attrCount; attrIdx++) {
				String yAttrName = chartStyle.getAttributeName(attrIdx);
				Number yValue = (Number) f.getAttribute(yAttrName);
				
				// Filter out NoDataValues
				yValue = chartStyle.filterNoDataValue(attrIdx, yValue);
				if (yValue == null)
					continue;

				/* Normalization of a range axis value */
				if (chartStyle.isAttributeNormalized(attrIdx))
					yValue = normalize(yValue, yAttrName,
							statisticsForNormalization);

				/*
				 * Normalization of the domain axis value (if they are not
				 * handled as categories)
				 */
				if (chartStyle.isAttributeNormalized(ChartStyle.DOMAIN_AXIS))
					xValue = normalize(xValue, xAttrName,
							statisticsForNormalization);

				xySeries[attrIdx].add(xValue, yValue);

				// Mapping between FID and data index in series
				mapping.setMapping(f.getID(), yAttrName, datasetIdx++);
			}
			
		}
		return dataset;
	}

	/**
	 * Calculates statistics needed to normalize data. If normalization is not
	 * used, this function returns an empty map.
	 * 
	 * @param fc
	 *            {@link FeatureCollection} where the data comes from
	 * @param chartStyle
	 *            {@link ChartStyle} to determine which attributes shall be
	 *            normalized.
	 */
	public static HashMap<String, StaticBin1D> calcStatisticsForNormalization(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc,
			FeatureChartStyle chartStyle) {
		// NORMALIZATION:
		// We have to create a descriptive statistic of all items before we can
		// transform the first one. So we iterate over all items before we do
		// optional sorting and insertion.
		/**
		 * Holds the statistics needed to normalize the attribute values. Key =
		 * attributeName
		 */
		HashMap<String, StaticBin1D> attribStats = new HashMap<String, StaticBin1D>();

		// First check if any attribute needs normalization. If not we can
		// skip the whole iteration
		boolean doNormalization = false;
		for (int attrIdx = 0; attrIdx < chartStyle.getAttributeCount(); attrIdx++) {
			if (chartStyle.isAttributeNormalized(attrIdx)) {
				doNormalization = true;
				break;
			}
		}

		if (doNormalization) {
			final FeatureIterator<SimpleFeature> fIt = fc.features();
			try {

				// Loop over all features, and collect the attribute values we
				// are interested in.
				while (fIt.hasNext()) {
					SimpleFeature f = fIt.next();

					for (int attrIdx = 0; attrIdx < chartStyle
							.getAttributeCount(); attrIdx++) {

						if (chartStyle.isAttributeNormalized(attrIdx)) {
							String attrName = chartStyle
									.getAttributeName(attrIdx);

							StaticBin1D stat = attribStats.get(attrName);
							if (stat == null) {
								stat = new DynamicBin1D();
								attribStats.put(attrName, stat);
							}

							// Filter out NoDataValues
							Object rawValue = chartStyle.filterNoDataValue(attrIdx, f.getAttribute(attrName));
							if (rawValue == null)
								continue;

//							rawValue = ,
//									rawValue);

							final Double doubleValue = new Double(
									((Number) rawValue).doubleValue());
							stat.add(doubleValue);
							// LOGGER.debug("Adding "+doubleValue+" and sd = "+stat.standardDeviation()+" and mean = "+stat.mean());
						}
					}
				} // Featuer iterator
			} finally {
				fc.close(fIt);
			}
		} // doNormlization

		return attribStats;
	}

	/**
	 * Do a z-transformation on the data item.
	 * 
	 * @param statisticsForNormalization
	 *            The statistics needed for normalization.
	 * 
	 * @see #calcStatisticsForNormalization(FeatureCollection,
	 *      FeatureChartStyle, int)
	 */
	private static Number normalize(Number yValue, String attrName,
			HashMap<String, StaticBin1D> statisticsForNormalization) {
		Number zValue;

		if (statisticsForNormalization.get(attrName) == null) {
			throw new IllegalArgumentException(
					"The statistics needed for the normlization need to be created before normalizing.");
		}

		StaticBin1D stats = statisticsForNormalization.get(attrName);

		zValue = (yValue.doubleValue() - stats.mean())
				/ stats.standardDeviation();

		return zValue;
	}

	/**
	 * Creates a {@link XYDataset} for 2 (or more) attributes of a
	 * {@link FeatureCollection}. XYDateset can only be created for
	 * <b>numeric</b> attributes.
	 * 
	 * @param fc
	 *            a {@link FeatureCollection}
	 * @param style
	 *            defines the attributes used to create the dataset from, as
	 *            well as the sorting and normalization properties
	 * @throws IllegalArgumentException
	 *             if less then 2 attributes are specified
	 * @throws UnsupportedOperationException
	 *             if attributes are not numeric
	 */
	public static DefaultCategoryDataset createCategoryDataset(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc,
			FeatureChartStyle chartStyle) {
		int attrCount = chartStyle.getAttributeCount();
		if (attrCount < 2)
			throw new IllegalArgumentException(
					"FeatureChartStyle must define at least 2 attributes to create CategoryDataset: "
							+ attrCount);
		String xAttrName = chartStyle.getAttributeName(0);
		// only check whether X attribute exists (numeric and not-numeric
		// allowed)
		checkAttributeType(fc.getSchema(), xAttrName, Comparable.class,
				"Domain attribute", "CategoryDataset");
		// check data types of Y attribute (only numeric allowed)
		for (int i = 1; i < attrCount; i++)
			checkAttributeType(fc.getSchema(), chartStyle.getAttributeName(i),
					Number.class, "Range attribute", "CategoryDataset");

		// Calculate any statistics needed for normalization.
		HashMap<String, StaticBin1D> statisticsForNormalization = calcStatisticsForNormalization(
				fc, chartStyle);

		// Create a new dataset and insert the series
		DefaultCategoryDataset dataset = new DefaultCategoryDataset();
		Feature2CategoryDatasetMapping mapping = new Feature2CategoryDatasetMapping(
				fc, dataset);
		dataset.setGroup(new FeatureDatasetMetaData(mapping));

		// If dataset should be sorted, the features must be sorted first
		// according to the domain attribute. CategoryDataset has no
		// "autoSort", and this - at all - would not work (see
		// createXYDataset(..))
		Iterator<SimpleFeature> fi = null;
		FeatureIterator<SimpleFeature> features = null;
		if (chartStyle.isSortDomainAxis()) {
			Vector<SimpleFeature> sortedFeatures = FeatureUtil.sortFeatures(fc,
					xAttrName);
			// Create an Iterator for the sorted Features
			fi = sortedFeatures.iterator();
		} else {
			features = fc.features();
			fi = new PipedFeatureIterator(features);
		}

		try {

			// Iterate the FeatureCollection and fill the series
			for (int fIdx = 0; fi.hasNext(); fIdx++) {
				SimpleFeature f = fi.next();
				// Determine the category (NULL not permitted!)
				Comparable<?> catValue = (Comparable<?>) f
						.getAttribute(xAttrName);
				
				// Filter out NODATA values
				catValue = chartStyle.filterNoDataValue(0, catValue);
				
				if (catValue == null)
					continue;
				
				// Determine the Y values and fill the dataset
				for (int attrIdx = 1; attrIdx < attrCount; attrIdx++) {
					String yAttrName = chartStyle.getAttributeName(attrIdx);
					Number yValue = (Number) f.getAttribute(yAttrName);

					// Filter out NoDataValues
					yValue = chartStyle.filterNoDataValue(attrIdx, yValue);
					if (yValue == null)
						continue;

					/* Normalization of a range axis value */
					if (chartStyle.isAttributeNormalized(attrIdx))
						yValue = normalize(yValue, yAttrName,
								statisticsForNormalization);

					// Add data to dataset
					dataset.addValue(yValue, yAttrName, catValue);

					// Mapping between FID and data index in series
					mapping.setMapping(f.getID(), yAttrName, catValue);
				}

			}
		} finally {
			if (features != null) {
				// this is a hint
				fc.close(features);
			}
		}

		return dataset;
	}

	/**
	 * After loading from XML, a {@link FeatureChartStyle} contains whatever is
	 * written in the XML. But the DBF Schema can change quickly by accident!
	 * This method checks an {@link FeatureChartStyle} against a schema and
	 * corrects upperCase/lowerCase problems where possible.
	 * 
	 * Returns <code>false</code> if attributes had to be removed from the
	 * {@link FeatureChartStyle}.
	 * 
	 * @author SK
	 */
	public static boolean correctAttributeNames(
			FeatureChartStyle featureChartStyle, SimpleFeatureType schema) {

		ArrayList<Integer> willRemove = new ArrayList<Integer>();

		// 1. Check.. all attributes in the atm should be in the schema as well.
		// maybe correct some upperCase/loweCase stuff

		for (int idx = 0; idx < featureChartStyle.getAttributeCount(); idx++) {

			String attributeName = featureChartStyle.getAttributeName(idx);

			AttributeDescriptor foundDescr = schema
					.getDescriptor(attributeName);
			if (foundDescr == null) {
				Name bestMatch = FeatureUtil.findBestMatchingAttribute(schema,
						attributeName);
				if (bestMatch == null)
					willRemove.add(idx);
				else
					featureChartStyle.setAttributeName(idx, bestMatch
							.getLocalPart());
			}
		}

		// Remove the ones that were not findable in the schema
		for (Integer removeIdx : willRemove) {
			featureChartStyle.removeAttribute(removeIdx);
		}

		// * Returns false if attributes had to be removed from the {@link
		// FeatureChartStyle}.
		return willRemove.isEmpty();
	}
}
