/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.jfree.feature.style;

import java.util.Set;

import org.geotools.feature.FeatureCollection;
import org.jfree.chart.JFreeChart;
import org.jfree.data.category.CategoryDataset;
import org.jfree.data.general.Dataset;
import org.jfree.data.xy.XYDataset;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.simple.SimpleFeatureType;

import schmitzm.jfree.chart.style.ChartStyle;
import schmitzm.jfree.chart.style.ScatterChartStyle;

/**
 * This class extends the {@link ScatterChartStyle} with the properties
 * of the {@link FeatureChartStyle}.<br>
 * After instantiation the 2 feature attributes used to define the chart
 * data must be set by {@link #setAttributeName(int, String)}. Without
 * setting the attributes the style can not be applied to a feature collection!
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 */
public class FeatureScatterChartStyle extends ScatterChartStyle implements FeatureChartStyle {
  /** Used to maintain the {@link FeatureChartStyle} properties. */
  protected Dummy dummyFeatureChartStyle = null;
  
  /**
   * Creates a scatter chart style a regression line shown.
   * @param id   a (unique) ID for the style
   */
  public FeatureScatterChartStyle(String id) {
    super(id);
    dummyFeatureChartStyle = new Dummy(id,2);
  }
  
  @Override
  /**
   * Nicer debugging
   */
	public String toString() {
	  
	  String titel = getTitleStyle() != null ? getTitleStyle().getLabel() : " - null title!!! - ";
		return "ID="+getID()+"   "+titel+"" + super.toString();
	}

  /**
   * Creates a scatter chart style a regression line shown.
   * @param id         a (unique) ID for the style
   * @param title      title for the chart
   * @param xAttrName  feature attribute used for domain axis
   * @param yAttrName  feature attribute used for range axis
   */
  public FeatureScatterChartStyle(String id, String title, String xAttrName, String yAttrName) {
    this(id);
    setAttributeName(0, xAttrName);
    setAttributeName(1, yAttrName);
    getTitleStyle().setLabel(title);
  }

  /**
   * Creates a (deep) clone of this style.
   */
  @Override
  public FeatureScatterChartStyle copy() {
    return (FeatureScatterChartStyle) copyTo( new FeatureScatterChartStyle("") );
  }
  
  /**
   * Copies all properties of this style to another one.
   * @param dest destination object (if {@code null} the copy
   *             is created by {@link #copy()})
   * @return {@code dest} or the new instance
   */
  @Override
  public ChartStyle copyTo(ChartStyle dest) {
    // copy the super class properties
    dest = super.copyTo(dest);
    // copy this classes properties
    if ( dest instanceof FeatureChartStyle )
      dummyFeatureChartStyle.copyTo(dest);

    return dest;
  }

  /**
   * Removes all style informations about an attribute and
   * reorganizes the attribute indexes so there is an continuous 
   * order. 
   * @param idx an attribute
   */
  @Override
  public void removeAttribute(int idx) {
    dummyFeatureChartStyle.removeAttribute(idx);
  }

  /**
   * Returns the maximum number of feature attributes that can be
   * specified by this style.
   * @return always 2
   */
  @Override
  public int getMaxAttributeCount() {
    return dummyFeatureChartStyle.getMaxAttributeCount();
  }

  /**
   * Returns the number of feature attributes defined in this style.
   */
  @Override
  public int getAttributeCount() {
    return dummyFeatureChartStyle.getAttributeCount();
  }

  /**
   * Returns the name of a feature attribute needed to create a
   * chart for this style.
   * @param idx attribute index (0=domain; 1=1st range series;
   *            2=2nd range series; ...)
   */
  @Override
  public String getAttributeName(int idx) {
    return dummyFeatureChartStyle.getAttributeName(idx);
  }

  /**
   * Sets the name of a feature attribute needed to create a
   * chart for this style.
   * @param idx attribute index (0=domain; 1=1st range series;
   *            2=2nd range series; ...)
   * @param attrName feature attribute name 
   */
  @Override
  public void setAttributeName(int idx, String attrName) {
    dummyFeatureChartStyle.setAttributeName(idx,attrName);
  }

  /**
   * Sets whether the features are sorted according to the
   * domain attribute (before creating a {@link Dataset}).
   */
  @Override
  public void setSortDomainAxis(boolean sortDomainAxis){
    dummyFeatureChartStyle.setSortDomainAxis(sortDomainAxis);
  }

  /**
   * Returns whether the features are sorted according to the
   * domain attribute (before creating a {@link Dataset}).
   * @return {@code false} as default
   */
  @Override
  public boolean isSortDomainAxis() {
    return dummyFeatureChartStyle.isSortDomainAxis();
  }

  /** 
   * Sets whether a {@link CategoryDataset} is forced for
   * a numeric domain attribute. The default is to create {@link XYDataset}
   * for a numeric and {@link CategoryDataset} for a non-numeric domain
   * attribute.
   */
  @Override
  public void setForceCategories(boolean forceCategories){
    dummyFeatureChartStyle.setForceCategories(forceCategories);
  }

  /** 
   * Returns whether a {@link CategoryDataset} is forced for
   * a numeric domain attribute. The default is to create {@link XYDataset}
   * for a numeric and {@link CategoryDataset} for a non-numeric domain
   * attribute.
   * @return {@code false} as default
   */
  @Override
  public boolean isForceCategories(){
    return dummyFeatureChartStyle.isForceCategories();
  }

  /** 
   * Sets whether the attribute data is normalized for an
   * attribute(before creating a {@link Dataset}).
   * @param idx attribute index (0=domain; 1=1st range series;
   *            2=2nd range series; ...)
   * @param normalize indicates the normalize property
   */  
  @Override
  public void setAttributeNormalized(int idx, Boolean normalize){
    dummyFeatureChartStyle.setAttributeNormalized(idx, normalize);
  }

  /** 
   * Returns whether the attribute data is normalized for an
   * attribute(before creating a {@link Dataset}).
   * @param idx attribute index (0=domain; 1=1st range series;
   *            2=2nd range series; ...)
   * @return {@code false} as default
   */  
  @Override
  public boolean isAttributeNormalized(int idx) {
       return dummyFeatureChartStyle.isAttributeNormalized(idx);
  }

  /**
   * Sets the values, which are interpreted as "No Data". 
   * @param idx attribute index the "No Data" values are set for
   * @param noDataValues the "No Data" values
   */
  public void setNoDataValues(int idx, Set<Object> noDataValues) {
    dummyFeatureChartStyle.setNoDataValues(idx, noDataValues);
  }
  
  /**
   * Returns the values, which are interpreted as "No Data". 
   * @param idx attribute index the "No Data" values are returned for
   */
  public Set<Object> getNoDataValues(int idx) {
    return dummyFeatureChartStyle.getNoDataValues(idx);
  }
  
  /**
   * Sets a value, which is interpreted as "No Data". 
   * @param idx attribute index the "No Data" value is set for
   * @param noDataValue the "No Data" value
   */
  public void addNoDataValue(int idx, Object noDataValue) {
    dummyFeatureChartStyle.addNoDataValue(idx, noDataValue);
  }

  /**
   * Removes a "No Data" value for an attribute. 
   * @param idx attribute index the "No Data" value is removed for
   * @param noDataValue the "No Data" value to remove
   * @return {@code false} if the value was not an "No Data" value 
   */
  public boolean removeNoDataValue(int idx, Object noDataValue) {
    return dummyFeatureChartStyle.removeNoDataValue(idx, noDataValue);
  }

  /**
   * Checks whether the given value is one of the "No data" values for
   * the attribute.
   * @param idx attribute index the "No Data" value is checked for
   * @param value an attribute value
   */
  public boolean isNoDataValue(int idx, Object value) {
    return dummyFeatureChartStyle.isNoDataValue(idx, value);
  }

  /**
   * Checks whether the given value is one of the "No data" values for
   * the attribute. In this case this method returns {@code null}, otherwise
   * the value itself.
   * @param idx attribute index the "No Data" value is checked for
   * @param value an attribute value
   * @return {@code null} if the given value is one of the "No data" values 
   */
  public <T> T filterNoDataValue(int idx, T value) {
    return dummyFeatureChartStyle.filterNoDataValue(idx, value);
  }

  /**
   * Creates an appropriate {@link Dataset} for the attributes defined
   * by this style (according to the attributes types in the given
   * {@link FeatureCollection}) and calls {@link #applyToDataset(Dataset)}.
   */
  @Override
  public JFreeChart applyToFeatureCollection(FeatureCollection<SimpleFeatureType, SimpleFeature> fc) {
    Dataset dataset = FeatureChartUtil.createDataset(fc,this);
    return applyToDataset(dataset);
  }
}
