/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.lang;

// nur fuer Doku
import java.util.TreeSet;

/**
 * Diese Klasse dient dazu, ein beliebiges Objekt <b>{@linkplain Comparable vergleichbar}</b>
 * zu machen. Hierzu verbindet eine diese <code>ComparableObject</code>
 * das Objekt mit einem {@linkplain Comparable vergleichbaren} Key, so dass
 * ein Tupel <K,V> entsteht, welches z.B. in ein {@link TreeSet} eingefuegt
 * werden kann.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ComparableObject<K extends Comparable,V> implements Comparable {
  private K       key    = null;
  private V       obj    = null;
  private boolean onlyObj = false;

  /**
   * Erzeugt ein neues Tupel
   * @param key  Key-Wert ueber den das Objekt verglichen wird
   *             (muss eine {@link Comparable}-Instanz sein!)
   * @param obj  beliebiges Objekt, dem der Key <code>key</code> zugeordnet ist
   */
  public ComparableObject(K key, V obj) {
    this(key,obj,false);
  }

  /**
   * Erzeugt ein neues Tupel
   * @param key     Key-Wert ueber den das Objekt verglichen wird
   *                (muss eine {@link Comparable}-Instanz sein!)
   * @param obj     beliebiges Objekt, dem der Key <code>key</code> zugeordnet ist
   * @param onlyObj wenn <code>true</code> wird in der {@link #equals(Object)}-Methode
   *                nur auf <code>obj</code> verglichen. Standardmaessig vergleicht
   *                die {@link #equals(Object)}-Methode <code>obj</code> und
   *                <code>key</code>.
   */
  public ComparableObject(K key, V obj, boolean onlyObj) {
    this.key = key;
    this.obj = obj;
    this.onlyObj = onlyObj;
  }

  /**
   * Liefert das Objekt.
   */
  public V getObject() {
    return obj;
  }

  /**
   * Liefert den Schluessel.
   */
  public K getKey() {
    return key;
  }

  /**
   * Vergleicht den Schluessel dieses Objekts mit einem anderen. Liefert nur 0,
   * wenn <code>o</code> mit <code>this</code> uebereinstimmt!
   * @param o zu vergleichendes Objekt (kann ein anderes <code>ComparableObject</code>
   *          sein oder eine {@link Comparable}-Instanz)
   * @return einen <b>negativen</b> Wert, falls dieses Objekt <i>kleiner</i> ist als <code>o</code>;<br>
   *         einen <b>positiven</b> Wert, falls dieses Objekt <i>groesser</i> ist als <code>o</code>;<br>
   *         <b>0</b>, falls die beiden Objekte identisch sind
   */
  public int compareTo(Object o) {
    if ( this == o || this.equals(o) )
      return 0;

    int comp = ( o instanceof ComparableObject ) ? key.compareTo( ((ComparableObject)o).getKey() ) : key.compareTo(o);
    if ( comp == 0 ) {
      comp = ( this.hashCode() < o.hashCode() ) ? -1 : 1;
    }
    return comp;
  }

  /**
   * Vergleicht das <code>ComparableObject</code> mit einem anderen. Wurde im
   * {@linkplain #ComparableObject(K,V,boolean) Konstruktor} das <code>onlyObj</code>-Flag
   * auf <code>true</code> gesetzt, wird an dieser Stelle nur das Objekt
   * auf Gleichheit verglichen, ansonsten auch der Key.
   * @param o ein anderes <code>ComparableObject</code>
   */
  public boolean equals(Object o) {
    if ( !(o instanceof ComparableObject) )
      return false;
    ComparableObject c = (ComparableObject)o;
    return this.obj.equals(c.getObject()) && (onlyObj || this.key.equals(c.getKey()));
  }

}
