/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.lang;

import java.util.Comparator;

/**
 * {@code Comparator} um beliebige Objekte ueber ihre {@code toString()}-Methode
 * zu vergleichen.
 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class DefaultComparator implements Comparator<Object> {

  /** Standard-Instanz dieses {@code Comparator}. */
  public static final DefaultComparator DEFAULT = new DefaultComparator();

  /**
   * Erzeugt einen neuen {@code Comparator}.
   */
  public DefaultComparator() {
  }

  /**
   * Fuehrt den Vergleich aus.
   * @param obj1 ein Objekt
   * @param obj2 ein anders Objekt
   * @return ein negativer Wert fuer {@code obj1 < obj2}, 0 für {@code obj1 = obj2}
   *         oder ein positiver Wert fuer {@code obj1 > obj2}
   */
  public int compare(Object obj1, Object obj2) {
    if ( obj1 == null && obj2 == null )
      return 0;
    if ( obj1 == null )
      return Integer.MAX_VALUE;
    if ( obj2 == null )
      return Integer.MIN_VALUE;
    return obj1.toString().compareTo(obj2.toString());
  }

  /**
   * Liefert einen {@link Comparator}, der die umgekehrte Reihenfolge eines
   * {@link Comparator} definiert. {@code null}-Werte sind trotzdem groesser
   * als alles andere und werden somit auch in der invertierten Reihenfolge
   * hinten angehaengt.
   * @param comparator ein Comparator
   */
  public static <T> Comparator<T> invert(final Comparator<T> comparator) {
    return new Comparator<T>() {
      public int compare(T obj1, T obj2) {
        if ( obj1 == null && obj2 == null )
          return 0;
        if ( obj1 == null )
          return Integer.MAX_VALUE;
        if ( obj2 == null )
          return Integer.MIN_VALUE;
        return comparator.compare(obj2,obj1);
      }
    };
  }
}
