/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.lang;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.ResourceBundle;

/**
 * Diese Klasse stellt ein als {@link Hashtable} organisiertes
 * {@link ResourceBundle} dar. Sub-Klassen muessen lediglich noch die Methode
 * {@link #getContents()} implementieren, die den Inhalt des ResourceBundle
 * liefert. Aus diesem wird die Hash-Tabelle zusammengesetzt.<br>
 * Das ResourceBundle kann so eingestellt werden, dass es Fehler ignoriert.
 * Normalerweise wirft {@link ResourceBundle#getObject} eine Exception, wenn
 * zu einem Key kein Objekt (auch in keinem Parent-ResourceBundle) zu finden ist.
 * Ueber die Methode {@link #setErrorTolerant(boolean)} kann das
 * <code>HashtableResourceBundle</code> so eingestellt werden, dass ein
 * solcher Fehler vermieden wird, in dem statt <code>null</code> ein Dummy-String
 * von der Methode {@link #handleGetObject(String)} geliefert wird.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class HashtableResourceBundle extends ResourceBundle {
  /** Speichert den Inhalt des {@link ResourceBundle}. */
  protected Hashtable content = new Hashtable<String,Object>();

  /** Speichert, ob das Fehlen einer angeforderten Ressource toleriert wird.<br>
   *  Default: <code>false</code> */
  protected static boolean errorTolerant = false;

  /**
   * Erzeugt ein neues ResourceBundle.
   */
  public HashtableResourceBundle() {
    Object[][] res = getContents();
    if ( res!=null ) {
      content = new Hashtable<String, Object> (res.length);
      for (int i = 0; i < res.length; i++) {
        if (res[i].length < 2)
          throw new UnsupportedOperationException("Return of HashtableResourceBundle.getContents() must be a two dimensional array!");
        if ( !(res[i][0] instanceof String) )
          throw new UnsupportedOperationException("HashtableResourceBundle.getContents() must contain Strings in first dimension!");
        content.put(res[i][0], res[i][1]);
      }
    }
  }

  /**
   * Liefert alle Keys zu denen eine Ressource hinterlegt ist.
   */
  public Enumeration<String> getKeys() {
    return content.keys();
  }

  /**
   * Stellt sie Toleranz des ResourceBundle bei fehlenden Ressourcen ein.
   */
  public static void setErrorTolerant(boolean tolerant) {
    errorTolerant = tolerant;
  }

  /**
   * Prueft, ob das ResourceBundle fehlertolerant auf fehlende Ressourcen
   * reagiert.
   */
  public static boolean isErrorTolerant() {
    return errorTolerant;
  }

  /**
   * Liefert eine Ressource zu einem Key aus der Hash-Tabelle.
   * @return <code>null</code> falls es den Key in der Hash-Tabelle nicht
   *         gibt und die Fehlerroleranz abgeschaltet ist
   */
  protected Object handleGetObject(String key) {
    Object obj = content.get(key);
    // Wenn kein Objekt gefunden wurde, wuerde die Standard-Methode
    // von ResourceBundle.getObject(String) eine Exception werfen
    // Dies soll aber nicht der Fall sein, wenn das Bundle Fehler-Tolerant ist
    if ( obj == null && parent == null && isErrorTolerant() )
      obj = "'"+key+"'-Resource not found!";
    return obj;
  }

  /**
   * Liefert die (Key,Ressource)-Paerchen. Der Key <b>muss</b> ein String sein.
   */
  public abstract Object[][] getContents();
}
