/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.lang;

import java.util.Hashtable;
import java.util.Vector;

/**
 * Diese Klasse erweitert die normale Hash-Tabelle um Kollisionslisten.
 * Pro Hashtable-Key koennen somit mehrere Werte gespeichert werden.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class HashtableWithCollisionList<K,V> extends Hashtable<K,Vector<V>> {

  /**
   * Erzeugt eine neue Hash-Tabelle.
   */
  public HashtableWithCollisionList() {
    super();
  }

  /**
   * Fuegt der Hash-Tabelle ein Objekt hinzu. Existiert bereits ein Element
   * mit dem Key, wird dieses nicht entfernt, sondern die entsprechende
   * Kollisionsliste erweitert.
   * @param key   Der Hashtable-Key
   * @param value Wert
   */
  public synchronized void add(K key, V value) {
    // Wenn noch keine Kollisionsliste zu dem Key existiert, wird
    // diese erzeugt
    if ( !this.containsKey(key) )
      super.put(key,new Vector<V>());
    // In der Liste das Objekt hinzufuegen
    ((Vector)this.get(key)).add(value);
  }

  /**
   * Ersetzt eine komplette Kollisionsliste fuer einen Key.
   * @param key   Der Hashtable-Key
   * @param value Kollisionsliste
   * @return immer <code>null</code>
   */
  public synchronized Vector<V> put(K key, Vector<V> value) {
    super.put(key,value);
    return null;
  }

  /**
   * Entfernt eine komplette Kollisionsliste zu einem Key. Es werden nur die
   * Element der Liste entfernt. Die Kollisionsliste selbst bleibt bestehen!
   * @param key   Der Hashtable-Key
   * @return immer <code>null</code>
   */
  public synchronized Vector<V> remove(Object key) {
    Vector list = super.get(key);
    if ( list!= null )
      list.clear();
    return null;
  }

  /**
   * Entfernt ein objekt aus der Hash-Tabelle.
   * @param key   Der Hashtable-Key
   * @return Das entfernte Objekt oder <code>null</code> falls das Objekt
   *         nicht in der Hash-Tabelle gespeichert war
   */
  public synchronized Object remove(Object key, Object value) {
    Vector list = super.get(key);
    if ( list == null )
      return null;
    list.remove(value);
    return value;
  }

}
