/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.lang;

import java.awt.Toolkit;
import java.awt.datatransfer.StringSelection;
import java.io.PrintStream;
import java.lang.reflect.Constructor;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Locale;
import java.util.TreeSet;
import java.util.Vector;

import net.jini.loader.pref.PreferredClassLoader;

import org.apache.log4j.Category;
import org.apache.log4j.Logger;

/**
 * Diese Klasse stellt Hilfsfunktionen fuer Basisoperationen bereit.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class LangUtil {

  /** {@link ResourceProvider}, der die Lokalisation fuer Komponenten
   *  des Package {@code schmitzm.lang} zur Verfuegung stellt. Diese sind
   *  in properties-Dateien unter {@code schmitzm.lang.resource.locales}
   *  hinterlegt. */
  public static ResourceProvider RESOURCE = new ResourceProvider( extendPackagePath(LangUtil.class,"resource.locales.LangResourceBundle"), Locale.ENGLISH );

  /**
   * Liefert den einfachen Namen einer Klasse (ohne Package) oder den
   * String "null", falls das Objekt {@code null} ist
   * @param object ein Objekt
   * @param niemals {@code null}
   */
  public static String getSimpleClassName(Object object) {
    if ( object == null )
      return "null";
    return object.getClass().getSimpleName();
  }


  /**
   * Prueft, ob ein {@link Enum} ein spezielles Feld besitzt.
   * @param enumClass enum typ
   * @param fieldName feld name
   */
  public static boolean checkEnumValue(Class<? extends Enum> enumClass, String fieldName) {
    try {
      Enum.valueOf(enumClass, fieldName);
      return true;
    } catch (IllegalArgumentException err) {
      return false;
    }
  }

  /**
   * Fuegt alle Elemente eines Arrays in einen {@link Vector} ein. Wird als Vector
   * <code>null</code> angegeben, wird ein neuer Vector erzeugt.
   * @param obj einzufuegende Objekte
   * @param vec Vector in den die Elemente eingefuegt werden (kann <code>null</code>
   *            sein)
   * @return den Vector <code>vec</code> oder eine neue {@link Vector}-Instanz
   */
  public static Vector addObjectsToVector(Object[] obj, Vector vec) {
    if ( vec == null )
      vec = new Vector();
    for (int i=0; i<obj.length; i++)
      vec.add(obj[i]);
    return vec;
  }

  /**
   * Ersetzt alle Tabulator-Vorkommen in einem String durch Leerzeichen.
   * @param source Quell-String
   * @param tabLen Anzahl an Zeichen die ein Tabulator darstellt
   */
  public static String replaceTabulators(String source, int tabLen) {
    String outString       = new String(source);
    char[] maxSpacePattern = new char[tabLen];
    Arrays.fill(maxSpacePattern,' ');

    for(int pos; (pos=outString.indexOf('\t')) >= 0;) {
      outString = outString.replaceFirst(
        "\t",
        String.valueOf(maxSpacePattern,0,tabLen-pos%tabLen)
      );
    }
    return outString;
  }

  /**
   * Haengt eine beliebige Anzahl von Strings hintereinander.
   * @param str aneinanderzuhaengende Stings
   */
  public static String stringConcat(String... str) {
    StringBuffer sb = new StringBuffer();
    for (int i=0; str!=null && i<str.length; i++)
      sb.append(str[i]);
    return sb.toString();
  }

  /**
   * Haengt eine beliebige Anzahl von Strings hintereinander und fuegt zwischen
   * den Strings einen Trenner ein.
   * @param sep Trenner
   * @param str aneinanderzuhaengende Stings
   */
  public static String stringConcatWithSep(String sep, String... str) {
    StringBuffer sb = new StringBuffer();
    for (int i=0; str!=null && i<str.length; i++) {
      if ( i>0 )
        sb.append(sep);
      sb.append(str[i]);
    }
    return sb.toString();
  }

  /**
   * Erzeugt einen vorbelegten Array.
   * @param count Groesse des Arrays
   * @param value Wert mit dem die Elemente vorbelegt werden
   */
  public static int[] createArray(int count, int value) {
    int[] array = new int[count];
    java.util.Arrays.fill(array,value);
    return array;
  }

  /**
   * Erzeugt einen vorbelegten Array.
   * @param count Groesse des Arrays
   * @param value Wert mit dem die Elemente vorbelegt werden
   */
  public static long[] createArray(int count, long value) {
    long[] array = new long[count];
    java.util.Arrays.fill(array,value);
    return array;
  }

  /**
   * Erzeugt einen vorbelegten Array.
   * @param count Groesse des Arrays
   * @param value Wert mit dem die Elemente vorbelegt werden
   */
  public static float[] createArray(int count, float value) {
    float[] array = new float[count];
    java.util.Arrays.fill(array,value);
    return array;
  }

  /**
   * Erzeugt einen vorbelegten Array.
   * @param count Groesse des Arrays
   * @param value Wert mit dem die Elemente vorbelegt werden
   */
  public static double[] createArray(int count, double value) {
    double[] array = new double[count];
    java.util.Arrays.fill(array,value);
    return array;
  }

  /**
   * Erzeugt einen vorbelegten Array.
   * @param count Groesse des Arrays
   * @param value Wert mit dem die Elemente vorbelegt werden
   */
  public static String[] createArray(int count, String value) {
    String[] array = new String[count];
    java.util.Arrays.fill(array,value);
    return array;
  }

  /**
   * Erzeugt einen mehr-dimensionalen {@link Object}-Array.
   * @param size Groesse der Array-Dimensionen
   * @return einen {@code Object[]..[]}
   */
  public static Object createArray(int... size) {
    switch ( size.length ) {
      case  1: return new Object[size[0]];
      case  2: return new Object[size[0]][size[1]];
      case  3: return new Object[size[0]][size[1]][size[2]];
      case  4: return new Object[size[0]][size[1]][size[2]][size[3]];
      case  5: return new Object[size[0]][size[1]][size[2]][size[3]][size[4]];
      case  6: return new Object[size[0]][size[1]][size[2]][size[3]][size[4]][size[5]];
      case  7: return new Object[size[0]][size[1]][size[2]][size[3]][size[4]][size[5]][size[6]];
      case  8: return new Object[size[0]][size[1]][size[2]][size[3]][size[4]][size[5]][size[6]][size[7]];
      case  9: return new Object[size[0]][size[1]][size[2]][size[3]][size[4]][size[5]][size[6]][size[7]][size[8]];
      case 10: return new Object[size[0]][size[1]][size[2]][size[3]][size[4]][size[5]][size[6]][size[7]][size[8]][size[9]];
      case 11: return new Object[size[0]][size[1]][size[2]][size[3]][size[4]][size[5]][size[6]][size[7]][size[8]][size[9]][size[10]];
      case 12: return new Object[size[0]][size[1]][size[2]][size[3]][size[4]][size[5]][size[6]][size[7]][size[8]][size[9]][size[10]][size[11]];
      case 13: return new Object[size[0]][size[1]][size[2]][size[3]][size[4]][size[5]][size[6]][size[7]][size[8]][size[9]][size[10]][size[11]][size[12]];
      case 14: return new Object[size[0]][size[1]][size[2]][size[3]][size[4]][size[5]][size[6]][size[7]][size[8]][size[9]][size[10]][size[11]][size[12]][size[13]];
      case 15: return new Object[size[0]][size[1]][size[2]][size[3]][size[4]][size[5]][size[6]][size[7]][size[8]][size[9]][size[10]][size[11]][size[12]][size[13]][size[14]];
      case 16: return new Object[size[0]][size[1]][size[2]][size[3]][size[4]][size[5]][size[6]][size[7]][size[8]][size[9]][size[10]][size[11]][size[12]][size[13]][size[14]][size[15]];
      case 17: return new Object[size[0]][size[1]][size[2]][size[3]][size[4]][size[5]][size[6]][size[7]][size[8]][size[9]][size[10]][size[11]][size[12]][size[13]][size[14]][size[15]][size[16]];
      case 18: return new Object[size[0]][size[1]][size[2]][size[3]][size[4]][size[5]][size[6]][size[7]][size[8]][size[9]][size[10]][size[11]][size[12]][size[13]][size[14]][size[15]][size[16]][size[17]];
      case 19: return new Object[size[0]][size[1]][size[2]][size[3]][size[4]][size[5]][size[6]][size[7]][size[8]][size[9]][size[10]][size[11]][size[12]][size[13]][size[14]][size[15]][size[16]][size[17]][size[18]];
      case 20: return new Object[size[0]][size[1]][size[2]][size[3]][size[4]][size[5]][size[6]][size[7]][size[8]][size[9]][size[10]][size[11]][size[12]][size[13]][size[14]][size[15]][size[16]][size[17]][size[18]][size[19]];
    }
    throw new IllegalArgumentException("Dimension not supported: "+size.length);
  }

  /**
   * Liefert die Dimension eines multi-dimensionalen {@link Object}-Array.<br>
   * <b>Diese Methode funktioniert nur, wenn das Element 0 in jeder Dimension
   * belegt ist!</b>
   * @param array multi-dim. Array (Instanz von {@code Object[]..[]})
   * @exception ArrayIndexOutOfBoundsException wenn das 0te Element in einer
   *            Dimension mit {@code null} belegt ist
   */
  public static int getArrayDimension(Object array) {
    int dim = 0;
    try {
      for (;; dim++) {
        array = ((Object[])array)[0];
      }
    } catch (ClassCastException err) {
    }
    return dim;
  }

  /**
   * Liefert die Groessen der einzelnen Dimensionen eines multi-dimensionalen
   * {@link Object}-Array.<br>
   * <b>Diese Methode funktioniert nur, wenn das Element 0 in jeder Dimension
   * belegt ist!</b>
   * @param array multi-dim. Array (Instanz von {@code Object[]..[]})
   * @exception ArrayIndexOutOfBoundsException wenn das 0te Element in einer
   *            Dimension mit {@code null} belegt ist
   */
  public static int[] getArrayLength(Object array) {
    int[] dimSize = new int[ getArrayDimension(array) ];
    for (int d=0; d<dimSize.length; d++) {
      dimSize[d] = ((Object[])array).length;
      array = ( (Object[]) array)[0];
    }
    return dimSize;
  }

  /**
   * Liefert einen Wert aus einem multi-dimensionalen {@link Object}-Array.
   * @param array multi-dim. Array (Instanz von {@code Object[]..[]})
   * @param coord Koordinaten des Arrays
   * @exception ArrayIndexOutOfBoundsException wenn eine fehlerhafte Koordinate
   *            angegeben wird
   */
  public static Object getArrayValue(Object array, int... coord) {
    for (int i=0; i<coord.length; i++)
      array = ((Object[])array)[ coord[i] ];
    return array;
  }

  /**
   * Setzt einen Wert aus einem multi-dimensionalen {@link Object}-Array.
   * @param array multi-dim. Array (Instanz von {@code Object[]..[]})
   * @param value zu setzender Wert (Instanz von {@code Object[]..[]})
   * @param coord Koordinaten des Arrays
   * @exception ArrayIndexOutOfBoundsException wenn eine fehlerhafte Koordinate
   *            angegeben wird
   */
  public static void setArrayValue(Object array, Object value, int... coord) {
    for (int i=0; i<coord.length-1; i++)
      array = ((Object[])array)[ coord[i] ];
    ((Object[])array)[ coord[coord.length-1] ] = value;
  }

  /**
   * Kopiert einen Array (aber nicht dessen Elemente). Darueberhinaus koennen
   * (einige) Elemente direkt neu belegt werden. Ist das Flag {@code replNull}
   * auf {@code false} gesetzt, werden {@code null}-Werte in {@code replElem}
   * ignoriert. So ist es z.B. moeglich, nur das 4. und 5. Element im neuen
   * Array neu zu belegen und die davor liegenden Elemente unangetastet zu lassen.
   * @param source   zu kopierender Array
   * @param replNull Bestimmt, ob {@code null}-Elemente von {@code replElem}
   *                 im neuen Array gesetzt werden ({@code true}) oder nicht
   * @param replElem Elemente (beginnend bei Index 0), die nach dem Kopieren
   *                 im Array neu gesetzt werden
   * @return Array von der Groesse {@code max(source.length, replElem.length)}
   */
  public static <T> T[] cloneArray(T[] source, boolean replNull, T... replElem) {
    T[] newArray =  java.util.Arrays.copyOf(source, Math.max(source.length, replElem.length));
    for (int i=0; i<replElem.length; i++)
      if (replNull || replElem[i] != null)
        newArray[i] = replElem[i];
    return newArray;
  }

  /**
   * Kopiert einen Array (aber nicht dessen Elemente). Darueberhinaus koennen
   * (einige) Elemente direkt neu belegt werden.
   * @param source   zu kopierender Array
   * @param replElem Elemente (beginnend bei Index 0), die nach dem Kopieren
   *                 im Array neu gesetzt werden
   * @return Array von der Groesse {@code max(source.length, replElem.length)}
   */
  public static <T> T[] cloneArray(T[] source, T... replElem) {
    return cloneArray(source,true,replElem);
  }

  /**
   * Kopiert einen Array und belegt (einige) Elemente direkt neu.
   * @param source   zu kopierender Array
   * @param replElem Elemente (beginnend bei Index 0), die nach dem Kopieren
   *                 im Array neu gesetzt werden
   * @return Array von der Groesse {@code max(source.length, replElem.length)}
   */
  public static long[] cloneArray(long[] source, long... replElem) {
    long[] newArray =  java.util.Arrays.copyOf(source, Math.max(source.length, replElem.length));
    for (int i=0; i<replElem.length; i++)
      newArray[i] = replElem[i];
    return newArray;
  }

  /**
   * Kopiert einen Array und belegt (einige) Elemente direkt neu.
   * @param source   zu kopierender Array
   * @param replElem Elemente (beginnend bei Index 0), die nach dem Kopieren
   *                 im Array neu gesetzt werden
   * @return Array von der Groesse {@code max(source.length, replElem.length)}
   */
  public static int[] cloneArray(int[] source, int... replElem) {
    int[] newArray =  java.util.Arrays.copyOf(source, Math.max(source.length, replElem.length));
    for (int i=0; i<replElem.length; i++)
      newArray[i] = replElem[i];
    return newArray;
  }

  /**
   * Kopiert einen Array und belegt (einige) Elemente direkt neu.
   * @param source   zu kopierender Array
   * @param replElem Elemente (beginnend bei Index 0), die nach dem Kopieren
   *                 im Array neu gesetzt werden
   * @return Array von der Groesse {@code max(source.length, replElem.length)}
   */
  public static short[] cloneArray(short[] source, short... replElem) {
    short[] newArray =  java.util.Arrays.copyOf(source, Math.max(source.length, replElem.length));
    for (int i=0; i<replElem.length; i++)
      newArray[i] = replElem[i];
    return newArray;
  }

  /**
   * Kopiert einen Array und belegt (einige) Elemente direkt neu.
   * @param source   zu kopierender Array
   * @param replElem Elemente (beginnend bei Index 0), die nach dem Kopieren
   *                 im Array neu gesetzt werden
   * @return Array von der Groesse {@code max(source.length, replElem.length)}
   */
  public static byte[] cloneArray(byte[] source, byte... replElem) {
    byte[] newArray =  java.util.Arrays.copyOf(source, Math.max(source.length, replElem.length));
    for (int i=0; i<replElem.length; i++)
      newArray[i] = replElem[i];
    return newArray;
  }

  /**
   * Kopiert einen Array und belegt (einige) Elemente direkt neu.
   * @param source   zu kopierender Array
   * @param replElem Elemente (beginnend bei Index 0), die nach dem Kopieren
   *                 im Array neu gesetzt werden
   * @return Array von der Groesse {@code max(source.length, replElem.length)}
   */
  public static double[] cloneArray(double[] source, double... replElem) {
    double[] newArray =  java.util.Arrays.copyOf(source, Math.max(source.length, replElem.length));
    for (int i=0; i<replElem.length; i++)
      newArray[i] = replElem[i];
    return newArray;
  }

  /**
   * Kopiert einen Array und belegt (einige) Elemente direkt neu.
   * @param source   zu kopierender Array
   * @param replElem Elemente (beginnend bei Index 0), die nach dem Kopieren
   *                 im Array neu gesetzt werden
   * @return Array von der Groesse {@code max(source.length, replElem.length)}
   */
  public static float[] cloneArray(float[] source, float... replElem) {
    float[] newArray =  java.util.Arrays.copyOf(source, Math.max(source.length, replElem.length));
    for (int i=0; i<replElem.length; i++)
      newArray[i] = replElem[i];
    return newArray;
  }

  /**
   * Kopiert einen Array und belegt (einige) Elemente direkt neu.
   * @param source   zu kopierender Array
   * @param replElem Elemente (beginnend bei Index 0), die nach dem Kopieren
   *                 im Array neu gesetzt werden
   * @return Array von der Groesse {@code max(source.length, replElem.length)}
   */
  public static boolean[] cloneArray(boolean[] source, boolean... replElem) {
    boolean[] newArray =  java.util.Arrays.copyOf(source, Math.max(source.length, replElem.length));
    for (int i=0; i<replElem.length; i++)
      newArray[i] = replElem[i];
    return newArray;
  }

  /**
   * Kopiert einen Array und belegt (einige) Elemente direkt neu.
   * @param source   zu kopierender Array
   * @param replElem Elemente (beginnend bei Index 0), die nach dem Kopieren
   *                 im Array neu gesetzt werden
   * @return Array von der Groesse {@code max(source.length, replElem.length)}
   */
  public static char[] cloneArray(char[] source, char... replElem) {
    char[] newArray =  java.util.Arrays.copyOf(source, Math.max(source.length, replElem.length));
    for (int i=0; i<replElem.length; i++)
      newArray[i] = replElem[i];
    return newArray;
  }

  /**
   * Erweitert einen Array.
   * @param source   zu kopierender Array
   * @param newElem  Elemente, die an das Ende des Arrays angehaengt werden
   * @return Array von der Groesse {@code source.length + newElem.length}
   */
  public static <T> T[] extendArray(T[] source, T... newElem) {
    T[] newArray =  java.util.Arrays.copyOf(source, source.length+newElem.length);
    for (int i=0; i<newElem.length; i++)
      newArray[source.length+i] = newElem[i];
    return newArray;
  }

  /**
   * Erweitert einen Array.
   * @param source   zu kopierender Array
   * @param newElem  Elemente, die an das Ende des Arrays angehaengt werden
   * @return Array von der Groesse {@code source.length + newElem.length}
   */
  public static long[] extendArray(long[] source, long... newElem) {
    long[] newArray =  java.util.Arrays.copyOf(source, source.length+newElem.length);
    for (int i=0; i<newElem.length; i++)
      newArray[source.length+i] = newElem[i];
    return newArray;
  }

  /**
   * Erweitert einen Array.
   * @param source   zu kopierender Array
   * @param newElem  Elemente, die an das Ende des Arrays angehaengt werden
   * @return Array von der Groesse {@code source.length + newElem.length}
   */
  public static int[] extendArray(int[] source, int... newElem) {
    int[] newArray = java.util.Arrays.copyOf(source, source.length+newElem.length);
    for (int i=0; i<newElem.length; i++)
      newArray[source.length+i] = newElem[i];
    return newArray;
  }

  /**
   * Erweitert einen Array.
   * @param source   zu kopierender Array
   * @param newElem  Elemente, die an das Ende des Arrays angehaengt werden
   * @return Array von der Groesse {@code source.length + newElem.length}
   */
  public static short[] extendArray(short[] source, short... newElem) {
    short[] newArray =  java.util.Arrays.copyOf(source, source.length+newElem.length);
    for (int i=0; i<newElem.length; i++)
      newArray[source.length+i] = newElem[i];
    return newArray;
  }

  /**
   * Erweitert einen Array.
   * @param source   zu kopierender Array
   * @param newElem  Elemente, die an das Ende des Arrays angehaengt werden
   * @return Array von der Groesse {@code source.length + newElem.length}
   */
  public static byte[] extendArray(byte[] source, byte... newElem) {
    byte[] newArray =  java.util.Arrays.copyOf(source, source.length+newElem.length);
    for (int i=0; i<newElem.length; i++)
      newArray[source.length+i] = newElem[i];
    return newArray;
  }

  /**
   * Erweitert einen Array.
   * @param source   zu kopierender Array
   * @param newElem  Elemente, die an das Ende des Arrays angehaengt werden
   * @return Array von der Groesse {@code source.length + newElem.length}
   */
  public static double[] extendArray(double[] source, double... newElem) {
    double[] newArray =  java.util.Arrays.copyOf(source, source.length+newElem.length);
    for (int i=0; i<newElem.length; i++)
      newArray[source.length+i] = newElem[i];
    return newArray;
  }

  /**
   * Erweitert einen Array.
   * @param source   zu kopierender Array
   * @param newElem  Elemente, die an das Ende des Arrays angehaengt werden
   * @return Array von der Groesse {@code source.length + newElem.length}
   */
  public static float[] extendArray(float[] source, float... newElem) {
    float[] newArray =  java.util.Arrays.copyOf(source, source.length+newElem.length);
    for (int i=0; i<newElem.length; i++)
      newArray[source.length+i] = newElem[i];
    return newArray;
  }

  /**
   * Erweitert einen Array.
   * @param source   zu kopierender Array
   * @param newElem  Elemente, die an das Ende des Arrays angehaengt werden
   * @return Array von der Groesse {@code source.length + newElem.length}
   */
  public static boolean[] extendArray(boolean[] source, boolean... newElem) {
    boolean[] newArray =  java.util.Arrays.copyOf(source, source.length+newElem.length);
    for (int i=0; i<newElem.length; i++)
      newArray[source.length+i] = newElem[i];
    return newArray;
  }

  /**
   * Erweitert einen Array.
   * @param source   zu kopierender Array
   * @param newElem  Elemente, die an das Ende des Arrays angehaengt werden
   * @return Array von der Groesse {@code source.length + newElem.length}
   */
  public static char[] extendArray(char[] source, char... newElem) {
    char[] newArray =  java.util.Arrays.copyOf(source, source.length+newElem.length);
    for (int i=0; i<newElem.length; i++)
      newArray[source.length+i] = newElem[i];
    return newArray;
  }

  /**
   * Liefert die Klassen-Namen zu einem Array von Klassen.
   * @param clazz  Klassen
   * @param simple gibt an, ob der einfache Klassenname ({@code true}) oder der
   *               komplette Klassenname inkl. Paketname ({@code false})
   *               zurueckgegeben wird.
   */
  public static String[] getClassNames(Class[] clazz, boolean simple) {
    String[] className = new String[clazz.length];
    for (int i=0; i<className.length; i++)
      className[i] = simple ? clazz[i].getSimpleName() : clazz[i].getName();
    return className;
  }

  /**
   * Ruft die Java Garbage Collection auf.
   * @return Anzahl an Bytes, die durch den Aufruf der Garbage Collection
   *         freigegeben wurden
   * @see Runtime#gc()
   * @see Runtime#freeMemory()
   */
  public static long gc() {
    Runtime runtime = Runtime.getRuntime();
    long prevFreeMem = runtime.freeMemory();
    runtime.gc();
    return runtime.freeMemory() - prevFreeMem;
  }

  /**
   * Ruft die Java Garbage Collection solange auf, bis kein Speicher mehr
   * freigegeben werden kann.
   * @return Gesamt-Anzahl an Bytes, die durch die Aufrufe der Garbage Collection
   *         freigegeben wurden
   * @see #gc()
   */
  public static long gcTotal() {
    long totalFreedMem = 0;
    for ( long freedMem=0; (freedMem = gc()) > 0; )
     totalFreedMem += freedMem;
    return totalFreedMem;
  }

  /**
   * Erzeugt einen neuen {@link URLClassLoader}, der vom
   * {@linkplain ClassLoader#getSystemClassLoader() System-ClassLoaders}
   * abgeleitet ist.
   * @param searchPath Pfade, in denen der ClassLoader nach Klassen sucht
   */
  public static ClassLoader createClassLoader(URL[] searchPath) {
    return URLClassLoader.newInstance( searchPath, ClassLoader.getSystemClassLoader() );
  }

  /**
   * Erzeugt einen neuen {@link URLClassLoader}, der vom
   * {@linkplain ClassLoader#getSystemClassLoader() System-ClassLoaders}
   * abgeleitet ist.
   * @param searchPath Pfad, in denm der ClassLoader nach Klassen sucht
   */
  public static ClassLoader createClassLoader(URL searchPath) {
    return createClassLoader( new URL[] {searchPath} );
  }

  /**
   * Laed eine Klasse. Ggf. wird hierfuer ein neuer {@link ClassLoader}
   * erzeugt.
   * @param clazz      Klassenname (inkl. Paket)
   * @param searchPath Pfade, in denen der ClassLoader nach Klassen sucht
   *                   (wenn <code>null</code>, wird der
   *                   {@linkplain ClassLoader#getSystemClassLoader() System-ClassLoader}
   *                   zum Laden der Klasse verwendet)
   */
  public static Class loadClass(String clazz, URL[] searchPath) throws ClassNotFoundException {
    if ( searchPath == null )
      return Class.forName(clazz);
    return createClassLoader(searchPath).loadClass(clazz);
  }

  /**
   * Laed eine Klasse. Ggf. wird hierfuer ein neuer {@link ClassLoader}
   * erzeugt.
   * @param clazz      Klassenname (inkl. Paket)
   * @param searchPath Pfad, in dem der ClassLoader nach Klassen sucht
   *                   (wenn <code>null</code>, wird der
   *                   {@linkplain ClassLoader#getSystemClassLoader() System-ClassLoader}
   *                   zum Laden der Klasse verwendet)
   */
  public static Class loadClass(String clazz, URL searchPath) throws ClassNotFoundException {
    if ( searchPath == null )
      return Class.forName(clazz);
    return createClassLoader(searchPath).loadClass(clazz);
  }

  /**
	 *
	 * Laed eine Klasse derart das Klassen im Cache ignoriert werden. Dazu ist es erforderlich
	 * das unter dem Parameter <code>searchPath</code> eine Verzeichnis "META-INF" vorhanden ist.
	 * Darin muss eine Datei "PREFERRED.LIST" sein. Wie genau diese Datei aufgebaut sein muss, lässt
	 * sich in der Dokumentation des {@link PreferredClassLoader PreferredClassLoaders} nachlesen.
	 *
	 * @param classname  Klassenname (inkl. Paket)
	 * @param searchPath Pfad, in dem der ClassLoader nach Klassen sucht
	 *                   (wenn <code>null</code>, wird der System-Klassenpfad
	 *                   {@linkplain ClassLoader#getSystemResource(String) ClassLoader.getSystemResource("")}
	 *                   zum Laden der Klasse verwendet)
	 * @throws ClassNotFoundException
	 * @see PreferredClassLoader
	 *
	 *  @author Dominik Appl
	 */
	public static Class loadPreferredClass(String classname, URL searchPath)
			throws ClassNotFoundException {
                if ( searchPath == null )
                  // System-Klassenpfad verwenden
                  searchPath = ClassLoader.getSystemResource("");
		PreferredClassLoader classLoader = new PreferredClassLoader(
				new URL[]{searchPath}, Thread.currentThread()
						.getContextClassLoader(), null, false);
		Class loadedClass = classLoader.loadClass(classname);
		if (loadedClass == null)
			return Class.forName(classname);

		return loadedClass;
	}

  /**
   * Laed eine Klasse neu und erzeugt eine Objekt-Instanz. Als Suchpfad fuer
   * die Klasse der System-Klassenpfad
   * {@linkplain ClassLoader#getSystemResource(String) ClassLoader.getSystemResource("")}
   * verwendet.
   * @param classname Klassenname (inkl. Paket)
   * @param param Parameter fuer den Konstruktor-Aufruf
   * @exception RuntimeException falls irgendein Fehler auftritt
   */
  public static Object loadPreferredObject(String classname, Object... param) {
    return loadPreferredObject(classname,null,param);
  }

  /**
   * Laed eine Klasse neu und erzeugt eine Objekt-Instanz.
   * @param classname Klassenname (inkl. Paket)
   * @param searchPath Pfad, in dem der ClassLoader nach Klassen sucht
   *                   (wenn <code>null</code>, wird der System-Klassenpfad
   *                   {@linkplain ClassLoader#getSystemResource(String) ClassLoader.getSystemResource("")}
   *                   zum Laden der Klasse verwendet)
   * @param param Parameter fuer den Konstruktor-Aufruf
   * @exception RuntimeException falls irgendein Fehler auftritt
   */
  public static Object loadPreferredObject(String classname, URL searchPath, Object... param) {
    try {
      // Klasse laden
      Class clazz = loadPreferredClass(classname,searchPath);
//      // Klassen der Konstruktor-Parameter ermitteln
//      Class[] paramClass = new Class[ param.length ];
//      for (int i=0; i<paramClass.length; i++)
//        paramClass[i] = param[i].getClass();
//      // Konstruktor fuer Parameter ermitteln (erfordert exakes Matching
//      // zwischen den Klassen der uebergebenen Parameter und der Klassen
//      // der Konstruktor-Parameter!)
//      Constructor constructor = clazz.getConstructor(paramClass);
      // Konstruktor fuer Parameter ausprobieren
      for ( Constructor constructor : clazz.getConstructors() )
        try {
          return constructor.newInstance(param);
        } catch (IllegalArgumentException err) {
        }
        // kein passender Konstruktor gefunden
        throw new UnsupportedOperationException("No matching constructor found...");
    } catch (Exception err) {
      throw new RuntimeException(err);
    }
  }


  /**
   * Liefert das Package einer Klasse erweitert um einen relativen Pfad
   * in Punkt-Notation.
   * @param clazz eine Klasse
   * @param pathExt Erweiterung des Pfads
   */
  public static String extendPackagePath(Class clazz, String pathExt) {
    String packagePath = clazz.getPackage().getName();
    if ( pathExt != null && !pathExt.trim().equals("") )
      packagePath += "." + pathExt;
    return packagePath;
  }


  /**
   * Prueft einen Identifier auf Korrektheit.
   * @param identifier Identifier
   * @exception IllegalArgumentException falls der Identifier nicht-erlaubte
   *            Zeichen enthaelt
   * @see Character#isJavaIdentifierStart(char)
   * @see Character#isJavaIdentifierPart(char)
   */
  public static void checkIdentifierAndError(String identifier) {
    if ( !Character.isJavaIdentifierStart( identifier.charAt(0) ) )
      throw new IllegalArgumentException("Illegal first character '"+identifier.charAt(0)+"'");
    for (int i=1; i<identifier.length(); i++) {
      char c = identifier.charAt(i);
      if ( !Character.isJavaIdentifierPart(c) )
        throw new IllegalArgumentException("Illegal inner character '"+c+"'");
    }
  }

  /**
   * Prueft einen Identifier auf Korrektheit.
   * @param identifier Identifier
   * @see #checkIdentifierAndError(String)
   */
  public static boolean checkIdentifier(String identifier) {
    try {
      checkIdentifierAndError(identifier);
      return true;
    } catch ( Exception err ) {
      return false;
    }
  }

  /**
   * Sortiert einen Array aus beliebigen Objekten nach der natuerlichen
   * Sortierung ihrer {@code toString()}-Werte.
   * @param objects Objekte
   */
  public static void sort(Object[] objects) {
    Arrays.sort( objects, new Comparator<Object>() {
      public int compare(Object c1, Object c2) {
        return c1.toString().compareTo(c2.toString());
      }
    });
  }
  
  /**
   * Returns the maximum value of some integer collections.
   * @param defMaxValue value taken as maximum if a collection is empty
   * @param collections some collections
   */
  public static int max(int defMaxValue, Collection<Integer>... collections) {
    int maxValue = defMaxValue;
    for (Collection<Integer> collection : collections) {
      int collMax = defMaxValue;
      if ( collection != null && !collection.isEmpty() )
        collMax = Collections.max(collection);
      maxValue = Math.max( maxValue, collMax );
    }
    return maxValue;
  }
  
  /**
   * Returns the maximum value of some integer collections. If
   * a collection is empty, 0 is taken as its maximum value.
   * @param collections some collections
   */
  public static int max(Collection<Integer>... collections) {
    return max(0,collections);
  }

  /**
   * Returns the minimum value of some integer collections.
   * @param defMinValue value taken as minimum if a collection is empty
   * @param collections some collections
   */
  public static int min(int defMinValue, Collection<Integer>... collections) {
    int minValue = defMinValue;
    for (Collection<Integer> collection : collections) {
      int collMin = defMinValue;
      if ( collection != null && !collection.isEmpty() )
        collMin = Collections.min(collection);
      minValue = Math.min( minValue, collMin );
    }
    return minValue;
  }
  
  /**
   * Returns the minimum value of some integer collections. If
   * a collection is empty, 0 is taken as its minimum value.
   * @param collections some collections
   */
  public static int min(Collection<Integer>... collections) {
    return min(0,collections);
  }

  /**
   * Rundet einen Wert auf Nach- oder Vorkommastellen.
   * @param value Wert
   * @param digits Anzahl an Nachkommastellen (negative Werte runden auf
   *               Vorkommastellen, also 10er-, 100er-, usw. Stelle)
   * @param mode Modus für das Runden (0 = normal Runden, 1 = Aufrunden, -1 = Abrunden;
   *             wenn nicht angegeben, wird normal gerundet)
   */
  public static double round(double value, int digits, int... mode) {
    int roundMode = 0; // Default: Normal-Runden
    if ( mode.length > 0 )
      roundMode = mode[0];

    // 10er-Potenz fuer Nachkommestellen bilden um vor dem Runden
    // die Nachkommastellen "vor das Komma" zu schieben
    double digitsFactor = Math.pow(10, digits);
    value = value * digitsFactor;
    // Runden
    if ( roundMode < 0 )
      value = Math.floor(value);
    else if ( roundMode > 0 )
      value = Math.ceil(value);
    else
      value = Math.round(value);

    // Nachkommastellen wieder hinter das Komma schieben
    return value / digitsFactor;
  }

  /**
   * Rundet alle Werte einer {@link Collection} auf Nach- oder Vorkommastellen.<br>
   * <b>Achtung:</b><br>
   * Es wird eine neue {@link Collection} des gleichen Typs erzeugt und die
   * gerundeten Werte darin in der Reihenfolge eingefuegt, wie sie von
   * {@link Collection#iterator()} geliefert werden! Unter umstaenden kann
   * die urspruengliche Sortierung verloren gehen! Diese Methode sollte also
   * im Zweifelsfall nur auf automatisch sortierte {@link Collection Collections}
   * angewandt werden (z.B. {@link TreeSet}).
   * @param values Werte
   * @param digits Anzahl an Nachkommastellen (negative Werte runden auf
   *               Vorkommastellen, also 10er-, 100er-, usw. Stelle)
   * @param mode Modus für das Runden (0 = normal Runden, 1 = Aufrunden, -1 = Abrunden;
   *             wenn nicht angegeben, wird normal gerundet)
   */
  public static Collection<Number> round(Collection<Number> values, int digits, int... mode) {
    try {
      Collection<Number> newCollection = values.getClass().newInstance();
      for (Number value : values) {
        newCollection.add( round(value.doubleValue(),digits,mode) );
      }
      return newCollection;
    } catch (Exception err) {
      throw new IllegalArgumentException("Can not handle collection: "+getSimpleClassName(values),err);
    }
  }
  
  /**
   * Puts a text to system clipboard.
   * @param str a text
   */
  public static void copyToClipboard(String str) {
    if ( str == null )
      return;
    StringSelection sel = new StringSelection( str );
    Toolkit.getDefaultToolkit().getSystemClipboard().setContents(sel,sel);
  }
  
  
  /**
   * Returns the class names of the top of a stack trace. Duplicate class names
   * are eliminated.
   * @param stackTrace stack trace to explore
   * @param depth      maximum count of classes, which are explored in the stack
   *                   trace
   */
  public static String[] getLastStackTraceClasses(StackTraceElement[] stackTrace, int depth) {
    Vector<String> lastClasses = new Vector<String>();
    for (StackTraceElement st : stackTrace) {
      String className = st.getClassName();
      if ( !lastClasses.contains(className) )
        lastClasses.add(className);
      if ( lastClasses.size() >= depth )
        break;
    }
    return lastClasses.toArray(new String[0]);
  }

  /**
   * Prints the last X classes from a stack trace.
   * @param stackTrace stack trace to explore
   * @param depth      maximum count of classes, which are explored in the stack
   *                   trace
   * @param out        output stream to write to
   */
  public static void printLastStackTraceClasses(StackTraceElement[] stackTrace, int depth, PrintStream out) {
    String[] lastClasses = getLastStackTraceClasses(stackTrace, depth);
    for (int i=0; i<lastClasses.length; i++) {
      if (i > 0)
        out.print("\t...");
      out.println(lastClasses[i]);
    }
  }

  /**
   * Prints the last X classes from a stack trace to the standard output.
   * @param stackTrace stack trace to explore
   * @param depth      maximum count of classes, which are explored in the stack
   *                   trace
   * @see System#out
   */
  public static void printLastStackTraceClasses(StackTraceElement[] stackTrace, int depth) {
    printLastStackTraceClasses(stackTrace, depth, System.out);
  }

  /**
   * Erzeugt einen Log4j-Logger fuer ein Objekt. Als Identifier fuer den Logger
   * wird der Klassenname des Objekts verwendet.
   * @param object ein Objekt
   * @return Logger mit dem Namen "NULL", falls das uebergebene Objekt {@code null}
   *         ist
   */
  public static Logger createLogger(Object object) {
    if ( object == null )
      return Logger.getLogger("");
    if ( object instanceof Class )
      return Logger.getLogger( ((Class)object).getName() );
    return Logger.getLogger( object.getClass().getName() );
  }

  /**
   * Erzeugt einen Error- und einen Debug-Eintrag in einem Logger. Der Error-Eintrag
   * erhaelt nur die Meldung, der Debug-Eintrag auch die detaillierten Informationen
   * des Fehlers.
   * @param logger ein Logger
   * @param mess   Fehlermeldung (wenn {@code null} wird die Message des Fehlers ausgegeben)
   * @param err    ein Fehler
   */
  public static void logDebugError(Category logger, Object mess, Throwable err) {
    if ( mess == null )
      mess = err.getClass().getName()+": "+err.getMessage();
    logger.error(mess);
    logger.debug(mess,err);
  }

  /**
   * Erzeugt einen Error- und einen Debug-Eintrag in einem Logger. Der Error-Eintrag
   * erhaelt nur die Meldung, der Debug-Eintrag auch die detaillierten Informationen
   * des Fehlers.
   * @param logger ein Logger
   * @param err    ein Fehler
   */
  public static void logDebugError(Category logger, Throwable err) {
    logDebugError(logger,null,err);
  }

  /**
   * Erzeugt einen Warn- und einen Debug-Eintrag in einem Logger. Der Warn-Eintrag
   * erhaelt nur die Meldung, der Debug-Eintrag auch die detaillierten Informationen
   * des Fehlers.
   * @param logger ein Logger
   * @param mess   Fehlermeldung (wenn {@code null} wird die Message des Fehlers ausgegeben)
   * @param err    ein Fehler
   */
  public static void logDebugWarn(Category logger, Object mess, Throwable err) {
    if ( mess == null )
      mess = err.getClass().getName()+": "+err.getMessage();
    logger.warn(mess);
    logger.debug(mess,err);
  }

  /**
   * Erzeugt einen Warn- und einen Debug-Eintrag in einem Logger. Der Warn-Eintrag
   * erhaelt nur die Meldung, der Debug-Eintrag auch die detaillierten Informationen
   * des Fehlers.
   * @param logger ein Logger
   * @param err    ein Fehler
   */
  public static void logDebugWarn(Category logger, Throwable err) {
    logDebugWarn(logger,null,err);
  }

  /**
   * Logs the last X classes from a stack trace as debug message.
   * @param logger     the logger
   * @param stackTrace stack trace to explore
   * @param depth      maximum count of classes, which are explored in the stack
   *                   trace
   */
  public static void logDebugLastStackTraceClasses(Category logger, StackTraceElement[] stackTrace, int depth) {
    String[] lastClasses = getLastStackTraceClasses(stackTrace, depth);
    for (int i=0; i<lastClasses.length; i++) {
      String message = "";    
      if (i > 0)
        message = "\t...";
      message += lastClasses[i];
      logger.debug(message);
    }
  }

}

