/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.lang;

import java.util.Collection;
import java.util.Vector;

import schmitzm.data.event.AbstractObjectTraceable;
import schmitzm.data.event.GeneralObjectChangeEvent;
import schmitzm.data.event.Invoker;
import schmitzm.data.event.ObjectChangeEvent;
import schmitzm.data.event.ObjectEvent;
import schmitzm.data.event.ObjectListener;
import schmitzm.data.event.ObjectTraceable;

/**
 * Diese Klasse stellt einen {@link Vector} dar, der auf eine bestimmte
 * Anzahl an Elementen begrenzt ist. Wird die maximale Anzahl ueberschritten
 * wird automatisch das <b>ERSTE</b> Element der Liste entfernt.<br>
 * Der {@code LimitedVector} eignet sich z.B. gut zum Speichern der letzten
 * 10 geladenen Dateien.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class LimitedVector<E> extends Vector<E> implements ObjectTraceable {
  /** Speichert die Anzahl der Elemente, die die Liste aufnehmen kann. */
  protected int maxLoad = 0;

  /** Flag, ob der Vector ein Objekt mehrmals enthalten darf.<br>
   *  Standard: {@code true} */
  protected boolean duplAllowed = true;

  // Einfache Implementierung von ObjectTraceable, um die Methoden der
  // Listener-Verwaltung nicht doppelt implementieren zu muessen
  private AbstractObjectTraceable objectTraceableProxy = new AbstractObjectTraceable() {
  };

  /**
   * Erzeugt eine leere Liste.
   * @param maxLoad maximale Anzahl an Elementen, die die Liste aufnehmen kann
   */
  public LimitedVector(int maxLoad) {
    super(maxLoad);
    setMaxLoad(maxLoad);
  }

  /**
   * Liefert die maximale Anzahl an Elementen, die die Liste aufnehmen kann.
   */
  public int getMaxLoad() {
    return maxLoad;
  }

  /**
   * Setzt die maximale Anzahl an Elementen, die die Liste aufnehmen kann.
   * Ist der neue Wert kleiner als der alte, so werden alle ueberzaehligen
   * Elemente sofort entfernt!
   * @param maxLoad Maximal-Anzahl an Elementen
   */
  public void setMaxLoad(int maxLoad) {
    this.maxLoad = maxLoad;
    truncate();
  }

  /**
   * Loescht solange das letzte Element der Liste, bis die maximale Anzahl
   * an Elementen erreicht ist.
   */
  protected void truncate() {
    while( size() > maxLoad )
      removeElementAt( 0 );
  }

  /**
   * Prueft, ob die Liste ein Objekt mehrmals enthalten darf.
   * @see #setDuplicatedAllowed(boolean)
   */
  public boolean getDuplicatedAllowed() {
    return this.duplAllowed;
  }

  /**
   * Bestimmt, ob die Liste ein Objekt mehrmals enthalten darf.<br>
   * Standard: {@code false}<br>
   * <b>Bemerkung:</b><br>
   * Die Aenderung dieser Eigenschaft wirkt sich NICHT auf den aktuellen Inhalt
   * der Liste aus. Bestehende Duplikate in der Liste bleiben erhalten!!
   * @param duplAllowed bei {@code false} wird vor dem Einfuegen ein etwaiges
   *                    Duplikat aus der Liste entfernt
   */
  public void setDuplicatedAllowed(boolean duplAllowed) {
    this.duplAllowed = duplAllowed;
  }

  ///////////////////////////////////////////////////////////////////
  /////////////////   Ueberschreiben von Vector   ///////////////////
  ///////////////////////////////////////////////////////////////////
  /**
   * Fuegt ein Element in die Liste ein. Danach werden alle ueberzaehligen
   * Element geloescht. Erzeugt ein {@link ObjectChangeEvent} falls sich der
   * Inhalt des Vectors aendert.<br>
   * <b>Bemerkung:</b><br>
   * Diese Methode macht nichts, falls sie bei bereits voll gefuellter Liste
   * mit Index 0 aufgerufen wird, da in diesem Fall das neue Element sofort
   * ueberzaehlig ist.<br>
   * <b>ACHTUNG:</b><br>
   * Bei automatischer Duplikat-Eleminierung kann diese Methode zu Problemen
   * fuehren, da ERST die Duplikate geloescht werden und dann die Einfuege-
   * Operation vorgenommen wird. {@code index} ist dann evt. nicht mehr die
   * gewuenschte Einfuege-Position!!
   * @param obj einzufuegendes Objekt
   * @param index Index an dem das Element eingefuegt wird
   */
  public synchronized void insertElementAt(E obj, int index) {
    if ( size() >= getMaxLoad() && index == 0 )
      return;

    // etwaige Duplikate entfernen
    if ( !duplAllowed )
      while( remove(obj) );

    int oldSize = size(); //
    super.insertElementAt(obj,index);
    if ( size() != oldSize ) {
      truncate();
      fireEvent( new ObjectChangeEvent(new Invoker(this),null,obj) );
    }
  }

  /**
   * Belegt eine Listenposition neu und erzeugt ein
   * {@link ObjectChangeEvent}, falls sich der Listen-Inhalt aendert.<br>
   * <b>ACHTUNG:</b><br>
   * Bei automatischer Duplikat-Eleminierung kann diese Methode zu Problemen
   * fuehren, da ERST die Duplikate geloescht werden und dann die Set-
   * Operation vorgenommen wird. {@code index} ist dann evt. nicht mehr die
   * gewuenschte Set-Position!!
   * @param obj neues Listen-Element
   * @param index Listen-Position
   */
  public synchronized void setElementAt(E obj, int index) {
    // etwaige Duplikate entfernen
    if ( !duplAllowed )
      while( remove(obj) );

    E oldObj = this.elementAt(index);
    super.setElementAt(obj,index);
    if ( oldObj != obj )
      fireEvent( new ObjectChangeEvent(new Invoker(this),oldObj,obj) );
  }

  /**
   * Belegt eine Listenposition neu und erzeugt ein
   * {@link ObjectChangeEvent}, falls sich der Listen-Inhalt aendert.<br>
   * <b>ACHTUNG:</b><br>
   * Bei automatischer Duplikat-Eleminierung kann diese Methode zu Problemen
   * fuehren, da ERST die Duplikate geloescht werden und dann die Set-
   * Operation vorgenommen wird. {@code index} ist dann evt. nicht mehr die
   * gewuenschte Set-Position!!
   * @param obj neues Listen-Element
   * @param index Listen-Position
   */
  public synchronized E set(int index, E obj) {
    // etwaige Duplikate entfernen
    if ( !duplAllowed )
      while( remove(obj) );

    E oldObj = super.set(index,obj);
    if ( oldObj != obj )
      fireEvent( new ObjectChangeEvent(new Invoker(this),oldObj,obj) );
    return oldObj;
  }

  /**
   * Fuegt ein Element in am Ende der Liste ein. Danach werden alle ueberzaehligen
   * Element geloescht. Erzeugt ein {@link ObjectChangeEvent} falls
   * sich der Inhalt des Vectors aendert.
   * @param obj einzufuegendes Objekt
   */
  public synchronized void addElement(E obj) {
    // etwaige Duplikate entfernen
    if ( !duplAllowed )
      while( remove(obj) );

    int oldSize = size();
    super.addElement(obj);
    if ( size() != oldSize ) {
      truncate();
      fireEvent( new ObjectChangeEvent(new Invoker(this),null,obj) );
    }
  }

  /**
   * Fuegt ein Element am Ende der Liste ein. Danach werden alle ueberzaehligen
   * Element geloescht. Erzeugt ein {@link ObjectChangeEvent} falls
   * sich der Inhalt des Vectors aendert.
   * @param o einzufuegendes Objekt
   */
  public synchronized boolean add(E o) {
    // etwaige Duplikate entfernen
    if ( !duplAllowed )
      while( remove(o) );

    boolean changed = super.add(o);
    if ( changed ) {
      truncate();
      fireEvent( new ObjectChangeEvent(new Invoker(this),null,o) );
    }
    return changed;
  }

  /**
   * Fuegt alle Elemente einer {@link Collection} am Ende der Liste ein.
   * Danach werden alle ueberzaehligen Element geloescht.  Erzeugt ein
   * {@link GeneralObjectChangeEvent} falls sich der Inhalt des Vectors aendert.
   * @param c einzufuegende Objekte
   */
  public synchronized boolean addAll(Collection<? extends E> c) {
    // etwaige Duplikate entfernen
    if ( !duplAllowed )
      while( removeAll(c) );

    boolean changed = super.addAll(c);
    if ( changed ) {
      truncate();
      fireEvent( new GeneralObjectChangeEvent(this) );
    }
    return changed;
  }

  /**
   * Fuegt alle Elemente einer {@link Collection} in die Liste ein.
   * Danach werden alle ueberzaehligen Element geloescht. Erzeugt ein
   * {@link GeneralObjectChangeEvent} falls sich der Inhalt des Vectors aendert.<br>
   * <b>Bemerkung:</b><br>
   * Diese Methode macht nichts, falls sie bei bereits voll gefuellter Liste
   * mit Index 0 aufgerufen wird, da in diesem Fall die neuen Elemente auch
   * die ueberzaehligen sind.
   * @param index Listen-Position ab der die Elemente eingefuegt werden
   * @param c einzufuegende Objekte
   */
  public synchronized boolean addAll(int index, Collection<? extends E> c) {
    // etwaiges Duplikat entfernen
    if ( !duplAllowed )
      while( removeAll(c) );

    if ( size() >= getMaxLoad() && index == 0 )
      return false;

    boolean changed = super.addAll(index,c);
    if ( changed ) {
      truncate();
      fireEvent( new GeneralObjectChangeEvent(this) );
    }
    return changed;
  }

  /**
   * Verhaelt sich wie {@link Vector#trimToSize()} und erzeugt ein
   * {@link GeneralObjectChangeEvent}.
   */
  public synchronized void trimToSize() {
    super.trimToSize();
    fireEvent( new GeneralObjectChangeEvent(this) );
  }

  /**
   * Verhaelt sich wie {@link Vector#setSize(int)} und erzeugt ein
   * {@link GeneralObjectChangeEvent}.
   */
  public synchronized void setSize(int newSize) {
    super.setSize(newSize);
    fireEvent( new GeneralObjectChangeEvent(this) );
  }

  /**
   * Verhaelt sich wie {@link Vector#removeElementAt(int)} und erzeugt ein
   * {@link ObjectChangeEvent}, falls sich der Listen-Inhalt aendert.
   */
  public synchronized void removeElementAt(int index) {
    E oldObj = this.elementAt(index);
    super.removeElementAt(index);
    fireEvent( new ObjectChangeEvent(new Invoker(this),oldObj,null) );
  }

  /**
   * Verhaelt sich wie {@link Vector#removeAllElements()} und erzeugt ein
   * {@link GeneralObjectChangeEvent}, falls sich der Listen-Inhalt aendert.
   */
  public synchronized void removeAllElements() {
    if ( size() == 0 )
      return;
    super.removeAllElements();
    fireEvent( new GeneralObjectChangeEvent(this) );
  }

  /**
   * Verhaelt sich wie {@link Vector#remove(int)} und erzeugt ein
   * {@link ObjectChangeEvent}, falls sich der Listen-Inhalt aendert.
   */
  public synchronized E remove(int index) {
    E oldObj = super.remove(index);
    fireEvent( new ObjectChangeEvent(new Invoker(this),oldObj,null) );
    return oldObj;
  }

  /**
   * Verhaelt sich wie {@link Vector#removeAll(Collection)} und erzeugt ein
   * {@link GeneralObjectChangeEvent}, falls sich der Listen-Inhalt aendert.
   */
  public synchronized boolean removeAll(Collection<?> c) {
    boolean changed =  super.removeAll(c);
    if ( changed )
      fireEvent(new GeneralObjectChangeEvent(this));
    return changed;
  }

  /**
   * Verhaelt sich wie {@link Vector#retainAll(Collection)} und erzeugt ein
   * {@link GeneralObjectChangeEvent}, falls sich der Listen-Inhalt aendert.
   */
  public synchronized boolean retainAll(Collection<?> c)  {
    boolean changed =  super.retainAll(c);
    if ( changed )
      fireEvent(new GeneralObjectChangeEvent(this));
    return changed;
  }

  /**
   * Verhaelt sich wie {@link Vector#removeRange(int,int)} und erzeugt ein
   * {@link GeneralObjectChangeEvent}, falls sich der Listen-Inhalt aendert.
   */
  protected synchronized void removeRange(int fromIndex, int toIndex) {
    int oldSize = size();
    super.removeRange(fromIndex,toIndex);
    if ( size() != oldSize )
      fireEvent(new GeneralObjectChangeEvent(this));
  }

  ///////////////////////////////////////////////////////////////////
  ////////////   Implementierung von ObjectTraceable   //////////////
  ///////////////////////////////////////////////////////////////////
  /**
   * Fuegt dem Objekt einen Listener hinzu, der bei Aenderungen informiert wird.
   */
  public void addObjectListener(ObjectListener listener) {
    this.objectTraceableProxy.addObjectListener(listener);
  }

  /**
   * Entfernt einen Listener von dem Objekt.
   */
  public void removeObjectListener(ObjectListener listener) {
    this.objectTraceableProxy.removeObjectListener(listener);
  }

  /**
   * Checkt, ob der angegebene Listener dem Objekt zugeordnet ist.
   */
  public boolean containsObjectListener(ObjectListener l) {
    return this.objectTraceableProxy.containsObjectListener(l);
  }

  /**
   * Liefert alle Listener eines bestimmten Typs.
   * @param type Art des Listeners (Filter)
   */
  public ObjectListener[] getObjectListener(Class type) {
    return this.objectTraceableProxy.getObjectListener(type);
  }

  /**
   * Informiert alle Listener, dass sich das Objekt (this) geaendert hat.
   */
  public void fireEvent(ObjectEvent e) {
    this.objectTraceableProxy.fireEvent(e);
  }

  /**
   * Informiert alle Listener eine bestimmten Klasse, dass sich das Objekt
   * (this) geaendert hat.
   */
  public void fireEvent(ObjectEvent e, Class c) {
    this.objectTraceableProxy.fireEvent(e, c);
  }

}
