/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.lang;

import java.lang.reflect.Array;

/**
 * Diese Klasse stellt ein Pendant zur Klasse {@link java.lang.reflect.Array} und
 * stellt direkten Element-Zugriff fuer mehrdimensionales Arrays zur Verfuegung.<br>
 * <b>Bemerke:</b><br>
 * Leider kann diese Klasse nicht von <code>java.lang.reflect.Array</code>
 * abgeleitet werden, da diese als final deklariert ist!!
 * @see java.lang.reflect.Array
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class MultiDimArray {

  /**
   * Erzeugt einen neuen mehrdimensionalen Array.
   * Ist aequivalent zu <code>java.lang.reflect.Array.newInstance(Class,int[])</code>.
   * @param componentType Typ, den der Array speichern kann. Kann auch ein build-in-Type
   *                      sein (z.B. <code>int.class</code>)
   * @param size          Groesse des Arrays in allen Dimensionen
   * @see java.lang.reflect.Array#newInstance(Class,int[])
   */
  public static Object newInstance(Class componentType, int[] size) {
    return Array.newInstance(componentType, size);
  }

  /**
   * Erzeugt einen native {@code Object[]...[]} aus einem {@code MultiDimArray}.
   * @param multiDimArray ein ueber {@code MultiDimArray#newInstance(.)} oder
   *                      {@code Array.newInstance(.)} erzeugter Array
   * @return ein {@code Object[]...[]}
   */
  public static Object wrapToNativeArray(Object multiDimArray) {
    int[] dimSizes = getLength(multiDimArray); // Size of the dimensions
    Object nativeArray = LangUtil.createArray(dimSizes);

    // Calculate the total number of elements in the matrix
    int elemCount = 1;
    for (int dimSize : dimSizes)
      elemCount *= dimSize;

    // put the elements in the array
    int[] coord = new int[dimSizes.length];
    for (int i = 0; i < elemCount; i++) {
      // set the element at coordinate tCoord
      LangUtil.setArrayValue(
          nativeArray,
          get(multiDimArray, coord),
          coord
      );

      // increment the coordinate for the element
      for (int d = 1; d < dimSizes.length; d++)
        if (++coord[d] < dimSizes[d])
          break; // Coordinates in dimensionen > d not influenced by the increment
        else
          coord[d] = 0; // Initialize coordinate in dimension d and continue
                        // to increment in dimension d+1 erhoehen
    }

    return nativeArray;
  }

  /**
   * Liefert den Array, in dem die angegebene Koordinate zu finden ist.
   * Der Wert an der Koordinate kann dann mit der jeweiligen getter/setter-Methode
   * ermittelt oder gesetzt werden.<br>
   * z.B. <code>Array.getInt( ArrayUtil.getArray(anArray,c), c[c.length-1] )</code>
   * @param array mehrdimensionaler Array
   * @param c     Koordinate
   */
  private static Object getArray(Object array, int c[]) {
    for (int d=0; d<c.length-1; d++)
      array = Array.get(array,c[d]);
    return array;
  }

  /**
   * Liefert die Dimension des Arrays.
   * @param array ein Array
   */
  public static int getDimension(Object array) {
    for (int d=0;; d++)
      try {
        array = Array.get(array,0);
      } catch (IllegalArgumentException err) {
        return d;
      }
  }

  /**
   * Liefert die Groesse des Arrays in allen Dimensionen.
   * @param array ein Array
   */
  public static int[] getLength(Object array) {
    int[] length = new int[ getDimension(array) ];
    for (int d=0; d<length.length; d++) {
      length[d] = Array.getLength( array );
      array = Array.get(array,0);
    }
    return length;
  }

  /**
   * Liefert die Groesse des Arrays in der angegebenen Dimension.
   * @param array ein Array
   * @param dim   Index einer Dimension (beginnend bei 0!)
   */
  public static int getLength(Object array, int dim) {
    return Array.getLength( getArray(array, new int[dim]) );
  }

  /**
   * Liefert ein Object aus einem mehrdimensionalen Array.
   * @param array ein Array
   * @param c     Koordinate
   */
  public static Object get(Object array, int[] c) {
    return Array.get( getArray(array,c), c[c.length-1] );
  }

  /**
   * Liefert ein Object aus einem mehrdimensionalen Array als <code>boolean</code>.
   * @param array ein Array
   * @param c     Koordinate
   */
  public static boolean getBoolean(Object array, int[] c) {
    return Array.getBoolean( getArray(array,c), c[c.length-1] );
  }

  /**
   * Liefert ein Object aus einem mehrdimensionalen Array als <code>byte</code>.
   * @param array ein Array
   * @param c     Koordinate
   */
  public static byte getByte(Object array, int[] c) {
    return Array.getByte( getArray(array,c), c[c.length-1] );
  }

  /**
   * Liefert ein Object aus einem mehrdimensionalen Array als <code>char</code>.
   * @param array ein Array
   * @param c     Koordinate
   */
  public static char getChar(Object array, int[] c) {
    return Array.getChar( getArray(array,c), c[c.length-1] );
  }

  /**
   * Liefert ein Object aus einem mehrdimensionalen Array als <code>double</code>.
   * @param array ein Array
   * @param c     Koordinate
   */
  public static double getDouble(Object array, int[] c) {
    return Array.getDouble( getArray(array,c), c[c.length-1] );
  }

  /**
   * Liefert ein Object aus einem mehrdimensionalen Array als <code>float</code>.
   * @param array ein Array
   * @param c     Koordinate
   */
  public static float getFloat(Object array, int[] c) {
    return Array.getFloat( getArray(array,c), c[c.length-1] );
  }

  /**
   * Liefert ein Object aus einem mehrdimensionalen Array als <code>int</code>.
   * @param array ein Array
   * @param c     Koordinate
   */
  public static int getInt(Object array, int[] c) {
    return Array.getInt( getArray(array,c), c[c.length-1] );
  }

  /**
   * Liefert ein Object aus einem mehrdimensionalen Array als <code>long</code>.
   * @param array ein Array
   * @param c     Koordinate
   */
  public static long getLong(Object array, int[] c) {
    return Array.getLong( getArray(array,c), c[c.length-1] );
  }

  /**
   * Liefert ein Object aus einem mehrdimensionalen Array als <code>short</code>.
   * @param array ein Array
   * @param c     Koordinate
   */
  public static short getShort(Object array, int[] c) {
    return Array.getShort( getArray(array,c), c[c.length-1] );
  }

  /**
   * Setzt eine Koordinate in einem mehrdimensionalen Array.
   * @param array ein Array
   * @param c     Koordinate
   * @param v     Objekt, da an die Koordinate gesetzt wird
   */
  public static void set(Object array, int[] c, Object v) {
    Array.set( getArray(array,c), c[c.length-1], v);
  }

  /**
   * Setzt eine Koordinate in einem mehrdimensionalen Array als <code>boolean</code>.
   * @param array ein Array
   * @param c     Koordinate
   * @param v     neuer Wert
   */
  public static void setBoolean(Object array, int[] c, boolean v) {
    Array.setBoolean( getArray(array,c), c[c.length-1], v);
  }

  /**
   * Setzt eine Koordinate in einem mehrdimensionalen Array als <code>byte</code>.
   * @param array ein Array
   * @param c     Koordinate
   * @param v     neuer Wert
   */
  public static void setByte(Object array, int[] c,  byte v) {
    Array.setByte( getArray(array,c), c[c.length-1], v);
  }

  /**
   * Setzt eine Koordinate in einem mehrdimensionalen Array als <code>char</code>.
   * @param array ein Array
   * @param c     Koordinate
   * @param v     neuer Wert
   */
  public static void setChar(Object array, int[] c, char v) {
    Array.setChar( getArray(array,c), c[c.length-1], v);
  }

  /**
   * Setzt eine Koordinate in einem mehrdimensionalen Array als <code>double</code>.
   * @param array ein Array
   * @param c     Koordinate
   * @param v     neuer Wert
   */
  public static void setDouble(Object array, int[] c, double v) {
    Array.setDouble( getArray(array,c), c[c.length-1], v);
  }

  /**
   * Setzt eine Koordinate in einem mehrdimensionalen Array als <code>float</code>.
   * @param array ein Array
   * @param c     Koordinate
   * @param v     neuer Wert
   */
  public static void setFloat(Object array, int[] c, float v) {
    Array.setFloat( getArray(array,c), c[c.length-1], v);
  }

  /**
   * Setzt eine Koordinate in einem mehrdimensionalen Array als <code>int</code>.
   * @param array ein Array
   * @param c     Koordinate
   * @param v     neuer Wert
   */
  public static void setInt(Object array, int[] c, int v) {
    Array.setInt( getArray(array,c), c[c.length-1], v);
  }

  /**
   * Setzt eine Koordinate in einem mehrdimensionalen Array als <code>long</code>.
   * @param array ein Array
   * @param c     Koordinate
   * @param v     neuer Wert
   */
  public static void setLong(Object array, int[] c, long v) {
    Array.setLong( getArray(array,c), c[c.length-1], v);
  }

  /**
   * Setzt eine Koordinate in einem mehrdimensionalen Array als <code>short</code>.
   * @param array ein Array
   * @param c     Koordinate
   * @param v     neuer Wert
   */
  public static void setShort(Object array, int[] c, short v) {
    Array.setShort( getArray(array,c), c[c.length-1], v);
  }


}
