/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.lang;

import java.util.StringTokenizer;
import java.util.Vector;

/**
 * Diese Klasse erweitert den {@link StringTokenizer} um eine PushBack-Funktion,
 * mit der ein Token zurueck auf den Tokenizer gelegt werden kann.<br>
 * <b>Bemerkung:</b><br>
 * Solange noch Token im Pushback-Speicher liegen, liefert die Methode
 * {@link #nextToken(String)} eines diese Token, <u>ungeachtet</u> von den
 * dieser Methode uebergebenen Delimitern!
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class PushbackStringTokenizer extends StringTokenizer {
  /** Verwaltet die bereits gelesenen Tokens. */
  protected Vector<String> readTokens = new Vector<String>();

  /** Verwaltet die zurueck gelegten Token. */
  protected Vector<String> pushedbackTokens = new Vector<String>();

  /**
   * Erzeugt einen neuen Tokenizer.
   * @param str          zu parsender String
   * @param delim        Delimiter
   * @param returnDelims bestimmt, ob die Delimiter ebenfalls als Token
   *                     zurueckgeliefert werden
   */
  public PushbackStringTokenizer(String str, String delim,  boolean returnDelims) {
    super(str, delim, returnDelims);
  }

  /**
   * Erzeugt einen neuen Tokenizer. Die Delimiter werden nicht als Token
   * zurueckgegeben.
   * @param str   zu parsender String
   * @param delim Delimiter
   */
  public PushbackStringTokenizer(String str, String delim) {
    super(str, delim);
  }

  /**
   * Erzeugt einen neuen Tokenizer. Als Delimiter werden {@code " \t\n\r\f"}
   * verwendet. Diese werden nicht als Token interpretiert.
   * zurueckgegeben.
   * @param str zu parsender String
   */
  public PushbackStringTokenizer(String str) {
    super(str);
  }

  /**
   * Prueft, ob noch Tokens gelesen werden koennen.
   */
  public boolean hasMoreTokens() {
    return !pushedbackTokens.isEmpty() || super.hasMoreTokens();
  }

  /**
   * Liefert das naechste Token.
   */
  public String nextToken() {
    String token = !pushedbackTokens.isEmpty() ? pushedbackTokens.remove(0) : super.nextToken();
    readTokens.add(0,token);
    return token;
  }

  /**
   * Liefert das naechste Token.<br>
   * <b>Diese Methode ist nur mit Bedacht zu verwenden, da sie - sofern vorhanden -
   * das letzte mittels {@link #pushback()} zurueckgelegte Token liefert, egal
   * welcher Delimiter uebergeben wurde!!</b>
   * @param delim Delimiter fuer das naechste Token
   */
  public String nextToken(String delim) {
    String token = !pushedbackTokens.isEmpty() ? pushedbackTokens.remove(0) : super.nextToken(delim);
    readTokens.add(0,token);
    return token;
  }

  /**
   * Liefert denselben Wert wie {@link #hasMoreTokens()}.
   */
  public boolean hasMoreElements() {
    return hasMoreTokens();
  }

  /**
   * Liefert denselben Wert wie {@link #nextToken()}.
   */
  public Object nextElement() {
    return nextToken();
  }

  /**
   * Liefert die Anzahl an Token, die noch gelesen werden koennen.
   */
  public int countTokens() {
    return pushedbackTokens.size() + super.countTokens();
  }

  /**
   * Legt das zuletzt gelesene Token zurueck in den Tokenizer.
   * @return das zurueckgelegte Token
   */
  public String pushback() {
    String pushbackedToken = readTokens.remove(0);
    pushedbackTokens.add(0, pushbackedToken );
    return pushbackedToken;
  }

  /**
   * Liefert alle bisher gelesenen Token als Zeichenkette. Diese Methode macht
   * nur Sinn, wenn auch die Delimiter als Token gelesen werden.
   */
  public String getReadTokens() {
    StringBuffer buffer = new StringBuffer();
    for (int i=readTokens.size()-1; i>=0; i--)
      buffer.append(readTokens.elementAt(i));
    return buffer.toString();
  }

}
