/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.lang;

import java.util.Arrays;
import java.util.Collection;
import java.util.Comparator;
import java.util.Vector;

/**
 * Diese Klasse erweitert den {@linkplain Vector Java-Vector} um eine
 * Sortier-Funktion.
 * Die Elemente werden (aus Effizienzgruenden) <b>nicht</b> automatisch
 * sortiert. Ein Aufruf von {@link #sort()} oder {@link #sortInverted()} ist
 * hierfuer notwendig.
 * @author <a href="mailto:Martin.Schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class SortableVector<E> extends Vector<E> {
  /** {@link Comparator}, der die Sortierung bestimmt. */
  protected Comparator comparator = null;
  /** {@link Comparator}, der die invertierte Sortierung bestimmt. */
  protected Comparator invertedComparator = null;

  /**
   * Erzeugt einen neuen {@code SortableVector}. Es wird der
   * {@link DefaultComparator} verwendet.
   */
  public SortableVector(int initialCapacity, int capacityIncrement) {
    this(null,initialCapacity, capacityIncrement);
  }

  /**
   * Erzeugt einen neuen {@code SortableVector}. Es wird der
   * {@link DefaultComparator} verwendet.
   */
  public SortableVector(int initialCapacity) {
    this(null,initialCapacity);
  }

  /**
   * Erzeugt einen neuen {@code SortableVector}. Es wird der
   * {@link DefaultComparator} verwendet.
   */
  public SortableVector() {
    this((Comparator)null);
  }

  /**
   * Erzeugt einen neuen {@code SortableVector}. Es wird der
   * {@link DefaultComparator} verwendet.
   */
  public SortableVector(Collection c) {
    this(null, c);
  }

  /**
   * Erzeugt einen neuen {@code SortableVector}
   * @param comparator bestimmt die Sortierung
   */
  public SortableVector(Comparator comparator, int initialCapacity, int capacityIncrement) {
    super(initialCapacity, capacityIncrement);
    setComparator(comparator);
  }

  /**
   * Erzeugt einen neuen {@code SortableVector}
   * @param comparator bestimmt die Sortierung
   */
  public SortableVector(Comparator comparator, int initialCapacity) {
    super(initialCapacity);
    setComparator(comparator);
  }

  /**
   * Erzeugt einen neuen {@code SortableVector}
   * @param comparator bestimmt die Sortierung
   */
  public SortableVector(Comparator comparator) {
    super();
    setComparator(comparator);
  }

  /**
   * Erzeugt einen neuen {@code SortableVector}
   * @param comparator bestimmt die Sortierung
   */
  public SortableVector(Comparator comparator, Collection c) {
    super(c);
    setComparator(comparator);
  }

  /**
   * Belegt die globale Variable {@link #comparator} mit einem neuen Wert.
   * Wird {@code null} uebergeben, wird der {@link DefaultComparator} verwendet
   * @param comparator ein Comparator
   */
  private void setComparator(Comparator comparator) {
    this.comparator = (comparator != null) ? comparator : DefaultComparator.DEFAULT;
    this.invertedComparator = DefaultComparator.invert(this.comparator);
  }

  /**
   * Sortiert den Vector.
   */
  public void sort() {
    Arrays.sort( this.elementData, comparator );
  }

  /**
   * Sortiert den Vector in der umgekehrten Reihenfolge.
   */
  public void sortInverted() {
    Arrays.sort( this.elementData, invertedComparator );
  }

}
