/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.lang;


/**
 * Diese Klasse stellt einen "arbeitenden" Thread dar (im Gegensatz zum
 * "konkurrierenden" Thread, der mit anderen Threads auf gemeinsame Ressourcen
 * zugreift). Die Arbeitsweise des <code>StepThread</code> erfolgt durch
 * immer wiederkehrende Arbeitsschritte. Der Thread kann
 * <ul>
 * <li>gestartet werden ({@link #start()})</li>
 * <li>angehalten werden ({@link #pause()})</li>
 * <li>fortgesetzt werden ({@link #resume()})</li>
 * <li>abgebrochen werden ({@link #terminate()})</li>
 * <li>schrittweise ausgefuehrt werden (step)</li>
 * </ul>
 * @author Martin Schmitz
 * @version 1.0
 */
public abstract class StepThread extends WorkingThread {
  /** Modus "Thread fuehrt aktuellen Schritt zu Ende und wechselt dann in
   *  'Pause'". */
  public static final int MODE_STEP = 5;

  /**
   * Erzeugt einen neuen Prozess.
   * @param name Name fuer den Thread
   */
  public StepThread(String name) {
    super(name);
  }

  /**
   * Erzeugt einen neuen Prozess.
   */
  public StepThread() {
    super();
  }

  /**
   * Fuehrt einen einzelnen Arbeitsschritt aus. Danach wird der Thread
   * wieder schlafen gelegt.
   */
  public void step() {
    boolean wasSleeping = isSleeping();

    this.mode = MODE_STEP;
    // beim allerersten Mal muss der Thread gestartet werden
    if ( !isAlive() )
      super.start();
    else
      if ( wasSleeping )
        wakeUp();
  }

  /**
   * Implementiert den konitinuierlichen Arbeitsablauf des Threads als eine
   * Schleife, die fortlaufend {@link #performStep(int)} aufruft.
   */
  protected void performWork() {
    for( int step=1;; step++ ) {
       try {
         checkBreakingCommands();
       }
       catch (ThreadDeath err) {
         return;
       }

       switch (mode) {
         case MODE_RUN:  performStep(step);
                         // wurde waehrend der Ausfuehrung des
                         // Schritts ein Step-Signal gegeben, wird nur
                         // der aktuelle Schritt beendet und kein neuer
                         // begonnen.
                         if ( mode == MODE_STEP )
                           mode = MODE_PAUSE;
                         break;
         case MODE_STEP: performStep(step);
                         mode = MODE_PAUSE;
                         break;
       }
     }

  }

  /**
   * Implementiert einen Arbeitsschritt des Threads.
   */
  protected abstract void performStep(int stepNo);
}
