/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.lang;

import java.util.Vector;

/**
 * Diese Klasse stellt einen "arbeitenden" Thread dar (im Gegensatz zum
 * "konkurrierenden" Thread, der mit anderen Threads auf gemeinsame Ressourcen
 * zugreift). Der <code>WorkingThread</code> kann
 * <ul>
 * <li>gestartet werden ({@link #start()})</li>
 * <li>angehalten werden ({@link #pause()})</li>
 * <li>fortgesetzt werden ({@link #resume()})</li>
 * <li>abgebrochen werden ({@link #terminate()})</li>
 * </ul>
 * @author Martin Schmitz
 * @version 1.0
 */
public abstract class WorkingThread extends Thread {
  /** Modus "Thread ist bereit, gestartet zu werden". */
  public static final int MODE_READY = 1;
  /** Modus "Thread ist initialisiert, und laeuft". */
  public static final int MODE_RUN = 2;
  /** Modus "Thread ist beendet".<br>
   *  <b>Der Thread kann nicht neu initialisiert oder gestartet werden!</b> */
  public static final int MODE_TERMINATE = 3;
  /** Modus "Thread pausiert und kann fortgesetzt werden". */
  public static final int MODE_PAUSE = 4;
//  /** Modus "Thread fuehrt aktuellen Schritt zu Ende und wechselt dann in
//   *  'Pause'". */
//  public static final int MODE_STEP = 5;

  /** Speichert den aktuellen Thread-Status. */
  protected int mode = MODE_READY;
  /** Speichert die Listener, die auf den Thread horchen. */
  protected Vector listeners = new Vector<WorkingThreadListener>();

  private String name = null;

  /**
   * Erzeugt einen neuen Thread.
   * @param name Bezeichnung fuer den Thread
   */
  public WorkingThread(String name) {
    this.name = name;
  }

  /**
   * Erzeugt einen neuen Thread.
   */
  public WorkingThread() {
    this(null);
  }

  /**
   * Liefert den Status, in dem sich der Thread gerade befindet.
   */
  public int getMode() {
    return mode;
  }

  /**
   * Liefert den Namen des Threads. Falls dieser nicht gesetzt ist, wird
   * <code>super.toString()</code> zurueckgegeben.
   */
  public String toString() {
    return this.name != null ? this.name : super.toString();
  }

  /**
   * Legt den Thread schlafen, bis er durch <code>wakeUp()</code>
   * wieder aktiviert wird. Loest {@link WorkingThread#fireThreadPaused()} aus.
   * @see #wakeUp()
   */
  protected synchronized void goSleep() {
    try {
      fireThreadPaused();
      wait();
    }
    catch ( InterruptedException err ) {
    }
  }

  /**
   * Aktiviert den Thread wieder, nachdem er durch <code>goSleep()</code>
   * Schlafen gelegt wurde. Loest {@link WorkingThread#fireThreadResumed()} aus.
   * @see #goSleep()
   */
  protected synchronized void wakeUp() {
    notify();
    fireThreadResumed();
  }

  /**
   * Startet den Thread (erneut). Beim allerersten Mal wird die
   * <code>performInit()</code>-Methode aufgerufen.
   * @see #performInit()
   */
  public void start() {
    boolean wasSleeping = isSleeping();

    this.mode = MODE_RUN;
    // beim allerersten Mal muss der Thread gestartet werden
    if ( !isAlive() ) {
      performInit();
      super.start();
    } else {
      if ( wasSleeping )
        wakeUp();
    }
  }

  /**
   * Beendet (temporaer) die Arbeit des Threads. Ueber die
   * <code>start()</code>- oder <code>step()</code>-Methode kann die Arbeit
   * wieder aufgenommen werden.
   */
  public void pause() {
    boolean wasSleeping = isSleeping();

    this.mode = MODE_PAUSE;
    // beim allerersten Mal muss der Thread gestartet werden
    if ( !isAlive() )
      super.start();
    else
      if ( wasSleeping )
        wakeUp();
    checkBreakingCommands();
  }

  /**
   * Beendet den Thread, in dem die Run-Methode auslaeuft. Danach kann der
   * Thread <b>nicht</b> mehr neu gestartet werden.
   *
   */
  public void terminate() {
    boolean wasSleeping = isSleeping();

    this.mode = MODE_TERMINATE;
    if ( !isAlive() && !isInterrupted() )
      super.start();
    else
      if ( wasSleeping )
        wakeUp();
  }

  /**
   * Checkt, ob der Thread aktuell am arbeiten ist.
   */
  public boolean isRunning() {
    return mode == MODE_RUN;
  }

  /**
   * Checkt, ob der Thread aktuell am warten ist. Dies ist er, wenn er
   * manuell in den Pause-Mode versetzt wurde.
   * @see #pause()
   */
  public boolean isSleeping() {
    return mode == MODE_PAUSE;
  }

  /**
   * Implementiert den Arbeitsablauf des Threads.
   * Sollte nicht manuell aufgerufen werden, da der aufrufende Prozess
   * sonst blockiert. Loest zunaechst {@link WorkingThread#fireThreadStarted()}
   * aus und {@link WorkingThread#fireThreadStopped()} bevor der Thread
   * auslaeuft.
   * @see #start()
   * @see #performWork()
   */
  public void run() {
    fireThreadStarted();
    try {
      // Thread wurde nur gestartet, um auszulaufen!
      if (mode != MODE_TERMINATE)
        performWork();
    } catch (ThreadDeath err) {
//      err.printStackTrace();
    }
    mode = MODE_TERMINATE;
    performDispose();
    fireThreadStopped();
  }

  /**
   * Prueft, on ein Unterbrechnung-Signal erfolgt ist (<code>pause()</code>
   * oder <code>terminate()</code>).<br>
   * Bei <code>pause()</code> wird der Thread sofort schlafen gelegt. Bei
   * <code>terminate()</code> wird eine <code>ThreadDeath</code>-Exception
   * geworfen. Diese Methode kann/sollte an geeigneten Stellen im
   * {@linkplain #performWork() Arbeitsablauf} aufgerufen werden.
   */
  protected void checkBreakingCommands() {
    switch (mode) {
      case MODE_TERMINATE: throw new ThreadDeath();
      case MODE_PAUSE: goSleep();
                       break;
    }
  }

  /**
   * Wird beim allerersten Aufruf der <code>start()</code>-Methode
   * aufgerufen, um Thread-Komponenten zu initialisieren.
   * <b>Sollte {@link #fireThreadInitialised()} aufrufen!</b>
   */
  protected abstract void performInit();

  /**
   * Implementiert den Arbeitsablauf des Threads.<br>
   * Zwischendurch sollte an geeigneter Stelle immer wieder die
   * {@link #checkBreakingCommands()}-Methode aufgerufen werden, um
   * auf Pause- oder Terminate-Signale zu reagieren.
   */
  protected abstract void performWork();

  /**
   * Implementiert alle Aktionen, die nach Auslaufen des Threads
   * durchzufuehren sind. z.B. sollte der Thread alle verteilten Listener
   * von anderen Objekten wieder entfernen, damit er nicht mehr informiert
   * wird. Diese Methode wird vor dem Auslaufen der {@link #run()}-Methode
   * aufgerufen.
   */
  protected abstract void performDispose();

  /**
   * Fuegt einen Listener hinzu.
   * @return <code>true</code> wenn dass Hinzufuegen erfolgreich war (siehe
   *         {@link Vector#add(Object)})
   */
  public boolean addThreadListener(WorkingThreadListener listener) {
    return listeners.add(listener);
  }

  /**
   * Entfernt einen Listener von dem Thread.
   * @return <code>true</code> wenn dass Entfernen erfolgreich war (siehe
   *         {@link Vector#remove(Object)})
   */
  public boolean removeThreadListener(WorkingThreadListener listener) {
    return listeners.remove(listener);
  }

  /**
   * Informiert alle {@link WorkingThreadListener} darueber, dass der Thread
   * initialisiert wurde.
   */
  protected void fireThreadInitialised() {
//    System.err.println("Thread init");
    for (int i=0; i<listeners.size(); i++)
      ((WorkingThreadListener)listeners.get(i)).threadInitialised(this);
  }

  /**
   * Informiert alle {@link WorkingThreadListener} darueber, dass der Thread
   * gestartet wurde.
   */
  protected void fireThreadStarted() {
//    System.err.println("Thread started");
    for (int i=0; i<listeners.size(); i++)
      ((WorkingThreadListener)listeners.get(i)).threadStarted(this);
  }

  /**
   * Informiert alle {@link WorkingThreadListener} darueber, dass der Thread
   * pausiert.
   */
  protected void fireThreadPaused() {
//    System.err.println("Thread paused");
    for (int i=0; i<listeners.size(); i++)
      ((WorkingThreadListener)listeners.get(i)).threadPaused(this);
  }

  /**
   * Informiert alle {@link WorkingThreadListener} darueber, dass der Thread
   * (nach einer Pause) wieder arbeitet.
   */
  protected void fireThreadResumed() {
//    System.err.println("Thread resumed");
    for (int i=0; i<listeners.size(); i++)
      ((WorkingThreadListener)listeners.get(i)).threadResumed(this);
  }

  /**
   * Informiert alle {@link WorkingThreadListener} darueber, dass der Thread
   * gestoppt wurde oder ausgelaufen ist.
   */
  protected void fireThreadStopped() {
//    System.err.println("Thread stopped");
   for (int i=0; i<listeners.size(); i++)
      ((WorkingThreadListener)listeners.get(i)).threadStopped(this);
  }
}
