/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.lang.tree;

import java.util.Vector;

/**
 * Diese Klasse stellt einen Knoten in einem Baum dar, der maximal 2 Kindknoten
 * hat, einen linken und einen rechten.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class BinaryTreeNode<E> extends TreeNode<E> {
  /** Speichert das linke Kind des Knoten. */
  protected BinaryTreeNode<E> leftChild = null;
  /** Speichert das rechte Kind des Knoten. */
  protected BinaryTreeNode<E> rightChild = null;

  /**
   * Erzeugt einen Wurzel-Knoten ohne Nachfolger.
   */
  public BinaryTreeNode() {
    this((E)null);
  }

  /**
   * Erzeugt einen Wurzel-Knoten ohne Nachfolger.
   * @param object Objekt, das in dem Knoten gespeichert wird
   */
  public BinaryTreeNode(E object) {
    this(object, null);
  }

  /**
   * Erzeugt einen Blatt-Knoten.
   * @param parent direkter Vorgaenger-Knoten
   */
  public BinaryTreeNode(BinaryTreeNode<E> parent) {
    this(null, parent);
  }

  /**
   * Erzeugt einen Blatt-Knoten.
   * @param object Objekt, das in dem Knoten gespeichert wird
   * @param parent direkter Vorgaenger-Knoten
   */
  public BinaryTreeNode(E object, BinaryTreeNode<E> parent) {
    this(object,parent,null,null);
  }

  /**
   * Erzeugt einen inneren Knoten.
   * @param object Objekt, das in dem Knoten gespeichert wird
   * @param parent direkter Vorgaenger-Knoten
   */
  public BinaryTreeNode(E object, BinaryTreeNode<E> parent, BinaryTreeNode<E> leftChild, BinaryTreeNode<E> rightChild) {
    super(object,parent);
    setLeftChild( leftChild );
    setRightChild( rightChild );
  }


  /**
   * Liefert den Wurzel-Knoten.
   */
  public BinaryTreeNode<E> getRoot() {
    return (BinaryTreeNode)super.getRoot();
  }

  /**
   * Liefert den direkten Vorgaenger-Knoten.
   */
  public BinaryTreeNode<E> getParent() {
    return (BinaryTreeNode)super.getParent();
  }

  /**
   * Prueft, ob es sich um einen Blattknoten handelt.
   * @return <code>true<code> gdw. beide Kinder {@code null} sind
   */
  public boolean isLeaf() {
    // Effizienter als die allgemeine super-Methode
    return leftChild == null & rightChild == null;
  }

  /**
   * Prueft, ob ein Knoten als Vater- oder Kind-Knoten fuer diesen Knoten
   * verwendet werden kann
   * @param node zu pruefende Knoten
   * @exception IllegalArgumentException falls der Knoten nicht geeignet ist
   */
  public void checkNode(TreeNode<E> node) {
    if ( node!=null && !(node instanceof BinaryTreeNode) )
      throw new IllegalArgumentException("Node must be an BinaryTreeNode!");
  }

  /**
   * Liefert die (maximale) Anzahl der Kindknoten.
   * @return immer 2, unabhaengig davon, ob diese {@code null} sind oder nicht
   */
  public int getChildCount() {
//    int count = 0;
//    if ( leftChild != null )
//      count++;
//    if ( rightChild != null )
//      count++;
//    return count;
      return 2;
  }

  /**
   * Liefert einen Kindknoten.
   * @param i Index (beginnend bei 0)
   * @return den linken Knoten fuer i = 1, den rechten Knoten sonst
   */
  public BinaryTreeNode<E> getChild(int i) {
    return ( i == 0 ) ? leftChild : rightChild;
  }

  /**
   * Setzt einen Kindknoten. Setzt den linken Kind-Knoten fuer i = 1, den rechten
   * Kind-Knoten sonst
   * @param i Index (beginnend bei 0)
   * @exception IllegalArgumentException falls {@code child} kein
   *            {@code BinaryTreeNode<E>} ist
   */
  public void setChild(int i, TreeNode<E> child) {
    checkNode(child);
    TreeNode oldChild = null;
    if ( i == 0 ) {
      oldChild  = leftChild;
      leftChild = (BinaryTreeNode<E>)child;
    } else {
      oldChild   = rightChild;
      rightChild = (BinaryTreeNode<E>)child;
    }
    // Vater des alten Kind-Konten loeschen
    if ( oldChild != null )
      oldChild.setParent(null);
    // Vater des neuen Kind-Konten setzen
    if ( child != null )
      child.setParent(this);
  }

  /**
   * Liefert den linken Kindknoten.
   */
  public BinaryTreeNode<E> getLeftChild() {
    return leftChild;
  }

  /**
   * Setzt den linken Kindknoten.
   */
  public void setLeftChild(BinaryTreeNode<E> child) {
    setChild(0,child);
  }

  /**
   * Liefert den rechten Kindknoten.
   */
  public BinaryTreeNode<E> getRightChild() {
    return rightChild;
  }

  /**
   * Setzt den rechten Kindknoten.
   */
  public void setRightChild(BinaryTreeNode<E> child) {
    setChild(1,child);
  }

  /**
   * Liefert die im (Teil-)Baum des Knotens gespeicherten Objekte in der
   * Inorder-Reihenfolge.
   * @param v Liste in die die Objekte eingefuegt werden (kann {@code null} sein!)
   */
  public Vector<E> getObjectsInorder(Vector<E> v) {
    if ( v == null )
      v = new Vector<E>();

    if ( leftChild != null )
      leftChild.getObjectsInorder(v);
    v.add(this.getObject());
    if ( rightChild != null )
      rightChild.getObjectsInorder(v);

    return v;
  }

  /**
   * Liefert die im (Teil-)Baum des Knotens gespeicherten Objekte in der
   * Preorder-Reihenfolge.
   * @param v Liste in die die Objekte eingefuegt werden (kann {@code null} sein!)
   */
  public Vector<E> getObjectsPreorder(Vector<E> v) {
    if ( v == null )
      v = new Vector<E>();

    v.add(this.getObject());
    if ( leftChild != null )
      leftChild.getObjectsPreorder(v);
    if ( rightChild != null )
      rightChild.getObjectsPreorder(v);

    return v;
  }

  /**
   * Liefert die im (Teil-)Baum des Knotens gespeicherten Objekte in der
   * Postorder-Reihenfolge.
   * @param v Liste in die die Objekte eingefuegt werden (kann {@code null} sein!)
   */
  public Vector<E> getObjectsPostorder(Vector<E> v) {
    if ( v == null )
      v = new Vector<E>();

    if ( leftChild != null )
      leftChild.getObjectsPostorder(v);
    if ( rightChild != null )
      rightChild.getObjectsPostorder(v);
    v.add(this.getObject());

    return v;
  }

}
