/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.lang.tree;

import java.util.Random;
import java.util.regex.Pattern;

import org.apache.log4j.Logger;

import schmitzm.lang.LangUtil;

/**
 * Diese Klasse stellt einen Operator-Baum dar. Dieser unterstuetzt folgende
 * Knoten:<br>
 * <ul>
 *   <li>{@link ConstantNode}: Blatt-Knoten, die einen konstanten Wert darstellen
 *                             (z.B. {@code double} oder {@link String}).</li>
 *   <li>{@link OperatorNode}: Innere Knoten, die fuer einen 2-stelligen Operator
 *       auf dem <i>linken</i> und <i>rechten</i> Sohn stehen:
 *       <ul>
 *         <li>Addition: {@code "+"}</li>
 *         <li>Subtraktion: {@code "-"}</li>
 *         <li>Multiplikation: {@code "*"}</li>
 *         <li>Division: {@code "/"}</li>
 *         <li>Potenzbildung: {@code "^"}</li>
 *         <li>boolesches UND: {@code "&"}</li>
 *         <li>boolesches ODER: {@code "|"}</li>
 *         <li>Gleich: {@code "="} oder {@code "=="}</li>
 *         <li>Ungleich: {@code "!="} oder {@code "<>"}</li>
 *         <li>Kleiner als: {@code "<"}</li>
 *         <li>Groesser als: {@code ">"}</li>
 *         <li>Kleiner gleich: {@code "<="}</li>
 *         <li>Groesser gleich: {@code ">="}</li>
 *         <li>String-Konkatenation: {@code "+"}</li>
 *         <li>Test auf regulaeren Ausdruck: {@code "regex(expression,regular expr)"}</li>
 *         <li>n-tes Ergebnis eines {@linkplain String#split(String) String-Splits} ueber regulaeren Ausdruck: {@code "split(expression,regular expr,n)"}</li>
 *         <li>String-Ersetzung: {@code "replAll(expression,pattern expr,repl expr)"}</li>
 *       </ul></li>
 *   <li>{@link UnaryOperatorNode}: Innere Knoten, die fuer einen 1-stelligen
 *       Operator auf dem <i>linken</i> Sohn stehen:
 *       <ul>
 *         <li>Betrag: {@code "abs(.)"}</li>
 *         <li>Wurzel: {@code "sqr(.)"} oder {@code "sqrt(.)"}</li>
 *         <li>Runden: {@code "rnd(.)"} oder {@code "round(.)"}</li>
 *         <li>Abrunden: {@code "int(.)"} oder {@code "trunc(.)"}</li>
 *         <li>boolesches Nicht: {@code "!(.)"}</li>
 *         <li>Test auf "Not a Number" (NaN): {@code "isNaN(.)"}</li>
 *         <li>Sinus: {@code "sin(.)"}</li>
 *         <li>Cosinus: {@code "cos(.)"}</li>
 *         <li>Tangens: {@code "tan(.)"}</li>
 *         <li>Arcus-Sinus: {@code "arcsin(.)"} oder {@code "asin(.)"}</li>
 *         <li>Arcus-Cosinus: {@code "arccos(.)"} oder {@code "acos(.)"}</li>
 *         <li>Arcus-Tangens: {@code "arctan(.)"} oder {@code "atan(.)"}</li>
 *         <li>Exponentialfunktion: {@code "exp(.)"}</li>
 *         <li>Logarithmus zur Basis e: {@code "ln(.)"}</li>
 *         <li>Logarithmus zur Basis 10: {@code "log(.)"}</li>
 *         <li>Umwandlung Zahl-zu-String: {@code "str(.)"}</li>
 *         <li>Umwandlung String-zu-Zahl: {@code "val(.)"}</li>
 *         <li>Umwandlung in Gross-Buchstaben: {@code "toupper(.)"}</li>
 *         <li>Umwandlung in Klein-Buchstaben: {@code "tolower(.)"}</li>
 *         </ul></li>
 *   <li>{@link ConstantAliasNode}: Blatt-Knoten, die einen Alias fuer eine
 *       Konstante darstellen:
 *       <ul>
 *         <li>Zufallszahl zwischen 0 und 1: {@code "rand"} oder {@code "random"}</li>
 *         <li>"Not a number" (NaN): {@code "NaN"}</li>
 *         <li>Aktuelle Zeit in ms ({@link System#currentTimeMillis()}): {@code "CURR_MILLIS"}</li>
 *       </ul></li>
 *   <li>{@link ITENode}: ITE-Operator:<br>
 *       Der 3-stellige ITE-Operator {@code ITE(.,.,.)} drueckt die Bedingung
 *       "If .. Then .. Else .." aus.</li>
 * </ul>
 * Damit boolesche Operatoren und arithmetische Ausdruecke kombiniert
 * werden koennen, werden die Operanden und das Ergebis einer booleschen
 * Operation nicht als TRUE oder FALSE, sondern numerisch codiert:<br>
 * <b>Operanden:</b> {@code op > 0} entspricht TRUE; {@code op = 0} entspricht FALSE.<br>
 * <b>Ergebnis:</b> TRUE wird als 1 codiert; FALSE wird als 0 codiert.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of  Bonn/Germany)
 * @version 1.0
 */
public class OperationTree {
  /** Logger fuer Debug-Ausgaben */
  protected Logger LOGGER = LangUtil.createLogger(this);

  /** Speichert den Wurzel-Knoten des Operator-Baums. */
  protected TreeNode rootNode = null;

  /**
   * Erzeugt einen neuen Operator-Baum.
   * @param root Wurzelknoten des Baums
   */
  protected OperationTree(TreeNode root) {
    this.rootNode = root;
  }

  /**
   * Erzeugt einen neuen Operator-Baum.
   * @param root Wurzelknoten des Baums
   */
  public OperationTree(OperatorNode root) {
    this((BinaryTreeNode)root);
  }

  /**
   * Erzeugt einen neuen Operator-Baum.
   * @param root Wurzelknoten des Baums
   */
  public OperationTree(ConstantNode root) {
    this((BinaryTreeNode) root);
  }

  /**
   * Liefert den Wurzelknoten des Operator-Baums.
   */
  public TreeNode getRoot() {
    return rootNode;
  }

  /**
   * Wertet den Operator-Baum aus.
   */
  public Object evaluate() {
    return evaluate(rootNode);
  }

  /**
   * Wertet einen Knoten des Operator-Baums aus.
   * @param opTreeNode BinaryTreeNode
   */
  protected Object evaluate(TreeNode opTreeNode) {
    if ( opTreeNode == null )
      throw new NullPointerException("null-TreeNode can not be evaluted");

    if ( opTreeNode instanceof ConstantNode )
      return ((ConstantNode)opTreeNode).getObject();

    if ( opTreeNode instanceof ConstantAliasNode )
      return performOperation(((ConstantAliasNode)opTreeNode).getObject());

    if ( opTreeNode instanceof UnaryOperatorNode ) {
      UnaryOperatorNode operatorNode = (UnaryOperatorNode)opTreeNode;
      String operator = operatorNode.getObject();
      Object operand  = evaluate( operatorNode.getLeftChild() );
      return performOperation(operator,operand);
    }

    if ( opTreeNode instanceof MultiParamOperatorNode ) {
      MultiParamOperatorNode operatorNode = (MultiParamOperatorNode)opTreeNode;
      String operator = operatorNode.getObject();
      // Operanden auswerten
      Object[] operand = new Object[operatorNode.getChildCount()];
      for (int i=0; i<operand.length; i++)
        operand[i] = evaluate( operatorNode.getChild(i) );
      return performOperation(operator,operand);
    }

    if ( opTreeNode instanceof ITENode ) {
      ITENode operatorNode = (ITENode)opTreeNode;
      Object result = evaluate( operatorNode.getIfNode() );
      if ( result != null &&
          (result instanceof Number && ((Number)result).doubleValue() != 0 ||
           result instanceof Boolean && ((Boolean)result).booleanValue())
         )
        return evaluate( operatorNode.getLeftChild() );  // THEN
      else
        return evaluate( operatorNode.getRightChild() ); // ELSE
    }

    if ( opTreeNode instanceof OperatorNode ) {
      OperatorNode operatorNode = (OperatorNode)opTreeNode;
      String operator = operatorNode.getObject();
      Object operand1 = evaluate( operatorNode.getLeftChild() );
      // Optimierung (2. Operand muss nicht mehr ausgewertet werden!)
      if ( (operator.equals("*") || operator.equals("/") ) &&
           (operand1 instanceof Number  && ((Number)operand1).doubleValue() == 0) )
        return 0;
      if ( operator.equals("&") &&
           (operand1 instanceof Number  && ((Number)operand1).doubleValue() == 0 ||
            operand1 instanceof Boolean && !((Boolean)operand1)) )
        return 0;
      if ( operator.equals("|") &&
           (operand1 instanceof Number  && ((Number)operand1).doubleValue() == 1 ||
            operand1 instanceof Boolean && ((Boolean)operand1)) )
        return 1;
      Object operand2 = evaluate( operatorNode.getRightChild() );
      return performOperation(operator,operand1,operand2);
    }

    throw new UnsupportedOperationException( LangUtil.RESOURCE.getString("OperationTree.err.UnknownNode",opTreeNode.getClass().getName()) );
  }

  /**
   * Wertet einen n-stelligen Operator aus. Als Operator unterstuetzt diese
   * Methode
   * <ul>
   *   <li><b>String-Operatoren</b>
   *       <ul>
   *         <li>Teil-String: {@code "substr(expression,startPos_incl,endPos_excl)"}</li>
   *       </ul></li>
   * </ul>
   * @param operator n-stelliger Operator
   * @param operand Operanden auf die der Operator angewandt wird
   * @exception UnsupportedOperationException falls der Operator {@code null} ist,
   *            unbekannt ist, oder auf die Operanden nicht angewendet werden kann
   */
  protected Object performOperation(String operator, Object[] operand) {
    if ( operator == null )
      throw new UnsupportedOperationException( LangUtil.RESOURCE.getString("OperationTree.err.UnknownOperator",operator) );

    operator = operator.toUpperCase();

    if ( operator.equals("SUBSTR") ) {
      if ( operand.length < 2 )
        throw new UnsupportedOperationException( LangUtil.RESOURCE.getString("OperationTree.err.LessParams1",2,operator) );
      if ( !(operand[0] instanceof String) )
        throw new UnsupportedOperationException( LangUtil.RESOURCE.getString("OperationTree.err.IllegalParam",1,operator,"String") );
      if ( !(operand[1] instanceof Number) )
        throw new UnsupportedOperationException( LangUtil.RESOURCE.getString("OperationTree.err.IllegalParam",2,operator,"Number") );
      if ( operand.length > 2 && !(operand[2] instanceof Number) )
        throw new UnsupportedOperationException( LangUtil.RESOURCE.getString("OperationTree.err.IllegalParam",3,operator,"Number") );

      String str      = (String)operand[0];
      int    startPos = ((Number)operand[1]).intValue();
      if ( operand.length < 3 )
        return str.substring(startPos);
      int    endPos = ((Number)operand[2]).intValue();
      return str.substring(startPos, endPos);
    }

    if ( operator.equals("SPLIT") ) {
      if ( operand.length < 2 )
        throw new UnsupportedOperationException( LangUtil.RESOURCE.getString("OperationTree.err.LessParams1",2,operator) );
      if ( !(operand[0] instanceof String) )
        throw new UnsupportedOperationException( LangUtil.RESOURCE.getString("OperationTree.err.IllegalParam",1,operator,"String") );
      if ( !(operand[1] instanceof String) )
        throw new UnsupportedOperationException( LangUtil.RESOURCE.getString("OperationTree.err.IllegalParam",2,operator,"String") );
      if ( operand.length > 2 && !(operand[2] instanceof Number) )
        throw new UnsupportedOperationException( LangUtil.RESOURCE.getString("OperationTree.err.IllegalParam",3,operator,"Number") );

      String[] result = operand[0].toString().split(operand[1].toString());
      int      idx    = operand.length > 2 ? ((Number)operand[2]).intValue() : 0;
      if ( idx < 0 || idx >= result.length )
        return "";
      return result[idx];
    }

    if ( operator.equals("REPLALL") ) {
      if ( operand.length != 3 )
        throw new UnsupportedOperationException( LangUtil.RESOURCE.getString("OperationTree.err.LessParams2",3,operator) );
      if ( !(operand[0] instanceof String) )
        throw new UnsupportedOperationException( LangUtil.RESOURCE.getString("OperationTree.err.IllegalParam",1,operator,"String") );
      if ( !(operand[1] instanceof String) )
        throw new UnsupportedOperationException( LangUtil.RESOURCE.getString("OperationTree.err.IllegalParam",2,operator,"String") );
      if ( !(operand[2] instanceof String) )
        throw new UnsupportedOperationException( LangUtil.RESOURCE.getString("OperationTree.err.IllegalParam",3,operator,"String") );

      String result = operand[0].toString().replaceAll(operand[1].toString(), operand[2].toString());
      return result;
    }

    String paramMess = "";
    for (Object op : operand) {
      if ( !paramMess.equals("") )
        paramMess += ", ";
      paramMess += (op != null) ? op.getClass().getSimpleName() : null;
    }
    throw new UnsupportedOperationException( LangUtil.RESOURCE.getString("OperationTree.err.UnknownOperator",operator+" ("+paramMess+")") );
}

  /**
   * Wertet einen 2-stelligen Operator aus. Als Operator unterstuetzt diese
   * Methode
   * <ul>
   *   <li><b>Numerische Operatoren (beide Operatoren numerisch oder boolean)</b>
   *       <ul>
   *         <li>Addition: {@code "+"}</li>
   *         <li>Subtraktion: {@code "-"}</li>
   *         <li>Multiplikation: {@code "*"}</li>
   *         <li>Division: {@code "/"}</li>
   *         <li>Potenzbildung: {@code "^"}</li>
   *         <li>Boolesches UND: {@code "&"}, {@code "AND"} oder {@code "UND"}</li>
   *         <li>Boolesches ODER: {@code "|"}, {@code "OR"} oder {@code "ODER"}</li>
   *       </ul></li>
   *   <li><b>Vergleichsoperatoren (beide Operatoren vom gleichen Typ!)</b>
   *       <ul>
   *         <li>Gleichheit: {@code "=="}, {@code "="} oder {@code "EQ"}</li>
   *         <li>Ungleichheit: {@code "!="}, {@code "<>"} oder {@code "NE"}</li>
   *       </ul></li>
   *   <li><b>Vergleichsoperatoren (beide Operatoren {@link Comparable})</b>
   *       <ul>
   *         <li>Kleiner als: {@code "<"} oder {@code "LT"}</li>
   *         <li>Groesser als: {@code ">"} oder {@code "GT"}</li>
   *         <li>Kleiner gleich: {@code "<="} oder {@code "LE"}</li>
   *         <li>Groesser gleich: {@code ">="} oder {@code GE"}</li>
   *       </ul></li>
   *   <li><b>String-Operatoren (beide Operatoren String!)</b>
   *       <ul>
   *         <li>String-Konkatenation: {@code "+"}</li>
   *         <li>Test auf regulaeren Ausdruck: {@code "regex(expression,regular expr)"}
   *       </ul></li>
   * </ul>
   * Die boolesche Logik wird im Ergebnis als 1 (TRUE) oder 0 (FALSE) codiert.
   * Operanden {@code op} werden als FALSE interpretiert, gdw. {@code op == 0}.
   * @param operator 2-stelliger Operator
   * @param operand1 linker Operand, auf den der Operator angewendet wird
   * @param operand2 linker Operand, auf den der Operator angewendet wird
   * @return {@code op1} <i>operator</i> {@code op2} oder {@code null}, wenn beide
   *         Operanden {@code null} sind
   * @exception UnsupportedOperationException falls der Operator {@code null} ist,
   *            unbekannt ist, oder auf die Operanden nicht angewendet werden kann
   */
  protected Object performOperation(String operator, Object operand1, Object operand2) {
    if ( operand1 == null && operand2 == null )
      return null;
    if ( operator == null )
      throw new UnsupportedOperationException( LangUtil.RESOURCE.getString("OperationTree.err.UnknownOperator",operator) );

    operator = operator.toUpperCase();

    // numerische und boolesche Operationen
    if ( (operand1 instanceof Number || operand1 instanceof Boolean) &&
          operand2 instanceof Number || operand2 instanceof Boolean) {
      // Operanden in double umwandeln
      double op1 = operand1 instanceof Number ? ((Number)operand1).doubleValue() : (Boolean)operand1 ? 1 : 0;
      double op2 = operand2 instanceof Number ? ((Number)operand2).doubleValue() : (Boolean)operand2 ? 1 : 0;
      if (operator.equals("+"))
        return op1 + op2;
      if (operator.equals("-"))
        return op1 - op2;
      if (operator.equals("*"))
        return op1 * op2;
      if (operator.equals("/"))
        return op1 / op2;
      if (operator.equals("^"))
        return Math.pow(op1, op2);
      if (operator.equals("&") || operator.equals("AND") || operator.equals("UND"))
        return op1 != 0 && op2 != 0 ? 1 : 0;
      if (operator.equals("|") || operator.equals("OR") || operator.equals("ODER"))
        return op1 != 0 || op2 != 0 ? 1 : 0;
    }

    // Boolean erst in numerischen Wert umwandeln
    if ( operand1 instanceof Boolean )
      operand1 = (Boolean)operand1 ? 1 : 0;
    if ( operand2 instanceof Boolean )
      operand2 = (Boolean)operand2 ? 1 : 0;
    // Vergleichsoperatoren funktionieren nicht korrekt, wenn die Objekte
    // nicht vom gleichen Typ sind (z.B. Double und Long)
    // -> in Double umwandeln
    if ( operand1 instanceof Number )
      operand1 = ((Number)operand1).doubleValue();
    if ( operand2 instanceof Number )
      operand2 = ((Number)operand2).doubleValue();

    // Un/Gleichheit (bei Typ-Gleichheit)
    if (operator.equals("=") || operator.equals("==") || operator.equals("EQ"))
      return operand1 != null && operand1.equals(operand2) ||
             operand2 != null && operand2.equals(operand1) ? 1 : 0;
    if (operator.equals("!=") || operator.equals("<>") || operator.equals("NE"))
      return operand1 != null && !operand1.equals(operand2) ||
             operand2 != null && !operand2.equals(operand1) ? 1 : 0;

    // Vergleichsoperatoren
    if ( operand1 instanceof Comparable && operand2 instanceof Comparable ) {
      Comparable op1 = (Comparable)operand1;
      Comparable op2 = (Comparable)operand2;
      if (operator.equals(">") || operator.equals("GT"))
        return op1 != null && op1.compareTo(op2) > 0 ? 1 : 0;
      if (operator.equals(">=") || operator.equals("GE"))
        return op1 != null && op1.compareTo(op2) >= 0 ? 1 : 0;
      if (operator.equals("<") || operator.equalsIgnoreCase("LT"))
        return op1 != null && op1.compareTo(op2) < 0 ? 1 : 0;
      if (operator.equals("<=") || operator.equalsIgnoreCase("LE"))
        return op1 != null && op1.compareTo(op2) <= 0 ? 1 : 0;
    }

    // String-Operationen
    if ( operand1 instanceof String || operand2 instanceof String ) {
      if ( operator.equals("+") )
        return operand1.toString() + operand2.toString();
      if ( operator.equals("REGEX") ) {
//        LOGGER.debug(++i+"   "+operand1);
//        boolean ret = Pattern.matches(operand2.toString(), operand1.toString());
        return Pattern.matches(operand2.toString(), operand1.toString()) ? 1 : 0;
      }
    }

    String op1Class = (operand1 != null) ? operand1.getClass().getSimpleName() : null;
    String op2Class = (operand2 != null) ? operand2.getClass().getSimpleName() : null;
    throw new UnsupportedOperationException( LangUtil.RESOURCE.getString("OperationTree.err.UnknownOperator",operator+" ("+op1Class+", "+op2Class+")") );
}

  /**
   * Wertet einen 1-stelligen Operator aus. Als Operator unterstuetzt diese
   * Methode
   * <ul>
   *   <li><b>Numerische Operatoren (Operand numerisch oder boolean)</b>
   *       <ul>
   *         <li>Betrag: {@code "abs(.)"}</li>
   *         <li>Wurzel: {@code "sqr(.)"} oder {@code "sqrt(.)"}</li>
   *         <li>Runden: {@code "rnd(.)"} oder {@code "round(.)"}</li>
   *         <li>Abrunden: {@code "int(.)"} oder {@code "trunc(.)"}</li>
   *         <li>Boolesches NICHT: {@code "!"}, {@code "NOT"} oder {@code "NICHT"}</li>
   *         <li>Test auf "Not a Number" (NaN): {@code "isNaN(.)"}</li>
   *         <li>Sinus: {@code "sin(.)"}</li>
   *         <li>Cosinus: {@code "cos(.)"}</li>
   *         <li>Tangens: {@code "tan(.)"}</li>
   *         <li>Arcus-Sinus: {@code "arcsin(.)"} oder {@code "asin(.)"}</li>
   *         <li>Arcus-Cosinus: {@code "arccos(.)"} oder {@code "acos(.)"}</li>
   *         <li>Arcus-Tangens: {@code "arctan(.)"} oder {@code "atan(.)"}</li>
   *         <li>Exponentialfunktion: {@code "exp(.)"}</li>
   *         <li>Logarithmus zur Basis e: {@code "ln(.)"}</li>
   *         <li>Logarithmus zur Basis 10: {@code "log(.)"}</li>
   *         <li>Umwandlung Zahl-zu-String: {@code "val(.)"}</li>
   *       </ul></li>
   *   <li><b>String Operatoren (Operand String)</b>
   *       <ul>
   *         <li>Umwandlung String-zu-Zahl: {@code "str(.)"}</li>
   *         <li>Umwandlung in Gross-Buchstaben: {@code "toupper(.)"}</li>
   *         <li>Umwandlung in Klein-Buchstaben: {@code "tolower(.)"}</li>
   *       </ul></li>
   * </ul>
   * Die boolesche Logik wird im Ergebnis als 1 (TRUE) oder 0 (FALSE) codiert.
   * Operanden {@code op} werden als FALSE interpretiert, gdw. {@code op == 0}.
   * @param operator 1-stelliger Operator
   * @param operand  Operand auf den der Operator angewendet wird
   * @return <i>operator</li>( {@code operand} )
   * @exception UnsupportedOperationException falls der Operator {@code null} ist,
   *            unbekannt ist, oder auf den Operanden nicht angewendet werden kann
   */
  protected Object performOperation(String operator, Object operand) {
    if ( operator == null )
      throw new UnsupportedOperationException( LangUtil.RESOURCE.getString("OperationTree.err.UnknownOperator",operator) );

    operator = operator.toUpperCase();

    // numerische und boolesche Operationen
    if ( operand instanceof Number || operand instanceof Boolean ) {
      // Operanden in double umwandeln
      double op = operand instanceof Number ? ((Number)operand).doubleValue() : (Boolean)operand ? 1 : 0;
      if ( operator.equals("ISNAN") )
        return Double.isNaN(op) ? 1: 0;
      if ( Double.isNaN(op) )
        return Double.NaN;
      if ( operator.equals("ABS") )
        return Math.abs(op);
      if ( operator.equals("SQR") || operator.equals("SQRT") )
        return Math.sqrt(op);
      if (operator.equals("RND") || operator.equals("ROUND"))
        return Math.round(op);
      if (operator.equals("INT") || operator.equals("TRUNC"))
        return (long) op;
      if (operator.equals("!") || operator.equals("NOT") || operator.equals("NICHT"))
        return op == 0 ? 1 : 0;
      if (operator.equals("SIN"))
        return Math.sin(op);
      if (operator.equals("COS"))
        return Math.cos(op);
      if (operator.equals("TAN"))
        return Math.tan(op);
      if (operator.equals("ARCSIN") || operator.equals("ASIN"))
        return Math.asin(op);
      if (operator.equals("ARCCOS") || operator.equals("ACOS"))
        return Math.acos(op);
      if (operator.equals("ARCTAN") || operator.equals("ATAN"))
        return Math.atan(op);
      if (operator.equals("EXP"))
        return Math.exp(op);
      if (operator.equals("LOG"))
        return Math.log10(op);
      if (operator.equals("LN"))
        return Math.log(op);
      if (operator.equals("STR"))
        return String.valueOf(op);
    }

    // String Operationen
    if ( operand instanceof String ) {
      String op = (String)operand;
      if (operator.equals("VAL"))
        return Double.parseDouble(op);
      if (operator.equals("LEN"))
        return op.length();
      if (operator.equals("TOUPPER"))
        return op.toUpperCase();
      if (operator.equals("TOLOWER"))
        return op.toLowerCase();
    }
    String opClass = (operand != null) ? operand.getClass().getSimpleName() : null;
    throw new UnsupportedOperationException( LangUtil.RESOURCE.getString("OperationTree.err.UnknownOperator",operator+"("+opClass+")") );
}

  /**
   * Wertet einen 0-stelligen Operator (Alias oder Variable) aus.
   * Als Operator unterstuetzt diese Methode
   * <ul>
   *   <li>Zufallszahl zwischen 0 und 1: {@code "rand"} oder {@code "random"}</li>
   * </ul>
   * @param operator 0-stelliger Operator
   */
  protected Object performOperation(String operator) {
    operator = operator.toUpperCase();

    if ( operator.equals("RAND") || operator.equals("RANDOM") )
      return new Random().nextDouble();
    if ( operator.equals("NAN") )
      return Double.NaN;
    if ( operator.equals("CURR_MILLIS") )
      return System.currentTimeMillis();
    throw new UnsupportedOperationException( LangUtil.RESOURCE.getString("OperationTree.err.UnknownAliasOperator",operator) );
  }

  ///////////////////////////////////////////////////////////////////
  /////////////////   Knoten im Operator-Baum   /////////////////////
  ///////////////////////////////////////////////////////////////////
  /**
   * Diese Knoten repraesentiert einen 2-stelligen Operator im Operatorbaum.
   * Der Operator wird durch eine Zeichenkette repraesentiert.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class OperatorNode extends BinaryTreeNode<String> {
    /**
     * Erzeugt einen neuen Operator-Knoten
     * @param operator     Operator
     * @param parent       Vater-Knoten
     * @param leftOperand  Knoten der den linker Operand repraesentiert
     * @param rightOperand Knoten der den rechten Operand repraesentiert
     */
    public OperatorNode(String operator, BinaryTreeNode parent, BinaryTreeNode leftOperand, BinaryTreeNode rightOperand) {
      super(operator, parent, leftOperand, rightOperand);
    }

    /**
     * Erzeugt einen neuen Operator-Knoten
     * @param operator     Operator
     * @param leftOperand  Knoten der den linker Operand repraesentiert
     * @param rightOperand Knoten der den rechten Operand repraesentiert
     */
    public OperatorNode(String operator, BinaryTreeNode leftOperand, BinaryTreeNode rightOperand) {
      this(operator, null, leftOperand, rightOperand);
    }
  }

  /**
   * Diese Knoten repraesentiert einen 1-stelligen Operator im Operatorbaum.
   * Dessen rechter Kind-Knoten ist immer {@code null}. Der Operator wird
   * durch eine Zeichenkette repraesentiert.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class UnaryOperatorNode extends OperatorNode {
    /**
     * Erzeugt einen neuen Operator-Knoten
     * @param operator     Operator
     * @param parent       Vater-Knoten
     * @param operand      Knoten der den (linker) Operand repraesentiert
     */
    public UnaryOperatorNode(String operator, BinaryTreeNode parent, BinaryTreeNode operand) {
      super(operator, parent, operand, null);
    }

    /**
     * Erzeugt einen neuen Operator-Knoten
     * @param operator     Operator
     * @param operand      Knoten der den (linker) Operand repraesentiert
     */
    public UnaryOperatorNode(String operator, BinaryTreeNode operand) {
      this(operator, null, operand);
    }

    /**
     * Setzt nur den linken Kind-Knoten. Ist der Parameter {@code i > 0}, wird
     * der rechte Kind-Knoten auf {@code null} gesetzt, unabhaengig vom
     * Parameter {@code child}.
     * @param i Index (beginnend bei 0)
     * @param child neuer Kind-Knoten
     */
    public void setChild(int i, TreeNode<String> child) {
      if ( i > 0 )
        child = null;
      super.setChild(i,child);
    }
  }

  /**
   * Diese Knoten repraesentiert den speziellen Operator ITE
   * (If-Then-Else) im Operatorbaum. Eigentlich ist dieser Operator 3-stellig.
   * Um ihn in den binaeren Operator-Baum einzubetten, wird der THEN-Teil im
   * linken, der ELSE-Teil im rechten Kind und die IF-Bedingung gesondert
   * im Knoten hinterlegt.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class ITENode extends OperatorNode {
    /** Speichert die IF-Bedingung des ITE-Operators. */
    protected TreeNode<String> ifNode = null;

    /**
     * Erzeugt einen neuen Operator-Knoten
     * @param operator     Operator
     * @param parent       Vater-Knoten
     * @param ifOperand    Knoten der die IF-Bedingung repraesentiert
     * @param thenOperand  Knoten der die THEN-Auswertung repraesentiert
     * @param elseOperand  Knoten der die ELSE-Auswertung repraesentiert
     */
    public ITENode(String operator, BinaryTreeNode parent, BinaryTreeNode ifOperand, BinaryTreeNode thenOperand, BinaryTreeNode elseOperand) {
      super(operator, parent, thenOperand, elseOperand);
      setIfNode(ifOperand);
    }

    /**
     * Erzeugt einen neuen Operator-Knoten
     * @param operator     Operator
     * @param ifOperand    Knoten der die IF-Bedingung repraesentiert
     * @param thenOperand  Knoten der die THEN-Auswertung repraesentiert
     * @param elseOperand  Knoten der die ELSE-Auswertung repraesentiert
     */
    public ITENode(String operator, BinaryTreeNode ifOperand, BinaryTreeNode thenOperand, BinaryTreeNode elseOperand) {
      this(operator, null, ifOperand, thenOperand, elseOperand);
    }

    /**
     * Setzt die IF-Bedingung.
     * @param cond Knoten der die IF-Bedingung repraesentiert
     */
    public void setIfNode(TreeNode<String> cond) {
      ifNode = cond;
    }

    /**
     * Liefert einen Knoten, der die IF-Bedingung repraesentiert
     */
    public TreeNode<String> getIfNode() {
      return ifNode;
    }
  }

  /**
   * Diese Knoten repraesentiert den speziellen Operator mit mehr als 2
   * Operanden.<br>
   * <b>Bemerkung:<b> Eigentlich "vergewaltigt" dieser Knoten den
   * {@link BinaryTreeNode} (bzw. {@link OperatorNode}), da er nicht mehr
   * binaer ist. Ich habe jedoch im Moment keine Zeit, den {@link OperationTreeParser}
   * so umzuprogrammieren, dass er nur noch auf dem allgemeinen {@link TreeNode}
   * basiert.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class MultiParamOperatorNode extends OperatorNode {
    /** Speichert die Parameter des Operators. */
    protected BinaryTreeNode<String>[] paramNodes = new BinaryTreeNode[0];

    /**
     * Erzeugt einen neuen Operator-Knoten
     * @param operator     Operator
     * @param parent       Vater-Knoten
     * @param paramCount   Anzahl an Parameters des Operators
     */
    public MultiParamOperatorNode(String operator, BinaryTreeNode parent, int paramCount) {
      super(operator, parent, null, null);
      paramNodes = new BinaryTreeNode[paramCount];
    }

    /**
     * Erzeugt einen neuen Operator-Knoten
     * @param operator     Operator
     * @param paramCount   Anzahl an Parameters des Operators
     */
    public MultiParamOperatorNode(String operator, int paramCount) {
      this(operator, null, paramCount);
    }

    /**
     * Erzeugt einen neuen Operator-Knoten
     * @param operator     Operator
     * @param parent       Vater-Knoten
     * @param parameter    Knoten die die Funktionsparameter darstellen
     */
    public MultiParamOperatorNode(String operator, BinaryTreeNode parent, BinaryTreeNode... parameter) {
      this(operator, parent, parameter.length);
      paramNodes = parameter;
    }

    /**
     * Setzt den Knoten fuer einen Funktionsparameter.
     * @param i int Nummer des Parameters
     * @param paramNode Knoten der den Parameter darstellt
     */
    public void setChild(int i, TreeNode<String> paramNode) {
      checkNode(paramNode);
      if ( paramNodes != null && i < paramNodes.length )
        this.paramNodes[i] = (BinaryTreeNode)paramNode;
    }

    /**
     * Liefert den Knoten fuer einen Funktionsparameter.
     * @param i int Nummer des Parameters
     */
    public BinaryTreeNode getChild(int i) {
      return paramNodes != null && i < paramNodes.length ? this.paramNodes[i] : null;
    }

    /**
     * Liefert die Anzahl an Parametern.
     */
    public int getChildCount() {
      return paramNodes != null ? this.paramNodes.length : 0;
    }
  }

  /**
   * Diese Knoten repraesentiert einen konstanten Wert
   * im Operatorbaum. Dessen Kind-Knoten sind immer {@code null}. Der Wert wird
   * durch einen {@code double} repraesentiert.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class ConstantNode extends BinaryTreeNode {
    /**
     * Erzeugt einen neuen Konstanten-Knoten
     * @param constant Wert der Konstante
     * @param parent   Vater-Knoten
     */
    public ConstantNode(Object constant, BinaryTreeNode parent) {
      super(constant, parent);
    }

    /**
     * Erzeugt einen neuen Konstanten-Knoten
     * @param constant Wert der Konstante
     */
    public ConstantNode(Object constant) {
      this(constant, null);
    }

    /**
     * Macht nichts, da {@code ConstantNode} immer einen Blatt-Knoten darstellt.
     * @param i Index (beginnend bei 0)
     * @param child neuer Kind-Knoten
     */
    public void setChild(int i, TreeNode child) {
    }
  }

  /**
   * Diese Knoten repraesentiert einen konstanten numerischen Wert
   * im Operatorbaum, der jedoch nicht direkt, sondern durch einen Alias
   * dargestellt wird (z.B. "rand" fuer eine Zufallszahl oder ein Variablenname).
   * Der Knoten hat keine Kind-Knoten (sind immer {@code null}). Der Wert wird
   * durch einen {@code String} repraesentiert.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class ConstantAliasNode extends BinaryTreeNode<String> {
    /**
     * Erzeugt einen neuen Alias-Knoten
     * @param alias    Alias der Konstante
     * @param parent   Vater-Knoten
     */
    public ConstantAliasNode(String alias, BinaryTreeNode parent) {
      super(alias, parent);
    }

    /**
     * Erzeugt einen neuen Alias-Knoten
     * @param alias Alias der Konstante
     */
    public ConstantAliasNode(String alias) {
      this(alias, null);
    }

    /**
     * Macht nichts, da {@code ConstantAliasNode} immer einen Blatt-Knoten darstellt.
     * @param i Index (beginnend bei 0)
     * @param child neuer Kind-Knoten
     */
    public void setChild(int i, TreeNode<String> child) {
    }
  }

}
