/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.lang.tree;

import java.util.HashMap;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Vector;

import org.apache.log4j.Logger;

import schmitzm.lang.LangUtil;
import schmitzm.lang.PushbackStringTokenizer;
import schmitzm.lang.tree.OperationTree.ConstantAliasNode;
import schmitzm.lang.tree.OperationTree.ConstantNode;
import schmitzm.lang.tree.OperationTree.ITENode;
import schmitzm.lang.tree.OperationTree.MultiParamOperatorNode;
import schmitzm.lang.tree.OperationTree.OperatorNode;
import schmitzm.lang.tree.OperationTree.UnaryOperatorNode;
import schmitzm.swing.ParserOperatorsHints;
import schmitzm.swing.SwingUtil;

/**
 * Diese Klasse stellt einen Parser fuer einen {@linkplain OperationTree Operator-Baum}
 * dar. Dieser erstellt einen {@linkplain OperationTree Operator-Baum} aus einem
 * Formel-String, der folgende Teilen bestehen darf:
 * <ul>
 *   <li>konstante numerische Werte: {@link ConstantNode}</li>
 *       <ul>
 *           <li>Zahlen im {@code double}-Bereich</li>
 *           <li>{@code rand} oder {@code random} als Alias fuer eine Zufallszahl
 *                zwischen 0 und 1: {@link ConstantAliasNode}</li>
 *           <li>"Not a number" (NaN): {@code "NaN"}</li>
 *           <li>boolesches WAHR (1): {@code "TRUE"}</li>
 *           <li>boolesches FALSCH (0): {@code "FALSE"}</li>
 *           <li>Aktuelle Zeit in ms ({@link System#currentTimeMillis()}): {@code "CURR_MILLIS"}</li>
 *       </ul></li>
 *   <li>konstante String Werte: {@link ConstantNode}</li>
 *       <ul>
 *           <li>Zeichenketten, die durch " oder ' eingeleitet und abgeschlossen
 *               werden</li>
 *       </ul></li>
 *   <li>den 2-stelligen arithmetischen Operatoren: {@link OperatorNode}
 *       <ul>
 *          <li>Addition: {@code "+"}</li>
 *          <li>Subtraktion: {@code "-"}</li>
 *          <li>Multiplikation: {@code "*"}</li>
 *          <li>Division: {@code "/"}</li>
 *          <li>Potenzbildung: {@code "^"}</li>
 *        </ul></li>
 *   <li>den 2-stelligen booleschen Operatoren: {@link OperatorNode}
 *       <ul>
 *          <li>boolesches UND: {@code "&"}</li>
 *          <li>boolesches ODER: {@code "|"}</li>
 *          <li>Gleich: {@code "="} oder {@code "=="}</li>
 *          <li>Ungleich: {@code "!="} oder {@code "<>"}</li>
 *          <li>Kleiner als: {@code "<"}</li>
 *          <li>Groesser als: {@code ">"}</li>
 *          <li>Kleiner gleich: {@code "<="}</li>
 *          <li>Groesser gleich: {@code ">="}</li>
 *        </ul></li>
 *   <li>den 2-stelligen Zeichenketten-Operatoren: {@link OperatorNode}
 *       <ul>
 *          <li>String-Konkatenation: {@code "+"}</li>
 *          <li>Test auf regulaeren Ausdruck: {@code "regex(expression,regular expr)"}</li>
 *          <li>n-tes Ergebnis eines {@linkplain String#split(String) String-Splits} ueber regulaeren Ausdruck: {@code "split(expression,regular expr,n)"}</li>
 *       </ul></li>
 *   <li>den mehr-stelligen Zeichenketten-Operatoren: {@link MultiParamOperatorNode}
 *       <ul>
 *          <li>Teil-String: {@code "substr(expression,startPos_inkl,endPos_exkl)"}</li>
 *          <li>String-Ersetzung: {@code "replAll(expression,pattern,replacement)"}</li>
 *       </ul></li>
 *   <li>die 1-stelligen Operatoren: {@link UnaryOperatorNode}
 *       <ul>
 *          <li>Betrag: {@code "abs(.)"}</li>
 *          <li>Wurzel: {@code "sqr(.)"} oder {@code "sqrt(.)"}</li>
 *          <li>Runden: {@code "rnd(.)"} oder {@code "round(.)"}</li>
 *          <li>Abschneiden: {@code "int(.)"} oder {@code "trunc(.)"}</li>
 *          <li>Test auf "Not a Number" (NaN): {@code "isNaN(.)"}</li>
 *          <li>Sinus: {@code "sin(.)"}</li>
 *          <li>Cosinus: {@code "cos(.)"}</li>
 *          <li>Tangens: {@code "tan(.)"}</li>
 *          <li>Arcus-Sinus: {@code "arcsin(.)"} oder {@code "asin(.)"}</li>
 *          <li>Arcus-Cosinus: {@code "arccos(.)"} oder {@code "acos(.)"}</li>
 *          <li>Arcus-Tangens: {@code "arctan(.)"} oder {@code "atan(.)"}</li>
 *          <li>Exponentialfunktion: {@code "exp(.)"}</li>
 *          <li>Logarithmus zur Basis e: {@code "ln(.)"}</li>
 *          <li>Logarithmus zur Basis 10: {@code "log(.)"}</li>
 *        </ul></li>
 *   <li>den 1-stelligen booleschen Operatoren: {@link UnaryOperatorNode}
 *       <ul>
 *          <li>boolesches NICHT: {@code "!(.)"}</li>
 *        </ul></li>
 *   <li>den 1-stelligen String-Operatoren: {@link UnaryOperatorNode}
 *       <ul>
 *          <li>Umwandlung Zahl-zu-String: {@code "str(.)"}</li>
 *          <li>Umwandlung String-zu-Zahl: {@code "val(.)"}</li>
 *          <li>Umwandlung in Gross-Buchstaben: {@code "toupper(.)"}</li>
 *          <li>Umwandlung in Klein-Buchstaben: {@code "tolower(.)"}</li>
 *          <li>Laenge eines Strings: {@code "len(.)"}</li>
 *        </ul></li>
 *   <li>dem 3-stelligen Operator {@code ITE(.,.,.)}: {@link ITENode}<br>
 *       Der ITE-Operator drueckt eine Bedingung "IF .. THEN .. ELSE .." aus.</li>
 * </ul>
 * <br>
 * Damit boolesche Operatoren und arithmetische Ausdruecke kombiniert
 * werden koennen, werden die Operanden und das Ergebis einer booleschen
 * Operation nicht als TRUE oder FALSE, sondern numerisch codiert:<br>
 * <b>Operanden:</b> {@code op > 0} entspricht TRUE; {@code op = 0} entspricht FALSE.<br>
 * <b>Ergebnis:</b> TRUE wird als 1 codiert; FALSE wird als 0 codiert.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class OperationTreeParser implements ParserOperatorsHints {
  private Logger LOG = Logger.getLogger(this.getClass());

  /** Ergebnis-Typ von {@link #checkRuleClosing(String)}. */
  static private enum RuleClosingType {
    none,       // Formel wird nicht beendet
    rule_part,  // Formel wird beendet
    func_pram   // Formel wird beendet UND glz. ein Funktions-Parameter
  };

  /** Stellt einen Stack dar, auf den die geoeffneten Klammern gelegt werden.
   *  Zeichen 0 stellt das oberste Stack-Element dar. */
  protected StringBuffer openBrackets = null;

  /** Tokenizer auf den Formel-String. Wird beim Starten des Parse-Vorgangs
   *  erzeugt.
   *  @see #parse(String)
   */
  protected PushbackStringTokenizer tok = null;
  /** Entaehlt die Delimiter des Tokenizers {@link #tok}. Diese setzen sich
   *  aus den oeffnenden und schliessenden Klammern zusammen, sowie aus den
   *  Operatoren. Wird bei der Instanziierung des {@code OperationTreeParser}
   *  befuellt.
   *  @see #getOpeningBracketChars()
   *  @see #getClosingBracketChars()
   *  @see #getOperatorChars()
   */
  protected String delimiter    = null;

  /** Contains the operators available to create a valid rule.<br>
   *  <b>Note: These operators do not effect the parsing process, but
   *     only used for user/GUI interaction. */
  protected Vector<String>  avOperators = new Vector<String>();

  /** Contains a resource key postfix for each operator in {@link #avOperators},
   *  because not all operators (e.g. "=") can be used as resource key in
   *  property files. */
  protected Map<String,String>  avOperatorsKeys = new HashMap<String,String>();

  /**
   * Erzeugt eine neue Instanz des Parsers.
   */
  public OperationTreeParser() {
    delimiter = getOpeningBracketChars() +
                getClosingBracketChars() +
                getStringEncapsulationChars() +
                getParameterSeperatorChars() +
                getOperatorChars() + "\n";
    initOperatorList();
  }

  ///////////////////////////////////////////////////////////////////
  /////////////////   Konfigurations-Methoden   /////////////////////
  ///////////////////////////////////////////////////////////////////
  /**
   * Liefert die Zeichen, die als oeffnende Klammern verwendet werden.
   * Unterklassen koennen diese Methode ueberschreiben und Zeichen hinzufuegen.
   * @return <code>"([{"</code>
   */
  public String getOpeningBracketChars() {
    return "([{";
  }

  /**
   * Prueft, ob es sich bei einem String um eine oeffnende Klammer handelt.
   * @param token Zeichenkette
   * @see #getOpeningBracketChars()
   */
  public boolean isOpeningBracket(String token) {
    return getOpeningBracketChars().contains(token);
  }

  /**
   * Liefert die Zeichen, die als schliessende Klammern verwendet werden.
   * Unterklassen koennen diese Methode ueberschreiben und Zeichen hinzufuegen.
   * @return <code>")]}"</code>
   */
  public String getClosingBracketChars() {
    return ")]}";
  }

  /**
   * Prueft, ob es sich bei einem String um eine schliessende Klammer handelt.
   * @param token Zeichenkette
   * @see #getClosingBracketChars()
   */
  public boolean isClosingBracket(String token) {
    return getClosingBracketChars().contains(token);
  }

  /**
   * Prueft, ob eine oeffnende und eine schliessende Klammer zueinander
   * korrespondieren.
   * Unterklassen muessen diese Methode ueberschreiben und erweitern, wenn
   * weitere Klammer-Zeichen verwendet werden.
   * @param openBracket oeffnende Klammer
   * @param closeBracket schliessende Klammer
   */
  public boolean checkCorrespondingBrackets(char openBracket, char closeBracket) {
    return openBracket == '(' && closeBracket == ')' ||
           openBracket == '[' && closeBracket == ']' ||
           openBracket == '{' && closeBracket == '}';
  }

  /**
   * Liefert die Zeichen, zwischen denen String-Konstanten gekapselt werden.
   * Unterklassen koennen diese Methode ueberschreiben und Zeichen hinzufuegen.
   * @return <code>"'</code>
   */
  public String getStringEncapsulationChars() {
    return "\"'";
  }

  /**
   * Prueft, ob es sich bei einem String um ein Zeichenketten-kapselndes
   * Zeichen handelt.
   * @param token Zeichenkette
   * @see #getStringEncapsulationChars()
   */
  public boolean isStringEncapsulationChar(String token) {
    return getStringEncapsulationChars().contains(token);
  }

  /**
   * Liefert die Zeichen, die als Trennzeichen zwischen Funktionsparametern
   * verwendet werden.
   * Unterklassen koennen diese Methode ueberschreiben und Zeichen hinzufuegen.
   * @return <code>",;"</code>
   */
  public String getParameterSeperatorChars() {
    return ",;";
  }

  /**
   * Prueft, ob es sich bei einem String um eine schliessende Klammer handelt.
   * @param token Zeichenkette
   * @see #getClosingBracketChars()
   */
  public boolean isParameterSeperator(String token) {
    return getParameterSeperatorChars().contains(token);
  }

  /**
   * Liefert die Zeichen, die als 2-stellige Operatoren verwendet werden.
   * Unterklassen koennen diese Methode ueberschreiben und Zeichen hinzufuegen.
   */
  public String getOperatorChars() {
    return getAdditionOperatorChars() +
           getMultiplyOperatorChars() +
           getCompareOperatorChars() +
           getBooleanOperatorChars();
  }

  /**
   * Prueft, ob es sich bei einem String um einen Operator handelt.
   * @param token Zeichenkette
   * @see #getOperatorChars()
   */
  public boolean isOperatorChar(String token) {
//    return getOperatorChars().contains(token);
    return isAdditionOperator(token)
        || isBooleanOperator(token)
        || isCompareOperator(token)
        || isMultiplyOperator(token);
  }


  /**
   * Liefert die Zeichen, die als arithmetische Punkt-Operatoren verwendet werden.
   * Unterklassen koennen diese Methode ueberschreiben und Zeichen hinzufuegen.
   * @return <code>"/*^"</code>
   */
  public String getMultiplyOperatorChars() {
    return "/*^";
  }

  /**
   * Prueft, ob es sich bei einem String um einen Punkt-Operator handelt.
   * @param token Zeichenkette
   * @see #getMultiplyOperatorChars()
   */
  public boolean isMultiplyOperator(String token) {
    return getMultiplyOperatorChars().contains(token);
  }

  /**
   * Liefert die Zeichen, die als arithmetische Strich-Operatoren verwendet werden.
   * Unterklassen koennen diese Methode ueberschreiben und Zeichen hinzufuegen.
   * @return <code>"+-"</code>
   */
  public String getAdditionOperatorChars() {
    return "+-";
  }

  /**
   * Prueft, ob es sich bei einem String um einen Strich-Operator handelt.
   * @param token Zeichenkette
   * @see #getAdditionOperatorChars()
   */
  public boolean isAdditionOperator(String token) {
    return getAdditionOperatorChars().contains(token);
  }

  /**
   * Liefert die Zeichen, die als boolesche Operatoren verwendet werden.
   * Unterklassen koennen diese Methode ueberschreiben und Zeichen hinzufuegen.
   * @return {@code "&|!"}
   */
  public String getBooleanOperatorChars() {
    return "&|!";
  }

  /**
   * Prueft, ob es sich bei einem String um einen booleschen Operator handelt.
   * @param token Zeichenkette
   */
  public boolean isBooleanOperator(String token) {
    return getBooleanOperatorChars().contains(token)
        || token.equalsIgnoreCase("AND")
        || token.equals("OR");
  }

  /**
   * Liefert die Zeichen, die fuer Vergleichsoperatoren verwendet werden.
   * Unterklassen koennen diese Methode ueberschreiben und Zeichen hinzufuegen.
   * @return {@code "=<>!"}
   */
  public String getCompareOperatorChars() {
    return "=<>!";
  }

  /**
   * Prueft, ob es sich bei einem String um einen Vergleichsoperator handelt.
   * @param token Zeichenkette
   */
  public boolean isCompareOperator(String token) {
    return token.equals("=") ||
           token.equals("==") ||
           token.equals("!=") ||
           token.equals("<>") ||
           token.equals(">") ||
           token.equals(">=") ||
           token.equals("<") ||
           token.equals("<=");
  }



  /**
   * Liefert einen Wert, der die Konnektivitaet (Bindung) eines Operators ausdrueckt (je
   * groesser, desto groesser die Bindung). Punkt-Operatoren haben die
   * groesste Bindung fuer die Operanden, werden also als erstes ausgewertet.
   * Darauf folgen die Strich-Operatoren, dann die Vergleichsoperatoren und
   * am Ende die booleschen Operatoren
   * @param operator Operator
   * @return 0, falls kein bekannter Operator uebergeben wird
   */
  public int getOperatorConnectivity(String operator) {
    // Punkt-Operatoren haben den hoechste verbindende Prioritaet
    if ( isMultiplyOperator(operator) )
      return 1000;
    // Strich-Operatoren haben Vorrang vor Vergleichen oder booleschen Operatoren
    if ( isAdditionOperator(operator) )
      return 100;
    // Vergleichs-Operatoren haben Vorrang vor booleschen Operatoren
    if ( isCompareOperator(operator) )
      return 10;
    // boolesche Operatoren haben den geringesten Rang
    if ( isBooleanOperator(operator) ) {
      if ( operator.equals("!") )
        return 3;
      if ( operator.equals("&") )
        return 2;
      if ( operator.equals("|") )
        return 1;
    }

    return 0;
  }

  /**
   * Liefert den Konnektivitaetswert des am meisten bindenden Operators.
   * Unterklasse muessen diese Methode u.U. ueberschreiben, wenn durch
   * Ueberschreiben der Methode {@link #getOperatorConnectivity(String)}
   * die Prioritaeten der Operatoren veraendert werden.
   */
  protected int getMaximumOperatorConnectivity() {
    return 1000;
  }

  /**
   * Liefert die 2-stellige Operatoren, die zwei Operanden aneinander
   * binden (Punkt-Operatoren).
   * Unterklassen koennen diese Methode ueberschreiben und Zeichen hinzufuegen.
   * @return <code>"/*^"</code>
   * @deprecated wegen booleschen Operatoren muss eine Verbindungs-Prioritaet
                 verwendet werden; ersetzt durch {@link #getOperatorConnectivity(String)}
   */
  public String getConnectingOperatorChars() {
    return "/*^";
  }

  /**
   * Prueft, ob es sich bei einem 2-stelligen Operator um einen verbindenen
   * Operator handelt (Punkt-Operator).
   * @param operator 2-stelliger Operator
   * @see #getConnectingOperatorChars()
   * @deprecated wegen booleschen Operatoren muss eine Verbindungs-Prioritaet
                 verwendet werden; ersetzt durch {@link #getOperatorConnectivity(String)}
   */
  public boolean isConnectingOperator(String operator) {
    return getConnectingOperatorChars().contains(operator);
  }

  ///////////////////////////////////////////////////////////////////
  ///////////////////   Methoden fuer Parsen  ///////////////////////
  ///////////////////////////////////////////////////////////////////
  /**
   * Erstellt einen Operator-Baum aus einem Formel-String.
   * @param rule Formel
   */
  public OperationTree parse(String rule) {
    openBrackets = new StringBuffer();
//    // Probleme, wenn Formel mit einem geklammerten Ausdruck beginnt
//    // z.B. "(17+7)*2+1" -> *2+1 wird nicht ausgewertet
//    // Loesung: Gesamten Ausdruck nochmal klammern
//    rule = "("+rule.trim()+")";
//    tok  = new PushbackStringTokenizer(rule,delimiter,true);
//    return new OperationTree( parseRulePart() );
    tok  = new PushbackStringTokenizer(rule,delimiter,true);
    try {
      return new OperationTree( parseRule() );
    } catch (NoSuchElementException err) {
      throwParseError( LangUtil.RESOURCE.getString("OperationTree.err.UnexpectedEOR") );
    } catch (Exception err) {
      throwParseError( LangUtil.RESOURCE.getString("OperationTree.err.UnexpectedError"), err );
    }
    // kann nicht vorkommen!!
    return null;
  }

//  /**
//   * Parst die naechste Teil-Formel aus dem Tokenizer {@link #tok}. Eine (geklammerter)
//   * Teilformel ist eine durch Strich-Operatoren verbundene Kette von
//   * {@linkplain #parseOperand() Operanden}.
//   */
//  protected final BinaryTreeNode parseRulePart() {
//    BinaryTreeNode rulePart    = null;
//    BinaryTreeNode nextOperand = null;
//    String         operator    = null;
//
//    // Ersten (linken) Operand parsen
//    String token = nextNonWSToken();
//    // oeffnende Klammer
//    if ( isOpeningBracket(token) )
//      openBrackets.insert(0,token);
//    else
//      tok.pushback();
//
//    rulePart = parseOperand();
//    while ( tok.hasMoreTokens() ) {
//      // Operator parsen (kann nur ein Strich- oder Bool-Operator sein!)
//      operator = nextNonWSToken();
//
//      // Sonderfall: schliessende Klammer
//      // -> Klammerung ueberpruefen und Teil-Formel beenden
//      if ( isClosingBracket(operator) ) {
//        if ( !checkCorrespondingBrackets(openBrackets.charAt(0),operator.charAt(0)) )
//          throwParseError("Bracket '" + operator + "' can not close '" + openBrackets.charAt(0) + "'");
//        openBrackets.deleteCharAt(0);
//        return rulePart;
//      }
//
//      // Pruefen, ob es wirklich ein Operator ist
//      if ( !isOperatorChar(operator) )
//        throwParseError("Illegal operator '" + operator + "'");
//
//      // Sonderfall: boolesche Operatoren u.U. 2-stellig (<=, >=, !=, <>, ==)
//      String nToken = nextNonWSToken();
//      if ( operator.equals("<") && nToken.equals("=") ||
//           operator.equals(">") && nToken.equals("=") ||
//           operator.equals("=") && nToken.equals("=") ||
//           operator.equals("!") && nToken.equals("=") ||
//           operator.equals("<") && nToken.equals(">") )
//        operator += nToken;
//      else
//        tok.pushback();
//
//      // Naechsten Operand parsen
//      nextOperand = parseOperand();
//
//      // Teil zusammensetzen
//      rulePart = new OperationTree.OperatorNode(operator,rulePart,nextOperand);
//    }
//    // keine Token mehr, aber noch offene Klammern
//    if ( openBrackets.length() > 0 )
//      throwParseError("Bracket"+(openBrackets.length() > 1 ? "s" : "")+" '"+openBrackets.toString()+"' not closed!");
//    return rulePart;
//  }
//
//  /**
//   * Parst den naechsten Operand aus dem Tokenizer {@link #tok}. Ein Operand ist eine
//   * durch Punkt-Operatoren verbundene Kette von {@linkplain #parseLiteral() Literalen}
//   * und geklammerten {@linkplain #parseRulePart() Teil-Formeln}.<br>
//   */
//  protected final BinaryTreeNode parseOperand() {
//    BinaryTreeNode operand = null;
//    BinaryTreeNode nextLiteral = null;
//    String operator = null;
//
//    // Erstes (linkes) Literal parsen
//    String token = nextNonWSToken();
//    // oeffnende Klammer direkt am Anfang
//    // -> Literal besteht aus kompletter Teilformel
//    if (isOpeningBracket(token)) {
//      tok.pushback();
//      operand = parseRulePart();
//    } else {
//      // Sonderfall: Operand startet mit Minus-Zeichen (z.B. '-(1+2)' oder '-abs(-4)')
//      //             -> folgende
//      if ( token.equals("-") )
//        operand = new OperationTree.OperatorNode("*", new OperationTree.ConstantNode(-1.0),  parseOperand());
//      else {
//        tok.pushback();
//        operand = parseLiteral();
//      }
//    }
//    while (tok.hasMoreTokens()) {
//      // Operator parsen
//      operator = nextNonWSToken();
//      // wenn schliessende Klammer oder Strich-/Bool-Operator
//      // -> Operand beenden
//      if (isClosingBracket(operator) || !isMultiplyOperator(operator)) {
//        tok.pushback();
//        return operand;
//      }
//      // Pruefen, ob es wirklich ein Operator ist
//      if ( !isOperatorChar(operator) )
//        throwParseError("Illegal operator '" + operator + "'");
//
//      // Naechstes Literal
//      token = nextNonWSToken();
//      tok.pushback();
//      // oeffnende Klammer
//      // -> naechstes Literal ist eine komplette Teilformel
//      if (isOpeningBracket(token))
//        nextLiteral = parseRulePart();
//      else
//        nextLiteral = parseLiteral();
//
//      // Operand zusammensetzen
//      operand = new OperationTree.OperatorNode(operator, operand, nextLiteral);
//    }
//
//    return operand;
//  }

  /**
   * Liesst ein weiteres Zeichen aus dem Tokenizer und fuegt dies gegebenfalls
   * dem Operator hinzu. Dies ist z.B. bei den Vergleichsoperatoren
   * <=, >=, ==, != und == notwendig.
   * @param operator Operator
   */
  private String expandOperator(String operator) {
      // Sonderfall: boolesche Operatoren u.U. 2-stellig (<=, >=, !=, <>, ==)
      String nToken = nextNonWSToken();
      if ( operator.equals("<") && nToken.equals("=") ||
           operator.equals(">") && nToken.equals("=") ||
           operator.equals("=") && nToken.equals("=") ||
           operator.equals("!") && nToken.equals("=") ||
           operator.equals("<") && nToken.equals(">") )
        operator += nToken;
      else
        pushbackWithWSToken();
      return operator;
  }

  /**
   * Prueft, ob der aktuelle Operand oder die Teil-Formel aufrund einer
   * schliessende Klammer oder einem Parameter-Trennzeichen abgeschlossen
   * werden muss.
   * @param token aktuelles Zeichen
   */
  private RuleClosingType checkRuleClosing(String token) {
    // Operand/Teil-Formel wird nur bei schliessender Klammer oder
    // Parameter-Trennzeichen beendet
    if ( !isClosingBracket(token) &&
         !isParameterSeperator(token) )
      return RuleClosingType.none;

    // aktuelle Klammerung ermitteln
    if ( openBrackets.length() == 0 )
        throwParseError( LangUtil.RESOURCE.getString("OperationTree.err.TokenNotExpected",token) );
    char openBracket = openBrackets.charAt(0);

    // Schliessende Klammer
    if ( isClosingBracket(token) ) {
        // Flag, ob Klammer eine Funktion mit Parametern f(.,.,...) beendet.
        // In diesem Fall liegt die verbleibende Anzahl (0) an Parametern
        // auf dem Stack
        boolean funcEnding = (openBracket == '0');
        if ( funcEnding )
          // korrespondierende oeffnende Klammer liegt vor Parameter-Wert
          // auf Stack
          openBracket = openBrackets.charAt(1);

      // Wenn keine oeffnende Klammer auf Stack liegt (sondern ein Parameter-Zaehler)
      // kann Klammer nicht schliessen, da noch ein Parameter erwartet wird
      if ( !isOpeningBracket(String.valueOf(openBracket)) )
        throwParseError( LangUtil.RESOURCE.getString("OperationTree.err.MoreParamsExpected",openBracket) );
      if ( !checkCorrespondingBrackets(openBracket,token.charAt(0)) )
        throwParseError( LangUtil.RESOURCE.getString("OperationTree.err.UnexpectedBracket",token,openBracket) );

      return funcEnding ? RuleClosingType.func_pram : RuleClosingType.rule_part;
    }

    // Trenner zwischen Funktions-Parametern
    if ( isParameterSeperator(token) ) {
      if ( isOpeningBracket(String.valueOf(openBracket)) )
        throwParseError( LangUtil.RESOURCE.getString("OperationTree.err.TokenNotExpected",token) );
      return RuleClosingType.func_pram;
    }

    return RuleClosingType.none;
  }

  /**
   * Prueft, ob noch offene Klammern auf dem Stack liegen, obwohl keine
   * Token mehr zu parsen sind. Ist dies der Fall wird eine Exception
   * geworfen.
   */
  private void checkRemainingBrackets() {
    // keine Token mehr, aber noch offene Klammern
    if ( !tok.hasMoreTokens() && openBrackets.length() > 0 )
      throwParseError( LangUtil.RESOURCE.getString("OperationTree.err.BracketsNotClosed",openBrackets.toString()) );
  }

  /**
   * Parst eine komplette Formel, bis zum Ende der Eingabe-Formel oder
   * einem Parameter-Trenner.
   */
  protected final BinaryTreeNode parseRule() {
    String         operator     = null;
    BinaryTreeNode rulePart     = null;
    BinaryTreeNode nextOperand  = null;

    // Ersten Formel-Teil parsen
    rulePart = parseRulePart();

    while ( tok.hasMoreTokens() ) {
      // Operator parsen
      operator = nextNonWSToken();

      // Ende der Formel pruefen, kann eigentlich nur das Ende eines
      // Funktions-Parameters sein, da schliessende Klammern sonst bereits
      // entfernt wurden...
      if ( checkRuleClosing(operator) != RuleClosingType.none ) {
        pushbackWithWSToken();
        return rulePart;
      }

      // Pruefen, ob es wirklich ein Operator ist
      if ( !isOperatorChar(operator) )
        throwParseError( LangUtil.RESOURCE.getString("OperationTree.err.IllegalOperator", operator) );
      // Sonderfall: boolesche Operatoren u.U. 2-stellig (<=, >=, !=, <>, ==)
      operator = expandOperator(operator);

      // Naechsten Operand parsen
      nextOperand = parseOperand( getOperatorConnectivity(operator) );

      // Teil zusammensetzen
      rulePart = new OperationTree.OperatorNode(operator,rulePart,nextOperand);
    }

    checkRemainingBrackets();

    return rulePart;
  }

  /**
   * Parst die naechste Teil-Formel aus dem Tokenizer {@link #tok}. Eine (geklammerter)
   * Teilformel ist eine durch Strich-Operatoren verbundene Kette von
   * {@linkplain #parseOperand(int) Operanden}.
   */
  protected final BinaryTreeNode parseRulePart() {
    BinaryTreeNode rulePart    = null;
    BinaryTreeNode nextOperand = null;
    String         operator    = null;

    // Ersten (linken) Operand parsen
    String token = nextNonWSToken();
    // oeffnende Klammer
    if ( isOpeningBracket(token) )
      openBrackets.insert(0,token);
    else
      pushbackWithWSToken();

    rulePart = parseOperand( getMaximumOperatorConnectivity() );
    while ( tok.hasMoreTokens() ) {
      // Operator parsen
      operator = nextNonWSToken();

      // schliessende Klammer oder Parameter-Trenner beendet die Teil-Formel
      RuleClosingType closingType = checkRuleClosing(operator);
      if ( closingType != RuleClosingType.none ) {
        // Wenn Teil-Formel beendet wird, dann schliessende Klammer
        // vom Stack entfernen
        if ( closingType == RuleClosingType.rule_part )
          openBrackets.deleteCharAt(0);
        // Wenn Functions-Parameter beendet wird, dann Token zuruecklegen,
        // damit Beendigung der Formel (in uebergeordneter Methode)
        // erkannt wird
        if ( closingType == RuleClosingType.func_pram )
          pushbackWithWSToken();
        // (Teil-)Formel beenden
        return rulePart;
      }

      // Pruefen, ob es wirklich ein Operator ist
      if ( !isOperatorChar(operator) )
        throwParseError( LangUtil.RESOURCE.getString("OperationTree.err.IllegalOperator", operator) );
      // Sonderfall: boolesche Operatoren u.U. 2-stellig (<=, >=, !=, <>, ==)
      operator = expandOperator(operator);

      // Naechsten Operand parsen
      nextOperand = parseOperand( getOperatorConnectivity(operator) );

      // Teil zusammensetzen
      rulePart = new OperationTree.OperatorNode(operator,rulePart,nextOperand);
    }

    checkRemainingBrackets();

    return rulePart;
  }

  /**
   * Parst den naechsten Operand aus dem Tokenizer {@link #tok}. Ein Operand ist eine
   * durch Operatoren gleicher Konnektivitaet verbundene Kette von
   * {@linkplain #parseLiteral() Literalen} und geklammerten
   * {@linkplain #parseRulePart() Teil-Formeln}.<br>
   * @param minOperandConn minimale Konnektivitaet fuer den Operanden (bei einem
   *                       Operator mit kleinerer Konnektivitaet wird der
   *                       Operand beendet)
   */
  protected final BinaryTreeNode parseOperand(int minOperatorConn) {
    BinaryTreeNode operand = null;
    BinaryTreeNode nextLiteral = null;
    String operator = null;
    int lastOperatorConn = this.getMaximumOperatorConnectivity();

    // Erstes (linkes) Literal parsen
    String token = nextNonWSToken();
    // oeffnende Klammer direkt am Anfang
    // -> Literal besteht aus kompletter Teilformel
    if (isOpeningBracket(token)) {
      pushbackWithWSToken();
      operand = parseRulePart();
    } else {
      // Sonderfall: Operand startet mit Minus-Zeichen (z.B. '-(1+2)' oder '-abs(-4)')
      //             -> folgende
      if ( token.equals("-") )
        operand = new OperationTree.OperatorNode("*", new OperationTree.ConstantNode(-1.0),  parseOperand(getOperatorConnectivity("*")));
      else {
        pushbackWithWSToken();
        operand = parseLiteral();
      }
    }
    while (tok.hasMoreTokens()) {
      // Operator parsen
      operator = nextNonWSToken();
      // schliessende Klammer oder Parameter-Trenner beendet den Operand
      if ( checkRuleClosing(operator) != RuleClosingType.none) {
        pushbackWithWSToken();
        return operand;
      }

      // Pruefen, ob es wirklich ein Operator ist
      if ( !isOperatorChar(operator) )
        throwParseError( LangUtil.RESOURCE.getString("OperationTree.err.IllegalOperator", operator) );
      // Sonderfall: boolesche Operatoren u.U. 2-stellig (<=, >=, !=, <>, ==)
      operator = expandOperator(operator);

      int operatorConn = getOperatorConnectivity(operator);
      // Operator mit geringerer Konnektivitaet beendet den Operand
      if ( operatorConn < minOperatorConn ) {
        for (int i=0; i<operator.length(); i++)
        pushbackWithWSToken();
        return operand;
      }

//      // Operator mit hoeherer Konnektivitaet erzeugt einen neuen Operand
//      // Ausnahme: erste Operation im Operand ('operand' hat noch keine Kinder)!
//      if ( operatorConn > minOperandConn && !operand.isLeaf() ) {

      // Operator mit hoeherer Konnektivitaet, als der letzte Operator, erzeugt
      // einen neuen Operand
      if ( operatorConn > lastOperatorConn ) {
        BinaryTreeNode operandHigher = parseOperand(operatorConn);
        // Das letzte Literal ist der erste Operand von 'operandHigher'
        operand.setRightChild(
          new OperationTree.OperatorNode(
                  operator,
                  operand.getRightChild(),
                  operandHigher
          )
        );
        // mit naechstem Operator fortfahren
        lastOperatorConn = operatorConn;
        continue;
      }

      // Naechstes Literal
      token = nextNonWSToken();
      pushbackWithWSToken();
      // oeffnende Klammer
      // -> naechstes Literal ist eine komplette Teilformel
      if (isOpeningBracket(token))
        nextLiteral = parseRulePart();
      else
        nextLiteral = parseLiteral();

      // Operand zusammensetzen
      operand = new OperationTree.OperatorNode(operator, operand, nextLiteral);

      lastOperatorConn = operatorConn;
    }

    return operand;
  }



  /**
   * Parst das naechste Literal aus dem Tokenizer {@link #tok}. Ein Literal ist eine numerische
   * Konstante, ein Alias oder ein 1-stelliger Operator auf einer geklammerten
   * {@linkplain #parseRulePart() Teil-Formel}. Diese Methode unterstuetzt
   * <ul>
   *   <li>als 1-stellige Operatoren
   *      <ul>
   *         <li>Betrag: {@code "abs(.)"}</li>
   *         <li>Wurzel: {@code "sqr(.)"} oder {@code "sqrt(.)"}</li>
   *         <li>Runden: {@code "rnd(.)"} oder {@code "round(.)"}</li>
   *         <li>Abschneiden: {@code "int(.)"} oder {@code "trunc(.)"}</li>
   *         <li>boolesches NICHT: {@code "!(.)"} oder {@code "not(.)"}</li>
   *         <li>Test auf "Not a Number" (NaN): {@code "isNaN(.)"}</li>
   *         <li>Sinus: {@code "sin(.)"}</li>
   *         <li>Cosinus: {@code "cos(.)"}</li>
   *         <li>Tangens: {@code "tan(.)"}</li>
   *         <li>Arcus-Sinus: {@code "arcsin(.)"} oder {@code "asin(.)"}</li>
   *         <li>Arcus-Cosinus: {@code "arccos(.)"} oder {@code "acos(.)"}</li>
   *         <li>Arcus-Tangens: {@code "arctan(.)"} oder {@code "atan(.)"}</li>
   *         <li>Exponentialfunktion: {@code "exp(.)"}</li>
   *         <li>Logarithmus zur Basis e: {@code "ln(.)"}</li>
   *         <li>Logarithmus zur Basis 10: {@code "log(.)"}</li>
   *         <li>Umwandlung Zahl-zu-String: {@code "str(.)"}</li>
   *         <li>Umwandlung String-zu-Zahl: {@code "val(.)"}</li>
   *         <li>Laenge eines Strings: {@code "len(.)"}</li>
   *         <li>Umwandlung in Gross-Buchstaben: {@code "toupper(.)"}</li>
   *         <li>Umwandlung in Klein-Buchstaben: {@code "tolower(.)"}</li>
   *       </ul></li>
   *   <li>als Alias (0-stellige Operatoren)
   *       <ul>
   *         <li>Zufallszahl zwischen 0 und 1: {@code "rand"} oder {@code "random"}</li>
   *         <li>"Not a number" (NaN): {@code "NaN"}</li>
   *       </ul></li>
   *   <li>als Sonderfaelle
   *       <ul>
   *         <li>IF .. THEN .. ELSE ..: {@code "ITE(.,.,.)"}</li>
   *       </ul></li>
   * </ul>
   * <b>Sub-Klassen muessen diese Methode erweitern, wenn z.B. (alphanumerische)
   * Variablen oder zusaetzliche 1-stellige Operatoren beruecksichtigt werden
   * sollen!!</b>
   */
  protected BinaryTreeNode parseLiteral() {
    String token = nextNonWSToken().toLowerCase();

    // Sonderfall: in Anfuehrungsstriche gekapselte String-Konstante
    if ( isStringEncapsulationChar(token) ) {
      // Token (Anfuehrungsstrichen) erweitern um alles, bis zum naechsten
      // Anfuehrungsstrich gleicher Art
      String encChar = token;
      // Bemerkung: tok.nextToken(encChar) funktioniert nicht so richtig,
      //            wohl wegen PushBack-Funktionalitaet. Zudem aendert
      //            dieser Aufruf die Delimiter dauerhaft!
      do {
        token += tok.nextToken();
      } while (!token.endsWith(encChar));
    }


    // Betrag, Wurzel, Runden
    if ( token.equals("abs") ||
         token.equals("sqr") || token.equals("sqrt") ||
         token.equals("rnd") || token.equals("round") ||
         token.equals("int") || token.equals("trunc") ||
         token.equals("!")   || token.equals("not") ||
         token.equals("isnan") ||
         token.equals("sin") ||
         token.equals("cos") ||
         token.equals("tan") ||
         token.equals("arcsin") || token.equals("asin") ||
         token.equals("arccos") || token.equals("acos") ||
         token.equals("arctan") || token.equals("atan") ||
         token.equals("exp") ||
         token.equals("log") ||
         token.equals("ln") ||
         token.equals("str")     || token.equals("val") ||
         token.equals("len") ||
         token.equals("toupper") || token.equals("tolower") ) {
      return new OperationTree.UnaryOperatorNode(token, parseRulePart());
    }
    // Zufallszahl oder NaN-Konstante
    if ( token.equals("rand") || token.equals("random") ||
         token.equals("nan")  || token.equals("curr_millis")) {
      return new OperationTree.ConstantAliasNode(token);
    }
    // ITE(a,b,c) = IF a THEN b ELSE c
    if ( token.equals("ite") ) {
        // Parameter fuer ITE parsen
        BinaryTreeNode[] iteParam = parseFunctionParameters("ITE",3);
        // Baum fuer ITE
        return new OperationTree.ITENode(token,iteParam[0],iteParam[1],iteParam[2]);
    }

    // REGEX(string,regexpr) = Regular Expression
    if ( token.equals("regex") ) {
      // Parameter fuer REGEX parsen
      BinaryTreeNode[] regexParam = parseFunctionParameters("REGEX",2);
      // Baum fuer REGEX
      return new OperationTree.OperatorNode(token,regexParam[0],regexParam[1]);
    }

    // SPLIT(string,regexpr) = String-Split
    if ( token.equals("split") ) {
      // Parameter fuer SPLIT parsen
      BinaryTreeNode[] splitParam = parseFunctionParameters("SPLIT",3);
      // Baum fuer SPLIT
      return new OperationTree.MultiParamOperatorNode(token,null,splitParam);
    }

    // SUBSTR(string,startPos_inkl,endPos_exkl) = Teil-String
    if ( token.equals("substr") ) {
      // Parameter fuer SUBSTR parsen
      BinaryTreeNode[] substrParam = parseFunctionParameters("SUBSTR",3);
      // Baum fuer SPLIT
      return new OperationTree.MultiParamOperatorNode(token,null,substrParam);
    }

    // REPLALL(string,pattern,replacement) = String-Ersetzung
    if ( token.equals("replall") ) {
      // Parameter fuer REPLALL parsen
      BinaryTreeNode[] substrParam = parseFunctionParameters("REPLALL",3);
      // Baum fuer REPLALL
      return new OperationTree.MultiParamOperatorNode(token,null,substrParam);
    }

    // sonst: Konstante
    return new OperationTree.ConstantNode(getConstantFromString(token));
  }

  /**
   * Parst die Funktions-Parameter fuer eine Funktion F(.,.,.,.).
   * @param operator Name des Operators (nur fuer Exception-Beschreibung)
   * @param paramCount Anzahl der erwarteten Parameter
   */
  private BinaryTreeNode[] parseFunctionParameters(String operator, int paramCount) {
    BinaryTreeNode[] paramNode = new BinaryTreeNode[paramCount];

    // naechstes Token muss eine oeffnende Klammer sein!
    String token = nextNonWSToken();
    if ( !token.equals("(") )
    throwParseError( LangUtil.RESOURCE.getString("OperationTree.err.IllegalOperatorSyntax", operator, token) );
    openBrackets.insert(0,"(");

    openBrackets.insert(0, String.valueOf(paramCount-1)); // noch X weitere Parameter (Trenner) erwartet
    // parCount-1 Parameter parsen
    for (int i=0; i<paramCount-1; i++) {
      // Parameter fuer ITE lesen
      paramNode[i] = parseRule();
      if ( !isParameterSeperator( tok.nextToken() ) ) // Seperator entfernen
        throwParseError( LangUtil.RESOURCE.getString("OperationTree.err.ParamSepExpected") );
      openBrackets.replace(0, 1, String.valueOf(paramCount-(i+2))); // noch X-1 weiterer Parameter (Trenner) erwartet
    }
    // den letzten Parameter parsen
    paramNode[paramNode.length-1] = parseRule();
    // Funktion-schliessende Klammer entfernen (Pruefung auf Konsistenz mit
    // oeffnender Klammer erfolgte bereits in checkRuleClosing(.)
    tok.nextToken();
    // Parameter-Zaehler von Stack entfernen
    openBrackets.deleteCharAt(0);
    // Funktion-oeffnende Klammer von Stack entfernen
    openBrackets.deleteCharAt(0);

    return paramNode;
  }

  ///////////////////////////////////////////////////////////////////
  //////////////////////   Hilfsmethoden   //////////////////////////
  ///////////////////////////////////////////////////////////////////
  /**
   * Wirft eine {@link IllegalArgumentException} mit der angegebenen Fehlermeldung.
   * Als zusaetzlicher Hinweis wird das Formel-Praefix ausgegeben nach dem
   * der Fehler aufgetreten ist.
   * @param mess Fehler-Meldung
   */
  protected void throwParseError(String mess) {
    throwParseError(mess,null);
  }

  /**
   * Wirft eine {@link IllegalArgumentException} mit der angegebenen Fehlermeldung.
   * Als zusaetzlicher Hinweis wird das Formel-Praefix ausgegeben nach dem
   * der Fehler aufgetreten ist.
   * @param mess Fehler-Meldung
   * @param err  verursachender Fehler
   */
  protected void throwParseError(String mess, Throwable err) {
    String errorPos = tok.getReadTokens();
//    // Die Klammer entfernen, die am Anfang kuenstlich eingefuegt wurde
//    errorPos = errorPos.substring(1);
    if ( err != null )
      throw new IllegalArgumentException( LangUtil.RESOURCE.getString("OperationTree.err.ParseErrorMess", errorPos, mess), err );
    else
      throw new IllegalArgumentException( LangUtil.RESOURCE.getString("OperationTree.err.ParseErrorMess", errorPos, mess) );
  }

  /**
   * Liefert einen Double- oder String-Wert aus einem String. String-Konstanten
   * muessen in " oder ' gekapselt sein, um als solche interpretiert zu werden.
   * Die Wildcards {@code TRUE} und {@code FALSE} werden in 1 bzw. 0 umgewandelt.
   * @param token umzuwandelnde Zeichenkette
   * @exception IllegalArgumentException falls die Umwandlung nicht vorgenommen
   *            werden kann
   */
  protected Object getConstantFromString(String token) {
    // Sonderfall: String-Konstante in Anfuehrungsstrichen
    if ( token.length()>0 && isStringEncapsulationChar(token.substring(0,1)) )
      return token.substring(1, token.length()-1);
    // Sonderfall: TRUE/FALSE als Wildcards fuer 1/0
    if ( token.equalsIgnoreCase("TRUE") )
      return 1;
    if ( token.equalsIgnoreCase("FALSE") )
      return 0;
    // Sonst eine numerische Konstante
    return getDoubleFromString(token);
  }

  /**
   * Liefert einen Integer-Wert aus einem String.
   * @param token umzuwandelnde Zeichenkette
   * @exception IllegalArgumentException falls die Umwandlung nicht vorgenommen
   *            werden kann
   */
  protected int getIntFromString(String token) {
    token = token.trim();
    try {
      return Integer.parseInt(token);
    }
    catch (Exception err) {
      throwParseError( LangUtil.RESOURCE.getString(
                                 "OperationTree.err.IllegalCharacter",
                                 token,
                                 LangUtil.RESOURCE.getString("OperationTree.Integer")
      ));
    }
    return 0;
  }

  /**
   * Liefert einen Double-Wert aus einem String.
   * @param token umzuwandelnde Zeichenkette
   * @exception IllegalArgumentException falls die Umwandlung nicht vorgenommen
   *            werden kann
   */
  protected double getDoubleFromString(String token) {
    token = token.trim();
    // Sonderfall: Bei alleinstehender negativer Konstante ist im Token nur
    //             das Minus-Zeichen enthalten! Zahl-Token muss angehaengt werden.
    if ( token.equals("-") )
      token += nextNonWSToken();

    try {
      return Double.parseDouble(token);
    }
    catch (Exception err) {
      throwParseError( LangUtil.RESOURCE.getString(
                               "OperationTree.err.IllegalCharacter",
                               token,
                               LangUtil.RESOURCE.getString("OperationTree.Number")
      ));
    }
    return 0;
  }

  /**
   * Liefert das naechste Token aus dem Tokenizer, das nicht ein White-Space
   * ist.
   */
  protected String nextNonWSToken() {
    String token = null;
    while ( (token = tok.nextToken()).trim().equals("") );
    return token.trim();
  }

  /**
   * Legt solange Token zurueck auf den Tokenizer, bis das zurueckgelegte
   * Token <b>kein</b> White-Space ist.
   */
  protected void pushbackWithWSToken() {
    while ( tok.pushback().trim().equals("") );
  }

  ////////////////////////////////////////////////////////////////
  /////////////////   ParserOperatorsHints   /////////////////////
  ////////////////////////////////////////////////////////////////
  /**
   * Initialises the list of available operators.
   * @see #getOperators()
   * @see #avOperators
   */
  protected void initOperatorList() {
    // Zur Verfuegung stehende Operatoren
    avOperators.clear();
    avOperators.add("&");           
    avOperators.add("|");           
    avOperators.add("!");           
    avOperators.add("abs");         
    avOperators.add("sqrt");        
    avOperators.add("round");       
    avOperators.add("trunc");       
    avOperators.add("isNaN");       
    avOperators.add("NaN");         
    avOperators.add("random");      
    avOperators.add("sin");         
    avOperators.add("cos");         
    avOperators.add("tan");         
    avOperators.add("asin");        
    avOperators.add("acos");        
    avOperators.add("atan");        
    avOperators.add("exp");         
    avOperators.add("ln");          
    avOperators.add("log");         
    avOperators.add("str");         
    avOperators.add("val");         
    avOperators.add("len");         
    avOperators.add("toupper");     
    avOperators.add("tolower");     
    avOperators.add("ITE");         
    avOperators.add("REGEX");       
    avOperators.add("SUBSTR");      
    // Die "einfachen Operatoren" and Ende setzen, da diese eigentlich
    // menuell eingegeben werden und nicht so oft gebraucht werden
    avOperators.add("+"); 
    avOperators.add("-"); 
    avOperators.add("*"); 
    avOperators.add("/"); 
    avOperators.add("^"); 
    avOperators.add("="); 
    avOperators.add("<>");
    avOperators.add("<"); 
    avOperators.add("<=");
    avOperators.add(">"); 
    avOperators.add(">=");
    
    initOperatorKeys();
  }
  
  /**
   * Initialises the localisation key list for the available operators.
   * @see #avOperatorsKeys
   */
  protected void initOperatorKeys() {
    avOperatorsKeys.clear();
    // Usually the operator can be uses as key for localization...
    for (String op : avOperators)
      avOperatorsKeys.put(op,op);
    // ... but some operators can not be used as keys, so they are replaced
    // with special keys
    avOperatorsKeys.put("&","and");
    avOperatorsKeys.put("|","or");
    avOperatorsKeys.put("!","not");
    avOperatorsKeys.put("+","plus");
    avOperatorsKeys.put("-","minus");
    avOperatorsKeys.put("*","multiply");
    avOperatorsKeys.put("/","divide");
    avOperatorsKeys.put("^","pow");
    avOperatorsKeys.put("=","eq");
    avOperatorsKeys.put("<>","ne");
    avOperatorsKeys.put("<","lt");
    avOperatorsKeys.put("<=","le");
    avOperatorsKeys.put(">","gt");
    avOperatorsKeys.put(">=","ge");
  }
  
  /**
   * Returns the operators, which are available for creating a
   * valid rule.<br> 
   * <b>This method is not used for the parsing process, but only
   * for GUI/User interaction!</b>  
   */
  @Override
  public Vector<String> getOperators() {
    return avOperators;
  }

  /**
   * Returns a short (internationalised) title for an operator. 
   * @param operator an operator
   * @return {@code null} if operator is undefined
   */
  @Override
  public String getOperatorTitle(String operator) {
    String opKey = avOperatorsKeys.get(operator);
    if ( opKey == null )
      return null;
    return SwingUtil.RESOURCE.getString("OperationTreePanel.OpDesc."+opKey);
  }

  /**
   * Returns an (internationalised) description for an operator. 
   * @param operator an operator
   * @return {@code null} if operator is undefined
   */
  @Override
  public String getOperatorDescription(String operator) {
    String opKey = avOperatorsKeys.get(operator);
    if ( opKey == null )
      return null;
    return SwingUtil.RESOURCE.getString("OperationTreePanel.OpTooltip."+opKey);
  }

  /**
   * Returns the number of bracked-enclosed parameters for an operator.
   * @param op an operator
   * @return {@code 0} if operator is undefined
   */
  @Override
  public int getOperatorParameterCount(String op) {
    // Funktionen mit einem geklammerten Parameter
    if (op.equalsIgnoreCase("abs") || op.equalsIgnoreCase("sqrt") ||
        op.equalsIgnoreCase("round") || op.equalsIgnoreCase("trunc") ||
        op.equalsIgnoreCase("isNaN") || op.equalsIgnoreCase("sin") ||
        op.equalsIgnoreCase("cos") || op.equalsIgnoreCase("tan") ||
        op.equalsIgnoreCase("asin") || op.equalsIgnoreCase("acos") ||
        op.equalsIgnoreCase("atan") || op.equalsIgnoreCase("exp") ||
        op.equalsIgnoreCase("ln") || op.equalsIgnoreCase("log") ||
        op.equalsIgnoreCase("val") || op.equalsIgnoreCase("str") ||
        op.equalsIgnoreCase("toupper") || op.equalsIgnoreCase("tolower") ||
        op.equalsIgnoreCase("!") )
      return 1;

    // Funktionen mit zwei geklammerten Parameter
    if (op.equalsIgnoreCase("REGEX") )
      return 2;

    // Funktionen mit drei geklammerten Parameter
    if (op.equalsIgnoreCase("ITE") || op.equalsIgnoreCase("SUBSTR") )
      return 3;

    // sonst: Konstante/Alias angenommen
    return 0;
  }


}
