/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.lang.tree;

/**
 * Diese Klasse stellt einen Knoten in einem Baum dar.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class TreeNode<E> {
  /** Speichert den Vater-Knoten. */
  protected TreeNode<E> parent = null;
  /** Speichert das Objekt, das in dem Knoten gespeichert wird. */
  protected E object = null;

  /**
   * Erzeugt einen Wurzel-Knoten ohne Nachfolger.
   */
  public TreeNode() {
    this((E)null);
  }

  /**
   * Erzeugt einen Wurzel-Knoten ohne Nachfolger.
   * @param object Objekt, das in dem Knoten gespeichert wird
   */
  public TreeNode(E object) {
    this(object, null);
  }

  /**
   * Erzeugt einen Blatt-Knoten.
   * @param parent direkter Vorgaenger-Knoten
   */
  public TreeNode(TreeNode<E> parent) {
    this(null, parent);
  }

  /**
   * Erzeugt einen Blatt-Knoten.
   * @param object Objekt, das in dem Knoten gespeichert wird
   * @param parent direkter Vorgaenger-Knoten
   */
  public TreeNode(E object, TreeNode<E> parent) {
    setObject( object );
    setParent( parent );
  }

  /**
   * Liefert den Wurzel-Knoten.
   */
  public TreeNode<E> getRoot() {
    if ( this.isRoot() )
      return this;
    return parent.getRoot();
  }

  /**
   * Prueft, ob ein Knoten als Vater- oder Kind-Knoten fuer diesen Knoten
   * verwendet werden kann
   * @param node zu pruefende Knoten
   * @exception RuntimeException falls der Knoten nicht geeignet ist
   */
  public abstract void checkNode(TreeNode<E> node);

  /**
   * Prueft, ob es sich um einen Wurzelknoten handelt.
   * @return <code>true<code> gdw. der Knoten keinen Vater hat
   * @see #getParent()
   */
  public boolean isRoot() {
    return getParent() == null;
  }

  /**
   * Prueft, ob es sich um einen Blattknoten handelt.
   * @return <code>true<code> gdw. alle Kinder {@code null} sind
   */
  public boolean isLeaf() {
    for (int i=0; i<getChildCount(); i++)
      if ( getChild(i) != null )
        return false;
    return true;
  }

  /**
   * Prueft, ob es sich um einen inneren Knoten handelt.
   * @return <code>true<code> gdw. der Knoten keine Wurzel und kein Blatt ist
   */
  public boolean isInnerNode() {
    return !isRoot() && !isLeaf();
  }

  /**
   * Liefert den direkten Vaterknoten.
   */
  public TreeNode<E> getParent() {
    return parent;
  }

  /**
   * Setzt den direkten Vorgaenger-Knoten
   * @param parent neuer direkter Vorgaenger
   */
  public void setParent(TreeNode<E> parent) {
    checkNode(parent);
    this.parent = parent;
  }

  /**
   * Liefert die Anzahl der Kindknoten.
   */
  public abstract int getChildCount();

  /**
   * Liefert einen Kindknoten.
   * @param i Index (beginnend bei 0)
   */
  public abstract TreeNode<E> getChild(int i);

  /**
   * Setzt einen Kindknoten.
   * @param i Index (beginnend bei 0)
   * @param child ein Kindknoten
   */
  public abstract void setChild(int i, TreeNode<E> child);

  /**
   * Liefert das Objekt, das in dem Knoten gespeichert ist.
   */
  public E getObject() {
    return object;
  }

  /**
   * Setzt das Objekt, das in dem Knoten gespeichert ist.
   */
  public void setObject(E object) {
    this.object = object;
  }
}
