/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing;

import javax.swing.JCheckBox;

/**
 * Diese Klasse stellt eine boolsche Eingabe-Option fuer das {@link MultipleOptionPane}
 * dar. Diese wird durch eine {@link JCheckBox} dargestellt.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class BooleanInputOption extends InputOption {
  /**
   * Erzeugt eine neue Eingabe-Option.
   * @param label    Beschreibung
   * @param defValue gibt an, ob die Checkbox standardmaessig aktiviert ist
   */
  public BooleanInputOption(String label, boolean defValue) {
    super(label,false);
    // Bei einer Checkbox soll die Beschreibung nicht oberhalb
    // des Eingabefeldes zu stehen, sondern mit der Checkbox in
    // einer Zeile
    // -> JLabel entfernen und statt dessen den Text der Checkbox setzen
    this.remove( this.descLabel );
    ((JCheckBox)inpComp).setText(label);
    // Standardwert setzen
    setValue( defValue );
  }

  /**
   * Erzeugt eine neue Eingabe-Option. Diese ist standardmaessig deaktiviert.
   * @param label Beschreibung
   */
  public BooleanInputOption(String label) {
    this(label,false);
  }

  /**
   * Erzeugt eine neue Instanz von {@link JCheckBox}.
   */
  protected JCheckBox createInputComponent() {
    return new JCheckBox();
  }

  /**
   * Liefert den Wert, der in der Option eingegeben wurde. Prueft zuerst
   * auf Gueltigkeit und ruft dann {@link #performGetValue()} auf.
   * @return <code>null</code> wenn die aktuelle Eingabe nicht zulaessig ist
   *         (siehe {@link #isInputValid()})
   */
  public Boolean getValue() {
    return (Boolean)super.getValue();
  }

  /**
   * Liefert immer <code>true</code>, da keine speziellen Anforderungen
   * an die Text-Eingabe gestellt werden.
   */
  protected boolean performIsInputValid() {
    return true;
  }

  /**
   * Liefert immer <code>false</code>, da es keine Leereingabe gibt. Eine nicht
   * aktivierte Checkbox bedeutet den Wert FALSE.
   */
  protected boolean performIsInputEmpty() {
    return false;
  }

  /**
   * Liefert <code>true</code>, falls die Checkbox aktiviert ist.
   * @return einen <code>boolean</code>-Wert (<b>kein</b> {@link Boolean}-Objekt!)
   */
  protected Boolean performGetValue() {
    return ((JCheckBox)inpComp).isSelected();
  }

  /**
   * (De)aktiviert die Checkbox.
   * @param active definiert den neuen Status (muss vom Typ {@link Boolean} sein!)
   * @return <code>false</code> falls kein {@link Boolean} uebergeben wird.
   */
  protected boolean performSetValue(Object active) {
    if ( !(active instanceof Boolean) )
      return false;
    ((JCheckBox)inpComp).setSelected((Boolean)active);
    return true;
  }
}
