/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing;

import java.awt.BorderLayout;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JButton;
import javax.swing.JTextField;

import schmitzm.swing.event.InputOptionAdapter;

/**
 * Die <code>BrowseInputOption</code> erweitert die {@linkplain ManualInputOption manuelle Text-Eingabe}
 * um eine Browse-Aktion. Welche konkrete Aktion beim Klick auf den Browse-Button
 * ausgefuehrt wird, bestimmt die jeweinige Implementierung.<br>
 * Der Wert dieser Eingabe-Option stellt keinen Text dar, sondern ein
 * zugehoeriges Objekt. Um ein Objekt sowohl ueber eine manuelle Texteingabe,
 * als auch ueber die Browse-Aktion angeben zu koennen, spezifiziert
 * <code>BrowseInputOption</code> zwei Methoden ({@link #convertFromString(String)},
 * {@link #convertToString(Object)}), ueber die eine entsprechende Umwandung zu
 * implementieren ist.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class BrowseInputOption  extends ManualInputOption {
  /** Speichert den aktuellen Wert der Eingabe-Option. */
  protected Object object = null;
  /** Container enthaelt das Text-Eingabefeld und den Browse-Button. */
  protected Container inpCompCont = new Container();
  /** Der Button fuer die Browse-Aktion. */
  protected JButton browseButton = new JButton("...");

  /**
   * Erzeugt eine neue Eingabe-Option.
   * @param label Beschreibung der Eingabe
   * @param inputNeeded gibt an, ob eine Eingabe in der Option erforderlich ist
   * @param defValue Standard-Belegung fuer die Eingabe-Option
   */
  public BrowseInputOption(String label, boolean inputNeeded, Object defValue) {
    super(label, inputNeeded, null);
    setValue(defValue);

    // Browse-Button formatieren
    browseButton.setPreferredSize( new Dimension(30,10) );
    browseButton.addActionListener( new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        browseAction();
      }
    });

    // Eingabe-Feld formatieren
    SwingUtil.setPreferredWidth(inpComp,200);
    this.addInputOptionListener( new InputOptionAdapter() {
      public void optionLostFocus(InputOption option) {
        try {
          setValue( convertFromString( ((JTextField)inpComp).getText() ) );
        } catch (Exception err) {
          ExceptionDialog.show(OPTION_COMPONENT,err,SwingUtil.RESOURCE.getString("Error"),SwingUtil.RESOURCE.getString("InvalidInputMess"));
          setValue( getValue() );
        }
      }
    });

    // Neben dem Eingabefeld noch ein Browse-Button
    inpCompCont.setLayout( new BorderLayout(10,0) );
    inpCompCont.add(this.inpComp, BorderLayout.CENTER);
    inpCompCont.add(this.browseButton, BorderLayout.EAST);
    // normale Eingabe-Komponente durch Container ersetzen
    remove(inpComp);
    add(inpCompCont,BorderLayout.SOUTH);
  }

  /**
   * Aktion des Browse-Button.
   */
  private void browseAction() {
    Object newObj = performBrowse(this.getValue());
    // Browse-Aktion abgebrochen
    if ( newObj == null )
      return;
    this.setValue(newObj);
  }

  /**
   * Liefert das zu der Text-Eingabe gehoerende Objekt.
   */
  protected Object performGetValue() {
    return object;
  }

  /**
   * Setzt den Wert der Eingabe-Option und befuellt entsprechend das
   * Text-Feld der Eingabe-Option.
   * @param newValue neuer Wert der Eingabe-Option (ein Objekt, <b>nicht</b> ein
   *                 Wert fuer das Textfeld!)
   * @return <code>false</code> falls das Objekt nicht in einen String umgewandelt
   *         werden konnte
   */
  protected boolean performSetValue(Object newValue) {
    try {
      ((JTextField)inpComp).setText(convertToString(newValue));
      this.object = newValue;
      return true;
    } catch (Exception err) {
      return false;
    }
  }

  /**
   * Implementiert die Browse-Aktion die ausgefuehrt wird, wenn der Button
   * der Eingabe-Option gedrueckt wird. Liefert ein neues Objekt fuer die
   * Eingabe-Option.
   * @param actValue aktueller Wert der Eingabe-Option
   * @return <code>null</code> falls die Browse-Aktion abgebrochen wird
   */
  public abstract Object performBrowse(Object actValue);

  /**
   * Liefert das Objekt zu der Text-Eingabe der Option.
   * @param objectStr Objekt-String
   */
  public abstract Object convertFromString(String objectStr);

  /**
   * Erzeugt einen (eindeutigen) String fuer ein Objekt.
   * @param object Objekt
   */
  public abstract String convertToString(Object object);
}
