/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing;

import java.awt.Component;
import java.awt.Container;
import java.awt.FlowLayout;
import java.awt.LayoutManager;
import java.util.HashMap;
import java.util.Map;

/**
 * Diese Klasse stellt ein normales {@link JPanel} dar, welches
 * jedoch die {@link CaptionsChangeable}-Funktionaelitaet beim Aufruf von
 * {@link #resetCaptions(Map)} an alle im Panel enthaltenen
 * {@link CaptionsChangeable}-Komponenten weiterleitet.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class CaptionsChangeablePanel extends JPanel implements CaptionsChangeable {
  /**
   * Erzeugt ein neues Panel.
   * @param layout Layout fuer das Panel
   * @param isDoubleBuffered {@code true} fuer "double-buffering", welches zwar
   *        zusaetzlichen Speicher benoetigt, aber schnelle und flicker-freie
   *        Updates gewaehrleistet
   */
  public CaptionsChangeablePanel(LayoutManager layout, boolean isDoubleBuffered) {
    super(layout, isDoubleBuffered);
  }

  /**
   * Erzeugt ein gepuffertes Panel.
   * @param layout Layout fuer das Panel
   */
  public CaptionsChangeablePanel(LayoutManager layout) {
    super(layout);
  }

  /**
   * Erzeugt ein {@link FlowLayout}-Panel.
   * @param isDoubleBuffered {@code true} fuer "double-buffering", welches zwar
   *        zusaetzlichen Speicher benoetigt, aber schnelle und flicker-freie
   *        Updates gewaehrleistet
   */
  public CaptionsChangeablePanel(boolean isDoubleBuffered) {
    super(isDoubleBuffered);
  }

  /**
   * Erzeugt ein gepuffertes {@link FlowLayout}-Panel.
   */
  public CaptionsChangeablePanel() {
    super();
  }

  /**
   * Startet die Aktualisierung der Beschriftungen an alle im Panel
   * enthaltenen {@link CaptionsChangeable}-Komponenten mit einer leeren
   * {@link HashMap}.
   */
  public void resetCaptions() {
    resetCaptions(new HashMap<String,Object>());
  }

  /**
   * Leitet die Aktualisierung der Beschriftungen an alle im Panel
   * enthaltenen {@link CaptionsChangeable}-Komponenten weiter.
   * @param captionMap enthaelt die neuen Beschriftungen
   */
  public void resetCaptions(Map<String, Object> captionMap) {
    resetCaptions(this,captionMap);
  }

  /**
   * Leitet die Aktualisierung der Beschriftungen an alle im Container {@code container}
   * enthaltenen {@link CaptionsChangeable}-Komponenten weiter.
   * Zudem wird die Auswertung rekursiv fuer alle in {@code container} enthaltenen
   * {@link Container} fortgefuehrt.
   * @param container  ein Container der weitere {@link Component Components}
   *                   enthaelt
   * @param captionMap enthaelt die neuen Beschriftungen
   */
  public static void resetCaptions(Container container, Map<String, Object> captionMap) {
    Component[] innerComponents = container.getComponents();
    for (Component c : innerComponents) {
      if (c instanceof CaptionsChangeable)
        ( (CaptionsChangeable) c).resetCaptions(captionMap);
      if (c instanceof Container)
        resetCaptions((Container)c, captionMap);
    }
  }

}
