/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing;

import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import javax.swing.BorderFactory;
import javax.swing.JColorChooser;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JLabel;

/**
 * Diese Klasse stellt eine Farb-Eingabe-Option fuer das {@link MultipleOptionPane}
 * dar. Diese wird durch eine {@link JLabel} in der jeweiligen Farbe dargestellt.
 * Bei einem Klick auf das Label erscheint ein {@link JColorChooser}, ueber den
 * der Optionswert (Farbe) geaendert werden kann.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ColorInputOption extends InputOption {
  private   JColorChooser colorChooser;
  private   JDialog       dialog;

  /**
   * Erzeugt eine neue Eingabe-Option.
   * @param label     Beschreibung
   * @param inpNeeded gibt an, ob eine Leer-Eingabe erlaubt ist
   */
  public ColorInputOption(String label, boolean inpNeeded) {
    this(label,inpNeeded,null);
  }

  /**
   * Erzeugt eine neue Eingabe-Option.
   * @param label     Beschreibung
   * @param inpNeeded gibt an, ob eine Leer-Eingabe erlaubt ist
   * @param defColor  vorgeblendeter Optionswert
   */
  public ColorInputOption(String label, boolean inpNeeded, Color defColor) {
    super(label, inpNeeded);
    setValue( (defColor != null) ? defColor : new Color(0, 0, 0, 0));

    // Dialog zum Farbe auswaehlen
    colorChooser = new JColorChooser();
    dialog = JColorChooser.createDialog(
        this,
        "Farbe auswaehlen",
        true, //modal
        colorChooser,
        new ActionListener() { //OK button handler
             public void actionPerformed(ActionEvent e) {
                 // wird auf OK geklickt wird die ausgewaehlte Farbe
                 // uebernommen
                 setValue(colorChooser.getColor());
               }
             },
        null // CANCEL button handler
    );
  }

  /**
   * Erzeugt eine neue Instanz von {@link JLabel}. Dieses ist transparent und
   * schwart umrandet.<br>
   * Bei einem Klick auf das Label oeffnet sich ein {@link JColorChooser}
   * ueber den eine neue Farbe ausgewaehlt werden kann.
   */
  protected JComponent createInputComponent() {
//    // Eigentlich sollte ein JLabel verwendet werden. Dieses nimmt aber
//    // nicht automatisch die maximale Container-Breite des uebergeordneten
//    // Containers (bei SpringLayout) an. Deshalb wird ein TextField "vergewaltigt".
//    JTextField l = new JTextField("");
//    l.setEditable(false);
//    l.setBorder(null);
    JLabel l = new JLabel(" ");
    //////////////
    l.setOpaque(true);
    //l.setBorder(BorderFactory.createLineBorder(Color.BLACK));
    l.setBorder(BorderFactory.createRaisedBevelBorder());
    //l.setBorder(BorderFactory.createLoweredBevelBorder());

    // Beim Klick auf das Label erscheint der Auswahl-Dialog
    // fuer die Farben
    l.addMouseListener( new MouseAdapter() {
      public void mouseClicked(MouseEvent e) {
        if ( e.getButton() == MouseEvent.BUTTON1 ) {
          colorChooser.setColor( inpComp.getBackground() );
          dialog.setVisible(true);
        }
      }
    });
    return l;
  }

  /**
   * Liefert den Wert, der in der Option eingegeben wurde. Prueft zuerst
   * auf Gueltigkeit und ruft dann {@link #performGetValue()} auf.
   * @return <code>null</code> wenn die aktuelle Eingabe nicht zulaessig ist
   *         (siehe {@link #isInputValid()})
   */
  public Color getValue() {
    return (Color)super.getValue();
  }

  /**
   * Liefert immer <code>true</code>, da keine speziellen Anforderungen
   * an die Farb-Eingabe gestellt werden.
   */
  protected boolean performIsInputValid() {
    return true;
  }

  /**
   * Liefert immer <code>true</code>, wenn keine Farbe (<code>null</code>)
   * ausgewaehlt ist.
   */
  protected boolean performIsInputEmpty() {
    return inpComp.getBackground() == null;
  }

  /**
   * Liefert die aktuell in der Option ausgewaehlte Farbe.
   */
  protected Color performGetValue() {
    return inpComp.getBackground();
  }

  /**
   * Setzt den Wert der Farb-Option.
   * @param newValue neue Farbe (muss eine {@link Color}-Instanz sein)
   * @return <code>false</code> falls keine {@link Color} uebergeben wird.
   */
  protected boolean performSetValue(Object newValue) {
    if ( !(newValue instanceof Color) )
      return false;
    Color color = (Color)newValue;
    this.inpComp.setBackground(color);
    this.inpComp.setToolTipText("RGB value: " + color.getRed() + ", "
                                              + color.getGreen() + ", "
                                              + color.getBlue());
    return true;
  }
}
