/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Graphics2D;
import java.awt.event.MouseListener;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.plot.CompassPlot;
import org.jfree.chart.plot.PlotRenderingInfo;
import org.jfree.chart.plot.PlotState;
import org.jfree.data.general.DefaultValueDataset;
import org.jfree.data.general.ValueDataset;

/**
 * Diese Komponente stellt eine Kompass-Nadel dar, deren Ausrichtung ueber
 * eine Grad-Angabe zwischen 0\u00b0 und 360\u00b0 (Nord = 0\u00b0; West = 90\u00b0) eingestellt wird.
 * {@link MouseListener} oder aehnliches koennen aus technischen Gruenden
 * nicht der Komponente direkt zugewiesen werden, sondern muessen dem
 * {@linkplain #getCompassPane() Content-Pane} zugewiesen werden.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public  class Compass extends JPanel {
  /** Ereignis, welches an die Listener gesendet wird, wenn sich der Kompass-Wert
   *  aendert. */
  protected ChangeEvent changeEvent = new ChangeEvent(this);

  private ChartPanel chartPanel;
  private double degree = 0;

  /**
   * Erzeugt eine neue Kompass-Komponente.
   */
  public Compass() {
    this(0);
  }

  /**
   * Erzeugt eine neue Kompass-Komponente.
   * @param degree angezeigte Grad-Angabe (Nord = 0\u00b0, West = 90\u00b0)
   */
  public Compass(double degree) {
    super();
    setLayout( new BorderLayout() );
    chartPanel = new ChartPanel(
      null,      // Chart
      100,       // Width
      100,       // Height
      0,         // Min. draw width
      0,         // Min. draw height
      9999999,   // Max. draw width
      9999999,   // Max. draw height
      true,      // Use buffer
      true,     // Properties
      false,     // Save
      false,     // Print
      false,     // Zoom
      false      // Tooltips
    );
    setValue(degree);
    this.add(chartPanel,BorderLayout.CENTER);
  }

  /**
   * Setzt die Hintergrund-Farbe (auch fuer alle Unterkomponenten).
   * @param bgColor Hintergrund-Farbe
   */
  public void setBackground(Color bgColor) {
    setBackground(bgColor, true);
    if ( chartPanel != null && chartPanel.getChart() != null )
      chartPanel.getChart().setBackgroundPaint( bgColor );
  }

  /**
   * Liefert die Komponente des Kompass fuer {@link MouseListener} oder aehnliches.
   */
  public Component getCompassPane() {
    return chartPanel;
  }

  /**
   * Setzt die angezeigte Grad-Angabe und erzeugt ein {@link ChangeEvent} fuer
   * alle angeschlossenen {@link ChangeListener}.
   * @param degree angezeigte Grad-Angabe (Nord = 0\u00b0, West = 90\u00b0)
   */
  public void setValue(double degree) {
    double oldDegree = this.degree;
    this.degree = degree % 360; // auf ]-360,+360[ normieren
    this.degree = (this.degree+360) % 360; // auf [0,360[ normierern
    chartPanel.setChart( createCompassChart( this.degree ) );
    if (this.degree != oldDegree)
      fireStateChanged();
  }

  /**
   * Liefert die angezeigte Grad-Angabe aus [0\u00b0,360\u00b0[.
   */
  public double getValue() {
    return degree;
  }

  /**
   * Fuegt dem Kompass einen {@link ChangeListener} hinzu. Dieser wird jedesmal
   * informtiert, wenn sich der Wert des Kompass aendert.
   * @param listener neuer {@link ChangeListener}
   */
  public void addChangeListener(ChangeListener listener) {
    this.listenerList.add( ChangeListener.class, listener );
  }

  /**
   * Entfernt einen {@link ChangeListener} vom Kompass.
   * @param listener zu entfernender {@link ChangeListener}
   */
  public void removeChangeListener(ChangeListener listener) {
    this.listenerList.remove( ChangeListener.class, listener );
  }

  /**
   * Liefert die Anzahl der an den Kompass angeschlossenen {@link ChangeListener}.
   */
  public int getListenerCount() {
    return listenerList.getListenerCount(ChangeListener.class);
  }

  /**
   * Liefert an den Kompass angeschlossenen {@link ChangeListener}.
   */
  public ChangeListener[] getChangeListeners() {
    return listenerList.getListeners(ChangeListener.class);
  }

  /**
   * Leitet ein {@link ChangeEvent} an alle angeschlossenen {@link ChangeListener}.
   * @see #addChangeListener(ChangeListener)
   * @see #removeChangeListener(ChangeListener)
   */
  protected void fireStateChanged() {
    ChangeListener[] listeners = getChangeListeners();
    for (ChangeListener l : listeners)
      l.stateChanged(changeEvent);
  }

//  public void setPreferredSize(Dimension d) {
//    int min = Math.min(d.height,d.width);
//    chartPanel.setPreferredSize( new Dimension(min,min) );
//    super.setPreferredSize(d);
//  }
//
//  public void setSize(Dimension d) {
//    int min = Math.min(d.height,d.width);
//    chartPanel.setPreferredSize( new Dimension(min,min) );
//    super.setSize(d);
//  }
//
//  public void setSize(int w, int h) {
//    int min = Math.min(w,h);
//    chartPanel.setSize( new Dimension(min,min) );
//    super.setSize(w,h);
//  }

  /**
   * Erzeugt ein Chart in Form einer Kompass-Nadel.
   * @param degree angezeigte Grad-Angabe (Nord = 0\u00b0, West = 90\u00b0)
   */
  private JFreeChart createCompassChart(double degree) {
    // Bei der Chart-Komponente ist Ost = 90\u00b0
    ValueDataset dataset = new DefaultValueDataset(-degree);
    CompassPlot plot = new CompassPlot(dataset) {
      public void draw(Graphics2D g, Rectangle2D area, Point2D anchor, PlotState parentState, PlotRenderingInfo info) {
//        System.out.println( area + "     " + g.getClipBounds() + "   " + g.getTransform().getScaleX() + "  " + g.getTransform().getScaleY());
//        double minSize  = Math.min(area.getWidth(), area.getHeight());
//        double minScale = Math.min(g.getTransform().getScaleX(), g.getTransform().getScaleY() );
//        double x        = (chartPanel.getWidth()-minSize)/2;
//        double y        = (chartPanel.getHeight()-minSize)/2;
//        area.setFrame( x, y, minSize, minSize);
//        g.setTransform( AffineTransform.getScaleInstance(minScale,minScale) );
//        System.out.println( area + "     " + g.getClipBounds() + "   " + g.getTransform().getScaleX() + "  " + g.getTransform().getScaleY());
//        System.out.println();
        super.draw(g,area,anchor,parentState,info);
      }
    };
    plot.setSeriesNeedle(4);
    plot.setSeriesPaint(0, Color.black);
    plot.setSeriesOutlinePaint(0, Color.black);
    plot.setRosePaint(Color.red);
    plot.setRoseHighlightPaint(Color.gray);
    plot.setRoseCenterPaint(Color.white);
    plot.setDrawBorder(false);
    JFreeChart chart = new JFreeChart(plot);
    chart.setBackgroundPaint( getBackground() );
    return chart;
  }
}
