/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Frame;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JToggleButton;

import schmitzm.lang.LangUtil;

/**
 * Diese Klasse stellt eine modale Fehler-Meldung dar. Diese besteht neben
 * einer Meldung aus einem Text-Feld, in dem der StackTrace des Fehlers
 * angezeigt wird.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ExceptionDialog extends JDialog {
  /** Speichert den angezeigten Fehler. */
  protected Throwable err = null;
  /** Label in dem die Meldung angezeigt wird. */
  protected JLabel  messageLabel = null;
  /** Button um den Dialog zu schliessen. */
  protected JButton okButton = null;
  /** Button um Details anzuzeigen. */
  protected JToggleButton detailsButton = null;
  /** Button um die Exception in die Zwischenablage zu kopieren. */
  protected JButton copyToClipboardButton = null;
  /** Button um die Exception auf die Console zu kopieren. */
  protected JButton copyToConsoleButton = null;
  /** Bereich, in dem die Details angezeigt werden. */
  protected JTextArea detailsTextArea = null;
  /** ScrollPane, in dem sich die TextArea fuer die Details befinden. */
  protected JScrollPane detailsScrollPane = null;
  /** Verbindung zwischen StackTrace-PrintStream und TextArea*/
  private TextAreaPrintStream detailsPrintStream= null;

  /**
   * Erzeugt einen neuen Fehler-Dialog. Dem Dialog wird zunaechst noch keine
   * Fehlermeldung zugeordnet.
   * @param parent   uebergeordnetes Fenster (kann <code>null</code> sein!)
   * @see #setError(Throwable)
   */
  public ExceptionDialog(Component parent) {
    this(parent,null,null,null);
  }

  /**
   * Erzeugt einen neuen Fehler-Dialog. Der Dialog wird relativ zum Parent-Fenster
   * zentriert.
   * @param parent   uebergeordnetes Fenster (kann <code>null</code> sein!)
   * @param err      darzustellender Fehler
   */
  public ExceptionDialog(Component parent, Throwable err) {
    this(parent,err,null,null);
  }

  /**
   * Erzeugt einen neuen Fehler-Dialog. Der Dialog wird relativ zum Parent-Fenster
   * zentriert.
   * @param parent     uebergeordnetes Fenster (kann <code>null</code> sein!)
   * @param err        darzustellender Fehler
   * @param title      Titel fuer das Fenster (kann <code>null</code> sein!)
   * @param errMessage Meldung, die zu dem Fehler angezeigt angezeigt wird
   *                   (kann <code>null</code> sein!)
   */
  public ExceptionDialog(Component parent, Throwable err, String title, String errMessage) {
    this(parent,err,title,errMessage,false);
  }
  
  /**
   * Erzeugt einen neuen Fehler-Dialog. Der Dialog wird relativ zum Parent-Fenster
   * zentriert.
   * @param parent     uebergeordnetes Fenster (kann <code>null</code> sein!)
   * @param err        darzustellender Fehler
   * @param title      Titel fuer das Fenster (kann <code>null</code> sein!)
   * @param errMessage Meldung, die zu dem Fehler angezeigt angezeigt wird
   *                   (kann <code>null</code> sein!)
   * @param showDetails wenn {@code true} werden die Details initial angezeigt
   */
  public ExceptionDialog(Component parent, Throwable err, String title, String errMessage, boolean showDetails) {
    super((Frame)null,true);
    if ( err != null && (title==null || title.trim().equals("")) )
      title = err.getClass().getSimpleName();
    if ( err != null && (errMessage==null || errMessage.trim().equals("")) )
      errMessage = err.getMessage();

    // Vorlagen-Dialog erzeugen
    this.err                   = err;
    this.messageLabel          = new JLabel(errMessage);
    this.okButton              = new JButton( SwingUtil.RESOURCE.getString("Ok") );
    this.copyToConsoleButton   = new JButton( SwingUtil.RESOURCE.getString("ExceptionDialog.CopyToConsole") );
    this.copyToClipboardButton = new JButton( SwingUtil.RESOURCE.getString("ExceptionDialog.CopyToClipboard") );
    this.detailsButton         = new JToggleButton( SwingUtil.RESOURCE.getString("Details"), showDetails );
    this.detailsTextArea       = new JTextArea(10,60);
    this.detailsTextArea.setEditable(false);
    this.detailsPrintStream    = new TextAreaPrintStream(detailsTextArea);
    this.detailsScrollPane     = new JScrollPane(detailsTextArea);
    this.detailsScrollPane.setVisible(showDetails);
    this.setError(err);
    JOptionPane  pane = new JOptionPane(
      messageLabel,
      JOptionPane.ERROR_MESSAGE,
      JOptionPane.DEFAULT_OPTION,
      null,
      new Object[] {okButton, detailsButton, copyToConsoleButton, copyToClipboardButton}
    );
    JDialog dialog = pane.createDialog(parent,title);

    // Dialog nach Vorlage initialisieren
    this.setTitle( dialog.getTitle() );
    this.setAlwaysOnTop(true); // Fehler soll auch oberhalb von "alwaysOnTop"-
                               // Komponenten erscheinen
    this.getContentPane().setLayout(new BorderLayout());
    this.getContentPane().add(dialog.getContentPane(), BorderLayout.NORTH);
    this.getContentPane().add(detailsScrollPane, BorderLayout.CENTER);
//    this.setDefaultCloseOperation( DO_NOTHING_ON_CLOSE );
    this.okButton.addActionListener( new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        setVisible(false);
      }
    } );
    this.detailsButton.addActionListener( new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        detailsScrollPane.setVisible( detailsButton.isSelected() );
        pack();
      }
    } );
    this.copyToClipboardButton.addActionListener( new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        LangUtil.copyToClipboard( detailsTextArea.getText() );
      }
    } );
    this.copyToConsoleButton.addActionListener( new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        if ( getError() != null )
          getError().printStackTrace();
      }
    } );

    pack();
    SwingUtil.setRelativeFramePosition(this,SwingUtil.getParentWindow(parent),0.5,0.5);
  }

  /**
   * Liefert die angezeigte Fehlermeldung.
   */
  public String getMessage() {
    return  messageLabel.getText();
  }

  /**
   * Setzt die anzuzeigende Fehlermeldung.
   */
  public void setMessage(String mess) {
    if ( mess == null || mess.trim().equals("") )
      mess = getError() != null ? getError().getMessage() : SwingUtil.RESOURCE.getString("Error");
    messageLabel.setText(mess);
  }

  /**
   * Liefert den angezeigten Fehler.
   */
  public Throwable getError() {
    return err;
  }

  /**
   * Setzt den angezeigten Fehler.
   */
  public void setError(Throwable err) {
    this.err = err;
    if ( err == null )
      detailsTextArea.setText("");
    else
      err.printStackTrace(detailsPrintStream);
    detailsTextArea.select(0,0);
  }

  /**
   * Zeigt einen Fehler-Dialog an.
   * @param parent      uebergeordnetes Fenster (kann <code>null</code> sein!)
   * @param err         darzustellender Fehler
   * @param title       Titel fuer das Fenster (kann <code>null</code> sein!)
   * @param errMessage  Meldung, die zu dem Fehler angezeigt angezeigt wird
   *                    (kann <code>null</code> sein!)
   */
  public static void show(Component parent, Throwable err, String title, String errMessage) {
    new ExceptionDialog(parent,err,title,errMessage).setVisible(true);
  }

  /**
   * Zeigt einen Fehler-Dialog an.
   * @param parent   uebergeordnetes Fenster (kann <code>null</code> sein!)
   * @param err      darzustellender Fehler
   */
  public static void show(Component parent, Throwable err) {
    new ExceptionDialog(parent,err).setVisible(true);
  }

  /**
   * Zeigt einen Fehler-Dialog an.
   * @param err darzustellender Fehler
   */
  public static void show(Throwable err) {
    new ExceptionDialog(null,err).setVisible(true);
  }

  /**
   * Zeigt einen Fehler-Dialog an.
   * @param parent      uebergeordnetes Fenster (kann <code>null</code> sein!)
   * @param err         darzustellender Fehler
   * @param title       Titel fuer das Fenster (kann <code>null</code> sein!)
   * @param errMessage  Meldung, die zu dem Fehler angezeigt angezeigt wird
   *                    (kann <code>null</code> sein!)
   * @param showDetails wenn {@code true} werden die Details initial angezeigt
   */
  public static void show(Component parent, Throwable err, String title, String errMessage, boolean showDetails) {
    new ExceptionDialog(parent,err,title,errMessage,showDetails).setVisible(true);
  }

  /**
   * Zeigt einen Fehler-Dialog an.
   * @param parent   uebergeordnetes Fenster (kann <code>null</code> sein!)
   * @param err      darzustellender Fehler
   * @param showDetails wenn {@code true} werden die Details initial angezeigt
   */
  public static void show(Component parent, Throwable err, boolean showDetails) {
    show(parent,err,null,null,showDetails);
  }

  /**
   * Zeigt einen Fehler-Dialog an.
   * @param err darzustellender Fehler
   * @param showDetails wenn {@code true} werden die Details initial angezeigt
   */
  public static void show(Throwable err, boolean showDetails) {
    show(null,err,showDetails);
  }
}
