/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing;

import java.awt.Color;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.geom.AffineTransform;
import java.text.DecimalFormat;
import java.text.NumberFormat;

import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JProgressBar;

/**
 * Ein ExpansionBar aehnelt einem {@link JProgressBar}, mit dem Unterschied,
 * dass die Auspraegung des ExpansionBar nicht nur in eine Richtung gehen
 * kann (unten nach oben, oder links nach rechts), sondern ausgehend von einem
 * vorgegebenen Mitte-Wert in beide Richtungen, je nachdem, ob der gesetzte
 * Wert groesser oder kleiner ist als der Mitte-Wert.<br>
 * Waehrend der {@link JProgressBar} als Status-Balken aufgefasst werden kann,
 * der stetig fortschreitet, dient der der ExpansionBar auch der Anzeige von
 * negativen Werten (z.B. Abweichungswerten oder Amplituden). Fuer die
 * Abweichung koennen je ein oberer und unterer Grenzwert festgelegt werden.
 * Liegt der aktuelle Wert des Balken innerhalb dieser Toleranz, wird er
 * gruen angezeigt, ansonsten rot.<br>
 * Optional koennen fuer den Balken Labels mit den Grenz-Werten (Minimum,
 * Mitte, Maximum) angezeigt werden (siehe {@link #setValueLabelsPainted(boolean)}.
 * Dies ist per Default jedoch deaktiviert.<br>
 * <b>Bemerke:</b><br>
 * Bei vertikaler Darstellung gibt es noch Probleme mit der Label-Darstellung!
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 * @todo Probleme mit vertikaler Label-Darstellung beheben
 */
public class ExpansionBar extends Container {
  /** Konstante fuer eine lineare Darstellung des Balken. */
  public static final int LINEAR = 9;
  /** Konstante fuer eine logarithmische Darstellung des Balken. */
  public static final int LOGARITHMIC = 8;
  /** Konstante fuer eine vertikale Ausrichtung des Balken.
   *  @see JProgressBar#VERTICAL */
  public static final int VERTICAL = JProgressBar.VERTICAL;
  /** Konstante fuer eine horizontale Ausrichtung des Balken.
   *  @see JProgressBar#HORIZONTAL */
  public static final int HORIZONTAL = JProgressBar.HORIZONTAL;

  /** Speichert den Umrechnungsfaktor, um von den gesetzten double-Werten
   *  auf Integer fuer die {@link JProgressBar}-Instanzen umzurechnen.
   *  @see #convertToProgressBarValue(double)
   */
  protected int fracFact = 0;

  /** Speichert die Farbe, in der der Balken angezeigt wird, wenn der
   *  Wert innerhalb der Toleranz liegt. <br>
   *  Default: {@linkplain Color#RED Gruen}*/
  protected Color toleranceColor = Color.GREEN;

  /** Speichert die Farbe, in der der Balken angezeigt wird, wenn der
   *  Wert ausserhalb der Toleranz liegt.<br>
   *  Default: {@linkplain Color#RED Rot} */
  protected Color intoleranceColor = Color.RED;

  // Beispiel-Schriftart fuer die Labels
  private Font sampleFont = new JLabel().getFont();

  // Label fuer Wert-Anzeige
  private JLabel       minLabel = new JLabel();
  private JLabel       midLabel = new JLabel();
  private JLabel       maxLabel = new JLabel();
  // Progressbars fuer die positive und negative Richtung
  private JProgressBar negBar;
  private JProgressBar posBar;
  private JPanel       barContainer = new JPanel();

  // Konfigurationsparameter
  private double       min           = -100;
  private int          barMin        = -100;
  private double       max           = 100;
  private int          barMax        = 100;
  private double       mid           = 0;
  private int          barMid        = 0;
  private double       minTol        = 0;
  private double       maxTol        = 0;
  private int          orient        = HORIZONTAL;
  private int          type          = LINEAR;
  private double       value         = 0;
  private boolean      stringPainted = true;
  private NumberFormat barNumFormat  = new DecimalFormat("0.00%");
  private NumberFormat infoNumFormat = new DecimalFormat("0.00%");

  /**
   * Erzeugt einen neuen linearen Balken. Als Mitte-Wert wird 0 gesetzt und die
   * Toleranz ist nach oben und unten gleich.<br>
   * Die Minimum-, Mitte- und Maximum-Label werden
   * standardmaessig nicht angezeigt, wohl aber der aktuelle Balkenwert.
   * @param orient    Orientierung des Balken (<code>ExpansionBar.HORIZONTAL</code>
   *                  oder <code>ExpansionBar.VERTICAL</code>
   * @param min       Minimalwert des Balkens
   * @param max       Maximalwert des Balkens
   * @param tol       Toleranzwert nach oben/unten bis zu dem der Balken gruen
   *                  dargestellt wird (muss zwischen <code>0</code> und
   *                  <code>max</code> liegen)
   * @param form      Bestimmt die Darstellung der Werte im Anzeige-Label des
   *                  Balken und in der etwaigen Info-Leiste
   * @see #setValueLabelsPainted(boolean)
   * @see #setStringPainted(boolean)
   * @exception IllegalArgumentException falls die Konstellation der angegeben
   *            Werte unzulaessig ist
   */
  public ExpansionBar(int orient, double min, double max, double tol, NumberFormat form) {
    this(orient,LINEAR,min,max,tol,form);
  }

  /**
   * Erzeugt einen neuen Balken. Als Mitte-Wert wird 0 gesetzt und die
   * Toleranz ist nach oben und unten gleich.<br>
   * Die Minimum-, Mitte- und Maximum-Label werden
   * standardmaessig nicht angezeigt, wohl aber der aktuelle Balkenwert.
   * @param orient    Orientierung des Balken (<code>ExpansionBar.HORIZONTAL</code>
   *                  oder <code>ExpansionBar.VERTICAL</code>
   * @param type      Darstellung des Balken (<code>ExpansionBar.LINEAR</code>
   *                  oder <code>ExpansionBar.LOGARITHMIC</code>
   * @param min       Minimalwert des Balkens
   * @param max       Maximalwert des Balkens
   * @param tol       Toleranzwert nach oben/unten bis zu dem der Balken gruen
   *                  dargestellt wird (muss zwischen <code>0</code> und
   *                  <code>max</code> liegen)
   * @param form      Bestimmt die Darstellung der Werte im Anzeige-Label des
   *                  Balken und in der etwaigen Info-Leiste
   * @see #setValueLabelsPainted(boolean)
   * @see #setStringPainted(boolean)
   * @exception IllegalArgumentException falls die Konstellation der angegeben
   *            Werte unzulaessig ist
   */
  public ExpansionBar(int orient, int type, double min, double max, double tol, NumberFormat form) {
    this(orient,type,min,max,0.0,-tol,tol,0.0,form,form);
  }

  /**
   * Erzeugt einen neuen linearen Balken. Die Toleranz vom Mitte-Wert ist nach oben
   * und unten gleich.<br>
   * Die Minimum-, Mitte- und Maximum-Label werden
   * standardmaessig nicht angezeigt, wohl aber der aktuelle Balkenwert.
   * @param orient    Orientierung des Balken (<code>ExpansionBar.HORIZONTAL</code>
   *                  oder <code>ExpansionBar.VERTICAL</code>
   * @param min       Minimalwert des Balkens
   * @param max       Maximalwert des Balkens
   * @param mid       Mitte-Wert des Balkens (muss zwischen <code>min</code> und
   *                  <code>max</code> liegen)
   * @param tol       Toleranzwert (relativ zur Mitte!) bis zu dem der Balken gruen
   *                  dargestellt wird (muss zwischen <code>0</code> und
   *                  <code>max</code> liegen)
   * @param form      Bestimmt die Darstellung der Werte im Anzeige-Label des
   *                  Balken und in der etwaigen Info-Leiste
   * @see #setValueLabelsPainted(boolean)
   * @see #setStringPainted(boolean)
   * @exception IllegalArgumentException falls die Konstellation der angegeben
   *            Werte unzulaessig ist
   */
  public ExpansionBar(int orient, double min, double max, double mid, double tol, NumberFormat form) {
    this(orient,LINEAR,min,max,mid,tol,form);
  }

  /**
   * Erzeugt einen neuen Balken. Die Toleranz vom Mitte-Wert ist nach oben
   * und unten gleich.<br>
   * Die Minimum-, Mitte- und Maximum-Label werden
   * standardmaessig nicht angezeigt, wohl aber der aktuelle Balkenwert.
   * @param orient    Orientierung des Balken (<code>ExpansionBar.HORIZONTAL</code>
   *                  oder <code>ExpansionBar.VERTICAL</code>
   * @param type      Darstellung des Balken (<code>ExpansionBar.LINEAR</code>
   *                  oder <code>ExpansionBar.LOGARITHMIC</code>
   * @param min       Minimalwert des Balkens
   * @param max       Maximalwert des Balkens
   * @param mid       Mitte-Wert des Balkens (muss zwischen <code>min</code> und
   *                  <code>max</code> liegen)
   * @param tol       Toleranzwert (relativ zur Mitte!) bis zu dem der Balken gruen
   *                  dargestellt wird (muss zwischen <code>0</code> und
   *                  <code>max</code> liegen)
   * @param form      Bestimmt die Darstellung der Werte im Anzeige-Label des
   *                  Balken und in der etwaigen Info-Leiste
   * @see #setValueLabelsPainted(boolean)
   * @see #setStringPainted(boolean)
   * @exception IllegalArgumentException falls die Konstellation der angegeben
   *            Werte unzulaessig ist
   */
  public ExpansionBar(int orient, int type, double min, double max, double mid, double tol, NumberFormat form) {
    this(orient,type,min,max,mid,mid-tol,mid+tol,mid,form,form);
  }

  /**
   * Erzeugt einen neuen Balken. Die Minimum-, Mitte- und Maximum-Label werden
   * standardmaessig nicht angezeigt, wohl aber der aktuelle Balkenwert.
   * @param orient    Orientierung des Balken (<code>ExpansionBar.HORIZONTAL</code>
   *                  oder <code>ExpansionBar.VERTICAL</code>
   * @param type      Darstellung des Balken (<code>ExpansionBar.LINEAR</code>
   *                  oder <code>ExpansionBar.LOGARITHMIC</code>
   * @param min       Minimalwert des Balkens
   * @param max       Maximalwert des Balkens
   * @param mid       Mitte-Wert des Balkens (muss zwischen <code>min</code> und
   *                  <code>max</code> liegen)
   * @param minTol    <b>Absoluter</b> Toleranzwert nach unten bis zu dem der Balken gruen
   *                  dargestellt wird (muss zwischen <code>min</code> und
   *                  <code>mid</code> liegen)
   * @param maxTol    <b>Absoluter</b> Toleranzwert nach oben bis zu dem der Balken gruen
   *                  dargestellt wird (muss zwischen <code>mid</code> und
   *                  <code>max</code> liegen)
   * @param initValue Initialer Wert den der Balken annimmt
   * @param barForm   Bestimmt die Darstellung der Werte im Anzeige-Label des
   *                  Balken
   * @param infoForm  Bestimmt die Darstellung der Werte im etwaigen Info-Label
   *                  unterhalb des Balken
   * @see #setValueLabelsPainted(boolean)
   * @see #setStringPainted(boolean)
   * @exception IllegalArgumentException falls die Konstellation der angegeben
   *            Werte unzulaessig ist
   */
  public ExpansionBar(int orient, int type, double min, double max, double mid, double minTol, double maxTol, double initValue, NumberFormat barForm, NumberFormat infoForm) {
    super();
    this.min       = min;
    this.max       = max;
    this.mid       = mid;
    this.minTol    = minTol;
    this.maxTol    = maxTol;
    this.orient    = orient;
    this.type      = type;
    if ( barForm != null )
      this.barNumFormat = barForm;
    if ( infoForm != null )
      this.infoNumFormat = infoForm;
    checkValues();

    // Der Progress-Bar kann nur Integers darstellen.
    // Damit auch kleinere Bereiche (z.B. zwischen 0 und 1) fliessend
    // dargestellt werden koennen, wird der Bereich auf (mind. 1000 Werte)
    // skaliert
    this.fracFact = 1;
    while(  (max-mid)*fracFact < 1000 || (mid-min)*fracFact < 1000 )
      fracFact *= 10;
    this.barMin   = convertToProgressBarValue(min);
    this.barMax   = convertToProgressBarValue(max);
    this.barMid   = convertToProgressBarValue(mid);

    // Label erzeugen
    if (orient == VERTICAL)
      // Label bekommen vertikale Schrift
      sampleFont = sampleFont.deriveFont(AffineTransform.getRotateInstance(Math.PI / 2));
    this.minLabel = new JLabel(infoNumFormat.format(min));
    this.midLabel = new JLabel(infoNumFormat.format(mid));
    this.maxLabel = new JLabel(infoNumFormat.format(max));
    minLabel.setFont(sampleFont);
    midLabel.setFont(sampleFont);
    maxLabel.setFont(sampleFont);
    setValueLabelsPainted(false); // Wert-Labels werden per Default NICHT angezeigt

    // Das Ausgeben des Wert-Strings bewirkt, dass die Balken nicht in
    // kleinen Bloecken dargestellt werden, sondern durchgehend (letztes
    // ist notwendig, damit das "Tricksen" mit den Vorder/Hintergrundfarben
    // beim Negatv-Balken den gewuenschten visuellen Effekt hat!)
    // --> Es soll aber trotzdem keine (oder hoechstens beim Positiv-Balken)
    //     Strings angezeigt werden
    // --> getString()-Methode ueberschreiben und NICHTS zurueckgeben
    this.posBar = new JProgressBar(orient,barMid,barMax) {
      public String getString() { return stringPainted ? barNumFormat.format(value) : ""; }
    };
    this.negBar = new JProgressBar(orient,barMin,barMid) {
      public String getString() { return ""; } };
     // WICHTIG wegen Farbeffekt!!
    posBar.setStringPainted(true);
    negBar.setStringPainted(true);
    posBar.setMinimumSize( new Dimension(0,0) );
    negBar.setMinimumSize( new Dimension(0,0) );
    // Negativ-Balken wird invertiert dargestellt, so dass es den Anschein
    // bekommt, er wuerde sich von rechts nach links (oben nach unten)
    // bewegen!!!
    negBar.setForeground( posBar.getBackground() );

    // ProgressBars erhalten keine eigene Border, sondern werden in einen
    // Container eingefuegt, der eine Border erhaelt (WCIHTIG, da sonst die
    // Border des Nagativ-Balken in der "Vordergrundfarbe" angezeigt wird und
    // immer sichtbar ist)
    posBar.setBorder( BorderFactory.createEmptyBorder() );
    negBar.setBorder( BorderFactory.createEmptyBorder() );
    barContainer.setBorder( BorderFactory.createLoweredBevelBorder() );
    barContainer.setLayout( new GridBagLayout() );
    // Zwischen die beiden Progress-Bars kommt ein Trennstrich
    JLabel divComp = new JLabel();
    divComp.setBorder( BorderFactory.createLineBorder(Color.DARK_GRAY));

    // Komponenten in Container einfuegen
    this.setLayout( new GridBagLayout() );
    if ( orient == HORIZONTAL ) {
      barContainer.add(negBar,new GridBagConstraints(0,0,1,1,0.5,1.0,GridBagConstraints.CENTER,GridBagConstraints.BOTH,new Insets(0,0,0,0),0,0));
      barContainer.add(divComp,new GridBagConstraints(1,0,1,1,0.0,1.0,GridBagConstraints.CENTER,GridBagConstraints.BOTH,new Insets(0,0,0,0),0,0));
      barContainer.add(posBar,new GridBagConstraints(2,0,1,1,0.5,1.0,GridBagConstraints.CENTER,GridBagConstraints.BOTH,new Insets(0,0,0,0),0,0));
      add(barContainer, new GridBagConstraints(0,0,1,1,1.0,1.0,GridBagConstraints.CENTER,GridBagConstraints.BOTH,new Insets(0,0,0,0),0,0));
      add(minLabel,new GridBagConstraints(0,1,1,1,1.0,1.0,GridBagConstraints.NORTHWEST,GridBagConstraints.NONE,new Insets(0,0,0,0),0,0));
      add(midLabel,new GridBagConstraints(0,1,1,1,1.0,1.0,GridBagConstraints.NORTH,GridBagConstraints.NONE,new Insets(0,0,0,0),0,0));
      add(maxLabel,new GridBagConstraints(0,1,1,1,1.0,1.0,GridBagConstraints.NORTHEAST,GridBagConstraints.NONE,new Insets(0,0,0,0),0,0));
    } else {
      barContainer.add(posBar,new GridBagConstraints(0,0,1,1,1.0,0.5,GridBagConstraints.CENTER,GridBagConstraints.BOTH,new Insets(0,0,0,0),0,0));
      barContainer.add(divComp,new GridBagConstraints(0,1,1,1,1.0,0.0,GridBagConstraints.CENTER,GridBagConstraints.BOTH,new Insets(0,0,0,0),0,0));
      barContainer.add(negBar,new GridBagConstraints(0,2,1,1,1.0,0.5,GridBagConstraints.CENTER,GridBagConstraints.BOTH,new Insets(0,0,0,0),0,0));
      add(barContainer, new GridBagConstraints(0,0,1,1,1.0,1.0,GridBagConstraints.CENTER,GridBagConstraints.BOTH,new Insets(0,0,0,0),0,0));
      add(minLabel,new GridBagConstraints(1,0,1,1,1.0,1.0,GridBagConstraints.NORTH,GridBagConstraints.NONE,new Insets(0,0,0,0),0,0));
      add(midLabel,new GridBagConstraints(1,0,1,1,1.0,1.0,GridBagConstraints.CENTER,GridBagConstraints.NONE,new Insets(0,0,0,0),0,0));
      add(maxLabel,new GridBagConstraints(1,0,1,1,1.0,1.0,GridBagConstraints.SOUTH,GridBagConstraints.NONE,new Insets(0,0,0,0),0,0));
    }
    // Default-Wert setzen
    setValue(initValue);
  }

  /**
   * Der ExpansionBar ist ueber zwei {@link JProgressBar}s realisiert. Diese
   * koennen nur Integers darstellen. Damit auch kleinere Bereiche (z.B.
   * zwischen 0 und 1) fliessend dargestellt werden koennen, wird der Bereich
   * auf (mind. 1000 Werte) skaliert. Desweiteren wird ggf. auf eine
   * logarithmische Darstellung umgerechnet.
   * @param val umzurechnender Wert
   * @see #fracFact
   */
  protected int convertToProgressBarValue(double val) {
    if ( type == LOGARITHMIC ) {
      if ( val >=0 )
        val = Math.log(1 + val);
      else
        val = -Math.log(1 + -val);
    }
    return (int)Math.round(val*fracFact);
  }

  /**
   * Prueft die durch den Konstruktor gesetzten Initialisierungswerte auf
   * korrektheit.
   * <ol>
   * <li><code>min <= max</code></li>
   * <li><code>min <= mid <= max</code></li>
   * <li><code>min <= minTol <= mid</code></li>
   * <li><code>mid <= maxTol <= max</code></li>
   * <li><code>orient == ExpansionBar.HORIZONTAL || orient == ExpansionBar.VERTICAL</code></li>
   * </ol>
   */
  protected void checkValues() {
    if ( min > max )
      throw new IllegalArgumentException("Minimum value must be less than maximum!");
    if ( mid < min || mid > max )
      throw new IllegalArgumentException("Middle value must be between minimum and maximum!");
    if ( minTol < min || minTol > mid )
      throw new IllegalArgumentException("Minimum tolerance must be between minimum and middle!");
    if ( maxTol < mid || maxTol > max )
      throw new IllegalArgumentException("Maximum tolerance must be between middle and maximum!");
    if ( orient != HORIZONTAL && orient != VERTICAL )
      throw new IllegalArgumentException("Orientation value must be ExpansionBar.HORIZONTAL or ExpansionBar.VERTICAL!");
  }

  /**
   * Liefert den Minimum-Wert des Balkens.
   */
  public double getMinimum() {
    return min;
  }

  /**
   * Liefert den Maximum-Wert des Balkens.
   */
  public double getMaximum() {
    return max;
  }

  /**
   * Liefert den Mitte-Wert des Balkens.
   */
  public double getMiddle() {
    return mid;
  }

  /**
   * Liefert den (absoluten) Wert nach unten, bis zu dessen Wert der
   * Balkens in gruen dargstellt wird.
   */
  public double getMinimumTolerance() {
    return minTol;
  }

  /**
   * Setzt den (absoluten) Wert nach unten, bis zu dessen Wert der
   * Balkens in gruen dargstellt wird.
   */
  public void setMinimumTolerance(double minTol) {
    if ( minTol < min || minTol > mid )
      throw new IllegalArgumentException("Minimum tolerance must be between minimum and middle!");
    this.minTol = minTol;;
    updateForeground();
  }

  /**
   * Liefert den (absoluten) Wert nach oben, bis zu dessen Wert der
   * Balkens in gruen dargstellt wird.
   */
  public double getMaximumTolerance() {
    return maxTol;
  }

  /**
   * Setzt den (absoluten) Wert nach oben, bis zu dessen Wert der
   * Balkens in gruen dargstellt wird.
   */
  public void setMaximumTolerance(double maxTol) {
    if ( maxTol < mid || maxTol > max )
      throw new IllegalArgumentException("Maximum tolerance must be between middle and maximum!");
    this.maxTol = maxTol;
    updateForeground();
  }

  /**
   * Liefert den aktuellen Wert des Balkens.
   */
  public double getValue() {
    return value;
  }

  /**
   * Setzt den aktuellen Wert des Balkens.
   */
  public void setValue(double value) {
    this.value = value;
    // Farbe aktualsieren
    updateForeground();

    // angezeigt wird nur bis zum Minimum/Maximum
    value = Math.min(value,max);
    value = Math.max(value,min);

    if (value > mid) {
      // Bei werden oberhalb der Mitte wird der Negativ-Balken auf
      // die Mitte (also "nichts") gesetzt und der Positiv-Balken
      // auf den Wert oberhalb der Mitte
      negBar.setValue(barMid);
      posBar.setValue(convertToProgressBarValue(value)-barMid);
    } else {
      // Bei werden unterhalb der Mitte wird der Positiv-Balken auf
      // die Mitte (also "nichts") gesetzt und der Negativ-Balken
      // auf den Wert oberhalb der Mitte
      posBar.setValue(barMid);
      negBar.setValue(convertToProgressBarValue(value));
    }
    // String neu setzten, damit er neu angezeigt wird. Diese Aktualisierung
    // erfolgt sonst nicht immer korrekt?!
    posBar.setString( posBar.getString() );
  }

  /**
   * Prueft, ob die Minimum/Mitte/Maximum-Labels angezeigt werden.
   */
  public boolean getValueLabelsPainted() {
    return minLabel.isVisible();
  }

  /**
   * Zeigt die Minimum/Mitte/Maximum-Labels an oder verbirgt sie.
   */
  public void setValueLabelsPainted(boolean visible) {
    minLabel.setVisible(visible);
    midLabel.setVisible(visible);
    maxLabel.setVisible(visible);
  }

  /**
   * Prueft, ob der Balken-Wert angezeigt wird.
   */
  public boolean getStringPainted() {
    return this.stringPainted;
  }

  /**
   * Zeigt den Balken-Wert an oder verbirgt ihn.
   */
  public void setStringPainted(boolean strPaint) {
    this.stringPainted = strPaint;
    // irgendwas "anzeigen" damit die Darstellung des
    // Progressbars aktualisiert wird
    posBar.setString("");
  }

  /**
   * Aktualisiert die Vordergrundfarbe des Balkens, je nachdem, ob der
   * Balkenwert aktuell innerhalb oder ausserhalb der Toleranz liegt.
   */
  protected void updateForeground() {
    if ( minTol <= value && value <= maxTol )
      setForeground( toleranceColor );
    else
      setForeground( intoleranceColor );
  }

  /**
   * Setzt die Farbe, die angezeigt wird, wenn der Balkenwert innerhalb der
   * Toleranz liegt.
   */
  public void setToleranceColor(Color color) {
    this.toleranceColor = color;
    updateForeground();
  }

  /**
   * Liefert die Farbe, die angezeigt wird, wenn der Balkenwert innerhalb der
   * Toleranz liegt.
   */
  public Color getToleranceColor() {
    return toleranceColor;
  }

  /**
   * Setzt die Farbe, die angezeigt wird, wenn der Balkenwert ausserhalb der
   * Toleranz liegt.
   */
  public void setIntoleranceColor(Color color) {
    this.intoleranceColor = color;
    updateForeground();
  }

  /**
   * Liefert die Farbe, die angezeigt wird, wenn der Balkenwert ausserhalb der
   * Toleranz liegt.
   */
  public Color getIntoleranceColor() {
    return intoleranceColor;
  }

  /**
   * Liefert die aktuelle Vordergrund-Farbe des Balkens.
   */
  public Color getForeground() {
    return posBar.getForeground();
  }

  /**
   * Setzt die Vordergrundfarbe des Balken unabhaengig davon, ob der
   * Wert innerhalb oder ausserhalb der Toleranz liegt.<br>
   * <b>Bemerke:</b><br>
   * Dies beeinflusst die (In)Toleranz-Farben nicht! Bei der naechsten
   * Aktualisierung (z.B. {@link #setValue(double)}) wird die Farbe wieder
   * entsprechend der (In)Toleranz-Einstellungen gesetzt.
   * @see #setToleranceColor(Color)
   * @see #setIntoleranceColor(Color)
   */
  public void setForeground(Color color) {
    posBar.setForeground( color );
    negBar.setBackground( color );
  }

  /**
   * Liefert die Hintergrund-Farbe des Balkens.
   */
  public Color getBackground() {
    return posBar.getBackground();
  }

  /**
   * Setzt die Hintergrund-Farbe des Balkens.
   */
  public void setBackground(Color color) {
    posBar.setBackground( color );
    negBar.setForeground( color );
  }

}
