/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing;

import java.io.File;

import javax.swing.JFileChooser;

/**
 * Diese Eingabe-Option dient dazu eine Datei-Angabe zu vorzunehmen. Dies
 * kann ueber manuelle Eingabe erfolgen, oder durch Browsen mittels einem
 * {@link JFileChooser}. Der Wert, den die Eingabe-Option repraesentiert stellt
 * ein {@link File}-Objekt dar.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class FileInputOption extends BrowseInputOption {
  private JFileChooser fileChooser = new JFileChooser();

  /**
   * Erzeugt eine neue Eingabe-Option.
   * @param label Beschreibung der Eingabe-Option
   * @param inputNeeded bestimmt, ob eine Eingabe erforderlich ist
   * @param defValue Standard-Wert der vorgeblendet wird
   */
  public FileInputOption(String label, boolean inputNeeded, File defValue) {
    super(label,inputNeeded,/*defValue == null ? new File(".") :*/ defValue);
  }

  /**
   * Erzeugt eine neue Eingabe-Option.
   * @param label Beschreibung der Eingabe-Option
   * @param inputNeeded bestimmt, ob eine Eingabe erforderlich ist
   */
  public FileInputOption(String label, boolean inputNeeded) {
    this(label,inputNeeded,null);
  }

  /**
   * Liefert den Wert, der in der Option eingegeben wurde. Prueft zuerst
   * auf Gueltigkeit und ruft dann {@link #performGetValue()} auf.
   * @return <code>null</code> wenn die aktuelle Eingabe nicht zulaessig ist
   *         (siehe {@link #isInputValid()})
   */
  public File getValue() {
    return (File)super.getValue();
  }

  /**
   * Liefert den Dateiauswahl-Dialog fuer die Browse-Aktion.
   * @see JFileChooser
   */
  public JFileChooser getFileChooser() {
    return fileChooser;
  }

  /**
   * Liefert das ein {@link File}-Objekt zu der Text-Eingabe (Dateipfad)
   * der Option.
   * @param objectStr Objekt-String
   */
  public File convertFromString(String objectStr) {
    return objectStr == null || objectStr.trim().equals("") ? null : new File(objectStr);
  }

  /**
   * Liefert den kompletten Datei-Pfad fuer die durch die Eingabe-Option
   * repraesentierten {@link File}.
   * @param object {@link File}-Instanz
   * @return {@code null}, fall kein {@link File}-Objekt uebergeben wird
   */
  public String convertToString(Object object) {
//    if ( !(object instanceof File) )
//      return null;
    return ((File)object).getAbsolutePath();
  }

  /**
   * Implementiert die Browse-Aktion die ausgefuehrt wird, wenn der Button der
   * Eingabe-Option gedrueckt wird. Oeffnet einen {@link JFileChooser}.
   * @param actValue aktueller Wert der Eingabe-Option
   * @return <code>null</code> falls die Browse-Aktion abgebrochen wird
   */
  public File performBrowse(Object actValue) {
    fileChooser.setSelectedFile( (File)actValue );
    int ret = -1;
    if ( fileChooser.getDialogType() == JFileChooser.OPEN_DIALOG )
      ret = fileChooser.showOpenDialog(this.getInputComponent());
    else
      ret = fileChooser.showSaveDialog(this.getInputComponent());
    if ( ret  == JFileChooser.APPROVE_OPTION )
      return fileChooser.getSelectedFile();
    return (File)actValue;
  }
}
