/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing;

import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.Point;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import javax.swing.JSpinner;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;


// fuer Doku

/**
 * Diese Komponente stellt eine Kompass-Nadel dar, deren Ausrichtung
 * (Nord = 0\u00b0; West = 90\u00b0) ueber die Maus eingestellt werden kann. Daneben
 * besitzt das Panel einen {@link JSpinner}, um den Kompass-Wert manuell
 * einzustellen.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class InputCompass extends Compass {
  /** Spinner zum Einstellen der Kompass-Einstellung. */
  protected JSpinner spinner = null;

  /**
   * Erzeugt einen neuen Kompass. Der Spinner wird unterhalb des Kompass
   * angezeigt.
   */
  public InputCompass() {
    this(0);
  }

  /**
   * Erzeugt einen neuen Kompass. Der Spinner wird unterhalb des Kompass
   * angezeigt.
   * @param degree angezeigte Grad-Angabe (Nord = 0\u00b0, West = 90\u00b0)
   */
  public InputCompass(double degree) {
    this(degree, true);
  }

  /**
   * Erzeugt einen neuen Kompass. Wird {@code null} als Spinner-Position
   * angegeben, wird KEIN Spinnter angezeigt!
   * @param degree angezeigte Grad-Angabe (Nord = 0\u00b0, West = 90\u00b0)
   * @param showSpinner wenn {@code false}, wird KEIN Spinner angezeigt
   */
  public InputCompass(double degree, boolean showSpinner) {
    this(degree,showSpinner,null,null);
  }

  /**
   * Erzeugt einen neuen Kompass. Wird {@code null} als Spinner-Position
   * angegeben, wird KEIN Spinnter angezeigt!
   * @param degree angezeigte Grad-Angabe (Nord = 0\u00b0, West = 90\u00b0)
   * @param showSpinner wenn {@code false}, wird KEIN Spinner angezeigt
   * @param compassConstr Layout-Constraints fuer den Kompass
   * @param spinnerConstr Layout-Constraints fuer den Spinner
   */
  public InputCompass(double degree, boolean showSpinner, GridBagConstraints compassConstr, GridBagConstraints spinnerConstr) {
    super(degree);

    if ( showSpinner ) {
      if ( compassConstr == null )
        compassConstr = new GridBagConstraints(
          0, 0, 1, 1, 1.0, 1.0, GridBagConstraints.CENTER, GridBagConstraints.BOTH, new Insets(0, 0, 0, 0), 30, 30);
      if ( spinnerConstr == null )
        spinnerConstr = new GridBagConstraints(
          0, 1, 1, 1, 1.0, 0.0, GridBagConstraints.NORTH, GridBagConstraints.NONE, new Insets(5, 0, 0, 0), 30, 0);

      this.remove( getCompassPane() );
      this.setLayout( new GridBagLayout() );
      this.add( getCompassPane(), compassConstr );

      this.spinner = new JSpinner(new RotationSpinnerNumberModel(degree, 0, 360, 0.1, true));
      this.spinner.setEditor( new JSpinner.NumberEditor(this.spinner,SwingUtil.getNumberFormatPattern(3)) );
      SwingUtil.setPreferredWidth( spinner, 100 );
      this.add( spinner, spinnerConstr );

      // ActionListener auf Spinner, um Kompass (THIS) zu aktualisieren
      this.spinner.addChangeListener( new ChangeListener() {
        public void stateChanged(ChangeEvent e) {
          double newValue = (Double)spinner.getValue();
          if ( newValue != InputCompass.this.getValue() )
            InputCompass.this.setValue(newValue);
        }
      });
      // ActionListener auf Kompass (THIS), um Spinner zu aktualisieren
      InputCompass.this.addChangeListener( new ChangeListener() {
        public void stateChanged(ChangeEvent e) {
          double newValue = InputCompass.this.getValue();
          if ( newValue != (Double)spinner.getValue() )
            spinner.setValue(newValue);
        }
      });
    }

    // Mouse-Listener fuer Kompass-Einstellung
    MouseAdapter listener = new MouseAdapter() {
      private boolean button1pressed = false;
      public void mouseDragged(MouseEvent e) {
        if ( button1pressed )
          setDegreeForMousePosition(e.getPoint());
      }
      public void mousePressed(MouseEvent e) {
        if ( e.getButton() == MouseEvent.BUTTON1 ) {
          button1pressed = true;
          setDegreeForMousePosition(e.getPoint());
        }
      }
      public void mouseReleased(MouseEvent e) {
        button1pressed = false;
      }
    };
    getCompassPane().addMouseMotionListener( listener );
    getCompassPane().addMouseListener( listener );
  }

  /**
   * Setzt die Grad-Einstellung der Kompass-Nadel entsprechend einer
   * Maus-Position
   * @param p Maus-Position relativ zur Kompass-Komponente ((0/0) ist oben-links)
   */
  private void setDegreeForMousePosition(Point p) {
    if ( !isEnabled() )
      return;

    int centerX = getCompassPane().getWidth()/2;
    int centerY = getCompassPane().getHeight()/2;

    // (0/0) von Komponente ist oben-links
    int dx = p.x - centerX;
    int dy = centerY - p.y;
    // Distanz zwischen Klickpunkt und Mittelpunkt
    double radius = Math.hypot( dx,dy );
    double degree = Math.toDegrees( Math.asin( dy/radius ) );
    if ( dx < 0 )
      degree = 180 - degree;
    // 0\u00b0 = Norden; 90\u00b0 = West
    setValue( degree-90 );
  }
}
