/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.util.Vector;

import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JPanel;

import schmitzm.swing.event.InputOptionListener;

/**
 * Diese Klasse stellt die allgemeine Oberklasse fuer eine Eingabe-Option
 * des {@link MultipleOptionPane} dar. Jede Option besteht aus einem
 * {@linkplain #descLabel Label} und einem {@linkplain #inpComp Eingabefeld}.
 * Die Art des Eingabefelds wird durch die Unterklassen spezifiziert.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class InputOption<E> extends JPanel {
  /** Speichert eine Referenz auf die Eingabe-Option (<code>this</code>), damit
   *  in inneren Klassen darauf referenziert werden kann. */
  protected final Component OPTION_COMPONENT = this;
  /** Speichert die Beschreibung der Option */
  protected JLabel    descLabel = null;
  /** Speichert die Eingabe-Komponente der Option. */
  protected JComponent inpComp   = null;
  /** Kann eine Fehlermeldung enthalten, wenn die Eingabe nicht
   *  valide ist. Kann direkt von {@link #performIsInputValid()} befuellt
   *  werden. Wird automatisch befuellt, wenn {@link #performIsInputValid()}
   *  eine Exception wirft. */
  protected String invalidInputMess = "";

  private   boolean   inpNeeded = false;
  private   Vector    listeners = new Vector();
  private   E         oldValue  = null;

  /**
   * Erzeugt eine neue Eingabe-Option.
   * @param label       Beschreibung (wird im Label angezeigt)
   * @param inputNeeded gibt an, ob eine Eingabe in der Option erforderlich ist
   */
  public InputOption(String label, boolean inputNeeded) {
    this(label,inputNeeded,false);
  }

  /**
   * Erzeugt eine neue Eingabe-Option.
   * @param label       Beschreibung (wird im Label angezeigt)
   * @param inputNeeded gibt an, ob eine Eingabe in der Option erforderlich ist
   * @param centerInputComp bestimmt, ob die Eingabe-Komponente (siehe {@link #createInputComponent()})
   *                        ins Zentrum des {@link BorderLayout} gelegt wird)
   */
  public InputOption(String label, boolean inputNeeded, boolean centerInputComp) {
    this.inpNeeded = inputNeeded;

    // Darstellung des Containers erzeugen
    setLayout(new BorderLayout());
    descLabel = new JLabel(label);
    inpComp   = createInputComponent();
    inpComp.addFocusListener( new FocusListener() {
      public void focusGained(FocusEvent e) {
        oldValue = getValue();
        fireFocusGained();
      }
      public void focusLost(FocusEvent e) {
        fireFocusLost();
        if ( oldValue == null && getValue() != null ||
             oldValue != null && !oldValue.equals( getValue() ) )
          fireOptionChanged(oldValue,getValue());
      }
    });

    // Elemente dem Container hinzufuegen
    if ( label != null ) {
      add( descLabel, BorderLayout.NORTH );
      add( inpComp, centerInputComp ? BorderLayout.CENTER : BorderLayout.SOUTH );
    } else {
      add( inpComp, centerInputComp ? BorderLayout.CENTER : BorderLayout.NORTH );
    }
    doLayout();
  }

  /**
   * Deaktiviert die Option, in dem die Eingabe-Komponente deaktiviert wird.
   */
  public void setEnabled(boolean enabled) {
    super.setEnabled(enabled);
    inpComp.setEnabled(enabled);
  }

  /**
   * Setzt den Hilfetext fuer Eingabe-Komponente und Beschreibungslabel.
   * @param text Hilfe-Text
   */
  public void setToolTipText(String text) {
    super.setToolTipText(text);
    if ( inpComp != null )
      inpComp.setToolTipText(text);
    if ( descLabel != null )
      descLabel.setToolTipText(text);
  }

  /**
   * Setzt den Fokus auf das Eingabe-Feld der {@code InputOption}.
   */
  public void grabFocus() {
    inpComp.grabFocus();
  }

  /**
   * Liefert den Wert, der in der Option eingegeben wurde. Prueft zuerst
   * auf Gueltigkeit und ruft dann {@link #performGetValue()} auf.
   * @return <code>null</code> wenn die aktuelle Eingabe nicht zulaessig ist
   *         (siehe {@link #isInputValid()})
   */
  public E getValue() {
    if ( !isInputValid() )
      return null;
    return performGetValue();
  }

  /**
   * Setzt einen neuen Wert, mit der die Eingabe-Option belegt wird.
   * Ruft {@link #performSetValue(Object)} auf und bewirkt ein
   * {@link InputOptionListener#optionChanged(Object,Object)} fuer alle
   * angeschlossenen {@link InputOptionListener}, falls sich der Wert der
   * Option geaendert hat.
   * @param newValue neuer Wert
   * @return <code>false</code> gdw. der Objekt-Typ fuer die Option nicht
   *         zulaessig ist
   */
  public boolean setValue(E newValue) {
    // alten Wert speichern
    oldValue = getValue();
    if ( !performSetValue(newValue) )
      return false;
    newValue = getValue();
    if ( oldValue==null && newValue!=null || oldValue!=null && newValue==null || oldValue!=null && !oldValue.equals( newValue ) )
      fireOptionChanged(oldValue,getValue());
    return true;
  }

  /**
   * Liefert die Beschreibung der Eingabe-Option.
   */
  public String getLabel() {
    return descLabel.getText();
  }

  /**
   * Prueft, ob fuer das Feld zwingend eine Eingabe erforderlich ist.
   */
  public boolean inputNeeded() {
    return inpNeeded;
  }

  /**
   * Bestimmt, ob fuer das Feld zwingend eine Eingabe erforderlich ist.
   */
  public void setInputNeeded(boolean inpNeeded) {
    this.inpNeeded = inpNeeded;
  }

  /**
   * Prueft, ob die Eingabe in dem Feld zulaessig ist. Die Implementierung
   * prueft zuerst auf Leereingabe und ruft dann {@link #performIsInputValid()}
   * auf.
   * @return <code>false</code> falls die Option leer ist, aber eine Eingabe
   *         erforderlich ist
   */
  public boolean isInputValid() {
    // Leereingabe
    if (  inputEmpty() )
      return !inputNeeded();
    try {
      this.invalidInputMess = "";
      return performIsInputValid();
    } catch (Exception err) {
      this.invalidInputMess = err.getMessage();
      return false;
    }
  }

  /**
   * Prueft, ob die Eingabe in dem Feld zulaessig ist und wird ggf. eine
   * Exception. Die Implementierung
   */
  public void checkInputAndError() throws Exception {
    if ( !isInputValid() )
      throw new Exception( getInvalidInputMessage() );
  }

  /**
   * Liefert die letzte von {@link #isInputValid()} erzeugte
   * Fehlermeldung.
   */
  public String getInvalidInputMessage() {
    return invalidInputMess;
  }

  /**
   * Prueft, ob das Feld eine Leereingabe beinhaltet.
   * @see #performIsInputEmpty()
   */
  public boolean inputEmpty() {
    return performIsInputEmpty();
  }

  /**
   * Liefert eine Referenz auf die Eingabe-Komponente.<br>
   * <b>Bemerke:</b><br>
   * Uber diese Referenz sollte hoechsten ihr Layout beeinflusst werden, aber
   * nicht ihr Verhalten, da ansonsten die Funktionalitaet der <code>InputOption</code>
   * negativ beeinflusst werden kann!!
   */
  public JComponent getInputComponent() {
    return inpComp;
  }

  /**
   * Liefert eine Referenz auf das Beschreibungs-Label.<br>
   * <b>Bemerke:</b><br>
   * Uber diese Referenz sollte hoechsten das Layout beeinflusst werden, aber
   * nicht das Verhalten, da ansonsten die Funktionalitaet der <code>InputOption</code>
   * negativ beeinflusst werden kann!!
   */
  public JLabel getDescriptionLabel() {
    return descLabel;
  }

  /**
   * Erzeugt eine neue Instanz der Eingabe-Komponente. z.B. ein Text-Eingabefeld
   * oder eine Combo-Box.
   */
  protected abstract JComponent createInputComponent();

  /**
   * Liefert den aktuellen Wert der Eingabe-Option.
   */
  protected abstract E performGetValue();

  /**
   * Setzt den aktuellen Wert der Eingabe-Option. Ist der Objekt-Typ fuer die
   * Option nicht zulaessig, sollte nichts gemacht werden und <code>false</code>
   * zurueckgegeben werden.
   * @param newValue neuer Wert
   * @return <code>false</code> gdw. der Objekt-Typ fuer die Eingabe-Option
   *         nicht zulaessig ist
   */
  protected abstract boolean performSetValue(E newValue);

  /**
   * Prueft, ob die aktuelle Eingabe leer ist.
   */
  protected abstract boolean performIsInputEmpty();

  /**
   * Prueft, ob die aktuelle Eingabe leer ist.
   */
  protected abstract boolean performIsInputValid();

  /**
   * Fuegt der Eingabeoption einen Listener hinzu.
   * @param l neuer Listener
   */
  public void addInputOptionListener(InputOptionListener l) {
    listeners.add(l);
  }

  /**
   * Entfernt einen Listener von der Eingabeoption.
   * @param l zu entfernender Listener
   */
  public void removeInputOptionListener(InputOptionListener l) {
    listeners.remove(l);
  }

  /**
   * Informiert alle {@link InputOptionListener}, dass die Eingabe-Option
   * den Fokus erhalten hat.
   */
  protected void fireFocusGained() {
    for (int i=0; i<listeners.size(); i++)
      if ( listeners.elementAt(i) instanceof InputOptionListener )
        ((InputOptionListener)listeners.elementAt(i)).optionGainedFocus(this);
  }

  /**
   * Informiert alle {@link InputOptionListener}, dass die Eingabe-Option
   * den Fokus verloren hat.
   */
  protected void fireFocusLost() {
    for (int i=0; i<listeners.size(); i++)
      if ( listeners.elementAt(i) instanceof InputOptionListener )
        ((InputOptionListener)listeners.elementAt(i)).optionLostFocus(this);
  }

  /**
   * Informiert alle {@link InputOptionListener}, dass sich der Wert der
   * Eingabeoption geaendert hat.
   * @param oldValue alter Optionswert
   * @param newValue neuer Optionswert
   */
  protected void fireOptionChanged(E oldValue, E newValue) {
    for (int i=0; i<listeners.size(); i++)
      if ( listeners.elementAt(i) instanceof InputOptionListener )
        ((InputOptionListener)listeners.elementAt(i)).optionChanged(this,oldValue,newValue);
  }

}
