/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing;

import java.awt.Color;
import java.awt.Component;
import java.awt.FlowLayout;
import java.awt.Graphics;
import java.awt.LayoutManager;

import javax.swing.BorderFactory;
import javax.swing.border.TitledBorder;

/**
 * Diese Klasse erweitert das {@link javax.swing.JPanel} aus Standard-Java um
 * einige (nuetzliche) Funktionen.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class JPanel extends javax.swing.JPanel {

	/** Wenn beim deaktivieren evt. ein {@link TitledBorder} ausgegraut wird, dann wird sich in dieser Variable die orginale Farbe gemerkt.**/
  private Color backupColor = null;

/**
   * Erzeugt ein neues Panel.
   * @param layout {@link LayoutManager} fuer das Panel
   * @param isDoubleBuffered Flag, ob zusaetzlicher Speicher verwendet werden soll um
   *        flicker-freie Updates zu gewaehrleisten
   */
  public JPanel(LayoutManager layout, boolean isDoubleBuffered) {
    super(layout, isDoubleBuffered);
  }

  /**
   * Erzeugt ein neues Panel.
   * @param layout {@link LayoutManager} fuer das Panel
   */
  public JPanel(LayoutManager layout) {
    super(layout);
  }

  /**
   * Erzeugt ein neues Panel mit einem {@link FlowLayout}.
   * @param isDoubleBuffered Flag, ob zusaetzlicher Speicher verwendet werden soll um
   *        flicker-freie Updates zu gewaehrleisten
   */
  public JPanel(boolean isDoubleBuffered) {
    super(isDoubleBuffered);
  }

  /**
   * Erzeugt ein neues Panel mit einem {@link FlowLayout}.
   */
  public JPanel() {
    super();
  }

  /**
	 * Erzeugt ein neues Panel mit angegebenen {@link LayoutManager} und einem
	 * {@link TitledBorder}
	 * 
	 * @param borderTitle
	 *            Wenn ungleich null wird {@link TitledBorder} für das Panel
	 *            erzeugt
	 */
	public JPanel(LayoutManager layout, String borderTitle) {
		super(layout);
		if (borderTitle != null) {
			setBorder(BorderFactory.createTitledBorder(borderTitle));
		}
	}

/**
   * Aktiviert und deaktiviert alle Komponenten des Panels. 
   */
  @Override
  public void setEnabled(boolean enabled) {
    super.setEnabled(enabled);
    
    if (getBorder() instanceof TitledBorder) {
    	TitledBorder tb = (TitledBorder)getBorder();
    	if(enabled) {
    		if (backupColor != null) tb.setTitleColor(backupColor);
    	} else {
    		backupColor = tb.getTitleColor();
    		tb.setTitleColor(backupColor.brighter().brighter());
    	}
    }
    
    for (Component c : getComponents())
      c.setEnabled(enabled);
  }

  /**
   * Setzt die Hintergrund-Farbe der Komponente und optional die der
   * untergeordneten Komponenten
   * @param color neue Hintergrund-Farbe
   * @param components bestimmt, ob auch die Hintergrundfarbe der Unterkomponenten
   *        gesetzt werden soll
   */
  public void setBackground(Color color, boolean components) {
    // Hintergrund von Panel selbst setzen
    super.setBackground(color);
    // wenn rekursive Verarbeitung, dann auch den Hintergrund
    // der Componenten aendern
    if ( components )
      for ( Component c : getComponents() )
        if ( c instanceof JPanel )
          ((JPanel)c).setBackground(color,components);
        else
          c.setBackground(color);
  }
  

	/**
	 * Nuetzlich wenn die Componente gedruckt (z.B. wenn ein Screenshot
	 * gemacht wird) wird. Dann werden wird der Hintergrund auf {@link Color#WHITE}
	 * gesetzt.
	 */
	@Override
	public void print(final Graphics g) {
		final Color orig = getBackground();
		setBackground(Color.WHITE, true);
		// wrap in try/finally so that we always restore the state
		try {
			super.print(g);
		} finally {
			setBackground(orig, true);
		}
	}

}
