/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing;

import javax.swing.JTextField;

/**
 * Diese Klasse stellt eine manuelle Eingabe-Option fuer das {@link MultipleOptionPane}
 * dar. Die Eingabe erfolgt ueber ein {@link JTextField}. Ausser der
 * "Input-Needed"-Restriktion, werden keine weiteren Anforderungen an die
 * Eingabe gestellt. Die Klasse kann nicht direkt instanziiert werden.
 * Statt dessen sind die eingebetteten Klassen
 * <ul>
 * <li>{@link ManualInputOption.Text    ManualInputOption.Text}</li>
 * <li>{@link ManualInputOption.Integer ManualInputOption.Integer}</li>
 * <li>{@link ManualInputOption.Double  ManualInputOption.Double}</li>
 * </ul>
 * zu verwenden.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ManualInputOption extends InputOption {
  /**
   * Erzeugt eine neue Eingabe-Option.
   * @param label       Beschreibung
   * @param inputNeeded gibt an, ob eine Eingabe erforderlich ist
   * @param defValue    Wert der im Textfeld vorgeblendet wird
   */
  protected ManualInputOption(String label, boolean inputNeeded, Object defValue) {
    super(label,inputNeeded);
    ((JTextField)inpComp).setText( defValue != null ? defValue.toString() : "");
  }

  /**
   * Erzeugt eine neue Instanz von {@link JTextField}.
   */
  protected JTextField createInputComponent() {
    return new JTextField();
  }

  /**
   * Liefert immer <code>true</code>, da keine speziellen Anforderungen
   * an die Text-Eingabe gestellt werden.
   */
  protected boolean performIsInputValid() {
    return true;
  }

  /**
   * Prueft, ob im Eingabefeld ein Leerstring eingegeben wurde.
   */
  protected boolean performIsInputEmpty() {
    return ((JTextField)inpComp).getText().trim().equals("");
  }

  /**
   * Liefert die aktuelle Eingabe im {@link JTextField}.
   */
  protected Object performGetValue() {
    return ((JTextField)inpComp).getText();
  }

  /**
   * Setzt die aktuelle Eingabe im {@link JTextField}.
   * @param newValue neuer Wert (muss vom Typ {@link String} sein!)
   * @return <code>false</code> wenn kein {@link String} uebergeben wird
   */
  protected boolean performSetValue(Object newValue) {
    if ( newValue == null )
      newValue = "";
    if ( !(newValue instanceof String) )
      return false;
    ((JTextField)inpComp).setText((String)newValue);
    return true;
  }

  /**
   * Diese Klasse stellt eine Eingabe-Option dar, in der ein beliebigen Text
   * eingegeben werden kann.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class Text extends ManualInputOption {
    /**
     * Erzeugt eine neue Eingabe-Option.
     * @param label       Beschreibung
     * @param inputNeeded gibt an, ob eine Eingabe erforderlich ist
     * @param defValue    Wert der im Textfeld vorgeblendet wird
     */
    public Text(String label, boolean inputNeeded, String defValue) {
      super(label,inputNeeded,defValue);
    }

    /**
     * Erzeugt eine neue Eingabe-Option.
     * @param label       Beschreibung
     * @param inputNeeded gibt an, ob eine Eingabe erforderlich ist
     */
    public Text(String label, boolean inputNeeded) {
      this(label,inputNeeded,"");
    }

    /**
     * Erzeugt eine neue Eingabe-Option, in der eine Eingabe zwingend erforderlich
     * ist.
     * @param label Beschreibung
     */
    public Text(String label) {
      this(label,true,"");
    }

    /**
     * Erzeugt eine neue Eingabe-Option, in der eine Eingabe zwingend erforderlich
     * ist.
     * @param label    Beschreibung
     * @param defValue Wert der im Textfeld vorgeblendet wird
     */
    public Text(String label, String defValue) {
      this(label,true,defValue);
    }

    /**
     * Liefert den Wert, der in der Option eingegeben wurde. Prueft zuerst
     * auf Gueltigkeit und ruft dann {@link #performGetValue()} auf.
     * @return <code>null</code> wenn die aktuelle Eingabe nicht zulaessig ist
     *         (siehe {@link #isInputValid()})
     */
    public String getValue() {
      return (String)super.getValue();
    }
  }

  /**
   * Diese Klasse stellt eine Eingabe-Option dar, in der ein Integer-Zahlen
   * eingegeben werden kann.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class Integer extends ManualInputOption {
    /**
     * Erzeugt eine neue Eingabe-Option.
     * @param label       Beschreibung
     * @param inputNeeded gibt an, ob eine Eingabe erforderlich ist
     * @param defValue    Wert der im Textfeld vorgeblendet wird
     */
    public Integer(String label, boolean inputNeeded, int defValue) {
      super(label,inputNeeded,String.valueOf(defValue));
    }

    /**
     * Erzeugt eine neue Eingabe-Option. Auch wenn die leere Eingabe u.U.
     * nicht zulaessig ist, wird ein Leerstring vorgeblendet.
     * @param label       Beschreibung
     * @param inputNeeded gibt an, ob eine Eingabe erforderlich ist
     */
    public Integer(String label, boolean inputNeeded) {
      super(label,inputNeeded,"");
    }

    /**
     * Erzeugt eine neue Eingabe-Option, in der eine Eingabe zwingend erforderlich
     * ist. Auch wenn die leere Eingabe u.U. nicht zulaessig ist, wird ein
     * Leerstring vorgeblendet.
     * @param label Beschreibung
     */
    public Integer(String label) {
      super(label,true,"");
    }

    /**
     * Erzeugt eine neue Eingabe-Option, in der eine Eingabe zwingend erforderlich
     * ist.
     * @param label    Beschreibung
     * @param defValue Wert der im Textfeld vorgeblendet wird
     */
    public Integer(String label, int defValue) {
      this(label,true,defValue);
    }

    /**
     * Liefert den Wert, der in der Option eingegeben wurde. Prueft zuerst
     * auf Gueltigkeit und ruft dann {@link #performGetValue()} auf.
     * @return <code>null</code> wenn die aktuelle Eingabe nicht zulaessig ist
     *         (siehe {@link #isInputValid()})
     */
    public java.lang.Integer getValue() {
      return (java.lang.Integer)super.getValue();
    }

    /**
     * Transformiert die Text-Eingabe in einen Integer.
     * @return <code>null</code> falls das Eingabefeld leer ist.
     */
    protected java.lang.Integer performGetValue() {
      if ( inputEmpty() )
        return null;
      return java.lang.Integer.parseInt(((JTextField)inpComp).getText());
    }

    /**
     * Prueft, ob ein gueltiger Integer-Wert im Feld eingegeben wurde.
     */
    protected boolean performIsInputValid() {
      try {
        java.lang.Integer.parseInt(((JTextField)inpComp).getText());
      } catch (NumberFormatException err) {
        return false;
      }
      return super.performIsInputValid();
    }
  }

  /**
   * Diese Klasse stellt eine Eingabe-Option dar, in der ein Double-Zahlen
   * eingegeben werden kann.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class Double extends ManualInputOption {
    /**
     * Erzeugt eine neue Eingabe-Option.
     * @param label       Beschreibung
     * @param inputNeeded gibt an, ob eine Eingabe erforderlich ist
     * @param defValue    Wert der im Textfeld vorgeblendet wird
     */
    public Double(String label, boolean inputNeeded, double defValue) {
      super(label,inputNeeded,String.valueOf(defValue));
    }

    /**
     * Erzeugt eine neue Eingabe-Option. Auch wenn die leere Eingabe u.U.
     * nicht zulaessig ist, wird ein Leerstring vorgeblendet.
     * @param label       Beschreibung
     * @param inputNeeded gibt an, ob eine Eingabe erforderlich ist
     */
    public Double(String label, boolean inputNeeded) {
      super(label,inputNeeded,"");
    }

    /**
     * Erzeugt eine neue Eingabe-Option, in der eine Eingabe zwingend erforderlich
     * ist. Auch wenn die leere Eingabe u.U. nicht zulaessig ist, wird ein
     * Leerstring vorgeblendet.
     * @param label Beschreibung
     */
    public Double(String label) {
      super(label,true,"");
    }

    /**
     * Erzeugt eine neue Eingabe-Option, in der eine Eingabe zwingend erforderlich
     * ist.
     * @param label    Beschreibung
     * @param defValue Wert der im Textfeld vorgeblendet wird
     */
    public Double(String label, int defValue) {
      this(label,true,defValue);
    }

    /**
     * Liefert den Wert, der in der Option eingegeben wurde. Prueft zuerst
     * auf Gueltigkeit und ruft dann {@link #performGetValue()} auf.
     * @return <code>null</code> wenn die aktuelle Eingabe nicht zulaessig ist
     *         (siehe {@link #isInputValid()})
     */
    public java.lang.Double getValue() {
      return (java.lang.Double)super.getValue();
    }

    /**
     * Transformiert die Text-Eingabe in einen Double.
     * @return <code>null</code> falls das Eingabefeld leer ist.
     */
    protected java.lang.Double performGetValue() {
      if ( inputEmpty() )
        return null;
      return java.lang.Double.parseDouble(((JTextField)inpComp).getText());
    }

    /**
     * Prueft, ob ein gueltiger Double-Wert im Feld eingegeben wurde.
     */
    protected boolean performIsInputValid() {
      try {
        java.lang.Double.parseDouble(((JTextField)inpComp).getText());
      } catch (NumberFormatException err) {
        return false;
      }
      return super.performIsInputValid();
    }
  }
}
