/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing;

import java.awt.BorderLayout;
import java.awt.Container;

import javax.swing.JComponent;
import javax.swing.JSplitPane;
import javax.swing.border.Border;

/**
 * <hr>
 * <small><i>Diese Klasse wurde urspruenglich 2002/2003 von <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *            fuer die ADAGIOS-Java-Library (Universitaet Bonn) erstellt, 2009 in das
 *            SCHMITZM-Projekt uebernommen und teilweise erweitert.</i></small>
 * <hr>
 * Ein <code>MultiSplitPane</code> ist eine Menge von horizontal oder
 * vertikal angeordneten Containern.<br>
 * Die Groesse dieser <code>Container</code> kann durch den Benutzer manuell
 * veraendert werden, in dem er die Trennlinien ("Dividers") zwischen den
 * Containern verschiebt.<br>
 * Das <code>MultiSplitPane</code> ist implementiert als eine baumartige
 * Verschachtelung von <code>JSplitPane</code>-Objekten, in denen immer
 * genau zwei Komponenten hinterlegt werden koennen.<br>
 * Aus diesem Grund kann die Darstellung der Randbereiche recht
 * unregelmaessig erscheinen. Deshalb ist es sinnvoll, den Rand entweder
 * komplett auszuschalten (<code>setInnerBorder(null)</code>), oder einen
 * unsichtbaren Rand zu verwenden (<code>setInnerBorder( new EmptyBorder(..) )</code>.
 * Dies gilt vor allem, wenn man ein <code>MultiSplitPane</code> mit vielen
 * Containern verwendet (mehr als 6 oder 7).
 * @see javax.swing.JSplitPane
 * @see javax.swing.JComponent
 * @see #setInnerBorder(Border)
 * @see javax.swing.border.EmptyBorder
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 */
public class MultiSplitPane extends JComponent {

  public static final int HORIZONTAL_SPLIT = JSplitPane.HORIZONTAL_SPLIT;
  public static final int VERTICAL_SPLIT   = JSplitPane.VERTICAL_SPLIT;

  private Container[]    container     = null;
  private SplitPaneTree  rootSplitPane = null;
  private int            orientation   = HORIZONTAL_SPLIT;
  private int            dividerSize   = 0;
  private Border         innerBorder   = new JSplitPane().getBorder();

  /**
   * Erzeugt ein neues MultiSplitPane. Dieses muss aus mindestens
   * 2 Containern bestehen. Diese werden horizontal angeordnet.
   * @param compCount   Anzahl an Containern im MultiSplitPane
   * @exception java.lang.IllegalArgumentException falls die
   *            Container-Anzahl < 2 ist.
   */
  public MultiSplitPane(int compCount) {
    this(compCount,HORIZONTAL_SPLIT);
  }

  /**
   * Erzeugt ein neues MultiSplitPane. Dieses muss aus mindestens
   * 2 Containern bestehen. Moegliche Orientierungen:
   * <ul>
   * <li><code>HORIZONTAL_SPLIT</code></li>
   * <li><code>VERTICAL_SPLIT</code></li>
   * </ul>
   * @param compCount   Anzahl an Containern im MultiSplitPane
   * @param orientation Orientierung des MultiSplitPane
   * @exception java.lang.IllegalArgumentException falls die Orientierung
   *            ungueltig ist oder die Container-Anzahl < 2 ist.
   */
  public MultiSplitPane(int compCount, int orientation) {
    super();
    this.setLayout( new BorderLayout() );

    if (compCount < 2)
      throw new IllegalArgumentException("MultiSplitPane must have at least 2 Components");
    // erstmal nur die Orientierung checken, setzen folgt nachdem
    // der SplitPaneTree aufgebaut ist
    checkOrientation(orientation);

    // Referenz auf die Container (der Effizienz wegen) nochmal
    // in einer Liste speichern
    container = new Container[compCount];
    for (int i=0; i<container.length; i++) {
      container[i] = new Container();
      container[i].setLayout( new BorderLayout() );
    }
    rootSplitPane = new SplitPaneTree(container);
    // Orientierung setzen
    setOrientation(orientation);
    // Basis-SplitPane im Haupt-Container einfuegen
    this.add(rootSplitPane.getSplitPane());
  }

  /**
   * Liefert die Anzahl an Komponenten, in die das SplitPane aufgeteilt ist.
   */
  public int getContainerCount() {
    return container.length;
  }

  /**
   * Liefert den Inhalt eine Teil-Panes.
   * @param i    Index des Teil-Panes (von links/oben nach rechts/unten)
   * @exception java.lang.ArrayIndexOutOfBoundsException falls ein ungueltiger
   *            Index angegeben wird
   */
  public Container getContainer(int i) {
    return container[i];
  }

  /**
   * Setzt den Inhalt eine Teil-Panes.
   * @param i    Index des Teil-Panes (von links/oben nach rechts/unten)
   * @param cont neuer Inhalt des Teil-Panes
   * @exception java.lang.ArrayIndexOutOfBoundsException falls ein ungueltiger
   *            Index angegeben wird
   */
  public void setContainer(int i, Container cont) {
    this.container[i] = cont;
    rootSplitPane.setContainer(i,cont);
  }

  /**
   * Liefert die Orientierung in der die Container des MultiSplitPanes
   * angeordnet sind.
   * @see #setOrientation(int)
   */
  public int getOrientation() {
    return this.orientation;
  }

  /**
   * Aendert die Orientierung des SplitPanes. Moegliche
   * Orientierungen:
   * <ul>
   * <li><code>HORIZONTAL_SPLIT</code></li>
   * <li><code>VERTICAL_SPLIT</code></li>
   * </ul>
   * @param orientation neue Orientierung
   * @exception java.lang.IllegalArgumentException falls die Orientierung
   *            ungueltig ist
   * @see #getOrientation()
   */
  public void setOrientation(int orientation) {
    checkOrientation(orientation);

    final int o = orientation;
    rootSplitPane.performAllSplitPanes(
        new SplitPaneTreePerformer() {
            public void performSplitPaneTree(SplitPaneTree tree) {
              tree.getSplitPane().setOrientation(o);
            }
        }
    );
    this.orientation = orientation;
  }

  /**
   * Checkt, ob die angegebene Orientierung gueltig ist. Andernfalls wird
   * eine Exception geworfen.
   * @param orientation Orientierung
   * @exception java.lang.IllegalArgumentException falls die Orientierung
   *            ungueltig ist
   */
  protected void checkOrientation(int orientation) {
    if (orientation != HORIZONTAL_SPLIT &&
        orientation != VERTICAL_SPLIT)
      throw new IllegalArgumentException("HORIZONTAL_SPLIT or VERTICAL_SPLIT expected!!");
  }

  /**
   * Liefert die Groesse des Dividers (in Pixeln).
   */
  public int getDividerSize() {
    return dividerSize;
  }

  /**
   * Setzt die Groesse der Divider-Balken. Die Darstellung des "laestigen"
   * einpixligen Strichs laesst sich zwar verhindern, indem man die
   * Divider-Groesse auf 0 setzt, leider sind dann aber auch die Groessen der
   * einzelnen Container nicht mehr veraenderbar!!
   * @param size Groesse in Pixeln
   * @see #getDividerSize()
   */
  public void setDividerSize(int size) {
    final int s = size;
    rootSplitPane.performAllSplitPanes(
        new SplitPaneTreePerformer() {
            public void performSplitPaneTree(SplitPaneTree tree) {
              tree.getSplitPane().setDividerSize(s);
            }
        }
    );
    this.dividerSize = size;
  }

  /**
   * Liefert das gesetzte Gewicht eines Containers, mit dessen Anteil
   * dieser bei einer Platzvergroesserung beruecksichtigt wird.
   * @param i Index des Teil-Panes (von links/oben nach rechts/unten)
   * @exception java.lang.ArrayIndexOutOfBoundsException falls ein ungueltiger
   *            Index angegeben wird
   * @see #setResizeWeigth(double[])
   */
  public double getResizeWeight(int i) {
    // Teilbaum (JSplitPane) suchen, in dem der Container mit
    // dem Index i abgelegt ist
    SplitPaneTree tree = rootSplitPane.getSplitPaneTree(i);

    // ResizeWeight-Anteil des SplitPanes
    double splitPaneWeight = tree.getResizeWeight();
    // Gewicht der linken Komponente (bezogen auf das eine SplitPane)
    double weight = tree.getSplitPane().getResizeWeight();
    // liegt der gesuchte Container im rechten Teil-Pane,
    // muss das Gegengewicht genommen werden
    if ( tree.getSplitPane().getLeftComponent() != getContainer(i) )
      weight = 1 - weight;

    return weight * splitPaneWeight;
  }

  /**
   * Liefert die Gewichte aller Containers, mit deren Anteil diese bei
   * einer Platzvergroesserung beruecksichtigt werden.
   * @see #getResizeWeight(int)
   * @see #setResizeWeigth(double[])
   */
  public double[] getResizeWeight() {
    double[] weight = new double[ getContainerCount() ];
    for (int i=0; i<weight.length; i++)
      weight[i] = getResizeWeight(i);
    return weight;
  }

  /**
   * Setzt die Container-Gewichtungen, mit deren Anteil diese bei einer
   * Platzvergroesserung beruecksichtigt werden. Bei der Angabe der
   * Gewichte sind folgende Punkte zu beachten:
   * <ul>
   * <li>Die Summe der angegebenen Gewichte darf 1 nicht ueberschreiten.</li>
   * <li>Es duerfen hoechstens so viele Gewichte angegeben werden, wie
   *     Container im <code>MultiSplitPane</code> vorhanden sind.</li>
   * <li>Werden genau so viele Gewichte angegeben, wie Container im
   *     <code>MultiSplitPane</code> vorhanden sind, muss die Summe der
   *     Gewichte 1 ergeben</li>
   * <li>Werden weniger Gewichte als vorhandene Container angegeben, so wird
   *     das restliche Gewicht gleichmaessig auf die uebrigen Container
   *     verteilt.</li>
   * </ul>
   * @param weight neue Gewichte
   * @exception java.lang.IllegalArgumentException falls eine der obigen
   *            Bedingungen verletzt ist.
   */
  public void setResizeWeigth(double[] weight) {
    if ( weight.length > getContainerCount() )
      throw new IllegalArgumentException("Es duerfen hoechstens so viele Gewichte, wie Container angegeben werden!");

    // Gewichte in einen Array kopieren, dessen Groesse genau der
    // Anzahl an Containern entspricht
    final double[] newWeight = new double[ getContainerCount() ];
          double   weightSum = 0.0;

    for (int i=0; i<newWeight.length; i++) {
      if ( i<weight.length ) {
        weightSum += weight[i]; // Gewichtssumme ermitteln
        if ( weightSum > 1.0 )
          throw new IllegalArgumentException("Die angegebnene Summe der Gewichte ist groesser als 1!!");
        newWeight[i] = weight[i]; // vorhandenes Gewicht kopieren
      }
      else {
        // Bei Angabe von zu wenig Gewichten, wird der Rest "gerecht"
        // verteilt
        newWeight[i] = (1-weightSum) / (newWeight.length - weight.length);
      }
    }

    if ( weight.length == newWeight.length && weightSum < 1 )
      throw new IllegalArgumentException("Werden Gewichte fuer alle Container angegeben, muss deren Summe 1 ergeben!!");

    rootSplitPane.performAllSplitPanes(
        new SplitPaneTreePerformer() {
            private int contCounter = 0;

            public void performSplitPaneTree(SplitPaneTree tree) {
              // Beinhaltet das SplitPane einen (finalen) Container
              // wird dessen neues Gewicht genommen, ansonsten das
              // Gewicht des gesamten Teilbaums (dieses ist aufgrund
              // des POSTORDER-Druchlaufs bereits neu berechnet!!)
              double lWeight = ( !tree.isSplittedLeft() )  ? newWeight[contCounter++] : tree.getLeftResizeWeight();
              double rWeight = ( !tree.isSplittedRight() ) ? newWeight[contCounter++] : tree.getRightResizeWeight();
              // neues Gewicht fuer den Teilbaum neu setzen
              tree.setResizeWeight( lWeight + rWeight );
              // neues Gewicht fuer das Root-SplitPane des Teilbaum neu setzen
              tree.getSplitPane().setResizeWeight( lWeight / tree.getResizeWeight() );
            }
        });
  }

  /**
   * Liefert die Border des <code>MultiSplitPane</code>
   */
  public Border getInnerBorder() {
    return this.innerBorder;
  }

  /**
   * Setzt die Border fuer alle Container.
   */
  public void setInnerBorder(final Border border) {
    rootSplitPane.performAllSplitPanes(
        new SplitPaneTreePerformer() {
            public void performSplitPaneTree(SplitPaneTree tree) {
              tree.getSplitPane().setBorder(border);
            }
        });
    this.innerBorder = border;
  }

  /////////////////////////////////////////////////////////////////////////
  ///// Interface   SplitPaneTreePerformer                            /////
  /////////////////////////////////////////////////////////////////////////
  private interface SplitPaneTreePerformer {
    public void performSplitPaneTree(SplitPaneTree tree);
  }

  /////////////////////////////////////////////////////////////////////////
  ///// Klasse   SplitPaneTree                                        /////
  /////////////////////////////////////////////////////////////////////////
  private class SplitPaneTree {
    private JSplitPane    splitPane    = null;
    private int           contCount    = 0;
    private SplitPaneTree leftTree     = null;
    private SplitPaneTree rightTree    = null;
    private double        resizeWeight = 0.0;

    /**
     * Interner Konstruktor zum Initialisieren ALLER Variablen.
     */
    private void init(JSplitPane splitPane, int contCount, double weight, SplitPaneTree left, SplitPaneTree right) {
      this.splitPane    = splitPane;
      this.contCount    = contCount;
      this.resizeWeight = weight;
      this.leftTree     = left;
      this.rightTree    = right;
    }

    /**
     * Interner Konstruktor zum Initialisieren eines (Teil-)Baums, der
     * sowohl links, als auch rechts weiter geplittet ist.
     */
    private void initInner(JSplitPane splitPane, double weight, SplitPaneTree left, SplitPaneTree right) {
      init(splitPane,left.getContainerCount()+right.getContainerCount(),weight,left,right);
    }

    /**
     * Erzeugt einen neuen (Teil-)Baum, der sowohl links, als auch rechts
     * weiter gesplittet ist. Beide Teil-Panes sind also weiter unterteilt.
     * @param splitPane <code>JSplitPane</code> fuer das der (Teil-)Baum steht
     * @param weight    Gewicht des Teilbaums bezogen auf den gesamten Baum
     *                  (Anteil am neuen Platz, den die Container im Teilbaum
     *                  erhalten, wenn sich das Pane vergroessert)
     * @param left      linker Teilbaum
     * @param right     rechter Teilbaum
     */
    public SplitPaneTree(JSplitPane splitPane, double weight, SplitPaneTree left, SplitPaneTree right) {
      initInner(splitPane,weight,left,right);
    }

    /**
     * Interner Konstruktor zum Initialisieren eines Blattes.
     */
    private void initLeaf(JSplitPane splitPane, double weight) {
      init(splitPane,2,weight,null,null);
    }

    /**
     * Erzeugt ein neues Blatt. Beide Teil-Panes enthalten also einen
     * <code>Container</code> und sind nicht weiter gesplittet.
     * @param splitPane <code>JSplitPane</code> fuer das der (Teil-)Baum steht
     * @param weight    Gewicht des Teilbaums bezogen auf den gesamten Baum
     *                  (Anteil am neuen Platz, den die Container im Teilbaum
     *                  erhalten, wenn sich das Pane vergroessert)
     */
    public SplitPaneTree(JSplitPane splitPane, double weight) {
      initLeaf(splitPane,weight);
    }

    /**
     * Interner Konstruktor zum Initialisieren eines (Teil-)Baums, der
     * links gesplittet ist und rechts ein Blatt hat.
     */
    private void initHalfLeaf(JSplitPane splitPane, double weight, SplitPaneTree left) {
      init(splitPane,left.getContainerCount()+1,weight,left,null);
    }

    /**
     * Erzeugt einen neuen (Teil-)Baum, der nach links gesplittet ist und rechts
     * endet ("Halb-Blatt"). Das linke Teil-Pane ist also weiter unterteilt, waehrend im
     * rechten Teil-Pane ein <code>Container</code> zu finden ist.
     * @param splitPane <code>JSplitPane</code> fuer das der (Teil-)Baum steht
     * @param weight    Gewicht des Teilbaums bezogen auf den gesamten Baum
     *                  (Anteil am neuen Platz, den die Container im Teilbaum
     *                  erhalten, wenn sich das Pane vergroessert)
     * @param left      linker Teilbaum
     */
    public SplitPaneTree(JSplitPane splitPane, double weight, SplitPaneTree left) {
      initHalfLeaf(splitPane,weight,left);
    }

    /**
     * Erzeugt einen neuen Baum.
     * @param container Container, die das SplitPane haben soll
     * @param start     Index im <code>Container</code>-Array, bei dem
     *                  begonnen wird (alle davor werden ignoriert)
     * @param weight    Gewicht des Teilbaums bezogen auf den gesamten Baum
     *                  (Anteil am neuen Platz, den die Container im Teilbaum
     *                  erhalten, wenn sich das Pane vergroessert)
     */
    public SplitPaneTree(Container[] container, int start, int count, double weight) {
//  System.out.println(start+"\t"+count+"\t"+weight);
      if ( count < 2 )
        throw new IllegalArgumentException("SplitPane must have at least 2 Containers");

      if ( start + count -1 >= container.length )
        throw new IllegalArgumentException("'count' too big for this start position in Container-Array");

      this.splitPane = new JSplitPane();

      // Aufteilung ermitteln
      int rightCount  = count / 2;
      int leftCount   = count - rightCount;
      // Vergroesserungs-Gewichte ermitteln:
      // --> Entsprechend der Aufteilung erhalten die beiden Teil-Panes
      //     einen Anteil am Gesamt-Gewicht des SplitPanes
      double rightWeight = weight/count * rightCount;
      double leftWeight  = weight/count * leftCount;

      // bei nur 2 Teil-Panes wird nicht mehr weiter gesplitten
      // --> Blatt erzeugen
      if ( leftCount == 1 && rightCount == 1 )
        initLeaf(splitPane,weight);

      // nur das linke Teil-Pane wird weitergesplittet
      // --> Halb-Blatt erzeugen
      if ( leftCount > 1 && rightCount == 1)
        initHalfLeaf(splitPane,weight,new SplitPaneTree(container,start,leftCount,leftWeight));

      // beider Teil-Panes werden weitergesplittet
      // --> Inneren Knoten erzeugen
      if ( leftCount > 1 && rightCount > 1)
        initInner(splitPane,weight, new SplitPaneTree(container,start,leftCount,leftWeight),
                                    new SplitPaneTree(container,start+leftCount,rightCount,rightWeight) );

      // Das Split-Pane initialisieren, fuer das dieser Knoten steht
//  System.out.println(start+"\t"+count+"\t"+isSplittedLeft()+"\t"+isSplittedRight());
      if ( this.isSplittedLeft() )
        this.splitPane.setLeftComponent( this.getLeftTree().getSplitPane() );
      else
        this.splitPane.setLeftComponent( container[ start ] );

      if ( this.isSplittedRight() )
        this.splitPane.setRightComponent( this.getRightTree().getSplitPane() );
      else
        this.splitPane.setRightComponent( container[ start+count-1 ] );

      this.splitPane.setResizeWeight(leftWeight/weight);
    }

    /**
     * Erzeugt einen neuen Baum.
     * @param container Container, die das SplitPane haben soll
     * @param start     Index im <code>Container</code>-Array, bei dem
     *                  begonnen wird (alle davor werden ignoriert)
     * @param weight    Gewicht des Teilbaums bezogen auf den gesamten Baum
     *                  (Anteil am neuen Platz, den die Container im Teilbaum
     *                  erhalten, wenn sich das Pane vergroessert)
     */
    public SplitPaneTree(Container[] container, int start, double weight) {
      this(container,start,container.length - start,weight);
    }

    /**
     * Erzeugt einen neuen Baum.
     * @param container Container, die das SplitPane haben soll
     */
    public SplitPaneTree(Container[] container) {
      this(container,0,1.0);
    }

    /**
     * Liefert das <code>JSplitPane</code> fuer das der Knoten steht.
     */
    public JSplitPane getSplitPane() {
      return splitPane;
    }

    /**
     * Liefert die Anzahl an Containern, die im Teil-Baum vorhanden sind.
     */
    public int getContainerCount() {
      return contCount;
    }

    /**
     * Checkt, ob das linke (obere) Teil-Pane wiederum gesplittet ist
     * (<code>true</code>) oder einen Container enthaelt (<code>false</code>).
     */
    public boolean isSplittedLeft() {
      return getLeftTree() != null;
    }

    /**
     * Liefert den linken Teilbaum des aktuellen Knoten.
     * @return <code>null</code>, falls das linke (obere) Teil-Pane einen
     *         Container enthaelt und nicht weiter gesplittet ist.
     */
    public SplitPaneTree getLeftTree() {
      return leftTree;
    }

    /**
     * Checkt, ob das rechte (untere) Teil-Pane wiederum gesplittet ist
     * (<code>true</code>) oder einen Container enthaelt (<code>false</code>).
     */
    public boolean isSplittedRight() {
      return getRightTree() != null;
    }

    /**
     * Liefert den rechten Teilbaum des aktuellen Knoten.
     * @return <code>null</code>, falls das rechte (untere) Teil-Pane einen
     *         Container enthaelt und nicht weiter gesplittet ist.
     */
    public SplitPaneTree getRightTree() {
      return rightTree;
    }

    /**
     * Aendert den Gesamt-Anteil am neuen Platz, den die Container im (Teil-Baum)
     * erhalten.<br>
     * <b>Diese Methode sollte nur in einem kontrollierten Durchlauf durch
     * alle Teilbaeume verwendet werden, in dem gewaehrleistet ist, dass
     * mit der Aenderung des Gewichts auch die Gewichte der anderen Teilbaeume
     * korrekt abgeaendert werden!! Die macht diese Methode naemlich
     * <u>nicht</u> automatisch!!</b>
     */
    public void setResizeWeight(double weight) {
      this.resizeWeight = weight;
    }

    /**
     * Liefert den Anteil am gesamten neuen Platz, den die Container erhalten,
     * die im (Teil-)Baums liegen.
     */
    public double getResizeWeight() {
      return resizeWeight;
    }

    /**
     * Liefert den Anteil am gesamten neuen Platz, den die Container erhalten,
     * die im linken Teil-Baum (also im linken/oberen Teil-Pane) liegen.
     */
    public double getLeftResizeWeight() {
      if ( isSplittedLeft() ) return getLeftTree().getResizeWeight();
      return getSplitPane().getResizeWeight() * getResizeWeight();
    }

    /**
     * Liefert den Anteil am gesamten neuen Platz, den die Container erhalten,
     * die im rechten Teil-Baum (also im rechten/unteren Teil-Pane) liegen.
     */
    public double getRightResizeWeight() {
      if ( isSplittedRight() ) return getRightTree().getResizeWeight();
      return (1-getSplitPane().getResizeWeight()) * getResizeWeight();
    }

    /**
     * Liefert den Container eines Blattes aus dem (Teil-)Baum. Dabei sind die
     * Container (mit 0 beginnend) von links nach recht im Baum durchnummeriert.
     * @param i Nummer des Containers
     * @exception lava.lang.ArrayIndexOutOfBoundsException falls in dem Teilbaum
     *            nicht genug Container vorhanden sind
     */
    public Container getContainer(int i) {
      if ( i>= getContainerCount() )
        throw new ArrayIndexOutOfBoundsException("not enough Containers in SplitPaneTree");

      int leftCount = 0;
      if ( isSplittedLeft() ) leftCount = getLeftTree().getContainerCount();

      int rightCount = 0;
      if ( isSplittedRight() ) rightCount = getRightTree().getContainerCount();

      // erster Container gesucht und kein linker Teilbaum
      // --> also Container im linken Teil-Pane ist der gesuchte
      if ( i == 0 && leftCount == 0 )
        return (Container)splitPane.getLeftComponent();
      // letzter Container gesucht und kein rechter Teilbaum
      // --> also Container im rechten Teil-Pane ist der gesuchte
      if ( i == getContainerCount()-1 && rightCount == 0 )
        return (Container)splitPane.getRightComponent();
      // gesuchter Container liegt im linken Teil-Pane
      if ( i < leftCount )
        return getLeftTree().getContainer(i);
      // gesuchter Container liegt im rechten Teil-Pane
      return getRightTree().getContainer(i-leftCount);
    }

    /**
     * Liefert den Container eines Blattes aus dem (Teil-)Baum. Dabei sind die
     * Container (mit 0 beginnend) von links nach recht im Baum durchnummeriert.
     * @param i Nummer des Containers
     * @exception lava.lang.ArrayIndexOutOfBoundsException falls in dem Teilbaum
     *            nicht genug Container vorhanden sind
     */
    public void setContainer(int i, Container cont) {
      if ( i>= getContainerCount() )
        throw new ArrayIndexOutOfBoundsException("not enough Containers in SplitPaneTree");

      int leftCount = 0;
      if ( isSplittedLeft() ) leftCount = getLeftTree().getContainerCount();

      int rightCount = 0;
      if ( isSplittedRight() ) rightCount = getRightTree().getContainerCount();

      // erster Container gesucht und kein linker Teilbaum
      // --> also Container im linken Teil-Pane ist der gesuchte
      if ( i == 0 && leftCount == 0 )
        splitPane.setLeftComponent(cont);
      else
        // letzter Container gesucht und kein rechter Teilbaum
        // --> also Container im rechten Teil-Pane ist der gesuchte
        if ( i == getContainerCount()-1 && rightCount == 0 )
          splitPane.setRightComponent(cont);
        else
          // gesuchter Container liegt im linken Teil-Pane
          if ( i < leftCount )
            getLeftTree().setContainer(i,cont);
          else
            // gesuchter Container liegt im rechten Teil-Pane
            getRightTree().setContainer(i-leftCount,cont);
    }

    /**
     * Liefert das <code>JSplitPane</code> in dessen linker oder rechter
     * (bzw. oberer oder unterer) Komponente der i-te Container gespeichert
     * ist (Zaehlung beginnt bei 0!!).
     * @exception java.lang.ArrayIndexOutOfBoundsException falls in dem
     *            (Teil-)Baum weniger als i+1 Container gespeichert sind
     */
    public SplitPaneTree getSplitPaneTree(int i) {
      if ( i>=getContainerCount() )
        throw new ArrayIndexOutOfBoundsException("not enough Containers in SplitPaneTree");

      int leftCount = 0;
      if ( isSplittedLeft() ) leftCount = getLeftTree().getContainerCount();

      int rightCount = 0;
      if ( isSplittedRight() ) rightCount = getRightTree().getContainerCount();

      // erster Container gesucht und kein linker Teilbaum ODER
      // letzter Container gesucht und kein rechter Teilbaum
      // --> gesuchter Container liegt im aktuellen SplitPane
      if ( i == 0                     && leftCount == 0 ||
           i == getContainerCount()-1 && rightCount == 0)
        return this;

      // gesuchter Container liegt im linken Teil-Pane
      if ( i < leftCount )  return getLeftTree().getSplitPaneTree(i);

      // gesuchter Container liegt im rechten Teil-Pane
      return getRightTree().getSplitPaneTree(i-leftCount);
    }

    /**
     * Aendert rekursiv alle SplitPanes (in einem Post-Order-Durchlauf).
     * Was an den SplitPanes geaendert wird, bestimmt der <code>SplitPaneTreePerformer</code>.
     * @param performer bestimmt die Verarbeitung der SplitPanes
     */
    public void performAllSplitPanes(SplitPaneTreePerformer performer) {
      if ( isSplittedLeft() )
        getLeftTree().performAllSplitPanes( performer );
      if ( isSplittedRight() )
        getRightTree().performAllSplitPanes( performer );
      // Orientierung des SplitPanes setzen
      performer.performSplitPaneTree(this);
    }

  }
}


