/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing;

import java.awt.Component;

import javax.swing.JDialog;
import javax.swing.JOptionPane;

/**
 * Diese Klasse erweitert {@link JOptionPane} um einen Ok/Abbrechen-Dialog, der
 * mehrfache Werte gleichzeitig abfragt. Die Art der Eingaben wird ueber
 * Instanzen der  Klasse {@link InputOption} spezifiziert.<br>
 * Ueber die statische Funktion
 * {@link #showMultipleInputDialog(Component,String,InputOption[])} kann ein
 * solcher Dialog angezeigt werden.<br>
 * Darueber hinaus bietet die statische Funktion
 * {@link #showClassAndDescInputDialog(Component,String,Class[],String)} einen
 * speziellen Dialog, der eine Auswahl an Klassennamen anzeigt und zu der
 * ausgewaehlten Klasse eine Beschreibung abfragt.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class MultipleOptionPane extends JOptionPane {
  /** Speichert die Eingabe-Optionen des MultipleOptionPane. */
  protected InputOption[] multipleOption = null;

  /**
   * Erzeugt einen neuen Ok/Abbrechen-Dialog mit mehrfachen Eingabefeldern.
   * @param multipleOption Eingabe-Optionen
   */
  public MultipleOptionPane(InputOption[] multipleOption) {
    super();
    this.multipleOption = multipleOption.clone();
    this.setMessageType(JOptionPane.INFORMATION_MESSAGE);
    this.setOptionType(JOptionPane.OK_CANCEL_OPTION);
    this.setMessage( this.multipleOption );
  }

  /**
   * Liefert die Eingabe-Optionen des <code>MultipleOptionPane</code>.
   */
  public InputOption[] getMultipleOptions() {
    return multipleOption.clone();
  }

  /**
   * Prueft, ob die Eingaben in allen Felden zulaessig sind.
   */
  public boolean multipleOptionsValid() {
    return getInvalidOption() == null;
  }

  /**
   * Prueft, ob alle Eingaben zulaessig sind und gibt die erste fehlerhafte
   * Option zurueck.
   * @return <code>null</code> wenn alle Eingaben korrekt sind
   */
  private InputOption getInvalidOption() {
    for (int i=0; i<multipleOption.length; i++)
      if ( !multipleOption[i].isInputValid() )
        return multipleOption[i];
    return null;
  }

  /**
   * Liefert die Eingaben, die in dem <code>MultipleOptionPane</code>
   * gemacht wurden.
   * @return <code>null</code> falls der Dialog noch aktiv ist oder
   *         nicht ueber die Ok-Option verlassen wurde
   */
  public Object[] getMultipleOptionValues() {
    if ( getValue() == null || (Integer)getValue() != JOptionPane.OK_OPTION)
      return null;
    Object[] value = new Object[ multipleOption.length ];
    for (int i=0; i<value.length; i++)
      value[i] = multipleOption[i].getValue();
    return value;
  }

  /**
   * Zeigt einen modalen Ok/Abbrechen-Dialog mit mehrfachen Eingabefeldern an.
   * @param parent Uebergeordnetes Fenster, in dem der Dialog angezeigt wird
   *               (kann <code>null</code> sein)
   * @param title  Titel fuer den Dialog
   * @param option Eingabefelder
   * @return die Werte, die in den einzelnen Feldern gemacht wurden, oder
   *         <code>null</code>, falls der Dialog nicht ueber die Ok-Option
   *         verlassen wurde
   */
  public static Object[] showMultipleInputDialog(Component parent, String title, InputOption... option) {
    MultipleOptionPane pane   = new MultipleOptionPane(option);
    JDialog            dialog = pane.createDialog(parent,title);
    InputOption        invalidOption = null;
    do {
      dialog.setVisible(true);
      boolean cancel = pane.getValue() == null || ((Integer)pane.getValue() != JOptionPane.OK_OPTION);
      invalidOption = !cancel ? pane.getInvalidOption() : null;

      if ( invalidOption != null ) {
        String invalidOptionName = invalidOption.getLabel();
        String errorDesc         = invalidOption.getInvalidInputMessage();
        String mess = SwingUtil.RESOURCE.getString("InvalidInputMess");
        if ( invalidOptionName != null && !invalidOptionName.trim().equals("") )
          mess = mess.concat(": '").concat(invalidOptionName).concat("'");
        if ( errorDesc != null && !errorDesc.trim().equals("") )
          mess = mess.concat("\n").concat(errorDesc);
        showMessageDialog(pane,mess,SwingUtil.RESOURCE.getString("Error"),JOptionPane.ERROR_MESSAGE);
      }

    } while ( invalidOption != null );
    return pane.getMultipleOptionValues();
  }

  /**
   * Zeigt einen modalen Ok/Abbrechen-Dialog mit einem Klassen-Auswahlfeld und
   * einem Text-Eingabefeld fuer eine Beschreibung an. Eine Eingabe in beiden
   * Feldern ist zwingend erforderlich.
   * @param parent      Uebergeordnetes Fenster, in dem der Dialog angezeigt wird
   *                    (kann <code>null</code> sein)
   * @param title       Titel fuer den Dialog
   * @param classOption Klassen, die im Dialog zur Auswahl gestellt werden
   * @param defaultText Default-Wert fuer das Text-Eingabefeld
   * @return Array mit 2 Werten (0 = ausgewaehlte Klasse [<code>Class</code>];
   *         1 = eingegebene Beschreibung [<code>String</code>]), oder
   *         <code>null</code>, falls der Dialog nicht ueber die Ok-Option
   *         verlassen wurde
   */
  public static Object[] showClassAndDescInputDialog(Component parent, String title, Class[] classOption, String defaultText) {
    String[] classDesc = new String[classOption.length];
    for (int i=0; i<classDesc.length; i++)
      classDesc[i] = classOption[i].getName();

    return showMultipleInputDialog(
      parent,
      title,
        new InputOption[] { new SelectionInputOption.Combo(SwingUtil.RESOURCE.getString("Class")+":",true,classOption,classDesc),
                            new ManualInputOption.Text(SwingUtil.RESOURCE.getString("Description")+":",true,defaultText) }
    );
  }


  /**
   * Zeigt einen Dialog mit mehreren Options-Buttons dar.
   * @param parent       Parent-Komponente, ueber der der Dialog angezeigt wird
   * @param type         Art der Meldung ({@link JOptionPane#ERROR_MESSAGE},
   *                     {@link JOptionPane#INFORMATION_MESSAGE} oder {@link JOptionPane#QUESTION_MESSAGE})
   * @param title        Titel des Dialog-Fensters
   * @param message      Meldung die angezeigt wird
   * @param options      Bezeichnungen der Button, die zur Auswahl stehen
   * @param initialValue Bezeichnung des standardmaessig ausgewaehlten Button
   * @return Die Bezeichnung ({@code String}) des Buttons, der ausgewaehlt wurde,
   *         oder {@code null}, wenn der Dialog ueber das System-Menu abgebrochen
   *         wurde
   */
  public static Object showMultipleOptionDialog(Component parent, int type, String title, Object message, String[] options, String initialValue) {
    JOptionPane optionPane = new JOptionPane();
    optionPane.setMessageType(type);
    optionPane.setOptionType(JOptionPane.OK_CANCEL_OPTION);
    optionPane.setMessage( message );
    optionPane.setOptions( options );
    optionPane.setInitialValue(initialValue);
    JDialog dialog = optionPane.createDialog(parent,title);
    dialog.setVisible(true);
    return optionPane.getValue();
  }
}
