/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing;

import java.awt.Component;
import java.awt.Container;

/**
 * Diese Klasse repraesentiert eine abstrakte Oberklasse fuer alle GUI-Komponenten,
 * die ein (allgemeines) Datenobjekt darstellen.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class ObjectDisplayContainer extends Container {
  /** Speichert eine Referent auf den Container (<code>this</code>). */
  protected final Component GUI_COMPONENT  = this;
  /** Speichert das dargestellte Objekt. */
  protected Object data = null;

  /**
   * Erzeugt eine neue Darstellungskomponente.
   */
  public ObjectDisplayContainer() {
    this(null);
  }

  /**
   * Erzeugt eine neue Darstellungskomponente und zeigt sofort ein
   * Objekt an.
   * @param data anzuzeigendes Objekt
   */
  public ObjectDisplayContainer(Object data) {
    setObject(data);
  }

  /**
   * Setzt das darzustellende Objekt. Wird <code>null</code> uebergeben,
   * wird das bisher dargestellte Objekt entfernt.
   * @exception UnsupportedOperationException falls das angegebene Objekt
   *            nicht dargestellt werden kann.
   * @see #canDisplay(Object)
   */
  public void setObject(Object data) {
    if ( data!=null && !canDisplay(data) )
      throw new UnsupportedOperationException( getClass().getSimpleName().concat(" can not display an instance of ").concat(data.getClass().getName()) );
    this.data = data;
    refresh();
  }

  /**
   * Liefert das aktuell dargestellte Objekt.
   * @return <code>null</code> falls aktuell kein Objekt dargestellt wird
   */
  public Object getObject() {
    return data;
  }

  /**
   * Prueft, ob ein Objekt darstellbar ist.
   */
  public abstract boolean canDisplay(Object data);

  /**
   * Aktualisiert die Darstellung des aktuell angezeigten Objekts.
   */
  public abstract void refresh();
}
