/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing;

import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.HashMap;
import java.util.Map;
import java.util.Vector;

import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JTextField;

import schmitzm.lang.tree.OperationTree;
import schmitzm.lang.tree.OperationTreeParser;

/**
 * Diese Klasse stellt eine Panel zur Vefuegung, mit der eine einfache arithmetische
 * (und boolsche) Berechnungsformel eingegeben werden kann.
 * @see OperationTree
 * @see OperationTreeParser
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class OperationTreePanel extends CaptionsChangeablePanel {
  /** Parser, der aus dem Formel-String einen Operator-Baum erstellt. */
  private static final OperationTreeParser parser = new OperationTreeParser();

  /** Speichert eine Referenz auf das Fenster-Objekt, um aus verschachtelten
   *  Klassen einfach darauf zugreifen zu koennen. */
  protected final OperationTreePanel THIS = this;

  /** Konstante fuer den Start-Button.
  *  @see #layoutConstraints
  *  @see #resetCaptions(Map) */
  public static final String START_BUTTON = OperationTreePanel.class.getName()+".StartButton";
  /** Konstante fuer das Regel-Label.
  *  @see #resetCaptions(Map) */
  public static final String RULE_LABEL = OperationTreePanel.class.getName()+".RuleLabel";
  /** Konstante fuer das Formel-Eingabefeld.
  *  @see #layoutConstraints */
  public static final String RULE_TEXTFIELD = OperationTreePanel.class.getName()+".RuleField";
  /** Konstante fuer den Tooltip des Regel-Felds.
  *  @see #resetCaptions(Map) */
  public static final String RULE_TOOLTIP = OperationTreePanel.class.getName()+".RuleTooltip";
  /** Konstante fuer das Operatoren-Label.
  *  @see #resetCaptions(Map) */
  public static final String OPERATOR_LABEL = OperationTreePanel.class.getName()+".OperatorLabel";
  /** Konstante fuer das Operatoren-Auswahlfeld.
  *  @see #layoutConstraints */
  public static final String OPERATOR_COMBOBOX = OperationTreePanel.class.getName()+".OperatorBox";
  /** Werte fuer die grafische Anordnung der Layout-Komponenten.
   *  Ueber die Konstanten {@link #START_BUTTON}, {@link #RULE_TEXTFIELD}
   *  und {@link #OPERATOR_COMBOBOX} koennen die Constraints (am Anfang von {@link #initGUI()})
   *  veraendert oder erweitert werden. */
  protected HashMap<String,GridBagConstraints> layoutConstraints = new HashMap<String,GridBagConstraints>();

  /** Eingabefeld fuer die Formel. */
  protected ManualInputOption.Text rule = null;
  /** Butten zum Starten der Berechnung. */
  protected JButton startButton = null;
  /** Auswahl-Liste der zur Verfuegung stehenden Operatoren und Konstanzen. */
  protected SelectionInputOption.Combo<String> operators = null;
  /** Liefert Informationen ueber die Operatoren, die dem Nutzer in der
   *  ComboBox zur Verfuegung gestellt werden. */
  protected ParserOperatorsHints opHints = null;
  /** Enthaelt die in der Operator-Auswahl zur Verfuegung gestellten Operatoren
   *  Konstanten */
  private Vector<String> avOperators     = new Vector<String>();
  /** Enthaelt die in der Operator-Auswahl angezeigen Operatoren Konstanten */
  private HashMap<String,String> avOperatorsTitle = new HashMap<String,String>();
  /** Enthaelt die in der Operator-Auswahl angezeigen Tooltips der Operatoren. */
  private HashMap<Object,String> avOperatorsToolTip = new HashMap<Object,String>();
 
  private JTextField ruleTextField = null;
  private JComboBox  operatorsComboBox = null;

  /**
   * Erzeugt eine neue Eingabe-GUI fuer den {@link OperationTree}.
   * @param opHints Liefert Informationen ueber die Operatoren, die dem Nutzer in der
   *                ComboBox zur Verfuegung gestellt werden.
   */
  public OperationTreePanel(ParserOperatorsHints opHints) {
    this(opHints,true);
  }

  /**
   * Erzeugt eine neue Eingabe-GUI fuer den {@link OperationTree}.
   * @param opHints Liefert Informationen ueber die Operatoren, die dem Nutzer in der
   *                ComboBox zur Verfuegung gestellt werden.
   * @param initGUI Flag, ob {@link #initGUI()} am Ende des Konstruktor
   *        aufgerufen werden soll (wenn {@code false} muss die explizit
   *        durch die Unterklasse erfolgen!)
   */
  protected OperationTreePanel(ParserOperatorsHints opHints, boolean initGUI) {
    super();
    this.opHints = opHints;
    this.setPreferredSize( new Dimension(500,400) );
    this.setLayout( new GridBagLayout() );

    // Standard-Werte fuer die Layout-Anordnung hinterlegen
    layoutConstraints.put(RULE_TEXTFIELD,    new GridBagConstraints(0,0,1,1,1.0,0.0,GridBagConstraints.WEST, GridBagConstraints.HORIZONTAL, new Insets(0,10,5,10),0,0));
    layoutConstraints.put(OPERATOR_COMBOBOX, new GridBagConstraints(1,0,1,1,0.0,0.0,GridBagConstraints.SOUTHEAST, GridBagConstraints.NONE, new Insets(0,10,5,10),0,0));
    layoutConstraints.put(START_BUTTON,      new GridBagConstraints(0,1,1,1,0.0,0.0,GridBagConstraints.WEST, GridBagConstraints.NONE, new Insets(0,10,5,10),0,0));

    
    // Operatoren filtern
    if ( this.opHints != null )
      for (String op : opHints.getOperators() )
        if ( acceptOperator(op) ) {
          avOperators.add(op);
          avOperatorsTitle.put(op, opHints.getOperatorTitle(op));
          avOperatorsToolTip.put(opHints.getOperatorTitle(op), opHints.getOperatorDescription(op));
        }
    
    if (initGUI)
      initGUI();
  }

  /**
   * Initalisiert die GUI des Fensters.
   */
  protected void initGUI() {
    // Formel
    rule = new ManualInputOption.Text(SwingUtil.RESOURCE.getString("Rule")+":",true,"");
    rule.setToolTipText(SwingUtil.RESOURCE.getString("RuleToolTip"));
    ruleTextField = (JTextField)rule.getInputComponent();
    // neue Caret-Implementierung, damit Selektion nicht verloren geht,
    // wenn Textfeld den Focus verliert
    ruleTextField.setCaret( new SelectionPreservingCaret(true) );

    // Button zum Starten
    startButton = new JButton(SwingUtil.RESOURCE.getString("Calculate"));
    startButton.addActionListener( new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        startCalculation();
      }
    });

    // Combo-Box fuer Operatoren
    String[] opArr  = (String[])avOperators.toArray(new String[0]);
    String[] opDesc = new String[opArr.length];
    for (int i=0; i<opArr.length; i++) {
      String desc = avOperatorsTitle.get(opArr[i]);
      opDesc[i] = (desc == null) ? opArr[i] : desc;
    }
    operators = new SelectionInputOption.Combo<String>(
      SwingUtil.RESOURCE.getString("Operators")+":",  // Label
      true,     // keine Null-Auswahl
      opArr,    // Operatoren
      null,     // Default-Wert
      opDesc,   // Angezeigte Operatoren
      avOperatorsToolTip // Angezeigte Tooltips fuer die Operatoren
    );
    SwingUtil.setPreferredWidth(operators, 150);
    operatorsComboBox = ((JComboBox)operators.getInputComponent());
    operatorsComboBox.addActionListener( new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        performOperatorInsert(
            (String)operators.getValue(),
            ruleTextField
        );
        // Operator-Auswahl auf "nichts" zuruecksetzen
        operators.setValue( null );
        // Fokus zurueck auf Formel-Feld
        rule.grabFocus();
      }
    } );

    // Komponenten in Fenster anordnen
    add( rule,        layoutConstraints.get(RULE_TEXTFIELD) );
    add( startButton, layoutConstraints.get(START_BUTTON) );
    add( operators,   layoutConstraints.get(OPERATOR_COMBOBOX) );
  }
  
  /**
   * Kann von Sub-Klassen ueberschrieben werden, um bestimmte Operatoren
   * in der Auswahl-Liste auszublenden.
   * @param operator ein Operator
   * @return immer {@code true}
   */
  protected boolean acceptOperator(String operator) {
    return true;
  }

  /**
   * Setzt die Labels des Panels neu. Die Beschriftungen werden in einer
   * Map hinterlegt und ueber die Konstanten {@link #RULE_LABEL}, {@link #RULE_TOOLTIP},
   * {@link #OPERATOR_LABEL} und {@link #START_BUTTON} angesprochen.
   * @param captionMap Map
   */
  public void resetCaptions(Map<String,Object> captionMap) {
    super.resetCaptions(captionMap);
    SwingUtil.resetCaption( rule.getDescriptionLabel(), captionMap.get( RULE_LABEL) );
    SwingUtil.resetCaption( operators.getDescriptionLabel(), captionMap.get( OPERATOR_LABEL ) );
    SwingUtil.resetCaption( startButton, captionMap.get( START_BUTTON ) );
    Object caption = captionMap.get( RULE_TOOLTIP );
    if ( caption != null )
      rule.setToolTipText( caption.toString() );
  }

  /**
   * Prueft die in der Maske spezifizierten Angaben auf Korrektheit. Ist eine
   * Angabe nicht korrekt, wird ein entsprechender Fehler-Dialog angezeigt.
   */
  private boolean checkInputs() {
    try {
      checkInputsAndError();
    } catch (Exception err) {
      ExceptionDialog.show(
          SwingUtil.getParentWindowComponent(this),
          err,
          (err!=null) ? err.getClass().getSimpleName() : SwingUtil.RESOURCE.getString("Error"),
          null
      );
      return false;
    }
    return true;
  }


  /**
   * Prueft die eingegebene Formel auf syntaktische Korrektheit. Unterklassen
   * koennen diese Methode ueberschreiben, wenn die Mask
   * @exception Exception falls ein Fehler gefunden wird
   */
  protected void checkInputsAndError() throws Exception {
    parser.parse(rule.getValue());
  }

  /**
   * Erzeugt einen {@link OperationTree} aus der Formel und wertet diese
   * aus.
   * @return das Ergebnis der Formel, als {@code double}-Wert
   * @see OperationTreeParser
   */
  protected Object performCalculation() throws Exception {
    return parser.parse( rule.getValue() ).evaluate();
  }

  /**
   * Startet die Berechnung, wenn der Start-Button betaetigt wurde.
   */
  private void startCalculation() {
    // Eingaben pruefen
    if ( !checkInputs() )
      return;

    // SwingWorker erzeugen, der die Generierung durchfuehrt
    SwingWorker worker = new SwingWorker(
      new SwingWorker.Work() {
        public Object execute() throws Exception {
          performCalculation();
          return null;
        }
        public void performError(Exception err) {
          // Fehler bei der Code-Generierung verarbeiten
          ExceptionDialog.show(
              SwingUtil.getParentWindowComponent(THIS),
              err,
              (err!=null) ? err.getClass().getSimpleName() : SwingUtil.RESOURCE.getString("Error"),
              null
          );
        }
      },
      SwingUtil.getParentFrame(this),
      SwingUtil.RESOURCE.getString("WaitMess")
    );

    // Berechnung starten
    worker.start();
  }

  /**
   * Fuegt einen Operator in die aktuelle Formel ein. Ueber
   * {@link JTextField#getSelectionStart()} und {@link JTextField#getSelectionEnd()}
   * kann ermittelt werden, an welcher Stelle der aktuelle Cursor steht.
   * @param op   einzufuegender Operator
   * @param textField Text-Feld in das der Operator eingefuegt werden soll
   * @see JTextField#getSelectedText()
   */
  protected void performOperatorInsert(String op, JTextField textField) {
    int selStart     = textField.getSelectionStart();
    int selEnd       = textField.getSelectionEnd();
    String selection = textField.getSelectedText();

    // Anzahl an geklammerten Parametern fuer den Operator ermitteln
    int parameterCount = opHints.getOperatorParameterCount(op);

    // Bei mehr-stelligen Funktionen, erhaelt die Funktion den selektierten
    // Bereich als ersten Operand
    if ( parameterCount >= 1 ) {
      // wenn nichts selektiert war, wird "_" fuer den Operanden
      // vorgeblendet
      if ( selection == null )
        selection = "_";
      // Operator um die Parameter erweitern
      op = op + "( " + selection;
      for (int i=1; i<parameterCount; i++)
        op = op + ", _";
      op = op + " )";
    }

    // Selektierten Bereich ersetzen (wenn nichts selektiert war, wird autom.
    // nur eingefuegt)
    textField.replaceSelection( op );

    // Wenn vorher ein Bereich ausgewaehlt war, wird nun der
    // neue Operator ausgewaehlt
    if ( selStart < selEnd ) {
      textField.setSelectionStart( selStart );
      textField.setSelectionEnd( selStart + op.length() );
    }
  }
}
