/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing;

import javax.swing.SpinnerNumberModel;

/**
 * Diese Klasse erweitert das {@link RotationSpinnerNumberModel} um eine
 * Rotation-Funktionalitaet. Ist diese eingeschaltet, wechselt der Spinner
 * automatisch auf den kleinsten Wert, wenn der Maximum-Wert ueberschritten
 * wird (und umgekehrt).
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class RotationSpinnerNumberModel extends SpinnerNumberModel {
  /** Flag fuer die Rotation-Funktionalitaet. */
  protected boolean rotation = false;

  /**
   * Erzeugt ein neues Model.
   * @param value    initialer Wert
   * @param minimum  Minimal-Wert
   * @param maximum  Maximal-Wert
   * @param stepSize Schrittweite
   * @param rotation falls {@code true} wechselt der Spinner auf den kleinsten Wert
   *                 wenn der Maximalwert ueberschritten wird (und umgekehrt)
   */
  public RotationSpinnerNumberModel(Number value, Comparable minimum, Comparable maximum, Number stepSize, boolean rotation) {
    super(value, minimum, maximum, stepSize);
    setRotation(rotation);
  }

  /**
   * Erzeugt ein neues Model.
   * @param value    initialer Wert
   * @param minimum  Minimal-Wert
   * @param maximum  Maximal-Wert
   * @param stepSize Schrittweite
   * @param rotation falls {@code true} wechselt der Spinner auf den kleinsten Wert
   *                 wenn der Maximalwert ueberschritten wird (und umgekehrt)
   */
  public RotationSpinnerNumberModel(int value, int minimum, int maximum, int stepSize, boolean rotation) {
    super(value, minimum, maximum, stepSize);
    setRotation(rotation);
  }

  /**
   * Erzeugt ein neues Model.
   * @param value    initialer Wert
   * @param minimum  Minimal-Wert
   * @param maximum  Maximal-Wert
   * @param stepSize Schrittweite
   * @param rotation falls {@code true} wechselt der Spinner auf den kleinsten Wert
   *                 wenn der Maximalwert ueberschritten wird (und umgekehrt)
   */
  public RotationSpinnerNumberModel(double value, double minimum, double maximum, double stepSize, boolean rotation) {
    super(value, minimum, maximum, stepSize);
    setRotation(rotation);
  }

  /**
   * Erzeugt ein neues Model ohne Minimum- und Maximumwert.
   */
  public RotationSpinnerNumberModel() {
    super();
    setRotation( false );
  }

  /**
   * Liefert den naechst-groesseren Wert.
   * @return {@link #getMinimum()} wenn der Maximalwert erreicht ist und
   *         die Rotation-Funktion eingestellt ist
   * @see #setRotation(boolean)
   */
  public Object getNextValue() {
    Object next = super.getNextValue();
    if ( next == null && rotation )
      next = getMinimum();
    return next;
  }

  /**
   * Liefert den naechst-kleineren Wert.
   * @return {@link #getMaximum()} wenn der Minimalwert erreicht ist und
   *         die Rotation-Funktion eingestellt ist
   * @see #setRotation(boolean)
   */
  public Object getPreviousValue() {
    Object next = super.getPreviousValue();
    if ( next == null && rotation )
      next = getMaximum();
    return next;
  }

  /**
   * (De)aktiviert die Rotation-Funktion.
   * @param rotation wenn {@code true} wird, die Rotation-Funktion aktiviert
   */
  public void setRotation(boolean rotation) {
    this.rotation = rotation;
  }

  /**
   * Prueft, ob die Rotation-Funktion aktiviert ist.
   */
  public boolean getRotation() {
    return this.rotation;
  }
}
