/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing;

import java.util.Vector;

import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.table.TableColumnModel;
import javax.swing.table.TableModel;

import schmitzm.swing.table.SelectionTableModel;

/**
 * Extends the {@link JTable} by redefining the {@link #changeSelection(int, int, boolean, boolean)}
 * method. If a {@link SelectionTableModel} is used, a click on the selection column must not
 * clear an existing selection.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 */
public class SelectableJTable extends JTable {
  /**
   * Creates an empty table.
   */
  public SelectableJTable() {
    super();
    initTable();
  }

  /**
   * Creates an empty table.
   */
  public SelectableJTable(int numRows, int numColumns) {
    super(numRows, numColumns);
    initTable();
  }

  /**
   * Creates a new table.
   */
  public SelectableJTable(Object[][] rowData, Object[] columnNames) {
    super(rowData, columnNames);
    initTable();
  }

  /**
   * Creates a new table.
   */
  public SelectableJTable(TableModel dm, TableColumnModel cm, ListSelectionModel sm) {
    super(dm, cm, sm);
    initTable();
  }

  /**
   * Creates a new table.
   */
  public SelectableJTable(TableModel dm, TableColumnModel cm) {
    super(dm, cm);
    initTable();
  }

  /**
   * Creates a new table.
   */
  public SelectableJTable(TableModel dm) {
    super(dm);
    initTable();
  }

  /**
   * Creates a new table.
   */
  public SelectableJTable(Vector rowData, Vector columnNames) {
    super(rowData, columnNames);
    initTable();
  }

  /**
   * Called by every constructor to initialize the extended
   * functionalities. Currently does nothing.
   */
  protected void initTable() {
  }

  /**
   * If a {@link SelectionTableModel} is used, a click on the selection column must not
   * clear an existing selection. So if ...
   * <ul>
   *   <li>{@link #getModel()} is an {@link SelectionTableModel}
   *   <li>{@code toggle} flag is {@code false}</li>
   *   <li>{@code extend} flag is {@code false}</li>
   *   <li>{@link #convertColumnIndexToModel(int) convertColumnIndexToModel(col)} is 0
   * </ul>
   * ... this method sets {@code toggle} to {@code true} (before calling the super-method)
   * to simulate that the Strg/Ctrl-key is pressed.
   */
  @Override
  public void changeSelection(int row, int col, boolean toggle, boolean extend) {
    if ( getModel() instanceof SelectionTableModel &&
         convertColumnIndexToModel(col) == 0 &&
         !toggle &&
         !extend )
      toggle = true;
    super.changeSelection(row, col, toggle, extend);
  }
}
