/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.HashMap;
import java.util.Map;
import java.util.Vector;

import javax.swing.BoxLayout;
import javax.swing.Icon;
import javax.swing.JComboBox;
import javax.swing.JRadioButton;

/**
 * Diese Klasse stellt eine Auswahl-Eingabe-Option fuer das {@link MultipleOptionPane}
 * dar. Die Klasse kann nicht direkt instanziiert werden.
 * Statt dessen sind die eingebetteten Klassen
 * <ul>
 * <li>{@link SelectionInputOption.Combo SelectionInputOption.Combo}</li>
 * </ul>
 * zu verwenden.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class SelectionInputOption<E> extends InputOption<E> {
  /** Speichert die Objekte, die zur Auswahl stehen. */
  protected E[] selectionObject = null;
  /** Speichert die Objekte, die fuer die Auswahl angezeigt werden. */
  protected Object[] displayObject   = null;
  /** Speichert fuer jedes {@link #selectionObject} einen ToolTip, der
   *  fuer das Auswahl-Objekt angezeigt wird. */
  protected Map<Object,String> displayToolTip  = new HashMap<Object, String>();

  /**
   * Erzeugt eine neue Eingabe-Option.
   * @param label       Beschreibung
   * @param inputNeeded gibt an, ob eine Eingabe erforderlich ist
   * @param value       die zur Auswahl stehenden Objekte
   * @param defIdx      Index der vorgeblendeten Auswahl
   * @param display     die anstelle der Auswahl-Objekte angezeigten Objekte
   *                    (kann <code>null</code> sein)
   * @exception IllegalArgumentException falls sich die Array-Groessen von
   *            <code>value</code> und <code>display</code> unterscheiden
   */
  protected SelectionInputOption(String label, boolean inputNeeded, E[] value, int defIdx, Object[] display) {
    super(label,inputNeeded);
    // Referenz auf Objekte merken und Komponente befuellen
    setSelectionObjects(value,display);
    setSelectedIndex(defIdx);
  }

  /**
   * Erzeugt eine neue Eingabe-Option.
   * @param label       Beschreibung
   * @param inputNeeded gibt an, ob eine Eingabe erforderlich ist
   * @param value       die zur Auswahl stehenden Objekte
   * @param defValue    vorgeblendetes Auswahlobjekt
   * @param display     die anstelle der Auswahl-Objekte angezeigten Objekte
   *                    (kann <code>null</code> sein)
   * @exception IllegalArgumentException falls sich die Array-Groessen von
   *            <code>value</code> und <code>display</code> unterscheiden
   */
  protected SelectionInputOption(String label, boolean inputNeeded, E[] value, E defValue, Object[] display) {
    this(label,inputNeeded,value,defValue,display,null);
  }

  /**
   * Erzeugt eine neue Eingabe-Option.
   * @param label       Beschreibung
   * @param inputNeeded gibt an, ob eine Eingabe erforderlich ist
   * @param value       die zur Auswahl stehenden Objekte
   * @param defValue    vorgeblendetes Auswahlobjekt
   * @param display     die anstelle der Auswahl-Objekte angezeigten Objekte
   *                    (kann <code>null</code> sein)
   * @param tooltip     die fuer jedes Auswahl-Objekt angezeigten Tooltips
   *                    (kann <code>null</code> sein)
   * @exception IllegalArgumentException falls sich die Array-Groessen von
   *            <code>value</code> und <code>display</code> unterscheiden
   */
  protected SelectionInputOption(String label, boolean inputNeeded, E[] value, E defValue, Object[] display, Map<Object,String> tooltip) {
    super(label,inputNeeded);
    // Referenz auf Objekte merken und Komponente befuellen
    setSelectionObjects(value,display,tooltip);
    setSelectedItem(defValue);
  }

  /**
   * Erzeugt eine neue Eingabe-Option. Es wird (sofern vorhanden) das erste
   * Auswahl-Objekt vorgeblendet.
   * @param label       Beschreibung
   * @param inputNeeded gibt an, ob eine Eingabe erforderlich ist
   * @param value       die zur Auswahl stehenden Objekte
   * @param display     die anstelle der Auswahl-Objekte angezeigten Objekte
   *                    (kann <code>null</code> sein)
   * @exception IllegalArgumentException falls sich die Array-Groessen von
   *            <code>value</code> und <code>display</code> unterscheiden
   */
  protected SelectionInputOption(String label, boolean inputNeeded, E[] value, Object[] display) {
    this(label,inputNeeded,value,value.length > 0 ? 0 : -1, display);
  }

  /**
   * Setzt die im Auswahlfeld zur Verfuegung stehenden Eintraege
   * @param value       die zur Auswahl stehenden Objekte
   * @param display     die anstelle der Auswahl-Objekte angezeigten Objekte
   *                    (kann <code>null</code> sein)
   * @exception IllegalArgumentException falls sich die Array-Groessen von
   *            <code>value</code> und <code>display</code> unterscheiden
   *
   */
  public void setSelectionObjects(E[] value, Object[] display) {
    setSelectionObjects(value,display,null);
  }
  
  /**
   * Setzt die im Auswahlfeld zur Verfuegung stehenden Eintraege
   * @param value       die zur Auswahl stehenden Objekte
   * @param display     die anstelle der Auswahl-Objekte angezeigten Objekte
   *                    (kann <code>null</code> sein)
   * @param tooltip     die fuer jedes Auswahl-Objekt angezeigten Tooltips
   *                    (kann <code>null</code> sein)
   * @exception IllegalArgumentException falls sich die Array-Groessen von
   *            <code>value</code> und <code>display</code> unterscheiden
   *
   */
  public void setSelectionObjects(E[] value, Object[] display, Map<Object,String> tooltip) {
    // werden keine Wert angegeben, wird dies wie eine leere
    // Menge behandelt
    if (value == null)
      value = (E[])new Object[0];
    // Wurden keine speziellen Anzeigewerte angegeben, werden die
    // Auswahl-Objekte verwendet
    if (display == null)
      display = value;
    // Haben der Auswahl- und der Anzeige-Array unterschiedliche
    // Laenge, wird dies nicht akzeptiert
    if ( ( (Object[]) display).length != ( (E[]) value).length)
      throw new IllegalArgumentException("display-Array and value-Array must have the same size!");

    Object lastSelection = getValue();
    this.selectionObject = value;
    this.displayObject   = display;
    this.displayToolTip  = tooltip;
    performSelectionUpdate();
    setSelectedItem(lastSelection);
  }

  /**
   * Aktualisiert die Objekte in der der Auswahlkomponente. Wird aufgerufen,
   * wenn dich die zur Auswahl stehenden Objekte aendern. Die Variablen
   * {@link #selectionObject} und {@link #displayObject} sind dann bereits
   * aktualisiert,
   */
  protected abstract void performSelectionUpdate();

  /**
   * Liefert immer <code>true</code>, da keine speziellen Anforderungen
   * an die Auswahl-Eingabe gestellt werden.
   */
  protected boolean performIsInputValid() {
    return true;
  }

  /**
   * Prueft, ob im Auswahlfeld ein Eintrag ausgewaehlt wurde.
   */
  protected boolean performIsInputEmpty() {
    return getSelectedIndex() < 0;
  }

  /**
   * Liefert das Objekt, das zum aktuell ausgewaehlten Eintrag der
   * Auswahl-Liste gehoert.
   * @return <code>null</code> falls kein Objekt ausgewaehlt ist
   */
  protected E performGetValue() {
    int idx = getSelectedIndex();
    return idx < 0 ? null : selectionObject[getSelectedIndex()];
  }

  /**
   * Setzt das Objekt, das in der Auswahlliste angewaehlt wird. Ruft
   * {@link #setSelectedItem(Object)} auf.
   * @return immer <code>true</code>
   */
  protected boolean performSetValue(E newValue) {
    setSelectedItem(newValue);
    return true;
  }

  /**
   * Setzt einen neuen Wert, mit der die Eingabe-Option belegt wird.
   * Ruft lediglich {@link #performSetValue(Object)} auf. Ueberschreibt die
   * Methode von {@link InputOption}, da die Ueberpruefung auf Aenderung
   * (und Event-Ausloesung) in {@link #setSelectedIndex(int)} erfolgt.
   * @param newValue neuer Wert
   * @return <code>false</code> gdw. der Objekt-Typ fuer die Option nicht
   *         zulaessig ist
   */
  public boolean setValue(E newValue) {
    if ( !performSetValue(newValue) )
      return false;
    return true;
  }

  /**
   * Liefert den Index, der in der Objekt-Liste ausgewaehlt wurde.
   * @return <code>-1</code> falls kein Objekt ausgewaehlt wurde
   */
  public abstract int getSelectedIndex();

  /**
   * Setzt den Index, der in der Objekt-Liste ausgewaehlt wurde. <b>Da dies die
   * Basis-Methode fuer alle Aenderungsoperationen darstellt, muss sie
   * die Ueberpruefung auf Aenderung implementieren und ggf.
   * {@link #fireOptionChanged(Object,Object)} ausloesen.</b>
   * @param idx Listen-Index (-1 um eine Leer-Auswahl zu erzeugen)
   */
  public abstract void setSelectedIndex(int idx);

  /**
   * Setzt die Auswahl auf ein bestimmtes Objekt. Ist dieses nicht vorhanden,
   * wird die Auswahlliste auf eine leere Auswahl eingestellt.
   */
  public void setSelectedItem(Object object) {
    // altes Object in der neuen Objektliste suchen
    int newIdx = -1;
    for (int i=0; i<selectionObject.length && newIdx < 0; i++)
//      if ( selectionObject[i] == object )
      if ( selectionObject[i]==null && object==null ||
           selectionObject[i]!=null && selectionObject[i].equals(object) )
        newIdx = i;
    setSelectedIndex(newIdx);
  }

  /**
   * Liefert das Anzeige-Objekt der akuellen Auswahl.
   */
  public Object getSelectedDisplayItem() {
    // altes Object in der neuen Objektliste suchen
    if ( getSelectedIndex() >= 0 )
      return displayObject[ getSelectedIndex() ];
   return null;
  }

  /**
   * Setzt die Auswahl auf ein bestimmtes Anzeige-Objekt. Ist dieses nicht vorhanden,
   * wird die Auswahlliste auf eine leere Auswahl eingestellt.
   */
  public void setSelectedDisplayItem(Object object) {
    // altes Object in der neuen Objektliste suchen
    int newIdx = -1;
    for (int i=0; i<displayObject.length && newIdx < 0; i++)
//      if ( selectionObject[i] == object )
      if ( displayObject[i]==null && object==null ||
           displayObject[i]!=null && displayObject[i].equals(object) )
        newIdx = i;
    setSelectedIndex(newIdx);
  }

  /**
   * Liefert die Anzahl der zur Auswahl stehenden Eintraege.
   */
  public int getSelectedItemCount() {
    return selectionObject != null ? selectionObject.length : 0;
  }

  /**
   * Diese Klasse stellt eine Auswahl-Option dar, die durch eine
   * {@link JComboBox} dargestellt wird.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class Combo<E> extends SelectionInputOption {
    /**
     * Erzeugt eine neue Auswahl-Option.
     * @param label       Beschreibung
     * @param inputNeeded gibt an, ob eine Eingabe erforderlich ist
     * @param value       die zur Auswahl stehenden Objekte
     * @param defIdx      Index der vorgeblendeten Auswahl
     * @param display     die anstelle der Auswahl-Objekte angezeigten Objekte
     *                    (kann <code>null</code> sein)
     * @exception IllegalArgumentException falls sich die Array-Groessen von
     *            <code>value</code> und <code>display</code> unterscheiden
     */
    public Combo(String label, boolean inputNeeded, E[] value, int defIdx, Object[] display) {
      super(label,inputNeeded,value,defIdx, display);
    }

    /**
     * Erzeugt eine neue Auswahl-Option.
     * @param label       Beschreibung
     * @param inputNeeded gibt an, ob eine Eingabe erforderlich ist
     * @param value       die zur Auswahl stehenden Objekte
     * @param defValue    vorgeblendetes Auswahlobjekt
     * @param display     die anstelle der Auswahl-Objekte angezeigten Objekte
     *                    (kann <code>null</code> sein)
     * @exception IllegalArgumentException falls sich die Array-Groessen von
     *            <code>value</code> und <code>display</code> unterscheiden
     */
    public Combo(String label, boolean inputNeeded, E[] value, E defValue, Object[] display) {
      super(label,inputNeeded,value,defValue,display);
    }

    /**
     * Erzeugt eine neue Auswahl-Option.
     * @param label       Beschreibung
     * @param inputNeeded gibt an, ob eine Eingabe erforderlich ist
     * @param value       die zur Auswahl stehenden Objekte
     * @param defValue    vorgeblendetes Auswahlobjekt
     * @param display     die anstelle der Auswahl-Objekte angezeigten Objekte
     *                    (kann <code>null</code> sein)
     * @param tooltip     die fuer jedes Auswahl-Objekt angezeigten Tooltips
     *                    (kann <code>null</code> sein)
     * @exception IllegalArgumentException falls sich die Array-Groessen von
     *            <code>value</code> und <code>display</code> unterscheiden
     */
    public Combo(String label, boolean inputNeeded, E[] value, E defValue, Object[] display, Map<Object,String> tooltip) {
      super(label,inputNeeded,value,defValue,display,tooltip);
    }

    /**
     * Erzeugt eine neue Auswahl-Option. Es wird das erste Auswahl-Objekt
     * vorgeblendet.
     * @param label       Beschreibung
     * @param inputNeeded gibt an, ob eine Eingabe erforderlich ist
     * @param value       die zur Auswahl stehenden Objekte
     * @param display     die anstelle der Auswahl-Objekte angezeigten Objekte
     *                    (kann <code>null</code> sein)
     * @exception IllegalArgumentException falls sich die Array-Groessen von
     *            <code>value</code> und <code>display</code> unterscheiden
     */
    public Combo(String label, boolean inputNeeded, E[] value, Object[] display) {
      super(label,inputNeeded,value,display);
    }

    /**
     * Erzeugt eine leere Auswahl-Option. Diese muss nachtraeglich ueber
     * {@link #setSelectionObjects(Object[],Object[])} befuellt werden.
     * @param label       Beschreibung
     * @param inputNeeded gibt an, ob eine Eingabe erforderlich ist
     */
    public Combo(String label, boolean inputNeeded) {
      super(label,inputNeeded,new Object[0],new Object[0]);
    }

    /**
     * Erzeugt eine neue Instanz von {@link JComboBox}.
     */
    protected JComboBox createInputComponent() {
      JComboBox comboBox = new JComboBox();
      // Change Listener
      comboBox.addItemListener( new ItemListener() {
        Object oldValue = null;
        public void itemStateChanged(ItemEvent e) {
          if ( e.getStateChange() == e.SELECTED ) {
            Object newValue = getValue();
            if ( oldValue != newValue )
              fireOptionChanged(oldValue,newValue);
            oldValue = newValue;
          }
        }
      });
      return comboBox;
    }

    /**
     * Befuellt die ComboBox-Liste neu.
     */
    protected void performSelectionUpdate() {
      // wenn eine Leereingabe erlaubt ist, wird am Anfang ein
      // null-Eintrag eingefuegt
      if ( !inputNeeded() ) {
        Object[] newSelObj = new Object[this.selectionObject.length+1];
        Object[] newDisObj = new Object[this.displayObject.length+1];

        newSelObj[0] = null;
        newDisObj[0] = null;
        for (int i=0; i<this.selectionObject.length; i++) {
          newSelObj[i+1] = this.selectionObject[i];
          newDisObj[i+1] = this.displayObject[i];
        }
        this.selectionObject = newSelObj;
        this.displayObject   = newDisObj;
      }
      // Combo-Box aktualisieren
      ((JComboBox)inpComp).removeAllItems();
      for (int i=0; i<displayObject.length; i++)
        ((JComboBox)inpComp).addItem( displayObject[i] );
      // Tooltips neu setzen
      ((JComboBox)inpComp).setRenderer( new ToolTipComboBoxRenderer(displayToolTip) );
    }

    /**
     * Liefert den Index, der in der ComboBox-Liste ausgewaehlt wurde.
     * @return <code>-1</code> falls kein Objekt ausgewaehlt wurde
     */
    public int getSelectedIndex() {
      return ((JComboBox)inpComp).getSelectedIndex();
    }

    /**
     * Setzt den Index, der in der ComboBox-Liste ausgewaehlt wurde.
     * @param idx Listen-Index (-1 um eine Leer-Auswahl zu erzeugen)
     */
    public void setSelectedIndex(int idx) {
      Object oldValue = getValue();
      ((JComboBox)inpComp).setSelectedIndex(idx);
      Object newValue = getValue();
      if ( oldValue != newValue || oldValue!=null && !oldValue.equals( getValue() ) )
        fireOptionChanged(oldValue,newValue);
    }
  }

  /**
   * Diese Klasse stellt eine Auswahl-Option dar, die durch ein {@link JPanel}
   * mit vertikal angeordneten {@link JRadioButton JRadioButtons} dargestellt wird.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class Radio<E> extends SelectionInputOption {
    /** Gruppe, in der die RadioButton agieren. */
    protected ButtonGroup buttonGroup;
    /** Speichert die letzte Auswahl. */
    protected Object lastSelection = null;
    /** Liste der Buttons. */
    protected Vector<JRadioButton> buttonList;

    /** ActionListener, der auf die Button-Klicks reagiert und ggf. Events
     *  feuert. */
    private ActionListener actionListener = new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        Object newValue = getValue();
        if ( lastSelection != newValue )
          fireOptionChanged(lastSelection,newValue);
        lastSelection = newValue;
      }
    };

    /**
     * Erzeugt eine neue Auswahl-Option.
     * @param label       Beschreibung
     * @param inputNeeded gibt an, ob eine Eingabe erforderlich ist
     * @param value       die zur Auswahl stehenden Objekte
     * @param defIdx      Index der vorgeblendeten Auswahl
     * @param display     die anstelle der Auswahl-Objekte angezeigten Objekte
     *                    (kann <code>null</code> sein)
     * @exception IllegalArgumentException falls sich die Array-Groessen von
     *            <code>value</code> und <code>display</code> unterscheiden
     */
    public Radio(String label, boolean inputNeeded, E[] value, int defIdx, Object[] display) {
      super(label,inputNeeded,value,defIdx, display);
    }

    /**
     * Erzeugt eine neue Auswahl-Option.
     * @param label       Beschreibung
     * @param inputNeeded gibt an, ob eine Eingabe erforderlich ist
     * @param value       die zur Auswahl stehenden Objekte
     * @param defValue    vorgeblendetes Auswahlobjekt
     * @param display     die anstelle der Auswahl-Objekte angezeigten Objekte
     *                    (kann <code>null</code> sein)
     * @exception IllegalArgumentException falls sich die Array-Groessen von
     *            <code>value</code> und <code>display</code> unterscheiden
     */
    public Radio(String label, boolean inputNeeded, E[] value, E defValue, Object[] display) {
      super(label,inputNeeded,value,defValue,display);
    }

    /**
     * Erzeugt eine neue Auswahl-Option.
     * @param label       Beschreibung
     * @param inputNeeded gibt an, ob eine Eingabe erforderlich ist
     * @param value       die zur Auswahl stehenden Objekte
     * @param defValue    vorgeblendetes Auswahlobjekt
     * @param display     die anstelle der Auswahl-Objekte angezeigten Objekte
     *                    (kann <code>null</code> sein)
     * @param tooltip     die fuer jedes Auswahl-Objekt angezeigten Tooltips
     *                    (kann <code>null</code> sein)
     * @exception IllegalArgumentException falls sich die Array-Groessen von
     *            <code>value</code> und <code>display</code> unterscheiden
     */
    public Radio(String label, boolean inputNeeded, E[] value, E defValue, Object[] display, Map<Object,String> tooltip) {
      super(label,inputNeeded,value,defValue,display,tooltip);
    }

    /**
     * Erzeugt eine neue Auswahl-Option. Es wird das erste Auswahl-Objekt
     * vorgeblendet.
     * @param label       Beschreibung
     * @param inputNeeded gibt an, ob eine Eingabe erforderlich ist
     * @param value       die zur Auswahl stehenden Objekte
     * @param display     die anstelle der Auswahl-Objekte angezeigten Objekte
     *                    (kann <code>null</code> sein)
     * @exception IllegalArgumentException falls sich die Array-Groessen von
     *            <code>value</code> und <code>display</code> unterscheiden
     */
    public Radio(String label, boolean inputNeeded, E[] value, Object[] display) {
      super(label,inputNeeded,value,display);
    }

    /**
     * Erzeugt eine leere Auswahl-Option. Diese muss nachtraeglich ueber
     * {@link #setSelectionObjects(Object[],Object[])} befuellt werden.
     * @param label       Beschreibung
     * @param inputNeeded gibt an, ob eine Eingabe erforderlich ist
     */
    public Radio(String label, boolean inputNeeded) {
      super(label,inputNeeded,new Object[0],new Object[0]);
    }

    /**
     * Erzeugt eine neues {@link JPanel}, in dem untereinander
     * {@link JRadioButton JRadioButtons} fuer jede Auswahl-Option
     * angeordnet werden.
     */
    protected JPanel createInputComponent() {
      this.buttonGroup = new ButtonGroup();
      this.buttonList  = new Vector<JRadioButton>();
      JPanel panel = new JPanel();
      panel.setLayout( new BoxLayout(panel,BoxLayout.Y_AXIS) );
      return panel;
    }

    /**
     * Befuellt das Panel neu mit RadioButtons.
     */
    protected void performSelectionUpdate() {
      JPanel buttonPanel = ((JPanel)inpComp);
      // Panel aktualisieren
      buttonGroup = new ButtonGroup();
      buttonPanel.removeAll();
      if ( buttonList != null )
        buttonList.clear();
      for (int i=0; i<displayObject.length; i++) {
        Object disObj = displayObject[i];
        JRadioButton button = null;
        if ( disObj instanceof Icon )
          button = new JRadioButton((Icon)disObj);
        else
          button = new JRadioButton(disObj.toString());
        button.addActionListener(actionListener);
        if (this.displayToolTip != null)
          button.setToolTipText( (String)displayToolTip.get(selectionObject[i]) );
        buttonPanel.add(button);
        buttonGroup.add(button);
        buttonList.add(button);
      }
    }

    /**
     * Liefert den Index, der in der Button-Liste ausgewaehlt wurde.
     * @return <code>-1</code> falls kein Objekt ausgewaehlt wurde
     */
    public int getSelectedIndex() {
      if ( buttonGroup.getSelection() == null )
        return -1;
      return buttonList.indexOf( buttonGroup.getSelectedButton() );
    }

    /**
     * Setzt den Index, der in der Button-Liste ausgewaehlt wurde.
     * @param idx Listen-Index (-1 um eine Leer-Auswahl zu erzeugen)
     */
    public void setSelectedIndex(int idx) {
// Event wird - glaube ich - bereits durch den ActionListener
// des RadioButtons realisiert!
//      Object oldValue = getValue();
      if ( idx == -1 || idx >= buttonList.size() )
        buttonGroup.setUnselected();
      else
        buttonList.elementAt(idx).setSelected(true);
//      Object newValue = getValue();
//      if ( oldValue != newValue || oldValue!=null && !oldValue.equals( getValue() ) )
//        fireOptionChanged(oldValue,newValue);
    }
  }
}
