/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing;

import java.awt.Font;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;

import javax.swing.JLabel;
import javax.swing.JSlider;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

/**
 * Dieses Panel enthaelt einen Slider, der mit einem Spinner verknuepft ist.
 * Zusaetzlich wird eine Ueberschrift ausgewiesen.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class SliderSpinnerPanel extends JPanel {
  /** Konstante fuer die <b>horizontale</b> Orientierung des Sliders.
   *  @see JSlider#HORIZONTAL */
  public static final int HORIZONTAL = JSlider.HORIZONTAL;
  /** Konstante fuer die <b>vertikale</b> Orientierung des Sliders.
   *  @see JSlider#VERTICAL */
  public static final int VERTICAL = JSlider.VERTICAL;

  /** Slider des Panels. */
  protected JSlider slider = null;
  /** Ueberschrift-Label des Panels. */
  protected JLabel headerLabel = null;
  /** Spinner des Panels. */
  protected JSpinner spinner = null;

  /**
   * Erzeugt ein neues Panel.
   * @param orientation Orientierung des Sliders
   * @param min Minimal einstellbarer Wert
   * @param max Maximal einstellbarer Wert
   * @param step Schrittweite fuer Spinner
   * @param value Initialer Wert
   * @param headerText Text fuer das Ueberschrift-Label (wenn {@code null} wird keine Ueberschrift ausgegeben!)
   */
  public SliderSpinnerPanel(int orientation, double min, double max, double step, double value, String headerText) {
    super();
    this.setLayout( new GridBagLayout() );
    // Ueberschrift
    if ( headerText != null ) {
      this.headerLabel = new JLabel( headerText );
      this.headerLabel.setFont( headerLabel.getFont().deriveFont(Font.BOLD) );
      this.add( headerLabel, new GridBagConstraints(
        0, 0, 1, 1, 0.0, 0.0, GridBagConstraints.SOUTHWEST, GridBagConstraints.NONE, new Insets(0, 0, 0, 0), 0, 0) );
    }
    // Slider
    this.slider = new JSlider(orientation);
    this.add( slider, new GridBagConstraints(
      0, 1, 1, 1, 1.0, 1.0, GridBagConstraints.CENTER, GridBagConstraints.BOTH, new Insets(0, 0, 0, 0), 0, 0) );
    // Spinner
    this.spinner = new JSpinner();
    SwingUtil.setPreferredWidth(this.spinner, 60);
    this.add( spinner, new GridBagConstraints(
      1, 1, 1, 1, 0.0, 0.0, GridBagConstraints.NORTHWEST, GridBagConstraints.HORIZONTAL, new Insets(0, 0, 0, 0), 0, 0) );

    // ActionListener auf Spinner, um Slider zu aktualisieren
    this.spinner.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        double newValue = ((Number)spinner.getValue()).doubleValue();
        if ( Math.round(newValue) != slider.getValue())
          slider.setValue( (int)Math.round(newValue) );
      }
    });

    // ActionListener auf Slider, um Spinner zu aktualisieren
    this.slider.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        int newValue = slider.getValue();
        if (newValue != Math.round( ((Number)spinner.getValue()).doubleValue() ))
          spinner.setValue(newValue);
      }
    });

    // Minimum, Maximum, StepSize und initalalier Wert einstellen
    setRestrictions(value,min,max,step);
  }

  /**
   * Erzeugt ein neues Panel. Minimum und Maximum werden automatisch auf 0 und
   * 100 eingestellt.
   * @param orientation Orientierung des Sliders
   * @param headerText Text fuer das Ueberschrift-Label (wenn {@code null} wird keine Ueberschrift ausgegeben!)
   */
  public SliderSpinnerPanel(int orientation, String headerText) {
    this(orientation, 0, 100, 1, 0, headerText);
  }

  /**
   * Liefert den eingestellten Spinner-Wert.
   */
  public double getValue() {
    return ((Number)spinner.getValue()).doubleValue();
  }

  /**
   * Setzt den eingestellten Spinner-Wert.
   * @param value Spinner-Wert
   */
  public void setValue(double value) {
    spinner.setValue( Math.max( Math.min(value,getMaxValue()), getMinValue() ) );
  }

  /**
   * Liefert den kleinsten einstellbaren Spinner-Wert.
   */
  public double getMinValue() {
    return ((Number)((SpinnerNumberModel)spinner.getModel()).getMinimum()).doubleValue();
  }

  /**
   * Setzt den kleinsten einstellbaren Spinner-Wert.
   * @param min kleinster einstellbarer Spinner-Wert
   */
  public void setMinValue(double min) {
    setRestrictions(
      getValue(),
      min,
      getMaxValue(),
      getStepSize()
    );
  }

  /**
   * Liefert den groessten einstellbaren Spinner-Wert.
   */
  public double getMaxValue() {
    return ((Number)((SpinnerNumberModel)spinner.getModel()).getMaximum()).doubleValue();
  }

  /**
   * Setzt den groessten einstellbaren Spinner-Wert.
   * @param max groesster einstellbarer Spinner-Wert
   */
  public void setMaxValue(double max) {
    setRestrictions(
      getValue(),
      getMinValue(),
      max,
      getStepSize()
    );
  }

  /**
   * Liefert die Schrittgroesse des Spinners.
   */
  public double getStepSize() {
    return ((SpinnerNumberModel)spinner.getModel()).getStepSize().doubleValue();
  }

  /**
   * Setzt die Schrittweiter des Spinners.
   * @param stepSize groesster einstellbarer Spinner-Wert
   */
  public void setStepSize(double stepSize) {
    setRestrictions(
      getValue(),
      getMinValue(),
      getMaxValue(),
      stepSize
    );
  }

  /**
   * Setzt alle Restriktionen des Sliders und Spinners neu.
   * @param value    angezeigter Wert
   * @param min      minimaler einstellbarer Wert
   * @param max      maximaler einstellbarer Wert
   * @param stepSize Schrittweite des Spinners
   */
  public void setRestrictions(double value, double min, double max, double stepSize) {
    // Spinner einstellen
    spinner.setModel( new SpinnerNumberModel(
      value,
      min,
      max,
      stepSize
    ) );
    spinner.setEditor( new JSpinner.NumberEditor(this.spinner,SwingUtil.getNumberFormatPattern(stepSize)) );
    // Slider einstellen
    slider.setMinimum( (int)Math.floor(min) );
    slider.setMaximum( (int)Math.ceil(max) );
    setValue( value );
  }

  /**
   * Liefert den Slider des Panels.
   */
  public JSlider getSlider() {
    return this.slider;
  }

  /**
   * Liefert das Ueberschrift-Label des Panels.
   */
  public JLabel getHeaderLabel() {
    return this.headerLabel;
  }

  /**
   * Liefert den Spinner des Panels.
   */
  public JSpinner getSpinner() {
    return this.spinner;
  }
}
