/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing;

import java.util.Vector;

import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.table.TableColumnModel;
import javax.swing.table.TableModel;

/**
 * Extends the {@link JTable} with automatic sort functionality.
 * This class adds the following methods so that manually calling
 * {@link JTable#convertRowIndexToModel(int)} and {@link JTable#convertRowIndexToView(int)}
 * is not necessary:
 * <ul>
 *   <li>{@link #getEditingModelRow()}</li>
 *   <li>{@link #getSelectedModelRow()}</li>
 *   <li>{@link #getSelectedModelRows()}</li>
 *   <li>{@link #addModelRowSelectionInterval(int, int)}</li>
 *   <li>{@link #setModelRowSelectionInterval(int, int)}</li>
 *   <li>{@link #removeRowSelectionInterval(int, int)}</li>
 *   <li>{@link #changeModelSelection(int, int, boolean, boolean)}</li>
 * </ul>   
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 */
public class SortableJTable extends SelectableJTable {

  /**
   * Creates an empty table.
   */
  public SortableJTable() {
    super();
    initTable();
  }

  /**
   * Creates an empty table.
   */
  public SortableJTable(int numRows, int numColumns) {
    super(numRows, numColumns);
    initTable();
  }

  /**
   * Creates a new table.
   */
  public SortableJTable(Object[][] rowData, Object[] columnNames) {
    super(rowData, columnNames);
    initTable();
  }

  /**
   * Creates a new table.
   */
  public SortableJTable(TableModel dm, TableColumnModel cm, ListSelectionModel sm) {
    super(dm, cm, sm);
    initTable();
  }

  /**
   * Creates a new table.
   */
  public SortableJTable(TableModel dm, TableColumnModel cm) {
    super(dm, cm);
    initTable();
  }

  /**
   * Creates a new table.
   */
  public SortableJTable(TableModel dm) {
    super(dm);
    initTable();
  }

  /**
   * Creates a new table.
   */
  public SortableJTable(Vector rowData, Vector columnNames) {
    super(rowData, columnNames);
    initTable();
  }
  
  /**
   * Called by every constructor to initialize the extended
   * functionalities.
   */
  protected void initTable() {
    setAutoCreateRowSorter(true);
  }
  
  /**
   * Returns the index of the currently edited row according to
   * the table model.
   */
  public int getEditingModelRow() {
    return convertRowIndexToModel(getEditingRow());
  }

  /**
   * Returns the index of the selected row according to
   * the table model.
   */
  public int getSelectedModelRow() {
    return convertRowIndexToModel(getSelectedRow());
  }

  /**
   * Returns the indices of the selected rows according to
   * the table model.
   */
  public int[] getSelectedModelRows() {
    int[] rows = super.getSelectedRows();
    for (int i=0; i<rows.length; i++)
      rows[i] = convertRowIndexToModel(rows[i]);
    return rows;
  }
  
  /**
   * Changes the selected cells.
   * @param row row index
   * @param col column index
   * @param toggle see {@link #changeSelection(int, int, boolean, boolean)}
   * @param extend see {@link #changeSelection(int, int, boolean, boolean)}
   */
  public void changeModelSelection(int row, int col, boolean toggle, boolean extend) {
    row = convertRowIndexToView(row);
    changeSelection(row, col, toggle, extend);
  }
  
  /**
   * Extends the current selection by the given index interval
   * according to the table model.
   * @param index0 start index
   * @param index1 end index (inclusive)
   */
  public void addModelRowSelectionInterval(int index0, int index1) {
    for (int i=index0; i<=index1; i++) {
      int viewIdx = convertRowIndexToView(i);
      addRowSelectionInterval(viewIdx, viewIdx);
    }
  }

  /**
   * Sets the current selection to the given index interval
   * according to the table model.
   * @param index0 start index
   * @param index1 end index (inclusive)
   */
  public void setModelRowSelectionInterval(int index0, int index1) {
    clearSelection();
    for (int i=index0; i<=index1; i++) {
      int viewIdx = convertRowIndexToView(i);
      addRowSelectionInterval(viewIdx, viewIdx);
    }
  }
  
  /**
   * Removes the given index interval of rows from the current selection
   * according to the table model.
   * @param index0 start index
   * @param index1 end index (inclusive)
   */
  public void removeModelRowSelectionInterval(int index0, int index1) {
    for (int i=index0; i<=index1; i++) {
      int viewIdx = convertRowIndexToView(i);
      removeRowSelectionInterval(viewIdx, viewIdx);
    }
  }
}
