/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.Vector;

import javax.swing.Icon;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JProgressBar;

/**
 * Diese Klasse stellt einen modalen Status-Dialog dar. Diese besteht neben
 * einer Meldung aus einem {@linkplain JProgressBar Status-Balken} der
 * standardmaessig auf 'indeterminate' eingestellt ist
 * (siehe {@link JProgressBar#setIndeterminate(boolean)}). Darueberhinaus
 * kann der Dialog ueber einen Abbrechen-Button beendet werden. Wie auf den
 * Abbruch reagiert wird, ist durch die aufrufende Klasse zu behandeln.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class StatusDialog extends JDialog {
  /** Dialog-Option "Abbrechen". */
  public static final int CANCEL_OPTION = JOptionPane.CANCEL_OPTION;
  /** Dialog-Option "Ok". */
  public static final int OK_OPTION = JOptionPane.OK_OPTION;

  /** Label in dem die Meldung angezeigt wird. */
  protected JLabel messageLabel = null;
  /** Status-Balken, der den Fortschritt anzeigt. */
  protected JProgressBar progressBar  = null;
  /** Button um den Dialog zu beenden. */
  protected JButton button = null;
  /** Flag signalisiert, ob der Dialog ueber den Button abgebrochen wurde. */
  protected boolean canceled = false;
  /**
   * Typ des Dialogs (Stanard: {@link #CANCEL_OPTION}).
   * @see #CANCEL_OPTION
   * @see #OK_OPTION
   * @see #setDialogOption(int)
   */
  protected int dialogOption = CANCEL_OPTION;

  /**
   * Erzeugt einen neuen Status-Dialog. Der Dialog wird relativ zum Parent-Fenster
   * zentriert.
   * @param parent   uebergeordnetes Fenster (kann <code>null</code> sein!)
   * @param title    Titel fuer das Fenster
   * @param message  Meldung, die zu dem Status-Balken angezeigt wird
   */
  public StatusDialog(Component parent, String title, String message) {
    this(parent,title,message,0.5,0.5);
  }

  /**
   * Erzeugt einen neuen Status-Dialog.
   * @param parent   uebergeordnetes Fenster (kann <code>null</code> sein!)
   * @param title    Titel fuer das Fenster
   * @param message  Meldung, die zu dem Status-Balken angezeigt wird
   * @param relX     relative horizontale Position zum Parent-Fenster
   * @param relY     relative vertikale Position zum Parent-Fenster
   */
  public StatusDialog(Component parent, String title, String message, double relX, double relY) {
    this(parent,title,message,null,relX,relY);
  }
  /**
   * Erzeugt einen neuen Status-Dialog.
   * @param parent   uebergeordnetes Fenster (kann <code>null</code> sein!)
   * @param title    Titel fuer das Fenster
   * @param message  Meldung, die zu dem Status-Balken angezeigt wird
   * @param icon     Icon fuer das Status-Fenster
   * @param relX     relative horizontale Position zum Parent-Fenster
   * @param relY     relative vertikale Position zum Parent-Fenster
   */
  public StatusDialog(Component parent, String title, String message, Icon icon, double relX, double relY) {
//	  War: super((Frame)parent,true); aber getParentWindow ist wohl besser
	  super( SwingUtil.getParentWindow(parent), ModalityType.APPLICATION_MODAL);
    
    // wenn kein uebergeordnetes Fenster angegeben ist, wird es immer
    // im Vordergrund angezeigt
    if ( parent==null )
      this.setAlwaysOnTop(true);

    // Vorlagen-Dialog erzeugen
    this.messageLabel = new JLabel(message);
    this.progressBar  = new JProgressBar(JProgressBar.HORIZONTAL,0,100);
    this.progressBar.setIndeterminate(true);
    this.progressBar.getPreferredSize().height = 50;
    this.button       = new JButton();
    setDialogOption( CANCEL_OPTION );

    Vector infoComponents = new Vector<Component>();
    insertInformationComponents(infoComponents);
    JOptionPane  pane = new JOptionPane(
      infoComponents.toArray(),
      JOptionPane.INFORMATION_MESSAGE,
      JOptionPane.DEFAULT_OPTION,
      icon,
      new Object[] {button}
    );
    JDialog dialog = pane.createDialog(parent,title);

    // Dialog nach Vorlage initialisieren
    this.setTitle( dialog.getTitle() );
    this.getContentPane().setLayout(new BorderLayout());
    this.getContentPane().add(dialog.getContentPane());
    this.setDefaultCloseOperation( DO_NOTHING_ON_CLOSE );
    this.button.addActionListener( new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        // Wenn die Option des Dialogs auf "OK" gesetzt wurde, handelt es
        // sich beim Button-Klick nicht um einen Abbruch!
        canceled = (dialogOption == CANCEL_OPTION);
        setVisible(false);
      }
    } );
    pack();
    SwingUtil.setRelativeFramePosition(this,SwingUtil.getParentWindow(parent),relX,relY);
  }

  /**
   * Fuegt die Informations-Komponenten des Dialogs (Label und Statusbalken)
   * in eine Liste ein. Diese werden dann (untereinander) im Dialog angezeigt.
   * Sub-Klassen koennen diese Methode ueberschreiben und weitere Komponenten
   * in die Liste einfuegen.<br>
   * <b>Bemerke:</b> Der Abbruch-Button ist eine Steuerungskomponente und
   * gehoert <b>nicht</b> zu den Elementen der Liste.
   * @param list Liste von Komponenten, die im Dialog angezeigt werden
   */
  protected void insertInformationComponents(Vector<Component> list) {
    list.add( messageLabel );
    list.add( progressBar );
  }

  /**
   * Liefert eine Referenz auf den Status-Balken, der im Dialog angezeigt wird.
   * Hierueber kann dieser formatiert werden.
   */
  public JProgressBar getProgressBar() {
    return progressBar;
  }

  /**
   * Setzt die Eigenschaft des Dialog-Buttons.
   * @param option {@link #CANCEL_OPTION} oder {@link #OK_OPTION}
   * @exception IllegalArgumentException falls keine der Optionen
   *            {@link #CANCEL_OPTION} oder {@link #OK_OPTION} angegeben wurde
   */
  public void setDialogOption(int option) {
    switch( option ) {
      case CANCEL_OPTION: this.button.setText(SwingUtil.RESOURCE.getString("Cancel"));
                          break;
      case OK_OPTION: this.button.setText(SwingUtil.RESOURCE.getString("Ok"));
                      break;
      default: throw new IllegalArgumentException("Unknown dialog option! Only CANCEL_OPTION or OK_OPTION allowed!");
    }
    this.dialogOption = option;
  }

  /**
   * Zeigt oder verbirgt den Dialog. Beim Angezeigen wird das {@link #canceled}-Flag
   * mit <code>false</code> initialisiert.
   */
  public void setVisible(boolean visible) {
    if ( !isVisible() && visible )
      canceled = false;
    super.setVisible(visible);
  }

  /**
   * Prueft, ob das Fenster durch den Button abgebrochen wurde.
   */
  public boolean isCanceled() {
    return canceled;
  }

}
