/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing;

import java.awt.Color;
import java.awt.Dimension;

import javax.swing.JButton;
import javax.swing.JPanel;

/**
 * Stellt eine horizontale (links rot/rechts gruen) oder vertikale
 * (oben rot/unten gruen) Rot/Gruen-Ampel dar.
 * Die Ampel wird aus zwei nebeneinander oder untereinander angeordneten
 * {@link JButton} erstellt, die deaktiviert sind, also nicht auf
 * Clicks reagieren.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class StoplightContainer extends JPanel {
  /** beide Ampel-Lichter sind aus */
  public static final int EMPTY = 0;
  /** das untere (rechte) Ampel-Licht ist Gruen */
  public static final int GREEN = 1;
  /** das obere (linke) Ampel-Licht ist Rot */
  public static final int RED = 2;

  private CircleIcon emptyIcon;
  private CircleIcon greenIcon;
  private CircleIcon redIcon;

  private JButton upperLight;
  private JButton lowerLight;
  private int mode;

  /**
   * Erzeugt eine horizontale oder vertikale Ampel.
   * @param mode initialer Modus (EMPTY/GREEN/RED)
   * @param width Breite der Ampel in Pixeln
   * @param height Hoehe der Ampel in Pixeln
   * @param vertical <code>true</code> = vertikale Anordnung; <code>false<code> = horizontale Anordnung
   */
  public StoplightContainer(int mode, int width, int height, boolean vertical) {
    super();
    setSize( new Dimension(width, height) );
    setLayout(null);

    // Durchmesser des Kreises ist 80% der Ampelbreite
    int cirWidth = (int) (vertical ? (width * 0.8) : (height * 0.8));
    emptyIcon = new CircleIcon(cirWidth, cirWidth, Color.lightGray);
    greenIcon = new CircleIcon(cirWidth, cirWidth, Color.green);
    redIcon = new CircleIcon(cirWidth, cirWidth, Color.red);
    if (vertical) {
      upperLight = createButton(width, height / 2, 0, 0);
      lowerLight = createButton(width, height / 2, 0, height / 2);
    }
    else {
      upperLight = createButton(width / 2, height, 0, 0);
      lowerLight = createButton(width / 2, height, width / 2, 0);
    }

    upperLight.setEnabled(false);
    lowerLight.setEnabled(false);
    add(upperLight);
    add(lowerLight);

    setMode(mode);
  }

  /**
   * Erzeugt eine vertikale Ampel.
   * @param mode initialer Modus (EMPTY/GREEN/RED)
   * @param width Breite der Ampel in Pixeln
   * @param height Hoehe der Ampel in Pixeln
   */
  public StoplightContainer(int mode, int width, int height) {
    this(mode, width, height, true);
  }

  private JButton createButton(int width, int height, int posX, int posY) {
    JButton b = new JButton();
    b.setSize(width, height);
    b.setLocation(posX, posY);
    return b;
  }

  /**
   * Setzt den Modus der Ampel
   * @param mode EMPTY, GREEN oder RED (andere Werte werden ignoriert)
   */
  public void setMode(int mode) {
    this.mode = mode;
    switch (mode) {
      case EMPTY: {
        upperLight.setIcon(emptyIcon);
        lowerLight.setIcon(emptyIcon);
        break;
      }
      case GREEN: {
        upperLight.setIcon(emptyIcon);
        lowerLight.setIcon(greenIcon);
        break;
      }
      case RED: {
        upperLight.setIcon(redIcon);
        lowerLight.setIcon(emptyIcon);
        break;
      }
    }
  }

  /**
   * Liefert den aktuellen Modus der Ampel.
   * @return EMPTY, GREEN oder RED.
   */
  public int getMode() {
    return mode;
  }
}
