/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Graphics2D;
import java.awt.GridBagConstraints;
import java.awt.Image;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.image.BufferedImage;
import java.text.NumberFormat;
import java.util.Enumeration;
import java.util.Locale;
import java.util.Random;
import java.util.StringTokenizer;

import javax.imageio.ImageIO;
import javax.swing.AbstractButton;
import javax.swing.ImageIcon;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JInternalFrame;
import javax.swing.JLabel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTree;
import javax.swing.table.TableCellRenderer;
import javax.swing.tree.TreeNode;
import javax.swing.tree.TreePath;

import org.apache.log4j.Logger;

import schmitzm.lang.LangUtil;
import schmitzm.lang.ResourceProvider;

/**
 * Diese Klasse beinhaltet statische Hilfsfunktionen fuer das Arbeiten mit
 * Swing-GUIs.
 * 
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *         (University of Bonn/Germany)
 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons Kr&uuml;ger</a>
 * 
 * @version 1.1
 */
public class SwingUtil {
	private static final Logger LOGGER = Logger.getLogger(SwingUtil.class
			.getClass().getName());

	/**
	 * {@link ResourceProvider}, der die Lokalisation fuer GUI-Komponenten des
	 * Package {@code schmitzm.swing} zur Verfuegung stellt. Diese sind in
	 * properties-Dateien unter {@code schmitzm.swing.resource.locales}
	 * hinterlegt.
	 */
	public static ResourceProvider RESOURCE = new ResourceProvider(LangUtil
			.extendPackagePath(SwingUtil.class,
					"resource.locales.SwingResourceBundle"), Locale.ENGLISH);

	/**
	 * Convenience method to access the translation resources.
	 */
	public static String R(String key, Object... values) {
		return RESOURCE.getString(key, values);
	}

	// ****************************************************************************
	// Diese Icons sind auf Basis der Icons von Gimp erstellt
	// Eine Sammlung aller Gimp-Icons liegt im svn: gimp-tool-cursors.xcf
	// ****************************************************************************
	/** Cursor in Form einer Lupe mit Plus Symbol */
	public static final Cursor ZOOMIN_CURSOR = createCursorFromResourcePath(
			"resource/cursor/zoom_in.gif", 8, 8, null);
	/** Cursor in Form einer Lupe mit Minus Symbol */
	public static final Cursor ZOOMOUT_CURSOR = createCursorFromResourcePath(
			"resource/cursor/zoom_out.gif", 8, 8, null);
	/** Cursor in Form einer Lupe ohne Symbol */
	public static final Cursor ZOOM_CURSOR = createCursorFromResourcePath(
			"resource/cursor/zoom.gif", 8, 8, null);
	/** Cursor in Form einer offenen Hand **/
	public static final Cursor PAN_CURSOR = createCursorFromResourcePath(
			"resource/cursor/hand_pan.png", 9, 11, null);
	/** Cursor in Form einer geschlossenen Hand **/
	public static final Cursor PANNING_CURSOR = createCursorFromResourcePath(
			"resource/cursor/hand_closed.png", 9, 11, null);
	/** Cursor in Form eines Zielkreuzes **/
	public static final Cursor CROSSHAIR_CURSOR = createCursorFromResourcePath(
			"resource/cursor/crosshair.gif", 10, 10, null);
	/** Cursor in Form eines Zielkreuzes mit kleinem i **/
	public static final Cursor INFO_CURSOR = createCursorFromResourcePath(
			"resource/cursor/crosshair_i.png", 10, 10, null);
	public static final Cursor SELECTION_ADD_CURSOR = createCursorFromResourcePath(
			"resource/cursor/selection_add.png", 10, 10, null);
	public static final Cursor SELECTION_SET_CURSOR = createCursorFromResourcePath(
			"resource/cursor/selection_set.png", 10, 10, null);
	public static final Cursor SELECTION_REMOVE_CURSOR = createCursorFromResourcePath(
			"resource/cursor/selection_remove.png", 10, 10, null);
	
	

	// public static final ImageIcon ICON_RASTER =
	// createImageIconFromResourcePath("resource/icon/small/raster.png","");
	// public static final ImageIcon ICON_VECTOR =
	// createImageIconFromResourcePath("resource/icon/small/vector.png","");

	/**
	 * Modus "Innen".
	 * 
	 * @see #setRelativeFramePosition(Window,Window,int,int)
	 */
	public static final int BOUNDS_INNER = 10;
	/**
	 * Modus "Aussen".
	 * 
	 * @see #setRelativeFramePosition(Window,Window,int,int)
	 */
	public static final int BOUNDS_OUTER = 20;
	/** Ausrichtung oben-mitte. */
	public static final int NORTH = GridBagConstraints.NORTH;
	/** Ausrichtung oben-links. */
	public static final int NORTHWEST = GridBagConstraints.NORTHWEST;
	/** Ausrichtung oben-rechts. */
	public static final int NORTHEAST = GridBagConstraints.NORTHEAST;
	/** Ausrichtung unten-mitte. */
	public static final int SOUTH = GridBagConstraints.SOUTH;
	/** Ausrichtung unten-links. */
	public static final int SOUTHWEST = GridBagConstraints.SOUTHWEST;
	/** Ausrichtung unten-rechts. */
	public static final int SOUTHEAST = GridBagConstraints.SOUTHEAST;
	/** Ausrichtung zentriert. */
	public static final int CENTER = GridBagConstraints.CENTER;
	/** Ausrichtung mitte-links. */
	public static final int WEST = GridBagConstraints.WEST;
	/** Ausrichtung mitte-rechts. */
	public static final int EAST = GridBagConstraints.EAST;

	/**
	 * Erzeugt ein Icon auf Basis einer relativen Pfad-Angabe. Als
	 * Basis-Verzeichnis wird der Classpath von {@code SwingUtil} verwendet.
	 * 
	 * @param imgPath
	 *            relativer Pfad des Icons
	 * @param imgDesc
	 *            Beschreibung fuer Icon
	 * @return {@code null}, wenn das Icon nicht gefunden wird
	 */
	public static ImageIcon createImageIconFromResourcePath(String imgPath,
			String imgDesc) {
		return createImageIconFromResourcePath(null, imgPath, imgDesc);
	}

	/**
	 * Erzeugt ein Icon auf Basis einer relativen Pfad-Angabe.
	 * 
	 * @param resourceBase
	 *            Klasse, deren Classpath als Basis-Verzeichnis verwendet wird
	 * @param imgPath
	 *            relativer Pfad des Icons
	 * @param imgDesc
	 *            Beschreibung fuer Icon
	 * @return {@code null}, wenn das Icon nicht gefunden wird
	 */
	public static ImageIcon createImageIconFromResourcePath(Class resourceBase,
			String imgPath, String imgDesc) {
		if (resourceBase == null)
			resourceBase = SwingUtil.class;
		java.net.URL imgURL = resourceBase.getResource(imgPath);
		if (imgURL == null) {
			LOGGER.error("Couldn't find image file: " + imgPath);
			return null;
		}
		return new ImageIcon(imgURL, imgDesc);
	}

	/**
	 * Erzeugt einen {@link Cursor} auf Basis einer relativen Pfad-Angabe. Als
	 * Basis-Verzeichnis wird der Classpath von {@code SwingUtil} verwendet.
	 * 
	 * @param imgPath
	 *            relativer Pfad des Icons
	 * @param x
	 *            X-Position im Image, die den Hotspot des Cursors darstellen
	 *            soll
	 * @param y
	 *            Y-Position im Image, die den Hotspot des Cursors darstellen
	 *            soll
	 * @param name
	 *            Bezeichnung fuer den Cursor
	 * @return {@code null}, wenn das Icon nicht gefunden wird
	 */
	public static Cursor createCursorFromResourcePath(String imgPath, int x,
			int y, String name) {
		java.net.URL imgURL = SwingUtil.class.getResource(imgPath);
		try {
			return Toolkit.getDefaultToolkit().createCustomCursor(
					ImageIO.read(imgURL), new Point(x, y), name);
		} catch (Exception err) {
			return null;
		}
	}

	/**
	 * Erzeugt ein neues Fenster mit {@link BorderLayout} und zeigt darin eine
	 * {@link Component} an.
	 * 
	 * @param comp
	 *            anzuzeigende Komponente (kann {@code null} sein)
	 * @param title
	 *            Titel des Fensters (kann {@code null} sein)
	 * @param icon
	 *            Icon-Image (kann {@code null} sein)
	 */
	public static JFrame createFrame(Component comp, String title, Image icon) {
		JFrame frame = new JFrame(title);
		frame.getContentPane().setLayout(new BorderLayout());
		if (comp != null)
			frame.getContentPane().add(comp, BorderLayout.CENTER);
		if (title != null)
			frame.setTitle(title);
		if (icon != null)
			frame.setIconImage(icon);
		frame.pack();
		return frame;
	}

	/**
	 * Zeigt ein neues Fenster einer {@link Component} an.
	 * 
	 * @param comp
	 *            anzuzeigende Komponente (kann {@code null} sein)
	 * @param title
	 *            Titel des Fensters (kann {@code null} sein)
	 * @param icon
	 *            Icon-Image (kann {@code null} sein)
	 * @see #createFrame(Component, String, Image)
	 */
	public static JFrame showFrame(Component comp, String title, Image icon) {
		JFrame frame = createFrame(comp, title, icon);
		frame.setVisible(true);
		return frame;
	}

	/**
	 * Liefert das Fenster, das eine GUI-Komponente beinhaltet. Wenn
	 * GUI-Komponente selber ein {@link Window} ist, wird diese zurückgelifert.
	 * 
	 * @param comp
	 *            eine GUI-Komponente
	 * @return <code>null</code> falls die Komponente in keinem Fenster
	 *         enthalten ist
	 */
	public static Window getParentWindow(Component comp) {
		// durch die Parents laufen, bis Window gefunden
		while (comp != null && !(comp instanceof Window))
			comp = comp.getParent();
		// wenn kein Window gefunden -> null zurueckgeben
		if (comp == null)
			return null;
		return (Window) comp;
	}

	/**
	 * Liefert den {@link Frame}, das eine Kompoenente beinhaltet.
	 * 
	 * @param comp
	 *            eine GUI-Komponente
	 * @return <code>null</code> falls die Komponente in keinem {@link Frame}
	 *         enthalten ist
	 */
	public static Frame getParentFrame(Component comp) {
		// durch die Parents laufen, bis Frame gefunden
		while (comp != null && !(comp instanceof Frame))
			comp = comp.getParent();
		// wenn kein Frame gefunden -> null zurueckgeben
		if (comp == null)
			return null;
		return (Frame) comp;
	}

	/**
	 * Liefert das Fenster, das eine Kompoenente beinhaltet. Dabei kann es sich
	 * auch um einen {@link JInternalFrame} handelt.
	 * 
	 * @param comp
	 *            eine GUI-Komponente
	 * @return <code>null</code> falls die Komponente in keinem Fenster
	 *         enthalten ist
	 */
	public static Component getParentWindowComponent(Component comp) {
		// durch die Parents laufen, bis Window gefunden
		while (comp != null && !(comp instanceof Window)
				&& !(comp instanceof JInternalFrame))
			comp = comp.getParent();
		// wenn kein Window gefunden -> null zurueckgeben
		if (comp == null)
			return null;
		return comp;
	}

	/**
	 * Packt das Fenster, in dem eine Kompoenente plaziert ist.
	 * 
	 * @param comp
	 *            eine GUI-Komponente
	 * @return <code>false</code> falls die Komponente in keinem Fenster
	 *         enthalten ist
	 * @see Window#pack()
	 */
	public static boolean packParentWindow(Component comp) {
		Window w = getParentWindow(comp);
		if (w == null)
			return false;
		w.pack();
		return true;
	}

	/**
	 * Prueft, ob eine Komponente eine Kind-Komponente einer anderen Komponente
	 * ist.
	 * 
	 * @param child
	 *            Component
	 * @param parent
	 *            Component
	 * @return {@code false} wenn {@code child == null}, {@code true} wenn
	 *         {@code child == parent}, {@code
	 *         isChildComponent(child.getParent(),parent)} sonst
	 */
	public static boolean isChildComponent(Component child, Component parent) {
		if (child == null)
			return false;
		if (child == parent)
			return true;
		return isChildComponent(child.getParent(), parent);
	}

	/**
	 * Zentriert ein Fenster auf dem Monitor.
	 * 
	 * @param window
	 *            das zu zentrierende Fenster
	 */
	public static void centerFrameOnScreen(Component comp) {
		setRelativeFramePosition(comp, 0.5, 0.5);
	}

	/**
	 * Zentriert ein Fenster auf dem Monitor.<br>
	 * <b>Bemerkung:</b> Da {@link Window} eine {@link Component} ist, ist diese
	 * Methode eigentlich ueberfluessig. Es koennte
	 * {@link #centerFrameOnScreen(Component)} verwendet werden.
	 * Merkwuerdigerweise macht XULU beim Starten von der Console aus aber
	 * Probleme, wenn die Methode {@code centerFrameOnScreen(Window)} fehlt!!
	 * TODO: Probleme beim Xulu-Start loesen.
	 * 
	 * @param window
	 *            das zu zentrierende Fenster
	 */
	public static void centerFrameOnScreen(Window comp) {
		setRelativeFramePosition(comp, 0.5, 0.5);
	}

	/**
	 * Zentriert ein Fenster auf dem Monitor, aber verrückt das Window per
	 * Zufall um 10 Prozenz
	 * 
	 * @param comp
	 *            Eine Componente des zu zentrierenden Fensters. Wenn comp kein
	 *            {@link Window} ist, wird das Parent {@link Window} ermittelt.
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static void centerFrameOnScreenRandom(Component comp) {
		Random r = new Random();
		setRelativeFramePosition(comp, 0.5 + (r.nextDouble() * 0.2 - 0.1),
				0.5 + (r.nextDouble() * 0.2 - 0.1));
	}

	/**
	 * Positioniert ein Fenster auf dem Monitor, relativ zu dessen Groesse.
	 * (0.5/0.5) positioniert das Fenster z.B. genau in der Mitte des Monitors;
	 * (0.25/0.5) setzt das Fenster vertikal in die Mitte und horizontal auf ein
	 * Viertel der Monitorbreite.
	 * 
	 * @param comp
	 *            Eine Komponente des zu positionierenden Fensters oder direckt
	 *            das {@link Window}.
	 * @param relX
	 *            Relationsfaktor fuer die horizontale Position (0 <
	 *            <code>relX</code> < 1)
	 * @param relY
	 *            Relationsfaktor fuer die vertikale Position (0 <
	 *            <code>relX</code> < 1)
	 */
	public static void setRelativeFramePosition(Component comp, double relX,
			double relY) {

		setRelativeFramePosition(comp, null, relX, relY);
	}

	/**
	 * Positioniert ein Fenster auf dem Monitor relativ zur Position eines
	 * anderen Fensters. (0.5/0.5) positioniert das Fenster z.B. genau in der
	 * Mitte anderen Fensters. (0.25/0.5) setzt das Fenster vertikal in die
	 * Mitte und horizontal auf ein Viertel des Referenz-Fensters.
	 * 
	 * @param comp
	 *            Eine Komponente des zu positionierenden Fensters oder direkt
	 *            das {@link Window}.
	 * @param relComp
	 *            Fenster (parent Fenster der übergebenen Componente) zu dem das
	 *            Fenster in Relation gesetzt wird (wenn <code>null</code> wird
	 *            der gesamte Monitor verwendet)
	 * @param relX
	 *            Relationsfaktor fuer die horizontale Position (0 <
	 *            <code>relX</code> < 1)
	 * @param relY
	 *            Relationsfaktor fuer die vertikale Position (0 <
	 *            <code>relY</code> < 1)
	 */
	public static void setRelativeFramePosition(Component comp,
			Component relComp, double relX, double relY) {

		// Added by SK, 17.4.2009
		Window window = SwingUtil.getParentWindow(comp);
		relComp = getParentWindow(relComp);

		Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
		Dimension relSize = (relComp == null) ? screenSize : relComp.getSize();
		Point relLocation = (relComp == null) ? new Point(0, 0) : relComp
				.getLocation();

		// Fenster auf Monitor-Groesse anpassen
		Dimension frameSize = window.getSize();
		if (frameSize.height > screenSize.height) {
			frameSize.height = screenSize.height;
		}
		if (frameSize.width > screenSize.width) {
			frameSize.width = screenSize.width;
		}
		// Fenster positionieren
		window.setLocation(relLocation.x
				+ Math.round((relSize.width - frameSize.width) * (float) relX),
				relLocation.y
						+ Math.round((relSize.height - frameSize.height)
								* (float) relY));
	}

	/**
	 * Positioniert ein Fenster auf dem Monitor relativ zur Position eines
	 * anderen Fensters. Das Fenster wird jedoch immer innerhalb des Bildschirms
	 * positioniert.
	 * 
	 * @param window
	 *            das zu positionierende Fenster
	 * @param relWindow
	 *            Fenster zu dem das Fenster in Relation gesetzt wird (wenn
	 *            <code>null</code> wird der gesamte Monitor verwendet)
	 * @param type
	 *            bestimmt, ob das Fenster innerhalb oder ausserhalb des
	 *            relativen Fensters positioniert wird ({@link #BOUNDS_INNER}
	 *            oder {@link #BOUNDS_OUTER}).
	 * @param position
	 *            Positionierung des Fensters ({@link #NORTH},
	 *            {@link #NORTHEAST}, ...)
	 */
	public static void setRelativeFramePosition(Component comp,
			Component relWindow, int type, int position) {

		// Added by SK, 23.10.2009
		Window window = SwingUtil.getParentWindow(comp);

		relWindow = SwingUtil.getParentWindow(relWindow);

		if (type == BOUNDS_INNER) {
			// An den inneren Grenzen des Frames ausrichten
			double relX = 0;
			double relY = 0;
			switch (position) {
			case NORTH:
				relX = 0.5;
				relY = 0.0;
				break;
			case NORTHWEST:
				relX = 0.0;
				relY = 0.0;
				break;
			case NORTHEAST:
				relX = 1.0;
				relY = 0.0;
				break;
			case SOUTH:
				relX = 0.5;
				relY = 1.0;
				break;
			case SOUTHWEST:
				relX = 0.0;
				relY = 1.0;
				break;
			case SOUTHEAST:
				relX = 1.0;
				relY = 1.0;
				break;
			case CENTER:
				relX = 0.5;
				relY = 0.5;
				break;
			case WEST:
				relX = 0.0;
				relY = 0.5;
				break;
			case EAST:
				relX = 1.0;
				relY = 0.5;
				break;
			default:
				throw new IllegalArgumentException(
						"Unsupported mode for 'position'.");
			}
			setRelativeFramePosition(window, relWindow, relX, relY);
			return;
		}
		if (type == BOUNDS_OUTER) {
			Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
			Dimension relSize = (relWindow == null) ? screenSize : relWindow
					.getSize();
			Point relLocation = (relWindow == null) ? new Point(0, 0)
					: relWindow.getLocation();
			// An den Aussen-Grenzen des Frames ausrichten
			Point loc = new Point(0, 0);
			switch (position) {
			case NORTH:
			case NORTHWEST:
			case NORTHEAST:
				loc.y = relLocation.y; // Oben bündig ist schöner, oder?
				// loc.y = relLocation.y - window.getHeight();
				break;
			case SOUTH:
			case SOUTHWEST:
			case SOUTHEAST:
				loc.y = relLocation.y + relSize.height;
				break;
			case CENTER:
			case WEST:
			case EAST:
				loc.y = relLocation.y + relSize.height / 2 - window.getHeight()
						/ 2;
				break;
			default:
				throw new IllegalArgumentException(
						"Unsupported mode for 'position'.");
			}
			switch (position) {
			case WEST:
			case NORTHWEST:
			case SOUTHWEST:
				loc.x = relLocation.x - window.getWidth();
				break;
			case EAST:
			case NORTHEAST:
			case SOUTHEAST:
				loc.x = relLocation.x + relSize.width;
				break;
			case CENTER:
			case NORTH:
			case SOUTH:
				loc.x = relLocation.x + relSize.width / 2 - window.getWidth()
						/ 2;
				break;
			default:
				throw new IllegalArgumentException(
						"Unsupported mode for 'position'. GridBagConstraints location expected!");
			}

			loc.x = Math.max(0, loc.x);
			loc.y = Math.max(0, loc.y);
			if (loc.x + window.getWidth() > screenSize.width)
				loc.x = screenSize.width - window.getWidth();
			if (loc.y + window.getHeight() > screenSize.height)
				loc.y = screenSize.height - window.getHeight();

			// Fenster positionieren
			window.setLocation(loc);
			return;
		}
		throw new IllegalArgumentException(
				"Unsupported mode for 'type'. BOUNDS_INNER or BOUNDS_OUTER expected!");
	}

	/**
	 * Aendert die Breite einer GUI-Komponente.
	 * 
	 * @param comp
	 *            GUI-Komponente
	 * @param w
	 *            Breite
	 */
	public static void setWidth(Component comp, int w) {
		Dimension d = comp.getPreferredSize();
		d.width = w;
		comp.setSize(d);
	}

	/**
	 * Aendert die Hoehe einer GUI-Komponente.
	 * 
	 * @param comp
	 *            GUI-Komponente
	 * @param h
	 *            Hoehe
	 */
	public static void setHeight(Component comp, int h) {
		Dimension d = comp.getPreferredSize();
		d.height = h;
		comp.setSize(d);
	}

	/**
	 * Aendert die bevorzugte Breite einer GUI-Komponente.
	 * 
	 * @param comp
	 *            GUI-Komponente
	 * @param w
	 *            Breite
	 */
	public static void setPreferredWidth(Component comp, int w) {
		Dimension d = comp.getPreferredSize();
		d.width = w;
		comp.setPreferredSize(d);
	}

	/**
	 * Aendert die bevorzugte Hoehe einer GUI-Komponente.
	 * 
	 * @param comp
	 *            GUI-Komponente
	 * @param h
	 *            Hoehe
	 */
	public static void setPreferredHeight(Component comp, int h) {
		Dimension d = comp.getPreferredSize();
		d.height = h;
		comp.setPreferredSize(d);
	}

	/**
	 * Aendert die minimal erlaubte Breite einer GUI-Komponente.
	 * 
	 * @param comp
	 *            GUI-Komponente
	 * @param w
	 *            Breite
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static void setMinimumWidth(Component comp, int w) {
		Dimension d = comp.getMinimumSize();
		d.width = w;
		comp.setMinimumSize(d);
	}

	/**
	 * Aendert die minimal erlaubte Hoehe einer GUI-Komponente.
	 * 
	 * @param comp
	 *            GUI-Komponente
	 * @param h
	 *            Hoehe
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static void setMinimumHeight(Component comp, int h) {
		Dimension d = comp.getMinimumSize();
		d.height = h;
		comp.setMinimumSize(d);
	}

	/**
	 * Aendert die maximal erlaubte Breite einer GUI-Komponente.
	 * 
	 * @param comp
	 *            GUI-Komponente
	 * @param w
	 *            Breite
	 */
	public static void setMaximumWidth(Component comp, int w) {
		Dimension d = comp.getMaximumSize();
		d.width = w;
		comp.setMaximumSize(d);
	}

	/**
	 * Aendert die maximal erlaubte Hoehe einer GUI-Komponente.
	 * 
	 * @param comp
	 *            GUI-Komponente
	 * @param h
	 *            Hoehe
	 */
	public static void setMaximumHeight(Component comp, int h) {
		Dimension d = comp.getMaximumSize();
		d.height = h;
		comp.setMaximumSize(d);
	}

	/**
	 * Fixiert die Groesse einer GUI-Komponente.
	 * 
	 * @param comp
	 *            GUI-Komponente
	 * @param d
	 *            Ausmasse (Hoehe und Breite)
	 */
	public static void fixComponentSize(Component comp, Dimension d) {
		comp.setPreferredSize(d);
		comp.setMinimumSize(d);
		comp.setMaximumSize(d);
	}

	/**
	 * Fixiert die Groesse einer GUI-Komponente mit der aktuell bevorzugten
	 * Groesse.
	 * 
	 * @param comp
	 *            GUI-Komponente
	 */
	public static void fixComponentSize(Component comp) {
		fixComponentSize(comp, comp.getPreferredSize());
	}

	/***
	 * Setzt die Hintergrundfarbe einer Komponente und aller darin enthaltener
	 * Komponenten.
	 * 
	 * @param comp
	 *            Komponente
	 * @param color
	 *            neue Hintergrund-Farbe
	 */
	public static void setAllBackground(Component comp, Color color) {
		comp.setBackground(color);
		if (comp instanceof JFrame)
			setAllBackground(((JFrame) comp).getContentPane(), color);
		if (comp instanceof JScrollPane)
			setAllBackground(((JScrollPane) comp).getViewport(), color);
		else if (comp instanceof Container)
			for (Component innerComp : ((Container) comp).getComponents())
				setAllBackground(innerComp, color);
	}

	/**
	 * Versucht, aus einem String eine Farbe zu erstellen. Drei Moeglichekeiten
	 * gibt es fuer das Format des Strings:
	 * <ol>
	 * <li><code>"RGB(<i>red</i>,<i>green</i>,<i>blue</i>)"</code><br>
	 * wobei <i>red</i>,<i>green</i> und <i>blue</i> dezimale Werte zwischen 0
	 * und 255 sind.</li>
	 * <li>Der String stellt einen Integer-Wert im dezimalen, oktalen oder
	 * hexadezimalen Format dar, aus dem die 3 RGB-Werte extrahiert werden
	 * (siehe {@link Color#decode(String) Color.decode(..)}).</li>
	 * <li>Der String spezifiziert ein statisches Feld der Klasse {@link Color}.
	 * <br>
	 * z.B. steht <code>"RED"</code> fuer {@link Color#RED Color.RED},
	 * <code>"darkGray"</code> fuer {@link Color#darkGray Color.darkGray} oder
	 * <code>"LIGHT_GRAY"</code> fuer {@link Color#LIGHT_GRAY Color.LIGHT_GRAY}</li>
	 * </ol>
	 * 
	 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
	 *         (University of Bonn/Germany)
	 * 
	 * @throws IllegalArgumentException
	 *             wenn der uebergebene String nicht interpretiert werden kann.
	 * 
	 */
	// Schmeisst nur noch IllegalArgumentException, vorher Exception. (SK)
	// 21.08.2007
	public static Color parseColor(String colorStr)
			throws IllegalArgumentException {
		// Wenn String mit "RGB(" startet, Format RGB(<r>,<b>,<g>) parsen
		if (colorStr.toUpperCase().startsWith("RGB("))
			try {
				StringTokenizer tok = new StringTokenizer(colorStr.substring(4));
				int r = Integer.parseInt(tok.nextToken(",()\n"));
				int g = Integer.parseInt(tok.nextToken(",()\n"));
				int b = Integer.parseInt(tok.nextToken(",()\n"));
				return new Color(r, g, b);
			} catch (Exception err) {
				throw new IllegalArgumentException(
						"Invalid RGB-specification. RGB(<r>,<g>,<b>) expected.",
						err);
			}

		// Versucht, die Farbe als dezimalen, oktalen oder hexadezimalen Integer
		// einzulesen
		Long testLong = null;
		try {
			testLong = Long.decode(colorStr);
			return Color.decode(colorStr);
		} catch (NumberFormatException err) {
			// wenn 'testLong' ein gueltiger Long ist, handelt es sich zwar um
			// eine gueltige Zahl, aber kein passendes Integer-Format nicht!
			if (testLong != null)
				throw new IllegalArgumentException(
						colorStr
								+ " is not a valid RGB-Integer in decimal, octal or hexadecimal format.");
		}

		// Wenn auch kein Integer angegeben wurde, wird versucht, die Farbe
		// als Feld von java.awt.Color zu interpretieren
		try {
			return (Color) Color.class.getDeclaredField(colorStr).get(null);
		} catch (Exception err1) {
			try {
				// als letztes noch versuchen, die Farbe in Gross-Buchstaben zu
				// finden
				return (Color) Color.class.getDeclaredField(
						colorStr.toUpperCase()).get(null);
			} catch (Exception err2) {
				throw new IllegalArgumentException(
						colorStr
								+ " is not a valid color name. Name of a static field of class java.awt.Color expected.");
			}
		}
	}

	/**
	 * Versucht, aus einem String eine Farbe zu erstellen. Drei Moeglichekeiten
	 * gibt es fuer das Format des Strings:
	 * <ol>
	 * <li><code>"RGB(<i>red</i>,<i>green</i>,<i>blue</i>)"</code><br>
	 * wobei <i>red</i>,<i>green</i> und <i>blue</i> dezimale Werte zwischen 0
	 * und 255 sind.</li>
	 * <li>Der String stellt einen Integer-Wert im dezimalen, oktalen oder
	 * hexadezimalen Format dar, aus dem die 3 RGB-Werte extrahiert werden
	 * (siehe {@link Color#decode(String) Color.decode(..)}).</li>
	 * <li>Der String spezifiziert ein statisches Feld der Klasse {@link Color}.
	 * <br>
	 * z.B. steht <code>"RED"</code> fuer {@link Color#RED Color.RED},
	 * <code>"darkGray"</code> fuer {@link Color#darkGray Color.darkGray} oder
	 * <code>"LIGHT_GRAY"</code> fuer {@link Color#LIGHT_GRAY Color.LIGHT_GRAY}</li>
	 * </ol>
	 * 
	 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
	 *         (University of Bonn/Germany)
	 * 
	 * @throws IllegalArgumentException
	 *             wenn der uebergebene String nicht interpretiert werden kann.
	 * 
	 */
	public static String convertColorToHex(Color color) {
		int r = color.getRed();
		int g = color.getGreen();
		int b = color.getBlue();

		// Convert RGB to 2-digit Hex
		String rHex = Integer.toHexString(r);
		if (rHex.length() < 2)
			rHex = "0" + rHex;
		String gHex = Integer.toHexString(g);
		if (gHex.length() < 2)
			gHex = "0" + gHex;
		String bHex = Integer.toHexString(b);
		if (bHex.length() < 2)
			bHex = "0" + bHex;

		// combine RGB to Hex string
		return "#" + rHex + gHex + bHex;
	}

	/**
	 * Setzt das Label eine Componente neu. Macht nichts, falls {@code newLabel}
	 * oder {@code comp} den Wert {@code null} hat.
	 * 
	 * @param comp
	 *            {@link JLabel}, {@link AbstractButton}, {@link JDialog} oder
	 *            {@link Frame}
	 * @param newLabel
	 *            neue Beschriftung
	 * @exception UnsupportedOperationException
	 *                falls {@code comp} nicht unterstuetzt wird.
	 * 
	 */
	public static void resetCaption(Component comp, Object newLabel) {
		if (comp == null || newLabel == null)
			return;
		if (comp instanceof JLabel)
			((JLabel) comp).setText(newLabel.toString());
		else if (comp instanceof AbstractButton)
			((AbstractButton) comp).setText(newLabel.toString());
		else if (comp instanceof Frame)
			((Frame) comp).setTitle(newLabel.toString());
		else if (comp instanceof JDialog)
			((JDialog) comp).setTitle(newLabel.toString());
		else
			throw new UnsupportedOperationException(SwingUtil.class
					.getSimpleName()
					+ "resetCaption(.) can not be applied to "
					+ comp.getClass().getSimpleName());
	}

	/**
	 * Erstellt das Pattern fuer ein {@link NumberFormat}.
	 * 
	 * @param sample
	 *            Beispiel-Wert, der die Anzahl der dargestellten
	 *            Nachkomma-Stellen bestimmt
	 */
	public static String getNumberFormatPattern(double sample) {
		String format = "0";
		for (int i = 0; sample != (int) sample; i++) {
			if (i == 0)
				format += ".";
			sample *= 10;
			format += "0";
		}
		return format;
	}

	/**
	 * Erstellt das Pattern fuer ein {@link NumberFormat}.
	 * 
	 * @param digits
	 *            Anzahl der dargestellten Nachkomma-Stellen
	 */
	public static String getNumberFormatPattern(int digits) {
		String format = "0";
		for (int i = 0; i < digits; i++) {
			if (i == 0)
				format += ".";
			format += "0";
		}
		return format;
	}

	/**
	 * Copied from
	 * http://www.exampledepot.com/egs/javax.swing.tree/ExpandAll.html e1029.
	 * "Expanding or Collapsing all Nodes in a JTree Component".<br/>
	 * If expand is true,expands all nodes in the tree. Otherwise, collapses all
	 * nodes in the tree.
	 * 
	 * @param tree
	 *            {@link JTree} to expand or collapse
	 * @param expand
	 *            If <code>true</code> all nodfes will be expanded, otherwise
	 *            they will be collapsed.
	 */
	public static void expandAll(JTree tree, boolean expand) {
		TreeNode root = (TreeNode) tree.getModel().getRoot();

		// Traverse tree from root
		expandAll(tree, new TreePath(root), expand);
	}

	/**
	 * If expand is <code>true</code>, expands all nodes on the way down the
	 * path.
	 * 
	 * @param tree
	 *            {@link JTree} to expand or collapse
	 */
	public static void expandAll(JTree tree, TreePath path, boolean expand) {
		// Traverse children
		TreeNode node = (TreeNode) path.getLastPathComponent();
		if (node.getChildCount() >= 0) {
			for (Enumeration e = node.children(); e.hasMoreElements();) {
				TreeNode n = (TreeNode) e.nextElement();
				TreePath path2 = path.pathByAddingChild(n);
				expandAll(tree, path2, expand);
			}
		}

		// Expansion or collapse must be done bottom-up
		if (expand) {
			tree.expandPath(path);
		} else {
			tree.collapsePath(path);
		}
	}

	/**
	 * This method maximizes a frame; the iconified bit is not affected Taken
	 * from e564. Iconifying and Maximizing a Frame,
	 * http://www.exampledepot.com/egs/java.awt/frame_FrameIconify.html
	 */
	public static void maximize(Frame frame) {
		int state = frame.getExtendedState();

		// Set the maximized bits
		state |= Frame.MAXIMIZED_BOTH;

		// Maximize the frame
		frame.setExtendedState(state);
	}

	/**
	 * This method deiconifies a frame; the maximized bits are not affected.
	 */
	public static void deiconify(JFrame frame) {
		int state = frame.getExtendedState();

		// Clear the iconified bit
		state &= ~Frame.ICONIFIED;

		// Deiconify the frame
		frame.setExtendedState(state);

	}

	/**
	 * Allows to define a renderer and min- and max width of a column with just
	 * one line of code. Just for convenience.
	 * 
	 * @param table
	 *            A {@link JTable} to access the column model.
	 * @param modelIdx
	 *            column index in model counting
	 * @param cellRenderer
	 *            <code>null</code> allowed
	 * @param minwidth
	 *            <code>null</code> allowed
	 * @param prefwidth
	 *            <code>null</code> allowed
	 * @param maxwidth
	 *            <code>null</code> allowed
	 */
	public static void setColumnLook(JTable table, Integer modelIdx,
			TableCellRenderer cellRenderer, Integer minWidth,
			Integer preferredWidth, Integer maxWidth) {

		int idx = table.convertColumnIndexToView(modelIdx);
		if (cellRenderer != null)
			table.getColumnModel().getColumn(idx).setCellRenderer(cellRenderer);
		if (minWidth != null)
			table.getColumnModel().getColumn(idx).setMinWidth(minWidth);
		if (preferredWidth != null)
			table.getColumnModel().getColumn(idx).setPreferredWidth(
					preferredWidth);
		if (maxWidth != null)
			table.getColumnModel().getColumn(idx).setMaxWidth(maxWidth);

	}

	/**
	 * Erase an image. This is much faster than recreating a new
	 * {@link BufferedImage}.
	 */
	public static void clearImage(BufferedImage image, Color bgColor) {
		Graphics2D graphics = image.createGraphics();
		if (bgColor != null)
			graphics.setBackground(bgColor);
		graphics.clearRect(0, 0, image.getWidth(), image.getHeight());
//		Composite composite = graphics.getComposite();
//		graphics.setComposite(AlphaComposite.Clear);
//		Rectangle2D.Double rect = new Rectangle2D.Double(0, 0,
//				image.getWidth(), image.getHeight());
//		graphics.fill(rect);
//		graphics.setComposite(composite);
	}

	/**
	 * Clears the four rectengular areas around a given inner {@link Rectangle}.
	 * Clearing is faster than painting.
	 * 
	 * @param paintArea
	 *            The {@link Rectangle} that shall not be cleared.
	 * @param clearArea
	 *            The {@link Rectangle} that shall be cleared, except for where
	 *            it intersects with the paintArea.
	 */
	public static void clearAround(Graphics2D graphics, Rectangle paintArea,
			Rectangle clearArea, Color bgColor) {
		
		if (bgColor != null)
			graphics.setBackground(bgColor);

		// Clear left of the painted area if needed
		if (clearArea.getMinX() < paintArea.getMinX()) {
			graphics.clearRect((int) clearArea.getMinX(), (int) clearArea
					.getMinY(), (int) paintArea.getMinX(), (int) clearArea
					.getMaxY());
		}

		// Clear right of the painted area if needed
		if (clearArea.getMaxX() > paintArea.getMaxX()) {
			graphics.clearRect((int) paintArea.getMaxX(), (int) clearArea
					.getMinY(), (int) clearArea.getMaxX(), (int) clearArea
					.getMaxY());
		}

		// Clear above of the painted area if needed
		if (clearArea.getMinY() < paintArea.getMinY()) {
			graphics.clearRect((int) clearArea.getMinX(), (int) clearArea
					.getMinY(), (int) clearArea.getMaxX(), (int) paintArea
					.getMinY());
		}

		// Clear below of the painted area if needed
		if (clearArea.getMaxY() > paintArea.getMaxY()) {
			graphics.clearRect((int) clearArea.getMinX(), (int) paintArea
					.getMaxY(), (int) clearArea.getMaxX(), (int) clearArea
					.getMaxY());
		}

	}

}
