/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing;

import java.awt.Frame;

import schmitzm.lang.WorkingThread;

/**
 * Diese Klasse stellt einen Thread dar, der (aufwendigere) Arbeiten innerhalb einer
 * GUI ausfuehrt, damit diese nicht blockiert.<br>
 * Dem SwingWorker <b>kann</b> ein {@link StatusDialog} (modal!) zugeordnet werden.
 * Ist dies der Fall kehrt der Thread nicht aus seiner <code>start()</code>-Routine
 * zurueck sondern sperrt mit dem Dialog die zugrunde liegende GUI.
 * Ist die "Arbeit" beendet schliesst der Thread automatisch den Dialog, so dass
 * die GUI wieder zugeaenglich wird. Zudem kann die Arbeit des SwingWorker jederzeit
 * ueber den Abbruch-Button des {@link StatusDialog} beendet werden.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class SwingWorker extends WorkingThread {
  /** Speichert den Arbeitsablauf fuer den Thread. */
  protected Work         work = null;
  /** Speichert den Dialog, der waehrend der Arbeit des SwingWorkers angezeigt wird. */
  protected StatusDialog dialog = null;
  /** Speichert, ob der Dialog nach Beendigung der Arbeit offen gehalten soll
   *  (Default: <code>false</code>).
   *  @see #setKeepOpen(boolean) */
  protected boolean keepOpen = false;
  /** Speichert, ob der Worker (durch einen Fehler) abgebrochen wurde ({@code true})
   *  oder seine Arbeit erfolgreich ausgefuehrt hat ({@code false}). */
  protected boolean canceled = false;
  /** Resultat von {@link Work#execute()}. Kann nach Beendigung des Threads
   *  ueber {@link #getWorkResult()} abgefragt werden.*/
  protected Object workResult = null;

  /**
   * Erzeugt einen neuen SwingWorker. Waehrend der Arbeit wird kein Dialog
   * angezeigt. Der Thread kehrt unmittelbar aus seiner <code>start()</code>-Routine
   * zurueck, so dass eine untergeordnete GUI <b>nicht</b> blockiert.
   * @param work auszufuehrende Arbeit
   */
  public SwingWorker(Work work) {
    this(work,null);
  }

  /**
   * Erzeugt einen neuen SwingWorker. Der Thread kehrt nicht aus seiner
   * <code>start()</code>-Routine zurueck, sondern zeigt den (modalen) Dialog
   * an, so dass eine untergeordnete GUI gesperrt wird.
   * @param work   auszufuehrende Arbeit
   * @param dialog anzuzeigender Dialog (modal!)
   */
  public SwingWorker(Work work, StatusDialog dialog) {
    this.work   = work;
    this.dialog = dialog;
    // Worker in der Work-Klasse setzten, so dass waehrend der
    // Ausfuehrung der Arbeit auf den Worker (Dialog) zugegriffen werden
    // kann
    work.worker = this;
  }

  /**
   * Erzeugt einen neuen SwingWorker. Der Thread kehrt nicht aus seiner
   * <code>start()</code>-Routine zurueck, sondern zeigt einen (modalen) Dialog
   * an, so dass eine untergeordnete GUI gesperrt wird.
   * @param work       auszufuehrende Arbeit
   * @param parent     uebergeordnete GUI (kann <code>null</code> sein!)
   * @param statusMess Meldung, die im Dialog angezeigt wird
   */
  public SwingWorker(Work work, Frame parent, String statusMess) {
    this(work,parent,statusMess,0.5,0.5);
  }

  /**
   * Erzeugt einen neuen SwingWorker. Der Thread kehrt nicht aus seiner
   * <code>start()</code>-Routine zurueck, sondern zeigt einen (modalen) Dialog
   * an, so dass eine untergeordnete GUI gesperrt wird.
   * @param work       auszufuehrende Arbeit
   * @param parent     uebergeordnete GUI (kann <code>null</code> sein!)
   * @param statusMess Meldung, die im Dialog angezeigt wird
   * @param relX       horizontale Positionierung des Dialogs relativ zum uebergeordneten Fenster
   * @param relY       vertikale Positionierung des Dialogs relativ zum uebergeordneten Fenster
   */
  public SwingWorker(Work work, Frame parent, String statusMess, double relX, double relY) {
    this(work,
         new StatusDialog(
           parent,
           SwingUtil.RESOURCE.getString("WaitMess"),
           statusMess,
           relX,
           relY
         )
    );
  }

  /**
   * Bestimmt, ob der Dialog nach Beendigung der Arbeit geoeffnet bleibt.
   * Standardmaessig wird der Dialog automatisch geschlossen.
   * @param keepOpen wenn <code>false</code> wird der Dialog nach Beendigung
   *                 der Arbeit automatisch geschlossen.
   */
  public void setKeepOpen(boolean keepOpen) {
    this.keepOpen = keepOpen;
  }

  /**
   * Prueft, ob der Dialog nach Beendigung der Arbeit geoeffnet bleibt.
   * Standardmaessig wird der Dialog automatisch geschlossen.
   * @see #setKeepOpen(boolean)
   */
  public boolean isKeptOpen() {
    return this.keepOpen;
  }

  /**
   * Startet den Thread. Wurde ein Dialog angegeben/erzeugt wird dieser
   * nach dem Thread-Start angezeigt und blockiert den untergeordneten
   * Ablauf. Der Thread kehrt erst nach Beendigung seiner Arbeit oder wenn
   * der Dialog abgebrochen wurde, aus dieser Methode zurueck.
   */
  public void start() {
    super.start();
    if ( dialog != null ) {
      dialog.setVisible(true);
      if ( dialog.isCanceled() && isRunning() )
        terminate();
    }
  }

  /**
   * Fuehrt die dem Thread zugeordnete Arbeit aus und verarbeitet auftretende
   * Fehler.
   * @see Work#execute()
   * @see Work#performError(Throwable)
   */
  public void performWork() {
    try {
      this.canceled = false;
      this.workResult = work.execute();
    } catch (Throwable err) {
      // Fehler die zum Abbruch der "Arbeit" fuehren
      this.canceled = true;
      this.workResult = null;
      work.performError(err);
    }
  }

  /**
   * Initialisierung des Threads. Macht nichts!
   */
  public void performInit() {
  }

  /**
   * Wird ausgefuehrt nachdem der Thread beendet wurde. Schliesst einen evt.
   * vorhandenen Dialog, so dass eine untergeordnete GUI wieder freigegeben wird.
   */
  public void performDispose() {
    if (dialog != null) {
      if ( dialog.isCanceled() )
        this.canceled = true;
      if ( !keepOpen )
        dialog.setVisible(false);
      else
        dialog.setDialogOption( StatusDialog.OK_OPTION );
    }
  }

  /**
   * Bricht die Arbeit des Threads <b>unmittelbar</b> und <b>unkontrolliert</b>
   * ab.
   * @see Thread#stop()
   */
  public void terminate() {
    stop();
  }

  /**
   * Liefert den Dialog, der waehrend der Arbeit des Threads angezeigt wird.
   * @return <code>null</code> falls kein Dialog angegeben wurde.
   */
  public StatusDialog getDialog() {
    return dialog;
  }

  /**
   * Liefert {@code true}, wenn der Worker (aufgrund eines Fehlers oder
   * manuell durch den Dialog) abgebrochen wurde.
   */
  public boolean isCanceled() {
    return canceled;
  }

  /**
   * Liefert nach der Beendigung des Threads dessen Ergebnis.
   * @see Work#execute()
   */
  public Object getWorkResult() {
    return this.workResult;
  }

  /**
   * Diese Klasse spezifiziert die Arbeit eines {@link SwingWorker}.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static abstract class Work {
    // speichert den SwingWorker, der die Arbeit verrichtet
    // WIRD VON DER AUSFUEHRENDEN SWINGWORKER-KLASSE GESETZT!!
    private SwingWorker worker = null;

    /**
     * Implementiert den Arbeitsablauf des {@link SwingWorker}.
     * @return "irgendwas" als Ergebnis
     */
    public abstract Object execute() throws Exception;

    /**
     * Implementiert die Fehlerbehandlung, wenn waehrend des Arbeitsablauf
     * ein Fehler auftritt. In dieser Basis-Implementierung wird lediglich
     * eine Fehlermeldung angezeigt.<br>
     * Kann gefahrlos von Unterklassen ueberschrieben werden.
     */
    public void performError(Throwable err) {
      ExceptionDialog.show(worker.dialog,err);//,SwingResource.DEFAULT.getString("Error"),err.getMessage());
    };

    /**
     * Liefert den {@link SwingWorker}, der die Arbeit verrichtet.
     */
    public SwingWorker getSwingWorker() {
      return worker;
    }
  }
}
