/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing.event;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.Vector;

/**
 * This class provides a super class for all classes which initiates
 * {@link PropertyChangeEvent PropertyChangeEvents} to
 * {@link PropertyChangeListener PropertyChangeListeners}. 
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 */
public abstract class PropertyChangeEmitter<E extends PropertyChangeListener> {
  /** Holds the connected listeners. */
  protected Vector<E> listenerList = new Vector<E>();
  /** Flag to (temporary) disable firing events. Useful if several changes
   *  should be performed and fired collectively. */
  protected boolean eventFiringEnabled = true;

  /**
   * Adds a listener to the listener list.
   * @param listener a listener
   */
  public void addListener(E listener) {
    listenerList.add(listener);
  }
  
  /**
   * Removes a listener from the listener list.
   * @param listener a listener
   */
  public boolean removeListener(E listener) {
    return listenerList.remove(listener);
  }
  
  /**
   * Fires the {@link PropertyChangeEvent} to all connected
   * {@link PropertyChangeListener PropertyChangeListeners} except
   * the specified ones. Does nothing if {@link #eventFiringEnabled} is
   * set to {@code false}.
   * @param e an event
   * @param exceptListeners listeners which are not notified (useful to avoid
   *                        event circles)
   */
  protected void firePropertyChangeEvent(PropertyChangeEvent e, PropertyChangeListener... exceptListeners) {
    if ( !eventFiringEnabled )
      return;

    Vector<PropertyChangeListener> exceptList = new Vector<PropertyChangeListener>();
    for (PropertyChangeListener l : exceptListeners)
      exceptList.add(l);
    
    for (PropertyChangeListener listener : listenerList){
      if ( !exceptList.contains(listener))
        listener.propertyChange(e);
    }
  }

}
