/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/

package schmitzm.swing.event;

import java.awt.Point;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.util.HashMap;
import java.util.Map;

/**
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 *
 */
public class SelectiveMouseAdapter extends MouseAdapter {
  /** Holds the activation of the several {@link MouseInputType MouseInputTypes}. */
  protected Map<MouseInputType, Boolean> activationMap = new HashMap<MouseInputType, Boolean>();
  /** Holds the button the drag is performed with. */
  protected int dragButton = MouseEvent.NOBUTTON;
  /** Holds the window position a drag is started on. */
  protected Point dragStartPos = null;
  /** Holds the window position of last {@link #mouseDragged(MouseEvent)}
   *  call (e.g. to remove a dragging rectangle). */
  protected Point dragLastPos = null;
  
  /**
   * Creates a new mouse adapter. All actions are activated initially.
   */
  public SelectiveMouseAdapter() {
    for ( MouseInputType type : MouseInputType.values() )
      setEnabled(type, true);
  }
  
  /**
   * Sets whether a specific input action of the mouse adapter is
   * enabled.
   * @param type mouse input type 
   * @param enabled {@code true} to enable the action
   */
  public void setEnabled(MouseInputType type, boolean enabled) {
    activationMap.put(type, enabled);
  }

  /**
   * Returns whether a specific input action of the mouse adapter is
   * enabled.
   * @param type mouse input type 
   */
  public boolean isEnabled(MouseInputType type) {
    Boolean enabled = activationMap.get(type);  
    return enabled != null && enabled;
  }
  
  /**
   * Returns the start point of a drag or window selection.
   */
  public Point getDragStartPoint() {
    return dragStartPos;
  }
  
  /**
   * Returns the window selection of the last {@link #mouseDragged(MouseEvent)}
   * call.
   */
  public Point getDragLastPoint() {
    return dragLastPos;
  }

  /**
   * Returns the button a drag is performed with.<br>
   * <b>Note:</b> This method only returns the button DURING a drag.
   * If no drag is performed {@link MouseEvent#NOBUTTON} is returned!
   * call.
   */
  public int getDragButton() {
    return dragButton;
  }

  /**
   * Calls {@link #performMouseClicked(MouseEvent)} if the mouse
   * action is enabled for the click.<br>
   * For actions which can not be specified by {@link MouseInputType} (e.g. 3rd
   * mouse button) the perform-method is called generally.<br>
   * In any case the drag start position is initialized with {@code null}
   * at the end of this method.
   */
  @Override
  public final void mouseClicked(MouseEvent e) {
    if ( isMouseClickEnabled(e) )
      performMouseClicked(e);
    dragStartPos = null;
    dragLastPos = null;
  }

  /**
   * Calls {@link #performMouseDragged(MouseEvent)} if the mouse
   * action is enabled for the drag.<br>
   * For actions which can not be specified by {@link MouseInputType} (e.g. 3rd
   * mouse button) the perform-method is called generally.
   */
  @Override
  public final void mouseDragged(MouseEvent e) {
    if ( isMouseDragEnabled(e) )
      performMouseDragged(e);
    // remember the position of this call
    dragLastPos = e.getPoint();
  }
  
  /**
   * Calls {@link #performMouseDragged(MouseEvent)} in any case, because
   * enter, exit and move can not be specified by {@link MouseInputType}.
   */
  @Override
  public final void mouseEntered(MouseEvent e) {
    performMouseEntered(e);
  }
  
  /**
   * Calls {@link #performMouseExited(MouseEvent)} in any case, because
   * enter, exit and move can not be specified by {@link MouseInputType}.
   */
  @Override
  public final void mouseExited(MouseEvent e) {
    performMouseExited(e);
  }

  /**
   * Calls {@link #performMouseExited(MouseEvent)} in any case, because
   * enter, exit and move can not be specified by {@link MouseInputType}.
   */
  @Override
  public final void mouseMoved(MouseEvent e) {
    performMouseMoved(e);
  }

  
  /**
   * Stores the current mouse location in {@link #dragStartPos}.
   * {@link #performMousePressed(MouseEvent)} is only called, if
   * the respective click action is enabled.<br>
   * For actions which can not be specified by {@link MouseInputType} (e.g. 3rd
   * mouse button) the perform-method is called generally.
   */
  @Override
  public final void mousePressed(MouseEvent e) {
    dragStartPos = e.getPoint();
    dragLastPos  = null;
    dragButton   = e.getButton();
    if ( isMouseClickEnabled(e) )
      performMousePressed(e);
  }

  /**
   * Calls {@link #performMouseReleased(MouseEvent)} only in the
   * following cases
   * <ul>
   *   <li>Startposition = Endposition and the mouse click action is activated</li>
   *   <li>Startposition != Endposition and the mouse drag or window action is activated</li>
   * </ul>
   * In any case the drag start position is initialized with {@code null}
   * at the end of this method.
   */
  @Override
  public final void mouseReleased(MouseEvent e) {
    Point dragEndPos = e.getPoint();
    
    if (  dragEndPos.equals(dragStartPos) ) {
      // Startposition = Endposition
      // --> nur wenn Klick aktiviert
      if ( isMouseClickEnabled(e) )
        performMouseReleased(e);
    } else {
      // Startposition <> Endposition
      // --> nur wenn Drag aktiviert
      if ( isMouseDragEnabled(e)  )
        performMouseReleased(e);
    }
  
    dragStartPos = null;
    dragLastPos  = null;
    dragButton   = MouseEvent.NOBUTTON;
  }

  /**
   * Calls {@link #performMouseWheelMoved(MouseWheelEvent)} if the mouse
   * action is enabled for mouse wheel events.<br>
   */
  @Override
  public final void mouseWheelMoved(MouseWheelEvent e) {
    if ( isEnabled(MouseInputType.Wheel) )
      performMouseWheelMoved(e);
  }

  /**
   * Empty implementation.  
   */
  public void performMouseClicked(MouseEvent e) {
  }

  /**
   * Empty implementation.  
   */
  public void performMouseDragged(MouseEvent e) {
  }
  
  /**
   * Empty implementation.  
   */
  public void performMouseEntered(MouseEvent e) {
  }
  
  /**
   * Empty implementation.  
   */
  public void performMouseExited(MouseEvent e) {
  }

  /**
   * Empty implementation.  
   */
  public void performMouseMoved(MouseEvent e) {
  }

  /**
   * Empty implementation.  
   */
  public void performMousePressed(MouseEvent e) {
  }

  /**
   * Empty implementation.  
   */
  public void performMouseReleased(MouseEvent e) {
  }

  /**
   * Empty implementation.  
   */
  public void performMouseWheelMoved(MouseWheelEvent e) {
  }

  /**
   * Checks whether the click is enabled for a mouse event.
   */
  private boolean isMouseClickEnabled(MouseEvent e) {
    switch ( e.getButton() ) {
      case MouseEvent.BUTTON1: return isEnabled(MouseInputType.LClick);
      case MouseEvent.BUTTON3: return isEnabled(MouseInputType.RClick);
    }
    return e.getButton() != MouseEvent.NOBUTTON;
  }
  
  /**
   * Checks whether dragging is enabled for a mouse event.
   */
  private boolean isMouseDragEnabled(MouseEvent e) {
    switch ( dragButton ) {
      case MouseEvent.BUTTON1: return isEnabled(MouseInputType.LDrag);
      case MouseEvent.BUTTON3: return isEnabled(MouseInputType.RDrag);
    }
    return dragButton != MouseEvent.NOBUTTON;
  }

  
}


