/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing.event;

import java.awt.Dialog;
import java.awt.Frame;
import java.awt.Window;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

/**
 * Dieser {@link WindowAdapter} verbindet ein (Unter-)Fenster mit den Aktionen
 * eines anderen (Haupt-)Fensters.<br>
 * Wird das Haupt-Fenster geschlossen oder minimiert, wird auch das verbundene
 * Unter-Fenster verborgen. Wird das Haupt-Fenster geoeffnet, wird auch das
 * Unter-Fenster angezeigt, sofern es beim Schliessen des Haupt-Fensters
 * geoeffnet war.<br>
 * Fuer jedes Unter-Fenster wird ein <code>WindowEventConnector</code> erzeugt
 * und als {@link Window#addWindowListener WindowListener} an das Haupt-Fenster
 * gekoppelt.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class WindowEventConnector extends WindowAdapter {
  /** Kontrolliertes Unter-Fenster. */
  protected Window    window     = null;
  /** Speichert den Status des Unter-Fensters, wenn das Haupt-Fenster geschlossen
   *  oder minimiert wird.*/
  protected boolean   wasVisible = false;

  /**
   * Erzeugt einen neuen Connector.
   * @param window durch den Connector gesteuertes (Unter-)Fenster
   */
  public WindowEventConnector(Window window) {
    super();
    this.window = window;
  }

  /**
   * Wird aufgerufen, wenn das Haupt-Fenster geschlossen wird. Bevor das Unter-Fenster
   * ebenfalls geschlossen wird, wird sein Status gesichtert.
   * @see #wasVisible
   */
  public void windowClosed(WindowEvent e) {
    wasVisible = window.isVisible();
    window.setVisible(false);
  }

  /**
   * Wird aufgerufen, wenn das Haupt-Fenster ueber das System-Menue geschlossen
   * wird. Bevor das Unter-Fenster ebenfalls geschlossen wird, wird sein Status
   * gesichert.
   * @see #wasVisible
   */
  public void windowClosing(WindowEvent e) {
    windowClosed(e);
  }

  /**
   * Wird aufgerufen, wenn das Haupt-Fenster geoeffnet wird. War das Unter-Fenster
   * zuvor geoeffnet, wird es ebenfalls wieder angezeigt.
   */
  public void windowOpened(WindowEvent e) {
    if ( !window.isVisible() )
      window.setVisible(wasVisible);
  }

// KLAPPT NICHT WIE GEWUENSCHT
//  /**
//   * Wird aufgerufen, wenn das Haupt-Fenster aktiviert wird.
//   */
//  public void windowActivated(WindowEvent e) {
//    windowOpened(e);
//  }

  /**
   * Wird aufgerufen, wenn das Haupt-Fenster minimiert wird. Handelt es sich
   * bei dem Unter-Fenster um einen {@link Frame}, wird auch dieses minimiert.
   * Andernfalls (z.B. bei einem {@link Dialog}) wird das Unter-Fenster
   * verborgen.
   */
  public void windowIconified(WindowEvent e) {
    if ( window instanceof Frame )
      ((Frame)window).setState(Frame.ICONIFIED);
    else
      windowClosed(e);
  }

  /**
   * Wird aufgerufen, wenn das Haupt-Fenster "deminimiert" wird. Handelt es sich
   * bei dem Unter-Fenster um einen {@link Frame}, wird auch dieses "deminimiert".
   * Andernfalls (z.B. bei einem {@link Dialog}) wird das Unter-Fenster
   * wieder angezeigt.
   */
  public void windowDeiconified(WindowEvent e) {
    if ( window instanceof Frame )
      ((Frame)window).setState(Frame.NORMAL);
    else
      windowOpened(e);
  }

}
