/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing.log4j;

import java.util.Comparator;
import java.util.Enumeration;
import java.util.TreeSet;
import java.util.Vector;

import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableModel;

import org.apache.log4j.Appender;
import org.apache.log4j.Level;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;

/**
 * Diese Klasse stellt ein {@link TableModel} dar, in dem die aktuell
 * im {@link LogManager} registrierten Log4j-Logger in 4 Spalten angezeigt werden.
 * <ol start=0>
 *   <li>{@linkplain Logger#getName() Logger-Name} (Typ {@link String})</li>
 *   <li>{@linkplain Logger#getLevel() Logger-Level} (Typ {@link Level})</li>
 *   <li>{@linkplain Logger#getAdditivity() Logger-Additiviaet (Rekursive Verarbeitung der Vater-Appender)} (Typ {@link Boolean})</li>
 *   <li>Appender-Liste (Typ {@link Appender Enumeration&lt;Appender&gt;})</li>
 * </ol>
 * <b><u>Beachte</u>:</b><br>
 * Nur die Spalten 1 (Level) und 2 (Additivitaet) sind als veraenderbar vorgesehen!
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class LoggerConfigurationTableModel extends AbstractTableModel {
  /** Spalten-Namen der Tabelle. */
  public final String[] COL_NAME = new String[] {"Logger","Level","Additivity","Appender list"};
  /** Spalten-Typen der Tabelle.
   *  <ol start=0>
   *    <li>{@link String}</li>
   *    <li>{@link Level}</li>
   *    <li>{@link Boolean}</li>
   *    <li>{@link Enumeration}</li>
   *  </ol> */
  public final Class[] COL_TYPE = new Class[] {String.class,Level.class,Boolean.class,Enumeration.class};

  /** {@link LogManager#getCurrentLoggers()} liefert eine {@link Enumeration}.
   *  Fuer einen wahlfreien Zugriff, werden die Logger bei der {@link #refresh()}-Aktion
   *  in ein {@link Vector} kopiert. */
  protected Vector<Logger> loggers = new Vector<Logger>();

  /**
   * Erzeugt ein neues Tabellenmodell.
   */
  public LoggerConfigurationTableModel() {
    super();
    refresh();
  }

  /**
   * Aktualisiert die Tabelle auf Basis der aktuelle registrierten
   * Logger.
   * @see LogManager#getCurrentLoggers()
   */
  public void refresh() {
    loggers.clear();
    loggers.add( LogManager.getRootLogger() );
    Enumeration<Logger> newLoggers = LogManager.getCurrentLoggers();
    // Fuer die Sortierung die Logger in ein TreeSet einfuegen
    TreeSet<Logger> sortedLoggers = new TreeSet<Logger>( new Comparator<Logger>() {
      public int compare(Logger logger1, Logger logger2) {
        return logger1.getName().compareTo( logger2.getName() );
      }
    });
    for (; newLoggers != null && newLoggers.hasMoreElements(); )
      sortedLoggers.add( newLoggers.nextElement() );

    // Die sortierten Logger der Reihe nach in die Logger-Liste einfuegen
    for (; !sortedLoggers.isEmpty();) {
      loggers.add(sortedLoggers.first());
      sortedLoggers.remove( sortedLoggers.first() );
    }
    this.fireTableDataChanged();
  }

  /**
   * Liefert die Anzahl an Zeilen (Loggern) der Tabelle.
   */
  public int getRowCount() {
    return loggers.size();
  }

  /**
   * Liefert die Spaltenanzahl der Tabelle. Immer 3.
   */
  public int getColumnCount() {
    return COL_NAME.length;
  }

  /**
   * Liefert einen Spaltennamen der Tabelle.
   */
  public String getColumnName(int col) {
    return COL_NAME[col];
  }

  /**
   * Liefert den Typ einer Tabellen-Spalte.
   */
  public Class getColumnClass(int col) {
    return COL_TYPE[col];
  }

  /**
   * Liefert einen Zellenwert der Tabelle.
   * @param row Zeilenindex (beginnend bei 0)
   * @param col Spaltenindex (beginnend bei 0)
   */
  public Object getValueAt(int row, int col) {
    Logger logger = loggers.elementAt(row);
    switch (col) {
      case 0: return logger.getName();
      case 1: return logger.getLevel();
      case 2: return logger.getAdditivity();
      case 3: return logger.getAllAppenders();
    }
    return null;
  }

  /**
   * Prueft, ob eine Zeile editierbar ist.
   * @param row Zeilenindex (beginnend bei 0)
   * @param col Spaltenindex (beginnend bei 0)
   * @return {@code true} nur fuer Spalte 1 (Logger-Level) und Spalte 3
   * (Additivitaet)
   */
  public boolean isCellEditable(int row, int col) {
    return col == 1 || col == 2;
  }

  /**
   * Setzt einen Zellenwert der Tabelle. Nur relevant fuer die Spalten 1 (Level)
   * und 2 (Additivitaet).
   * @param value neue Wert
   * @param row Zeilenindex (beginnend bei 0)
   * @param col Spaltenindex (beginnend bei 0)
   */
  public void setValueAt(Object value, int row, int col) {
    Logger logger = loggers.elementAt(row);
    switch (col) {
      case 1: logger.setLevel( (Level)value );
              break;
      case 2: logger.setAdditivity( (Boolean)value );
              break;
    }
  }
}
