/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing.log4j;

import java.awt.Component;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.util.Enumeration;

import javax.swing.DefaultCellEditor;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JFrame;
import javax.swing.JScrollPane;
import javax.swing.JTable;

import org.apache.log4j.Appender;
import org.apache.log4j.ConsoleAppender;
import org.apache.log4j.FileAppender;
import org.apache.log4j.Level;
import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.apache.log4j.PropertyConfigurator;
import org.apache.log4j.spi.LoggerRepository;

import schmitzm.swing.ExceptionDialog;
import schmitzm.swing.FileInputOption;
import schmitzm.swing.SwingUtil;
import schmitzm.swing.table.ComponentRenderer;

/**
 * Diese Klasse stellt ein Fenster dar, in dem die aktuell im {@link LogManager}
 * registrierten Log4j-Logger als Tabelle dargestellt sind.
 * <ol>
 *   <li>Logger-Name</li>
 *   <li>Level</li>
 *   <li>Additivitaet (ob rekursiv auch die Vater-Appender ausgefuehrt werden)</li>
 *   <li>Dateinamen der {@link FileAppender}, "CONSOLE" fuer {@link ConsoleAppender}</li>
 * </ol>
 * Ueber einen {@linkplain #refresh() Refresh-Button} kann die Tabelle jederzeit aktualisiert
 * werden. Es koennen <b>keine</b> Logger hinzugefuegt oder entfernt werden.
 * Veraendert werden koennen die Logger nur hinsichtlich ihres Levels und ihrer
 * Addititvitaet.<br>
 * Die Eingabe-Maske bietet jedoch die Moeglichkeit, eine Logger-Konfiguration
 * aus einer {@linkplain PropertyConfigurator#doConfigure(String,LoggerRepository) Log4j-Konfigurationsdatei}
 * (neu) zu laden.<br>
 * <b><u>Beachte</u>:</b><br>
 * Beim Neu-Laden werden lediglich die <u>explizit</u> in der Datei spezifizierten
 * Logger-Konfigurationen aktualisiert. Vom der Applikation erzeugte Logger bleiben
 * erhalten (und koennen auch nicht mehr geloescht werden)!
 * Ebenso bleiben alle Logger-Properties unveraendert erhalten, die nicht
 * <u>explizit</u> in der Konfigurationsdatei angegeben sind. Wird z.B. die
 * Additivitaet eines Loggers in der Datei nicht gesetzt, wird sie beim ersten
 * Einlesen per Default auf {@code true} gesetzt. Wird sie durch den Anwender
 * abgeaendert, bleibt diese Einstellung erhalten, auch wenn die Datei neu
 * eingelesen wird.
 * @see LoggerConfigurationTableModel
 * @see Logger
 * @see LogManager
 * @see PropertyConfigurator#configure(String)
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class LoggerFrame extends JFrame {
  /** Eingabe-Feld fuer die Log4j-Konfigurationsdatei. */
  protected FileInputOption configFileField =  new FileInputOption(null,false);
  /** Butten, ueber den die Log4j-Konfigurationsdatei neu geladen wird.
   *  @see #reload() */
  protected JButton reloadButton = new JButton( SwingUtil.RESOURCE.getString("Reload") );
  /** Butten, ueber den die Logger-Tabelle aktualisiert wird.
   *  @see #refresh() */
  protected JButton refreshButton = new JButton( SwingUtil.RESOURCE.getString("Refresh") );
  /** Tabelle, in der die Logger angezeigt werden.
   *  @see #loggerTableModel */
  protected JTable loggerTable = null;
  /** Datenbasis fuer die Logger-Tabelle. */
  protected LoggerConfigurationTableModel loggerTableModel = null;

  /**
   * Erzeugt ein neues Fenster.
   * @param file vorgeblendete Log4j-Konfigurationsdatei
   */
  public LoggerFrame(File file) {
    super("Log4j Logger configuration");
    initGUI();
    configFileField.setValue(file);
  }

  /**
   * Erzeugt ein neues Fenster.
   */
  public LoggerFrame() {
    this(null);
  }

  /**
   * Initialisiert das Fenster-Layout.
   */
  protected void initGUI() {
    this.setLayout( new GridBagLayout() );
    this.setSize(400,300);
    // Logger-Tabelle
    loggerTableModel = new LoggerConfigurationTableModel();
    loggerTable = new JTable(loggerTableModel);
    loggerTable.setCellSelectionEnabled(false);
    loggerTable.setRowHeight( new JComboBox().getPreferredSize().height );
    loggerTable.setDefaultRenderer( Enumeration.class, new ComponentRenderer() {
      public Component createRendererComponent( JTable table,
                                                Object value,
                                                boolean isSelected,
                                                boolean hasFocus,
                                                int row,
                                                int column) {

        // Appender in String umwandeln
        if ( value instanceof Enumeration )
          value = getStringFromAppenders((Enumeration<Appender>)value);
        return super.createRendererComponent(table,value,isSelected,hasFocus,row,column);
      }
    });
    loggerTable.setDefaultRenderer( Boolean.class, new ComponentRenderer.JCheckBox());
    loggerTable.setDefaultEditor( Boolean.class, new DefaultCellEditor( new JCheckBox() {
      public int getHorizontalAlignment() {
        return this.CENTER;
      }
      public int getVerticalAlignment() {
        return this.CENTER;
      }
    }));

    loggerTable.setDefaultRenderer( Level.class, new ComponentRenderer.JComboBox() );
    loggerTable.setDefaultEditor( Level.class, new DefaultCellEditor( new JComboBox( new Object[] {
        Level.ALL,
        Level.DEBUG,
        Level.ERROR,
        Level.FATAL,
        Level.INFO,
        Level.OFF,
        Level.WARN
    })));

    // Button-Aktionen
    refreshButton.addActionListener( new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        refresh();
      }
    });
    reloadButton.addActionListener( new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        reload();
      }
    });

    // Componenten in GUI einbinden
    this.getContentPane().add( reloadButton,                   new GridBagConstraints(0,0,1,1,0.0,0.0,GridBagConstraints.WEST,GridBagConstraints.NONE,new Insets(5,5,5,5),0,0) );
    this.getContentPane().add( configFileField,                new GridBagConstraints(1,0,1,1,1.0,0.0,GridBagConstraints.WEST,GridBagConstraints.HORIZONTAL,new Insets(5,5,5,5),0,0) );
    this.getContentPane().add( new JScrollPane( loggerTable ), new GridBagConstraints(0,1,2,1,1.0,1.0,GridBagConstraints.CENTER,GridBagConstraints.BOTH,new Insets(5,5,5,5),0,0) );
    this.getContentPane().add( refreshButton,                  new GridBagConstraints(0,2,1,1,0.0,0.0,GridBagConstraints.EAST,GridBagConstraints.NONE,new Insets(5,5,5,5),0,0) );
  }

  /**
   * Aktualisiert die Logger auf Basis der Konfigurationsdatei. Dies entspricht
   * der Aktion des {@linkplain #reloadButton Reload-Button}.
   */
  public void reload() {
    if ( loggerTableModel != null && configFileField.getValue() != null ) {
      try {
        String configFileName = configFileField.getValue().getAbsolutePath();
        PropertyConfigurator.configure(configFileName);
        refresh();
      } catch (Exception err) {
        ExceptionDialog.show(this,err);
      }
    }
  }

  /**
   * Aktualisiert die Logger-Tabelle auf Basis der aktuelle geladenen
   * Log4j-Logger.  Dies entspricht der Aktion des
   * {@linkplain #refreshButton Refresh-Button}.
   * @see LoggerConfigurationTableModel#refresh()
   */
  public void refresh() {
    if ( loggerTableModel != null ) {
      loggerTableModel.refresh();
    }
  }

  /**
   * Liefert einen String, der die {@link Appender} eines Loggers darstellt.
   * <ul>
   *   <li>{@link FileAppender} werden durch den Dateinamen dargestellt</li>
   *   <li>{@link ConsoleAppender} werden durch das Schluesselwort "CONSOLE" dargestellt</li>
   *   <li>Alle anderen {@link Appender} werden NICHT dargestellt</li>
   * </ul>
   * Sub-Klassen koennen diese Methode ueberschreiben, um auch andere
   * Appender-Arten darzustellen.
   * @param appenders {@link Appender}-Liste
   */
  public String getStringFromAppenders(Enumeration<Appender> appenders) {
    StringBuffer buffer = new StringBuffer();
    for(;appenders != null && appenders.hasMoreElements();) {
      Appender appender = appenders.nextElement();
      if ( appender instanceof ConsoleAppender )
        buffer.append("CONSOLE");
      if ( appender instanceof FileAppender )
        buffer.append( ((FileAppender)appender).getFile() );
      if ( appenders.hasMoreElements() )
        buffer.append("; ");
    }
    return buffer.toString();
  }

//    /**
//     * Ermittelt  einen String, der die {@link Appender} eines Loggers darstellt.
//     * {@link FileAppender} werden durch den Dateinamen dargestellt und
//     * {@link ConsoleAppender} durch das Schluesselwort "CONSOLE".
//     * @param appenders {@link Appender}-Liste
//     */
//    private void setAppendersFromString(String appendersStr, Logger logger) {
//      logger.removeAllAppenders();
//      StringTokenizer tok = new StringTokenizer(appendersStr);
//      for(;tok.hasMoreTokens();) {
//        String appenderStr = tok.nextToken(" ;\t\n");
//        Appender appender = null;
//        if ( appenderStr.equalsIgnoreCase("CONSOLE") )
//          appender = new ConsoleAppender();
//        else {
//          appender = new RollingFileAppender();
//          ((RollingFileAppender)appender).setFile(appenderStr);
//        }
//        logger.addAppender( appender );
//      }
//    }
}
