/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing.menu;

import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.util.Vector;

import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;

/**
 * <hr>
 * <small><i>Diese Klasse wurde urspruenglich 2002/2003 von <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *            fuer die ADAGIOS-Java-Library (Universitaet Bonn) erstellt, 2009 in das
 *            SCHMITZM-Projekt uebernommen und teilweise erweitert.</i></small>
 * <hr>
 * Diese Klasse erweitert das <code>JPopupMenu</code>.
 * Menueeintraege (Instanzen von <code>JMenuItem</code>) werden automatisch
 * in der <code>add(String)</code>-Methode erzeugt und dem Menue hinzugefuegt.
 * Es muss nicht jedem Menueeintragen einzeln ein eigener <code>ActionListener</code>
 * hinzugefuegt werden, um auf eine Aktion zu reagieren. Das
 * <code>ConnectedPopupMenu</code> selbst "achtet" auf eine Aktion und
 * informiert ueber <b>einen</b> <code>ActionListener</code> die verbundenen
 * Anwendungen. In deren <code>actionPerformed(..)</code>-Methode kann die
 * Anwendung ueber <code>getProcessedItem()</code> ermittelt werden, welcher
 * Menueeintrag gewaehlt wurde.
 * @see javax.swing.JPopupMenu
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 */
public class ConnectedPopupMenu extends JPopupMenu implements MouseListener, ActionListener {
  private Object popupClickSource = null;
  private Vector listeners        = new Vector();
  private int    processedItem    = -1;

  /**
   * Fuegt dem Menue einen <code>ActionListener</code> hinzu.
   * Dieser wird informiert, sobald ein Menueeintrag des PopupMenue
   * ausgewaehlt wurde.
   */
  public void addActionListener(ActionListener listener) {
    listeners.add( listener );
  }

  /**
   * Entfernt einen <code>ActionListener</code>.
   */
  public void removeActionListener(ActionListener listener) {
    listeners.remove( listener );
  }

  /**
   * Wird aufgerufen, sobald eine Maustaste auf einer mit dem Menue
   * verbundene Komponente gedrueckt wurde.
   * Die Methode checkt, ob es sich bei dem Klick um einen PopupTrigger
   * (i.A. Rechtsklick) handelt.
   * Ist dies der Fall, wird das PopupMenue geoeffnet.
   */
  public void mousePressed(MouseEvent e) {
    checkPopupSignal(e);
  }

  /**
   * Wird aufgerufen, sobald eine Maustaste auf einer mit dem Menue
   * verbundene Komponente losgelassen wurde.
   * Die Methode checkt, ob es sich bei dem Loslassen um einen PopupTrigger
   * (i.A. Rechtsklick) handelt.
   * Ist dies der Fall, wird das PopupMenue geoeffnet.
   */
  public void mouseReleased(MouseEvent e) {
    checkPopupSignal(e);
  }

  /**
   * Checkt, ob ein PopupTrigger (i.A. Rechtsklick) auf eine mit dem
   * PopupMenue verbundene Komponente stattgefunden hat.
   * Ist dies der Fall, wird das PopupMenue geoeffnet.
   */
  private void checkPopupSignal(MouseEvent e) {
    if ( e.isPopupTrigger() ) {
      this.show( (Component)e.getSource(), e.getX(), e.getY() );
      popupClickSource = e.getSource();
      e.consume();
    }
  }

  /**
   * Liefert den Index des Menueeintrags, der ausgewaehlt wurde.
   * Ist nur zugaenglich, waehrend ein <code>ActionListerner</code> ueber
   * die Auswahl eines Menueeintrags informiert wird!! Zu allen anderen
   * Zeitpunkten liefert diese Methode -1.
   */
  public int getProcessedItem() {
    return processedItem;
  }

  /**
   * Checkt, auf welchen Menueeintrag geklickt wurde und informiert die
   * angeschlossenen <code>ActionListener</code> ueber das Ereignis.<br>
   * Diese koennen (in ihrer <code>actionPerformed</code>-Methode ueber die
   * Methode <code>getProcessedItem()</code> ermitteln, auf welchen
   * Menueeintrag geklickt wurde.<br>
   * Das an die <code>ActionListener</code> weiter gereichte
   * <code>ActionEvent</code> erhaelt als Quelle (<code>getSource()</code>)
   * dasjenige Objekt, auf dem das PopupMenue (mittels Linksklick)
   * geoeffnet wurde (nicht das <code>ConnectedPopupMenu</code>!!)
   * @see #addMouseListener(MouseListener)
   */
  public void actionPerformed(ActionEvent e) {
    // herausfinden, weches Menueitem geklickt wurde
    for (int i=0; i<this.getComponentCount(); i++) {
      if ( this.getComponent(i) == e.getSource() ) {
        processedItem = i;
        for (int j=0; j<listeners.size(); j++)
          ((ActionListener)listeners.elementAt(j)).actionPerformed( new ActionEvent( popupClickSource, e.getID(), "" ) );
        processedItem = -1;
        break;
      }
    }
  }

  /**
   * Verbindet das Popup-Menue mit einer Komponente.
   * Beim (Rechts-)Klick auf diese Komponente wird das Popup-Menue geoeffnet.
   */
  public void connectTo(Component comp) {
    comp.addMouseListener(this);
  }

  /**
   * Erzeugt einen Menueeintrag und fuegt ihn dem Menue hinzu.
   * Als <code>ActionListener</code> fuer den Menueeintrag wird die
   * <code>ConnectedPopupMenue</code>-Instanz selbst verwendet, welcher
   * wiederum die angeschlossenen <code>ActionListener</code> informiert.
   * @param desc Beschreibung fuer den Menueeintrag
   * @see #actionPerformed(ActionEvent)
   */
  public JMenuItem add(String desc) {
    JMenuItem item = super.add(desc);
    item.addActionListener(this);
    return item;
  }

  /**
   * Implementiert nichts.
   */
  public void mouseClicked(MouseEvent e) {}

  /**
   * Implementiert nichts.
   */
  public void mouseEntered(MouseEvent e) {}

  /**
   * Implementiert nichts.
   */
  public void mouseExited(MouseEvent e) {}
}
