/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing.menu;

import java.awt.event.ActionListener;

import javax.swing.JMenuItem;

/**
 * Diese Klasse stellt einen Menupunkt dar, der gleichzeitig auch ein
 * Benutzer-Objekt beinhalten kann.<br>
 * Jedem Menuepunkt ist ein ActionCode zugeordnet, so dass nach Auswahl des
 * Menuepunkts eine Aktionskette entsteht, welche in einem einheitlichen
 * <code>ActionListener</code> interpretiert werden kann.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ObjectMenuItem extends JMenuItem implements ActionStructure {
  private   int               actionCode = 0;
  private   ActionStructure   parent;
  private   Object            object;

  /**
   * Erzeugt einen neuen Menuepunkt.
   * @param name       Anzeige-Name
   * @param object     Benutzerobjekt
   * @param actionCode codiert die Aktion die mit diesem Menue verknuepft ist
   * @param parent     uebergeordnete Aktion (z.B. uebergeordnetes <code>ObjectSubMenu</code>)
   */
  public ObjectMenuItem(String name, Object object, int actionCode, ActionStructure parent) {
    super(name);
    this.parent     = parent;
    this.actionCode = actionCode;
    this.object     = object;
  }

  /**
   * Erzeugt einen neuen Menuepunkt. Diesem ist keine spezielle Aktion
   * zugeordnet.
   * @param name    Anzeige-Name
   * @param object  Benutzerobjekt
   * @param parent  uebergeordnete Aktion (z.B. uebergeordnetes <code>ObjectSubMenu</code>)
   * @see schmitzm.swing.menu.ActionStructure#ACTION_NONE
   */
  public ObjectMenuItem(String name, Object object, ActionStructure parent) {
    this(name,object,ACTION_NONE,parent);
  }

  /**
   * Erzeugt einen neuen Menuepunkt. Diesem ist kein Objekt zugeordnet, sondern
   * nur eine Aktion.
   * @param name       Anzeige-Name
   * @param listener   Listener, der auf den Menuepunkt horcht
   * @param actionCode codiert die Aktion die mit diesem Menue verknuepft ist
   * @param parent     uebergeordnete Aktion (z.B. uebergeordnetes <code>ObjectSubMenu</code>)
   */
  public ObjectMenuItem(String name, ActionListener listener, int actionCode, ActionStructure parent) {
    this(name,(Object)null,actionCode,parent);
    this.addActionListener(listener);
  }


  /**
   * Liefert das Benutzer-Objekt des Menupunkts.
   */
  public Object getObject() {
    return object;
  }

  /**
   * Liefert die dem Menue uebergeordnete <code>ActionStructure</code>.
   */
  public ActionStructure getParentActionStructure() {
    return parent;
  }

  /**
   * Liefert die einzelne Aktion, die diesem Menue zugeordnet ist.
   */
  public int getActionCode() {
    return actionCode;
  }

  /**
   * Liefert die gesamte Aktionskette, die diesem Menue zugeordnet ist, in dem
   * die einzelnen Aktionen der Kette logisch verodert werden.
   */
  public int getCompleteActionCode() {
    if ( parent == null )
      return getActionCode();
    return getActionCode() | getParentActionStructure().getCompleteActionCode();
  }

  /**
   * Prueft, ob zwei Actionscodes gemeinsame Komponenten beinhalten.
   * Dies geschiet durch ein logisches UND zwischen den Aktionen, sowie
   * einem Vergleich, ob das Ergebnis >0 ist.
   * @param actionCode1 Aktionscode
   * @param actionCode2 Aktionscode
   * @return <code>true</code> gdw. die beiden Gesamt-Aktionen mindestens
   *         eine gemeinsame Aktion haben.
   */
  public boolean compareActions(int actionCode1, int actionCode2) {
    return (actionCode1 & actionCode2) > 0;
  }

}
