/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing.menu;

import java.awt.event.ActionListener;

import javax.swing.JMenu;

/**
 * Diese Klasse stellt ein Untermenue in einer Menue-Struktur dar.<br>
 * Jedem Menue ist ein ActionCode zugeordnet, so dass nach Auswahl eines
 * Menuepunkts eine Aktionskette entsteht, welche in einem einheitlichen
 * <code>ActionListener</code> interpretiert werden kann.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class ObjectSubMenu extends JMenu implements ActionStructure {
  private    int              actionCode = 0;
  private    ActionStructure  parent;
  /** Speichert den (einheitlichen) <code>AktionListener</code>, der fuer
   *  die Interpretation einer Menue-Auswahl verwendet wird. Dieser wird
   *  einem <code>MenuItem</code> <b>nicht</b> automatisch zugeordnet!.
   *  Die Speicherung in dieser abstrakten Oberklasse dient lediglich
   *  der Handhabbarkeit der Unterklassen. */
  protected  ActionListener   itemActionListener;

  /**
   * Erzeugt ein neues Untermenue.
   * @param name       Anzeige-Name fuer das Untermenue.
   * @param actionCode codiert die Aktion die mit diesem Menue verknuepft ist
   * @param parent     uebergeordnete Aktion (z.B. uebergeordnetes <code>ObjectSubMenu</code>)
   * @param itemActionListener ActionListener, der fuer die Interpretation des
   *                           ausgewaehlten Menuepunkts herangezogen wird (wird
   *                           den Menuepunkten <b>nicht</b> automatisch zugeordnet!
   * @param reorganize bestimmt, ob die Struktur dieses Menues bereits im Konstruktor
   *                   erzeugt werden soll (siehe {@link #reorganize()}). In
   *                   vielen Faellen wird diese Funktion manuell durch eine
   *                   uebergeordnetes Menue ausgeloest, was eine
   *                   Struktur-Erzeugung bei der Instanziierung ueberfluessig macht.
   */
  public ObjectSubMenu(String name, int actionCode, ActionStructure parent, ActionListener itemActionListener, boolean reorganize) {
    super(name);
    this.parent     = parent;
    this.actionCode = actionCode;
    this.itemActionListener = itemActionListener;
    if (reorganize)
      reorganize();
  }

  /**
   * Erzeugt ein neues Wurzel-Menue. Diesem ist keine Aktion zugeordnet und
   * es hat keine Uebergeordneten Menues. In der Regel ist es deshalb fuer
   * die Interpretation und Verarbeitung der ausgewaehlten Menueaktion
   * zustaendig (ist also selbst der <code>ActionListener</code>!).
   * @param name Anzeige-Name fuer das Untermenue.
   */
  public ObjectSubMenu(String name) {
    this(name,ActionStructure.ACTION_NONE,null,null,false);
  }

  /**
   * Erstellt die Struktur des Menues neu und ruft <code>reorganize()</code>
   * fuer alle Untermenues auf.
   */
  public void reorganize() {
    for (int i = 0; i < getItemCount(); i++)
      if ( getItem(i) instanceof ObjectSubMenu )
        ((ObjectSubMenu)getItem(i)).reorganize();
  }

  /**
   * Liefert den ActionListener, der den Menuepunkten zugeordnet werden kann.
   */
  public ActionListener getItemActionListener() {
    return this.itemActionListener;
  }

  /**
   * Liefert die dem Menue uebergeordnete <code>ActionStructure</code>.
   */
  public ActionStructure getParentActionStructure() {
    return parent;
  }

  /**
   * Liefert die einzelne Aktion, die diesem Menue zugeordnet ist.
   */
  public int getActionCode() {
    return actionCode;
  }

  /**
   * Liefert die gesamte Aktionskette, die diesem Menue zugeordnet ist, in dem
   * die einzelnen Aktionen der Kette logisch verodert werden.
   */
  public int getCompleteActionCode() {
    if ( parent == null )
      return getActionCode();
    return getActionCode() | getParentActionStructure().getCompleteActionCode();
  }

  /**
   * Prueft, ob zwei Actionscodes gemeinsame Komponenten beinhalten.
   * Dies geschiet durch ein logisches UND zwischen den Aktionen, sowie
   * einem Vergleich, ob das Ergebnis >0 ist.
   * @param actionCode1 Aktionscode
   * @param actionCode2 Aktionscode
   * @return <code>true</code> gdw. die beiden Gesamt-Aktionen mindestens
   *         eine gemeinsame Aktion haben.
   */
  public boolean compareActions(int actionCode1, int actionCode2) {
    return (actionCode1 & actionCode2) > 0;
  }
}
