/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing.table;

import java.awt.Color;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.AbstractCellEditor;
import javax.swing.JButton;
import javax.swing.JColorChooser;
import javax.swing.JDialog;
import javax.swing.JTable;
import javax.swing.table.TableCellEditor;

/**
 * Diese Klasse stellt einen Editor fuer eine Tabellen-Zelle dar, welche
 * ein {@link Color}-Objekt beinhaltet. Dieser wird durch einen Button
 * repraesentiert. Wird dieser aktiviert erscheint ein {@link JColorChooser}
 * ueber den eine neue Farbe ausgewaehlt werden kann.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of  Bonn/Germany)
 * @version 1.0
 */
public class ColorEditor extends AbstractCellEditor implements TableCellEditor, ActionListener {
  private Color         currentColor;
  private JButton       button;
  private JColorChooser colorChooser;
  private JDialog       dialog;
  protected static final String EDIT = "edit";

  /**
   * Erzeugt einen neuen Editor.
   */
  public ColorEditor() {
    //Set up the editor (from the table's point of view),
    //which is a button.
    //This button brings up the color chooser dialog,
    //which is the editor from the user's point of view.
    button = new JButton();
    button.setActionCommand(EDIT);
    button.addActionListener(this);
    button.setBorderPainted(false);

    //Set up the dialog that the button brings up.
    colorChooser = new JColorChooser();
    dialog = JColorChooser.createDialog(button,
                                        "Pick a Color",
                                        true, //modal
                                        colorChooser,
                                        this, //OK button handler
                                        null); //no CANCEL button handler
  }

  /**
   * Verarbeitet den "Klick" auf die Tabellenzelle (Button). Ruft einen
   * {@link JColorChooser} auf, ueber den eine Farbe ausgewaehlt werden
   * kann.
   */
  public void actionPerformed(ActionEvent e) {
    if (EDIT.equals(e.getActionCommand())) {
      //The user has clicked the cell, so
      //bring up the dialog.
      button.setBackground(currentColor);
      colorChooser.setColor(currentColor);
      dialog.setVisible(true);
      //Make the renderer reappear.
      fireEditingStopped();
    } else {
      //User pressed dialog's "OK" button.
      currentColor = colorChooser.getColor();
    }
  }

  /**
   * Liefert die aktuell ausgewaehlte Farbe.
   * @return <code>Color</code>-Instanz
   */
  public Object getCellEditorValue() {
    return currentColor;
  }

  /**
   * Liefert die Komponente, die den Editor repraesentiert.
   * Hierbei handelt es sich um einen {@link JButton}.
   */
  public Component getTableCellEditorComponent(JTable table,
                                               Object value,
                                               boolean isSelected,
                                               int row,
                                               int column) {
    currentColor = (Color) value;
    return button;
  }
}
