/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing.table;

import java.awt.Color;
import java.awt.Component;

import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JTable;
import javax.swing.border.Border;
import javax.swing.table.TableCellRenderer;

/**
 * Diese Klasse stellt einen Renderer fuer Tabellenzellen dar, in denen
 * eine Farbe (also eine <code>Color</code>-Instanz) dargestellt werden soll.<br>
 * Die Farbe wird durch ein {@link JLabel} in entsprechender Hintergrundfarbe
 * dargestellt.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ColorRenderer extends JLabel implements TableCellRenderer {
    private   Border  unselectedBorder = null;
    private   Border  selectedBorder   = null;
    /** Speichert, ob die Zelle einen inneren Rand erhaelt */
    protected boolean isBordered = true;

  /**
   * Erzeugt einen neuen Zellen-Renderer.
   * @param isBordered gibt an ob die Zelle komplett ausgefuellt wird, oder
   *                   ein innerer Rand erzeugt wird
   */
  public ColorRenderer(boolean isBordered) {
    this.isBordered = isBordered;
    setOpaque(true);
  }

  /**
   * Liefert einen {@link JLabel}, das die in der Tabellenzelle enthaltene
   * Farbe darstellt.
   * @param table      Tabelle in der die Zelle liegt
   * @param color      Farbe, die in der Zelle darzustellen ist (<b>muss eine
   *                   Instanz von <code>Color</code> sein!!</b>)
   * @param isSelected gibt an ob die Zelle aktuell selektiert ist
   * @param hasFocus   gibt an ob die Zelle aktuell den Fokus besitzt
   * @param row        Tabellenzeile der Zelle
   * @param column     Tabellenspalte der Zelle
   * @return ein farbiges {@link JLabel}
   */
  public Component getTableCellRendererComponent(
                               JTable table,
                               Object color,
                               boolean isSelected,
                               boolean hasFocus,
                               int row,
                               int column) {
    // Randarten erzeugen, falls dies noch nicht geschehen ist
    if (selectedBorder == null) {
      selectedBorder = BorderFactory.createMatteBorder(
          2, 5, 2, 5,
          table.getSelectionBackground()
      );
    }
    if (unselectedBorder == null) {
      unselectedBorder = BorderFactory.createMatteBorder(
          2, 5, 2, 5,
          table.getBackground()
      );
    }
    // Rand setzen
    if (isBordered) {
      if (isSelected)
        setBorder(selectedBorder);
      else
        setBorder(unselectedBorder);
    }
    // Label erzeugen/initialisieren
    return createColorLabel((Color)color,this);
  }

  /**
   * Liefert einen {@link JLabel}, das eine Farbe darstellt. Ist der uebergebene
   * <code>label</code>-Parameter <code>null</code> wird eine neue
   * {@link JLabel}-Instanz erzeugt.
   * @param color Farbe, die durch das Label dargestellt wird
   * @param label Label, das entsprechend der Farbe umformatiert (kann
   *              <code>null</code> sein)
   */
  public static JLabel createColorLabel(Color color, JLabel label) {
    if (label == null)
      label = new JLabel();

    label.setOpaque(true);
    label.setBackground(color);
    if ( color != null )
      label.setToolTipText("RGB value: " + color.getRed() + ", "
                           + color.getGreen() + ", "
                           + color.getBlue());
    return label;
  }
}
