/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing.table;

import java.awt.Color;
import java.awt.Component;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;

import javax.swing.BorderFactory;
import javax.swing.JComponent;
import javax.swing.JLabel;
import javax.swing.JTable;
import javax.swing.border.Border;

/**
 * Diese Klasse stellt einen Renderer fuer Tabellenzellen dar, in denen
 * mehrere Farben (also ein <code>Color[]</code>) dargestellt werden sollen.<br>
 * Jede Farbe wird durch ein {@link JLabel} in entsprechender Hintergrundfarbe
 * dargestellt. Die Label werden horizontal angeordnet.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ColorsRenderer extends ColorRenderer {
  // Rand um Farb-Label
  private Border labelBorder = BorderFactory.createLineBorder(Color.BLACK);

  /**
   * Erzeugt einen neuen Zellen-Renderer.
   * @param isBordered gibt an ob die Zelle komplett ausgefuellt wird, oder
   *                   ein Rand erzeugt wird
   */
  public ColorsRenderer(boolean isBordered) {
    super(isBordered);
    this.setLayout( new GridBagLayout() );
  }

  /**
   * Liefert einen Container, in dem fuer jede Farbe ein {@link JLabel}
   * enthalten ist.
   * @param table      Tabelle in der die Zelle liegt
   * @param object     Farben, die in der Zelle darzustellen sind (<b>muss eine
   *                   Instanz von <code>Color[]</code> sein!!</b>)
   * @param isSelected gibt an ob die Zelle aktuell selektiert ist
   * @param hasFocus   gibt an ob die Zelle aktuell den Fokus besitzt
   * @param row        Tabellenzeile der Zelle
   * @param column     Tabellenspalte der Zelle
   * @return eine {@link JComponent} mit horizontal angeordneten {@link JLabel}s
   */
  public Component getTableCellRendererComponent(
                               JTable table,
                               Object object,
                               boolean isSelected,
                               boolean hasFocus,
                               int row,
                               int column) {
    // Methode der Oberklasse (mit Dummy-Farbe) zum erzeugen/setzen
    // des Rands
    super.getTableCellRendererComponent(table,null,isSelected,hasFocus,row,column);

    Color[] color = (Color[]) object;
    // Fuer jede Farbe ein Label erzeugen
    this.removeAll();
    for (int i=0; i<color.length; i++) {
      JLabel label = ColorRenderer.createColorLabel(color[i],null);
      if ( isBordered )
        label.setBorder( labelBorder );
      add(label, new GridBagConstraints(i,0,1,1,1,1.0,GridBagConstraints.CENTER,GridBagConstraints.BOTH,new Insets(0,0,0,0),0,0));
    }
    return this;
  }
}
