/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing.table;

import java.awt.Component;

import javax.swing.AbstractButton;
import javax.swing.JTable;
import javax.swing.SwingConstants;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.TableCellRenderer;

/**
 * Diese Klasse stellt einen Tabellen-Renderer fuer jede Art von
 * {@link Component}-Objekten dar. Um z.B. einen Button oder eine
 * Checkbox in einer {@link JTable}-Zelle darzustellen, muss mittels der
 * Methode {@link JTable#setDefaultRenderer(Class,TableCellRenderer)} dieser
 * Renderer fuer die entsprechende Zellwert-Klasse zugeordnet werden.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class ComponentRenderer implements TableCellRenderer, SwingConstants {
  /** Horizontale Ausrichtung der Componente innerhalb der Tabellenzelle
   *  (Default: {@link SwingConstants#CENTER}). */
  protected int horAlign = CENTER;
  /** Vertikale Ausrichtung der Componente innerhalb der Tabellenzelle
   *  (Default: {@link SwingConstants#CENTER}). */
  protected int vertAlign = CENTER;

  /**
   * Erzeugt einen neuen Zellen-Renderer. <b>Die Ausrichtung ist nur relevant
   * fuer Komponenten vom Typ {@link AbstractButton}!</b>
   * @param horAlign horizontale Ausrichtung der Komponente innerhalb der
   *                 Tabellenzelle (RIGHT, LEFT, CENTER, LEADING, TRAILING aus
   *                 {@link SwingConstants}; Default = CENTER)
   * @param vertAlign vertikale Ausrichtung der Komponente innerhalb der
   *                  Tabellenzelle (CENTER, TOP, BOTTOM aus
   *                  {@link SwingConstants}; Default = CENTER)
   */
  public ComponentRenderer(int horAlign, int vertAlign) {
    super();
    this.horAlign = horAlign;
    this.vertAlign = vertAlign;
  }

  /**
   * Erzeugt einen neuen Zellen-Renderer.
   */
  public ComponentRenderer() {
    this(CENTER,CENTER);
  }

  /**
   * Liefert die horizontale Ausrichtung der Komponente innerhalb der
   * Tabellenzelle.
   * @return RIGHT, LEFT, CENTER, LEADING oder TRAILING aus {@link SwingConstants}; Default = CENTER
   */
  public int getHorizontalAlignment() {
    return horAlign;
  }

  /**
   * Setzt die horizontale Ausrichtung der Komponente innerhalb der
   * Tabellenzelle. Wirkt sich nur auf Komponenten vom Typ
   * {@link AbstractButton} aus!
   * @param horAlign (RIGHT, LEFT, CENTER, LEADING oder TRAILING aus {@link SwingConstants})
   */
  public void setHorizontalAlignment(int horAlign) {
    this.horAlign = horAlign;
  }

  /**
   * Liefert die vertikale Ausrichtung der Komponente innerhalb der
   * Tabellenzelle.
   * @return CENTER, TOP oder BOTTOM aus {@link SwingConstants}; Default = CENTER
   */
  public int getVerticalAlignment() {
    return vertAlign;
  }

  /**
   * Setzt die vertikale Ausrichtung der Komponente innerhalb der
   * Tabellenzelle. Wirkt sich nur auf Komponenten vom Typ
   * {@link AbstractButton} aus!
   * @param vertAlign (CENTER, TOP oder BOTTOM aus {@link SwingConstants})
   */
  public void setVerticalAlignment(int vertAlign) {
    this.vertAlign = vertAlign;
  }

  /**
   * Die von der Methode {@link #createRendererComponent(JTable,Object,boolean,boolean,int,int)}
   * gelieferte Komponente wird farblich hintelegt, wenn die Tabellenzeile
   * selektiert ist.
   * @param table      Tabelle in der die Zelle liegt
   * @param value      Zu rendernder Wert
   * @param isSelected gibt an ob die Zelle aktuell selektiert ist
   * @param hasFocus   gibt an ob die Zelle aktuell den Fokus besitzt
   * @param row        Tabellenzeile der Zelle
   * @param column     Tabellenspalte der Zelle
   */
  public Component getTableCellRendererComponent(
                               JTable table,
                               Object value,
                               boolean isSelected,
                               boolean hasFocus,
                               int row,
                               int column) {
    Component rendComp = createRendererComponent(table,value,isSelected,hasFocus,row,column);
    if ( rendComp instanceof AbstractButton ) {
      ((AbstractButton)rendComp).setHorizontalAlignment(horAlign);
      ((AbstractButton)rendComp).setVerticalAlignment(vertAlign);
    }
    if ( isSelected ) {
      rendComp.setBackground(table.getSelectionBackground());
      rendComp.setForeground(table.getSelectionForeground());
    }
    return rendComp;
  }

  /**
   * Liefert einfach das zu {@link Component} gecastete Objekt {@code value} zurueck.
   * Handelt es sich bei <code>comp</code> nicht um ein {@link Component}-Objekt, wird
   * der {@link DefaultTableCellRenderer} herangezogen.
   * @param table      Tabelle in der die Zelle liegt
   * @param value      Zu rendernde Komponente (<b>muss eine
   *                   Instanz von <code>Component</code> sein!!</b>)
   * @param isSelected gibt an ob die Zelle aktuell selektiert ist
   * @param hasFocus   gibt an ob die Zelle aktuell den Fokus besitzt
   * @param row        Tabellenzeile der Zelle
   * @param column     Tabellenspalte der Zelle
   */
  protected Component createRendererComponent( JTable table,
                                               Object value,
                                               boolean isSelected,
                                               boolean hasFocus,
                                               int row,
                                               int column) {

    if ( !(value instanceof Component) )
      return new DefaultTableCellRenderer().getTableCellRendererComponent(table,value,isSelected,hasFocus,row,column);
    return (Component)value;
  }

  /**
   * Dieser Renderer stellt das jeweilige Objekt in einem {@link javax.swing.JTextField} dar.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class JTextField extends ComponentRenderer {
    /**
     * Erzeugt einen neuen Zellen-Renderer in Form einer {@link JCheckBox}.
     */
    public JTextField() {
      super();
    }

    /**
     * Liefert ein {@link javax.swing.JTextField}, das als Inhalt den Wert der
     * Tabellenzelle hat.
     * @param table      Tabelle in der die Zelle liegt
     * @param value      Zu renderndes Objekt
     * @param isSelected gibt an ob die Zelle aktuell selektiert ist
     * @param hasFocus   gibt an ob die Zelle aktuell den Fokus besitzt
     * @param row        Tabellenzeile der Zelle
     * @param column     Tabellenspalte der Zelle
     */
    public Component createRendererComponent( JTable table,
                                              Object value,
                                              boolean isSelected,
                                              boolean hasFocus,
                                              int row,
                                              int column) {
      return new javax.swing.JTextField(value != null ? value.toString() : "");
    }
  }

  /**
   * Dieser Renderer stellt das jeweilige Objekt in einer {@link javax.swing.JCheckBox} dar.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class JCheckBox extends ComponentRenderer  {
    /**
     * Erzeugt einen neuen Zellen-Renderer in Form einer {@link JCheckBox}.
     * @param horAlign horizontale Ausrichtung der Checkbox innerhalb der
     *                 Tabellenzelle (RIGHT, LEFT, CENTER, LEADING, TRAILING aus
     *                 {@link SwingConstants}; Default = CENTER)
     * @param vertAlign vertikale Ausrichtung der Checkbox innerhalb der
     *                  Tabellenzelle (CENTER, TOP, BOTTOM aus
     *                  {@link SwingConstants}; Default = CENTER)
     */
    public JCheckBox(int horAlign, int vertAlign) {
      super(horAlign, vertAlign);
    }

    /**
     * Erzeugt einen neuen Zellen-Renderer in Form einer {@link JCheckBox}.
     */
    public JCheckBox() {
      this(CENTER, CENTER);
    }

    /**
     * Liefert eine {@link javax.swing.JCheckBox}, die aktiviert ist, wenn
     * es sich bei dem darzustellenden Objekt um einen {@link Boolean} handelt,
     * der den Wert {@code true} hat.
     * @param table      Tabelle in der die Zelle liegt
     * @param value      Zu renderndes Objekt
     * @param isSelected gibt an ob die Zelle aktuell selektiert ist
     * @param hasFocus   gibt an ob die Zelle aktuell den Fokus besitzt
     * @param row        Tabellenzeile der Zelle
     * @param column     Tabellenspalte der Zelle
     */
    public Component createRendererComponent( JTable table,
                                              Object value,
                                              boolean isSelected,
                                              boolean hasFocus,
                                              int row,
                                              int column) {
      return new javax.swing.JCheckBox("", value instanceof Boolean && value != null ? (Boolean)value : false);
    }
  }

  /**
   * Dieser Renderer stellt das jeweilige Objekt in Form einer
   * {@link javax.swing.JComboBox} dar.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  public static class JComboBox extends ComponentRenderer {
    /**
     * Erzeugt einen neuen Zellen-Renderer in Form einer {@link JComboBox}.
     * @param horAlign horizontale Ausrichtung der Combobox innerhalb der
     *                 Tabellenzelle (RIGHT, LEFT, CENTER, LEADING, TRAILING aus
     *                 {@link SwingConstants}; Default = CENTER)
     * @param vertAlign vertikale Ausrichtung der Combobox innerhalb der
     *                  Tabellenzelle (CENTER, TOP, BOTTOM aus
     *                  {@link SwingConstants}; Default = CENTER)
     */
    public JComboBox(int horAlign, int vertAlign) {
      super(horAlign, vertAlign);
    }

    /**
     * Erzeugt einen neuen Zellen-Renderer in Form einer {@link JComboBox}.
     */
    public JComboBox() {
      this(CENTER, CENTER);
    }

    /**
     * Liefert eine {@link javax.swing.JComboBox}, die als Inhalt den Wert der
     * Tabellenzelle hat.
     * @param table      Tabelle in der die Zelle liegt
     * @param value      Zu renderndes Objekt
     * @param isSelected gibt an ob die Zelle aktuell selektiert ist
     * @param hasFocus   gibt an ob die Zelle aktuell den Fokus besitzt
     * @param row        Tabellenzeile der Zelle
     * @param column     Tabellenspalte der Zelle
     */
    public Component createRendererComponent( JTable table,
                                              Object value,
                                              boolean isSelected,
                                              boolean hasFocus,
                                              int row,
                                              int column) {
      return new javax.swing.JComboBox( new Object[] {value} );
    }
  }

}
