/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing.table;

import java.awt.Color;
import java.awt.Component;
import java.text.NumberFormat;

import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JTable;
import javax.swing.border.Border;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.TableCellRenderer;

/**
 * Diese Klasse stellt einen Renderer fuer Tabellenzellen dar, in denen
 * eine {@link Double}-Wert dargestellt werden soll.<br>
 * Das {@linkplain NumberFormat Format} kann individuell gesetzt werden.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class DoubleRenderer extends DefaultTableCellRenderer.UIResource {
  /** Speichert das Format, in dem der {@link Double}-Wert <b>in der Zelle</b>
   *  dargestellt wird. */
  protected NumberFormat cellFormatter = null;

  /** Speichert das Format, in dem der {@link Double}-Wert <b>im Tooltip</b>
   *  dargestellt wird. */
  protected NumberFormat tooltipFormatter = null;
  
  /** Speichert, ob ein Tooltip angezeigt wird. */
  protected boolean tooltipEnabled = true;

  /**
   * Erzeugt einen neuen Zellen-Renderer. Wird {@code null} als Format uebergeben,
   * wird der {@link Double}-Wert unformatiert mittels {@link Double#toString()}
   * dargestellt.
   * @param cellFormatter bestimmt das Format, in dem der {@link Double}-Wert
   *                      <b>in der Zelle</b> dargestellt wird.
   * @param tooltipFormatter bestimmt das Format, in dem der {@link Double}-Wert
   *                         <b>im Tooltip</b> dargestellt wird.
   */
  public DoubleRenderer(NumberFormat cellFormatter, NumberFormat tooltipFormatter) {
    super();
    setHorizontalAlignment(JLabel.RIGHT);
    setCellFormat(cellFormatter);
  }

  /**
   * Setzt das Format, in dem der {@link Double}-Wert in der Zelle dargestellt wird.
   * @param formatter Format, in dem der {@link Double}-Wert in der Zelle
   *                  dargestellt wird (wenn {@code null}, wird der Tooltip
   *                  unformatiert mittels {@link Double#toString()} dargestellt).
   */
  public void setCellFormat(NumberFormat formatter) {
    this.cellFormatter = formatter;
  }

  /**
   * Liefert das Format, in dem der {@link Double}-Wert in der Zelle dargestellt wird.
   */
  public NumberFormat getCellFormat() {
    return this.cellFormatter;
  }

  /**
   * Setzt das Format, in dem der {@link Double}-Wert im Tooltip dargestellt wird.
   * @param formatter Format, in dem der {@link Double}-Wert im Tooltip
   *                  dargestellt wird (wenn {@code null}, wird der Tooltip
   *                  unformatiert mittels {@link Double#toString()} dargestellt).
   */
  public void setTooltipFormat(NumberFormat formatter) {
    this.tooltipFormatter = formatter;
  }

  /**
   * Liefert das Format, in dem der {@link Double}-Wert im Tooltip dargestellt wird.
   */
  public NumberFormat getTooltipFormat() {
    return this.tooltipFormatter;
  }

  /**
   * Setzt, ob der Tooltip fuer die Zelle angezeigt wird.
   */
  public void setTooltipEnabled(boolean enabled) {
    this.tooltipEnabled = enabled;
  }
  
  /**
   * Prueft, ob der Tooltip fuer die Zelle angezeigt wird.
   */
  public boolean isTooltipEnabled() {
    return tooltipEnabled;
  }
  
  /**
   * Formatiert einen Wert und belegt entsprechend das Label. 
   * @see JLabel#setText(String)
   */
  @Override
  public void setValue(Object value) {
    if ( value == null ) {
      setText( "" );
      setToolTipText( "<null>" );
    } else {
      setText( cellFormatter != null ? cellFormatter.format(value) : value.toString() );
      setToolTipText( tooltipFormatter != null ? tooltipFormatter.format(value) : value.toString() );
    }
    
    // Tooltip initialisieren, wenn nicht aktiviert
    if ( !tooltipEnabled )
      setToolTipText(null);
  }
  
}
