/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing.table;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JPopupMenu;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.event.MenuEvent;
import javax.swing.event.MenuListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableModel;

import schmitzm.swing.event.PopupMenuListener;


/**
 * Diese Klasse stellt eine veraenderbare Tabelle dar. Sie erhaelt ein
 * Kontextmenue, ueber das der Tabelle Datensaetze hinzugefuegt, entfernt oder
 * geaendert werden koennen. Zudem reagiert die Tabelle automatisch auf einen
 * Doppelklick mit der Aenderungsaktion.
 * @see MutableTableModel
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class MutableTable extends JTable {
  /** Konstante fuer den Menuepunkt "Hinzufuegen". */
  public static final int ITEM_ADD    = 2;
  /** Konstante fuer den Menuepunkt "Loeschen". */
  public static final int ITEM_REMOVE = 4;
  /** Konstante fuer den Menuepunkt "Aendern". */
  public static final int ITEM_CHANGE = 8;
  /** Konstante, die alle Menuepunkte beinhaltet. */
  public static final int ITEM_ALL    = ITEM_ADD | ITEM_REMOVE | ITEM_CHANGE;

  /** Speichert die Datenbasis fuer die Tabelle */
  protected MutableTableModel model = null;

  /** Speichert, welche Menuepunkte fuer das Kontextmenue aktiv sind */
  protected int mask = 0;

  /**
   * Erzeugt eine neue Tabelle.
   * @param model Datenbasis fuer die Tabelle
   * @param mask  Definiert die Eintraege im Kontextmenue (logische
   *              ODER-Verknuepfung der Konstanten <code>ITEM_ADD</code>,
   *              <code>ITEM_REMOVE</code> und <code>ITEM_CHANGE</code>)
   * @see #ITEM_ADD
   * @see #ITEM_REMOVE
   * @see #ITEM_CHANGE
   */
  public MutableTable(MutableTableModel model, int mask) {
    super(model);
    this.mask  = mask;
    setModel(model);

    // MouseListener fuer PopupMenu
    this.addMouseListener( createPopupMenuListener(mask) );
    getTableHeader().addMouseListener( createPopupMenuListener(mask) );

    // MouseListener fuer Doppelklick
    this.addMouseListener(new MouseAdapter() {
      public void mouseClicked(MouseEvent e) {
        if (e.getButton() == MouseEvent.BUTTON1 &&
            e.getClickCount() == 2)
          performDoubleClick();
      }
    });
  }

  /**
   * Erzeugt eine neue Tabelle mit komplettem Kontextmenue.
   * @param model Datenbasis fuer die Tabelle
   * @see #ITEM_ALL
   */
  public MutableTable(MutableTableModel model) {
    this(model,ITEM_ALL);
  }

  /**
   * Setzt das Datenmodell fuer die Tabelle und ruft {@link MutableTableModel#initTable(JTable)}
   * auf.
   * @exception UnsupportedOperationException falls kein {@link MutableTableModel}
   *            uebergeben wird
   */
  @Override
  public void setModel(TableModel model) {
    if ( !(model instanceof MutableTableModel) )
      throw new UnsupportedOperationException("MutableTableModel expected for MutableTable!");
    super.setModel(model);
    this.model = (MutableTableModel)model;
    this.model.initTable(this);
  }

  /**
   * Erzeugt einen {@link MouseListener} der bei einem Popup-Signal (z.B.
   * Rechtsklick) das Tabellen-Menue oeffnet. In diesem sind alle Menuepunkte
   * (Aendern, Loeschen, Hinzufuegen) enthalten.
   * @see #createPopupMenuListener(int)
   * @see #ITEM_ALL
   */
  public MouseListener createPopupMenuListener() {
    return createPopupMenuListener(ITEM_ALL);
  }

  /**
   * Erzeugt einen {@link MouseListener} der bei einem Popup-Signal (z.B.
   * Rechtsklick) das Tabellen-Kontextmenue oeffnet.
   * @param mask  Definiert die Eintraege im Kontextmenue (logische
   *              ODER-Verknuepfung der Konstanten <code>ITEM_ADD</code>,
   *              <code>ITEM_REMOVE</code> und <code>ITEM_CHANGE</code>)
   * @see #ITEM_ADD
   * @see #ITEM_REMOVE
   * @see #ITEM_CHANGE
   */
  public MouseListener createPopupMenuListener(int mask) {
    return new PopupMenuListener( createPopupMenu(mask).getPopupMenu() );
  }

  /**
   * Erzeugt eine Instanz des {@link PopupMenu}. Unterklassen koennen diese
   * Methode ueberschreiben, um das Menu zu erweitern.
   * @param mask  Definiert die Eintraege im Kontextmenue (logische
   *              ODER-Verknuepfung der Konstanten <code>ITEM_ADD</code>,
   *              <code>ITEM_REMOVE</code> und <code>ITEM_CHANGE</code>)
   * @see #ITEM_ADD
   * @see #ITEM_REMOVE
   * @see #ITEM_CHANGE
   */
  protected PopupMenu createPopupMenu(int mask) {
    return new PopupMenu(this,mask);
  }

  /**
   * Liefert die Bit-Maske, welche Punkt im Kontextmenue aktiv sind.
   * @see #ITEM_ADD
   * @see #ITEM_REMOVE
   * @see #ITEM_CHANGE
   * @see #ITEM_ALL
   */
  public int getMask() {
    return mask;
  }

  /**
   * Wird aufgerufen, wenn der Menuepunkt "Ändern" gewaehlt wird.
   * Ruft fuer jede selektierte Zelle {@link MutableTableModel#performChangeData(int,int) MutableTableModel#performChangeData(..)}
   * auf. Ist nur eine zeilenweise Auswahl in der Tabelle erlaubt, so
   * wird die obige Methode nur fuer jede selektierte Zeile aufgerufen.
   * Der Aufrufparameter fuer die Spalte ist in diesem Fall die Spalte, auf die geklickt
   * wurde.<br>
   * Erwirkt anschliessend ein {@link AbstractTableModel#fireTableDataChanged()}
   * @see #setRowSelectionAllowed(boolean)
   */
  protected void performChange() {
    int selCol[] = this.getSelectedColumns();
    int selRow[] = this.getSelectedRows();

    for (int r=0; r<selRow.length; r++) {
      // werden Zellen einzeln selektiert, wird fuer jede Zelle
      // einzeln die Aenderungsaktion ausgefuehrt, sonst nur einmal
      // pro Zeile
      if ( !this.rowSelectionAllowed )
        for (int c=0; c<selCol.length; c++)
          model.performChangeData(selRow[r],selCol[c]);
      else
        // Als Spalte wird diejenige angegeben, auf die geklickt wurde
        model.performChangeData(selRow[r], this.getSelectedColumn());
    }
    model.fireTableDataChanged();
  }

  /**
   * Wird aufgerufen, wenn der Menuepunkt "Löschen" gewaehlt wird.
   * Ruft fuer jede selektierte Zeile {@link MutableTableModel#performRemoveRow(int) MutableTableModel#performRemoveRow(..)}
   * auf. Dabei wird <b>rueckwaerts</b> vorgegangen - also die letzte selektierte
   * Zeile zuerst -, damit der Zeilenindex ueber alle
   * {@link MutableTableModel#performRemoveRow(int)}-Aufrufe konsistent bleibt.
   * Erwirkt anschliessend automatisch ein {@link AbstractTableModel#fireTableDataChanged()}

   */
  protected void performRemove() {
    int selRow[] = this.getSelectedRows();
    for (int r=selRow.length-1; r>=0; r--)
      model.performRemoveRow(selRow[r]);
    model.fireTableDataChanged();
  }

  /**
   * Wird aufgerufen, wenn der Menuepunkt "Hinzufügen" gewaehlt wird.
   * Ruft {@link MutableTableModel#performAddRow() MutableTableModel#performAddRow()}
   * auf.<br>
   * Erwirkt anschliessend automatisch ein {@link AbstractTableModel#fireTableDataChanged()}
   */
  protected void performAdd() {
    model.performAddRow();
    model.fireTableDataChanged();
  }

  /**
   * Wird aufgerufen, wenn ein Doppelklick auf die Tabelle vorgenommen wird.
   * Standardmaessig wird {@link #performChange()} aufgerufen (sofern
   * der Aendern-Menuepunkt aktiv ist!).
   * Unterklassen koennen diese Methode ueberschreiben, um diesem Ereignis
   * eins andere Aktion zuzuweisen oder das Ereignis gaenzlich zu unterdruecken.
   */
  public void performDoubleClick() {
    if ( (mask & ITEM_CHANGE) > 0 )
      performChange();
  }

  /**
   * Erzeugt ein {@link JScrollPane} fuer die Tabelle. Dieses erhaelt automatisch
   * den gleichen ToolTip-Text und Popupmenue-Listener wie die Tabelle.
   * Der Anwender kann somit auch auf das ScrollPane klicken, um die Tabellen-Optionen
   * anzuwaehlen.
   */
  public JScrollPane createScrollPane() {
    JScrollPane scrollPane = new JScrollPane( this );
    scrollPane.setToolTipText( this.getToolTipText() );
    scrollPane.addMouseListener( this.createPopupMenuListener( this.getMask() ) );
    return scrollPane;
  }

  /**
   * Diese Klasse stellt das Kontextmenue fuer einen {@link MutableTable}
   * dar.
   * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
   * @version 1.0
   */
  protected class PopupMenu extends JMenu {
    /** Menueeintrag fuer Aendern-Aktion */
    protected JMenuItem changeItem = new JMenuItem("Ändern");
    /** Menueeintrag fuer Entfernen-Aktion */
    protected JMenuItem removeItem = new JMenuItem("Entfernen");
    /** Menueeintrag fuer Hinzufuegen-Aktion */
    protected JMenuItem addItem = new JMenuItem("Hinzufügen");

    private MutableTable table = null;

    /**
     * Erzeugt ein neues Kontextmenue.
     * @param table Tabelle auf die sich die Aktionen des Menus beziehen
     * @param mask  Definiert die Eintraege im Kontextmenue (logische
     *              ODER-Verknuepfung der Konstanten <code>ITEM_ADD</code>,
     *              <code>ITEM_REMOVE</code> und <code>ITEM_CHANGE</code>)
     * @see #ITEM_ADD
     * @see #ITEM_REMOVE
     * @see #ITEM_CHANGE
     */
    public PopupMenu(final MutableTable table, int mask) {
      super();
      this.table = table;
      changeItem.addActionListener( new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          table.performChange();
        }
      });
      addItem.addActionListener( new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          table.performAdd();
        }
      });
      removeItem.addActionListener( new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          table.performRemove();
        }
      });

      // Menue-Punkte hinzufuegen
      if ( (mask & ITEM_CHANGE) > 0 )
        this.add( changeItem );
      if ( (mask & ITEM_ADD) > 0 )
        this.add( addItem );
      if ( (mask & ITEM_REMOVE) > 0 )
        this.add( removeItem );

      this.addMenuListener( new MenuListener() {
        public void menuCanceled(MenuEvent e) {}
        public void menuDeselected(MenuEvent e) {}
        public void menuSelected(MenuEvent e) {
          setEnabled(ITEM_REMOVE,table.getSelectedRowCount() > 0);
          setEnabled(ITEM_CHANGE,table.getSelectedRowCount() > 0);
        }

      });
    }

    /**
     * Erzeugt ein komplettes Kontextmenue mit allen moeglichen Aktionen.
     */
    public PopupMenu(final MutableTable table) {
      this(table,ITEM_ALL);
    }

    /**
     * (De)aktiviert ein oder mehrere Menueintraege.
     * @param mask definiert die betroffenen Eintraege (logische
     *              ODER-Verknuepfung der Konstanten <code>ITEM_ADD</code>,
     *              <code>ITEM_REMOVE</code> und <code>ITEM_CHANGE</code>)
     * @param enabled bestimmt ob der Eintrag aktiviert oder deaktiviert wird
     */
    public void setEnabled(int mask, boolean enabled) {
      if ( (mask & ITEM_CHANGE) > 0 )
        changeItem.setEnabled( enabled );
      if ( (mask & ITEM_ADD) > 0 )
        addItem.setEnabled( enabled );
      if ( (mask & ITEM_REMOVE) > 0 )
        removeItem.setEnabled( enabled );
    }

    /**
     * Erzeugt ein dem Menue entsprechendes {@link JPopupMenu}.
     */
    public JPopupMenu getPopupMenu() {
      JPopupMenu menu = super.getPopupMenu();
      menu.addPopupMenuListener( new javax.swing.event.PopupMenuListener() {
        public void popupMenuCanceled(javax.swing.event.PopupMenuEvent e) {}
        public void popupMenuWillBecomeInvisible(javax.swing.event.PopupMenuEvent e) {}
        public void popupMenuWillBecomeVisible(javax.swing.event.PopupMenuEvent e) {
          setEnabled(ITEM_REMOVE,table.getSelectedRowCount() > 0);
          setEnabled(ITEM_CHANGE,table.getSelectedRowCount() > 0);
        }
      });
      return menu;
    }

  }

}
