/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing.table;

import javax.swing.JTable;
import javax.swing.event.TableModelListener;
import javax.swing.table.AbstractTableModel;
import javax.swing.table.TableModel;

import schmitzm.swing.SelectableJTable;

/**
 * This table model extends an other {@link TableModel} by inserting an
 * additional column 0 which shows and controls the row selection.
 * To realize this functionality it is necessary to connect this table model
 * to an explicit {@link JTable}. Besides it is necessary that this table
 * is a {@link SelectableJTable} to overcome the "normal" click behavior (clear
 * selection when no Ctrl/Shift key is hold), when clicking the selection column. 
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 */
public class SelectionTableModel extends AbstractTableModel implements PipedTableModel {

  /** Holds the table which controls the selection. */
  protected SelectableJTable table = null;
  /** Holds the base data. */
  protected TableModel model = null;
//!! SOLVED BY USING SelectableJTable !!
//  /** Used to realize the selection column. */
//  protected SelectionRestorer selectionRestorer = null;
  
  
  /**
   * Creates a new table model
   * @param model provides the data
   * @param table used to determine the row selection 
   */
  public SelectionTableModel(TableModel model, SelectableJTable table) {
    super();
    this.model = model;
    this.table = table;
    table.getColumnModel().setColumnSelectionAllowed(false);
    
//!! SOLVED BY USING SelectableJTable !!
//    // Goal: By selection column cells, rows should be selected individually,
//    //       without holding Strg/Shift key.
//    // Problem: Also on a click in the selection column, the selection
//    //          is first cleared
//    // Solution: After every key stroke and every mouse click, the current
//    //           selection is remembered. On click on selection column
//    //           (the selection is already cleared!) the selection is
//    //           restored. After that the click (TableModel.setValue(..)) is
//    //           progressed, which extends the selection.
//    this.selectionRestorer = new SelectionRestorer(table);
  }

//!! SOLVED BY USING SelectableJTable !!
//  /**
//   * Forces a update of the {@link SelectionRestorer} to renew the
//   * remembered selected cells.
//   */
//  public void updateSelectionRestorer() {
//    if ( selectionRestorer != null )
//      selectionRestorer.update();
//  }
  
  /**
   * Returns the original {@link TableModel}.
   */
  @Override
  public TableModel getPipedModel() {
    return model;
  }
  
  /**
   * Adds a listener to the list that's notified each time a change
   * to the data model occurs.
   */
  @Override
  public void addTableModelListener(TableModelListener listener) {
    model.addTableModelListener(listener);
  }
  
  /**
   * Removes a listener from the list that's notified each time a change
   * to the data model occurs.
   */
  @Override
  public void removeTableModelListener(TableModelListener listener) {
    model.removeTableModelListener(listener);
  }

  /**
   * Returns the type of a column.
   * @return {@code Boolean.class} for column 0
   */
  @Override
  public Class<?> getColumnClass(int col) {
    if ( col == 0 )
      return Boolean.class;
   return model.getColumnClass(col - 1); 
  }
  
  /**
   * Returns the type of a column.
   * @return {@code Boolean.class} for column 0
   */
  @Override
  public int getColumnCount() {
   return model.getColumnCount() + 1; 
  }
  
  /**
   * Returns the name of a column.
   * @return empty string for column 0
   */
  @Override
  public String getColumnName(int col) {
    if ( col == 0 )
      return "";
   return model.getColumnName(col - 1); 
  }
  
  /**
   * Returns the number of rows in the model.
   */
  @Override
  public int getRowCount() {
    return model.getRowCount();
  }

  /**
   * Returns the value for a cell.
   * @param row row index
   * @param col column index
   */
  @Override
  public Object getValueAt(int row, int col) {
    if ( col == 0 )
      return table.isRowSelected( table.convertRowIndexToView(row) );
    return model.getValueAt(row, col - 1);
  }
  
  /**
   * Sets a value for a cell.
   * @param value new cell value
   * @param row row index
   * @param col column index
   */
  @Override
  public void setValueAt(Object value, int row, int col) {
    if ( col == 0 ) {
      if ( value instanceof Boolean ) {
        boolean select   = ((Boolean)value).booleanValue();
        int     tableRow = table.convertRowIndexToView(row);
        if ( select )
          table.addRowSelectionInterval(tableRow, tableRow);
        else
          table.removeRowSelectionInterval(tableRow, tableRow);
      }
    } else
      model.setValueAt(value, row, col - 1);
  }

  /**
   * Returns the value for a cell.
   * @param row row index
   * @param col column index
   * @param {@code true} for column 0
   */
  @Override
  public boolean isCellEditable(int row, int col) {
    if ( col == 0 )
      return true;
    return model.isCellEditable(row, col - 1);
  }

// !!! BY USING SelectableJTable THIS PROCEDURE IS NOT NEEDED ANYMORE !!!  
//  /**
//   * This listener must be added as {@link MouseListener} and {@link KeyListener}.
//   * It stored the current table selection (rows) on every mouse click and key
//   * stroke. On mouse click on selection column 0, the remembered selection
//   * is restored to avoid the selection clear.<br><br>
//   * 
//   * <b>Goal:</b> By selection column cells, rows should be selected individually,
//   *              without holding Strg/Shift key.<br>
//   * <b>Problem:</b> Also on a click in the selection column, the selection
//   *                 is first cleared.<br>
//   * <b>Solution:</b> After every key stroke and every mouse click, the current
//   *                  selection is remembered. On click on selection column
//   *                  (the selection is already cleared!) the selection is
//   *                  restored. After that the click (TableModel.setValue(..)) is
//   *                  progressed, which extends the selection.
//   */
//  private static class SelectionRestorer extends MouseAdapter implements KeyListener, ListSelectionListener {
//    /** Holds the table the selection is remembered from. */
//    protected JTable table = null;
//    /** Holds the remembered rows according to the model. */
//    protected int[] selectedModelRows = null;
//
//    /**
//     * Created a new selection rememberer and adds the listener to
//     * the {@link JTable}.
//     * @param table a table
//     */
//    public SelectionRestorer(JTable table) {
//      super();
//      connect(table);
//    }
//
//    /**
//     * Connects the {@link SelectionRestorer} to a {@link JTable}
//     * as {@link MouseListener}, {@link KeyListener} and {@link ListSelectionListener}.
//     * @param table a table
//     */
//    protected void connect(JTable table) {
//      this.table = table;
//      // remove all previously connected SelectionRestorers
//      disconnect();
//      // add Listeners
//      table.addMouseListener(this);
//      table.addKeyListener(this);
//      table.getSelectionModel().addListSelectionListener(this);
//    }
//    
//    /**
//     * Disconnects the {@link SelectionRestorer} from the {@link JTable}.
//     */
//    public void disconnect() {
//      // remove all previously connected SelectionRestorers
//      SelectionRestorer[] selRest = table.getListeners(SelectionRestorer.class);
//      for (SelectionRestorer selectionRestorer : selRest) {
//        table.removeMouseListener(selectionRestorer);
//        table.removeKeyListener(selectionRestorer);
//        table.getSelectionModel().removeListSelectionListener(selectionRestorer);
//      }
//    }
//
//    /**
//     * Updates the remembered selection. Calls {@link #storeSelection()}.
//     */
//    public void update() {
//      storeSelection();
//    }
//    
//    /**
//     * Stores the current selection (according to the model)
//     * in {@link #selectedModelRows}.
//     */
//    protected void storeSelection() {
//      selectedModelRows = table.getSelectedRows();
//      for (int i=0; i < selectedModelRows.length; i++)
//        selectedModelRows[i] = table.convertRowIndexToModel(selectedModelRows[i]);
//    }
//    
//    /**
//     * Restores the remembered selection.
//     */
//    protected void restoreSelection() {
//      if ( selectedModelRows == null )
//        return;
//      
//      for (int row : selectedModelRows) {
//        int tableRow = table.convertRowIndexToView(row);
//        table.addRowSelectionInterval(tableRow,tableRow);
//      }
//    }
//    
//    /**
//     * On mouse click (without holding Strg/Shift) the selection was
//     * cleared. So we restore the remembered selection.
//     */
//    @Override
//    public void mousePressed(MouseEvent e) {
//      int modelCol = table.convertColumnIndexToModel(
//                        table.columnAtPoint(e.getPoint()) );
//      if ( modelCol == 0 )
//        restoreSelection();
//    }
//    
//    /**
//     * After every click, remember the currently selected rows
//     * according to the model.
//     */
//    @Override
//    public void mouseReleased(MouseEvent e) {
//      storeSelection();
//    }
//
//    /**
//     * After every key stroke, remember the currently selected rows
//     * according to the model.
//     */
//    @Override
//    public void keyReleased(KeyEvent e) {
//      storeSelection();
//    }
//
//    /** Does nothing. */
//    @Override
//    public void keyPressed(KeyEvent e) {}
//    
//    /** Does nothing. */
//    @Override
//    public void keyTyped(KeyEvent e) {}
//    
//    /**
//     * Does currently nothing.
//     */
//    @Override
//    public void valueChanged(ListSelectionEvent e) {
//    }
//  }
}
