/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing.table;

import java.awt.Component;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;

import javax.swing.JTable;
import javax.swing.SwingUtilities;
import javax.swing.table.TableColumnModel;

/**
 * Dieser MouseListener schleust Maus-Ereignisse auf eine Tabellenzelle
 * an die darin dargestellten {@link Component}-Objekte (z.B. einen Button) durch.
 * Hierzu muss der {@link JTable} eine Instanz dieses Listeners hinzugefuegt werden.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class TableComponentMouseListener implements MouseListener {
  private JTable table;

  /**
   * Erzeugt einen neuen Listener fuer eine Tabelle.
   * @param table Tabelle fuer die die Maus-Ereignisse durchgeschleust werden
   */
  public TableComponentMouseListener(JTable table) {
    this.table = table;
  }

  private void forwardEventToButton(MouseEvent e) {
    TableColumnModel columnModel = table.getColumnModel();

    // Feststellen, auf welche Zelle geklickt wurde
    int column = columnModel.getColumnIndexAtX(e.getX());
    int row    = e.getY() / table.getRowHeight();
    if( row    >= table.getRowCount()    | row < 0 ||
        column >= table.getColumnCount() || column < 0)
      return;
    Object value = table.getValueAt(row, column);

    // nichts tun, wenn es sich nicht um eine Componente handelt
    if(!(value instanceof Component))
      return;
    Component comp = (Component)value;

    MouseEvent buttonEvent = (MouseEvent)SwingUtilities.convertMouseEvent(table, e, comp);
    comp.dispatchEvent(buttonEvent);
    // This is necessary so that when a button is pressed and released
    // it gets rendered properly.  Otherwise, the button may still appear
    // pressed down when it has been released.
    table.repaint();
  }

  /**
   * Leitet das MouseClicked-Ereignis an die {@link Component} weiter.
   * @param e MouseEvent
   */
  public void mouseClicked(MouseEvent e) {
    forwardEventToButton(e);
  }

  /**
   * Leitet das MouseEntered-Ereignis an die {@link Component} weiter.
   * @param e MouseEvent
   */
  public void mouseEntered(MouseEvent e) {
    forwardEventToButton(e);
  }

  /**
   * Leitet das MouseExited-Ereignis an die {@link Component} weiter.
   * @param e MouseEvent
   */
  public void mouseExited(MouseEvent e) {
    forwardEventToButton(e);
  }

  /**
   * Leitet das MousePressed-Ereignis an die {@link Component} weiter.
   * @param e MouseEvent
   */
  public void mousePressed(MouseEvent e) {
    forwardEventToButton(e);
  }

  /**
   * Leitet das MouseReleased-Ereignis an die {@link Component} weiter.
   * @param e MouseEvent
   */
  public void mouseReleased(MouseEvent e) {
    forwardEventToButton(e);
  }
}
