/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.swing.tree;

import javax.swing.tree.MutableTreeNode;

/**
 * Diese Klasse stellt einen inneren Knoten dar, der selbst kein Objekt
 * beinhaltet, fuer den jedoch potentielle Kinder vorgesehen sind.
 * Enthaelt er keine "richtigen" Kinder, erhaelt er einen Pseudo-Nachfolger
 * "<leer>".
 */
public class EmptyInnerNode extends EmptyNode {
  private final EmptyNode DUMMY_CHILD = new EmptyNode("<leer>");
  /**
   * Erzeugt einen neuen inneren Knoten.
   * @param desc Beschreibung
   */
  public EmptyInnerNode(String desc) {
    super(desc);
    addDummyChild();
  }

  /**
   * Fuegt den Pseudo-Nachfolger ein.
   */
  private void addDummyChild() {
    super.add(DUMMY_CHILD);
  }

  /**
   * Fuegt dem Knoten ein Kind hinzu. Der Pseudo-Nachfolger wird zuvor
   * entfernt (sofern er vorhanden war).
   */
  public void add(MutableTreeNode newChild) {
    if ( getChildCount()==1 && getFirstChild()==DUMMY_CHILD )
      super.remove(0);
    super.add(newChild);
  }

  /**
   * Entfernt ein Kind des Knotens. Falls es sich um den letzten Kindknoten
   * handelt, wird der Pseudo-Nachfolger hinzugefuegt. Der Pseude-Nachfolger
   * selbst kann nicht entfernt werden.
   */
  public void remove(MutableTreeNode aChild) {
    // Pseudo-Nachfolger kann nicht entfernt werden
    if ( aChild==DUMMY_CHILD )
      return;
    super.remove(aChild);
    if ( getChildCount() == 0 )
      addDummyChild();
  }

  /**
   * Entfernt ein Kind des Knotens. Falls es sich um den letzten Kindknoten
   * handelt, wird der Pseudo-Nachfolger hinzugefuegt. Der Pseude-Nachfolger
   * selbst kann nicht entfernt werden.
   */
  public void remove(int index) {
    // Pseudo-Nachfolger kann nicht entfernt werden
    if (getChildAt(index) == DUMMY_CHILD)
      return;
    super.remove(index);
    // Wenn der entfernte Knoten der letzte war, wird wieder ein
    // Pseudo-Nachfolger hinzugefuegt
    if ( getChildCount() == 0 )
      addDummyChild();
  }

  /**
   * Entfernt alle Kinder des Knotens und fuegt dann den Pseude-Nachfolger
   * wieder hinzu.
   */
  public void removeAllChildren() {
    super.removeAllChildren();
    // Dummy wird nur hinzugefuegt, wenn die super-Methode ihn
    // nicht bereits hinzugefuegt hat (das waere der Fall, wenn
    // die super-Methode sukzessive eine andere remove-Methode
    // aufruft.
    if ( getChildCount() == 0 )
      addDummyChild();
  }
}
