/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package schmitzm.xml;

import java.awt.Color;

import org.apache.log4j.Logger;
import org.jdom.Element;
import org.jdom.adapters.JAXPDOMAdapter;
import org.jdom.output.Format;
import org.jdom.output.XMLOutputter;

import schmitzm.swing.SwingUtil;
import skrueger.i8n.Translation;


/**
 * Diese Klasse enthaelt statische Helper-Methoden fuer das Arbeiten mit XML.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class XMLUtil {
  /** Logger, to give debug and warning messaged. */
  private static Logger LOGGER = Logger.getLogger( XMLUtil.class.getName() );

  /** Wrapper from JDOM to W3C. */
  public static final JAXPDOMAdapter JDOM_TO_JAX = new JAXPDOMAdapter();
  /** Writes XML element to file. */
  public static final XMLOutputter XML_OUTPUTTER = new XMLOutputter( Format.getPrettyFormat() );

  /**
   * Gets the attribute value from element.
   * @param element  element where the attribute is determined from
   * @param attrName name of the attribute
   * @param defValue default value returned if attribute is not found (or empty)
   * @return {@code null} if {@code element} is {@code null} or attribute is not
   *         specified in element
   */
  public static String getAttribute(Element element, String attrName, String... defValue) {
    if ( element == null )
      return null;
    String attrValue = element.getAttributeValue(attrName);
    if ( (attrValue == null || "".equals(attrValue)) && defValue.length >= 1 )
      return defValue[0];
    return attrValue;
  }

  /**
   * Gets the attribute value from element as {@link Boolean}.
   * @param element  element where the attribute is determined from
   * @param attrName name of the attribute
   * @param defValue optional default value returned if attribute is not found (or empty)
   * @return {@code null} if {@code element} is {@code null} or attribute is not
   *         specified in element
   */
  public static Boolean getBooleanAttribute(Element element, String attrName, Boolean... defValue) {
    String value = getAttribute(element, attrName);
    if ( value != null )
      return Boolean.parseBoolean(value);
    if ( defValue.length > 0 )
      return defValue[0];
    return null;
  }
  
  /**
   * Gets the attribute value from element as {@link Integer}.
   * @param element  element where the attribute is determined from
   * @param attrName name of the attribute
   * @param defValue optional default value returned if attribute is not found (or empty)
   * @return {@code null} if {@code element} is {@code null} or attribute is not
   *         specified in element
   */
  public static Integer getIntegerAttribute(Element element, String attrName, Integer... defValue) {
    String value = getAttribute(element, attrName);
    if ( value != null )
      return Integer.parseInt(value);
    if ( defValue.length > 0 )
      return defValue[0];
    return null;
  }

  /**
   * Gets the attribute value from element as {@link Double}.
   * @param element  element where the attribute is determined from
   * @param attrName name of the attribute
   * @param defValue optional default value returned if attribute is not found (or empty)
   * @return {@code null} if {@code element} is {@code null} or attribute is not
   *         specified in element
   */
  public static Double getDoubleAttribute(Element element, String attrName, Double... defValue) {
    String value = getAttribute(element, attrName);
    if ( value != null )
      return Double.parseDouble(value);
    if ( defValue.length > 0 )
      return defValue[0];
    return null;
  }

  /**
   * Gets the attribute value from element as {@link Float}.
   * @param element  element where the attribute is determined from
   * @param attrName name of the attribute
   * @param defValue optional default value returned if attribute is not found (or empty)
   * @return {@code null} if {@code element} is {@code null} or attribute is not
   *         specified in element
   */
  public static Float getFloatAttribute(Element element, String attrName, Float... defValue) {
    String value = getAttribute(element, attrName);
    if ( value != null )
      return Float.parseFloat(value);
    if ( defValue.length > 0 )
      return defValue[0];
    return null;
  }

  /**
   * Gets the attribute value from element as {@link Color}.
   * @param element  element where the attribute is determined from
   * @param attrName name of the attribute
   * @param defValue optional default value returned if attribute is not found (or empty)
   * @return {@code null} if {@code element} is {@code null} or attribute is not
   *         specified in element
   */
  public static Color getColorAttribute(Element element, String attrName, Color... defValue) {
    String value = getAttribute(element, attrName);
    if ( value != null )
      return SwingUtil.parseColor(value);
    if ( defValue.length > 0 )
      return defValue[0];
    return null;
  }

  /**
   * Sets an attribute value. The {@code .toString()} method is used
   * to convert the object value to string.<br>
   * Exceptions:
   * <ul>
   *   <li>{@link Color}: the color is converted to hex rgb string</li>
   *   <li>{@link Translation}: the translation is converted {@linkplain Translation#toOneLine() to one line}</li>
   * </ul>
   * @param element      element the attribute is set for
   * @param attrName     name of the attribute to set
   * @param value        attribute value
   * @param inclNullAttr if {@code false} and {@code value} is {@code null}
   *                     the attribute is NOT SET
   * @return {@code true} if the attribute is set
   */
  public static boolean setAttribute(Element element, String attrName, Object value, boolean inclNullAttr) {
    if ( value == null && !inclNullAttr )
      return( false );
    
    if ( value == null )
      value = "";
    
    // Default: use the internal string conversion
    String valueStr = value.toString();
    // Special: convert color to hex rgb string
    if ( value instanceof Color )
      valueStr = SwingUtil.convertColorToHex((Color)value);
    // Special: convert Translation to one line
    if ( value instanceof Translation )
      valueStr = ((Translation)value).toOneLine();
    
    element.setAttribute(attrName, valueStr);
  
    return true;
  }
  
  /**
   * Sets an attribute value if the value is not {@code null}.
   * The {@code .toString()} method is used to convert the object
   * value to string.<br>
   * Exceptions:
   * <ul>
   *   <li>{@link Color}: the color is converted to hex rgb string</li>
   * </ul>
   * @param element    element the attribute is set for
   * @param attrName   name of the attribute to set
   * @param value      attribute value
   * @return {@code true} if the attribute is set
   */
  public static boolean setNotNullAttribute(Element element, String attrName, Object value) {
    return setAttribute(element, attrName, value, false);
  }
  
//  /**
//   * Returns a {@link DocumentBuilder}. If validating is true,
//   * the contents is validated against the XSD specified in the file.
//   */
//  private final static DocumentBuilder getDocumentBuilder(final boolean validating) throws SAXException, IOException, ParserConfigurationException {
//    // Create a builder factory
//    final DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
//
//    factory.setNamespaceAware(true);
//    final String JAXP_SCHEMA_LANGUAGE = "http://java.sun.com/xml/jaxp/properties/schemaLanguage";
//    final String W3C_XML_SCHEMA = "http://www.w3.org/2001/XMLSchema";
//    factory.setAttribute(JAXP_SCHEMA_LANGUAGE, W3C_XML_SCHEMA);
////    factory.setValidating(validating);
//
//    // Create the builder and parse the file
//    final DocumentBuilder documentBuilder = factory.newDocumentBuilder();
////    documentBuilder.setErrorHandler(new ErrorHandler() {
////      public void error(final SAXParseException exception) throws SAXException {
////        LOGGER.error("ErrorHandler.error");
////        throw exception;
////      }
////      public void fatalError(final SAXParseException exception) throws SAXException {
////        LOGGER.error("ErrorHandler.fataError");
////        throw exception;
////      }
////      public void warning(final SAXParseException exception) throws SAXException {
////        LOGGER.warn("ErrorHandler.warning", exception);
////      }
////    });
////
//    return documentBuilder;
//  }

//  /**
//   * Wraps a {@linkplain Element JDOM-Element} to a {@linkplain Node W3C-Node}.
//   * @param element a JDOM-Element
//   */
//  public static Node wrapElement(Element element) {
////    String  xmlDefinition = new XMLOutputter().outputString( new Document(element,element.getDocument().getDocType()) );
////    Document document = new DefaultJDOMFactory().document( element );
//    String  xmlString = new XMLOutputter().outputString( element );
//
//    try {
//      InputStream xmlStream = new ByteArrayInputStream(xmlString.getBytes());
//      DocumentBuilder builder = getDocumentBuilder(false);
//      org.w3c.dom.Document node = builder.parse(xmlStream);
//
////      org.w3c.dom.Document node = JDOM_TO_JAX.getDocument(xmlStream,true);
//
//      LOGGER.debug( xmlString );
//      LOGGER.debug( node.hasAttributes() );
//      xmlStream.close();
//      return node;
//    } catch ( Exception err ) {
//      throw new UnsupportedOperationException("Error during XML wrapping", err);
//    }
//  }

}
