/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package skrueger.geotools;

import javax.swing.ImageIcon;

import org.apache.log4j.Logger;
import org.geotools.styling.Style;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import schmitzm.lang.LangUtil;
import skrueger.i8n.Translation;

import com.vividsolutions.jts.geom.Envelope;

/**
 * This class is a default implementation of {@link StyledLayerInterface}.
 * {@link StyledLayerInterface#dispose()} and {@link StyledLayerInterface#uncache()}
 * must be implemented by the sub class. This class only implements the "hold"
 * of an geo object of type {@code <E>}.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public abstract class AbstractStyledLayer<E> implements StyledLayerInterface<E> {
  /** Logger for warning- and error messages. */
  protected Logger LOGGER = LangUtil.createLogger(this);

  /** Holds the unique ID of the geo object. */
  protected String id = null;
  /** Holds a short (language-specific) description of the geo object. */
  protected Translation title = null;
  /** Holds a long (language-specific) description of the geo object. */
  protected Translation desc = null;
  /** Holds the (language-specific) keywords to describe the geo object. */
  protected Translation keywords = null;
  /** Holds an icon to represent the geo object */
  protected ImageIcon icon = null;
  /** Holds the geo object represeneted by the map */
  protected E geoObject = null;
  /** Holds the CRS of the geo object */
  protected CoordinateReferenceSystem crs = null;
  /** Holds the bounds of the geo object */
  protected Envelope envelope = null;
  /** Holds the display style for the geo object */
  protected Style style = null;

  /**
   * Creates a language specific styled layer.
   * @param geoObject the geo object
   * @param envelope the bounds of the geo object
   * @param crs the CRS of the geo object
   * @param id a unique ID for the geo object
   * @param title a (language-specific) short description
   * @param desc a (language-specific) long description
   * @param keywords (language-specific) keywords for the geo objects
   * @param style a display style
   * @param icon an icon for the object
   * @exception IllegalArgumentException if {@code null} is given as ID or
   *            geo object
   */
  public AbstractStyledLayer(E geoObject, Envelope envelope, CoordinateReferenceSystem crs, String id, Translation title, Translation desc, Translation keywords, Style style, ImageIcon icon) {
    if ( id == null )
      throw new IllegalArgumentException("ID is not allowed to be null!");
    if ( geoObject == null )
      throw new IllegalArgumentException("The GeoObject is not allowed to be null!");
    this.id        = id;
    this.geoObject = geoObject;
    this.crs       = crs;
    this.envelope  = envelope;
    setTitle( title );
    setDesc( desc );
    setKeywords( keywords );
    setStyle( style );
    setImageIcon( icon );
  }

  /**
   * Creates a non-translated styled layer.
   * @param geoObject the geo object
   * @param envelope the bounds of the geo object
   * @param crs the CRS of the geo object
   * @param id a unique ID for the geo object
   * @param title a short description
   * @param desc a long description
   * @param keywords keywords for the geo objects
   * @param style a display style
   * @param icon an icon for the object
   * @exception IllegalArgumentException if {@code null} is given as ID
   */
  public AbstractStyledLayer(E geoObject, Envelope envelope, CoordinateReferenceSystem crs, String id, String title, String desc, String keywords, Style style, ImageIcon icon ) {
    this(geoObject, envelope, crs, id, (Translation)null, null, null, style, icon);
    setTitle( title );
    setDesc( desc );
    setKeywords( keywords );
  }

  /**
   * Returns a ID for the geo object. The ID should be unique in a map ob
   * {@linkplain StyledLayerInterface styled layer objects}
   */
  public String getId() {
    return id;
  }

  /**
   * Returns a short (language-specific) description of the geo object.
   */
  public Translation getTitle() {
    return title;
  }

  /**
   * Sets a short (language-specific) description of the geo object.
   * If {@code title} is {@code null} an untranslated default title is set, so
   * {@link #getTitle()} never returns {@code null}.
   * @param title new description for the geo object
   */
  public void setTitle(Translation title) {
    this.title = (title != null) ? title : new Translation("untitled");
  }

  /**
   * Sets a short (non-translated) description of the geo object.
   * If {@code title} is {@code null} an untranslated default title is set, so
   * {@link #getTitle()} never returns {@code null}.
   * @param title new description for the geo object
   */
  public void setTitle(String title) {
    setTitle( title != null ? new Translation(title): (Translation)null );
  }

  /**
   * Returns a long (language-specific) description of the object.
   */
  public Translation getDesc() {
    return desc;
  }

  /**
   * Sets a long (language-specific) description of the object.
   * If {@code desc} is {@code null} an (untranslated) empty description is set, so
   * {@link #getDesc()} never returns {@code null}.
   * @param desc new description for the geo object
  */
  public void setDesc(Translation desc) {
    this.desc = (desc != null) ? desc : new Translation("");
  }

  /**
   * Sets a long (non-translated) description of the object.
   * If {@code desc} is {@code null} an (untranslated) empty description is set, so
   * {@link #getDesc()} never returns {@code null}.
   * @param desc new description for the geo object
  */
  public void setDesc(String desc) {
    setDesc( desc != null ? new Translation(desc) : (Translation)null);
  }

  /**
   * Returns a (language-specific) key word sequence for the geo object.
   */
  public Translation getKeywords() {
    return keywords;
  }

  /**
   * Sets a (language-specific) key word sequence for the geo object.
   * If {@code keywords} is {@code null} an (untranslated) empty string is set, so
   * {@link #getKeywords()} never returns {@code null}.
   * @param keywords Keywords
   */
  public void setKeywords(Translation keywords) {
    this.keywords = (keywords != null) ? keywords : new Translation("");
  }

  /**
   * Sets a (non-translated) key word sequence for the geo object.
   * If {@code keywords} is {@code null} an (untranslated) empty string is set, so
   * {@link #getKeywords()} never returns {@code null}.
   * @param keywords Keywords
   */
  public void setKeywords(String keywords) {
    setKeywords( keywords != null ? new Translation(keywords) : (Translation)null);
  }

  /**
   * Returns the geo object representet in the map. Sub classes must override
   * this method to implement "late loading" on first call.
   * @return {@link #geoObject}
   */
  public E getGeoObject() {
    return geoObject;
  }

  /**
   * Returns the bounds of the geo object.
   */
  public Envelope getEnvelope() {
    return envelope;
  }

  /**
   * Returns the {@link CoordinateReferenceSystem} of the geo object.
   */
  public CoordinateReferenceSystem getCrs() {
    return crs;
  }

  /**
   * Returns {@link #crs CoordinateReferenceSystem.toString()}. This method
   * can be overriden to create a "nicer" description.
   */
  public String getCRSString() {
    return crs.toString();
  }

  /**
   * Returns an icon, which represents the geo object.
   */
  public ImageIcon getImageIcon() {
    return null;
  }

  /**
   * Sets an icon, which represents the geo object.
   * @param icon an icon
   */
  public void setImageIcon(ImageIcon icon) {
    this.icon = icon;
  }

  /**
   * Returns the display style for the geo object.
   */
  public Style getStyle() {
    return style;
  }

  /**
   * Sets the display style for the geo object.
   * If {@code style} is {@code null} an default style is set, so
   * {@link #getStyle()} never returns {@code null}.
   * @see #createDefaultStyle()
   */
  public void setStyle(Style style) {
    this.style = (style != null) ? style : createDefaultStyle();
  }

  /**
   * Creates a default style for the geo object. This style is used whenever
   * the style is set to {@code null}.
   * @see #setStyle(Style)
   */
  protected abstract Style createDefaultStyle();
}
