package skrueger.geotools;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.apache.log4j.Logger;
import org.geotools.feature.NameImpl;
import org.opengis.feature.type.Name;

import skrueger.AttributeMetadata;
import skrueger.i8n.Translation;

/**
 * The {@link AttributeMetadataMap} is a {@link Map} holding
 * {@link AttributeMetadata} object for {@link Name} keys. It's an extension of
 * a {@link TreeMap}. It's copyable in the sense of the {@link Copyable}
 * interface.<br>
 * The {@link #get(Name)} and {@link #get(String)} methods will never return
 * <code>null</code>, but rather create a default {@link AttributeMetadata}
 * on-the-fly.
 */
public class AttributeMetadataMap extends TreeMap<Name, AttributeMetadata>
		implements Copyable<AttributeMetadataMap> {
	static private final Logger LOGGER = Logger
			.getLogger(AttributeMetadataMap.class);

	/**
	 * A list of default languages used when creating default
	 * {@link AttributeMetadata} on-the-fly. If not initialized by a
	 * constructor, it will be {@link Translation#getActiveLang()} only.
	 **/
	private final List<String> langs;

	/**
	 @Deprecated Use another constructor. AttributeMetadataMap is based on {@link Translation} and
	 *             would like to know how many languages are supported, so the
	 *             needed defaults can be created.
	 */
	public AttributeMetadataMap() {
		langs = new ArrayList<String>();
		langs.add(Translation.getActiveLang());
	}

	/**
	 * Creates an {@link AttributeMetadataMap} and sets the list of default
	 * languages.
	 */
	public AttributeMetadataMap(final List<String> defLanguages) {
		langs = defLanguages;
	}

	/**
	 * Creates an {@link AttributeMetadataMap} and sets the list of default
	 * languages.
	 */
	public AttributeMetadataMap(final String[] strings) {
		langs = new ArrayList<String>(java.util.Arrays.asList(strings));
	}

	/**
	 * Returns a deep-copy. @see {@link Copyable} interface
	 */
	@Override
	public AttributeMetadataMap copy() {
		final AttributeMetadataMap copy = new AttributeMetadataMap(langs);
		return copyTo(copy);
	}

	/**
	 * Copies all its values to another {@link AttributeMetadataMap}. @see
	 * {@link Copyable} interface.
	 */
	@Override
	public AttributeMetadataMap copyTo(final AttributeMetadataMap amdMap) {

		amdMap.clear();

		for (final Name key : keySet()) {
			final AttributeMetadata attributeMetaData = get(key);
			amdMap.put(key, attributeMetaData.copy());
		}
		return amdMap;
	}

	/**
	 * Returns the {@link AttributeMetadata} for a given {@link Name}. Never
	 * returns <code>null</code>, but rather creates a default
	 * {@link AttributeMetadata} on the fly.
	 */
	public AttributeMetadata get(final Name name) {
		final AttributeMetadata attributeMetadata = super.get(name);
		if (attributeMetadata == null && name != null
				&& !name.getLocalPart().trim().isEmpty()) {
			put(name, new AttributeMetadata(name, langs));
			return super.get(name);
		}
		return attributeMetadata;
	}

	/**
	 * @Deprecated use get(Name name) or get(String localName)
	 */
	@Override
	@Deprecated
	public AttributeMetadata get(final Object key) {
		LOGGER.warn("PLEASE DONT USE get(Object) any MORE!");
		LOGGER.warn("PLEASE DONT USE get(Object) any MORE!");
		LOGGER.warn("PLEASE DONT USE get(Object) any MORE!");
		return super.get(key);
	}

	/**
	 * Returns the {@link AttributeMetadata} for a given {@link Name}. Never
	 * returns <code>null</code>, but rather creates a default
	 * {@link AttributeMetadata} on the fly.
	 */
	public AttributeMetadata get(final String localName) {
		if (localName == null) return null;
		return this.get(new NameImpl(localName));
	}

	public List<String> getLanguages() {
		return langs;
	}

	/**
	 * @return a number between 0. (bad) and 1. (good) that is calculated from
	 *         the amount of translation available in the visible attributes
	 */
	public double getQuality(final List<String> languages) {
		int allVisible = 0;
		double colQmSum = 0.;
		for (final AttributeMetadata oneCol : values()) {

			if (oneCol.isVisible()) {
				allVisible++;
				colQmSum += oneCol.getQuality(languages);
			}
		}

		if (allVisible > 0)
			return colQmSum / allVisible;
		else
			return 1.;

	}


	/**
	 * @return List of {@link AttributeMetadata} objects ordered by their
	 *         weight. (heavier => further down)
	 */
	public List<AttributeMetadata> sortedValues() {
		final ArrayList<AttributeMetadata> list = new ArrayList<AttributeMetadata>();
		list.addAll(values());
		Collections.sort(list);
		return list;
	}

	/**
	 * @return List of only the visible {@link AttributeMetadata} objects
	 *         ordered by their weight. (heavier => further down)
	 */
	public List<AttributeMetadata> sortedValuesVisibleOnly() {
		final ArrayList<AttributeMetadata> list = new ArrayList<AttributeMetadata>();
		for (final AttributeMetadata atm : sortedValues()) {
			if (atm.isVisible())
				list.add(atm);
		}
		return list;
	}
}
