package skrueger.geotools;

import java.awt.geom.Point2D;

import org.apache.log4j.Logger;
import org.geotools.data.FeatureSource;
import org.geotools.filter.AbstractFilter;
import org.geotools.filter.GeometryFilterImpl;
import org.geotools.geometry.jts.ReferencedEnvelope;
import org.geotools.referencing.CRS;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.simple.SimpleFeatureType;
import org.opengis.feature.type.GeometryDescriptor;
import org.opengis.filter.expression.Expression;
import org.opengis.filter.spatial.BinarySpatialOperator;
import org.opengis.geometry.DirectPosition;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import schmitzm.geotools.FilterUtil;
import schmitzm.geotools.JTSUtil;
import schmitzm.geotools.io.GeoImportUtil;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Envelope;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.MultiPolygon;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.geom.Polygon;

/**
 * The {@link GeomFilterGenerator} prepares a {@link BinarySpatialOperator}
 * filter for multiple use. Only the "right" argument is prepared. The "left"
 * argument (the geometry attribute of the {@link FeatureSource} to filter) is
 * first set on calling {@link #adaptFilter(FeatureSource)} for a specific
 * {@link FeatureSource}. This method also takes care to recreate the filter (or
 * its "right" argument) if the given {@link FeatureSource} has another
 * {@link CoordinateReferenceSystem} than the base constraint.<br>
 * The type of filter (e.g. distance or bounding box) is specified by the
 * subclass implemenation of {@link #prepareFilter(CoordinateReferenceSystem)} .
 * 
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 */
public abstract class GeomFilterGenerator {
	private final static Logger LOGGER = Logger
			.getLogger(GeomFilterGenerator.class);

	/**
	 * Holds the {@link CoordinateReferenceSystem} the filter constraint bases
	 * on.
	 */
	protected CoordinateReferenceSystem baseCRS = null;

	// /**
	// * Caches the base filter constraint for several
	// * {@link CoordinateReferenceSystem CoordinateReferenceSystems} and the
	// Geometry-Type of
	// */
	// protected Map<Integer, GeometryFilterImpl> filterCache = new
	// HashMap<Integer, GeometryFilterImpl>();

	/**
	 * Creates a new filter generator
	 * 
	 * @param crs
	 *            {@link CoordinateReferenceSystem} the base constraint ("right"
	 *            filter argument is relative to)
	 */
	public GeomFilterGenerator(CoordinateReferenceSystem crs) {
		this.baseCRS = crs;
	}

	/**
	 * Creates a filter containing the base constraint ("right" argument)
	 * transformed to the given {@link CoordinateReferenceSystem}.<br>
	 * Depending on the Geometry type of the targeted layer, the filter may look
	 * different. For example checking POINT INTERSECTS POLYGON is easy, where
	 * POINT INTERSECTS LINE will hardly ever hit, so we do
	 * SMALLBBOXARROUND(POINT) INTERSECTS POLYGON.
	 * 
	 * @param crs
	 *            the {@link CoordinateReferenceSystem} the base constraint is
	 *            transformed to
	 * @param Geometry
	 *            class of the features we want to filter against.
	 * @param Class
	 *            the geometry class we are testing against. For point against
	 *            polygon for example, an intersects is much faster that the
	 *            bbox filter
	 */
	protected abstract GeometryFilterImpl prepareFilter(
			CoordinateReferenceSystem crs, Class<?> geomClass);

	/**
	 * Completes the filter with its "left" argument for a concrete
	 * {@link FeatureSource}. If the {@link FeatureSource FeatureSource's} CRS
	 * differs from the CRS the base constraint is specified in, first a new
	 * filter is created by calling
	 * {@link #prepareFilter(CoordinateReferenceSystem)}.
	 * 
	 * @param fs
	 *            {@link FeatureSource} the filter is adaped to
	 * @return
	 */
	public GeometryFilterImpl adaptFilter(
			FeatureSource<SimpleFeatureType, SimpleFeature> fs) {
		GeometryDescriptor geomDescr = fs.getSchema().getGeometryDescriptor();
		CoordinateReferenceSystem fsCRS = geomDescr
				.getCoordinateReferenceSystem();
		if (fsCRS == null)
			fsCRS = GeoImportUtil.getDefaultCRS();
		Class<?> binding = geomDescr.getType().getBinding();

		GeometryFilterImpl filter;
		// = filterCache.get(fsCRS.hashCode()+binding.hashCode());
		// if (filter == null) {
		filter = prepareFilter(fsCRS, binding);
		// filterCache.put(fsCRS.hashCode()+binding.hashCode(), filter);
		// }
		Expression geometry = FilterUtil.FILTER_FAC2.property(geomDescr
				.getLocalName());
		filter.setExpression1(geometry);
		return filter;
	}

	/**
	 * {@link GeomFilterGenerator} for a bounding box constraint.
	 * 
	 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
	 */
	public static class BoundingBoxFilterGenerator extends GeomFilterGenerator {
		private final static Logger LOGGER = Logger
				.getLogger(BoundingBoxFilterGenerator.class);

		/**
		 * Holds the base constraint (bounding box {@link Envelope}) relative to
		 * the {@linkplain GeomFilterGenerator#baseCRS base CRS}.
		 */
		protected Envelope baseEnv = null;

		/**
		 * Creates a new filter generator.
		 * 
		 * @param baseEnv
		 *            defines the bounding box
		 * @param baseCrs
		 *            defines the CRS of the bounding box
		 */
		public BoundingBoxFilterGenerator(Envelope baseEnv,
				CoordinateReferenceSystem baseCrs) {
			super(baseCrs);
			this.baseEnv = baseEnv;
		}

		/**
		 * Prepares a filter with the bounding box transformed to the given
		 * {@link CoordinateReferenceSystem} as the "right" argument.
		 * 
		 * @param crs
		 *            the {@link CoordinateReferenceSystem} the bounding box is
		 *            transformed to * @param Class the geometry class we are
		 *            testing against. For point against polygon for example, an
		 *            intersects is much faster that the bbox filter
		 */
		protected GeometryFilterImpl prepareFilter(
				CoordinateReferenceSystem crs, Class<?> geomClass) {
			Envelope bbEnv = baseEnv;
			if (!CRS.equalsIgnoreMetadata(baseCRS, crs))
				bbEnv = JTSUtil.transformEnvelope(baseEnv, baseCRS, crs);
			Expression bbox = FilterUtil.FILTER_FAC.createBBoxExpression(bbEnv);
			GeometryFilterImpl bboxFilter = (GeometryFilterImpl) FilterUtil.FILTER_FAC
					.createGeometryFilter(AbstractFilter.GEOMETRY_INTERSECTS);
			bboxFilter.setExpression2(bbox);
			return bboxFilter;
		}
	}

	/**
	 * {@link GeomFilterGenerator} for a "near distance" constraint.
	 * 
	 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
	 */
	public static class PointFilterGenerator extends GeomFilterGenerator {
		private final static Logger LOGGER = Logger
				.getLogger(PointFilterGenerator.class);

		private static final GeometryFactory GF = new GeometryFactory(); // TODO
		// do
		// we
		// have
		// a
		// application
		// wide
		// one?

		/**
		 * Holds the base constraint (coordinate) relative to the
		 * {@linkplain GeomFilterGenerator#baseCRS base CRS}.
		 */
		protected Point basePoint = null;

		private ReferencedEnvelope baseEnv;

		/**
		 * Creates a new filter generator.
		 * 
		 * @param basePoint
		 *            defines the point for the "near point" constraint
		 * @param dist
		 *            defines the distance around the base point
		 * @param crs
		 *            defines the CRS of base point
		 */
		public PointFilterGenerator(Point basePoint, double radius,
				CoordinateReferenceSystem crs) {
			super(crs);
			this.basePoint = basePoint;
			this.baseEnv = new ReferencedEnvelope(basePoint.getX() - radius,
					basePoint.getX() + radius, basePoint.getY() - radius,
					basePoint.getY() + radius, crs);
		}

		/**
		 * Creates a new filter generator.
		 * 
		 * @param basePoint
		 *            defines the point for the "near point" constraint
		 * @param dist
		 *            defines the distance around the base point in base point
		 *            CRS
		 * @param crs
		 *            defines the CRS of base point
		 */
		public PointFilterGenerator(Point2D basePoint, double dist,
				CoordinateReferenceSystem crs) {
			this(GF.createPoint(new Coordinate(basePoint.getX(), basePoint
					.getY())), dist, crs);
		}
		

		/**
		 * Creates a new filter generator.
		 * 
		 * @param basePoint
		 *            defines the Position for the "near point" constraint
		 * @param dist
		 *            defines the distance around the base point in base point
		 *            CRS
		 */
		public PointFilterGenerator(DirectPosition basePos, double dist) {
			this(GF.createPoint(new Coordinate(basePos.getCoordinate()[0], basePos
					.getCoordinate()[1])), dist, basePos.getCoordinateReferenceSystem());
		}

		/**
		 * Prepares a filter with the base point and distance transformed to the
		 * given {@link CoordinateReferenceSystem} as the "right" argument.
		 * 
		 * @param crs
		 *            the {@link CoordinateReferenceSystem} the point and
		 *            distance is transformed to
		 */
		protected GeometryFilterImpl prepareFilter(
				CoordinateReferenceSystem crs, Class<?> geomClass) {
			Point localCrsPoint = basePoint;

			boolean isPolygonGeometry = (Polygon.class
					.isAssignableFrom(geomClass) || MultiPolygon.class
					.isAssignableFrom(geomClass));

			LOGGER.debug("is polygon = " + isPolygonGeometry);

			GeometryFilterImpl filter = (GeometryFilterImpl) FilterUtil.FILTER_FAC
					.createGeometryFilter(AbstractFilter.GEOMETRY_INTERSECTS);

			if (isPolygonGeometry) {

				localCrsPoint = (Point) JTSUtil.transformGeometry(
						localCrsPoint, baseCRS, crs);

				filter.setExpression2(FilterUtil.FILTER_FAC2
						.literal(localCrsPoint));

			} else {
				// we check against lines and points with a bbox

				Envelope bbEnv = JTSUtil.transformEnvelope(baseEnv, baseCRS,
						crs);

				// Filter fuer Envelope zusammenstellen
				Expression bbox = FilterUtil.FILTER_FAC
						.createBBoxExpression(bbEnv);
				filter.setExpression2(bbox);
			}

			return filter;

		}

	}

}