/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package skrueger.geotools;
import java.util.List;

import org.geotools.map.MapContext;
import org.geotools.map.MapLayer;
import org.geotools.map.event.MapLayerListListener;

import skrueger.AttributeMetadata;
import skrueger.RasterLegendData;

/**
 * Implementations of this class can can be used to fill/insert/remove a mapContext
 * with {@link StyledLayerInterface} objects.
 *
 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons Kr&uuml;ger</a>
 *
 */
public interface MapContextManagerInterface {

	/**
	 * Add a {@link StyledLayerInterface} object into the underlying {@link MapContext}
	 * as the topmost layer
	 *
	 * @param styledLayer Layer to add to the map
	 */
	boolean addStyledLayer (StyledLayerInterface<?> styledLayer);

	/**
	 * @param mapContextIdx Index in the {@link MapContext}, bottom first
	 * @return successful?
	 */
	boolean removeStyledLayer (int mapContextIdx);

	/**
	 * Inserts a {@link StyledLayerInterface} object into the underlying {@link MapContext}
	 * at the given position
	 */
	boolean insertStyledLayer (StyledLayerInterface<?> styledLayer, int mapContextIdx);

	/** Add {@link MapLayerListListener} */
	void addMapLayerListListener( MapLayerListListener listener);

	/** Remove {@link MapLayerListListener} */
	void removeMapLayerListListener( MapLayerListListener listener);

	/**
	 * Help the GC by disposing this Component
	 */
	void dispose();

	/**
	 * Returns a ordered list of the layers that are contained in the underlying {@link MapContext}
	 */
	List<StyledLayerInterface<?>> getStyledObjects();

	/**
	 * Returns a list of {@link AttributeMetadata} that shall be shown (e.g. when the Mouse clicked into the map)
	 * Returns an empty list if the layer doesn't exist or is not backed by a {@link StyledFeatureCollectionInterface}
	 */
	List<AttributeMetadata> getVisibleAttribsFor(MapLayer mapLayer);

	/**
	 * Returns the title of the layer
	 * @param mapLayer {@link MapLayer}
	 * @return <code>null</code>, if the layer is unknown
	 */
	String getTitleFor(MapLayer mapLayer);

	/**
	 * Returns the description of the layer
	 * @param mapLayer {@link MapLayer}
	 * @return <code>null</code>, if the layer is unknown. Empty String if the description is empty
	 */
	String getDescFor(MapLayer mapLayer);

	/**
	 * Returns the {@link RasterLegendData} object for the layer.
	 * @return null, if the layer is not found or of type raster
	 */
	RasterLegendData getLegendMetaData(MapLayer mapLayer);

	/**
	 * Returns the {@link StyledLayerInterface} object that is associated with the layer or NULL if the layer can't be found.
	 */
	StyledLayerInterface<?> getStyledObjectFor(MapLayer layer);
	
	/**
	 * @return The {@link MapContext} managed by this {@link MapContextManagerInterface} 
	 */
	MapContext getMapContext();

}
