package skrueger.geotools;

import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.geom.AffineTransform;

import org.apache.log4j.Logger;
import org.geotools.geometry.jts.ReferencedEnvelope;
import org.geotools.renderer.GTRenderer;
import org.geotools.renderer.RenderListener;
import org.opengis.feature.simple.SimpleFeature;

/**
 * This class is used by {@link XMapPane} to start and stop the rendering a
 * {@link Thread} for rendering.
 */
class RenderingExecutor {
	private final static Logger LOGGER = Logger.getLogger(RenderingExecutor.class);
	/**
	 * Instance to a {@link RenderThread} doing any work. It's volatile so the
	 * correct value will always be visible to any {@link Thread}
	 **/
	private volatile RenderThread renderThread;

	private final XMapPane mapPane;

	public RenderingExecutor(XMapPane mapPane) {
		this.mapPane = mapPane;
	}

	/**
	 * Submit a new rendering task. If no rendering task is presently running
	 * this new job will be accepted; otherwise it will be rejected and it
	 * returns <code>false</code>.
	 * 
	 * @param envelope
	 *            the map area (world coordinates) to be rendered.
	 * @param graphics
	 *            the graphics object to draw on.
	 * @param paintArea
	 *            size of the area to paint the world into.
	 * @param worldToScreen
	 *            the {@link AffineTransform} from world coordinates to screen
	 *            coordinates.
	 * @param renderer
	 *            the {@link GTRenderer} to use.
	 * 
	 * @return true if the rendering task was accepted; false if it was rejected
	 */
	public synchronized boolean submit(ReferencedEnvelope envelope,
			Rectangle paintArea, Graphics2D graphics, final GTRenderer renderer
	// , AffineTransform worldToScreen
	) {
		if (renderThread == null || !renderThread.isAlive()) {
			// System.out.println("is vacant... starting thread!");
			renderThread = null;

			renderThread = new RenderThread(paintArea, graphics, renderer,
			// worldToScreen,
					envelope);
			renderThread.start();

			return true;
		} else {
			// System.out.println("is busy... requesting stop!");
			renderThread.getRenderer().stopRendering();
			return false;
		}
	}

	/**
	 * For every new rendering job submitted and accepted, an instance of this
	 * {@link Thread} will be started.
	 * 
	 */
	class RenderThread extends Thread {

		private final GTRenderer renderer;

		public RenderThread(final Rectangle paintArea,
				final Graphics2D graphics, GTRenderer renderer,
				// AffineTransform worldToScreen,
				ReferencedEnvelope mapEnv) {
			super(new RenderRun(paintArea, graphics, renderer, mapEnv
			// , worldToScreen
					));
			this.renderer = renderer;

			setName("Render" + getName());

			// System.out.println("starting render thread " + getName());
		}

		public GTRenderer getRenderer() {
			return renderer;
		}

	}

	/**
	 * This {@link Runnable} will actually start the rendering
	 */
	class RenderRun implements Runnable, RenderListener {
		private final Rectangle paintArea;
		private final Graphics2D graphics;
		// private final AffineTransform worldToScreen;
		private final GTRenderer renderer;
		private final ReferencedEnvelope mapEnv;

		public RenderRun(Rectangle paintArea, Graphics2D graphics,
				GTRenderer renderer, ReferencedEnvelope mapEnv
		// ,
		// AffineTransform worldToScreen
		) {
			this.paintArea = paintArea;
			this.graphics = graphics;
			this.renderer = renderer;
			this.mapEnv = mapEnv;
			// this.worldToScreen = worldToScreen;
		}

		@Override
		public void run() {
			long startT = System.currentTimeMillis();
			try {
				renderer.addRenderListener(this);
//				LOGGER.debug("start rendering...");

				// Clear the graphics context
				graphics.setBackground(mapPane.getMapBackgroundColor());
				graphics.clearRect(paintArea.x, paintArea.y, paintArea.width,
						paintArea.height);

				renderer.paint(graphics, paintArea, mapEnv);

				// Kill the reference to this Thread so #isRunning will say
				// false directly
				renderThread = null;
				mapPane.onRenderingCompleted(System.currentTimeMillis()-startT);
			} catch (Exception e) {
				mapPane.onRenderingFailed(e);
			} finally {
				renderer.removeRenderListener(this);
			}
		}

		@Override
		public void errorOccurred(Exception e) {
			mapPane.onRenderingFailed(e);
		}

		@Override
		public void featureRenderer(SimpleFeature feature) {
		}

	}

	/**
	 * Ask to stop the rendering. May be called often.
	 */
	public void cancelTask() {
		if (renderThread != null)
			synchronized (renderThread) {
				if (renderThread.isAlive()) {
					if (renderThread.getRenderer() != null)
						renderThread.getRenderer().stopRendering();
				}
			}
	}

	/**
	 * @return <code>true</code> if the {@link Thread} is busy rendering.
	 */
	public boolean isRunning() {
		return (renderThread != null && renderThread.isAlive());
	}

	/**
	 * Will stop rendering and remove the reference to the {@link Thread}.
	 */
	public void dispose() {
		if (renderThread != null) {
			renderThread.renderer.stopRendering();
			renderThread = null;
		}
	}

}
