/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package skrueger.geotools;

import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.util.Date;
import java.util.Random;

import javax.swing.ImageIcon;
import javax.swing.JPanel;

import org.apache.log4j.Logger;
import org.geotools.data.FeatureSource;
import org.geotools.feature.FeatureCollection;
import org.geotools.styling.Style;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.simple.SimpleFeatureType;
import org.opengis.feature.type.AttributeDescriptor;
import org.opengis.filter.Filter;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import schmitzm.geotools.io.GeoImportUtil;
import schmitzm.geotools.styling.StylingUtil;
import skrueger.AttributeMetadata;
import skrueger.i8n.Translation;

import com.vividsolutions.jts.geom.Envelope;

/**
 * This class enables a non Atlas context to use the Atlas LayerPanel
 * {@link JPanel} as a {@link MapContextManagerInterface}
 * 
 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons Kr&uuml;ger</a>
 * 
 * TODO Rename to StyledShapefile
 */
public class StyledFS implements StyledFeatureSourceInterface {
	private static final Logger LOGGER = Logger.getLogger(StyledFS.class);

	private final FeatureSource<SimpleFeatureType, SimpleFeature> fs;
	
	/** Caching the CRS of the layer **/
	CoordinateReferenceSystem crs = null;

	/**
	 * A unique ID which identifies the Layer in the Atlas. It's more important
	 * than it should be ;-)
	 */
	private String id;

	private Style style;

	private Translation title;

	private Translation desc;

	private File sldFile;

	private AttributeMetadataMap map;

	private Filter filter = Filter.INCLUDE;

	/**
	 * This class enables a non Atlas context to use the Atlas LayerPanel
	 * {@link JPanel} as a {@link MapContextManagerInterface}
	 * 
	 * @param fs
	 *            {@link FeatureSource} that is beeing styled.
	 * 
	 * @param sldFile
	 *            may be <code>null</code>. Otherwise the SLD {@link File} to
	 *            import and associate with this {@link StyledFS}
	 */
	public StyledFS(FeatureSource<SimpleFeatureType, SimpleFeature> fs,
			File sldFile) {

		this.fs = fs;

		id = StyledFS.class.getSimpleName()
				+ new Random(new Date().getTime()).nextInt(10000000);

		this.sldFile = sldFile;

		// datei existiert, dann lesen
		if (sldFile != null && sldFile.exists()) {
			try {
				style = StylingUtil.loadSLD(sldFile)[0];
			} catch (Exception e) {
				LOGGER.warn("Reading SLD failed: " + sldFile, e);
				style = null;
			}
		}

		title = new Translation();
		desc = new Translation();

		if (sldFile != null) {
			title.fromOneLine(sldFile.getName());
			desc.fromOneLine(sldFile.getAbsolutePath());
		}

	}

	public StyledFS(FeatureSource<SimpleFeatureType, SimpleFeature> fs) {
		this(fs, null);
	}

	public void dispose() {
	}

	/**
	 * Returns human readable {@link String} of the CRS natively used by this
	 * {@link DpLayer}
	 * 
	 * If CRS == null, it will call {@link #getGeoObject()}
	 * 
	 */
	public String getCRSString() {
		if (getCrs() == null)
			return "CRS?";

		return getCrs().getName().getCode();
	}

	public CoordinateReferenceSystem getCrs() {
		if (crs == null) {
			crs = fs.getSchema()
					.getCoordinateReferenceSystem();
			if (fs.getSchema().getCoordinateReferenceSystem() == null) {
				LOGGER.warn("Could not determine the CRS of " + getTitle()
						+ ". Using default " + GeoImportUtil.getDefaultCRS());
				crs = GeoImportUtil.getDefaultCRS();
			}
		}
		return crs;
	}

	public Translation getDesc() {
		return desc;
	}

	public Envelope getEnvelope() {
		try {
			return fs.getBounds();
		} catch (IOException e) {
			e.printStackTrace();
			return null;
		}
	}

	public FeatureSource<SimpleFeatureType, SimpleFeature> getGeoObject() {
		return fs;
	}

	public String getId() {
		return id;
	}

	public ImageIcon getImageIcon() {
		return null;
	}

	public URL getInfoURL() {
		return null;
	}

	public Translation getKeywords() {
		return null;
	}

	public Style getStyle() {
		return style;
	}

	public Translation getTitle() {
		return title;
	}

	public boolean isDisposed() {
		return false;
	}

	/**
	 * If true, this layer will not be shown in the legend. Default = false
	 */
	/**
	 * 
	 * Killed by SK: 6. April 09: Ein Layer soll nicht generell auf
	 * verstecken/nicht verstecken gestellt werden können. Das sind
	 * Eigenschaften der Karte/MapContext, ebenso wie die Reihenfolge der Layer.
	 * Im Atlas verwaltet deshalb nun die Klasse skrueger.atlas.Map welche Layer
	 * nicht in der Legende auftauchen sollen. Meines Wissens hat keiner bisher
	 * die Funktion genutzt.
	 * 
	 * // public boolean isHideInLegend() { // return false; // }
	 */

	public void setDesc(Translation dec) {
		this.desc = dec;
	}

	public void setImageIcon(ImageIcon icon) {
	}

	public void setKeywords(Translation keywords) {
	}

	public void setStyle(Style style) {
		this.style = style;

	}

	public void setTitle(Translation title) {
		this.title = title;

	}

	public void uncache() {
	}

	/**
	 * 
	 */
	public AttributeMetadataMap getAttributeMetaDataMap() {
		if (map == null) {

			map = new AttributeMetadataMap(new String[] { Translation
					.getActiveLang() });

			// Leaving out the first one, it will be the_geom
			for (int i = 1; i < fs.getSchema().getAttributeCount(); i++) {
				AttributeDescriptor attDesc = fs.getSchema().getDescriptor(i);

				AttributeMetadata attMetaData = new AttributeMetadata(attDesc
						.getName(), map.getLanguages());
				map.put(attDesc.getName(), attMetaData);
			}
		}
		return map;
	}

	/**
	 * @return The {@link File} where the SLD was loaded from or
	 *         <code>null</code> if there didn't exist a {@link File}.
	 * 
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public File getSldFile() {
		return sldFile;
	}
	
	public void setSldFile(File sldFile) {
		this.sldFile = sldFile;
	}

	/**
	 * Returns the features of the {@link FeatureSource}.
	 * 
	 * @see {@link StyledFeaturesInterface}
	 */
	@Override
	public FeatureCollection<SimpleFeatureType, SimpleFeature> getFeatureCollection() {
		FeatureCollection<SimpleFeatureType, SimpleFeature> features;
		try {
			features = getGeoObject().getFeatures();
		} catch (IOException e) {
			throw new RuntimeException(
					"Error getting the features of the  FeatureSource");
		}
		return features;
	}

	/**
	 * Same as {@link #getGeoObject()} method, but complies to the
	 * {@link StyledFeaturesInterface}
	 * 
	 * @see {@link StyledFeaturesInterface}
	 */
	@Override
	public FeatureSource<SimpleFeatureType, SimpleFeature> getFeatureSource() {
		return getGeoObject();
	}

	@Override
	public FeatureCollection<SimpleFeatureType, SimpleFeature> getFeatureCollectionFiltered() {
		// final FeatureCollection<SimpleFeatureType, SimpleFeature> fc =
		// getFeatureCollection();
		// if (filter == Filter.EXCLUDE)
		// return new EmptyFeatureCollection(fc.getSchema());
		// if (filter == Filter.INCLUDE)
		// return fc;
		// return fc.subCollection(filter);

		try {
			return getFeatureSource().getFeatures(filter);
		} catch (IOException e) {
			throw new RuntimeException(e);
		}
	}

	@Override
	public Filter getFilter() {
		return filter;
	}

	@Override
	public void setFilter(Filter filter) {
		this.filter = filter;
	}

	@Override
	public SimpleFeatureType getSchema() {
		return getGeoObject().getSchema();
	}

}
