/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package skrueger.geotools;

import java.io.IOException;
import java.net.URL;

import javax.swing.ImageIcon;

import org.geotools.data.FeatureSource;
import org.geotools.data.collection.CollectionDataStore;
import org.geotools.data.store.EmptyFeatureCollection;
import org.geotools.feature.FeatureCollection;
import org.geotools.feature.collection.SubFeatureCollection;
import org.geotools.styling.Style;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.simple.SimpleFeatureType;
import org.opengis.feature.type.AttributeDescriptor;
import org.opengis.filter.Filter;

import schmitzm.geotools.feature.FeatureUtil;
import skrueger.AttributeMetadata;
import skrueger.i8n.Translation;

/**
 * This class provides a simple implementation of {@link StyledLayerInterface}
 * for {@link FeatureCollection}. The uncache functionality is not supported,
 * because this class bases on an existing {@link FeatureCollection} object in
 * memory.
 * 
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *         (University of Bonn/Germany)
 * @version 1.0
 */
public class StyledFeatureCollection
		extends
		AbstractStyledLayer<FeatureCollection<SimpleFeatureType, SimpleFeature>>
		implements StyledFeatureCollectionInterface {

	/** Holds the meta data for displaying a legend. */
	protected AttributeMetadataMap attrMetaData = null;

	/**
	 * We be filled with a "virtual" {@link FeatureSource} on demand.
	 */
	private FeatureSource<SimpleFeatureType, SimpleFeature> featureSource = null;

	/**
	 * A Filter that will be applied internally applied to this
	 * FeatureCollection
	 **/
	private Filter filter = Filter.INCLUDE;

	/**
	 * Creates a styled {@link FeatureCollection} with language-specific
	 * informations.
	 * 
	 * @param fc
	 *            the {@link FeatureCollection}
	 * @param id
	 *            a unique ID for the object
	 * @param title
	 *            a (language-specific) short description
	 * @param desc
	 *            a (language-specific) long description
	 * @param keywords
	 *            (language-specific) keywords for the geo objects
	 * @param style
	 *            a display style (if {@code null}, a default style is created)
	 * @param attrMetaData
	 *            meta data for displaying a legend
	 * @param icon
	 *            an icon for the object (can be {@code null})
	 * @exception IllegalArgumentException
	 *                if {@code null} is given as ID or geo object
	 */
	public StyledFeatureCollection(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc, String id,
			Translation title, Translation desc, Translation keywords,
			Style style, AttributeMetadataMap attrMetaData, ImageIcon icon) {
		super(fc, fc.getBounds(), fc.getSchema().getGeometryDescriptor()
				.getCoordinateReferenceSystem(), id, title, desc, keywords,
				style, icon);
		setAttributeMetaData(attrMetaData);
	}

	/**
	 * Creates a styled {@link FeatureCollection} with language-specific
	 * informations.
	 * 
	 * @param fc
	 *            the {@link FeatureCollection}
	 * @param id
	 *            a unique ID for the object
	 * @param title
	 *            a (language-specific) short description
	 * @param desc
	 *            a (language-specific) long description
	 * @param keywords
	 *            (language-specific) keywords for the geo objects
	 * @param style
	 *            a display style with attribute meta data information
	 * @param icon
	 *            an icon for the object (can be {@code null})
	 * @exception IllegalArgumentException
	 *                if {@code null} is given as ID or geo object
	 */
	public StyledFeatureCollection(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc, String id,
			Translation title, Translation desc, Translation keywords,
			StyledLayerStyle<AttributeMetadataMap> style, ImageIcon icon) {
		super(fc, fc.getBounds(), fc.getSchema().getGeometryDescriptor()
				.getCoordinateReferenceSystem(), id, title, desc, keywords,
				style != null ? style.getGeoObjectStyle() : null, icon);
		setAttributeMetaData(style != null ? style.getMetaData() : null);
	}

	/**
	 * Creates a styled {@link FeatureCollection} with a language-specific
	 * title, no long description, no keywords, default attribute meta data and
	 * no icon.
	 * 
	 * @param fc
	 *            the {@link FeatureCollection}
	 * @param id
	 *            a unique ID for the object
	 * @param title
	 *            a short description
	 * @param style
	 *            a display style (if {@code null}, a default style is created)
	 * @exception IllegalArgumentException
	 *                if {@code null} is given as ID or geo object
	 * @see #createDefaultAttributeMetaDataMap(FeatureCollection)
	 */
	public StyledFeatureCollection(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc, String id,
			Translation title, Style style) {
		this(fc, id, title, null, null, style, null, null);
	}

	/**
	 * Creates a styled {@link FeatureCollection} with non-translated
	 * informations.
	 * 
	 * @param fc
	 *            the {@link FeatureCollection}
	 * @param id
	 *            a unique ID for the object
	 * @param title
	 *            a short description
	 * @param desc
	 *            a long description
	 * @param keywords
	 *            keywords for the geo objects
	 * @param style
	 *            a display style (if {@code null}, a default style is created)
	 * @param attrMetaData
	 *            meta data for displaying a legend
	 * @param icon
	 *            an icon for the object (can be {@code null})
	 * @exception IllegalArgumentException
	 *                if {@code null} is given as ID or geo object
	 */
	public StyledFeatureCollection(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc, String id,
			String title, String desc, String keywords, Style style,
			AttributeMetadataMap attrMetaData, ImageIcon icon) {
		this(fc, id, (Translation) null, null, null, style, attrMetaData, icon);
		setTitle(title);
		setDesc(desc);
		setKeywords(keywords);
	}

	/**
	 * Creates a styled {@link FeatureCollection} with non-translated
	 * informations.
	 * 
	 * @param fc
	 *            the {@link FeatureCollection}
	 * @param id
	 *            a unique ID for the object
	 * @param title
	 *            a short description
	 * @param desc
	 *            a long description
	 * @param keywords
	 *            keywords for the geo objects
	 * @param style
	 *            a display style with attribute meta data information
	 * @param icon
	 *            an icon for the object (can be {@code null})
	 * @exception IllegalArgumentException
	 *                if {@code null} is given as ID or geo object
	 */
	public StyledFeatureCollection(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc, String id,
			String title, String desc, String keywords,
			StyledLayerStyle<AttributeMetadataMap> style, ImageIcon icon) {
		this(fc, id, title, desc, keywords, style != null ? style
				.getGeoObjectStyle() : null, style != null ? style
				.getMetaData() : null, icon);
	}

	/**
	 * Creates a styled {@link FeatureCollection} with a non-translated title,
	 * no long description, no keywords, default attribute meta data and no
	 * icon.
	 * 
	 * @param fc
	 *            the {@link FeatureCollection}
	 * @param id
	 *            a unique ID for the object
	 * @param title
	 *            a short description
	 * @param style
	 *            a display style (if {@code null}, a default style is created)
	 * @exception IllegalArgumentException
	 *                if {@code null} is given as ID or geo object
	 * @see #createDefaultAttributeMetaDataMap(FeatureCollection)
	 */
	public StyledFeatureCollection(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc, String id,
			String title, Style style) {
		this(fc, id, title, null, null, style, null, null);
	}

	/**
	 * Creates a styled {@link FeatureCollection} with a non-translated title,
	 * no long description, no keywords, default attribute meta data and no
	 * icon.
	 * 
	 * @param fc
	 *            the {@link FeatureCollection}
	 * @param id
	 *            a unique ID for the object
	 * @param title
	 *            a short description
	 * @param style
	 *            a display style (if {@code null}, a default style is created)
	 * @exception IllegalArgumentException
	 *                if {@code null} is given as ID or geo object
	 * @see #createDefaultAttributeMetaDataMap(FeatureCollection)
	 */
	public StyledFeatureCollection(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc, String id,
			String title, StyledLayerStyle<AttributeMetadataMap> style) {
		this(fc, id, title, null, null, style != null ? style
				.getGeoObjectStyle() : null, style != null ? style
				.getMetaData() : null, null);
	}

	/**
	 * Creates a default style for the {@link FeatureCollection}.
	 * 
	 * @see FeatureUtil#createDefaultStyle(FeatureCollection)
	 */
	protected Style createDefaultStyle() {
		return FeatureUtil.createDefaultStyle(geoObject);
	}

	/**
	 * Returns the meta data needed for displaying a legend.
	 */
	public AttributeMetadataMap getAttributeMetaDataMap() {
		return attrMetaData;
	}

	/**
	 * Sets the meta data needed for displaying a legend. If {@code legendData}
	 * is {@code null} an empty map is set, so
	 * {@link #getAttributeMetaDataMap()} never returns {@code null}.
	 * 
	 * @param attrMetaData
	 *            map of attribute meta data
	 */
	public void setAttributeMetaData(AttributeMetadataMap attrMetaData) {
		this.attrMetaData = (attrMetaData != null) ? attrMetaData
				: createDefaultAttributeMetaDataMap(geoObject);
	}

	/**
	 * Creates non-translated default meta data for a {@link FeatureCollection}
	 * with all attributes visible and no unit set.
	 * 
	 * @param fc
	 *            a {@link FeatureCollection}
	 */
	public static AttributeMetadataMap createDefaultAttributeMetaDataMap(
			FeatureCollection<SimpleFeatureType, SimpleFeature> fc) {
		AttributeMetadataMap metaDataMap = new AttributeMetadataMap();
		SimpleFeatureType ftype = fc.getSchema();
		for (int i = 0; i < ftype.getAttributeCount(); i++) {
			AttributeDescriptor aDesc = ftype.getAttributeDescriptors().get(i);
			if (aDesc != ftype.getGeometryDescriptor())
				metaDataMap.put(aDesc.getName(), new AttributeMetadata(aDesc.getName(), 
						true, // visible
						new Translation(aDesc.getLocalName()), // Column name
						new Translation(aDesc.getLocalName()), // description
						"" // Unit
				));
		}
		return metaDataMap;
	}

	/**
	 * Simply sets the {@link #geoObject}, {@link #crs}, {@link #envelope} and
	 * {@link #attrMetaData} to {@code null}.
	 */
	public void dispose() {
		this.geoObject = null;
		this.envelope = null;
		this.crs = null;
		this.attrMetaData = null;
	}

	/**
	 * Tests whether the geo object is disposed.
	 */
	public boolean isDisposed() {
		return geoObject == null;
	}

	/**
	 * Does nothing, because the {@link AbstractStyledLayer} bases on existing
	 * objects (in memory) which can not be uncached and reloaded.
	 */
	public void uncache() {

		/** It will be recreated on the next getFetureSource() **/
		featureSource = null;

		LOGGER
				.warn("Uncache onyl uncached any virtual FeatureSource. Object remains in memory.");
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see skrueger.geotools.StyledLayerInterface#getInfoURL()
	 */
	public URL getInfoURL() {
		return null;
	}

	/**
	 * Same as {@link #getGeoObject()} method, but complies to the
	 * {@link StyledFeaturesInterface}. The associated {@link Filter} is NOT
	 * automatically applied.
	 * 
	 * @see {@link StyledFeaturesInterface}
	 * @see #getFeatureCollectionFiltered()
	 */
	@Override
	public FeatureCollection<SimpleFeatureType, SimpleFeature> getFeatureCollection() {
		return getGeoObject();
	}

	/**
	 * Same as {@link #getGeoObject()} method, but complies to the
	 * {@link StyledFeaturesInterface}. The associated {@link Filter} is
	 * automatically applied by creating a {@link SubFeatureCollection}.
	 * 
	 * @see {@link StyledFeaturesInterface}
	 * @see #getFeatureCollectionFiltered()
	 */
	@Override
	public FeatureCollection<SimpleFeatureType, SimpleFeature> getFeatureCollectionFiltered() {
		final FeatureCollection<SimpleFeatureType, SimpleFeature> fc = getFeatureCollection();
		if (filter == Filter.EXCLUDE)
			return new EmptyFeatureCollection(fc.getSchema());
		if (filter == Filter.INCLUDE)
			return fc;
		return fc.subCollection(filter);
	}

	/**
	 * Returns a virtual {@link FeatureSource} to access the
	 * {@link FeatureCollection}. Once created, it will be reused until
	 * {@link #uncache()} is called.<br/>
	 * 
	 * @see {@link StyledFeaturesInterface}
	 */
	@Override
	public FeatureSource<SimpleFeatureType, SimpleFeature> getFeatureSource() {
		if (featureSource == null) {
			CollectionDataStore store = new CollectionDataStore(getGeoObject());
			try {
				featureSource = store.getFeatureSource(store.getTypeNames()[0]);
			} catch (IOException e) {
				throw new RuntimeException(
						"Could not create a FeatureSource from the CollectionDataStore:",
						e);
			}
		}
		return featureSource;
	}

	@Override
	public Filter getFilter() {
		return filter;
	}

	@Override
	public void setFilter(Filter filter) {
		this.filter = filter;
	}

	@Override
	public SimpleFeatureType getSchema() {
		return getGeoObject().getSchema();
	}


}
