/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package skrueger.geotools;

import java.util.HashMap;
import java.util.List;
import java.util.Vector;

import org.apache.log4j.Logger;
import org.geotools.data.DefaultQuery;
import org.geotools.data.FeatureSource;
import org.geotools.data.Query;
import org.geotools.feature.FeatureCollection;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.simple.SimpleFeatureType;
import org.opengis.feature.type.AttributeDescriptor;
import org.opengis.filter.Filter;

import schmitzm.geotools.feature.FeatureUtil;
import schmitzm.geotools.gui.FeatureCollectionTableModel;
import skrueger.AttributeMetadata;

import com.vividsolutions.jts.geom.Envelope;

/**
 * This class extends the the {@link FeatureCollectionTableModel} with the
 * functionalities of the {@link AttributeMetadata}.
 * <ul>
 * <li>column names are translated according to
 * {@link AttributeMetadata#getTitle()}</li>
 * <li>columns are hidden according to {@link AttributeMetaData#isVisible()()}</li>
 * <li>Any filter defined in the {@link StyledFeaturesInterface} will be
 * applied.</li>
 * </ul>
 * 
 * @author Stefan A. Krüger
 */
public class StyledFeatureCollectionTableModel extends
		FeatureCollectionTableModel {
	final static private Logger LOGGER = Logger
			.getLogger(StyledFeatureCollectionTableModel.class);
	/** Contains the complete {@link AttributeMetadata}-Map of the styled layer. */
	protected AttributeMetadataMap amdMap = null;
	/** Holds the current filter on the table */
	protected Filter filter = null;
	/** Holds the Bounds for all features. Only set once during the constructor **/
	protected Envelope bounds;
	/**
	 * Tooltips für die Spaltennamen. Wird nur beim Aufruf von
	 * {@link #reorganize} befuellt.
	 */
	protected String[] colTooltips = null;

	/** A cache for the #sortedValuesVisibleOnly() **/
	protected List<AttributeMetadata> amdMapVisibleOnly = null;

	/**
	 * Creates a new table model for a styled layer.
	 * 
	 * @param styledFeatures
	 *            the styled layer
	 * @param filter
	 *            filter applied to the table
	 */
	public StyledFeatureCollectionTableModel(
			StyledFeaturesInterface<?> styledFeatures) {
		setStyledFeatures(styledFeatures);
	}

	/**
	 * This overwritten method filters the values for NODATA-values defined in
	 * the {@link AttributeMetadata}
	 */
	@Override
	public Object getValueAt(int row, int col) {
		Object rawValue = super.getValueAt(row, col);
		return amdMap.sortedValuesVisibleOnly().get(col).fiterNodata(rawValue);
	}
	
	/**
	 * Sets a new data source for the table.
	 * 
	 * @param fs
	 *            the feature source
	 * @param amdm
	 *            {@link AttributeMetadata}-Map to define the visible attributes
	 *            and translation
	 */
	protected void setFeatureSource(
			FeatureSource<SimpleFeatureType, SimpleFeature> fs,
			AttributeMetadataMap amdm, Filter filter) throws Exception {

		if (filter == null)
			filter = Filter.INCLUDE;

		// this.featureSource = fs;
		this.filter = filter;
		this.amdMap = amdm;
		this.amdMapVisibleOnly = amdMap.sortedValuesVisibleOnly();

		FeatureCollection<SimpleFeatureType, SimpleFeature> fc = null;
		if (fs != null) {

			bounds = fs.getBounds();

			final SimpleFeatureType schema = fs.getSchema();
			Query query = new DefaultQuery(schema.getTypeName(), filter);
			if (amdm != null) {
				Vector<String> visibleAttrNames = new Vector<String>();

				// Add the column with the geometry (usually "the_geom") always
				visibleAttrNames.add(schema.getGeometryDescriptor()
						.getLocalName());

				// Add other visible attributes as ordered by weights
				for (AttributeMetadata a : amdMapVisibleOnly) {
					visibleAttrNames.add(a.getLocalName());
				}

				// Tested with 2.6.x trunk from 2009-11-26 and it now works. So
				// we only request the properties we need!
				// /**
				// * I got NPEs when properties contained only [the_geom]
				// ?!??!!??
				// */
				// if (properties.length > 1) {
				query = new DefaultQuery(schema.getTypeName(), filter,
						visibleAttrNames.toArray(new String[] {}));
				// } else {
				// query = new DefaultQuery(schema.getTypeName(), filter);
				// }
			}
			fc = fs.getFeatures(query);
		}
		setFeatureCollection(fc);
	}

	/**
	 * Converts the {@code StyledFeatureCollection} to a {@code FeatureSource}
	 * and sets this as the new data source for the table.
	 * 
	 * @param fs
	 *            the feature source
	 * @param amd
	 *            {@link AttributeMetadata}-Map to define the visible attributes
	 *            and translation
	 */
	public void setStyledFeatures(StyledFeaturesInterface<?> styledFeatures) {
		try {
			if (styledFeatures == null)
				setFeatureSource(null, null, null);
			else {
				setFeatureSource(styledFeatures.getFeatureSource(),
						styledFeatures.getAttributeMetaDataMap(),
						styledFeatures.getFilter());
			}
		} catch (Exception err) {
			throw new RuntimeException(err);
		}
	}

	/**
	 * After calling {@code super.reorganize(.)} this method replaced the column
	 * descriptions with the titles of the {@code AttributeMetaData}.
	 * 
	 * @param fireTableStructureChanged
	 *            indicates whether a table event is initiated after reorganize
	 */
	@Override
	protected void reorganize(boolean fireTableStructureChanged) {

		featureArray = FeatureUtil.featuresToArray(featureTable);
		if (featureArray == null || featureArray.length == 0) {
			colNames = new String[0];
			colTooltips = new String[0]; // Only set and used in
			// StyledFeatureCollectionTableModel
			colClass = new Class[0];
		} else {
			// Struktur der Tabelle vom AttributeMetaDtaaMap übernehmen
			SimpleFeatureType schema = featureArray[0].getFeatureType();
			// Pruefen, welche Attribute angezeigt werden
			attrTypes.clear();
			for (AttributeMetadata amd : amdMapVisibleOnly) {
				AttributeDescriptor type = schema.getDescriptor(amd.getName());
				if (attrFilter == null || attrFilter.accept(type))
					attrTypes.add(type);
			}
			// Namen und Attribut-Indizes der angezeigten Spalten ermitteln
			colNames = new String[attrTypes.size()];
			colTooltips = new String[attrTypes.size()]; // Only set and used in
			// StyledFeatureCollectionTableModel
			colClass = new Class[attrTypes.size()];
			attrIdxForCol = new int[attrTypes.size()];
			for (int i = 0; i < colNames.length; i++) {
				AttributeDescriptor descriptor = schema
						.getDescriptor(amdMapVisibleOnly.get(i).getName());

				// Not so nice in 26: find the index of an attribute...
				int idx = schema.getAttributeDescriptors().indexOf(descriptor);
				attrIdxForCol[i] = idx;

				String attName = schema.getAttributeDescriptors().get(idx)
						.getLocalName();
				colNames[i] = amdMap.get(attName).getTitle().toString();
				AttributeMetadata amd = amdMap.get(attName);
				colTooltips[i] = "<html>" + amd.getDesc().toString() + "<br>"
						+ amd.getName() + "</html>";
				colClass[i] = schema.getAttributeDescriptors().get(idx)
						.getType().getBinding();
			}
		}

		// store feature indexes in HashMap to optimize findFeature(.)
		featureIdx = new HashMap<String, Integer>();
		for (int i = 0; i < featureArray.length; i++)
			if (featureArray[i] != null)
				featureIdx.put(featureArray[i].getID(), i);
		//
		// // translate the column names
		// if (amdMap != null) {
		// for (int i = 0; i < colNames.length; i++) {
		// colTooltips[i] = amdMap.get(colNames[i]).getDesc().toString()
		// + "<br>" + colNames[i];
		// colNames[i] = amdMap.get(colNames[i]).getTitle().toString();
		//
		// }
		// }
		if (fireTableStructureChanged)
			fireTableStructureChanged();

	}

	/**
	 * @return Cached bounds for the whole dataset (without applying the filter)
	 *         or <code>null</code>
	 */
	public Envelope getBounds() {
		return bounds;
	}
}
