/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package skrueger.geotools;

import java.net.URL;

import javax.swing.ImageIcon;

import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.styling.Style;

import schmitzm.geotools.JTSUtil;
import schmitzm.geotools.grid.GridUtil;
import skrueger.RasterLegendData;
import skrueger.i8n.Translation;

/**
 * This class provides a simple implementation of {@link StyledLayerInterface}
 * for {@link GridCoverage2D}. The uncache functionality is not supported,
 * because this class bases on an existing {@link GridCoverage2D} object in
 * memory.
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 * @version 1.0
 */
public class StyledGridCoverage extends AbstractStyledLayer<GridCoverage2D> implements StyledGridCoverageInterface {

  /** Holds the meta data for displaying a legend. */
  protected RasterLegendData legendData = null;

  /**
   * Creates a styled grid with language-specific informations.
   * @param gc the grid
   * @param id a unique ID for the object
   * @param title a (language-specific) short description
   * @param desc a (language-specific) long description
   * @param keywords (language-specific) keywords for the geo objects
   * @param style a display style (if {@code null}, a default style is created)
   * @param legendData meta data for displaying a legend
   * @param icon an icon for the object (can be {@code null})
   * @exception IllegalArgumentException if {@code null} is given as ID or geo object
   */
  public StyledGridCoverage(GridCoverage2D gc, String id, Translation title, Translation desc, Translation keywords, Style style, RasterLegendData legendData, ImageIcon icon) {
    super(gc, JTSUtil.createEnvelope(gc.getEnvelope()), gc.getCoordinateReferenceSystem(), id, title, desc, keywords, style, icon);
    setLegendMetaData(legendData);
  }

  /**
   * Creates a styled grid with language-specific informations.
   * @param gc the grid
   * @param id a unique ID for the object
   * @param title a (language-specific) short description
   * @param desc a (language-specific) long description
   * @param keywords (language-specific) keywords for the geo objects
   * @param style a display style with legend information
   * @param icon an icon for the object (can be {@code null})
   * @exception IllegalArgumentException if {@code null} is given as ID or geo object
   */
  public StyledGridCoverage(GridCoverage2D gc, String id, Translation title, Translation desc, Translation keywords, StyledLayerStyle<RasterLegendData> style, ImageIcon icon) {
    super(gc, JTSUtil.createEnvelope(gc.getEnvelope()), gc.getCoordinateReferenceSystem(), id, title, desc, keywords, style != null ? style.getGeoObjectStyle() : null, icon);
    setLegendMetaData( style != null ? style.getMetaData() : null );
  }

  /**
   * Creates a styled grid with a language-specific title, no long description, no
   * keywords and no icon.
   * @param gc the grid
   * @param id a unique ID for the object
   * @param title a short description
   * @param style a display style (if {@code null}, a default style is created)
   * @param legendData meta data for displaying a legend
   * @exception IllegalArgumentException if {@code null} is given as ID or geo object
   */
  public StyledGridCoverage(GridCoverage2D gc, String id, Translation title, Style style, RasterLegendData legendData) {
    this(gc, id, title, null, null, style, legendData, null);
  }

  /**
   * Creates a styled grid with non-translated informations.
   * @param gc the grid
   * @param id a unique ID for the object
   * @param title a short description
   * @param desc a long description
   * @param keywords keywords for the geo objects
   * @param style a display style (if {@code null}, a default style is created)
   * @param legendData meta data for displaying a legend
   * @param icon an icon for the object (can be {@code null})
   * @exception IllegalArgumentException if {@code null} is given as ID or geo object
   */
  public StyledGridCoverage(GridCoverage2D gc, String id, String title, String desc, String keywords, Style style, RasterLegendData legendData, ImageIcon icon) {
    this(gc, id, (Translation)null, null, null, style, legendData, icon);
    setTitle(title);
    setDesc(desc);
    setKeywords(keywords);
  }

  /**
   * Creates a styled grid with non-translated informations.
   * @param gc the grid
   * @param id a unique ID for the object
   * @param title a short description
   * @param desc a long description
   * @param keywords keywords for the geo objects
   * @param style a display style with legend information
   * @param icon an icon for the object (can be {@code null})
   * @exception IllegalArgumentException if {@code null} is given as ID or geo object
   */
  public StyledGridCoverage(GridCoverage2D gc, String id, String title, String desc, String keywords, StyledLayerStyle<RasterLegendData> style, ImageIcon icon) {
    this(gc,
         id,
         title,
         desc,
         keywords,
         style != null ? style.getGeoObjectStyle() : null,
         style != null ? style.getMetaData() : null,
         icon
    );
  }

  /**
   * Creates a styled grid with a non-translated title, no long description, no
   * keywords and no icon.
   * @param gc the grid
   * @param id a unique ID for the object
   * @param title a short description
   * @param style a display style (if {@code null}, a default style is created)
   * @exception IllegalArgumentException if {@code null} is given as ID or geo object
   */
  public StyledGridCoverage(GridCoverage2D gc, String id, String title, Style style, RasterLegendData legendData) {
    this(gc, id, title, null, null, style, legendData, null);
  }

  /**
   * Creates a styled grid with a non-translated title, no long description, no
   * keywords and no icon.
   * @param gc the grid
   * @param id a unique ID for the object
   * @param title a short description
   * @param style a display style with legend information
   * @exception IllegalArgumentException if {@code null} is given as ID or geo object
   */
  public StyledGridCoverage(GridCoverage2D gc, String id, String title, StyledLayerStyle<RasterLegendData> style) {
    this(gc,
         id,
         title,
         null,
         null,
         style != null ? style.getGeoObjectStyle() : null,
         style != null ? style.getMetaData() : null,
         null
    );
  }

  /**
   * Creates a default style for a {@link GridCoverage2D}.
   * @see GridUtil#createDefaultStyle()
   */
  protected Style createDefaultStyle() {
    return GridUtil.createDefaultStyle();
  }

  /**
   * Returns the meta data needed for displaying a legend.
   */
  public RasterLegendData getLegendMetaData() {
    return legendData;
  }

  /**
   * Sets the meta data needed for displaying a legend.
   * If {@code legendData} is {@code null} an empty {@link RasterLegendData}
   * (without gaps) is set, so {@link #getLegendMetaData()} never returns {@code null}.
   * @param legendData legend meta data
   */
  public void setLegendMetaData(RasterLegendData legendData) {
    this.legendData = (legendData != null) ? legendData : new RasterLegendData(false);
  }

  /**
   * Simply sets the {@link #geoObject}, {@link #crs}, {@link #envelope} and
   * {@link #legendData} to {@code null}.
   */
  public void dispose() {
    this.geoObject  = null;
    this.envelope   = null;
    this.crs        = null;
    this.legendData = null;
  }

  /**
   * Tests whether the geo object is disposed.
   * @return boolean
   */
  public boolean isDisposed() {
    return geoObject == null;
  }

  /**
   * Does nothing, because the {@link AbstractStyledLayer} bases on existing
   * objects (in memory) which can not be uncached and reloaded.
   */
  public void uncache() {
    LOGGER.warn("Uncache functionality is not supported. Object remains in memory.");
  }

  /*
   * (non-Javadoc)
   * @see skrueger.geotools.StyledLayerInterface#getInfoURL()
   */
  public URL getInfoURL() {
	return null;
  }

  /**
	 * If true, this layer will not be shown in the legend. Default = false
	 */
	/**
	 * 
	 * Killed by SK: 6. April 09: Ein Layer soll nicht generell auf
	 * verstecken/nicht verstecken gestellt werden können. Das sind
	 * Eigenschaften der Karte/MapContext, ebenso wie die Reihenfolge der Layer.
	 * Im Atlas verwaltet deshalb nun die Klasse skrueger.atlas.Map welche Layer
	 * nicht in der Legende auftauchen sollen. Meines Wissens hat keiner bisher
	 * die Funktion genutzt.
	 * 
  public boolean isHideInLegend() {
	return false;
   }
	 */
  
}
