/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package skrueger.geotools;

import java.net.URL;

import javax.swing.ImageIcon;

import org.geotools.feature.FeatureCollection;
import org.geotools.styling.Style;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import skrueger.AttributeMetadata;
import skrueger.RasterLegendData;
import skrueger.i8n.Translation;

import com.vividsolutions.jts.geom.Envelope;

/**
 * This class is the top interface for styled objects to be managed in
 * {@link MapContextManagerInterface}. The (rough) classe structure is the
 * following:
 * <ul>
 * <li><b>{@link StyledLayerInterface StyledLayerInterface<E>}</b>
 * <ul>
 * <li>{@link #getId()} -> String</li>
 * <li>{@link #getKeywords() get/setKeywords()} -> {@link Translation}</li>
 * <li>{@link #getTitle() set/getTitle()} -> {@link Translation} (short
 * description for layer list)</li>
 * <li>{@link #getDesc() set/getDesc()} -> {@link Translation} (long description
 * for details)</li>
 * <li>{@link #getCrs()} -> {@link CoordinateReferenceSystem}</li>
 * <li>{@link #getCRSString()} -> String (readable description of CRS)</li>
 * <li>{@link #getEnvelope()} -> {@link Envelope} (JTS-Envelope)</li>
 * <li>{@link #getGeoObject()} -> E (GridCoverage/FeatureCollection/...)</li>
 * <li>{@link #getStyle() set/getStyle()} -> {@link Style}</li>
 * <li>{@link #uncache()}</li>
 * <li>{@link #dispose()}</li>
 * </ul>
 * </li>
 * <li><b>{@link StyledFeatureCollectionInterface} extends
 * {@link StyledLayerInterface StyledLayerInterface<FeatureCollection>}</b>
 * <ul>
 * <li>{@link StyledFeatureCollectionInterface#getAttributeMetaDataMap()} ->
 * Map<Integer,AttributeMetaData></li>
 * </ul>
 * </li>
 * <li><b>{@link StyledRasterInterface} extends {@link StyledLayerInterface
 * StyledLayerInterface<GridCoverage2D>}</b>
 * <ul>
 * <li>{@link StyledRasterInterface#getLegendMetaData()} ->
 * {@link RasterLegendData}</li>
 * </ul>
 * </li>
 * </ul>
 * <br>
 * <b>Restrictions:</b>
 * <ul>
 * <li>layer list only depends on {@link StyledLayerInterface}</li>
 * <li>methods returning {@link Translation} must not return {@code null}</li>
 * <li>methods returning {@link AttributeMetadata}-Map must not return {@code
 * null}</li>
 * <li>static helper method to get a new {@link AttributeMetadata}-map withe the
 * visible attributes only</li>
 * <li>static helper method to create a "default" {@link AttributeMetadata}-map
 * for a {@link FeatureCollection} with all attributes visible and without real
 * translations, but the attribute name as description.</li>
 * </ul>
 */
public interface StyledLayerInterface<E> {
	public String getId();

	public Translation getTitle();

	public void setTitle(Translation title);

	public Translation getDesc();

	public void setDesc(Translation dec);

	public Translation getKeywords();

	public void setKeywords(Translation keywords);

	public CoordinateReferenceSystem getCrs();

	public String getCRSString();

	public Envelope getEnvelope();

	/**
	 * @return return an ImageIcon - <code>null</code> is valid and no icon or a
	 *         default icon will then be shown
	 */
	public ImageIcon getImageIcon();

	public void setImageIcon(ImageIcon icon);

	/**
	 * Returns the underlying GeoTools Object
	 * 
	 * @throws RuntimeException
	 */
	public E getGeoObject();

	public Style getStyle();

	public void setStyle(Style style);

	/**
	 * Returns the {@link URL} to a (HTML) file that provides more information
	 * about this layer. If no HTML if associated with this
	 * {@link StyledLayerInterface}, then <code>null</code> will be returned.
	 * 
	 * @return null or an {@link URL}
	 */
	public URL getInfoURL();
	
	/**
	 * Should be called when this Object is not needed anymore.
	 */
	public void dispose();

	/** Is the object already disposed? * */
	public boolean isDisposed();

	/**
	 * Clears any caches. For example the GeoObject could be released, and
	 * reread on next call of getGeoObject()
	 */
	public void uncache();


}
