/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package skrueger.geotools;

import java.awt.Point;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;

import org.opengis.geometry.DirectPosition;

import schmitzm.swing.event.MouseInputType;

/**
 * Defines an action (e.g. Zoom in, zoom out, drag) when a click, drag or window
 * selection is performed on a {@link XMapPane}.
 * 
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *         (University of Bonn/Germany)
 */
public interface XMapPaneAction {

	/**
	 * This action can be assigned to any {@link MouseInputType} to perform
	 * zoom-in
	 **/
	public static XMapPaneAction_Zoom ZOOM_IN = new XMapPaneAction_Zoom.In();

	/**
	 * This action can be assigned to any {@link MouseInputType} to perform
	 * zoom-out
	 **/
	public static XMapPaneAction_Zoom ZOOM_OUT = new XMapPaneAction_Zoom.Out();
	
	/**
	 * This action can be assigned to any {@link MouseInputType} to perform
	 * panning on a Map
	 **/
	public static XMapPaneAction_Pan PAN = new XMapPaneAction_Pan();
	
	/**
	 * This action can be assigned to fire events for any number of feature from all layers.
	 **/
	public static XMapPaneAction_Select.All SELECT_ALL = new XMapPaneAction_Select.All();	

	/**
	 * This action can be assigned to fire selection events for one feature from the top layer.
	 **/
	public static XMapPaneAction_Select.OneFromTop SELECT_ONE_FROM_TOP = new XMapPaneAction_Select.OneFromTop();	

	/**
	 * This action can be assigned to fire events for any number of feature from the top layer. 
	 **/
	public static XMapPaneAction_Select.Top SELECT_TOP = new XMapPaneAction_Select.Top();	
	
	/**
	 * Defines the action in case of a single click on the map. Called by ###
	 * XMapPaneMouseAdapter ### on {@link MouseInputType#LClick} and
	 * {@link MouseInputType#RClick}.
	 * 
	 * @param mapPane
	 *            map pane the action should be performed on
	 * @param ev
	 *            mouse event of the action
	 * @param coord
	 *            geo coordinate the click is performed on
	 */
	public void performClick(XMapPane mapPane, MouseEvent ev,
			DirectPosition coord);

	/**
	 * Defines the action in case of a mouse drag on the map. This method is
	 * called on every mouse motion.
	 * 
	 * @param mapPane
	 *            map pane the action should be performed on
	 * @param ev
	 *            mouse event of the action
	 * @param dragStartPos
	 *            window position the drag was started (the current position can
	 *            be determined from the mouse event)
	 * @param startCoord
	 *            geo coordinate the drag was started
	 * @param endCoord
	 *            geo coordinate the drag is currently moved over
	 */
	public void performDragging(XMapPane mapPane, MouseEvent ev,
			Point dragStartPos, Point dragLastPos, DirectPosition startCoord,
			DirectPosition endCoord);

	/**
	 * Defines the action in case of a window selection on the map (the moment a
	 * drag ends).
	 * 
	 * @param mapPane
	 *            map pane the action should be performed on
	 * @param ev
	 *            mouse event of the action
	 * @param dragStartPos
	 *            window position the window starts (the end position can be
	 *            determined from the mouse event)
	 * @param startCoord
	 *            geo coordinate the window starts
	 * @param endCoord
	 *            geo coordinate the window ends
	 */
	public void performDragged(XMapPane mapPane, MouseEvent ev,
			Point dragStartPos, Point dragLastPos, DirectPosition startCoord,
			DirectPosition endCoord);

	/**
	 * Defines the action in case of a mouse wheel action on the map (the moment
	 * a drag ends).
	 * 
	 * @param mapPane
	 *            map pane the action should be performed on
	 * @param ev
	 *            mouse event of the action
	 * @param coord
	 *            geo coordinate the wheel is turned on
	 */
	public void performWheel(XMapPane mapPane, MouseWheelEvent ev,
			DirectPosition coord);

	/**
	 * Defines what happens if this action has been triggered by a keyboard-key.
	 * 
	 * @param mapPane
	 *            map pane the action should be performed on
	 *            
	 * @param param
	 *            An optinal paramter that can be defined.            
	 */
	public void performKeyboard(XMapPane mapPane, Object param);

}
