/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package skrueger.geotools;

import java.awt.Cursor;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;

import org.opengis.geometry.DirectPosition;

import schmitzm.swing.SwingUtil;

import com.vividsolutions.jts.geom.Envelope;

public class XMapPaneAction_Pan implements XMapPaneAction {

	public static enum Direction {
		UP, DOWN, RIGHT, LEFT
	}

	/**
	 * If this action is triggered by keyboard, defines the amount of panning in
	 * screen pixels
	 **/
	private static final int KEYBOARD_PAN_LENGTH = 15;

	/**
	 * This variable can be used to backup the active cursor of the mapPane, if
	 * the actions changes the cursor during dragging
	 */
	public Cursor backupCursor = null;

	/**
	 * Performs a pan action. During panning, the displacement is stored in
	 * {@link #imageOrigin} object. Calling {@link #performPan()} will reset the
	 * offset and call {@link #setMapArea(Envelope)}.
	 */
	@Override
	public void performDragged(XMapPane mapPane, MouseEvent ev,
			Point dragStartPos, Point dragLastPos, DirectPosition startCoord,
			DirectPosition endCoord) {

		Point translation = mapPane.getImageOrigin();

		doPan(mapPane, translation);
	}

	private void doPan(XMapPane mapPane, Point translation) {

		final Rectangle winBounds = mapPane.getVisibleRect();
		winBounds.translate(-translation.x, -translation.y);
		final Envelope newMapArea = mapPane.tranformWindowToGeo(winBounds.x,
				winBounds.y, winBounds.x + winBounds.width, winBounds.y
						+ winBounds.height);

		translation.x = 0;
		translation.y = 0;

		if (!mapPane.setMapArea(newMapArea)) {
			/**
			 * If setMapArea returns true, the finalImage is updated anyways.
			 * This if-case exists to ensure that we repaint a correct image
			 * even if the new panning area has been denied.
			 */
			mapPane.updateFinalImage();
			mapPane.repaint();
		}

		if (mapPane.getCursor() == SwingUtil.PANNING_CURSOR) {
			mapPane.setCursor(backupCursor);
		}
	}

	/**
	 * TODO: Stop dragging if the expected map area would not be valid.
	 */
	@Override
	public void performDragging(XMapPane mapPane, MouseEvent ev,
			Point dragStartPos, Point dragLastPos, DirectPosition startCoord,
			DirectPosition endCoord) {

		if (mapPane.getCursor() != SwingUtil.PANNING_CURSOR) {
			backupCursor = mapPane.getCursor();
			mapPane.setCursor(SwingUtil.PANNING_CURSOR);
		}

		// Variables for the translation in screen pixels
		int dX;
		int dY;

		if (dragLastPos == null) {
			dX = ev.getPoint().x - dragStartPos.x;
			dY = ev.getPoint().y - dragStartPos.y;
		} else {
			dX = ev.getPoint().x - dragLastPos.x;
			dY = ev.getPoint().y - dragLastPos.y;
		}

		if (dX != 0 && dY != 0)
			mapPane.pan(dX, dY);

	}

	@Override
	public void performWheel(XMapPane mapPane, MouseWheelEvent ev,
			DirectPosition coord) {
	}

	@Override
	public void performClick(XMapPane mapPane, MouseEvent ev,
			DirectPosition coord) {
	}

	@Override
	public void performKeyboard(XMapPane mapPane, Object param) {
		if (param == null)
			return;

		if (!(param instanceof Direction)) {
			throw new IllegalArgumentException("param = " + param
					+ " is of expected type XMapPaneAction_Pan.Direction");
		}

		// Variables for the translation in screen pixels
		int dX = 0;
		int dY = 0;

		switch ((Direction) param) {
		case DOWN:
			dY = KEYBOARD_PAN_LENGTH;
			break;
		case UP:
			dY = -KEYBOARD_PAN_LENGTH;
			break;
		case LEFT:
			dX = -KEYBOARD_PAN_LENGTH;
			break;
		case RIGHT:
			dX = KEYBOARD_PAN_LENGTH;
			break;
		}

		doPan(mapPane, new Point(dX, dY));
	}

}
