package skrueger.geotools;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.util.HashMap;
import java.util.Map;

import javax.swing.JComponent;
import javax.swing.KeyStroke;

import org.geotools.geometry.DirectPosition2D;

import schmitzm.swing.event.MouseInputType;
import schmitzm.swing.event.SelectiveMouseAdapter;

public class XMapPaneMouseListener extends SelectiveMouseAdapter {

	/**
	 * Holds the configuration of which {@link XMapPaneAction} is associated
	 * with which {@link MouseInputType}
	 **/
	Map<MouseInputType, XMapPaneAction> actions = new HashMap<MouseInputType, XMapPaneAction>(
			MouseInputType.values().length);

	/**
	 * A reference to the {@link XMapPane} that the {@link XMapPaneAction}s will
	 * be performed on.
	 */
	private final XMapPane xMapPane;

	public XMapPaneMouseListener(XMapPane xMapPane) {
		this.xMapPane = xMapPane;
	}

	/**
	 * Configures the instance with definitions from a {@link XMapPaneTool}
	 */
	public void configure(final XMapPaneTool tool) {
		actions = tool.mouseActions;

		// TODO Remove any KEYSrokes registered previously

		// Adding configured keyboard actions
		for (final KeyStroke stroke : tool.keyAction.keySet()) {
			xMapPane.registerKeyboardAction(new ActionListener() {

				public void actionPerformed(final ActionEvent e) {
					XMapPaneAction action = tool.keyAction.get(stroke);
					action.performKeyboard(xMapPane, tool.keyActionParams
							.get(stroke));
				}

			}, stroke, JComponent.WHEN_IN_FOCUSED_WINDOW);
		}
	}

	@Override
	public void performMouseClicked(MouseEvent mEv) {

		MouseInputType key = null;

		switch (mEv.getButton()) {
		case MouseEvent.BUTTON1:
			key = MouseInputType.LClick;
			break;
		case MouseEvent.BUTTON3:
			key = MouseInputType.RClick;
			break;
		default:
			return;
		}

		XMapPaneAction action = actions.get(key);

		if (action == null) {
			// No action has been defined for this - no problem
			return;
		}

		// Calculate the coordinates in the map's CRS
		DirectPosition2D coord = XMapPane.getMapCoordinatesFromEvent(mEv);

		/** Call the XMapPaneAction */
		if (coord != null)
			action.performClick(xMapPane, mEv, coord);
	}

	@Override
	public void performMouseDragged(MouseEvent mEv) {

		MouseInputType key = null;

		switch (dragButton) {
		case MouseEvent.BUTTON1:
			key = MouseInputType.LDrag;
			break;
		case MouseEvent.BUTTON3:
			key = MouseInputType.RDrag;
			break;
		default:
			return;
		}

		XMapPaneAction action = actions.get(key);

		if (action == null) {
			// No action has been defined for this - no problem
			return;
		}

		// Calculate the coordinates in the map's CRS
		DirectPosition2D endCoord = XMapPane.getMapCoordinatesFromEvent(mEv);

		DirectPosition2D startCoord = getDragStartCoord();

		/** Call the XMapPaneAction */
		if (startCoord != null && endCoord != null)
			action.performDragging(xMapPane, mEv, dragStartPos, dragLastPos,
					startCoord, endCoord);
	}

	@Override
	public void performMouseWheelMoved(MouseWheelEvent mEv) {

		XMapPaneAction action = actions.get(MouseInputType.Wheel);

		if (action == null)
			return;

		// Calculate the coordinates in the map's CRS
		DirectPosition2D coord = XMapPane.getMapCoordinatesFromEvent(mEv);

		/** Call the XMapPaneAction */
		if (coord != null)
			action.performWheel(xMapPane, mEv, coord);
	}

	public void performMouseReleased(MouseEvent mEv) {
	  MouseInputType key;
		switch (mEv.getButton()) {
		case MouseEvent.BUTTON1:
			key = MouseInputType.LDrag;
			break;
		case MouseEvent.BUTTON3:
			key = MouseInputType.RDrag;
			break;
		default:
			return;
		}

		XMapPaneAction action = actions.get(key);

		if (action == null)
			return;

		// Transform window coordinates to MapContext coordinates

		// Calculate the coordinates in the map's CRS
		DirectPosition2D endCoord = XMapPane.getMapCoordinatesFromEvent(mEv);
		if (endCoord == null)
			return;

		action.performDragged(xMapPane, mEv, dragStartPos, dragLastPos,
				getDragStartCoord(), endCoord);
	}

	/**
	 * Returns the drag position in map CRS
	 */
	public DirectPosition2D getDragStartCoord() {

		final AffineTransform at = xMapPane.getScreenToWorld();
		if (at == null)
			return null;

		Point2D transformed = at.transform(dragStartPos, null);
		DirectPosition2D startCoord = new DirectPosition2D(xMapPane
				.getMapContext().getCoordinateReferenceSystem(), transformed
				.getX(), transformed.getY());
		return startCoord;
	}

}
