/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package skrueger.geotools.io;

import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.net.URL;

import javax.imageio.IIOException;
import javax.imageio.ImageIO;

import org.apache.log4j.Logger;
import org.geotools.coverage.grid.GridCoverage2D;
import org.geotools.coverage.grid.GridCoverageFactory;
import org.geotools.factory.Hints;
import org.geotools.gce.geotiff.GeoTiffReader;
import org.geotools.geometry.Envelope2D;
import org.opengis.parameter.GeneralParameterValue;
import org.opengis.referencing.crs.CoordinateReferenceSystem;

import schmitzm.geotools.io.GeoImportUtil;
import schmitzm.io.IOUtil;

/**
 * Erweiterungen von Martin's {@link GeoImportUtil} classe fuer die konsequente
 * Benutzung mit {@link URL}s. TODO Diese Klasse sollte vielleicht mit der
 * {@link GeoImportUtil} zusammengefuegt werden.
 * 
 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons Kr&uuml;ger</a>
 * 
 */
public class GeoImportUtilURL extends GeoImportUtil {
	final static private Logger LOGGER = Logger
			.getLogger(GeoImportUtilURL.class);

	/**
	 * Read a {@link GridCoverage2D} from an image file. .prj and .wld files are
	 * usually expected
	 */
	public static GridCoverage2D readGridFromImage(URL url) throws IOException {
		return readGridFromImage(url, null);
	}

	/**
	 * Read a {@link GridCoverage2D} from an image file. .wld file is usually
	 * expected also. The CRS can be given as the crs parameter. null is valid.
	 * 
	 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
	 *         (University of Bonn/Germany)
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 */
	public static GridCoverage2D readGridFromImage(URL url,
			CoordinateReferenceSystem crs) throws IOException {
		GridCoverage2D gc = null;

		BufferedImage im = ImageIO.read(url);
		if (im == null)
			throw new IIOException("No image reader found for this image type!");

		// World-File einlesen
		double[] tfwInfo = null;

		for (WORLD_POSTFIXES pf : WORLD_POSTFIXES.values()) {
			if (tfwInfo == null) {
				try {
					tfwInfo = readWorldFile(IOUtil.changeUrlExt(url,
							pf.toString()).openStream());
				} catch (Exception e) {
				}
			}
		}

		if (tfwInfo == null)
			throw new IllegalArgumentException(
					"No georeferencing information found.\n"
							+ "Attach a .wld file to " + url + "please.");

		float w = (float) (im.getWidth() * tfwInfo[0]); // reale Breite =
		// RasterSpalten *
		// hor. Aufloesung
		float h = (float) (im.getHeight() * (-tfwInfo[3])); // reale Hoehe =
		// RasterZeilen
		// * vert.
		// Aufloesung
		float x = (float) tfwInfo[4]; // Suedwestliche Ecke!
		float y = (float) tfwInfo[5] - h; // Suedwestliche Ecke (im
		// tfw-File steht die
		// Nordwestliche!)
		// ggf. Projektion einlesen
		if (crs == null) {
			crs = determineProjection(IOUtil.changeUrlExt(url, "prj"));
		}
		Envelope2D envelope = new Envelope2D(crs, new Rectangle2D.Float(x, y,
				w, h));
		// WICHTIG: Name des Rasters sollte Leer-String sein, da ansonsten
		// das
		// Coloring des Rasters nicht klappt.
		// --> Name der Categories und ColorMapEntries-Labels muss (warum auch
		// immer) mit dem Namen
		// des Rasters uebereinstimmen!!!
		gc = new GridCoverageFactory().create("", im, envelope);

		return gc;
	}

	/**
	 * @param input
	 *            URL, File oder InputStream to a GeoTIFF-File
	 * @param crs
	 *            erzwungenes CoordinateReferenceSystem fuer das Raster (kann
	 *            {@code null} sein)
	 * 
	 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
	 *         (University of Bonn/Germany)
	 * @author <a href="mailto:skpublic@wikisquare.de">Stefan Alfons
	 *         Kr&uuml;ger</a>
	 * 
	 * @param readParams
	 *            May be {@link NullPointerException}. Otherwise an array of
	 *            {@link GeneralParameterValue}s
	 */
	public static GridCoverage2D readGridFromGeoTiff(Object input,
			CoordinateReferenceSystem crs, GeneralParameterValue[] readParams)
			throws IOException {
		GridCoverage2D gc = null;
		LOGGER.debug("Loading GeoTiff from URL = " + input + " now.");

		// Versuchen Geo-Information aus Tiff zu lesen

		// Wenn CRS angegeben, dieses durch Hint erzwingen
		Hints hints = new Hints();

		if (crs != null)
			hints.put(Hints.DEFAULT_COORDINATE_REFERENCE_SYSTEM, crs);
		// Reader (mit Metadaten) erzeugen

		LOGGER.debug("First try to create GeoTiff reader");

		/**
		 * Attention, Attention: The GeoTiffReader fails for URLs like:
		 * 
		 * jar:file:/home/sdsd/ad/atlas/raster_textur00949608497.jar!/ad/data/
		 * raster_textur00949608497/textur.tif
		 * 
		 * It should be converted to something like:
		 * jar:http:///atlas/raster_textur00949608497
		 * .jar!/ad/data/raster_textur00949608497/textur.tif
		 * jar:http://www.wikisquare
		 * .de/atlas/raster_textur00949608497.jar!/ad/data
		 * /raster_textur00949608497/textur.tif
		 */

		try {

			GeoTiffReader reader = new GeoTiffReader(input, hints);

			gc = (GridCoverage2D) reader.read(readParams);
			LOGGER.debug("... was successfull!");
			return gc;
		} catch (Exception e) {
			// keine Geo-Informationen in Tiff
			// --> Raster ueber ImageIO einlesen und WorldFile/ProjectionFile
			// verwenden
			LOGGER.warn("No geo information found in '" + input
					+ "'. Using world- and prj-file...");

			BufferedImage im = null;
			double[] tfwInfo = null ;
			if (input instanceof File) {
				im = ImageIO.read((File) input);
				tfwInfo = readWorldFile(IOUtil.changeFileExt((File)input, "tfw"));
				if (crs == null)
					crs = determineProjection((File)input);
			} else if (input instanceof URL) {
				tfwInfo = readWorldFile(IOUtil.changeUrlExt((URL)input, "tfw").openStream());
				im = ImageIO.read((URL) input);
				if (crs == null)
					crs = determineProjection((URL)input);
			}

			if (im == null)
				throw new IIOException(
						"No image reader found for this image type!");
			// World-File einlesen
			
			float w = (float) (im.getWidth() * tfwInfo[0]); // reale Breite =
			// RasterSpalten * hor.
			// Aufloesung
			float h = (float) (im.getHeight() * (-tfwInfo[3])); // reale Hoehe =
			// RasterZeilen *
			// vert. Aufloesung
			float x = (float) tfwInfo[4]; // Suedwestliche Ecke!
			float y = (float) tfwInfo[5] - h; // Suedwestliche Ecke (im tfw-File
			// steht die Nordwestliche!)
			// ggf. Projektion einlesen
			Envelope2D envelope = new Envelope2D(crs, new Rectangle2D.Float(x,
					y, w, h));
			// WICHTIG: Name des Rasters sollte Leer-String sein, da ansonsten
			// das
			// Coloring des Rasters nicht klappt.
			// --> Name der Categories muss (warum auch immer) mit dem Namen
			// des Rasters uebereinstimmen!!!
			gc = new GridCoverageFactory().create("", im.copyData(null),
					envelope);
			return gc;
		}
	}

	//		
	// ArcGridRaster reader = new ArcGridRaster(new BufferedReader(
	// new InputStreamReader(url.openStream())), false);
	// WritableRaster raster = reader.readRaster();
	// if (crs == null)
	// crs = determineProjection(IOUtil.changeUrlExt(url, "prj"));
	//
	// LOGGER.debug("Position  " + reader.getXlCorner() + " / "
	// + reader.getYlCorner());
	// LOGGER.debug("Size      " + reader.getNCols() + " / "
	// + reader.getNRows());
	// LOGGER.debug("Min/Max   " + reader.getMinValue() + " / "
	// + reader.getMaxValue());
	// LOGGER.debug("NoData    " + reader.getNoData());
	// LOGGER.debug("CellSize  " + reader.getCellSize());
	//
	// Double x = reader.getXlCorner(); // Suedwestliche Ecke!
	// Double y = reader.getYlCorner(); // Suedwestliche Ecke!
	// if (y.isNaN() || x.isNaN()) {
	// throw new IllegalArgumentException(
	// "The parameters xlCorner and/or ylCorner could not be read. Hint: Check if the file defines 'xlcenter' instead of 'xlcorner'. If so, just replace all 'center' with 'corner' and try again.");
	// }
	//
	// Double w = (reader.getNCols() * reader.getCellSize()); // reale Breite =
	// // RasterSpalten
	// // * Aufloesung
	// Double h = (reader.getNRows() * reader.getCellSize()); // reale Hoehe =
	// // RasterZeilen
	// // * Aufloesung
	// Envelope2D envelope = new Envelope2D(crs, new Rectangle2D.Double(x, y,
	// w, h));
	//
	// // WICHTIG: Name des Rasters sollte Leer-String sein, da ansonsten das
	// // Coloring des Rasters nicht klappt.
	// // --> Name der Categories muss (warum auch immer) mit dem Namen
	// // des Rasters uebereinstimmen!!!
	// return new GridCoverageFactory().create("", raster, envelope);
	// //
	// // === OHNE ArcGridRaster ===
	// // // ArcGridReader reader = new ArcGridReader( new BufferedReader( new
	// // InputStreamReader( new FileInputStream(file) ) ));
	// // ArcGridReader reader = new ArcGridReader( file );
	// // return (GridCoverage2D)reader.read(null);

}
