/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package skrueger.geotools.labelsearch;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;

import org.apache.log4j.Logger;
import org.geotools.data.DefaultQuery;
import org.geotools.data.FeatureSource;
import org.geotools.feature.FeatureCollection;
import org.geotools.filter.text.cql2.CQL;
import org.geotools.filter.text.cql2.CQLException;
import org.geotools.map.MapLayer;
import org.geotools.styling.Style;
import org.geotools.styling.TextSymbolizer;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.simple.SimpleFeatureType;
import org.opengis.filter.Filter;
import org.opengis.filter.expression.PropertyName;

import schmitzm.geotools.feature.FeatureUtil;
import schmitzm.geotools.styling.StylingUtil;
import schmitzm.lang.LangUtil;
import schmitzm.lang.ResourceProvider;

/**
 * This class allows to search for a {@link String} in a map. The algorithm will
 * analyze the {@link Style} of every visible(?) layer and determine the label
 * attribute. This field is then searched for every feature.
 * 
 * @author Stefan A. Krüger
 * 
 */
public class LabelSearch {
	final static private Logger LOGGER = Logger.getLogger(LabelSearch.class);

	public static ResourceProvider RESOURCE = new ResourceProvider(LangUtil
			.extendPackagePath(LabelSearch.class, "labelsearch"),
			Locale.ENGLISH);

	public static String R(String key, Object... values) {
		return RESOURCE.getString(key, values);
	}

	protected final schmitzm.geotools.gui.SelectableXMapPane mapPane;

	public LabelSearch(final schmitzm.geotools.gui.SelectableXMapPane mapPane) {
		this.mapPane = mapPane;
	}

	public List<SearchResult> search(final String string) {

		final String searchMe = string.toUpperCase();

		final ArrayList<SearchResult> hits = new ArrayList<SearchResult>();

		for (final MapLayer ml : mapPane.getMapContext().getLayers()) {
			try {

				// System.out.println("layer = "+ml.getTitle());

				if (!ml.isVisible())
					continue;

				final List<TextSymbolizer> allTS = StylingUtil
						.getVisibleTextSymbolizers(ml.getStyle());
				if (allTS.size() == 0) {
					// A layer without any TextSymbolizer doesn't have to be
					// searched any more.
					continue;
				}

				// We only deal with one TextSymbolizer so far:
				TextSymbolizer ts = allTS.get(0);

				final FeatureSource<SimpleFeatureType, SimpleFeature> featureSource = (FeatureSource<SimpleFeatureType, SimpleFeature>) ml
						.getFeatureSource();

				SimpleFeatureType schema = featureSource.getSchema();

				final String typeName = schema.getName().getLocalPart();

				PropertyName prop1 = StylingUtil.getFirstPropertyName(schema,
						ts);
				PropertyName prop2 = StylingUtil.getSecondPropertyName(schema,
						ts);

				if (StylingUtil.getFirstPropertyName(schema, ts) == null) {
					// At least one property field we need
					continue;
				}

				Filter searchFilter;
				String[] properties = new String[] { schema.getGeometryDescriptor().getLocalName(), prop1.getPropertyName() };
				// Only one property used...

				searchFilter = CQL.toFilter("strToUpperCase "
						+ prop1.getPropertyName() + " LIKE '%" + searchMe
						+ "%'");

				if (prop2 != null) {
					Filter searchFilter2 = CQL.toFilter("strToUpperCase "
							+ prop2.getPropertyName() + " LIKE '%" + searchMe
							+ "%'");

					searchFilter = FeatureUtil.FILTER_FACTORY2.or(searchFilter,
							searchFilter2);

					properties = LangUtil.extendArray(properties, prop2
							.getPropertyName());
				}

				// Add the layer's filter if it exists
				Filter layerFilter = ml.getQuery().getFilter();
				if (layerFilter != null && layerFilter != Filter.INCLUDE) {
					searchFilter = FeatureUtil.FILTER_FACTORY2.and(layerFilter,
							searchFilter);
				}
				
//				LOGGER.info("Searching for "+searchFilter.toString());

				FeatureCollection<SimpleFeatureType, SimpleFeature> features = (FeatureCollection<SimpleFeatureType, SimpleFeature>) ml
						.getFeatureSource().getFeatures(
								new DefaultQuery(typeName, searchFilter,
										properties));

				final Iterator<SimpleFeature> fi = features.iterator();
				try {
					while (fi.hasNext()) {
						final SimpleFeature f = fi.next();

						String valueString = "";
						valueString = f.getAttribute(prop1.getPropertyName())
								.toString();

						if (prop2 != null) {
							String valueString2 = f.getAttribute(
									prop2.getPropertyName()).toString();

							if (valueString2 != null && !valueString2.isEmpty())
								valueString += ", " + valueString2;
						}

						hits.add(createSearchResult(f, valueString, ml
								.getTitle(), ml));
					}
				} finally {
					features.close(fi);
				}
			} catch (final IOException e) {
				// Searching this layer failed
				LOGGER.error("",e);
			} catch (CQLException e) {
				LOGGER.error("",e);
			}
		} // next layer

		// Hits from the top-most layer should appear first.
		Collections.reverse(hits);

		return hits;
	}

	protected SearchResult createSearchResult(final SimpleFeature f,
			final String title, final String inTitle, MapLayer ml) {
		return new SearchResultFeature(f, title, inTitle, mapPane, ml);
	}

}
