/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
/*
 * SearchMapDialog.java
 *
 * Created on __DATE__, __TIME__
 */

package skrueger.geotools.labelsearch;

import java.awt.BorderLayout;
import java.awt.Cursor;
import java.awt.Point;
import java.awt.Window;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.List;
import java.util.concurrent.ExecutionException;

import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.SwingWorker;

import net.miginfocom.swing.MigLayout;

import org.apache.log4j.Logger;

import schmitzm.geotools.gui.SelectableXMapPane;
import schmitzm.swing.ExceptionDialog;
import schmitzm.swing.SortableJTable;
import schmitzm.swing.SwingUtil;

/**
 * @author Stefan A. Krueger
 */
public class SearchMapDialog extends javax.swing.JDialog {
	final static private Logger LOGGER = Logger
			.getLogger(SearchMapDialog.class);

	private final LabelSearch labelSearch;
	/** This JPanel shows the results to the user if there is more than one **/
	final private JTextField searchTextField = new JTextField(25);
	final private JScrollPane scrollPane = new JScrollPane();

	public String getSearchText() {
		return searchTextField.getText();
	}

	public void setSearchText(String text) {
		searchTextField.setText(text);
	}

	private final SelectableXMapPane mapPane;

	/**
	 * The dialog will be relative to the {@link SelectableXMapPane}s parent
	 * {@link Window}.
	 * 
	 * @param parent
	 * @param labelSearch
	 * @param mapPane
	 * @param windowTitle
	 *            A title to be used for the dialog.
	 */
	public SearchMapDialog(LabelSearch labelSearch, SelectableXMapPane mapPane,
			final String windowTitle) {
		super(SwingUtil.getParentWindow(mapPane), windowTitle);
		this.labelSearch = labelSearch;
		this.mapPane = mapPane;
		initialize();
	}

	// Pressing ESC disposes the Dialog
	KeyAdapter keyEscDispose = new KeyAdapter() {
		public void keyPressed(KeyEvent e) {
			int key = e.getKeyCode();
			if (key == KeyEvent.VK_ESCAPE) {
				dispose();
			}
		}
	};

	// Pressing ENTER in the JTextfield automatically starts the search
	KeyAdapter keyListenerStartSearch = new KeyAdapter() {
		public void keyPressed(KeyEvent e) {
			int key = e.getKeyCode();
			if (key == KeyEvent.VK_ENTER) {
				if (searchTextField.getText().trim().length() > 0)
					search(searchTextField.getText());
			}
		}
	};

	private void initialize() {

		JPanel cp = new JPanel(new MigLayout("wrap 1","[grow]","[][][grow]"));

		JLabel label = new JLabel(LabelSearch.R("SearchMapDialog.searchString.Label"));
		label.setLabelFor(searchTextField);
		searchTextField.setToolTipText(LabelSearch.R("SearchMapDialog.searchString.tt"));
		searchTextField.addKeyListener(keyListenerStartSearch);
		// Pressing ESC disposes the Dialog
		searchTextField.addKeyListener(keyEscDispose);
		cp.add(label, "split 2");
		cp.add(searchTextField);
		
		cp.add(new JLabel(LabelSearch.R("SearchMapDialog.Explanation")));
		
		cp.add(scrollPane);

		cp.add(scrollPane, BorderLayout.SOUTH);

		cp.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));

		setContentPane(cp);
		pack();

		Point mapPanelocationOnScreen = mapPane.getLocationOnScreen();
		mapPanelocationOnScreen.x -= mapPane.getLocation().x;
		mapPanelocationOnScreen.y -= mapPane.getLocation().y;
		setLocation(mapPanelocationOnScreen);

		// SwingUtil.centerFrameOnScreenRandom(this);
	}

	/**
	 * Performes a search on the {@link SelectableXMapPane}'s labels and outputs the
	 * possible numerouse results.
	 * 
	 * @param text
	 *            Text to search
	 */
	public void search(final String text) {

		final Cursor backupCursor = getContentPane().getCursor();
		getContentPane().setCursor(
				Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
		searchTextField.setEnabled(false);

		SwingWorker<List<SearchResult>, String> swingWorker = new SwingWorker<List<SearchResult>, String>() {
			@Override
			protected List<SearchResult> doInBackground() throws Exception {
				return labelSearch.search(text);
			}

			@Override
			protected void done() {
				getContentPane().setCursor(backupCursor);
				searchTextField.setEnabled(true);
				super.done();
				try {
					showMultipleResults(get());
				} catch (InterruptedException e) {
					LOGGER.error(e);
					ExceptionDialog.show(SearchMapDialog.this, e);
				} catch (ExecutionException e) {
					LOGGER.error(e);
					ExceptionDialog.show(SearchMapDialog.this, e);
				}
			}
		};
		swingWorker.execute();

	}

	private void showMultipleResults(final List<SearchResult> searchResultsList) {

		final SortableJTable resultsTable = new SortableJTable();
		resultsTable.setModel(new SearchResultTableModel(searchResultsList));
		
		resultsTable.addKeyListener(keyEscDispose);

		// set up the table
		resultsTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

//		resultsTable.setRowSorter( new TableRowSorter<SearchResultTableModel>((SearchResultTableModel)resultsTable.getModel()));
//
		resultsTable.setToolTipText(LabelSearch.R("SearchMapDialog.resultsTable.tt"));
		
//		// The first column should be small
//		resultsTable.getColumnModel().getColumn(0).setMinWidth(20);
//		resultsTable.getColumnModel().getColumn(0).setMaxWidth(25);

		resultsTable.addMouseListener(new MouseAdapter() {

			@Override
			public void mouseClicked(MouseEvent e) {
				int index = resultsTable.getSelectedModelRow();
				SearchResult searchResult = searchResultsList.get(index);
				Cursor backupCursor = scrollPane.getCursor();
				scrollPane.setCursor(Cursor
						.getPredefinedCursor(Cursor.WAIT_CURSOR));
				searchResult.open();
				scrollPane.setCursor(backupCursor);
			}
		});

		scrollPane.setViewportView(resultsTable);
		SwingUtil.setPreferredHeight(scrollPane, 140);
		SwingUtil.setMaximumHeight(scrollPane, 140);

		pack();
	}

}
