/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package skrueger.geotools.selection;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.Vector;

import org.jfree.data.general.Dataset;

import schmitzm.jfree.chart.renderer.SelectionRenderer;
import schmitzm.jfree.chart.selection.DatasetSelectionChangeEvent;
import schmitzm.jfree.chart.selection.DatasetSelectionListener;
import schmitzm.jfree.chart.selection.DatasetSelectionModel;
import schmitzm.jfree.chart.selection.DatasetSelectionModelProvider;
import schmitzm.jfree.feature.FeatureDatasetSelectionModel;

/**
 * This class keeps the selection of a {@link Dataset} (based on feature
 * attributes) synchronized with the {@link StyledLayerSelectionModel} of a layer.
 * This is done by implementing:
 * <ul>
 * <li>a {@link PropertyChangeListener} which listens to the
 * {@link StyledLayerSelectionModel} and accordingly changes the {@link SelectionRenderer}
 * selection</li>
 * <li>a {@link DatasetSelectionModel} which listens to the {@link SelectionRenderer} and
 * accordingly changes the {@link StyledLayerSelectionModel} selection</li>
 * </ul>
 * After creating, the instance of this synchronizer must be added as listener
 * to both, the {@link StyledLayerSelectionModel} and the chart's
 * {@link DatasetSelectionModel} (e.g. the renderer).
 * @see DatasetSelectionModelProvider
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 */
public class ChartSelectionSynchronizer extends StyledLayerSelectionModelSynchronizer<StyledFeatureLayerSelectionModel>
                                        implements DatasetSelectionListener {

  /** Holds the chart datset to keep synchronized with the layer selection model. */
  protected FeatureDatasetSelectionModel<?,?,?> datasetSelModel = null;

  /**
   * Creates a new synchronizer.
   * @param layerSelModel
   *             layer selection model to keep synchronized with the dataset
   *             selection model
   * @param datasetSelModel
   *            dataset selection model to keep synchronized with the layer
   *            selection model
   */
  public ChartSelectionSynchronizer(StyledFeatureLayerSelectionModel layerSelModel, FeatureDatasetSelectionModel<?,?,?> datasetSelModel) {
    super(layerSelModel);
    this.datasetSelModel = datasetSelModel;
  }

  /**
   * Called by {@link StyledLayerSelectionModel} when the selection on other
   * selection components (map, table, ...) has changed. When calling this
   * method changes the dataset selection according to the
   * {@link StyledLayerSelectionModel} selection.
   * @param evt an event
   */
  @Override
  public void propertyChange(PropertyChangeEvent evt) {
	 
	if (!isEnabled()) 
		return;
	  
    if (!(evt instanceof StyledLayerSelectionEvent))
      return;
    StyledLayerSelectionEvent selEvt = (StyledLayerSelectionEvent) evt;
    // Only react on own layer and ignore events invoked by
    // this component itself
    if (selEvt.getEmitter() != layerSelModel || selectionChangeCausedByMe)
        return;
    // Apply new selection on chart (except this event is one of
    // some more events)
    Vector<String> newSelection = layerSelModel.getSelection();
    if (newSelection == null)
      return;

    // Avoid event circles in valueChanged(..)
    selectionChangeCausedByMe = true;

    datasetSelModel.setValueIsAdjusting(true);
    datasetSelModel.clearSelection();
    for (String fID : newSelection)
      datasetSelModel.setItemSelected(fID, true);
    datasetSelModel.setValueIsAdjusting(false); // event is fired autmatically!

    // Danger of event circles in valueChanged(..) banned
    selectionChangeCausedByMe = false;
  }

  /**
   * Called when the chart selection is changed by the user. When calling this
   * method changes the selection of the {@link StyledLayerSelectionModel}.
   * @param evt an event
   */
  @Override
  public void selectionChanged(DatasetSelectionChangeEvent evt) {
    // ignore event if it is part of multiple events
    if (evt != null && evt.getSource().getValueIsAdjusting())
        return;
    // ignore event if it is caused by the ChartSelectionSynchronizer
    if (selectionChangeCausedByMe)
        return;

    // Avoid event circles in propertyChange(..)
    selectionChangeCausedByMe = true;

    // reset the selection of the DpLayerSelectionModel
    layerSelModel.setValueIsAdjusting(true);
    layerSelModel.clearSelection();
    for (String featureID : datasetSelModel.getSelectedFeatures())
      layerSelModel.addSelection(featureID);
    layerSelModel.setValueIsAdjusting(false);

    // Danger of event circles in propertyChange(..) banned
    selectionChangeCausedByMe = false;
  }
  

}
