/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
/** 
 Copyright 2008 Stefan Alfons Krüger 

 atlas-framework - This file is part of the Atlas Framework

 This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
 This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

 Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
 Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
 Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/
package skrueger.geotools.selection;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.Vector;

import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionListener;

import org.geotools.feature.FeatureCollection;
import org.geotools.feature.FeatureIterator;
import org.geotools.filter.FidFilterImpl;
import org.geotools.map.MapLayer;
import org.geotools.styling.FeatureTypeStyle;
import org.geotools.styling.Style;
import org.geotools.styling.visitor.DuplicatingStyleVisitor;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.simple.SimpleFeatureType;
import org.opengis.filter.Filter;
import org.opengis.filter.Not;
import org.opengis.filter.identity.FeatureId;

import schmitzm.geotools.FilterUtil;
import schmitzm.geotools.gui.SelectableXMapPane;
import schmitzm.geotools.map.event.FeatureSelectedEvent;
import schmitzm.geotools.map.event.JMapPaneListener;
import schmitzm.geotools.map.event.MapPaneEvent;
import schmitzm.geotools.styling.StylingUtil;
import skrueger.geotools.MapPaneToolBar;
import skrueger.geotools.StyledFeaturesInterface;

/**
 * This class keeps the selection of a (feature) {@link JTable} synchronized
 * with the {@link StyledLayerSelectionModel} of a layer. This is done by
 * implementing:
 * <ul>
 * <li>a {@link PropertyChangeListener} which listens to the
 * {@link StyledLayerSelectionModel} and accordingly changes the {@link JTable}
 * selection</li>
 * <li>a {@link ListSelectionListener} which listens to the {@link JTable} and
 * accordingly changes the {@link StyledLayerSelectionModel} selection</li>
 * </ul>
 * After creating, the instance of this synchronizer must be added as listener
 * to both, the {@link StyledLayerSelectionModel} and the table's
 * {@link ListSelectionModel}.
 * 
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *         (University of Bonn/Germany)
 */
public class FeatureMapLayerSelectionSynchronizer extends
		StyledLayerSelectionModelSynchronizer<StyledFeatureLayerSelectionModel>
		implements JMapPaneListener {

	/**
	 * This constant is set as the {@link FeatureTypeStyle#getName()} attribute
	 * in the {@link FeatureTypeStyle}s that only exist to present the selected
	 * features
	 **/
	public static final String SELECTION_STYLING_FTS_NAME = "SELECTION";

	/**
	 * Holds the {@link MapLayer} to keep synchronized with the layer selection
	 * model.
	 */
	protected final MapLayer mapLayer;
	protected final StyledFeaturesInterface<?> styledLayer;
	protected final SelectableXMapPane mapPane;
	private final MapPaneToolBar toolBar;

	/**
	 * Creates a new synchronizer
	 * 
	 * @param layerSelModel
	 *            layer selection model to keep synchronized with the
	 *            {@link MapLayer}
	 * 
	 * @param mapLayer
	 *            {@link MapLayer} to keep synchronized with.
	 */
	public FeatureMapLayerSelectionSynchronizer(
			StyledFeatureLayerSelectionModel layerSelModel,
			StyledFeaturesInterface<?> styledLayer, MapLayer mapLayer,
			SelectableXMapPane mapPane, MapPaneToolBar toolBar) {

		super(layerSelModel);
		this.styledLayer = styledLayer;

		this.mapLayer = mapLayer;
		this.mapPane = mapPane;
		this.toolBar = toolBar;
	}

	/**
	 * Called by {@link StyledLayerSelectionModel} when a the selection on other
	 * selection components (map, chart, ...) has changed. When calling this
	 * 
	 * method changes the {@link MapLayer} selection according to the
	 * {@link StyledLayerSelectionModel} selection.
	 * 
	 * @param evt
	 *            an event
	 */
	@Override
	public void propertyChange(PropertyChangeEvent evt) {

		if (!isEnabled())
			return;

		if (!(evt instanceof StyledLayerSelectionEvent))
			return;
		StyledLayerSelectionEvent selEvt = (StyledLayerSelectionEvent) evt;
		// Only react on own layer and ignore events invoked by
		// this component itself
		if (selEvt.getEmitter() != layerSelModel || selectionChangeCausedByMe)
			return;
		// Apply new selection on table (except this event is one of
		// some more events)
		Vector<String> newSelection = layerSelModel.getSelection();
		if (newSelection == null)
			return;

		// Avoid event circles in valueChanged(..)
		selectionChangeCausedByMe = true;

		changeLayerStyle(newSelection);

		// Danger of event circles in valueChanged(..) banned
		selectionChangeCausedByMe = false;
	}

	/**
	 * Changes the Style of the {@link MapLayer} to reflect changes of the
	 * selection.
	 * 
	 * @param newSelection
	 *            A {@link Vector} of SimpleFeature-IDs that are selected.
	 */
	private void changeLayerStyle(final Vector<String> newSelection) {
		try {

			Style originalStyle = mapLayer.getStyle();

			if (newSelection.isEmpty()) {

				// Check if the Style contains a SELECTION FTS

				FeatureTypeStyle[] clone = originalStyle.featureTypeStyles()
						.toArray(new FeatureTypeStyle[] {}).clone();

				for (FeatureTypeStyle fts : clone) {
					if (fts.getName() != null
							&& fts.getName().equals(SELECTION_STYLING_FTS_NAME)) {
						originalStyle.featureTypeStyles().remove(fts);

						mapLayer.setStyle(originalStyle);
						// replaceRenderer();
						// // mapPane.refresh();

						return;
					}
				}

			} else {

				LOGGER.debug("SELECTION .. change style");

				// Saving a repaint if the selection didn't change
				if (!selectionChanged(newSelection, originalStyle)) {
					return;
				}

				// We take Style from the MapLayer that is displayed at the
				// moment. We do not use the styledLayer.getStyle, because in
				// the atlas, this always return the default style, but
				// additional styles might be selected.
				// Taking the style from the mapLayer indicated, that we have to
				// remove any selection rules first.

				FeatureTypeStyle selectionFTStyle = StylingUtil
						.createSelectionStyle(styledLayer.getGeoObject());

				selectionFTStyle.setName(SELECTION_STYLING_FTS_NAME);

				/**
				 * 
				 * Add a Filter to the selectionMapStyle, so that it is only
				 * used on objects that are selected. <br/>
				 * To optimize rendering speed, the filter checks whether more
				 * than half of the features are selected. If so, the filter is
				 * inverted to be shorter.
				 * 
				 */
				final FeatureCollection<SimpleFeatureType, SimpleFeature> featureCollectionFiltered = styledLayer
						.getFeatureCollectionFiltered();
				if (newSelection.size() > featureCollectionFiltered.size() / 2) {
					FeatureIterator<SimpleFeature> iterator = featureCollectionFiltered
							.features();
					Set<FeatureId> antiFids = new HashSet<FeatureId>();
					try {
						while (iterator.hasNext()) {
							SimpleFeature next = iterator.next();
							if (!newSelection.contains(next.getID()))
								antiFids.add(FilterUtil.FILTER_FAC2
										.featureId(next.getID()));
						}

						selectionFTStyle.rules().get(0).setFilter(
								FilterUtil.FILTER_FAC2
										.not(FilterUtil.FILTER_FAC2
												.id(antiFids)));

					} finally {
						featureCollectionFiltered.close(iterator);
					}
				} else {
					Set<FeatureId> fids = new HashSet<FeatureId>();
					for (String fid : newSelection) {
						if (fid != null)
							fids.add(FilterUtil.FILTER_FAC2.featureId(fid));
					}

					selectionFTStyle.rules().get(0).setFilter(
							FilterUtil.FILTER_FAC2.id(fids));
				}

				// Maybe there has already been another selection
				// FeatureTypeStyle... Let's replace it...
				boolean foundAndReplaced = false;
				for (FeatureTypeStyle fts : originalStyle.featureTypeStyles()) {
					if (fts.getName() != null
							&& fts.getName().equals(SELECTION_STYLING_FTS_NAME)) {
						foundAndReplaced = true;
						fts.rules().clear();
						fts.rules().addAll(selectionFTStyle.rules());
						break;
					}
				}
				if (!foundAndReplaced) {
					originalStyle.featureTypeStyles().add(selectionFTStyle);
				}

				// Refresh the map
				// mapPane.refresh();

				DuplicatingStyleVisitor dsv = new DuplicatingStyleVisitor();
				dsv.visit(originalStyle);
				Style newStyle = (Style) dsv.getCopy();

				// SK-Debug
				try {
					//
					StylingUtil.saveStyleToSLD(newStyle, new File(
							"/home/stefan/Desktop/selection.sld"));
				} catch (Exception e) {
				}

				// DuplicatingStyleVisitor dsv = new DuplicatingStyleVisitor();
				// dsv.visit(originalStyle);
				// Style newStyle = originalStyle;

				// Style newStyle = StylingUtil.STYLE_BUILDER.createStyle();
				// newStyle.featureTypeStyles().addAll(originalStyle.featureTypeStyles());
				mapLayer.setStyle(newStyle);

				replaceRenderer();
			}

		} catch (Exception e) {
			LOGGER.error("Error while trying to create a selection style", e);
		}
	}

	/**
	 * Analyzes whether the selection has changed in comparison to the selection
	 * stored in the mapLayer.Style
	 * 
	 * @param newSelection a List<String> of all newly selected FIDs
	 * @param originalStyle the original {@link Style} that has an earlier selection coded into the {@link Style}
	 * @return <code>true</code> if changed.
	 */
	private boolean selectionChanged(Vector<String> newSelection,
			Style originalStyle) {

		boolean SELECTION_STYLING_foundInMapStyle = false;

		/**
		 * testing, whether the selection really changed. If not, we can save
		 * one rendering!
		 */
		for (FeatureTypeStyle fts : originalStyle.featureTypeStyles()) {

			if (fts.getName() != null
					&& fts.getName().equals(SELECTION_STYLING_FTS_NAME)) {

				SELECTION_STYLING_foundInMapStyle = true;

				Filter filter = fts.rules().get(0).getFilter();
				if (filter instanceof Not) {
					FidFilterImpl antiOrigFidsFilter = (FidFilterImpl) ((Not) filter)
							.getFilter();

					// Check one way
					final Set<String> antiFids = antiOrigFidsFilter
							.getFidsSet();

					if (antiFids.isEmpty() && !newSelection.isEmpty())
						return true;

					for (String fid : antiFids) {
						if (newSelection.contains(fid)) {
							return true;
						}
					}

					// Check the other way
					for (String fid : newSelection) {
						if (antiFids.contains(fid)) {
							return true;
						}
					}

					if (antiFids.size() + newSelection.size() != styledLayer
							.getFeatureCollectionFiltered().size())
						return true;

				} else {
					FidFilterImpl origFidsFilter = (FidFilterImpl) filter;

					// Check one way
					final Set<String> fids = origFidsFilter.getFidsSet();

					if (fids.isEmpty() && newSelection.isEmpty())
						return false;
					if (fids.size() != newSelection.size())
						return true;

					for (String fid : fids) {
						if (!newSelection.contains(fid)) {
							return true;
						}
					}

					// Check the other way
					for (String fid : newSelection) {
						if (!fids.contains(fid)) {
							return true;
						}
					}

				}

				break;
			}
		}

		if (!SELECTION_STYLING_foundInMapStyle && !newSelection.isEmpty())
			return true;

		return false;
	}

	/**
	 * Replaces the local renderer
	 */
	private void replaceRenderer() {
	}

	/**
	 * Used to synchronize {@link FeatureSelectedEvent}s with the
	 * {@link StyledFeatureLayerSelectionModel}
	 */
	@Override
	public void performMapPaneEvent(MapPaneEvent e) {

		// Ignore event if it is caused by us or the synchronizer is disabled.
		if (!isEnabled() || selectionChangeCausedByMe)
			return;

		if (!(e instanceof FeatureSelectedEvent)) {
			// LOGGER.debug("Ignoring event " + e);
			return;
		}

		/**
		 * Only listen to FeatureSelectedEvents if an appropriate tool is
		 * selected.
		 */
		final int selectedTool = toolBar.getSelectedTool();
		if (selectedTool != MapPaneToolBar.TOOL_SELECTION_ADD
				&& selectedTool != MapPaneToolBar.TOOL_SELECTION_REMOVE
				&& selectedTool != MapPaneToolBar.TOOL_SELECTION_SET) {
			return;
		}

		// only listen to events directly coming from JMapPane
		// selection (ignore selections from FilterDialog)
		if (e.getSourceObject() != this.mapPane)
			return;

		/**
		 * Checking, that the FeatureSelectedEvent actually contains features
		 * from this layer
		 */
		FeatureSelectedEvent fse = (FeatureSelectedEvent) e;
		final String sourceID = fse.getSourceLayer().getTitle();
		final String syncForID = mapLayer.getTitle();
		if (sourceID != null && syncForID != null
				&& !sourceID.equals(syncForID)) {
			LOGGER.debug("Ignoring a FeatureSelectedEvent from " + sourceID);
			return;
		}

		// LOGGER.debug("do event " + fse);

		// Avoid event circles in propertyChange(..)
		selectionChangeCausedByMe = true;

		final FeatureCollection<SimpleFeatureType, SimpleFeature> selectionResult = fse
				.getSelectionResult();
		Iterator<SimpleFeature> fi = selectionResult.iterator();
		try {

			// reset the selection of the DpLayerSelectionModel
			// layerSelModel.setValueIsAdjusting(true);

			if (selectedTool == MapPaneToolBar.TOOL_SELECTION_ADD) {
				layerSelModel.setValueIsAdjusting(true);

				for (int fIdx = 0; fi.hasNext(); fIdx++) {
					SimpleFeature f = fi.next();
					layerSelModel.addSelection(f.getID());
				}
				layerSelModel.setValueIsAdjusting(false);
			} else if (selectedTool == MapPaneToolBar.TOOL_SELECTION_SET) {
				layerSelModel.setValueIsAdjusting(true);
				layerSelModel.clearSelection();

				for (int fIdx = 0; fi.hasNext(); fIdx++) {
					SimpleFeature f = fi.next();
					layerSelModel.addSelection(f.getID());
				}

				// LOGGER.debug("Setting selection to " + fi.());

				layerSelModel.setValueIsAdjusting(false);
			} else if (selectedTool == MapPaneToolBar.TOOL_SELECTION_REMOVE) {
				layerSelModel.setValueIsAdjusting(true);
				for (int fIdx = 0; fi.hasNext(); fIdx++) {
					SimpleFeature f = fi.next();
					layerSelModel.removeSelection(f.getID());
				}
				layerSelModel.setValueIsAdjusting(false);
			}

		} finally {
			selectionResult.close(fi);
		}

		// Show selected features in the map, which is not automatically done by
		// the origin: FeatureSelectedEvent
		changeLayerStyle(layerSelModel.getSelection());

		// Danger of event circles in propertyChange(..) banned
		selectionChangeCausedByMe = false;
	}

}
