/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
/**
 Copyright 2008 Stefan Alfons Krüger

 atlas-framework - This file is part of the Atlas Framework

 This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
 This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

 Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
 Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
 Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/

package skrueger.geotools.selection;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.Collection;
import java.util.HashSet;
import java.util.Vector;

import javax.swing.event.EventListenerList;

import org.apache.log4j.Logger;
import org.opengis.feature.simple.SimpleFeature;

import schmitzm.swing.event.PropertyChangeEmitter;
import skrueger.geotools.StyledLayerInterface;

/**
 * This manager holds a set of objects which are <i>selected</i> for a {@link StyledLayerInterface}.
 * Several components can connect to this model to keep their selection synchronized.
 * @see #addSelectionListener(java.beans.PropertyChangeListener)
 * @see StyledLayerSelectionModelSynchronizer
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a> (University of Bonn/Germany)
 */
public abstract class StyledLayerSelectionModel<E> extends PropertyChangeEmitter<StyledLayerSelectionModelSynchronizer<StyledLayerSelectionModel<E>>> {
  /** A static logger for this class  */
  protected final Logger LOGGER = Logger.getLogger(this.getClass());

  /** Indicates that the selection has changed in {@link PropertyChangeEvent}. */
  public final static String SELECTION = "SELECTION";


  /** Holds the connected listeners. */
  protected EventListenerList listenerList = new EventListenerList();

  /** Holds the current selection (e.g. {@link SimpleFeature Features}). */
  protected final HashSet<E> selectionObjects;

  /** Holds the styled map whose selection is managed by this class. */
  private final StyledLayerInterface<?> styledLayer;


  /** Indicates whether a selection change is a part of multiple
   *  changes. If {@code true} {@link #refreshSelection()} has NO effect
   *  until {@link #setValueIsAdjusting(boolean) setValueIsAdjusting(false)}
   *  is called. */
  private boolean valueIsAdjusting = false;

  /**
   * Creates a new selection model.
   * @param styledLayer styled layer the selection is controlled of
   */
  public StyledLayerSelectionModel(StyledLayerInterface<?> styledLayer) {
	this.selectionObjects = new HashSet<E>();
	this.styledLayer         = styledLayer;
  }

  /**
   * Fires an event to every connected listener to update their selection
   * according to the selection of this model.<br>
   * <b>Does nothing if adjusting is <code>true</code></b>.
   */
  public void refreshSelection() {
    if ( getValueIsAdjusting() )
      return;
    firePropertyChangeEvent( createSelectionChangeEvent(null) );
  }

  /**
   * Clears the selection and fires an event to every connected listener
   * to update their selection.
   */
  public void clearSelection() {
    selectionObjects.clear();
    refreshSelection();
  }

  /**
   * Creates a simple {@link PropertyChangeEvent} for property {@link #SELECTION}
   * with no new and old value, just to inform the listeners THAT
   * some selection has changed.
   * Subclasses may overwrite this method to create more sophisticated
   * events.
   * @param respObject object which is responsible for the selection change; this
   *                   object should ignore this event to avoid event circles (can be {@code null}
   *                   in case of general refresh events; in this case all connected
   *                   components will refresh their selection)
   *
   */
  public StyledLayerSelectionEvent createSelectionChangeEvent(Object respObject) {
    if ( respObject == null )
      respObject = this;
    return new StyledLayerSelectionEvent(this,respObject,SELECTION,null,null);
  }


  /**
   * Returns whether the current selection change is a part of multiple
   * changes. If {@code true} {@link #refreshSelection()} has no effect
   * and no {@link StyledLayerSelectionEvent} is fired until
   * {@link #setValueIsAdjusting(boolean) setValueIsAdjusting(false)}
   * is called.
   */
  public boolean getValueIsAdjusting() {
    return this.valueIsAdjusting;
  }

  /**
   * Sets whether the following selection changes are part of multiple
   * changes. If {@code true} {@link #refreshSelection()} has no effect
   * and no {@link StyledLayerSelectionEvent} event is fired.
   * If the value is set from {@code true} to {@code false} an automatic
   * refresh is initiated.
   * @see #refreshSelection()
   */
  public void setValueIsAdjusting(boolean valueIsAdjusting) {
    if ( valueIsAdjusting != this.valueIsAdjusting ) {
      this.valueIsAdjusting = valueIsAdjusting;
      if ( !this.valueIsAdjusting )
        refreshSelection();
    }
  }


  /**
   * Extends the current selection by an individual object.
   * @param selectedObject defines the (new) selected object
   * @return {@code true} if selection has changed by calling this method
   */
  public boolean addSelection(E selectedObject) {
	 
    if ( selectionObjects.add(selectedObject) ) {
      refreshSelection();
      return true;
    }
    return false;
  }

  /**
   * Removes an individual object from the current selection.
   * @param unselectedObject defines an object which will be unselected
   * @return {@code true} if selection has changed by calling this method
   */
  public boolean removeSelection(E unselectedObject) {
    if ( selectionObjects.remove(unselectedObject) ) {
      refreshSelection();
      return true;
    }
    return false;
  }

  /**
   * Extends the current selection by individual objects.
   * @param selectedObjects defines some objects which will be selected
   * @return {@code true} if selection has changed by calling this method
   */
  public boolean addSelection(Collection<E> selectedObjects) {
    if ( selectionObjects.addAll(selectedObjects) ) {
      refreshSelection();
      return true;
    }
    return false;
  }

  /**
   * Removes individual objects from the current selection.
   * @param unselectedObjects defines some objects which will be unselected
   * @return {@code true} if selection has changed by calling this method
   */
  public boolean removeSelection(Collection<E> unselectedObjects) {
    if ( selectionObjects.removeAll(unselectedObjects) ) {
      refreshSelection();
      return true;
    }
    return false;
  }

  /**
   * Returns all selected objects as list.
   * @return a copy of the original list, so changes have no effect!
   */
  public Vector<E> getSelection() {
    return new Vector<E>(selectionObjects);
  }

  /**
   * Adds a listener to the listener list. Only alias for {@link #addListener(PropertyChangeListener)}.
   * @param listener a listener
   */
  public void addSelectionListener(StyledLayerSelectionModelSynchronizer<StyledLayerSelectionModel<E>> listener) {
    this.addListener(listener);
  }

  /**
   * Removes a listener from the listener list. Only alias for {@link #removeListener(PropertyChangeListener)}.
   * @param listener a listener
   */
  public boolean removeSelectionListener(StyledLayerSelectionModelSynchronizer<StyledLayerSelectionModel<E>> listener) {
    return this.removeListener(listener);
  }

public StyledLayerInterface<?> getStyledLayer() {
	return styledLayer;
}


}
