/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
/** 
 Copyright 2008 Stefan Alfons Krüger 
 
 atlas-framework - This file is part of the Atlas Framework

 This library is free software; you can redistribute it and/or modify it under the terms of the GNU Lesser General Public License as published by the Free Software Foundation; either version 2.1 of the License, or (at your option) any later version.
 This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 You should have received a copy of the GNU Lesser General Public License along with this library; if not, write to the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA

 Diese Bibliothek ist freie Software; Sie dürfen sie unter den Bedingungen der GNU Lesser General Public License, wie von der Free Software Foundation veröffentlicht, weiterverteilen und/oder modifizieren; entweder gemäß Version 2.1 der Lizenz oder (nach Ihrer Option) jeder späteren Version.
 Diese Bibliothek wird in der Hoffnung weiterverbreitet, daß sie nützlich sein wird, jedoch OHNE IRGENDEINE GARANTIE, auch ohne die implizierte Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN BESTIMMTEN ZWECK. Mehr Details finden Sie in der GNU Lesser General Public License.
 Sie sollten eine Kopie der GNU Lesser General Public License zusammen mit dieser Bibliothek erhalten haben; falls nicht, schreiben Sie an die Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
 **/
package skrueger.geotools.selection;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.Vector;

import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;
import javax.swing.table.TableModel;

import org.opengis.feature.simple.SimpleFeature;

import schmitzm.geotools.gui.FeatureCollectionTableModel;
import schmitzm.swing.table.PipedTableModel;

/**
 * This class keeps the selection of a (feature) {@link JTable} synchronized
 * with the {@link StyledLayerSelectionModel} of a layer. This is done by
 * implementing:
 * <ul>
 * <li>a {@link PropertyChangeListener} which listens to the
 * {@link StyledLayerSelectionModel} and accordingly changes the {@link JTable}
 * selection</li>
 * <li>a {@link ListSelectionListener} which listens to the {@link JTable} and
 * accordingly changes the {@link StyledLayerSelectionModel} selection</li>
 * </ul>
 * After creating, the instance of this synchronizer must be added as listener
 * to both, the {@link StyledLayerSelectionModel} and the table's
 * {@link ListSelectionModel}.
 * 
 * @author <a href="mailto:martin.schmitz@koeln.de">Martin Schmitz</a>
 *         (University of Bonn/Germany)
 */
public class TableSelectionSynchronizer extends
		StyledLayerSelectionModelSynchronizer<StyledFeatureLayerSelectionModel>
		implements ListSelectionListener {
	/**
	 * Holds the table to keep synchronized with the layer selection model.
	 */
	protected JTable featureTable = null;
	/** Holds the table's data. */
	protected FeatureCollectionTableModel featureTableModel = null;

	/**
	 * Creates a new synchronizer
	 * 
	 * @param layerSelModel
	 *            layer selection model to keep synchronized with the feature
	 *            table
	 * @param table
	 *            table to keep synchronized with the layer selection model
	 */
	public TableSelectionSynchronizer(
			StyledFeatureLayerSelectionModel layerSelModel, JTable table) {
		super(layerSelModel);

		TableModel model = table.getModel();
		if (model instanceof PipedTableModel)
			model = ((PipedTableModel) table.getModel()).getPipedModel();

		if (!(model instanceof FeatureCollectionTableModel))
			throw new IllegalArgumentException(
					"Table must have a model instance of FeatureCollectionTableModel");

		this.featureTableModel = (FeatureCollectionTableModel) model;
		this.featureTable = table;
	}

	/**
	 * Called by {@link StyledLayerSelectionModel} when a the selection on other
	 * selection components (map, chart, ...) has changed. When calling this
	 * method changes the table selection according to the
	 * {@link StyledLayerSelectionModel} selection.
	 * 
	 * @param evt
	 *            an event
	 */
	@Override
	public void propertyChange(PropertyChangeEvent evt) {
		if (!(evt instanceof StyledLayerSelectionEvent))
			return;
		StyledLayerSelectionEvent selEvt = (StyledLayerSelectionEvent) evt;
		// Only react on own layer and ignore events invoked by
		// this component itself
		if (selEvt.getEmitter() != layerSelModel || selectionChangeCausedByMe)
			return;
		// Apply new selection on table (except this event is one of
		// some more events)
		Vector<String> newSelection = layerSelModel.getSelection();
		if (newSelection == null)
			return;

		// LOGGER.debug("Table pop changed "+selEvt);
		// LOGGER.debug("Table pop changed to "+newSelection);

		// Avoid event circles in valueChanged(..)
		selectionChangeCausedByMe = true;

		ListSelectionModel tableSelModel = featureTable.getSelectionModel();
		tableSelModel.setValueIsAdjusting(true);
		tableSelModel.clearSelection();
		for (String fid : newSelection) {
			int modelIdx = featureTableModel.findFeature(fid);
			if (modelIdx >= 0) {
				int tableIdx = featureTable.convertRowIndexToView(modelIdx);
				tableSelModel.addSelectionInterval(tableIdx, tableIdx);
			} else {
				LOGGER
						.warn("Something that is not visible in the Table should be selected");
			}
		}
		tableSelModel.setValueIsAdjusting(false); // event is fired
		// autmatically!

		// Danger of event circles in valueChanged(..) banned
		selectionChangeCausedByMe = false;
	}

	/**
	 * Called when the table selection is changed by the user. When calling this
	 * method changes the selection of the {@link StyledLayerSelectionModel}.
	 * 
	 * @param evt
	 *            an event
	 */
	@Override
	public void valueChanged(ListSelectionEvent evt) {
		// ignore event if it is part of multiple events
		if (evt != null && evt.getValueIsAdjusting())
			return;
		// ignore event if it is caused by the DpLayerVectorSelectionModel
		if (selectionChangeCausedByMe)
			return;

		// Avoid event circles in propertyChange(..)
		selectionChangeCausedByMe = true;

		// reset the selection of the DpLayerSelectionModel
		layerSelModel.setValueIsAdjusting(true);
		LOGGER.debug("valueCahnged in TableSyncronizer vom Index "
				+ evt.getFirstIndex() + " to " + evt.getLastIndex());

		if (evt == null || (evt != null && evt.getFirstIndex() == -1)){
			// If the value is changing because the filter has been changed (e.g.
			// table structure change) the evt.getFirstIndex() is -1, but evt.getLastIndex is > -1
			layerSelModel.clearSelection();
		} else{
			
			for (int i = evt.getFirstIndex(); i <= evt.getLastIndex(); i++) {
				int featureIndex = featureTable.convertRowIndexToModel(i);
				SimpleFeature changedFeature = featureTableModel.getFeature(featureIndex);
				if (featureTable.isRowSelected(i))
					layerSelModel.addSelection(changedFeature.getID());
				else
					layerSelModel.removeSelection(changedFeature.getID());
			}
		}
		layerSelModel.setValueIsAdjusting(false);

		// Danger of event circles in propertyChange(..) banned
		selectionChangeCausedByMe = false;
	}
}
