/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package skrueger.i8n;

import java.util.List;
import java.util.Locale;
import java.util.Set;
import java.util.TreeSet;

import org.apache.log4j.Logger;

public class I8NUtil {
	static final Logger LOGGER = Logger.getLogger(I8NUtil.class);

	private static Set<String> languageCodes = new TreeSet<String>();
	static {
		for (Locale locale : java.util.Locale.getAvailableLocales()) {
			getLanguageCodes().add(locale.getLanguage());
		}
		// for (String code : java.util.Locale.getISOLanguages()) {
		// getLanguageCodes().add(code);
		// }
	}

	/**
	 * Creates a {@link Translation}<br/>
	 * <p>
	 * <li>If <code>oneLineCoded</code> is empty or null, NO TRANSLATION is set.
	 * <li>If format can't be recognized, the {@link String} is interpreted as
	 * the translation in the <code>{@value #DEFAULT_KEY}</code> language
	 * 
	 * @author Stefan Alfons Krüger
	 */
	public static Translation createFromOneLIne(final String oneLineCoded) {
		Translation result = new Translation();
		result.fromOneLine(oneLineCoded);
		return result;
	}

	/**
	 * Returns the Translation to a String of the Format: "de{Baum}en{tree}" <br/>
	 * 
	 * 
	 * @author Stefan Alfons Krüger
	 */
	public static String toOneLine(Translation source) {
		StringBuffer oneLine = new StringBuffer();
		for (String key : source.keySet()) {
			oneLine.append(key + "{" + source.get(key) + "}");
		}
		return oneLine.toString();
	}

	/**
	 * @author Stefan Alfons Krüger
	 * @param code
	 * @return true if the code paramter is a valid ISO Language code
	 */
	public static boolean isValidISOLangCode(String code) {
		return getLanguageCodes().contains(code);
	}

	/**
	 * @return All language codes available in
	 *         java.util.Locale.getISOLanguages() without duplicates.
	 */
	public static Set<String> getLanguageCodes() {
		return languageCodes;
	}

	/**
	 * @param code
	 *            A two-letter language code.
	 * @return <code>null</code> or one (of many possible) {@link Locale} that
	 *         uses this language.
	 */
	public static Locale getLocaleFor(String code) {
		for (Locale l : Locale.getAvailableLocales()) {
			if (l.getLanguage().equals(code.toLowerCase())) {
				return l;
			}
			// LOGGER.debug(l.getLanguage() + " not = " + code);
		}

		LOGGER.error("Can't create a Locale for code " + code
				+ "! Returning the system default locale to avoid NPEs.");

		return Locale.getDefault();
	}

	/**
	 * A convenience method that checks if the {@link Translation} object
	 * contains a translation for the active language. A {@link String}
	 * containing only spaces will return <code>false</code>.
	 * 
	 * @param trans
	 *            {@link Translation} to check.
	 */
	public static boolean isEmpty(final Translation trans) {
		if (trans == null)
			return true;
		return isEmpty(trans.toString());
	}

	/**
	 * A convenience method that checks if the {@link String} returned by from a
	 * {@link Translation} object contains a "valid" translation for the active
	 * language. A {@link String} containing only spaces or equals
	 * {@link Translation}.NO_TRANSLATION will return <code>false</code>.
	 * 
	 * @param transString
	 *            {@link String} to check.
	 */
	public static boolean isEmpty(final String transString) {
		if (transString == null)
			return true;
		if (transString.trim().isEmpty())
			return true;
		if (transString.equals(Translation.NO_TRANSLATION))
			return true;
		return false;
	}

	/**
	 * @return a {@link Double} between 0 and 1 representing the part of the
	 *         given {@link Translation} that has been filled.
	 * @param ac
	 *            {@link AtlasConfig} to determine the languages to expect.
	 * @param trans
	 *            The {@link Translation} to check.
	 */
	public static double qmTranslation(final List<String> languages,
			final Translation trans) {
	
		if (trans == null)
			return 0.;
	
		Integer cunt = 0;
		for (final String l : languages) {
			final String t = trans.get(l);
			if (!isEmpty(t))
				cunt++;
		}
		return cunt.doubleValue() / (double) languages.size();
	}
}
