/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package skrueger.swing;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.HashSet;
import java.util.List;
import java.util.Random;
import java.util.Set;
import java.util.WeakHashMap;

import javax.swing.BorderFactory;
import javax.swing.JLabel;
import javax.swing.JTextField;
import javax.swing.SpringLayout;
import javax.swing.SwingConstants;

import org.apache.log4j.Logger;

import schmitzm.swing.JPanel;
import schmitzm.swing.SpringUtilities;
import schmitzm.swing.SwingUtil;
import skrueger.i8n.Translation;

/**
 * A {@link JPanel} that asks the user for the translations of a String in
 * several languages. Use {@link TranslationAskJDialog} to display.<br/>
 * The class does not implement any backup/clong strategies. The
 * {@link Translation} object is manipulated directly.<br/>
 * {@link TranslationEditJPanel}s. {@link TranslationAskJDialog} implements a
 * transparent Apply/Cancel logic.
 * 
 * @author Stefan Alfons Krüger
 */
public class TranslationEditJPanel extends JPanel {
	static final protected Logger LOGGER = Logger
			.getLogger(TranslationEditJPanel.class);

	private final List<String> languages;
	private JPanel translationGrid;
	private Translation trans;

	/**
	 * Remembers all {@link JTextField} that have been created.
	 */
	 private Set<TranslationJTextField> langTextFields = new HashSet<TranslationJTextField>();
	private WeakHashMap<ActionListener, ActionListener> actionListeners = new WeakHashMap<ActionListener, ActionListener>();

	/**
	 * Creates a {@link JPanel} that asks the user for the translation of a
	 * String in several languages
	 */
	public TranslationEditJPanel(Translation trans, List<String> languages_) {
		this(null, trans, languages_);
	}

	/**
	 * Creates a {@link JPanel} that asks the user for the translation of a
	 * String in several languages and additionally puts a {@link JLabel} with a
	 * question at the {@link JPanel}'s first row.
	 */
	public TranslationEditJPanel(String question, Translation trans,
			List<String> languages_) {
		super(new BorderLayout());
		
		SwingUtil.setMinimumWidth(this, 400);

		if (trans == null)
			trans = new Translation();

		this.trans = trans;
		this.languages = languages_;

		add(getTranslationGrid(), BorderLayout.CENTER);

		if (question != null) {
			JLabel questionLable = new JLabel(question);
			questionLable
					.setBorder(BorderFactory.createEmptyBorder(6, 6, 6, 6));
			add(questionLable, BorderLayout.NORTH);
		}

		/**
		 * Add listeners
		 */
	}

	/**
	 * Creates a {@link JPanel} that asks the user for the translation of a
	 * String in several languages and additionally puts a {@link JLabel} with a
	 * question at the {@link JPanel}'s first row.
	 * 
	 * <br/>
	 * This constructor also sets a TitledBorder with the given title.
	 */
	public TranslationEditJPanel(Translation title, List<String> languages2,
			String borderTitle) {
		this(title, languages2);

		setBorder(BorderFactory.createTitledBorder(borderTitle));
	}

	private JPanel getTranslationGrid() {
		if (translationGrid == null) {
			translationGrid = new JPanel(new SpringLayout());

			for (String langId : languages) {

				// language code : entry field for translation
				JLabel langDesc = new JLabel(langId.toUpperCase() + " :"); // i8n
				langDesc.setHorizontalAlignment(SwingConstants.RIGHT);
				langDesc.setVerticalAlignment(SwingConstants.NORTH);

				TranslationJTextField langTextField = new TranslationJTextField(
						trans, langId);
				// Setting a size
				langTextField.setPreferredSize(new Dimension(360, 22));
				langDesc.setLabelFor(langTextField);
				translationGrid.add(langDesc);
				translationGrid.add(langTextField);

				 langTextFields.add(langTextField);
			}

			// Lay out the panel.
			SpringUtilities.makeCompactGrid(translationGrid, languages.size(), // rows,
					2, // cols
					6, 6, // initX, initY
					6, 6); // xPad, yPad

		}
		return translationGrid;
	}

	/**
	 * @return The {@link Translation} that this {@link TranslationEditJPanel}
	 *         deals with.
	 */
	public Translation getTranslation() {
		return trans;
	}

	/**
	 * TODO Is never called?!
	 */
	public void dispose() {
		for (TranslationJTextField f: langTextFields) {
			f.dispose();
		}
	}

	public void addTranslationChangeListener(final ActionListener al) {
		final ActionListener actionListener = new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				al.actionPerformed(new ActionEvent(TranslationEditJPanel.this,
						new Random().nextInt(), ""));
			}

		};
		actionListeners.put(al, actionListener);
		getTranslation().addTranslationChangeListener(actionListener);
	}

	public void removeTranslationChangeListener(ActionListener al) {
		if (actionListeners.get(al) != null) {
			ActionListener actionListener = actionListeners.get(al);
			getTranslation().removeTranslationChangeListener(actionListener);
			actionListeners.remove(actionListener);
		}
	}

}
