/*******************************************************************************
 * Copyright (c) 2009 Martin O. J. Schmitz.
 * 
 * This file is part of the SCHMITZM library - a collection of utility 
 * classes based on Java 1.6, focusing (not only) on Java Swing 
 * and the Geotools library.
 * 
 * The SCHMITZM project is hosted at:
 * http://wald.intevation.org/projects/schmitzm/
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License (license.txt)
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * or try this link: http://www.gnu.org/licenses/lgpl.html
 * 
 * Contributors:
 *     Martin O. J. Schmitz - initial API and implementation
 *     Stefan A. Krüger - additional utility classes
 ******************************************************************************/
package skrueger.swing;

import java.awt.Color;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;

import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;

import org.apache.log4j.Logger;

import skrueger.i8n.Translation;

/**
 * A {@link JTextArea} that signals red when it is not filled by a value. A
 * listener is automatically registered with the {@link Translation} to update
 * on {@link Translation} changes.
 * 
 * @author Stefan Alfons Krüger
 * 
 */
public class TranslationJTextField extends JTextField {
	static final protected Logger LOGGER = Logger
			.getLogger(TranslationJTextField.class);
	int wasValid = 99; // 0 = false, 1 = true, 99 = undefined
	private String langCode;
	private Translation trans;
	/**
	 * Listens to changes in the {@link Translation} and update the
	 * {@link JTextField}
	 */
	private final ActionListener translationChangeActionListener;

	/**
	 * This RED indicates a missing translation
	 * 
	 * TODO to properties ;-)
	 */
	final static public Color emptyColor = new Color(240, 190, 190);

	/**
	 * Creates a new {@link TranslationJTextField}
	 * 
	 * @param trans
	 * @param langCode
	 */
	public TranslationJTextField(final Translation trans, final String langCode) {
		super(trans.get(langCode));
		this.trans = trans;
		this.langCode = langCode;

		/**
		 * SK: Change 26.Mai Use the default for an empty field
		 **/
		if (trans.get(langCode) == null) {
			String defaultTrans = trans.get(Translation.DEFAULT_KEY);
			trans.put(langCode, defaultTrans);
			setText(defaultTrans);
		}

		checkValid();

		// This Listener colors the JTextfiel red if it is empty
		getDocument().addDocumentListener(new DocumentListener() {

			// This method is called after an insert into the document
			public void insertUpdate(DocumentEvent evt) {
				checkValid();
			}

			// This method is called after a removal from the document
			public void removeUpdate(DocumentEvent evt) {
				checkValid();
			}

			// This method is called after one or more attributes have changed.
			// This method is not called when characters are inserted with
			// attributes.
			public void changedUpdate(DocumentEvent evt) {
				checkValid();
			}
		});

		/**
		 * Add a Listener to the Translation, so the JTextField can update on
		 * Translation changes.
		 */
		translationChangeActionListener = new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
//				LOGGER.debug("ActionListener called ");

				if (e.getSource() == trans) 
//						&& langCode.equals(e.getActionCommand())) 
						{
//					LOGGER.debug(" and omittet!\n");
					return;
				}
				LOGGER.debug(" and performed!\n");
				String newString = trans.get(langCode);
				if (newString == null)
					newString = "";
				if ((newString != null && !newString.equals(getText()))) {
					LOGGER.debug("Setting text to " + newString
							+ " becuse oldString was " + getText());
					setText(newString);
				}
			}

		};

		trans.addTranslationChangeListener(translationChangeActionListener);
	}

	/**
	 * TODO Is never called. Should be called!
	 */
	public void dispose() {
		boolean b = trans
				.removeTranslationChangeListener(translationChangeActionListener);
		LOGGER.debug("removing translationChangeActionListener = " + b);
	}

	/**
	 * If the getText().equals("") sets background color to red. The method also
	 * stored the changed string into the {@link Translation} object.
	 */
	protected void checkValid() {

		String trimmedText = getText().trim();

		if (trimmedText.equals("")) {
			// Not valid
			if (wasValid != 0) {
				setBackground(emptyColor);
				repaint();
				wasValid = 0;
			}

		} else {
			// valid
			if (wasValid != 1) {
				setBackground(Color.white);
				repaint();
				wasValid = 1;
			}
			// Directly store the change in the Translation Object
		}

		// LOGGER.debug("putting '"+trimmedText+"' into the translation");
		trans.put(langCode, trimmedText);
	}
}
